/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "include_dirs": [
            "/private/var/folders/24/8k48jl6d249_n_qfxwsl6xvm0000gn/T/pip-build-env-ykxdc412/overlay/lib/python3.9/site-packages/numpy/core/include"
        ],
        "name": "ms_deisotope._c.peak_dependency_network.intervals",
        "sources": [
            "src/ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        ]
    },
    "module_name": "ms_deisotope._c.peak_dependency_network.intervals"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_deisotope___c__peak_dependency_network__intervals
#define __PYX_HAVE_API__ms_deisotope___c__peak_dependency_network__intervals
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "pythread.h"
#include <stdlib.h>
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src/ms_deisotope/_c/peak_dependency_network/intervals.pyx",
  "src/ms_deisotope/_c/peak_dependency_network/intervals.pxd",
  "stringsource",
  "type.pxd",
};
/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_incr_aligned(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && CYTHON_COMPILING_IN_NOGIL
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #pragma intrinsic (_InterlockedExchangeAdd)
    #define __pyx_atomic_incr_aligned(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_aligned(value) _InterlockedExchangeAdd(value, -1)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview))
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview))
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/*--- Type declarations ---*/
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D;
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":1
 * cdef class SpanningMixin(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double start
 */

#line 1 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin {

#line 1 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject_HEAD

#line 1 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_vtab;

#line 1 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double start;

#line 1 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double end;

#line 1 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
};

#line 1 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":19
 * 
 * 
 * cdef class Interval(SpanningMixin):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object members
 */

#line 19 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval {

#line 19 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;

#line 19 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *members;

#line 19 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *data;

#line 19 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
};

#line 19 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":24
 *         public dict data
 * 
 * cdef class SimpleInterval(SpanningMixin):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef SimpleInterval _create(double start, double end)
 */

#line 24 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval {

#line 24 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;

#line 24 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
};

#line 24 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":28
 *     cdef SimpleInterval _create(double start, double end)
 * 
 * cdef class IntervalTreeNode(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double center
 */

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode {

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject_HEAD

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_vtab;

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double center;

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double start;

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double end;

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *left;

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *right;

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *contained;

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int level;

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *parent;

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
};

#line 28 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":45
 * 
 * 
 * cdef class SpanningMixin2D(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double[2] _start
 */

#line 45 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D {

#line 45 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject_HEAD

#line 45 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_vtab;

#line 45 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double _start[2];

#line 45 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double _end[2];

#line 45 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
};

#line 45 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":56
 * 
 * 
 * cdef class Interval2D(SpanningMixin2D):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list members
 */

#line 56 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D {

#line 56 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D __pyx_base;

#line 56 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *members;

#line 56 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *data;

#line 56 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
};

#line 56 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":62
 * 
 * 
 * cdef class IntervalTreeNode2D(IntervalTreeNode):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object inner_organizer
 */

#line 62 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D {

#line 62 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode __pyx_base;

#line 62 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *inner_organizer;

#line 62 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *organized;

#line 62 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
};

#line 62 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


/* "View.MemoryView":106
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

#line 106 "View.MemoryView"
struct __pyx_array_obj {

#line 106 "View.MemoryView"
  PyObject_HEAD

#line 106 "View.MemoryView"
  struct __pyx_vtabstruct_array *__pyx_vtab;

#line 106 "View.MemoryView"
  char *data;

#line 106 "View.MemoryView"
  Py_ssize_t len;

#line 106 "View.MemoryView"
  char *format;

#line 106 "View.MemoryView"
  int ndim;

#line 106 "View.MemoryView"
  Py_ssize_t *_shape;

#line 106 "View.MemoryView"
  Py_ssize_t *_strides;

#line 106 "View.MemoryView"
  Py_ssize_t itemsize;

#line 106 "View.MemoryView"
  PyObject *mode;

#line 106 "View.MemoryView"
  PyObject *_format;

#line 106 "View.MemoryView"
  void (*callback_free_data)(void *);

#line 106 "View.MemoryView"
  int free_data;

#line 106 "View.MemoryView"
  int dtype_is_object;

#line 106 "View.MemoryView"
};

#line 106 "View.MemoryView"


/* "View.MemoryView":280
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */

#line 280 "View.MemoryView"
struct __pyx_MemviewEnum_obj {

#line 280 "View.MemoryView"
  PyObject_HEAD

#line 280 "View.MemoryView"
  PyObject *name;

#line 280 "View.MemoryView"
};

#line 280 "View.MemoryView"


/* "View.MemoryView":331
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

#line 331 "View.MemoryView"
struct __pyx_memoryview_obj {

#line 331 "View.MemoryView"
  PyObject_HEAD

#line 331 "View.MemoryView"
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;

#line 331 "View.MemoryView"
  PyObject *obj;

#line 331 "View.MemoryView"
  PyObject *_size;

#line 331 "View.MemoryView"
  PyObject *_array_interface;

#line 331 "View.MemoryView"
  PyThread_type_lock lock;

#line 331 "View.MemoryView"
  __pyx_atomic_int acquisition_count[2];

#line 331 "View.MemoryView"
  __pyx_atomic_int *acquisition_count_aligned_p;

#line 331 "View.MemoryView"
  Py_buffer view;

#line 331 "View.MemoryView"
  int flags;

#line 331 "View.MemoryView"
  int dtype_is_object;

#line 331 "View.MemoryView"
  __Pyx_TypeInfo *typeinfo;

#line 331 "View.MemoryView"
};

#line 331 "View.MemoryView"


/* "View.MemoryView":967
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

#line 967 "View.MemoryView"
struct __pyx_memoryviewslice_obj {

#line 967 "View.MemoryView"
  struct __pyx_memoryview_obj __pyx_base;

#line 967 "View.MemoryView"
  __Pyx_memviewslice from_slice;

#line 967 "View.MemoryView"
  PyObject *from_object;

#line 967 "View.MemoryView"
  PyObject *(*to_object_func)(char *);

#line 967 "View.MemoryView"
  int (*to_dtype_func)(char *, PyObject *);

#line 967 "View.MemoryView"
};

#line 967 "View.MemoryView"


#line 967 "View.MemoryView"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":7
 * 
 * 
 * cdef class SpanningMixin(object):             # <<<<<<<<<<<<<<
 *     """Provides methods for checking whether an entity
 *     which has a defined start and end point over a single
 */

#line 7 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 7 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin {

#line 7 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int (*_contains)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, double);

#line 7 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int (*contains)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, double, int __pyx_skip_dispatch);

#line 7 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int (*overlaps)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);

#line 7 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double (*overlap_size)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);

#line 7 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int (*is_contained_in_interval)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);

#line 7 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int (*contains_interval)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch);

#line 7 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
};

#line 7 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;

#line 7 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":136
 * 
 * 
 * cdef class Interval(SpanningMixin):             # <<<<<<<<<<<<<<
 *     def __init__(self, start, end, members=None, **kwargs):
 *         if members is None:
 */

#line 136 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 136 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval {

#line 136 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;

#line 136 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
};

#line 136 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;

#line 136 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":120
 * 
 * 
 * cdef class SimpleInterval(SpanningMixin):             # <<<<<<<<<<<<<<
 *     def __init__(self, start, end):
 *         self.start = start
 */

#line 120 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 120 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval {

#line 120 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_base;

#line 120 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *(*_create)(double, double);

#line 120 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
};

#line 120 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval;

#line 120 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":174
 * 
 * 
 * cdef class IntervalTreeNode(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, double center, IntervalTreeNode left, list contained, IntervalTreeNode right,
 */

#line 174 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 174 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode {

#line 174 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *(*contains_point)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, int __pyx_skip_dispatch);

#line 174 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *(*_overlaps_interval)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, double, int __pyx_skip_dispatch);

#line 174 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *(*overlaps)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, double, int __pyx_skip_dispatch);

#line 174 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, int __pyx_skip_dispatch);

#line 174 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
};

#line 174 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;

#line 174 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":472
 * 
 * 
 * cdef class SpanningMixin2D(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, start, end):
 */

#line 472 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 472 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D {

#line 472 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int (*_contains)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, double *);

#line 472 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int (*contains)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, __Pyx_memviewslice, int __pyx_skip_dispatch);

#line 472 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int (*overlaps)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, int __pyx_skip_dispatch);

#line 472 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int (*contains_interval)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, int __pyx_skip_dispatch);

#line 472 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
};

#line 472 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D;

#line 472 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":595
 * 
 * 
 * cdef class Interval2D(SpanningMixin2D):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, start, end, members=None, **kwargs):
 */

#line 595 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 595 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D {

#line 595 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D __pyx_base;

#line 595 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
};

#line 595 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D;

#line 595 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":633
 * 
 * 
 * cdef class IntervalTreeNode2D(IntervalTreeNode):             # <<<<<<<<<<<<<<
 *     def __init__(self, center, left, contained, right, level=0, parent=None, inner_organizer=None):
 *         super().__init__(center, left, contained, right, level=level, parent=parent)
 */

#line 633 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 633 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D {

#line 633 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode __pyx_base;

#line 633 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *(*_overlaps_interval_2d)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *, __Pyx_memviewslice, __Pyx_memviewslice, int __pyx_skip_dispatch);

#line 633 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *(*overlaps_2d)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *, __Pyx_memviewslice, __Pyx_memviewslice, int __pyx_skip_dispatch);

#line 633 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
};

#line 633 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D;

#line 633 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


/* "View.MemoryView":106
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

#line 106 "View.MemoryView"


#line 106 "View.MemoryView"
struct __pyx_vtabstruct_array {

#line 106 "View.MemoryView"
  PyObject *(*get_memview)(struct __pyx_array_obj *);

#line 106 "View.MemoryView"
};

#line 106 "View.MemoryView"
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;

#line 106 "View.MemoryView"


/* "View.MemoryView":331
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

#line 331 "View.MemoryView"


#line 331 "View.MemoryView"
struct __pyx_vtabstruct_memoryview {

#line 331 "View.MemoryView"
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);

#line 331 "View.MemoryView"
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);

#line 331 "View.MemoryView"
};

#line 331 "View.MemoryView"
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;

#line 331 "View.MemoryView"


/* "View.MemoryView":967
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

#line 967 "View.MemoryView"


#line 967 "View.MemoryView"
struct __pyx_vtabstruct__memoryviewslice {

#line 967 "View.MemoryView"
  struct __pyx_vtabstruct_memoryview __pyx_base;

#line 967 "View.MemoryView"
};

#line 967 "View.MemoryView"
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

/* ModInt[long].proto */
static CYTHON_INLINE long __Pyx_mod_long(long, long);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* DivInt[Py_ssize_t].proto */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t, Py_ssize_t);

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* PyObject_Unicode.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))
#else
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Unicode(obj))
#endif

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?\
 PyFloat_AS_DOUBLE(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* ClassMethod.proto */
#include "descrobject.h"
static CYTHON_UNUSED PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_double(PyObject *, int writable_flag);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_double(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_double(const char *itemp, PyObject *obj);

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* TypeInfoToFormat.proto */
struct __pyx_typeinfo_string {
    char string[3];
};
static struct __pyx_typeinfo_string __Pyx_TypeInfoToFormat(__Pyx_TypeInfo *type);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE short __Pyx_PyInt_As_short(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin__contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlap_size(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_is_contained_in_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval__create(double __pyx_v_start, double __pyx_v_end); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contains_point(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__overlaps_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__eq(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D__contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, double *__pyx_v_i); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_i, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_interval, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_interval, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D__overlaps_interval_2d(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_starts, __Pyx_memviewslice __pyx_v_ends, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_overlaps_2d(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_starts, __Pyx_memviewslice __pyx_v_ends, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'ms_deisotope._c.peak_dependency_network.intervals' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D = 0;
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static double __pyx_v_12ms_deisotope_2_c_23peak_dependency_network_9intervals_INF;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(PyObject *, double, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point_2d(PyObject *, double, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SimpleInterval__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin2D__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval2D__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode2D__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__Pyx_carray_to_py_double(double *, Py_ssize_t); /*proto*/
static CYTHON_INLINE PyObject *__Pyx_carray_to_tuple_double(double *, Py_ssize_t); /*proto*/
static int __Pyx_carray_from_py_double(PyObject *, double *, Py_ssize_t); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
static PyObject *__pyx_format_from_typeinfo(__Pyx_TypeInfo *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "ms_deisotope._c.peak_dependency_network.intervals"
extern int __pyx_module_is_main_ms_deisotope___c__peak_dependency_network__intervals;
int __pyx_module_is_main_ms_deisotope___c__peak_dependency_network__intervals = 0;

/* Implementation of 'ms_deisotope._c.peak_dependency_network.intervals' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_OverflowError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static const char __pyx_k_O[] = "O";
static const char __pyx_k_T[] = "T{";
  static const char __pyx_k_c[] = "c";
  static const char __pyx_k_s[] = "(%s)";
  static const char __pyx_k_x[] = "x";
  static const char __pyx_k_eq[] = "_eq";
  static const char __pyx_k_id[] = "id";
  static const char __pyx_k__26[] = "^";
  static const char __pyx_k__27[] = "";
  static const char __pyx_k__28[] = ":";
static const char __pyx_k__29[] = "}";
static const char __pyx_k__30[] = ",";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_ends[] = "ends";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_left[] = "left";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_build[] = "build";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_deque[] = "deque";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_level[] = "level";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_right[] = "right";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_center[] = "center";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_insert[] = "insert";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_parent[] = "parent";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_starts[] = "starts";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_flatten[] = "flatten";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_members[] = "members";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_Interval[] = "Interval";
static const char __pyx_k_contains[] = "contains";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_organize[] = "organize";
static const char __pyx_k_overlaps[] = "overlaps";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_contained[] = "contained";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_intervals[] = "intervals";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_Interval2D[] = "Interval2D";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_overlaps_2d[] = "overlaps_2d";
static const char __pyx_k_overlap_size[] = "overlap_size";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_OverflowError[] = "OverflowError";
static const char __pyx_k_SpanningMixin[] = "SpanningMixin";
static const char __pyx_k_node_contains[] = "node_contains";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_set_organizer[] = "set_organizer";
static const char __pyx_k_update_bounds[] = "_update_bounds";
static const char __pyx_k_SimpleInterval[] = "SimpleInterval";
static const char __pyx_k_contains_point[] = "contains_point";
static const char __pyx_k_SpanningMixin2D[] = "SpanningMixin2D";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_inner_organizer[] = "inner_organizer";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_IntervalTreeNode[] = "IntervalTreeNode";
static const char __pyx_k_contains_interval[] = "contains_interval";
static const char __pyx_k_overlaps_interval[] = "_overlaps_interval";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_IntervalTreeNode2D[] = "IntervalTreeNode2D";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_organizer_callback[] = "organizer_callback";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_overlaps_interval_2d[] = "_overlaps_interval_2d";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_pyx_unpickle_Interval[] = "__pyx_unpickle_Interval";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_s_start_r_end_r_data_r[] = "%s(start=%r, end=%r, data=%r)";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_pyx_unpickle_Interval2D[] = "__pyx_unpickle_Interval2D";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_is_contained_in_interval[] = "is_contained_in_interval";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_pyx_unpickle_SpanningMixin[] = "__pyx_unpickle_SpanningMixin";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_pyx_unpickle_SimpleInterval[] = "__pyx_unpickle_SimpleInterval";
static const char __pyx_k_pyx_unpickle_SpanningMixin2D[] = "__pyx_unpickle_SpanningMixin2D";
static const char __pyx_k_pyx_unpickle_IntervalTreeNode[] = "__pyx_unpickle_IntervalTreeNode";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_pyx_unpickle_IntervalTreeNode2[] = "__pyx_unpickle_IntervalTreeNode2D";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_IntervalTreeNode_level_d_center[] = "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)";
static const char __pyx_k_ms_deisotope__c_peak_dependency[] = "ms_deisotope._c.peak_dependency_network.intervals";
static const char __pyx_k_self___class_____name___self_st[] = "{self.__class__.__name__}({self.start}, {self.end})";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0xeace574, 0x24a0b9b, 0x4c97279) = (data, end, members, start))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0x262a9ad, 0x0629e3a, 0xfab189d) = (center, contained, end, left, level, parent, right, start))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_4[] = "Incompatible checksums (0x%x vs (0xfeaf07a, 0x2a19646, 0xea81723) = (_end, _start))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_5[] = "Incompatible checksums (0x%x vs (0x688f3f3, 0xc0bd0b9, 0x6f637d9) = (_end, _start, data, members))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_6[] = "Incompatible checksums (0x%x vs (0x0745b49, 0xc11802f, 0xc16f891) = (center, contained, end, inner_organizer, left, level, organized, parent, right, start))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_7[] = "Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))";
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_n_s_Interval;
static PyObject *__pyx_n_s_Interval2D;
static PyObject *__pyx_n_s_IntervalTreeNode;
static PyObject *__pyx_n_s_IntervalTreeNode2D;
static PyObject *__pyx_kp_s_IntervalTreeNode_level_d_center;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_OverflowError;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_SimpleInterval;
static PyObject *__pyx_n_s_SpanningMixin;
static PyObject *__pyx_n_s_SpanningMixin2D;
static PyObject *__pyx_kp_b_T;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_kp_b__26;
static PyObject *__pyx_kp_b__27;
static PyObject *__pyx_kp_b__28;
static PyObject *__pyx_kp_b__29;
static PyObject *__pyx_kp_u__30;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_build;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_center;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_contained;
static PyObject *__pyx_n_s_contains;
static PyObject *__pyx_n_s_contains_interval;
static PyObject *__pyx_n_s_contains_point;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_deque;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_ends;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_eq;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_flatten;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_inner_organizer;
static PyObject *__pyx_n_s_insert;
static PyObject *__pyx_n_s_intervals;
static PyObject *__pyx_n_s_is_contained_in_interval;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_left;
static PyObject *__pyx_n_s_level;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_members;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_ms_deisotope__c_peak_dependency;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_node_contains;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_organize;
static PyObject *__pyx_n_s_organizer_callback;
static PyObject *__pyx_n_s_overlap_size;
static PyObject *__pyx_n_s_overlaps;
static PyObject *__pyx_n_s_overlaps_2d;
static PyObject *__pyx_n_s_overlaps_interval;
static PyObject *__pyx_n_s_overlaps_interval_2d;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_parent;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_unpickle_Interval;
static PyObject *__pyx_n_s_pyx_unpickle_Interval2D;
static PyObject *__pyx_n_s_pyx_unpickle_IntervalTreeNode;
static PyObject *__pyx_n_s_pyx_unpickle_IntervalTreeNode2;
static PyObject *__pyx_n_s_pyx_unpickle_SimpleInterval;
static PyObject *__pyx_n_s_pyx_unpickle_SpanningMixin;
static PyObject *__pyx_n_s_pyx_unpickle_SpanningMixin2D;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_right;
static PyObject *__pyx_kp_u_s;
static PyObject *__pyx_kp_s_s_start_r_end_r_data_r;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self___class_____name___self_st;
static PyObject *__pyx_n_s_set_organizer;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_starts;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_update_bounds;
static PyObject *__pyx_n_s_x;
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_4contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_6overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_8overlap_size(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_10is_contained_in_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_12contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_14__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_16__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_2__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_4__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_6__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_members, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_2__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4__getitem__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_k); /* proto */
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_6__len__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_8__iter__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_10__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_12__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_intervals, double __pyx_v_x); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_center, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_left, PyObject *__pyx_v_contained, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_right, int __pyx_v_level, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_parent); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_2contains_point(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4_overlaps_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_8__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_10__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_12_eq(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_other); /* proto */
static Py_hash_t __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_14__hash__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_16build(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_intervals); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_18_update_bounds(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_20insert(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_22flatten(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_24balance(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_26__iter__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_28__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_30__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_start); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_end); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_2__contains__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_8contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_10__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_12__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_members, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_2__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4__getitem__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self, PyObject *__pyx_v_k); /* proto */
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_6__len__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_8__iter__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_10__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_12__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_2intervals_containing_point_2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_intervals, double __pyx_v_x, int __pyx_v_level); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, PyObject *__pyx_v_center, PyObject *__pyx_v_left, PyObject *__pyx_v_contained, PyObject *__pyx_v_right, PyObject *__pyx_v_level, PyObject *__pyx_v_parent, PyObject *__pyx_v_inner_organizer); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_2set_organizer(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, PyObject *__pyx_v_organizer); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_4organize(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_6_overlaps_interval_2d(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_starts, __Pyx_memviewslice __pyx_v_ends); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_8overlaps_2d(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_starts, __Pyx_memviewslice __pyx_v_ends); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_10build(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_intervals, PyObject *__pyx_v_organizer_callback); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_12__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_14__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_4__pyx_unpickle_SpanningMixin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_6__pyx_unpickle_SimpleInterval(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8__pyx_unpickle_Interval(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10__pyx_unpickle_IntervalTreeNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_12__pyx_unpickle_SpanningMixin2D(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14__pyx_unpickle_Interval2D(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16__pyx_unpickle_IntervalTreeNode2D(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_5726981;
static PyObject *__pyx_int_6463034;
static PyObject *__pyx_int_7625545;
static PyObject *__pyx_int_38407067;
static PyObject *__pyx_int_40020397;
static PyObject *__pyx_int_44144198;
static PyObject *__pyx_int_80310905;
static PyObject *__pyx_int_108041459;
static PyObject *__pyx_int_109638643;
static PyObject *__pyx_int_112105877;
static PyObject *__pyx_int_116799449;
static PyObject *__pyx_int_136983863;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_202100921;
static PyObject *__pyx_int_202473519;
static PyObject *__pyx_int_202832017;
static PyObject *__pyx_int_214708752;
static PyObject *__pyx_int_245896995;
static PyObject *__pyx_int_246211956;
static PyObject *__pyx_int_262871197;
static PyObject *__pyx_int_267055226;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__21;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__51;
/* Late includes */

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":14
 *     """
 * 
 *     def __init__(self, start, end):             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_start = 0
#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_end = 0
#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,0};

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[2] = {0,0};

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  2: 
#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else 
#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); 
#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 14, __pyx_L3_error)

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 14, __pyx_L3_error)

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_argtuple_error;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_start = values[0];

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_end = values[1];

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 14, __pyx_L3_error)

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return -1;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin___init__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), __pyx_v_start, __pyx_v_end);

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end) {

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_1;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":15
 * 
 *     def __init__(self, start, end):
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 * 
 */

#line 15 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_start); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 15, __pyx_L1_error)

#line 15 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->start = __pyx_t_1;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":16
 *     def __init__(self, start, end):
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, i):
 */

#line 16 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_end); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L1_error)

#line 16 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->end = __pyx_t_1;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":14
 *     """
 * 
 *     def __init__(self, start, end):             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 14 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":18
 *         self.end = end
 * 
 *     def __contains__(self, i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__[] = "Tests for point inclusion, `start <= i <= end`\n\n        Parameters\n        ----------\n        i : Number\n            The point to be tested\n\n        Returns\n        -------\n        bool\n        ";

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
#if CYTHON_UPDATE_DESCRIPTOR_DOC

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
struct wrapperbase __pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
#endif

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = 
#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((PyObject *)__pyx_v_i));

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_i) {

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_1;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":30
 *         bool
 *         """
 *         return self._contains(i)             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _contains(self, double i):
 */

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_i); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 30, __pyx_L1_error)

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self->__pyx_vtab)->_contains(__pyx_v_self, __pyx_t_1);

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":18
 *         self.end = end
 * 
 *     def __contains__(self, i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 18 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":32
 *         return self._contains(i)
 * 
 *     cdef bint _contains(self, double i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin__contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i) {

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_1;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_contains", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":44
 *         bool
 *         """
 *         return self.start <= i <= self.end             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint contains(self, double i):
 */

#line 44 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_v_self->start <= __pyx_v_i);

#line 44 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_1) {

#line 44 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = (__pyx_v_i <= __pyx_v_self->end);

#line 44 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 44 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 44 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":32
 *         return self._contains(i)
 * 
 *     cdef bint _contains(self, double i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":46
 *         return self.start <= i <= self.end
 * 
 *     cpdef bint contains(self, double i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5contains(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i, int __pyx_skip_dispatch) {

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains", 0);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_contains); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5contains)) {

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 46, __pyx_L1_error)

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_5 = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_5)) {

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 46, __pyx_L1_error)

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = __pyx_t_6;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":58
 *         bool
 *         """
 *         return self._contains(i)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint overlaps(self, SpanningMixin interval):
 */

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self->__pyx_vtab)->_contains(__pyx_v_self, __pyx_v_i);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":46
 *         return self.start <= i <= self.end
 * 
 *     cpdef bint contains(self, double i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.contains", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5contains(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_4contains[] = "SpanningMixin.contains(self, double i) -> bool\nTests for point inclusion, `start <= i <= end`\n\n        Parameters\n        ----------\n        i : double\n            The point to be tested\n\n        Returns\n        -------\n        bool\n        ";

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5contains(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  double __pyx_v_i
#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains (wrapper)", 0);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  assert(__pyx_arg_i); {

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_i = __pyx_PyFloat_AsDouble(__pyx_arg_i); if (unlikely((__pyx_v_i == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 46, __pyx_L3_error)

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.contains", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_4contains(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((double)__pyx_v_i));

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_4contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, double __pyx_v_i) {

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains", 0);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains(__pyx_v_self, __pyx_v_i, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.contains", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":60
 *         return self._contains(i)
 * 
 *     cpdef bint overlaps(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether another spanning entity with
 *         a defined start and end point overlaps with
 */

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch) {
  int __pyx_v_cond
#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_5;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps", 0);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlaps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L1_error)

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_7overlaps)) {

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_4 = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_4)) {

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_interval));

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 60, __pyx_L1_error)

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 60, __pyx_L1_error)

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = __pyx_t_5;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":74
 *         """
 *         cdef bint cond
 *         cond = ((self.start <= interval.start and self.end >= interval.start) or (             # <<<<<<<<<<<<<<
 *             self.start >= interval.start and self.end <= interval.end) or (
 *             self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (
 */

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->start <= __pyx_v_interval->start) != 0);

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!__pyx_t_6) {

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L4_next_or;

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->start) != 0);

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!__pyx_t_6) {

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L3_bool_binop_done;

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 74 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_next_or:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":75
 *         cdef bint cond
 *         cond = ((self.start <= interval.start and self.end >= interval.start) or (
 *             self.start >= interval.start and self.end <= interval.end) or (             # <<<<<<<<<<<<<<
 *             self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (
 *             self.start <= interval.start and self.end >= interval.start) or (
 */

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->start >= __pyx_v_interval->start) != 0);

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!__pyx_t_6) {

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L6_next_or;

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->end <= __pyx_v_interval->end) != 0);

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!__pyx_t_6) {

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L3_bool_binop_done;

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 75 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L6_next_or:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":76
 *         cond = ((self.start <= interval.start and self.end >= interval.start) or (
 *             self.start >= interval.start and self.end <= interval.end) or (
 *             self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (             # <<<<<<<<<<<<<<
 *             self.start <= interval.start and self.end >= interval.start) or (
 *             self.start <= interval.end and self.end >= interval.end))
 */

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->start >= __pyx_v_interval->start) != 0);

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!__pyx_t_6) {

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L8_next_or;

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->end) != 0);

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!__pyx_t_6) {

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L8_next_or;

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->start <= __pyx_v_interval->end) != 0);

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!__pyx_t_6) {

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L3_bool_binop_done;

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 76 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L8_next_or:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":77
 *             self.start >= interval.start and self.end <= interval.end) or (
 *             self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (
 *             self.start <= interval.start and self.end >= interval.start) or (             # <<<<<<<<<<<<<<
 *             self.start <= interval.end and self.end >= interval.end))
 *         return cond
 */

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->start <= __pyx_v_interval->start) != 0);

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!__pyx_t_6) {

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L11_next_or;

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->start) != 0);

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!__pyx_t_6) {

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L3_bool_binop_done;

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 77 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L11_next_or:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":78
 *             self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (
 *             self.start <= interval.start and self.end >= interval.start) or (
 *             self.start <= interval.end and self.end >= interval.end))             # <<<<<<<<<<<<<<
 *         return cond
 * 
 */

#line 78 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->start <= __pyx_v_interval->end) != 0);

#line 78 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_6) {

#line 78 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 78 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 78 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L3_bool_binop_done;

#line 78 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 78 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->end) != 0);

#line 78 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = __pyx_t_6;

#line 78 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_bool_binop_done:;

#line 78 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_cond = __pyx_t_5;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":79
 *             self.start <= interval.start and self.end >= interval.start) or (
 *             self.start <= interval.end and self.end >= interval.end))
 *         return cond             # <<<<<<<<<<<<<<
 * 
 *     cpdef double overlap_size(self, SpanningMixin interval):
 */

#line 79 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_v_cond;

#line 79 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":60
 *         return self._contains(i)
 * 
 *     cpdef bint overlaps(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether another spanning entity with
 *         a defined start and end point overlaps with
 */

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_6overlaps[] = "SpanningMixin.overlaps(self, SpanningMixin interval) -> bool\nTests whether another spanning entity with\n        a defined start and end point overlaps with\n        this spanning entity\n\n        Parameters\n        ----------\n        interval : SpanningMixin\n\n        Returns\n        -------\n        bool\n        ";

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps (wrapper)", 0);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, 1, "interval", 0))) __PYX_ERR(0, 60, __pyx_L1_error)
  __pyx_r = 
#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_6overlaps(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_interval));

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_6overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval) {

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps", 0);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlaps(__pyx_v_self, __pyx_v_interval, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L1_error)

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 60 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":81
 *         return cond
 * 
 *     cpdef double overlap_size(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:
 *             return interval.end - interval.start
 */

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_9overlap_size(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static double __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlap_size(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch) {

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_r;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_5;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_8;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlap_size", 0);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlap_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_9overlap_size)) {

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_4 = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_4)) {

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_interval));

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = __pyx_t_5;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":82
 * 
 *     cpdef double overlap_size(self, SpanningMixin interval):
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:             # <<<<<<<<<<<<<<
 *             return interval.end - interval.start
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 */

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = (__pyx_v_self->start <= __pyx_v_interval->start);

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_7) {

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (__pyx_v_interval->start <= __pyx_v_self->end);

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_8 = (__pyx_t_7 != 0);

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_8) {

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = __pyx_t_8;

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L4_bool_binop_done;

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_8 = (__pyx_v_self->start <= __pyx_v_interval->end);

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_8) {

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_8 = (__pyx_v_interval->end <= __pyx_v_self->end);

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = (__pyx_t_8 != 0);

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = __pyx_t_7;

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_bool_binop_done:;

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":83
 *     cpdef double overlap_size(self, SpanningMixin interval):
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:
 *             return interval.end - interval.start             # <<<<<<<<<<<<<<
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 *             return self.end - self.start
 */

#line 83 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = (__pyx_v_interval->end - __pyx_v_interval->start);

#line 83 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":82
 * 
 *     cpdef double overlap_size(self, SpanningMixin interval):
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:             # <<<<<<<<<<<<<<
 *             return interval.end - interval.start
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 */

#line 82 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":84
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:
 *             return interval.end - interval.start
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:             # <<<<<<<<<<<<<<
 *             return self.end - self.start
 *         if self.end >= interval.start:
 */

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = (__pyx_v_interval->start <= __pyx_v_self->start);

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_7) {

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (__pyx_v_self->start <= __pyx_v_interval->end);

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_8 = (__pyx_t_7 != 0);

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_8) {

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = __pyx_t_8;

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L6_bool_binop_done;

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_8 = (__pyx_v_interval->start <= __pyx_v_self->end);

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_8) {

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_8 = (__pyx_v_self->end <= __pyx_v_interval->end);

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = (__pyx_t_8 != 0);

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = __pyx_t_7;

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L6_bool_binop_done:;

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":85
 *             return interval.end - interval.start
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 *             return self.end - self.start             # <<<<<<<<<<<<<<
 *         if self.end >= interval.start:
 *             return self.end - interval.start
 */

#line 85 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = (__pyx_v_self->end - __pyx_v_self->start);

#line 85 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":84
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:
 *             return interval.end - interval.start
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:             # <<<<<<<<<<<<<<
 *             return self.end - self.start
 *         if self.end >= interval.start:
 */

#line 84 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":86
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 *             return self.end - self.start
 *         if self.end >= interval.start:             # <<<<<<<<<<<<<<
 *             return self.end - interval.start
 *         elif self.start >= interval.end:
 */

#line 86 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->start) != 0);

#line 86 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":87
 *             return self.end - self.start
 *         if self.end >= interval.start:
 *             return self.end - interval.start             # <<<<<<<<<<<<<<
 *         elif self.start >= interval.end:
 *             return self.start - interval.end
 */

#line 87 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = (__pyx_v_self->end - __pyx_v_interval->start);

#line 87 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":86
 *         elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
 *             return self.end - self.start
 *         if self.end >= interval.start:             # <<<<<<<<<<<<<<
 *             return self.end - interval.start
 *         elif self.start >= interval.end:
 */

#line 86 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":88
 *         if self.end >= interval.start:
 *             return self.end - interval.start
 *         elif self.start >= interval.end:             # <<<<<<<<<<<<<<
 *             return self.start - interval.end
 * 
 */

#line 88 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->start >= __pyx_v_interval->end) != 0);

#line 88 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":89
 *             return self.end - interval.start
 *         elif self.start >= interval.end:
 *             return self.start - interval.end             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_contained_in_interval(self, SpanningMixin interval):
 */

#line 89 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = (__pyx_v_self->start - __pyx_v_interval->end);

#line 89 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":88
 *         if self.end >= interval.start:
 *             return self.end - interval.start
 *         elif self.start >= interval.end:             # <<<<<<<<<<<<<<
 *             return self.start - interval.end
 * 
 */

#line 88 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":81
 *         return cond
 * 
 *     cpdef double overlap_size(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:
 *             return interval.end - interval.start
 */

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.overlap_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_9overlap_size(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_8overlap_size[] = "SpanningMixin.overlap_size(self, SpanningMixin interval) -> double";

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_9overlap_size(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlap_size (wrapper)", 0);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, 1, "interval", 0))) __PYX_ERR(0, 81, __pyx_L1_error)
  __pyx_r = 
#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_8overlap_size(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_interval));

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_8overlap_size(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval) {

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlap_size", 0);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlap_size(__pyx_v_self, __pyx_v_interval, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.overlap_size", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 81 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":91
 *             return self.start - interval.end
 * 
 *     cpdef bint is_contained_in_interval(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether this spanning entity is
 *         completely contained inside the other entity
 */

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_11is_contained_in_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_is_contained_in_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch) {

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_5;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("is_contained_in_interval", 0);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_contained_in_interval); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_11is_contained_in_interval)) {

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_4 = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_4)) {

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_interval));

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 91, __pyx_L1_error)

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = __pyx_t_5;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":103
 *         bool
 *         """
 *         return self.start >= interval.start and self.end <= interval.end             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint contains_interval(self, SpanningMixin interval):
 */

#line 103 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->start >= __pyx_v_interval->start) != 0);

#line 103 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_6) {

#line 103 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 103 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 103 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L3_bool_binop_done;

#line 103 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 103 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->end <= __pyx_v_interval->end) != 0);

#line 103 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = __pyx_t_6;

#line 103 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_bool_binop_done:;

#line 103 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_5;

#line 103 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":91
 *             return self.start - interval.end
 * 
 *     cpdef bint is_contained_in_interval(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether this spanning entity is
 *         completely contained inside the other entity
 */

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.is_contained_in_interval", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_11is_contained_in_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_10is_contained_in_interval[] = "SpanningMixin.is_contained_in_interval(self, SpanningMixin interval) -> bool\nTests whether this spanning entity is\n        completely contained inside the other entity\n\n        Parameters\n        ----------\n        interval : SpanningMixin\n\n        Returns\n        -------\n        bool\n        ";

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_11is_contained_in_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("is_contained_in_interval (wrapper)", 0);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, 1, "interval", 0))) __PYX_ERR(0, 91, __pyx_L1_error)
  __pyx_r = 
#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_10is_contained_in_interval(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_interval));

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_10is_contained_in_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval) {

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("is_contained_in_interval", 0);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_is_contained_in_interval(__pyx_v_self, __pyx_v_interval, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.is_contained_in_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 91 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":105
 *         return self.start >= interval.start and self.end <= interval.end
 * 
 *     cpdef bint contains_interval(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether the other spanning entity is
 *         completely contained inside this entity
 */

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_13contains_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval, int __pyx_skip_dispatch) {

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_5;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains_interval", 0);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_contains_interval); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_13contains_interval)) {

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_4 = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_4)) {

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_interval));

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 105, __pyx_L1_error)

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = __pyx_t_5;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":117
 *         bool
 *         """
 *         return self.start <= interval.start and self.end >= interval.end             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 117 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->start <= __pyx_v_interval->start) != 0);

#line 117 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_6) {

#line 117 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 117 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 117 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L3_bool_binop_done;

#line 117 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 117 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = ((__pyx_v_self->end >= __pyx_v_interval->end) != 0);

#line 117 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = __pyx_t_6;

#line 117 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_bool_binop_done:;

#line 117 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_5;

#line 117 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":105
 *         return self.start >= interval.start and self.end <= interval.end
 * 
 *     cpdef bint contains_interval(self, SpanningMixin interval):             # <<<<<<<<<<<<<<
 *         """Tests whether the other spanning entity is
 *         completely contained inside this entity
 */

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.contains_interval", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_13contains_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_12contains_interval[] = "SpanningMixin.contains_interval(self, SpanningMixin interval) -> bool\nTests whether the other spanning entity is\n        completely contained inside this entity\n\n        Parameters\n        ----------\n        interval : SpanningMixin\n\n        Returns\n        -------\n        bool\n        ";

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_13contains_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains_interval (wrapper)", 0);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, 1, "interval", 0))) __PYX_ERR(0, 105, __pyx_L1_error)
  __pyx_r = 
#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_12contains_interval(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_interval));

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_12contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval) {

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains_interval", 0);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains_interval(__pyx_v_self, __pyx_v_interval, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.contains_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 105 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":3
 * cdef class SpanningMixin(object):
 *     cdef:
 *         public double start             # <<<<<<<<<<<<<<
 *         public double end
 * 
 */

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_1__get__(PyObject *__pyx_v_self) {

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self));

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self) {

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3, __pyx_L1_error)

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_t_1;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.start.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = NULL;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_value) {

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double __pyx_t_1;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L1_error)

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->start = __pyx_t_1;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.start.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 3 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":4
 *     cdef:
 *         public double start
 *         public double end             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _contains(self, double i)
 */

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_1__get__(PyObject *__pyx_v_self) {

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self));

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self) {

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_t_1;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.end.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = NULL;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v_value) {

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double __pyx_t_1;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->end = __pyx_t_1;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.end.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_14__reduce_cython__[] = "SpanningMixin.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_14__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_14__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_4;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.end, self.start)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.end, self.start)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_3;

#line 6 "(tree fragment)"
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.end, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_5 = (__pyx_t_4 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));

#line 8 "(tree fragment)"
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.end, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = 0;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_5 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_5) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_SpanningMixin); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 13 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_214708752);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_214708752);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_214708752);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_2 = 0;

#line 13 "(tree fragment)"
    __pyx_t_3 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_1;

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, None), state
 *     else:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SpanningMixin__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_SpanningMixin); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_214708752);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_214708752);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_214708752);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    __pyx_t_3 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_2;

#line 15 "(tree fragment)"
    __pyx_t_2 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SpanningMixin__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_16__setstate_cython__[] = "SpanningMixin.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_16__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_16__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SpanningMixin__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SpanningMixin, (type(self), 0xccc3210, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SpanningMixin__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":121
 * 
 * cdef class SimpleInterval(SpanningMixin):
 *     def __init__(self, start, end):             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_start = 0
#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_end = 0
#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,0};

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[2] = {0,0};

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  2: 
#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else 
#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); 
#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 121, __pyx_L3_error)

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 121, __pyx_L3_error)

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_argtuple_error;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_start = values[0];

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_end = values[1];

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 121, __pyx_L3_error)

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SimpleInterval.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return -1;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval___init__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *)__pyx_v_self), __pyx_v_start, __pyx_v_end);

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end) {

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_1;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":122
 * cdef class SimpleInterval(SpanningMixin):
 *     def __init__(self, start, end):
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 * 
 */

#line 122 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_start); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 122, __pyx_L1_error)

#line 122 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->__pyx_base.start = __pyx_t_1;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":123
 *     def __init__(self, start, end):
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 123 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_end); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->__pyx_base.end = __pyx_t_1;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":121
 * 
 * cdef class SimpleInterval(SpanningMixin):
 *     def __init__(self, start, end):             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SimpleInterval.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 121 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":126
 * 
 *     @staticmethod
 *     cdef SimpleInterval _create(double start, double end):             # <<<<<<<<<<<<<<
 *         cdef SimpleInterval self = SimpleInterval.__new__(SimpleInterval)
 *         self.start = start
 */

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval__create(double __pyx_v_start, double __pyx_v_end) {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v_self = 0
#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_r = NULL;

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":127
 *     @staticmethod
 *     cdef SimpleInterval _create(double start, double end):
 *         cdef SimpleInterval self = SimpleInterval.__new__(SimpleInterval)             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */

#line 127 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

#line 127 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *)__pyx_t_1);

#line 127 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":128
 *     cdef SimpleInterval _create(double start, double end):
 *         cdef SimpleInterval self = SimpleInterval.__new__(SimpleInterval)
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 *         return self
 */

#line 128 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->__pyx_base.start = __pyx_v_start;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":129
 *         cdef SimpleInterval self = SimpleInterval.__new__(SimpleInterval)
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 129 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->__pyx_base.end = __pyx_v_end;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":130
 *         self.start = start
 *         self.end = end
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 130 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 130 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 130 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_v_self;

#line 130 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":126
 * 
 *     @staticmethod
 *     cdef SimpleInterval _create(double start, double end):             # <<<<<<<<<<<<<<
 *         cdef SimpleInterval self = SimpleInterval.__new__(SimpleInterval)
 *         self.start = start
 */

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SimpleInterval._create", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_self);

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 126 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":132
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({self.start}, {self.end})".format(self=self)
 * 
 */

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_3__repr__(PyObject *__pyx_v_self); /*proto*/

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_3__repr__(PyObject *__pyx_v_self) {

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_2__repr__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *)__pyx_v_self));

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_2__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v_self) {

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":133
 * 
 *     def __repr__(self):
 *         return "{self.__class__.__name__}({self.start}, {self.end})".format(self=self)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_self___class_____name___self_st, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_3;

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = 0;

#line 133 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":132
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({self.start}, {self.end})".format(self=self)
 * 
 */

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SimpleInterval.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 132 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_4__reduce_cython__[] = "SimpleInterval.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_4__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_4__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_4;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.end, self.start)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.end); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.start); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.end, self.start)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_3;

#line 6 "(tree fragment)"
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.end, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_5 = (__pyx_t_4 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));

#line 8 "(tree fragment)"
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.end, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_SimpleInterval, (type(self), 0xccc3210, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = 0;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SimpleInterval, (type(self), 0xccc3210, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_5 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_5) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_SimpleInterval, (type(self), 0xccc3210, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_SimpleInterval, (type(self), 0xccc3210, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_SimpleInterval); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 13 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_214708752);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_214708752);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_214708752);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_2 = 0;

#line 13 "(tree fragment)"
    __pyx_t_3 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_1;

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SimpleInterval, (type(self), 0xccc3210, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_SimpleInterval, (type(self), 0xccc3210, None), state
 *     else:
 *         return __pyx_unpickle_SimpleInterval, (type(self), 0xccc3210, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SimpleInterval__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_SimpleInterval); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_214708752);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_214708752);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_214708752);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    __pyx_t_3 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_2;

#line 15 "(tree fragment)"
    __pyx_t_2 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SimpleInterval.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SimpleInterval, (type(self), 0xccc3210, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SimpleInterval__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_6__setstate_cython__[] = "SimpleInterval.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_6__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_6__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_SimpleInterval, (type(self), 0xccc3210, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SimpleInterval__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SimpleInterval__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SimpleInterval, (type(self), 0xccc3210, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SimpleInterval__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SimpleInterval.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":137
 * 
 * cdef class Interval(SpanningMixin):
 *     def __init__(self, start, end, members=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if members is None:
 *             members = []
 */

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_start = 0
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_end = 0
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_members = 0
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_kwargs);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,&__pyx_n_s_members,0};

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[3] = {0,0,0};

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    values[2] = ((PyObject *)Py_None);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  3: 
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else 
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); 
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 137, __pyx_L3_error)

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  2:

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (kw_args > 0) {

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_members);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 137, __pyx_L3_error)

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        break;
        default: 
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_start = values[0];

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_end = values[1];

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_members = values[2];

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 137, __pyx_L3_error)

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return -1;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval___init__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self), __pyx_v_start, __pyx_v_end, __pyx_v_members, __pyx_v_kwargs);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_members, PyObject *__pyx_v_kwargs) {

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_1;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_2;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_4;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_members);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":138
 * cdef class Interval(SpanningMixin):
 *     def __init__(self, start, end, members=None, **kwargs):
 *         if members is None:             # <<<<<<<<<<<<<<
 *             members = []
 *         self.start = start
 */

#line 138 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_v_members == Py_None);

#line 138 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 138 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":139
 *     def __init__(self, start, end, members=None, **kwargs):
 *         if members is None:
 *             members = []             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */

#line 139 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 139 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF_SET(__pyx_v_members, __pyx_t_3);

#line 139 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":138
 * cdef class Interval(SpanningMixin):
 *     def __init__(self, start, end, members=None, **kwargs):
 *         if members is None:             # <<<<<<<<<<<<<<
 *             members = []
 *         self.start = start
 */

#line 138 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":140
 *         if members is None:
 *             members = []
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 *         self.members = members
 */

#line 140 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_start); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 140, __pyx_L1_error)

#line 140 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->__pyx_base.start = __pyx_t_4;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":141
 *             members = []
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 *         self.members = members
 *         self.data = kwargs
 */

#line 141 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_end); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 141, __pyx_L1_error)

#line 141 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->__pyx_base.end = __pyx_t_4;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":142
 *         self.start = start
 *         self.end = end
 *         self.members = members             # <<<<<<<<<<<<<<
 *         self.data = kwargs
 * 
 */

#line 142 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_members);

#line 142 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_members);

#line 142 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->members);

#line 142 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_v_self->members);

#line 142 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->members = __pyx_v_members;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":143
 *         self.end = end
 *         self.members = members
 *         self.data = kwargs             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 143 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_kwargs);

#line 143 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_kwargs);

#line 143 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->data);

#line 143 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_v_self->data);

#line 143 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->data = __pyx_v_kwargs;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":137
 * 
 * cdef class Interval(SpanningMixin):
 *     def __init__(self, start, end, members=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if members is None:
 *             members = []
 */

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_members);

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 137 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":145
 *         self.data = kwargs
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)
 * 
 */

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_3__repr__(PyObject *__pyx_v_self); /*proto*/

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_3__repr__(PyObject *__pyx_v_self) {

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_2__repr__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_2__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":146
 * 
 *     def __repr__(self):
 *         return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, k):
 */

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L1_error)

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L1_error)

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 146, __pyx_L1_error)

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 146, __pyx_L1_error)

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_self->data);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_self->data);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_self->data);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = 0;

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = 0;

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_s_start_r_end_r_data_r, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 146, __pyx_L1_error)

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_3;

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = 0;

#line 146 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":145
 *         self.data = kwargs
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)
 * 
 */

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 145 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":148
 *         return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)
 * 
 *     def __getitem__(self, k):             # <<<<<<<<<<<<<<
 *         return self.members[k]
 * 
 */

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_k); /*proto*/

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_k) {

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4__getitem__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self), ((PyObject *)__pyx_v_k));

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4__getitem__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_k) {

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":149
 * 
 *     def __getitem__(self, k):
 *         return self.members[k]             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 149 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 149 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->members, __pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)

#line 149 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 149 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 149 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 149 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":148
 *         return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)
 * 
 *     def __getitem__(self, k):             # <<<<<<<<<<<<<<
 *         return self.members[k]
 * 
 */

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 148 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":151
 *         return self.members[k]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.members)
 * 
 */

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7__len__(PyObject *__pyx_v_self); /*proto*/

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7__len__(PyObject *__pyx_v_self) {

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_r;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_6__len__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_6__len__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_r;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_2;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":152
 * 
 *     def __len__(self):
 *         return len(self.members)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */

#line 152 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_v_self->members;

#line 152 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 152 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 152, __pyx_L1_error)

#line 152 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 152 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_2;

#line 152 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":151
 *         return self.members[k]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.members)
 * 
 */

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 151 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":154
 *         return len(self.members)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.members)
 * 
 */

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_9__iter__(PyObject *__pyx_v_self); /*proto*/

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_9__iter__(PyObject *__pyx_v_self) {

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = 
#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_8__iter__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_8__iter__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":155
 * 
 *     def __iter__(self):
 *         return iter(self.members)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 155 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 155 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_v_self->members;

#line 155 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 155 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 155, __pyx_L1_error)

#line 155 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 155 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 155 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_2;

#line 155 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = 0;

#line 155 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":154
 *         return len(self.members)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.members)
 * 
 */

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 154 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":21
 * cdef class Interval(SpanningMixin):
 *     cdef:
 *         public object members             # <<<<<<<<<<<<<<
 *         public dict data
 * 
 */

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_1__get__(PyObject *__pyx_v_self) {

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_v_self->members);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_v_self->members;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_value) {

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_v_value);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(__pyx_v_value);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->members);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->members);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->members = __pyx_v_value;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_5__del__(PyObject *__pyx_v_self) {

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(Py_None);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(Py_None);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->members);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->members);

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->members = Py_None;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":22
 *     cdef:
 *         public object members
 *         public dict data             # <<<<<<<<<<<<<<
 * 
 * cdef class SimpleInterval(SpanningMixin):
 */

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_1__get__(PyObject *__pyx_v_self) {

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_v_self->data);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_v_self->data;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v_value) {

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 22, __pyx_L1_error)

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->data);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->data);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->data = ((PyObject*)__pyx_t_1);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.data.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_5__del__(PyObject *__pyx_v_self) {

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(Py_None);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(Py_None);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->data);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->data);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->data = ((PyObject*)Py_None);

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_10__reduce_cython__[] = "Interval.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_10__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_10__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_4;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  int __pyx_t_6;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.data, self.end, self.members, self.start)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.end); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.start); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->data);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->data);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->data);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->members);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->members);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->members);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.data, self.end, self.members, self.start)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_3;

#line 6 "(tree fragment)"
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.data, self.end, self.members, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_5 = (__pyx_t_4 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));

#line 8 "(tree fragment)"
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.data is not None or self.members is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.data, self.end, self.members, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.data is not None or self.members is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_4 = (__pyx_v_self->data != ((PyObject*)Py_None));

#line 11 "(tree fragment)"
    __pyx_t_6 = (__pyx_t_4 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_6) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_5 = __pyx_t_6;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_6 = (__pyx_v_self->members != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6 != 0);

#line 11 "(tree fragment)"
    __pyx_t_5 = __pyx_t_4;

#line 11 "(tree fragment)"
    __pyx_L4_bool_binop_done:;

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_5;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None or self.members is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_5 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_5) {

    /* "(tree fragment)":13
 *         use_setstate = self.data is not None or self.members is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_Interval); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 13 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_246211956);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_246211956);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_246211956);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_2 = 0;

#line 13 "(tree fragment)"
    __pyx_t_3 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_1;

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None or self.members is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, None), state
 *     else:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Interval__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Interval); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_246211956);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_246211956);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_246211956);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    __pyx_t_3 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_2;

#line 15 "(tree fragment)"
    __pyx_t_2 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Interval__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_12__setstate_cython__[] = "Interval.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_12__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_12__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Interval__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Interval, (type(self), 0xeace574, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Interval__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":158
 * 
 * 
 * cpdef list intervals_containing_point(list intervals, double x):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_1intervals_containing_point(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(PyObject *__pyx_v_intervals, double __pyx_v_x, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interval = 0
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_result = 0
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_2;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_3;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_4;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_8;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("intervals_containing_point", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":163
 *         SpanningMixin interval
 *         list result
 *     result = []             # <<<<<<<<<<<<<<
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)
 */

#line 163 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)

#line 163 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 163 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_result = ((PyObject*)__pyx_t_1);

#line 163 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":164
 *         list result
 *     result = []
 *     for i in range(PyList_Size(intervals)):             # <<<<<<<<<<<<<<
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)
 *         if interval.start <= x <= interval.end:
 */

#line 164 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = PyList_Size(__pyx_v_intervals); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 164, __pyx_L1_error)

#line 164 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 164 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 164 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":165
 *     result = []
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)             # <<<<<<<<<<<<<<
 *         if interval.start <= x <= interval.end:
 *             result.append(interval)
 */

#line 165 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = PyList_GetItem(__pyx_v_intervals, __pyx_v_i); if (unlikely(__pyx_t_5 == ((PyObject *)NULL))) __PYX_ERR(0, 165, __pyx_L1_error)

#line 165 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_5);

#line 165 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 165 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF_SET(__pyx_v_interval, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_t_1));

#line 165 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":166
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)
 *         if interval.start <= x <= interval.end:             # <<<<<<<<<<<<<<
 *             result.append(interval)
 *     return result
 */

#line 166 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = (__pyx_v_interval->start <= __pyx_v_x);

#line 166 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_6) {

#line 166 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_6 = (__pyx_v_x <= __pyx_v_interval->end);

#line 166 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 166 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (__pyx_t_6 != 0);

#line 166 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":167
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)
 *         if interval.start <= x <= interval.end:
 *             result.append(interval)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */

#line 167 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_interval)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 167, __pyx_L1_error)

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":166
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin>PyList_GetItem(intervals, i)
 *         if interval.start <= x <= interval.end:             # <<<<<<<<<<<<<<
 *             result.append(interval)
 *     return result
 */

#line 166 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 166 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":168
 *         if interval.start <= x <= interval.end:
 *             result.append(interval)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 168 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 168 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_result);

#line 168 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_v_result;

#line 168 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":158
 * 
 * 
 * cpdef list intervals_containing_point(list intervals, double x):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.intervals_containing_point", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_interval);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_1intervals_containing_point(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point[] = "intervals_containing_point(list intervals, double x) -> list";

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_1intervals_containing_point(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_intervals = 0
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  double __pyx_v_x
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("intervals_containing_point (wrapper)", 0);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_intervals,&__pyx_n_s_x,0};

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[2] = {0,0};

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  2: 
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_intervals)) != 0)) kw_args--;
        else 
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("intervals_containing_point", 1, 2, 2, 1); 
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 158, __pyx_L3_error)

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "intervals_containing_point") < 0)) __PYX_ERR(0, 158, __pyx_L3_error)

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_argtuple_error;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_intervals = ((PyObject*)values[0]);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 158, __pyx_L3_error)

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("intervals_containing_point", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 158, __pyx_L3_error)

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.intervals_containing_point", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return NULL;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_intervals), (&PyList_Type), 1, "intervals", 1))) __PYX_ERR(0, 158, __pyx_L1_error)
  __pyx_r = 
#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(__pyx_self, __pyx_v_intervals, __pyx_v_x);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_intervals, double __pyx_v_x) {

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("intervals_containing_point", 0);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(__pyx_v_intervals, __pyx_v_x, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.intervals_containing_point", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 158 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":176
 * cdef class IntervalTreeNode(object):
 * 
 *     def __init__(self, double center, IntervalTreeNode left, list contained, IntervalTreeNode right,             # <<<<<<<<<<<<<<
 *                  int level=0, IntervalTreeNode parent=None):
 *         cdef:
 */

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_center
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_left = 0
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_contained = 0
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_right = 0
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  int __pyx_v_level
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_parent = 0
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_center,&__pyx_n_s_left,&__pyx_n_s_contained,&__pyx_n_s_right,&__pyx_n_s_level,&__pyx_n_s_parent,0};

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":177
 * 
 *     def __init__(self, double center, IntervalTreeNode left, list contained, IntervalTreeNode right,
 *                  int level=0, IntervalTreeNode parent=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double start, end, i_start, i_end
 */

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    values[5] = (PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  6: 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_center)) != 0)) kw_args--;
        else 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 1); 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 176, __pyx_L3_error)

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  2:

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_contained)) != 0)) kw_args--;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 2); 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 176, __pyx_L3_error)

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  3:

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 3); 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 176, __pyx_L3_error)

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  4:

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (kw_args > 0) {

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_level);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  5:

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (kw_args > 0) {

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parent);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (value) { values[5] = value; kw_args--; }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 176, __pyx_L3_error)

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        break;
        default: 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_center = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_center == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 176, __pyx_L3_error)

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)values[1]);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_contained = ((PyObject*)values[2]);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)values[3]);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (values[4]) {

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_level = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_level == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 177, __pyx_L3_error)

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_level = ((int)0);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)values[5]);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, PyTuple_GET_SIZE(__pyx_args)); 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 176, __pyx_L3_error)

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return -1;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, 1, "left", 0))) __PYX_ERR(0, 176, __pyx_L1_error)

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_contained), (&PyList_Type), 1, "contained", 1))) __PYX_ERR(0, 176, __pyx_L1_error)

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, 1, "right", 0))) __PYX_ERR(0, 176, __pyx_L1_error)

#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_parent), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, 1, "parent", 0))) __PYX_ERR(0, 177, __pyx_L1_error)
  __pyx_r = 
#line 177 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode___init__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), __pyx_v_center, __pyx_v_left, __pyx_v_contained, __pyx_v_right, __pyx_v_level, __pyx_v_parent);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":176
 * cdef class IntervalTreeNode(object):
 * 
 *     def __init__(self, double center, IntervalTreeNode left, list contained, IntervalTreeNode right,             # <<<<<<<<<<<<<<
 *                  int level=0, IntervalTreeNode parent=None):
 *         cdef:
 */

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_center, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_left, PyObject *__pyx_v_contained, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_right, int __pyx_v_level, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_parent) {
  double __pyx_v_start
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  double __pyx_v_end
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  double __pyx_v_i_start
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  double __pyx_v_i_end
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  size_t __pyx_v_i
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_interval = NULL
#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_2;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_4;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_5;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":182
 *             size_t i
 * 
 *         self.center = center             # <<<<<<<<<<<<<<
 *         self.left = left
 *         self.contained = contained
 */

#line 182 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->center = __pyx_v_center;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":183
 * 
 *         self.center = center
 *         self.left = left             # <<<<<<<<<<<<<<
 *         self.contained = contained
 *         self.right = right
 */

#line 183 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_left));

#line 183 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));

#line 183 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->left);

#line 183 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->left));

#line 183 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->left = __pyx_v_left;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":184
 *         self.center = center
 *         self.left = left
 *         self.contained = contained             # <<<<<<<<<<<<<<
 *         self.right = right
 *         self.level = level
 */

#line 184 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_contained);

#line 184 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_contained);

#line 184 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->contained);

#line 184 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_v_self->contained);

#line 184 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->contained = __pyx_v_contained;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":185
 *         self.left = left
 *         self.contained = contained
 *         self.right = right             # <<<<<<<<<<<<<<
 *         self.level = level
 *         self.parent = parent
 */

#line 185 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_right));

#line 185 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));

#line 185 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->right);

#line 185 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->right));

#line 185 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->right = __pyx_v_right;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":186
 *         self.contained = contained
 *         self.right = right
 *         self.level = level             # <<<<<<<<<<<<<<
 *         self.parent = parent
 * 
 */

#line 186 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->level = __pyx_v_level;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":187
 *         self.right = right
 *         self.level = level
 *         self.parent = parent             # <<<<<<<<<<<<<<
 * 
 *         start = INF
 */

#line 187 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_parent));

#line 187 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_parent));

#line 187 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->parent);

#line 187 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->parent));

#line 187 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->parent = __pyx_v_parent;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":189
 *         self.parent = parent
 * 
 *         start = INF             # <<<<<<<<<<<<<<
 *         end = -INF
 *         i = 0
 */

#line 189 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_start = __pyx_v_12ms_deisotope_2_c_23peak_dependency_network_9intervals_INF;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":190
 * 
 *         start = INF
 *         end = -INF             # <<<<<<<<<<<<<<
 *         i = 0
 * 
 */

#line 190 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_end = (-__pyx_v_12ms_deisotope_2_c_23peak_dependency_network_9intervals_INF);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":191
 *         start = INF
 *         end = -INF
 *         i = 0             # <<<<<<<<<<<<<<
 * 
 *         for interval in self.contained:
 */

#line 191 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":193
 *         i = 0
 * 
 *         for interval in self.contained:             # <<<<<<<<<<<<<<
 *             i += 1
 *             i_start = interval.start
 */

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_v_self->contained == Py_None)) {

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __PYX_ERR(0, 193, __pyx_L1_error)

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_v_self->contained; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  for (;;) {

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 193, __pyx_L1_error)

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #else

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 193, __pyx_L1_error)

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF_SET(__pyx_v_interval, __pyx_t_3);

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":194
 * 
 *         for interval in self.contained:
 *             i += 1             # <<<<<<<<<<<<<<
 *             i_start = interval.start
 *             if i_start < start:
 */

#line 194 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_i = (__pyx_v_i + 1);

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":195
 *         for interval in self.contained:
 *             i += 1
 *             i_start = interval.start             # <<<<<<<<<<<<<<
 *             if i_start < start:
 *                 start = i_start
 */

#line 195 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 195, __pyx_L1_error)

#line 195 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 195 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L1_error)

#line 195 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 195 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_i_start = __pyx_t_4;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":196
 *             i += 1
 *             i_start = interval.start
 *             if i_start < start:             # <<<<<<<<<<<<<<
 *                 start = i_start
 *             i_end = interval.end
 */

#line 196 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = ((__pyx_v_i_start < __pyx_v_start) != 0);

#line 196 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":197
 *             i_start = interval.start
 *             if i_start < start:
 *                 start = i_start             # <<<<<<<<<<<<<<
 *             i_end = interval.end
 *             if i_end > end:
 */

#line 197 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_start = __pyx_v_i_start;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":196
 *             i += 1
 *             i_start = interval.start
 *             if i_start < start:             # <<<<<<<<<<<<<<
 *                 start = i_start
 *             i_end = interval.end
 */

#line 196 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":198
 *             if i_start < start:
 *                 start = i_start
 *             i_end = interval.end             # <<<<<<<<<<<<<<
 *             if i_end > end:
 *                 end = i_end
 */

#line 198 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 198, __pyx_L1_error)

#line 198 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 198 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 198, __pyx_L1_error)

#line 198 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 198 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_i_end = __pyx_t_4;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":199
 *                 start = i_start
 *             i_end = interval.end
 *             if i_end > end:             # <<<<<<<<<<<<<<
 *                 end = i_end
 * 
 */

#line 199 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = ((__pyx_v_i_end > __pyx_v_end) != 0);

#line 199 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":200
 *             i_end = interval.end
 *             if i_end > end:
 *                 end = i_end             # <<<<<<<<<<<<<<
 * 
 *         if i > 0:
 */

#line 200 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_end = __pyx_v_i_end;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":199
 *                 start = i_start
 *             i_end = interval.end
 *             if i_end > end:             # <<<<<<<<<<<<<<
 *                 end = i_end
 * 
 */

#line 199 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":193
 *         i = 0
 * 
 *         for interval in self.contained:             # <<<<<<<<<<<<<<
 *             i += 1
 *             i_start = interval.start
 */

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 193 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":202
 *                 end = i_end
 * 
 *         if i > 0:             # <<<<<<<<<<<<<<
 *             self.start = start
 *             self.end = end
 */

#line 202 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = ((__pyx_v_i > 0) != 0);

#line 202 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":203
 * 
 *         if i > 0:
 *             self.start = start             # <<<<<<<<<<<<<<
 *             self.end = end
 *         else:
 */

#line 203 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_self->start = __pyx_v_start;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":204
 *         if i > 0:
 *             self.start = start
 *             self.end = end             # <<<<<<<<<<<<<<
 *         else:
 *             self.start = self.end = center
 */

#line 204 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_self->end = __pyx_v_end;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":202
 *                 end = i_end
 * 
 *         if i > 0:             # <<<<<<<<<<<<<<
 *             self.start = start
 *             self.end = end
 */

#line 202 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L7;

#line 202 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":206
 *             self.end = end
 *         else:
 *             self.start = self.end = center             # <<<<<<<<<<<<<<
 * 
 *     cpdef list contains_point(self, double x):
 */

#line 206 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /*else*/ {

#line 206 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_self->start = __pyx_v_center;

#line 206 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_self->end = __pyx_v_center;

#line 206 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 206 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L7:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":176
 * cdef class IntervalTreeNode(object):
 * 
 *     def __init__(self, double center, IntervalTreeNode left, list contained, IntervalTreeNode right,             # <<<<<<<<<<<<<<
 *                  int level=0, IntervalTreeNode parent=None):
 *         cdef:
 */

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_interval);

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 176 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":208
 *             self.start = self.end = center
 * 
 *     cpdef list contains_point(self, double x):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list inner
 */

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3contains_point(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contains_point(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_inner = 0
#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_8;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_9;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains_point", 0);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_contains_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3contains_point)) {

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L1_error)

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_5 = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_5)) {

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L1_error)

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 208, __pyx_L1_error)

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":211
 *         cdef:
 *             list inner
 *         inner = intervals_containing_point(self.contained, x)             # <<<<<<<<<<<<<<
 *         if self.left is not None and self.left.start <= x <= self.left.end:
 *             inner.extend(self.left.contains_point(x))
 */

#line 211 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_v_self->contained;

#line 211 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 211 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point(((PyObject*)__pyx_t_1), __pyx_v_x, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 211, __pyx_L1_error)

#line 211 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 211 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 211 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_inner = ((PyObject*)__pyx_t_2);

#line 211 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":212
 *             list inner
 *         inner = intervals_containing_point(self.contained, x)
 *         if self.left is not None and self.left.start <= x <= self.left.end:             # <<<<<<<<<<<<<<
 *             inner.extend(self.left.contains_point(x))
 *             return inner
 */

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = (((PyObject *)__pyx_v_self->left) != Py_None);

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_8 = (__pyx_t_7 != 0);

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_8) {

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = __pyx_t_8;

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L4_bool_binop_done;

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_8 = (__pyx_v_self->left->start <= __pyx_v_x);

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_8) {

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_8 = (__pyx_v_x <= __pyx_v_self->left->end);

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = (__pyx_t_8 != 0);

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = __pyx_t_7;

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_bool_binop_done:;

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":213
 *         inner = intervals_containing_point(self.contained, x)
 *         if self.left is not None and self.left.start <= x <= self.left.end:
 *             inner.extend(self.left.contains_point(x))             # <<<<<<<<<<<<<<
 *             return inner
 *         if self.right is not None and self.right.start <= x <= self.right.end:
 */

#line 213 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_v_inner == Py_None)) {

#line 213 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");

#line 213 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __PYX_ERR(0, 213, __pyx_L1_error)

#line 213 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 213 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->left->__pyx_vtab)->contains_point(__pyx_v_self->left, __pyx_v_x, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)

#line 213 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 213 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_inner, __pyx_t_2); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 213, __pyx_L1_error)

#line 213 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":214
 *         if self.left is not None and self.left.start <= x <= self.left.end:
 *             inner.extend(self.left.contains_point(x))
 *             return inner             # <<<<<<<<<<<<<<
 *         if self.right is not None and self.right.start <= x <= self.right.end:
 *             inner.extend(self.right.contains_point(x))
 */

#line 214 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 214 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_v_inner);

#line 214 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = __pyx_v_inner;

#line 214 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":212
 *             list inner
 *         inner = intervals_containing_point(self.contained, x)
 *         if self.left is not None and self.left.start <= x <= self.left.end:             # <<<<<<<<<<<<<<
 *             inner.extend(self.left.contains_point(x))
 *             return inner
 */

#line 212 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":215
 *             inner.extend(self.left.contains_point(x))
 *             return inner
 *         if self.right is not None and self.right.start <= x <= self.right.end:             # <<<<<<<<<<<<<<
 *             inner.extend(self.right.contains_point(x))
 *             return inner
 */

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = (((PyObject *)__pyx_v_self->right) != Py_None);

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_8 = (__pyx_t_7 != 0);

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_8) {

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = __pyx_t_8;

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L7_bool_binop_done;

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_8 = (__pyx_v_self->right->start <= __pyx_v_x);

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_8) {

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_8 = (__pyx_v_x <= __pyx_v_self->right->end);

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = (__pyx_t_8 != 0);

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = __pyx_t_7;

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L7_bool_binop_done:;

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":216
 *             return inner
 *         if self.right is not None and self.right.start <= x <= self.right.end:
 *             inner.extend(self.right.contains_point(x))             # <<<<<<<<<<<<<<
 *             return inner
 *         else:
 */

#line 216 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_v_inner == Py_None)) {

#line 216 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");

#line 216 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 216 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->right->__pyx_vtab)->contains_point(__pyx_v_self->right, __pyx_v_x, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 216 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_inner, __pyx_t_2); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":217
 *         if self.right is not None and self.right.start <= x <= self.right.end:
 *             inner.extend(self.right.contains_point(x))
 *             return inner             # <<<<<<<<<<<<<<
 *         else:
 *             return inner
 */

#line 217 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 217 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_v_inner);

#line 217 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = __pyx_v_inner;

#line 217 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":215
 *             inner.extend(self.left.contains_point(x))
 *             return inner
 *         if self.right is not None and self.right.start <= x <= self.right.end:             # <<<<<<<<<<<<<<
 *             inner.extend(self.right.contains_point(x))
 *             return inner
 */

#line 215 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":219
 *             return inner
 *         else:
 *             return inner             # <<<<<<<<<<<<<<
 * 
 *     cpdef list _overlaps_interval(self, double start, double end):
 */

#line 219 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /*else*/ {

#line 219 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 219 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_v_inner);

#line 219 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = __pyx_v_inner;

#line 219 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

#line 219 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":208
 *             self.start = self.end = center
 * 
 *     cpdef list contains_point(self, double x):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list inner
 */

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.contains_point", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_inner);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3contains_point(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_2contains_point[] = "IntervalTreeNode.contains_point(self, double x) -> list";

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3contains_point(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  double __pyx_v_x
#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains_point (wrapper)", 0);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  assert(__pyx_arg_x); {

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_x = __pyx_PyFloat_AsDouble(__pyx_arg_x); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 208, __pyx_L3_error)

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.contains_point", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_2contains_point(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((double)__pyx_v_x));

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_2contains_point(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_x) {

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains_point", 0);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contains_point(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.contains_point", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 208 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":221
 *             return inner
 * 
 *     cpdef list _overlaps_interval(self, double start, double end):             # <<<<<<<<<<<<<<
 *         cdef:
 *             SpanningMixin interv
 */

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5_overlaps_interval(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__overlaps_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end, int __pyx_skip_dispatch) {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interv = 0
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v_query = 0
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  size_t __pyx_v_i
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_result = NULL
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_9;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_10;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_11;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_12;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_13;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_14;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_overlaps_interval", 0);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlaps_interval); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 221, __pyx_L1_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5_overlaps_interval)) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 221, __pyx_L1_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L1_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; 
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_6 = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_7 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_6)) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_6);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_5, function);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_t_7 = 1;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #if CYTHON_FAST_PYCALL

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (PyFunction_Check(__pyx_t_5)) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        } else

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #endif

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #if CYTHON_FAST_PYCCALL

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        } else

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #endif

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 221, __pyx_L1_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_8);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (__pyx_t_6) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_3 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 221, __pyx_L1_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":226
 *             SimpleInterval query
 *             size_t i
 *         query = SimpleInterval._create(start, end)             # <<<<<<<<<<<<<<
 *         result = []
 *         for i in range(PyList_Size(self.contained)):
 */

#line 226 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval__create(__pyx_v_start, __pyx_v_end)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)

#line 226 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 226 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_query = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *)__pyx_t_1);

#line 226 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":227
 *             size_t i
 *         query = SimpleInterval._create(start, end)
 *         result = []             # <<<<<<<<<<<<<<
 *         for i in range(PyList_Size(self.contained)):
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)
 */

#line 227 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)

#line 227 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 227 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_result = ((PyObject*)__pyx_t_1);

#line 227 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":228
 *         query = SimpleInterval._create(start, end)
 *         result = []
 *         for i in range(PyList_Size(self.contained)):             # <<<<<<<<<<<<<<
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)
 *             if query.overlaps(interv):
 */

#line 228 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_v_self->contained;

#line 228 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 228 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_9 = PyList_Size(__pyx_t_1); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 228, __pyx_L1_error)

#line 228 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 228 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_10 = __pyx_t_9;

#line 228 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {

#line 228 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_i = __pyx_t_11;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":229
 *         result = []
 *         for i in range(PyList_Size(self.contained)):
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)             # <<<<<<<<<<<<<<
 *             if query.overlaps(interv):
 *                 result.append(interv)
 */

#line 229 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = __pyx_v_self->contained;

#line 229 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 229 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_12 = PyList_GetItem(__pyx_t_1, __pyx_v_i); if (unlikely(__pyx_t_12 == ((PyObject *)NULL))) __PYX_ERR(0, 229, __pyx_L1_error)

#line 229 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 229 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_12);

#line 229 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 229 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF_SET(__pyx_v_interv, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_t_1));

#line 229 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":230
 *         for i in range(PyList_Size(self.contained)):
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)
 *             if query.overlaps(interv):             # <<<<<<<<<<<<<<
 *                 result.append(interv)
 *         return result
 */

#line 230 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_13 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *)__pyx_v_query->__pyx_base.__pyx_vtab)->__pyx_base.overlaps(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_query), __pyx_v_interv, 0) != 0);

#line 230 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_13) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":231
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)
 *             if query.overlaps(interv):
 *                 result.append(interv)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */

#line 231 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_14 = __Pyx_PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_interv)); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 231, __pyx_L1_error)

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":230
 *         for i in range(PyList_Size(self.contained)):
 *             interv = <SpanningMixin>PyList_GetItem(self.contained, i)
 *             if query.overlaps(interv):             # <<<<<<<<<<<<<<
 *                 result.append(interv)
 *         return result
 */

#line 230 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 230 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":232
 *             if query.overlaps(interv):
 *                 result.append(interv)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef list overlaps(self, double start, double end):
 */

#line 232 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 232 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_result);

#line 232 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_v_result;

#line 232 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":221
 *             return inner
 * 
 *     cpdef list _overlaps_interval(self, double start, double end):             # <<<<<<<<<<<<<<
 *         cdef:
 *             SpanningMixin interv
 */

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._overlaps_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_interv);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_query);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5_overlaps_interval(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4_overlaps_interval[] = "IntervalTreeNode._overlaps_interval(self, double start, double end) -> list";

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5_overlaps_interval(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_start
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  double __pyx_v_end
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_overlaps_interval (wrapper)", 0);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,0};

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[2] = {0,0};

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  2: 
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else 
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_overlaps_interval", 1, 2, 2, 1); 
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 221, __pyx_L3_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_overlaps_interval") < 0)) __PYX_ERR(0, 221, __pyx_L3_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_argtuple_error;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_start = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_start == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 221, __pyx_L3_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_end = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_end == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 221, __pyx_L3_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_overlaps_interval", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 221, __pyx_L3_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._overlaps_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4_overlaps_interval(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), __pyx_v_start, __pyx_v_end);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4_overlaps_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end) {

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_overlaps_interval", 0);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__overlaps_interval(__pyx_v_self, __pyx_v_start, __pyx_v_end, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 221, __pyx_L1_error)

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._overlaps_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 221 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":234
 *         return result
 * 
 *     cpdef list overlaps(self, double start, double end):             # <<<<<<<<<<<<<<
 *         """Returns the list of all contained intervals which
 *         overlap with the interval described by `start` and `end`
 */

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_result = 0
#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_9;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_10;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_11;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_12;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps", 0);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlaps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_7overlaps)) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; 
#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_6 = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_7 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_6)) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_6);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_5, function);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_t_7 = 1;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #if CYTHON_FAST_PYCALL

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (PyFunction_Check(__pyx_t_5)) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        } else

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #endif

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #if CYTHON_FAST_PYCCALL

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        } else

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #endif

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_8);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (__pyx_t_6) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_3 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":251
 *             from all spanning nodes' `contained` list.
 *         """
 *         cdef list result = []             # <<<<<<<<<<<<<<
 *         if start > self.end:
 *             return result
 */

#line 251 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)

#line 251 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 251 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_result = ((PyObject*)__pyx_t_1);

#line 251 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":252
 *         """
 *         cdef list result = []
 *         if start > self.end:             # <<<<<<<<<<<<<<
 *             return result
 *         if end < self.start:
 */

#line 252 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_9 = ((__pyx_v_start > __pyx_v_self->end) != 0);

#line 252 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_9) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":253
 *         cdef list result = []
 *         if start > self.end:
 *             return result             # <<<<<<<<<<<<<<
 *         if end < self.start:
 *             return result
 */

#line 253 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 253 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_v_result);

#line 253 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = __pyx_v_result;

#line 253 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":252
 *         """
 *         cdef list result = []
 *         if start > self.end:             # <<<<<<<<<<<<<<
 *             return result
 *         if end < self.start:
 */

#line 252 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":254
 *         if start > self.end:
 *             return result
 *         if end < self.start:             # <<<<<<<<<<<<<<
 *             return result
 *         elif start <= self.start:
 */

#line 254 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_9 = ((__pyx_v_end < __pyx_v_self->start) != 0);

#line 254 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_9) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":255
 *             return result
 *         if end < self.start:
 *             return result             # <<<<<<<<<<<<<<
 *         elif start <= self.start:
 *             if end < self.start:
 */

#line 255 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 255 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_v_result);

#line 255 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = __pyx_v_result;

#line 255 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":254
 *         if start > self.end:
 *             return result
 *         if end < self.start:             # <<<<<<<<<<<<<<
 *             return result
 *         elif start <= self.start:
 */

#line 254 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":256
 *         if end < self.start:
 *             return result
 *         elif start <= self.start:             # <<<<<<<<<<<<<<
 *             if end < self.start:
 *                 return result
 */

#line 256 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_9 = ((__pyx_v_start <= __pyx_v_self->start) != 0);

#line 256 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_9) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":257
 *             return result
 *         elif start <= self.start:
 *             if end < self.start:             # <<<<<<<<<<<<<<
 *                 return result
 *             else:
 */

#line 257 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 = ((__pyx_v_end < __pyx_v_self->start) != 0);

#line 257 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_9) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":258
 *         elif start <= self.start:
 *             if end < self.start:
 *                 return result             # <<<<<<<<<<<<<<
 *             else:
 *                 if self.left is not None:
 */

#line 258 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 258 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_v_result);

#line 258 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_r = __pyx_v_result;

#line 258 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":257
 *             return result
 *         elif start <= self.start:
 *             if end < self.start:             # <<<<<<<<<<<<<<
 *                 return result
 *             else:
 */

#line 257 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":260
 *                 return result
 *             else:
 *                 if self.left is not None:             # <<<<<<<<<<<<<<
 *                     result.extend(self.left.overlaps(start, end))
 *                 result.extend(self._overlaps_interval(start, end))
 */

#line 260 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    /*else*/ {

#line 260 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_9 = (((PyObject *)__pyx_v_self->left) != Py_None);

#line 260 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_10 = (__pyx_t_9 != 0);

#line 260 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_10) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":261
 *             else:
 *                 if self.left is not None:
 *                     result.extend(self.left.overlaps(start, end))             # <<<<<<<<<<<<<<
 *                 result.extend(self._overlaps_interval(start, end))
 *                 if self.right is not None and end >= self.right.start:
 */

#line 261 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->left->__pyx_vtab)->overlaps(__pyx_v_self->left, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)

#line 261 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 261 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 261, __pyx_L1_error)

#line 261 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":260
 *                 return result
 *             else:
 *                 if self.left is not None:             # <<<<<<<<<<<<<<
 *                     result.extend(self.left.overlaps(start, end))
 *                 result.extend(self._overlaps_interval(start, end))
 */

#line 260 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":262
 *                 if self.left is not None:
 *                     result.extend(self.left.overlaps(start, end))
 *                 result.extend(self._overlaps_interval(start, end))             # <<<<<<<<<<<<<<
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend(self.right.overlaps(start, end))
 */

#line 262 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->__pyx_vtab)->_overlaps_interval(__pyx_v_self, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)

#line 262 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 262 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 262, __pyx_L1_error)

#line 262 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":263
 *                     result.extend(self.left.overlaps(start, end))
 *                 result.extend(self._overlaps_interval(start, end))
 *                 if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:
 */

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_9 = (((PyObject *)__pyx_v_self->right) != Py_None);

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_12 = (__pyx_t_9 != 0);

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_12) {

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      } else {

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_10 = __pyx_t_12;

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L8_bool_binop_done;

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_12 = ((__pyx_v_end >= __pyx_v_self->right->start) != 0);

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_10 = __pyx_t_12;

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_L8_bool_binop_done:;

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_10) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":264
 *                 result.extend(self._overlaps_interval(start, end))
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend(self.right.overlaps(start, end))             # <<<<<<<<<<<<<<
 *         elif start > self.start:
 *             if self.left is not None and self.left.end >= start:
 */

#line 264 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->right->__pyx_vtab)->overlaps(__pyx_v_self->right, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)

#line 264 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 264 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 264, __pyx_L1_error)

#line 264 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":263
 *                     result.extend(self.left.overlaps(start, end))
 *                 result.extend(self._overlaps_interval(start, end))
 *                 if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:
 */

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 263 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":256
 *         if end < self.start:
 *             return result
 *         elif start <= self.start:             # <<<<<<<<<<<<<<
 *             if end < self.start:
 *                 return result
 */

#line 256 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L4;

#line 256 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":265
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:             # <<<<<<<<<<<<<<
 *             if self.left is not None and self.left.end >= start:
 *                 result.extend(self.left.overlaps(start, end))
 */

#line 265 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_10 = ((__pyx_v_start > __pyx_v_self->start) != 0);

#line 265 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":266
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:
 *             if self.left is not None and self.left.end >= start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 */

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_12 = (((PyObject *)__pyx_v_self->left) != Py_None);

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 = (__pyx_t_12 != 0);

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_9) {

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_10 = __pyx_t_9;

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L11_bool_binop_done;

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 = ((__pyx_v_self->left->end >= __pyx_v_start) != 0);

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = __pyx_t_9;

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L11_bool_binop_done:;

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":267
 *         elif start > self.start:
 *             if self.left is not None and self.left.end >= start:
 *                 result.extend(self.left.overlaps(start, end))             # <<<<<<<<<<<<<<
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:
 */

#line 267 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->left->__pyx_vtab)->overlaps(__pyx_v_self->left, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error)

#line 267 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 267 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 267, __pyx_L1_error)

#line 267 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":266
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:
 *             if self.left is not None and self.left.end >= start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 */

#line 266 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":268
 *             if self.left is not None and self.left.end >= start:
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))             # <<<<<<<<<<<<<<
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))
 */

#line 268 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->__pyx_vtab)->_overlaps_interval(__pyx_v_self, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)

#line 268 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 268 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 268, __pyx_L1_error)

#line 268 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":269
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:
 */

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 = (((PyObject *)__pyx_v_self->right) != Py_None);

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_12 = (__pyx_t_9 != 0);

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_12) {

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_10 = __pyx_t_12;

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L14_bool_binop_done;

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_12 = ((__pyx_v_end >= __pyx_v_self->right->start) != 0);

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = __pyx_t_12;

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L14_bool_binop_done:;

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":270
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))             # <<<<<<<<<<<<<<
 *         elif end > self.start:
 *             if self.left is not None:
 */

#line 270 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->right->__pyx_vtab)->overlaps(__pyx_v_self->right, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)

#line 270 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 270 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 270, __pyx_L1_error)

#line 270 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":269
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:
 */

#line 269 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":265
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend(self.right.overlaps(start, end))
 *         elif start > self.start:             # <<<<<<<<<<<<<<
 *             if self.left is not None and self.left.end >= start:
 *                 result.extend(self.left.overlaps(start, end))
 */

#line 265 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L4;

#line 265 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":271
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:             # <<<<<<<<<<<<<<
 *             if self.left is not None:
 *                 result.extend(self.left.overlaps(start, end))
 */

#line 271 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_10 = ((__pyx_v_end > __pyx_v_self->start) != 0);

#line 271 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":272
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:
 *             if self.left is not None:             # <<<<<<<<<<<<<<
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 */

#line 272 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = (((PyObject *)__pyx_v_self->left) != Py_None);

#line 272 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_12 = (__pyx_t_10 != 0);

#line 272 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_12) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":273
 *         elif end > self.start:
 *             if self.left is not None:
 *                 result.extend(self.left.overlaps(start, end))             # <<<<<<<<<<<<<<
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:
 */

#line 273 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->left->__pyx_vtab)->overlaps(__pyx_v_self->left, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 273 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":272
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:
 *             if self.left is not None:             # <<<<<<<<<<<<<<
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 */

#line 272 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":274
 *             if self.left is not None:
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))             # <<<<<<<<<<<<<<
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))
 */

#line 274 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->__pyx_vtab)->_overlaps_interval(__pyx_v_self, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":275
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.right.overlaps(start, end))
 *         return result
 */

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = (((PyObject *)__pyx_v_self->right) != Py_None);

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 = (__pyx_t_10 != 0);

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_9) {

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_12 = __pyx_t_9;

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L18_bool_binop_done;

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 = ((__pyx_v_end >= __pyx_v_self->right->start) != 0);

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_12 = __pyx_t_9;

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L18_bool_binop_done:;

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_12) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":276
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))             # <<<<<<<<<<<<<<
 *         return result
 * 
 */

#line 276 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->right->__pyx_vtab)->overlaps(__pyx_v_self->right, __pyx_v_start, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L1_error)

#line 276 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 276 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 276, __pyx_L1_error)

#line 276 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":275
 *                 result.extend(self.left.overlaps(start, end))
 *             result.extend(self._overlaps_interval(start, end))
 *             if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                 result.extend(self.right.overlaps(start, end))
 *         return result
 */

#line 275 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":271
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))
 *         elif end > self.start:             # <<<<<<<<<<<<<<
 *             if self.left is not None:
 *                 result.extend(self.left.overlaps(start, end))
 */

#line 271 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 271 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":277
 *             if self.right is not None and end >= self.right.start:
 *                 result.extend(self.right.overlaps(start, end))
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 277 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 277 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_result);

#line 277 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_v_result;

#line 277 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":234
 *         return result
 * 
 *     cpdef list overlaps(self, double start, double end):             # <<<<<<<<<<<<<<
 *         """Returns the list of all contained intervals which
 *         overlap with the interval described by `start` and `end`
 */

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6overlaps[] = "IntervalTreeNode.overlaps(self, double start, double end) -> list\nReturns the list of all contained intervals which\n        overlap with the interval described by `start` and `end`\n\n        Parameters\n        ----------\n        start : float\n            The start of the query interval\n        end : float\n            The end of the query interval\n\n        Returns\n        -------\n        list\n            A list of all objects which overlap the argument interval\n            from all spanning nodes' `contained` list.\n        ";

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_start
#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  double __pyx_v_end
#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps (wrapper)", 0);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,0};

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[2] = {0,0};

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  2: 
#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else 
#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("overlaps", 1, 2, 2, 1); 
#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 234, __pyx_L3_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "overlaps") < 0)) __PYX_ERR(0, 234, __pyx_L3_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_argtuple_error;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_start = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_start == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 234, __pyx_L3_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_end = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_end == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 234, __pyx_L3_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("overlaps", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 234, __pyx_L3_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6overlaps(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), __pyx_v_start, __pyx_v_end);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, double __pyx_v_start, double __pyx_v_end) {

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps", 0);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_overlaps(__pyx_v_self, __pyx_v_start, __pyx_v_end, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 234 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":279
 *         return result
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (
 *             self.level, self.center, self.start, self.end)
 */

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9__repr__(PyObject *__pyx_v_self); /*proto*/

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9__repr__(PyObject *__pyx_v_self) {

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_8__repr__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_8__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":280
 * 
 *     def __repr__(self):
 *         return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (             # <<<<<<<<<<<<<<
 *             self.level, self.center, self.start, self.end)
 * 
 */

#line 280 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":281
 *     def __repr__(self):
 *         return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (
 *             self.level, self.center, self.start, self.end)             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, other, int code):
 */

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->level); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->center); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = 0;

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = 0;

#line 281 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":280
 * 
 *     def __repr__(self):
 *         return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (             # <<<<<<<<<<<<<<
 *             self.level, self.center, self.start, self.end)
 * 
 */

#line 280 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_IntervalTreeNode_level_d_center, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 280, __pyx_L1_error)

#line 280 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 280 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 280 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_4;

#line 280 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = 0;

#line 280 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":279
 *         return result
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (
 *             self.level, self.center, self.start, self.end)
 */

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 279 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":283
 *             self.level, self.center, self.start, self.end)
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             return self._eq(other)
 */

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_11__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /*proto*/

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_11__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = 
#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_10__richcmp__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_code));

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_10__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":284
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */

#line 284 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  switch (__pyx_v_code) {

#line 284 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    case 2:

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":285
 *     def __richcmp__(self, other, int code):
 *         if code == 2:
 *             return self._eq(other)             # <<<<<<<<<<<<<<
 *         elif code == 3:
 *             return not (self._eq(other))
 */

#line 285 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 285 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 285, __pyx_L1_error)

#line 285 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->__pyx_vtab)->_eq(__pyx_v_self, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)

#line 285 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 285 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = __pyx_t_1;

#line 285 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = 0;

#line 285 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":284
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */

#line 284 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    break;

#line 284 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    case 3:

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":287
 *             return self._eq(other)
 *         elif code == 3:
 *             return not (self._eq(other))             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, IntervalTreeNode other):
 */

#line 287 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 287 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 287, __pyx_L1_error)

#line 287 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = __Pyx_PyBool_FromLong((!(((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->__pyx_vtab)->_eq(__pyx_v_self, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_other), 0) != 0))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)

#line 287 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 287 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = __pyx_t_1;

#line 287 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = 0;

#line 287 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":286
 *         if code == 2:
 *             return self._eq(other)
 *         elif code == 3:             # <<<<<<<<<<<<<<
 *             return not (self._eq(other))
 * 
 */

#line 286 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    break;

#line 286 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    default: break;

#line 286 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":283
 *             self.level, self.center, self.start, self.end)
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             return self._eq(other)
 */

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 283 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":289
 *             return not (self._eq(other))
 * 
 *     cpdef bint _eq(self, IntervalTreeNode other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_13_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__eq(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_result = NULL
#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_5;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_eq", 0);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_13_eq)) {

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_4 = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_4)) {

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_other)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other));

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 289, __pyx_L1_error)

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = __pyx_t_5;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":290
 * 
 *     cpdef bint _eq(self, IntervalTreeNode other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */

#line 290 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = (((PyObject *)__pyx_v_other) == Py_None);

#line 290 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 290 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":291
 *     cpdef bint _eq(self, IntervalTreeNode other):
 *         if other is None:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             result = self.contained == other.contained
 */

#line 291 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = 0;

#line 291 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":290
 * 
 *     cpdef bint _eq(self, IntervalTreeNode other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */

#line 290 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":293
 *             return False
 *         else:
 *             result = self.contained == other.contained             # <<<<<<<<<<<<<<
 *             if result:
 *                 result = self.left == other.left
 */

#line 293 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /*else*/ {

#line 293 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->contained, __pyx_v_other->contained, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)

#line 293 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_result = __pyx_t_1;

#line 293 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":294
 *         else:
 *             result = self.contained == other.contained
 *             if result:             # <<<<<<<<<<<<<<
 *                 result = self.left == other.left
 *                 if result:
 */

#line 294 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":295
 *             result = self.contained == other.contained
 *             if result:
 *                 result = self.left == other.left             # <<<<<<<<<<<<<<
 *                 if result:
 *                     result = self.right == other.right
 */

#line 295 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self->left), ((PyObject *)__pyx_v_other->left), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_1);

#line 295 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":296
 *             if result:
 *                 result = self.left == other.left
 *                 if result:             # <<<<<<<<<<<<<<
 *                     result = self.right == other.right
 *             return result
 */

#line 296 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 296, __pyx_L1_error)

#line 296 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_6) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":297
 *                 result = self.left == other.left
 *                 if result:
 *                     result = self.right == other.right             # <<<<<<<<<<<<<<
 *             return result
 * 
 */

#line 297 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self->right), ((PyObject *)__pyx_v_other->right), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 297, __pyx_L1_error)

#line 297 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_1);

#line 297 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":296
 *             if result:
 *                 result = self.left == other.left
 *                 if result:             # <<<<<<<<<<<<<<
 *                     result = self.right == other.right
 *             return result
 */

#line 296 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":294
 *         else:
 *             result = self.contained == other.contained
 *             if result:             # <<<<<<<<<<<<<<
 *                 result = self.left == other.left
 *                 if result:
 */

#line 294 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":298
 *                 if result:
 *                     result = self.right == other.right
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */

#line 298 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L1_error)

#line 298 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = __pyx_t_6;

#line 298 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

#line 298 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":289
 *             return not (self._eq(other))
 * 
 *     cpdef bint _eq(self, IntervalTreeNode other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._eq", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_13_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_12_eq[] = "IntervalTreeNode._eq(self, IntervalTreeNode other) -> bool";

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_13_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_eq (wrapper)", 0);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, 1, "other", 0))) __PYX_ERR(0, 289, __pyx_L1_error)
  __pyx_r = 
#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_12_eq(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_other));

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_12_eq(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_other) {

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_eq", 0);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__eq(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._eq", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 289 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":300
 *             return result
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.start, self.center, self.right, self.level))
 * 
 */

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static Py_hash_t __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_15__hash__(PyObject *__pyx_v_self); /*proto*/

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static Py_hash_t __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_15__hash__(PyObject *__pyx_v_self) {

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_hash_t __pyx_r;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = 
#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_14__hash__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static Py_hash_t __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_14__hash__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_hash_t __pyx_r;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_hash_t __pyx_t_5;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":301
 * 
 *     def __hash__(self):
 *         return hash((self.start, self.center, self.right, self.level))             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->center); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->level); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 301, __pyx_L1_error)

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right));

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right));

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)__pyx_v_self->right));

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_3);

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = 0;

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = 0;

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = PyObject_Hash(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_hash_t)-1))) __PYX_ERR(0, 301, __pyx_L1_error)

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_5;

#line 301 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":300
 *             return result
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.start, self.center, self.right, self.level))
 * 
 */

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 300 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":304
 * 
 *     @classmethod
 *     def build(cls, intervals):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalTreeNode root, parent, node, up
 */

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_17build(PyObject *__pyx_v_cls, PyObject *__pyx_v_intervals); /*proto*/

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_16build[] = "IntervalTreeNode.build(type cls, intervals)";

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_17build(PyObject *__pyx_v_cls, PyObject *__pyx_v_intervals) {

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("build (wrapper)", 0);
  __pyx_r = 
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_16build(((PyTypeObject*)__pyx_v_cls), ((PyObject *)__pyx_v_intervals));

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_16build(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_intervals) {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_root = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_parent = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_node = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_up = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_stack = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_members = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_left = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_right = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_contained = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_intervals_ = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_bunch = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_side = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  double __pyx_v_center
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  double __pyx_v_accumulator
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_i = 0
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  size_t __pyx_v_j
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  size_t __pyx_v_n
#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_3;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_4;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_5;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_6;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_7;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_8;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_9;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_10;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_12;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_13;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_14;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("build", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":313
 *             SpanningMixin i
 *             size_t j, n
 *         stack = []             # <<<<<<<<<<<<<<
 *         root = cls(0, None, [], None, -1)
 * 
 */

#line 313 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L1_error)

#line 313 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 313 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_stack = ((PyObject*)__pyx_t_1);

#line 313 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":314
 *             size_t j, n
 *         stack = []
 *         root = cls(0, None, [], None, -1)             # <<<<<<<<<<<<<<
 * 
 *         intervals_ = list(intervals)
 */

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 314, __pyx_L1_error)

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_int_0);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_int_0);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(Py_None);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(Py_None);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(Py_None);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(Py_None);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 3, Py_None);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_int_neg_1);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_int_neg_1);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_int_neg_1);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_v_cls), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 314, __pyx_L1_error)

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_root = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 314 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":316
 *         root = cls(0, None, [], None, -1)
 * 
 *         intervals_ = list(intervals)             # <<<<<<<<<<<<<<
 *         if not intervals_:
 *             return root
 */

#line 316 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PySequence_List(__pyx_v_intervals); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L1_error)

#line 316 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 316 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_intervals_ = ((PyObject*)__pyx_t_1);

#line 316 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":317
 * 
 *         intervals_ = list(intervals)
 *         if not intervals_:             # <<<<<<<<<<<<<<
 *             return root
 *         stack.append((root, intervals_, "left"))
 */

#line 317 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_intervals_) != 0);

#line 317 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = ((!__pyx_t_3) != 0);

#line 317 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_4) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":318
 *         intervals_ = list(intervals)
 *         if not intervals_:
 *             return root             # <<<<<<<<<<<<<<
 *         stack.append((root, intervals_, "left"))
 * 
 */

#line 318 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 318 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_root));

#line 318 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = ((PyObject *)__pyx_v_root);

#line 318 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":317
 * 
 *         intervals_ = list(intervals)
 *         if not intervals_:             # <<<<<<<<<<<<<<
 *             return root
 *         stack.append((root, intervals_, "left"))
 */

#line 317 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":319
 *         if not intervals_:
 *             return root
 *         stack.append((root, intervals_, "left"))             # <<<<<<<<<<<<<<
 * 
 *         while stack:
 */

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L1_error)

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_root));

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_root));

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_root));

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_intervals_);

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_intervals_);

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_intervals_);

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_n_s_left);

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_n_s_left);

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_left);

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_stack, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 319, __pyx_L1_error)

#line 319 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":321
 *         stack.append((root, intervals_, "left"))
 * 
 *         while stack:             # <<<<<<<<<<<<<<
 *             # parent, members, side = stack.pop()
 *             bunch = <tuple>stack.pop()
 */

#line 321 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  while (1) {

#line 321 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = (PyList_GET_SIZE(__pyx_v_stack) != 0);

#line 321 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!__pyx_t_4) break;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":323
 *         while stack:
 *             # parent, members, side = stack.pop()
 *             bunch = <tuple>stack.pop()             # <<<<<<<<<<<<<<
 *             parent = <IntervalTreeNode>PyTuple_GetItem(bunch, 0)
 *             members = <list>PyTuple_GetItem(bunch, 1)
 */

#line 323 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = __Pyx_PyList_Pop(__pyx_v_stack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L1_error)

#line 323 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 323 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = __pyx_t_1;

#line 323 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 323 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 323 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF_SET(__pyx_v_bunch, ((PyObject*)__pyx_t_2));

#line 323 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":324
 *             # parent, members, side = stack.pop()
 *             bunch = <tuple>stack.pop()
 *             parent = <IntervalTreeNode>PyTuple_GetItem(bunch, 0)             # <<<<<<<<<<<<<<
 *             members = <list>PyTuple_GetItem(bunch, 1)
 *             side = <str>PyTuple_GetItem(bunch, 2)
 */

#line 324 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = PyTuple_GetItem(__pyx_v_bunch, 0); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 324, __pyx_L1_error)

#line 324 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = ((PyObject *)__pyx_t_6);

#line 324 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 324 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF_SET(__pyx_v_parent, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_2));

#line 324 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":325
 *             bunch = <tuple>stack.pop()
 *             parent = <IntervalTreeNode>PyTuple_GetItem(bunch, 0)
 *             members = <list>PyTuple_GetItem(bunch, 1)             # <<<<<<<<<<<<<<
 *             side = <str>PyTuple_GetItem(bunch, 2)
 *             n = PyList_Size(members)
 */

#line 325 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = PyTuple_GetItem(__pyx_v_bunch, 1); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 325, __pyx_L1_error)

#line 325 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = ((PyObject *)__pyx_t_6);

#line 325 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 325 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF_SET(__pyx_v_members, ((PyObject*)__pyx_t_2));

#line 325 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":326
 *             parent = <IntervalTreeNode>PyTuple_GetItem(bunch, 0)
 *             members = <list>PyTuple_GetItem(bunch, 1)
 *             side = <str>PyTuple_GetItem(bunch, 2)             # <<<<<<<<<<<<<<
 *             n = PyList_Size(members)
 *             if n > 0:
 */

#line 326 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = PyTuple_GetItem(__pyx_v_bunch, 2); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 326, __pyx_L1_error)

#line 326 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = ((PyObject *)__pyx_t_6);

#line 326 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 326 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF_SET(__pyx_v_side, ((PyObject*)__pyx_t_2));

#line 326 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":327
 *             members = <list>PyTuple_GetItem(bunch, 1)
 *             side = <str>PyTuple_GetItem(bunch, 2)
 *             n = PyList_Size(members)             # <<<<<<<<<<<<<<
 *             if n > 0:
 *                 accumulator = 0.
 */

#line 327 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = PyList_Size(__pyx_v_members); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 327, __pyx_L1_error)

#line 327 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_n = __pyx_t_7;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":328
 *             side = <str>PyTuple_GetItem(bunch, 2)
 *             n = PyList_Size(members)
 *             if n > 0:             # <<<<<<<<<<<<<<
 *                 accumulator = 0.
 *                 j = 0
 */

#line 328 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = ((__pyx_v_n > 0) != 0);

#line 328 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_4) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":329
 *             n = PyList_Size(members)
 *             if n > 0:
 *                 accumulator = 0.             # <<<<<<<<<<<<<<
 *                 j = 0
 *                 for j in range(n):
 */

#line 329 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_accumulator = 0.;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":330
 *             if n > 0:
 *                 accumulator = 0.
 *                 j = 0             # <<<<<<<<<<<<<<
 *                 for j in range(n):
 *                     i = <SpanningMixin>PyList_GetItem(members, j)
 */

#line 330 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_j = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":331
 *                 accumulator = 0.
 *                 j = 0
 *                 for j in range(n):             # <<<<<<<<<<<<<<
 *                     i = <SpanningMixin>PyList_GetItem(members, j)
 *                     accumulator += (i.start + i.end) / 2.
 */

#line 331 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_8 = __pyx_v_n;

#line 331 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_9 = __pyx_t_8;

#line 331 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 331 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_v_j = __pyx_t_10;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":332
 *                 j = 0
 *                 for j in range(n):
 *                     i = <SpanningMixin>PyList_GetItem(members, j)             # <<<<<<<<<<<<<<
 *                     accumulator += (i.start + i.end) / 2.
 *                 if j > 0:
 */

#line 332 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_6 = PyList_GetItem(__pyx_v_members, __pyx_v_j); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 332, __pyx_L1_error)

#line 332 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = ((PyObject *)__pyx_t_6);

#line 332 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_2);

#line 332 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF_SET(__pyx_v_i, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_t_2));

#line 332 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":333
 *                 for j in range(n):
 *                     i = <SpanningMixin>PyList_GetItem(members, j)
 *                     accumulator += (i.start + i.end) / 2.             # <<<<<<<<<<<<<<
 *                 if j > 0:
 *                     center = accumulator / (j + 1)
 */

#line 333 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_v_accumulator = (__pyx_v_accumulator + ((__pyx_v_i->start + __pyx_v_i->end) / 2.));

#line 333 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":334
 *                     i = <SpanningMixin>PyList_GetItem(members, j)
 *                     accumulator += (i.start + i.end) / 2.
 *                 if j > 0:             # <<<<<<<<<<<<<<
 *                     center = accumulator / (j + 1)
 *                 else:
 */

#line 334 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = ((__pyx_v_j > 0) != 0);

#line 334 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_4) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":335
 *                     accumulator += (i.start + i.end) / 2.
 *                 if j > 0:
 *                     center = accumulator / (j + 1)             # <<<<<<<<<<<<<<
 *                 else:
 *                     center = 0.0
 */

#line 335 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_8 = (__pyx_v_j + 1);

#line 335 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__pyx_t_8 == 0)) {

#line 335 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 335 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __PYX_ERR(0, 335, __pyx_L1_error)

#line 335 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 335 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_v_center = (__pyx_v_accumulator / __pyx_t_8);

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":334
 *                     i = <SpanningMixin>PyList_GetItem(members, j)
 *                     accumulator += (i.start + i.end) / 2.
 *                 if j > 0:             # <<<<<<<<<<<<<<
 *                     center = accumulator / (j + 1)
 *                 else:
 */

#line 334 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L9;

#line 334 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":337
 *                     center = accumulator / (j + 1)
 *                 else:
 *                     center = 0.0             # <<<<<<<<<<<<<<
 * 
 *                 left = []
 */

#line 337 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      /*else*/ {

#line 337 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_v_center = 0.0;

#line 337 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 337 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_L9:;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":339
 *                     center = 0.0
 * 
 *                 left = []             # <<<<<<<<<<<<<<
 *                 right = []
 *                 contained = []
 */

#line 339 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 339, __pyx_L1_error)

#line 339 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 339 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF_SET(__pyx_v_left, ((PyObject*)__pyx_t_2));

#line 339 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":340
 * 
 *                 left = []
 *                 right = []             # <<<<<<<<<<<<<<
 *                 contained = []
 * 
 */

#line 340 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 340, __pyx_L1_error)

#line 340 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 340 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF_SET(__pyx_v_right, ((PyObject*)__pyx_t_2));

#line 340 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":341
 *                 left = []
 *                 right = []
 *                 contained = []             # <<<<<<<<<<<<<<
 * 
 *                 if n > 20:
 */

#line 341 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 341, __pyx_L1_error)

#line 341 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 341 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF_SET(__pyx_v_contained, ((PyObject*)__pyx_t_2));

#line 341 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":343
 *                 contained = []
 * 
 *                 if n > 20:             # <<<<<<<<<<<<<<
 *                     j = 0
 *                     for j in range(n):
 */

#line 343 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = ((__pyx_v_n > 20) != 0);

#line 343 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_4) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":344
 * 
 *                 if n > 20:
 *                     j = 0             # <<<<<<<<<<<<<<
 *                     for j in range(n):
 *                         i = <SpanningMixin>PyList_GetItem(members, j)
 */

#line 344 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_v_j = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":345
 *                 if n > 20:
 *                     j = 0
 *                     for j in range(n):             # <<<<<<<<<<<<<<
 *                         i = <SpanningMixin>PyList_GetItem(members, j)
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
 */

#line 345 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_8 = __pyx_v_n;

#line 345 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_9 = __pyx_t_8;

#line 345 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 345 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_v_j = __pyx_t_10;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":346
 *                     j = 0
 *                     for j in range(n):
 *                         i = <SpanningMixin>PyList_GetItem(members, j)             # <<<<<<<<<<<<<<
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
 *                             contained.append(i)
 */

#line 346 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_6 = PyList_GetItem(__pyx_v_members, __pyx_v_j); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 346, __pyx_L1_error)

#line 346 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = ((PyObject *)__pyx_t_6);

#line 346 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_INCREF(__pyx_t_2);

#line 346 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_XDECREF_SET(__pyx_v_i, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_t_2));

#line 346 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = 0;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":347
 *                     for j in range(n):
 *                         i = <SpanningMixin>PyList_GetItem(members, j)
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:             # <<<<<<<<<<<<<<
 *                             contained.append(i)
 *                         elif center > i.end:
 */

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_3 = ((fabs((__pyx_v_i->start - __pyx_v_center)) < 1e-6) != 0);

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (__pyx_t_3) {

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          } else {

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_t_4 = __pyx_t_3;

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            goto __pyx_L14_bool_binop_done;

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_3 = ((fabs((__pyx_v_i->end - __pyx_v_center)) < 1e-6) != 0);

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = __pyx_t_3;

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_L14_bool_binop_done:;

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (__pyx_t_4) {

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":348
 *                         i = <SpanningMixin>PyList_GetItem(members, j)
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
 *                             contained.append(i)             # <<<<<<<<<<<<<<
 *                         elif center > i.end:
 *                             left.append(i)
 */

#line 348 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_contained, ((PyObject *)__pyx_v_i)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 348, __pyx_L1_error)

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":347
 *                     for j in range(n):
 *                         i = <SpanningMixin>PyList_GetItem(members, j)
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:             # <<<<<<<<<<<<<<
 *                             contained.append(i)
 *                         elif center > i.end:
 */

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            goto __pyx_L13;

#line 347 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":349
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
 *                             contained.append(i)
 *                         elif center > i.end:             # <<<<<<<<<<<<<<
 *                             left.append(i)
 *                         elif center < i.start:
 */

#line 349 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = ((__pyx_v_center > __pyx_v_i->end) != 0);

#line 349 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (__pyx_t_4) {

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":350
 *                             contained.append(i)
 *                         elif center > i.end:
 *                             left.append(i)             # <<<<<<<<<<<<<<
 *                         elif center < i.start:
 *                             right.append(i)
 */

#line 350 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_left, ((PyObject *)__pyx_v_i)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 350, __pyx_L1_error)

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":349
 *                         if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
 *                             contained.append(i)
 *                         elif center > i.end:             # <<<<<<<<<<<<<<
 *                             left.append(i)
 *                         elif center < i.start:
 */

#line 349 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            goto __pyx_L13;

#line 349 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":351
 *                         elif center > i.end:
 *                             left.append(i)
 *                         elif center < i.start:             # <<<<<<<<<<<<<<
 *                             right.append(i)
 *                         else:
 */

#line 351 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = ((__pyx_v_center < __pyx_v_i->start) != 0);

#line 351 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (__pyx_t_4) {

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":352
 *                             left.append(i)
 *                         elif center < i.start:
 *                             right.append(i)             # <<<<<<<<<<<<<<
 *                         else:
 *                             contained.append(i)
 */

#line 352 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_right, ((PyObject *)__pyx_v_i)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 352, __pyx_L1_error)

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":351
 *                         elif center > i.end:
 *                             left.append(i)
 *                         elif center < i.start:             # <<<<<<<<<<<<<<
 *                             right.append(i)
 *                         else:
 */

#line 351 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            goto __pyx_L13;

#line 351 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":354
 *                             right.append(i)
 *                         else:
 *                             contained.append(i)             # <<<<<<<<<<<<<<
 *                 else:
 *                     contained = list(members)
 */

#line 354 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          /*else*/ {

#line 354 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_contained, ((PyObject *)__pyx_v_i)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 354, __pyx_L1_error)

#line 354 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 354 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_L13:;

#line 354 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":343
 *                 contained = []
 * 
 *                 if n > 20:             # <<<<<<<<<<<<<<
 *                     j = 0
 *                     for j in range(n):
 */

#line 343 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L10;

#line 343 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":356
 *                             contained.append(i)
 *                 else:
 *                     contained = list(members)             # <<<<<<<<<<<<<<
 * 
 *                 # If this node itself contains nothing
 */

#line 356 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      /*else*/ {

#line 356 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = PySequence_List(__pyx_v_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 356, __pyx_L1_error)

#line 356 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 356 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_v_contained, ((PyObject*)__pyx_t_2));

#line 356 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = 0;

#line 356 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 356 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_L10:;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":359
 * 
 *                 # If this node itself contains nothing
 *                 if not contained:             # <<<<<<<<<<<<<<
 *                     # If only one child contains something
 *                     if not (left and right):
 */

#line 359 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = (PyList_GET_SIZE(__pyx_v_contained) != 0);

#line 359 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_3 = ((!__pyx_t_4) != 0);

#line 359 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_3) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":361
 *                 if not contained:
 *                     # If only one child contains something
 *                     if not (left and right):             # <<<<<<<<<<<<<<
 *                         # store that side's elements on this node instead of
 *                         # creating an extra layer
 */

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_4 = (PyList_GET_SIZE(__pyx_v_left) != 0);

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (__pyx_t_4) {

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        } else {

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_3 = __pyx_t_4;

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          goto __pyx_L18_bool_binop_done;

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_4 = (PyList_GET_SIZE(__pyx_v_right) != 0);

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = __pyx_t_4;

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_L18_bool_binop_done:;

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_4 = ((!__pyx_t_3) != 0);

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (__pyx_t_4) {

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":364
 *                         # store that side's elements on this node instead of
 *                         # creating an extra layer
 *                         contained = left + right             # <<<<<<<<<<<<<<
 *                         left = []
 *                         right = []
 */

#line 364 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = PyNumber_Add(__pyx_v_left, __pyx_v_right); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)

#line 364 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 364 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF_SET(__pyx_v_contained, ((PyObject*)__pyx_t_2));

#line 364 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = 0;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":365
 *                         # creating an extra layer
 *                         contained = left + right
 *                         left = []             # <<<<<<<<<<<<<<
 *                         right = []
 * 
 */

#line 365 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)

#line 365 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 365 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF_SET(__pyx_v_left, ((PyObject*)__pyx_t_2));

#line 365 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = 0;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":366
 *                         contained = left + right
 *                         left = []
 *                         right = []             # <<<<<<<<<<<<<<
 * 
 *                 node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)
 */

#line 366 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 366, __pyx_L1_error)

#line 366 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 366 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF_SET(__pyx_v_right, ((PyObject*)__pyx_t_2));

#line 366 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = 0;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":361
 *                 if not contained:
 *                     # If only one child contains something
 *                     if not (left and right):             # <<<<<<<<<<<<<<
 *                         # store that side's elements on this node instead of
 *                         # creating an extra layer
 */

#line 361 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":359
 * 
 *                 # If this node itself contains nothing
 *                 if not contained:             # <<<<<<<<<<<<<<
 *                     # If only one child contains something
 *                     if not (left and right):
 */

#line 359 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":368
 *                         right = []
 * 
 *                 node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)             # <<<<<<<<<<<<<<
 *                 if side == 'left':
 *                     parent.left = node
 */

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = PyFloat_FromDouble(__pyx_v_center); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GIVEREF(__pyx_t_2);

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = 0;

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_left, Py_None) < 0) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_contained, __pyx_v_contained) < 0) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_right, Py_None) < 0) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = __Pyx_PyInt_From_long((__pyx_v_parent->level + 1)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_11);

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_level, __pyx_t_11) < 0) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_parent, ((PyObject *)__pyx_v_parent)) < 0) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = __Pyx_PyObject_Call(((PyObject *)__pyx_v_cls), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_11);

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!(likely(((__pyx_t_11) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_11, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_11));

#line 368 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":369
 * 
 *                 node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)
 *                 if side == 'left':             # <<<<<<<<<<<<<<
 *                     parent.left = node
 *                     parent.start = min(node.start, parent.start)
 */

#line 369 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_side, __pyx_n_s_left, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 369, __pyx_L1_error)

#line 369 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_3 = (__pyx_t_4 != 0);

#line 369 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_3) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":370
 *                 node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)
 *                 if side == 'left':
 *                     parent.left = node             # <<<<<<<<<<<<<<
 *                     parent.start = min(node.start, parent.start)
 *                     up = parent.parent
 */

#line 370 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(((PyObject *)__pyx_v_node));

#line 370 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GIVEREF(((PyObject *)__pyx_v_node));

#line 370 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_v_parent->left);

#line 370 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(((PyObject *)__pyx_v_parent->left));

#line 370 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_v_parent->left = __pyx_v_node;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":371
 *                 if side == 'left':
 *                     parent.left = node
 *                     parent.start = min(node.start, parent.start)             # <<<<<<<<<<<<<<
 *                     up = parent.parent
 *                     while up is not None:
 */

#line 371 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_12 = __pyx_v_parent->start;

#line 371 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_13 = __pyx_v_node->start;

#line 371 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (((__pyx_t_12 < __pyx_t_13) != 0)) {

#line 371 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_14 = __pyx_t_12;

#line 371 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        } else {

#line 371 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_14 = __pyx_t_13;

#line 371 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 371 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_v_parent->start = __pyx_t_14;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":372
 *                     parent.left = node
 *                     parent.start = min(node.start, parent.start)
 *                     up = parent.parent             # <<<<<<<<<<<<<<
 *                     while up is not None:
 *                         if up.start > node.start:
 */

#line 372 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_11 = ((PyObject *)__pyx_v_parent->parent);

#line 372 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_11);

#line 372 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF_SET(__pyx_v_up, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_11));

#line 372 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_11 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":373
 *                     parent.start = min(node.start, parent.start)
 *                     up = parent.parent
 *                     while up is not None:             # <<<<<<<<<<<<<<
 *                         if up.start > node.start:
 *                             up.start = node.start
 */

#line 373 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        while (1) {

#line 373 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_3 = (((PyObject *)__pyx_v_up) != Py_None);

#line 373 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = (__pyx_t_3 != 0);

#line 373 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (!__pyx_t_4) break;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":374
 *                     up = parent.parent
 *                     while up is not None:
 *                         if up.start > node.start:             # <<<<<<<<<<<<<<
 *                             up.start = node.start
 *                         up = up.parent
 */

#line 374 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = ((__pyx_v_up->start > __pyx_v_node->start) != 0);

#line 374 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (__pyx_t_4) {

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":375
 *                     while up is not None:
 *                         if up.start > node.start:
 *                             up.start = node.start             # <<<<<<<<<<<<<<
 *                         up = up.parent
 *                 elif side == 'right':
 */

#line 375 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_t_14 = __pyx_v_node->start;

#line 375 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_v_up->start = __pyx_t_14;

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":374
 *                     up = parent.parent
 *                     while up is not None:
 *                         if up.start > node.start:             # <<<<<<<<<<<<<<
 *                             up.start = node.start
 *                         up = up.parent
 */

#line 374 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":376
 *                         if up.start > node.start:
 *                             up.start = node.start
 *                         up = up.parent             # <<<<<<<<<<<<<<
 *                 elif side == 'right':
 *                     parent.right = node
 */

#line 376 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_11 = ((PyObject *)__pyx_v_up->parent);

#line 376 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_INCREF(__pyx_t_11);

#line 376 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF_SET(__pyx_v_up, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_11));

#line 376 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_11 = 0;

#line 376 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":369
 * 
 *                 node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)
 *                 if side == 'left':             # <<<<<<<<<<<<<<
 *                     parent.left = node
 *                     parent.start = min(node.start, parent.start)
 */

#line 369 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L20;

#line 369 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":377
 *                             up.start = node.start
 *                         up = up.parent
 *                 elif side == 'right':             # <<<<<<<<<<<<<<
 *                     parent.right = node
 *                     parent.end = max(node.end, parent.end)
 */

#line 377 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_side, __pyx_n_s_right, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 377, __pyx_L1_error)

#line 377 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_3 = (__pyx_t_4 != 0);

#line 377 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (likely(__pyx_t_3)) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":378
 *                         up = up.parent
 *                 elif side == 'right':
 *                     parent.right = node             # <<<<<<<<<<<<<<
 *                     parent.end = max(node.end, parent.end)
 *                     up = parent.parent
 */

#line 378 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(((PyObject *)__pyx_v_node));

#line 378 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GIVEREF(((PyObject *)__pyx_v_node));

#line 378 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_v_parent->right);

#line 378 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(((PyObject *)__pyx_v_parent->right));

#line 378 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_v_parent->right = __pyx_v_node;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":379
 *                 elif side == 'right':
 *                     parent.right = node
 *                     parent.end = max(node.end, parent.end)             # <<<<<<<<<<<<<<
 *                     up = parent.parent
 *                     while up is not None:
 */

#line 379 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_14 = __pyx_v_parent->end;

#line 379 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_12 = __pyx_v_node->end;

#line 379 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (((__pyx_t_14 > __pyx_t_12) != 0)) {

#line 379 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_13 = __pyx_t_14;

#line 379 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        } else {

#line 379 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_13 = __pyx_t_12;

#line 379 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 379 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_v_parent->end = __pyx_t_13;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":380
 *                     parent.right = node
 *                     parent.end = max(node.end, parent.end)
 *                     up = parent.parent             # <<<<<<<<<<<<<<
 *                     while up is not None:
 *                         if up.end < node.end:
 */

#line 380 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_11 = ((PyObject *)__pyx_v_parent->parent);

#line 380 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_11);

#line 380 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF_SET(__pyx_v_up, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_11));

#line 380 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_11 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":381
 *                     parent.end = max(node.end, parent.end)
 *                     up = parent.parent
 *                     while up is not None:             # <<<<<<<<<<<<<<
 *                         if up.end < node.end:
 *                             up.end = node.end
 */

#line 381 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        while (1) {

#line 381 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_3 = (((PyObject *)__pyx_v_up) != Py_None);

#line 381 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = (__pyx_t_3 != 0);

#line 381 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (!__pyx_t_4) break;

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":382
 *                     up = parent.parent
 *                     while up is not None:
 *                         if up.end < node.end:             # <<<<<<<<<<<<<<
 *                             up.end = node.end
 *                         up = up.parent
 */

#line 382 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = ((__pyx_v_up->end < __pyx_v_node->end) != 0);

#line 382 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (__pyx_t_4) {

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":383
 *                     while up is not None:
 *                         if up.end < node.end:
 *                             up.end = node.end             # <<<<<<<<<<<<<<
 *                         up = up.parent
 *                 else:
 */

#line 383 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_t_13 = __pyx_v_node->end;

#line 383 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_v_up->end = __pyx_t_13;

            /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":382
 *                     up = parent.parent
 *                     while up is not None:
 *                         if up.end < node.end:             # <<<<<<<<<<<<<<
 *                             up.end = node.end
 *                         up = up.parent
 */

#line 382 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":384
 *                         if up.end < node.end:
 *                             up.end = node.end
 *                         up = up.parent             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(side)
 */

#line 384 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_11 = ((PyObject *)__pyx_v_up->parent);

#line 384 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_INCREF(__pyx_t_11);

#line 384 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF_SET(__pyx_v_up, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_11));

#line 384 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_11 = 0;

#line 384 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":377
 *                             up.start = node.start
 *                         up = up.parent
 *                 elif side == 'right':             # <<<<<<<<<<<<<<
 *                     parent.right = node
 *                     parent.end = max(node.end, parent.end)
 */

#line 377 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L20;

#line 377 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":386
 *                         up = up.parent
 *                 else:
 *                     raise ValueError(side)             # <<<<<<<<<<<<<<
 *                 if left:
 *                     stack.append((node, left, "left"))
 */

#line 386 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      /*else*/ {

#line 386 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_side); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 386, __pyx_L1_error)

#line 386 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_11);

#line 386 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);

#line 386 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 386 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __PYX_ERR(0, 386, __pyx_L1_error)

#line 386 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 386 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_L20:;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":387
 *                 else:
 *                     raise ValueError(side)
 *                 if left:             # <<<<<<<<<<<<<<
 *                     stack.append((node, left, "left"))
 *                 if right:
 */

#line 387 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = (PyList_GET_SIZE(__pyx_v_left) != 0);

#line 387 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_4) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":388
 *                     raise ValueError(side)
 *                 if left:
 *                     stack.append((node, left, "left"))             # <<<<<<<<<<<<<<
 *                 if right:
 *                     stack.append((node, right, "right"))
 */

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 388, __pyx_L1_error)

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_11);

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(((PyObject *)__pyx_v_node));

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GIVEREF(((PyObject *)__pyx_v_node));

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyTuple_SET_ITEM(__pyx_t_11, 0, ((PyObject *)__pyx_v_node));

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_v_left);

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GIVEREF(__pyx_v_left);

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_left);

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_n_s_left);

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GIVEREF(__pyx_n_s_left);

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_n_s_left);

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_stack, __pyx_t_11); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 388, __pyx_L1_error)

#line 388 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":387
 *                 else:
 *                     raise ValueError(side)
 *                 if left:             # <<<<<<<<<<<<<<
 *                     stack.append((node, left, "left"))
 *                 if right:
 */

#line 387 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":389
 *                 if left:
 *                     stack.append((node, left, "left"))
 *                 if right:             # <<<<<<<<<<<<<<
 *                     stack.append((node, right, "right"))
 *         return root.left
 */

#line 389 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = (PyList_GET_SIZE(__pyx_v_right) != 0);

#line 389 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_4) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":390
 *                     stack.append((node, left, "left"))
 *                 if right:
 *                     stack.append((node, right, "right"))             # <<<<<<<<<<<<<<
 *         return root.left
 * 
 */

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 390, __pyx_L1_error)

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_11);

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(((PyObject *)__pyx_v_node));

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GIVEREF(((PyObject *)__pyx_v_node));

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyTuple_SET_ITEM(__pyx_t_11, 0, ((PyObject *)__pyx_v_node));

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_v_right);

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GIVEREF(__pyx_v_right);

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_right);

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_n_s_right);

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GIVEREF(__pyx_n_s_right);

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_n_s_right);

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_stack, __pyx_t_11); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 390, __pyx_L1_error)

#line 390 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":389
 *                 if left:
 *                     stack.append((node, left, "left"))
 *                 if right:             # <<<<<<<<<<<<<<
 *                     stack.append((node, right, "right"))
 *         return root.left
 */

#line 389 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":328
 *             side = <str>PyTuple_GetItem(bunch, 2)
 *             n = PyList_Size(members)
 *             if n > 0:             # <<<<<<<<<<<<<<
 *                 accumulator = 0.
 *                 j = 0
 */

#line 328 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 328 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":391
 *                 if right:
 *                     stack.append((node, right, "right"))
 *         return root.left             # <<<<<<<<<<<<<<
 * 
 *     def _update_bounds(self):
 */

#line 391 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 391 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_root->left));

#line 391 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = ((PyObject *)__pyx_v_root->left);

#line 391 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":304
 * 
 *     @classmethod
 *     def build(cls, intervals):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalTreeNode root, parent, node, up
 */

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_11);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.build", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_root);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_parent);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_up);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_stack);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_members);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_left);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_right);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_contained);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_intervals_);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_bunch);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_side);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_i);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":393
 *         return root.left
 * 
 *     def _update_bounds(self):             # <<<<<<<<<<<<<<
 *         changed = False
 *         if self.left is not None and self.left.start < self.start:
 */

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_19_update_bounds(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_18_update_bounds[] = "IntervalTreeNode._update_bounds(self)";

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_19_update_bounds(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_update_bounds (wrapper)", 0);
  __pyx_r = 
#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_18_update_bounds(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_18_update_bounds(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  int __pyx_v_changed
#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_1;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_2;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_3;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_4;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_update_bounds", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":394
 * 
 *     def _update_bounds(self):
 *         changed = False             # <<<<<<<<<<<<<<
 *         if self.left is not None and self.left.start < self.start:
 *             changed = True
 */

#line 394 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_changed = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":395
 *     def _update_bounds(self):
 *         changed = False
 *         if self.left is not None and self.left.start < self.start:             # <<<<<<<<<<<<<<
 *             changed = True
 *             self.start = self.left.start
 */

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = (((PyObject *)__pyx_v_self->left) != Py_None);

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_3) {

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = __pyx_t_3;

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L4_bool_binop_done;

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = ((__pyx_v_self->left->start < __pyx_v_self->start) != 0);

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_t_3;

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_bool_binop_done:;

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":396
 *         changed = False
 *         if self.left is not None and self.left.start < self.start:
 *             changed = True             # <<<<<<<<<<<<<<
 *             self.start = self.left.start
 *         if self.right is not None and self.right.end > self.end:
 */

#line 396 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_changed = 1;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":397
 *         if self.left is not None and self.left.start < self.start:
 *             changed = True
 *             self.start = self.left.start             # <<<<<<<<<<<<<<
 *         if self.right is not None and self.right.end > self.end:
 *             changed = True
 */

#line 397 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = __pyx_v_self->left->start;

#line 397 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_self->start = __pyx_t_4;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":395
 *     def _update_bounds(self):
 *         changed = False
 *         if self.left is not None and self.left.start < self.start:             # <<<<<<<<<<<<<<
 *             changed = True
 *             self.start = self.left.start
 */

#line 395 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":398
 *             changed = True
 *             self.start = self.left.start
 *         if self.right is not None and self.right.end > self.end:             # <<<<<<<<<<<<<<
 *             changed = True
 *             self.end = self.right.end
 */

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = (((PyObject *)__pyx_v_self->right) != Py_None);

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_2) {

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = __pyx_t_2;

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L7_bool_binop_done;

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = ((__pyx_v_self->right->end > __pyx_v_self->end) != 0);

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_t_2;

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L7_bool_binop_done:;

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":399
 *             self.start = self.left.start
 *         if self.right is not None and self.right.end > self.end:
 *             changed = True             # <<<<<<<<<<<<<<
 *             self.end = self.right.end
 *         if self.parent is not None:
 */

#line 399 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_changed = 1;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":400
 *         if self.right is not None and self.right.end > self.end:
 *             changed = True
 *             self.end = self.right.end             # <<<<<<<<<<<<<<
 *         if self.parent is not None:
 *             self.parent._update_bounds()
 */

#line 400 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = __pyx_v_self->right->end;

#line 400 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_self->end = __pyx_t_4;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":398
 *             changed = True
 *             self.start = self.left.start
 *         if self.right is not None and self.right.end > self.end:             # <<<<<<<<<<<<<<
 *             changed = True
 *             self.end = self.right.end
 */

#line 398 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":401
 *             changed = True
 *             self.end = self.right.end
 *         if self.parent is not None:             # <<<<<<<<<<<<<<
 *             self.parent._update_bounds()
 *         return changed
 */

#line 401 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (((PyObject *)__pyx_v_self->parent) != Py_None);

#line 401 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 401 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":402
 *             self.end = self.right.end
 *         if self.parent is not None:
 *             self.parent._update_bounds()             # <<<<<<<<<<<<<<
 *         return changed
 * 
 */

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->parent), __pyx_n_s_update_bounds); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 402, __pyx_L1_error)

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = NULL;

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (likely(__pyx_t_7)) {

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_7);

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(function);

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_t_6, function);

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 402, __pyx_L1_error)

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 402 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":401
 *             changed = True
 *             self.end = self.right.end
 *         if self.parent is not None:             # <<<<<<<<<<<<<<
 *             self.parent._update_bounds()
 *         return changed
 */

#line 401 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":403
 *         if self.parent is not None:
 *             self.parent._update_bounds()
 *         return changed             # <<<<<<<<<<<<<<
 * 
 *     def insert(self, interval):
 */

#line 403 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 403 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_changed); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 403, __pyx_L1_error)

#line 403 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 403 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_5;

#line 403 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = 0;

#line 403 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":393
 *         return root.left
 * 
 *     def _update_bounds(self):             # <<<<<<<<<<<<<<
 *         changed = False
 *         if self.left is not None and self.left.start < self.start:
 */

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode._update_bounds", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 393 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":405
 *         return changed
 * 
 *     def insert(self, interval):             # <<<<<<<<<<<<<<
 *         insert_in_self = False
 *         if self.node_contains(interval.start):
 */

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_21insert(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_20insert[] = "IntervalTreeNode.insert(self, interval)";

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_21insert(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  __pyx_r = 
#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_20insert(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_interval));

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_20insert(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_interval) {
  int __pyx_v_insert_in_self
#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  int __pyx_v_changed
#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_5;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_8;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_9;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("insert", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":406
 * 
 *     def insert(self, interval):
 *         insert_in_self = False             # <<<<<<<<<<<<<<
 *         if self.node_contains(interval.start):
 *             if self.left is not None and self.left.node_contains(interval.end):
 */

#line 406 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_insert_in_self = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":407
 *     def insert(self, interval):
 *         insert_in_self = False
 *         if self.node_contains(interval.start):             # <<<<<<<<<<<<<<
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)
 */

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 407, __pyx_L1_error)

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 407, __pyx_L1_error)

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = NULL;

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (likely(__pyx_t_4)) {

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(function);

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 407, __pyx_L1_error)

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":408
 *         insert_in_self = False
 *         if self.node_contains(interval.start):
 *             if self.left is not None and self.left.node_contains(interval.end):             # <<<<<<<<<<<<<<
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and
 */

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = (((PyObject *)__pyx_v_self->left) != Py_None);

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (__pyx_t_6 != 0);

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_7) {

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = __pyx_t_7;

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_bool_binop_done;

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 408, __pyx_L1_error)

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 408, __pyx_L1_error)

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = NULL;

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (likely(__pyx_t_4)) {

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_4);

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(function);

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_t_2, function);

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 408, __pyx_L1_error)

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_7;

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L5_bool_binop_done:;

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":409
 *         if self.node_contains(interval.start):
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)             # <<<<<<<<<<<<<<
 *             elif (self.right is not None and self.right.node_contains(interval.end) and
 *                   self.right.node_contains(interval.start)):
 */

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left), __pyx_n_s_insert); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 409, __pyx_L1_error)

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_3 = NULL;

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely(__pyx_t_3)) {

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_INCREF(__pyx_t_3);

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_INCREF(function);

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF_SET(__pyx_t_2, function);

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_interval) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_interval);

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_r = __pyx_t_1;

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = 0;

#line 409 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":408
 *         insert_in_self = False
 *         if self.node_contains(interval.start):
 *             if self.left is not None and self.left.node_contains(interval.end):             # <<<<<<<<<<<<<<
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and
 */

#line 408 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":410
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and             # <<<<<<<<<<<<<<
 *                   self.right.node_contains(interval.start)):
 *                 return self.right.insert(interval)
 */

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((PyObject *)__pyx_v_self->right) != Py_None);

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = (__pyx_t_7 != 0);

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_6) {

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = __pyx_t_6;

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L7_bool_binop_done;

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 410, __pyx_L1_error)

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = NULL;

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (likely(__pyx_t_4)) {

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_4);

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(function);

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_t_2, function);

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 410, __pyx_L1_error)

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_6) {

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = __pyx_t_6;

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L7_bool_binop_done;

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":411
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and
 *                   self.right.node_contains(interval.start)):             # <<<<<<<<<<<<<<
 *                 return self.right.insert(interval)
 *             else:
 */

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 411, __pyx_L1_error)

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 411, __pyx_L1_error)

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = NULL;

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (likely(__pyx_t_4)) {

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_4);

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(function);

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_t_2, function);

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 411, __pyx_L1_error)

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 411 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L7_bool_binop_done:;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":410
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and             # <<<<<<<<<<<<<<
 *                   self.right.node_contains(interval.start)):
 *                 return self.right.insert(interval)
 */

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":412
 *             elif (self.right is not None and self.right.node_contains(interval.end) and
 *                   self.right.node_contains(interval.start)):
 *                 return self.right.insert(interval)             # <<<<<<<<<<<<<<
 *             else:
 *                 insert_in_self = True
 */

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right), __pyx_n_s_insert); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 412, __pyx_L1_error)

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_3 = NULL;

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely(__pyx_t_3)) {

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_INCREF(__pyx_t_3);

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_INCREF(function);

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF_SET(__pyx_t_2, function);

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_interval) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_interval);

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_r = __pyx_t_1;

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = 0;

#line 412 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":410
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)
 *             elif (self.right is not None and self.right.node_contains(interval.end) and             # <<<<<<<<<<<<<<
 *                   self.right.node_contains(interval.start)):
 *                 return self.right.insert(interval)
 */

#line 410 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":414
 *                 return self.right.insert(interval)
 *             else:
 *                 insert_in_self = True             # <<<<<<<<<<<<<<
 *         elif self.node_contains(interval.end):
 *             if self.right is not None and self.right.node_contains(interval.start):
 */

#line 414 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    /*else*/ {

#line 414 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_insert_in_self = 1;

#line 414 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":407
 *     def insert(self, interval):
 *         insert_in_self = False
 *         if self.node_contains(interval.start):             # <<<<<<<<<<<<<<
 *             if self.left is not None and self.left.node_contains(interval.end):
 *                 return self.left.insert(interval)
 */

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L3;

#line 407 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":415
 *             else:
 *                 insert_in_self = True
 *         elif self.node_contains(interval.end):             # <<<<<<<<<<<<<<
 *             if self.right is not None and self.right.node_contains(interval.start):
 *                 return self.right.insert(interval)
 */

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L1_error)

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 415, __pyx_L1_error)

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = NULL;

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (likely(__pyx_t_4)) {

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(function);

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 415, __pyx_L1_error)

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":416
 *                 insert_in_self = True
 *         elif self.node_contains(interval.end):
 *             if self.right is not None and self.right.node_contains(interval.start):             # <<<<<<<<<<<<<<
 *                 return self.right.insert(interval)
 *             else:
 */

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = (((PyObject *)__pyx_v_self->right) != Py_None);

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (__pyx_t_6 != 0);

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_7) {

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = __pyx_t_7;

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L11_bool_binop_done;

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right), __pyx_n_s_node_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L1_error)

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = NULL;

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (likely(__pyx_t_4)) {

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_4);

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(function);

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_t_2, function);

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 416, __pyx_L1_error)

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_7;

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L11_bool_binop_done:;

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":417
 *         elif self.node_contains(interval.end):
 *             if self.right is not None and self.right.node_contains(interval.start):
 *                 return self.right.insert(interval)             # <<<<<<<<<<<<<<
 *             else:
 *                 insert_in_self = True
 */

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right), __pyx_n_s_insert); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 417, __pyx_L1_error)

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_3 = NULL;

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely(__pyx_t_3)) {

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_INCREF(__pyx_t_3);

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_INCREF(function);

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF_SET(__pyx_t_2, function);

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_interval) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_interval);

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_r = __pyx_t_1;

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = 0;

#line 417 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":416
 *                 insert_in_self = True
 *         elif self.node_contains(interval.end):
 *             if self.right is not None and self.right.node_contains(interval.start):             # <<<<<<<<<<<<<<
 *                 return self.right.insert(interval)
 *             else:
 */

#line 416 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":419
 *                 return self.right.insert(interval)
 *             else:
 *                 insert_in_self = True             # <<<<<<<<<<<<<<
 *         if not insert_in_self and self.parent is None:
 *             insert_in_self = True
 */

#line 419 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    /*else*/ {

#line 419 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_insert_in_self = 1;

#line 419 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":415
 *             else:
 *                 insert_in_self = True
 *         elif self.node_contains(interval.end):             # <<<<<<<<<<<<<<
 *             if self.right is not None and self.right.node_contains(interval.start):
 *                 return self.right.insert(interval)
 */

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 415 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":420
 *             else:
 *                 insert_in_self = True
 *         if not insert_in_self and self.parent is None:             # <<<<<<<<<<<<<<
 *             insert_in_self = True
 *         if insert_in_self:
 */

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = ((!(__pyx_v_insert_in_self != 0)) != 0);

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_7) {

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  } else {

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_7;

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L14_bool_binop_done;

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = (((PyObject *)__pyx_v_self->parent) == Py_None);

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_6 = (__pyx_t_7 != 0);

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = __pyx_t_6;

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L14_bool_binop_done:;

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":421
 *                 insert_in_self = True
 *         if not insert_in_self and self.parent is None:
 *             insert_in_self = True             # <<<<<<<<<<<<<<
 *         if insert_in_self:
 *             self.contained.append(interval)
 */

#line 421 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_insert_in_self = 1;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":420
 *             else:
 *                 insert_in_self = True
 *         if not insert_in_self and self.parent is None:             # <<<<<<<<<<<<<<
 *             insert_in_self = True
 *         if insert_in_self:
 */

#line 420 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":422
 *         if not insert_in_self and self.parent is None:
 *             insert_in_self = True
 *         if insert_in_self:             # <<<<<<<<<<<<<<
 *             self.contained.append(interval)
 *             changed = False
 */

#line 422 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = (__pyx_v_insert_in_self != 0);

#line 422 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":423
 *             insert_in_self = True
 *         if insert_in_self:
 *             self.contained.append(interval)             # <<<<<<<<<<<<<<
 *             changed = False
 *             if interval.start < self.start:
 */

#line 423 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_v_self->contained == Py_None)) {

#line 423 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");

#line 423 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __PYX_ERR(0, 423, __pyx_L1_error)

#line 423 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 423 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_self->contained, __pyx_v_interval); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 423, __pyx_L1_error)

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":424
 *         if insert_in_self:
 *             self.contained.append(interval)
 *             changed = False             # <<<<<<<<<<<<<<
 *             if interval.start < self.start:
 *                 self.start = interval.start
 */

#line 424 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_changed = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":425
 *             self.contained.append(interval)
 *             changed = False
 *             if interval.start < self.start:             # <<<<<<<<<<<<<<
 *                 self.start = interval.start
 *                 changed = True
 */

#line 425 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 425, __pyx_L1_error)

#line 425 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 425 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 425, __pyx_L1_error)

#line 425 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 425 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 425, __pyx_L1_error)

#line 425 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 425 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 425 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 425, __pyx_L1_error)

#line 425 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 425 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":426
 *             changed = False
 *             if interval.start < self.start:
 *                 self.start = interval.start             # <<<<<<<<<<<<<<
 *                 changed = True
 *             if interval.end > self.end:
 */

#line 426 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 426, __pyx_L1_error)

#line 426 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_3);

#line 426 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 426, __pyx_L1_error)

#line 426 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 426 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_self->start = __pyx_t_9;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":427
 *             if interval.start < self.start:
 *                 self.start = interval.start
 *                 changed = True             # <<<<<<<<<<<<<<
 *             if interval.end > self.end:
 *                 self.end = interval.end
 */

#line 427 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_changed = 1;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":425
 *             self.contained.append(interval)
 *             changed = False
 *             if interval.start < self.start:             # <<<<<<<<<<<<<<
 *                 self.start = interval.start
 *                 changed = True
 */

#line 425 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":428
 *                 self.start = interval.start
 *                 changed = True
 *             if interval.end > self.end:             # <<<<<<<<<<<<<<
 *                 self.end = interval.end
 *                 changed = True
 */

#line 428 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 428, __pyx_L1_error)

#line 428 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 428 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 428, __pyx_L1_error)

#line 428 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 428 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)

#line 428 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 428 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 428 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 428, __pyx_L1_error)

#line 428 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 428 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":429
 *                 changed = True
 *             if interval.end > self.end:
 *                 self.end = interval.end             # <<<<<<<<<<<<<<
 *                 changed = True
 *             if changed:
 */

#line 429 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)

#line 429 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 429 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 429, __pyx_L1_error)

#line 429 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 429 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_self->end = __pyx_t_9;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":430
 *             if interval.end > self.end:
 *                 self.end = interval.end
 *                 changed = True             # <<<<<<<<<<<<<<
 *             if changed:
 *                 self._update_bounds()
 */

#line 430 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_changed = 1;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":428
 *                 self.start = interval.start
 *                 changed = True
 *             if interval.end > self.end:             # <<<<<<<<<<<<<<
 *                 self.end = interval.end
 *                 changed = True
 */

#line 428 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":431
 *                 self.end = interval.end
 *                 changed = True
 *             if changed:             # <<<<<<<<<<<<<<
 *                 self._update_bounds()
 * 
 */

#line 431 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = (__pyx_v_changed != 0);

#line 431 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":432
 *                 changed = True
 *             if changed:
 *                 self._update_bounds()             # <<<<<<<<<<<<<<
 * 
 *     def flatten(self):
 */

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update_bounds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_3 = NULL;

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely(__pyx_t_3)) {

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_INCREF(__pyx_t_3);

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_INCREF(function);

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF_SET(__pyx_t_2, function);

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 432 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":431
 *                 self.end = interval.end
 *                 changed = True
 *             if changed:             # <<<<<<<<<<<<<<
 *                 self._update_bounds()
 * 
 */

#line 431 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":422
 *         if not insert_in_self and self.parent is None:
 *             insert_in_self = True
 *         if insert_in_self:             # <<<<<<<<<<<<<<
 *             self.contained.append(interval)
 *             changed = False
 */

#line 422 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":405
 *         return changed
 * 
 *     def insert(self, interval):             # <<<<<<<<<<<<<<
 *         insert_in_self = False
 *         if self.node_contains(interval.start):
 */

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 405 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":434
 *                 self._update_bounds()
 * 
 *     def flatten(self):             # <<<<<<<<<<<<<<
 *         # perform an infix traversal of the tree and collect
 *         # the elements
 */

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_23flatten(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_22flatten[] = "IntervalTreeNode.flatten(self)";

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_23flatten(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("flatten (wrapper)", 0);
  __pyx_r = 
#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_22flatten(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_22flatten(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_v_items = NULL
#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_stack = NULL
#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_current = NULL
#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  int __pyx_v_done
#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_5;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("flatten", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":437
 *         # perform an infix traversal of the tree and collect
 *         # the elements
 *         items = []             # <<<<<<<<<<<<<<
 *         stack = deque([])
 *         current = self
 */

#line 437 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L1_error)

#line 437 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 437 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_items = ((PyObject*)__pyx_t_1);

#line 437 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":438
 *         # the elements
 *         items = []
 *         stack = deque([])             # <<<<<<<<<<<<<<
 *         current = self
 *         done = False
 */

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_deque); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 438, __pyx_L1_error)

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = NULL;

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (likely(__pyx_t_4)) {

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(function);

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 438, __pyx_L1_error)

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_stack = __pyx_t_1;

#line 438 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":439
 *         items = []
 *         stack = deque([])
 *         current = self             # <<<<<<<<<<<<<<
 *         done = False
 *         while not done:
 */

#line 439 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 439 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_current = ((PyObject *)__pyx_v_self);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":440
 *         stack = deque([])
 *         current = self
 *         done = False             # <<<<<<<<<<<<<<
 *         while not done:
 *             if current is not None:
 */

#line 440 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_done = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":441
 *         current = self
 *         done = False
 *         while not done:             # <<<<<<<<<<<<<<
 *             if current is not None:
 *                 stack.append(current)
 */

#line 441 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  while (1) {

#line 441 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = ((!(__pyx_v_done != 0)) != 0);

#line 441 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!__pyx_t_5) break;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":442
 *         done = False
 *         while not done:
 *             if current is not None:             # <<<<<<<<<<<<<<
 *                 stack.append(current)
 *                 current = current.left
 */

#line 442 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = (__pyx_v_current != Py_None);

#line 442 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = (__pyx_t_5 != 0);

#line 442 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":443
 *         while not done:
 *             if current is not None:
 *                 stack.append(current)             # <<<<<<<<<<<<<<
 *                 current = current.left
 *             else:
 */

#line 443 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_7 = __Pyx_PyObject_Append(__pyx_v_stack, __pyx_v_current); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 443, __pyx_L1_error)

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":444
 *             if current is not None:
 *                 stack.append(current)
 *                 current = current.left             # <<<<<<<<<<<<<<
 *             else:
 *                 if stack:
 */

#line 444 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_current, __pyx_n_s_left); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error)

#line 444 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 444 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_v_current, __pyx_t_1);

#line 444 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":442
 *         done = False
 *         while not done:
 *             if current is not None:             # <<<<<<<<<<<<<<
 *                 stack.append(current)
 *                 current = current.left
 */

#line 442 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5;

#line 442 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":446
 *                 current = current.left
 *             else:
 *                 if stack:             # <<<<<<<<<<<<<<
 *                     current = stack.pop()
 *                     items.extend(current.contained)
 */

#line 446 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    /*else*/ {

#line 446 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_stack); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 446, __pyx_L1_error)

#line 446 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_6) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":447
 *             else:
 *                 if stack:
 *                     current = stack.pop()             # <<<<<<<<<<<<<<
 *                     items.extend(current.contained)
 *                     current = current.right
 */

#line 447 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = __Pyx_PyObject_Pop(__pyx_v_stack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)

#line 447 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 447 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_v_current, __pyx_t_1);

#line 447 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":448
 *                 if stack:
 *                     current = stack.pop()
 *                     items.extend(current.contained)             # <<<<<<<<<<<<<<
 *                     current = current.right
 *                 else:
 */

#line 448 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_current, __pyx_n_s_contained); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)

#line 448 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 448 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_7 = __Pyx_PyList_Extend(__pyx_v_items, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 448, __pyx_L1_error)

#line 448 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":449
 *                     current = stack.pop()
 *                     items.extend(current.contained)
 *                     current = current.right             # <<<<<<<<<<<<<<
 *                 else:
 *                     done = True
 */

#line 449 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_current, __pyx_n_s_right); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)

#line 449 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 449 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_v_current, __pyx_t_1);

#line 449 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":446
 *                 current = current.left
 *             else:
 *                 if stack:             # <<<<<<<<<<<<<<
 *                     current = stack.pop()
 *                     items.extend(current.contained)
 */

#line 446 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L6;

#line 446 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":451
 *                     current = current.right
 *                 else:
 *                     done = True             # <<<<<<<<<<<<<<
 *         return items
 * 
 */

#line 451 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      /*else*/ {

#line 451 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_v_done = 1;

#line 451 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 451 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_L6:;

#line 451 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 451 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L5:;

#line 451 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":452
 *                 else:
 *                     done = True
 *         return items             # <<<<<<<<<<<<<<
 * 
 *     def balance(self):
 */

#line 452 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 452 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_items);

#line 452 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_v_items;

#line 452 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":434
 *                 self._update_bounds()
 * 
 *     def flatten(self):             # <<<<<<<<<<<<<<
 *         # perform an infix traversal of the tree and collect
 *         # the elements
 */

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.flatten", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_items);

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_stack);

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_current);

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 434 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":454
 *         return items
 * 
 *     def balance(self):             # <<<<<<<<<<<<<<
 *         items = self.flatten()
 *         tree = self.build(items)
 */

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_25balance(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_24balance[] = "IntervalTreeNode.balance(self)";

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_25balance(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("balance (wrapper)", 0);
  __pyx_r = 
#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_24balance(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_24balance(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_v_items = NULL
#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_tree = NULL
#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_4;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("balance", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":455
 * 
 *     def balance(self):
 *         items = self.flatten()             # <<<<<<<<<<<<<<
 *         tree = self.build(items)
 *         self.left = tree.left
 */

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_flatten); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = NULL;

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (likely(__pyx_t_3)) {

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(function);

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_items = __pyx_t_1;

#line 455 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":456
 *     def balance(self):
 *         items = self.flatten()
 *         tree = self.build(items)             # <<<<<<<<<<<<<<
 *         self.left = tree.left
 *         self.right = tree.right
 */

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_build); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L1_error)

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = NULL;

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (likely(__pyx_t_3)) {

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(function);

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_items) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_items);

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_tree = __pyx_t_1;

#line 456 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":457
 *         items = self.flatten()
 *         tree = self.build(items)
 *         self.left = tree.left             # <<<<<<<<<<<<<<
 *         self.right = tree.right
 *         self.contained = tree.contained
 */

#line 457 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_left); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)

#line 457 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 457 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 457, __pyx_L1_error)

#line 457 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 457 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->left);

#line 457 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->left));

#line 457 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 457 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":458
 *         tree = self.build(items)
 *         self.left = tree.left
 *         self.right = tree.right             # <<<<<<<<<<<<<<
 *         self.contained = tree.contained
 *         self.start = tree.start
 */

#line 458 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_right); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)

#line 458 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 458 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 458, __pyx_L1_error)

#line 458 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 458 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->right);

#line 458 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->right));

#line 458 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 458 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":459
 *         self.left = tree.left
 *         self.right = tree.right
 *         self.contained = tree.contained             # <<<<<<<<<<<<<<
 *         self.start = tree.start
 *         self.end = tree.end
 */

#line 459 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_contained); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 459, __pyx_L1_error)

#line 459 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 459 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 459, __pyx_L1_error)

#line 459 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 459 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->contained);

#line 459 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_v_self->contained);

#line 459 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->contained = ((PyObject*)__pyx_t_1);

#line 459 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":460
 *         self.right = tree.right
 *         self.contained = tree.contained
 *         self.start = tree.start             # <<<<<<<<<<<<<<
 *         self.end = tree.end
 *         self.center = tree.center
 */

#line 460 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L1_error)

#line 460 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 460 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 460, __pyx_L1_error)

#line 460 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 460 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->start = __pyx_t_4;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":461
 *         self.contained = tree.contained
 *         self.start = tree.start
 *         self.end = tree.end             # <<<<<<<<<<<<<<
 *         self.center = tree.center
 *         self.left.parent = self
 */

#line 461 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)

#line 461 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 461 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 461, __pyx_L1_error)

#line 461 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 461 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->end = __pyx_t_4;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":462
 *         self.start = tree.start
 *         self.end = tree.end
 *         self.center = tree.center             # <<<<<<<<<<<<<<
 *         self.left.parent = self
 *         self.right.parent = self
 */

#line 462 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tree, __pyx_n_s_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L1_error)

#line 462 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 462 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 462, __pyx_L1_error)

#line 462 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 462 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->center = __pyx_t_4;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":463
 *         self.end = tree.end
 *         self.center = tree.center
 *         self.left.parent = self             # <<<<<<<<<<<<<<
 *         self.right.parent = self
 *         self._update_bounds()
 */

#line 463 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 463 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 463 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->left->parent);

#line 463 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->left->parent));

#line 463 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->left->parent = __pyx_v_self;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":464
 *         self.center = tree.center
 *         self.left.parent = self
 *         self.right.parent = self             # <<<<<<<<<<<<<<
 *         self._update_bounds()
 * 
 */

#line 464 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 464 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 464 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->right->parent);

#line 464 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->right->parent));

#line 464 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->right->parent = __pyx_v_self;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":465
 *         self.left.parent = self
 *         self.right.parent = self
 *         self._update_bounds()             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update_bounds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 465, __pyx_L1_error)

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = NULL;

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (likely(__pyx_t_3)) {

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(function);

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 465 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":454
 *         return items
 * 
 *     def balance(self):             # <<<<<<<<<<<<<<
 *         items = self.flatten()
 *         tree = self.build(items)
 */

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.balance", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_items);

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_tree);

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 454 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":467
 *         self._update_bounds()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         flat = self.flatten()
 *         return iter(flat)
 */

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_27__iter__(PyObject *__pyx_v_self); /*proto*/

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_27__iter__(PyObject *__pyx_v_self) {

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = 
#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_26__iter__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_26__iter__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_v_flat = NULL
#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":468
 * 
 *     def __iter__(self):
 *         flat = self.flatten()             # <<<<<<<<<<<<<<
 *         return iter(flat)
 * 
 */

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_flatten); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 468, __pyx_L1_error)

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = NULL;

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (likely(__pyx_t_3)) {

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(function);

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_flat = __pyx_t_1;

#line 468 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":469
 *     def __iter__(self):
 *         flat = self.flatten()
 *         return iter(flat)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 469 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 469 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyObject_GetIter(__pyx_v_flat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)

#line 469 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 469 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 469 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 469 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":467
 *         self._update_bounds()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         flat = self.flatten()
 *         return iter(flat)
 */

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_flat);

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 467 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":30
 * cdef class IntervalTreeNode(object):
 *     cdef:
 *         public double center             # <<<<<<<<<<<<<<
 *         public double start
 *         public double end
 */

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_1__get__(PyObject *__pyx_v_self) {

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 30, __pyx_L1_error)

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_t_1;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = NULL;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double __pyx_t_1;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 30, __pyx_L1_error)

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->center = __pyx_t_1;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.center.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 30 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":31
 *     cdef:
 *         public double center
 *         public double start             # <<<<<<<<<<<<<<
 *         public double end
 *         public IntervalTreeNode left
 */

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_1__get__(PyObject *__pyx_v_self) {

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 31, __pyx_L1_error)

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_t_1;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.start.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = NULL;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double __pyx_t_1;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 31, __pyx_L1_error)

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->start = __pyx_t_1;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.start.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 31 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":32
 *         public double center
 *         public double start
 *         public double end             # <<<<<<<<<<<<<<
 *         public IntervalTreeNode left
 *         public IntervalTreeNode right
 */

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_1__get__(PyObject *__pyx_v_self) {

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 32, __pyx_L1_error)

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_t_1;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.end.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = NULL;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double __pyx_t_1;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 32, __pyx_L1_error)

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->end = __pyx_t_1;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.end.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 32 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":33
 *         public double start
 *         public double end
 *         public IntervalTreeNode left             # <<<<<<<<<<<<<<
 *         public IntervalTreeNode right
 *         public list contained
 */

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_1__get__(PyObject *__pyx_v_self) {

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left));

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->left);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(1, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->left);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->left));

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.left.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_5__del__(PyObject *__pyx_v_self) {

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(Py_None);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(Py_None);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->left);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->left));

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None);

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 33 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":34
 *         public double end
 *         public IntervalTreeNode left
 *         public IntervalTreeNode right             # <<<<<<<<<<<<<<
 *         public list contained
 *         public int level
 */

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_1__get__(PyObject *__pyx_v_self) {

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right));

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->right);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(1, 34, __pyx_L1_error)

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->right);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->right));

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.right.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_5__del__(PyObject *__pyx_v_self) {

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(Py_None);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(Py_None);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->right);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->right));

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None);

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 34 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":35
 *         public IntervalTreeNode left
 *         public IntervalTreeNode right
 *         public list contained             # <<<<<<<<<<<<<<
 *         public int level
 *         public IntervalTreeNode parent
 */

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_1__get__(PyObject *__pyx_v_self) {

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_v_self->contained);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_v_self->contained;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 35, __pyx_L1_error)

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->contained);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->contained);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->contained = ((PyObject*)__pyx_t_1);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.contained.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_5__del__(PyObject *__pyx_v_self) {

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(Py_None);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(Py_None);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->contained);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->contained);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->contained = ((PyObject*)Py_None);

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 35 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":36
 *         public IntervalTreeNode right
 *         public list contained
 *         public int level             # <<<<<<<<<<<<<<
 *         public IntervalTreeNode parent
 * 
 */

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_1__get__(PyObject *__pyx_v_self) {

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->level); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 36, __pyx_L1_error)

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_t_1;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.level.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = NULL;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_t_1;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 36, __pyx_L1_error)

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->level = __pyx_t_1;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.level.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 36 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":37
 *         public list contained
 *         public int level
 *         public IntervalTreeNode parent             # <<<<<<<<<<<<<<
 * 
 *     cpdef list contains_point(self, double x)
 */

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_1__get__(PyObject *__pyx_v_self) {

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->parent));

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->parent);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(1, 37, __pyx_L1_error)

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->parent);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->parent));

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.parent.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_5__del__(PyObject *__pyx_v_self) {

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(Py_None);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(Py_None);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->parent);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->parent));

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None);

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 37 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_29__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_28__reduce_cython__[] = "IntervalTreeNode.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_29__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_28__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_28__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_6;

#line 1 "(tree fragment)"
  int __pyx_t_7;

#line 1 "(tree fragment)"
  int __pyx_t_8;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center, self.contained, self.end, self.left, self.level, self.parent, self.right, self.start)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->center); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->end); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->level); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->start); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
  __pyx_t_5 = PyTuple_New(8); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_5);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->contained);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->contained);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self->contained);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2);

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 3, ((PyObject *)__pyx_v_self->left));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_3);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_3);

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->parent));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->parent));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 5, ((PyObject *)__pyx_v_self->parent));

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 6, ((PyObject *)__pyx_v_self->right));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_4);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_t_4);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

#line 5 "(tree fragment)"
  __pyx_t_4 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_5);

#line 5 "(tree fragment)"
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center, self.contained, self.end, self.left, self.level, self.parent, self.right, self.start)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_5;

#line 6 "(tree fragment)"
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.center, self.contained, self.end, self.left, self.level, self.parent, self.right, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_6 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_7 = (__pyx_t_6 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_7) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));

#line 8 "(tree fragment)"
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.contained is not None or self.left is not None or self.parent is not None or self.right is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center, self.contained, self.end, self.left, self.level, self.parent, self.right, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.contained is not None or self.left is not None or self.parent is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_6 = (__pyx_v_self->contained != ((PyObject*)Py_None));

#line 11 "(tree fragment)"
    __pyx_t_8 = (__pyx_t_6 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_8) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_7 = __pyx_t_8;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_8 = (((PyObject *)__pyx_v_self->left) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_6 = (__pyx_t_8 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_6) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_7 = __pyx_t_6;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_6 = (((PyObject *)__pyx_v_self->parent) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_8 = (__pyx_t_6 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_8) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_7 = __pyx_t_8;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_8 = (((PyObject *)__pyx_v_self->right) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_6 = (__pyx_t_8 != 0);

#line 11 "(tree fragment)"
    __pyx_t_7 = __pyx_t_6;

#line 11 "(tree fragment)"
    __pyx_L4_bool_binop_done:;

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_7;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.contained is not None or self.left is not None or self.parent is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_7 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_7) {

    /* "(tree fragment)":13
 *         use_setstate = self.contained is not None or self.left is not None or self.parent is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_IntervalTreeNode); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_40020397);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_40020397);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_40020397);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_5);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_t_5 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_3;

#line 13 "(tree fragment)"
    __pyx_t_3 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.contained is not None or self.left is not None or self.parent is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, None), state
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntervalTreeNode__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_IntervalTreeNode); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 15 "(tree fragment)"
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_40020397);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_40020397);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_40020397);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_5);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);

#line 15 "(tree fragment)"
    __pyx_t_3 = 0;

#line 15 "(tree fragment)"
    __pyx_t_5 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_4;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntervalTreeNode__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_31__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_30__setstate_cython__[] = "IntervalTreeNode.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_31__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_30__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_30__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntervalTreeNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode, (type(self), 0x262a9ad, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntervalTreeNode__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":474
 * cdef class SpanningMixin2D(object):
 * 
 *     def __init__(self, start, end):             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_start = 0
#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_end = 0
#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,0};

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[2] = {0,0};

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  2: 
#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else 
#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); 
#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 474, __pyx_L3_error)

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 474, __pyx_L3_error)

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_argtuple_error;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_start = values[0];

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_end = values[1];

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 474, __pyx_L3_error)

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return -1;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D___init__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self), __pyx_v_start, __pyx_v_end);

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end) {

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":475
 * 
 *     def __init__(self, start, end):
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 * 
 */

#line 475 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_start, __pyx_v_start) < 0) __PYX_ERR(0, 475, __pyx_L1_error)

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":476
 *     def __init__(self, start, end):
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 476 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_end, __pyx_v_end) < 0) __PYX_ERR(0, 476, __pyx_L1_error)

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":474
 * cdef class SpanningMixin2D(object):
 * 
 *     def __init__(self, start, end):             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 474 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":479
 * 
 *     @property
 *     def start(self):             # <<<<<<<<<<<<<<
 *         cdef double[:] view = self._start
 *         return view
 */

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start_1__get__(PyObject *__pyx_v_self) {

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self));

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_view = { 0, 0, { 0 }, { 0 }, { 0 } }
#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double *__pyx_t_1;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  struct __pyx_array_obj *__pyx_t_2 = NULL;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":480
 *     @property
 *     def start(self):
 *         cdef double[:] view = self._start             # <<<<<<<<<<<<<<
 *         return view
 * 
 */

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_v_self->_start;

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!__pyx_t_1) {

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyErr_SetString(PyExc_ValueError,"Cannot create cython.array from NULL pointer");

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __PYX_ERR(0, 480, __pyx_L1_error)

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __pyx_format_from_typeinfo(&__Pyx_TypeInfo_double); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 480, __pyx_L1_error)

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = Py_BuildValue((char*) "("  __PYX_BUILD_PY_SSIZE_T  ")", ((Py_ssize_t)2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 480, __pyx_L1_error)

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __pyx_array_new(__pyx_t_3, sizeof(double), PyBytes_AS_STRING(__pyx_t_4), (char *) "fortran", (char *) __pyx_t_1);

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 480, __pyx_L1_error)

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(((PyObject *)__pyx_t_2), PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 480, __pyx_L1_error)

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_view = __pyx_t_5;

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5.memview = NULL;

#line 480 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5.data = NULL;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":481
 *     def start(self):
 *         cdef double[:] view = self._start
 *         return view             # <<<<<<<<<<<<<<
 * 
 *     @start.setter
 */

#line 481 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 481 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_view, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 481, __pyx_L1_error)

#line 481 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 481 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_4;

#line 481 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = 0;

#line 481 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":479
 * 
 *     @property
 *     def start(self):             # <<<<<<<<<<<<<<
 *         cdef double[:] view = self._start
 *         return view
 */

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_t_2));

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.start.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_v_view, 1);

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 479 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":484
 * 
 *     @start.setter
 *     def start(self, start):             # <<<<<<<<<<<<<<
 *         self._start = start
 * 
 */

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_start); /*proto*/

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_start) {

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self), ((PyObject *)__pyx_v_start));

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_start) {

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_1[2];

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":485
 *     @start.setter
 *     def start(self, start):
 *         self._start = start             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 485 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__Pyx_carray_from_py_double(__pyx_v_start, __pyx_t_1, 2) < 0)) __PYX_ERR(0, 485, __pyx_L1_error)

#line 485 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  memcpy(&(__pyx_v_self->_start[0]), __pyx_t_1, sizeof(__pyx_v_self->_start[0]) * (2));

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":484
 * 
 *     @start.setter
 *     def start(self, start):             # <<<<<<<<<<<<<<
 *         self._start = start
 * 
 */

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.start.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 484 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":488
 * 
 *     @property
 *     def end(self):             # <<<<<<<<<<<<<<
 *         cdef double[:] view = self._end
 *         return view
 */

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end_1__get__(PyObject *__pyx_v_self) {

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self));

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_view = { 0, 0, { 0 }, { 0 }, { 0 } }
#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double *__pyx_t_1;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  struct __pyx_array_obj *__pyx_t_2 = NULL;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":489
 *     @property
 *     def end(self):
 *         cdef double[:] view = self._end             # <<<<<<<<<<<<<<
 *         return view
 * 
 */

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_v_self->_end;

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!__pyx_t_1) {

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyErr_SetString(PyExc_ValueError,"Cannot create cython.array from NULL pointer");

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __PYX_ERR(0, 489, __pyx_L1_error)

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __pyx_format_from_typeinfo(&__Pyx_TypeInfo_double); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 489, __pyx_L1_error)

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = Py_BuildValue((char*) "("  __PYX_BUILD_PY_SSIZE_T  ")", ((Py_ssize_t)2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __pyx_array_new(__pyx_t_3, sizeof(double), PyBytes_AS_STRING(__pyx_t_4), (char *) "fortran", (char *) __pyx_t_1);

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error)

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(((PyObject *)__pyx_t_2), PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 489, __pyx_L1_error)

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_view = __pyx_t_5;

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5.memview = NULL;

#line 489 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5.data = NULL;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":490
 *     def end(self):
 *         cdef double[:] view = self._end
 *         return view             # <<<<<<<<<<<<<<
 * 
 *     @end.setter
 */

#line 490 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 490 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_view, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 490, __pyx_L1_error)

#line 490 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 490 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_4;

#line 490 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = 0;

#line 490 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":488
 * 
 *     @property
 *     def end(self):             # <<<<<<<<<<<<<<
 *         cdef double[:] view = self._end
 *         return view
 */

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_t_2));

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.end.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_v_view, 1);

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 488 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":493
 * 
 *     @end.setter
 *     def end(self, end):             # <<<<<<<<<<<<<<
 *         self._end = end
 * 
 */

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_end); /*proto*/

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_end) {

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self), ((PyObject *)__pyx_v_end));

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_end) {

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_1[2];

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":494
 *     @end.setter
 *     def end(self, end):
 *         self._end = end             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, i):
 */

#line 494 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__Pyx_carray_from_py_double(__pyx_v_end, __pyx_t_1, 2) < 0)) __PYX_ERR(0, 494, __pyx_L1_error)

#line 494 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  memcpy(&(__pyx_v_self->_end[0]), __pyx_t_1, sizeof(__pyx_v_self->_end[0]) * (2));

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":493
 * 
 *     @end.setter
 *     def end(self, end):             # <<<<<<<<<<<<<<
 *         self._end = end
 * 
 */

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.end.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 493 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":496
 *         self._end = end
 * 
 *     def __contains__(self, i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_2__contains__[] = "Tests for point inclusion, `start <= i <= end`\n\n        Parameters\n        ----------\n        i : Number\n            The point to be tested\n\n        Returns\n        -------\n        bool\n        ";

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
#if CYTHON_UPDATE_DESCRIPTOR_DOC

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
struct wrapperbase __pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_2__contains__;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
#endif

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = 
#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_2__contains__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self), ((PyObject *)__pyx_v_i));

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_2__contains__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_i) {
  double __pyx_v_v[2]
#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  double __pyx_t_1[2];

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":511
 *             double[2] v
 * 
 *         v = i             # <<<<<<<<<<<<<<
 *         return self._contains(v)
 * 
 */

#line 511 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__Pyx_carray_from_py_double(__pyx_v_i, __pyx_t_1, 2) < 0)) __PYX_ERR(0, 511, __pyx_L1_error)

#line 511 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  memcpy(&(__pyx_v_v[0]), __pyx_t_1, sizeof(__pyx_v_v[0]) * (2));

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":512
 * 
 *         v = i
 *         return self._contains(v)             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _contains(self, double[2] i):
 */

#line 512 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self->__pyx_vtab)->_contains(__pyx_v_self, __pyx_v_v);

#line 512 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":496
 *         self._end = end
 * 
 *     def __contains__(self, i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 496 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":514
 *         return self._contains(v)
 * 
 *     cdef bint _contains(self, double[2] i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D__contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, double *__pyx_v_i) {
  PyObject *__pyx_v_v = NULL
#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_1;

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_contains", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":526
 *         bool
 *         """
 *         v = self._start[0] <= i[0] <= self._end[0]             # <<<<<<<<<<<<<<
 *         if v:
 *             v = self._start[1] <= i[1] <= self._end[1]
 */

#line 526 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = ((__pyx_v_self->_start[0]) <= (__pyx_v_i[0]));

#line 526 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_1) {

#line 526 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = ((__pyx_v_i[0]) <= (__pyx_v_self->_end[0]));

#line 526 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 526 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 526, __pyx_L1_error)

#line 526 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 526 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_v = __pyx_t_2;

#line 526 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":527
 *         """
 *         v = self._start[0] <= i[0] <= self._end[0]
 *         if v:             # <<<<<<<<<<<<<<
 *             v = self._start[1] <= i[1] <= self._end[1]
 *         return v
 */

#line 527 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_v); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 527, __pyx_L1_error)

#line 527 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":528
 *         v = self._start[0] <= i[0] <= self._end[0]
 *         if v:
 *             v = self._start[1] <= i[1] <= self._end[1]             # <<<<<<<<<<<<<<
 *         return v
 * 
 */

#line 528 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = ((__pyx_v_self->_start[1]) <= (__pyx_v_i[1]));

#line 528 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_1) {

#line 528 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((__pyx_v_i[1]) <= (__pyx_v_self->_end[1]));

#line 528 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 528 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)

#line 528 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 528 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF_SET(__pyx_v_v, __pyx_t_2);

#line 528 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":527
 *         """
 *         v = self._start[0] <= i[0] <= self._end[0]
 *         if v:             # <<<<<<<<<<<<<<
 *             v = self._start[1] <= i[1] <= self._end[1]
 *         return v
 */

#line 527 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":529
 *         if v:
 *             v = self._start[1] <= i[1] <= self._end[1]
 *         return v             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint contains(self, double[:] i):
 */

#line 529 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_v); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 529, __pyx_L1_error)

#line 529 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 529 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":514
 *         return self._contains(v)
 * 
 *     cdef bint _contains(self, double[2] i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D._contains", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_v);

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 514 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":531
 *         return v
 * 
 *     cpdef bint contains(self, double[:] i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5contains(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_i, int __pyx_skip_dispatch) {
  double __pyx_v_view[2]
#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_7;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_8;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains", 0);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_contains); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5contains)) {

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_v_i.memview)) { __Pyx_RaiseUnboundLocalError("i"); __PYX_ERR(0, 531, __pyx_L1_error) }

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_i, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 531, __pyx_L1_error)

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_5 = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_5)) {

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 531, __pyx_L1_error)

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = __pyx_t_6;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":546
 *             double[2] view
 * 
 *         view[0] = i[0]             # <<<<<<<<<<<<<<
 *         view[1] = i[1]
 *         return self._contains(view)
 */

#line 546 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = 0;

#line 546 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_8 = -1;

#line 546 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_7 < 0) {

#line 546 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 += __pyx_v_i.shape[0];

#line 546 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 0;
  } else 
#line 546 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
if (unlikely(__pyx_t_7 >= __pyx_v_i.shape[0])) __pyx_t_8 = 0;

#line 546 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_t_8 != -1)) {

#line 546 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_RaiseBufferIndexError(__pyx_t_8);

#line 546 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __PYX_ERR(0, 546, __pyx_L1_error)

#line 546 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 546 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  (__pyx_v_view[0]) = (*((double *) ( /* dim=0 */ (__pyx_v_i.data + __pyx_t_7 * __pyx_v_i.strides[0]) )));

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":547
 * 
 *         view[0] = i[0]
 *         view[1] = i[1]             # <<<<<<<<<<<<<<
 *         return self._contains(view)
 * 
 */

#line 547 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = 1;

#line 547 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_8 = -1;

#line 547 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_7 < 0) {

#line 547 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 += __pyx_v_i.shape[0];

#line 547 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_t_7 < 0)) __pyx_t_8 = 0;
  } else 
#line 547 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
if (unlikely(__pyx_t_7 >= __pyx_v_i.shape[0])) __pyx_t_8 = 0;

#line 547 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_t_8 != -1)) {

#line 547 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_RaiseBufferIndexError(__pyx_t_8);

#line 547 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __PYX_ERR(0, 547, __pyx_L1_error)

#line 547 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 547 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  (__pyx_v_view[1]) = (*((double *) ( /* dim=0 */ (__pyx_v_i.data + __pyx_t_7 * __pyx_v_i.strides[0]) )));

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":548
 *         view[0] = i[0]
 *         view[1] = i[1]
 *         return self._contains(view)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint overlaps(self, SpanningMixin2D interval):
 */

#line 548 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self->__pyx_vtab)->_contains(__pyx_v_self, __pyx_v_view);

#line 548 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":531
 *         return v
 * 
 *     cpdef bint contains(self, double[:] i):             # <<<<<<<<<<<<<<
 *         """Tests for point inclusion, `start <= i <= end`
 * 
 */

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.contains", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5contains(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4contains[] = "SpanningMixin2D.contains(self, double[:] i) -> bool\nTests for point inclusion, `start <= i <= end`\n\n        Parameters\n        ----------\n        i : double\n            The point to be tested\n\n        Returns\n        -------\n        bool\n        ";

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5contains(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  __Pyx_memviewslice __pyx_v_i = { 0, 0, { 0 }, { 0 }, { 0 } }
#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains (wrapper)", 0);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  assert(__pyx_arg_i); {

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_i = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_arg_i, PyBUF_WRITABLE); if (unlikely(!__pyx_v_i.memview)) __PYX_ERR(0, 531, __pyx_L3_error)

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.contains", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4contains(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self), __pyx_v_i);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4contains(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_i) {

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains", 0);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_v_i.memview)) { __Pyx_RaiseUnboundLocalError("i"); __PYX_ERR(0, 531, __pyx_L1_error) }

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_contains(__pyx_v_self, __pyx_v_i, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.contains", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_v_i, 1);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 531 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":550
 *         return self._contains(view)
 * 
 *     cpdef bint overlaps(self, SpanningMixin2D interval):             # <<<<<<<<<<<<<<
 *         """Tests whether another spanning entity with
 *         a defined start and end point overlaps with
 */

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_interval, int __pyx_skip_dispatch) {
  int __pyx_v_cond
#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  short __pyx_v_i
#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_5;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  short __pyx_t_6;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps", 0);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlaps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_7overlaps)) {

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_4 = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_4)) {

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_interval));

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = __pyx_t_5;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":565
 *         cdef bint cond
 *         cdef short i
 *         cond = True             # <<<<<<<<<<<<<<
 *         for i in range(2):
 *             cond &= ((self._start[i] <= interval._start[i] and self._end[i] >= interval._start[i]) or (
 */

#line 565 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_cond = 1;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":566
 *         cdef short i
 *         cond = True
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             cond &= ((self._start[i] <= interval._start[i] and self._end[i] >= interval._start[i]) or (
 *                 self._start[i] >= interval._start[i] and self._end[i] <= interval._end[i]) or (
 */

#line 566 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  for (__pyx_t_6 = 0; __pyx_t_6 < 2; __pyx_t_6+=1) {

#line 566 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_i = __pyx_t_6;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":567
 *         cond = True
 *         for i in range(2):
 *             cond &= ((self._start[i] <= interval._start[i] and self._end[i] >= interval._start[i]) or (             # <<<<<<<<<<<<<<
 *                 self._start[i] >= interval._start[i] and self._end[i] <= interval._end[i]) or (
 *                 self._start[i] >= interval._start[i] and self._end[i] >= interval._end[i] and self._start[i] <= interval._end[i]) or (
 */

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_start[__pyx_v_i]) <= (__pyx_v_interval->_start[__pyx_v_i])) != 0);

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!__pyx_t_7) {

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L6_next_or;

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_end[__pyx_v_i]) >= (__pyx_v_interval->_start[__pyx_v_i])) != 0);

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!__pyx_t_7) {

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = __pyx_t_7;

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_bool_binop_done;

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L6_next_or:;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":568
 *         for i in range(2):
 *             cond &= ((self._start[i] <= interval._start[i] and self._end[i] >= interval._start[i]) or (
 *                 self._start[i] >= interval._start[i] and self._end[i] <= interval._end[i]) or (             # <<<<<<<<<<<<<<
 *                 self._start[i] >= interval._start[i] and self._end[i] >= interval._end[i] and self._start[i] <= interval._end[i]) or (
 *                 self._start[i] <= interval._start[i] and self._end[i] >= interval._start[i]) or (
 */

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_start[__pyx_v_i]) >= (__pyx_v_interval->_start[__pyx_v_i])) != 0);

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!__pyx_t_7) {

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L8_next_or;

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_end[__pyx_v_i]) <= (__pyx_v_interval->_end[__pyx_v_i])) != 0);

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!__pyx_t_7) {

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = __pyx_t_7;

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_bool_binop_done;

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 568 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L8_next_or:;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":569
 *             cond &= ((self._start[i] <= interval._start[i] and self._end[i] >= interval._start[i]) or (
 *                 self._start[i] >= interval._start[i] and self._end[i] <= interval._end[i]) or (
 *                 self._start[i] >= interval._start[i] and self._end[i] >= interval._end[i] and self._start[i] <= interval._end[i]) or (             # <<<<<<<<<<<<<<
 *                 self._start[i] <= interval._start[i] and self._end[i] >= interval._start[i]) or (
 *                 self._start[i] <= interval._end[i] and self._end[i] >= interval._end[i]))
 */

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_start[__pyx_v_i]) >= (__pyx_v_interval->_start[__pyx_v_i])) != 0);

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!__pyx_t_7) {

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L10_next_or;

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_end[__pyx_v_i]) >= (__pyx_v_interval->_end[__pyx_v_i])) != 0);

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!__pyx_t_7) {

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L10_next_or;

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_start[__pyx_v_i]) <= (__pyx_v_interval->_end[__pyx_v_i])) != 0);

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!__pyx_t_7) {

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = __pyx_t_7;

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_bool_binop_done;

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 569 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L10_next_or:;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":570
 *                 self._start[i] >= interval._start[i] and self._end[i] <= interval._end[i]) or (
 *                 self._start[i] >= interval._start[i] and self._end[i] >= interval._end[i] and self._start[i] <= interval._end[i]) or (
 *                 self._start[i] <= interval._start[i] and self._end[i] >= interval._start[i]) or (             # <<<<<<<<<<<<<<
 *                 self._start[i] <= interval._end[i] and self._end[i] >= interval._end[i]))
 *         return cond
 */

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_start[__pyx_v_i]) <= (__pyx_v_interval->_start[__pyx_v_i])) != 0);

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!__pyx_t_7) {

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L13_next_or;

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_end[__pyx_v_i]) >= (__pyx_v_interval->_start[__pyx_v_i])) != 0);

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!__pyx_t_7) {

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = __pyx_t_7;

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_bool_binop_done;

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 570 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L13_next_or:;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":571
 *                 self._start[i] >= interval._start[i] and self._end[i] >= interval._end[i] and self._start[i] <= interval._end[i]) or (
 *                 self._start[i] <= interval._start[i] and self._end[i] >= interval._start[i]) or (
 *                 self._start[i] <= interval._end[i] and self._end[i] >= interval._end[i]))             # <<<<<<<<<<<<<<
 *         return cond
 * 
 */

#line 571 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_start[__pyx_v_i]) <= (__pyx_v_interval->_end[__pyx_v_i])) != 0);

#line 571 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_7) {

#line 571 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 571 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = __pyx_t_7;

#line 571 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_bool_binop_done;

#line 571 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 571 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_end[__pyx_v_i]) >= (__pyx_v_interval->_end[__pyx_v_i])) != 0);

#line 571 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_7;

#line 571 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L5_bool_binop_done:;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":567
 *         cond = True
 *         for i in range(2):
 *             cond &= ((self._start[i] <= interval._start[i] and self._end[i] >= interval._start[i]) or (             # <<<<<<<<<<<<<<
 *                 self._start[i] >= interval._start[i] and self._end[i] <= interval._end[i]) or (
 *                 self._start[i] >= interval._start[i] and self._end[i] >= interval._end[i] and self._start[i] <= interval._end[i]) or (
 */

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_cond = (__pyx_v_cond & __pyx_t_5);

#line 567 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":572
 *                 self._start[i] <= interval._start[i] and self._end[i] >= interval._start[i]) or (
 *                 self._start[i] <= interval._end[i] and self._end[i] >= interval._end[i]))
 *         return cond             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint contains_interval(self, SpanningMixin2D interval):
 */

#line 572 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_v_cond;

#line 572 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":550
 *         return self._contains(view)
 * 
 *     cpdef bint overlaps(self, SpanningMixin2D interval):             # <<<<<<<<<<<<<<
 *         """Tests whether another spanning entity with
 *         a defined start and end point overlaps with
 */

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6overlaps[] = "SpanningMixin2D.overlaps(self, SpanningMixin2D interval) -> bool\nTests whether another spanning entity with\n        a defined start and end point overlaps with\n        this spanning entity\n\n        Parameters\n        ----------\n        interval : SpanningMixin2D\n\n        Returns\n        -------\n        bool\n        ";

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_7overlaps(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps (wrapper)", 0);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D, 1, "interval", 0))) __PYX_ERR(0, 550, __pyx_L1_error)
  __pyx_r = 
#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6overlaps(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_interval));

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6overlaps(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_interval) {

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps", 0);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_overlaps(__pyx_v_self, __pyx_v_interval, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 550 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":574
 *         return cond
 * 
 *     cpdef bint contains_interval(self, SpanningMixin2D interval):             # <<<<<<<<<<<<<<
 *         """Tests whether the other spanning entity is
 *         completely contained inside this entity
 */

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_9contains_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_interval, int __pyx_skip_dispatch) {
  short __pyx_v_i
#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  int __pyx_v_cond
#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_5;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  short __pyx_t_6;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains_interval", 0);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_contains_interval); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 574, __pyx_L1_error)

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_9contains_interval)) {

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_4 = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_4)) {

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_interval));

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 574, __pyx_L1_error)

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 574, __pyx_L1_error)

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = __pyx_t_5;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":589
 *             short i
 *             bint cond
 *         cond = True             # <<<<<<<<<<<<<<
 *         for i in range(2):
 *             cond &= self._start[i] <= interval._start[i] and self._end[i] >= interval._end[i]
 */

#line 589 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_cond = 1;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":590
 *             bint cond
 *         cond = True
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             cond &= self._start[i] <= interval._start[i] and self._end[i] >= interval._end[i]
 *         return cond
 */

#line 590 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  for (__pyx_t_6 = 0; __pyx_t_6 < 2; __pyx_t_6+=1) {

#line 590 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_i = __pyx_t_6;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":591
 *         cond = True
 *         for i in range(2):
 *             cond &= self._start[i] <= interval._start[i] and self._end[i] >= interval._end[i]             # <<<<<<<<<<<<<<
 *         return cond
 * 
 */

#line 591 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_start[__pyx_v_i]) <= (__pyx_v_interval->_start[__pyx_v_i])) != 0);

#line 591 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_7) {

#line 591 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 591 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = __pyx_t_7;

#line 591 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_bool_binop_done;

#line 591 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 591 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (((__pyx_v_self->_end[__pyx_v_i]) >= (__pyx_v_interval->_end[__pyx_v_i])) != 0);

#line 591 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __pyx_t_7;

#line 591 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L5_bool_binop_done:;

#line 591 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_cond = (__pyx_v_cond & __pyx_t_5);

#line 591 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":592
 *         for i in range(2):
 *             cond &= self._start[i] <= interval._start[i] and self._end[i] >= interval._end[i]
 *         return cond             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 592 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_v_cond;

#line 592 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":574
 *         return cond
 * 
 *     cpdef bint contains_interval(self, SpanningMixin2D interval):             # <<<<<<<<<<<<<<
 *         """Tests whether the other spanning entity is
 *         completely contained inside this entity
 */

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.contains_interval", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_9contains_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_8contains_interval[] = "SpanningMixin2D.contains_interval(self, SpanningMixin2D interval) -> bool\nTests whether the other spanning entity is\n        completely contained inside this entity\n\n        Parameters\n        ----------\n        interval : SpanningMixin2D\n\n        Returns\n        -------\n        bool\n        ";

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_9contains_interval(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains_interval (wrapper)", 0);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D, 1, "interval", 0))) __PYX_ERR(0, 574, __pyx_L1_error)
  __pyx_r = 
#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_8contains_interval(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_interval));

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_8contains_interval(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_interval) {

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("contains_interval", 0);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_contains_interval(__pyx_v_self, __pyx_v_interval, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 574, __pyx_L1_error)

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.contains_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 574 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":47
 * cdef class SpanningMixin2D(object):
 *     cdef:
 *         public double[2] _start             # <<<<<<<<<<<<<<
 *         public double[2] _end
 * 
 */

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start_1__get__(PyObject *__pyx_v_self) {

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self));

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self) {

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __Pyx_carray_to_py_double(__pyx_v_self->_start, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_t_1;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D._start.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = NULL;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_value) {

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double __pyx_t_1[2];

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  if (unlikely(__Pyx_carray_from_py_double(__pyx_v_value, __pyx_t_1, 2) < 0)) __PYX_ERR(1, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  memcpy(&(__pyx_v_self->_start[0]), __pyx_t_1, sizeof(__pyx_v_self->_start[0]) * (2));

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D._start.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 47 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":48
 *     cdef:
 *         public double[2] _start
 *         public double[2] _end             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _contains(self, double[2] i)
 */

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end_1__get__(PyObject *__pyx_v_self) {

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self));

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self) {

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __Pyx_carray_to_py_double(__pyx_v_self->_end, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 48, __pyx_L1_error)

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_t_1;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D._end.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = NULL;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v_value) {

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  double __pyx_t_1[2];

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  if (unlikely(__Pyx_carray_from_py_double(__pyx_v_value, __pyx_t_1, 2) < 0)) __PYX_ERR(1, 48, __pyx_L1_error)

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  memcpy(&(__pyx_v_self->_end[0]), __pyx_t_1, sizeof(__pyx_v_self->_end[0]) * (2));

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D._end.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 48 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_10__reduce_cython__[] = "SpanningMixin2D.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_10__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_10__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_4;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._end, self._start)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = __Pyx_carray_to_py_double(__pyx_v_self->_end, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = __Pyx_carray_to_py_double(__pyx_v_self->_start, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._end, self._start)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_3;

#line 6 "(tree fragment)"
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self._end, self._start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_5 = (__pyx_t_4 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));

#line 8 "(tree fragment)"
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._end, self._start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_SpanningMixin2D, (type(self), 0xfeaf07a, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = 0;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SpanningMixin2D, (type(self), 0xfeaf07a, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_5 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_5) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_SpanningMixin2D, (type(self), 0xfeaf07a, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_SpanningMixin2D, (type(self), 0xfeaf07a, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_SpanningMixin2D); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 13 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_267055226);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_267055226);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_267055226);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_2 = 0;

#line 13 "(tree fragment)"
    __pyx_t_3 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_1;

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SpanningMixin2D, (type(self), 0xfeaf07a, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_SpanningMixin2D, (type(self), 0xfeaf07a, None), state
 *     else:
 *         return __pyx_unpickle_SpanningMixin2D, (type(self), 0xfeaf07a, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SpanningMixin2D__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_SpanningMixin2D); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_267055226);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_267055226);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_267055226);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    __pyx_t_3 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_2;

#line 15 "(tree fragment)"
    __pyx_t_2 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SpanningMixin2D, (type(self), 0xfeaf07a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SpanningMixin2D__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_12__setstate_cython__[] = "SpanningMixin2D.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_12__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_12__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_SpanningMixin2D, (type(self), 0xfeaf07a, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SpanningMixin2D__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin2D__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SpanningMixin2D, (type(self), 0xfeaf07a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SpanningMixin2D__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":597
 * cdef class Interval2D(SpanningMixin2D):
 * 
 *     def __init__(self, start, end, members=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if members is None:
 *             members = []
 */

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_start = 0
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_end = 0
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_members = 0
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_kwargs);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,&__pyx_n_s_members,0};

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[3] = {0,0,0};

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    values[2] = ((PyObject *)Py_None);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  3: 
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else 
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); 
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 597, __pyx_L3_error)

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  2:

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (kw_args > 0) {

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_members);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 597, __pyx_L3_error)

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        break;
        default: 
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_start = values[0];

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_end = values[1];

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_members = values[2];

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 597, __pyx_L3_error)

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval2D.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return -1;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D___init__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self), __pyx_v_start, __pyx_v_end, __pyx_v_members, __pyx_v_kwargs);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_members, PyObject *__pyx_v_kwargs) {

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_1;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_2;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_members);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":598
 * 
 *     def __init__(self, start, end, members=None, **kwargs):
 *         if members is None:             # <<<<<<<<<<<<<<
 *             members = []
 *         self.start = start
 */

#line 598 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_v_members == Py_None);

#line 598 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 598 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":599
 *     def __init__(self, start, end, members=None, **kwargs):
 *         if members is None:
 *             members = []             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */

#line 599 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 599, __pyx_L1_error)

#line 599 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 599 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF_SET(__pyx_v_members, __pyx_t_3);

#line 599 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":598
 * 
 *     def __init__(self, start, end, members=None, **kwargs):
 *         if members is None:             # <<<<<<<<<<<<<<
 *             members = []
 *         self.start = start
 */

#line 598 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":600
 *         if members is None:
 *             members = []
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 *         self.members = members
 */

#line 600 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_start, __pyx_v_start) < 0) __PYX_ERR(0, 600, __pyx_L1_error)

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":601
 *             members = []
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 *         self.members = members
 *         self.data = kwargs
 */

#line 601 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_end, __pyx_v_end) < 0) __PYX_ERR(0, 601, __pyx_L1_error)

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":602
 *         self.start = start
 *         self.end = end
 *         self.members = members             # <<<<<<<<<<<<<<
 *         self.data = kwargs
 * 
 */

#line 602 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!(likely(PyList_CheckExact(__pyx_v_members))||((__pyx_v_members) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_members)->tp_name), 0))) __PYX_ERR(0, 602, __pyx_L1_error)

#line 602 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __pyx_v_members;

#line 602 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_t_3);

#line 602 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 602 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->members);

#line 602 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_v_self->members);

#line 602 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->members = ((PyObject*)__pyx_t_3);

#line 602 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":603
 *         self.end = end
 *         self.members = members
 *         self.data = kwargs             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 603 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_kwargs);

#line 603 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_kwargs);

#line 603 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->data);

#line 603 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_v_self->data);

#line 603 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->data = __pyx_v_kwargs;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":597
 * cdef class Interval2D(SpanningMixin2D):
 * 
 *     def __init__(self, start, end, members=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if members is None:
 *             members = []
 */

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval2D.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_members);

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 597 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":605
 *         self.data = kwargs
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(start=%r, end=%r, data=%r)" % (
 *             self.__class__.__name__,
 */

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_3__repr__(PyObject *__pyx_v_self); /*proto*/

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_3__repr__(PyObject *__pyx_v_self) {

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_2__repr__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self));

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_2__repr__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self) {

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":606
 * 
 *     def __repr__(self):
 *         return "%s(start=%r, end=%r, data=%r)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             list(self.start), list(self.end), self.data)
 */

#line 606 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":607
 *     def __repr__(self):
 *         return "%s(start=%r, end=%r, data=%r)" % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             list(self.start), list(self.end), self.data)
 * 
 */

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 607, __pyx_L1_error)

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 607, __pyx_L1_error)

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":608
 *         return "%s(start=%r, end=%r, data=%r)" % (
 *             self.__class__.__name__,
 *             list(self.start), list(self.end), self.data)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, k):
 */

#line 608 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L1_error)

#line 608 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 608 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 608, __pyx_L1_error)

#line 608 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 608 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 608 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L1_error)

#line 608 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 608 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 608, __pyx_L1_error)

#line 608 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 608 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":607
 *     def __repr__(self):
 *         return "%s(start=%r, end=%r, data=%r)" % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             list(self.start), list(self.end), self.data)
 * 
 */

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 607, __pyx_L1_error)

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_self->data);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_self->data);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->data);

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = 0;

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = 0;

#line 607 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":606
 * 
 *     def __repr__(self):
 *         return "%s(start=%r, end=%r, data=%r)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             list(self.start), list(self.end), self.data)
 */

#line 606 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_s_start_r_end_r_data_r, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 606, __pyx_L1_error)

#line 606 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 606 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 606 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_4;

#line 606 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = 0;

#line 606 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":605
 *         self.data = kwargs
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(start=%r, end=%r, data=%r)" % (
 *             self.__class__.__name__,
 */

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval2D.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 605 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":610
 *             list(self.start), list(self.end), self.data)
 * 
 *     def __getitem__(self, k):             # <<<<<<<<<<<<<<
 *         return self.members[k]
 * 
 */

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_k); /*proto*/

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_k) {

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4__getitem__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self), ((PyObject *)__pyx_v_k));

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4__getitem__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self, PyObject *__pyx_v_k) {

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":611
 * 
 *     def __getitem__(self, k):
 *         return self.members[k]             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 611 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 611 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_v_self->members == Py_None)) {

#line 611 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 611 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __PYX_ERR(0, 611, __pyx_L1_error)

#line 611 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 611 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->members, __pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 611, __pyx_L1_error)

#line 611 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 611 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 611 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 611 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":610
 *             list(self.start), list(self.end), self.data)
 * 
 *     def __getitem__(self, k):             # <<<<<<<<<<<<<<
 *         return self.members[k]
 * 
 */

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval2D.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 610 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":613
 *         return self.members[k]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.members)
 * 
 */

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7__len__(PyObject *__pyx_v_self); /*proto*/

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7__len__(PyObject *__pyx_v_self) {

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_r;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_6__len__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self));

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_6__len__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self) {

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_r;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_2;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":614
 * 
 *     def __len__(self):
 *         return len(self.members)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */

#line 614 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_v_self->members;

#line 614 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_t_1 == Py_None)) {

#line 614 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 614 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 614 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_2;

#line 614 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":613
 *         return self.members[k]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.members)
 * 
 */

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval2D.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 613 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":616
 *         return len(self.members)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.members)
 * 
 */

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_9__iter__(PyObject *__pyx_v_self); /*proto*/

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_9__iter__(PyObject *__pyx_v_self) {

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = 
#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_8__iter__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self));

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_8__iter__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self) {

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":617
 * 
 *     def __iter__(self):
 *         return iter(self.members)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 617 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 617 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_v_self->members;

#line 617 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 617 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 617, __pyx_L1_error)

#line 617 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 617 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 617 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_2;

#line 617 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = 0;

#line 617 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":616
 *         return len(self.members)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.members)
 * 
 */

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval2D.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 616 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":58
 * cdef class Interval2D(SpanningMixin2D):
 *     cdef:
 *         public list members             # <<<<<<<<<<<<<<
 *         public dict data
 * 
 */

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_1__get__(PyObject *__pyx_v_self) {

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self));

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self) {

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_v_self->members);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_v_self->members;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self, PyObject *__pyx_v_value) {

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->members);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->members);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->members = ((PyObject*)__pyx_t_1);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval2D.members.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_5__del__(PyObject *__pyx_v_self) {

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self));

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self) {

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(Py_None);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(Py_None);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->members);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->members);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->members = ((PyObject*)Py_None);

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":59
 *     cdef:
 *         public list members
 *         public dict data             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_1__get__(PyObject *__pyx_v_self) {

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self));

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self) {

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_v_self->data);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_v_self->data;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self, PyObject *__pyx_v_value) {

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 59, __pyx_L1_error)

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->data);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->data);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->data = ((PyObject*)__pyx_t_1);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval2D.data.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_5__del__(PyObject *__pyx_v_self) {

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self));

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self) {

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(Py_None);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(Py_None);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->data);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->data);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->data = ((PyObject*)Py_None);

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 59 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_10__reduce_cython__[] = "Interval2D.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_10__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_10__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_4;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  int __pyx_t_6;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._end, self._start, self.data, self.members)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = __Pyx_carray_to_py_double(__pyx_v_self->__pyx_base._end, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = __Pyx_carray_to_py_double(__pyx_v_self->__pyx_base._start, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->data);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->data);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->data);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->members);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->members);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->members);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._end, self._start, self.data, self.members)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_3;

#line 6 "(tree fragment)"
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self._end, self._start, self.data, self.members)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_5 = (__pyx_t_4 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));

#line 8 "(tree fragment)"
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.data is not None or self.members is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._end, self._start, self.data, self.members)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.data is not None or self.members is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Interval2D, (type(self), 0x688f3f3, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_4 = (__pyx_v_self->data != ((PyObject*)Py_None));

#line 11 "(tree fragment)"
    __pyx_t_6 = (__pyx_t_4 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_6) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_5 = __pyx_t_6;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_6 = (__pyx_v_self->members != ((PyObject*)Py_None));

#line 11 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6 != 0);

#line 11 "(tree fragment)"
    __pyx_t_5 = __pyx_t_4;

#line 11 "(tree fragment)"
    __pyx_L4_bool_binop_done:;

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_5;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None or self.members is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Interval2D, (type(self), 0x688f3f3, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_5 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_5) {

    /* "(tree fragment)":13
 *         use_setstate = self.data is not None or self.members is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Interval2D, (type(self), 0x688f3f3, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Interval2D, (type(self), 0x688f3f3, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_Interval2D); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 13 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_109638643);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_109638643);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_109638643);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_2 = 0;

#line 13 "(tree fragment)"
    __pyx_t_3 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_1;

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None or self.members is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Interval2D, (type(self), 0x688f3f3, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Interval2D, (type(self), 0x688f3f3, None), state
 *     else:
 *         return __pyx_unpickle_Interval2D, (type(self), 0x688f3f3, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Interval2D__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Interval2D); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_109638643);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_109638643);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_109638643);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    __pyx_t_3 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_2;

#line 15 "(tree fragment)"
    __pyx_t_2 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval2D.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Interval2D, (type(self), 0x688f3f3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Interval2D__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_12__setstate_cython__[] = "Interval2D.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_12__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_12__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Interval2D, (type(self), 0x688f3f3, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Interval2D__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval2D__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Interval2D, (type(self), 0x688f3f3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Interval2D__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.Interval2D.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":620
 * 
 * 
 * cpdef list intervals_containing_point_2d(list intervals, double x, int level):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_3intervals_containing_point_2d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point_2d(PyObject *__pyx_v_intervals, double __pyx_v_x, int __pyx_v_level, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v_interval = 0
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_result = 0
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_2;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_3;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_4;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_6;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_8;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("intervals_containing_point_2d", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":625
 *         SpanningMixin2D interval
 *         list result
 *     result = []             # <<<<<<<<<<<<<<
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin2D>PyList_GetItem(intervals, i)
 */

#line 625 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)

#line 625 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 625 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_result = ((PyObject*)__pyx_t_1);

#line 625 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":626
 *         list result
 *     result = []
 *     for i in range(PyList_Size(intervals)):             # <<<<<<<<<<<<<<
 *         interval = <SpanningMixin2D>PyList_GetItem(intervals, i)
 *         if interval._start[level % 2] <= x <= interval._end[level % 2]:
 */

#line 626 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = PyList_Size(__pyx_v_intervals); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 626, __pyx_L1_error)

#line 626 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 626 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 626 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":627
 *     result = []
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin2D>PyList_GetItem(intervals, i)             # <<<<<<<<<<<<<<
 *         if interval._start[level % 2] <= x <= interval._end[level % 2]:
 *             result.append(interval)
 */

#line 627 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = PyList_GetItem(__pyx_v_intervals, __pyx_v_i); if (unlikely(__pyx_t_5 == ((PyObject *)NULL))) __PYX_ERR(0, 627, __pyx_L1_error)

#line 627 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_5);

#line 627 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 627 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF_SET(__pyx_v_interval, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_t_1));

#line 627 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":628
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin2D>PyList_GetItem(intervals, i)
 *         if interval._start[level % 2] <= x <= interval._end[level % 2]:             # <<<<<<<<<<<<<<
 *             result.append(interval)
 *     return result
 */

#line 628 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_6 = ((__pyx_v_interval->_start[__Pyx_mod_long(__pyx_v_level, 2)]) <= __pyx_v_x);

#line 628 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_6) {

#line 628 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_6 = (__pyx_v_x <= (__pyx_v_interval->_end[__Pyx_mod_long(__pyx_v_level, 2)]));

#line 628 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 628 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = (__pyx_t_6 != 0);

#line 628 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":629
 *         interval = <SpanningMixin2D>PyList_GetItem(intervals, i)
 *         if interval._start[level % 2] <= x <= interval._end[level % 2]:
 *             result.append(interval)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */

#line 629 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_interval)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 629, __pyx_L1_error)

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":628
 *     for i in range(PyList_Size(intervals)):
 *         interval = <SpanningMixin2D>PyList_GetItem(intervals, i)
 *         if interval._start[level % 2] <= x <= interval._end[level % 2]:             # <<<<<<<<<<<<<<
 *             result.append(interval)
 *     return result
 */

#line 628 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 628 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":630
 *         if interval._start[level % 2] <= x <= interval._end[level % 2]:
 *             result.append(interval)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 630 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 630 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_result);

#line 630 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_v_result;

#line 630 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":620
 * 
 * 
 * cpdef list intervals_containing_point_2d(list intervals, double x, int level):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.intervals_containing_point_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_interval);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_3intervals_containing_point_2d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_2intervals_containing_point_2d[] = "intervals_containing_point_2d(list intervals, double x, int level) -> list";

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_3intervals_containing_point_2d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_intervals = 0
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  double __pyx_v_x
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  int __pyx_v_level
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("intervals_containing_point_2d (wrapper)", 0);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_intervals,&__pyx_n_s_x,&__pyx_n_s_level,0};

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[3] = {0,0,0};

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  3: 
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_intervals)) != 0)) kw_args--;
        else 
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("intervals_containing_point_2d", 1, 3, 3, 1); 
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 620, __pyx_L3_error)

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  2:

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_level)) != 0)) kw_args--;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("intervals_containing_point_2d", 1, 3, 3, 2); 
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 620, __pyx_L3_error)

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "intervals_containing_point_2d") < 0)) __PYX_ERR(0, 620, __pyx_L3_error)

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_argtuple_error;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_intervals = ((PyObject*)values[0]);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 620, __pyx_L3_error)

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_level = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_level == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 620, __pyx_L3_error)

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("intervals_containing_point_2d", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 620, __pyx_L3_error)

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.intervals_containing_point_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return NULL;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_intervals), (&PyList_Type), 1, "intervals", 1))) __PYX_ERR(0, 620, __pyx_L1_error)
  __pyx_r = 
#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_2intervals_containing_point_2d(__pyx_self, __pyx_v_intervals, __pyx_v_x, __pyx_v_level);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_2intervals_containing_point_2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_intervals, double __pyx_v_x, int __pyx_v_level) {

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("intervals_containing_point_2d", 0);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point_2d(__pyx_v_intervals, __pyx_v_x, __pyx_v_level, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 620, __pyx_L1_error)

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.intervals_containing_point_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 620 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":634
 * 
 * cdef class IntervalTreeNode2D(IntervalTreeNode):
 *     def __init__(self, center, left, contained, right, level=0, parent=None, inner_organizer=None):             # <<<<<<<<<<<<<<
 *         super().__init__(center, left, contained, right, level=level, parent=parent)
 *         self.inner_organizer = inner_organizer
 */

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_center = 0
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_left = 0
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_contained = 0
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_right = 0
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_level = 0
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_parent = 0
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_inner_organizer = 0
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_center,&__pyx_n_s_left,&__pyx_n_s_contained,&__pyx_n_s_right,&__pyx_n_s_level,&__pyx_n_s_parent,&__pyx_n_s_inner_organizer,0};

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[7] = {0,0,0,0,0,0,0};

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    values[4] = ((PyObject *)__pyx_int_0);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    values[5] = ((PyObject *)Py_None);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    values[6] = ((PyObject *)Py_None);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  7: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[6] = PyTuple_GET_ITEM(__pyx_args, 6);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  6: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_center)) != 0)) kw_args--;
        else 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 7, 1); 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 634, __pyx_L3_error)

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  2:

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_contained)) != 0)) kw_args--;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 7, 2); 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 634, __pyx_L3_error)

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  3:

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 7, 3); 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 634, __pyx_L3_error)

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  4:

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (kw_args > 0) {

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_level);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  5:

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (kw_args > 0) {

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parent);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (value) { values[5] = value; kw_args--; }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  6:

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (kw_args > 0) {

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inner_organizer);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (value) { values[6] = value; kw_args--; }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 634, __pyx_L3_error)

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[6] = PyTuple_GET_ITEM(__pyx_args, 6);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  6: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        break;
        default: 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_center = values[0];

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_left = values[1];

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_contained = values[2];

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_right = values[3];

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_level = values[4];

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_parent = values[5];

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_inner_organizer = values[6];

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 7, PyTuple_GET_SIZE(__pyx_args)); 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 634, __pyx_L3_error)

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return -1;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D___init__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self), __pyx_v_center, __pyx_v_left, __pyx_v_contained, __pyx_v_right, __pyx_v_level, __pyx_v_parent, __pyx_v_inner_organizer);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D___init__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, PyObject *__pyx_v_center, PyObject *__pyx_v_left, PyObject *__pyx_v_contained, PyObject *__pyx_v_right, PyObject *__pyx_v_level, PyObject *__pyx_v_parent, PyObject *__pyx_v_inner_organizer) {

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_r;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":635
 * cdef class IntervalTreeNode2D(IntervalTreeNode):
 *     def __init__(self, center, left, contained, right, level=0, parent=None, inner_organizer=None):
 *         super().__init__(center, left, contained, right, level=level, parent=parent)             # <<<<<<<<<<<<<<
 *         self.inner_organizer = inner_organizer
 *         self.organize()
 */

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D));

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D));

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D));

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self));

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_center);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_center);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_center);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_left);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_left);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_left);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_contained);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_contained);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_contained);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_right);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_right);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_v_right);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_level, __pyx_v_level) < 0) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_parent, __pyx_v_parent) < 0) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 635 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":636
 *     def __init__(self, center, left, contained, right, level=0, parent=None, inner_organizer=None):
 *         super().__init__(center, left, contained, right, level=level, parent=parent)
 *         self.inner_organizer = inner_organizer             # <<<<<<<<<<<<<<
 *         self.organize()
 * 
 */

#line 636 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_inner_organizer);

#line 636 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_inner_organizer);

#line 636 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->inner_organizer);

#line 636 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_v_self->inner_organizer);

#line 636 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->inner_organizer = __pyx_v_inner_organizer;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":637
 *         super().__init__(center, left, contained, right, level=level, parent=parent)
 *         self.inner_organizer = inner_organizer
 *         self.organize()             # <<<<<<<<<<<<<<
 * 
 *     def set_organizer(self, organizer):
 */

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_organize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 637, __pyx_L1_error)

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = NULL;

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (likely(__pyx_t_2)) {

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_2);

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(function);

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 637, __pyx_L1_error)

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 637 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":634
 * 
 * cdef class IntervalTreeNode2D(IntervalTreeNode):
 *     def __init__(self, center, left, contained, right, level=0, parent=None, inner_organizer=None):             # <<<<<<<<<<<<<<
 *         super().__init__(center, left, contained, right, level=level, parent=parent)
 *         self.inner_organizer = inner_organizer
 */

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = -1;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 634 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":639
 *         self.organize()
 * 
 *     def set_organizer(self, organizer):             # <<<<<<<<<<<<<<
 *         self.inner_organizer = organizer
 *         if self.left is not None:
 */

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_3set_organizer(PyObject *__pyx_v_self, PyObject *__pyx_v_organizer); /*proto*/

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_2set_organizer[] = "IntervalTreeNode2D.set_organizer(self, organizer)";

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_3set_organizer(PyObject *__pyx_v_self, PyObject *__pyx_v_organizer) {

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("set_organizer (wrapper)", 0);
  __pyx_r = 
#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_2set_organizer(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self), ((PyObject *)__pyx_v_organizer));

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_2set_organizer(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, PyObject *__pyx_v_organizer) {

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_1;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_2;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("set_organizer", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":640
 * 
 *     def set_organizer(self, organizer):
 *         self.inner_organizer = organizer             # <<<<<<<<<<<<<<
 *         if self.left is not None:
 *             self.left.set_organizer(organizer)
 */

#line 640 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_organizer);

#line 640 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_v_organizer);

#line 640 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_v_self->inner_organizer);

#line 640 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_v_self->inner_organizer);

#line 640 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_self->inner_organizer = __pyx_v_organizer;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":641
 *     def set_organizer(self, organizer):
 *         self.inner_organizer = organizer
 *         if self.left is not None:             # <<<<<<<<<<<<<<
 *             self.left.set_organizer(organizer)
 *         if self.right is not None:
 */

#line 641 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (((PyObject *)__pyx_v_self->__pyx_base.left) != Py_None);

#line 641 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 641 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":642
 *         self.inner_organizer = organizer
 *         if self.left is not None:
 *             self.left.set_organizer(organizer)             # <<<<<<<<<<<<<<
 *         if self.right is not None:
 *             self.right.set_organizer(organizer)
 */

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.left), __pyx_n_s_set_organizer); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 642, __pyx_L1_error)

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = NULL;

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (likely(__pyx_t_5)) {

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_5);

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(function);

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_t_4, function);

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_organizer) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_organizer);

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 642, __pyx_L1_error)

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 642 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":641
 *     def set_organizer(self, organizer):
 *         self.inner_organizer = organizer
 *         if self.left is not None:             # <<<<<<<<<<<<<<
 *             self.left.set_organizer(organizer)
 *         if self.right is not None:
 */

#line 641 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":643
 *         if self.left is not None:
 *             self.left.set_organizer(organizer)
 *         if self.right is not None:             # <<<<<<<<<<<<<<
 *             self.right.set_organizer(organizer)
 *         self.organize()
 */

#line 643 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = (((PyObject *)__pyx_v_self->__pyx_base.right) != Py_None);

#line 643 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_t_2 != 0);

#line 643 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":644
 *             self.left.set_organizer(organizer)
 *         if self.right is not None:
 *             self.right.set_organizer(organizer)             # <<<<<<<<<<<<<<
 *         self.organize()
 * 
 */

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.right), __pyx_n_s_set_organizer); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 644, __pyx_L1_error)

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = NULL;

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (likely(__pyx_t_5)) {

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_5);

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(function);

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_t_4, function);

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_organizer) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_organizer);

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 644, __pyx_L1_error)

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 644 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":643
 *         if self.left is not None:
 *             self.left.set_organizer(organizer)
 *         if self.right is not None:             # <<<<<<<<<<<<<<
 *             self.right.set_organizer(organizer)
 *         self.organize()
 */

#line 643 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":645
 *         if self.right is not None:
 *             self.right.set_organizer(organizer)
 *         self.organize()             # <<<<<<<<<<<<<<
 * 
 *     def organize(self):
 */

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_organize); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 645, __pyx_L1_error)

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_5 = NULL;

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (likely(__pyx_t_5)) {

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_5);

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(function);

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_t_4, function);

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 645, __pyx_L1_error)

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 645 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":639
 *         self.organize()
 * 
 *     def set_organizer(self, organizer):             # <<<<<<<<<<<<<<
 *         self.inner_organizer = organizer
 *         if self.left is not None:
 */

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.set_organizer", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 639 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":647
 *         self.organize()
 * 
 *     def organize(self):             # <<<<<<<<<<<<<<
 *         if self.inner_organizer is not None:
 *             self.organized = self.inner_organizer(self.contained)
 */

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_5organize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_4organize[] = "IntervalTreeNode2D.organize(self)";

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_5organize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("organize (wrapper)", 0);
  __pyx_r = 
#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_4organize(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self));

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_4organize(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self) {

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_1;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_2;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("organize", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":648
 * 
 *     def organize(self):
 *         if self.inner_organizer is not None:             # <<<<<<<<<<<<<<
 *             self.organized = self.inner_organizer(self.contained)
 *         else:
 */

#line 648 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_v_self->inner_organizer != Py_None);

#line 648 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 648 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":649
 *     def organize(self):
 *         if self.inner_organizer is not None:
 *             self.organized = self.inner_organizer(self.contained)             # <<<<<<<<<<<<<<
 *         else:
 *             self.organized = None
 */

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_v_self->inner_organizer);
    __pyx_t_4 = __pyx_v_self->inner_organizer; 
#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_5 = NULL;

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (likely(__pyx_t_5)) {

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_5);

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(function);

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_t_4, function);

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_self->__pyx_base.contained) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self->__pyx_base.contained);

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 649, __pyx_L1_error)

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(0, 649, __pyx_L1_error)

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GIVEREF(__pyx_t_3);

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_v_self->organized);

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(((PyObject *)__pyx_v_self->organized));

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_self->organized = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_3);

#line 649 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":648
 * 
 *     def organize(self):
 *         if self.inner_organizer is not None:             # <<<<<<<<<<<<<<
 *             self.organized = self.inner_organizer(self.contained)
 *         else:
 */

#line 648 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L3;

#line 648 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":651
 *             self.organized = self.inner_organizer(self.contained)
 *         else:
 *             self.organized = None             # <<<<<<<<<<<<<<
 * 
 *     cpdef list _overlaps_interval_2d(self, double[::1] starts, double[::1] ends):
 */

#line 651 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /*else*/ {

#line 651 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(Py_None);

#line 651 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GIVEREF(Py_None);

#line 651 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_v_self->organized);

#line 651 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(((PyObject *)__pyx_v_self->organized));

#line 651 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_self->organized = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None);

#line 651 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 651 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":647
 *         self.organize()
 * 
 *     def organize(self):             # <<<<<<<<<<<<<<
 *         if self.inner_organizer is not None:
 *             self.organized = self.inner_organizer(self.contained)
 */

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.organize", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 647 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":653
 *             self.organized = None
 * 
 *     cpdef list _overlaps_interval_2d(self, double[::1] starts, double[::1] ends):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list overlaps_1d
 */

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_7_overlaps_interval_2d(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D__overlaps_interval_2d(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_starts, __Pyx_memviewslice __pyx_v_ends, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_overlaps_1d = 0
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v_interv2 = 0
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v_interv = 0
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  size_t __pyx_v_i
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  size_t __pyx_v_j
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  size_t __pyx_v_n
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  size_t __pyx_v_m
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v_query = NULL
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_result = NULL
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_9;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_10;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_11;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_12;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_13;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_14;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_15;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_16;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_17;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_18;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_19;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  size_t __pyx_t_20;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_21;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_overlaps_interval_2d", 0);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlaps_interval_2d); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 653, __pyx_L1_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_7_overlaps_interval_2d)) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_v_starts.memview)) { __Pyx_RaiseUnboundLocalError("starts"); __PYX_ERR(0, 653, __pyx_L1_error) }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_starts, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 653, __pyx_L1_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_v_ends.memview)) { __Pyx_RaiseUnboundLocalError("ends"); __PYX_ERR(0, 653, __pyx_L1_error) }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_ends, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 653, __pyx_L1_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; 
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_6 = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_7 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_6)) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_6);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_5, function);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_t_7 = 1;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #if CYTHON_FAST_PYCALL

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (PyFunction_Check(__pyx_t_5)) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L1_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        } else

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #endif

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #if CYTHON_FAST_PYCCALL

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L1_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        } else

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #endif

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 653, __pyx_L1_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_8);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (__pyx_t_6) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_3 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L1_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 653, __pyx_L1_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":659
 *             SpanningMixin interv
 *             size_t i, j, n, m
 *         query = SimpleInterval._create(starts[0], ends[0])             # <<<<<<<<<<<<<<
 *         result = []
 *         if self.organized is None:
 */

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_9 = 0;

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = -1;

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_9 < 0) {

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 += __pyx_v_starts.shape[0];

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_t_9 < 0)) __pyx_t_7 = 0;
  } else 
#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
if (unlikely(__pyx_t_9 >= __pyx_v_starts.shape[0])) __pyx_t_7 = 0;

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_t_7 != -1)) {

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_RaiseBufferIndexError(__pyx_t_7);

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __PYX_ERR(0, 659, __pyx_L1_error)

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_10 = 0;

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = -1;

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_10 < 0) {

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 += __pyx_v_ends.shape[0];

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_t_10 < 0)) __pyx_t_7 = 0;
  } else 
#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
if (unlikely(__pyx_t_10 >= __pyx_v_ends.shape[0])) __pyx_t_7 = 0;

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_t_7 != -1)) {

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_RaiseBufferIndexError(__pyx_t_7);

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __PYX_ERR(0, 659, __pyx_L1_error)

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval__create((*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_starts.data) + __pyx_t_9)) ))), (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_ends.data) + __pyx_t_10)) ))))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_query = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *)__pyx_t_1);

#line 659 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":660
 *             size_t i, j, n, m
 *         query = SimpleInterval._create(starts[0], ends[0])
 *         result = []             # <<<<<<<<<<<<<<
 *         if self.organized is None:
 *             return super()._overlaps_interval(starts[0], ends[0])
 */

#line 660 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 660, __pyx_L1_error)

#line 660 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 660 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_result = ((PyObject*)__pyx_t_1);

#line 660 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":661
 *         query = SimpleInterval._create(starts[0], ends[0])
 *         result = []
 *         if self.organized is None:             # <<<<<<<<<<<<<<
 *             return super()._overlaps_interval(starts[0], ends[0])
 *         else:
 */

#line 661 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_11 = (((PyObject *)__pyx_v_self->organized) == Py_None);

#line 661 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_12 = (__pyx_t_11 != 0);

#line 661 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_12) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":662
 *         result = []
 *         if self.organized is None:
 *             return super()._overlaps_interval(starts[0], ends[0])             # <<<<<<<<<<<<<<
 *         else:
 *             overlaps_1d = self.organized.overlaps(starts[1], ends[1])
 */

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_super); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_overlaps_interval); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = -1;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_10 < 0) {

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_10 += __pyx_v_starts.shape[0];

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_t_10 < 0)) __pyx_t_7 = 0;
    } else 
#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
if (unlikely(__pyx_t_10 >= __pyx_v_starts.shape[0])) __pyx_t_7 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_t_7 != -1)) {

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_7);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_2 = PyFloat_FromDouble((*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_starts.data) + __pyx_t_10)) )))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = -1;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_10 < 0) {

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_10 += __pyx_v_ends.shape[0];

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_t_10 < 0)) __pyx_t_7 = 0;
    } else 
#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
if (unlikely(__pyx_t_10 >= __pyx_v_ends.shape[0])) __pyx_t_7 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_t_7 != -1)) {

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_7);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_8 = PyFloat_FromDouble((*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_ends.data) + __pyx_t_10)) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_8);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_4 = NULL;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (likely(__pyx_t_4)) {

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_4);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(function);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF_SET(__pyx_t_5, function);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_7 = 1;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_FAST_PYCALL

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (PyFunction_Check(__pyx_t_5)) {

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_t_8};

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_FAST_PYCCALL

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_t_8};

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    {

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_3 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_3);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_4) {

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GIVEREF(__pyx_t_2);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_7, __pyx_t_2);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GIVEREF(__pyx_t_8);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_7, __pyx_t_8);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_2 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_8 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = ((PyObject*)__pyx_t_1);

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = 0;

#line 662 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":661
 *         query = SimpleInterval._create(starts[0], ends[0])
 *         result = []
 *         if self.organized is None:             # <<<<<<<<<<<<<<
 *             return super()._overlaps_interval(starts[0], ends[0])
 *         else:
 */

#line 661 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":664
 *             return super()._overlaps_interval(starts[0], ends[0])
 *         else:
 *             overlaps_1d = self.organized.overlaps(starts[1], ends[1])             # <<<<<<<<<<<<<<
 *             n = PyList_Size(overlaps_1d)
 *             for i in range(n):
 */

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /*else*/ {

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = 1;

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = -1;

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_10 < 0) {

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_10 += __pyx_v_starts.shape[0];

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_t_10 < 0)) __pyx_t_7 = 0;
    } else 
#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
if (unlikely(__pyx_t_10 >= __pyx_v_starts.shape[0])) __pyx_t_7 = 0;

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_t_7 != -1)) {

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_7);

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __PYX_ERR(0, 664, __pyx_L1_error)

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 = 1;

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_7 = -1;

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_9 < 0) {

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_9 += __pyx_v_ends.shape[0];

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_t_9 < 0)) __pyx_t_7 = 0;
    } else 
#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
if (unlikely(__pyx_t_9 >= __pyx_v_ends.shape[0])) __pyx_t_7 = 0;

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_t_7 != -1)) {

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_7);

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __PYX_ERR(0, 664, __pyx_L1_error)

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v_self->organized->__pyx_vtab)->overlaps(__pyx_v_self->organized, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_starts.data) + __pyx_t_10)) ))), (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_ends.data) + __pyx_t_9)) ))), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 664, __pyx_L1_error)

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_overlaps_1d = ((PyObject*)__pyx_t_1);

#line 664 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":665
 *         else:
 *             overlaps_1d = self.organized.overlaps(starts[1], ends[1])
 *             n = PyList_Size(overlaps_1d)             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 interv2 = <Interval>PyList_GetItem(overlaps_1d, i)
 */

#line 665 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_13 = PyList_Size(__pyx_v_overlaps_1d); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 665, __pyx_L1_error)

#line 665 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_n = __pyx_t_13;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":666
 *             overlaps_1d = self.organized.overlaps(starts[1], ends[1])
 *             n = PyList_Size(overlaps_1d)
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 interv2 = <Interval>PyList_GetItem(overlaps_1d, i)
 *                 m = PyList_Size(interv2.members)
 */

#line 666 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_14 = __pyx_v_n;

#line 666 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_15 = __pyx_t_14;

#line 666 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {

#line 666 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_i = __pyx_t_16;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":667
 *             n = PyList_Size(overlaps_1d)
 *             for i in range(n):
 *                 interv2 = <Interval>PyList_GetItem(overlaps_1d, i)             # <<<<<<<<<<<<<<
 *                 m = PyList_Size(interv2.members)
 *                 for j in range(m):
 */

#line 667 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_17 = PyList_GetItem(__pyx_v_overlaps_1d, __pyx_v_i); if (unlikely(__pyx_t_17 == ((PyObject *)NULL))) __PYX_ERR(0, 667, __pyx_L1_error)

#line 667 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_t_17);

#line 667 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 667 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF_SET(__pyx_v_interv2, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_t_1));

#line 667 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":668
 *             for i in range(n):
 *                 interv2 = <Interval>PyList_GetItem(overlaps_1d, i)
 *                 m = PyList_Size(interv2.members)             # <<<<<<<<<<<<<<
 *                 for j in range(m):
 *                     interv = <SpanningMixin>PyList_GetItem(interv2.members, j)
 */

#line 668 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __pyx_v_interv2->members;

#line 668 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 668 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_13 = PyList_Size(__pyx_t_1); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 668, __pyx_L1_error)

#line 668 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 668 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_v_m = __pyx_t_13;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":669
 *                 interv2 = <Interval>PyList_GetItem(overlaps_1d, i)
 *                 m = PyList_Size(interv2.members)
 *                 for j in range(m):             # <<<<<<<<<<<<<<
 *                     interv = <SpanningMixin>PyList_GetItem(interv2.members, j)
 *                     if interv.overlaps(query):
 */

#line 669 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_18 = __pyx_v_m;

#line 669 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_19 = __pyx_t_18;

#line 669 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {

#line 669 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_v_j = __pyx_t_20;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":670
 *                 m = PyList_Size(interv2.members)
 *                 for j in range(m):
 *                     interv = <SpanningMixin>PyList_GetItem(interv2.members, j)             # <<<<<<<<<<<<<<
 *                     if interv.overlaps(query):
 *                         result.append(interv)
 */

#line 670 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = __pyx_v_interv2->members;

#line 670 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);

#line 670 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_17 = PyList_GetItem(__pyx_t_1, __pyx_v_j); if (unlikely(__pyx_t_17 == ((PyObject *)NULL))) __PYX_ERR(0, 670, __pyx_L1_error)

#line 670 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 670 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = ((PyObject *)__pyx_t_17);

#line 670 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);

#line 670 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF_SET(__pyx_v_interv, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_t_1));

#line 670 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":671
 *                 for j in range(m):
 *                     interv = <SpanningMixin>PyList_GetItem(interv2.members, j)
 *                     if interv.overlaps(query):             # <<<<<<<<<<<<<<
 *                         result.append(interv)
 *         return result
 */

#line 671 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_12 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_interv->__pyx_vtab)->overlaps(__pyx_v_interv, ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v_query), 0) != 0);

#line 671 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (__pyx_t_12) {

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":672
 *                     interv = <SpanningMixin>PyList_GetItem(interv2.members, j)
 *                     if interv.overlaps(query):
 *                         result.append(interv)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */

#line 672 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_interv)); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 672, __pyx_L1_error)

          /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":671
 *                 for j in range(m):
 *                     interv = <SpanningMixin>PyList_GetItem(interv2.members, j)
 *                     if interv.overlaps(query):             # <<<<<<<<<<<<<<
 *                         result.append(interv)
 *         return result
 */

#line 671 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 671 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 671 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 671 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":673
 *                     if interv.overlaps(query):
 *                         result.append(interv)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef list overlaps_2d(self, double[::1] starts, double[::1] ends):
 */

#line 673 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 673 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_result);

#line 673 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_v_result;

#line 673 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":653
 *             self.organized = None
 * 
 *     cpdef list _overlaps_interval_2d(self, double[::1] starts, double[::1] ends):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list overlaps_1d
 */

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D._overlaps_interval_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_overlaps_1d);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_interv2);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_interv);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_query);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_7_overlaps_interval_2d(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_6_overlaps_interval_2d[] = "IntervalTreeNode2D._overlaps_interval_2d(self, double[::1] starts, double[::1] ends) -> list";

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_7_overlaps_interval_2d(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_starts = { 0, 0, { 0 }, { 0 }, { 0 } }
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  __Pyx_memviewslice __pyx_v_ends = { 0, 0, { 0 }, { 0 }, { 0 } }
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_overlaps_interval_2d (wrapper)", 0);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_starts,&__pyx_n_s_ends,0};

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[2] = {0,0};

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  2: 
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_starts)) != 0)) kw_args--;
        else 
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ends)) != 0)) kw_args--;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_overlaps_interval_2d", 1, 2, 2, 1); 
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 653, __pyx_L3_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_overlaps_interval_2d") < 0)) __PYX_ERR(0, 653, __pyx_L3_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_argtuple_error;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_starts = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_starts.memview)) __PYX_ERR(0, 653, __pyx_L3_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_ends = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_ends.memview)) __PYX_ERR(0, 653, __pyx_L3_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_overlaps_interval_2d", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 653, __pyx_L3_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D._overlaps_interval_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_6_overlaps_interval_2d(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self), __pyx_v_starts, __pyx_v_ends);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_6_overlaps_interval_2d(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_starts, __Pyx_memviewslice __pyx_v_ends) {

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("_overlaps_interval_2d", 0);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_v_starts.memview)) { __Pyx_RaiseUnboundLocalError("starts"); __PYX_ERR(0, 653, __pyx_L1_error) }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_v_ends.memview)) { __Pyx_RaiseUnboundLocalError("ends"); __PYX_ERR(0, 653, __pyx_L1_error) }

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D__overlaps_interval_2d(__pyx_v_self, __pyx_v_starts, __pyx_v_ends, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 653, __pyx_L1_error)

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D._overlaps_interval_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_v_starts, 1);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_v_ends, 1);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 653 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":675
 *         return result
 * 
 *     cpdef list overlaps_2d(self, double[::1] starts, double[::1] ends):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double start, end
 */

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9overlaps_2d(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_overlaps_2d(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_starts, __Pyx_memviewslice __pyx_v_ends, int __pyx_skip_dispatch) {
  double __pyx_v_start
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  double __pyx_v_end
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_result = 0
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_7;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  Py_ssize_t __pyx_t_9;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_10;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_11;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_12;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_t_13;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps_2d", 0);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if called by wrapper */

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* Check if overridden in Python */

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlaps_2d); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9overlaps_2d)) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_v_starts.memview)) { __Pyx_RaiseUnboundLocalError("starts"); __PYX_ERR(0, 675, __pyx_L1_error) }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_starts, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 675, __pyx_L1_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(!__pyx_v_ends.memview)) { __Pyx_RaiseUnboundLocalError("ends"); __PYX_ERR(0, 675, __pyx_L1_error) }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_ends, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 675, __pyx_L1_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; 
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_t_6 = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_7 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (likely(__pyx_t_6)) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(__pyx_t_6);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_INCREF(function);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_DECREF_SET(__pyx_t_5, function);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __pyx_t_7 = 1;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #if CYTHON_FAST_PYCALL

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (PyFunction_Check(__pyx_t_5)) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        } else

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #endif

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #if CYTHON_FAST_PYCCALL

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        } else

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        #endif

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 675, __pyx_L1_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_8);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (__pyx_t_6) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_3 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_4 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 675, __pyx_L1_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_2 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #endif

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    #endif

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":680
 *             list result
 * 
 *         start = starts[0]             # <<<<<<<<<<<<<<
 *         end = ends[0]
 *         result = []
 */

#line 680 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_9 = 0;

#line 680 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = -1;

#line 680 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_9 < 0) {

#line 680 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 += __pyx_v_starts.shape[0];

#line 680 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_t_9 < 0)) __pyx_t_7 = 0;
  } else 
#line 680 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
if (unlikely(__pyx_t_9 >= __pyx_v_starts.shape[0])) __pyx_t_7 = 0;

#line 680 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_t_7 != -1)) {

#line 680 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_RaiseBufferIndexError(__pyx_t_7);

#line 680 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __PYX_ERR(0, 680, __pyx_L1_error)

#line 680 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 680 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_start = (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_starts.data) + __pyx_t_9)) )));

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":681
 * 
 *         start = starts[0]
 *         end = ends[0]             # <<<<<<<<<<<<<<
 *         result = []
 *         if start > self.end:
 */

#line 681 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_9 = 0;

#line 681 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_7 = -1;

#line 681 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_9 < 0) {

#line 681 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_9 += __pyx_v_ends.shape[0];

#line 681 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_t_9 < 0)) __pyx_t_7 = 0;
  } else 
#line 681 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
if (unlikely(__pyx_t_9 >= __pyx_v_ends.shape[0])) __pyx_t_7 = 0;

#line 681 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(__pyx_t_7 != -1)) {

#line 681 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_RaiseBufferIndexError(__pyx_t_7);

#line 681 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __PYX_ERR(0, 681, __pyx_L1_error)

#line 681 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 681 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_end = (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_ends.data) + __pyx_t_9)) )));

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":682
 *         start = starts[0]
 *         end = ends[0]
 *         result = []             # <<<<<<<<<<<<<<
 *         if start > self.end:
 *             return result
 */

#line 682 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 682, __pyx_L1_error)

#line 682 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 682 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_result = ((PyObject*)__pyx_t_1);

#line 682 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":683
 *         end = ends[0]
 *         result = []
 *         if start > self.end:             # <<<<<<<<<<<<<<
 *             return result
 *         if end < self.start:
 */

#line 683 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_10 = ((__pyx_v_start > __pyx_v_self->__pyx_base.end) != 0);

#line 683 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":684
 *         result = []
 *         if start > self.end:
 *             return result             # <<<<<<<<<<<<<<
 *         if end < self.start:
 *             return result
 */

#line 684 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 684 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_v_result);

#line 684 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = __pyx_v_result;

#line 684 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":683
 *         end = ends[0]
 *         result = []
 *         if start > self.end:             # <<<<<<<<<<<<<<
 *             return result
 *         if end < self.start:
 */

#line 683 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":685
 *         if start > self.end:
 *             return result
 *         if end < self.start:             # <<<<<<<<<<<<<<
 *             return result
 *         elif start <= self.start:
 */

#line 685 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_10 = ((__pyx_v_end < __pyx_v_self->__pyx_base.start) != 0);

#line 685 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":686
 *             return result
 *         if end < self.start:
 *             return result             # <<<<<<<<<<<<<<
 *         elif start <= self.start:
 *             if end < self.start:
 */

#line 686 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 686 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_INCREF(__pyx_v_result);

#line 686 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_r = __pyx_v_result;

#line 686 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":685
 *         if start > self.end:
 *             return result
 *         if end < self.start:             # <<<<<<<<<<<<<<
 *             return result
 *         elif start <= self.start:
 */

#line 685 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":687
 *         if end < self.start:
 *             return result
 *         elif start <= self.start:             # <<<<<<<<<<<<<<
 *             if end < self.start:
 *                 return result
 */

#line 687 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_10 = ((__pyx_v_start <= __pyx_v_self->__pyx_base.start) != 0);

#line 687 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":688
 *             return result
 *         elif start <= self.start:
 *             if end < self.start:             # <<<<<<<<<<<<<<
 *                 return result
 *             else:
 */

#line 688 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = ((__pyx_v_end < __pyx_v_self->__pyx_base.start) != 0);

#line 688 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":689
 *         elif start <= self.start:
 *             if end < self.start:
 *                 return result             # <<<<<<<<<<<<<<
 *             else:
 *                 if self.left is not None:
 */

#line 689 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 689 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_v_result);

#line 689 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_r = __pyx_v_result;

#line 689 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":688
 *             return result
 *         elif start <= self.start:
 *             if end < self.start:             # <<<<<<<<<<<<<<
 *                 return result
 *             else:
 */

#line 688 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":691
 *                 return result
 *             else:
 *                 if self.left is not None:             # <<<<<<<<<<<<<<
 *                     result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *                 result.extend(self._overlaps_interval_2d(starts, ends))
 */

#line 691 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    /*else*/ {

#line 691 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_10 = (((PyObject *)__pyx_v_self->__pyx_base.left) != Py_None);

#line 691 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = (__pyx_t_10 != 0);

#line 691 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_11) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":692
 *             else:
 *                 if self.left is not None:
 *                     result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))             # <<<<<<<<<<<<<<
 *                 result.extend(self._overlaps_interval_2d(starts, ends))
 *                 if self.right is not None and end >= self.right.start:
 */

#line 692 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.left)->__pyx_base.__pyx_vtab)->overlaps_2d(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.left), __pyx_v_starts, __pyx_v_ends, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 692, __pyx_L1_error)

#line 692 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 692 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_12 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 692, __pyx_L1_error)

#line 692 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":691
 *                 return result
 *             else:
 *                 if self.left is not None:             # <<<<<<<<<<<<<<
 *                     result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *                 result.extend(self._overlaps_interval_2d(starts, ends))
 */

#line 691 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":693
 *                 if self.left is not None:
 *                     result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *                 result.extend(self._overlaps_interval_2d(starts, ends))             # <<<<<<<<<<<<<<
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 */

#line 693 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.__pyx_vtab)->_overlaps_interval_2d(__pyx_v_self, __pyx_v_starts, __pyx_v_ends, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)

#line 693 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 693 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_12 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 693, __pyx_L1_error)

#line 693 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":694
 *                     result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *                 result.extend(self._overlaps_interval_2d(starts, ends))
 *                 if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                     result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif start > self.start:
 */

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_10 = (((PyObject *)__pyx_v_self->__pyx_base.right) != Py_None);

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_13 = (__pyx_t_10 != 0);

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_13) {

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      } else {

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_11 = __pyx_t_13;

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        goto __pyx_L8_bool_binop_done;

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_13 = ((__pyx_v_end >= __pyx_v_self->__pyx_base.right->start) != 0);

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = __pyx_t_13;

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_L8_bool_binop_done:;

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (__pyx_t_11) {

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":695
 *                 result.extend(self._overlaps_interval_2d(starts, ends))
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))             # <<<<<<<<<<<<<<
 *         elif start > self.start:
 *             if (<IntervalTreeNode2D>self.left) is not None and (<IntervalTreeNode2D>self.left).end >= start:
 */

#line 695 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.right)->__pyx_base.__pyx_vtab)->overlaps_2d(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.right), __pyx_v_starts, __pyx_v_ends, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 695, __pyx_L1_error)

#line 695 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 695 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __pyx_t_12 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 695, __pyx_L1_error)

#line 695 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":694
 *                     result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *                 result.extend(self._overlaps_interval_2d(starts, ends))
 *                 if self.right is not None and end >= self.right.start:             # <<<<<<<<<<<<<<
 *                     result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif start > self.start:
 */

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 694 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":687
 *         if end < self.start:
 *             return result
 *         elif start <= self.start:             # <<<<<<<<<<<<<<
 *             if end < self.start:
 *                 return result
 */

#line 687 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L4;

#line 687 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":696
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif start > self.start:             # <<<<<<<<<<<<<<
 *             if (<IntervalTreeNode2D>self.left) is not None and (<IntervalTreeNode2D>self.left).end >= start:
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 */

#line 696 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_11 = ((__pyx_v_start > __pyx_v_self->__pyx_base.start) != 0);

#line 696 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_11) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":697
 *                     result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif start > self.start:
 *             if (<IntervalTreeNode2D>self.left) is not None and (<IntervalTreeNode2D>self.left).end >= start:             # <<<<<<<<<<<<<<
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 */

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_13 = (((PyObject *)__pyx_v_self->__pyx_base.left) != Py_None);

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = (__pyx_t_13 != 0);

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_10) {

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = __pyx_t_10;

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L11_bool_binop_done;

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = ((((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.left)->__pyx_base.end >= __pyx_v_start) != 0);

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_11 = __pyx_t_10;

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L11_bool_binop_done:;

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_11) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":698
 *         elif start > self.start:
 *             if (<IntervalTreeNode2D>self.left) is not None and (<IntervalTreeNode2D>self.left).end >= start:
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))             # <<<<<<<<<<<<<<
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:
 */

#line 698 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.left)->__pyx_base.__pyx_vtab)->overlaps_2d(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.left), __pyx_v_starts, __pyx_v_ends, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)

#line 698 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 698 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_12 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 698, __pyx_L1_error)

#line 698 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":697
 *                     result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif start > self.start:
 *             if (<IntervalTreeNode2D>self.left) is not None and (<IntervalTreeNode2D>self.left).end >= start:             # <<<<<<<<<<<<<<
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 */

#line 697 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":699
 *             if (<IntervalTreeNode2D>self.left) is not None and (<IntervalTreeNode2D>self.left).end >= start:
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *             result.extend(self._overlaps_interval_2d(starts, ends))             # <<<<<<<<<<<<<<
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 */

#line 699 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.__pyx_vtab)->_overlaps_interval_2d(__pyx_v_self, __pyx_v_starts, __pyx_v_ends, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 699, __pyx_L1_error)

#line 699 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 699 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_12 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 699, __pyx_L1_error)

#line 699 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":700
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:             # <<<<<<<<<<<<<<
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif end > self.start:
 */

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = (((PyObject *)__pyx_v_self->__pyx_base.right) != Py_None);

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_13 = (__pyx_t_10 != 0);

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_13) {

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_11 = __pyx_t_13;

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L14_bool_binop_done;

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_13 = ((__pyx_v_end >= ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.right)->__pyx_base.start) != 0);

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_11 = __pyx_t_13;

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L14_bool_binop_done:;

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_11) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":701
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))             # <<<<<<<<<<<<<<
 *         elif end > self.start:
 *             if (<IntervalTreeNode2D>self.left) is not None:
 */

#line 701 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.right)->__pyx_base.__pyx_vtab)->overlaps_2d(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.right), __pyx_v_starts, __pyx_v_ends, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 701, __pyx_L1_error)

#line 701 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 701 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_12 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 701, __pyx_L1_error)

#line 701 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":700
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:             # <<<<<<<<<<<<<<
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif end > self.start:
 */

#line 700 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":696
 *                 if self.right is not None and end >= self.right.start:
 *                     result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif start > self.start:             # <<<<<<<<<<<<<<
 *             if (<IntervalTreeNode2D>self.left) is not None and (<IntervalTreeNode2D>self.left).end >= start:
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 */

#line 696 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    goto __pyx_L4;

#line 696 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":702
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif end > self.start:             # <<<<<<<<<<<<<<
 *             if (<IntervalTreeNode2D>self.left) is not None:
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 */

#line 702 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_11 = ((__pyx_v_end > __pyx_v_self->__pyx_base.start) != 0);

#line 702 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (__pyx_t_11) {

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":703
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif end > self.start:
 *             if (<IntervalTreeNode2D>self.left) is not None:             # <<<<<<<<<<<<<<
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 */

#line 703 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.left) != Py_None);

#line 703 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_13 = (__pyx_t_11 != 0);

#line 703 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_13) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":704
 *         elif end > self.start:
 *             if (<IntervalTreeNode2D>self.left) is not None:
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))             # <<<<<<<<<<<<<<
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:
 */

#line 704 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.left)->__pyx_base.__pyx_vtab)->overlaps_2d(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.left), __pyx_v_starts, __pyx_v_ends, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)

#line 704 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 704 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_12 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 704, __pyx_L1_error)

#line 704 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":703
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif end > self.start:
 *             if (<IntervalTreeNode2D>self.left) is not None:             # <<<<<<<<<<<<<<
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 */

#line 703 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":705
 *             if (<IntervalTreeNode2D>self.left) is not None:
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *             result.extend(self._overlaps_interval_2d(starts, ends))             # <<<<<<<<<<<<<<
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 */

#line 705 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.__pyx_vtab)->_overlaps_interval_2d(__pyx_v_self, __pyx_v_starts, __pyx_v_ends, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 705, __pyx_L1_error)

#line 705 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 705 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_12 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 705, __pyx_L1_error)

#line 705 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":706
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:             # <<<<<<<<<<<<<<
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         return result
 */

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.right) != Py_None);

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = (__pyx_t_11 != 0);

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_10) {

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_13 = __pyx_t_10;

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L18_bool_binop_done;

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_10 = ((__pyx_v_end >= ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.right)->__pyx_base.start) != 0);

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_13 = __pyx_t_10;

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_L18_bool_binop_done:;

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (__pyx_t_13) {

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":707
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))             # <<<<<<<<<<<<<<
 *         return result
 * 
 */

#line 707 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.right)->__pyx_base.__pyx_vtab)->overlaps_2d(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self->__pyx_base.right), __pyx_v_starts, __pyx_v_ends, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)

#line 707 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 707 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __pyx_t_12 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 707, __pyx_L1_error)

#line 707 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":706
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 *             result.extend(self._overlaps_interval_2d(starts, ends))
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:             # <<<<<<<<<<<<<<
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         return result
 */

#line 706 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

    /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":702
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         elif end > self.start:             # <<<<<<<<<<<<<<
 *             if (<IntervalTreeNode2D>self.left) is not None:
 *                 result.extend((<IntervalTreeNode2D>self.left).overlaps_2d(starts, ends))
 */

#line 702 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 702 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4:;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":708
 *             if (<IntervalTreeNode2D>self.right) is not None and end >= (<IntervalTreeNode2D>self.right).start:
 *                 result.extend((<IntervalTreeNode2D>self.right).overlaps_2d(starts, ends))
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */

#line 708 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 708 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_v_result);

#line 708 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_v_result;

#line 708 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":675
 *         return result
 * 
 *     cpdef list overlaps_2d(self, double[::1] starts, double[::1] ends):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double start, end
 */

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.overlaps_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9overlaps_2d(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_8overlaps_2d[] = "IntervalTreeNode2D.overlaps_2d(self, double[::1] starts, double[::1] ends) -> list";

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9overlaps_2d(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_starts = { 0, 0, { 0 }, { 0 }, { 0 } }
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  __Pyx_memviewslice __pyx_v_ends = { 0, 0, { 0 }, { 0 }, { 0 } }
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps_2d (wrapper)", 0);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_starts,&__pyx_n_s_ends,0};

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[2] = {0,0};

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  2: 
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_starts)) != 0)) kw_args--;
        else 
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ends)) != 0)) kw_args--;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("overlaps_2d", 1, 2, 2, 1); 
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 675, __pyx_L3_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "overlaps_2d") < 0)) __PYX_ERR(0, 675, __pyx_L3_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      goto __pyx_L5_argtuple_error;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_starts = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_starts.memview)) __PYX_ERR(0, 675, __pyx_L3_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_ends = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_ends.memview)) __PYX_ERR(0, 675, __pyx_L3_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("overlaps_2d", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 675, __pyx_L3_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.overlaps_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_8overlaps_2d(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self), __pyx_v_starts, __pyx_v_ends);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_8overlaps_2d(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, __Pyx_memviewslice __pyx_v_starts, __Pyx_memviewslice __pyx_v_ends) {

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("overlaps_2d", 0);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_v_starts.memview)) { __Pyx_RaiseUnboundLocalError("starts"); __PYX_ERR(0, 675, __pyx_L1_error) }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_v_ends.memview)) { __Pyx_RaiseUnboundLocalError("ends"); __PYX_ERR(0, 675, __pyx_L1_error) }

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_overlaps_2d(__pyx_v_self, __pyx_v_starts, __pyx_v_ends, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = __pyx_t_1;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.overlaps_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_v_starts, 1);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_v_ends, 1);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 675 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":711
 * 
 *     @classmethod
 *     def build(cls, intervals, organizer_callback=None):             # <<<<<<<<<<<<<<
 *         root: IntervalTreeNode2D = super(IntervalTreeNode2D, cls).build(intervals)
 *         root.set_organizer(organizer_callback)
 */

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
/* Python wrapper */

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_11build(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_10build[] = "IntervalTreeNode2D.build(type cls, intervals, organizer_callback=None)";

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_11build(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_intervals = 0
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;
  PyObject *__pyx_v_organizer_callback = 0
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = 0;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("build (wrapper)", 0);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  {

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_intervals,&__pyx_n_s_organizer_callback,0};

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    PyObject* values[2] = {0,0};

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    values[1] = ((PyObject *)Py_None);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (unlikely(__pyx_kwds)) {

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      Py_ssize_t kw_args;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {
        case  2: 
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0: break;
        default: 
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (pos_args) {

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  0:

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_intervals)) != 0)) kw_args--;
        else 
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        case  1:

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (kw_args > 0) {

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_organizer_callback);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        }

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      if (unlikely(kw_args > 0)) {

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "build") < 0)) __PYX_ERR(0, 711, __pyx_L3_error)

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    } else {

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: 
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
        break;
        default: 
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
goto __pyx_L5_argtuple_error;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      }

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_intervals = values[0];

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_v_organizer_callback = values[1];

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("build", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__PYX_ERR(0, 711, __pyx_L3_error)

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L3_error:;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.build", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return NULL;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_10build(((PyTypeObject*)__pyx_v_cls), __pyx_v_intervals, __pyx_v_organizer_callback);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_10build(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_intervals, PyObject *__pyx_v_organizer_callback) {
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_root = 0
#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_r = NULL;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyDeclarations

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_lineno = 0;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  const char *__pyx_filename = NULL;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  int __pyx_clineno = 0;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannySetupContext("build", 0);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":712
 *     @classmethod
 *     def build(cls, intervals, organizer_callback=None):
 *         root: IntervalTreeNode2D = super(IntervalTreeNode2D, cls).build(intervals)             # <<<<<<<<<<<<<<
 *         root.set_organizer(organizer_callback)
 *         return root
 */

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 712, __pyx_L1_error)

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D));

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D));

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D));

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_cls));

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_cls));

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_cls));

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 712, __pyx_L1_error)

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_build); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 712, __pyx_L1_error)

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = NULL;

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (likely(__pyx_t_3)) {

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(function);

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_intervals) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_intervals);

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 712, __pyx_L1_error)

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D))))) __PYX_ERR(0, 712, __pyx_L1_error)

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_root = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_t_1);

#line 712 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":713
 *     def build(cls, intervals, organizer_callback=None):
 *         root: IntervalTreeNode2D = super(IntervalTreeNode2D, cls).build(intervals)
 *         root.set_organizer(organizer_callback)             # <<<<<<<<<<<<<<
 *         return root
 */

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_root), __pyx_n_s_set_organizer); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 713, __pyx_L1_error)

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = NULL;

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    if (likely(__pyx_t_3)) {

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_INCREF(function);

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
    }

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  }

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_organizer_callback) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_organizer_callback);

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 713, __pyx_L1_error)

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 713 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":714
 *         root: IntervalTreeNode2D = super(IntervalTreeNode2D, cls).build(intervals)
 *         root.set_organizer(organizer_callback)
 *         return root             # <<<<<<<<<<<<<<
 */

#line 714 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 714 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_root));

#line 714 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = ((PyObject *)__pyx_v_root);

#line 714 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":711
 * 
 *     @classmethod
 *     def build(cls, intervals, organizer_callback=None):             # <<<<<<<<<<<<<<
 *         root: IntervalTreeNode2D = super(IntervalTreeNode2D, cls).build(intervals)
 *         root.set_organizer(organizer_callback)
 */

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"


#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  /* function exit code */

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L1_error:;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.build", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_r = NULL;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_L0:;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_root);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_RefNannyFinishContext();

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  return __pyx_r;

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":64
 * cdef class IntervalTreeNode2D(IntervalTreeNode):
 *     cdef:
 *         public object inner_organizer             # <<<<<<<<<<<<<<
 *         public IntervalTreeNode organized
 * 
 */

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_1__get__(PyObject *__pyx_v_self) {

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self));

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self) {

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_v_self->inner_organizer);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = __pyx_v_self->inner_organizer;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, PyObject *__pyx_v_value) {

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_v_value);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(__pyx_v_value);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->inner_organizer);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->inner_organizer);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->inner_organizer = __pyx_v_value;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_5__del__(PyObject *__pyx_v_self) {

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self));

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self) {

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(Py_None);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(Py_None);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->inner_organizer);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(__pyx_v_self->inner_organizer);

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->inner_organizer = Py_None;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 64 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "ms_deisotope/_c/peak_dependency_network/intervals.pxd":65
 *     cdef:
 *         public object inner_organizer
 *         public IntervalTreeNode organized             # <<<<<<<<<<<<<<
 * 
 *     cpdef list _overlaps_interval_2d(self, double[::1] starts, double[::1] ends)
 */

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_1__get__(PyObject *__pyx_v_self) {

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = 0;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized___get__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self));

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized___get__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self) {

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_r = NULL;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->organized));

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->organized);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_2__set__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, PyObject *__pyx_v_value) {

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_lineno = 0;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  const char *__pyx_filename = NULL;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_clineno = 0;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(1, 65, __pyx_L1_error)

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->organized);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->organized));

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->organized = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_t_1 = 0;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  goto __pyx_L0;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L1_error:;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.organized.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = -1;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_L0:;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
/* Python wrapper */

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_5__del__(PyObject *__pyx_v_self) {

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self));

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
static int __pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_4__del__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self) {

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  int __pyx_r;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyDeclarations

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_INCREF(Py_None);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GIVEREF(Py_None);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_GOTREF(__pyx_v_self->organized);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->organized));

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_v_self->organized = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None);

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"


#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  /* function exit code */

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __pyx_r = 0;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  __Pyx_RefNannyFinishContext();

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
  return __pyx_r;

#line 65 "ms_deisotope/_c/peak_dependency_network/intervals.pxd"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_12__reduce_cython__[] = "IntervalTreeNode2D.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_12__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_12__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_6;

#line 1 "(tree fragment)"
  int __pyx_t_7;

#line 1 "(tree fragment)"
  int __pyx_t_8;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center, self.contained, self.end, self.inner_organizer, self.left, self.level, self.organized, self.parent, self.right, self.start)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.center); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.end); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->__pyx_base.level); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.start); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
  __pyx_t_5 = PyTuple_New(10); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_5);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->__pyx_base.contained);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.contained);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self->__pyx_base.contained);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->inner_organizer);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->inner_organizer);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_v_self->inner_organizer);

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.left));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.left));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 4, ((PyObject *)__pyx_v_self->__pyx_base.left));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_3);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_t_3);

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->organized));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->organized));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 6, ((PyObject *)__pyx_v_self->organized));

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.parent));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.parent));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 7, ((PyObject *)__pyx_v_self->__pyx_base.parent));

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.right));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.right));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 8, ((PyObject *)__pyx_v_self->__pyx_base.right));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_4);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_5, 9, __pyx_t_4);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

#line 5 "(tree fragment)"
  __pyx_t_4 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_5);

#line 5 "(tree fragment)"
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center, self.contained, self.end, self.inner_organizer, self.left, self.level, self.organized, self.parent, self.right, self.start)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_5;

#line 6 "(tree fragment)"
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.center, self.contained, self.end, self.inner_organizer, self.left, self.level, self.organized, self.parent, self.right, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_6 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_7 = (__pyx_t_6 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_7) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));

#line 8 "(tree fragment)"
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.contained is not None or self.inner_organizer is not None or self.left is not None or self.organized is not None or self.parent is not None or self.right is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center, self.contained, self.end, self.inner_organizer, self.left, self.level, self.organized, self.parent, self.right, self.start)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.contained is not None or self.inner_organizer is not None or self.left is not None or self.organized is not None or self.parent is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IntervalTreeNode2D, (type(self), 0x0745b49, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_6 = (__pyx_v_self->__pyx_base.contained != ((PyObject*)Py_None));

#line 11 "(tree fragment)"
    __pyx_t_8 = (__pyx_t_6 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_8) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_7 = __pyx_t_8;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_8 = (__pyx_v_self->inner_organizer != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_6 = (__pyx_t_8 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_6) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_7 = __pyx_t_6;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_6 = (((PyObject *)__pyx_v_self->__pyx_base.left) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_8 = (__pyx_t_6 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_8) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_7 = __pyx_t_8;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_8 = (((PyObject *)__pyx_v_self->organized) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_6 = (__pyx_t_8 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_6) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_7 = __pyx_t_6;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_6 = (((PyObject *)__pyx_v_self->__pyx_base.parent) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_8 = (__pyx_t_6 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_8) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_7 = __pyx_t_8;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_8 = (((PyObject *)__pyx_v_self->__pyx_base.right) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_6 = (__pyx_t_8 != 0);

#line 11 "(tree fragment)"
    __pyx_t_7 = __pyx_t_6;

#line 11 "(tree fragment)"
    __pyx_L4_bool_binop_done:;

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_7;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.contained is not None or self.inner_organizer is not None or self.left is not None or self.organized is not None or self.parent is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntervalTreeNode2D, (type(self), 0x0745b49, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_7 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_7) {

    /* "(tree fragment)":13
 *         use_setstate = self.contained is not None or self.inner_organizer is not None or self.left is not None or self.organized is not None or self.parent is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IntervalTreeNode2D, (type(self), 0x0745b49, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode2D, (type(self), 0x0745b49, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_IntervalTreeNode2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_7625545);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_7625545);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_7625545);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_5);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_t_5 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_3;

#line 13 "(tree fragment)"
    __pyx_t_3 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.contained is not None or self.inner_organizer is not None or self.left is not None or self.organized is not None or self.parent is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntervalTreeNode2D, (type(self), 0x0745b49, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IntervalTreeNode2D, (type(self), 0x0745b49, None), state
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode2D, (type(self), 0x0745b49, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntervalTreeNode2D__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_IntervalTreeNode2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 15 "(tree fragment)"
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_7625545);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_7625545);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_7625545);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_5);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);

#line 15 "(tree fragment)"
    __pyx_t_3 = 0;

#line 15 "(tree fragment)"
    __pyx_t_5 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_4;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode2D, (type(self), 0x0745b49, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntervalTreeNode2D__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_14__setstate_cython__[] = "IntervalTreeNode2D.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_14__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_14__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IntervalTreeNode2D, (type(self), 0x0745b49, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntervalTreeNode2D__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode2D__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntervalTreeNode2D, (type(self), 0x0745b49, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntervalTreeNode2D__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_SpanningMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_5__pyx_unpickle_SpanningMixin(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_4__pyx_unpickle_SpanningMixin[] = "__pyx_unpickle_SpanningMixin(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_5__pyx_unpickle_SpanningMixin = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_SpanningMixin", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_5__pyx_unpickle_SpanningMixin, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_4__pyx_unpickle_SpanningMixin};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_5__pyx_unpickle_SpanningMixin(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpanningMixin (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpanningMixin", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpanningMixin", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_SpanningMixin") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpanningMixin", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SpanningMixin", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_4__pyx_unpickle_SpanningMixin(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_4__pyx_unpickle_SpanningMixin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpanningMixin", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xccc3210, 0x0576305, 0x67094f3):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple_, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xccc3210, 0x0576305, 0x67094f3):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xccc3210, 0x0576305, 0x67094f3):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xccc3210, 0x0576305, 0x67094f3):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin__set_state(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_SpanningMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SpanningMixin", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  double __pyx_t_2;

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpanningMixin__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->end = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->start = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = ((__pyx_t_4 > 2) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_5) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_3 = __pyx_t_5;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 13 "(tree fragment)"
  __pyx_t_3 = __pyx_t_6;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_9 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {

#line 14 "(tree fragment)"
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_9)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_8, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SpanningMixin__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_SimpleInterval(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_7__pyx_unpickle_SimpleInterval(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_6__pyx_unpickle_SimpleInterval[] = "__pyx_unpickle_SimpleInterval(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_7__pyx_unpickle_SimpleInterval = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_SimpleInterval", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_7__pyx_unpickle_SimpleInterval, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_6__pyx_unpickle_SimpleInterval};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_7__pyx_unpickle_SimpleInterval(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SimpleInterval (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SimpleInterval", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SimpleInterval", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_SimpleInterval") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SimpleInterval", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SimpleInterval", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_6__pyx_unpickle_SimpleInterval(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_6__pyx_unpickle_SimpleInterval(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SimpleInterval", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xccc3210, 0x0576305, 0x67094f3):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple_, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xccc3210, 0x0576305, 0x67094f3):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 *     __pyx_result = SimpleInterval.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xccc3210, 0x0576305, 0x67094f3):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = SimpleInterval.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xccc3210, 0x0576305, 0x67094f3):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 *     __pyx_result = SimpleInterval.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SimpleInterval__set_state(<SimpleInterval> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 *     __pyx_result = SimpleInterval.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SimpleInterval__set_state(<SimpleInterval> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = SimpleInterval.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SimpleInterval__set_state(<SimpleInterval> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_SimpleInterval__set_state(SimpleInterval __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SimpleInterval__set_state(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 *     __pyx_result = SimpleInterval.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SimpleInterval__set_state(<SimpleInterval> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SimpleInterval__set_state(<SimpleInterval> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_SimpleInterval__set_state(SimpleInterval __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_SimpleInterval(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SimpleInterval", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_SimpleInterval__set_state(<SimpleInterval> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SimpleInterval__set_state(SimpleInterval __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SimpleInterval__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  double __pyx_t_2;

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SimpleInterval__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_SimpleInterval__set_state(SimpleInterval __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.end = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.start = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_SimpleInterval__set_state(SimpleInterval __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = ((__pyx_t_4 > 2) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_5) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_3 = __pyx_t_5;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 13 "(tree fragment)"
  __pyx_t_3 = __pyx_t_6;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_9 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {

#line 14 "(tree fragment)"
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_9)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_8, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_SimpleInterval__set_state(SimpleInterval __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_SimpleInterval__set_state(<SimpleInterval> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SimpleInterval__set_state(SimpleInterval __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SimpleInterval__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Interval(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_9__pyx_unpickle_Interval(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8__pyx_unpickle_Interval[] = "__pyx_unpickle_Interval(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_9__pyx_unpickle_Interval = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_Interval", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_9__pyx_unpickle_Interval, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8__pyx_unpickle_Interval};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_9__pyx_unpickle_Interval(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Interval (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Interval", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Interval", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Interval") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Interval", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_Interval", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8__pyx_unpickle_Interval(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8__pyx_unpickle_Interval(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Interval", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xeace574, 0x24a0b9b, 0x4c97279):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeace574, 0x24a0b9b, 0x4c97279) = (data, end, members, start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__2, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xeace574, 0x24a0b9b, 0x4c97279):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeace574, 0x24a0b9b, 0x4c97279) = (data, end, members, start))" % __pyx_checksum)
 *     __pyx_result = Interval.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xeace574, 0x24a0b9b, 0x4c97279):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeace574, 0x24a0b9b, 0x4c97279) = (data, end, members, start))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Interval.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xeace574, 0x24a0b9b, 0x4c97279):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeace574, 0x24a0b9b, 0x4c97279) = (data, end, members, start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeace574, 0x24a0b9b, 0x4c97279) = (data, end, members, start))" % __pyx_checksum)
 *     __pyx_result = Interval.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeace574, 0x24a0b9b, 0x4c97279) = (data, end, members, start))" % __pyx_checksum)
 *     __pyx_result = Interval.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Interval.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval__set_state(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeace574, 0x24a0b9b, 0x4c97279) = (data, end, members, start))" % __pyx_checksum)
 *     __pyx_result = Interval.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Interval(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_Interval", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  double __pyx_t_2;

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Interval__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->data);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->data);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->data = ((PyObject*)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.end = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->members);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->members);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->members = __pyx_t_1;

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.start = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = ((__pyx_t_4 > 4) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_5) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_3 = __pyx_t_5;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 13 "(tree fragment)"
  __pyx_t_3 = __pyx_t_6;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_9 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {

#line 14 "(tree fragment)"
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_9)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_8, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_Interval__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IntervalTreeNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_11__pyx_unpickle_IntervalTreeNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10__pyx_unpickle_IntervalTreeNode[] = "__pyx_unpickle_IntervalTreeNode(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_11__pyx_unpickle_IntervalTreeNode = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_IntervalTreeNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_11__pyx_unpickle_IntervalTreeNode, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10__pyx_unpickle_IntervalTreeNode};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_11__pyx_unpickle_IntervalTreeNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTreeNode (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTreeNode", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTreeNode", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_IntervalTreeNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTreeNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_IntervalTreeNode", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10__pyx_unpickle_IntervalTreeNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10__pyx_unpickle_IntervalTreeNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTreeNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x262a9ad, 0x0629e3a, 0xfab189d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x262a9ad, 0x0629e3a, 0xfab189d) = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__3, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x262a9ad, 0x0629e3a, 0xfab189d):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x262a9ad, 0x0629e3a, 0xfab189d) = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x262a9ad, 0x0629e3a, 0xfab189d):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x262a9ad, 0x0629e3a, 0xfab189d) = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x262a9ad, 0x0629e3a, 0xfab189d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x262a9ad, 0x0629e3a, 0xfab189d) = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x262a9ad, 0x0629e3a, 0xfab189d) = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x262a9ad, 0x0629e3a, 0xfab189d) = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode__set_state(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x262a9ad, 0x0629e3a, 0xfab189d) = (center, contained, end, left, level, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IntervalTreeNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_IntervalTreeNode", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  double __pyx_t_2;

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  int __pyx_t_4;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  int __pyx_t_7;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_10 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTreeNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->center = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->contained);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->contained);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->contained = ((PyObject*)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->end = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->left);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->level = __pyx_t_3;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->parent);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->parent));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->right);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->start = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = ((__pyx_t_5 > 8) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_6) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_4 = __pyx_t_6;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_7 = (__pyx_t_6 != 0);

#line 13 "(tree fragment)"
  __pyx_t_4 = __pyx_t_7;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_9);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __pyx_t_10 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {

#line 14 "(tree fragment)"
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_10)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_10);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_9, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IntervalTreeNode__set_state(<IntervalTreeNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode__set_state(IntervalTreeNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.left = __pyx_state[3]; __pyx_result.level = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.start = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_10);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_IntervalTreeNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_SpanningMixin2D(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13__pyx_unpickle_SpanningMixin2D(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_12__pyx_unpickle_SpanningMixin2D[] = "__pyx_unpickle_SpanningMixin2D(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13__pyx_unpickle_SpanningMixin2D = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_SpanningMixin2D", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13__pyx_unpickle_SpanningMixin2D, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_12__pyx_unpickle_SpanningMixin2D};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13__pyx_unpickle_SpanningMixin2D(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpanningMixin2D (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpanningMixin2D", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpanningMixin2D", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_SpanningMixin2D") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpanningMixin2D", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SpanningMixin2D", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_12__pyx_unpickle_SpanningMixin2D(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_12__pyx_unpickle_SpanningMixin2D(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpanningMixin2D", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfeaf07a, 0x2a19646, 0xea81723):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfeaf07a, 0x2a19646, 0xea81723) = (_end, _start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__4, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfeaf07a, 0x2a19646, 0xea81723):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfeaf07a, 0x2a19646, 0xea81723) = (_end, _start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin2D.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xfeaf07a, 0x2a19646, 0xea81723):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfeaf07a, 0x2a19646, 0xea81723) = (_end, _start))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = SpanningMixin2D.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfeaf07a, 0x2a19646, 0xea81723):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfeaf07a, 0x2a19646, 0xea81723) = (_end, _start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfeaf07a, 0x2a19646, 0xea81723) = (_end, _start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin2D.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpanningMixin2D__set_state(<SpanningMixin2D> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfeaf07a, 0x2a19646, 0xea81723) = (_end, _start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin2D.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SpanningMixin2D__set_state(<SpanningMixin2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = SpanningMixin2D.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpanningMixin2D__set_state(<SpanningMixin2D> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin2D__set_state(SpanningMixin2D __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin2D__set_state(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfeaf07a, 0x2a19646, 0xea81723) = (_end, _start))" % __pyx_checksum)
 *     __pyx_result = SpanningMixin2D.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SpanningMixin2D__set_state(<SpanningMixin2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpanningMixin2D__set_state(<SpanningMixin2D> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_SpanningMixin2D__set_state(SpanningMixin2D __pyx_result, tuple __pyx_state):
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_SpanningMixin2D(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SpanningMixin2D", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_SpanningMixin2D__set_state(<SpanningMixin2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin2D__set_state(SpanningMixin2D __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_SpanningMixin2D__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  double __pyx_t_2[2];

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpanningMixin2D__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin2D__set_state(SpanningMixin2D __pyx_result, tuple __pyx_state):
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (unlikely(__Pyx_carray_from_py_double(__pyx_t_1, __pyx_t_2, 2) < 0)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  memcpy(&(__pyx_v___pyx_result->_end[0]), __pyx_t_2, sizeof(__pyx_v___pyx_result->_end[0]) * (2));

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (unlikely(__Pyx_carray_from_py_double(__pyx_t_1, __pyx_t_2, 2) < 0)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  memcpy(&(__pyx_v___pyx_result->_start[0]), __pyx_t_2, sizeof(__pyx_v___pyx_result->_start[0]) * (2));

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_SpanningMixin2D__set_state(SpanningMixin2D __pyx_result, tuple __pyx_state):
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = ((__pyx_t_4 > 2) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_5) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_3 = __pyx_t_5;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 13 "(tree fragment)"
  __pyx_t_3 = __pyx_t_6;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_9 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {

#line 14 "(tree fragment)"
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_9)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_8, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_SpanningMixin2D__set_state(SpanningMixin2D __pyx_result, tuple __pyx_state):
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_SpanningMixin2D__set_state(<SpanningMixin2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin2D__set_state(SpanningMixin2D __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_SpanningMixin2D__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Interval2D(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15__pyx_unpickle_Interval2D(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14__pyx_unpickle_Interval2D[] = "__pyx_unpickle_Interval2D(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15__pyx_unpickle_Interval2D = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_Interval2D", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15__pyx_unpickle_Interval2D, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14__pyx_unpickle_Interval2D};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15__pyx_unpickle_Interval2D(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Interval2D (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Interval2D", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Interval2D", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Interval2D") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Interval2D", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_Interval2D", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14__pyx_unpickle_Interval2D(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14__pyx_unpickle_Interval2D(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Interval2D", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x688f3f3, 0xc0bd0b9, 0x6f637d9):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x688f3f3, 0xc0bd0b9, 0x6f637d9) = (_end, _start, data, members))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__5, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x688f3f3, 0xc0bd0b9, 0x6f637d9):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x688f3f3, 0xc0bd0b9, 0x6f637d9) = (_end, _start, data, members))" % __pyx_checksum)
 *     __pyx_result = Interval2D.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x688f3f3, 0xc0bd0b9, 0x6f637d9):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x688f3f3, 0xc0bd0b9, 0x6f637d9) = (_end, _start, data, members))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Interval2D.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x688f3f3, 0xc0bd0b9, 0x6f637d9):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x688f3f3, 0xc0bd0b9, 0x6f637d9) = (_end, _start, data, members))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x688f3f3, 0xc0bd0b9, 0x6f637d9) = (_end, _start, data, members))" % __pyx_checksum)
 *     __pyx_result = Interval2D.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Interval2D__set_state(<Interval2D> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x688f3f3, 0xc0bd0b9, 0x6f637d9) = (_end, _start, data, members))" % __pyx_checksum)
 *     __pyx_result = Interval2D.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Interval2D__set_state(<Interval2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Interval2D.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Interval2D__set_state(<Interval2D> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval2D__set_state(Interval2D __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval2D__set_state(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x688f3f3, 0xc0bd0b9, 0x6f637d9) = (_end, _start, data, members))" % __pyx_checksum)
 *     __pyx_result = Interval2D.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Interval2D__set_state(<Interval2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Interval2D__set_state(<Interval2D> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Interval2D__set_state(Interval2D __pyx_result, tuple __pyx_state):
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]; __pyx_result.data = __pyx_state[2]; __pyx_result.members = __pyx_state[3]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Interval2D(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_Interval2D", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Interval2D__set_state(<Interval2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval2D__set_state(Interval2D __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]; __pyx_result.data = __pyx_state[2]; __pyx_result.members = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_Interval2D__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  double __pyx_t_2[2];

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Interval2D__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval2D__set_state(Interval2D __pyx_result, tuple __pyx_state):
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]; __pyx_result.data = __pyx_state[2]; __pyx_result.members = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (unlikely(__Pyx_carray_from_py_double(__pyx_t_1, __pyx_t_2, 2) < 0)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  memcpy(&(__pyx_v___pyx_result->__pyx_base._end[0]), __pyx_t_2, sizeof(__pyx_v___pyx_result->__pyx_base._end[0]) * (2));

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (unlikely(__Pyx_carray_from_py_double(__pyx_t_1, __pyx_t_2, 2) < 0)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  memcpy(&(__pyx_v___pyx_result->__pyx_base._start[0]), __pyx_t_2, sizeof(__pyx_v___pyx_result->__pyx_base._start[0]) * (2));

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->data);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->data);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->data = ((PyObject*)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->members);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->members);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->members = ((PyObject*)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Interval2D__set_state(Interval2D __pyx_result, tuple __pyx_state):
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]; __pyx_result.data = __pyx_state[2]; __pyx_result.members = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = ((__pyx_t_4 > 4) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_5) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_3 = __pyx_t_5;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 13 "(tree fragment)"
  __pyx_t_3 = __pyx_t_6;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]; __pyx_result.data = __pyx_state[2]; __pyx_result.members = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_9 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {

#line 14 "(tree fragment)"
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_9)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_8, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Interval2D__set_state(Interval2D __pyx_result, tuple __pyx_state):
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]; __pyx_result.data = __pyx_state[2]; __pyx_result.members = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Interval2D__set_state(<Interval2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval2D__set_state(Interval2D __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]; __pyx_result.data = __pyx_state[2]; __pyx_result.members = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_Interval2D__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IntervalTreeNode2D(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_17__pyx_unpickle_IntervalTreeNode2D(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16__pyx_unpickle_IntervalTreeNode2D[] = "__pyx_unpickle_IntervalTreeNode2D(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_17__pyx_unpickle_IntervalTreeNode2D = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_IntervalTreeNode2D", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_17__pyx_unpickle_IntervalTreeNode2D, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16__pyx_unpickle_IntervalTreeNode2D};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_17__pyx_unpickle_IntervalTreeNode2D(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTreeNode2D (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTreeNode2D", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTreeNode2D", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_IntervalTreeNode2D") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTreeNode2D", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_IntervalTreeNode2D", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16__pyx_unpickle_IntervalTreeNode2D(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16__pyx_unpickle_IntervalTreeNode2D(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTreeNode2D", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0745b49, 0xc11802f, 0xc16f891):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x0745b49, 0xc11802f, 0xc16f891) = (center, contained, end, inner_organizer, left, level, organized, parent, right, start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__6, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0745b49, 0xc11802f, 0xc16f891):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x0745b49, 0xc11802f, 0xc16f891) = (center, contained, end, inner_organizer, left, level, organized, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode2D.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x0745b49, 0xc11802f, 0xc16f891):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x0745b49, 0xc11802f, 0xc16f891) = (center, contained, end, inner_organizer, left, level, organized, parent, right, start))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = IntervalTreeNode2D.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0745b49, 0xc11802f, 0xc16f891):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x0745b49, 0xc11802f, 0xc16f891) = (center, contained, end, inner_organizer, left, level, organized, parent, right, start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x0745b49, 0xc11802f, 0xc16f891) = (center, contained, end, inner_organizer, left, level, organized, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode2D.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTreeNode2D__set_state(<IntervalTreeNode2D> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x0745b49, 0xc11802f, 0xc16f891) = (center, contained, end, inner_organizer, left, level, organized, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode2D.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntervalTreeNode2D__set_state(<IntervalTreeNode2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = IntervalTreeNode2D.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTreeNode2D__set_state(<IntervalTreeNode2D> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode2D__set_state(IntervalTreeNode2D __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode2D__set_state(((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x0745b49, 0xc11802f, 0xc16f891) = (center, contained, end, inner_organizer, left, level, organized, parent, right, start))" % __pyx_checksum)
 *     __pyx_result = IntervalTreeNode2D.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntervalTreeNode2D__set_state(<IntervalTreeNode2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTreeNode2D__set_state(<IntervalTreeNode2D> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IntervalTreeNode2D__set_state(IntervalTreeNode2D __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.inner_organizer = __pyx_state[3]; __pyx_result.left = __pyx_state[4]; __pyx_result.level = __pyx_state[5]; __pyx_result.organized = __pyx_state[6]; __pyx_result.parent = __pyx_state[7]; __pyx_result.right = __pyx_state[8]; __pyx_result.start = __pyx_state[9]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IntervalTreeNode2D(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_IntervalTreeNode2D", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IntervalTreeNode2D__set_state(<IntervalTreeNode2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode2D__set_state(IntervalTreeNode2D __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.inner_organizer = __pyx_state[3]; __pyx_result.left = __pyx_state[4]; __pyx_result.level = __pyx_state[5]; __pyx_result.organized = __pyx_state[6]; __pyx_result.parent = __pyx_state[7]; __pyx_result.right = __pyx_state[8]; __pyx_result.start = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals___pyx_unpickle_IntervalTreeNode2D__set_state(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  double __pyx_t_2;

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  int __pyx_t_4;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  int __pyx_t_7;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_10 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTreeNode2D__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode2D__set_state(IntervalTreeNode2D __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.inner_organizer = __pyx_state[3]; __pyx_result.left = __pyx_state[4]; __pyx_result.level = __pyx_state[5]; __pyx_result.organized = __pyx_state[6]; __pyx_result.parent = __pyx_state[7]; __pyx_result.right = __pyx_state[8]; __pyx_result.start = __pyx_state[9]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.center = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.contained);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.contained);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.contained = ((PyObject*)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.end = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->inner_organizer);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->inner_organizer);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->inner_organizer = __pyx_t_1;

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.left);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->__pyx_base.left));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.level = __pyx_t_3;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->organized);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->organized));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->organized = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.parent);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->__pyx_base.parent));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode))))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.right);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->__pyx_base.right));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.start = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IntervalTreeNode2D__set_state(IntervalTreeNode2D __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.inner_organizer = __pyx_state[3]; __pyx_result.left = __pyx_state[4]; __pyx_result.level = __pyx_state[5]; __pyx_result.organized = __pyx_state[6]; __pyx_result.parent = __pyx_state[7]; __pyx_result.right = __pyx_state[8]; __pyx_result.start = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = ((__pyx_t_5 > 10) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_6) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_4 = __pyx_t_6;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_7 = (__pyx_t_6 != 0);

#line 13 "(tree fragment)"
  __pyx_t_4 = __pyx_t_7;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.inner_organizer = __pyx_state[3]; __pyx_result.left = __pyx_state[4]; __pyx_result.level = __pyx_state[5]; __pyx_result.organized = __pyx_state[6]; __pyx_result.parent = __pyx_state[7]; __pyx_result.right = __pyx_state[8]; __pyx_result.start = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[10])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_9);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __pyx_t_10 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {

#line 14 "(tree fragment)"
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_10)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_10);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_9, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IntervalTreeNode2D__set_state(IntervalTreeNode2D __pyx_result, tuple __pyx_state):
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.inner_organizer = __pyx_state[3]; __pyx_result.left = __pyx_state[4]; __pyx_result.level = __pyx_state[5]; __pyx_result.organized = __pyx_state[6]; __pyx_result.parent = __pyx_state[7]; __pyx_result.right = __pyx_state[8]; __pyx_result.start = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IntervalTreeNode2D__set_state(<IntervalTreeNode2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTreeNode2D__set_state(IntervalTreeNode2D __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center = __pyx_state[0]; __pyx_result.contained = __pyx_state[1]; __pyx_result.end = __pyx_state[2]; __pyx_result.inner_organizer = __pyx_state[3]; __pyx_result.left = __pyx_state[4]; __pyx_result.level = __pyx_state[5]; __pyx_result.organized = __pyx_state[6]; __pyx_result.parent = __pyx_state[7]; __pyx_result.right = __pyx_state[8]; __pyx_result.start = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_10);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.peak_dependency_network.intervals.__pyx_unpickle_IntervalTreeNode2D__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "carray.to_py":112
 * 
 * @cname("__Pyx_carray_to_py_double")
 * cdef inline list __Pyx_carray_to_py_double(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

#line 112 "carray.to_py"


#line 112 "carray.to_py"
static CYTHON_INLINE PyObject *__Pyx_carray_to_py_double(double *__pyx_v_v, Py_ssize_t __pyx_v_length) {
  size_t __pyx_v_i
#line 112 "carray.to_py"
;
  PyObject *__pyx_v_value = 0
#line 112 "carray.to_py"
;
  PyObject *__pyx_v_l = NULL
#line 112 "carray.to_py"
;

#line 112 "carray.to_py"
  PyObject *__pyx_r = NULL;

#line 112 "carray.to_py"
  __Pyx_RefNannyDeclarations

#line 112 "carray.to_py"
  PyObject *__pyx_t_1 = NULL;

#line 112 "carray.to_py"
  size_t __pyx_t_2;

#line 112 "carray.to_py"
  size_t __pyx_t_3;

#line 112 "carray.to_py"
  size_t __pyx_t_4;

#line 112 "carray.to_py"
  int __pyx_lineno = 0;

#line 112 "carray.to_py"
  const char *__pyx_filename = NULL;

#line 112 "carray.to_py"
  int __pyx_clineno = 0;

#line 112 "carray.to_py"
  __Pyx_RefNannySetupContext("__Pyx_carray_to_py_double", 0);

  /* "carray.to_py":115
 *     cdef size_t i
 *     cdef object value
 *     l = PyList_New(length)             # <<<<<<<<<<<<<<
 *     for i in range(<size_t>length):
 *         value = v[i]
 */

#line 115 "carray.to_py"
  __pyx_t_1 = PyList_New(__pyx_v_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 115, __pyx_L1_error)

#line 115 "carray.to_py"
  __Pyx_GOTREF(__pyx_t_1);

#line 115 "carray.to_py"
  __pyx_v_l = ((PyObject*)__pyx_t_1);

#line 115 "carray.to_py"
  __pyx_t_1 = 0;

  /* "carray.to_py":116
 *     cdef object value
 *     l = PyList_New(length)
 *     for i in range(<size_t>length):             # <<<<<<<<<<<<<<
 *         value = v[i]
 *         Py_INCREF(value)
 */

#line 116 "carray.to_py"
  __pyx_t_2 = ((size_t)__pyx_v_length);

#line 116 "carray.to_py"
  __pyx_t_3 = __pyx_t_2;

#line 116 "carray.to_py"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 116 "carray.to_py"
    __pyx_v_i = __pyx_t_4;

    /* "carray.to_py":117
 *     l = PyList_New(length)
 *     for i in range(<size_t>length):
 *         value = v[i]             # <<<<<<<<<<<<<<
 *         Py_INCREF(value)
 *         PyList_SET_ITEM(l, i, value)
 */

#line 117 "carray.to_py"
    __pyx_t_1 = PyFloat_FromDouble((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 117, __pyx_L1_error)

#line 117 "carray.to_py"
    __Pyx_GOTREF(__pyx_t_1);

#line 117 "carray.to_py"
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);

#line 117 "carray.to_py"
    __pyx_t_1 = 0;

    /* "carray.to_py":118
 *     for i in range(<size_t>length):
 *         value = v[i]
 *         Py_INCREF(value)             # <<<<<<<<<<<<<<
 *         PyList_SET_ITEM(l, i, value)
 *     return l
 */

#line 118 "carray.to_py"
    Py_INCREF(__pyx_v_value);

    /* "carray.to_py":119
 *         value = v[i]
 *         Py_INCREF(value)
 *         PyList_SET_ITEM(l, i, value)             # <<<<<<<<<<<<<<
 *     return l
 * 
 */

#line 119 "carray.to_py"
    PyList_SET_ITEM(__pyx_v_l, __pyx_v_i, __pyx_v_value);

#line 119 "carray.to_py"
  }

  /* "carray.to_py":120
 *         Py_INCREF(value)
 *         PyList_SET_ITEM(l, i, value)
 *     return l             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 120 "carray.to_py"
  __Pyx_XDECREF(__pyx_r);

#line 120 "carray.to_py"
  __Pyx_INCREF(__pyx_v_l);

#line 120 "carray.to_py"
  __pyx_r = __pyx_v_l;

#line 120 "carray.to_py"
  goto __pyx_L0;

  /* "carray.to_py":112
 * 
 * @cname("__Pyx_carray_to_py_double")
 * cdef inline list __Pyx_carray_to_py_double(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

#line 112 "carray.to_py"


#line 112 "carray.to_py"
  /* function exit code */

#line 112 "carray.to_py"
  __pyx_L1_error:;

#line 112 "carray.to_py"
  __Pyx_XDECREF(__pyx_t_1);

#line 112 "carray.to_py"
  __Pyx_AddTraceback("carray.to_py.__Pyx_carray_to_py_double", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 112 "carray.to_py"
  __pyx_r = 0;

#line 112 "carray.to_py"
  __pyx_L0:;

#line 112 "carray.to_py"
  __Pyx_XDECREF(__pyx_v_value);

#line 112 "carray.to_py"
  __Pyx_XDECREF(__pyx_v_l);

#line 112 "carray.to_py"
  __Pyx_XGIVEREF(__pyx_r);

#line 112 "carray.to_py"
  __Pyx_RefNannyFinishContext();

#line 112 "carray.to_py"
  return __pyx_r;

#line 112 "carray.to_py"
}

/* "carray.to_py":124
 * 
 * @cname("__Pyx_carray_to_tuple_double")
 * cdef inline tuple __Pyx_carray_to_tuple_double(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

#line 124 "carray.to_py"


#line 124 "carray.to_py"
static CYTHON_INLINE PyObject *__Pyx_carray_to_tuple_double(double *__pyx_v_v, Py_ssize_t __pyx_v_length) {
  size_t __pyx_v_i
#line 124 "carray.to_py"
;
  PyObject *__pyx_v_value = 0
#line 124 "carray.to_py"
;
  PyObject *__pyx_v_t = NULL
#line 124 "carray.to_py"
;

#line 124 "carray.to_py"
  PyObject *__pyx_r = NULL;

#line 124 "carray.to_py"
  __Pyx_RefNannyDeclarations

#line 124 "carray.to_py"
  PyObject *__pyx_t_1 = NULL;

#line 124 "carray.to_py"
  size_t __pyx_t_2;

#line 124 "carray.to_py"
  size_t __pyx_t_3;

#line 124 "carray.to_py"
  size_t __pyx_t_4;

#line 124 "carray.to_py"
  int __pyx_lineno = 0;

#line 124 "carray.to_py"
  const char *__pyx_filename = NULL;

#line 124 "carray.to_py"
  int __pyx_clineno = 0;

#line 124 "carray.to_py"
  __Pyx_RefNannySetupContext("__Pyx_carray_to_tuple_double", 0);

  /* "carray.to_py":127
 *     cdef size_t i
 *     cdef object value
 *     t = PyTuple_New(length)             # <<<<<<<<<<<<<<
 *     for i in range(<size_t>length):
 *         value = v[i]
 */

#line 127 "carray.to_py"
  __pyx_t_1 = PyTuple_New(__pyx_v_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 127, __pyx_L1_error)

#line 127 "carray.to_py"
  __Pyx_GOTREF(__pyx_t_1);

#line 127 "carray.to_py"
  __pyx_v_t = ((PyObject*)__pyx_t_1);

#line 127 "carray.to_py"
  __pyx_t_1 = 0;

  /* "carray.to_py":128
 *     cdef object value
 *     t = PyTuple_New(length)
 *     for i in range(<size_t>length):             # <<<<<<<<<<<<<<
 *         value = v[i]
 *         Py_INCREF(value)
 */

#line 128 "carray.to_py"
  __pyx_t_2 = ((size_t)__pyx_v_length);

#line 128 "carray.to_py"
  __pyx_t_3 = __pyx_t_2;

#line 128 "carray.to_py"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 128 "carray.to_py"
    __pyx_v_i = __pyx_t_4;

    /* "carray.to_py":129
 *     t = PyTuple_New(length)
 *     for i in range(<size_t>length):
 *         value = v[i]             # <<<<<<<<<<<<<<
 *         Py_INCREF(value)
 *         PyTuple_SET_ITEM(t, i, value)
 */

#line 129 "carray.to_py"
    __pyx_t_1 = PyFloat_FromDouble((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 129, __pyx_L1_error)

#line 129 "carray.to_py"
    __Pyx_GOTREF(__pyx_t_1);

#line 129 "carray.to_py"
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);

#line 129 "carray.to_py"
    __pyx_t_1 = 0;

    /* "carray.to_py":130
 *     for i in range(<size_t>length):
 *         value = v[i]
 *         Py_INCREF(value)             # <<<<<<<<<<<<<<
 *         PyTuple_SET_ITEM(t, i, value)
 *     return t
 */

#line 130 "carray.to_py"
    Py_INCREF(__pyx_v_value);

    /* "carray.to_py":131
 *         value = v[i]
 *         Py_INCREF(value)
 *         PyTuple_SET_ITEM(t, i, value)             # <<<<<<<<<<<<<<
 *     return t
 */

#line 131 "carray.to_py"
    PyTuple_SET_ITEM(__pyx_v_t, __pyx_v_i, __pyx_v_value);

#line 131 "carray.to_py"
  }

  /* "carray.to_py":132
 *         Py_INCREF(value)
 *         PyTuple_SET_ITEM(t, i, value)
 *     return t             # <<<<<<<<<<<<<<
 */

#line 132 "carray.to_py"
  __Pyx_XDECREF(__pyx_r);

#line 132 "carray.to_py"
  __Pyx_INCREF(__pyx_v_t);

#line 132 "carray.to_py"
  __pyx_r = __pyx_v_t;

#line 132 "carray.to_py"
  goto __pyx_L0;

  /* "carray.to_py":124
 * 
 * @cname("__Pyx_carray_to_tuple_double")
 * cdef inline tuple __Pyx_carray_to_tuple_double(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

#line 124 "carray.to_py"


#line 124 "carray.to_py"
  /* function exit code */

#line 124 "carray.to_py"
  __pyx_L1_error:;

#line 124 "carray.to_py"
  __Pyx_XDECREF(__pyx_t_1);

#line 124 "carray.to_py"
  __Pyx_AddTraceback("carray.to_py.__Pyx_carray_to_tuple_double", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 124 "carray.to_py"
  __pyx_r = 0;

#line 124 "carray.to_py"
  __pyx_L0:;

#line 124 "carray.to_py"
  __Pyx_XDECREF(__pyx_v_value);

#line 124 "carray.to_py"
  __Pyx_XDECREF(__pyx_v_t);

#line 124 "carray.to_py"
  __Pyx_XGIVEREF(__pyx_r);

#line 124 "carray.to_py"
  __Pyx_RefNannyFinishContext();

#line 124 "carray.to_py"
  return __pyx_r;

#line 124 "carray.to_py"
}

/* "carray.from_py":77
 * 
 * @cname("__Pyx_carray_from_py_double")
 * cdef int __Pyx_carray_from_py_double(object o, base_type *v, Py_ssize_t length) except -1:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i = length
 *     try:
 */

#line 77 "carray.from_py"


#line 77 "carray.from_py"
static int __Pyx_carray_from_py_double(PyObject *__pyx_v_o, double *__pyx_v_v, Py_ssize_t __pyx_v_length) {
  Py_ssize_t __pyx_v_i
#line 77 "carray.from_py"
;
  PyObject *__pyx_v_item = NULL
#line 77 "carray.from_py"
;

#line 77 "carray.from_py"
  int __pyx_r;

#line 77 "carray.from_py"
  __Pyx_RefNannyDeclarations

#line 77 "carray.from_py"
  PyObject *__pyx_t_1 = NULL;

#line 77 "carray.from_py"
  PyObject *__pyx_t_2 = NULL;

#line 77 "carray.from_py"
  PyObject *__pyx_t_3 = NULL;

#line 77 "carray.from_py"
  Py_ssize_t __pyx_t_4;

#line 77 "carray.from_py"
  int __pyx_t_5;

#line 77 "carray.from_py"
  int __pyx_t_6;

#line 77 "carray.from_py"
  PyObject *__pyx_t_7 = NULL;

#line 77 "carray.from_py"
  Py_ssize_t __pyx_t_8;

#line 77 "carray.from_py"
  PyObject *(*__pyx_t_9)(PyObject *);

#line 77 "carray.from_py"
  PyObject *__pyx_t_10 = NULL;

#line 77 "carray.from_py"
  double __pyx_t_11;

#line 77 "carray.from_py"
  char const *__pyx_t_12;

#line 77 "carray.from_py"
  int __pyx_lineno = 0;

#line 77 "carray.from_py"
  const char *__pyx_filename = NULL;

#line 77 "carray.from_py"
  int __pyx_clineno = 0;

#line 77 "carray.from_py"
  __Pyx_RefNannySetupContext("__Pyx_carray_from_py_double", 0);

  /* "carray.from_py":78
 * @cname("__Pyx_carray_from_py_double")
 * cdef int __Pyx_carray_from_py_double(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length             # <<<<<<<<<<<<<<
 *     try:
 *         i = len(o)
 */

#line 78 "carray.from_py"
  __pyx_v_i = __pyx_v_length;

  /* "carray.from_py":79
 * cdef int __Pyx_carray_from_py_double(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */

#line 79 "carray.from_py"
  {

#line 79 "carray.from_py"
    __Pyx_PyThreadState_declare

#line 79 "carray.from_py"
    __Pyx_PyThreadState_assign

#line 79 "carray.from_py"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 79 "carray.from_py"
    __Pyx_XGOTREF(__pyx_t_1);

#line 79 "carray.from_py"
    __Pyx_XGOTREF(__pyx_t_2);

#line 79 "carray.from_py"
    __Pyx_XGOTREF(__pyx_t_3);

#line 79 "carray.from_py"
    /*try:*/ {

      /* "carray.from_py":80
 *     cdef Py_ssize_t i = length
 *     try:
 *         i = len(o)             # <<<<<<<<<<<<<<
 *     except (TypeError, OverflowError):
 *         pass
 */

#line 80 "carray.from_py"
      __pyx_t_4 = PyObject_Length(__pyx_v_o); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 80, __pyx_L3_error)

#line 80 "carray.from_py"
      __pyx_v_i = __pyx_t_4;

      /* "carray.from_py":79
 * cdef int __Pyx_carray_from_py_double(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */

#line 79 "carray.from_py"
    }

#line 79 "carray.from_py"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 79 "carray.from_py"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 79 "carray.from_py"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 79 "carray.from_py"
    goto __pyx_L8_try_end;

#line 79 "carray.from_py"
    __pyx_L3_error:;

    /* "carray.from_py":81
 *     try:
 *         i = len(o)
 *     except (TypeError, OverflowError):             # <<<<<<<<<<<<<<
 *         pass
 *     if i == length:
 */

#line 81 "carray.from_py"
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OverflowError);

#line 81 "carray.from_py"
    if (__pyx_t_5) {

#line 81 "carray.from_py"
      __Pyx_ErrRestore(0,0,0);

#line 81 "carray.from_py"
      goto __pyx_L4_exception_handled;

#line 81 "carray.from_py"
    }

#line 81 "carray.from_py"
    goto __pyx_L5_except_error;

#line 81 "carray.from_py"
    __pyx_L5_except_error:;

    /* "carray.from_py":79
 * cdef int __Pyx_carray_from_py_double(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */

#line 79 "carray.from_py"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 79 "carray.from_py"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 79 "carray.from_py"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 79 "carray.from_py"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 79 "carray.from_py"
    goto __pyx_L1_error;

#line 79 "carray.from_py"
    __pyx_L4_exception_handled:;

#line 79 "carray.from_py"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 79 "carray.from_py"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 79 "carray.from_py"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 79 "carray.from_py"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 79 "carray.from_py"
    __pyx_L8_try_end:;

#line 79 "carray.from_py"
  }

  /* "carray.from_py":83
 *     except (TypeError, OverflowError):
 *         pass
 *     if i == length:             # <<<<<<<<<<<<<<
 *         for i, item in enumerate(o):
 *             if i >= length:
 */

#line 83 "carray.from_py"
  __pyx_t_6 = ((__pyx_v_i == __pyx_v_length) != 0);

#line 83 "carray.from_py"
  if (__pyx_t_6) {

    /* "carray.from_py":84
 *         pass
 *     if i == length:
 *         for i, item in enumerate(o):             # <<<<<<<<<<<<<<
 *             if i >= length:
 *                 break
 */

#line 84 "carray.from_py"
    __pyx_t_4 = 0;

#line 84 "carray.from_py"
    if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {

#line 84 "carray.from_py"
      __pyx_t_7 = __pyx_v_o; __Pyx_INCREF(__pyx_t_7); __pyx_t_8 = 0;

#line 84 "carray.from_py"
      __pyx_t_9 = NULL;

#line 84 "carray.from_py"
    } else {
      __pyx_t_8 = -1; 
#line 84 "carray.from_py"
__pyx_t_7 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 84, __pyx_L1_error)

#line 84 "carray.from_py"
      __Pyx_GOTREF(__pyx_t_7);

#line 84 "carray.from_py"
      __pyx_t_9 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 84, __pyx_L1_error)

#line 84 "carray.from_py"
    }

#line 84 "carray.from_py"
    for (;;) {

#line 84 "carray.from_py"
      if (likely(!__pyx_t_9)) {

#line 84 "carray.from_py"
        if (likely(PyList_CheckExact(__pyx_t_7))) {

#line 84 "carray.from_py"
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_7)) break;

#line 84 "carray.from_py"
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 84 "carray.from_py"
          __pyx_t_10 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(2, 84, __pyx_L1_error)

#line 84 "carray.from_py"
          #else

#line 84 "carray.from_py"
          __pyx_t_10 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 84, __pyx_L1_error)

#line 84 "carray.from_py"
          __Pyx_GOTREF(__pyx_t_10);

#line 84 "carray.from_py"
          #endif

#line 84 "carray.from_py"
        } else {

#line 84 "carray.from_py"
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_7)) break;

#line 84 "carray.from_py"
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 84 "carray.from_py"
          __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(2, 84, __pyx_L1_error)

#line 84 "carray.from_py"
          #else

#line 84 "carray.from_py"
          __pyx_t_10 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 84, __pyx_L1_error)

#line 84 "carray.from_py"
          __Pyx_GOTREF(__pyx_t_10);

#line 84 "carray.from_py"
          #endif

#line 84 "carray.from_py"
        }
      } else 
#line 84 "carray.from_py"
{

#line 84 "carray.from_py"
        __pyx_t_10 = __pyx_t_9(__pyx_t_7);

#line 84 "carray.from_py"
        if (unlikely(!__pyx_t_10)) {

#line 84 "carray.from_py"
          PyObject* exc_type = PyErr_Occurred();

#line 84 "carray.from_py"
          if (exc_type) {

#line 84 "carray.from_py"
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 84 "carray.from_py"
            else __PYX_ERR(2, 84, __pyx_L1_error)

#line 84 "carray.from_py"
          }

#line 84 "carray.from_py"
          break;

#line 84 "carray.from_py"
        }

#line 84 "carray.from_py"
        __Pyx_GOTREF(__pyx_t_10);

#line 84 "carray.from_py"
      }

#line 84 "carray.from_py"
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_10);

#line 84 "carray.from_py"
      __pyx_t_10 = 0;

#line 84 "carray.from_py"
      __pyx_v_i = __pyx_t_4;

#line 84 "carray.from_py"
      __pyx_t_4 = (__pyx_t_4 + 1);

      /* "carray.from_py":85
 *     if i == length:
 *         for i, item in enumerate(o):
 *             if i >= length:             # <<<<<<<<<<<<<<
 *                 break
 *             v[i] = item
 */

#line 85 "carray.from_py"
      __pyx_t_6 = ((__pyx_v_i >= __pyx_v_length) != 0);

#line 85 "carray.from_py"
      if (__pyx_t_6) {

        /* "carray.from_py":86
 *         for i, item in enumerate(o):
 *             if i >= length:
 *                 break             # <<<<<<<<<<<<<<
 *             v[i] = item
 *         else:
 */

#line 86 "carray.from_py"
        goto __pyx_L11_break;

        /* "carray.from_py":85
 *     if i == length:
 *         for i, item in enumerate(o):
 *             if i >= length:             # <<<<<<<<<<<<<<
 *                 break
 *             v[i] = item
 */

#line 85 "carray.from_py"
      }

      /* "carray.from_py":87
 *             if i >= length:
 *                 break
 *             v[i] = item             # <<<<<<<<<<<<<<
 *         else:
 *             i += 1  # convert index to length
 */

#line 87 "carray.from_py"
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_v_item); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 87, __pyx_L1_error)

#line 87 "carray.from_py"
      (__pyx_v_v[__pyx_v_i]) = __pyx_t_11;

      /* "carray.from_py":84
 *         pass
 *     if i == length:
 *         for i, item in enumerate(o):             # <<<<<<<<<<<<<<
 *             if i >= length:
 *                 break
 */

#line 84 "carray.from_py"
    }

#line 84 "carray.from_py"
    /*else*/ {

      /* "carray.from_py":89
 *             v[i] = item
 *         else:
 *             i += 1  # convert index to length             # <<<<<<<<<<<<<<
 *             if i == length:
 *                 return 0
 */

#line 89 "carray.from_py"
      __pyx_v_i = (__pyx_v_i + 1);

      /* "carray.from_py":90
 *         else:
 *             i += 1  # convert index to length
 *             if i == length:             # <<<<<<<<<<<<<<
 *                 return 0
 * 
 */

#line 90 "carray.from_py"
      __pyx_t_6 = ((__pyx_v_i == __pyx_v_length) != 0);

#line 90 "carray.from_py"
      if (__pyx_t_6) {

        /* "carray.from_py":91
 *             i += 1  # convert index to length
 *             if i == length:
 *                 return 0             # <<<<<<<<<<<<<<
 * 
 *     PyErr_Format(
 */

#line 91 "carray.from_py"
        __pyx_r = 0;

#line 91 "carray.from_py"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 91 "carray.from_py"
        goto __pyx_L0;

        /* "carray.from_py":90
 *         else:
 *             i += 1  # convert index to length
 *             if i == length:             # <<<<<<<<<<<<<<
 *                 return 0
 * 
 */

#line 90 "carray.from_py"
      }

#line 90 "carray.from_py"
    }

    /* "carray.from_py":84
 *         pass
 *     if i == length:
 *         for i, item in enumerate(o):             # <<<<<<<<<<<<<<
 *             if i >= length:
 *                 break
 */

#line 84 "carray.from_py"
    __pyx_L11_break:;

#line 84 "carray.from_py"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "carray.from_py":83
 *     except (TypeError, OverflowError):
 *         pass
 *     if i == length:             # <<<<<<<<<<<<<<
 *         for i, item in enumerate(o):
 *             if i >= length:
 */

#line 83 "carray.from_py"
  }

  /* "carray.from_py":96
 *         IndexError,
 *         ("too many values found during array assignment, expected %zd"
 *          if i >= length else             # <<<<<<<<<<<<<<
 *          "not enough values found during array assignment, expected %zd, got %zd"),
 *         length, i)
 */

#line 96 "carray.from_py"
  if (((__pyx_v_i >= __pyx_v_length) != 0)) {

#line 96 "carray.from_py"
    __pyx_t_12 = ((char const *)"too many values found during array assignment, expected %zd");

#line 96 "carray.from_py"
  } else {

#line 96 "carray.from_py"
    __pyx_t_12 = ((char const *)"not enough values found during array assignment, expected %zd, got %zd");

#line 96 "carray.from_py"
  }

  /* "carray.from_py":93
 *                 return 0
 * 
 *     PyErr_Format(             # <<<<<<<<<<<<<<
 *         IndexError,
 *         ("too many values found during array assignment, expected %zd"
 */

#line 93 "carray.from_py"
  __pyx_t_7 = PyErr_Format(__pyx_builtin_IndexError, __pyx_t_12, __pyx_v_length, __pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 93, __pyx_L1_error)

#line 93 "carray.from_py"
  __Pyx_GOTREF(__pyx_t_7);

#line 93 "carray.from_py"
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "carray.from_py":77
 * 
 * @cname("__Pyx_carray_from_py_double")
 * cdef int __Pyx_carray_from_py_double(object o, base_type *v, Py_ssize_t length) except -1:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i = length
 *     try:
 */

#line 77 "carray.from_py"


#line 77 "carray.from_py"
  /* function exit code */

#line 77 "carray.from_py"
  __pyx_r = 0;

#line 77 "carray.from_py"
  goto __pyx_L0;

#line 77 "carray.from_py"
  __pyx_L1_error:;

#line 77 "carray.from_py"
  __Pyx_XDECREF(__pyx_t_7);

#line 77 "carray.from_py"
  __Pyx_XDECREF(__pyx_t_10);

#line 77 "carray.from_py"
  __Pyx_AddTraceback("carray.from_py.__Pyx_carray_from_py_double", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 77 "carray.from_py"
  __pyx_r = -1;

#line 77 "carray.from_py"
  __pyx_L0:;

#line 77 "carray.from_py"
  __Pyx_XDECREF(__pyx_v_item);

#line 77 "carray.from_py"
  __Pyx_RefNannyFinishContext();

#line 77 "carray.from_py"
  return __pyx_r;

#line 77 "carray.from_py"
}

/* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
/* Python wrapper */

#line 123 "View.MemoryView"
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 123 "View.MemoryView"
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0
#line 123 "View.MemoryView"
;
  Py_ssize_t __pyx_v_itemsize
#line 123 "View.MemoryView"
;
  PyObject *__pyx_v_format = 0
#line 123 "View.MemoryView"
;
  PyObject *__pyx_v_mode = 0
#line 123 "View.MemoryView"
;
  int __pyx_v_allocate_buffer
#line 123 "View.MemoryView"
;

#line 123 "View.MemoryView"
  int __pyx_lineno = 0;

#line 123 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 123 "View.MemoryView"
  int __pyx_clineno = 0;

#line 123 "View.MemoryView"
  int __pyx_r;

#line 123 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 123 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);

#line 123 "View.MemoryView"
  {

#line 123 "View.MemoryView"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};

#line 123 "View.MemoryView"
    PyObject* values[5] = {0,0,0,0,0};

#line 123 "View.MemoryView"
    values[3] = ((PyObject *)__pyx_n_s_c);

#line 123 "View.MemoryView"
    if (unlikely(__pyx_kwds)) {

#line 123 "View.MemoryView"
      Py_ssize_t kw_args;

#line 123 "View.MemoryView"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 123 "View.MemoryView"
      switch (pos_args) {
        case  5: 
#line 123 "View.MemoryView"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 123 "View.MemoryView"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 123 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 123 "View.MemoryView"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 123 "View.MemoryView"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  0: break;
        default: 
#line 123 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
      kw_args = PyDict_Size(__pyx_kwds);

#line 123 "View.MemoryView"
      switch (pos_args) {

#line 123 "View.MemoryView"
        case  0:

#line 123 "View.MemoryView"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else 
#line 123 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  1:

#line 123 "View.MemoryView"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;

#line 123 "View.MemoryView"
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); 
#line 123 "View.MemoryView"
__PYX_ERR(2, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  2:

#line 123 "View.MemoryView"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;

#line 123 "View.MemoryView"
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); 
#line 123 "View.MemoryView"
__PYX_ERR(2, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  3:

#line 123 "View.MemoryView"
        if (kw_args > 0) {

#line 123 "View.MemoryView"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);

#line 123 "View.MemoryView"
          if (value) { values[3] = value; kw_args--; }

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  4:

#line 123 "View.MemoryView"
        if (kw_args > 0) {

#line 123 "View.MemoryView"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);

#line 123 "View.MemoryView"
          if (value) { values[4] = value; kw_args--; }

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
      if (unlikely(kw_args > 0)) {

#line 123 "View.MemoryView"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
    } else {

#line 123 "View.MemoryView"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: 
#line 123 "View.MemoryView"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 123 "View.MemoryView"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 123 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 123 "View.MemoryView"
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 123 "View.MemoryView"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 123 "View.MemoryView"
        break;
        default: 
#line 123 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
    }

#line 123 "View.MemoryView"
    __pyx_v_shape = ((PyObject*)values[0]);

#line 123 "View.MemoryView"
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
    __pyx_v_format = values[2];

#line 123 "View.MemoryView"
    __pyx_v_mode = values[3];

#line 123 "View.MemoryView"
    if (values[4]) {

#line 123 "View.MemoryView"
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 124, __pyx_L3_error)

#line 123 "View.MemoryView"
    } else {

      /* "View.MemoryView":124
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */

#line 124 "View.MemoryView"
      __pyx_v_allocate_buffer = ((int)1);

#line 124 "View.MemoryView"
    }

#line 124 "View.MemoryView"
  }

#line 124 "View.MemoryView"
  goto __pyx_L4_argument_unpacking_done;

#line 124 "View.MemoryView"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); 
#line 124 "View.MemoryView"
__PYX_ERR(2, 123, __pyx_L3_error)

#line 124 "View.MemoryView"
  __pyx_L3_error:;

#line 124 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 124 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 124 "View.MemoryView"
  return -1;

#line 124 "View.MemoryView"
  __pyx_L4_argument_unpacking_done:;

#line 124 "View.MemoryView"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(2, 123, __pyx_L1_error)

#line 124 "View.MemoryView"
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {

#line 124 "View.MemoryView"
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(2, 123, __pyx_L1_error)

#line 124 "View.MemoryView"
  }
  __pyx_r = 
#line 124 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
  /* function exit code */

#line 123 "View.MemoryView"
  goto __pyx_L0;

#line 123 "View.MemoryView"
  __pyx_L1_error:;

#line 123 "View.MemoryView"
  __pyx_r = -1;

#line 123 "View.MemoryView"
  __pyx_L0:;

#line 123 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 123 "View.MemoryView"
  return __pyx_r;

#line 123 "View.MemoryView"
}

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx
#line 123 "View.MemoryView"
;
  Py_ssize_t __pyx_v_i
#line 123 "View.MemoryView"
;
  Py_ssize_t __pyx_v_dim
#line 123 "View.MemoryView"
;
  PyObject **__pyx_v_p
#line 123 "View.MemoryView"
;
  char __pyx_v_order
#line 123 "View.MemoryView"
;

#line 123 "View.MemoryView"
  int __pyx_r;

#line 123 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 123 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 123 "View.MemoryView"
  int __pyx_t_2;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 123 "View.MemoryView"
  int __pyx_t_4;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 123 "View.MemoryView"
  char *__pyx_t_7;

#line 123 "View.MemoryView"
  int __pyx_t_8;

#line 123 "View.MemoryView"
  Py_ssize_t __pyx_t_9;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_10 = NULL;

#line 123 "View.MemoryView"
  Py_ssize_t __pyx_t_11;

#line 123 "View.MemoryView"
  int __pyx_lineno = 0;

#line 123 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 123 "View.MemoryView"
  int __pyx_clineno = 0;

#line 123 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__", 0);

#line 123 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":130
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */

#line 130 "View.MemoryView"
  if (unlikely(__pyx_v_shape == Py_None)) {

#line 130 "View.MemoryView"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 130 "View.MemoryView"
    __PYX_ERR(2, 130, __pyx_L1_error)

#line 130 "View.MemoryView"
  }

#line 130 "View.MemoryView"
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(2, 130, __pyx_L1_error)

#line 130 "View.MemoryView"
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":131
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */

#line 131 "View.MemoryView"
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":133
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */

#line 133 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);

#line 133 "View.MemoryView"
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":134
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */

#line 134 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 134, __pyx_L1_error)

#line 134 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 134 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 134 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 134 "View.MemoryView"
    __PYX_ERR(2, 134, __pyx_L1_error)

    /* "View.MemoryView":133
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */

#line 133 "View.MemoryView"
  }

  /* "View.MemoryView":136
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */

#line 136 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);

#line 136 "View.MemoryView"
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":137
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */

#line 137 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 137, __pyx_L1_error)

#line 137 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 137 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 137 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 137 "View.MemoryView"
    __PYX_ERR(2, 137, __pyx_L1_error)

    /* "View.MemoryView":136
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */

#line 136 "View.MemoryView"
  }

  /* "View.MemoryView":139
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */

#line 139 "View.MemoryView"
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 

#line 139 "View.MemoryView"
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);

#line 139 "View.MemoryView"
  if (__pyx_t_4) {

    /* "View.MemoryView":140
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */

#line 140 "View.MemoryView"
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 140, __pyx_L1_error)

#line 140 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 140 "View.MemoryView"
    __pyx_t_6 = NULL;

#line 140 "View.MemoryView"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {

#line 140 "View.MemoryView"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 140 "View.MemoryView"
      if (likely(__pyx_t_6)) {

#line 140 "View.MemoryView"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 140 "View.MemoryView"
        __Pyx_INCREF(__pyx_t_6);

#line 140 "View.MemoryView"
        __Pyx_INCREF(function);

#line 140 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_t_5, function);

#line 140 "View.MemoryView"
      }

#line 140 "View.MemoryView"
    }

#line 140 "View.MemoryView"
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);

#line 140 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 140 "View.MemoryView"
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 140, __pyx_L1_error)

#line 140 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 140 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 140 "View.MemoryView"
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);

#line 140 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":139
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */

#line 139 "View.MemoryView"
  }

  /* "View.MemoryView":141
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */

#line 141 "View.MemoryView"
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(2, 141, __pyx_L1_error)

#line 141 "View.MemoryView"
  __pyx_t_3 = __pyx_v_format;

#line 141 "View.MemoryView"
  __Pyx_INCREF(__pyx_t_3);

#line 141 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 141 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_self->_format);

#line 141 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_self->_format);

#line 141 "View.MemoryView"
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);

#line 141 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":142
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 142 "View.MemoryView"
  if (unlikely(__pyx_v_self->_format == Py_None)) {

#line 142 "View.MemoryView"
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");

#line 142 "View.MemoryView"
    __PYX_ERR(2, 142, __pyx_L1_error)

#line 142 "View.MemoryView"
  }

#line 142 "View.MemoryView"
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(2, 142, __pyx_L1_error)

#line 142 "View.MemoryView"
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":145
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */

#line 145 "View.MemoryView"
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":146
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */

#line 146 "View.MemoryView"
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":148
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */

#line 148 "View.MemoryView"
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);

#line 148 "View.MemoryView"
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":149
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 149 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 149, __pyx_L1_error)

#line 149 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 149 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 149 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 149 "View.MemoryView"
    __PYX_ERR(2, 149, __pyx_L1_error)

    /* "View.MemoryView":148
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */

#line 148 "View.MemoryView"
  }

  /* "View.MemoryView":152
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */

#line 152 "View.MemoryView"
  __pyx_t_8 = 0;

#line 152 "View.MemoryView"
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;

#line 152 "View.MemoryView"
  for (;;) {

#line 152 "View.MemoryView"
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;

#line 152 "View.MemoryView"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 152 "View.MemoryView"
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(2, 152, __pyx_L1_error)

#line 152 "View.MemoryView"
    #else

#line 152 "View.MemoryView"
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 152, __pyx_L1_error)

#line 152 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 152 "View.MemoryView"
    #endif

#line 152 "View.MemoryView"
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 152, __pyx_L1_error)

#line 152 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 152 "View.MemoryView"
    __pyx_v_dim = __pyx_t_9;

#line 152 "View.MemoryView"
    __pyx_v_idx = __pyx_t_8;

#line 152 "View.MemoryView"
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":153
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */

#line 153 "View.MemoryView"
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);

#line 153 "View.MemoryView"
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":154
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */

#line 154 "View.MemoryView"
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 154 "View.MemoryView"
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 154 "View.MemoryView"
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_10);

#line 154 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_t_5);

#line 154 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);

#line 154 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_t_6);

#line 154 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);

#line 154 "View.MemoryView"
      __pyx_t_5 = 0;

#line 154 "View.MemoryView"
      __pyx_t_6 = 0;

#line 154 "View.MemoryView"
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 154 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 154 "View.MemoryView"
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_10);

#line 154 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 154 "View.MemoryView"
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);

#line 154 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 154 "View.MemoryView"
      __PYX_ERR(2, 154, __pyx_L1_error)

      /* "View.MemoryView":153
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */

#line 153 "View.MemoryView"
    }

    /* "View.MemoryView":155
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */

#line 155 "View.MemoryView"
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":152
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */

#line 152 "View.MemoryView"
  }

#line 152 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":158
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */

#line 158 "View.MemoryView"
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(2, 158, __pyx_L1_error)

#line 158 "View.MemoryView"
  if (__pyx_t_4) {

    /* "View.MemoryView":159
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */

#line 159 "View.MemoryView"
    __pyx_v_order = 'F';

    /* "View.MemoryView":160
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */

#line 160 "View.MemoryView"
    __Pyx_INCREF(__pyx_n_u_fortran);

#line 160 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_n_u_fortran);

#line 160 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_self->mode);

#line 160 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_self->mode);

#line 160 "View.MemoryView"
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":158
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */

#line 158 "View.MemoryView"
    goto __pyx_L10;

#line 158 "View.MemoryView"
  }

  /* "View.MemoryView":161
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */

#line 161 "View.MemoryView"
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(2, 161, __pyx_L1_error)

#line 161 "View.MemoryView"
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":162
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */

#line 162 "View.MemoryView"
    __pyx_v_order = 'C';

    /* "View.MemoryView":163
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */

#line 163 "View.MemoryView"
    __Pyx_INCREF(__pyx_n_u_c);

#line 163 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_n_u_c);

#line 163 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_self->mode);

#line 163 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_self->mode);

#line 163 "View.MemoryView"
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":161
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */

#line 161 "View.MemoryView"
    goto __pyx_L10;

#line 161 "View.MemoryView"
  }

  /* "View.MemoryView":165
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */

#line 165 "View.MemoryView"
  /*else*/ {

#line 165 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 165, __pyx_L1_error)

#line 165 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 165 "View.MemoryView"
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 165, __pyx_L1_error)

#line 165 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_10);

#line 165 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 165 "View.MemoryView"
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);

#line 165 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 165 "View.MemoryView"
    __PYX_ERR(2, 165, __pyx_L1_error)

#line 165 "View.MemoryView"
  }

#line 165 "View.MemoryView"
  __pyx_L10:;

  /* "View.MemoryView":167
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */

#line 167 "View.MemoryView"
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":170
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */

#line 170 "View.MemoryView"
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":171
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */

#line 171 "View.MemoryView"
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 171, __pyx_L1_error)

#line 171 "View.MemoryView"
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 171, __pyx_L1_error)

#line 171 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 171 "View.MemoryView"
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":172
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 172 "View.MemoryView"
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);

#line 172 "View.MemoryView"
  if (__pyx_t_4) {

    /* "View.MemoryView":175
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */

#line 175 "View.MemoryView"
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":176
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */

#line 176 "View.MemoryView"
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);

#line 176 "View.MemoryView"
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":177
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */

#line 177 "View.MemoryView"
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 177, __pyx_L1_error)

#line 177 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_10);

#line 177 "View.MemoryView"
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);

#line 177 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 177 "View.MemoryView"
      __PYX_ERR(2, 177, __pyx_L1_error)

      /* "View.MemoryView":176
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */

#line 176 "View.MemoryView"
    }

    /* "View.MemoryView":179
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */

#line 179 "View.MemoryView"
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);

#line 179 "View.MemoryView"
    if (__pyx_t_4) {

      /* "View.MemoryView":180
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */

#line 180 "View.MemoryView"
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":181
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */

#line 181 "View.MemoryView"
      if (unlikely(__pyx_v_itemsize == 0)) {

#line 181 "View.MemoryView"
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");

#line 181 "View.MemoryView"
        __PYX_ERR(2, 181, __pyx_L1_error)

#line 181 "View.MemoryView"
      }

#line 181 "View.MemoryView"
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {

#line 181 "View.MemoryView"
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");

#line 181 "View.MemoryView"
        __PYX_ERR(2, 181, __pyx_L1_error)

#line 181 "View.MemoryView"
      }

#line 181 "View.MemoryView"
      __pyx_t_1 = __Pyx_div_Py_ssize_t(__pyx_v_self->len, __pyx_v_itemsize);

#line 181 "View.MemoryView"
      __pyx_t_9 = __pyx_t_1;

#line 181 "View.MemoryView"
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {

#line 181 "View.MemoryView"
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":182
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */

#line 182 "View.MemoryView"
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":183
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */

#line 183 "View.MemoryView"
        Py_INCREF(Py_None);

#line 183 "View.MemoryView"
      }

      /* "View.MemoryView":179
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */

#line 179 "View.MemoryView"
    }

    /* "View.MemoryView":172
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 172 "View.MemoryView"
  }

  /* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
  /* function exit code */

#line 123 "View.MemoryView"
  __pyx_r = 0;

#line 123 "View.MemoryView"
  goto __pyx_L0;

#line 123 "View.MemoryView"
  __pyx_L1_error:;

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_10);

#line 123 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 123 "View.MemoryView"
  __pyx_r = -1;

#line 123 "View.MemoryView"
  __pyx_L0:;

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_format);

#line 123 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 123 "View.MemoryView"
  return __pyx_r;

#line 123 "View.MemoryView"
}

/* "View.MemoryView":186
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
/* Python wrapper */

#line 186 "View.MemoryView"
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/

#line 186 "View.MemoryView"
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {

#line 186 "View.MemoryView"
  int __pyx_r;

#line 186 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 186 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = 
#line 186 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
  /* function exit code */

#line 186 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 186 "View.MemoryView"
  return __pyx_r;

#line 186 "View.MemoryView"
}

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode
#line 186 "View.MemoryView"
;

#line 186 "View.MemoryView"
  int __pyx_r;

#line 186 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 186 "View.MemoryView"
  int __pyx_t_1;

#line 186 "View.MemoryView"
  int __pyx_t_2;

#line 186 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 186 "View.MemoryView"
  char *__pyx_t_4;

#line 186 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 186 "View.MemoryView"
  int __pyx_t_6;

#line 186 "View.MemoryView"
  Py_ssize_t *__pyx_t_7;

#line 186 "View.MemoryView"
  int __pyx_lineno = 0;

#line 186 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 186 "View.MemoryView"
  int __pyx_clineno = 0;

#line 186 "View.MemoryView"
  if (__pyx_v_info == NULL) {

#line 186 "View.MemoryView"
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");

#line 186 "View.MemoryView"
    return -1;

#line 186 "View.MemoryView"
  }

#line 186 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__", 0);

#line 186 "View.MemoryView"
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);

#line 186 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":187
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */

#line 187 "View.MemoryView"
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":188
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */

#line 188 "View.MemoryView"
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 188, __pyx_L1_error)

#line 188 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 188 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":189
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */

#line 189 "View.MemoryView"
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":188
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */

#line 188 "View.MemoryView"
    goto __pyx_L3;

#line 188 "View.MemoryView"
  }

  /* "View.MemoryView":190
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */

#line 190 "View.MemoryView"
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 190, __pyx_L1_error)

#line 190 "View.MemoryView"
  __pyx_t_1 = (__pyx_t_2 != 0);

#line 190 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":191
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */

#line 191 "View.MemoryView"
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":190
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */

#line 190 "View.MemoryView"
  }

#line 190 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":192
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */

#line 192 "View.MemoryView"
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);

#line 192 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":193
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */

#line 193 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 193, __pyx_L1_error)

#line 193 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 193 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 193 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 193 "View.MemoryView"
    __PYX_ERR(2, 193, __pyx_L1_error)

    /* "View.MemoryView":192
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */

#line 192 "View.MemoryView"
  }

  /* "View.MemoryView":194
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */

#line 194 "View.MemoryView"
  __pyx_t_4 = __pyx_v_self->data;

#line 194 "View.MemoryView"
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":195
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */

#line 195 "View.MemoryView"
  __pyx_t_5 = __pyx_v_self->len;

#line 195 "View.MemoryView"
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":196
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */

#line 196 "View.MemoryView"
  __pyx_t_6 = __pyx_v_self->ndim;

#line 196 "View.MemoryView"
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":197
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */

#line 197 "View.MemoryView"
  __pyx_t_7 = __pyx_v_self->_shape;

#line 197 "View.MemoryView"
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */

#line 198 "View.MemoryView"
  __pyx_t_7 = __pyx_v_self->_strides;

#line 198 "View.MemoryView"
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":199
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */

#line 199 "View.MemoryView"
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":200
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */

#line 200 "View.MemoryView"
  __pyx_t_5 = __pyx_v_self->itemsize;

#line 200 "View.MemoryView"
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":201
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */

#line 201 "View.MemoryView"
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":203
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */

#line 203 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);

#line 203 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":204
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */

#line 204 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->format;

#line 204 "View.MemoryView"
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":203
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */

#line 203 "View.MemoryView"
    goto __pyx_L5;

#line 203 "View.MemoryView"
  }

  /* "View.MemoryView":206
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */

#line 206 "View.MemoryView"
  /*else*/ {

#line 206 "View.MemoryView"
    __pyx_v_info->format = NULL;

#line 206 "View.MemoryView"
  }

#line 206 "View.MemoryView"
  __pyx_L5:;

  /* "View.MemoryView":208
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */

#line 208 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 208 "View.MemoryView"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 208 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_info->obj);

#line 208 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_info->obj);

#line 208 "View.MemoryView"
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":186
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
  /* function exit code */

#line 186 "View.MemoryView"
  __pyx_r = 0;

#line 186 "View.MemoryView"
  goto __pyx_L0;

#line 186 "View.MemoryView"
  __pyx_L1_error:;

#line 186 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 186 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 186 "View.MemoryView"
  __pyx_r = -1;

#line 186 "View.MemoryView"
  if (__pyx_v_info->obj != NULL) {

#line 186 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 186 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 186 "View.MemoryView"
  }

#line 186 "View.MemoryView"
  goto __pyx_L2;

#line 186 "View.MemoryView"
  __pyx_L0:;

#line 186 "View.MemoryView"
  if (__pyx_v_info->obj == Py_None) {

#line 186 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 186 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 186 "View.MemoryView"
  }

#line 186 "View.MemoryView"
  __pyx_L2:;

#line 186 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 186 "View.MemoryView"
  return __pyx_r;

#line 186 "View.MemoryView"
}

/* "View.MemoryView":212
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
/* Python wrapper */

#line 212 "View.MemoryView"
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/

#line 212 "View.MemoryView"
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {

#line 212 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 212 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);

#line 212 "View.MemoryView"
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
  /* function exit code */

#line 212 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 212 "View.MemoryView"
}

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {

#line 212 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 212 "View.MemoryView"
  int __pyx_t_1;

#line 212 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":213
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */

#line 213 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);

#line 213 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":214
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */

#line 214 "View.MemoryView"
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":213
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */

#line 213 "View.MemoryView"
    goto __pyx_L3;

#line 213 "View.MemoryView"
  }

  /* "View.MemoryView":215
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */

#line 215 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->free_data != 0);

#line 215 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":216
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */

#line 216 "View.MemoryView"
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);

#line 216 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":217
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */

#line 217 "View.MemoryView"
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":216
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */

#line 216 "View.MemoryView"
    }

    /* "View.MemoryView":219
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */

#line 219 "View.MemoryView"
    free(__pyx_v_self->data);

    /* "View.MemoryView":215
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */

#line 215 "View.MemoryView"
  }

#line 215 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":220
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 220 "View.MemoryView"
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":212
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
  /* function exit code */

#line 212 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 212 "View.MemoryView"
}

/* "View.MemoryView":223
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
/* Python wrapper */

#line 223 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 223 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {

#line 223 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 223 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 223 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 223 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
  /* function exit code */

#line 223 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 223 "View.MemoryView"
  return __pyx_r;

#line 223 "View.MemoryView"
}

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {

#line 223 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 223 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 223 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 223 "View.MemoryView"
  int __pyx_lineno = 0;

#line 223 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 223 "View.MemoryView"
  int __pyx_clineno = 0;

#line 223 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":224
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */

#line 224 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 224 "View.MemoryView"
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 224, __pyx_L1_error)

#line 224 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 224 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 224 "View.MemoryView"
  __pyx_t_1 = 0;

#line 224 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":223
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
  /* function exit code */

#line 223 "View.MemoryView"
  __pyx_L1_error:;

#line 223 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 223 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 223 "View.MemoryView"
  __pyx_r = NULL;

#line 223 "View.MemoryView"
  __pyx_L0:;

#line 223 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 223 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 223 "View.MemoryView"
  return __pyx_r;

#line 223 "View.MemoryView"
}

/* "View.MemoryView":227
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

#line 227 "View.MemoryView"


#line 227 "View.MemoryView"
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags
#line 227 "View.MemoryView"
;

#line 227 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 227 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 227 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 227 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 227 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 227 "View.MemoryView"
  int __pyx_lineno = 0;

#line 227 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 227 "View.MemoryView"
  int __pyx_clineno = 0;

#line 227 "View.MemoryView"
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":228
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */

#line 228 "View.MemoryView"
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":229
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 229 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 229 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 229 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 229 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 229 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 229 "View.MemoryView"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 229 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));

#line 229 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 229 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);

#line 229 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 229 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 229 "View.MemoryView"
  __pyx_t_1 = 0;

#line 229 "View.MemoryView"
  __pyx_t_2 = 0;

#line 229 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 229 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 229 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 229 "View.MemoryView"
  __pyx_t_2 = 0;

#line 229 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":227
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

#line 227 "View.MemoryView"


#line 227 "View.MemoryView"
  /* function exit code */

#line 227 "View.MemoryView"
  __pyx_L1_error:;

#line 227 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 227 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 227 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 227 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 227 "View.MemoryView"
  __pyx_r = 0;

#line 227 "View.MemoryView"
  __pyx_L0:;

#line 227 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 227 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 227 "View.MemoryView"
  return __pyx_r;

#line 227 "View.MemoryView"
}

/* "View.MemoryView":231
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
/* Python wrapper */

#line 231 "View.MemoryView"
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/

#line 231 "View.MemoryView"
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {

#line 231 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 231 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 231 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 231 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
  /* function exit code */

#line 231 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 231 "View.MemoryView"
  return __pyx_r;

#line 231 "View.MemoryView"
}

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {

#line 231 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 231 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 231 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":232
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */

#line 232 "View.MemoryView"
  __pyx_r = (__pyx_v_self->_shape[0]);

#line 232 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":231
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
  /* function exit code */

#line 231 "View.MemoryView"
  __pyx_L0:;

#line 231 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 231 "View.MemoryView"
  return __pyx_r;

#line 231 "View.MemoryView"
}

/* "View.MemoryView":234
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
/* Python wrapper */

#line 234 "View.MemoryView"
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/

#line 234 "View.MemoryView"
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {

#line 234 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 234 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 234 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = 
#line 234 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
  /* function exit code */

#line 234 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 234 "View.MemoryView"
  return __pyx_r;

#line 234 "View.MemoryView"
}

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {

#line 234 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 234 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 234 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 234 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 234 "View.MemoryView"
  int __pyx_lineno = 0;

#line 234 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 234 "View.MemoryView"
  int __pyx_clineno = 0;

#line 234 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":235
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */

#line 235 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 235 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 235, __pyx_L1_error)

#line 235 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 235 "View.MemoryView"
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 235, __pyx_L1_error)

#line 235 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 235 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 235 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 235 "View.MemoryView"
  __pyx_t_2 = 0;

#line 235 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":234
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
  /* function exit code */

#line 234 "View.MemoryView"
  __pyx_L1_error:;

#line 234 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 234 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 234 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 234 "View.MemoryView"
  __pyx_r = NULL;

#line 234 "View.MemoryView"
  __pyx_L0:;

#line 234 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 234 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 234 "View.MemoryView"
  return __pyx_r;

#line 234 "View.MemoryView"
}

/* "View.MemoryView":237
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
/* Python wrapper */

#line 237 "View.MemoryView"
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/

#line 237 "View.MemoryView"
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {

#line 237 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 237 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 237 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 237 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
  /* function exit code */

#line 237 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 237 "View.MemoryView"
  return __pyx_r;

#line 237 "View.MemoryView"
}

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {

#line 237 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 237 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 237 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 237 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 237 "View.MemoryView"
  int __pyx_lineno = 0;

#line 237 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 237 "View.MemoryView"
  int __pyx_clineno = 0;

#line 237 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":238
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */

#line 238 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 238 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 238, __pyx_L1_error)

#line 238 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 238 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 238, __pyx_L1_error)

#line 238 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 238 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 238 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 238 "View.MemoryView"
  __pyx_t_2 = 0;

#line 238 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":237
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
  /* function exit code */

#line 237 "View.MemoryView"
  __pyx_L1_error:;

#line 237 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 237 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 237 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 237 "View.MemoryView"
  __pyx_r = NULL;

#line 237 "View.MemoryView"
  __pyx_L0:;

#line 237 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 237 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 237 "View.MemoryView"
  return __pyx_r;

#line 237 "View.MemoryView"
}

/* "View.MemoryView":240
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
/* Python wrapper */

#line 240 "View.MemoryView"
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/

#line 240 "View.MemoryView"
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {

#line 240 "View.MemoryView"
  int __pyx_r;

#line 240 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 240 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = 
#line 240 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
  /* function exit code */

#line 240 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 240 "View.MemoryView"
  return __pyx_r;

#line 240 "View.MemoryView"
}

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {

#line 240 "View.MemoryView"
  int __pyx_r;

#line 240 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 240 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 240 "View.MemoryView"
  int __pyx_lineno = 0;

#line 240 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 240 "View.MemoryView"
  int __pyx_clineno = 0;

#line 240 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":241
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 241 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 241, __pyx_L1_error)

#line 241 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 241 "View.MemoryView"
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(2, 241, __pyx_L1_error)

#line 241 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":240
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
  /* function exit code */

#line 240 "View.MemoryView"
  __pyx_r = 0;

#line 240 "View.MemoryView"
  goto __pyx_L0;

#line 240 "View.MemoryView"
  __pyx_L1_error:;

#line 240 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 240 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 240 "View.MemoryView"
  __pyx_r = -1;

#line 240 "View.MemoryView"
  __pyx_L0:;

#line 240 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 240 "View.MemoryView"
  return __pyx_r;

#line 240 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 2 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 2 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 2 "(tree fragment)"
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
/* Python wrapper */

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 3 "(tree fragment)"
__pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 3 "(tree fragment)"
  int __pyx_lineno = 0;

#line 3 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 3 "(tree fragment)"
  int __pyx_clineno = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __pyx_L1_error:;

#line 3 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "(tree fragment)"
  __pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

/* "View.MemoryView":245
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

#line 245 "View.MemoryView"


#line 245 "View.MemoryView"
static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0
#line 245 "View.MemoryView"
;

#line 245 "View.MemoryView"
  struct __pyx_array_obj *__pyx_r = NULL;

#line 245 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 245 "View.MemoryView"
  int __pyx_t_1;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 245 "View.MemoryView"
  int __pyx_lineno = 0;

#line 245 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 245 "View.MemoryView"
  int __pyx_clineno = 0;

#line 245 "View.MemoryView"
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":249
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */

#line 249 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);

#line 249 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":250
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */

#line 250 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 250 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 250 "View.MemoryView"
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 250 "View.MemoryView"
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 250 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_shape);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_shape);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_2);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_3);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_4);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);

#line 250 "View.MemoryView"
    __pyx_t_2 = 0;

#line 250 "View.MemoryView"
    __pyx_t_3 = 0;

#line 250 "View.MemoryView"
    __pyx_t_4 = 0;

#line 250 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 250 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 250 "View.MemoryView"
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);

#line 250 "View.MemoryView"
    __pyx_t_4 = 0;

    /* "View.MemoryView":249
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */

#line 249 "View.MemoryView"
    goto __pyx_L3;

#line 249 "View.MemoryView"
  }

  /* "View.MemoryView":252
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */

#line 252 "View.MemoryView"
  /*else*/ {

#line 252 "View.MemoryView"
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 252 "View.MemoryView"
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 252 "View.MemoryView"
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 252 "View.MemoryView"
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 252 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_shape);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_shape);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_4);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_5);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_3);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);

#line 252 "View.MemoryView"
    __pyx_t_4 = 0;

#line 252 "View.MemoryView"
    __pyx_t_5 = 0;

#line 252 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":253
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */

#line 253 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 253, __pyx_L1_error)

#line 253 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 253 "View.MemoryView"
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(2, 253, __pyx_L1_error)

    /* "View.MemoryView":252
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */

#line 252 "View.MemoryView"
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 252 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 252 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 252 "View.MemoryView"
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);

#line 252 "View.MemoryView"
    __pyx_t_5 = 0;

    /* "View.MemoryView":254
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */

#line 254 "View.MemoryView"
    __pyx_v_result->data = __pyx_v_buf;

#line 254 "View.MemoryView"
  }

#line 254 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":256
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 256 "View.MemoryView"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 256 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 256 "View.MemoryView"
  __pyx_r = __pyx_v_result;

#line 256 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":245
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

#line 245 "View.MemoryView"


#line 245 "View.MemoryView"
  /* function exit code */

#line 245 "View.MemoryView"
  __pyx_L1_error:;

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 245 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 245 "View.MemoryView"
  __pyx_r = 0;

#line 245 "View.MemoryView"
  __pyx_L0:;

#line 245 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 245 "View.MemoryView"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 245 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 245 "View.MemoryView"
  return __pyx_r;

#line 245 "View.MemoryView"
}

/* "View.MemoryView":282
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
/* Python wrapper */

#line 282 "View.MemoryView"
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 282 "View.MemoryView"
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0
#line 282 "View.MemoryView"
;

#line 282 "View.MemoryView"
  int __pyx_lineno = 0;

#line 282 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 282 "View.MemoryView"
  int __pyx_clineno = 0;

#line 282 "View.MemoryView"
  int __pyx_r;

#line 282 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 282 "View.MemoryView"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 282 "View.MemoryView"
  {

#line 282 "View.MemoryView"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,0};

#line 282 "View.MemoryView"
    PyObject* values[1] = {0};

#line 282 "View.MemoryView"
    if (unlikely(__pyx_kwds)) {

#line 282 "View.MemoryView"
      Py_ssize_t kw_args;

#line 282 "View.MemoryView"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 282 "View.MemoryView"
      switch (pos_args) {
        case  1: 
#line 282 "View.MemoryView"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 282 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 282 "View.MemoryView"
        case  0: break;
        default: 
#line 282 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 282 "View.MemoryView"
      }

#line 282 "View.MemoryView"
      kw_args = PyDict_Size(__pyx_kwds);

#line 282 "View.MemoryView"
      switch (pos_args) {

#line 282 "View.MemoryView"
        case  0:

#line 282 "View.MemoryView"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2)) != 0)) kw_args--;
        else 
#line 282 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 282 "View.MemoryView"
      }

#line 282 "View.MemoryView"
      if (unlikely(kw_args > 0)) {

#line 282 "View.MemoryView"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(2, 282, __pyx_L3_error)

#line 282 "View.MemoryView"
      }

#line 282 "View.MemoryView"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {

#line 282 "View.MemoryView"
      goto __pyx_L5_argtuple_error;

#line 282 "View.MemoryView"
    } else {

#line 282 "View.MemoryView"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 282 "View.MemoryView"
    }

#line 282 "View.MemoryView"
    __pyx_v_name = values[0];

#line 282 "View.MemoryView"
  }

#line 282 "View.MemoryView"
  goto __pyx_L4_argument_unpacking_done;

#line 282 "View.MemoryView"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 282 "View.MemoryView"
__PYX_ERR(2, 282, __pyx_L3_error)

#line 282 "View.MemoryView"
  __pyx_L3_error:;

#line 282 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 282 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 282 "View.MemoryView"
  return -1;

#line 282 "View.MemoryView"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 282 "View.MemoryView"
__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
  /* function exit code */

#line 282 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 282 "View.MemoryView"
  return __pyx_r;

#line 282 "View.MemoryView"
}

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {

#line 282 "View.MemoryView"
  int __pyx_r;

#line 282 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 282 "View.MemoryView"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":283
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */

#line 283 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_name);

#line 283 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_name);

#line 283 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_self->name);

#line 283 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_self->name);

#line 283 "View.MemoryView"
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":282
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
  /* function exit code */

#line 282 "View.MemoryView"
  __pyx_r = 0;

#line 282 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 282 "View.MemoryView"
  return __pyx_r;

#line 282 "View.MemoryView"
}

/* "View.MemoryView":284
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
/* Python wrapper */

#line 284 "View.MemoryView"
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/

#line 284 "View.MemoryView"
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {

#line 284 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 284 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 284 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 284 "View.MemoryView"
__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
  /* function exit code */

#line 284 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 284 "View.MemoryView"
  return __pyx_r;

#line 284 "View.MemoryView"
}

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {

#line 284 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 284 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 284 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":285
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */

#line 285 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 285 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->name);

#line 285 "View.MemoryView"
  __pyx_r = __pyx_v_self->name;

#line 285 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":284
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
  /* function exit code */

#line 284 "View.MemoryView"
  __pyx_L0:;

#line 284 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 284 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 284 "View.MemoryView"
  return __pyx_r;

#line 284 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->name);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->name);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_1;

#line 6 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_2 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));

#line 8 "(tree fragment)"
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_3 = (__pyx_v_self->name != Py_None);

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_3;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_3 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_184977713);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_184977713);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_5;

#line 13 "(tree fragment)"
    __pyx_t_5 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 15 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_184977713);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_184977713);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_5);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);

#line 15 "(tree fragment)"
    __pyx_t_5 = 0;

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_4;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "View.MemoryView":299
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

#line 299 "View.MemoryView"


#line 299 "View.MemoryView"
static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p
#line 299 "View.MemoryView"
;
  size_t __pyx_v_offset
#line 299 "View.MemoryView"
;

#line 299 "View.MemoryView"
  void *__pyx_r;

#line 299 "View.MemoryView"
  int __pyx_t_1;

  /* "View.MemoryView":301
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */

#line 301 "View.MemoryView"
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":305
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */

#line 305 "View.MemoryView"
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":307
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */

#line 307 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);

#line 307 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":308
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */

#line 308 "View.MemoryView"
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":307
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */

#line 307 "View.MemoryView"
  }

  /* "View.MemoryView":310
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 310 "View.MemoryView"
  __pyx_r = ((void *)__pyx_v_aligned_p);

#line 310 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":299
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

#line 299 "View.MemoryView"


#line 299 "View.MemoryView"
  /* function exit code */

#line 299 "View.MemoryView"
  __pyx_L0:;

#line 299 "View.MemoryView"
  return __pyx_r;

#line 299 "View.MemoryView"
}

/* "View.MemoryView":346
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
/* Python wrapper */

#line 346 "View.MemoryView"
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 346 "View.MemoryView"
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0
#line 346 "View.MemoryView"
;
  int __pyx_v_flags
#line 346 "View.MemoryView"
;
  int __pyx_v_dtype_is_object
#line 346 "View.MemoryView"
;

#line 346 "View.MemoryView"
  int __pyx_lineno = 0;

#line 346 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 346 "View.MemoryView"
  int __pyx_clineno = 0;

#line 346 "View.MemoryView"
  int __pyx_r;

#line 346 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 346 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);

#line 346 "View.MemoryView"
  {

#line 346 "View.MemoryView"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};

#line 346 "View.MemoryView"
    PyObject* values[3] = {0,0,0};

#line 346 "View.MemoryView"
    if (unlikely(__pyx_kwds)) {

#line 346 "View.MemoryView"
      Py_ssize_t kw_args;

#line 346 "View.MemoryView"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 346 "View.MemoryView"
      switch (pos_args) {
        case  3: 
#line 346 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 346 "View.MemoryView"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 346 "View.MemoryView"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 346 "View.MemoryView"
        case  0: break;
        default: 
#line 346 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
      kw_args = PyDict_Size(__pyx_kwds);

#line 346 "View.MemoryView"
      switch (pos_args) {

#line 346 "View.MemoryView"
        case  0:

#line 346 "View.MemoryView"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else 
#line 346 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 346 "View.MemoryView"
        case  1:

#line 346 "View.MemoryView"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;

#line 346 "View.MemoryView"
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); 
#line 346 "View.MemoryView"
__PYX_ERR(2, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
        }

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 346 "View.MemoryView"
        case  2:

#line 346 "View.MemoryView"
        if (kw_args > 0) {

#line 346 "View.MemoryView"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);

#line 346 "View.MemoryView"
          if (value) { values[2] = value; kw_args--; }

#line 346 "View.MemoryView"
        }

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
      if (unlikely(kw_args > 0)) {

#line 346 "View.MemoryView"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
    } else {

#line 346 "View.MemoryView"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 346 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 346 "View.MemoryView"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 346 "View.MemoryView"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 346 "View.MemoryView"
        break;
        default: 
#line 346 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
    }

#line 346 "View.MemoryView"
    __pyx_v_obj = values[0];

#line 346 "View.MemoryView"
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
    if (values[2]) {

#line 346 "View.MemoryView"
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
    } else {

#line 346 "View.MemoryView"
      __pyx_v_dtype_is_object = ((int)0);

#line 346 "View.MemoryView"
    }

#line 346 "View.MemoryView"
  }

#line 346 "View.MemoryView"
  goto __pyx_L4_argument_unpacking_done;

#line 346 "View.MemoryView"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 346 "View.MemoryView"
__PYX_ERR(2, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
  __pyx_L3_error:;

#line 346 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 346 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 346 "View.MemoryView"
  return -1;

#line 346 "View.MemoryView"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 346 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
  /* function exit code */

#line 346 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 346 "View.MemoryView"
  return __pyx_r;

#line 346 "View.MemoryView"
}

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {

#line 346 "View.MemoryView"
  int __pyx_r;

#line 346 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 346 "View.MemoryView"
  int __pyx_t_1;

#line 346 "View.MemoryView"
  int __pyx_t_2;

#line 346 "View.MemoryView"
  int __pyx_t_3;

#line 346 "View.MemoryView"
  int __pyx_t_4;

#line 346 "View.MemoryView"
  int __pyx_lineno = 0;

#line 346 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 346 "View.MemoryView"
  int __pyx_clineno = 0;

#line 346 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":347
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */

#line 347 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_obj);

#line 347 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_obj);

#line 347 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_self->obj);

#line 347 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_self->obj);

#line 347 "View.MemoryView"
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":348
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */

#line 348 "View.MemoryView"
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":349
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */

#line 349 "View.MemoryView"
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));

#line 349 "View.MemoryView"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 349 "View.MemoryView"
  if (!__pyx_t_3) {

#line 349 "View.MemoryView"
  } else {

#line 349 "View.MemoryView"
    __pyx_t_1 = __pyx_t_3;

#line 349 "View.MemoryView"
    goto __pyx_L4_bool_binop_done;

#line 349 "View.MemoryView"
  }

#line 349 "View.MemoryView"
  __pyx_t_3 = (__pyx_v_obj != Py_None);

#line 349 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 349 "View.MemoryView"
  __pyx_t_1 = __pyx_t_2;

#line 349 "View.MemoryView"
  __pyx_L4_bool_binop_done:;

#line 349 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":350
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */

#line 350 "View.MemoryView"
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 350, __pyx_L1_error)

    /* "View.MemoryView":351
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */

#line 351 "View.MemoryView"
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);

#line 351 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":352
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */

#line 352 "View.MemoryView"
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":353
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 */

#line 353 "View.MemoryView"
      Py_INCREF(Py_None);

      /* "View.MemoryView":351
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */

#line 351 "View.MemoryView"
    }

    /* "View.MemoryView":349
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */

#line 349 "View.MemoryView"
  }

  /* "View.MemoryView":355
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 */

#line 355 "View.MemoryView"
  __pyx_t_1 = ((!(__PYX_CYTHON_ATOMICS_ENABLED() != 0)) != 0);

#line 355 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":357
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */

#line 357 "View.MemoryView"
    __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);

#line 357 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":358
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 */

#line 358 "View.MemoryView"
      __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

      /* "View.MemoryView":359
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 */

#line 359 "View.MemoryView"
      __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

      /* "View.MemoryView":357
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */

#line 357 "View.MemoryView"
    }

    /* "View.MemoryView":360
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */

#line 360 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);

#line 360 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":361
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *                 if self.lock is NULL:
 *                     raise MemoryError
 */

#line 361 "View.MemoryView"
      __pyx_v_self->lock = PyThread_allocate_lock();

      /* "View.MemoryView":362
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */

#line 362 "View.MemoryView"
      __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);

#line 362 "View.MemoryView"
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":363
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 *                     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */

#line 363 "View.MemoryView"
        PyErr_NoMemory(); __PYX_ERR(2, 363, __pyx_L1_error)

        /* "View.MemoryView":362
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */

#line 362 "View.MemoryView"
      }

      /* "View.MemoryView":360
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */

#line 360 "View.MemoryView"
    }

    /* "View.MemoryView":355
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 */

#line 355 "View.MemoryView"
  }

  /* "View.MemoryView":365
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */

#line 365 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);

#line 365 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":366
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */

#line 366 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);

#line 366 "View.MemoryView"
    if (__pyx_t_2) {

#line 366 "View.MemoryView"
    } else {

#line 366 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 366 "View.MemoryView"
      goto __pyx_L12_bool_binop_done;

#line 366 "View.MemoryView"
    }

#line 366 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);

#line 366 "View.MemoryView"
    __pyx_t_1 = __pyx_t_2;

#line 366 "View.MemoryView"
    __pyx_L12_bool_binop_done:;

#line 366 "View.MemoryView"
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":365
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */

#line 365 "View.MemoryView"
    goto __pyx_L11;

#line 365 "View.MemoryView"
  }

  /* "View.MemoryView":368
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */

#line 368 "View.MemoryView"
  /*else*/ {

#line 368 "View.MemoryView"
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;

#line 368 "View.MemoryView"
  }

#line 368 "View.MemoryView"
  __pyx_L11:;

  /* "View.MemoryView":370
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */

#line 370 "View.MemoryView"
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":372
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */

#line 372 "View.MemoryView"
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":346
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
  /* function exit code */

#line 346 "View.MemoryView"
  __pyx_r = 0;

#line 346 "View.MemoryView"
  goto __pyx_L0;

#line 346 "View.MemoryView"
  __pyx_L1_error:;

#line 346 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 346 "View.MemoryView"
  __pyx_r = -1;

#line 346 "View.MemoryView"
  __pyx_L0:;

#line 346 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 346 "View.MemoryView"
  return __pyx_r;

#line 346 "View.MemoryView"
}

/* "View.MemoryView":374
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
/* Python wrapper */

#line 374 "View.MemoryView"
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/

#line 374 "View.MemoryView"
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {

#line 374 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 374 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);

#line 374 "View.MemoryView"
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
  /* function exit code */

#line 374 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 374 "View.MemoryView"
}

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i
#line 374 "View.MemoryView"
;

#line 374 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 374 "View.MemoryView"
  int __pyx_t_1;

#line 374 "View.MemoryView"
  int __pyx_t_2;

#line 374 "View.MemoryView"
  int __pyx_t_3;

#line 374 "View.MemoryView"
  int __pyx_t_4;

#line 374 "View.MemoryView"
  int __pyx_t_5;

#line 374 "View.MemoryView"
  PyThread_type_lock __pyx_t_6;

#line 374 "View.MemoryView"
  PyThread_type_lock __pyx_t_7;

#line 374 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":375
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */

#line 375 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);

#line 375 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 375 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":376
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */

#line 376 "View.MemoryView"
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":375
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */

#line 375 "View.MemoryView"
    goto __pyx_L3;

#line 375 "View.MemoryView"
  }

  /* "View.MemoryView":377
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */

#line 377 "View.MemoryView"
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);

#line 377 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":379
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */

#line 379 "View.MemoryView"
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":380
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */

#line 380 "View.MemoryView"
    Py_DECREF(Py_None);

    /* "View.MemoryView":377
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */

#line 377 "View.MemoryView"
  }

#line 377 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":384
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */

#line 384 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);

#line 384 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":385
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */

#line 385 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;

#line 385 "View.MemoryView"
    __pyx_t_4 = __pyx_t_3;

#line 385 "View.MemoryView"
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {

#line 385 "View.MemoryView"
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":386
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */

#line 386 "View.MemoryView"
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);

#line 386 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":387
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */

#line 387 "View.MemoryView"
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":388
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */

#line 388 "View.MemoryView"
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);

#line 388 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":390
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */

#line 390 "View.MemoryView"
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

#line 390 "View.MemoryView"
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":389
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */

#line 389 "View.MemoryView"
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;

#line 389 "View.MemoryView"
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":388
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */

#line 388 "View.MemoryView"
        }

        /* "View.MemoryView":391
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */

#line 391 "View.MemoryView"
        goto __pyx_L6_break;

        /* "View.MemoryView":386
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */

#line 386 "View.MemoryView"
      }

#line 386 "View.MemoryView"
    }

#line 386 "View.MemoryView"
    /*else*/ {

      /* "View.MemoryView":393
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */

#line 393 "View.MemoryView"
      PyThread_free_lock(__pyx_v_self->lock);

#line 393 "View.MemoryView"
    }

#line 393 "View.MemoryView"
    __pyx_L6_break:;

    /* "View.MemoryView":384
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */

#line 384 "View.MemoryView"
  }

  /* "View.MemoryView":374
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
  /* function exit code */

#line 374 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 374 "View.MemoryView"
}

/* "View.MemoryView":395
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

#line 395 "View.MemoryView"


#line 395 "View.MemoryView"
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim
#line 395 "View.MemoryView"
;
  char *__pyx_v_itemp
#line 395 "View.MemoryView"
;
  PyObject *__pyx_v_idx = NULL
#line 395 "View.MemoryView"
;

#line 395 "View.MemoryView"
  char *__pyx_r;

#line 395 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 395 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 395 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 395 "View.MemoryView"
  Py_ssize_t __pyx_t_3;

#line 395 "View.MemoryView"
  PyObject *(*__pyx_t_4)(PyObject *);

#line 395 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 395 "View.MemoryView"
  Py_ssize_t __pyx_t_6;

#line 395 "View.MemoryView"
  char *__pyx_t_7;

#line 395 "View.MemoryView"
  int __pyx_lineno = 0;

#line 395 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 395 "View.MemoryView"
  int __pyx_clineno = 0;

#line 395 "View.MemoryView"
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":397
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */

#line 397 "View.MemoryView"
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":399
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */

#line 399 "View.MemoryView"
  __pyx_t_1 = 0;

#line 399 "View.MemoryView"
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {

#line 399 "View.MemoryView"
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;

#line 399 "View.MemoryView"
    __pyx_t_4 = NULL;

#line 399 "View.MemoryView"
  } else {
    __pyx_t_3 = -1; 
#line 399 "View.MemoryView"
__pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 399 "View.MemoryView"
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
  }

#line 399 "View.MemoryView"
  for (;;) {

#line 399 "View.MemoryView"
    if (likely(!__pyx_t_4)) {

#line 399 "View.MemoryView"
      if (likely(PyList_CheckExact(__pyx_t_2))) {

#line 399 "View.MemoryView"
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;

#line 399 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 399 "View.MemoryView"
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        #else

#line 399 "View.MemoryView"
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_5);

#line 399 "View.MemoryView"
        #endif

#line 399 "View.MemoryView"
      } else {

#line 399 "View.MemoryView"
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;

#line 399 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 399 "View.MemoryView"
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        #else

#line 399 "View.MemoryView"
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_5);

#line 399 "View.MemoryView"
        #endif

#line 399 "View.MemoryView"
      }
    } else 
#line 399 "View.MemoryView"
{

#line 399 "View.MemoryView"
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);

#line 399 "View.MemoryView"
      if (unlikely(!__pyx_t_5)) {

#line 399 "View.MemoryView"
        PyObject* exc_type = PyErr_Occurred();

#line 399 "View.MemoryView"
        if (exc_type) {

#line 399 "View.MemoryView"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 399 "View.MemoryView"
          else __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        }

#line 399 "View.MemoryView"
        break;

#line 399 "View.MemoryView"
      }

#line 399 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 399 "View.MemoryView"
    }

#line 399 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);

#line 399 "View.MemoryView"
    __pyx_t_5 = 0;

#line 399 "View.MemoryView"
    __pyx_v_dim = __pyx_t_1;

#line 399 "View.MemoryView"
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":400
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */

#line 400 "View.MemoryView"
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 400, __pyx_L1_error)

#line 400 "View.MemoryView"
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(2, 400, __pyx_L1_error)

#line 400 "View.MemoryView"
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":399
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */

#line 399 "View.MemoryView"
  }

#line 399 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":402
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 402 "View.MemoryView"
  __pyx_r = __pyx_v_itemp;

#line 402 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":395
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

#line 395 "View.MemoryView"


#line 395 "View.MemoryView"
  /* function exit code */

#line 395 "View.MemoryView"
  __pyx_L1_error:;

#line 395 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 395 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 395 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 395 "View.MemoryView"
  __pyx_r = NULL;

#line 395 "View.MemoryView"
  __pyx_L0:;

#line 395 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_idx);

#line 395 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 395 "View.MemoryView"
  return __pyx_r;

#line 395 "View.MemoryView"
}

/* "View.MemoryView":405
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
/* Python wrapper */

#line 405 "View.MemoryView"
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/

#line 405 "View.MemoryView"
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {

#line 405 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 405 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 405 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 405 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
  /* function exit code */

#line 405 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 405 "View.MemoryView"
  return __pyx_r;

#line 405 "View.MemoryView"
}

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL
#line 405 "View.MemoryView"
;
  PyObject *__pyx_v_indices = NULL
#line 405 "View.MemoryView"
;
  char *__pyx_v_itemp
#line 405 "View.MemoryView"
;

#line 405 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 405 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 405 "View.MemoryView"
  int __pyx_t_1;

#line 405 "View.MemoryView"
  int __pyx_t_2;

#line 405 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 405 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 405 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 405 "View.MemoryView"
  char *__pyx_t_6;

#line 405 "View.MemoryView"
  int __pyx_lineno = 0;

#line 405 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 405 "View.MemoryView"
  int __pyx_clineno = 0;

#line 405 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":406
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */

#line 406 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);

#line 406 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 406 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":407
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */

#line 407 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 407 "View.MemoryView"
    __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 407 "View.MemoryView"
    __pyx_r = ((PyObject *)__pyx_v_self);

#line 407 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":406
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */

#line 406 "View.MemoryView"
  }

  /* "View.MemoryView":409
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */

#line 409 "View.MemoryView"
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 409 "View.MemoryView"
  if (likely(__pyx_t_3 != Py_None)) {

#line 409 "View.MemoryView"
    PyObject* sequence = __pyx_t_3;

#line 409 "View.MemoryView"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 409 "View.MemoryView"
    if (unlikely(size != 2)) {

#line 409 "View.MemoryView"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 409 "View.MemoryView"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 409 "View.MemoryView"
      __PYX_ERR(2, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
    }

#line 409 "View.MemoryView"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 409 "View.MemoryView"
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 

#line 409 "View.MemoryView"
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 

#line 409 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_4);

#line 409 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_5);

#line 409 "View.MemoryView"
    #else

#line 409 "View.MemoryView"
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 409 "View.MemoryView"
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 409 "View.MemoryView"
    #endif

#line 409 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 409 "View.MemoryView"
  } else {

#line 409 "View.MemoryView"
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
  }

#line 409 "View.MemoryView"
  __pyx_v_have_slices = __pyx_t_4;

#line 409 "View.MemoryView"
  __pyx_t_4 = 0;

#line 409 "View.MemoryView"
  __pyx_v_indices = __pyx_t_5;

#line 409 "View.MemoryView"
  __pyx_t_5 = 0;

  /* "View.MemoryView":412
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */

#line 412 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 412, __pyx_L1_error)

#line 412 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":413
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */

#line 413 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 413 "View.MemoryView"
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 413, __pyx_L1_error)

#line 413 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 413 "View.MemoryView"
    __pyx_r = __pyx_t_3;

#line 413 "View.MemoryView"
    __pyx_t_3 = 0;

#line 413 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":412
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */

#line 412 "View.MemoryView"
  }

  /* "View.MemoryView":415
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */

#line 415 "View.MemoryView"
  /*else*/ {

#line 415 "View.MemoryView"
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(2, 415, __pyx_L1_error)

#line 415 "View.MemoryView"
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":416
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */

#line 416 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 416 "View.MemoryView"
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 416, __pyx_L1_error)

#line 416 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 416 "View.MemoryView"
    __pyx_r = __pyx_t_3;

#line 416 "View.MemoryView"
    __pyx_t_3 = 0;

#line 416 "View.MemoryView"
    goto __pyx_L0;

#line 416 "View.MemoryView"
  }

  /* "View.MemoryView":405
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
  /* function exit code */

#line 405 "View.MemoryView"
  __pyx_L1_error:;

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 405 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 405 "View.MemoryView"
  __pyx_r = NULL;

#line 405 "View.MemoryView"
  __pyx_L0:;

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_have_slices);

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_indices);

#line 405 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 405 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 405 "View.MemoryView"
  return __pyx_r;

#line 405 "View.MemoryView"
}

/* "View.MemoryView":418
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
/* Python wrapper */

#line 418 "View.MemoryView"
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/

#line 418 "View.MemoryView"
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {

#line 418 "View.MemoryView"
  int __pyx_r;

#line 418 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 418 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = 
#line 418 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
  /* function exit code */

#line 418 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 418 "View.MemoryView"
  return __pyx_r;

#line 418 "View.MemoryView"
}

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL
#line 418 "View.MemoryView"
;
  PyObject *__pyx_v_obj = NULL
#line 418 "View.MemoryView"
;

#line 418 "View.MemoryView"
  int __pyx_r;

#line 418 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 418 "View.MemoryView"
  int __pyx_t_1;

#line 418 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 418 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 418 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 418 "View.MemoryView"
  int __pyx_lineno = 0;

#line 418 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 418 "View.MemoryView"
  int __pyx_clineno = 0;

#line 418 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__", 0);

#line 418 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":419
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */

#line 419 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);

#line 419 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":420
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */

#line 420 "View.MemoryView"
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 420, __pyx_L1_error)

#line 420 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 420 "View.MemoryView"
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 420 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 420 "View.MemoryView"
    __PYX_ERR(2, 420, __pyx_L1_error)

    /* "View.MemoryView":419
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */

#line 419 "View.MemoryView"
  }

  /* "View.MemoryView":422
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */

#line 422 "View.MemoryView"
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 422 "View.MemoryView"
  if (likely(__pyx_t_2 != Py_None)) {

#line 422 "View.MemoryView"
    PyObject* sequence = __pyx_t_2;

#line 422 "View.MemoryView"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 422 "View.MemoryView"
    if (unlikely(size != 2)) {

#line 422 "View.MemoryView"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 422 "View.MemoryView"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 422 "View.MemoryView"
      __PYX_ERR(2, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
    }

#line 422 "View.MemoryView"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 422 "View.MemoryView"
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 

#line 422 "View.MemoryView"
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 

#line 422 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 422 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_4);

#line 422 "View.MemoryView"
    #else

#line 422 "View.MemoryView"
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 422 "View.MemoryView"
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 422 "View.MemoryView"
    #endif

#line 422 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 422 "View.MemoryView"
  } else {

#line 422 "View.MemoryView"
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
  }

#line 422 "View.MemoryView"
  __pyx_v_have_slices = __pyx_t_3;

#line 422 "View.MemoryView"
  __pyx_t_3 = 0;

#line 422 "View.MemoryView"
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);

#line 422 "View.MemoryView"
  __pyx_t_4 = 0;

  /* "View.MemoryView":424
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */

#line 424 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 424, __pyx_L1_error)

#line 424 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":425
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */

#line 425 "View.MemoryView"
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 425, __pyx_L1_error)

#line 425 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 425 "View.MemoryView"
    __pyx_v_obj = __pyx_t_2;

#line 425 "View.MemoryView"
    __pyx_t_2 = 0;

    /* "View.MemoryView":426
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */

#line 426 "View.MemoryView"
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 426, __pyx_L1_error)

#line 426 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":427
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */

#line 427 "View.MemoryView"
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 427, __pyx_L1_error)

#line 427 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_2);

#line 427 "View.MemoryView"
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 427, __pyx_L1_error)

#line 427 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 427 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 427 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":426
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */

#line 426 "View.MemoryView"
      goto __pyx_L5;

#line 426 "View.MemoryView"
    }

    /* "View.MemoryView":429
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */

#line 429 "View.MemoryView"
    /*else*/ {

#line 429 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 429, __pyx_L1_error)

#line 429 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 429 "View.MemoryView"
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(2, 429, __pyx_L1_error)

#line 429 "View.MemoryView"
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 429, __pyx_L1_error)

#line 429 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_2);

#line 429 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 429 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 429 "View.MemoryView"
    }

#line 429 "View.MemoryView"
    __pyx_L5:;

    /* "View.MemoryView":424
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */

#line 424 "View.MemoryView"
    goto __pyx_L4;

#line 424 "View.MemoryView"
  }

  /* "View.MemoryView":431
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */

#line 431 "View.MemoryView"
  /*else*/ {

#line 431 "View.MemoryView"
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 431, __pyx_L1_error)

#line 431 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 431 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 431 "View.MemoryView"
  }

#line 431 "View.MemoryView"
  __pyx_L4:;

  /* "View.MemoryView":418
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
  /* function exit code */

#line 418 "View.MemoryView"
  __pyx_r = 0;

#line 418 "View.MemoryView"
  goto __pyx_L0;

#line 418 "View.MemoryView"
  __pyx_L1_error:;

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 418 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 418 "View.MemoryView"
  __pyx_r = -1;

#line 418 "View.MemoryView"
  __pyx_L0:;

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_have_slices);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_obj);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_index);

#line 418 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 418 "View.MemoryView"
  return __pyx_r;

#line 418 "View.MemoryView"
}

/* "View.MemoryView":433
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

#line 433 "View.MemoryView"


#line 433 "View.MemoryView"
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {

#line 433 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 433 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 433 "View.MemoryView"
  int __pyx_t_1;

#line 433 "View.MemoryView"
  int __pyx_t_2;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_8 = NULL;

#line 433 "View.MemoryView"
  int __pyx_t_9;

#line 433 "View.MemoryView"
  int __pyx_lineno = 0;

#line 433 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 433 "View.MemoryView"
  int __pyx_clineno = 0;

#line 433 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_slice", 0);

#line 433 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":434
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */

#line 434 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 

#line 434 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);

#line 434 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */

#line 435 "View.MemoryView"
    {

#line 435 "View.MemoryView"
      __Pyx_PyThreadState_declare

#line 435 "View.MemoryView"
      __Pyx_PyThreadState_assign

#line 435 "View.MemoryView"
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);

#line 435 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_3);

#line 435 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_4);

#line 435 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_5);

#line 435 "View.MemoryView"
      /*try:*/ {

        /* "View.MemoryView":436
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */

#line 436 "View.MemoryView"
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 436, __pyx_L4_error)

#line 436 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */

#line 437 "View.MemoryView"
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 437, __pyx_L4_error)

#line 437 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":436
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */

#line 436 "View.MemoryView"
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 436, __pyx_L4_error)

#line 436 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_8);

#line 436 "View.MemoryView"
        __Pyx_INCREF(__pyx_v_obj);

#line 436 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_v_obj);

#line 436 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);

#line 436 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_6);

#line 436 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);

#line 436 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_7);

#line 436 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);

#line 436 "View.MemoryView"
        __pyx_t_6 = 0;

#line 436 "View.MemoryView"
        __pyx_t_7 = 0;

#line 436 "View.MemoryView"
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 436, __pyx_L4_error)

#line 436 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 436 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 436 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);

#line 436 "View.MemoryView"
        __pyx_t_7 = 0;

        /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */

#line 435 "View.MemoryView"
      }

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 435 "View.MemoryView"
      goto __pyx_L9_try_end;

#line 435 "View.MemoryView"
      __pyx_L4_error:;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":438
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */

#line 438 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);

#line 438 "View.MemoryView"
      if (__pyx_t_9) {

#line 438 "View.MemoryView"
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 438 "View.MemoryView"
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(2, 438, __pyx_L6_except_error)

#line 438 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 438 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_8);

#line 438 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":439
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */

#line 439 "View.MemoryView"
        __Pyx_XDECREF(__pyx_r);

#line 439 "View.MemoryView"
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 439 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 439 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 439 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 439 "View.MemoryView"
        goto __pyx_L7_except_return;

#line 439 "View.MemoryView"
      }

#line 439 "View.MemoryView"
      goto __pyx_L6_except_error;

#line 439 "View.MemoryView"
      __pyx_L6_except_error:;

      /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_3);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_4);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_5);

#line 435 "View.MemoryView"
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);

#line 435 "View.MemoryView"
      goto __pyx_L1_error;

#line 435 "View.MemoryView"
      __pyx_L7_except_return:;

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_3);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_4);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_5);

#line 435 "View.MemoryView"
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);

#line 435 "View.MemoryView"
      goto __pyx_L0;

#line 435 "View.MemoryView"
      __pyx_L9_try_end:;

#line 435 "View.MemoryView"
    }

    /* "View.MemoryView":434
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */

#line 434 "View.MemoryView"
  }

  /* "View.MemoryView":441
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */

#line 441 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 441 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_obj);

#line 441 "View.MemoryView"
  __pyx_r = __pyx_v_obj;

#line 441 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":433
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

#line 433 "View.MemoryView"


#line 433 "View.MemoryView"
  /* function exit code */

#line 433 "View.MemoryView"
  __pyx_L1_error:;

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_7);

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_8);

#line 433 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 433 "View.MemoryView"
  __pyx_r = 0;

#line 433 "View.MemoryView"
  __pyx_L0:;

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_obj);

#line 433 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 433 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 433 "View.MemoryView"
  return __pyx_r;

#line 433 "View.MemoryView"
}

/* "View.MemoryView":443
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

#line 443 "View.MemoryView"


#line 443 "View.MemoryView"
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice
#line 443 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_src_slice
#line 443 "View.MemoryView"
;

#line 443 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 443 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 443 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 443 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_2;

#line 443 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 443 "View.MemoryView"
  int __pyx_t_4;

#line 443 "View.MemoryView"
  int __pyx_t_5;

#line 443 "View.MemoryView"
  int __pyx_t_6;

#line 443 "View.MemoryView"
  int __pyx_lineno = 0;

#line 443 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 443 "View.MemoryView"
  int __pyx_clineno = 0;

#line 443 "View.MemoryView"
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":447
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */

#line 447 "View.MemoryView"
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(2, 447, __pyx_L1_error)

#line 447 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 447, __pyx_L1_error)

  /* "View.MemoryView":448
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */

#line 448 "View.MemoryView"
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(2, 448, __pyx_L1_error)

#line 448 "View.MemoryView"
  __pyx_t_2 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_2 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 448, __pyx_L1_error)

  /* "View.MemoryView":449
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */

#line 449 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 449 "View.MemoryView"
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 449 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 449 "View.MemoryView"
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":447
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */

#line 447 "View.MemoryView"
  __pyx_t_6 = __pyx_memoryview_copy_contents((__pyx_t_1[0]), (__pyx_t_2[0]), __pyx_t_4, __pyx_t_5, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 447, __pyx_L1_error)

  /* "View.MemoryView":443
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

#line 443 "View.MemoryView"


#line 443 "View.MemoryView"
  /* function exit code */

#line 443 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 443 "View.MemoryView"
  goto __pyx_L0;

#line 443 "View.MemoryView"
  __pyx_L1_error:;

#line 443 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 443 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 443 "View.MemoryView"
  __pyx_r = 0;

#line 443 "View.MemoryView"
  __pyx_L0:;

#line 443 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 443 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 443 "View.MemoryView"
  return __pyx_r;

#line 443 "View.MemoryView"
}

/* "View.MemoryView":451
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

#line 451 "View.MemoryView"


#line 451 "View.MemoryView"
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80]
#line 451 "View.MemoryView"
;
  void *__pyx_v_tmp
#line 451 "View.MemoryView"
;
  void *__pyx_v_item
#line 451 "View.MemoryView"
;
  __Pyx_memviewslice *__pyx_v_dst_slice
#line 451 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp_slice
#line 451 "View.MemoryView"
;

#line 451 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 451 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 451 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 451 "View.MemoryView"
  int __pyx_t_2;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 451 "View.MemoryView"
  int __pyx_t_4;

#line 451 "View.MemoryView"
  int __pyx_t_5;

#line 451 "View.MemoryView"
  char const *__pyx_t_6;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_8 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_9 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_10 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_11 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_12 = NULL;

#line 451 "View.MemoryView"
  int __pyx_lineno = 0;

#line 451 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 451 "View.MemoryView"
  int __pyx_clineno = 0;

#line 451 "View.MemoryView"
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":453
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */

#line 453 "View.MemoryView"
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":458
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */

#line 458 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 458, __pyx_L1_error)

#line 458 "View.MemoryView"
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":460
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */

#line 460 "View.MemoryView"
  __pyx_t_2 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);

#line 460 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":461
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */

#line 461 "View.MemoryView"
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":462
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */

#line 462 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);

#line 462 "View.MemoryView"
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":463
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */

#line 463 "View.MemoryView"
      PyErr_NoMemory(); __PYX_ERR(2, 463, __pyx_L1_error)

      /* "View.MemoryView":462
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */

#line 462 "View.MemoryView"
    }

    /* "View.MemoryView":464
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */

#line 464 "View.MemoryView"
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":460
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */

#line 460 "View.MemoryView"
    goto __pyx_L3;

#line 460 "View.MemoryView"
  }

  /* "View.MemoryView":466
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */

#line 466 "View.MemoryView"
  /*else*/ {

#line 466 "View.MemoryView"
    __pyx_v_item = ((void *)__pyx_v_array);

#line 466 "View.MemoryView"
  }

#line 466 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":468
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */

#line 468 "View.MemoryView"
  /*try:*/ {

    /* "View.MemoryView":469
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */

#line 469 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_self->dtype_is_object != 0);

#line 469 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":470
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */

#line 470 "View.MemoryView"
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":469
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */

#line 469 "View.MemoryView"
      goto __pyx_L8;

#line 469 "View.MemoryView"
    }

    /* "View.MemoryView":472
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 472 "View.MemoryView"
    /*else*/ {

#line 472 "View.MemoryView"
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 472, __pyx_L6_error)

#line 472 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 472 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 472 "View.MemoryView"
    }

#line 472 "View.MemoryView"
    __pyx_L8:;

    /* "View.MemoryView":476
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */

#line 476 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_self->view.suboffsets != NULL) != 0);

#line 476 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":477
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */

#line 477 "View.MemoryView"
      __pyx_t_3 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 477, __pyx_L6_error)

#line 477 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 477 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":476
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */

#line 476 "View.MemoryView"
    }

    /* "View.MemoryView":478
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */

#line 478 "View.MemoryView"
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);

#line 478 "View.MemoryView"
  }

  /* "View.MemoryView":481
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */

#line 481 "View.MemoryView"
  /*finally:*/ {

#line 481 "View.MemoryView"
    /*normal exit:*/{

#line 481 "View.MemoryView"
      PyMem_Free(__pyx_v_tmp);

#line 481 "View.MemoryView"
      goto __pyx_L7;

#line 481 "View.MemoryView"
    }

#line 481 "View.MemoryView"
    __pyx_L6_error:;

#line 481 "View.MemoryView"
    /*exception exit:*/{

#line 481 "View.MemoryView"
      __Pyx_PyThreadState_declare

#line 481 "View.MemoryView"
      __Pyx_PyThreadState_assign

#line 481 "View.MemoryView"
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;

#line 481 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 481 "View.MemoryView"
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);

#line 481 "View.MemoryView"
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_7);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_8);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_9);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_10);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_11);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_12);

#line 481 "View.MemoryView"
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;

#line 481 "View.MemoryView"
      {

#line 481 "View.MemoryView"
        PyMem_Free(__pyx_v_tmp);

#line 481 "View.MemoryView"
      }

#line 481 "View.MemoryView"
      if (PY_MAJOR_VERSION >= 3) {

#line 481 "View.MemoryView"
        __Pyx_XGIVEREF(__pyx_t_10);

#line 481 "View.MemoryView"
        __Pyx_XGIVEREF(__pyx_t_11);

#line 481 "View.MemoryView"
        __Pyx_XGIVEREF(__pyx_t_12);

#line 481 "View.MemoryView"
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);

#line 481 "View.MemoryView"
      }

#line 481 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_7);

#line 481 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_8);

#line 481 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_9);

#line 481 "View.MemoryView"
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);

#line 481 "View.MemoryView"
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;

#line 481 "View.MemoryView"
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;

#line 481 "View.MemoryView"
      goto __pyx_L1_error;

#line 481 "View.MemoryView"
    }

#line 481 "View.MemoryView"
    __pyx_L7:;

#line 481 "View.MemoryView"
  }

  /* "View.MemoryView":451
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

#line 451 "View.MemoryView"


#line 451 "View.MemoryView"
  /* function exit code */

#line 451 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 451 "View.MemoryView"
  goto __pyx_L0;

#line 451 "View.MemoryView"
  __pyx_L1_error:;

#line 451 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 451 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 451 "View.MemoryView"
  __pyx_r = 0;

#line 451 "View.MemoryView"
  __pyx_L0:;

#line 451 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 451 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 451 "View.MemoryView"
  return __pyx_r;

#line 451 "View.MemoryView"
}

/* "View.MemoryView":483
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

#line 483 "View.MemoryView"


#line 483 "View.MemoryView"
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp
#line 483 "View.MemoryView"
;

#line 483 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 483 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 483 "View.MemoryView"
  char *__pyx_t_1;

#line 483 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 483 "View.MemoryView"
  int __pyx_lineno = 0;

#line 483 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 483 "View.MemoryView"
  int __pyx_clineno = 0;

#line 483 "View.MemoryView"
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":484
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */

#line 484 "View.MemoryView"
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(2, 484, __pyx_L1_error)

#line 484 "View.MemoryView"
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":485
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */

#line 485 "View.MemoryView"
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 485, __pyx_L1_error)

#line 485 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 485 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":483
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

#line 483 "View.MemoryView"


#line 483 "View.MemoryView"
  /* function exit code */

#line 483 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 483 "View.MemoryView"
  goto __pyx_L0;

#line 483 "View.MemoryView"
  __pyx_L1_error:;

#line 483 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 483 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 483 "View.MemoryView"
  __pyx_r = 0;

#line 483 "View.MemoryView"
  __pyx_L0:;

#line 483 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 483 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 483 "View.MemoryView"
  return __pyx_r;

#line 483 "View.MemoryView"
}

/* "View.MemoryView":487
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 487 "View.MemoryView"


#line 487 "View.MemoryView"
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL
#line 487 "View.MemoryView"
;
  PyObject *__pyx_v_bytesitem = 0
#line 487 "View.MemoryView"
;
  PyObject *__pyx_v_result = NULL
#line 487 "View.MemoryView"
;

#line 487 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 487 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 487 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 487 "View.MemoryView"
  int __pyx_t_8;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_9 = NULL;

#line 487 "View.MemoryView"
  size_t __pyx_t_10;

#line 487 "View.MemoryView"
  int __pyx_t_11;

#line 487 "View.MemoryView"
  int __pyx_lineno = 0;

#line 487 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 487 "View.MemoryView"
  int __pyx_clineno = 0;

#line 487 "View.MemoryView"
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":490
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */

#line 490 "View.MemoryView"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 490, __pyx_L1_error)

#line 490 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 490 "View.MemoryView"
  __pyx_v_struct = __pyx_t_1;

#line 490 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":493
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */

#line 493 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 493, __pyx_L1_error)

#line 493 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 493 "View.MemoryView"
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);

#line 493 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */

#line 494 "View.MemoryView"
  {

#line 494 "View.MemoryView"
    __Pyx_PyThreadState_declare

#line 494 "View.MemoryView"
    __Pyx_PyThreadState_assign

#line 494 "View.MemoryView"
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);

#line 494 "View.MemoryView"
    __Pyx_XGOTREF(__pyx_t_2);

#line 494 "View.MemoryView"
    __Pyx_XGOTREF(__pyx_t_3);

#line 494 "View.MemoryView"
    __Pyx_XGOTREF(__pyx_t_4);

#line 494 "View.MemoryView"
    /*try:*/ {

      /* "View.MemoryView":495
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */

#line 495 "View.MemoryView"
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 495 "View.MemoryView"
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 495 "View.MemoryView"
      __pyx_t_7 = NULL;

#line 495 "View.MemoryView"
      __pyx_t_8 = 0;

#line 495 "View.MemoryView"
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {

#line 495 "View.MemoryView"
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);

#line 495 "View.MemoryView"
        if (likely(__pyx_t_7)) {

#line 495 "View.MemoryView"
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 495 "View.MemoryView"
          __Pyx_INCREF(__pyx_t_7);

#line 495 "View.MemoryView"
          __Pyx_INCREF(function);

#line 495 "View.MemoryView"
          __Pyx_DECREF_SET(__pyx_t_5, function);

#line 495 "View.MemoryView"
          __pyx_t_8 = 1;

#line 495 "View.MemoryView"
        }

#line 495 "View.MemoryView"
      }

#line 495 "View.MemoryView"
      #if CYTHON_FAST_PYCALL

#line 495 "View.MemoryView"
      if (PyFunction_Check(__pyx_t_5)) {

#line 495 "View.MemoryView"
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};

#line 495 "View.MemoryView"
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 495 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 495 "View.MemoryView"
      } else

#line 495 "View.MemoryView"
      #endif

#line 495 "View.MemoryView"
      #if CYTHON_FAST_PYCCALL

#line 495 "View.MemoryView"
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 495 "View.MemoryView"
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};

#line 495 "View.MemoryView"
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 495 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 495 "View.MemoryView"
      } else

#line 495 "View.MemoryView"
      #endif

#line 495 "View.MemoryView"
      {

#line 495 "View.MemoryView"
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_9);

#line 495 "View.MemoryView"
        if (__pyx_t_7) {

#line 495 "View.MemoryView"
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;

#line 495 "View.MemoryView"
        }

#line 495 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_6);

#line 495 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);

#line 495 "View.MemoryView"
        __Pyx_INCREF(__pyx_v_bytesitem);

#line 495 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_v_bytesitem);

#line 495 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);

#line 495 "View.MemoryView"
        __pyx_t_6 = 0;

#line 495 "View.MemoryView"
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 495 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 495 "View.MemoryView"
      }

#line 495 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 495 "View.MemoryView"
      __pyx_v_result = __pyx_t_1;

#line 495 "View.MemoryView"
      __pyx_t_1 = 0;

      /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */

#line 494 "View.MemoryView"
    }

    /* "View.MemoryView":499
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */

#line 499 "View.MemoryView"
    /*else:*/ {

#line 499 "View.MemoryView"
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 

#line 499 "View.MemoryView"
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);

#line 499 "View.MemoryView"
      if (__pyx_t_11) {

        /* "View.MemoryView":500
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */

#line 500 "View.MemoryView"
        __Pyx_XDECREF(__pyx_r);

#line 500 "View.MemoryView"
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 500, __pyx_L5_except_error)

#line 500 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 500 "View.MemoryView"
        __pyx_r = __pyx_t_1;

#line 500 "View.MemoryView"
        __pyx_t_1 = 0;

#line 500 "View.MemoryView"
        goto __pyx_L6_except_return;

        /* "View.MemoryView":499
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */

#line 499 "View.MemoryView"
      }

      /* "View.MemoryView":501
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */

#line 501 "View.MemoryView"
      __Pyx_XDECREF(__pyx_r);

#line 501 "View.MemoryView"
      __Pyx_INCREF(__pyx_v_result);

#line 501 "View.MemoryView"
      __pyx_r = __pyx_v_result;

#line 501 "View.MemoryView"
      goto __pyx_L6_except_return;

#line 501 "View.MemoryView"
    }

#line 501 "View.MemoryView"
    __pyx_L3_error:;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":496
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */

#line 496 "View.MemoryView"
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);

#line 496 "View.MemoryView"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 496, __pyx_L5_except_error)

#line 496 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 496 "View.MemoryView"
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);

#line 496 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 496 "View.MemoryView"
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);

#line 496 "View.MemoryView"
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;

#line 496 "View.MemoryView"
    if (__pyx_t_8) {

#line 496 "View.MemoryView"
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 496 "View.MemoryView"
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(2, 496, __pyx_L5_except_error)

#line 496 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 496 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 496 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":497
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */

#line 497 "View.MemoryView"
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 497, __pyx_L5_except_error)

#line 497 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 497 "View.MemoryView"
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 497 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 497 "View.MemoryView"
      __PYX_ERR(2, 497, __pyx_L5_except_error)

#line 497 "View.MemoryView"
    }

#line 497 "View.MemoryView"
    goto __pyx_L5_except_error;

#line 497 "View.MemoryView"
    __pyx_L5_except_error:;

    /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_4);

#line 494 "View.MemoryView"
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);

#line 494 "View.MemoryView"
    goto __pyx_L1_error;

#line 494 "View.MemoryView"
    __pyx_L6_except_return:;

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_4);

#line 494 "View.MemoryView"
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);

#line 494 "View.MemoryView"
    goto __pyx_L0;

#line 494 "View.MemoryView"
  }

  /* "View.MemoryView":487
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 487 "View.MemoryView"


#line 487 "View.MemoryView"
  /* function exit code */

#line 487 "View.MemoryView"
  __pyx_L1_error:;

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_7);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_9);

#line 487 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 487 "View.MemoryView"
  __pyx_r = 0;

#line 487 "View.MemoryView"
  __pyx_L0:;

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_struct);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_bytesitem);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_result);

#line 487 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 487 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 487 "View.MemoryView"
  return __pyx_r;

#line 487 "View.MemoryView"
}

/* "View.MemoryView":503
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 503 "View.MemoryView"


#line 503 "View.MemoryView"
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL
#line 503 "View.MemoryView"
;
  char __pyx_v_c
#line 503 "View.MemoryView"
;
  PyObject *__pyx_v_bytesvalue = 0
#line 503 "View.MemoryView"
;
  Py_ssize_t __pyx_v_i
#line 503 "View.MemoryView"
;

#line 503 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 503 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 503 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 503 "View.MemoryView"
  int __pyx_t_2;

#line 503 "View.MemoryView"
  int __pyx_t_3;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 503 "View.MemoryView"
  int __pyx_t_7;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_8 = NULL;

#line 503 "View.MemoryView"
  Py_ssize_t __pyx_t_9;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_10 = NULL;

#line 503 "View.MemoryView"
  char *__pyx_t_11;

#line 503 "View.MemoryView"
  char *__pyx_t_12;

#line 503 "View.MemoryView"
  char *__pyx_t_13;

#line 503 "View.MemoryView"
  char *__pyx_t_14;

#line 503 "View.MemoryView"
  int __pyx_lineno = 0;

#line 503 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 503 "View.MemoryView"
  int __pyx_clineno = 0;

#line 503 "View.MemoryView"
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":506
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */

#line 506 "View.MemoryView"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 506, __pyx_L1_error)

#line 506 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 506 "View.MemoryView"
  __pyx_v_struct = __pyx_t_1;

#line 506 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":511
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */

#line 511 "View.MemoryView"
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 

#line 511 "View.MemoryView"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 511 "View.MemoryView"
  if (__pyx_t_3) {

    /* "View.MemoryView":512
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */

#line 512 "View.MemoryView"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_1);

#line 512 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 512 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_4);

#line 512 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_4 = 0;

#line 512 "View.MemoryView"
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 512 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 512 "View.MemoryView"
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_4 = 0;

    /* "View.MemoryView":511
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */

#line 511 "View.MemoryView"
    goto __pyx_L3;

#line 511 "View.MemoryView"
  }

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */

#line 514 "View.MemoryView"
  /*else*/ {

#line 514 "View.MemoryView"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 514 "View.MemoryView"
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_1);

#line 514 "View.MemoryView"
    __pyx_t_5 = NULL;

#line 514 "View.MemoryView"
    __pyx_t_7 = 0;

#line 514 "View.MemoryView"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {

#line 514 "View.MemoryView"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);

#line 514 "View.MemoryView"
      if (likely(__pyx_t_5)) {

#line 514 "View.MemoryView"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);

#line 514 "View.MemoryView"
        __Pyx_INCREF(__pyx_t_5);

#line 514 "View.MemoryView"
        __Pyx_INCREF(function);

#line 514 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_t_6, function);

#line 514 "View.MemoryView"
        __pyx_t_7 = 1;

#line 514 "View.MemoryView"
      }

#line 514 "View.MemoryView"
    }

#line 514 "View.MemoryView"
    #if CYTHON_FAST_PYCALL

#line 514 "View.MemoryView"
    if (PyFunction_Check(__pyx_t_6)) {

#line 514 "View.MemoryView"
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};

#line 514 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 514 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 514 "View.MemoryView"
    } else

#line 514 "View.MemoryView"
    #endif

#line 514 "View.MemoryView"
    #if CYTHON_FAST_PYCCALL

#line 514 "View.MemoryView"
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {

#line 514 "View.MemoryView"
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};

#line 514 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 514 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 514 "View.MemoryView"
    } else

#line 514 "View.MemoryView"
    #endif

#line 514 "View.MemoryView"
    {

#line 514 "View.MemoryView"
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_8);

#line 514 "View.MemoryView"
      if (__pyx_t_5) {

#line 514 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;

#line 514 "View.MemoryView"
      }

#line 514 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_t_1);

#line 514 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);

#line 514 "View.MemoryView"
      __Pyx_INCREF(__pyx_v_value);

#line 514 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_v_value);

#line 514 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);

#line 514 "View.MemoryView"
      __pyx_t_1 = 0;

#line 514 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 514 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 514 "View.MemoryView"
    }

#line 514 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 514 "View.MemoryView"
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);

#line 514 "View.MemoryView"
    __pyx_t_4 = 0;

#line 514 "View.MemoryView"
  }

#line 514 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */

#line 516 "View.MemoryView"
  __pyx_t_9 = 0;

#line 516 "View.MemoryView"
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {

#line 516 "View.MemoryView"
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");

#line 516 "View.MemoryView"
    __PYX_ERR(2, 516, __pyx_L1_error)

#line 516 "View.MemoryView"
  }

#line 516 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_bytesvalue);

#line 516 "View.MemoryView"
  __pyx_t_10 = __pyx_v_bytesvalue;

#line 516 "View.MemoryView"
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);

#line 516 "View.MemoryView"
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));

#line 516 "View.MemoryView"
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {

#line 516 "View.MemoryView"
    __pyx_t_11 = __pyx_t_14;

#line 516 "View.MemoryView"
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":517
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */

#line 517 "View.MemoryView"
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */

#line 516 "View.MemoryView"
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":517
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */

#line 517 "View.MemoryView"
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;

#line 517 "View.MemoryView"
  }

#line 517 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":503
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 503 "View.MemoryView"


#line 503 "View.MemoryView"
  /* function exit code */

#line 503 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 503 "View.MemoryView"
  goto __pyx_L0;

#line 503 "View.MemoryView"
  __pyx_L1_error:;

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_8);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_10);

#line 503 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 503 "View.MemoryView"
  __pyx_r = 0;

#line 503 "View.MemoryView"
  __pyx_L0:;

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_struct);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_bytesvalue);

#line 503 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 503 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 503 "View.MemoryView"
  return __pyx_r;

#line 503 "View.MemoryView"
}

/* "View.MemoryView":520
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
/* Python wrapper */

#line 520 "View.MemoryView"
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/

#line 520 "View.MemoryView"
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {

#line 520 "View.MemoryView"
  int __pyx_r;

#line 520 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 520 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = 
#line 520 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
  /* function exit code */

#line 520 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 520 "View.MemoryView"
  return __pyx_r;

#line 520 "View.MemoryView"
}

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {

#line 520 "View.MemoryView"
  int __pyx_r;

#line 520 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 520 "View.MemoryView"
  int __pyx_t_1;

#line 520 "View.MemoryView"
  int __pyx_t_2;

#line 520 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 520 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 520 "View.MemoryView"
  char *__pyx_t_5;

#line 520 "View.MemoryView"
  void *__pyx_t_6;

#line 520 "View.MemoryView"
  int __pyx_t_7;

#line 520 "View.MemoryView"
  Py_ssize_t __pyx_t_8;

#line 520 "View.MemoryView"
  int __pyx_lineno = 0;

#line 520 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 520 "View.MemoryView"
  int __pyx_clineno = 0;

#line 520 "View.MemoryView"
  if (__pyx_v_info == NULL) {

#line 520 "View.MemoryView"
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");

#line 520 "View.MemoryView"
    return -1;

#line 520 "View.MemoryView"
  }

#line 520 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__", 0);

#line 520 "View.MemoryView"
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);

#line 520 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":521
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */

#line 521 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);

#line 521 "View.MemoryView"
  if (__pyx_t_2) {

#line 521 "View.MemoryView"
  } else {

#line 521 "View.MemoryView"
    __pyx_t_1 = __pyx_t_2;

#line 521 "View.MemoryView"
    goto __pyx_L4_bool_binop_done;

#line 521 "View.MemoryView"
  }

#line 521 "View.MemoryView"
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);

#line 521 "View.MemoryView"
  __pyx_t_1 = __pyx_t_2;

#line 521 "View.MemoryView"
  __pyx_L4_bool_binop_done:;

#line 521 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":522
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */

#line 522 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 522, __pyx_L1_error)

#line 522 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 522 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 522 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 522 "View.MemoryView"
    __PYX_ERR(2, 522, __pyx_L1_error)

    /* "View.MemoryView":521
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */

#line 521 "View.MemoryView"
  }

  /* "View.MemoryView":524
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */

#line 524 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);

#line 524 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":525
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */

#line 525 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->view.shape;

#line 525 "View.MemoryView"
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":524
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */

#line 524 "View.MemoryView"
    goto __pyx_L6;

#line 524 "View.MemoryView"
  }

  /* "View.MemoryView":527
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */

#line 527 "View.MemoryView"
  /*else*/ {

#line 527 "View.MemoryView"
    __pyx_v_info->shape = NULL;

#line 527 "View.MemoryView"
  }

#line 527 "View.MemoryView"
  __pyx_L6:;

  /* "View.MemoryView":529
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */

#line 529 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);

#line 529 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":530
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */

#line 530 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->view.strides;

#line 530 "View.MemoryView"
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":529
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */

#line 529 "View.MemoryView"
    goto __pyx_L7;

#line 529 "View.MemoryView"
  }

  /* "View.MemoryView":532
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */

#line 532 "View.MemoryView"
  /*else*/ {

#line 532 "View.MemoryView"
    __pyx_v_info->strides = NULL;

#line 532 "View.MemoryView"
  }

#line 532 "View.MemoryView"
  __pyx_L7:;

  /* "View.MemoryView":534
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */

#line 534 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);

#line 534 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":535
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */

#line 535 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->view.suboffsets;

#line 535 "View.MemoryView"
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":534
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */

#line 534 "View.MemoryView"
    goto __pyx_L8;

#line 534 "View.MemoryView"
  }

  /* "View.MemoryView":537
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */

#line 537 "View.MemoryView"
  /*else*/ {

#line 537 "View.MemoryView"
    __pyx_v_info->suboffsets = NULL;

#line 537 "View.MemoryView"
  }

#line 537 "View.MemoryView"
  __pyx_L8:;

  /* "View.MemoryView":539
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */

#line 539 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);

#line 539 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":540
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */

#line 540 "View.MemoryView"
    __pyx_t_5 = __pyx_v_self->view.format;

#line 540 "View.MemoryView"
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":539
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */

#line 539 "View.MemoryView"
    goto __pyx_L9;

#line 539 "View.MemoryView"
  }

  /* "View.MemoryView":542
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */

#line 542 "View.MemoryView"
  /*else*/ {

#line 542 "View.MemoryView"
    __pyx_v_info->format = NULL;

#line 542 "View.MemoryView"
  }

#line 542 "View.MemoryView"
  __pyx_L9:;

  /* "View.MemoryView":544
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */

#line 544 "View.MemoryView"
  __pyx_t_6 = __pyx_v_self->view.buf;

#line 544 "View.MemoryView"
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":545
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */

#line 545 "View.MemoryView"
  __pyx_t_7 = __pyx_v_self->view.ndim;

#line 545 "View.MemoryView"
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":546
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */

#line 546 "View.MemoryView"
  __pyx_t_8 = __pyx_v_self->view.itemsize;

#line 546 "View.MemoryView"
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":547
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */

#line 547 "View.MemoryView"
  __pyx_t_8 = __pyx_v_self->view.len;

#line 547 "View.MemoryView"
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":548
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */

#line 548 "View.MemoryView"
  __pyx_t_1 = __pyx_v_self->view.readonly;

#line 548 "View.MemoryView"
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":549
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */

#line 549 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 549 "View.MemoryView"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 549 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_info->obj);

#line 549 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_info->obj);

#line 549 "View.MemoryView"
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":520
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
  /* function exit code */

#line 520 "View.MemoryView"
  __pyx_r = 0;

#line 520 "View.MemoryView"
  goto __pyx_L0;

#line 520 "View.MemoryView"
  __pyx_L1_error:;

#line 520 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 520 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 520 "View.MemoryView"
  __pyx_r = -1;

#line 520 "View.MemoryView"
  if (__pyx_v_info->obj != NULL) {

#line 520 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 520 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 520 "View.MemoryView"
  }

#line 520 "View.MemoryView"
  goto __pyx_L2;

#line 520 "View.MemoryView"
  __pyx_L0:;

#line 520 "View.MemoryView"
  if (__pyx_v_info->obj == Py_None) {

#line 520 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 520 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 520 "View.MemoryView"
  }

#line 520 "View.MemoryView"
  __pyx_L2:;

#line 520 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 520 "View.MemoryView"
  return __pyx_r;

#line 520 "View.MemoryView"
}

/* "View.MemoryView":555
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
/* Python wrapper */

#line 555 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 555 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {

#line 555 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 555 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 555 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 555 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
  /* function exit code */

#line 555 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 555 "View.MemoryView"
  return __pyx_r;

#line 555 "View.MemoryView"
}

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0
#line 555 "View.MemoryView"
;

#line 555 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 555 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 555 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 555 "View.MemoryView"
  int __pyx_t_2;

#line 555 "View.MemoryView"
  int __pyx_lineno = 0;

#line 555 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 555 "View.MemoryView"
  int __pyx_clineno = 0;

#line 555 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":556
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */

#line 556 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 556, __pyx_L1_error)

#line 556 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 556 "View.MemoryView"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(2, 556, __pyx_L1_error)

#line 556 "View.MemoryView"
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);

#line 556 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":557
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */

#line 557 "View.MemoryView"
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 557, __pyx_L1_error)

  /* "View.MemoryView":558
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 558 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 558 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 558 "View.MemoryView"
  __pyx_r = ((PyObject *)__pyx_v_result);

#line 558 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":555
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
  /* function exit code */

#line 555 "View.MemoryView"
  __pyx_L1_error:;

#line 555 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 555 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 555 "View.MemoryView"
  __pyx_r = NULL;

#line 555 "View.MemoryView"
  __pyx_L0:;

#line 555 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 555 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 555 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 555 "View.MemoryView"
  return __pyx_r;

#line 555 "View.MemoryView"
}

/* "View.MemoryView":561
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
/* Python wrapper */

#line 561 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 561 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {

#line 561 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 561 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 561 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 561 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
  /* function exit code */

#line 561 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 561 "View.MemoryView"
  return __pyx_r;

#line 561 "View.MemoryView"
}

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 561 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 561 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 561 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":562
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 562 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 562 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->obj);

#line 562 "View.MemoryView"
  __pyx_r = __pyx_v_self->obj;

#line 562 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":561
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
  /* function exit code */

#line 561 "View.MemoryView"
  __pyx_L0:;

#line 561 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 561 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 561 "View.MemoryView"
  return __pyx_r;

#line 561 "View.MemoryView"
}

/* "View.MemoryView":565
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
/* Python wrapper */

#line 565 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 565 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {

#line 565 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 565 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 565 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 565 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
  /* function exit code */

#line 565 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 565 "View.MemoryView"
  return __pyx_r;

#line 565 "View.MemoryView"
}

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length
#line 565 "View.MemoryView"
;

#line 565 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 565 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 565 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 565 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 565 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 565 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 565 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 565 "View.MemoryView"
  int __pyx_lineno = 0;

#line 565 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 565 "View.MemoryView"
  int __pyx_clineno = 0;

#line 565 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":566
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 566 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 566 "View.MemoryView"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 566 "View.MemoryView"
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);

#line 566 "View.MemoryView"
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {

#line 566 "View.MemoryView"
    __pyx_t_2 = __pyx_t_4;

#line 566 "View.MemoryView"
    __pyx_v_length = (__pyx_t_2[0]);

#line 566 "View.MemoryView"
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 566 "View.MemoryView"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(2, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 566 "View.MemoryView"
  }

#line 566 "View.MemoryView"
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_5);

#line 566 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 566 "View.MemoryView"
  __pyx_r = __pyx_t_5;

#line 566 "View.MemoryView"
  __pyx_t_5 = 0;

#line 566 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":565
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
  /* function exit code */

#line 565 "View.MemoryView"
  __pyx_L1_error:;

#line 565 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 565 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 565 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 565 "View.MemoryView"
  __pyx_r = NULL;

#line 565 "View.MemoryView"
  __pyx_L0:;

#line 565 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 565 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 565 "View.MemoryView"
  return __pyx_r;

#line 565 "View.MemoryView"
}

/* "View.MemoryView":569
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
/* Python wrapper */

#line 569 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 569 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {

#line 569 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 569 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 569 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 569 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
  /* function exit code */

#line 569 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 569 "View.MemoryView"
  return __pyx_r;

#line 569 "View.MemoryView"
}

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride
#line 569 "View.MemoryView"
;

#line 569 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 569 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 569 "View.MemoryView"
  int __pyx_t_1;

#line 569 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 569 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 569 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 569 "View.MemoryView"
  Py_ssize_t *__pyx_t_5;

#line 569 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 569 "View.MemoryView"
  int __pyx_lineno = 0;

#line 569 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 569 "View.MemoryView"
  int __pyx_clineno = 0;

#line 569 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":570
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */

#line 570 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);

#line 570 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":572
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */

#line 572 "View.MemoryView"
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 572, __pyx_L1_error)

#line 572 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 572 "View.MemoryView"
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 572 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 572 "View.MemoryView"
    __PYX_ERR(2, 572, __pyx_L1_error)

    /* "View.MemoryView":570
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */

#line 570 "View.MemoryView"
  }

  /* "View.MemoryView":574
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 574 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 574 "View.MemoryView"
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 574 "View.MemoryView"
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);

#line 574 "View.MemoryView"
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {

#line 574 "View.MemoryView"
    __pyx_t_3 = __pyx_t_5;

#line 574 "View.MemoryView"
    __pyx_v_stride = (__pyx_t_3[0]);

#line 574 "View.MemoryView"
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 574 "View.MemoryView"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(2, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 574 "View.MemoryView"
  }

#line 574 "View.MemoryView"
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_6);

#line 574 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 574 "View.MemoryView"
  __pyx_r = __pyx_t_6;

#line 574 "View.MemoryView"
  __pyx_t_6 = 0;

#line 574 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":569
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
  /* function exit code */

#line 569 "View.MemoryView"
  __pyx_L1_error:;

#line 569 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 569 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 569 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 569 "View.MemoryView"
  __pyx_r = NULL;

#line 569 "View.MemoryView"
  __pyx_L0:;

#line 569 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 569 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 569 "View.MemoryView"
  return __pyx_r;

#line 569 "View.MemoryView"
}

/* "View.MemoryView":577
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
/* Python wrapper */

#line 577 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 577 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {

#line 577 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 577 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 577 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 577 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
  /* function exit code */

#line 577 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 577 "View.MemoryView"
  return __pyx_r;

#line 577 "View.MemoryView"
}

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset
#line 577 "View.MemoryView"
;

#line 577 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 577 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 577 "View.MemoryView"
  int __pyx_t_1;

#line 577 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 577 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 577 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 577 "View.MemoryView"
  Py_ssize_t *__pyx_t_5;

#line 577 "View.MemoryView"
  Py_ssize_t *__pyx_t_6;

#line 577 "View.MemoryView"
  int __pyx_lineno = 0;

#line 577 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 577 "View.MemoryView"
  int __pyx_clineno = 0;

#line 577 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":578
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */

#line 578 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);

#line 578 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":579
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */

#line 579 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 579 "View.MemoryView"
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 579, __pyx_L1_error)

#line 579 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 579 "View.MemoryView"
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__18, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 579, __pyx_L1_error)

#line 579 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 579 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 579 "View.MemoryView"
    __pyx_r = __pyx_t_3;

#line 579 "View.MemoryView"
    __pyx_t_3 = 0;

#line 579 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":578
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */

#line 578 "View.MemoryView"
  }

  /* "View.MemoryView":581
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 581 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 581 "View.MemoryView"
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 581 "View.MemoryView"
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);

#line 581 "View.MemoryView"
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {

#line 581 "View.MemoryView"
    __pyx_t_4 = __pyx_t_6;

#line 581 "View.MemoryView"
    __pyx_v_suboffset = (__pyx_t_4[0]);

#line 581 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 581 "View.MemoryView"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(2, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 581 "View.MemoryView"
  }

#line 581 "View.MemoryView"
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 581 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 581 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 581 "View.MemoryView"
  __pyx_t_2 = 0;

#line 581 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":577
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
  /* function exit code */

#line 577 "View.MemoryView"
  __pyx_L1_error:;

#line 577 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 577 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 577 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 577 "View.MemoryView"
  __pyx_r = NULL;

#line 577 "View.MemoryView"
  __pyx_L0:;

#line 577 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 577 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 577 "View.MemoryView"
  return __pyx_r;

#line 577 "View.MemoryView"
}

/* "View.MemoryView":584
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
/* Python wrapper */

#line 584 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 584 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {

#line 584 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 584 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 584 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 584 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
  /* function exit code */

#line 584 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 584 "View.MemoryView"
  return __pyx_r;

#line 584 "View.MemoryView"
}

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 584 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 584 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 584 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 584 "View.MemoryView"
  int __pyx_lineno = 0;

#line 584 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 584 "View.MemoryView"
  int __pyx_clineno = 0;

#line 584 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":585
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 585 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 585 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 585, __pyx_L1_error)

#line 585 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 585 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 585 "View.MemoryView"
  __pyx_t_1 = 0;

#line 585 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":584
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
  /* function exit code */

#line 584 "View.MemoryView"
  __pyx_L1_error:;

#line 584 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 584 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 584 "View.MemoryView"
  __pyx_r = NULL;

#line 584 "View.MemoryView"
  __pyx_L0:;

#line 584 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 584 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 584 "View.MemoryView"
  return __pyx_r;

#line 584 "View.MemoryView"
}

/* "View.MemoryView":588
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
/* Python wrapper */

#line 588 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 588 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {

#line 588 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 588 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 588 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 588 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
  /* function exit code */

#line 588 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 588 "View.MemoryView"
  return __pyx_r;

#line 588 "View.MemoryView"
}

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 588 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 588 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 588 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 588 "View.MemoryView"
  int __pyx_lineno = 0;

#line 588 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 588 "View.MemoryView"
  int __pyx_clineno = 0;

#line 588 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":589
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 589 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 589 "View.MemoryView"
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 589, __pyx_L1_error)

#line 589 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 589 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 589 "View.MemoryView"
  __pyx_t_1 = 0;

#line 589 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":588
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
  /* function exit code */

#line 588 "View.MemoryView"
  __pyx_L1_error:;

#line 588 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 588 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 588 "View.MemoryView"
  __pyx_r = NULL;

#line 588 "View.MemoryView"
  __pyx_L0:;

#line 588 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 588 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 588 "View.MemoryView"
  return __pyx_r;

#line 588 "View.MemoryView"
}

/* "View.MemoryView":592
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
/* Python wrapper */

#line 592 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 592 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {

#line 592 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 592 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 592 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 592 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
  /* function exit code */

#line 592 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 592 "View.MemoryView"
  return __pyx_r;

#line 592 "View.MemoryView"
}

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 592 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 592 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 592 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 592 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 592 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 592 "View.MemoryView"
  int __pyx_lineno = 0;

#line 592 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 592 "View.MemoryView"
  int __pyx_clineno = 0;

#line 592 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":593
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 593 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 593 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 593, __pyx_L1_error)

#line 593 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 593 "View.MemoryView"
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 593, __pyx_L1_error)

#line 593 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 593 "View.MemoryView"
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 593, __pyx_L1_error)

#line 593 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 593 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 593 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 593 "View.MemoryView"
  __pyx_r = __pyx_t_3;

#line 593 "View.MemoryView"
  __pyx_t_3 = 0;

#line 593 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":592
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
  /* function exit code */

#line 592 "View.MemoryView"
  __pyx_L1_error:;

#line 592 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 592 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 592 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 592 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 592 "View.MemoryView"
  __pyx_r = NULL;

#line 592 "View.MemoryView"
  __pyx_L0:;

#line 592 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 592 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 592 "View.MemoryView"
  return __pyx_r;

#line 592 "View.MemoryView"
}

/* "View.MemoryView":596
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
/* Python wrapper */

#line 596 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 596 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {

#line 596 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 596 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 596 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 596 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
  /* function exit code */

#line 596 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 596 "View.MemoryView"
  return __pyx_r;

#line 596 "View.MemoryView"
}

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL
#line 596 "View.MemoryView"
;
  PyObject *__pyx_v_length = NULL
#line 596 "View.MemoryView"
;

#line 596 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 596 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 596 "View.MemoryView"
  int __pyx_t_1;

#line 596 "View.MemoryView"
  int __pyx_t_2;

#line 596 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 596 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 596 "View.MemoryView"
  Py_ssize_t *__pyx_t_5;

#line 596 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 596 "View.MemoryView"
  int __pyx_lineno = 0;

#line 596 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 596 "View.MemoryView"
  int __pyx_clineno = 0;

#line 596 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":597
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */

#line 597 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);

#line 597 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 597 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":598
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */

#line 598 "View.MemoryView"
    __Pyx_INCREF(__pyx_int_1);

#line 598 "View.MemoryView"
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":600
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */

#line 600 "View.MemoryView"
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);

#line 600 "View.MemoryView"
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {

#line 600 "View.MemoryView"
      __pyx_t_3 = __pyx_t_5;

#line 600 "View.MemoryView"
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 600, __pyx_L1_error)

#line 600 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 600 "View.MemoryView"
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);

#line 600 "View.MemoryView"
      __pyx_t_6 = 0;

      /* "View.MemoryView":601
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */

#line 601 "View.MemoryView"
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 601, __pyx_L1_error)

#line 601 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 601 "View.MemoryView"
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);

#line 601 "View.MemoryView"
      __pyx_t_6 = 0;

#line 601 "View.MemoryView"
    }

    /* "View.MemoryView":603
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */

#line 603 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_result);

#line 603 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_result);

#line 603 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_self->_size);

#line 603 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_self->_size);

#line 603 "View.MemoryView"
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":597
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */

#line 597 "View.MemoryView"
  }

  /* "View.MemoryView":605
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 605 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 605 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->_size);

#line 605 "View.MemoryView"
  __pyx_r = __pyx_v_self->_size;

#line 605 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":596
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
  /* function exit code */

#line 596 "View.MemoryView"
  __pyx_L1_error:;

#line 596 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 596 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 596 "View.MemoryView"
  __pyx_r = NULL;

#line 596 "View.MemoryView"
  __pyx_L0:;

#line 596 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_result);

#line 596 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_length);

#line 596 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 596 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 596 "View.MemoryView"
  return __pyx_r;

#line 596 "View.MemoryView"
}

/* "View.MemoryView":607
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
/* Python wrapper */

#line 607 "View.MemoryView"
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/

#line 607 "View.MemoryView"
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {

#line 607 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 607 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 607 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 607 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
  /* function exit code */

#line 607 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 607 "View.MemoryView"
  return __pyx_r;

#line 607 "View.MemoryView"
}

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 607 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 607 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 607 "View.MemoryView"
  int __pyx_t_1;

#line 607 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":608
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */

#line 608 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);

#line 608 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":609
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */

#line 609 "View.MemoryView"
    __pyx_r = (__pyx_v_self->view.shape[0]);

#line 609 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":608
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */

#line 608 "View.MemoryView"
  }

  /* "View.MemoryView":611
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 611 "View.MemoryView"
  __pyx_r = 0;

#line 611 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":607
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
  /* function exit code */

#line 607 "View.MemoryView"
  __pyx_L0:;

#line 607 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 607 "View.MemoryView"
  return __pyx_r;

#line 607 "View.MemoryView"
}

/* "View.MemoryView":613
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
/* Python wrapper */

#line 613 "View.MemoryView"
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/

#line 613 "View.MemoryView"
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {

#line 613 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 613 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 613 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 613 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
  /* function exit code */

#line 613 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 613 "View.MemoryView"
  return __pyx_r;

#line 613 "View.MemoryView"
}

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 613 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 613 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 613 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 613 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 613 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 613 "View.MemoryView"
  int __pyx_lineno = 0;

#line 613 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 613 "View.MemoryView"
  int __pyx_clineno = 0;

#line 613 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":614
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */

#line 614 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 614 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 614 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":615
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */

#line 615 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 615, __pyx_L1_error)

#line 615 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":614
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */

#line 614 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 614 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 614 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 614 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 614 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);

#line 614 "View.MemoryView"
  __pyx_t_1 = 0;

#line 614 "View.MemoryView"
  __pyx_t_2 = 0;

#line 614 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 614 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 614 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 614 "View.MemoryView"
  __pyx_t_2 = 0;

#line 614 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":613
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
  /* function exit code */

#line 613 "View.MemoryView"
  __pyx_L1_error:;

#line 613 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 613 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 613 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 613 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 613 "View.MemoryView"
  __pyx_r = NULL;

#line 613 "View.MemoryView"
  __pyx_L0:;

#line 613 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 613 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 613 "View.MemoryView"
  return __pyx_r;

#line 613 "View.MemoryView"
}

/* "View.MemoryView":617
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
/* Python wrapper */

#line 617 "View.MemoryView"
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/

#line 617 "View.MemoryView"
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {

#line 617 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 617 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 617 "View.MemoryView"
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = 
#line 617 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
  /* function exit code */

#line 617 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 617 "View.MemoryView"
  return __pyx_r;

#line 617 "View.MemoryView"
}

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 617 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 617 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 617 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 617 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 617 "View.MemoryView"
  int __pyx_lineno = 0;

#line 617 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 617 "View.MemoryView"
  int __pyx_clineno = 0;

#line 617 "View.MemoryView"
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":618
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 618 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 618 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 618 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 618 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 618 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 618 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 618 "View.MemoryView"
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 618 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 618 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 618 "View.MemoryView"
  __pyx_t_1 = 0;

#line 618 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 618 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 618 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 618 "View.MemoryView"
  __pyx_t_1 = 0;

#line 618 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":617
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
  /* function exit code */

#line 617 "View.MemoryView"
  __pyx_L1_error:;

#line 617 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 617 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 617 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 617 "View.MemoryView"
  __pyx_r = NULL;

#line 617 "View.MemoryView"
  __pyx_L0:;

#line 617 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 617 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 617 "View.MemoryView"
  return __pyx_r;

#line 617 "View.MemoryView"
}

/* "View.MemoryView":621
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
/* Python wrapper */

#line 621 "View.MemoryView"
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 621 "View.MemoryView"
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 621 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 621 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 621 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = 
#line 621 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
  /* function exit code */

#line 621 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 621 "View.MemoryView"
  return __pyx_r;

#line 621 "View.MemoryView"
}

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice
#line 621 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp
#line 621 "View.MemoryView"
;

#line 621 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 621 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 621 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 621 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 621 "View.MemoryView"
  int __pyx_lineno = 0;

#line 621 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 621 "View.MemoryView"
  int __pyx_clineno = 0;

#line 621 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":624
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */

#line 624 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 624, __pyx_L1_error)

#line 624 "View.MemoryView"
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":625
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */

#line 625 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 625 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 625, __pyx_L1_error)

#line 625 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 625 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 625 "View.MemoryView"
  __pyx_t_2 = 0;

#line 625 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":621
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
  /* function exit code */

#line 621 "View.MemoryView"
  __pyx_L1_error:;

#line 621 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 621 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 621 "View.MemoryView"
  __pyx_r = NULL;

#line 621 "View.MemoryView"
  __pyx_L0:;

#line 621 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 621 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 621 "View.MemoryView"
  return __pyx_r;

#line 621 "View.MemoryView"
}

/* "View.MemoryView":627
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
/* Python wrapper */

#line 627 "View.MemoryView"
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 627 "View.MemoryView"
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 627 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 627 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 627 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = 
#line 627 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
  /* function exit code */

#line 627 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 627 "View.MemoryView"
  return __pyx_r;

#line 627 "View.MemoryView"
}

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice
#line 627 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp
#line 627 "View.MemoryView"
;

#line 627 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 627 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 627 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 627 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 627 "View.MemoryView"
  int __pyx_lineno = 0;

#line 627 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 627 "View.MemoryView"
  int __pyx_clineno = 0;

#line 627 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":630
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */

#line 630 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 630, __pyx_L1_error)

#line 630 "View.MemoryView"
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":631
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */

#line 631 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 631 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 631, __pyx_L1_error)

#line 631 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 631 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 631 "View.MemoryView"
  __pyx_t_2 = 0;

#line 631 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":627
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
  /* function exit code */

#line 627 "View.MemoryView"
  __pyx_L1_error:;

#line 627 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 627 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 627 "View.MemoryView"
  __pyx_r = NULL;

#line 627 "View.MemoryView"
  __pyx_L0:;

#line 627 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 627 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 627 "View.MemoryView"
  return __pyx_r;

#line 627 "View.MemoryView"
}

/* "View.MemoryView":633
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
/* Python wrapper */

#line 633 "View.MemoryView"
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 633 "View.MemoryView"
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 633 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 633 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 633 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = 
#line 633 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
  /* function exit code */

#line 633 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 633 "View.MemoryView"
  return __pyx_r;

#line 633 "View.MemoryView"
}

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice
#line 633 "View.MemoryView"
;
  int __pyx_v_flags
#line 633 "View.MemoryView"
;

#line 633 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 633 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 633 "View.MemoryView"
  __Pyx_memviewslice __pyx_t_1;

#line 633 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 633 "View.MemoryView"
  int __pyx_lineno = 0;

#line 633 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 633 "View.MemoryView"
  int __pyx_clineno = 0;

#line 633 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":635
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */

#line 635 "View.MemoryView"
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":637
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */

#line 637 "View.MemoryView"
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":638
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */

#line 638 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 638, __pyx_L1_error)

#line 638 "View.MemoryView"
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":643
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */

#line 643 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 643 "View.MemoryView"
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 643, __pyx_L1_error)

#line 643 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 643 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 643 "View.MemoryView"
  __pyx_t_2 = 0;

#line 643 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":633
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
  /* function exit code */

#line 633 "View.MemoryView"
  __pyx_L1_error:;

#line 633 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 633 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 633 "View.MemoryView"
  __pyx_r = NULL;

#line 633 "View.MemoryView"
  __pyx_L0:;

#line 633 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 633 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 633 "View.MemoryView"
  return __pyx_r;

#line 633 "View.MemoryView"
}

/* "View.MemoryView":645
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
/* Python wrapper */

#line 645 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 645 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 645 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 645 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 645 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = 
#line 645 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
  /* function exit code */

#line 645 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 645 "View.MemoryView"
  return __pyx_r;

#line 645 "View.MemoryView"
}

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src
#line 645 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_dst
#line 645 "View.MemoryView"
;
  int __pyx_v_flags
#line 645 "View.MemoryView"
;

#line 645 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 645 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 645 "View.MemoryView"
  __Pyx_memviewslice __pyx_t_1;

#line 645 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 645 "View.MemoryView"
  int __pyx_lineno = 0;

#line 645 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 645 "View.MemoryView"
  int __pyx_clineno = 0;

#line 645 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":647
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */

#line 647 "View.MemoryView"
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":649
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */

#line 649 "View.MemoryView"
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":650
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */

#line 650 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 650, __pyx_L1_error)

#line 650 "View.MemoryView"
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":655
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 655 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 655 "View.MemoryView"
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 655, __pyx_L1_error)

#line 655 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 655 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 655 "View.MemoryView"
  __pyx_t_2 = 0;

#line 655 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":645
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
  /* function exit code */

#line 645 "View.MemoryView"
  __pyx_L1_error:;

#line 645 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 645 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 645 "View.MemoryView"
  __pyx_r = NULL;

#line 645 "View.MemoryView"
  __pyx_L0:;

#line 645 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 645 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 645 "View.MemoryView"
  return __pyx_r;

#line 645 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 2 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 2 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 2 "(tree fragment)"
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
/* Python wrapper */

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 3 "(tree fragment)"
__pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 3 "(tree fragment)"
  int __pyx_lineno = 0;

#line 3 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 3 "(tree fragment)"
  int __pyx_clineno = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __pyx_L1_error:;

#line 3 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "(tree fragment)"
  __pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

/* "View.MemoryView":659
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

#line 659 "View.MemoryView"


#line 659 "View.MemoryView"
static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0
#line 659 "View.MemoryView"
;

#line 659 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 659 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 659 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 659 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 659 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 659 "View.MemoryView"
  int __pyx_lineno = 0;

#line 659 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 659 "View.MemoryView"
  int __pyx_clineno = 0;

#line 659 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":660
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */

#line 660 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 660 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 660 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 660 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_o);

#line 660 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_o);

#line 660 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);

#line 660 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 660 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);

#line 660 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 660 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 660 "View.MemoryView"
  __pyx_t_1 = 0;

#line 660 "View.MemoryView"
  __pyx_t_2 = 0;

#line 660 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 660 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 660 "View.MemoryView"
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);

#line 660 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":661
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */

#line 661 "View.MemoryView"
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":662
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */

#line 662 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 662 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 662 "View.MemoryView"
  __pyx_r = ((PyObject *)__pyx_v_result);

#line 662 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":659
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

#line 659 "View.MemoryView"


#line 659 "View.MemoryView"
  /* function exit code */

#line 659 "View.MemoryView"
  __pyx_L1_error:;

#line 659 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 659 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 659 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 659 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 659 "View.MemoryView"
  __pyx_r = 0;

#line 659 "View.MemoryView"
  __pyx_L0:;

#line 659 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 659 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 659 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 659 "View.MemoryView"
  return __pyx_r;

#line 659 "View.MemoryView"
}

/* "View.MemoryView":665
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

#line 665 "View.MemoryView"


#line 665 "View.MemoryView"
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {

#line 665 "View.MemoryView"
  int __pyx_r;

#line 665 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 665 "View.MemoryView"
  int __pyx_t_1;

#line 665 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":666
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */

#line 666 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 

#line 666 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 666 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":665
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

#line 665 "View.MemoryView"


#line 665 "View.MemoryView"
  /* function exit code */

#line 665 "View.MemoryView"
  __pyx_L0:;

#line 665 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 665 "View.MemoryView"
  return __pyx_r;

#line 665 "View.MemoryView"
}

/* "View.MemoryView":668
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

#line 668 "View.MemoryView"


#line 668 "View.MemoryView"
static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL
#line 668 "View.MemoryView"
;
  PyObject *__pyx_v_result = NULL
#line 668 "View.MemoryView"
;
  int __pyx_v_have_slices
#line 668 "View.MemoryView"
;
  int __pyx_v_seen_ellipsis
#line 668 "View.MemoryView"
;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL
#line 668 "View.MemoryView"
;
  PyObject *__pyx_v_item = NULL
#line 668 "View.MemoryView"
;
  Py_ssize_t __pyx_v_nslices
#line 668 "View.MemoryView"
;

#line 668 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 668 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 668 "View.MemoryView"
  int __pyx_t_1;

#line 668 "View.MemoryView"
  int __pyx_t_2;

#line 668 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 668 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 668 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 668 "View.MemoryView"
  PyObject *(*__pyx_t_6)(PyObject *);

#line 668 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 668 "View.MemoryView"
  Py_ssize_t __pyx_t_8;

#line 668 "View.MemoryView"
  int __pyx_t_9;

#line 668 "View.MemoryView"
  int __pyx_t_10;

#line 668 "View.MemoryView"
  PyObject *__pyx_t_11 = NULL;

#line 668 "View.MemoryView"
  int __pyx_lineno = 0;

#line 668 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 668 "View.MemoryView"
  int __pyx_clineno = 0;

#line 668 "View.MemoryView"
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":673
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */

#line 673 "View.MemoryView"
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 

#line 673 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);

#line 673 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":674
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */

#line 674 "View.MemoryView"
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 674, __pyx_L1_error)

#line 674 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 674 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_index);

#line 674 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_index);

#line 674 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);

#line 674 "View.MemoryView"
    __pyx_v_tup = __pyx_t_3;

#line 674 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":673
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */

#line 673 "View.MemoryView"
    goto __pyx_L3;

#line 673 "View.MemoryView"
  }

  /* "View.MemoryView":676
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */

#line 676 "View.MemoryView"
  /*else*/ {

#line 676 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_index);

#line 676 "View.MemoryView"
    __pyx_v_tup = __pyx_v_index;

#line 676 "View.MemoryView"
  }

#line 676 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":678
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */

#line 678 "View.MemoryView"
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 678, __pyx_L1_error)

#line 678 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 678 "View.MemoryView"
  __pyx_v_result = ((PyObject*)__pyx_t_3);

#line 678 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":679
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */

#line 679 "View.MemoryView"
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":680
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */

#line 680 "View.MemoryView"
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":681
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */

#line 681 "View.MemoryView"
  __Pyx_INCREF(__pyx_int_0);

#line 681 "View.MemoryView"
  __pyx_t_3 = __pyx_int_0;

#line 681 "View.MemoryView"
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {

#line 681 "View.MemoryView"
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;

#line 681 "View.MemoryView"
    __pyx_t_6 = NULL;

#line 681 "View.MemoryView"
  } else {
    __pyx_t_5 = -1; 
#line 681 "View.MemoryView"
__pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 681 "View.MemoryView"
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
  }

#line 681 "View.MemoryView"
  for (;;) {

#line 681 "View.MemoryView"
    if (likely(!__pyx_t_6)) {

#line 681 "View.MemoryView"
      if (likely(PyList_CheckExact(__pyx_t_4))) {

#line 681 "View.MemoryView"
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;

#line 681 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 681 "View.MemoryView"
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        #else

#line 681 "View.MemoryView"
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
        #endif

#line 681 "View.MemoryView"
      } else {

#line 681 "View.MemoryView"
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;

#line 681 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 681 "View.MemoryView"
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        #else

#line 681 "View.MemoryView"
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
        #endif

#line 681 "View.MemoryView"
      }
    } else 
#line 681 "View.MemoryView"
{

#line 681 "View.MemoryView"
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);

#line 681 "View.MemoryView"
      if (unlikely(!__pyx_t_7)) {

#line 681 "View.MemoryView"
        PyObject* exc_type = PyErr_Occurred();

#line 681 "View.MemoryView"
        if (exc_type) {

#line 681 "View.MemoryView"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 681 "View.MemoryView"
          else __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        }

#line 681 "View.MemoryView"
        break;

#line 681 "View.MemoryView"
      }

#line 681 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
    }

#line 681 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);

#line 681 "View.MemoryView"
    __pyx_t_7 = 0;

#line 681 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 681 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);

#line 681 "View.MemoryView"
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3);

#line 681 "View.MemoryView"
    __pyx_t_3 = __pyx_t_7;

#line 681 "View.MemoryView"
    __pyx_t_7 = 0;

    /* "View.MemoryView":682
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */

#line 682 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);

#line 682 "View.MemoryView"
    __pyx_t_1 = (__pyx_t_2 != 0);

#line 682 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":683
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */

#line 683 "View.MemoryView"
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);

#line 683 "View.MemoryView"
      if (__pyx_t_1) {

        /* "View.MemoryView":684
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */

#line 684 "View.MemoryView"
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(2, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 684 "View.MemoryView"
        { Py_ssize_t __pyx_temp;

#line 684 "View.MemoryView"
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {

#line 684 "View.MemoryView"
            __Pyx_INCREF(__pyx_slice__21);

#line 684 "View.MemoryView"
            __Pyx_GIVEREF(__pyx_slice__21);

#line 684 "View.MemoryView"
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__21);

#line 684 "View.MemoryView"
          }

#line 684 "View.MemoryView"
        }

#line 684 "View.MemoryView"
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":685
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */

#line 685 "View.MemoryView"
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":683
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */

#line 683 "View.MemoryView"
        goto __pyx_L7;

#line 683 "View.MemoryView"
      }

      /* "View.MemoryView":687
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */

#line 687 "View.MemoryView"
      /*else*/ {

#line 687 "View.MemoryView"
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__21); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 687, __pyx_L1_error)

#line 687 "View.MemoryView"
      }

#line 687 "View.MemoryView"
      __pyx_L7:;

      /* "View.MemoryView":688
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */

#line 688 "View.MemoryView"
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":682
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */

#line 682 "View.MemoryView"
      goto __pyx_L6;

#line 682 "View.MemoryView"
    }

    /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */

#line 690 "View.MemoryView"
    /*else*/ {

#line 690 "View.MemoryView"
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 

#line 690 "View.MemoryView"
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);

#line 690 "View.MemoryView"
      if (__pyx_t_10) {

#line 690 "View.MemoryView"
      } else {

#line 690 "View.MemoryView"
        __pyx_t_1 = __pyx_t_10;

#line 690 "View.MemoryView"
        goto __pyx_L9_bool_binop_done;

#line 690 "View.MemoryView"
      }

#line 690 "View.MemoryView"
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);

#line 690 "View.MemoryView"
      __pyx_t_1 = __pyx_t_10;

#line 690 "View.MemoryView"
      __pyx_L9_bool_binop_done:;

#line 690 "View.MemoryView"
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":691
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */

#line 691 "View.MemoryView"
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 691, __pyx_L1_error)

#line 691 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 691 "View.MemoryView"
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 691, __pyx_L1_error)

#line 691 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_11);

#line 691 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 691 "View.MemoryView"
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);

#line 691 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 691 "View.MemoryView"
        __PYX_ERR(2, 691, __pyx_L1_error)

        /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */

#line 690 "View.MemoryView"
      }

      /* "View.MemoryView":693
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */

#line 693 "View.MemoryView"
      __pyx_t_10 = (__pyx_v_have_slices != 0);

#line 693 "View.MemoryView"
      if (!__pyx_t_10) {

#line 693 "View.MemoryView"
      } else {

#line 693 "View.MemoryView"
        __pyx_t_1 = __pyx_t_10;

#line 693 "View.MemoryView"
        goto __pyx_L11_bool_binop_done;

#line 693 "View.MemoryView"
      }

#line 693 "View.MemoryView"
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 

#line 693 "View.MemoryView"
      __pyx_t_2 = (__pyx_t_10 != 0);

#line 693 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 693 "View.MemoryView"
      __pyx_L11_bool_binop_done:;

#line 693 "View.MemoryView"
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":694
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */

#line 694 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 694, __pyx_L1_error)

#line 694 "View.MemoryView"
    }

#line 694 "View.MemoryView"
    __pyx_L6:;

    /* "View.MemoryView":681
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */

#line 681 "View.MemoryView"
  }

#line 681 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 681 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":696
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */

#line 696 "View.MemoryView"
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 696, __pyx_L1_error)

#line 696 "View.MemoryView"
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":697
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */

#line 697 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_nslices != 0);

#line 697 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":698
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */

#line 698 "View.MemoryView"
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 698, __pyx_L1_error)

#line 698 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 698 "View.MemoryView"
    { Py_ssize_t __pyx_temp;

#line 698 "View.MemoryView"
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {

#line 698 "View.MemoryView"
        __Pyx_INCREF(__pyx_slice__21);

#line 698 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_slice__21);

#line 698 "View.MemoryView"
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__21);

#line 698 "View.MemoryView"
      }

#line 698 "View.MemoryView"
    }

#line 698 "View.MemoryView"
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 698, __pyx_L1_error)

#line 698 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":697
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */

#line 697 "View.MemoryView"
  }

  /* "View.MemoryView":700
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */

#line 700 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 700 "View.MemoryView"
  if (!__pyx_v_have_slices) {

#line 700 "View.MemoryView"
  } else {

#line 700 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 700 "View.MemoryView"
    __pyx_t_3 = __pyx_t_4;

#line 700 "View.MemoryView"
    __pyx_t_4 = 0;

#line 700 "View.MemoryView"
    goto __pyx_L14_bool_binop_done;

#line 700 "View.MemoryView"
  }

#line 700 "View.MemoryView"
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 700 "View.MemoryView"
  __pyx_t_3 = __pyx_t_4;

#line 700 "View.MemoryView"
  __pyx_t_4 = 0;

#line 700 "View.MemoryView"
  __pyx_L14_bool_binop_done:;

#line 700 "View.MemoryView"
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 700 "View.MemoryView"
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_11);

#line 700 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 700 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);

#line 700 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_4);

#line 700 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);

#line 700 "View.MemoryView"
  __pyx_t_3 = 0;

#line 700 "View.MemoryView"
  __pyx_t_4 = 0;

#line 700 "View.MemoryView"
  __pyx_r = ((PyObject*)__pyx_t_11);

#line 700 "View.MemoryView"
  __pyx_t_11 = 0;

#line 700 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":668
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

#line 668 "View.MemoryView"


#line 668 "View.MemoryView"
  /* function exit code */

#line 668 "View.MemoryView"
  __pyx_L1_error:;

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_7);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_11);

#line 668 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 668 "View.MemoryView"
  __pyx_r = 0;

#line 668 "View.MemoryView"
  __pyx_L0:;

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_tup);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_result);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_idx);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_item);

#line 668 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 668 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 668 "View.MemoryView"
  return __pyx_r;

#line 668 "View.MemoryView"
}

/* "View.MemoryView":702
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

#line 702 "View.MemoryView"


#line 702 "View.MemoryView"
static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset
#line 702 "View.MemoryView"
;

#line 702 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 702 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 702 "View.MemoryView"
  Py_ssize_t *__pyx_t_1;

#line 702 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 702 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 702 "View.MemoryView"
  int __pyx_t_4;

#line 702 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 702 "View.MemoryView"
  int __pyx_lineno = 0;

#line 702 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 702 "View.MemoryView"
  int __pyx_clineno = 0;

#line 702 "View.MemoryView"
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":703
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */

#line 703 "View.MemoryView"
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);

#line 703 "View.MemoryView"
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {

#line 703 "View.MemoryView"
    __pyx_t_1 = __pyx_t_3;

#line 703 "View.MemoryView"
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":704
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */

#line 704 "View.MemoryView"
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);

#line 704 "View.MemoryView"
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":705
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 705 "View.MemoryView"
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 705, __pyx_L1_error)

#line 705 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 705 "View.MemoryView"
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);

#line 705 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 705 "View.MemoryView"
      __PYX_ERR(2, 705, __pyx_L1_error)

      /* "View.MemoryView":704
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */

#line 704 "View.MemoryView"
    }

#line 704 "View.MemoryView"
  }

  /* "View.MemoryView":702
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

#line 702 "View.MemoryView"


#line 702 "View.MemoryView"
  /* function exit code */

#line 702 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 702 "View.MemoryView"
  goto __pyx_L0;

#line 702 "View.MemoryView"
  __pyx_L1_error:;

#line 702 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 702 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 702 "View.MemoryView"
  __pyx_r = 0;

#line 702 "View.MemoryView"
  __pyx_L0:;

#line 702 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 702 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 702 "View.MemoryView"
  return __pyx_r;

#line 702 "View.MemoryView"
}

/* "View.MemoryView":712
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

#line 712 "View.MemoryView"


#line 712 "View.MemoryView"
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim
#line 712 "View.MemoryView"
;
  int __pyx_v_suboffset_dim
#line 712 "View.MemoryView"
;
  int __pyx_v_dim
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_src
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_dst
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice *__pyx_v_p_src
#line 712 "View.MemoryView"
;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice *__pyx_v_p_dst
#line 712 "View.MemoryView"
;
  int *__pyx_v_p_suboffset_dim
#line 712 "View.MemoryView"
;
  Py_ssize_t __pyx_v_start
#line 712 "View.MemoryView"
;
  Py_ssize_t __pyx_v_stop
#line 712 "View.MemoryView"
;
  Py_ssize_t __pyx_v_step
#line 712 "View.MemoryView"
;
  int __pyx_v_have_start
#line 712 "View.MemoryView"
;
  int __pyx_v_have_stop
#line 712 "View.MemoryView"
;
  int __pyx_v_have_step
#line 712 "View.MemoryView"
;
  PyObject *__pyx_v_index = NULL
#line 712 "View.MemoryView"
;

#line 712 "View.MemoryView"
  struct __pyx_memoryview_obj *__pyx_r = NULL;

#line 712 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 712 "View.MemoryView"
  int __pyx_t_1;

#line 712 "View.MemoryView"
  int __pyx_t_2;

#line 712 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 712 "View.MemoryView"
  struct __pyx_memoryview_obj *__pyx_t_4;

#line 712 "View.MemoryView"
  char *__pyx_t_5;

#line 712 "View.MemoryView"
  int __pyx_t_6;

#line 712 "View.MemoryView"
  Py_ssize_t __pyx_t_7;

#line 712 "View.MemoryView"
  PyObject *(*__pyx_t_8)(PyObject *);

#line 712 "View.MemoryView"
  PyObject *__pyx_t_9 = NULL;

#line 712 "View.MemoryView"
  Py_ssize_t __pyx_t_10;

#line 712 "View.MemoryView"
  int __pyx_t_11;

#line 712 "View.MemoryView"
  Py_ssize_t __pyx_t_12;

#line 712 "View.MemoryView"
  int __pyx_lineno = 0;

#line 712 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 712 "View.MemoryView"
  int __pyx_clineno = 0;

#line 712 "View.MemoryView"
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":713
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */

#line 713 "View.MemoryView"
  __pyx_v_new_ndim = 0;

#line 713 "View.MemoryView"
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":720
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */

#line 720 "View.MemoryView"
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":724
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */

#line 724 "View.MemoryView"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 724 "View.MemoryView"
  if (unlikely(!Py_OptimizeFlag)) {

#line 724 "View.MemoryView"
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {

#line 724 "View.MemoryView"
      PyErr_SetNone(PyExc_AssertionError);

#line 724 "View.MemoryView"
      __PYX_ERR(2, 724, __pyx_L1_error)

#line 724 "View.MemoryView"
    }

#line 724 "View.MemoryView"
  }

#line 724 "View.MemoryView"
  #endif

  /* "View.MemoryView":726
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */

#line 726 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 726 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 726 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":727
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */

#line 727 "View.MemoryView"
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(2, 727, __pyx_L1_error)

#line 727 "View.MemoryView"
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);

#line 727 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 727 "View.MemoryView"
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);

#line 727 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":728
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */

#line 728 "View.MemoryView"
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":726
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */

#line 726 "View.MemoryView"
    goto __pyx_L3;

#line 726 "View.MemoryView"
  }

  /* "View.MemoryView":730
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */

#line 730 "View.MemoryView"
  /*else*/ {

#line 730 "View.MemoryView"
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":731
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 731 "View.MemoryView"
    __pyx_v_p_src = (&__pyx_v_src);

#line 731 "View.MemoryView"
  }

#line 731 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":737
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */

#line 737 "View.MemoryView"
  __pyx_t_4 = __pyx_v_p_src->memview;

#line 737 "View.MemoryView"
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":738
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 738 "View.MemoryView"
  __pyx_t_5 = __pyx_v_p_src->data;

#line 738 "View.MemoryView"
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":743
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */

#line 743 "View.MemoryView"
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":744
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */

#line 744 "View.MemoryView"
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":748
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */

#line 748 "View.MemoryView"
  __pyx_t_6 = 0;

#line 748 "View.MemoryView"
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {

#line 748 "View.MemoryView"
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;

#line 748 "View.MemoryView"
    __pyx_t_8 = NULL;

#line 748 "View.MemoryView"
  } else {
    __pyx_t_7 = -1; 
#line 748 "View.MemoryView"
__pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 748 "View.MemoryView"
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
  }

#line 748 "View.MemoryView"
  for (;;) {

#line 748 "View.MemoryView"
    if (likely(!__pyx_t_8)) {

#line 748 "View.MemoryView"
      if (likely(PyList_CheckExact(__pyx_t_3))) {

#line 748 "View.MemoryView"
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;

#line 748 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 748 "View.MemoryView"
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        #else

#line 748 "View.MemoryView"
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_9);

#line 748 "View.MemoryView"
        #endif

#line 748 "View.MemoryView"
      } else {

#line 748 "View.MemoryView"
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;

#line 748 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 748 "View.MemoryView"
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        #else

#line 748 "View.MemoryView"
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_9);

#line 748 "View.MemoryView"
        #endif

#line 748 "View.MemoryView"
      }
    } else 
#line 748 "View.MemoryView"
{

#line 748 "View.MemoryView"
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);

#line 748 "View.MemoryView"
      if (unlikely(!__pyx_t_9)) {

#line 748 "View.MemoryView"
        PyObject* exc_type = PyErr_Occurred();

#line 748 "View.MemoryView"
        if (exc_type) {

#line 748 "View.MemoryView"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 748 "View.MemoryView"
          else __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        }

#line 748 "View.MemoryView"
        break;

#line 748 "View.MemoryView"
      }

#line 748 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 748 "View.MemoryView"
    }

#line 748 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);

#line 748 "View.MemoryView"
    __pyx_t_9 = 0;

#line 748 "View.MemoryView"
    __pyx_v_dim = __pyx_t_6;

#line 748 "View.MemoryView"
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":749
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */

#line 749 "View.MemoryView"
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);

#line 749 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":753
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */

#line 753 "View.MemoryView"
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 753, __pyx_L1_error)

      /* "View.MemoryView":750
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */

#line 750 "View.MemoryView"
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(2, 750, __pyx_L1_error)

      /* "View.MemoryView":749
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */

#line 749 "View.MemoryView"
      goto __pyx_L6;

#line 749 "View.MemoryView"
    }

    /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */

#line 756 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_index == Py_None);

#line 756 "View.MemoryView"
    __pyx_t_1 = (__pyx_t_2 != 0);

#line 756 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":757
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */

#line 757 "View.MemoryView"
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":758
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */

#line 758 "View.MemoryView"
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":759
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */

#line 759 "View.MemoryView"
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":760
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */

#line 760 "View.MemoryView"
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */

#line 756 "View.MemoryView"
      goto __pyx_L6;

#line 756 "View.MemoryView"
    }

    /* "View.MemoryView":762
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */

#line 762 "View.MemoryView"
    /*else*/ {

#line 762 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 762, __pyx_L1_error)

#line 762 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 762 "View.MemoryView"
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 762, __pyx_L1_error)

#line 762 "View.MemoryView"
      if (!__pyx_t_1) {

#line 762 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 762 "View.MemoryView"
      } else {

#line 762 "View.MemoryView"
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 762, __pyx_L1_error)

#line 762 "View.MemoryView"
        __pyx_t_10 = __pyx_t_12;

#line 762 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 762 "View.MemoryView"
        goto __pyx_L7_bool_binop_done;

#line 762 "View.MemoryView"
      }

#line 762 "View.MemoryView"
      __pyx_t_10 = 0;

#line 762 "View.MemoryView"
      __pyx_L7_bool_binop_done:;

#line 762 "View.MemoryView"
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":763
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */

#line 763 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 763, __pyx_L1_error)

#line 763 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 763 "View.MemoryView"
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 763, __pyx_L1_error)

#line 763 "View.MemoryView"
      if (!__pyx_t_1) {

#line 763 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 763 "View.MemoryView"
      } else {

#line 763 "View.MemoryView"
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 763, __pyx_L1_error)

#line 763 "View.MemoryView"
        __pyx_t_10 = __pyx_t_12;

#line 763 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 763 "View.MemoryView"
        goto __pyx_L9_bool_binop_done;

#line 763 "View.MemoryView"
      }

#line 763 "View.MemoryView"
      __pyx_t_10 = 0;

#line 763 "View.MemoryView"
      __pyx_L9_bool_binop_done:;

#line 763 "View.MemoryView"
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":764
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */

#line 764 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 764, __pyx_L1_error)

#line 764 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 764 "View.MemoryView"
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 764, __pyx_L1_error)

#line 764 "View.MemoryView"
      if (!__pyx_t_1) {

#line 764 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 764 "View.MemoryView"
      } else {

#line 764 "View.MemoryView"
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 764, __pyx_L1_error)

#line 764 "View.MemoryView"
        __pyx_t_10 = __pyx_t_12;

#line 764 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 764 "View.MemoryView"
        goto __pyx_L11_bool_binop_done;

#line 764 "View.MemoryView"
      }

#line 764 "View.MemoryView"
      __pyx_t_10 = 0;

#line 764 "View.MemoryView"
      __pyx_L11_bool_binop_done:;

#line 764 "View.MemoryView"
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":766
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */

#line 766 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 766, __pyx_L1_error)

#line 766 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 766 "View.MemoryView"
      __pyx_t_1 = (__pyx_t_9 != Py_None);

#line 766 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 766 "View.MemoryView"
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":767
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */

#line 767 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 767, __pyx_L1_error)

#line 767 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 767 "View.MemoryView"
      __pyx_t_1 = (__pyx_t_9 != Py_None);

#line 767 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 767 "View.MemoryView"
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */

#line 768 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 768, __pyx_L1_error)

#line 768 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 768 "View.MemoryView"
      __pyx_t_1 = (__pyx_t_9 != Py_None);

#line 768 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 768 "View.MemoryView"
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":770
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */

#line 770 "View.MemoryView"
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(2, 770, __pyx_L1_error)

      /* "View.MemoryView":776
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */

#line 776 "View.MemoryView"
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

#line 776 "View.MemoryView"
    }

#line 776 "View.MemoryView"
    __pyx_L6:;

    /* "View.MemoryView":748
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */

#line 748 "View.MemoryView"
  }

#line 748 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */

#line 778 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 778 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 778 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */

#line 779 "View.MemoryView"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":780
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */

#line 780 "View.MemoryView"
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(2, 780, __pyx_L1_error) }

    /* "View.MemoryView":781
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */

#line 781 "View.MemoryView"
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(2, 781, __pyx_L1_error) }

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */

#line 779 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 779, __pyx_L1_error)

#line 779 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 779 "View.MemoryView"
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(2, 779, __pyx_L1_error)

#line 779 "View.MemoryView"
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);

#line 779 "View.MemoryView"
    __pyx_t_3 = 0;

#line 779 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */

#line 778 "View.MemoryView"
  }

  /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */

#line 784 "View.MemoryView"
  /*else*/ {

#line 784 "View.MemoryView"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":785
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 785 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 784, __pyx_L1_error)

#line 785 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */

#line 784 "View.MemoryView"
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(2, 784, __pyx_L1_error)

#line 784 "View.MemoryView"
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);

#line 784 "View.MemoryView"
    __pyx_t_3 = 0;

#line 784 "View.MemoryView"
    goto __pyx_L0;

#line 784 "View.MemoryView"
  }

  /* "View.MemoryView":712
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

#line 712 "View.MemoryView"


#line 712 "View.MemoryView"
  /* function exit code */

#line 712 "View.MemoryView"
  __pyx_L1_error:;

#line 712 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 712 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_9);

#line 712 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 712 "View.MemoryView"
  __pyx_r = 0;

#line 712 "View.MemoryView"
  __pyx_L0:;

#line 712 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);

#line 712 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_index);

#line 712 "View.MemoryView"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 712 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 712 "View.MemoryView"
  return __pyx_r;

#line 712 "View.MemoryView"
}

/* "View.MemoryView":809
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

#line 809 "View.MemoryView"


#line 809 "View.MemoryView"
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape
#line 809 "View.MemoryView"
;
  int __pyx_v_negative_step
#line 809 "View.MemoryView"
;

#line 809 "View.MemoryView"
  int __pyx_r;

#line 809 "View.MemoryView"
  int __pyx_t_1;

#line 809 "View.MemoryView"
  int __pyx_t_2;

#line 809 "View.MemoryView"
  int __pyx_t_3;

#line 809 "View.MemoryView"
  int __pyx_lineno = 0;

#line 809 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 809 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":829
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */

#line 829 "View.MemoryView"
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);

#line 829 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":831
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */

#line 831 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);

#line 831 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":832
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */

#line 832 "View.MemoryView"
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":831
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */

#line 831 "View.MemoryView"
    }

    /* "View.MemoryView":833
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */

#line 833 "View.MemoryView"
    __pyx_t_1 = (0 <= __pyx_v_start);

#line 833 "View.MemoryView"
    if (__pyx_t_1) {

#line 833 "View.MemoryView"
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);

#line 833 "View.MemoryView"
    }

#line 833 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);

#line 833 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":834
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */

#line 834 "View.MemoryView"
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 834, __pyx_L1_error)

      /* "View.MemoryView":833
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */

#line 833 "View.MemoryView"
    }

    /* "View.MemoryView":829
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */

#line 829 "View.MemoryView"
    goto __pyx_L3;

#line 829 "View.MemoryView"
  }

  /* "View.MemoryView":837
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */

#line 837 "View.MemoryView"
  /*else*/ {

#line 837 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);

#line 837 "View.MemoryView"
    if (__pyx_t_1) {

#line 837 "View.MemoryView"
    } else {

#line 837 "View.MemoryView"
      __pyx_t_2 = __pyx_t_1;

#line 837 "View.MemoryView"
      goto __pyx_L6_bool_binop_done;

#line 837 "View.MemoryView"
    }

#line 837 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);

#line 837 "View.MemoryView"
    __pyx_t_2 = __pyx_t_1;

#line 837 "View.MemoryView"
    __pyx_L6_bool_binop_done:;

#line 837 "View.MemoryView"
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":839
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */

#line 839 "View.MemoryView"
    __pyx_t_1 = (__pyx_v_have_step != 0);

#line 839 "View.MemoryView"
    if (__pyx_t_1) {

#line 839 "View.MemoryView"
    } else {

#line 839 "View.MemoryView"
      __pyx_t_2 = __pyx_t_1;

#line 839 "View.MemoryView"
      goto __pyx_L9_bool_binop_done;

#line 839 "View.MemoryView"
    }

#line 839 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);

#line 839 "View.MemoryView"
    __pyx_t_2 = __pyx_t_1;

#line 839 "View.MemoryView"
    __pyx_L9_bool_binop_done:;

#line 839 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":840
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 840 "View.MemoryView"
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 840, __pyx_L1_error)

      /* "View.MemoryView":839
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */

#line 839 "View.MemoryView"
    }

    /* "View.MemoryView":843
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */

#line 843 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_have_start != 0);

#line 843 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":844
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */

#line 844 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);

#line 844 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":845
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */

#line 845 "View.MemoryView"
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":846
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */

#line 846 "View.MemoryView"
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);

#line 846 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":847
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */

#line 847 "View.MemoryView"
          __pyx_v_start = 0;

          /* "View.MemoryView":846
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */

#line 846 "View.MemoryView"
        }

        /* "View.MemoryView":844
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */

#line 844 "View.MemoryView"
        goto __pyx_L12;

#line 844 "View.MemoryView"
      }

      /* "View.MemoryView":848
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */

#line 848 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);

#line 848 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":849
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */

#line 849 "View.MemoryView"
        __pyx_t_2 = (__pyx_v_negative_step != 0);

#line 849 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":850
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */

#line 850 "View.MemoryView"
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":849
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */

#line 849 "View.MemoryView"
          goto __pyx_L14;

#line 849 "View.MemoryView"
        }

        /* "View.MemoryView":852
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */

#line 852 "View.MemoryView"
        /*else*/ {

#line 852 "View.MemoryView"
          __pyx_v_start = __pyx_v_shape;

#line 852 "View.MemoryView"
        }

#line 852 "View.MemoryView"
        __pyx_L14:;

        /* "View.MemoryView":848
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */

#line 848 "View.MemoryView"
      }

#line 848 "View.MemoryView"
      __pyx_L12:;

      /* "View.MemoryView":843
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */

#line 843 "View.MemoryView"
      goto __pyx_L11;

#line 843 "View.MemoryView"
    }

    /* "View.MemoryView":854
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */

#line 854 "View.MemoryView"
    /*else*/ {

#line 854 "View.MemoryView"
      __pyx_t_2 = (__pyx_v_negative_step != 0);

#line 854 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":855
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */

#line 855 "View.MemoryView"
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":854
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */

#line 854 "View.MemoryView"
        goto __pyx_L15;

#line 854 "View.MemoryView"
      }

      /* "View.MemoryView":857
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */

#line 857 "View.MemoryView"
      /*else*/ {

#line 857 "View.MemoryView"
        __pyx_v_start = 0;

#line 857 "View.MemoryView"
      }

#line 857 "View.MemoryView"
      __pyx_L15:;

#line 857 "View.MemoryView"
    }

#line 857 "View.MemoryView"
    __pyx_L11:;

    /* "View.MemoryView":859
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */

#line 859 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_have_stop != 0);

#line 859 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":860
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */

#line 860 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);

#line 860 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":861
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */

#line 861 "View.MemoryView"
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":862
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */

#line 862 "View.MemoryView"
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);

#line 862 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":863
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */

#line 863 "View.MemoryView"
          __pyx_v_stop = 0;

          /* "View.MemoryView":862
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */

#line 862 "View.MemoryView"
        }

        /* "View.MemoryView":860
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */

#line 860 "View.MemoryView"
        goto __pyx_L17;

#line 860 "View.MemoryView"
      }

      /* "View.MemoryView":864
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */

#line 864 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);

#line 864 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":865
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */

#line 865 "View.MemoryView"
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":864
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */

#line 864 "View.MemoryView"
      }

#line 864 "View.MemoryView"
      __pyx_L17:;

      /* "View.MemoryView":859
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */

#line 859 "View.MemoryView"
      goto __pyx_L16;

#line 859 "View.MemoryView"
    }

    /* "View.MemoryView":867
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */

#line 867 "View.MemoryView"
    /*else*/ {

#line 867 "View.MemoryView"
      __pyx_t_2 = (__pyx_v_negative_step != 0);

#line 867 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":868
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */

#line 868 "View.MemoryView"
        __pyx_v_stop = -1L;

        /* "View.MemoryView":867
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */

#line 867 "View.MemoryView"
        goto __pyx_L19;

#line 867 "View.MemoryView"
      }

      /* "View.MemoryView":870
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */

#line 870 "View.MemoryView"
      /*else*/ {

#line 870 "View.MemoryView"
        __pyx_v_stop = __pyx_v_shape;

#line 870 "View.MemoryView"
      }

#line 870 "View.MemoryView"
      __pyx_L19:;

#line 870 "View.MemoryView"
    }

#line 870 "View.MemoryView"
    __pyx_L16:;

    /* "View.MemoryView":872
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */

#line 872 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);

#line 872 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":873
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 873 "View.MemoryView"
      __pyx_v_step = 1;

      /* "View.MemoryView":872
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */

#line 872 "View.MemoryView"
    }

    /* "View.MemoryView":877
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */

#line 877 "View.MemoryView"
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":879
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */

#line 879 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);

#line 879 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":880
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */

#line 880 "View.MemoryView"
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":879
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */

#line 879 "View.MemoryView"
    }

    /* "View.MemoryView":882
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */

#line 882 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);

#line 882 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":883
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 883 "View.MemoryView"
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":882
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */

#line 882 "View.MemoryView"
    }

    /* "View.MemoryView":886
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */

#line 886 "View.MemoryView"
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":887
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */

#line 887 "View.MemoryView"
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":888
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 888 "View.MemoryView"
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;

#line 888 "View.MemoryView"
  }

#line 888 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":891
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */

#line 891 "View.MemoryView"
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);

#line 891 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":892
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */

#line 892 "View.MemoryView"
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":891
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */

#line 891 "View.MemoryView"
    goto __pyx_L23;

#line 891 "View.MemoryView"
  }

  /* "View.MemoryView":894
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */

#line 894 "View.MemoryView"
  /*else*/ {

#line 894 "View.MemoryView"
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);

#line 894 "View.MemoryView"
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));

#line 894 "View.MemoryView"
  }

#line 894 "View.MemoryView"
  __pyx_L23:;

  /* "View.MemoryView":896
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */

#line 896 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);

#line 896 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":897
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */

#line 897 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);

#line 897 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":898
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */

#line 898 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);

#line 898 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":899
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */

#line 899 "View.MemoryView"
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":898
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */

#line 898 "View.MemoryView"
        goto __pyx_L26;

#line 898 "View.MemoryView"
      }

      /* "View.MemoryView":901
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */

#line 901 "View.MemoryView"
      /*else*/ {

        /* "View.MemoryView":902
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */

#line 902 "View.MemoryView"
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 901, __pyx_L1_error)

#line 902 "View.MemoryView"
      }

#line 902 "View.MemoryView"
      __pyx_L26:;

      /* "View.MemoryView":897
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */

#line 897 "View.MemoryView"
      goto __pyx_L25;

#line 897 "View.MemoryView"
    }

    /* "View.MemoryView":904
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */

#line 904 "View.MemoryView"
    /*else*/ {

#line 904 "View.MemoryView"
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;

#line 904 "View.MemoryView"
    }

#line 904 "View.MemoryView"
    __pyx_L25:;

    /* "View.MemoryView":896
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */

#line 896 "View.MemoryView"
  }

  /* "View.MemoryView":906
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 906 "View.MemoryView"
  __pyx_r = 0;

#line 906 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":809
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

#line 809 "View.MemoryView"


#line 809 "View.MemoryView"
  /* function exit code */

#line 809 "View.MemoryView"
  __pyx_L1_error:;

#line 809 "View.MemoryView"
  {

#line 809 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 809 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 809 "View.MemoryView"
    #endif

#line 809 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 809 "View.MemoryView"
    #ifdef WITH_THREAD

#line 809 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 809 "View.MemoryView"
    #endif

#line 809 "View.MemoryView"
  }

#line 809 "View.MemoryView"
  __pyx_r = -1;

#line 809 "View.MemoryView"
  __pyx_L0:;

#line 809 "View.MemoryView"
  return __pyx_r;

#line 809 "View.MemoryView"
}

/* "View.MemoryView":912
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

#line 912 "View.MemoryView"


#line 912 "View.MemoryView"
static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape
#line 912 "View.MemoryView"
;
  Py_ssize_t __pyx_v_stride
#line 912 "View.MemoryView"
;
  Py_ssize_t __pyx_v_suboffset
#line 912 "View.MemoryView"
;
  Py_ssize_t __pyx_v_itemsize
#line 912 "View.MemoryView"
;
  char *__pyx_v_resultp
#line 912 "View.MemoryView"
;

#line 912 "View.MemoryView"
  char *__pyx_r;

#line 912 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 912 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 912 "View.MemoryView"
  int __pyx_t_2;

#line 912 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 912 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 912 "View.MemoryView"
  int __pyx_lineno = 0;

#line 912 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 912 "View.MemoryView"
  int __pyx_clineno = 0;

#line 912 "View.MemoryView"
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":914
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */

#line 914 "View.MemoryView"
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":915
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */

#line 915 "View.MemoryView"
  __pyx_t_1 = __pyx_v_view->itemsize;

#line 915 "View.MemoryView"
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":918
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */

#line 918 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);

#line 918 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":919
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */

#line 919 "View.MemoryView"
    if (unlikely(__pyx_v_itemsize == 0)) {

#line 919 "View.MemoryView"
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");

#line 919 "View.MemoryView"
      __PYX_ERR(2, 919, __pyx_L1_error)

#line 919 "View.MemoryView"
    }

#line 919 "View.MemoryView"
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {

#line 919 "View.MemoryView"
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");

#line 919 "View.MemoryView"
      __PYX_ERR(2, 919, __pyx_L1_error)

#line 919 "View.MemoryView"
    }

#line 919 "View.MemoryView"
    __pyx_v_shape = __Pyx_div_Py_ssize_t(__pyx_v_view->len, __pyx_v_itemsize);

    /* "View.MemoryView":920
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */

#line 920 "View.MemoryView"
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":918
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */

#line 918 "View.MemoryView"
    goto __pyx_L3;

#line 918 "View.MemoryView"
  }

  /* "View.MemoryView":922
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */

#line 922 "View.MemoryView"
  /*else*/ {

#line 922 "View.MemoryView"
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":923
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */

#line 923 "View.MemoryView"
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":924
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */

#line 924 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);

#line 924 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":925
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */

#line 925 "View.MemoryView"
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":924
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */

#line 924 "View.MemoryView"
    }

#line 924 "View.MemoryView"
  }

#line 924 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":927
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */

#line 927 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);

#line 927 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":928
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */

#line 928 "View.MemoryView"
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":929
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 929 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);

#line 929 "View.MemoryView"
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":930
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */

#line 930 "View.MemoryView"
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 930, __pyx_L1_error)

#line 930 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 930 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 930, __pyx_L1_error)

#line 930 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 930 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 930 "View.MemoryView"
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 930, __pyx_L1_error)

#line 930 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 930 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 930 "View.MemoryView"
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 930 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 930 "View.MemoryView"
      __PYX_ERR(2, 930, __pyx_L1_error)

      /* "View.MemoryView":929
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 929 "View.MemoryView"
    }

    /* "View.MemoryView":927
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */

#line 927 "View.MemoryView"
  }

  /* "View.MemoryView":932
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 932 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);

#line 932 "View.MemoryView"
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":933
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */

#line 933 "View.MemoryView"
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 933, __pyx_L1_error)

#line 933 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 933 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 933, __pyx_L1_error)

#line 933 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 933 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 933 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 933, __pyx_L1_error)

#line 933 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 933 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 933 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 933 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 933 "View.MemoryView"
    __PYX_ERR(2, 933, __pyx_L1_error)

    /* "View.MemoryView":932
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 932 "View.MemoryView"
  }

  /* "View.MemoryView":935
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */

#line 935 "View.MemoryView"
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":936
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */

#line 936 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);

#line 936 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":937
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */

#line 937 "View.MemoryView"
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":936
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */

#line 936 "View.MemoryView"
  }

  /* "View.MemoryView":939
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 939 "View.MemoryView"
  __pyx_r = __pyx_v_resultp;

#line 939 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":912
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

#line 912 "View.MemoryView"


#line 912 "View.MemoryView"
  /* function exit code */

#line 912 "View.MemoryView"
  __pyx_L1_error:;

#line 912 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 912 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 912 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 912 "View.MemoryView"
  __pyx_r = NULL;

#line 912 "View.MemoryView"
  __pyx_L0:;

#line 912 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 912 "View.MemoryView"
  return __pyx_r;

#line 912 "View.MemoryView"
}

/* "View.MemoryView":945
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

#line 945 "View.MemoryView"


#line 945 "View.MemoryView"
static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim
#line 945 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_shape
#line 945 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_strides
#line 945 "View.MemoryView"
;
  int __pyx_v_i
#line 945 "View.MemoryView"
;
  int __pyx_v_j
#line 945 "View.MemoryView"
;

#line 945 "View.MemoryView"
  int __pyx_r;

#line 945 "View.MemoryView"
  int __pyx_t_1;

#line 945 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 945 "View.MemoryView"
  long __pyx_t_3;

#line 945 "View.MemoryView"
  long __pyx_t_4;

#line 945 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 945 "View.MemoryView"
  Py_ssize_t __pyx_t_6;

#line 945 "View.MemoryView"
  int __pyx_t_7;

#line 945 "View.MemoryView"
  int __pyx_t_8;

#line 945 "View.MemoryView"
  int __pyx_t_9;

#line 945 "View.MemoryView"
  int __pyx_lineno = 0;

#line 945 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 945 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":946
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */

#line 946 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;

#line 946 "View.MemoryView"
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":948
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */

#line 948 "View.MemoryView"
  __pyx_t_2 = __pyx_v_memslice->shape;

#line 948 "View.MemoryView"
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":949
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 949 "View.MemoryView"
  __pyx_t_2 = __pyx_v_memslice->strides;

#line 949 "View.MemoryView"
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":953
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */

#line 953 "View.MemoryView"
  __pyx_t_3 = __Pyx_div_long(__pyx_v_ndim, 2);

#line 953 "View.MemoryView"
  __pyx_t_4 = __pyx_t_3;

#line 953 "View.MemoryView"
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {

#line 953 "View.MemoryView"
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":954
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */

#line 954 "View.MemoryView"
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":955
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */

#line 955 "View.MemoryView"
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);

#line 955 "View.MemoryView"
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);

#line 955 "View.MemoryView"
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;

#line 955 "View.MemoryView"
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":956
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */

#line 956 "View.MemoryView"
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);

#line 956 "View.MemoryView"
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);

#line 956 "View.MemoryView"
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;

#line 956 "View.MemoryView"
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":958
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */

#line 958 "View.MemoryView"
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);

#line 958 "View.MemoryView"
    if (!__pyx_t_8) {

#line 958 "View.MemoryView"
    } else {

#line 958 "View.MemoryView"
      __pyx_t_7 = __pyx_t_8;

#line 958 "View.MemoryView"
      goto __pyx_L6_bool_binop_done;

#line 958 "View.MemoryView"
    }

#line 958 "View.MemoryView"
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);

#line 958 "View.MemoryView"
    __pyx_t_7 = __pyx_t_8;

#line 958 "View.MemoryView"
    __pyx_L6_bool_binop_done:;

#line 958 "View.MemoryView"
    if (__pyx_t_7) {

      /* "View.MemoryView":959
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */

#line 959 "View.MemoryView"
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 959, __pyx_L1_error)

      /* "View.MemoryView":958
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */

#line 958 "View.MemoryView"
    }

#line 958 "View.MemoryView"
  }

  /* "View.MemoryView":961
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 961 "View.MemoryView"
  __pyx_r = 1;

#line 961 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":945
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

#line 945 "View.MemoryView"


#line 945 "View.MemoryView"
  /* function exit code */

#line 945 "View.MemoryView"
  __pyx_L1_error:;

#line 945 "View.MemoryView"
  {

#line 945 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 945 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 945 "View.MemoryView"
    #endif

#line 945 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 945 "View.MemoryView"
    #ifdef WITH_THREAD

#line 945 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 945 "View.MemoryView"
    #endif

#line 945 "View.MemoryView"
  }

#line 945 "View.MemoryView"
  __pyx_r = 0;

#line 945 "View.MemoryView"
  __pyx_L0:;

#line 945 "View.MemoryView"
  return __pyx_r;

#line 945 "View.MemoryView"
}

/* "View.MemoryView":978
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
/* Python wrapper */

#line 978 "View.MemoryView"
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/

#line 978 "View.MemoryView"
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {

#line 978 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 978 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);

#line 978 "View.MemoryView"
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
  /* function exit code */

#line 978 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 978 "View.MemoryView"
}

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {

#line 978 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 978 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":979
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */

#line 979 "View.MemoryView"
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":978
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
  /* function exit code */

#line 978 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 978 "View.MemoryView"
}

/* "View.MemoryView":981
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

#line 981 "View.MemoryView"


#line 981 "View.MemoryView"
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {

#line 981 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 981 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 981 "View.MemoryView"
  int __pyx_t_1;

#line 981 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 981 "View.MemoryView"
  int __pyx_lineno = 0;

#line 981 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 981 "View.MemoryView"
  int __pyx_clineno = 0;

#line 981 "View.MemoryView"
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":982
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */

#line 982 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);

#line 982 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":983
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */

#line 983 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 983 "View.MemoryView"
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 983, __pyx_L1_error)

#line 983 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 983 "View.MemoryView"
    __pyx_r = __pyx_t_2;

#line 983 "View.MemoryView"
    __pyx_t_2 = 0;

#line 983 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":982
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */

#line 982 "View.MemoryView"
  }

  /* "View.MemoryView":985
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */

#line 985 "View.MemoryView"
  /*else*/ {

#line 985 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 985 "View.MemoryView"
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 985, __pyx_L1_error)

#line 985 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 985 "View.MemoryView"
    __pyx_r = __pyx_t_2;

#line 985 "View.MemoryView"
    __pyx_t_2 = 0;

#line 985 "View.MemoryView"
    goto __pyx_L0;

#line 985 "View.MemoryView"
  }

  /* "View.MemoryView":981
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

#line 981 "View.MemoryView"


#line 981 "View.MemoryView"
  /* function exit code */

#line 981 "View.MemoryView"
  __pyx_L1_error:;

#line 981 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 981 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 981 "View.MemoryView"
  __pyx_r = 0;

#line 981 "View.MemoryView"
  __pyx_L0:;

#line 981 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 981 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 981 "View.MemoryView"
  return __pyx_r;

#line 981 "View.MemoryView"
}

/* "View.MemoryView":987
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

#line 987 "View.MemoryView"


#line 987 "View.MemoryView"
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {

#line 987 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 987 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 987 "View.MemoryView"
  int __pyx_t_1;

#line 987 "View.MemoryView"
  int __pyx_t_2;

#line 987 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 987 "View.MemoryView"
  int __pyx_lineno = 0;

#line 987 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 987 "View.MemoryView"
  int __pyx_clineno = 0;

#line 987 "View.MemoryView"
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":988
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */

#line 988 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);

#line 988 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":989
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */

#line 989 "View.MemoryView"
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 989, __pyx_L1_error)

    /* "View.MemoryView":988
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */

#line 988 "View.MemoryView"
    goto __pyx_L3;

#line 988 "View.MemoryView"
  }

  /* "View.MemoryView":991
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 991 "View.MemoryView"
  /*else*/ {

#line 991 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 991, __pyx_L1_error)

#line 991 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 991 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 991 "View.MemoryView"
  }

#line 991 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":987
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

#line 987 "View.MemoryView"


#line 987 "View.MemoryView"
  /* function exit code */

#line 987 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 987 "View.MemoryView"
  goto __pyx_L0;

#line 987 "View.MemoryView"
  __pyx_L1_error:;

#line 987 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 987 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 987 "View.MemoryView"
  __pyx_r = 0;

#line 987 "View.MemoryView"
  __pyx_L0:;

#line 987 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 987 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 987 "View.MemoryView"
  return __pyx_r;

#line 987 "View.MemoryView"
}

/* "View.MemoryView":994
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
/* Python wrapper */

#line 994 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 994 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {

#line 994 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 994 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 994 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 994 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
  /* function exit code */

#line 994 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 994 "View.MemoryView"
  return __pyx_r;

#line 994 "View.MemoryView"
}

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {

#line 994 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 994 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 994 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":995
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */

#line 995 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 995 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->from_object);

#line 995 "View.MemoryView"
  __pyx_r = __pyx_v_self->from_object;

#line 995 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":994
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
  /* function exit code */

#line 994 "View.MemoryView"
  __pyx_L0:;

#line 994 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 994 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 994 "View.MemoryView"
  return __pyx_r;

#line 994 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 2 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 2 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 2 "(tree fragment)"
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
/* Python wrapper */

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 3 "(tree fragment)"
__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 3 "(tree fragment)"
  int __pyx_lineno = 0;

#line 3 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 3 "(tree fragment)"
  int __pyx_clineno = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __pyx_L1_error:;

#line 3 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "(tree fragment)"
  __pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

/* "View.MemoryView":1001
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

#line 1001 "View.MemoryView"


#line 1001 "View.MemoryView"
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0
#line 1001 "View.MemoryView"
;
  Py_ssize_t __pyx_v_suboffset
#line 1001 "View.MemoryView"
;
  PyObject *__pyx_v_length = NULL
#line 1001 "View.MemoryView"
;

#line 1001 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 1001 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1001 "View.MemoryView"
  int __pyx_t_1;

#line 1001 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1001 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1001 "View.MemoryView"
  __Pyx_TypeInfo *__pyx_t_4;

#line 1001 "View.MemoryView"
  Py_buffer __pyx_t_5;

#line 1001 "View.MemoryView"
  Py_ssize_t *__pyx_t_6;

#line 1001 "View.MemoryView"
  Py_ssize_t *__pyx_t_7;

#line 1001 "View.MemoryView"
  Py_ssize_t *__pyx_t_8;

#line 1001 "View.MemoryView"
  Py_ssize_t __pyx_t_9;

#line 1001 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1001 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1001 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1001 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1009
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

#line 1009 "View.MemoryView"
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);

#line 1009 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1010
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1010 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 1010 "View.MemoryView"
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 1010 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1009
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

#line 1009 "View.MemoryView"
  }

  /* "View.MemoryView":1015
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */

#line 1015 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1015, __pyx_L1_error)

#line 1015 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1015 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1015, __pyx_L1_error)

#line 1015 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1015 "View.MemoryView"
  __Pyx_INCREF(Py_None);

#line 1015 "View.MemoryView"
  __Pyx_GIVEREF(Py_None);

#line 1015 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);

#line 1015 "View.MemoryView"
  __Pyx_INCREF(__pyx_int_0);

#line 1015 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_int_0);

#line 1015 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);

#line 1015 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 1015 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 1015 "View.MemoryView"
  __pyx_t_2 = 0;

#line 1015 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1015, __pyx_L1_error)

#line 1015 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1015 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1015 "View.MemoryView"
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);

#line 1015 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":1017
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */

#line 1017 "View.MemoryView"
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1018
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */

#line 1018 "View.MemoryView"
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1020
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */

#line 1020 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1020, __pyx_L1_error)

#line 1020 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1020 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 1020 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_result->from_object);

#line 1020 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_result->from_object);

#line 1020 "View.MemoryView"
  __pyx_v_result->from_object = __pyx_t_2;

#line 1020 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":1021
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */

#line 1021 "View.MemoryView"
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;

#line 1021 "View.MemoryView"
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1023
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */

#line 1023 "View.MemoryView"
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;

#line 1023 "View.MemoryView"
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1024
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */

#line 1024 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1025
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */

#line 1025 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1026
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */

#line 1026 "View.MemoryView"
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1027
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */

#line 1027 "View.MemoryView"
  Py_INCREF(Py_None);

  /* "View.MemoryView":1029
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */

#line 1029 "View.MemoryView"
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);

#line 1029 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1030
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */

#line 1030 "View.MemoryView"
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1029
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */

#line 1029 "View.MemoryView"
    goto __pyx_L4;

#line 1029 "View.MemoryView"
  }

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */

#line 1032 "View.MemoryView"
  /*else*/ {

#line 1032 "View.MemoryView"
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;

#line 1032 "View.MemoryView"
  }

#line 1032 "View.MemoryView"
  __pyx_L4:;

  /* "View.MemoryView":1034
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */

#line 1034 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1035
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1035 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1038
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

#line 1038 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1039
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */

#line 1039 "View.MemoryView"
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);

#line 1039 "View.MemoryView"
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {

#line 1039 "View.MemoryView"
    __pyx_t_6 = __pyx_t_8;

#line 1039 "View.MemoryView"
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1040
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */

#line 1040 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);

#line 1040 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":1041
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */

#line 1041 "View.MemoryView"
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1042
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */

#line 1042 "View.MemoryView"
      goto __pyx_L6_break;

      /* "View.MemoryView":1040
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */

#line 1040 "View.MemoryView"
    }

#line 1040 "View.MemoryView"
  }

#line 1040 "View.MemoryView"
  __pyx_L6_break:;

  /* "View.MemoryView":1044
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */

#line 1044 "View.MemoryView"
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;

#line 1044 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1045
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */

#line 1045 "View.MemoryView"
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);

#line 1045 "View.MemoryView"
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {

#line 1045 "View.MemoryView"
    __pyx_t_6 = __pyx_t_8;

#line 1045 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1045, __pyx_L1_error)

#line 1045 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 1045 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);

#line 1045 "View.MemoryView"
    __pyx_t_2 = 0;

    /* "View.MemoryView":1046
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */

#line 1046 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1046, __pyx_L1_error)

#line 1046 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 1046 "View.MemoryView"
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1046, __pyx_L1_error)

#line 1046 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 1046 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1046 "View.MemoryView"
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 1046, __pyx_L1_error)

#line 1046 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1046 "View.MemoryView"
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

#line 1046 "View.MemoryView"
  }

  /* "View.MemoryView":1048
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */

#line 1048 "View.MemoryView"
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1049
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */

#line 1049 "View.MemoryView"
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1051
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */

#line 1051 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 1051 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 1051 "View.MemoryView"
  __pyx_r = ((PyObject *)__pyx_v_result);

#line 1051 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1001
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

#line 1001 "View.MemoryView"


#line 1001 "View.MemoryView"
  /* function exit code */

#line 1001 "View.MemoryView"
  __pyx_L1_error:;

#line 1001 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1001 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1001 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1001 "View.MemoryView"
  __pyx_r = 0;

#line 1001 "View.MemoryView"
  __pyx_L0:;

#line 1001 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 1001 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_length);

#line 1001 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 1001 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1001 "View.MemoryView"
  return __pyx_r;

#line 1001 "View.MemoryView"
}

/* "View.MemoryView":1054
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

#line 1054 "View.MemoryView"


#line 1054 "View.MemoryView"
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0
#line 1054 "View.MemoryView"
;

#line 1054 "View.MemoryView"
  __Pyx_memviewslice *__pyx_r;

#line 1054 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1054 "View.MemoryView"
  int __pyx_t_1;

#line 1054 "View.MemoryView"
  int __pyx_t_2;

#line 1054 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1054 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1054 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1054 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1054 "View.MemoryView"
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1057
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */

#line 1057 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 1057 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 1057 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1058
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */

#line 1058 "View.MemoryView"
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(2, 1058, __pyx_L1_error)

#line 1058 "View.MemoryView"
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);

#line 1058 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 1058 "View.MemoryView"
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);

#line 1058 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":1059
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */

#line 1059 "View.MemoryView"
    __pyx_r = (&__pyx_v_obj->from_slice);

#line 1059 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1057
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */

#line 1057 "View.MemoryView"
  }

  /* "View.MemoryView":1061
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */

#line 1061 "View.MemoryView"
  /*else*/ {

#line 1061 "View.MemoryView"
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1062
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */

#line 1062 "View.MemoryView"
    __pyx_r = __pyx_v_mslice;

#line 1062 "View.MemoryView"
    goto __pyx_L0;

#line 1062 "View.MemoryView"
  }

  /* "View.MemoryView":1054
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

#line 1054 "View.MemoryView"


#line 1054 "View.MemoryView"
  /* function exit code */

#line 1054 "View.MemoryView"
  __pyx_L1_error:;

#line 1054 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1054 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1054 "View.MemoryView"
  __pyx_r = NULL;

#line 1054 "View.MemoryView"
  __pyx_L0:;

#line 1054 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);

#line 1054 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1054 "View.MemoryView"
  return __pyx_r;

#line 1054 "View.MemoryView"
}

/* "View.MemoryView":1065
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

#line 1065 "View.MemoryView"


#line 1065 "View.MemoryView"
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim
#line 1065 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_shape
#line 1065 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_strides
#line 1065 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_suboffsets
#line 1065 "View.MemoryView"
;

#line 1065 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1065 "View.MemoryView"
  Py_ssize_t *__pyx_t_1;

#line 1065 "View.MemoryView"
  int __pyx_t_2;

#line 1065 "View.MemoryView"
  int __pyx_t_3;

#line 1065 "View.MemoryView"
  int __pyx_t_4;

#line 1065 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 1065 "View.MemoryView"
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1069
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */

#line 1069 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memview->view.shape;

#line 1069 "View.MemoryView"
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1070
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */

#line 1070 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memview->view.strides;

#line 1070 "View.MemoryView"
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1071
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */

#line 1071 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;

#line 1071 "View.MemoryView"
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1073
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */

#line 1073 "View.MemoryView"
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1074
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */

#line 1074 "View.MemoryView"
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1076
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */

#line 1076 "View.MemoryView"
  __pyx_t_2 = __pyx_v_memview->view.ndim;

#line 1076 "View.MemoryView"
  __pyx_t_3 = __pyx_t_2;

#line 1076 "View.MemoryView"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1076 "View.MemoryView"
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1077
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */

#line 1077 "View.MemoryView"
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1078
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */

#line 1078 "View.MemoryView"
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1079
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */

#line 1079 "View.MemoryView"
    if ((__pyx_v_suboffsets != 0)) {

#line 1079 "View.MemoryView"
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);

#line 1079 "View.MemoryView"
    } else {

#line 1079 "View.MemoryView"
      __pyx_t_5 = -1L;

#line 1079 "View.MemoryView"
    }

#line 1079 "View.MemoryView"
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;

#line 1079 "View.MemoryView"
  }

  /* "View.MemoryView":1065
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

#line 1065 "View.MemoryView"


#line 1065 "View.MemoryView"
  /* function exit code */

#line 1065 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1065 "View.MemoryView"
}

/* "View.MemoryView":1082
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

#line 1082 "View.MemoryView"


#line 1082 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice
#line 1082 "View.MemoryView"
;

#line 1082 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 1082 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1082 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 1082 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1082 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1082 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1082 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1085
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */

#line 1085 "View.MemoryView"
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1086
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */

#line 1086 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 1086 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1086, __pyx_L1_error)

#line 1086 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 1086 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 1086 "View.MemoryView"
  __pyx_t_1 = 0;

#line 1086 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1082
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

#line 1082 "View.MemoryView"


#line 1082 "View.MemoryView"
  /* function exit code */

#line 1082 "View.MemoryView"
  __pyx_L1_error:;

#line 1082 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 1082 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1082 "View.MemoryView"
  __pyx_r = 0;

#line 1082 "View.MemoryView"
  __pyx_L0:;

#line 1082 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 1082 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1082 "View.MemoryView"
  return __pyx_r;

#line 1082 "View.MemoryView"
}

/* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

#line 1089 "View.MemoryView"


#line 1089 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *)
#line 1089 "View.MemoryView"
;
  int (*__pyx_v_to_dtype_func)(char *, PyObject *)
#line 1089 "View.MemoryView"
;

#line 1089 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 1089 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1089 "View.MemoryView"
  int __pyx_t_1;

#line 1089 "View.MemoryView"
  int __pyx_t_2;

#line 1089 "View.MemoryView"
  PyObject *(*__pyx_t_3)(char *);

#line 1089 "View.MemoryView"
  int (*__pyx_t_4)(char *, PyObject *);

#line 1089 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 1089 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1089 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1089 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1089 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1096
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */

#line 1096 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 1096 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 1096 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1097
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */

#line 1097 "View.MemoryView"
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;

#line 1097 "View.MemoryView"
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1098
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */

#line 1098 "View.MemoryView"
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;

#line 1098 "View.MemoryView"
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1096
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */

#line 1096 "View.MemoryView"
    goto __pyx_L3;

#line 1096 "View.MemoryView"
  }

  /* "View.MemoryView":1100
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */

#line 1100 "View.MemoryView"
  /*else*/ {

#line 1100 "View.MemoryView"
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1101
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */

#line 1101 "View.MemoryView"
    __pyx_v_to_dtype_func = NULL;

#line 1101 "View.MemoryView"
  }

#line 1101 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1103
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */

#line 1103 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1105
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1105 "View.MemoryView"
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 1103, __pyx_L1_error)

#line 1105 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_5);

#line 1105 "View.MemoryView"
  __pyx_r = __pyx_t_5;

#line 1105 "View.MemoryView"
  __pyx_t_5 = 0;

#line 1105 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

#line 1089 "View.MemoryView"


#line 1089 "View.MemoryView"
  /* function exit code */

#line 1089 "View.MemoryView"
  __pyx_L1_error:;

#line 1089 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 1089 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1089 "View.MemoryView"
  __pyx_r = 0;

#line 1089 "View.MemoryView"
  __pyx_L0:;

#line 1089 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 1089 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1089 "View.MemoryView"
  return __pyx_r;

#line 1089 "View.MemoryView"
}

/* "View.MemoryView":1111
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

#line 1111 "View.MemoryView"


#line 1111 "View.MemoryView"
static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {

#line 1111 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 1111 "View.MemoryView"
  int __pyx_t_1;

  /* "View.MemoryView":1112
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */

#line 1112 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);

#line 1112 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1113
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */

#line 1113 "View.MemoryView"
    __pyx_r = (-__pyx_v_arg);

#line 1113 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1112
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */

#line 1112 "View.MemoryView"
  }

  /* "View.MemoryView":1115
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */

#line 1115 "View.MemoryView"
  /*else*/ {

#line 1115 "View.MemoryView"
    __pyx_r = __pyx_v_arg;

#line 1115 "View.MemoryView"
    goto __pyx_L0;

#line 1115 "View.MemoryView"
  }

  /* "View.MemoryView":1111
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

#line 1111 "View.MemoryView"


#line 1111 "View.MemoryView"
  /* function exit code */

#line 1111 "View.MemoryView"
  __pyx_L0:;

#line 1111 "View.MemoryView"
  return __pyx_r;

#line 1111 "View.MemoryView"
}

/* "View.MemoryView":1118
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

#line 1118 "View.MemoryView"


#line 1118 "View.MemoryView"
static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i
#line 1118 "View.MemoryView"
;
  Py_ssize_t __pyx_v_c_stride
#line 1118 "View.MemoryView"
;
  Py_ssize_t __pyx_v_f_stride
#line 1118 "View.MemoryView"
;

#line 1118 "View.MemoryView"
  char __pyx_r;

#line 1118 "View.MemoryView"
  int __pyx_t_1;

#line 1118 "View.MemoryView"
  int __pyx_t_2;

#line 1118 "View.MemoryView"
  int __pyx_t_3;

#line 1118 "View.MemoryView"
  int __pyx_t_4;

  /* "View.MemoryView":1123
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */

#line 1123 "View.MemoryView"
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1124
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */

#line 1124 "View.MemoryView"
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1126
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */

#line 1126 "View.MemoryView"
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {

#line 1126 "View.MemoryView"
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1127
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */

#line 1127 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);

#line 1127 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1128
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */

#line 1128 "View.MemoryView"
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1129
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */

#line 1129 "View.MemoryView"
      goto __pyx_L4_break;

      /* "View.MemoryView":1127
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */

#line 1127 "View.MemoryView"
    }

#line 1127 "View.MemoryView"
  }

#line 1127 "View.MemoryView"
  __pyx_L4_break:;

  /* "View.MemoryView":1131
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */

#line 1131 "View.MemoryView"
  __pyx_t_1 = __pyx_v_ndim;

#line 1131 "View.MemoryView"
  __pyx_t_3 = __pyx_t_1;

#line 1131 "View.MemoryView"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1131 "View.MemoryView"
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1132
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */

#line 1132 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);

#line 1132 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1133
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */

#line 1133 "View.MemoryView"
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1134
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */

#line 1134 "View.MemoryView"
      goto __pyx_L7_break;

      /* "View.MemoryView":1132
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */

#line 1132 "View.MemoryView"
    }

#line 1132 "View.MemoryView"
  }

#line 1132 "View.MemoryView"
  __pyx_L7_break:;

  /* "View.MemoryView":1136
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */

#line 1136 "View.MemoryView"
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);

#line 1136 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1137
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */

#line 1137 "View.MemoryView"
    __pyx_r = 'C';

#line 1137 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1136
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */

#line 1136 "View.MemoryView"
  }

  /* "View.MemoryView":1139
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */

#line 1139 "View.MemoryView"
  /*else*/ {

#line 1139 "View.MemoryView"
    __pyx_r = 'F';

#line 1139 "View.MemoryView"
    goto __pyx_L0;

#line 1139 "View.MemoryView"
  }

  /* "View.MemoryView":1118
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

#line 1118 "View.MemoryView"


#line 1118 "View.MemoryView"
  /* function exit code */

#line 1118 "View.MemoryView"
  __pyx_L0:;

#line 1118 "View.MemoryView"
  return __pyx_r;

#line 1118 "View.MemoryView"
}

/* "View.MemoryView":1142
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

#line 1142 "View.MemoryView"


#line 1142 "View.MemoryView"
static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i
#line 1142 "View.MemoryView"
;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent
#line 1142 "View.MemoryView"
;
  Py_ssize_t __pyx_v_dst_extent
#line 1142 "View.MemoryView"
;
  Py_ssize_t __pyx_v_src_stride
#line 1142 "View.MemoryView"
;
  Py_ssize_t __pyx_v_dst_stride
#line 1142 "View.MemoryView"
;

#line 1142 "View.MemoryView"
  int __pyx_t_1;

#line 1142 "View.MemoryView"
  int __pyx_t_2;

#line 1142 "View.MemoryView"
  int __pyx_t_3;

#line 1142 "View.MemoryView"
  Py_ssize_t __pyx_t_4;

#line 1142 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 1142 "View.MemoryView"
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1149
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */

#line 1149 "View.MemoryView"
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */

#line 1150 "View.MemoryView"
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1151
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */

#line 1151 "View.MemoryView"
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */

#line 1152 "View.MemoryView"
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1154
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */

#line 1154 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);

#line 1154 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */

#line 1155 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);

#line 1155 "View.MemoryView"
    if (__pyx_t_2) {

#line 1155 "View.MemoryView"
    } else {

#line 1155 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 1155 "View.MemoryView"
      goto __pyx_L5_bool_binop_done;

#line 1155 "View.MemoryView"
    }

#line 1155 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);

#line 1155 "View.MemoryView"
    if (__pyx_t_2) {

#line 1155 "View.MemoryView"
    } else {

#line 1155 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 1155 "View.MemoryView"
      goto __pyx_L5_bool_binop_done;

#line 1155 "View.MemoryView"
    }

    /* "View.MemoryView":1156
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */

#line 1156 "View.MemoryView"
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);

#line 1156 "View.MemoryView"
    if (__pyx_t_2) {

#line 1156 "View.MemoryView"
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));

#line 1156 "View.MemoryView"
    }

#line 1156 "View.MemoryView"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 1156 "View.MemoryView"
    __pyx_t_1 = __pyx_t_3;

#line 1156 "View.MemoryView"
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */

#line 1155 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":1157
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */

#line 1157 "View.MemoryView"
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */

#line 1155 "View.MemoryView"
      goto __pyx_L4;

#line 1155 "View.MemoryView"
    }

    /* "View.MemoryView":1159
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */

#line 1159 "View.MemoryView"
    /*else*/ {

#line 1159 "View.MemoryView"
      __pyx_t_4 = __pyx_v_dst_extent;

#line 1159 "View.MemoryView"
      __pyx_t_5 = __pyx_t_4;

#line 1159 "View.MemoryView"
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1159 "View.MemoryView"
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1160
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */

#line 1160 "View.MemoryView"
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1161
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */

#line 1161 "View.MemoryView"
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1162
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */

#line 1162 "View.MemoryView"
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);

#line 1162 "View.MemoryView"
      }

#line 1162 "View.MemoryView"
    }

#line 1162 "View.MemoryView"
    __pyx_L4:;

    /* "View.MemoryView":1154
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */

#line 1154 "View.MemoryView"
    goto __pyx_L3;

#line 1154 "View.MemoryView"
  }

  /* "View.MemoryView":1164
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */

#line 1164 "View.MemoryView"
  /*else*/ {

#line 1164 "View.MemoryView"
    __pyx_t_4 = __pyx_v_dst_extent;

#line 1164 "View.MemoryView"
    __pyx_t_5 = __pyx_t_4;

#line 1164 "View.MemoryView"
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1164 "View.MemoryView"
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1165
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */

#line 1165 "View.MemoryView"
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1169
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */

#line 1169 "View.MemoryView"
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1170
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */

#line 1170 "View.MemoryView"
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);

#line 1170 "View.MemoryView"
    }

#line 1170 "View.MemoryView"
  }

#line 1170 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1142
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

#line 1142 "View.MemoryView"


#line 1142 "View.MemoryView"
  /* function exit code */

#line 1142 "View.MemoryView"
}

/* "View.MemoryView":1172
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

#line 1172 "View.MemoryView"


#line 1172 "View.MemoryView"
static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1175
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */

#line 1175 "View.MemoryView"
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1172
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

#line 1172 "View.MemoryView"


#line 1172 "View.MemoryView"
  /* function exit code */

#line 1172 "View.MemoryView"
}

/* "View.MemoryView":1179
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

#line 1179 "View.MemoryView"


#line 1179 "View.MemoryView"
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape
#line 1179 "View.MemoryView"
;
  Py_ssize_t __pyx_v_size
#line 1179 "View.MemoryView"
;

#line 1179 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 1179 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1179 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 1179 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 1179 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1181
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */

#line 1181 "View.MemoryView"
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;

#line 1181 "View.MemoryView"
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1183
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */

#line 1183 "View.MemoryView"
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);

#line 1183 "View.MemoryView"
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {

#line 1183 "View.MemoryView"
    __pyx_t_2 = __pyx_t_4;

#line 1183 "View.MemoryView"
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1184
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */

#line 1184 "View.MemoryView"
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);

#line 1184 "View.MemoryView"
  }

  /* "View.MemoryView":1186
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */

#line 1186 "View.MemoryView"
  __pyx_r = __pyx_v_size;

#line 1186 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1179
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

#line 1179 "View.MemoryView"


#line 1179 "View.MemoryView"
  /* function exit code */

#line 1179 "View.MemoryView"
  __pyx_L0:;

#line 1179 "View.MemoryView"
  return __pyx_r;

#line 1179 "View.MemoryView"
}

/* "View.MemoryView":1189
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

#line 1189 "View.MemoryView"


#line 1189 "View.MemoryView"
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx
#line 1189 "View.MemoryView"
;

#line 1189 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 1189 "View.MemoryView"
  int __pyx_t_1;

#line 1189 "View.MemoryView"
  int __pyx_t_2;

#line 1189 "View.MemoryView"
  int __pyx_t_3;

#line 1189 "View.MemoryView"
  int __pyx_t_4;

  /* "View.MemoryView":1198
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */

#line 1198 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);

#line 1198 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1199
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */

#line 1199 "View.MemoryView"
    __pyx_t_2 = __pyx_v_ndim;

#line 1199 "View.MemoryView"
    __pyx_t_3 = __pyx_t_2;

#line 1199 "View.MemoryView"
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1199 "View.MemoryView"
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1200
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */

#line 1200 "View.MemoryView"
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1201
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */

#line 1201 "View.MemoryView"
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));

#line 1201 "View.MemoryView"
    }

    /* "View.MemoryView":1198
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */

#line 1198 "View.MemoryView"
    goto __pyx_L3;

#line 1198 "View.MemoryView"
  }

  /* "View.MemoryView":1203
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */

#line 1203 "View.MemoryView"
  /*else*/ {

#line 1203 "View.MemoryView"
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {

#line 1203 "View.MemoryView"
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1204
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */

#line 1204 "View.MemoryView"
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1205
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */

#line 1205 "View.MemoryView"
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));

#line 1205 "View.MemoryView"
    }

#line 1205 "View.MemoryView"
  }

#line 1205 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1207
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */

#line 1207 "View.MemoryView"
  __pyx_r = __pyx_v_stride;

#line 1207 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1189
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

#line 1189 "View.MemoryView"


#line 1189 "View.MemoryView"
  /* function exit code */

#line 1189 "View.MemoryView"
  __pyx_L0:;

#line 1189 "View.MemoryView"
  return __pyx_r;

#line 1189 "View.MemoryView"
}

/* "View.MemoryView":1210
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

#line 1210 "View.MemoryView"


#line 1210 "View.MemoryView"
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i
#line 1210 "View.MemoryView"
;
  void *__pyx_v_result
#line 1210 "View.MemoryView"
;
  size_t __pyx_v_itemsize
#line 1210 "View.MemoryView"
;
  size_t __pyx_v_size
#line 1210 "View.MemoryView"
;

#line 1210 "View.MemoryView"
  void *__pyx_r;

#line 1210 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1210 "View.MemoryView"
  int __pyx_t_2;

#line 1210 "View.MemoryView"
  int __pyx_t_3;

#line 1210 "View.MemoryView"
  struct __pyx_memoryview_obj *__pyx_t_4;

#line 1210 "View.MemoryView"
  int __pyx_t_5;

#line 1210 "View.MemoryView"
  int __pyx_t_6;

#line 1210 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1210 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1210 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":1221
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */

#line 1221 "View.MemoryView"
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;

#line 1221 "View.MemoryView"
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1222
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */

#line 1222 "View.MemoryView"
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1224
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */

#line 1224 "View.MemoryView"
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1225
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */

#line 1225 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);

#line 1225 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1226
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1226 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 1226, __pyx_L1_error)

    /* "View.MemoryView":1225
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */

#line 1225 "View.MemoryView"
  }

  /* "View.MemoryView":1229
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */

#line 1229 "View.MemoryView"
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1230
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */

#line 1230 "View.MemoryView"
  __pyx_t_4 = __pyx_v_src->memview;

#line 1230 "View.MemoryView"
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1231
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */

#line 1231 "View.MemoryView"
  __pyx_t_3 = __pyx_v_ndim;

#line 1231 "View.MemoryView"
  __pyx_t_5 = __pyx_t_3;

#line 1231 "View.MemoryView"
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1231 "View.MemoryView"
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1232
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */

#line 1232 "View.MemoryView"
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1233
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */

#line 1233 "View.MemoryView"
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;

#line 1233 "View.MemoryView"
  }

  /* "View.MemoryView":1235
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */

#line 1235 "View.MemoryView"
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1239
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */

#line 1239 "View.MemoryView"
  __pyx_t_3 = __pyx_v_ndim;

#line 1239 "View.MemoryView"
  __pyx_t_5 = __pyx_t_3;

#line 1239 "View.MemoryView"
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1239 "View.MemoryView"
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1240
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */

#line 1240 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);

#line 1240 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1241
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */

#line 1241 "View.MemoryView"
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1240
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */

#line 1240 "View.MemoryView"
    }

#line 1240 "View.MemoryView"
  }

  /* "View.MemoryView":1243
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */

#line 1243 "View.MemoryView"
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);

#line 1243 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1244
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */

#line 1244 "View.MemoryView"
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1243
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */

#line 1243 "View.MemoryView"
    goto __pyx_L9;

#line 1243 "View.MemoryView"
  }

  /* "View.MemoryView":1246
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */

#line 1246 "View.MemoryView"
  /*else*/ {

#line 1246 "View.MemoryView"
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);

#line 1246 "View.MemoryView"
  }

#line 1246 "View.MemoryView"
  __pyx_L9:;

  /* "View.MemoryView":1248
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1248 "View.MemoryView"
  __pyx_r = __pyx_v_result;

#line 1248 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1210
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

#line 1210 "View.MemoryView"


#line 1210 "View.MemoryView"
  /* function exit code */

#line 1210 "View.MemoryView"
  __pyx_L1_error:;

#line 1210 "View.MemoryView"
  {

#line 1210 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 1210 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1210 "View.MemoryView"
    #endif

#line 1210 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1210 "View.MemoryView"
    #ifdef WITH_THREAD

#line 1210 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1210 "View.MemoryView"
    #endif

#line 1210 "View.MemoryView"
  }

#line 1210 "View.MemoryView"
  __pyx_r = NULL;

#line 1210 "View.MemoryView"
  __pyx_L0:;

#line 1210 "View.MemoryView"
  return __pyx_r;

#line 1210 "View.MemoryView"
}

/* "View.MemoryView":1253
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

#line 1253 "View.MemoryView"


#line 1253 "View.MemoryView"
static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {

#line 1253 "View.MemoryView"
  int __pyx_r;

#line 1253 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 1253 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1253 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1253 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1253 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1253 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1253 "View.MemoryView"
  #endif

#line 1253 "View.MemoryView"
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1256
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */

#line 1256 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 1256 "View.MemoryView"
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1256 "View.MemoryView"
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1256 "View.MemoryView"
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 1256 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 1256 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);

#line 1256 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 1256 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);

#line 1256 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 1256 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);

#line 1256 "View.MemoryView"
  __pyx_t_1 = 0;

#line 1256 "View.MemoryView"
  __pyx_t_2 = 0;

#line 1256 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":1255
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */

#line 1255 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1255, __pyx_L1_error)

#line 1255 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1255 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1255 "View.MemoryView"
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1255, __pyx_L1_error)

#line 1255 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 1255 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1255 "View.MemoryView"
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 1255 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1255 "View.MemoryView"
  __PYX_ERR(2, 1255, __pyx_L1_error)

  /* "View.MemoryView":1253
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

#line 1253 "View.MemoryView"


#line 1253 "View.MemoryView"
  /* function exit code */

#line 1253 "View.MemoryView"
  __pyx_L1_error:;

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 1253 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1253 "View.MemoryView"
  __pyx_r = -1;

#line 1253 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1253 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1253 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1253 "View.MemoryView"
  #endif

#line 1253 "View.MemoryView"
  return __pyx_r;

#line 1253 "View.MemoryView"
}

/* "View.MemoryView":1259
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

#line 1259 "View.MemoryView"


#line 1259 "View.MemoryView"
static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {

#line 1259 "View.MemoryView"
  int __pyx_r;

#line 1259 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 1259 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1259 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1259 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1259 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1259 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1259 "View.MemoryView"
  #endif

#line 1259 "View.MemoryView"
  __Pyx_RefNannySetupContext("_err_dim", 0);

#line 1259 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1260
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */

#line 1260 "View.MemoryView"
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1260 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1260 "View.MemoryView"
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1260 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; 
#line 1260 "View.MemoryView"
__pyx_t_2 = NULL;

#line 1260 "View.MemoryView"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 1260 "View.MemoryView"
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);

#line 1260 "View.MemoryView"
    if (likely(__pyx_t_2)) {

#line 1260 "View.MemoryView"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 1260 "View.MemoryView"
      __Pyx_INCREF(__pyx_t_2);

#line 1260 "View.MemoryView"
      __Pyx_INCREF(function);

#line 1260 "View.MemoryView"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 1260 "View.MemoryView"
    }

#line 1260 "View.MemoryView"
  }

#line 1260 "View.MemoryView"
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);

#line 1260 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1260 "View.MemoryView"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1260 "View.MemoryView"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1260 "View.MemoryView"
  __PYX_ERR(2, 1260, __pyx_L1_error)

  /* "View.MemoryView":1259
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

#line 1259 "View.MemoryView"


#line 1259 "View.MemoryView"
  /* function exit code */

#line 1259 "View.MemoryView"
  __pyx_L1_error:;

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 1259 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1259 "View.MemoryView"
  __pyx_r = -1;

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_error);

#line 1259 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1259 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1259 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1259 "View.MemoryView"
  #endif

#line 1259 "View.MemoryView"
  return __pyx_r;

#line 1259 "View.MemoryView"
}

/* "View.MemoryView":1263
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

#line 1263 "View.MemoryView"


#line 1263 "View.MemoryView"
static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {

#line 1263 "View.MemoryView"
  int __pyx_r;

#line 1263 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1263 "View.MemoryView"
  int __pyx_t_1;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 1263 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1263 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1263 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1263 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1263 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1263 "View.MemoryView"
  #endif

#line 1263 "View.MemoryView"
  __Pyx_RefNannySetupContext("_err", 0);

#line 1263 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1264
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */

#line 1264 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);

#line 1264 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1265
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */

#line 1265 "View.MemoryView"
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1265, __pyx_L1_error)

#line 1265 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 1265 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; 
#line 1265 "View.MemoryView"
__pyx_t_5 = NULL;

#line 1265 "View.MemoryView"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 1265 "View.MemoryView"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 1265 "View.MemoryView"
      if (likely(__pyx_t_5)) {

#line 1265 "View.MemoryView"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 1265 "View.MemoryView"
        __Pyx_INCREF(__pyx_t_5);

#line 1265 "View.MemoryView"
        __Pyx_INCREF(function);

#line 1265 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_t_4, function);

#line 1265 "View.MemoryView"
      }

#line 1265 "View.MemoryView"
    }

#line 1265 "View.MemoryView"
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 1265 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 1265 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1265 "View.MemoryView"
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1265, __pyx_L1_error)

#line 1265 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 1265 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1265 "View.MemoryView"
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 1265 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1265 "View.MemoryView"
    __PYX_ERR(2, 1265, __pyx_L1_error)

    /* "View.MemoryView":1264
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */

#line 1264 "View.MemoryView"
  }

  /* "View.MemoryView":1267
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */

#line 1267 "View.MemoryView"
  /*else*/ {

#line 1267 "View.MemoryView"
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);

#line 1267 "View.MemoryView"
    __PYX_ERR(2, 1267, __pyx_L1_error)

#line 1267 "View.MemoryView"
  }

  /* "View.MemoryView":1263
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

#line 1263 "View.MemoryView"


#line 1263 "View.MemoryView"
  /* function exit code */

#line 1263 "View.MemoryView"
  __pyx_L1_error:;

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 1263 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1263 "View.MemoryView"
  __pyx_r = -1;

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_error);

#line 1263 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1263 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1263 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1263 "View.MemoryView"
  #endif

#line 1263 "View.MemoryView"
  return __pyx_r;

#line 1263 "View.MemoryView"
}

/* "View.MemoryView":1270
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

#line 1270 "View.MemoryView"


#line 1270 "View.MemoryView"
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata
#line 1270 "View.MemoryView"
;
  size_t __pyx_v_itemsize
#line 1270 "View.MemoryView"
;
  int __pyx_v_i
#line 1270 "View.MemoryView"
;
  char __pyx_v_order
#line 1270 "View.MemoryView"
;
  int __pyx_v_broadcasting
#line 1270 "View.MemoryView"
;
  int __pyx_v_direct_copy
#line 1270 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp
#line 1270 "View.MemoryView"
;
  int __pyx_v_ndim
#line 1270 "View.MemoryView"
;

#line 1270 "View.MemoryView"
  int __pyx_r;

#line 1270 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1270 "View.MemoryView"
  int __pyx_t_2;

#line 1270 "View.MemoryView"
  int __pyx_t_3;

#line 1270 "View.MemoryView"
  int __pyx_t_4;

#line 1270 "View.MemoryView"
  int __pyx_t_5;

#line 1270 "View.MemoryView"
  int __pyx_t_6;

#line 1270 "View.MemoryView"
  void *__pyx_t_7;

#line 1270 "View.MemoryView"
  int __pyx_t_8;

#line 1270 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1270 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1270 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":1278
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */

#line 1278 "View.MemoryView"
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1279
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */

#line 1279 "View.MemoryView"
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;

#line 1279 "View.MemoryView"
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1281
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */

#line 1281 "View.MemoryView"
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1282
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */

#line 1282 "View.MemoryView"
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1283
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */

#line 1283 "View.MemoryView"
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1286
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */

#line 1286 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);

#line 1286 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1287
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */

#line 1287 "View.MemoryView"
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1286
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */

#line 1286 "View.MemoryView"
    goto __pyx_L3;

#line 1286 "View.MemoryView"
  }

  /* "View.MemoryView":1288
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */

#line 1288 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);

#line 1288 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1289
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */

#line 1289 "View.MemoryView"
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1288
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */

#line 1288 "View.MemoryView"
  }

#line 1288 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1291
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */

#line 1291 "View.MemoryView"
  __pyx_t_3 = __pyx_v_dst_ndim;

#line 1291 "View.MemoryView"
  __pyx_t_4 = __pyx_v_src_ndim;

#line 1291 "View.MemoryView"
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {

#line 1291 "View.MemoryView"
    __pyx_t_5 = __pyx_t_3;

#line 1291 "View.MemoryView"
  } else {

#line 1291 "View.MemoryView"
    __pyx_t_5 = __pyx_t_4;

#line 1291 "View.MemoryView"
  }

#line 1291 "View.MemoryView"
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1293
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */

#line 1293 "View.MemoryView"
  __pyx_t_5 = __pyx_v_ndim;

#line 1293 "View.MemoryView"
  __pyx_t_3 = __pyx_t_5;

#line 1293 "View.MemoryView"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1293 "View.MemoryView"
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1294
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */

#line 1294 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);

#line 1294 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1295
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */

#line 1295 "View.MemoryView"
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);

#line 1295 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":1296
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */

#line 1296 "View.MemoryView"
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1297
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */

#line 1297 "View.MemoryView"
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1295
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */

#line 1295 "View.MemoryView"
        goto __pyx_L7;

#line 1295 "View.MemoryView"
      }

      /* "View.MemoryView":1299
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */

#line 1299 "View.MemoryView"
      /*else*/ {

#line 1299 "View.MemoryView"
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 1299, __pyx_L1_error)

#line 1299 "View.MemoryView"
      }

#line 1299 "View.MemoryView"
      __pyx_L7:;

      /* "View.MemoryView":1294
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */

#line 1294 "View.MemoryView"
    }

    /* "View.MemoryView":1301
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */

#line 1301 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);

#line 1301 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1302
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */

#line 1302 "View.MemoryView"
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 1302, __pyx_L1_error)

      /* "View.MemoryView":1301
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */

#line 1301 "View.MemoryView"
    }

#line 1301 "View.MemoryView"
  }

  /* "View.MemoryView":1304
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */

#line 1304 "View.MemoryView"
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);

#line 1304 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1306
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */

#line 1306 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);

#line 1306 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1307
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */

#line 1307 "View.MemoryView"
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1306
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */

#line 1306 "View.MemoryView"
    }

    /* "View.MemoryView":1309
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */

#line 1309 "View.MemoryView"
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(2, 1309, __pyx_L1_error)

#line 1309 "View.MemoryView"
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1310
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */

#line 1310 "View.MemoryView"
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1304
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */

#line 1304 "View.MemoryView"
  }

  /* "View.MemoryView":1312
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1312 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);

#line 1312 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1315
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */

#line 1315 "View.MemoryView"
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);

#line 1315 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1316
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */

#line 1316 "View.MemoryView"
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1315
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */

#line 1315 "View.MemoryView"
      goto __pyx_L12;

#line 1315 "View.MemoryView"
    }

    /* "View.MemoryView":1317
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */

#line 1317 "View.MemoryView"
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);

#line 1317 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */

#line 1318 "View.MemoryView"
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1317
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */

#line 1317 "View.MemoryView"
    }

#line 1317 "View.MemoryView"
    __pyx_L12:;

    /* "View.MemoryView":1320
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */

#line 1320 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_direct_copy != 0);

#line 1320 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1322
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */

#line 1322 "View.MemoryView"
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1323
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */

#line 1323 "View.MemoryView"
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1324
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */

#line 1324 "View.MemoryView"
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1325
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */

#line 1325 "View.MemoryView"
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1326
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */

#line 1326 "View.MemoryView"
      __pyx_r = 0;

#line 1326 "View.MemoryView"
      goto __pyx_L0;

      /* "View.MemoryView":1320
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */

#line 1320 "View.MemoryView"
    }

    /* "View.MemoryView":1312
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1312 "View.MemoryView"
  }

  /* "View.MemoryView":1328
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1328 "View.MemoryView"
  __pyx_t_2 = (__pyx_v_order == 'F');

#line 1328 "View.MemoryView"
  if (__pyx_t_2) {

#line 1328 "View.MemoryView"
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));

#line 1328 "View.MemoryView"
  }

#line 1328 "View.MemoryView"
  __pyx_t_8 = (__pyx_t_2 != 0);

#line 1328 "View.MemoryView"
  if (__pyx_t_8) {

    /* "View.MemoryView":1331
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */

#line 1331 "View.MemoryView"
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(2, 1331, __pyx_L1_error)

    /* "View.MemoryView":1332
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */

#line 1332 "View.MemoryView"
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(2, 1332, __pyx_L1_error)

    /* "View.MemoryView":1328
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1328 "View.MemoryView"
  }

  /* "View.MemoryView":1334
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */

#line 1334 "View.MemoryView"
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1335
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */

#line 1335 "View.MemoryView"
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1336
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */

#line 1336 "View.MemoryView"
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1338
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */

#line 1338 "View.MemoryView"
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1339
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */

#line 1339 "View.MemoryView"
  __pyx_r = 0;

#line 1339 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1270
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

#line 1270 "View.MemoryView"


#line 1270 "View.MemoryView"
  /* function exit code */

#line 1270 "View.MemoryView"
  __pyx_L1_error:;

#line 1270 "View.MemoryView"
  {

#line 1270 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 1270 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1270 "View.MemoryView"
    #endif

#line 1270 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1270 "View.MemoryView"
    #ifdef WITH_THREAD

#line 1270 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1270 "View.MemoryView"
    #endif

#line 1270 "View.MemoryView"
  }

#line 1270 "View.MemoryView"
  __pyx_r = -1;

#line 1270 "View.MemoryView"
  __pyx_L0:;

#line 1270 "View.MemoryView"
  return __pyx_r;

#line 1270 "View.MemoryView"
}

/* "View.MemoryView":1342
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

#line 1342 "View.MemoryView"


#line 1342 "View.MemoryView"
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i
#line 1342 "View.MemoryView"
;
  int __pyx_v_offset
#line 1342 "View.MemoryView"
;

#line 1342 "View.MemoryView"
  int __pyx_t_1;

#line 1342 "View.MemoryView"
  int __pyx_t_2;

#line 1342 "View.MemoryView"
  int __pyx_t_3;

  /* "View.MemoryView":1346
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */

#line 1346 "View.MemoryView"
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1348
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */

#line 1348 "View.MemoryView"
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {

#line 1348 "View.MemoryView"
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1349
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */

#line 1349 "View.MemoryView"
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1350
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */

#line 1350 "View.MemoryView"
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1351
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */

#line 1351 "View.MemoryView"
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);

#line 1351 "View.MemoryView"
  }

  /* "View.MemoryView":1353
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */

#line 1353 "View.MemoryView"
  __pyx_t_1 = __pyx_v_offset;

#line 1353 "View.MemoryView"
  __pyx_t_2 = __pyx_t_1;

#line 1353 "View.MemoryView"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 1353 "View.MemoryView"
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1354
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */

#line 1354 "View.MemoryView"
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1355
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */

#line 1355 "View.MemoryView"
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1356
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1356 "View.MemoryView"
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;

#line 1356 "View.MemoryView"
  }

  /* "View.MemoryView":1342
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

#line 1342 "View.MemoryView"


#line 1342 "View.MemoryView"
  /* function exit code */

#line 1342 "View.MemoryView"
}

/* "View.MemoryView":1364
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

#line 1364 "View.MemoryView"


#line 1364 "View.MemoryView"
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {

#line 1364 "View.MemoryView"
  int __pyx_t_1;

  /* "View.MemoryView":1368
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */

#line 1368 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);

#line 1368 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1369
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */

#line 1369 "View.MemoryView"
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1368
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */

#line 1368 "View.MemoryView"
  }

  /* "View.MemoryView":1364
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

#line 1364 "View.MemoryView"


#line 1364 "View.MemoryView"
  /* function exit code */

#line 1364 "View.MemoryView"
}

/* "View.MemoryView":1373
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

#line 1373 "View.MemoryView"


#line 1373 "View.MemoryView"
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {

#line 1373 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1373 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1373 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1373 "View.MemoryView"
  #endif

#line 1373 "View.MemoryView"
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1376
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */

#line 1376 "View.MemoryView"
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1373
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

#line 1373 "View.MemoryView"


#line 1373 "View.MemoryView"
  /* function exit code */

#line 1373 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1373 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1373 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1373 "View.MemoryView"
  #endif

#line 1373 "View.MemoryView"
}

/* "View.MemoryView":1379
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

#line 1379 "View.MemoryView"


#line 1379 "View.MemoryView"
static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i
#line 1379 "View.MemoryView"
;

#line 1379 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1379 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1379 "View.MemoryView"
  Py_ssize_t __pyx_t_2;

#line 1379 "View.MemoryView"
  Py_ssize_t __pyx_t_3;

#line 1379 "View.MemoryView"
  int __pyx_t_4;

#line 1379 "View.MemoryView"
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1383
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */

#line 1383 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_shape[0]);

#line 1383 "View.MemoryView"
  __pyx_t_2 = __pyx_t_1;

#line 1383 "View.MemoryView"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 1383 "View.MemoryView"
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1384
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */

#line 1384 "View.MemoryView"
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);

#line 1384 "View.MemoryView"
    if (__pyx_t_4) {

      /* "View.MemoryView":1385
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */

#line 1385 "View.MemoryView"
      __pyx_t_4 = (__pyx_v_inc != 0);

#line 1385 "View.MemoryView"
      if (__pyx_t_4) {

        /* "View.MemoryView":1386
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */

#line 1386 "View.MemoryView"
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1385
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */

#line 1385 "View.MemoryView"
        goto __pyx_L6;

#line 1385 "View.MemoryView"
      }

      /* "View.MemoryView":1388
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */

#line 1388 "View.MemoryView"
      /*else*/ {

#line 1388 "View.MemoryView"
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));

#line 1388 "View.MemoryView"
      }

#line 1388 "View.MemoryView"
      __pyx_L6:;

      /* "View.MemoryView":1384
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */

#line 1384 "View.MemoryView"
      goto __pyx_L5;

#line 1384 "View.MemoryView"
    }

    /* "View.MemoryView":1390
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */

#line 1390 "View.MemoryView"
    /*else*/ {

      /* "View.MemoryView":1391
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */

#line 1391 "View.MemoryView"
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);

#line 1391 "View.MemoryView"
    }

#line 1391 "View.MemoryView"
    __pyx_L5:;

    /* "View.MemoryView":1393
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1393 "View.MemoryView"
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));

#line 1393 "View.MemoryView"
  }

  /* "View.MemoryView":1379
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

#line 1379 "View.MemoryView"


#line 1379 "View.MemoryView"
  /* function exit code */

#line 1379 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1379 "View.MemoryView"
}

/* "View.MemoryView":1399
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

#line 1399 "View.MemoryView"


#line 1399 "View.MemoryView"
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1402
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */

#line 1402 "View.MemoryView"
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1403
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */

#line 1403 "View.MemoryView"
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1405
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1405 "View.MemoryView"
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1399
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

#line 1399 "View.MemoryView"


#line 1399 "View.MemoryView"
  /* function exit code */

#line 1399 "View.MemoryView"
}

/* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

#line 1409 "View.MemoryView"


#line 1409 "View.MemoryView"
static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i
#line 1409 "View.MemoryView"
;
  Py_ssize_t __pyx_v_stride
#line 1409 "View.MemoryView"
;
  Py_ssize_t __pyx_v_extent
#line 1409 "View.MemoryView"
;

#line 1409 "View.MemoryView"
  int __pyx_t_1;

#line 1409 "View.MemoryView"
  Py_ssize_t __pyx_t_2;

#line 1409 "View.MemoryView"
  Py_ssize_t __pyx_t_3;

#line 1409 "View.MemoryView"
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1413
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */

#line 1413 "View.MemoryView"
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1414
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */

#line 1414 "View.MemoryView"
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */

#line 1416 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);

#line 1416 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1417
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */

#line 1417 "View.MemoryView"
    __pyx_t_2 = __pyx_v_extent;

#line 1417 "View.MemoryView"
    __pyx_t_3 = __pyx_t_2;

#line 1417 "View.MemoryView"
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1417 "View.MemoryView"
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1418
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */

#line 1418 "View.MemoryView"
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1419
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */

#line 1419 "View.MemoryView"
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);

#line 1419 "View.MemoryView"
    }

    /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */

#line 1416 "View.MemoryView"
    goto __pyx_L3;

#line 1416 "View.MemoryView"
  }

  /* "View.MemoryView":1421
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */

#line 1421 "View.MemoryView"
  /*else*/ {

#line 1421 "View.MemoryView"
    __pyx_t_2 = __pyx_v_extent;

#line 1421 "View.MemoryView"
    __pyx_t_3 = __pyx_t_2;

#line 1421 "View.MemoryView"
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1421 "View.MemoryView"
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1422
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */

#line 1422 "View.MemoryView"
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1424
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1424 "View.MemoryView"
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);

#line 1424 "View.MemoryView"
    }

#line 1424 "View.MemoryView"
  }

#line 1424 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

#line 1409 "View.MemoryView"


#line 1409 "View.MemoryView"
  /* function exit code */

#line 1409 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__25, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  int __pyx_t_2;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_3;

#line 11 "(tree fragment)"
  int __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->name);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->name);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->name = __pyx_t_1;

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_4) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_2 = __pyx_t_4;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = (__pyx_t_4 != 0);

#line 13 "(tree fragment)"
  __pyx_t_2 = __pyx_t_5;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 14 "(tree fragment)"
    __pyx_t_8 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {

#line 14 "(tree fragment)"
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_8)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_7, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "BufferFormatFromTypeInfo":1465
 * 
 * @cname('__pyx_format_from_typeinfo')
 * cdef bytes format_from_typeinfo(__Pyx_TypeInfo *type):             # <<<<<<<<<<<<<<
 *     cdef __Pyx_StructField *field
 *     cdef __pyx_typeinfo_string fmt
 */

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_format_from_typeinfo(__Pyx_TypeInfo *__pyx_v_type) {
  __Pyx_StructField *__pyx_v_field
#line 1465 "BufferFormatFromTypeInfo"
;
  struct __pyx_typeinfo_string __pyx_v_fmt
#line 1465 "BufferFormatFromTypeInfo"
;
  PyObject *__pyx_v_part = 0
#line 1465 "BufferFormatFromTypeInfo"
;
  PyObject *__pyx_v_result = 0
#line 1465 "BufferFormatFromTypeInfo"
;
  PyObject *__pyx_v_alignment = NULL
#line 1465 "BufferFormatFromTypeInfo"
;
  PyObject *__pyx_v_parts = NULL
#line 1465 "BufferFormatFromTypeInfo"
;
  PyObject *__pyx_v_extents = NULL
#line 1465 "BufferFormatFromTypeInfo"
;
  int __pyx_v_i
#line 1465 "BufferFormatFromTypeInfo"
;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *__pyx_r = NULL;

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_RefNannyDeclarations

#line 1465 "BufferFormatFromTypeInfo"
  int __pyx_t_1;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *__pyx_t_2 = NULL;

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_StructField *__pyx_t_3;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *__pyx_t_4 = NULL;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *__pyx_t_5 = NULL;

#line 1465 "BufferFormatFromTypeInfo"
  int __pyx_t_6;

#line 1465 "BufferFormatFromTypeInfo"
  int __pyx_t_7;

#line 1465 "BufferFormatFromTypeInfo"
  int __pyx_t_8;

#line 1465 "BufferFormatFromTypeInfo"
  int __pyx_t_9;

#line 1465 "BufferFormatFromTypeInfo"
  int __pyx_lineno = 0;

#line 1465 "BufferFormatFromTypeInfo"
  const char *__pyx_filename = NULL;

#line 1465 "BufferFormatFromTypeInfo"
  int __pyx_clineno = 0;

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_RefNannySetupContext("format_from_typeinfo", 0);

  /* "BufferFormatFromTypeInfo":1470
 *     cdef bytes part, result
 * 
 *     if type.typegroup == 'S':             # <<<<<<<<<<<<<<
 *         assert type.fields != NULL
 *         assert type.fields.type != NULL
 */

#line 1470 "BufferFormatFromTypeInfo"
  __pyx_t_1 = ((__pyx_v_type->typegroup == 'S') != 0);

#line 1470 "BufferFormatFromTypeInfo"
  if (__pyx_t_1) {

    /* "BufferFormatFromTypeInfo":1471
 * 
 *     if type.typegroup == 'S':
 *         assert type.fields != NULL             # <<<<<<<<<<<<<<
 *         assert type.fields.type != NULL
 * 
 */

#line 1471 "BufferFormatFromTypeInfo"
    #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 1471 "BufferFormatFromTypeInfo"
    if (unlikely(!Py_OptimizeFlag)) {

#line 1471 "BufferFormatFromTypeInfo"
      if (unlikely(!((__pyx_v_type->fields != NULL) != 0))) {

#line 1471 "BufferFormatFromTypeInfo"
        PyErr_SetNone(PyExc_AssertionError);

#line 1471 "BufferFormatFromTypeInfo"
        __PYX_ERR(2, 1471, __pyx_L1_error)

#line 1471 "BufferFormatFromTypeInfo"
      }

#line 1471 "BufferFormatFromTypeInfo"
    }

#line 1471 "BufferFormatFromTypeInfo"
    #endif

    /* "BufferFormatFromTypeInfo":1472
 *     if type.typegroup == 'S':
 *         assert type.fields != NULL
 *         assert type.fields.type != NULL             # <<<<<<<<<<<<<<
 * 
 *         if type.flags & __PYX_BUF_FLAGS_PACKED_STRUCT:
 */

#line 1472 "BufferFormatFromTypeInfo"
    #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 1472 "BufferFormatFromTypeInfo"
    if (unlikely(!Py_OptimizeFlag)) {

#line 1472 "BufferFormatFromTypeInfo"
      if (unlikely(!((__pyx_v_type->fields->type != NULL) != 0))) {

#line 1472 "BufferFormatFromTypeInfo"
        PyErr_SetNone(PyExc_AssertionError);

#line 1472 "BufferFormatFromTypeInfo"
        __PYX_ERR(2, 1472, __pyx_L1_error)

#line 1472 "BufferFormatFromTypeInfo"
      }

#line 1472 "BufferFormatFromTypeInfo"
    }

#line 1472 "BufferFormatFromTypeInfo"
    #endif

    /* "BufferFormatFromTypeInfo":1474
 *         assert type.fields.type != NULL
 * 
 *         if type.flags & __PYX_BUF_FLAGS_PACKED_STRUCT:             # <<<<<<<<<<<<<<
 *             alignment = b'^'
 *         else:
 */

#line 1474 "BufferFormatFromTypeInfo"
    __pyx_t_1 = ((__pyx_v_type->flags & __PYX_BUF_FLAGS_PACKED_STRUCT) != 0);

#line 1474 "BufferFormatFromTypeInfo"
    if (__pyx_t_1) {

      /* "BufferFormatFromTypeInfo":1475
 * 
 *         if type.flags & __PYX_BUF_FLAGS_PACKED_STRUCT:
 *             alignment = b'^'             # <<<<<<<<<<<<<<
 *         else:
 *             alignment = b''
 */

#line 1475 "BufferFormatFromTypeInfo"
      __Pyx_INCREF(__pyx_kp_b__26);

#line 1475 "BufferFormatFromTypeInfo"
      __pyx_v_alignment = __pyx_kp_b__26;

      /* "BufferFormatFromTypeInfo":1474
 *         assert type.fields.type != NULL
 * 
 *         if type.flags & __PYX_BUF_FLAGS_PACKED_STRUCT:             # <<<<<<<<<<<<<<
 *             alignment = b'^'
 *         else:
 */

#line 1474 "BufferFormatFromTypeInfo"
      goto __pyx_L4;

#line 1474 "BufferFormatFromTypeInfo"
    }

    /* "BufferFormatFromTypeInfo":1477
 *             alignment = b'^'
 *         else:
 *             alignment = b''             # <<<<<<<<<<<<<<
 * 
 *         parts = [b"T{"]
 */

#line 1477 "BufferFormatFromTypeInfo"
    /*else*/ {

#line 1477 "BufferFormatFromTypeInfo"
      __Pyx_INCREF(__pyx_kp_b__27);

#line 1477 "BufferFormatFromTypeInfo"
      __pyx_v_alignment = __pyx_kp_b__27;

#line 1477 "BufferFormatFromTypeInfo"
    }

#line 1477 "BufferFormatFromTypeInfo"
    __pyx_L4:;

    /* "BufferFormatFromTypeInfo":1479
 *             alignment = b''
 * 
 *         parts = [b"T{"]             # <<<<<<<<<<<<<<
 *         field = type.fields
 * 
 */

#line 1479 "BufferFormatFromTypeInfo"
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1479, __pyx_L1_error)

#line 1479 "BufferFormatFromTypeInfo"
    __Pyx_GOTREF(__pyx_t_2);

#line 1479 "BufferFormatFromTypeInfo"
    __Pyx_INCREF(__pyx_kp_b_T);

#line 1479 "BufferFormatFromTypeInfo"
    __Pyx_GIVEREF(__pyx_kp_b_T);

#line 1479 "BufferFormatFromTypeInfo"
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_kp_b_T);

#line 1479 "BufferFormatFromTypeInfo"
    __pyx_v_parts = ((PyObject*)__pyx_t_2);

#line 1479 "BufferFormatFromTypeInfo"
    __pyx_t_2 = 0;

    /* "BufferFormatFromTypeInfo":1480
 * 
 *         parts = [b"T{"]
 *         field = type.fields             # <<<<<<<<<<<<<<
 * 
 *         while field.type:
 */

#line 1480 "BufferFormatFromTypeInfo"
    __pyx_t_3 = __pyx_v_type->fields;

#line 1480 "BufferFormatFromTypeInfo"
    __pyx_v_field = __pyx_t_3;

    /* "BufferFormatFromTypeInfo":1482
 *         field = type.fields
 * 
 *         while field.type:             # <<<<<<<<<<<<<<
 *             part = format_from_typeinfo(field.type)
 *             parts.append(part + b':' + field.name + b':')
 */

#line 1482 "BufferFormatFromTypeInfo"
    while (1) {

#line 1482 "BufferFormatFromTypeInfo"
      __pyx_t_1 = (__pyx_v_field->type != 0);

#line 1482 "BufferFormatFromTypeInfo"
      if (!__pyx_t_1) break;

      /* "BufferFormatFromTypeInfo":1483
 * 
 *         while field.type:
 *             part = format_from_typeinfo(field.type)             # <<<<<<<<<<<<<<
 *             parts.append(part + b':' + field.name + b':')
 *             field += 1
 */

#line 1483 "BufferFormatFromTypeInfo"
      __pyx_t_2 = __pyx_format_from_typeinfo(__pyx_v_field->type); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1483, __pyx_L1_error)

#line 1483 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_2);

#line 1483 "BufferFormatFromTypeInfo"
      __Pyx_XDECREF_SET(__pyx_v_part, ((PyObject*)__pyx_t_2));

#line 1483 "BufferFormatFromTypeInfo"
      __pyx_t_2 = 0;

      /* "BufferFormatFromTypeInfo":1484
 *         while field.type:
 *             part = format_from_typeinfo(field.type)
 *             parts.append(part + b':' + field.name + b':')             # <<<<<<<<<<<<<<
 *             field += 1
 * 
 */

#line 1484 "BufferFormatFromTypeInfo"
      __pyx_t_2 = PyNumber_Add(__pyx_v_part, __pyx_kp_b__28); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1484, __pyx_L1_error)

#line 1484 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_2);

#line 1484 "BufferFormatFromTypeInfo"
      __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_field->name); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1484, __pyx_L1_error)

#line 1484 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_4);

#line 1484 "BufferFormatFromTypeInfo"
      __pyx_t_5 = PyNumber_Add(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 1484, __pyx_L1_error)

#line 1484 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_5);

#line 1484 "BufferFormatFromTypeInfo"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1484 "BufferFormatFromTypeInfo"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1484 "BufferFormatFromTypeInfo"
      __pyx_t_4 = PyNumber_Add(__pyx_t_5, __pyx_kp_b__28); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1484, __pyx_L1_error)

#line 1484 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_4);

#line 1484 "BufferFormatFromTypeInfo"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 1484 "BufferFormatFromTypeInfo"
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_parts, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 1484, __pyx_L1_error)

#line 1484 "BufferFormatFromTypeInfo"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "BufferFormatFromTypeInfo":1485
 *             part = format_from_typeinfo(field.type)
 *             parts.append(part + b':' + field.name + b':')
 *             field += 1             # <<<<<<<<<<<<<<
 * 
 *         result = alignment.join(parts) + b'}'
 */

#line 1485 "BufferFormatFromTypeInfo"
      __pyx_v_field = (__pyx_v_field + 1);

#line 1485 "BufferFormatFromTypeInfo"
    }

    /* "BufferFormatFromTypeInfo":1487
 *             field += 1
 * 
 *         result = alignment.join(parts) + b'}'             # <<<<<<<<<<<<<<
 *     else:
 *         fmt = __Pyx_TypeInfoToFormat(type)
 */

#line 1487 "BufferFormatFromTypeInfo"
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_v_alignment, __pyx_v_parts); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1487, __pyx_L1_error)

#line 1487 "BufferFormatFromTypeInfo"
    __Pyx_GOTREF(__pyx_t_4);

#line 1487 "BufferFormatFromTypeInfo"
    __pyx_t_5 = PyNumber_Add(__pyx_t_4, __pyx_kp_b__29); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 1487, __pyx_L1_error)

#line 1487 "BufferFormatFromTypeInfo"
    __Pyx_GOTREF(__pyx_t_5);

#line 1487 "BufferFormatFromTypeInfo"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1487 "BufferFormatFromTypeInfo"
    if (!(likely(PyBytes_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(2, 1487, __pyx_L1_error)

#line 1487 "BufferFormatFromTypeInfo"
    __pyx_v_result = ((PyObject*)__pyx_t_5);

#line 1487 "BufferFormatFromTypeInfo"
    __pyx_t_5 = 0;

    /* "BufferFormatFromTypeInfo":1470
 *     cdef bytes part, result
 * 
 *     if type.typegroup == 'S':             # <<<<<<<<<<<<<<
 *         assert type.fields != NULL
 *         assert type.fields.type != NULL
 */

#line 1470 "BufferFormatFromTypeInfo"
    goto __pyx_L3;

#line 1470 "BufferFormatFromTypeInfo"
  }

  /* "BufferFormatFromTypeInfo":1489
 *         result = alignment.join(parts) + b'}'
 *     else:
 *         fmt = __Pyx_TypeInfoToFormat(type)             # <<<<<<<<<<<<<<
 *         if type.arraysize[0]:
 *             extents = [unicode(type.arraysize[i]) for i in range(type.ndim)]
 */

#line 1489 "BufferFormatFromTypeInfo"
  /*else*/ {

#line 1489 "BufferFormatFromTypeInfo"
    __pyx_v_fmt = __Pyx_TypeInfoToFormat(__pyx_v_type);

    /* "BufferFormatFromTypeInfo":1490
 *     else:
 *         fmt = __Pyx_TypeInfoToFormat(type)
 *         if type.arraysize[0]:             # <<<<<<<<<<<<<<
 *             extents = [unicode(type.arraysize[i]) for i in range(type.ndim)]
 *             result = (u"(%s)" % u','.join(extents)).encode('ascii') + fmt.string
 */

#line 1490 "BufferFormatFromTypeInfo"
    __pyx_t_1 = ((__pyx_v_type->arraysize[0]) != 0);

#line 1490 "BufferFormatFromTypeInfo"
    if (__pyx_t_1) {

      /* "BufferFormatFromTypeInfo":1491
 *         fmt = __Pyx_TypeInfoToFormat(type)
 *         if type.arraysize[0]:
 *             extents = [unicode(type.arraysize[i]) for i in range(type.ndim)]             # <<<<<<<<<<<<<<
 *             result = (u"(%s)" % u','.join(extents)).encode('ascii') + fmt.string
 *         else:
 */

#line 1491 "BufferFormatFromTypeInfo"
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 1491, __pyx_L1_error)

#line 1491 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_5);

#line 1491 "BufferFormatFromTypeInfo"
      __pyx_t_7 = __pyx_v_type->ndim;

#line 1491 "BufferFormatFromTypeInfo"
      __pyx_t_8 = __pyx_t_7;

#line 1491 "BufferFormatFromTypeInfo"
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 1491 "BufferFormatFromTypeInfo"
        __pyx_v_i = __pyx_t_9;

#line 1491 "BufferFormatFromTypeInfo"
        __pyx_t_4 = __Pyx_PyInt_FromSize_t((__pyx_v_type->arraysize[__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1491, __pyx_L1_error)

#line 1491 "BufferFormatFromTypeInfo"
        __Pyx_GOTREF(__pyx_t_4);

#line 1491 "BufferFormatFromTypeInfo"
        __pyx_t_2 = __Pyx_PyObject_Unicode(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1491, __pyx_L1_error)

#line 1491 "BufferFormatFromTypeInfo"
        __Pyx_GOTREF(__pyx_t_2);

#line 1491 "BufferFormatFromTypeInfo"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1491 "BufferFormatFromTypeInfo"
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_2))) __PYX_ERR(2, 1491, __pyx_L1_error)

#line 1491 "BufferFormatFromTypeInfo"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1491 "BufferFormatFromTypeInfo"
      }

#line 1491 "BufferFormatFromTypeInfo"
      __pyx_v_extents = ((PyObject*)__pyx_t_5);

#line 1491 "BufferFormatFromTypeInfo"
      __pyx_t_5 = 0;

      /* "BufferFormatFromTypeInfo":1492
 *         if type.arraysize[0]:
 *             extents = [unicode(type.arraysize[i]) for i in range(type.ndim)]
 *             result = (u"(%s)" % u','.join(extents)).encode('ascii') + fmt.string             # <<<<<<<<<<<<<<
 *         else:
 *             result = fmt.string
 */

#line 1492 "BufferFormatFromTypeInfo"
      __pyx_t_5 = PyUnicode_Join(__pyx_kp_u__30, __pyx_v_extents); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 1492, __pyx_L1_error)

#line 1492 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_5);

#line 1492 "BufferFormatFromTypeInfo"
      __pyx_t_2 = PyUnicode_Format(__pyx_kp_u_s, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1492, __pyx_L1_error)

#line 1492 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_2);

#line 1492 "BufferFormatFromTypeInfo"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 1492 "BufferFormatFromTypeInfo"
      __pyx_t_5 = PyUnicode_AsASCIIString(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 1492, __pyx_L1_error)

#line 1492 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_5);

#line 1492 "BufferFormatFromTypeInfo"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1492 "BufferFormatFromTypeInfo"
      __pyx_t_2 = __Pyx_PyObject_FromString(__pyx_v_fmt.string); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1492, __pyx_L1_error)

#line 1492 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_2);

#line 1492 "BufferFormatFromTypeInfo"
      __pyx_t_4 = PyNumber_Add(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1492, __pyx_L1_error)

#line 1492 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_4);

#line 1492 "BufferFormatFromTypeInfo"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 1492 "BufferFormatFromTypeInfo"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1492 "BufferFormatFromTypeInfo"
      if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(2, 1492, __pyx_L1_error)

#line 1492 "BufferFormatFromTypeInfo"
      __pyx_v_result = ((PyObject*)__pyx_t_4);

#line 1492 "BufferFormatFromTypeInfo"
      __pyx_t_4 = 0;

      /* "BufferFormatFromTypeInfo":1490
 *     else:
 *         fmt = __Pyx_TypeInfoToFormat(type)
 *         if type.arraysize[0]:             # <<<<<<<<<<<<<<
 *             extents = [unicode(type.arraysize[i]) for i in range(type.ndim)]
 *             result = (u"(%s)" % u','.join(extents)).encode('ascii') + fmt.string
 */

#line 1490 "BufferFormatFromTypeInfo"
      goto __pyx_L7;

#line 1490 "BufferFormatFromTypeInfo"
    }

    /* "BufferFormatFromTypeInfo":1494
 *             result = (u"(%s)" % u','.join(extents)).encode('ascii') + fmt.string
 *         else:
 *             result = fmt.string             # <<<<<<<<<<<<<<
 * 
 *     return result
 */

#line 1494 "BufferFormatFromTypeInfo"
    /*else*/ {

#line 1494 "BufferFormatFromTypeInfo"
      __pyx_t_4 = __Pyx_PyObject_FromString(__pyx_v_fmt.string); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1494, __pyx_L1_error)

#line 1494 "BufferFormatFromTypeInfo"
      __Pyx_GOTREF(__pyx_t_4);

#line 1494 "BufferFormatFromTypeInfo"
      __pyx_v_result = ((PyObject*)__pyx_t_4);

#line 1494 "BufferFormatFromTypeInfo"
      __pyx_t_4 = 0;

#line 1494 "BufferFormatFromTypeInfo"
    }

#line 1494 "BufferFormatFromTypeInfo"
    __pyx_L7:;

#line 1494 "BufferFormatFromTypeInfo"
  }

#line 1494 "BufferFormatFromTypeInfo"
  __pyx_L3:;

  /* "BufferFormatFromTypeInfo":1496
 *             result = fmt.string
 * 
 *     return result             # <<<<<<<<<<<<<<
 */

#line 1496 "BufferFormatFromTypeInfo"
  __Pyx_XDECREF(__pyx_r);

#line 1496 "BufferFormatFromTypeInfo"
  __Pyx_INCREF(__pyx_v_result);

#line 1496 "BufferFormatFromTypeInfo"
  __pyx_r = __pyx_v_result;

#line 1496 "BufferFormatFromTypeInfo"
  goto __pyx_L0;

  /* "BufferFormatFromTypeInfo":1465
 * 
 * @cname('__pyx_format_from_typeinfo')
 * cdef bytes format_from_typeinfo(__Pyx_TypeInfo *type):             # <<<<<<<<<<<<<<
 *     cdef __Pyx_StructField *field
 *     cdef __pyx_typeinfo_string fmt
 */

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
  /* function exit code */

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_L1_error:;

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_XDECREF(__pyx_t_2);

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_XDECREF(__pyx_t_4);

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_XDECREF(__pyx_t_5);

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_AddTraceback("BufferFormatFromTypeInfo.format_from_typeinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_r = 0;

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_L0:;

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_XDECREF(__pyx_v_part);

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_XDECREF(__pyx_v_result);

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_XDECREF(__pyx_v_alignment);

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_XDECREF(__pyx_v_parts);

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_XDECREF(__pyx_v_extents);

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_XGIVEREF(__pyx_r);

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_RefNannyFinishContext();

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_r;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *p;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *o;

#line 1465 "BufferFormatFromTypeInfo"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1465 "BufferFormatFromTypeInfo"
    o = (*t->tp_alloc)(t, 0);

#line 1465 "BufferFormatFromTypeInfo"
  } else {

#line 1465 "BufferFormatFromTypeInfo"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(!o)) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *)o);

#line 1465 "BufferFormatFromTypeInfo"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;

#line 1465 "BufferFormatFromTypeInfo"
  return o;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TP_FINALIZE

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {

#line 1465 "BufferFormatFromTypeInfo"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  (*Py_TYPE(o)->tp_free)(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_start(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_start(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5start_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_end(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_end(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3end_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"contains", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_5contains, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_4contains},

#line 1465 "BufferFormatFromTypeInfo"
  {"overlaps", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_7overlaps, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_6overlaps},

#line 1465 "BufferFormatFromTypeInfo"
  {"overlap_size", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_9overlap_size, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_8overlap_size},

#line 1465 "BufferFormatFromTypeInfo"
  {"is_contained_in_interval", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_11is_contained_in_interval, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_10is_contained_in_interval},

#line 1465 "BufferFormatFromTypeInfo"
  {"contains_interval", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_13contains_interval, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_12contains_interval},

#line 1465 "BufferFormatFromTypeInfo"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_15__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_14__reduce_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_17__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_16__setstate_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"start", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_start, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_start, (char *)"start: 'double'", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"end", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_end, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_end, (char *)"end: 'double'", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PySequenceMethods __pyx_tp_as_sequence_SpanningMixin = {

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_length*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3__contains__, /*sq_contains*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin = {

#line 1465 "BufferFormatFromTypeInfo"
  PyVarObject_HEAD_INIT(0, 0)

#line 1465 "BufferFormatFromTypeInfo"
  "ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin", /*tp_name*/

#line 1465 "BufferFormatFromTypeInfo"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin), /*tp_basicsize*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_itemsize*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, /*tp_dealloc*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX < 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall_offset*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattr*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_compare*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_async*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_number*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_sequence_SpanningMixin, /*tp_as_sequence*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_mapping*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_hash*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_call*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_buffer*/

#line 1465 "BufferFormatFromTypeInfo"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  "SpanningMixin(start, end)\nProvides methods for checking whether an entity\n    which has a defined start and end point over a single\n    dimension contains or overlaps with another entity in\n    that same dimension.\n    ", /*tp_doc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_traverse*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_clear*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_richcompare*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklistoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iternext*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, /*tp_methods*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_members*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, /*tp_getset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_base*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dict*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_get*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_set*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dictoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_1__init__, /*tp_init*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_alloc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, /*tp_new*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_free*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_is_gc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_bases*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_mro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_cache*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_subclasses*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklist*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_del*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_version_tag*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_finalize*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_pypy_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *p;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin(t, a, k);

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(!o)) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)o);

#line 1465 "BufferFormatFromTypeInfo"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin*)__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;

#line 1465 "BufferFormatFromTypeInfo"
  p->members = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  p->data = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  return o;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)o;

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TP_FINALIZE

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1465 "BufferFormatFromTypeInfo"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  PyObject_GC_UnTrack(o);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->members);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->data);

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TYPE_SLOTS

#line 1465 "BufferFormatFromTypeInfo"
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  PyObject_GC_Track(o);

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyObject *o, visitproc v, void *a) {

#line 1465 "BufferFormatFromTypeInfo"
  int e;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)o;

#line 1465 "BufferFormatFromTypeInfo"
  e = ((likely(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin)) ? ((__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin->tp_traverse) ? __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval)); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  if (p->members) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->members, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (p->data) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->data, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject* tmp;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval *)o;

#line 1465 "BufferFormatFromTypeInfo"
  if (likely(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin)) { if (__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin->tp_clear) __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->members);

#line 1465 "BufferFormatFromTypeInfo"
  p->members = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->data);

#line 1465 "BufferFormatFromTypeInfo"
  p->data = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_sq_item_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval(PyObject *o, Py_ssize_t i) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *r;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 1465 "BufferFormatFromTypeInfo"
  Py_DECREF(x);

#line 1465 "BufferFormatFromTypeInfo"
  return r;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_members(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_members(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7members_5__del__(o);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_data(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_data(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_4data_5__del__(o);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_11__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_10__reduce_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_13__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_12__setstate_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"members", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_members, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_members, (char *)"members: object", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"data", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_data, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_data, (char *)"data: dict", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PySequenceMethods __pyx_tp_as_sequence_Interval = {

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7__len__, /*sq_length*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_sq_item_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*sq_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_3__contains__, /*sq_contains*/

#line 1465 "BufferFormatFromTypeInfo"
  #else

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_contains*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMappingMethods __pyx_tp_as_mapping_Interval = {

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_7__len__, /*mp_length*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_5__getitem__, /*mp_subscript*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*mp_ass_subscript*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval = {

#line 1465 "BufferFormatFromTypeInfo"
  PyVarObject_HEAD_INIT(0, 0)

#line 1465 "BufferFormatFromTypeInfo"
  "ms_deisotope._c.peak_dependency_network.intervals.Interval", /*tp_name*/

#line 1465 "BufferFormatFromTypeInfo"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval), /*tp_basicsize*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_itemsize*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_dealloc*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX < 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall_offset*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattr*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_compare*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_async*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_3__repr__, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_number*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_sequence_Interval, /*tp_as_sequence*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_mapping_Interval, /*tp_as_mapping*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_hash*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_call*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_buffer*/

#line 1465 "BufferFormatFromTypeInfo"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  "Interval(start, end, members=None, **kwargs)", /*tp_doc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_traverse*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_clear*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_richcompare*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklistoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_9__iter__, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iternext*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_methods*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_members*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_getset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_base*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dict*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_get*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_set*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dictoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_8Interval_1__init__, /*tp_init*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_alloc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval, /*tp_new*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_free*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_is_gc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_bases*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_mro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_cache*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_subclasses*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklist*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_del*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_version_tag*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_finalize*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_pypy_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval;

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *p;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin(t, a, k);

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(!o)) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *)o);

#line 1465 "BufferFormatFromTypeInfo"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin*)__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval;

#line 1465 "BufferFormatFromTypeInfo"
  return o;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_5__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_4__reduce_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_7__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_6__setstate_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval = {

#line 1465 "BufferFormatFromTypeInfo"
  PyVarObject_HEAD_INIT(0, 0)

#line 1465 "BufferFormatFromTypeInfo"
  "ms_deisotope._c.peak_dependency_network.intervals.SimpleInterval", /*tp_name*/

#line 1465 "BufferFormatFromTypeInfo"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval), /*tp_basicsize*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_itemsize*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, /*tp_dealloc*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX < 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall_offset*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattr*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_compare*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_async*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_3__repr__, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_number*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_sequence*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_mapping*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_hash*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_call*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_buffer*/

#line 1465 "BufferFormatFromTypeInfo"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  "SimpleInterval(start, end)", /*tp_doc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_traverse*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_clear*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_richcompare*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklistoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iternext*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval, /*tp_methods*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_members*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_base*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dict*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_get*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_set*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dictoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval_1__init__, /*tp_init*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_alloc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval, /*tp_new*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_free*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_is_gc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_bases*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_mro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_cache*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_subclasses*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklist*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_del*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_version_tag*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_finalize*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_pypy_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *p;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *o;

#line 1465 "BufferFormatFromTypeInfo"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1465 "BufferFormatFromTypeInfo"
    o = (*t->tp_alloc)(t, 0);

#line 1465 "BufferFormatFromTypeInfo"
  } else {

#line 1465 "BufferFormatFromTypeInfo"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(!o)) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)o);

#line 1465 "BufferFormatFromTypeInfo"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;

#line 1465 "BufferFormatFromTypeInfo"
  p->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  p->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  p->contained = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  p->parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  return o;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)o;

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TP_FINALIZE

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1465 "BufferFormatFromTypeInfo"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  PyObject_GC_UnTrack(o);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->left);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->right);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->contained);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->parent);

#line 1465 "BufferFormatFromTypeInfo"
  (*Py_TYPE(o)->tp_free)(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(PyObject *o, visitproc v, void *a) {

#line 1465 "BufferFormatFromTypeInfo"
  int e;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)o;

#line 1465 "BufferFormatFromTypeInfo"
  if (p->left) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(((PyObject *)p->left), a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (p->right) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(((PyObject *)p->right), a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (p->contained) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->contained, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (p->parent) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(((PyObject *)p->parent), a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject* tmp;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)o;

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->left);

#line 1465 "BufferFormatFromTypeInfo"
  p->left = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->right);

#line 1465 "BufferFormatFromTypeInfo"
  p->right = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->contained);

#line 1465 "BufferFormatFromTypeInfo"
  p->contained = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->parent);

#line 1465 "BufferFormatFromTypeInfo"
  p->parent = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_center(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_center(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6center_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_start(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_start(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5start_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_end(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_end(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3end_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_left(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_left(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4left_5__del__(o);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_right(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_right(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5right_5__del__(o);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contained(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contained(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9contained_5__del__(o);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_level(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_level(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5level_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_parent(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_parent(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6parent_5__del__(o);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"contains_point", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_3contains_point, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_2contains_point},

#line 1465 "BufferFormatFromTypeInfo"
  {"_overlaps_interval", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_5_overlaps_interval, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_4_overlaps_interval},

#line 1465 "BufferFormatFromTypeInfo"
  {"overlaps", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_7overlaps, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_6overlaps},

#line 1465 "BufferFormatFromTypeInfo"
  {"_eq", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_13_eq, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_12_eq},

#line 1465 "BufferFormatFromTypeInfo"
  {"build", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_17build, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_16build},

#line 1465 "BufferFormatFromTypeInfo"
  {"_update_bounds", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_19_update_bounds, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_18_update_bounds},

#line 1465 "BufferFormatFromTypeInfo"
  {"insert", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_21insert, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_20insert},

#line 1465 "BufferFormatFromTypeInfo"
  {"flatten", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_23flatten, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_22flatten},

#line 1465 "BufferFormatFromTypeInfo"
  {"balance", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_25balance, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_24balance},

#line 1465 "BufferFormatFromTypeInfo"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_29__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_28__reduce_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_31__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_30__setstate_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"center", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_center, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_center, (char *)"center: 'double'", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"start", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_start, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_start, (char *)"start: 'double'", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"end", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_end, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_end, (char *)"end: 'double'", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"left", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_left, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_left, (char *)"left: ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"right", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_right, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_right, (char *)"right: ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"contained", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contained, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contained, (char *)"contained: list", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"level", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_level, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_level, (char *)"level: 'int'", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"parent", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_parent, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_parent, (char *)"parent: ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode = {

#line 1465 "BufferFormatFromTypeInfo"
  PyVarObject_HEAD_INIT(0, 0)

#line 1465 "BufferFormatFromTypeInfo"
  "ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode", /*tp_name*/

#line 1465 "BufferFormatFromTypeInfo"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode), /*tp_basicsize*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_itemsize*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_dealloc*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX < 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall_offset*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattr*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_compare*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_async*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9__repr__, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_number*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_sequence*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_mapping*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_15__hash__, /*tp_hash*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_call*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_buffer*/

#line 1465 "BufferFormatFromTypeInfo"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  "IntervalTreeNode(double center, IntervalTreeNode left, list contained, IntervalTreeNode right, int level=0, IntervalTreeNode parent=None)", /*tp_doc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_traverse*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_clear*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_11__richcmp__, /*tp_richcompare*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklistoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_27__iter__, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iternext*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_methods*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_members*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_getset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_base*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dict*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_get*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_set*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dictoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_1__init__, /*tp_init*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_alloc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, /*tp_new*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_free*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_is_gc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_bases*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_mro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_cache*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_subclasses*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklist*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_del*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_version_tag*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_finalize*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_pypy_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D;

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *p;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *o;

#line 1465 "BufferFormatFromTypeInfo"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1465 "BufferFormatFromTypeInfo"
    o = (*t->tp_alloc)(t, 0);

#line 1465 "BufferFormatFromTypeInfo"
  } else {

#line 1465 "BufferFormatFromTypeInfo"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(!o)) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *)o);

#line 1465 "BufferFormatFromTypeInfo"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D;

#line 1465 "BufferFormatFromTypeInfo"
  return o;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TP_FINALIZE

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {

#line 1465 "BufferFormatFromTypeInfo"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  (*Py_TYPE(o)->tp_free)(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_start(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_start(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5start_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_end(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_end(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3end_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D__start(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D__start(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6_start_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D__end(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D__end(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4_end_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"contains", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_5contains, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_4contains},

#line 1465 "BufferFormatFromTypeInfo"
  {"overlaps", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_7overlaps, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_6overlaps},

#line 1465 "BufferFormatFromTypeInfo"
  {"contains_interval", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_9contains_interval, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_8contains_interval},

#line 1465 "BufferFormatFromTypeInfo"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_11__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_10__reduce_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_13__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_12__setstate_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"start", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_start, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_start, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"end", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_end, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_end, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"_start", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D__start, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D__start, (char *)"_start: 'double [2]'", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"_end", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D__end, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D__end, (char *)"_end: 'double [2]'", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PySequenceMethods __pyx_tp_as_sequence_SpanningMixin2D = {

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_length*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3__contains__, /*sq_contains*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D = {

#line 1465 "BufferFormatFromTypeInfo"
  PyVarObject_HEAD_INIT(0, 0)

#line 1465 "BufferFormatFromTypeInfo"
  "ms_deisotope._c.peak_dependency_network.intervals.SpanningMixin2D", /*tp_name*/

#line 1465 "BufferFormatFromTypeInfo"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D), /*tp_basicsize*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_itemsize*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D, /*tp_dealloc*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX < 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall_offset*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattr*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_compare*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_async*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_number*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_sequence_SpanningMixin2D, /*tp_as_sequence*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_mapping*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_hash*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_call*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_buffer*/

#line 1465 "BufferFormatFromTypeInfo"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  "SpanningMixin2D(start, end)", /*tp_doc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_traverse*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_clear*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_richcompare*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklistoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iternext*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D, /*tp_methods*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_members*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D, /*tp_getset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_base*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dict*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_get*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_set*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dictoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_1__init__, /*tp_init*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_alloc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D, /*tp_new*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_free*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_is_gc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_bases*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_mro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_cache*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_subclasses*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklist*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_del*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_version_tag*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_finalize*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_pypy_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D;

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *p;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D(t, a, k);

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(!o)) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)o);

#line 1465 "BufferFormatFromTypeInfo"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D*)__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D;

#line 1465 "BufferFormatFromTypeInfo"
  p->members = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  p->data = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  return o;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)o;

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TP_FINALIZE

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1465 "BufferFormatFromTypeInfo"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  PyObject_GC_UnTrack(o);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->members);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->data);

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TYPE_SLOTS

#line 1465 "BufferFormatFromTypeInfo"
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  PyObject_GC_Track(o);

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D(PyObject *o, visitproc v, void *a) {

#line 1465 "BufferFormatFromTypeInfo"
  int e;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)o;

#line 1465 "BufferFormatFromTypeInfo"
  e = ((likely(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D)) ? ((__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D->tp_traverse) ? __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D)); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  if (p->members) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->members, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (p->data) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->data, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject* tmp;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D *)o;

#line 1465 "BufferFormatFromTypeInfo"
  if (likely(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D)) { if (__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D->tp_clear) __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->members);

#line 1465 "BufferFormatFromTypeInfo"
  p->members = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->data);

#line 1465 "BufferFormatFromTypeInfo"
  p->data = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_sq_item_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D(PyObject *o, Py_ssize_t i) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *r;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 1465 "BufferFormatFromTypeInfo"
  Py_DECREF(x);

#line 1465 "BufferFormatFromTypeInfo"
  return r;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_members(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_members(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7members_5__del__(o);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_data(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_data(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_4data_5__del__(o);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_11__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_10__reduce_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_13__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_12__setstate_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"members", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_members, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_members, (char *)"members: list", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"data", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_data, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_data, (char *)"data: dict", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PySequenceMethods __pyx_tp_as_sequence_Interval2D = {

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7__len__, /*sq_length*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_sq_item_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D, /*sq_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_3__contains__, /*sq_contains*/

#line 1465 "BufferFormatFromTypeInfo"
  #else

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_contains*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMappingMethods __pyx_tp_as_mapping_Interval2D = {

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_7__len__, /*mp_length*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_5__getitem__, /*mp_subscript*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*mp_ass_subscript*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D = {

#line 1465 "BufferFormatFromTypeInfo"
  PyVarObject_HEAD_INIT(0, 0)

#line 1465 "BufferFormatFromTypeInfo"
  "ms_deisotope._c.peak_dependency_network.intervals.Interval2D", /*tp_name*/

#line 1465 "BufferFormatFromTypeInfo"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D), /*tp_basicsize*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_itemsize*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D, /*tp_dealloc*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX < 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall_offset*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattr*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_compare*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_async*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_3__repr__, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_number*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_sequence_Interval2D, /*tp_as_sequence*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_mapping_Interval2D, /*tp_as_mapping*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_hash*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_call*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_buffer*/

#line 1465 "BufferFormatFromTypeInfo"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  "Interval2D(start, end, members=None, **kwargs)", /*tp_doc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D, /*tp_traverse*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D, /*tp_clear*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_richcompare*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklistoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_9__iter__, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iternext*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D, /*tp_methods*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_members*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D, /*tp_getset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_base*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dict*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_get*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_set*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dictoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_10Interval2D_1__init__, /*tp_init*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_alloc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D, /*tp_new*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_free*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_is_gc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_bases*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_mro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_cache*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_subclasses*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklist*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_del*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_version_tag*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_finalize*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_pypy_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D;

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *p;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(t, a, k);

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(!o)) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)o);

#line 1465 "BufferFormatFromTypeInfo"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode*)__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D;

#line 1465 "BufferFormatFromTypeInfo"
  p->inner_organizer = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  p->organized = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  return o;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)o;

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TP_FINALIZE

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1465 "BufferFormatFromTypeInfo"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  PyObject_GC_UnTrack(o);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->inner_organizer);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->organized);

#line 1465 "BufferFormatFromTypeInfo"
  PyObject_GC_Track(o);

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D(PyObject *o, visitproc v, void *a) {

#line 1465 "BufferFormatFromTypeInfo"
  int e;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)o;

#line 1465 "BufferFormatFromTypeInfo"
  e = __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(o, v, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  if (p->inner_organizer) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->inner_organizer, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (p->organized) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(((PyObject *)p->organized), a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject* tmp;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *p = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *)o;

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode(o);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->inner_organizer);

#line 1465 "BufferFormatFromTypeInfo"
  p->inner_organizer = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->organized);

#line 1465 "BufferFormatFromTypeInfo"
  p->organized = ((struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_inner_organizer(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_inner_organizer(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15inner_organizer_5__del__(o);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_organized(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_organized(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_3__set__(o, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9organized_5__del__(o);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"set_organizer", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_3set_organizer, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_2set_organizer},

#line 1465 "BufferFormatFromTypeInfo"
  {"organize", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_5organize, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_4organize},

#line 1465 "BufferFormatFromTypeInfo"
  {"_overlaps_interval_2d", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_7_overlaps_interval_2d, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_6_overlaps_interval_2d},

#line 1465 "BufferFormatFromTypeInfo"
  {"overlaps_2d", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_9overlaps_2d, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_8overlaps_2d},

#line 1465 "BufferFormatFromTypeInfo"
  {"build", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_11build, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_10build},

#line 1465 "BufferFormatFromTypeInfo"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_13__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_12__reduce_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_15__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_14__setstate_cython__},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"inner_organizer", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_inner_organizer, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_inner_organizer, (char *)"inner_organizer: object", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"organized", __pyx_getprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_organized, __pyx_setprop_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_organized, (char *)"organized: ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode", 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D = {

#line 1465 "BufferFormatFromTypeInfo"
  PyVarObject_HEAD_INIT(0, 0)

#line 1465 "BufferFormatFromTypeInfo"
  "ms_deisotope._c.peak_dependency_network.intervals.IntervalTreeNode2D", /*tp_name*/

#line 1465 "BufferFormatFromTypeInfo"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D), /*tp_basicsize*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_itemsize*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D, /*tp_dealloc*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX < 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall_offset*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattr*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_compare*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_async*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_9__repr__, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  #else

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_number*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_sequence*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_mapping*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_hash*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_call*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_buffer*/

#line 1465 "BufferFormatFromTypeInfo"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  "IntervalTreeNode2D(center, left, contained, right, level=0, parent=None, inner_organizer=None)", /*tp_doc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_traverse_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D, /*tp_traverse*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_clear_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D, /*tp_clear*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_richcompare*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklistoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_27__iter__, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  #else

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iternext*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_methods_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D, /*tp_methods*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_members*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_getsets_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D, /*tp_getset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_base*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dict*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_get*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_set*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dictoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_1__init__, /*tp_init*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_alloc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_new_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D, /*tp_new*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_free*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_is_gc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_bases*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_mro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_cache*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_subclasses*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklist*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_del*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_version_tag*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_finalize*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_pypy_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"
static struct __pyx_vtabstruct_array __pyx_vtable_array;

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_array_obj *p;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *o;

#line 1465 "BufferFormatFromTypeInfo"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1465 "BufferFormatFromTypeInfo"
    o = (*t->tp_alloc)(t, 0);

#line 1465 "BufferFormatFromTypeInfo"
  } else {

#line 1465 "BufferFormatFromTypeInfo"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(!o)) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  p = ((struct __pyx_array_obj *)o);

#line 1465 "BufferFormatFromTypeInfo"
  p->__pyx_vtab = __pyx_vtabptr_array;

#line 1465 "BufferFormatFromTypeInfo"
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;

#line 1465 "BufferFormatFromTypeInfo"
  return o;

#line 1465 "BufferFormatFromTypeInfo"
  bad:

#line 1465 "BufferFormatFromTypeInfo"
  Py_DECREF(o); o = 0;

#line 1465 "BufferFormatFromTypeInfo"
  return NULL;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static void __pyx_tp_dealloc_array(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TP_FINALIZE

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {

#line 1465 "BufferFormatFromTypeInfo"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  {

#line 1465 "BufferFormatFromTypeInfo"
    PyObject *etype, *eval, *etb;

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_Fetch(&etype, &eval, &etb);

#line 1465 "BufferFormatFromTypeInfo"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);

#line 1465 "BufferFormatFromTypeInfo"
    __pyx_array___dealloc__(o);

#line 1465 "BufferFormatFromTypeInfo"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_Restore(etype, eval, etb);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->mode);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->_format);

#line 1465 "BufferFormatFromTypeInfo"
  (*Py_TYPE(o)->tp_free)(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *r;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 1465 "BufferFormatFromTypeInfo"
  Py_DECREF(x);

#line 1465 "BufferFormatFromTypeInfo"
  return r;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_array___setitem__(o, i, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_Format(PyExc_NotImplementedError,

#line 1465 "BufferFormatFromTypeInfo"
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);

#line 1465 "BufferFormatFromTypeInfo"
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_Clear();

#line 1465 "BufferFormatFromTypeInfo"
    v = __pyx_array___getattr__(o, n);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  return v;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods_array[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static struct PyGetSetDef __pyx_getsets_array[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PySequenceMethods __pyx_tp_as_sequence_array = {

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_array___len__, /*sq_length*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_sq_item_array, /*sq_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_contains*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMappingMethods __pyx_tp_as_mapping_array = {

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_array___len__, /*mp_length*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_array___getitem__, /*mp_subscript*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyBufferProcs __pyx_tp_as_buffer_array = {

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*bf_getreadbuffer*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*bf_getwritebuffer*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*bf_getsegcount*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*bf_getcharbuffer*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_array_getbuffer, /*bf_getbuffer*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*bf_releasebuffer*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyTypeObject __pyx_type___pyx_array = {

#line 1465 "BufferFormatFromTypeInfo"
  PyVarObject_HEAD_INIT(0, 0)

#line 1465 "BufferFormatFromTypeInfo"
  "ms_deisotope._c.peak_dependency_network.intervals.array", /*tp_name*/

#line 1465 "BufferFormatFromTypeInfo"
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_itemsize*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_array, /*tp_dealloc*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX < 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall_offset*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattr*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_compare*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_async*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_number*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_hash*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_call*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_getattro_array, /*tp_getattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattro*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/

#line 1465 "BufferFormatFromTypeInfo"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_doc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_traverse*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_clear*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_richcompare*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklistoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iternext*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_methods_array, /*tp_methods*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_members*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_getsets_array, /*tp_getset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_base*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dict*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_get*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_set*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dictoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_init*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_alloc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_new_array, /*tp_new*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_free*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_is_gc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_bases*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_mro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_cache*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_subclasses*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklist*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_del*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_version_tag*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_finalize*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_pypy_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_MemviewEnum_obj *p;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *o;

#line 1465 "BufferFormatFromTypeInfo"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1465 "BufferFormatFromTypeInfo"
    o = (*t->tp_alloc)(t, 0);

#line 1465 "BufferFormatFromTypeInfo"
  } else {

#line 1465 "BufferFormatFromTypeInfo"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(!o)) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  p = ((struct __pyx_MemviewEnum_obj *)o);

#line 1465 "BufferFormatFromTypeInfo"
  p->name = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  return o;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static void __pyx_tp_dealloc_Enum(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TP_FINALIZE

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1465 "BufferFormatFromTypeInfo"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  PyObject_GC_UnTrack(o);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->name);

#line 1465 "BufferFormatFromTypeInfo"
  (*Py_TYPE(o)->tp_free)(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {

#line 1465 "BufferFormatFromTypeInfo"
  int e;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;

#line 1465 "BufferFormatFromTypeInfo"
  if (p->name) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->name, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_clear_Enum(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject* tmp;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->name);

#line 1465 "BufferFormatFromTypeInfo"
  p->name = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods_Enum[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyTypeObject __pyx_type___pyx_MemviewEnum = {

#line 1465 "BufferFormatFromTypeInfo"
  PyVarObject_HEAD_INIT(0, 0)

#line 1465 "BufferFormatFromTypeInfo"
  "ms_deisotope._c.peak_dependency_network.intervals.Enum", /*tp_name*/

#line 1465 "BufferFormatFromTypeInfo"
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_itemsize*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX < 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall_offset*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattr*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_compare*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_async*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_MemviewEnum___repr__, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_number*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_sequence*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_mapping*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_hash*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_call*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_buffer*/

#line 1465 "BufferFormatFromTypeInfo"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_doc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_traverse_Enum, /*tp_traverse*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_clear_Enum, /*tp_clear*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_richcompare*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklistoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iternext*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_methods_Enum, /*tp_methods*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_members*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_base*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dict*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_get*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_set*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dictoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_MemviewEnum___init__, /*tp_init*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_alloc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_new_Enum, /*tp_new*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_free*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_is_gc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_bases*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_mro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_cache*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_subclasses*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklist*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_del*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_version_tag*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_finalize*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_pypy_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_memoryview_obj *p;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *o;

#line 1465 "BufferFormatFromTypeInfo"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1465 "BufferFormatFromTypeInfo"
    o = (*t->tp_alloc)(t, 0);

#line 1465 "BufferFormatFromTypeInfo"
  } else {

#line 1465 "BufferFormatFromTypeInfo"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(!o)) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  p = ((struct __pyx_memoryview_obj *)o);

#line 1465 "BufferFormatFromTypeInfo"
  p->__pyx_vtab = __pyx_vtabptr_memoryview;

#line 1465 "BufferFormatFromTypeInfo"
  p->obj = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  p->_size = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  p->_array_interface = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  p->view.obj = NULL;

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;

#line 1465 "BufferFormatFromTypeInfo"
  return o;

#line 1465 "BufferFormatFromTypeInfo"
  bad:

#line 1465 "BufferFormatFromTypeInfo"
  Py_DECREF(o); o = 0;

#line 1465 "BufferFormatFromTypeInfo"
  return NULL;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static void __pyx_tp_dealloc_memoryview(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TP_FINALIZE

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1465 "BufferFormatFromTypeInfo"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  PyObject_GC_UnTrack(o);

#line 1465 "BufferFormatFromTypeInfo"
  {

#line 1465 "BufferFormatFromTypeInfo"
    PyObject *etype, *eval, *etb;

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_Fetch(&etype, &eval, &etb);

#line 1465 "BufferFormatFromTypeInfo"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);

#line 1465 "BufferFormatFromTypeInfo"
    __pyx_memoryview___dealloc__(o);

#line 1465 "BufferFormatFromTypeInfo"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_Restore(etype, eval, etb);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->obj);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->_size);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->_array_interface);

#line 1465 "BufferFormatFromTypeInfo"
  (*Py_TYPE(o)->tp_free)(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {

#line 1465 "BufferFormatFromTypeInfo"
  int e;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;

#line 1465 "BufferFormatFromTypeInfo"
  if (p->obj) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->obj, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (p->_size) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->_size, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (p->_array_interface) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->_array_interface, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  if (p->view.obj) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->view.obj, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_clear_memoryview(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject* tmp;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->obj);

#line 1465 "BufferFormatFromTypeInfo"
  p->obj = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->_size);

#line 1465 "BufferFormatFromTypeInfo"
  p->_size = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->_array_interface);

#line 1465 "BufferFormatFromTypeInfo"
  p->_array_interface = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->view.obj);

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *r;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 1465 "BufferFormatFromTypeInfo"
  Py_DECREF(x);

#line 1465 "BufferFormatFromTypeInfo"
  return r;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {

#line 1465 "BufferFormatFromTypeInfo"
  if (v) {

#line 1465 "BufferFormatFromTypeInfo"
    return __pyx_memoryview___setitem__(o, i, v);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  else {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_Format(PyExc_NotImplementedError,

#line 1465 "BufferFormatFromTypeInfo"
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);

#line 1465 "BufferFormatFromTypeInfo"
    return -1;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods_memoryview[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static struct PyGetSetDef __pyx_getsets_memoryview[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PySequenceMethods __pyx_tp_as_sequence_memoryview = {

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_memoryview___len__, /*sq_length*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_sq_item_memoryview, /*sq_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_item*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_ass_slice*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_contains*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_concat*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*sq_inplace_repeat*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMappingMethods __pyx_tp_as_mapping_memoryview = {

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_memoryview___len__, /*mp_length*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_memoryview___getitem__, /*mp_subscript*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyBufferProcs __pyx_tp_as_buffer_memoryview = {

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*bf_getreadbuffer*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*bf_getwritebuffer*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*bf_getsegcount*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*bf_getcharbuffer*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*bf_releasebuffer*/

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyTypeObject __pyx_type___pyx_memoryview = {

#line 1465 "BufferFormatFromTypeInfo"
  PyVarObject_HEAD_INIT(0, 0)

#line 1465 "BufferFormatFromTypeInfo"
  "ms_deisotope._c.peak_dependency_network.intervals.memoryview", /*tp_name*/

#line 1465 "BufferFormatFromTypeInfo"
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_itemsize*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX < 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall_offset*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattr*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_compare*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_async*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_memoryview___repr__, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_number*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_hash*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_call*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_memoryview___str__, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattro*/

#line 1465 "BufferFormatFromTypeInfo"
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/

#line 1465 "BufferFormatFromTypeInfo"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_doc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_traverse_memoryview, /*tp_traverse*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_clear_memoryview, /*tp_clear*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_richcompare*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklistoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iternext*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_methods_memoryview, /*tp_methods*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_members*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_getsets_memoryview, /*tp_getset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_base*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dict*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_get*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_set*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dictoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_init*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_alloc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_new_memoryview, /*tp_new*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_free*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_is_gc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_bases*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_mro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_cache*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_subclasses*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklist*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_del*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_version_tag*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_finalize*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_pypy_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_memoryviewslice_obj *p;

#line 1465 "BufferFormatFromTypeInfo"
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(!o)) return 0;

#line 1465 "BufferFormatFromTypeInfo"
  p = ((struct __pyx_memoryviewslice_obj *)o);

#line 1465 "BufferFormatFromTypeInfo"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;

#line 1465 "BufferFormatFromTypeInfo"
  p->from_object = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  p->from_slice.memview = NULL;

#line 1465 "BufferFormatFromTypeInfo"
  return o;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_USE_TP_FINALIZE

#line 1465 "BufferFormatFromTypeInfo"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1465 "BufferFormatFromTypeInfo"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  PyObject_GC_UnTrack(o);

#line 1465 "BufferFormatFromTypeInfo"
  {

#line 1465 "BufferFormatFromTypeInfo"
    PyObject *etype, *eval, *etb;

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_Fetch(&etype, &eval, &etb);

#line 1465 "BufferFormatFromTypeInfo"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);

#line 1465 "BufferFormatFromTypeInfo"
    __pyx_memoryviewslice___dealloc__(o);

#line 1465 "BufferFormatFromTypeInfo"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_Restore(etype, eval, etb);

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  Py_CLEAR(p->from_object);

#line 1465 "BufferFormatFromTypeInfo"
  PyObject_GC_Track(o);

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc_memoryview(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {

#line 1465 "BufferFormatFromTypeInfo"
  int e;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;

#line 1465 "BufferFormatFromTypeInfo"
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  if (p->from_object) {

#line 1465 "BufferFormatFromTypeInfo"
    e = (*v)(p->from_object, a); if (e) return e;

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_tp_clear__memoryviewslice(PyObject *o) {

#line 1465 "BufferFormatFromTypeInfo"
  PyObject* tmp;

#line 1465 "BufferFormatFromTypeInfo"
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_clear_memoryview(o);

#line 1465 "BufferFormatFromTypeInfo"
  tmp = ((PyObject*)p->from_object);

#line 1465 "BufferFormatFromTypeInfo"
  p->from_object = Py_None; Py_INCREF(Py_None);

#line 1465 "BufferFormatFromTypeInfo"
  Py_XDECREF(tmp);

#line 1465 "BufferFormatFromTypeInfo"
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);

#line 1465 "BufferFormatFromTypeInfo"
  return 0;

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods__memoryviewslice[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyTypeObject __pyx_type___pyx_memoryviewslice = {

#line 1465 "BufferFormatFromTypeInfo"
  PyVarObject_HEAD_INIT(0, 0)

#line 1465 "BufferFormatFromTypeInfo"
  "ms_deisotope._c.peak_dependency_network.intervals._memoryviewslice", /*tp_name*/

#line 1465 "BufferFormatFromTypeInfo"
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_itemsize*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX < 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall_offset*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattr*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattr*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION < 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_compare*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_async*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_memoryview___repr__, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  #else

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_repr*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_number*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_sequence*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_mapping*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_hash*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_call*/

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_memoryview___str__, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  #else

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_str*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_getattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_setattro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_as_buffer*/

#line 1465 "BufferFormatFromTypeInfo"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  "Internal class for passing memoryview slices to Python", /*tp_doc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_richcompare*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklistoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iter*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_iternext*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_methods__memoryviewslice, /*tp_methods*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_members*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_getsets__memoryviewslice, /*tp_getset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_base*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dict*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_get*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_descr_set*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_dictoffset*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_init*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_alloc*/

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_tp_new__memoryviewslice, /*tp_new*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_free*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_is_gc*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_bases*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_mro*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_cache*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_subclasses*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_weaklist*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_del*/

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_version_tag*/

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_finalize*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_vectorcall*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_print*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1465 "BufferFormatFromTypeInfo"
  0, /*tp_pypy_flags*/

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static PyMethodDef __pyx_methods[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {"intervals_containing_point", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_1intervals_containing_point, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_intervals_containing_point},

#line 1465 "BufferFormatFromTypeInfo"
  {"intervals_containing_point_2d", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_23peak_dependency_network_9intervals_3intervals_containing_point_2d, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_2intervals_containing_point_2d},

#line 1465 "BufferFormatFromTypeInfo"
  {0, 0, 0, 0}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
#if PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
#if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1465 "BufferFormatFromTypeInfo"
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/

#line 1465 "BufferFormatFromTypeInfo"
static int __pyx_pymod_exec_intervals(PyObject* module); /*proto*/

#line 1465 "BufferFormatFromTypeInfo"
static PyModuleDef_Slot __pyx_moduledef_slots[] = {

#line 1465 "BufferFormatFromTypeInfo"
  {Py_mod_create, (void*)__pyx_pymod_create},

#line 1465 "BufferFormatFromTypeInfo"
  {Py_mod_exec, (void*)__pyx_pymod_exec_intervals},

#line 1465 "BufferFormatFromTypeInfo"
  {0, NULL}

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"
#endif

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
static struct PyModuleDef __pyx_moduledef = {

#line 1465 "BufferFormatFromTypeInfo"
    PyModuleDef_HEAD_INIT,

#line 1465 "BufferFormatFromTypeInfo"
    "intervals",

#line 1465 "BufferFormatFromTypeInfo"
    0, /* m_doc */

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1465 "BufferFormatFromTypeInfo"
    0, /* m_size */

#line 1465 "BufferFormatFromTypeInfo"
  #else

#line 1465 "BufferFormatFromTypeInfo"
    -1, /* m_size */

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
    __pyx_methods /* m_methods */,

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1465 "BufferFormatFromTypeInfo"
    __pyx_moduledef_slots, /* m_slots */

#line 1465 "BufferFormatFromTypeInfo"
  #else

#line 1465 "BufferFormatFromTypeInfo"
    NULL, /* m_reload */

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
    NULL, /* m_traverse */

#line 1465 "BufferFormatFromTypeInfo"
    NULL, /* m_clear */

#line 1465 "BufferFormatFromTypeInfo"
    NULL /* m_free */

#line 1465 "BufferFormatFromTypeInfo"
};

#line 1465 "BufferFormatFromTypeInfo"
#endif

#line 1465 "BufferFormatFromTypeInfo"
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_k_Incompatible_checksums_0x_x_vs_0_4, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_4), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_k_Incompatible_checksums_0x_x_vs_0_5, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_5), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_k_Incompatible_checksums_0x_x_vs_0_6, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_6), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_k_Incompatible_checksums_0x_x_vs_0_7, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_7), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_n_s_Interval, __pyx_k_Interval, sizeof(__pyx_k_Interval), 0, 0, 1, 1},
  {&__pyx_n_s_Interval2D, __pyx_k_Interval2D, sizeof(__pyx_k_Interval2D), 0, 0, 1, 1},
  {&__pyx_n_s_IntervalTreeNode, __pyx_k_IntervalTreeNode, sizeof(__pyx_k_IntervalTreeNode), 0, 0, 1, 1},
  {&__pyx_n_s_IntervalTreeNode2D, __pyx_k_IntervalTreeNode2D, sizeof(__pyx_k_IntervalTreeNode2D), 0, 0, 1, 1},
  {&__pyx_kp_s_IntervalTreeNode_level_d_center, __pyx_k_IntervalTreeNode_level_d_center, sizeof(__pyx_k_IntervalTreeNode_level_d_center), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_n_s_OverflowError, __pyx_k_OverflowError, sizeof(__pyx_k_OverflowError), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_SimpleInterval, __pyx_k_SimpleInterval, sizeof(__pyx_k_SimpleInterval), 0, 0, 1, 1},
  {&__pyx_n_s_SpanningMixin, __pyx_k_SpanningMixin, sizeof(__pyx_k_SpanningMixin), 0, 0, 1, 1},
  {&__pyx_n_s_SpanningMixin2D, __pyx_k_SpanningMixin2D, sizeof(__pyx_k_SpanningMixin2D), 0, 0, 1, 1},
  {&__pyx_kp_b_T, __pyx_k_T, sizeof(__pyx_k_T), 0, 0, 0, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_kp_b__26, __pyx_k__26, sizeof(__pyx_k__26), 0, 0, 0, 0},
  {&__pyx_kp_b__27, __pyx_k__27, sizeof(__pyx_k__27), 0, 0, 0, 0},
  {&__pyx_kp_b__28, __pyx_k__28, sizeof(__pyx_k__28), 0, 0, 0, 0},
  {&__pyx_kp_b__29, __pyx_k__29, sizeof(__pyx_k__29), 0, 0, 0, 0},
  {&__pyx_kp_u__30, __pyx_k__30, sizeof(__pyx_k__30), 0, 1, 0, 0},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_build, __pyx_k_build, sizeof(__pyx_k_build), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_center, __pyx_k_center, sizeof(__pyx_k_center), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_contained, __pyx_k_contained, sizeof(__pyx_k_contained), 0, 0, 1, 1},
  {&__pyx_n_s_contains, __pyx_k_contains, sizeof(__pyx_k_contains), 0, 0, 1, 1},
  {&__pyx_n_s_contains_interval, __pyx_k_contains_interval, sizeof(__pyx_k_contains_interval), 0, 0, 1, 1},
  {&__pyx_n_s_contains_point, __pyx_k_contains_point, sizeof(__pyx_k_contains_point), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_deque, __pyx_k_deque, sizeof(__pyx_k_deque), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_ends, __pyx_k_ends, sizeof(__pyx_k_ends), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_flatten, __pyx_k_flatten, sizeof(__pyx_k_flatten), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_inner_organizer, __pyx_k_inner_organizer, sizeof(__pyx_k_inner_organizer), 0, 0, 1, 1},
  {&__pyx_n_s_insert, __pyx_k_insert, sizeof(__pyx_k_insert), 0, 0, 1, 1},
  {&__pyx_n_s_intervals, __pyx_k_intervals, sizeof(__pyx_k_intervals), 0, 0, 1, 1},
  {&__pyx_n_s_is_contained_in_interval, __pyx_k_is_contained_in_interval, sizeof(__pyx_k_is_contained_in_interval), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_left, __pyx_k_left, sizeof(__pyx_k_left), 0, 0, 1, 1},
  {&__pyx_n_s_level, __pyx_k_level, sizeof(__pyx_k_level), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_members, __pyx_k_members, sizeof(__pyx_k_members), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_ms_deisotope__c_peak_dependency, __pyx_k_ms_deisotope__c_peak_dependency, sizeof(__pyx_k_ms_deisotope__c_peak_dependency), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_node_contains, __pyx_k_node_contains, sizeof(__pyx_k_node_contains), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_organize, __pyx_k_organize, sizeof(__pyx_k_organize), 0, 0, 1, 1},
  {&__pyx_n_s_organizer_callback, __pyx_k_organizer_callback, sizeof(__pyx_k_organizer_callback), 0, 0, 1, 1},
  {&__pyx_n_s_overlap_size, __pyx_k_overlap_size, sizeof(__pyx_k_overlap_size), 0, 0, 1, 1},
  {&__pyx_n_s_overlaps, __pyx_k_overlaps, sizeof(__pyx_k_overlaps), 0, 0, 1, 1},
  {&__pyx_n_s_overlaps_2d, __pyx_k_overlaps_2d, sizeof(__pyx_k_overlaps_2d), 0, 0, 1, 1},
  {&__pyx_n_s_overlaps_interval, __pyx_k_overlaps_interval, sizeof(__pyx_k_overlaps_interval), 0, 0, 1, 1},
  {&__pyx_n_s_overlaps_interval_2d, __pyx_k_overlaps_interval_2d, sizeof(__pyx_k_overlaps_interval_2d), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_parent, __pyx_k_parent, sizeof(__pyx_k_parent), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Interval, __pyx_k_pyx_unpickle_Interval, sizeof(__pyx_k_pyx_unpickle_Interval), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Interval2D, __pyx_k_pyx_unpickle_Interval2D, sizeof(__pyx_k_pyx_unpickle_Interval2D), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_IntervalTreeNode, __pyx_k_pyx_unpickle_IntervalTreeNode, sizeof(__pyx_k_pyx_unpickle_IntervalTreeNode), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_IntervalTreeNode2, __pyx_k_pyx_unpickle_IntervalTreeNode2, sizeof(__pyx_k_pyx_unpickle_IntervalTreeNode2), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_SimpleInterval, __pyx_k_pyx_unpickle_SimpleInterval, sizeof(__pyx_k_pyx_unpickle_SimpleInterval), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_SpanningMixin, __pyx_k_pyx_unpickle_SpanningMixin, sizeof(__pyx_k_pyx_unpickle_SpanningMixin), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_SpanningMixin2D, __pyx_k_pyx_unpickle_SpanningMixin2D, sizeof(__pyx_k_pyx_unpickle_SpanningMixin2D), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_right, __pyx_k_right, sizeof(__pyx_k_right), 0, 0, 1, 1},
  {&__pyx_kp_u_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 1, 0, 0},
  {&__pyx_kp_s_s_start_r_end_r_data_r, __pyx_k_s_start_r_end_r_data_r, sizeof(__pyx_k_s_start_r_end_r_data_r), 0, 0, 1, 0},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self___class_____name___self_st, __pyx_k_self___class_____name___self_st, sizeof(__pyx_k_self___class_____name___self_st), 0, 0, 1, 0},
  {&__pyx_n_s_set_organizer, __pyx_k_set_organizer, sizeof(__pyx_k_set_organizer), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_starts, __pyx_k_starts, sizeof(__pyx_k_starts), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_update_bounds, __pyx_k_update_bounds, sizeof(__pyx_k_update_bounds), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 164, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 386, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 635, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(2, 81, __pyx_L1_error)
  __pyx_builtin_OverflowError = __Pyx_GetBuiltinName(__pyx_n_s_OverflowError); if (!__pyx_builtin_OverflowError) __PYX_ERR(2, 81, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(2, 84, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(2, 94, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(2, 149, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(2, 406, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(2, 615, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xccc3210, 0x0576305, 0x67094f3):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xccc3210, 0x0576305, 0x67094f3) = (end, start))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_tuple_ = PyTuple_Pack(3, __pyx_int_214708752, __pyx_int_5726981, __pyx_int_108041459); if (unlikely(!__pyx_tuple_)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple_);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple_);

#line 4 "(tree fragment)"
  __pyx_tuple__2 = PyTuple_Pack(3, __pyx_int_246211956, __pyx_int_38407067, __pyx_int_80310905); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__2);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__2);

#line 4 "(tree fragment)"
  __pyx_tuple__3 = PyTuple_Pack(3, __pyx_int_40020397, __pyx_int_6463034, __pyx_int_262871197); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__3);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__3);

#line 4 "(tree fragment)"
  __pyx_tuple__4 = PyTuple_Pack(3, __pyx_int_267055226, __pyx_int_44144198, __pyx_int_245896995); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__4);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__4);

#line 4 "(tree fragment)"
  __pyx_tuple__5 = PyTuple_Pack(3, __pyx_int_109638643, __pyx_int_202100921, __pyx_int_116799449); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__5);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__5);

#line 4 "(tree fragment)"
  __pyx_tuple__6 = PyTuple_Pack(3, __pyx_int_7625545, __pyx_int_202473519, __pyx_int_202832017); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__6);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "View.MemoryView":134
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */

#line 134 "View.MemoryView"
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(2, 134, __pyx_L1_error)

#line 134 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__7);

#line 134 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "View.MemoryView":137
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */

#line 137 "View.MemoryView"
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(2, 137, __pyx_L1_error)

#line 137 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__8);

#line 137 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "View.MemoryView":149
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 149 "View.MemoryView"
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(2, 149, __pyx_L1_error)

#line 149 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__9);

#line 149 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "View.MemoryView":177
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */

#line 177 "View.MemoryView"
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(2, 177, __pyx_L1_error)

#line 177 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__10);

#line 177 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "View.MemoryView":193
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */

#line 193 "View.MemoryView"
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(2, 193, __pyx_L1_error)

#line 193 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__11);

#line 193 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__12);

#line 2 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__13);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "View.MemoryView":420
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */

#line 420 "View.MemoryView"
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(2, 420, __pyx_L1_error)

#line 420 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__14);

#line 420 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "View.MemoryView":497
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */

#line 497 "View.MemoryView"
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(2, 497, __pyx_L1_error)

#line 497 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__15);

#line 497 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "View.MemoryView":522
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */

#line 522 "View.MemoryView"
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(2, 522, __pyx_L1_error)

#line 522 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__16);

#line 522 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "View.MemoryView":572
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */

#line 572 "View.MemoryView"
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(2, 572, __pyx_L1_error)

#line 572 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__17);

#line 572 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "View.MemoryView":579
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */

#line 579 "View.MemoryView"
  __pyx_tuple__18 = PyTuple_New(1); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(2, 579, __pyx_L1_error)

#line 579 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__18);

#line 579 "View.MemoryView"
  __Pyx_INCREF(__pyx_int_neg_1);

#line 579 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_int_neg_1);

#line 579 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_tuple__18, 0, __pyx_int_neg_1);

#line 579 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__19);

#line 2 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__20);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "View.MemoryView":684
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */

#line 684 "View.MemoryView"
  __pyx_slice__21 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__21)) __PYX_ERR(2, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
  __Pyx_GOTREF(__pyx_slice__21);

#line 684 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_slice__21);

  /* "View.MemoryView":705
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 705 "View.MemoryView"
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(2, 705, __pyx_L1_error)

#line 705 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__22);

#line 705 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__23);

#line 2 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__24);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__24);

#line 4 "(tree fragment)"
  __pyx_tuple__25 = PyTuple_Pack(3, __pyx_int_184977713, __pyx_int_136983863, __pyx_int_112105877); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__25);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "(tree fragment)":1
 * def __pyx_unpickle_SpanningMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_tuple__31 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__31);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__31);

#line 1 "(tree fragment)"
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_SpanningMixin, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __pyx_tuple__33 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__33);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__33);

#line 1 "(tree fragment)"
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_SimpleInterval, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __pyx_tuple__35 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__35);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__35);

#line 1 "(tree fragment)"
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Interval, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __pyx_tuple__37 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__37);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__37);

#line 1 "(tree fragment)"
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IntervalTreeNode, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __pyx_tuple__39 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__39);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__39);

#line 1 "(tree fragment)"
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_SpanningMixin2D, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __pyx_tuple__41 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__41);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__41);

#line 1 "(tree fragment)"
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Interval2D, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __pyx_tuple__43 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__43);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__43);

#line 1 "(tree fragment)"
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IntervalTreeNode2, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "View.MemoryView":287
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */

#line 287 "View.MemoryView"
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(2, 287, __pyx_L1_error)

#line 287 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__45);

#line 287 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "View.MemoryView":288
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */

#line 288 "View.MemoryView"
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(2, 288, __pyx_L1_error)

#line 288 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__46);

#line 288 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "View.MemoryView":289
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 289 "View.MemoryView"
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(2, 289, __pyx_L1_error)

#line 289 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__47);

#line 289 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "View.MemoryView":292
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */

#line 292 "View.MemoryView"
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(2, 292, __pyx_L1_error)

#line 292 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__48);

#line 292 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "View.MemoryView":293
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 293 "View.MemoryView"
  __pyx_tuple__49 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(2, 293, __pyx_L1_error)

#line 293 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__49);

#line 293 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__49);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_tuple__50 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__50);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__50);

#line 1 "(tree fragment)"
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return 0;

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return -1;

#line 1 "(tree fragment)"
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5726981 = PyInt_FromLong(5726981L); if (unlikely(!__pyx_int_5726981)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6463034 = PyInt_FromLong(6463034L); if (unlikely(!__pyx_int_6463034)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7625545 = PyInt_FromLong(7625545L); if (unlikely(!__pyx_int_7625545)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_38407067 = PyInt_FromLong(38407067L); if (unlikely(!__pyx_int_38407067)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_40020397 = PyInt_FromLong(40020397L); if (unlikely(!__pyx_int_40020397)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_44144198 = PyInt_FromLong(44144198L); if (unlikely(!__pyx_int_44144198)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_80310905 = PyInt_FromLong(80310905L); if (unlikely(!__pyx_int_80310905)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_108041459 = PyInt_FromLong(108041459L); if (unlikely(!__pyx_int_108041459)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_109638643 = PyInt_FromLong(109638643L); if (unlikely(!__pyx_int_109638643)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_112105877 = PyInt_FromLong(112105877L); if (unlikely(!__pyx_int_112105877)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_116799449 = PyInt_FromLong(116799449L); if (unlikely(!__pyx_int_116799449)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_136983863 = PyInt_FromLong(136983863L); if (unlikely(!__pyx_int_136983863)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_202100921 = PyInt_FromLong(202100921L); if (unlikely(!__pyx_int_202100921)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_202473519 = PyInt_FromLong(202473519L); if (unlikely(!__pyx_int_202473519)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_202832017 = PyInt_FromLong(202832017L); if (unlikely(!__pyx_int_202832017)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_214708752 = PyInt_FromLong(214708752L); if (unlikely(!__pyx_int_214708752)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_245896995 = PyInt_FromLong(245896995L); if (unlikely(!__pyx_int_245896995)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_246211956 = PyInt_FromLong(246211956L); if (unlikely(!__pyx_int_246211956)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_262871197 = PyInt_FromLong(262871197L); if (unlikely(!__pyx_int_262871197)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_267055226 = PyInt_FromLong(267055226L); if (unlikely(!__pyx_int_267055226)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin = &__pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin._contains = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, double))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin__contains;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.contains = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.overlaps = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlaps;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.overlap_size = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_overlap_size;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.is_contained_in_interval = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_is_contained_in_interval;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.contains_interval = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_contains_interval;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin, "__contains__"); if (unlikely(!wrapper)) __PYX_ERR(0, 7, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__.doc = __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13SpanningMixin_2__contains__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SpanningMixin, (PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin = &__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval = &__pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_base = __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Interval, (PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval = &__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval;
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval = &__pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval._create = (struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval *(*)(double, double))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_14SimpleInterval__create;
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval.tp_base = __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SimpleInterval, (PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval = &__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SimpleInterval;
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode = &__pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.contains_point = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_contains_point;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode._overlaps_interval = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__overlaps_interval;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.overlaps = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, double, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode_overlaps;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode._eq = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_16IntervalTreeNode__eq;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IntervalTreeNode, (PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode = &__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D = &__pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D._contains = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, double *))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D__contains;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D.contains = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, __Pyx_memviewslice, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_contains;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D.overlaps = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_overlaps;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D.contains_interval = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_contains_interval;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D) < 0) __PYX_ERR(0, 472, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D, "__contains__"); if (unlikely(!wrapper)) __PYX_ERR(0, 472, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_2__contains__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_2__contains__.doc = __pyx_doc_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_2__contains__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15SpanningMixin2D_2__contains__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D) < 0) __PYX_ERR(0, 472, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SpanningMixin2D, (PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D) < 0) __PYX_ERR(0, 472, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D) < 0) __PYX_ERR(0, 472, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D = &__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D;
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D = &__pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D;
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D.tp_base = __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_SpanningMixin2D;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D) < 0) __PYX_ERR(0, 595, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D) < 0) __PYX_ERR(0, 595, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Interval2D, (PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D) < 0) __PYX_ERR(0, 595, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D) < 0) __PYX_ERR(0, 595, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D = &__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_Interval2D;
  __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D = &__pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D._overlaps_interval_2d = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *, __Pyx_memviewslice, __Pyx_memviewslice, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D__overlaps_interval_2d;
  __pyx_vtable_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D.overlaps_2d = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D *, __Pyx_memviewslice, __Pyx_memviewslice, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_23peak_dependency_network_9intervals_18IntervalTreeNode2D_overlaps_2d;
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D.tp_base = __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D) < 0) __PYX_ERR(0, 633, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D) < 0) __PYX_ERR(0, 633, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IntervalTreeNode2D, (PyObject *)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D) < 0) __PYX_ERR(0, 633, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D) < 0) __PYX_ERR(0, 633, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D = &__pyx_type_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D;
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(2, 106, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(2, 106, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(2, 106, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(2, 280, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(2, 280, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(2, 331, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(2, 331, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(2, 331, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(2, 967, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(2, 967, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(2, 967, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initintervals(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initintervals(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_intervals(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_intervals(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_intervals(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  static PyThread_type_lock __pyx_t_4[8];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'intervals' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_intervals(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("intervals", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_deisotope___c__peak_dependency_network__intervals) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_deisotope._c.peak_dependency_network.intervals")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_deisotope._c.peak_dependency_network.intervals", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":4
 * from cpython.tuple cimport PyTuple_GetItem
 * 
 * from collections import deque             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_INCREF(__pyx_n_s_deque);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GIVEREF(__pyx_n_s_deque);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_deque);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_deque); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_deque, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":171
 * 
 * 
 * cdef double INF = float('inf')             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 171 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_3 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_3 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_v_12ms_deisotope_2_c_23peak_dependency_network_9intervals_INF = __pyx_t_3;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":304
 * 
 *     @classmethod
 *     def build(cls, intervals):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalTreeNode root, parent, node, up
 */

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode, __pyx_n_s_build); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 304, __pyx_L1_error)

#line 304 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":303
 *         return hash((self.start, self.center, self.right, self.level))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def build(cls, intervals):
 *         cdef:
 */

#line 303 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_1 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L1_error)

#line 303 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 303 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 303 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (PyDict_SetItem((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode->tp_dict, __pyx_n_s_build, __pyx_t_1) < 0) __PYX_ERR(0, 304, __pyx_L1_error)

#line 303 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 303 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyType_Modified(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":711
 * 
 *     @classmethod
 *     def build(cls, intervals, organizer_callback=None):             # <<<<<<<<<<<<<<
 *         root: IntervalTreeNode2D = super(IntervalTreeNode2D, cls).build(intervals)
 *         root.set_organizer(organizer_callback)
 */

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GetNameInClass(__pyx_t_1, (PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D, __pyx_n_s_build); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)

#line 711 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_1);

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":710
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def build(cls, intervals, organizer_callback=None):
 *         root: IntervalTreeNode2D = super(IntervalTreeNode2D, cls).build(intervals)
 */

#line 710 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 710, __pyx_L1_error)

#line 710 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 710 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 710 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (PyDict_SetItem((PyObject *)__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D->tp_dict, __pyx_n_s_build, __pyx_t_2) < 0) __PYX_ERR(0, 711, __pyx_L1_error)

#line 710 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 710 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  PyType_Modified(__pyx_ptype_12ms_deisotope_2_c_23peak_dependency_network_9intervals_IntervalTreeNode2D);

  /* "(tree fragment)":1
 * def __pyx_unpickle_SpanningMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_5__pyx_unpickle_SpanningMixin, NULL, __pyx_n_s_ms_deisotope__c_peak_dependency); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_SpanningMixin, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_SpanningMixin__set_state(<SpanningMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin__set_state(SpanningMixin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_7__pyx_unpickle_SimpleInterval, NULL, __pyx_n_s_ms_deisotope__c_peak_dependency); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 11 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_SimpleInterval, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Interval(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_9__pyx_unpickle_Interval, NULL, __pyx_n_s_ms_deisotope__c_peak_dependency); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Interval, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Interval__set_state(<Interval> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Interval__set_state(Interval __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end = __pyx_state[1]; __pyx_result.members = __pyx_state[2]; __pyx_result.start = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_11__pyx_unpickle_IntervalTreeNode, NULL, __pyx_n_s_ms_deisotope__c_peak_dependency); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 11 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IntervalTreeNode, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_SpanningMixin2D(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_13__pyx_unpickle_SpanningMixin2D, NULL, __pyx_n_s_ms_deisotope__c_peak_dependency); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_SpanningMixin2D, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_SpanningMixin2D__set_state(<SpanningMixin2D> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SpanningMixin2D__set_state(SpanningMixin2D __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._end = __pyx_state[0]; __pyx_result._start = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_15__pyx_unpickle_Interval2D, NULL, __pyx_n_s_ms_deisotope__c_peak_dependency); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 11 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Interval2D, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IntervalTreeNode2D(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_23peak_dependency_network_9intervals_17__pyx_unpickle_IntervalTreeNode2D, NULL, __pyx_n_s_ms_deisotope__c_peak_dependency); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IntervalTreeNode2, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/peak_dependency_network/intervals.pyx":1
 * from cpython.list cimport PyList_GetItem, PyList_Size             # <<<<<<<<<<<<<<
 * from cpython.tuple cimport PyTuple_GetItem
 * 
 */

#line 1 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 1 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/peak_dependency_network/intervals.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":210
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */

#line 210 "View.MemoryView"
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 210, __pyx_L1_error)

#line 210 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 210 "View.MemoryView"
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(2, 210, __pyx_L1_error)

#line 210 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 210 "View.MemoryView"
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":287
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */

#line 287 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 287, __pyx_L1_error)

#line 287 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 287 "View.MemoryView"
  __Pyx_XGOTREF(generic);

#line 287 "View.MemoryView"
  __Pyx_DECREF_SET(generic, __pyx_t_2);

#line 287 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 287 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":288
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */

#line 288 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 288, __pyx_L1_error)

#line 288 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 288 "View.MemoryView"
  __Pyx_XGOTREF(strided);

#line 288 "View.MemoryView"
  __Pyx_DECREF_SET(strided, __pyx_t_2);

#line 288 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 288 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":289
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 289 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 289, __pyx_L1_error)

#line 289 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 289 "View.MemoryView"
  __Pyx_XGOTREF(indirect);

#line 289 "View.MemoryView"
  __Pyx_DECREF_SET(indirect, __pyx_t_2);

#line 289 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 289 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":292
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */

#line 292 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 292, __pyx_L1_error)

#line 292 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 292 "View.MemoryView"
  __Pyx_XGOTREF(contiguous);

#line 292 "View.MemoryView"
  __Pyx_DECREF_SET(contiguous, __pyx_t_2);

#line 292 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 292 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":293
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 293 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__49, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 293, __pyx_L1_error)

#line 293 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 293 "View.MemoryView"
  __Pyx_XGOTREF(indirect_contiguous);

#line 293 "View.MemoryView"
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_2);

#line 293 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 293 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":317
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */

#line 317 "View.MemoryView"
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":318
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */

#line 318 "View.MemoryView"
  __pyx_t_4[0] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[1] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[2] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[3] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[4] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[5] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[6] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[7] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_4, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":551
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 551 "View.MemoryView"
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 551, __pyx_L1_error)

#line 551 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 551 "View.MemoryView"
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(2, 551, __pyx_L1_error)

#line 551 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 551 "View.MemoryView"
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":997
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 997 "View.MemoryView"
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 997, __pyx_L1_error)

#line 997 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 997 "View.MemoryView"
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(2, 997, __pyx_L1_error)

#line 997 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 997 "View.MemoryView"
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "BufferFormatFromTypeInfo":1465
 * 
 * @cname('__pyx_format_from_typeinfo')
 * cdef bytes format_from_typeinfo(__Pyx_TypeInfo *type):             # <<<<<<<<<<<<<<
 *     cdef __Pyx_StructField *field
 *     cdef __pyx_typeinfo_string fmt
 */

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
  /*--- Wrapped vars code ---*/

#line 1465 "BufferFormatFromTypeInfo"


#line 1465 "BufferFormatFromTypeInfo"
  goto __pyx_L0;

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_L1_error:;

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_XDECREF(__pyx_t_1);

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_XDECREF(__pyx_t_2);

#line 1465 "BufferFormatFromTypeInfo"
  if (__pyx_m) {

#line 1465 "BufferFormatFromTypeInfo"
    if (__pyx_d) {

#line 1465 "BufferFormatFromTypeInfo"
      __Pyx_AddTraceback("init ms_deisotope._c.peak_dependency_network.intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1465 "BufferFormatFromTypeInfo"
    }

#line 1465 "BufferFormatFromTypeInfo"
    Py_CLEAR(__pyx_m);

#line 1465 "BufferFormatFromTypeInfo"
  } else if (!PyErr_Occurred()) {

#line 1465 "BufferFormatFromTypeInfo"
    PyErr_SetString(PyExc_ImportError, "init ms_deisotope._c.peak_dependency_network.intervals");

#line 1465 "BufferFormatFromTypeInfo"
  }

#line 1465 "BufferFormatFromTypeInfo"
  __pyx_L0:;

#line 1465 "BufferFormatFromTypeInfo"
  __Pyx_RefNannyFinishContext();

#line 1465 "BufferFormatFromTypeInfo"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1465 "BufferFormatFromTypeInfo"
  return (__pyx_m != NULL) ? 0 : -1;

#line 1465 "BufferFormatFromTypeInfo"
  #elif PY_MAJOR_VERSION >= 3

#line 1465 "BufferFormatFromTypeInfo"
  return __pyx_m;

#line 1465 "BufferFormatFromTypeInfo"
  #else

#line 1465 "BufferFormatFromTypeInfo"
  return;

#line 1465 "BufferFormatFromTypeInfo"
  #endif

#line 1465 "BufferFormatFromTypeInfo"
}

#line 1 "cython_utility"

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (Py_TYPE(L) == &PySet_Type) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* MemviewSliceInit */
static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#if PY_VERSION_HEX >= 0x030A0000 || defined(HAVE_STDARG_PROTOTYPES)
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None))
        return;
    if (unlikely(__pyx_get_slice_count(memview) < 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (unlikely(first_time)) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    if (unlikely(__pyx_get_slice_count(memview) <= 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (unlikely(last_time)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* BufferIndexError */
static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* ModInt[long] */
static CYTHON_INLINE long __Pyx_mod_long(long a, long b) {
    long r = a % b;
    r += ((r != 0) & ((r ^ b) < 0)) * b;
    return r;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* DivInt[Py_ssize_t] */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t a, Py_ssize_t b) {
    Py_ssize_t q = a / b;
    Py_ssize_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* DivInt[long] */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* StringJoin */
#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* pyobject_as_double */
static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
    float_value = PyNumber_Float(obj);  if ((0)) goto bad;
#else
    PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
    if (likely(nb) && likely(nb->nb_float)) {
        float_value = nb->nb_float(obj);
        if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
            PyErr_Format(PyExc_TypeError,
                "__float__ returned non-float (type %.200s)",
                Py_TYPE(float_value)->tp_name);
            Py_DECREF(float_value);
            goto bad;
        }
    } else if (PyUnicode_CheckExact(obj) || PyBytes_CheckExact(obj)) {
#if PY_MAJOR_VERSION >= 3
        float_value = PyFloat_FromString(obj);
#else
        float_value = PyFloat_FromString(obj, 0);
#endif
    } else {
        PyObject* args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        PyTuple_SET_ITEM(args, 0, obj);
        float_value = PyObject_Call((PyObject*)&PyFloat_Type, args, 0);
        PyTuple_SET_ITEM(args, 0, 0);
        Py_DECREF(args);
    }
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
bad:
    return (double)-1;
}

/* ClassMethod */
static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYSTON || CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    #if PY_MAJOR_VERSION == 2
    static PyTypeObject *methoddescr_type = NULL;
    if (methoddescr_type == NULL) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (!meth) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    #else
    PyTypeObject *methoddescr_type = &PyMethodDescr_Type;
    #endif
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else {
        return PyClassMethod_New(method);
    }
}

/* GetNameInClass */
static PyObject *__Pyx_GetGlobalNameAfterAttributeLookup(PyObject *name) {
    PyObject *result;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result) {
        result = __Pyx_GetGlobalNameAfterAttributeLookup(name);
    }
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


/* MemviewSliceIsContig */
static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* TypeInfoCompare */
  static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
  static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (unlikely(buf->strides[dim] != sizeof(void *))) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (unlikely(buf->strides[dim] != buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (unlikely(stride < buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (unlikely(spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(spec & (__Pyx_MEMVIEW_PTR))) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(buf->suboffsets)) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, CYTHON_UNUSED int ndim, int spec)
{
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (unlikely(buf->suboffsets && buf->suboffsets[dim] >= 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (unlikely(!buf->suboffsets || (buf->suboffsets[dim] < 0))) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (unlikely(buf->ndim != ndim)) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (unlikely(!__Pyx_BufFmt_CheckString(&ctx, buf->format))) goto fail;
    }
    if (unlikely((unsigned) buf->itemsize != dtype->size)) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    if (buf->len > 0) {
        for (i = 0; i < ndim; i++) {
            spec = axes_specs[i];
            if (unlikely(!__pyx_check_strides(buf, i, ndim, spec)))
                goto fail;
            if (unlikely(!__pyx_check_suboffsets(buf, i, ndim, spec)))
                goto fail;
        }
        if (unlikely(buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag)))
            goto fail;
    }
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 1,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_double(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(double *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_double(const char *itemp, PyObject *obj) {
    double value = __pyx_PyFloat_AsDouble(obj);
    if ((value == (double)-1) && PyErr_Occurred())
        return 0;
    *(double *) itemp = value;
    return 1;
}

/* MemviewSliceCopyTemplate */
  static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* TypeInfoToFormat */
  static struct __pyx_typeinfo_string __Pyx_TypeInfoToFormat(__Pyx_TypeInfo *type) {
    struct __pyx_typeinfo_string result = { {0} };
    char *buf = (char *) result.string;
    size_t size = type->size;
    switch (type->typegroup) {
        case 'H':
            *buf = 'c';
            break;
        case 'I':
        case 'U':
            if (size == 1)
                *buf = (type->is_unsigned) ? 'B' : 'b';
            else if (size == 2)
                *buf = (type->is_unsigned) ? 'H' : 'h';
            else if (size == 4)
                *buf = (type->is_unsigned) ? 'I' : 'i';
            else if (size == 8)
                *buf = (type->is_unsigned) ? 'Q' : 'q';
            break;
        case 'P':
            *buf = 'P';
            break;
        case 'C':
         {
            __Pyx_TypeInfo complex_type = *type;
            complex_type.typegroup = 'R';
            complex_type.size /= 2;
            *buf++ = 'Z';
            *buf = __Pyx_TypeInfoToFormat(&complex_type).string[0];
            break;
         }
        case 'R':
            if (size == 4)
                *buf = 'f';
            else if (size == 8)
                *buf = 'd';
            else
                *buf = 'g';
            break;
    }
    return result;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE short __Pyx_PyInt_As_short(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const short neg_one = (short) -1, const_zero = (short) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(short) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(short, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (short) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (short) 0;
                case  1: __PYX_VERIFY_RETURN_INT(short, digit, digits[0])
                case 2:
                    if (8 * sizeof(short) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) >= 2 * PyLong_SHIFT) {
                            return (short) (((((short)digits[1]) << PyLong_SHIFT) | (short)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(short) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) >= 3 * PyLong_SHIFT) {
                            return (short) (((((((short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(short) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) >= 4 * PyLong_SHIFT) {
                            return (short) (((((((((short)digits[3]) << PyLong_SHIFT) | (short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (short) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(short) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(short) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (short) 0;
                case -1: __PYX_VERIFY_RETURN_INT(short, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(short,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(short) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 2 * PyLong_SHIFT) {
                            return (short) (((short)-1)*(((((short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(short) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 2 * PyLong_SHIFT) {
                            return (short) ((((((short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(short) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 3 * PyLong_SHIFT) {
                            return (short) (((short)-1)*(((((((short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(short) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 3 * PyLong_SHIFT) {
                            return (short) ((((((((short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(short) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 4 * PyLong_SHIFT) {
                            return (short) (((short)-1)*(((((((((short)digits[3]) << PyLong_SHIFT) | (short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(short) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 4 * PyLong_SHIFT) {
                            return (short) ((((((((((short)digits[3]) << PyLong_SHIFT) | (short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(short) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(short) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            short val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (short) -1;
        }
    } else {
        short val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (short) -1;
        val = __Pyx_PyInt_As_short(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to short");
    return (short) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to short");
    return (short) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
