/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/npy_math.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/ms_peak_picker/_c",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/brainpy/_c"
        ],
        "name": "ms_deisotope._c.feature_map.feature_processor",
        "sources": [
            "src/ms_deisotope/_c/feature_map/feature_processor.pyx"
        ]
    },
    "module_name": "ms_deisotope._c.feature_map.feature_processor"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_deisotope___c__feature_map__feature_processor
#define __PYX_HAVE_API__ms_deisotope___c__feature_map__feature_processor
/* Early includes */
#include <string.h>
#include <stdlib.h>
#include "numpy/npy_math.h"
#include <stdio.h>
#include "pythread.h"
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "src/ms_deisotope/_c/feature_map/feature_processor.pyx",
  "stringsource",
  "__init__.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "peak_set.pxd",
  "composition.pxd",
  "isotopic_distribution.pxd",
  "peak_index.pxd",
  "src/ms_deisotope/_c/averagine.pxd",
  "src/ms_deisotope/_c/deconvoluter_base.pxd",
  "src/ms_deisotope/_c/scoring.pxd",
  "src/ms_deisotope/_c/feature_map/lcms_feature.pxd",
  "src/ms_deisotope/_c/feature_map/feature_map.pxd",
  "src/ms_deisotope/_c/peak_set.pxd",
  "src/ms_deisotope/_c/feature_map/feature_fit.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "brainpy/_c/composition.pxd":3
 * from brainpy.mass_dict import nist_mass as __nist_mass
 * 
 * ctypedef long count_type             # <<<<<<<<<<<<<<
 * 
 * cdef dict nist_mass
 */

#line 3 "brainpy/_c/composition.pxd"
typedef long __pyx_t_7brainpy_2_c_11composition_count_type;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */

#line 689 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */

#line 690 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */

#line 691 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */

#line 692 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */

#line 696 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */

#line 697 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */

#line 698 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */

#line 699 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */

#line 703 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */

#line 704 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */

#line 713 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */

#line 714 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */

#line 715 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */

#line 717 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */

#line 718 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */

#line 719 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */

#line 721 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */

#line 722 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */

#line 724 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */

#line 725 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */

#line 726 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition;
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted;

/* "ms_peak_picker/_c/peak_set.pxd":42
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout)
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self)
 */

#line 42 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak {

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":50
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak getitem(self, size_t i)
 */

#line 50 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for {

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  double error_tolerance;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":64
 * 
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */

#line 64 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell {

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  double center_value;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t start;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t end;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":70
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */

#line 70 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list {

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell *index;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  size_t size;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double low;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double high;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":90
 *     cpdef _allocate_index(self, size_t interval_index_size)
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=*, size_t hint_start=*, size_t hint_end=*)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n)
 */

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted {

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_start;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_end;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
};

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Isotope;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Element;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Composition;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop;

/* "brainpy/_c/composition.pxd":10
 * cdef double PROTON = nist_mass["H+"][0][0]
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 */

#line 10 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass {

#line 10 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 10 "brainpy/_c/composition.pxd"
  double charge_carrier;

#line 10 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":11
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 * 
 */

#line 11 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio {

#line 11 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 11 "brainpy/_c/composition.pxd"
  double charge_carrier;

#line 11 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":17
 * # Isotope and IsotopeMap Declarations
 * 
 * cdef struct Isotope:             # <<<<<<<<<<<<<<
 *     double mass
 *     double abundance
 */

#line 17 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Isotope {

#line 17 "brainpy/_c/composition.pxd"
  double mass;

#line 17 "brainpy/_c/composition.pxd"
  double abundance;

#line 17 "brainpy/_c/composition.pxd"
  int neutrons;

#line 17 "brainpy/_c/composition.pxd"
  int neutron_shift;

#line 17 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":23
 *     int neutron_shift
 * 
 * cdef struct IsotopeMap:             # <<<<<<<<<<<<<<
 *     Isotope* bins
 *     size_t size
 */

#line 23 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap {

#line 23 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *bins;

#line 23 "brainpy/_c/composition.pxd"
  size_t size;

#line 23 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":36
 * # Element Declarations
 * 
 * cdef struct Element:             # <<<<<<<<<<<<<<
 *     char* symbol
 *     IsotopeMap* isotopes
 */

#line 36 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Element {

#line 36 "brainpy/_c/composition.pxd"
  char *symbol;

#line 36 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *isotopes;

#line 36 "brainpy/_c/composition.pxd"
  int monoisotopic_isotope_index;

#line 36 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":56
 * # ElementHashTable and ElementHashBucket Declarations
 * 
 * cdef struct ElementHashBucket:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */

#line 56 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket {

#line 56 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;

#line 56 "brainpy/_c/composition.pxd"
  size_t used;

#line 56 "brainpy/_c/composition.pxd"
  size_t size;

#line 56 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":65
 * 
 * 
 * cdef struct ElementHashTable:             # <<<<<<<<<<<<<<
 *     ElementHashBucket* buckets
 *     size_t size
 */

#line 65 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable {

#line 65 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *buckets;

#line 65 "brainpy/_c/composition.pxd"
  size_t size;

#line 65 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":94
 * # Composition Declarations
 * 
 * cdef struct Composition:             # <<<<<<<<<<<<<<
 *     char** elements
 *     count_type* counts
 */

#line 94 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Composition {

#line 94 "brainpy/_c/composition.pxd"
  char **elements;

#line 94 "brainpy/_c/composition.pxd"
  __pyx_t_7brainpy_2_c_11composition_count_type *counts;

#line 94 "brainpy/_c/composition.pxd"
  size_t size;

#line 94 "brainpy/_c/composition.pxd"
  size_t used;

#line 94 "brainpy/_c/composition.pxd"
  int max_variants;

#line 94 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":130
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base)
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _initialize_from_formula(self, str formula)
 */

#line 130 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl {

#line 130 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 130 "brainpy/_c/composition.pxd"
  int free_existing;

#line 130 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":144
 *     cpdef list items(self)
 * 
 *     cpdef pop(self, str key, object default=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef count_type getitem(self, str key)
 */

#line 144 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop {

#line 144 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 144 "brainpy/_c/composition.pxd"
  PyObject *__pyx_default;

#line 144 "brainpy/_c/composition.pxd"
};

#line 144 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector;

/* "brainpy/_c/double_vector.pxd":1
 * cdef struct DoubleVector:             # <<<<<<<<<<<<<<
 *     double* v
 *     size_t used
 */

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector {

#line 1 "brainpy/_c/double_vector.pxd"
  double *v;

#line 1 "brainpy/_c/double_vector.pxd"
  size_t used;

#line 1 "brainpy/_c/double_vector.pxd"
  size_t size;

#line 1 "brainpy/_c/double_vector.pxd"
};

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters;

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants;

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants;

/* "brainpy/_c/isotopic_constants.pxd":6
 * from brainpy._c.double_vector cimport DoubleVector
 * 
 * ctypedef DoubleVector dvec             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 6 "brainpy/_c/isotopic_constants.pxd"
typedef struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector __pyx_t_7brainpy_2_c_18isotopic_constants_dvec;

/* "brainpy/_c/isotopic_constants.pxd":9
 * 
 * 
 * cdef struct PolynomialParameters:             # <<<<<<<<<<<<<<
 *     dvec* elementary_symmetric_polynomial
 *     dvec* power_sum
 */

#line 9 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters {

#line 9 "brainpy/_c/isotopic_constants.pxd"
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *elementary_symmetric_polynomial;

#line 9 "brainpy/_c/isotopic_constants.pxd"
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *power_sum;

#line 9 "brainpy/_c/isotopic_constants.pxd"
};

/* "brainpy/_c/isotopic_constants.pxd":14
 * 
 * 
 * cdef struct PhiConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     Element* element
 */

#line 14 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants {

#line 14 "brainpy/_c/isotopic_constants.pxd"
  int order;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element *element;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *element_coefficients;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *mass_coefficients;

#line 14 "brainpy/_c/isotopic_constants.pxd"
};

/* "brainpy/_c/isotopic_constants.pxd":21
 * 
 * 
 * cdef struct IsotopicConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     PhiConstants** constants
 */

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants {

#line 21 "brainpy/_c/isotopic_constants.pxd"
  int order;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants **constants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  size_t size;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  size_t used;

#line 21 "brainpy/_c/isotopic_constants.pxd"
};

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx;

/* "brainpy/_c/isotopic_distribution.pxd":7
 * 
 * 
 * cdef struct Peak:             # <<<<<<<<<<<<<<
 *     double mz
 *     double intensity
 */

#line 7 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak {

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  double mz;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  double intensity;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":13
 * 
 * 
 * cdef struct PeakList:             # <<<<<<<<<<<<<<
 *     Peak* peaks
 *     size_t used
 */

#line 13 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList {

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *peaks;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  size_t used;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  size_t size;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":35
 * 
 * 
 * cdef struct ElementCache:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */

#line 35 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache {

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  size_t used;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  size_t size;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *source;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":49
 * 
 * 
 * cdef struct IsotopicDistribution:             # <<<<<<<<<<<<<<
 *     Composition* composition
 *     IsotopicConstants* _isotopic_constants
 */

#line 49 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution {

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Composition *composition;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *_isotopic_constants;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  int order;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  double average_mass;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *monoisotopic_peak;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":58
 * 
 * 
 * cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil
 * 
 */

#line 58 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution {

#line 58 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 58 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 58 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":77
 * 
 * cdef int max_variants(Composition* composition, ElementCache* cache) nogil
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)
 */

#line 77 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks {

#line 77 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 77 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 77 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":79
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 79 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants {

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  PyObject *npeaks;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  double charge_carrier;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":82
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=*, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 82 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants {

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  double charge_carrier;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":85
 * 
 * 
 * cdef size_t max_variants_approx(double mass, double lambda_factor=*, size_t maxiter=*, double threshold=*) nogil             # <<<<<<<<<<<<<<
 */

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx {

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  double lambda_factor;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  size_t maxiter;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  double threshold;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
};

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */

#line 728 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */

#line 729 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */

#line 730 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_complex_t;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for;

/* "ms_peak_picker/_c/peak_index.pxd":11
 *         public PeakSet peaks
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 */

#line 11 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak {

#line 11 "ms_peak_picker/_c/peak_index.pxd"
  int __pyx_n;

#line 11 "ms_peak_picker/_c/peak_index.pxd"
  double tolerance;

#line 11 "ms_peak_picker/_c/peak_index.pxd"
};

/* "ms_peak_picker/_c/peak_index.pxd":14
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self)
 */

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for {

#line 14 "ms_peak_picker/_c/peak_index.pxd"
  int __pyx_n;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
  double error_tolerance;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
};

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate;

/* "ms_deisotope/_c/averagine.pxd":7
 * cdef double PROTON
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)
 * cdef void slide(double mz, list peaklist)
 */

#line 7 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio {

#line 7 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 7 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 7 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":8
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef void slide(double mz, list peaklist)
 * cdef dict scale_dict(dict data, double factor)
 */

#line 8 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass {

#line 8 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 8 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 8 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":17
 *         public dict base_composition
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 */

#line 17 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale {

#line 17 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 17 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 17 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 17 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":18
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */

#line 18 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster {

#line 18 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 18 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 18 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":19
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 19 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster {

#line 19 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 19 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 19 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":39
 *     cdef bint _eq_inst(self, TheoreticalIsotopicPattern other)
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 */

#line 39 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below {

#line 39 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 39 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 39 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":40
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)
 */

#line 40 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after {

#line 40 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 40 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 40 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":42
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern scale_raw(self, double scale_factor)
 *     cpdef double drop_last_peak(self)
 */

#line 42 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale {

#line 42 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 42 "ms_deisotope/_c/averagine.pxd"
  PyObject *method;

#line 42 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":53
 * 
 *     @cython.final
 *     cdef inline TheoreticalIsotopicPattern _scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 * 
 *     @cython.final
 */

#line 53 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale {

#line 53 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 53 "ms_deisotope/_c/averagine.pxd"
  PyObject *method;

#line 53 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":88
 *     cdef double _make_cache_key(self, double mz)
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */

#line 88 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair {

#line 88 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 88 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 88 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":89
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 89 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster {

#line 89 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 89 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 89 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":92
 * 
 * 
 * cpdef double isotopic_shift(int charge=*)             # <<<<<<<<<<<<<<
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)
 */

#line 92 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift {

#line 92 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 92 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 92 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":94
 * cpdef double isotopic_shift(int charge=*)
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)             # <<<<<<<<<<<<<<
 */

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate {

#line 94 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 94 "ms_deisotope/_c/averagine.pxd"
  double lambda_factor;

#line 94 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 94 "ms_deisotope/_c/averagine.pxd"
};

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs;

/* "ms_deisotope/_c/deconvoluter_base.pxd":42
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance)
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 */

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution {

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":45
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 *     cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound)
 * 
 */

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction {

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":50
 *     cpdef bint _check_fit(self, IsotopicFitRecord fit)
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 */

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak {

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":51
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 */

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak {

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":53
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 * 
 */

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace {

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":54
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace {

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":61
 *         public AveragineCache averagine
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution {

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":65
 *                                                          double ignore_below=*)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges {

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":73
 *         public list averagines
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution {

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":76
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges {

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":83
 * 
 * 
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=*,             # <<<<<<<<<<<<<<
 *                                      object charge_range=*,
 *                                      int left_search_limit=*, int right_search_limit=*,
 */

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs {

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *charge_range;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int left_search_limit;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int right_search_limit;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int recalculate_starting_peak;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int use_quick_charge;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference;

/* "ms_deisotope/_c/scoring.pxd":29
 * 
 *     @staticmethod
 *     cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,             # <<<<<<<<<<<<<<
 *                                    list experimental, object data=*, int missed_peaks=*)
 * 
 */

#line 29 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create {

#line 29 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 29 "ms_deisotope/_c/scoring.pxd"
  PyObject *data;

#line 29 "ms_deisotope/_c/scoring.pxd"
  int missed_peaks;

#line 29 "ms_deisotope/_c/scoring.pxd"
};

/* "ms_deisotope/_c/scoring.pxd":75
 *     cdef public double mass_error_tolerance
 * 
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=*, double minimum_signal_to_noise=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 75 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak {

#line 75 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 75 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 75 "ms_deisotope/_c/scoring.pxd"
  double minimum_signal_to_noise;

#line 75 "ms_deisotope/_c/scoring.pxd"
};

/* "ms_deisotope/_c/scoring.pxd":92
 *     cdef public PeakSet peaklist
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=*, double upper=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference {

#line 92 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 92 "ms_deisotope/_c/scoring.pxd"
  double lower;

#line 92 "ms_deisotope/_c/scoring.pxd"
  double upper;

#line 92 "ms_deisotope/_c/scoring.pxd"
};

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone;

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":98
 *     cpdef insert_node(self, LCMSFeatureTreeNode node)
 *     cpdef insert(self, PeakBase peak, double time)
 *     cpdef _invalidate(self, bint reaverage=*)             # <<<<<<<<<<<<<<
 *     cpdef list split_sparse(self, double delta_rt=*)
 *     cpdef LCMSFeature clone(self, deep=*, cls=*)
 */

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate {

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_n;

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int reaverage;

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":99
 *     cpdef insert(self, PeakBase peak, double time)
 *     cpdef _invalidate(self, bint reaverage=*)
 *     cpdef list split_sparse(self, double delta_rt=*)             # <<<<<<<<<<<<<<
 *     cpdef LCMSFeature clone(self, deep=*, cls=*)
 *     cpdef double max_intensity(self)
 */

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse {

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_n;

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double delta_rt;

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":100
 *     cpdef _invalidate(self, bint reaverage=*)
 *     cpdef list split_sparse(self, double delta_rt=*)
 *     cpdef LCMSFeature clone(self, deep=*, cls=*)             # <<<<<<<<<<<<<<
 *     cpdef double max_intensity(self)
 * 
 */

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone {

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_n;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *deep;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *cls;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse;

/* "ms_deisotope/_c/feature_map/feature_map.pxd":11
 *     cdef Py_ssize_t get_size(self)
 *     cdef LCMSFeature get(self, size_t i)
 *     cpdef list find_all(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef LCMSFeature search(self, double mz, double error_tolerance=*)
 *     cpdef list between(self, double lo, double hi, double error_tolerance=*)
 */

#line 11 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all {

#line 11 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_n;

#line 11 "ms_deisotope/_c/feature_map/feature_map.pxd"
  double error_tolerance;

#line 11 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

/* "ms_deisotope/_c/feature_map/feature_map.pxd":12
 *     cdef LCMSFeature get(self, size_t i)
 *     cpdef list find_all(self, double mz, double error_tolerance=*)
 *     cpdef LCMSFeature search(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list between(self, double lo, double hi, double error_tolerance=*)
 *     cpdef list spanning_time(self, double time_point)
 */

#line 12 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search {

#line 12 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_n;

#line 12 "ms_deisotope/_c/feature_map/feature_map.pxd"
  double error_tolerance;

#line 12 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

/* "ms_deisotope/_c/feature_map/feature_map.pxd":13
 *     cpdef list find_all(self, double mz, double error_tolerance=*)
 *     cpdef LCMSFeature search(self, double mz, double error_tolerance=*)
 *     cpdef list between(self, double lo, double hi, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list spanning_time(self, double time_point)
 *     cpdef LCMSFeatureMap clone(self, bint deep=*)
 */

#line 13 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between {

#line 13 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_n;

#line 13 "ms_deisotope/_c/feature_map/feature_map.pxd"
  double error_tolerance;

#line 13 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

/* "ms_deisotope/_c/feature_map/feature_map.pxd":15
 *     cpdef list between(self, double lo, double hi, double error_tolerance=*)
 *     cpdef list spanning_time(self, double time_point)
 *     cpdef LCMSFeatureMap clone(self, bint deep=*)             # <<<<<<<<<<<<<<
 * 
 * cpdef tuple binary_search_with_flag(list array, double mz, double error_tolerance)
 */

#line 15 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone {

#line 15 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_n;

#line 15 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int deep;

#line 15 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

/* "ms_deisotope/_c/feature_map/feature_map.pxd":21
 * cdef void search_sweep(list array, double mz, double error_tolerance, long* loout, long* hiout)
 * 
 * cpdef list split_sparse(LCMSFeatureMap self, double delta_rt=*, size_t min_size=*)             # <<<<<<<<<<<<<<
 */

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse {

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_n;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
  double delta_rt;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
  size_t min_size;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_all_peaks_for;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_has_peak;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for;

/* "ms_deisotope/_c/peak_set.pxd":138
 * 
 * 
 * cdef enum CPeaksFlags:             # <<<<<<<<<<<<<<
 *     borrowing = 0
 *     owns_peaks = 1
 */

#line 138 "ms_deisotope/_c/peak_set.pxd"
enum __pyx_t_12ms_deisotope_2_c_8peak_set_CPeaksFlags {

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_borrowing = 0,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_peaks = 1,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_index = 2,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_peaks_and_index = 3

#line 138 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":87
 *     cpdef DeconvolutedPeakSet copy(self)
 * 
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 */

#line 87 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak {

#line 87 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 87 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 87 "ms_deisotope/_c/peak_set.pxd"
  int use_mz;

#line 87 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":89
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)
 */

#line 89 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak {

#line 89 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 89 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 89 "ms_deisotope/_c/peak_set.pxd"
  int use_mz;

#line 89 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":91
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef DeconvolutedPeak _get_nearest_peak(self, double neutral_mass, double* errout)
 */

#line 91 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for {

#line 91 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 91 "ms_deisotope/_c/peak_set.pxd"
  double tolerance;

#line 91 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":118
 * cdef size_t INTERVAL_INDEX_SIZE
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */

#line 118 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell {

#line 118 "ms_deisotope/_c/peak_set.pxd"
  double center_value;

#line 118 "ms_deisotope/_c/peak_set.pxd"
  size_t start;

#line 118 "ms_deisotope/_c/peak_set.pxd"
  size_t end;

#line 118 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":124
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */

#line 124 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list {

#line 124 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell *index;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  size_t size;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  double low;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  double high;

#line 124 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":131
 * 
 * 
 * cdef struct deconvoluted_peak_t:             # <<<<<<<<<<<<<<
 *     double neutral_mass
 *     double intensity
 */

#line 131 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t {

#line 131 "ms_deisotope/_c/peak_set.pxd"
  double neutral_mass;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  double intensity;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  int charge;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  unsigned int index;

#line 131 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":144
 *     owns_peaks_and_index = 3
 * 
 * cdef struct deconvoluted_peak_set_t:             # <<<<<<<<<<<<<<
 *     deconvoluted_peak_t* peaks
 *     double* mass_index
 */

#line 144 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t {

#line 144 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *peaks;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  double *mass_index;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  size_t size;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  enum __pyx_t_12ms_deisotope_2_c_8peak_set_CPeaksFlags flags;

#line 144 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":154
 * cdef int free_deconvoluted_peak_set_t(deconvoluted_peak_set_t* destination) nogil
 * 
 * cdef deconvoluted_peak_set_t deconvoluted_peak_set_all_peaks_for(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * cdef deconvoluted_peak_t* deconvoluted_peak_set_has_peak(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil
 * 
 */

#line 154 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_all_peaks_for {

#line 154 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 154 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 154 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":155
 * 
 * cdef deconvoluted_peak_set_t deconvoluted_peak_set_all_peaks_for(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil
 * cdef deconvoluted_peak_t* deconvoluted_peak_set_has_peak(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cdef int deconvoluted_peak_eq(deconvoluted_peak_t* self, deconvoluted_peak_t* other) nogil
 */

#line 155 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_has_peak {

#line 155 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 155 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 155 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":167
 *     cdef deconvoluted_peak_t* getitem(self, size_t i) nogil
 * 
 *     cdef deconvoluted_peak_t* _has_peak(self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil
 * 
 */

#line 167 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak {

#line 167 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 167 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 167 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":168
 * 
 *     cdef deconvoluted_peak_t* _has_peak(self, double neutral_mass, double error_tolerance=*) nogil
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)
 */

#line 168 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for {

#line 168 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 168 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 168 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":170
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double error_tolerance=*)
 */

#line 170 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak {

#line 170 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 170 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 170 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":171
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double error_tolerance=*)             # <<<<<<<<<<<<<<
 */

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for {

#line 171 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 171 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 171 "ms_deisotope/_c/peak_set.pxd"
};

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition;

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":169
 * 
 * 
 * cdef tuple _conform_envelopes(list experimental, list base_theoretical, size_t* n_missing_out,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.05):
 *     cdef:
 */

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes {

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_n;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double minimum_theoretical_abundance;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":201
 *         self.n_missing = 0
 * 
 *     cdef inline void conform(self, double minimum_theoretical_abundance=0.05):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total = 0
 */

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform {

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_n;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double minimum_theoretical_abundance;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":235
 * 
 * 
 * cpdef tuple conform_envelopes(list experimental, list base_theoretical,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.01):
 *     cdef:
 */

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_n;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double minimum_theoretical_abundance;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":269
 *         public AveragineCache averagine
 * 
 *     cpdef TheoreticalIsotopicPattern create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8,             # <<<<<<<<<<<<<<
 *                                                                      double ignore_below=0.05):
 *         cdef:
 */

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_n;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double charge_carrier;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double truncate_after;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double ignore_below;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":278
 *         return base_tid
 * 
 *     cpdef list find_all_features(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         return self.feature_map.find_all(mz, error_tolerance)
 * 
 */

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_n;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double error_tolerance;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":281
 *         return self.feature_map.find_all(mz, error_tolerance)
 * 
 *     cpdef list find_features(self, double mz, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list f, overlapped
 */

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_n;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double error_tolerance;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *interval;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":311
 *         return f
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         """Given a list of theoretical peaks, find their counterparts in :attr:`peaklist` within `error_tolerance`
 *         ppm error. If no experimental peak is found, a placeholder will be used in its stead.
 */

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_n;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double error_tolerance;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *interval;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":372
 *         return acc / count
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,             # <<<<<<<<<<<<<<
 *                                                     int charge, double charge_carrier=PROTON, int max_missed_peaks=1,
 *                                                     double threshold_scale=0.3):
 */

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_n;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double charge_carrier;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int max_missed_peaks;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double threshold_scale;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":428
 *         return fit
 * 
 *     cdef list _fit_theoretical_distribution_on_features(self, double mz, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                         TheoreticalIsotopicPattern base_tid,
 *                                                         double charge_carrier=PROTON,
 */

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features {

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_n;

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double charge_carrier;

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double truncate_after;

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int max_missed_peaks;

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double threshold_scale;

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *feature;

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":550
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_n;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int left_search;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int right_search;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double charge_carrier;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double truncate_after;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int max_missed_peaks;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double threshold_scale;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *feature;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":566
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_n;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double charge_carrier;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double truncate_after;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double ignore_below;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int max_missed_peaks;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double threshold_scale;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *feature;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtab;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double mz;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double intensity;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double area;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double signal_to_noise;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double full_width_at_half_max;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double left_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double right_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long peak_count;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long index;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtab;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *peaks;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int indexed;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  double *mz_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *interval_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

#line 123 "brainpy/_c/composition.pxd"
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition {

#line 123 "brainpy/_c/composition.pxd"
  PyObject_HEAD

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtab;

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Composition *impl;

#line 123 "brainpy/_c/composition.pxd"
  double cached_mass;

#line 123 "brainpy/_c/composition.pxd"
  int _clean;

#line 123 "brainpy/_c/composition.pxd"
};

#line 123 "brainpy/_c/composition.pxd"


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 63 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  PyObject_HEAD

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtab;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  double mz;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  double intensity;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
};

#line 63 "brainpy/_c/isotopic_distribution.pxd"


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

#line 5 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex {

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyObject_HEAD

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtab;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyArrayObject *mz_array;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyArrayObject *intensity_array;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaks;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
};

#line 5 "ms_peak_picker/_c/peak_index.pxd"


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

#line 12 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine {

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtab;

#line 12 "ms_deisotope/_c/averagine.pxd"
  double base_mass;

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject *base_composition;

#line 12 "ms_deisotope/_c/averagine.pxd"
};

#line 12 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

#line 23 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtab;

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *peaklist;

#line 23 "ms_deisotope/_c/averagine.pxd"
  double origin;

#line 23 "ms_deisotope/_c/averagine.pxd"
  double offset;

#line 23 "ms_deisotope/_c/averagine.pxd"
};

#line 23 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

#line 79 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache {

#line 79 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtab;

#line 79 "ms_deisotope/_c/averagine.pxd"
  PyObject *backend;

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *averagine;

#line 79 "ms_deisotope/_c/averagine.pxd"
  double cache_truncation;

#line 79 "ms_deisotope/_c/averagine.pxd"
  int enabled;

#line 79 "ms_deisotope/_c/averagine.pxd"
};

#line 79 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject_HEAD

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtab;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int use_subtraction;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *scale_method;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int merge_isobaric_peaks;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  double minimum_intensity;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *scorer;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int verbose;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *_slice_cache;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *averagine;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *averagines;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject_HEAD

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtab;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int lower;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int upper;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int sign;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int *values;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t size;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t index;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

#line 16 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtab;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *seed_peak;

#line 16 "ms_deisotope/_c/scoring.pxd"
  double score;

#line 16 "ms_deisotope/_c/scoring.pxd"
  int charge;

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject *experimental;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *theoretical;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *monoisotopic_peak;

#line 16 "ms_deisotope/_c/scoring.pxd"
  int missed_peaks;

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject *data;

#line 16 "ms_deisotope/_c/scoring.pxd"
  Py_hash_t _hash;

#line 16 "ms_deisotope/_c/scoring.pxd"
};

#line 16 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

#line 38 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase {

#line 38 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtab;

#line 38 "ms_deisotope/_c/scoring.pxd"
  double minimum_score;

#line 38 "ms_deisotope/_c/scoring.pxd"
};

#line 38 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

#line 50 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {

#line 50 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtab;

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *select;

#line 50 "ms_deisotope/_c/scoring.pxd"
};

#line 50 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 62 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {

#line 62 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 62 "ms_deisotope/_c/scoring.pxd"
};

#line 62 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 66 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter {

#line 66 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 66 "ms_deisotope/_c/scoring.pxd"
};

#line 66 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 69 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {

#line 69 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 69 "ms_deisotope/_c/scoring.pxd"
};

#line 69 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

#line 72 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter {

#line 72 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 72 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 72 "ms_deisotope/_c/scoring.pxd"
};

#line 72 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

#line 78 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {

#line 78 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 78 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 78 "ms_deisotope/_c/scoring.pxd"
  double penalty_factor;

#line 78 "ms_deisotope/_c/scoring.pxd"
};

#line 78 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

#line 84 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer {

#line 84 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 84 "ms_deisotope/_c/scoring.pxd"
  PyObject *function;

#line 84 "ms_deisotope/_c/scoring.pxd"
};

#line 84 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

#line 89 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection {

#line 89 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 89 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtab;

#line 89 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;

#line 89 "ms_deisotope/_c/scoring.pxd"
};

#line 89 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

#line 95 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *interference_detector;

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *g_test_scaled;

#line 95 "ms_deisotope/_c/scoring.pxd"
  double peak_count_scale;

#line 95 "ms_deisotope/_c/scoring.pxd"
  double domain_scale;

#line 95 "ms_deisotope/_c/scoring.pxd"
};

#line 95 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

#line 102 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 102 "ms_deisotope/_c/scoring.pxd"
  double scale_factor;

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *scorer;

#line 102 "ms_deisotope/_c/scoring.pxd"
};

#line 102 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

#line 112 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter {

#line 112 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 112 "ms_deisotope/_c/scoring.pxd"
};

#line 112 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":4
 * 
 * 
 * cdef class LCMSFeatureTreeList(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list roots
 */

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList {

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_vtab;

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *roots;

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *_node_id_hash;

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":23
 * 
 * 
 * cdef class LCMSFeatureTreeNodeBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double time
 */

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase {

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_vtab;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double time;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *members;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *_most_abundant_member;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *node_id;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":44
 * 
 * 
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):             # <<<<<<<<<<<<<<
 *     cdef:
 * 
 */

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode {

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase __pyx_base;

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _mz;

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":56
 * 
 * 
 * cdef class FeatureBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public LCMSFeatureTreeList nodes
 */

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase {

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_vtab;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *nodes;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _mz;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _start_time;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _end_time;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":73
 * 
 * 
 * cdef class LCMSFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double _total_intensity
 */

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature {

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _total_intensity;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _last_mz;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *_times;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *_peaks;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *adducts;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *used_as_adduct;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *created_at;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *feature_id;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *_peak_averager;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":104
 * 
 * 
 * cdef class EmptyFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef EmptyFeature _create(double mz)
 */

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature {

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":109
 * 
 * 
 * cdef class FeatureSetIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator {

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_vtab;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *features;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *real_features;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double start_time;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double end_time;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double last_time_seen;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t *index_list;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":137
 * 
 * 
 * cdef class RunningWeightedAverage(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double current_mean
 */

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage {

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_vtab;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double current_mean;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t current_count;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double total_weight;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":154
 * 
 * 
 * cdef class RunningWeightedAverageNeutralMass(RunningWeightedAverage):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass {

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage __pyx_base;

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/feature_map.pxd":5
 * 
 * 
 * cdef class LCMSFeatureMap(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap {

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  PyObject_HEAD

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_vtab;

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  PyObject *features;

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */

#line 5 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index {

#line 5 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 5 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtab;

#line 5 "ms_deisotope/_c/peak_set.pxd"
  size_t neutral_mass;

#line 5 "ms_deisotope/_c/peak_set.pxd"
  size_t mz;

#line 5 "ms_deisotope/_c/peak_set.pxd"
};

#line 5 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 14 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair {

#line 14 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 14 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtab;

#line 14 "ms_deisotope/_c/peak_set.pxd"
  double mz;

#line 14 "ms_deisotope/_c/peak_set.pxd"
  double intensity;

#line 14 "ms_deisotope/_c/peak_set.pxd"
};

#line 14 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */

#line 25 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope {

#line 25 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtab;

#line 25 "ms_deisotope/_c/peak_set.pxd"
  PyObject *pairs;

#line 25 "ms_deisotope/_c/peak_set.pxd"
};

#line 25 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */

#line 40 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double neutral_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double signal_to_noise;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *_index;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double full_width_at_half_max;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int charge;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double a_to_a2_ratio;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double most_abundant_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double average_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double score;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *envelope;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *fit;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int chosen_for_msms;

#line 40 "ms_deisotope/_c/peak_set.pxd"
};

#line 40 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class IonMobilityDeconvolutedPeak(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double drift_time
 */

#line 66 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak {

#line 66 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 66 "ms_deisotope/_c/peak_set.pxd"
  double drift_time;

#line 66 "ms_deisotope/_c/peak_set.pxd"
};

#line 66 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":71
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */

#line 71 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {

#line 71 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 71 "ms_deisotope/_c/peak_set.pxd"
  PyObject *solution;

#line 71 "ms_deisotope/_c/peak_set.pxd"
};

#line 71 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":75
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 75 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtab;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *peaks;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *_mz_ordered;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  int indexed;

#line 75 "ms_deisotope/_c/peak_set.pxd"
};

#line 75 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":104
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */

#line 104 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  double *neutral_mass_array;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  double *mz_array;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list *interval_index;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  size_t _size;

#line 104 "ms_deisotope/_c/peak_set.pxd"
};

#line 104 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":161
 * 
 * 
 * cdef class _CPeakSet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         deconvoluted_peak_set_t* ptr
 */

#line 161 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet {

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet *__pyx_vtab;

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t *ptr;

#line 161 "ms_deisotope/_c/peak_set.pxd"
};

#line 161 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/feature_map/feature_fit.pxd":8
 * import numpy as np
 * 
 * cdef class map_coord:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord {

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  PyObject_HEAD

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord *__pyx_vtab;

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  double mz;

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  double time;

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
};

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"


/* "ms_deisotope/_c/feature_map/feature_fit.pxd":22
 * 
 * 
 * cdef class LCMSFeatureSetFit(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit {

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  PyObject_HEAD

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_vtab;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  PyObject *features;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *theoretical;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  PyObject *supporters;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  double score;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  double mz;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  double neutral_mass;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  int charge;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  PyObject *data;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  size_t missing_features;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  size_t n_points;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *monoisotopic_feature;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  PyObject *scores;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  PyObject *times;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
};

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":65
 * 
 * @cython.final
 * cdef class CartesianProductIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list collections
 */

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator {

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject_HEAD

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_vtab;

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *collections;

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t size;

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int *lengths;

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int *indices;

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int done;

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t total_combinations;

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":186
 * 
 * @cython.final
 * cdef class envelope_conformer:             # <<<<<<<<<<<<<<
 *     cdef:
 *         list experimental
 */

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer {

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject_HEAD

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_vtab;

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *experimental;

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *theoretical;

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t n_missing;

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":263
 * 
 * 
 * cdef class LCMSFeatureProcessorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public LCMSFeatureMap feature_map
 */

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase {

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject_HEAD

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_vtab;

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *feature_map;

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *scorer;

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *averagine;

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"


/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"


#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"


#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_create)(double, double, double, double, double, double, long, long, double);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak;

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"


#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_create)(PyObject *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*reindex)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_get_nearest_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_between_bounds)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double, size_t *, size_t *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getitem)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"


#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*_allocate_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, size_t, int __pyx_skip_dispatch);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak_hinted)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

#line 123 "brainpy/_c/composition.pxd"


#line 123 "brainpy/_c/composition.pxd"
struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition {

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*_create)(struct __pyx_t_7brainpy_2_c_11composition_Composition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*_set_impl)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args);

#line 123 "brainpy/_c/composition.pxd"
  void (*_initialize_from_formula)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);

#line 123 "brainpy/_c/composition.pxd"
  double (*mass)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  int (*__pyx___equality_pycomposition)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  int (*__pyx___equality_dict)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*copy)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*update)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*keys)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*values)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*items)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*pop)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop *__pyx_optional_args);

#line 123 "brainpy/_c/composition.pxd"
  __pyx_t_7brainpy_2_c_11composition_count_type (*getitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);

#line 123 "brainpy/_c/composition.pxd"
  void (*setitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);

#line 123 "brainpy/_c/composition.pxd"
  void (*increment)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);

#line 123 "brainpy/_c/composition.pxd"
  void (*add_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*subtract_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*scale_by)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, long);

#line 123 "brainpy/_c/composition.pxd"
};

#line 123 "brainpy/_c/composition.pxd"
static struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition;

#line 123 "brainpy/_c/composition.pxd"


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 63 "brainpy/_c/isotopic_distribution.pxd"


#line 63 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*_create)(double, double, int);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  int (*_eq)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*clone)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
};

#line 63 "brainpy/_c/isotopic_distribution.pxd"
static struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;

#line 63 "brainpy/_c/isotopic_distribution.pxd"


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

#line 5 "ms_peak_picker/_c/peak_index.pxd"


#line 5 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex {

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak *__pyx_optional_args);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, double);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for *__pyx_optional_args);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
};

#line 5 "ms_peak_picker/_c/peak_index.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex;

#line 5 "ms_peak_picker/_c/peak_index.pxd"


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

#line 12 "ms_deisotope/_c/averagine.pxd"


#line 12 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine {

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
};

#line 12 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine;

#line 12 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

#line 23 "ms_deisotope/_c/averagine.pxd"


#line 23 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_create)(PyObject *, double, double);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  int (*_eq_inst)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*ignore_below)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*truncate_after)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale_raw)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*drop_last_peak)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*total)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*normalize)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*get)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, Py_ssize_t);

#line 23 "ms_deisotope/_c/averagine.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*get_monoisotopic_mz)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*get_processed_peaks)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  size_t (*basepeak_index)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_drop_last)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
};

#line 23 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;

#line 23 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

#line 79 "ms_deisotope/_c/averagine.pxd"


#line 79 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache {

#line 79 "ms_deisotope/_c/averagine.pxd"
  double (*_make_cache_key)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double);

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*has_mz_charge_pair)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair *__pyx_optional_args);

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster *__pyx_optional_args);

#line 79 "ms_deisotope/_c/averagine.pxd"
};

#line 79 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache;

#line 79 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*between)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*match_theoretical_isotopic_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*scale_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_evaluate_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*subtraction)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*fit_incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_check_fit)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_find_next_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_find_previous_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_find_next_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_find_previous_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*_create)(int, int);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_tuple)(PyObject *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_quickcharge)(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*sequence_from_quickcharge)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*make_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  void (*release_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  void (*set_bounds)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int, int);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

#line 16 "ms_deisotope/_c/scoring.pxd"


#line 16 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_create)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create *__pyx_optional_args);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_lt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_gt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
};

#line 16 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;

#line 16 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

#line 38 "ms_deisotope/_c/scoring.pxd"


#line 38 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase {

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*best)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_best_from_set)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, double, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
};

#line 38 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase;

#line 38 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

#line 50 "ms_deisotope/_c/scoring.pxd"


#line 50 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {

#line 50 "ms_deisotope/_c/scoring.pxd"
  double (*_evaluate)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, double, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *(*_configure)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
};

#line 50 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;

#line 50 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 62 "ms_deisotope/_c/scoring.pxd"


#line 62 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {

#line 62 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 62 "ms_deisotope/_c/scoring.pxd"
};

#line 62 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;

#line 62 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 66 "ms_deisotope/_c/scoring.pxd"


#line 66 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter {

#line 66 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 66 "ms_deisotope/_c/scoring.pxd"
};

#line 66 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter;

#line 66 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 69 "ms_deisotope/_c/scoring.pxd"


#line 69 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {

#line 69 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 69 "ms_deisotope/_c/scoring.pxd"
};

#line 69 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;

#line 69 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

#line 72 "ms_deisotope/_c/scoring.pxd"


#line 72 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter {

#line 72 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 72 "ms_deisotope/_c/scoring.pxd"
  double (*score_peak)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak *__pyx_optional_args);

#line 72 "ms_deisotope/_c/scoring.pxd"
};

#line 72 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter;

#line 72 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

#line 78 "ms_deisotope/_c/scoring.pxd"


#line 78 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {

#line 78 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 78 "ms_deisotope/_c/scoring.pxd"
};

#line 78 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;

#line 78 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

#line 84 "ms_deisotope/_c/scoring.pxd"


#line 84 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer {

#line 84 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 84 "ms_deisotope/_c/scoring.pxd"
};

#line 84 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer;

#line 84 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

#line 89 "ms_deisotope/_c/scoring.pxd"


#line 89 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection {

#line 89 "ms_deisotope/_c/scoring.pxd"
  double (*detect_interference)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference *__pyx_optional_args);

#line 89 "ms_deisotope/_c/scoring.pxd"
};

#line 89 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection;

#line 89 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

#line 95 "ms_deisotope/_c/scoring.pxd"


#line 95 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 95 "ms_deisotope/_c/scoring.pxd"
};

#line 95 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;

#line 95 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

#line 102 "ms_deisotope/_c/scoring.pxd"


#line 102 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 102 "ms_deisotope/_c/scoring.pxd"
  double (*_calculate_scale_factor)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 102 "ms_deisotope/_c/scoring.pxd"
  void (*scale_fitted_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);

#line 102 "ms_deisotope/_c/scoring.pxd"
  void (*scale_theoretical_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);

#line 102 "ms_deisotope/_c/scoring.pxd"
};

#line 102 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;

#line 102 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

#line 112 "ms_deisotope/_c/scoring.pxd"


#line 112 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter {

#line 112 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 112 "ms_deisotope/_c/scoring.pxd"
};

#line 112 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter;

#line 112 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":4
 * 
 * 
 * cdef class LCMSFeatureTreeList(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list roots
 */

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList {

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *(*_create)(PyObject *);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  void (*_invalidate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*_make_node)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, PyObject *, int __pyx_skip_dispatch);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, int __pyx_skip_dispatch);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*_find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, size_t *);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, size_t);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":23
 * 
 * 
 * cdef class LCMSFeatureTreeNodeBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double time
 */

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase {

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*_total_intensity_members)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*max_intensity)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*total_intensity)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*_find_most_abundant_member)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_calculate_most_abundant_member)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_recalculate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, size_t);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t (*get_members_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":44
 * 
 * 
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):             # <<<<<<<<<<<<<<
 *     cdef:
 * 
 */

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode {

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase __pyx_base;

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getpeak)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, size_t);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_mz)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":56
 * 
 * 
 * cdef class FeatureBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public LCMSFeatureTreeList nodes
 */

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase {

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_mz)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_neutral_mass)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_start_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_end_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, double, int __pyx_skip_dispatch);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *(*_find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, double, size_t *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":73
 * 
 * 
 * cdef class LCMSFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double _total_intensity
 */

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature {

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, PyObject *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, PyObject *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*overlaps_in_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*spans_in_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, double, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, size_t);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_initialize_averager)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_update_from_averager)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_reaverage_and_update)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*insert_node)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*insert)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, double, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_invalidate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate *__pyx_optional_args);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*split_sparse)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse *__pyx_optional_args);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone *__pyx_optional_args);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*max_intensity)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":104
 * 
 * 
 * cdef class EmptyFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef EmptyFeature _create(double mz)
 */

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature {

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *(*_create)(double);

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":109
 * 
 * 
 * cdef class FeatureSetIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator {

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *(*_create)(PyObject *);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *(*_create_with_threshold)(PyObject *, PyObject *, double);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*_initialize)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, PyObject *);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*init_indices)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_next_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_current_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*get_peaks_for_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, double, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, size_t);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":137
 * 
 * 
 * cdef class RunningWeightedAverage(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double current_mean
 */

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage {

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, int __pyx_skip_dispatch);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_initialize)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, int __pyx_skip_dispatch);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*add)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*update)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, PyObject *, int __pyx_skip_dispatch);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*feed_from_feature)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*_create)(PyObject *);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  void (*_update)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, PyObject *);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":154
 * 
 * 
 * cdef class RunningWeightedAverageNeutralMass(RunningWeightedAverage):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass {

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage __pyx_base;

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *(*_create)(PyObject *);

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/feature_map.pxd":5
 * 
 * 
 * cdef class LCMSFeatureMap(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap {

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  Py_ssize_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *);

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *(*get)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, size_t);

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  PyObject *(*find_all)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all *__pyx_optional_args);

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *(*search)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search *__pyx_optional_args);

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  PyObject *(*between)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between *__pyx_optional_args);

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  PyObject *(*spanning_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, int __pyx_skip_dispatch);

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone *__pyx_optional_args);

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap;

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */

#line 5 "ms_deisotope/_c/peak_set.pxd"


#line 5 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index {

#line 5 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *(*_create)(size_t, size_t);

#line 5 "ms_deisotope/_c/peak_set.pxd"
};

#line 5 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index;

#line 5 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 14 "ms_deisotope/_c/peak_set.pxd"


#line 14 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair {

#line 14 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, int __pyx_skip_dispatch);

#line 14 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*_create)(double, double);

#line 14 "ms_deisotope/_c/peak_set.pxd"
};

#line 14 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair;

#line 14 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */

#line 25 "ms_deisotope/_c/peak_set.pxd"


#line 25 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope {

#line 25 "ms_deisotope/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, size_t);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*_create)(PyObject *);

#line 25 "ms_deisotope/_c/peak_set.pxd"
};

#line 25 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope;

#line 25 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */

#line 40 "ms_deisotope/_c/peak_set.pxd"


#line 40 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, int __pyx_skip_dispatch);

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_create_simple)(double, double, int, double, double, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);

#line 40 "ms_deisotope/_c/peak_set.pxd"
};

#line 40 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;

#line 40 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class IonMobilityDeconvolutedPeak(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double drift_time
 */

#line 66 "ms_deisotope/_c/peak_set.pxd"


#line 66 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak {

#line 66 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 66 "ms_deisotope/_c/peak_set.pxd"
};

#line 66 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak;

#line 66 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":71
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */

#line 71 "ms_deisotope/_c/peak_set.pxd"


#line 71 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {

#line 71 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 71 "ms_deisotope/_c/peak_set.pxd"
};

#line 71 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;

#line 71 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":75
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 75 "ms_deisotope/_c/peak_set.pxd"


#line 75 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*reindex)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*copy)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_get_nearest_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, double *);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*getslice)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t, size_t);

#line 75 "ms_deisotope/_c/peak_set.pxd"
};

#line 75 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;

#line 75 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":104
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */

#line 104 "ms_deisotope/_c/peak_set.pxd"


#line 104 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  void (*_build_index_arrays)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *);

#line 104 "ms_deisotope/_c/peak_set.pxd"
  int (*_interval_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *, double, double, size_t *, size_t *);

#line 104 "ms_deisotope/_c/peak_set.pxd"
};

#line 104 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;

#line 104 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":161
 * 
 * 
 * cdef class _CPeakSet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         deconvoluted_peak_set_t* ptr
 */

#line 161 "ms_deisotope/_c/peak_set.pxd"


#line 161 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet {

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, size_t);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t (*_all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
};

#line 161 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet;

#line 161 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/feature_map/feature_fit.pxd":8
 * import numpy as np
 * 
 * cdef class map_coord:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"


#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord {

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  int (*ge)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord *);

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  int (*le)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord *);

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord *(*_create)(double, double);

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord *(*copy)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord *, int __pyx_skip_dispatch);

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
};

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord;

#line 8 "ms_deisotope/_c/feature_map/feature_fit.pxd"


/* "ms_deisotope/_c/feature_map/feature_fit.pxd":22
 * 
 * 
 * cdef class LCMSFeatureSetFit(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"


#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit {

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, int __pyx_skip_dispatch);

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, int __pyx_skip_dispatch);

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  int (*_lt)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, int __pyx_skip_dispatch);

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  int (*_gt)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, int __pyx_skip_dispatch);

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  int (*count_null_features)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, int __pyx_skip_dispatch);

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  int (*has_multiple_real_features)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *, int __pyx_skip_dispatch);

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord *(*get_start)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *);

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord *(*get_end)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *);

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *(*_create)(PyObject *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int, size_t, PyObject *, PyObject *, double, size_t, PyObject *, PyObject *);

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
};

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit;

#line 22 "ms_deisotope/_c/feature_map/feature_fit.pxd"


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":65
 * 
 * @cython.final
 * cdef class CartesianProductIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list collections
 */

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator {

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int (*_initialize)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *, PyObject *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *(*_create)(PyObject *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *(*compose_next_value)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  void (*advance)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *(*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator__initialize(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *, PyObject *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator__create(PyObject *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_has_more(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_compose_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_advance(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_get_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *);

#line 65 "ms_deisotope/_c/feature_map/feature_processor.pyx"


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":186
 * 
 * @cython.final
 * cdef class envelope_conformer:             # <<<<<<<<<<<<<<
 *     cdef:
 *         list experimental
 */

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer {

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *(*_create)(void);

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  void (*acquire)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, PyObject *, PyObject *);

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  void (*conform)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform *__pyx_optional_args);

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create(void);

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_acquire(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, PyObject *, PyObject *);

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform *__pyx_optional_args);

#line 186 "ms_deisotope/_c/feature_map/feature_processor.pyx"


/* "ms_deisotope/_c/feature_map/feature_processor.pyx":263
 * 
 * 
 * cdef class LCMSFeatureProcessorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public LCMSFeatureMap feature_map
 */

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase {

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*create_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution *__pyx_optional_args);

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *(*find_all_features)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features *__pyx_optional_args);

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *(*find_features)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features *__pyx_optional_args);

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *(*match_theoretical_isotopic_distribution)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution *__pyx_optional_args);

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *(*conform_envelopes)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, int __pyx_skip_dispatch);

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *(*_conform_envelopes)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, size_t *);

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double (*_find_thresholded_score)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *, double);

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *(*_fit_single_feature_set)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set *__pyx_optional_args);

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *(*_fit_theoretical_distribution_on_features)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features *__pyx_optional_args);

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *(*_fit_feature_set)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set *__pyx_optional_args);

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *(*_fit_composition)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition *__pyx_optional_args);

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
};

#line 263 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

#define __Pyx_BufPtrCContig1d(type, buf, i0, s0) ((type)buf + i0)
/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?\
 PyFloat_AS_DOUBLE(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static CYTHON_INLINE int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator__initialize(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_collections); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator__create(PyObject *__pyx_v_collections); /* proto*/
static CYTHON_INLINE int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_has_more(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_compose_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto*/
static CYTHON_INLINE void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_advance(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_get_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create(void); /* proto*/
static CYTHON_INLINE void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_acquire(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical); /* proto*/
static CYTHON_INLINE void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform *__pyx_optional_args); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_conform_envelopes(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__conform_envelopes(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, size_t *__pyx_v_n_missing); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__find_thresholded_score(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_scores, double __pyx_v_percentage); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_features, PyObject *__pyx_v_base_tid, CYTHON_UNUSED double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_base_tid, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_composition, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition *__pyx_optional_args); /* proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cython' */

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'ms_peak_picker._c.peak_set' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = 0;
static size_t *__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'brainpy._c.composition' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition_PyComposition = 0;
static PyObject **__pyx_vp_7brainpy_2_c_11composition_nist_mass = 0;
#define __pyx_v_7brainpy_2_c_11composition_nist_mass (*__pyx_vp_7brainpy_2_c_11composition_nist_mass)
static double *__pyx_vp_7brainpy_2_c_11composition_PROTON = 0;
#define __pyx_v_7brainpy_2_c_11composition_PROTON (*__pyx_vp_7brainpy_2_c_11composition_PROTON)
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable **__pyx_vp_7brainpy_2_c_11composition__ElementTable = 0;
#define __pyx_v_7brainpy_2_c_11composition__ElementTable (*__pyx_vp_7brainpy_2_c_11composition__ElementTable)

/* Module declarations from 'brainpy._c.double_vector' */
static struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *(*__pyx_f_7brainpy_2_c_13double_vector_make_double_vector)(void); /*proto*/
static int (*__pyx_f_7brainpy_2_c_13double_vector_double_vector_append)(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *, double); /*proto*/
static void (*__pyx_f_7brainpy_2_c_13double_vector_free_double_vector)(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *); /*proto*/
static PyObject *(*__pyx_f_7brainpy_2_c_13double_vector_double_vector_to_list)(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *); /*proto*/

/* Module declarations from 'brainpy._c.isotopic_constants' */
static size_t *__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE = 0;
#define __pyx_v_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE (*__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE)

/* Module declarations from 'brainpy._c.isotopic_distribution' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = 0;
static PyObject *(*__pyx_f_7brainpy_2_c_21isotopic_distribution__isotopic_variants)(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants *__pyx_optional_args); /*proto*/

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'ms_peak_picker._c.peak_index' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = 0;

/* Module declarations from 'ms_deisotope._c.averagine' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = 0;
static double *__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON = 0;
#define __pyx_v_12ms_deisotope_2_c_9averagine_PROTON (*__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON)
static double (*__pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass)(double, int, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass *__pyx_optional_args); /*proto*/

/* Module declarations from 'ms_deisotope._c.deconvoluter_base' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = 0;

/* Module declarations from 'ms_deisotope._c.scoring' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = 0;
static double *__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY = 0;
#define __pyx_v_12ms_deisotope_2_c_7scoring_INFINITY (*__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY)

/* Module declarations from 'ms_deisotope._c.feature_map.lcms_feature' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = 0;

/* Module declarations from 'ms_deisotope._c.feature_map.feature_map' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap = 0;

/* Module declarations from 'ms_deisotope._c.peak_set' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet = 0;
static size_t *__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'ms_deisotope._c.feature_map.feature_fit' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit = 0;

/* Module declarations from 'ms_deisotope._c.feature_map.feature_processor' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase = 0;
static PyObject *__pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros = 0;
static double __pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_INF;
static size_t __pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_MAX_COMBINATIONS;
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes(PyObject *, PyObject *, size_t *, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes *__pyx_optional_args); /*proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_has_no_valid_features(PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_envelope_conformer__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_LCMSFeatureProcessorBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "ms_deisotope._c.feature_map.feature_processor"
extern int __pyx_module_is_main_ms_deisotope___c__feature_map__feature_processor;
int __pyx_module_is_main_ms_deisotope___c__feature_map__feature_processor = 0;

/* Implementation of 'ms_deisotope._c.feature_map.feature_processor' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k__6[] = "-";
static const char __pyx_k_mz[] = "mz";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_charge[] = "charge";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_feature[] = "feature";
static const char __pyx_k_base_tid[] = "base_tid";
static const char __pyx_k_end_time[] = "end_time";
static const char __pyx_k_features[] = "features";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_interval[] = "interval";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_0_2f_0_2f[] = "%0.2f-%0.2f";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_start_time[] = "start_time";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_composition[] = "composition";
static const char __pyx_k_left_search[] = "left_search";
static const char __pyx_k_experimental[] = "experimental";
static const char __pyx_k_ignore_below[] = "ignore_below";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_right_search[] = "right_search";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_find_features[] = "find_features";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_charge_carrier[] = "charge_carrier";
static const char __pyx_k_truncate_after[] = "truncate_after";
static const char __pyx_k_error_tolerance[] = "error_tolerance";
static const char __pyx_k_fit_composition[] = "_fit_composition";
static const char __pyx_k_fit_feature_set[] = "_fit_feature_set";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_threshold_scale[] = "threshold_scale";
static const char __pyx_k_base_theoretical[] = "base_theoretical";
static const char __pyx_k_max_missed_peaks[] = "max_missed_peaks";
static const char __pyx_k_conform_envelopes[] = "conform_envelopes";
static const char __pyx_k_find_all_features[] = "find_all_features";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_envelope_conformer[] = "envelope_conformer";
static const char __pyx_k_fit_single_feature_set[] = "_fit_single_feature_set";
static const char __pyx_k_CartesianProductIterator[] = "CartesianProductIterator";
static const char __pyx_k_LCMSFeatureProcessorBase[] = "LCMSFeatureProcessorBase";
static const char __pyx_k_theoretical_distribution[] = "theoretical_distribution";
static const char __pyx_k_minimum_theoretical_abundance[] = "minimum_theoretical_abundance";
static const char __pyx_k_pyx_unpickle_LCMSFeatureProces[] = "__pyx_unpickle_LCMSFeatureProcessorBase";
static const char __pyx_k_pyx_unpickle_envelope_conforme[] = "__pyx_unpickle_envelope_conformer";
static const char __pyx_k_create_theoretical_distribution[] = "create_theoretical_distribution";
static const char __pyx_k_d_combinations_have_been_consid[] = "\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xba9b7dc, 0x30f7136, 0x9f9ee19) = (experimental, n_missing, theoretical))";
static const char __pyx_k_match_theoretical_isotopic_distr[] = "match_theoretical_isotopic_distribution";
static const char __pyx_k_ms_deisotope__c_feature_map_feat[] = "ms_deisotope._c.feature_map.feature_processor";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_self_indices_self_lengths_cannot[] = "self.indices,self.lengths cannot be converted to a Python object for pickling";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0xa9723cd, 0xcb14c85, 0xe2c6adb) = (averagine, feature_map, scorer))";
static PyObject *__pyx_kp_s_0_2f_0_2f;
static PyObject *__pyx_n_s_CartesianProductIterator;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
static PyObject *__pyx_n_s_LCMSFeatureProcessorBase;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_StopIteration;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s__6;
static PyObject *__pyx_n_s_base_theoretical;
static PyObject *__pyx_n_s_base_tid;
static PyObject *__pyx_n_s_charge;
static PyObject *__pyx_n_s_charge_carrier;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_composition;
static PyObject *__pyx_n_s_conform_envelopes;
static PyObject *__pyx_n_s_create_theoretical_distribution;
static PyObject *__pyx_kp_s_d_combinations_have_been_consid;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_end_time;
static PyObject *__pyx_n_s_envelope_conformer;
static PyObject *__pyx_n_s_error_tolerance;
static PyObject *__pyx_n_s_experimental;
static PyObject *__pyx_n_s_feature;
static PyObject *__pyx_n_s_features;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_find_all_features;
static PyObject *__pyx_n_s_find_features;
static PyObject *__pyx_n_s_fit_composition;
static PyObject *__pyx_n_s_fit_feature_set;
static PyObject *__pyx_n_s_fit_single_feature_set;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_ignore_below;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_interval;
static PyObject *__pyx_n_s_left_search;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_map;
static PyObject *__pyx_n_s_match_theoretical_isotopic_distr;
static PyObject *__pyx_n_s_max_missed_peaks;
static PyObject *__pyx_n_s_minimum_theoretical_abundance;
static PyObject *__pyx_n_s_ms_deisotope__c_feature_map_feat;
static PyObject *__pyx_n_s_mz;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_LCMSFeatureProces;
static PyObject *__pyx_n_s_pyx_unpickle_envelope_conforme;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_right_search;
static PyObject *__pyx_kp_s_self_indices_self_lengths_cannot;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_start_time;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_theoretical_distribution;
static PyObject *__pyx_n_s_threshold_scale;
static PyObject *__pyx_n_s_truncate_after;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_zeros;
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_collections); /* proto */
static void __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_2__dealloc__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4get_lengths(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_6get_indices(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_8__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_10__next__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer___reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_2__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, double __pyx_v_minimum_theoretical_abundance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_2find_all_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_4find_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, double __pyx_v_error_tolerance, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_8conform_envelopes(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_10_fit_single_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_features, PyObject *__pyx_v_base_tid, double __pyx_v_error_tolerance, int __pyx_v_charge, double __pyx_v_charge_carrier, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_12_fit_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_v_left_search, int __pyx_v_right_search, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_14_fit_composition(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_composition, double __pyx_v_error_tolerance, int __pyx_v_charge, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_16__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_18__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_2__pyx_unpickle_envelope_conformer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_4__pyx_unpickle_LCMSFeatureProcessorBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_51343670;
static PyObject *__pyx_int_167374361;
static PyObject *__pyx_int_177677261;
static PyObject *__pyx_int_195672028;
static PyObject *__pyx_int_212946053;
static PyObject *__pyx_int_237791963;
static double __pyx_k__3;
static double __pyx_k__4;
static double __pyx_k__5;
static double __pyx_k__7;
static double __pyx_k__8;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_codeobj__14;
static PyObject *__pyx_codeobj__16;
/* Late includes */

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":47
 * 
 * 
 * cdef object double_vector_to_ndarray(dvec* vec):             # <<<<<<<<<<<<<<
 *     cdef:
 *         np.ndarray[double, ndim=1, mode='c'] out
 */

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_vec) {
  PyArrayObject *__pyx_v_out = 0
#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  npy_intp __pyx_v_shape[1]
#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_i
#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_n
#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_out
#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_out
#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_1;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyArrayObject *__pyx_t_6 = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_7;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_11;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_12;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_13;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("double_vector_to_ndarray", 0);

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_pybuffer_out.pybuffer.buf = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_pybuffer_out.refcount = 0;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_pybuffernd_out.data = NULL;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_pybuffernd_out.rcbuffer = &__pyx_pybuffer_out;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":53
 *         size_t i, n
 * 
 *     n = shape[0] = vec.used             # <<<<<<<<<<<<<<
 *     # out = np.PyArray_SimpleNew(1, shape, np.NPY_DOUBLE)
 *     out = zeros(n)
 */

#line 53 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_vec->used;

#line 53 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_n = __pyx_t_1;

#line 53 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  (__pyx_v_shape[0]) = __pyx_t_1;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":55
 *     n = shape[0] = vec.used
 *     # out = np.PyArray_SimpleNew(1, shape, np.NPY_DOUBLE)
 *     out = zeros(n)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         out[i] = vec.v[i]
 */

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 55, __pyx_L1_error)

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros);
  __pyx_t_4 = __pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros; 
#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_t_5 = NULL;

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (likely(__pyx_t_5)) {

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(__pyx_t_5);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(function);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF_SET(__pyx_t_4, function);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 55, __pyx_L1_error)

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_2);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  {

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_C_CONTIGUOUS| PyBUF_WRITABLE, 1, 0, __pyx_stack);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_t_7 < 0)) {

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_v_out, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_C_CONTIGUOUS| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      } else {

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_8 = __pyx_t_9 = __pyx_t_10 = 0;

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_pybuffernd_out.diminfo[0].strides = __pyx_pybuffernd_out.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_out.diminfo[0].shape = __pyx_pybuffernd_out.rcbuffer->pybuffer.shape[0];

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 55, __pyx_L1_error)

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_6 = 0;

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_out = ((PyArrayObject *)__pyx_t_2);

#line 55 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":56
 *     # out = np.PyArray_SimpleNew(1, shape, np.NPY_DOUBLE)
 *     out = zeros(n)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         out[i] = vec.v[i]
 *     # The NDArray returned here seems to get corrupted somehow.
 */

#line 56 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 56 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_11 = __pyx_t_1;

#line 56 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {

#line 56 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_12;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":57
 *     out = zeros(n)
 *     for i in range(n):
 *         out[i] = vec.v[i]             # <<<<<<<<<<<<<<
 *     # The NDArray returned here seems to get corrupted somehow.
 *     # Converting it to a managed object here and converting it back
 */

#line 57 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_13 = __pyx_v_i;

#line 57 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_7 = -1;

#line 57 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_7 = 0;

#line 57 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_t_7 != -1)) {

#line 57 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_7);

#line 57 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __PYX_ERR(0, 57, __pyx_L1_error)

#line 57 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 57 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    *__Pyx_BufPtrCContig1d(double *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_out.diminfo[0].strides) = (__pyx_v_vec->v[__pyx_v_i]);

#line 57 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":61
 *     # Converting it to a managed object here and converting it back
 *     # into a NumPy array later at smooth_leveled seems to prevent it.
 *     return tuple(out)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 61 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 61 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = PySequence_Tuple(((PyObject *)__pyx_v_out)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)

#line 61 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 61 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_2;

#line 61 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = 0;

#line 61 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":47
 * 
 * 
 * cdef object double_vector_to_ndarray(dvec* vec):             # <<<<<<<<<<<<<<
 *     cdef:
 *         np.ndarray[double, ndim=1, mode='c'] out
 */

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_PyThreadState_declare

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_PyThreadState_assign

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.double_vector_to_ndarray", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L2;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L2:;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_out);

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 47 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":74
 *         public size_t total_combinations
 * 
 *     cdef inline int _initialize(self, list collections) except 1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator__initialize(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_collections) {
  size_t __pyx_v_i
#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_sublist = 0
#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_2;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_3;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_4;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_5;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_6;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_initialize", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":79
 *             list sublist
 * 
 *         self.collections = collections             # <<<<<<<<<<<<<<
 *         self.size = PyList_GET_SIZE(self.collections)
 *         self.lengths = <int*>malloc(sizeof(int) * self.size)
 */

#line 79 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_collections);

#line 79 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_v_collections);

#line 79 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->collections);

#line 79 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_v_self->collections);

#line 79 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->collections = __pyx_v_collections;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":80
 * 
 *         self.collections = collections
 *         self.size = PyList_GET_SIZE(self.collections)             # <<<<<<<<<<<<<<
 *         self.lengths = <int*>malloc(sizeof(int) * self.size)
 *         if self.lengths == NULL:
 */

#line 80 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_self->collections;

#line 80 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 80 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->size = PyList_GET_SIZE(__pyx_t_1);

#line 80 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":81
 *         self.collections = collections
 *         self.size = PyList_GET_SIZE(self.collections)
 *         self.lengths = <int*>malloc(sizeof(int) * self.size)             # <<<<<<<<<<<<<<
 *         if self.lengths == NULL:
 *             PyErr_SetString(MemoryError, "Failed to allocate length array in CartesianProductIterator")
 */

#line 81 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->lengths = ((int *)malloc(((sizeof(int)) * __pyx_v_self->size)));

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":82
 *         self.size = PyList_GET_SIZE(self.collections)
 *         self.lengths = <int*>malloc(sizeof(int) * self.size)
 *         if self.lengths == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_SetString(MemoryError, "Failed to allocate length array in CartesianProductIterator")
 *         self.indices = <int*>malloc(sizeof(int) * self.size)
 */

#line 82 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = ((__pyx_v_self->lengths == NULL) != 0);

#line 82 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":83
 *         self.lengths = <int*>malloc(sizeof(int) * self.size)
 *         if self.lengths == NULL:
 *             PyErr_SetString(MemoryError, "Failed to allocate length array in CartesianProductIterator")             # <<<<<<<<<<<<<<
 *         self.indices = <int*>malloc(sizeof(int) * self.size)
 *         if self.indices == NULL:
 */

#line 83 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyErr_SetString(__pyx_builtin_MemoryError, ((char *)"Failed to allocate length array in CartesianProductIterator"));

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":82
 *         self.size = PyList_GET_SIZE(self.collections)
 *         self.lengths = <int*>malloc(sizeof(int) * self.size)
 *         if self.lengths == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_SetString(MemoryError, "Failed to allocate length array in CartesianProductIterator")
 *         self.indices = <int*>malloc(sizeof(int) * self.size)
 */

#line 82 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":84
 *         if self.lengths == NULL:
 *             PyErr_SetString(MemoryError, "Failed to allocate length array in CartesianProductIterator")
 *         self.indices = <int*>malloc(sizeof(int) * self.size)             # <<<<<<<<<<<<<<
 *         if self.indices == NULL:
 *             PyErr_SetString(MemoryError, "Failed to allocate indices array in CartesianProductIterator")
 */

#line 84 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->indices = ((int *)malloc(((sizeof(int)) * __pyx_v_self->size)));

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":85
 *             PyErr_SetString(MemoryError, "Failed to allocate length array in CartesianProductIterator")
 *         self.indices = <int*>malloc(sizeof(int) * self.size)
 *         if self.indices == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_SetString(MemoryError, "Failed to allocate indices array in CartesianProductIterator")
 *         self.done = False
 */

#line 85 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = ((__pyx_v_self->indices == NULL) != 0);

#line 85 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":86
 *         self.indices = <int*>malloc(sizeof(int) * self.size)
 *         if self.indices == NULL:
 *             PyErr_SetString(MemoryError, "Failed to allocate indices array in CartesianProductIterator")             # <<<<<<<<<<<<<<
 *         self.done = False
 *         self.total_combinations = 1
 */

#line 86 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyErr_SetString(__pyx_builtin_MemoryError, ((char *)"Failed to allocate indices array in CartesianProductIterator"));

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":85
 *             PyErr_SetString(MemoryError, "Failed to allocate length array in CartesianProductIterator")
 *         self.indices = <int*>malloc(sizeof(int) * self.size)
 *         if self.indices == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_SetString(MemoryError, "Failed to allocate indices array in CartesianProductIterator")
 *         self.done = False
 */

#line 85 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":87
 *         if self.indices == NULL:
 *             PyErr_SetString(MemoryError, "Failed to allocate indices array in CartesianProductIterator")
 *         self.done = False             # <<<<<<<<<<<<<<
 *         self.total_combinations = 1
 * 
 */

#line 87 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->done = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":88
 *             PyErr_SetString(MemoryError, "Failed to allocate indices array in CartesianProductIterator")
 *         self.done = False
 *         self.total_combinations = 1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.size):
 */

#line 88 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->total_combinations = 1;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":90
 *         self.total_combinations = 1
 * 
 *         for i in range(self.size):             # <<<<<<<<<<<<<<
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             self.lengths[i] = PyList_GET_SIZE(sublist)
 */

#line 90 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = __pyx_v_self->size;

#line 90 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_4 = __pyx_t_3;

#line 90 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {

#line 90 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_5;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":91
 * 
 *         for i in range(self.size):
 *             sublist = <list>PyList_GetItem(self.collections, i)             # <<<<<<<<<<<<<<
 *             self.lengths[i] = PyList_GET_SIZE(sublist)
 *             self.total_combinations *= self.lengths[i]
 */

#line 91 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = __pyx_v_self->collections;

#line 91 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 91 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_6 = PyList_GetItem(__pyx_t_1, __pyx_v_i); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 91, __pyx_L1_error)

#line 91 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 91 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_6);

#line 91 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 91 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_sublist, ((PyObject*)__pyx_t_1));

#line 91 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":92
 *         for i in range(self.size):
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             self.lengths[i] = PyList_GET_SIZE(sublist)             # <<<<<<<<<<<<<<
 *             self.total_combinations *= self.lengths[i]
 *             if self.lengths[i] == 0:
 */

#line 92 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    (__pyx_v_self->lengths[__pyx_v_i]) = PyList_GET_SIZE(__pyx_v_sublist);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":93
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             self.lengths[i] = PyList_GET_SIZE(sublist)
 *             self.total_combinations *= self.lengths[i]             # <<<<<<<<<<<<<<
 *             if self.lengths[i] == 0:
 *                 self.done = True
 */

#line 93 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_self->total_combinations = (__pyx_v_self->total_combinations * (__pyx_v_self->lengths[__pyx_v_i]));

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":94
 *             self.lengths[i] = PyList_GET_SIZE(sublist)
 *             self.total_combinations *= self.lengths[i]
 *             if self.lengths[i] == 0:             # <<<<<<<<<<<<<<
 *                 self.done = True
 *             self.indices[i] = 0
 */

#line 94 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = (((__pyx_v_self->lengths[__pyx_v_i]) == 0) != 0);

#line 94 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":95
 *             self.total_combinations *= self.lengths[i]
 *             if self.lengths[i] == 0:
 *                 self.done = True             # <<<<<<<<<<<<<<
 *             self.indices[i] = 0
 * 
 */

#line 95 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_self->done = 1;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":94
 *             self.lengths[i] = PyList_GET_SIZE(sublist)
 *             self.total_combinations *= self.lengths[i]
 *             if self.lengths[i] == 0:             # <<<<<<<<<<<<<<
 *                 self.done = True
 *             self.indices[i] = 0
 */

#line 94 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":96
 *             if self.lengths[i] == 0:
 *                 self.done = True
 *             self.indices[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 96 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    (__pyx_v_self->indices[__pyx_v_i]) = 0;

#line 96 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":74
 *         public size_t total_combinations
 * 
 *     cdef inline int _initialize(self, list collections) except 1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator._initialize", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 1;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_sublist);

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 74 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":99
 * 
 *     @staticmethod
 *     cdef CartesianProductIterator _create(list collections):             # <<<<<<<<<<<<<<
 *         cdef CartesianProductIterator self = CartesianProductIterator.__new__(CartesianProductIterator)
 *         self._initialize(collections)
 */

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator__create(PyObject *__pyx_v_collections) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self = 0
#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_r = NULL;

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_2;

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":100
 *     @staticmethod
 *     cdef CartesianProductIterator _create(list collections):
 *         cdef CartesianProductIterator self = CartesianProductIterator.__new__(CartesianProductIterator)             # <<<<<<<<<<<<<<
 *         self._initialize(collections)
 *         return self
 */

#line 100 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)

#line 100 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

#line 100 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_t_1);

#line 100 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":101
 *     cdef CartesianProductIterator _create(list collections):
 *         cdef CartesianProductIterator self = CartesianProductIterator.__new__(CartesianProductIterator)
 *         self._initialize(collections)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 101 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator__initialize(__pyx_v_self, __pyx_v_collections); if (unlikely(__pyx_t_2 == ((int)1))) __PYX_ERR(0, 101, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":102
 *         cdef CartesianProductIterator self = CartesianProductIterator.__new__(CartesianProductIterator)
 *         self._initialize(collections)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, collections):
 */

#line 102 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 102 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 102 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_self;

#line 102 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":99
 * 
 *     @staticmethod
 *     cdef CartesianProductIterator _create(list collections):             # <<<<<<<<<<<<<<
 *         cdef CartesianProductIterator self = CartesianProductIterator.__new__(CartesianProductIterator)
 *         self._initialize(collections)
 */

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator._create", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_self);

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 99 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":104
 *         return self
 * 
 *     def __init__(self, collections):             # <<<<<<<<<<<<<<
 *         self.collections = list(map(list, collections))
 *         self._initialize(self.collections)
 */

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_collections = 0
#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  {

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_collections,0};

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyObject* values[1] = {0};

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_kwds)) {

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      Py_ssize_t kw_args;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {
        case  1: 
#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0: break;
        default: 
#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0:

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_collections)) != 0)) kw_args--;
        else 
#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(kw_args > 0)) {

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 104, __pyx_L3_error)

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L5_argtuple_error;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_collections = values[0];

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 104, __pyx_L3_error)

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L3_error:;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return -1;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator___init__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), __pyx_v_collections);

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_collections) {

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_3;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":105
 * 
 *     def __init__(self, collections):
 *         self.collections = list(map(list, collections))             # <<<<<<<<<<<<<<
 *         self._initialize(self.collections)
 * 
 */

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(((PyObject *)(&PyList_Type)));

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(((PyObject *)(&PyList_Type)));

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&PyList_Type)));

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_collections);

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_v_collections);

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_collections);

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->collections);

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_v_self->collections);

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->collections = ((PyObject*)__pyx_t_1);

#line 105 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":106
 *     def __init__(self, collections):
 *         self.collections = list(map(list, collections))
 *         self._initialize(self.collections)             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */

#line 106 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_self->collections;

#line 106 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 106 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator__initialize(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(__pyx_t_3 == ((int)1))) __PYX_ERR(0, 106, __pyx_L1_error)

#line 106 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":104
 *         return self
 * 
 *     def __init__(self, collections):             # <<<<<<<<<<<<<<
 *         self.collections = list(map(list, collections))
 *         self._initialize(self.collections)
 */

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = -1;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 104 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":108
 *         self._initialize(self.collections)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free(self.lengths)
 *         free(self.indices)
 */

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static void __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_3__dealloc__(PyObject *__pyx_v_self); /*proto*/

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static void __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_3__dealloc__(PyObject *__pyx_v_self) {

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_2__dealloc__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static void __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_2__dealloc__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":109
 * 
 *     def __dealloc__(self):
 *         free(self.lengths)             # <<<<<<<<<<<<<<
 *         free(self.indices)
 * 
 */

#line 109 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  free(__pyx_v_self->lengths);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":110
 *     def __dealloc__(self):
 *         free(self.lengths)
 *         free(self.indices)             # <<<<<<<<<<<<<<
 * 
 *     def get_lengths(self):
 */

#line 110 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  free(__pyx_v_self->indices);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":108
 *         self._initialize(self.collections)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free(self.lengths)
 *         free(self.indices)
 */

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 108 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":112
 *         free(self.indices)
 * 
 *     def get_lengths(self):             # <<<<<<<<<<<<<<
 *         return [self.lengths[i] for i in range(self.size)]
 * 
 */

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_5get_lengths(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4get_lengths[] = "CartesianProductIterator.get_lengths(self)";

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_5get_lengths(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("get_lengths (wrapper)", 0);
  __pyx_r = 
#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4get_lengths(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4get_lengths(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  size_t __pyx_v_i
#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_2;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_3;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_4;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("get_lengths", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":113
 * 
 *     def get_lengths(self):
 *         return [self.lengths[i] for i in range(self.size)]             # <<<<<<<<<<<<<<
 * 
 *     def get_indices(self):
 */

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = __pyx_v_self->size;

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_4;

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = __Pyx_PyInt_From_int((__pyx_v_self->lengths[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 113, __pyx_L1_error)

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 113, __pyx_L1_error)

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":112
 *         free(self.indices)
 * 
 *     def get_lengths(self):             # <<<<<<<<<<<<<<
 *         return [self.lengths[i] for i in range(self.size)]
 * 
 */

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.get_lengths", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":115
 *         return [self.lengths[i] for i in range(self.size)]
 * 
 *     def get_indices(self):             # <<<<<<<<<<<<<<
 *         return [self.indices[i] for i in range(self.size)]
 * 
 */

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_7get_indices(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_6get_indices[] = "CartesianProductIterator.get_indices(self)";

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_7get_indices(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("get_indices (wrapper)", 0);
  __pyx_r = 
#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_6get_indices(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_6get_indices(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  size_t __pyx_v_i
#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_2;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_3;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_4;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("get_indices", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":116
 * 
 *     def get_indices(self):
 *         return [self.indices[i] for i in range(self.size)]             # <<<<<<<<<<<<<<
 * 
 *     cdef inline bint has_more(self):
 */

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 116, __pyx_L1_error)

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = __pyx_v_self->size;

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_4;

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = __Pyx_PyInt_From_int((__pyx_v_self->indices[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 116, __pyx_L1_error)

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 116, __pyx_L1_error)

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 116 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":115
 *         return [self.lengths[i] for i in range(self.size)]
 * 
 *     def get_indices(self):             # <<<<<<<<<<<<<<
 *         return [self.indices[i] for i in range(self.size)]
 * 
 */

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.get_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 115 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":118
 *         return [self.indices[i] for i in range(self.size)]
 * 
 *     cdef inline bint has_more(self):             # <<<<<<<<<<<<<<
 *         return not self.done
 * 
 */

#line 118 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 118 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_has_more(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {

#line 118 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 118 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 118 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("has_more", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":119
 * 
 *     cdef inline bint has_more(self):
 *         return not self.done             # <<<<<<<<<<<<<<
 * 
 *     cdef inline list compose_next_value(self):
 */

#line 119 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = (!(__pyx_v_self->done != 0));

#line 119 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":118
 *         return [self.indices[i] for i in range(self.size)]
 * 
 *     cdef inline bint has_more(self):             # <<<<<<<<<<<<<<
 *         return not self.done
 * 
 */

#line 118 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 118 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 118 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 118 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 118 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 118 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":121
 *         return not self.done
 * 
 *     cdef inline list compose_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i, ix
 */

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_compose_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  int __pyx_v_i
#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  int __pyx_v_ix
#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_result = 0
#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_sublist = 0
#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_value = 0
#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_1;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_3;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_4;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_5;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_6;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_7;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("compose_next_value", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":126
 *             list result, sublist
 *             object value
 *         if self.done:             # <<<<<<<<<<<<<<
 *             return None
 *         result = []
 */

#line 126 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = (__pyx_v_self->done != 0);

#line 126 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":127
 *             object value
 *         if self.done:
 *             return None             # <<<<<<<<<<<<<<
 *         result = []
 *         for i in range(self.size):
 */

#line 127 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 127 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);

#line 127 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":126
 *             list result, sublist
 *             object value
 *         if self.done:             # <<<<<<<<<<<<<<
 *             return None
 *         result = []
 */

#line 126 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":128
 *         if self.done:
 *             return None
 *         result = []             # <<<<<<<<<<<<<<
 *         for i in range(self.size):
 *             sublist = <list>PyList_GetItem(self.collections, i)
 */

#line 128 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 128, __pyx_L1_error)

#line 128 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 128 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_result = ((PyObject*)__pyx_t_2);

#line 128 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":129
 *             return None
 *         result = []
 *         for i in range(self.size):             # <<<<<<<<<<<<<<
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             ix = self.indices[i]
 */

#line 129 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = __pyx_v_self->size;

#line 129 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_4 = __pyx_t_3;

#line 129 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {

#line 129 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_5;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":130
 *         result = []
 *         for i in range(self.size):
 *             sublist = <list>PyList_GetItem(self.collections, i)             # <<<<<<<<<<<<<<
 *             ix = self.indices[i]
 *             value = <object>PyList_GetItem(sublist, ix)
 */

#line 130 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = __pyx_v_self->collections;

#line 130 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 130 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_6 = PyList_GetItem(__pyx_t_2, __pyx_v_i); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 130, __pyx_L1_error)

#line 130 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 130 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = ((PyObject *)__pyx_t_6);

#line 130 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 130 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_sublist, ((PyObject*)__pyx_t_2));

#line 130 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":131
 *         for i in range(self.size):
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             ix = self.indices[i]             # <<<<<<<<<<<<<<
 *             value = <object>PyList_GetItem(sublist, ix)
 *             result.append(value)
 */

#line 131 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_ix = (__pyx_v_self->indices[__pyx_v_i]);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":132
 *             sublist = <list>PyList_GetItem(self.collections, i)
 *             ix = self.indices[i]
 *             value = <object>PyList_GetItem(sublist, ix)             # <<<<<<<<<<<<<<
 *             result.append(value)
 *         return result
 */

#line 132 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_6 = PyList_GetItem(__pyx_v_sublist, __pyx_v_ix); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(0, 132, __pyx_L1_error)

#line 132 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = ((PyObject *)__pyx_t_6);

#line 132 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 132 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_2);

#line 132 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":133
 *             ix = self.indices[i]
 *             value = <object>PyList_GetItem(sublist, ix)
 *             result.append(value)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */

#line 133 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_value); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":134
 *             value = <object>PyList_GetItem(sublist, ix)
 *             result.append(value)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void advance(self):
 */

#line 134 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 134 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_result);

#line 134 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_result;

#line 134 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":121
 *         return not self.done
 * 
 *     cdef inline list compose_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i, ix
 */

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.compose_next_value", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_sublist);

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_value);

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 121 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":136
 *         return result
 * 
 *     cdef inline void advance(self):             # <<<<<<<<<<<<<<
 *         for i in range(self.size - 1, -1, -1):
 *             if self.indices[i] == self.lengths[i] - 1:
 */

#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_advance(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  size_t __pyx_v_i
#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_1;

#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_2;

#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_3;

#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("advance", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":137
 * 
 *     cdef inline void advance(self):
 *         for i in range(self.size - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             if self.indices[i] == self.lengths[i] - 1:
 *                 self.indices[i] = 0
 */

#line 137 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_1 = (__pyx_v_self->size - 1) + 1; __pyx_t_1 > -1L + 1; ) { __pyx_t_1-=1;

#line 137 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_1;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":138
 *     cdef inline void advance(self):
 *         for i in range(self.size - 1, -1, -1):
 *             if self.indices[i] == self.lengths[i] - 1:             # <<<<<<<<<<<<<<
 *                 self.indices[i] = 0
 *                 if i == 0:
 */

#line 138 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = (((__pyx_v_self->indices[__pyx_v_i]) == ((__pyx_v_self->lengths[__pyx_v_i]) - 1)) != 0);

#line 138 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":139
 *         for i in range(self.size - 1, -1, -1):
 *             if self.indices[i] == self.lengths[i] - 1:
 *                 self.indices[i] = 0             # <<<<<<<<<<<<<<
 *                 if i == 0:
 *                     self.done = True
 */

#line 139 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      (__pyx_v_self->indices[__pyx_v_i]) = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":140
 *             if self.indices[i] == self.lengths[i] - 1:
 *                 self.indices[i] = 0
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     self.done = True
 *             else:
 */

#line 140 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_2 = ((__pyx_v_i == 0) != 0);

#line 140 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_t_2) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":141
 *                 self.indices[i] = 0
 *                 if i == 0:
 *                     self.done = True             # <<<<<<<<<<<<<<
 *             else:
 *                 self.indices[i] += 1
 */

#line 141 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_v_self->done = 1;

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":140
 *             if self.indices[i] == self.lengths[i] - 1:
 *                 self.indices[i] = 0
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     self.done = True
 *             else:
 */

#line 140 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":138
 *     cdef inline void advance(self):
 *         for i in range(self.size - 1, -1, -1):
 *             if self.indices[i] == self.lengths[i] - 1:             # <<<<<<<<<<<<<<
 *                 self.indices[i] = 0
 *                 if i == 0:
 */

#line 138 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L5;

#line 138 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":143
 *                     self.done = True
 *             else:
 *                 self.indices[i] += 1             # <<<<<<<<<<<<<<
 *                 break
 * 
 */

#line 143 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    /*else*/ {

#line 143 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_3 = __pyx_v_i;

#line 143 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      (__pyx_v_self->indices[__pyx_t_3]) = ((__pyx_v_self->indices[__pyx_t_3]) + 1);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":144
 *             else:
 *                 self.indices[i] += 1
 *                 break             # <<<<<<<<<<<<<<
 * 
 *     cdef inline list get_next_value(self):
 */

#line 144 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L4_break;

#line 144 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 144 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_L5:;

#line 144 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 144 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_break:;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":136
 *         return result
 * 
 *     cdef inline void advance(self):             # <<<<<<<<<<<<<<
 *         for i in range(self.size - 1, -1, -1):
 *             if self.indices[i] == self.lengths[i] - 1:
 */

#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 136 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":146
 *                 break
 * 
 *     cdef inline list get_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef list value
 *         if self.done:
 */

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_get_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_v_value = 0
#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_1;

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("get_next_value", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":148
 *     cdef inline list get_next_value(self):
 *         cdef list value
 *         if self.done:             # <<<<<<<<<<<<<<
 *             return None
 *         else:
 */

#line 148 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = (__pyx_v_self->done != 0);

#line 148 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":149
 *         cdef list value
 *         if self.done:
 *             return None             # <<<<<<<<<<<<<<
 *         else:
 *             value = self.compose_next_value()
 */

#line 149 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 149 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);

#line 149 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":148
 *     cdef inline list get_next_value(self):
 *         cdef list value
 *         if self.done:             # <<<<<<<<<<<<<<
 *             return None
 *         else:
 */

#line 148 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":151
 *             return None
 *         else:
 *             value = self.compose_next_value()             # <<<<<<<<<<<<<<
 *             self.advance()
 *             return value
 */

#line 151 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /*else*/ {

#line 151 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_compose_next_value(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 151, __pyx_L1_error)

#line 151 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 151 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_value = ((PyObject*)__pyx_t_2);

#line 151 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":152
 *         else:
 *             value = self.compose_next_value()
 *             self.advance()             # <<<<<<<<<<<<<<
 *             return value
 * 
 */

#line 152 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_advance(__pyx_v_self);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":153
 *             value = self.compose_next_value()
 *             self.advance()
 *             return value             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */

#line 153 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 153 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_v_value);

#line 153 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_r = __pyx_v_value;

#line 153 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    goto __pyx_L0;

#line 153 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":146
 *                 break
 * 
 *     cdef inline list get_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef list value
 *         if self.done:
 */

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.get_next_value", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_value);

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 146 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":155
 *             return value
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_9__iter__(PyObject *__pyx_v_self); /*proto*/

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_9__iter__(PyObject *__pyx_v_self) {

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = 
#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_8__iter__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_8__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":156
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */

#line 156 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 156 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 156 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = ((PyObject *)__pyx_v_self);

#line 156 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":155
 *             return value
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 155 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":158
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         if self.done:
 *             raise StopIteration()
 */

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11__next__(PyObject *__pyx_v_self); /*proto*/

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11__next__(PyObject *__pyx_v_self) {

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = 
#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_10__next__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_10__next__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {
  PyObject *__pyx_v_value = NULL
#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_1;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_3;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":159
 * 
 *     def __next__(self):
 *         if self.done:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         value = self.compose_next_value()
 */

#line 159 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = (__pyx_v_self->done != 0);

#line 159 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(__pyx_t_1)) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":160
 *     def __next__(self):
 *         if self.done:
 *             raise StopIteration()             # <<<<<<<<<<<<<<
 *         value = self.compose_next_value()
 *         if value is None:
 */

#line 160 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_StopIteration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 160, __pyx_L1_error)

#line 160 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 160 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 160 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 160 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __PYX_ERR(0, 160, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":159
 * 
 *     def __next__(self):
 *         if self.done:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         value = self.compose_next_value()
 */

#line 159 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":161
 *         if self.done:
 *             raise StopIteration()
 *         value = self.compose_next_value()             # <<<<<<<<<<<<<<
 *         if value is None:
 *             raise StopIteration()
 */

#line 161 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_compose_next_value(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)

#line 161 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 161 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_value = ((PyObject*)__pyx_t_2);

#line 161 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":162
 *             raise StopIteration()
 *         value = self.compose_next_value()
 *         if value is None:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         else:
 */

#line 162 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = (__pyx_v_value == ((PyObject*)Py_None));

#line 162 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = (__pyx_t_1 != 0);

#line 162 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(__pyx_t_3)) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":163
 *         value = self.compose_next_value()
 *         if value is None:
 *             raise StopIteration()             # <<<<<<<<<<<<<<
 *         else:
 *             self.advance()
 */

#line 163 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_StopIteration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 163, __pyx_L1_error)

#line 163 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 163 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 163 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 163 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __PYX_ERR(0, 163, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":162
 *             raise StopIteration()
 *         value = self.compose_next_value()
 *         if value is None:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         else:
 */

#line 162 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":165
 *             raise StopIteration()
 *         else:
 *             self.advance()             # <<<<<<<<<<<<<<
 *         return value
 * 
 */

#line 165 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /*else*/ {

#line 165 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_advance(__pyx_v_self);

#line 165 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":166
 *         else:
 *             self.advance()
 *         return value             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 166 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 166 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_value);

#line 166 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_value;

#line 166 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":158
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         if self.done:
 *             raise StopIteration()
 */

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_value);

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 158 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":67
 * cdef class CartesianProductIterator(object):
 *     cdef:
 *         public list collections             # <<<<<<<<<<<<<<
 *         public size_t size
 *         int* lengths
 */

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_1__get__(PyObject *__pyx_v_self) {

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_self->collections);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_self->collections;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value) {

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 67, __pyx_L1_error)

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_value;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->collections);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_v_self->collections);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->collections = ((PyObject*)__pyx_t_1);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.collections.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = -1;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_5__del__(PyObject *__pyx_v_self) {

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(Py_None);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(Py_None);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->collections);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_v_self->collections);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->collections = ((PyObject*)Py_None);

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 67 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":68
 *     cdef:
 *         public list collections
 *         public size_t size             # <<<<<<<<<<<<<<
 *         int* lengths
 *         int* indices
 */

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_1__get__(PyObject *__pyx_v_self) {

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value) {

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_1;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 68, __pyx_L1_error)

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->size = __pyx_t_1;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.size.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = -1;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 68 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":71
 *         int* lengths
 *         int* indices
 *         public bint done             # <<<<<<<<<<<<<<
 *         public size_t total_combinations
 * 
 */

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_1__get__(PyObject *__pyx_v_self) {

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->done); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.done.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value) {

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_1;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 71, __pyx_L1_error)

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->done = __pyx_t_1;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.done.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = -1;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 71 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":72
 *         int* indices
 *         public bint done
 *         public size_t total_combinations             # <<<<<<<<<<<<<<
 * 
 *     cdef inline int _initialize(self, list collections) except 1:
 */

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_1__get__(PyObject *__pyx_v_self) {

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->total_combinations); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.total_combinations.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, PyObject *__pyx_v_value) {

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_1;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 72, __pyx_L1_error)

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->total_combinations = __pyx_t_1;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.total_combinations.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = -1;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 72 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_12__reduce_cython__[] = "CartesianProductIterator.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_12__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")
 */

#line 2 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 2 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 2 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 2 "(tree fragment)"
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
/* Python wrapper */

#line 3 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 3 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_14__setstate_cython__[] = "CartesianProductIterator.__setstate_cython__(self, __pyx_state)";

#line 3 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 3 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_14__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 3 "(tree fragment)"
  int __pyx_lineno = 0;

#line 3 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 3 "(tree fragment)"
  int __pyx_clineno = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __pyx_L1_error:;

#line 3 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "(tree fragment)"
  __pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":169
 * 
 * 
 * cdef tuple _conform_envelopes(list experimental, list base_theoretical, size_t* n_missing_out,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.05):
 *     cdef:
 */

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes(PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, size_t *__pyx_v_n_missing_out, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes *__pyx_optional_args) {

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_minimum_theoretical_abundance = ((double)0.05);
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_conformer = 0
#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_cleaned_eid = 0
#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_tid = 0
#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform __pyx_t_2;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_3;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_conform_envelopes", 0);

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_optional_args) {

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_minimum_theoretical_abundance = __pyx_optional_args->minimum_theoretical_abundance;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":175
 *         list cleaned_eid, tid
 * 
 *     conformer = envelope_conformer._create()             # <<<<<<<<<<<<<<
 *     conformer.acquire(experimental, base_theoretical)
 *     conformer.conform(minimum_theoretical_abundance)
 */

#line 175 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 175 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_conformer = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_t_1);

#line 175 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":176
 * 
 *     conformer = envelope_conformer._create()
 *     conformer.acquire(experimental, base_theoretical)             # <<<<<<<<<<<<<<
 *     conformer.conform(minimum_theoretical_abundance)
 * 
 */

#line 176 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_acquire(__pyx_v_conformer, __pyx_v_experimental, __pyx_v_base_theoretical);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":177
 *     conformer = envelope_conformer._create()
 *     conformer.acquire(experimental, base_theoretical)
 *     conformer.conform(minimum_theoretical_abundance)             # <<<<<<<<<<<<<<
 * 
 *     cleaned_eid = conformer.experimental
 */

#line 177 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 177 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.minimum_theoretical_abundance = __pyx_v_minimum_theoretical_abundance;

#line 177 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform(__pyx_v_conformer, &__pyx_t_2); 

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":179
 *     conformer.conform(minimum_theoretical_abundance)
 * 
 *     cleaned_eid = conformer.experimental             # <<<<<<<<<<<<<<
 *     tid = conformer.theoretical
 *     n_missing_out[0] = conformer.n_missing
 */

#line 179 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_conformer->experimental;

#line 179 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 179 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_cleaned_eid = ((PyObject*)__pyx_t_1);

#line 179 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":180
 * 
 *     cleaned_eid = conformer.experimental
 *     tid = conformer.theoretical             # <<<<<<<<<<<<<<
 *     n_missing_out[0] = conformer.n_missing
 *     return (cleaned_eid, tid)
 */

#line 180 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_conformer->theoretical;

#line 180 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 180 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_tid = ((PyObject*)__pyx_t_1);

#line 180 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":181
 *     cleaned_eid = conformer.experimental
 *     tid = conformer.theoretical
 *     n_missing_out[0] = conformer.n_missing             # <<<<<<<<<<<<<<
 *     return (cleaned_eid, tid)
 * 
 */

#line 181 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = __pyx_v_conformer->n_missing;

#line 181 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  (__pyx_v_n_missing_out[0]) = __pyx_t_3;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":182
 *     tid = conformer.theoretical
 *     n_missing_out[0] = conformer.n_missing
 *     return (cleaned_eid, tid)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_cleaned_eid);

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_v_cleaned_eid);

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_cleaned_eid);

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_tid);

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_v_tid);

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_tid);

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = ((PyObject*)__pyx_t_1);

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 182 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":169
 * 
 * 
 * cdef tuple _conform_envelopes(list experimental, list base_theoretical, size_t* n_missing_out,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.05):
 *     cdef:
 */

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor._conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_conformer);

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_cleaned_eid);

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_tid);

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 169 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":193
 * 
 *     @staticmethod
 *     cdef envelope_conformer _create():             # <<<<<<<<<<<<<<
 *         return envelope_conformer.__new__(envelope_conformer)
 * 
 */

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create(void) {

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_r = NULL;

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":194
 *     @staticmethod
 *     cdef envelope_conformer _create():
 *         return envelope_conformer.__new__(envelope_conformer)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void acquire(self, list experimental, list theoretical):
 */

#line 194 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 194 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)

#line 194 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

#line 194 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_t_1);

#line 194 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 194 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":193
 * 
 *     @staticmethod
 *     cdef envelope_conformer _create():             # <<<<<<<<<<<<<<
 *         return envelope_conformer.__new__(envelope_conformer)
 * 
 */

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.envelope_conformer._create", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 193 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":196
 *         return envelope_conformer.__new__(envelope_conformer)
 * 
 *     cdef inline void acquire(self, list experimental, list theoretical):             # <<<<<<<<<<<<<<
 *         self.experimental = experimental
 *         self.theoretical = theoretical
 */

#line 196 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 196 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_acquire(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical) {

#line 196 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 196 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("acquire", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":197
 * 
 *     cdef inline void acquire(self, list experimental, list theoretical):
 *         self.experimental = experimental             # <<<<<<<<<<<<<<
 *         self.theoretical = theoretical
 *         self.n_missing = 0
 */

#line 197 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_experimental);

#line 197 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_v_experimental);

#line 197 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->experimental);

#line 197 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_v_self->experimental);

#line 197 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->experimental = __pyx_v_experimental;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":198
 *     cdef inline void acquire(self, list experimental, list theoretical):
 *         self.experimental = experimental
 *         self.theoretical = theoretical             # <<<<<<<<<<<<<<
 *         self.n_missing = 0
 * 
 */

#line 198 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_theoretical);

#line 198 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_v_theoretical);

#line 198 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->theoretical);

#line 198 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_v_self->theoretical);

#line 198 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->theoretical = __pyx_v_theoretical;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":199
 *         self.experimental = experimental
 *         self.theoretical = theoretical
 *         self.n_missing = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void conform(self, double minimum_theoretical_abundance=0.05):
 */

#line 199 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->n_missing = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":196
 *         return envelope_conformer.__new__(envelope_conformer)
 * 
 *     cdef inline void acquire(self, list experimental, list theoretical):             # <<<<<<<<<<<<<<
 *         self.experimental = experimental
 *         self.theoretical = theoretical
 */

#line 196 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 196 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 196 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 196 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":201
 *         self.n_missing = 0
 * 
 *     cdef inline void conform(self, double minimum_theoretical_abundance=0.05):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total = 0
 */

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static CYTHON_INLINE void __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform *__pyx_optional_args) {

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_minimum_theoretical_abundance = ((double)0.05);
  double __pyx_v_total
#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_n_missing
#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_i
#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_cleaned_eid = 0
#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_tid = 0
#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_fpeak = 0
#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_tpeak = 0
#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_peak = 0
#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  Py_ssize_t __pyx_t_2;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  Py_ssize_t __pyx_t_3;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_4;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_6;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_7;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_8;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("conform", 0);

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_optional_args) {

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_minimum_theoretical_abundance = __pyx_optional_args->minimum_theoretical_abundance;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":203
 *     cdef inline void conform(self, double minimum_theoretical_abundance=0.05):
 *         cdef:
 *             double total = 0             # <<<<<<<<<<<<<<
 *             size_t n_missing = 0
 *             size_t i = 0
 */

#line 203 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_total = 0.0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":204
 *         cdef:
 *             double total = 0
 *             size_t n_missing = 0             # <<<<<<<<<<<<<<
 *             size_t i = 0
 *             list cleaned_eid
 */

#line 204 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_n_missing = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":205
 *             double total = 0
 *             size_t n_missing = 0
 *             size_t i = 0             # <<<<<<<<<<<<<<
 *             list cleaned_eid
 *             list tid
 */

#line 205 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":211
 *             TheoreticalPeak tpeak, peak
 * 
 *         cleaned_eid = []             # <<<<<<<<<<<<<<
 * 
 *         for i in range(PyList_GET_SIZE(self.experimental)):
 */

#line 211 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 211, __pyx_L1_error)

#line 211 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 211 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_cleaned_eid = ((PyObject*)__pyx_t_1);

#line 211 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":213
 *         cleaned_eid = []
 * 
 *         for i in range(PyList_GET_SIZE(self.experimental)):             # <<<<<<<<<<<<<<
 *             fpeak = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 *             if fpeak is None:
 */

#line 213 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_self->experimental;

#line 213 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 213 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1);

#line 213 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 213 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 213 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 213 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":214
 * 
 *         for i in range(PyList_GET_SIZE(self.experimental)):
 *             fpeak = <FittedPeak>PyList_GET_ITEM(self.experimental, i)             # <<<<<<<<<<<<<<
 *             if fpeak is None:
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 */

#line 214 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = __pyx_v_self->experimental;

#line 214 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 214 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 214 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 214 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_5);

#line 214 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 214 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_fpeak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 214 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":215
 *         for i in range(PyList_GET_SIZE(self.experimental)):
 *             fpeak = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 *             if fpeak is None:             # <<<<<<<<<<<<<<
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 */

#line 215 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_6 = (((PyObject *)__pyx_v_fpeak) == Py_None);

#line 215 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_7 = (__pyx_t_6 != 0);

#line 215 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":216
 *             fpeak = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 *             if fpeak is None:
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)             # <<<<<<<<<<<<<<
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 *                 if peak.intensity > minimum_theoretical_abundance:
 */

#line 216 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = __pyx_v_self->theoretical;

#line 216 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 216 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_t_5);

#line 216 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 216 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":217
 *             if fpeak is None:
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)             # <<<<<<<<<<<<<<
 *                 if peak.intensity > minimum_theoretical_abundance:
 *                     n_missing += 1
 */

#line 217 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak->_create(__pyx_v_peak->mz, 1.0, 1.0, -1.0, -1.0, 0.0, 1, 0, 0.0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)

#line 217 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 217 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF_SET(__pyx_v_fpeak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 217 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":218
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 *                 if peak.intensity > minimum_theoretical_abundance:             # <<<<<<<<<<<<<<
 *                     n_missing += 1
 *             total += fpeak.intensity
 */

#line 218 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_7 = ((__pyx_v_peak->intensity > __pyx_v_minimum_theoretical_abundance) != 0);

#line 218 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_t_7) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":219
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 *                 if peak.intensity > minimum_theoretical_abundance:
 *                     n_missing += 1             # <<<<<<<<<<<<<<
 *             total += fpeak.intensity
 *             cleaned_eid.append(fpeak)
 */

#line 219 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_v_n_missing = (__pyx_v_n_missing + 1);

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":218
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 *                 if peak.intensity > minimum_theoretical_abundance:             # <<<<<<<<<<<<<<
 *                     n_missing += 1
 *             total += fpeak.intensity
 */

#line 218 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":215
 *         for i in range(PyList_GET_SIZE(self.experimental)):
 *             fpeak = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 *             if fpeak is None:             # <<<<<<<<<<<<<<
 *                 peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *                 fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
 */

#line 215 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":220
 *                 if peak.intensity > minimum_theoretical_abundance:
 *                     n_missing += 1
 *             total += fpeak.intensity             # <<<<<<<<<<<<<<
 *             cleaned_eid.append(fpeak)
 *             i += 1
 */

#line 220 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_total = (__pyx_v_total + __pyx_v_fpeak->__pyx_base.intensity);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":221
 *                     n_missing += 1
 *             total += fpeak.intensity
 *             cleaned_eid.append(fpeak)             # <<<<<<<<<<<<<<
 *             i += 1
 * 
 */

#line 221 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_cleaned_eid, ((PyObject *)__pyx_v_fpeak)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 221, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":222
 *             total += fpeak.intensity
 *             cleaned_eid.append(fpeak)
 *             i += 1             # <<<<<<<<<<<<<<
 * 
 *         tid = []
 */

#line 222 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = (__pyx_v_i + 1);

#line 222 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":224
 *             i += 1
 * 
 *         tid = []             # <<<<<<<<<<<<<<
 *         for i in range(PyList_GET_SIZE(self.theoretical)):
 *             tpeak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 */

#line 224 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L1_error)

#line 224 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 224 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_tid = ((PyObject*)__pyx_t_1);

#line 224 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":225
 * 
 *         tid = []
 *         for i in range(PyList_GET_SIZE(self.theoretical)):             # <<<<<<<<<<<<<<
 *             tpeak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *             peak = tpeak.clone()
 */

#line 225 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_self->theoretical;

#line 225 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 225 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1);

#line 225 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 225 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 225 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 225 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":226
 *         tid = []
 *         for i in range(PyList_GET_SIZE(self.theoretical)):
 *             tpeak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)             # <<<<<<<<<<<<<<
 *             peak = tpeak.clone()
 *             peak.intensity *= total
 */

#line 226 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = __pyx_v_self->theoretical;

#line 226 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 226 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 226 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 226 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_5);

#line 226 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 226 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_tpeak, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 226 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":227
 *         for i in range(PyList_GET_SIZE(self.theoretical)):
 *             tpeak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *             peak = tpeak.clone()             # <<<<<<<<<<<<<<
 *             peak.intensity *= total
 *             tid.append(peak)
 */

#line 227 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_tpeak->__pyx_vtab)->clone(__pyx_v_tpeak, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)

#line 227 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 227 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 227 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":228
 *             tpeak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
 *             peak = tpeak.clone()
 *             peak.intensity *= total             # <<<<<<<<<<<<<<
 *             tid.append(peak)
 *         self.experimental = cleaned_eid
 */

#line 228 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_peak->intensity = (__pyx_v_peak->intensity * __pyx_v_total);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":229
 *             peak = tpeak.clone()
 *             peak.intensity *= total
 *             tid.append(peak)             # <<<<<<<<<<<<<<
 *         self.experimental = cleaned_eid
 *         self.theoretical = tid
 */

#line 229 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_tid, ((PyObject *)__pyx_v_peak)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 229, __pyx_L1_error)

#line 229 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":230
 *             peak.intensity *= total
 *             tid.append(peak)
 *         self.experimental = cleaned_eid             # <<<<<<<<<<<<<<
 *         self.theoretical = tid
 *         self.n_missing = n_missing
 */

#line 230 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_cleaned_eid);

#line 230 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_v_cleaned_eid);

#line 230 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->experimental);

#line 230 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_v_self->experimental);

#line 230 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->experimental = __pyx_v_cleaned_eid;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":231
 *             tid.append(peak)
 *         self.experimental = cleaned_eid
 *         self.theoretical = tid             # <<<<<<<<<<<<<<
 *         self.n_missing = n_missing
 * 
 */

#line 231 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_tid);

#line 231 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_v_tid);

#line 231 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->theoretical);

#line 231 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_v_self->theoretical);

#line 231 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->theoretical = __pyx_v_tid;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":232
 *         self.experimental = cleaned_eid
 *         self.theoretical = tid
 *         self.n_missing = n_missing             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 232 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->n_missing = __pyx_v_n_missing;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":201
 *         self.n_missing = 0
 * 
 *     cdef inline void conform(self, double minimum_theoretical_abundance=0.05):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total = 0
 */

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.feature_processor.envelope_conformer.conform", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_cleaned_eid);

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_tid);

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_fpeak);

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_tpeak);

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 201 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer___reduce_cython__[] = "envelope_conformer.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer___reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer___reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  int __pyx_t_4;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.experimental, self.n_missing, self.theoretical)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->n_missing); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->experimental);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->experimental);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->experimental);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->theoretical);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->theoretical);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->theoretical);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.experimental, self.n_missing, self.theoretical)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_2;

#line 6 "(tree fragment)"
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.experimental, self.n_missing, self.theoretical)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_3 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));

#line 8 "(tree fragment)"
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.experimental is not None or self.theoretical is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.experimental, self.n_missing, self.theoretical)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.experimental is not None or self.theoretical is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_3 = (__pyx_v_self->experimental != ((PyObject*)Py_None));

#line 11 "(tree fragment)"
    __pyx_t_5 = (__pyx_t_3 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_5) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_4 = __pyx_t_5;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_5 = (__pyx_v_self->theoretical != ((PyObject*)Py_None));

#line 11 "(tree fragment)"
    __pyx_t_3 = (__pyx_t_5 != 0);

#line 11 "(tree fragment)"
    __pyx_t_4 = __pyx_t_3;

#line 11 "(tree fragment)"
    __pyx_L4_bool_binop_done:;

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_4;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.experimental is not None or self.theoretical is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_4 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_4) {

    /* "(tree fragment)":13
 *         use_setstate = self.experimental is not None or self.theoretical is not None
 *     if use_setstate:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_envelope_conforme); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_195672028);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_195672028);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_195672028);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    __pyx_t_2 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_6;

#line 13 "(tree fragment)"
    __pyx_t_6 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.experimental is not None or self.theoretical is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, None), state
 *     else:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_envelope_conformer__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_envelope_conforme); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 15 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_195672028);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_195672028);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_195672028);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_6);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);

#line 15 "(tree fragment)"
    __pyx_t_6 = 0;

#line 15 "(tree fragment)"
    __pyx_t_2 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_1;

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.envelope_conformer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_envelope_conformer__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_2__setstate_cython__[] = "envelope_conformer.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_2__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_2__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_envelope_conformer__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_envelope_conformer__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_envelope_conformer, (type(self), 0xba9b7dc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_envelope_conformer__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.envelope_conformer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":235
 * 
 * 
 * cpdef tuple conform_envelopes(list experimental, list base_theoretical,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.01):
 *     cdef:
 */

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_1conform_envelopes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes *__pyx_optional_args) {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_minimum_theoretical_abundance = ((double)0.01);
  size_t __pyx_v_n_missing
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_out = 0
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes __pyx_t_2;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("conform_envelopes", 0);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_optional_args) {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_minimum_theoretical_abundance = __pyx_optional_args->minimum_theoretical_abundance;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":241
 *         tuple out
 * 
 *     out = _conform_envelopes(experimental, base_theoretical, &n_missing, minimum_theoretical_abundance)             # <<<<<<<<<<<<<<
 *     out += (n_missing,)
 *     return out
 */

#line 241 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 241 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.minimum_theoretical_abundance = __pyx_v_minimum_theoretical_abundance;

#line 241 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes(__pyx_v_experimental, __pyx_v_base_theoretical, (&__pyx_v_n_missing), &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)

#line 241 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 241 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_out = ((PyObject*)__pyx_t_1);

#line 241 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":242
 * 
 *     out = _conform_envelopes(experimental, base_theoretical, &n_missing, minimum_theoretical_abundance)
 *     out += (n_missing,)             # <<<<<<<<<<<<<<
 *     return out
 * 
 */

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_n_missing); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 242, __pyx_L1_error)

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_out, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF_SET(__pyx_v_out, ((PyObject*)__pyx_t_1));

#line 242 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":243
 *     out = _conform_envelopes(experimental, base_theoretical, &n_missing, minimum_theoretical_abundance)
 *     out += (n_missing,)
 *     return out             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 243 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 243 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_out);

#line 243 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_out;

#line 243 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":235
 * 
 * 
 * cpdef tuple conform_envelopes(list experimental, list base_theoretical,             # <<<<<<<<<<<<<<
 *                               double minimum_theoretical_abundance=0.01):
 *     cdef:
 */

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_out);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_1conform_envelopes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes[] = "conform_envelopes(list experimental, list base_theoretical, double minimum_theoretical_abundance=0.01) -> tuple";

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_1conform_envelopes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_experimental = 0
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_base_theoretical = 0
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_minimum_theoretical_abundance
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("conform_envelopes (wrapper)", 0);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_experimental,&__pyx_n_s_base_theoretical,&__pyx_n_s_minimum_theoretical_abundance,0};

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyObject* values[3] = {0,0,0};

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_kwds)) {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      Py_ssize_t kw_args;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {
        case  3: 
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0: break;
        default: 
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0:

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental)) != 0)) kw_args--;
        else 
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  1:

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_base_theoretical)) != 0)) kw_args--;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("conform_envelopes", 0, 2, 3, 1); 
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 235, __pyx_L3_error)

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  2:

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_theoretical_abundance);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(kw_args > 0)) {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "conform_envelopes") < 0)) __PYX_ERR(0, 235, __pyx_L3_error)

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        break;
        default: 
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_experimental = ((PyObject*)values[0]);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_base_theoretical = ((PyObject*)values[1]);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[2]) {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_minimum_theoretical_abundance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_minimum_theoretical_abundance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 236, __pyx_L3_error)

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_minimum_theoretical_abundance = ((double)0.01);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("conform_envelopes", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 235, __pyx_L3_error)

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L3_error:;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return NULL;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental), (&PyList_Type), 1, "experimental", 1))) __PYX_ERR(0, 235, __pyx_L1_error)

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_base_theoretical), (&PyList_Type), 1, "base_theoretical", 1))) __PYX_ERR(0, 235, __pyx_L1_error)
  __pyx_r = 
#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(__pyx_self, __pyx_v_experimental, __pyx_v_base_theoretical, __pyx_v_minimum_theoretical_abundance);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, double __pyx_v_minimum_theoretical_abundance) {

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes __pyx_t_2;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("conform_envelopes", 0);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.minimum_theoretical_abundance = __pyx_v_minimum_theoretical_abundance;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(__pyx_v_experimental, __pyx_v_base_theoretical, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 235 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":246
 * 
 * 
 * cdef int has_no_valid_features(list features):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n, cnt
 */

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_has_no_valid_features(PyObject *__pyx_v_features) {
  size_t __pyx_v_i
#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_n
#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_cnt
#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_feature = 0
#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_1;

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_2;

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_3;

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_4;

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_6;

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_7;

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("has_no_valid_features", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":251
 *         object feature
 * 
 *     n = PyList_GET_SIZE(features)             # <<<<<<<<<<<<<<
 *     cnt = 0
 *     for i in range(n):
 */

#line 251 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_features);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":252
 * 
 *     n = PyList_GET_SIZE(features)
 *     cnt = 0             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         feature = <object>PyList_GET_ITEM(features, i)
 */

#line 252 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_cnt = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":253
 *     n = PyList_GET_SIZE(features)
 *     cnt = 0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         feature = <object>PyList_GET_ITEM(features, i)
 *         if feature is None:
 */

#line 253 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 253 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 253 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 253 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":254
 *     cnt = 0
 *     for i in range(n):
 *         feature = <object>PyList_GET_ITEM(features, i)             # <<<<<<<<<<<<<<
 *         if feature is None:
 *             cnt += 1
 */

#line 254 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_features, __pyx_v_i);

#line 254 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = ((PyObject *)__pyx_t_4);

#line 254 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 254 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_feature, __pyx_t_5);

#line 254 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":255
 *     for i in range(n):
 *         feature = <object>PyList_GET_ITEM(features, i)
 *         if feature is None:             # <<<<<<<<<<<<<<
 *             cnt += 1
 *     return cnt == n
 */

#line 255 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_6 = (__pyx_v_feature == Py_None);

#line 255 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_7 = (__pyx_t_6 != 0);

#line 255 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":256
 *         feature = <object>PyList_GET_ITEM(features, i)
 *         if feature is None:
 *             cnt += 1             # <<<<<<<<<<<<<<
 *     return cnt == n
 * 
 */

#line 256 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_cnt = (__pyx_v_cnt + 1);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":255
 *     for i in range(n):
 *         feature = <object>PyList_GET_ITEM(features, i)
 *         if feature is None:             # <<<<<<<<<<<<<<
 *             cnt += 1
 *     return cnt == n
 */

#line 255 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 255 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":257
 *         if feature is None:
 *             cnt += 1
 *     return cnt == n             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 257 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = (__pyx_v_cnt == __pyx_v_n);

#line 257 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":246
 * 
 * 
 * cdef int has_no_valid_features(list features):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n, cnt
 */

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_feature);

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 246 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":269
 *         public AveragineCache averagine
 * 
 *     cpdef TheoreticalIsotopicPattern create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8,             # <<<<<<<<<<<<<<
 *                                                                      double ignore_below=0.05):
 *         cdef:
 */

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_1create_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution *__pyx_optional_args) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_charge_carrier = __pyx_k__3;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_truncate_after = ((double)0.8);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_ignore_below = ((double)0.05);
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_base_tid = 0
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_r = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_10;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster __pyx_t_12;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("create_theoretical_distribution", 0);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_optional_args) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_v_truncate_after = __pyx_optional_args->truncate_after;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__pyx_optional_args->__pyx_n > 2) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_v_ignore_below = __pyx_optional_args->ignore_below;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if called by wrapper */

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if overridden in Python */

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_create_theoretical_distribution); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_1create_theoretical_distribution)) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_5);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_truncate_after); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_ignore_below); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_8 = __pyx_t_1; 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_t_9 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_10 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (likely(__pyx_t_9)) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(__pyx_t_9);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(function);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_DECREF_SET(__pyx_t_8, function);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __pyx_t_10 = 1;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCALL

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (PyFunction_Check(__pyx_t_8)) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 5+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCCALL

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[6] = {__pyx_t_9, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 5+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_11 = PyTuple_New(5+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_11);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (__pyx_t_9) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_t_3);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_4);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_5);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_t_5);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_6);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_10, __pyx_t_6);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_7);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 4+__pyx_t_10, __pyx_t_7);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_3 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_4 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_5 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_6 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_7 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern))))) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_2);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_2 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        goto __pyx_L0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #endif

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":273
 *         cdef:
 *             TheoreticalIsotopicPattern base_tid
 *         base_tid = self.averagine.isotopic_cluster(             # <<<<<<<<<<<<<<
 *             mz, charge, truncate_after=truncate_after, ignore_below=ignore_below,
 *             charge_carrier=charge_carrier)
 */

#line 273 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_12.__pyx_n = 4;

#line 273 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_12.charge = __pyx_v_charge;

#line 273 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_12.charge_carrier = __pyx_v_charge_carrier;

#line 273 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_12.truncate_after = __pyx_v_truncate_after;

#line 273 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_12.ignore_below = __pyx_v_ignore_below;

#line 273 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *)__pyx_v_self->averagine->__pyx_vtab)->isotopic_cluster(__pyx_v_self->averagine, __pyx_v_mz, 0, &__pyx_t_12)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 273 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_base_tid = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_1);

#line 273 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":276
 *             mz, charge, truncate_after=truncate_after, ignore_below=ignore_below,
 *             charge_carrier=charge_carrier)
 *         return base_tid             # <<<<<<<<<<<<<<
 * 
 *     cpdef list find_all_features(self, double mz, double error_tolerance=2e-5):
 */

#line 276 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 276 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_base_tid));

#line 276 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_base_tid;

#line 276 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":269
 *         public AveragineCache averagine
 * 
 *     cpdef TheoreticalIsotopicPattern create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8,             # <<<<<<<<<<<<<<
 *                                                                      double ignore_below=0.05):
 *         cdef:
 */

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_9);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_11);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.create_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_base_tid);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_1create_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution[] = "LCMSFeatureProcessorBase.create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8, double ignore_below=0.05) -> TheoreticalIsotopicPattern";

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_1create_theoretical_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  int __pyx_v_charge
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_charge_carrier
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_truncate_after
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_ignore_below
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("create_theoretical_distribution (wrapper)", 0);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_charge,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_ignore_below,0};

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyObject* values[5] = {0,0,0,0,0};

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_kwds)) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      Py_ssize_t kw_args;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {
        case  5: 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0: break;
        default: 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0:

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  1:

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("create_theoretical_distribution", 0, 2, 5, 1); 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 269, __pyx_L3_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  2:

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  3:

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[3] = value; kw_args--; }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  4:

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ignore_below);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(kw_args > 0)) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_theoretical_distribution") < 0)) __PYX_ERR(0, 269, __pyx_L3_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        break;
        default: 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 269, __pyx_L3_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_charge = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 269, __pyx_L3_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[2]) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 269, __pyx_L3_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_k__3;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[3]) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_truncate_after = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_truncate_after == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 269, __pyx_L3_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_truncate_after = ((double)0.8);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[4]) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_ignore_below = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_ignore_below == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 270, __pyx_L3_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_ignore_below = ((double)0.05);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_theoretical_distribution", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 269, __pyx_L3_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L3_error:;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.create_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_mz, __pyx_v_charge, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_ignore_below);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_v_charge, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below) {

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution __pyx_t_2;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("create_theoretical_distribution", 0);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.__pyx_n = 3;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.ignore_below = __pyx_v_ignore_below;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->create_theoretical_distribution(__pyx_v_self, __pyx_v_mz, __pyx_v_charge, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.create_theoretical_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":278
 *         return base_tid
 * 
 *     cpdef list find_all_features(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         return self.feature_map.find_all(mz, error_tolerance)
 * 
 */

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_3find_all_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features *__pyx_optional_args) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_error_tolerance = ((double)2e-5);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_7;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all __pyx_t_9;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("find_all_features", 0);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_optional_args) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if called by wrapper */

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if overridden in Python */

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_all_features); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 278, __pyx_L1_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_3find_all_features)) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 278, __pyx_L1_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 278, __pyx_L1_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; 
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_t_6 = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_7 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (likely(__pyx_t_6)) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(__pyx_t_6);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(function);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_DECREF_SET(__pyx_t_5, function);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __pyx_t_7 = 1;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCALL

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (PyFunction_Check(__pyx_t_5)) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 278, __pyx_L1_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCCALL

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 278, __pyx_L1_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 278, __pyx_L1_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_8);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (__pyx_t_6) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_3 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_4 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 278, __pyx_L1_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 278, __pyx_L1_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_2 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        goto __pyx_L0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #endif

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":279
 * 
 *     cpdef list find_all_features(self, double mz, double error_tolerance=2e-5):
 *         return self.feature_map.find_all(mz, error_tolerance)             # <<<<<<<<<<<<<<
 * 
 *     cpdef list find_features(self, double mz, double error_tolerance=2e-5, LCMSFeature interval=None):
 */

#line 279 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 279 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_9.__pyx_n = 1;

#line 279 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_9.error_tolerance = __pyx_v_error_tolerance;

#line 279 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->feature_map->__pyx_vtab)->find_all(__pyx_v_self->feature_map, __pyx_v_mz, 0, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)

#line 279 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 279 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = ((PyObject*)__pyx_t_1);

#line 279 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 279 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":278
 *         return base_tid
 * 
 *     cpdef list find_all_features(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         return self.feature_map.find_all(mz, error_tolerance)
 * 
 */

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_all_features", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_3find_all_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_2find_all_features[] = "LCMSFeatureProcessorBase.find_all_features(self, double mz, double error_tolerance=2e-5) -> list";

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_3find_all_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_error_tolerance
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("find_all_features (wrapper)", 0);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_error_tolerance,0};

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyObject* values[2] = {0,0};

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_kwds)) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      Py_ssize_t kw_args;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {
        case  2: 
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0: break;
        default: 
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0:

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else 
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  1:

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(kw_args > 0)) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_all_features") < 0)) __PYX_ERR(0, 278, __pyx_L3_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: 
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        break;
        default: 
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 278, __pyx_L3_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[1]) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 278, __pyx_L3_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_error_tolerance = ((double)2e-5);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_all_features", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 278, __pyx_L3_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L3_error:;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_all_features", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_2find_all_features(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_mz, __pyx_v_error_tolerance);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_2find_all_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance) {

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features __pyx_t_2;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("find_all_features", 0);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->find_all_features(__pyx_v_self, __pyx_v_mz, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 278, __pyx_L1_error)

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_all_features", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 278 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":281
 *         return self.feature_map.find_all(mz, error_tolerance)
 * 
 *     cpdef list find_features(self, double mz, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list f, overlapped
 */

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_5find_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features *__pyx_optional_args) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_error_tolerance = ((double)2e-5);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
  PyObject *__pyx_v_f = 0
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_overlapped = 0
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_search_width
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_hit_width
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_el = 0
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_i
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_n
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  CYTHON_UNUSED int __pyx_v_has_miss
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_7;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features __pyx_t_9;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_10;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_11;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_12;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_13;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_14;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_15;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_16;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("find_features", 0);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_optional_args) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_v_interval = __pyx_optional_args->interval;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if called by wrapper */

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if overridden in Python */

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_features); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_5find_features)) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_t_6 = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_7 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (likely(__pyx_t_6)) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(__pyx_t_6);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(function);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_DECREF_SET(__pyx_t_5, function);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __pyx_t_7 = 1;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCALL

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (PyFunction_Check(__pyx_t_5)) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)};

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCCALL

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)};

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_8 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_8);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (__pyx_t_6) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_interval));

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_interval));

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, ((PyObject *)__pyx_v_interval));

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_3 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_4 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_2 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        goto __pyx_L0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #endif

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":288
 *             size_t i, n
 *             bint has_miss
 *         has_miss = False             # <<<<<<<<<<<<<<
 *         f = self.find_all_features(mz, error_tolerance)
 *         if not f:
 */

#line 288 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_has_miss = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":289
 *             bint has_miss
 *         has_miss = False
 *         f = self.find_all_features(mz, error_tolerance)             # <<<<<<<<<<<<<<
 *         if not f:
 *             f = [None]
 */

#line 289 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_9.__pyx_n = 1;

#line 289 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_9.error_tolerance = __pyx_v_error_tolerance;

#line 289 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->find_all_features(__pyx_v_self, __pyx_v_mz, 0, &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)

#line 289 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 289 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_f = ((PyObject*)__pyx_t_1);

#line 289 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":290
 *         has_miss = False
 *         f = self.find_all_features(mz, error_tolerance)
 *         if not f:             # <<<<<<<<<<<<<<
 *             f = [None]
 *             has_miss = True
 */

#line 290 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_10 = (__pyx_v_f != Py_None)&&(PyList_GET_SIZE(__pyx_v_f) != 0);

#line 290 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_11 = ((!__pyx_t_10) != 0);

#line 290 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_t_11) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":291
 *         f = self.find_all_features(mz, error_tolerance)
 *         if not f:
 *             f = [None]             # <<<<<<<<<<<<<<
 *             has_miss = True
 *         elif interval is not None:
 */

#line 291 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)

#line 291 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 291 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(Py_None);

#line 291 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GIVEREF(Py_None);

#line 291 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyList_SET_ITEM(__pyx_t_1, 0, Py_None);

#line 291 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF_SET(__pyx_v_f, ((PyObject*)__pyx_t_1));

#line 291 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":292
 *         if not f:
 *             f = [None]
 *             has_miss = True             # <<<<<<<<<<<<<<
 *         elif interval is not None:
 *             overlapped = []
 */

#line 292 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_has_miss = 1;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":290
 *         has_miss = False
 *         f = self.find_all_features(mz, error_tolerance)
 *         if not f:             # <<<<<<<<<<<<<<
 *             f = [None]
 *             has_miss = True
 */

#line 290 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    goto __pyx_L3;

#line 290 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":293
 *             f = [None]
 *             has_miss = True
 *         elif interval is not None:             # <<<<<<<<<<<<<<
 *             overlapped = []
 *             n = PyList_GET_SIZE(f)
 */

#line 293 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_11 = (((PyObject *)__pyx_v_interval) != Py_None);

#line 293 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_10 = (__pyx_t_11 != 0);

#line 293 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":294
 *             has_miss = True
 *         elif interval is not None:
 *             overlapped = []             # <<<<<<<<<<<<<<
 *             n = PyList_GET_SIZE(f)
 *             search_width = interval.get_end_time() - interval.get_start_time()
 */

#line 294 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 294 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_overlapped = ((PyObject*)__pyx_t_1);

#line 294 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":295
 *         elif interval is not None:
 *             overlapped = []
 *             n = PyList_GET_SIZE(f)             # <<<<<<<<<<<<<<
 *             search_width = interval.get_end_time() - interval.get_start_time()
 *             if search_width == 0:
 */

#line 295 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_n = PyList_GET_SIZE(__pyx_v_f);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":296
 *             overlapped = []
 *             n = PyList_GET_SIZE(f)
 *             search_width = interval.get_end_time() - interval.get_start_time()             # <<<<<<<<<<<<<<
 *             if search_width == 0:
 *                 return [None]
 */

#line 296 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_search_width = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_interval->__pyx_base.__pyx_vtab)->__pyx_base.get_end_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_interval)) - ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_interval->__pyx_base.__pyx_vtab)->__pyx_base.get_start_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_interval)));

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":297
 *             n = PyList_GET_SIZE(f)
 *             search_width = interval.get_end_time() - interval.get_start_time()
 *             if search_width == 0:             # <<<<<<<<<<<<<<
 *                 return [None]
 *             for i in range(n):
 */

#line 297 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_10 = ((__pyx_v_search_width == 0.0) != 0);

#line 297 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":298
 *             search_width = interval.get_end_time() - interval.get_start_time()
 *             if search_width == 0:
 *                 return [None]             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)
 */

#line 298 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 298 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)

#line 298 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 298 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(Py_None);

#line 298 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GIVEREF(Py_None);

#line 298 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PyList_SET_ITEM(__pyx_t_1, 0, Py_None);

#line 298 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_r = ((PyObject*)__pyx_t_1);

#line 298 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = 0;

#line 298 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":297
 *             n = PyList_GET_SIZE(f)
 *             search_width = interval.get_end_time() - interval.get_start_time()
 *             if search_width == 0:             # <<<<<<<<<<<<<<
 *                 return [None]
 *             for i in range(n):
 */

#line 297 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":299
 *             if search_width == 0:
 *                 return [None]
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)
 *                 if el.overlaps_in_time(interval):
 */

#line 299 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_12 = __pyx_v_n;

#line 299 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_13 = __pyx_t_12;

#line 299 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {

#line 299 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_i = __pyx_t_14;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":300
 *                 return [None]
 *             for i in range(n):
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)             # <<<<<<<<<<<<<<
 *                 if el.overlaps_in_time(interval):
 *                     hit_width = el.get_end_time() - el.get_start_time()
 */

#line 300 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_15 = PyList_GET_ITEM(__pyx_v_f, __pyx_v_i);

#line 300 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_t_15);

#line 300 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 300 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_XDECREF_SET(__pyx_v_el, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1));

#line 300 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":301
 *             for i in range(n):
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)
 *                 if el.overlaps_in_time(interval):             # <<<<<<<<<<<<<<
 *                     hit_width = el.get_end_time() - el.get_start_time()
 *                     if (hit_width / search_width) < 0.05:
 */

#line 301 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_10 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_el->__pyx_base.__pyx_vtab)->overlaps_in_time(__pyx_v_el, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_interval), 0) != 0);

#line 301 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_t_10) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":302
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)
 *                 if el.overlaps_in_time(interval):
 *                     hit_width = el.get_end_time() - el.get_start_time()             # <<<<<<<<<<<<<<
 *                     if (hit_width / search_width) < 0.05:
 *                         continue
 */

#line 302 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_v_hit_width = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_el->__pyx_base.__pyx_vtab)->__pyx_base.get_end_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_el)) - ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_el->__pyx_base.__pyx_vtab)->__pyx_base.get_start_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_el)));

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":303
 *                 if el.overlaps_in_time(interval):
 *                     hit_width = el.get_end_time() - el.get_start_time()
 *                     if (hit_width / search_width) < 0.05:             # <<<<<<<<<<<<<<
 *                         continue
 *                     overlapped.append(el)
 */

#line 303 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (unlikely(__pyx_v_search_width == 0)) {

#line 303 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 303 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __PYX_ERR(0, 303, __pyx_L1_error)

#line 303 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 303 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_10 = (((__pyx_v_hit_width / __pyx_v_search_width) < 0.05) != 0);

#line 303 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__pyx_t_10) {

          /* "ms_deisotope/_c/feature_map/feature_processor.pyx":304
 *                     hit_width = el.get_end_time() - el.get_start_time()
 *                     if (hit_width / search_width) < 0.05:
 *                         continue             # <<<<<<<<<<<<<<
 *                     overlapped.append(el)
 *             f = overlapped
 */

#line 304 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          goto __pyx_L5_continue;

          /* "ms_deisotope/_c/feature_map/feature_processor.pyx":303
 *                 if el.overlaps_in_time(interval):
 *                     hit_width = el.get_end_time() - el.get_start_time()
 *                     if (hit_width / search_width) < 0.05:             # <<<<<<<<<<<<<<
 *                         continue
 *                     overlapped.append(el)
 */

#line 303 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":305
 *                     if (hit_width / search_width) < 0.05:
 *                         continue
 *                     overlapped.append(el)             # <<<<<<<<<<<<<<
 *             f = overlapped
 *             if PyList_GET_SIZE(f) == 0:
 */

#line 305 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_overlapped, ((PyObject *)__pyx_v_el)); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 305, __pyx_L1_error)

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":301
 *             for i in range(n):
 *                 el = <LCMSFeature>PyList_GET_ITEM(f, i)
 *                 if el.overlaps_in_time(interval):             # <<<<<<<<<<<<<<
 *                     hit_width = el.get_end_time() - el.get_start_time()
 *                     if (hit_width / search_width) < 0.05:
 */

#line 301 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 301 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_L5_continue:;

#line 301 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":306
 *                         continue
 *                     overlapped.append(el)
 *             f = overlapped             # <<<<<<<<<<<<<<
 *             if PyList_GET_SIZE(f) == 0:
 *                 f = [None]
 */

#line 306 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_v_overlapped);

#line 306 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF_SET(__pyx_v_f, __pyx_v_overlapped);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":307
 *                     overlapped.append(el)
 *             f = overlapped
 *             if PyList_GET_SIZE(f) == 0:             # <<<<<<<<<<<<<<
 *                 f = [None]
 *         return f
 */

#line 307 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_10 = ((PyList_GET_SIZE(__pyx_v_f) == 0) != 0);

#line 307 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":308
 *             f = overlapped
 *             if PyList_GET_SIZE(f) == 0:
 *                 f = [None]             # <<<<<<<<<<<<<<
 *         return f
 * 
 */

#line 308 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 308, __pyx_L1_error)

#line 308 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 308 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(Py_None);

#line 308 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GIVEREF(Py_None);

#line 308 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PyList_SET_ITEM(__pyx_t_1, 0, Py_None);

#line 308 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF_SET(__pyx_v_f, ((PyObject*)__pyx_t_1));

#line 308 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":307
 *                     overlapped.append(el)
 *             f = overlapped
 *             if PyList_GET_SIZE(f) == 0:             # <<<<<<<<<<<<<<
 *                 f = [None]
 *         return f
 */

#line 307 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":293
 *             f = [None]
 *             has_miss = True
 *         elif interval is not None:             # <<<<<<<<<<<<<<
 *             overlapped = []
 *             n = PyList_GET_SIZE(f)
 */

#line 293 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 293 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L3:;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":309
 *             if PyList_GET_SIZE(f) == 0:
 *                 f = [None]
 *         return f             # <<<<<<<<<<<<<<
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5, LCMSFeature interval=None):
 */

#line 309 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 309 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_f);

#line 309 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_f;

#line 309 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":281
 *         return self.feature_map.find_all(mz, error_tolerance)
 * 
 *     cpdef list find_features(self, double mz, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list f, overlapped
 */

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_features", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_f);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_overlapped);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_el);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_5find_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_4find_features[] = "LCMSFeatureProcessorBase.find_features(self, double mz, double error_tolerance=2e-5, LCMSFeature interval=None) -> list";

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_5find_features(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_error_tolerance
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval = 0
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("find_features (wrapper)", 0);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_error_tolerance,&__pyx_n_s_interval,0};

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyObject* values[3] = {0,0,0};

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    values[2] = (PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_kwds)) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      Py_ssize_t kw_args;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {
        case  3: 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0: break;
        default: 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0:

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  1:

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  2:

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_interval);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(kw_args > 0)) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_features") < 0)) __PYX_ERR(0, 281, __pyx_L3_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        break;
        default: 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 281, __pyx_L3_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[1]) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 281, __pyx_L3_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_error_tolerance = ((double)2e-5);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_interval = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)values[2]);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_features", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 281, __pyx_L3_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L3_error:;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_features", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, 1, "interval", 0))) __PYX_ERR(0, 281, __pyx_L1_error)
  __pyx_r = 
#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_4find_features(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_mz, __pyx_v_error_tolerance, __pyx_v_interval);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_4find_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval) {

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features __pyx_t_2;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("find_features", 0);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.__pyx_n = 2;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.interval = __pyx_v_interval;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->find_features(__pyx_v_self, __pyx_v_mz, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.find_features", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 281 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":311
 *         return f
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         """Given a list of theoretical peaks, find their counterparts in :attr:`peaklist` within `error_tolerance`
 *         ppm error. If no experimental peak is found, a placeholder will be used in its stead.
 */

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_7match_theoretical_isotopic_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution *__pyx_optional_args) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_error_tolerance = ((double)2e-5);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
  PyObject *__pyx_v_experimental_distribution = 0
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_found = 0
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_i
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_n
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_p = 0
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_6;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_8;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_9;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_10;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_11;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features __pyx_t_12;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_13;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("match_theoretical_isotopic_distribution", 0);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_optional_args) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_v_interval = __pyx_optional_args->interval;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if called by wrapper */

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if overridden in Python */

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_match_theoretical_isotopic_distr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_7match_theoretical_isotopic_distribution)) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_t_5 = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_6 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (likely(__pyx_t_5)) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(function);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __pyx_t_6 = 1;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCALL

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (PyFunction_Check(__pyx_t_4)) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_theoretical_distribution, __pyx_t_3, ((PyObject *)__pyx_v_interval)};

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCCALL

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_theoretical_distribution, __pyx_t_3, ((PyObject *)__pyx_v_interval)};

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_7);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (__pyx_t_5) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_INCREF(__pyx_v_theoretical_distribution);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_v_theoretical_distribution);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_theoretical_distribution);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_interval));

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_interval));

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, ((PyObject *)__pyx_v_interval));

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_3 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_2 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        goto __pyx_L0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #endif

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":332
 *             TheoreticalPeak p
 * 
 *         experimental_distribution = []             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(theoretical_distribution)
 *         for i in range(n):
 */

#line 332 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)

#line 332 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 332 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_experimental_distribution = ((PyObject*)__pyx_t_1);

#line 332 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":333
 * 
 *         experimental_distribution = []
 *         n = PyList_GET_SIZE(theoretical_distribution)             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 */

#line 333 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_theoretical_distribution);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":334
 *         experimental_distribution = []
 *         n = PyList_GET_SIZE(theoretical_distribution)
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 *             found = self.find_features(p.mz, error_tolerance, interval=interval)
 */

#line 334 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_8 = __pyx_v_n;

#line 334 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_9 = __pyx_t_8;

#line 334 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 334 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":335
 *         n = PyList_GET_SIZE(theoretical_distribution)
 *         for i in range(n):
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)             # <<<<<<<<<<<<<<
 *             found = self.find_features(p.mz, error_tolerance, interval=interval)
 *             experimental_distribution.append(found)
 */

#line 335 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_theoretical_distribution, __pyx_v_i);

#line 335 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_11);

#line 335 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 335 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_p, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 335 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":336
 *         for i in range(n):
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 *             found = self.find_features(p.mz, error_tolerance, interval=interval)             # <<<<<<<<<<<<<<
 *             experimental_distribution.append(found)
 *         return experimental_distribution
 */

#line 336 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_12.__pyx_n = 2;

#line 336 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_12.error_tolerance = __pyx_v_error_tolerance;

#line 336 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_12.interval = __pyx_v_interval;

#line 336 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->find_features(__pyx_v_self, __pyx_v_p->mz, 0, &__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)

#line 336 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 336 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_found, ((PyObject*)__pyx_t_1));

#line 336 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":337
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 *             found = self.find_features(p.mz, error_tolerance, interval=interval)
 *             experimental_distribution.append(found)             # <<<<<<<<<<<<<<
 *         return experimental_distribution
 * 
 */

#line 337 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_experimental_distribution, __pyx_v_found); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 337, __pyx_L1_error)

#line 337 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":338
 *             found = self.find_features(p.mz, error_tolerance, interval=interval)
 *             experimental_distribution.append(found)
 *         return experimental_distribution             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple conform_envelopes(self, list experimental, list base_theoretical):
 */

#line 338 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 338 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_experimental_distribution);

#line 338 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_experimental_distribution;

#line 338 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":311
 *         return f
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5, LCMSFeature interval=None):             # <<<<<<<<<<<<<<
 *         """Given a list of theoretical peaks, find their counterparts in :attr:`peaklist` within `error_tolerance`
 *         ppm error. If no experimental peak is found, a placeholder will be used in its stead.
 */

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.match_theoretical_isotopic_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_experimental_distribution);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_found);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_p);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_7match_theoretical_isotopic_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6match_theoretical_isotopic_distribution[] = "LCMSFeatureProcessorBase.match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5, LCMSFeature interval=None) -> list\nGiven a list of theoretical peaks, find their counterparts in :attr:`peaklist` within `error_tolerance`\n        ppm error. If no experimental peak is found, a placeholder will be used in its stead.\n\n        Parameters\n        ----------\n        theoretical_distribution : list of TheoreticalPeak\n            The theoretical isotopic pattern to match\n        error_tolerance : float, optional\n            Parts-per-million error tolerance to permit in searching for matches\n\n        Returns\n        -------\n        list of list of LCMSFeature\n            The list of matched features for each theoretical m/z\n        ";

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_7match_theoretical_isotopic_distribution(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_theoretical_distribution = 0
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_error_tolerance
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval = 0
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("match_theoretical_isotopic_distribution (wrapper)", 0);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_theoretical_distribution,&__pyx_n_s_error_tolerance,&__pyx_n_s_interval,0};

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyObject* values[3] = {0,0,0};

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    values[2] = (PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_kwds)) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      Py_ssize_t kw_args;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {
        case  3: 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0: break;
        default: 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0:

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theoretical_distribution)) != 0)) kw_args--;
        else 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  1:

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  2:

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_interval);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(kw_args > 0)) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "match_theoretical_isotopic_distribution") < 0)) __PYX_ERR(0, 311, __pyx_L3_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        break;
        default: 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_theoretical_distribution = ((PyObject*)values[0]);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[1]) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 311, __pyx_L3_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_error_tolerance = ((double)2e-5);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_interval = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)values[2]);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("match_theoretical_isotopic_distribution", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 311, __pyx_L3_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L3_error:;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.match_theoretical_isotopic_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_theoretical_distribution), (&PyList_Type), 1, "theoretical_distribution", 1))) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, 1, "interval", 0))) __PYX_ERR(0, 311, __pyx_L1_error)
  __pyx_r = 
#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6match_theoretical_isotopic_distribution(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_theoretical_distribution, __pyx_v_error_tolerance, __pyx_v_interval);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6match_theoretical_isotopic_distribution(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_theoretical_distribution, double __pyx_v_error_tolerance, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_interval) {

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution __pyx_t_2;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("match_theoretical_isotopic_distribution", 0);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.__pyx_n = 2;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.interval = __pyx_v_interval;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->match_theoretical_isotopic_distribution(__pyx_v_self, __pyx_v_theoretical_distribution, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.match_theoretical_isotopic_distribution", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 311 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":340
 *         return experimental_distribution
 * 
 *     cpdef tuple conform_envelopes(self, list experimental, list base_theoretical):             # <<<<<<<<<<<<<<
 *         return conform_envelopes(experimental, base_theoretical)
 * 
 */

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9conform_envelopes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_conform_envelopes(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, int __pyx_skip_dispatch) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_5;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("conform_envelopes", 0);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if called by wrapper */

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if overridden in Python */

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_conform_envelopes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9conform_envelopes)) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_t_4 = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_5 = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (likely(__pyx_t_4)) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(function);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __pyx_t_5 = 1;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCALL

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_experimental, __pyx_v_base_theoretical};

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 340, __pyx_L1_error)

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCCALL

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_experimental, __pyx_v_base_theoretical};

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 340, __pyx_L1_error)

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 340, __pyx_L1_error)

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (__pyx_t_4) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_INCREF(__pyx_v_experimental);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_v_experimental);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_experimental);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_INCREF(__pyx_v_base_theoretical);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_v_base_theoretical);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_base_theoretical);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 340, __pyx_L1_error)

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (!(likely(PyTuple_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 340, __pyx_L1_error)

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_2 = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        goto __pyx_L0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #endif

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":341
 * 
 *     cpdef tuple conform_envelopes(self, list experimental, list base_theoretical):
 *         return conform_envelopes(experimental, base_theoretical)             # <<<<<<<<<<<<<<
 * 
 *     cdef tuple _conform_envelopes(self, list experimental, list base_theoretical, size_t* n_missing):
 */

#line 341 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 341 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes(__pyx_v_experimental, __pyx_v_base_theoretical, 0, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)

#line 341 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 341 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = ((PyObject*)__pyx_t_1);

#line 341 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 341 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":340
 *         return experimental_distribution
 * 
 *     cpdef tuple conform_envelopes(self, list experimental, list base_theoretical):             # <<<<<<<<<<<<<<
 *         return conform_envelopes(experimental, base_theoretical)
 * 
 */

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9conform_envelopes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_8conform_envelopes[] = "LCMSFeatureProcessorBase.conform_envelopes(self, list experimental, list base_theoretical) -> tuple";

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9conform_envelopes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_experimental = 0
#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_base_theoretical = 0
#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("conform_envelopes (wrapper)", 0);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_experimental,&__pyx_n_s_base_theoretical,0};

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyObject* values[2] = {0,0};

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_kwds)) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      Py_ssize_t kw_args;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {
        case  2: 
#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0: break;
        default: 
#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0:

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental)) != 0)) kw_args--;
        else 
#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  1:

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_base_theoretical)) != 0)) kw_args--;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("conform_envelopes", 1, 2, 2, 1); 
#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 340, __pyx_L3_error)

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(kw_args > 0)) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "conform_envelopes") < 0)) __PYX_ERR(0, 340, __pyx_L3_error)

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L5_argtuple_error;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_experimental = ((PyObject*)values[0]);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_base_theoretical = ((PyObject*)values[1]);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("conform_envelopes", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 340, __pyx_L3_error)

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L3_error:;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental), (&PyList_Type), 1, "experimental", 1))) __PYX_ERR(0, 340, __pyx_L1_error)

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_base_theoretical), (&PyList_Type), 1, "base_theoretical", 1))) __PYX_ERR(0, 340, __pyx_L1_error)
  __pyx_r = 
#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_8conform_envelopes(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_experimental, __pyx_v_base_theoretical);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_8conform_envelopes(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical) {

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("conform_envelopes", 0);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_conform_envelopes(__pyx_v_self, __pyx_v_experimental, __pyx_v_base_theoretical, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 340 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":343
 *         return conform_envelopes(experimental, base_theoretical)
 * 
 *     cdef tuple _conform_envelopes(self, list experimental, list base_theoretical, size_t* n_missing):             # <<<<<<<<<<<<<<
 *         return _conform_envelopes(experimental, base_theoretical, n_missing)
 * 
 */

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__conform_envelopes(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_experimental, PyObject *__pyx_v_base_theoretical, size_t *__pyx_v_n_missing) {

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_conform_envelopes", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":344
 * 
 *     cdef tuple _conform_envelopes(self, list experimental, list base_theoretical, size_t* n_missing):
 *         return _conform_envelopes(experimental, base_theoretical, n_missing)             # <<<<<<<<<<<<<<
 * 
 *     cdef double _find_thresholded_score(self, dvec* scores, double percentage):
 */

#line 344 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 344 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor__conform_envelopes(__pyx_v_experimental, __pyx_v_base_theoretical, __pyx_v_n_missing, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 344, __pyx_L1_error)

#line 344 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 344 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = ((PyObject*)__pyx_t_1);

#line 344 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 344 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":343
 *         return conform_envelopes(experimental, base_theoretical)
 * 
 *     cdef tuple _conform_envelopes(self, list experimental, list base_theoretical, size_t* n_missing):             # <<<<<<<<<<<<<<
 *         return _conform_envelopes(experimental, base_theoretical, n_missing)
 * 
 */

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._conform_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 343 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":346
 *         return _conform_envelopes(experimental, base_theoretical, n_missing)
 * 
 *     cdef double _find_thresholded_score(self, dvec* scores, double percentage):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n, count
 */

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__find_thresholded_score(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_scores, double __pyx_v_percentage) {
  size_t __pyx_v_i
#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_n
#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_count
#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_maximum
#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_val
#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_threshold
#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_acc
#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_r;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_1;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_2;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_3;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_4;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_find_thresholded_score", 0);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":351
 *             double maximum, val, threshold, acc
 * 
 *         maximum = -INF             # <<<<<<<<<<<<<<
 *         count = 0
 *         acc = 0
 */

#line 351 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_maximum = (-__pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_INF);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":352
 * 
 *         maximum = -INF
 *         count = 0             # <<<<<<<<<<<<<<
 *         acc = 0
 *         n = scores.used
 */

#line 352 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_count = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":353
 *         maximum = -INF
 *         count = 0
 *         acc = 0             # <<<<<<<<<<<<<<
 *         n = scores.used
 *         for i in range(n):
 */

#line 353 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_acc = 0.0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":354
 *         count = 0
 *         acc = 0
 *         n = scores.used             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             val = scores.v[i]
 */

#line 354 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_scores->used;

#line 354 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_n = __pyx_t_1;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":355
 *         acc = 0
 *         n = scores.used
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             val = scores.v[i]
 *             if val > maximum:
 */

#line 355 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 355 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 355 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 355 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":356
 *         n = scores.used
 *         for i in range(n):
 *             val = scores.v[i]             # <<<<<<<<<<<<<<
 *             if val > maximum:
 *                 maximum = val
 */

#line 356 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_val = (__pyx_v_scores->v[__pyx_v_i]);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":357
 *         for i in range(n):
 *             val = scores.v[i]
 *             if val > maximum:             # <<<<<<<<<<<<<<
 *                 maximum = val
 *         threshold = maximum * percentage
 */

#line 357 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_4 = ((__pyx_v_val > __pyx_v_maximum) != 0);

#line 357 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_4) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":358
 *             val = scores.v[i]
 *             if val > maximum:
 *                 maximum = val             # <<<<<<<<<<<<<<
 *         threshold = maximum * percentage
 *         for i in range(n):
 */

#line 358 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_maximum = __pyx_v_val;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":357
 *         for i in range(n):
 *             val = scores.v[i]
 *             if val > maximum:             # <<<<<<<<<<<<<<
 *                 maximum = val
 *         threshold = maximum * percentage
 */

#line 357 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 357 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":359
 *             if val > maximum:
 *                 maximum = val
 *         threshold = maximum * percentage             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             val = scores.v[i]
 */

#line 359 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_threshold = (__pyx_v_maximum * __pyx_v_percentage);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":360
 *                 maximum = val
 *         threshold = maximum * percentage
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             val = scores.v[i]
 *             if val > threshold:
 */

#line 360 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 360 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 360 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 360 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":361
 *         threshold = maximum * percentage
 *         for i in range(n):
 *             val = scores.v[i]             # <<<<<<<<<<<<<<
 *             if val > threshold:
 *                 acc += val
 */

#line 361 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_val = (__pyx_v_scores->v[__pyx_v_i]);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":362
 *         for i in range(n):
 *             val = scores.v[i]
 *             if val > threshold:             # <<<<<<<<<<<<<<
 *                 acc += val
 *                 count += 1
 */

#line 362 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_4 = ((__pyx_v_val > __pyx_v_threshold) != 0);

#line 362 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_4) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":363
 *             val = scores.v[i]
 *             if val > threshold:
 *                 acc += val             # <<<<<<<<<<<<<<
 *                 count += 1
 *         if count == 0:
 */

#line 363 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_acc = (__pyx_v_acc + __pyx_v_val);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":364
 *             if val > threshold:
 *                 acc += val
 *                 count += 1             # <<<<<<<<<<<<<<
 *         if count == 0:
 *             return 0
 */

#line 364 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_count = (__pyx_v_count + 1);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":362
 *         for i in range(n):
 *             val = scores.v[i]
 *             if val > threshold:             # <<<<<<<<<<<<<<
 *                 acc += val
 *                 count += 1
 */

#line 362 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 362 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":365
 *                 acc += val
 *                 count += 1
 *         if count == 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         if count == 1:
 */

#line 365 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_4 = ((__pyx_v_count == 0) != 0);

#line 365 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_t_4) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":366
 *                 count += 1
 *         if count == 0:
 *             return 0             # <<<<<<<<<<<<<<
 *         if count == 1:
 *             # Penalize singletons? This may harm more than help
 */

#line 366 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_r = 0.0;

#line 366 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":365
 *                 acc += val
 *                 count += 1
 *         if count == 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         if count == 1:
 */

#line 365 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":367
 *         if count == 0:
 *             return 0
 *         if count == 1:             # <<<<<<<<<<<<<<
 *             # Penalize singletons? This may harm more than help
 *             count += 1
 */

#line 367 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_4 = ((__pyx_v_count == 1) != 0);

#line 367 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_t_4) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":369
 *         if count == 1:
 *             # Penalize singletons? This may harm more than help
 *             count += 1             # <<<<<<<<<<<<<<
 *         return acc / count
 * 
 */

#line 369 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_count = (__pyx_v_count + 1);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":367
 *         if count == 0:
 *             return 0
 *         if count == 1:             # <<<<<<<<<<<<<<
 *             # Penalize singletons? This may harm more than help
 *             count += 1
 */

#line 367 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":370
 *             # Penalize singletons? This may harm more than help
 *             count += 1
 *         return acc / count             # <<<<<<<<<<<<<<
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,
 */

#line 370 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(__pyx_v_count == 0)) {

#line 370 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 370 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __PYX_ERR(0, 370, __pyx_L1_error)

#line 370 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 370 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = (__pyx_v_acc / __pyx_v_count);

#line 370 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":346
 *         return _conform_envelopes(experimental, base_theoretical, n_missing)
 * 
 *     cdef double _find_thresholded_score(self, dvec* scores, double percentage):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n, count
 */

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._find_thresholded_score", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 346 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":372
 *         return acc / count
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,             # <<<<<<<<<<<<<<
 *                                                     int charge, double charge_carrier=PROTON, int max_missed_peaks=1,
 *                                                     double threshold_scale=0.3):
 */

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11_fit_single_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_features, PyObject *__pyx_v_base_tid, CYTHON_UNUSED double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set *__pyx_optional_args) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_charge_carrier = __pyx_k__4;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_v_max_missed_peaks = ((int)1);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_threshold_scale = ((double)0.3);
  double __pyx_v_score
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_final_score
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_score_acc
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_neutral_mass
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_conformer = 0
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_score_vec
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_time_vec
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyArrayObject *__pyx_v_scores_array = 0
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyArrayObject *__pyx_v_times_array = 0
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_feat_iter = 0
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_feat_i
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_feat_n
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_n_missing
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_missing_features
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_counter
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_eid = NULL
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_cleaned_eid = NULL
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_tid = NULL
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_f = NULL
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_v_fit = NULL
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_r = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_10;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_12;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_13;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_14;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_15;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_16;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_17;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_t_18;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass __pyx_t_19;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_fit_single_feature_set", 0);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_optional_args) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_v_max_missed_peaks = __pyx_optional_args->max_missed_peaks;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__pyx_optional_args->__pyx_n > 2) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_v_threshold_scale = __pyx_optional_args->threshold_scale;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if called by wrapper */

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if overridden in Python */

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_single_feature_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11_fit_single_feature_set)) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_5);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_max_missed_peaks); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_threshold_scale); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_8 = __pyx_t_1; 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_t_9 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_10 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (likely(__pyx_t_9)) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(__pyx_t_9);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(function);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_DECREF_SET(__pyx_t_8, function);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __pyx_t_10 = 1;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCALL

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (PyFunction_Check(__pyx_t_8)) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[8] = {__pyx_t_9, __pyx_v_features, __pyx_v_base_tid, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 7+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCCALL

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[8] = {__pyx_t_9, __pyx_v_features, __pyx_v_base_tid, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 7+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_11 = PyTuple_New(7+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_11);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (__pyx_t_9) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_INCREF(__pyx_v_features);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_v_features);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_v_features);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_INCREF(__pyx_v_base_tid);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_v_base_tid);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_v_base_tid);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_t_3);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_10, __pyx_t_4);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_5);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 4+__pyx_t_10, __pyx_t_5);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_6);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 5+__pyx_t_10, __pyx_t_6);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_7);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_11, 6+__pyx_t_10, __pyx_t_7);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_3 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_4 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_5 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_6 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_7 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit))))) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *)__pyx_t_2);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_2 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        goto __pyx_L0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #endif

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":384
 *             size_t feat_i, feat_n, n_missing, missing_features, counter
 * 
 *         conformer = envelope_conformer._create()             # <<<<<<<<<<<<<<
 *         feat_iter = FeatureSetIterator._create(features)
 *         score_acc = 0
 */

#line 384 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)

#line 384 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 384 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_conformer = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_t_1);

#line 384 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":385
 * 
 *         conformer = envelope_conformer._create()
 *         feat_iter = FeatureSetIterator._create(features)             # <<<<<<<<<<<<<<
 *         score_acc = 0
 *         counter = 0
 */

#line 385 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator->_create(__pyx_v_features)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)

#line 385 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 385 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_feat_iter = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_t_1);

#line 385 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":386
 *         conformer = envelope_conformer._create()
 *         feat_iter = FeatureSetIterator._create(features)
 *         score_acc = 0             # <<<<<<<<<<<<<<
 *         counter = 0
 *         score_vec = make_double_vector()
 */

#line 386 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_score_acc = 0.0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":387
 *         feat_iter = FeatureSetIterator._create(features)
 *         score_acc = 0
 *         counter = 0             # <<<<<<<<<<<<<<
 *         score_vec = make_double_vector()
 *         time_vec = make_double_vector()
 */

#line 387 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_counter = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":388
 *         score_acc = 0
 *         counter = 0
 *         score_vec = make_double_vector()             # <<<<<<<<<<<<<<
 *         time_vec = make_double_vector()
 *         while feat_iter.has_more():
 */

#line 388 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_score_vec = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector();

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":389
 *         counter = 0
 *         score_vec = make_double_vector()
 *         time_vec = make_double_vector()             # <<<<<<<<<<<<<<
 *         while feat_iter.has_more():
 *             eid = feat_iter.get_next_value()
 */

#line 389 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_time_vec = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector();

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":390
 *         score_vec = make_double_vector()
 *         time_vec = make_double_vector()
 *         while feat_iter.has_more():             # <<<<<<<<<<<<<<
 *             eid = feat_iter.get_next_value()
 *             if eid is None:
 */

#line 390 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  while (1) {

#line 390 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_12 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->has_more(__pyx_v_feat_iter, 0) != 0);

#line 390 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (!__pyx_t_12) break;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":391
 *         time_vec = make_double_vector()
 *         while feat_iter.has_more():
 *             eid = feat_iter.get_next_value()             # <<<<<<<<<<<<<<
 *             if eid is None:
 *                 continue
 */

#line 391 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->get_next_value(__pyx_v_feat_iter, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)

#line 391 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 391 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_eid, ((PyObject*)__pyx_t_1));

#line 391 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":392
 *         while feat_iter.has_more():
 *             eid = feat_iter.get_next_value()
 *             if eid is None:             # <<<<<<<<<<<<<<
 *                 continue
 *             conformer.acquire(eid, base_tid)
 */

#line 392 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_12 = (__pyx_v_eid == ((PyObject*)Py_None));

#line 392 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_13 = (__pyx_t_12 != 0);

#line 392 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_13) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":393
 *             eid = feat_iter.get_next_value()
 *             if eid is None:
 *                 continue             # <<<<<<<<<<<<<<
 *             conformer.acquire(eid, base_tid)
 *             conformer.conform()
 */

#line 393 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":392
 *         while feat_iter.has_more():
 *             eid = feat_iter.get_next_value()
 *             if eid is None:             # <<<<<<<<<<<<<<
 *                 continue
 *             conformer.acquire(eid, base_tid)
 */

#line 392 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":394
 *             if eid is None:
 *                 continue
 *             conformer.acquire(eid, base_tid)             # <<<<<<<<<<<<<<
 *             conformer.conform()
 *             cleaned_eid = conformer.experimental
 */

#line 394 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_acquire(__pyx_v_conformer, __pyx_v_eid, __pyx_v_base_tid);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":395
 *                 continue
 *             conformer.acquire(eid, base_tid)
 *             conformer.conform()             # <<<<<<<<<<<<<<
 *             cleaned_eid = conformer.experimental
 *             tid = conformer.theoretical
 */

#line 395 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform(__pyx_v_conformer, NULL);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":396
 *             conformer.acquire(eid, base_tid)
 *             conformer.conform()
 *             cleaned_eid = conformer.experimental             # <<<<<<<<<<<<<<
 *             tid = conformer.theoretical
 *             n_missing = conformer.n_missing
 */

#line 396 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = __pyx_v_conformer->experimental;

#line 396 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 396 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_cleaned_eid, ((PyObject*)__pyx_t_1));

#line 396 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":397
 *             conformer.conform()
 *             cleaned_eid = conformer.experimental
 *             tid = conformer.theoretical             # <<<<<<<<<<<<<<
 *             n_missing = conformer.n_missing
 *             if n_missing > max_missed_peaks:
 */

#line 397 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = __pyx_v_conformer->theoretical;

#line 397 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 397 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_tid, ((PyObject*)__pyx_t_1));

#line 397 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":398
 *             cleaned_eid = conformer.experimental
 *             tid = conformer.theoretical
 *             n_missing = conformer.n_missing             # <<<<<<<<<<<<<<
 *             if n_missing > max_missed_peaks:
 *                 continue
 */

#line 398 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_14 = __pyx_v_conformer->n_missing;

#line 398 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_n_missing = __pyx_t_14;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":399
 *             tid = conformer.theoretical
 *             n_missing = conformer.n_missing
 *             if n_missing > max_missed_peaks:             # <<<<<<<<<<<<<<
 *                 continue
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 */

#line 399 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_13 = ((__pyx_v_n_missing > __pyx_v_max_missed_peaks) != 0);

#line 399 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_13) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":400
 *             n_missing = conformer.n_missing
 *             if n_missing > max_missed_peaks:
 *                 continue             # <<<<<<<<<<<<<<
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 *             if npy_isnan(score):
 */

#line 400 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":399
 *             tid = conformer.theoretical
 *             n_missing = conformer.n_missing
 *             if n_missing > max_missed_peaks:             # <<<<<<<<<<<<<<
 *                 continue
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 */

#line 399 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":401
 *             if n_missing > max_missed_peaks:
 *                 continue
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)             # <<<<<<<<<<<<<<
 *             if npy_isnan(score):
 *                 continue
 */

#line 401 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_score = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->scorer->__pyx_vtab)->_evaluate(__pyx_v_self->scorer, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)Py_None), __pyx_v_cleaned_eid, __pyx_v_tid, 0);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":402
 *                 continue
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 *             if npy_isnan(score):             # <<<<<<<<<<<<<<
 *                 continue
 *             score_acc += score
 */

#line 402 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_13 = (npy_isnan(__pyx_v_score) != 0);

#line 402 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_13) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":403
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 *             if npy_isnan(score):
 *                 continue             # <<<<<<<<<<<<<<
 *             score_acc += score
 *             counter += 1
 */

#line 403 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":402
 *                 continue
 *             score = self.scorer._evaluate(None, cleaned_eid, tid)
 *             if npy_isnan(score):             # <<<<<<<<<<<<<<
 *                 continue
 *             score_acc += score
 */

#line 402 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":404
 *             if npy_isnan(score):
 *                 continue
 *             score_acc += score             # <<<<<<<<<<<<<<
 *             counter += 1
 *             double_vector_append(score_vec, score)
 */

#line 404 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_score_acc = (__pyx_v_score_acc + __pyx_v_score);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":405
 *                 continue
 *             score_acc += score
 *             counter += 1             # <<<<<<<<<<<<<<
 *             double_vector_append(score_vec, score)
 *             double_vector_append(time_vec, feat_iter.get_current_time())
 */

#line 405 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_counter = (__pyx_v_counter + 1);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":406
 *             score_acc += score
 *             counter += 1
 *             double_vector_append(score_vec, score)             # <<<<<<<<<<<<<<
 *             double_vector_append(time_vec, feat_iter.get_current_time())
 *         # Compute feature score from score_vec
 */

#line 406 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_score_vec, __pyx_v_score));

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":407
 *             counter += 1
 *             double_vector_append(score_vec, score)
 *             double_vector_append(time_vec, feat_iter.get_current_time())             # <<<<<<<<<<<<<<
 *         # Compute feature score from score_vec
 *         final_score = self._find_thresholded_score(score_vec, threshold_scale)
 */

#line 407 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_time_vec, ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->get_current_time(__pyx_v_feat_iter, 0)));

#line 407 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_L3_continue:;

#line 407 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":409
 *             double_vector_append(time_vec, feat_iter.get_current_time())
 *         # Compute feature score from score_vec
 *         final_score = self._find_thresholded_score(score_vec, threshold_scale)             # <<<<<<<<<<<<<<
 *         missing_features = 0
 *         feat_n = PyList_GET_SIZE(features)
 */

#line 409 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_final_score = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->_find_thresholded_score(__pyx_v_self, __pyx_v_score_vec, __pyx_v_threshold_scale);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":410
 *         # Compute feature score from score_vec
 *         final_score = self._find_thresholded_score(score_vec, threshold_scale)
 *         missing_features = 0             # <<<<<<<<<<<<<<
 *         feat_n = PyList_GET_SIZE(features)
 *         for feat_i in range(feat_n):
 */

#line 410 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_missing_features = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":411
 *         final_score = self._find_thresholded_score(score_vec, threshold_scale)
 *         missing_features = 0
 *         feat_n = PyList_GET_SIZE(features)             # <<<<<<<<<<<<<<
 *         for feat_i in range(feat_n):
 *             f = <LCMSFeature?>PyList_GET_ITEM(features, feat_i)
 */

#line 411 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_feat_n = PyList_GET_SIZE(__pyx_v_features);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":412
 *         missing_features = 0
 *         feat_n = PyList_GET_SIZE(features)
 *         for feat_i in range(feat_n):             # <<<<<<<<<<<<<<
 *             f = <LCMSFeature?>PyList_GET_ITEM(features, feat_i)
 *             if f is None:
 */

#line 412 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_14 = __pyx_v_feat_n;

#line 412 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_15 = __pyx_t_14;

#line 412 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {

#line 412 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_feat_i = __pyx_t_16;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":413
 *         feat_n = PyList_GET_SIZE(features)
 *         for feat_i in range(feat_n):
 *             f = <LCMSFeature?>PyList_GET_ITEM(features, feat_i)             # <<<<<<<<<<<<<<
 *             if f is None:
 *                 missing_features += 1
 */

#line 413 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_17 = PyList_GET_ITEM(__pyx_v_features, __pyx_v_feat_i);

#line 413 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_17);

#line 413 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 413 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1));

#line 413 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":414
 *         for feat_i in range(feat_n):
 *             f = <LCMSFeature?>PyList_GET_ITEM(features, feat_i)
 *             if f is None:             # <<<<<<<<<<<<<<
 *                 missing_features += 1
 *         f = <LCMSFeature?>PyList_GET_ITEM(features, 0)
 */

#line 414 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_13 = (((PyObject *)__pyx_v_f) == Py_None);

#line 414 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_12 = (__pyx_t_13 != 0);

#line 414 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_12) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":415
 *             f = <LCMSFeature?>PyList_GET_ITEM(features, feat_i)
 *             if f is None:
 *                 missing_features += 1             # <<<<<<<<<<<<<<
 *         f = <LCMSFeature?>PyList_GET_ITEM(features, 0)
 *         neutral_mass = calc_neutral_mass(
 */

#line 415 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_missing_features = (__pyx_v_missing_features + 1);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":414
 *         for feat_i in range(feat_n):
 *             f = <LCMSFeature?>PyList_GET_ITEM(features, feat_i)
 *             if f is None:             # <<<<<<<<<<<<<<
 *                 missing_features += 1
 *         f = <LCMSFeature?>PyList_GET_ITEM(features, 0)
 */

#line 414 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 414 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":416
 *             if f is None:
 *                 missing_features += 1
 *         f = <LCMSFeature?>PyList_GET_ITEM(features, 0)             # <<<<<<<<<<<<<<
 *         neutral_mass = calc_neutral_mass(
 *                 f.get_mz(), charge, charge_carrier)
 */

#line 416 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_17 = PyList_GET_ITEM(__pyx_v_features, 0);

#line 416 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_t_17);

#line 416 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 416 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1));

#line 416 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":417
 *                 missing_features += 1
 *         f = <LCMSFeature?>PyList_GET_ITEM(features, 0)
 *         neutral_mass = calc_neutral_mass(             # <<<<<<<<<<<<<<
 *                 f.get_mz(), charge, charge_carrier)
 *         scores_array = double_vector_to_ndarray(score_vec)
 */

#line 417 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_19.__pyx_n = 1;

#line 417 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_19.charge_carrier = __pyx_v_charge_carrier;

#line 417 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_18 = __pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass(((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_f->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_f)), __pyx_v_charge, &__pyx_t_19); 

#line 417 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_neutral_mass = __pyx_t_18;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":419
 *         neutral_mass = calc_neutral_mass(
 *                 f.get_mz(), charge, charge_carrier)
 *         scores_array = double_vector_to_ndarray(score_vec)             # <<<<<<<<<<<<<<
 *         times_array = double_vector_to_ndarray(time_vec)
 *         fit = LCMSFeatureSetFit._create(
 */

#line 419 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(__pyx_v_score_vec); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)

#line 419 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 419 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 419, __pyx_L1_error)

#line 419 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_scores_array = ((PyArrayObject *)__pyx_t_1);

#line 419 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":420
 *                 f.get_mz(), charge, charge_carrier)
 *         scores_array = double_vector_to_ndarray(score_vec)
 *         times_array = double_vector_to_ndarray(time_vec)             # <<<<<<<<<<<<<<
 *         fit = LCMSFeatureSetFit._create(
 *             features, base_tid, final_score, charge, missing_features,
 */

#line 420 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(__pyx_v_time_vec); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)

#line 420 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 420 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 420, __pyx_L1_error)

#line 420 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_times_array = ((PyArrayObject *)__pyx_t_1);

#line 420 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":422
 *         times_array = double_vector_to_ndarray(time_vec)
 *         fit = LCMSFeatureSetFit._create(
 *             features, base_tid, final_score, charge, missing_features,             # <<<<<<<<<<<<<<
 *             [], None, neutral_mass, counter, scores_array, times_array)
 *         free_double_vector(score_vec)
 */

#line 422 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (!(likely(((__pyx_v_base_tid) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_base_tid, __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern))))) __PYX_ERR(0, 422, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":423
 *         fit = LCMSFeatureSetFit._create(
 *             features, base_tid, final_score, charge, missing_features,
 *             [], None, neutral_mass, counter, scores_array, times_array)             # <<<<<<<<<<<<<<
 *         free_double_vector(score_vec)
 *         free_double_vector(time_vec)
 */

#line 423 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)

#line 423 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":421
 *         scores_array = double_vector_to_ndarray(score_vec)
 *         times_array = double_vector_to_ndarray(time_vec)
 *         fit = LCMSFeatureSetFit._create(             # <<<<<<<<<<<<<<
 *             features, base_tid, final_score, charge, missing_features,
 *             [], None, neutral_mass, counter, scores_array, times_array)
 */

#line 421 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit->_create(__pyx_v_features, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_base_tid), __pyx_v_final_score, __pyx_v_charge, __pyx_v_missing_features, ((PyObject*)__pyx_t_1), Py_None, __pyx_v_neutral_mass, __pyx_v_counter, ((PyObject *)__pyx_v_scores_array), ((PyObject *)__pyx_v_times_array))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 421, __pyx_L1_error)

#line 421 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 421 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 421 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_fit = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *)__pyx_t_2);

#line 421 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":424
 *             features, base_tid, final_score, charge, missing_features,
 *             [], None, neutral_mass, counter, scores_array, times_array)
 *         free_double_vector(score_vec)             # <<<<<<<<<<<<<<
 *         free_double_vector(time_vec)
 *         return fit
 */

#line 424 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_score_vec);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":425
 *             [], None, neutral_mass, counter, scores_array, times_array)
 *         free_double_vector(score_vec)
 *         free_double_vector(time_vec)             # <<<<<<<<<<<<<<
 *         return fit
 * 
 */

#line 425 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_time_vec);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":426
 *         free_double_vector(score_vec)
 *         free_double_vector(time_vec)
 *         return fit             # <<<<<<<<<<<<<<
 * 
 *     cdef list _fit_theoretical_distribution_on_features(self, double mz, double error_tolerance, int charge,
 */

#line 426 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 426 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_fit));

#line 426 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_fit;

#line 426 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":372
 *         return acc / count
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,             # <<<<<<<<<<<<<<
 *                                                     int charge, double charge_carrier=PROTON, int max_missed_peaks=1,
 *                                                     double threshold_scale=0.3):
 */

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_9);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_11);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_single_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_conformer);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_scores_array);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_times_array);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_feat_iter);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_eid);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_cleaned_eid);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_tid);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_f);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_fit);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11_fit_single_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_10_fit_single_feature_set[] = "LCMSFeatureProcessorBase._fit_single_feature_set(self, list features, list base_tid, double error_tolerance, int charge, double charge_carrier=PROTON, int max_missed_peaks=1, double threshold_scale=0.3) -> LCMSFeatureSetFit";

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11_fit_single_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_features = 0
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_base_tid = 0
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_error_tolerance
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  int __pyx_v_charge
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_charge_carrier
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  int __pyx_v_max_missed_peaks
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_threshold_scale
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_fit_single_feature_set (wrapper)", 0);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_features,&__pyx_n_s_base_tid,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge,&__pyx_n_s_charge_carrier,&__pyx_n_s_max_missed_peaks,&__pyx_n_s_threshold_scale,0};

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyObject* values[7] = {0,0,0,0,0,0,0};

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_kwds)) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      Py_ssize_t kw_args;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {
        case  7: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[6] = PyTuple_GET_ITEM(__pyx_args, 6);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  6: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0: break;
        default: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0:

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  1:

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_base_tid)) != 0)) kw_args--;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_single_feature_set", 0, 4, 7, 1); 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 372, __pyx_L3_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  2:

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_single_feature_set", 0, 4, 7, 2); 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 372, __pyx_L3_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  3:

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_single_feature_set", 0, 4, 7, 3); 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 372, __pyx_L3_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  4:

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  5:

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_missed_peaks);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[5] = value; kw_args--; }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  6:

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_threshold_scale);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[6] = value; kw_args--; }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(kw_args > 0)) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_fit_single_feature_set") < 0)) __PYX_ERR(0, 372, __pyx_L3_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[6] = PyTuple_GET_ITEM(__pyx_args, 6);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  6: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        break;
        default: 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_features = ((PyObject*)values[0]);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_base_tid = ((PyObject*)values[1]);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 372, __pyx_L3_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_charge = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 373, __pyx_L3_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[4]) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 373, __pyx_L3_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_k__4;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[5]) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_max_missed_peaks = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_max_missed_peaks == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 373, __pyx_L3_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_max_missed_peaks = ((int)1);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[6]) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_threshold_scale = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_threshold_scale == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 374, __pyx_L3_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_threshold_scale = ((double)0.3);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_fit_single_feature_set", 0, 4, 7, PyTuple_GET_SIZE(__pyx_args)); 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 372, __pyx_L3_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L3_error:;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_single_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_features), (&PyList_Type), 1, "features", 1))) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_base_tid), (&PyList_Type), 1, "base_tid", 1))) __PYX_ERR(0, 372, __pyx_L1_error)
  __pyx_r = 
#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_10_fit_single_feature_set(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_features, __pyx_v_base_tid, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_charge_carrier, __pyx_v_max_missed_peaks, __pyx_v_threshold_scale);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_10_fit_single_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_features, PyObject *__pyx_v_base_tid, double __pyx_v_error_tolerance, int __pyx_v_charge, double __pyx_v_charge_carrier, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale) {

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set __pyx_t_2;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_fit_single_feature_set", 0);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.__pyx_n = 3;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.max_missed_peaks = __pyx_v_max_missed_peaks;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.threshold_scale = __pyx_v_threshold_scale;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->_fit_single_feature_set(__pyx_v_self, __pyx_v_features, __pyx_v_base_tid, __pyx_v_error_tolerance, __pyx_v_charge, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_single_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":428
 *         return fit
 * 
 *     cdef list _fit_theoretical_distribution_on_features(self, double mz, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                         TheoreticalIsotopicPattern base_tid,
 *                                                         double charge_carrier=PROTON,
 */

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_base_tid, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features *__pyx_optional_args) {

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_charge_carrier = __pyx_k__5;

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_v_max_missed_peaks = ((int)1);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_threshold_scale = ((double)0.3);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":432
 *                                                         double charge_carrier=PROTON,
 *                                                         double truncate_after=0.8, int max_missed_peaks=1,
 *                                                         double threshold_scale=0.3, LCMSFeature feature=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double score, final_score, score_acc, neutral_mass
 */

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
  double __pyx_v_score
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_final_score
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_score_acc
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_neutral_mass
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_feature_groups = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_feature_fits = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_features = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_eid = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_cleaned_eid = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_tid = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_scores_array = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_times_array = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_combn_i
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_n_missing
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_missing_features
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_counter
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_snapped_tid = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_feat_i
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_feat_n
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_feat_iter = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_f = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *__pyx_v_combn_iter = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v_conformer = 0
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_score_vec
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *__pyx_v_time_vec
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *__pyx_v_fit = NULL
#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution __pyx_t_3;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_4;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_5;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_t_11;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_12;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_13;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_14;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_15;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass __pyx_t_16;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_17;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_fit_theoretical_distribution_on_features", 0);

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_optional_args) {

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_optional_args->__pyx_n > 2) {

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_v_max_missed_peaks = __pyx_optional_args->max_missed_peaks;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__pyx_optional_args->__pyx_n > 3) {

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_v_threshold_scale = __pyx_optional_args->threshold_scale;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (__pyx_optional_args->__pyx_n > 4) {

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __pyx_v_feature = __pyx_optional_args->feature;

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 432 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":450
 * 
 *         feature_groups = self.match_theoretical_isotopic_distribution(
 *             base_tid.peaklist, error_tolerance, interval=feature)             # <<<<<<<<<<<<<<
 *         feature_fits = []
 * 
 */

#line 450 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_base_tid->peaklist;

#line 450 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":449
 *             dvec* time_vec
 * 
 *         feature_groups = self.match_theoretical_isotopic_distribution(             # <<<<<<<<<<<<<<
 *             base_tid.peaklist, error_tolerance, interval=feature)
 *         feature_fits = []
 */

#line 449 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3.__pyx_n = 2;

#line 449 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3.error_tolerance = __pyx_v_error_tolerance;

#line 449 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3.interval = __pyx_v_feature;

#line 449 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->match_theoretical_isotopic_distribution(__pyx_v_self, ((PyObject*)__pyx_t_1), 0, &__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 449, __pyx_L1_error)

#line 449 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 449 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 449 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_feature_groups = ((PyObject*)__pyx_t_2);

#line 449 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":451
 *         feature_groups = self.match_theoretical_isotopic_distribution(
 *             base_tid.peaklist, error_tolerance, interval=feature)
 *         feature_fits = []             # <<<<<<<<<<<<<<
 * 
 *         conformer = envelope_conformer._create()
 */

#line 451 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)

#line 451 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 451 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_feature_fits = ((PyObject*)__pyx_t_2);

#line 451 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":453
 *         feature_fits = []
 * 
 *         conformer = envelope_conformer._create()             # <<<<<<<<<<<<<<
 * 
 *         combn_iter = CartesianProductIterator._create(feature_groups)
 */

#line 453 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 453, __pyx_L1_error)

#line 453 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 453 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_conformer = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_t_2);

#line 453 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":455
 *         conformer = envelope_conformer._create()
 * 
 *         combn_iter = CartesianProductIterator._create(feature_groups)             # <<<<<<<<<<<<<<
 *         combn_i = 0
 *         while combn_iter.has_more():
 */

#line 455 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator__create(__pyx_v_feature_groups)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)

#line 455 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 455 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_combn_iter = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)__pyx_t_2);

#line 455 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":456
 * 
 *         combn_iter = CartesianProductIterator._create(feature_groups)
 *         combn_i = 0             # <<<<<<<<<<<<<<
 *         while combn_iter.has_more():
 *             features = combn_iter.get_next_value()
 */

#line 456 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_combn_i = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":457
 *         combn_iter = CartesianProductIterator._create(feature_groups)
 *         combn_i = 0
 *         while combn_iter.has_more():             # <<<<<<<<<<<<<<
 *             features = combn_iter.get_next_value()
 *             if features is None:
 */

#line 457 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  while (1) {

#line 457 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_4 = (__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_has_more(__pyx_v_combn_iter) != 0);

#line 457 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (!__pyx_t_4) break;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":458
 *         combn_i = 0
 *         while combn_iter.has_more():
 *             features = combn_iter.get_next_value()             # <<<<<<<<<<<<<<
 *             if features is None:
 *                 break
 */

#line 458 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_get_next_value(__pyx_v_combn_iter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 458, __pyx_L1_error)

#line 458 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 458 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_features, ((PyObject*)__pyx_t_2));

#line 458 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":459
 *         while combn_iter.has_more():
 *             features = combn_iter.get_next_value()
 *             if features is None:             # <<<<<<<<<<<<<<
 *                 break
 *             combn_i += 1
 */

#line 459 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_4 = (__pyx_v_features == ((PyObject*)Py_None));

#line 459 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = (__pyx_t_4 != 0);

#line 459 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":460
 *             features = combn_iter.get_next_value()
 *             if features is None:
 *                 break             # <<<<<<<<<<<<<<
 *             combn_i += 1
 *             if combn_i > MAX_COMBINATIONS:
 */

#line 460 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L4_break;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":459
 *         while combn_iter.has_more():
 *             features = combn_iter.get_next_value()
 *             if features is None:             # <<<<<<<<<<<<<<
 *                 break
 *             combn_i += 1
 */

#line 459 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":461
 *             if features is None:
 *                 break
 *             combn_i += 1             # <<<<<<<<<<<<<<
 *             if combn_i > MAX_COMBINATIONS:
 *                 break
 */

#line 461 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_combn_i = (__pyx_v_combn_i + 1);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":462
 *                 break
 *             combn_i += 1
 *             if combn_i > MAX_COMBINATIONS:             # <<<<<<<<<<<<<<
 *                 break
 *             elif combn_i % (100000) == 0:
 */

#line 462 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = ((__pyx_v_combn_i > __pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_MAX_COMBINATIONS) != 0);

#line 462 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":463
 *             combn_i += 1
 *             if combn_i > MAX_COMBINATIONS:
 *                 break             # <<<<<<<<<<<<<<
 *             elif combn_i % (100000) == 0:
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 */

#line 463 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L4_break;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":462
 *                 break
 *             combn_i += 1
 *             if combn_i > MAX_COMBINATIONS:             # <<<<<<<<<<<<<<
 *                 break
 *             elif combn_i % (100000) == 0:
 */

#line 462 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":464
 *             if combn_i > MAX_COMBINATIONS:
 *                 break
 *             elif combn_i % (100000) == 0:             # <<<<<<<<<<<<<<
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 *                     combn_i, mz, charge, "-" if feature is None else (
 */

#line 464 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = (((__pyx_v_combn_i % 0x186A0) == 0) != 0);

#line 464 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":466
 *             elif combn_i % (100000) == 0:
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 *                     combn_i, mz, charge, "-" if feature is None else (             # <<<<<<<<<<<<<<
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),
 *                     100. * combn_i / combn_iter.total_combinations))
 */

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_combn_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 466, __pyx_L1_error)

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 466, __pyx_L1_error)

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_5 = (((PyObject *)__pyx_v_feature) == Py_None);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if ((__pyx_t_5 != 0)) {

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_INCREF(__pyx_kp_s__6);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_7 = __pyx_kp_s__6;

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      } else {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":467
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 *                     combn_i, mz, charge, "-" if feature is None else (
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),             # <<<<<<<<<<<<<<
 *                     100. * combn_i / combn_iter.total_combinations))
 * 
 */

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_feature), __pyx_n_s_start_time); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 467, __pyx_L1_error)

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_8);

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_feature), __pyx_n_s_end_time); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 467, __pyx_L1_error)

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_9);

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 467, __pyx_L1_error)

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_10);

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GIVEREF(__pyx_t_8);

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8);

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GIVEREF(__pyx_t_9);

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_9);

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_8 = 0;

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_9 = 0;

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_9 = __Pyx_PyString_Format(__pyx_kp_s_0_2f_0_2f, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 467, __pyx_L1_error)

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_9);

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_7 = __pyx_t_9;

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_9 = 0;

#line 467 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":468
 *                     combn_i, mz, charge, "-" if feature is None else (
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),
 *                     100. * combn_i / combn_iter.total_combinations))             # <<<<<<<<<<<<<<
 * 
 *             if has_no_valid_features(features):
 */

#line 468 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_11 = (100. * __pyx_v_combn_i);

#line 468 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(__pyx_v_combn_iter->total_combinations == 0)) {

#line 468 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 468 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __PYX_ERR(0, 468, __pyx_L1_error)

#line 468 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 468 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = PyFloat_FromDouble((__pyx_t_11 / __pyx_v_combn_iter->total_combinations)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 468, __pyx_L1_error)

#line 468 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_9);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":466
 *             elif combn_i % (100000) == 0:
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 *                     combn_i, mz, charge, "-" if feature is None else (             # <<<<<<<<<<<<<<
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),
 *                     100. * combn_i / combn_iter.total_combinations))
 */

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_10 = PyTuple_New(5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 466, __pyx_L1_error)

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_10);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GIVEREF(__pyx_t_2);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GIVEREF(__pyx_t_1);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GIVEREF(__pyx_t_6);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_6);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GIVEREF(__pyx_t_7);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_7);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GIVEREF(__pyx_t_9);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_9);

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_2 = 0;

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = 0;

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_6 = 0;

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_7 = 0;

#line 466 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":465
 *                 break
 *             elif combn_i % (100000) == 0:
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (             # <<<<<<<<<<<<<<
 *                     combn_i, mz, charge, "-" if feature is None else (
 *                         "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),
 */

#line 465 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = __Pyx_PyString_Format(__pyx_kp_s_d_combinations_have_been_consid, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 465, __pyx_L1_error)

#line 465 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_9);

#line 465 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 465 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__Pyx_PrintOne(0, __pyx_t_9) < 0) __PYX_ERR(0, 465, __pyx_L1_error)

#line 465 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":464
 *             if combn_i > MAX_COMBINATIONS:
 *                 break
 *             elif combn_i % (100000) == 0:             # <<<<<<<<<<<<<<
 *                 print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
 *                     combn_i, mz, charge, "-" if feature is None else (
 */

#line 464 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":470
 *                     100. * combn_i / combn_iter.total_combinations))
 * 
 *             if has_no_valid_features(features):             # <<<<<<<<<<<<<<
 *                 continue
 *             # If the monoisotopic feature wasn't actually observed, create a dummy feature
 */

#line 470 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = (__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_has_no_valid_features(__pyx_v_features) != 0);

#line 470 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":471
 * 
 *             if has_no_valid_features(features):
 *                 continue             # <<<<<<<<<<<<<<
 *             # If the monoisotopic feature wasn't actually observed, create a dummy feature
 *             # since the monoisotopic feature cannot be None
 */

#line 471 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":470
 *                     100. * combn_i / combn_iter.total_combinations))
 * 
 *             if has_no_valid_features(features):             # <<<<<<<<<<<<<<
 *                 continue
 *             # If the monoisotopic feature wasn't actually observed, create a dummy feature
 */

#line 470 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":475
 *             # since the monoisotopic feature cannot be None
 *             # print("Checking monoisotopic feature")
 *             if features[0] is None:             # <<<<<<<<<<<<<<
 *                 # print("Creating monoisotopic feature stub")
 *                 features = list(features)
 */

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_v_features == Py_None)) {

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __PYX_ERR(0, 475, __pyx_L1_error)

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_9 = __Pyx_GetItemInt_List(__pyx_v_features, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 475, __pyx_L1_error)

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_9);

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = (__pyx_t_9 == Py_None);

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_4 = (__pyx_t_5 != 0);

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_4) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":477
 *             if features[0] is None:
 *                 # print("Creating monoisotopic feature stub")
 *                 features = list(features)             # <<<<<<<<<<<<<<
 *                 features[0] = EmptyFeature._create(mz)
 *                 snapped_tid = base_tid.clone().shift(mz)
 */

#line 477 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = PySequence_List(__pyx_v_features); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 477, __pyx_L1_error)

#line 477 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_9);

#line 477 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF_SET(__pyx_v_features, ((PyObject*)__pyx_t_9));

#line 477 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":478
 *                 # print("Creating monoisotopic feature stub")
 *                 features = list(features)
 *                 features[0] = EmptyFeature._create(mz)             # <<<<<<<<<<<<<<
 *                 snapped_tid = base_tid.clone().shift(mz)
 *             else:
 */

#line 478 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature->_create(__pyx_v_mz)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_9);

#line 478 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_features, 0, __pyx_t_9, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":479
 *                 features = list(features)
 *                 features[0] = EmptyFeature._create(mz)
 *                 snapped_tid = base_tid.clone().shift(mz)             # <<<<<<<<<<<<<<
 *             else:
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 */

#line 479 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_base_tid->__pyx_vtab)->clone(__pyx_v_base_tid, 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 479, __pyx_L1_error)

#line 479 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_9);

#line 479 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_9)->__pyx_vtab)->shift(((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_9), __pyx_v_mz, 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 479, __pyx_L1_error)

#line 479 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_10);

#line 479 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 479 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_XDECREF_SET(__pyx_v_snapped_tid, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_10));

#line 479 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_10 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":475
 *             # since the monoisotopic feature cannot be None
 *             # print("Checking monoisotopic feature")
 *             if features[0] is None:             # <<<<<<<<<<<<<<
 *                 # print("Creating monoisotopic feature stub")
 *                 features = list(features)
 */

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L8;

#line 475 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":481
 *                 snapped_tid = base_tid.clone().shift(mz)
 *             else:
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, 0)             # <<<<<<<<<<<<<<
 *                 snapped_tid = base_tid.clone().shift(f.get_mz())
 *             # feat_iter = FeatureSetIterator._create_with_threshold(features, base_tid, 0.1)
 */

#line 481 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    /*else*/ {

#line 481 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_12 = PyList_GET_ITEM(__pyx_v_features, 0);

#line 481 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_10 = ((PyObject *)__pyx_t_12);

#line 481 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(__pyx_t_10);

#line 481 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_10));

#line 481 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_10 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":482
 *             else:
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 *                 snapped_tid = base_tid.clone().shift(f.get_mz())             # <<<<<<<<<<<<<<
 *             # feat_iter = FeatureSetIterator._create_with_threshold(features, base_tid, 0.1)
 *             feat_iter = FeatureSetIterator._create(features)
 */

#line 482 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_base_tid->__pyx_vtab)->clone(__pyx_v_base_tid, 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 482, __pyx_L1_error)

#line 482 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_10);

#line 482 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_10)->__pyx_vtab)->shift(((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_10), ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_f->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_f)), 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 482, __pyx_L1_error)

#line 482 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_9);

#line 482 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 482 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_XDECREF_SET(__pyx_v_snapped_tid, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_9));

#line 482 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = 0;

#line 482 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 482 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_L8:;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":484
 *                 snapped_tid = base_tid.clone().shift(f.get_mz())
 *             # feat_iter = FeatureSetIterator._create_with_threshold(features, base_tid, 0.1)
 *             feat_iter = FeatureSetIterator._create(features)             # <<<<<<<<<<<<<<
 *             counter = 0
 *             score_acc = 0
 */

#line 484 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_9 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator->_create(__pyx_v_features)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_9);

#line 484 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_feat_iter, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_t_9));

#line 484 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_9 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":485
 *             # feat_iter = FeatureSetIterator._create_with_threshold(features, base_tid, 0.1)
 *             feat_iter = FeatureSetIterator._create(features)
 *             counter = 0             # <<<<<<<<<<<<<<
 *             score_acc = 0
 * 
 */

#line 485 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_counter = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":486
 *             feat_iter = FeatureSetIterator._create(features)
 *             counter = 0
 *             score_acc = 0             # <<<<<<<<<<<<<<
 * 
 *             score_vec = make_double_vector()
 */

#line 486 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_score_acc = 0.0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":488
 *             score_acc = 0
 * 
 *             score_vec = make_double_vector()             # <<<<<<<<<<<<<<
 *             time_vec = make_double_vector()
 *             # print("Scoring features")
 */

#line 488 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_score_vec = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector();

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":489
 * 
 *             score_vec = make_double_vector()
 *             time_vec = make_double_vector()             # <<<<<<<<<<<<<<
 *             # print("Scoring features")
 *             while feat_iter.has_more():
 */

#line 489 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_time_vec = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector();

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":491
 *             time_vec = make_double_vector()
 *             # print("Scoring features")
 *             while feat_iter.has_more():             # <<<<<<<<<<<<<<
 *                 eid = feat_iter.get_next_value()
 *                 if eid is None:
 */

#line 491 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    while (1) {

#line 491 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_4 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->has_more(__pyx_v_feat_iter, 0) != 0);

#line 491 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (!__pyx_t_4) break;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":492
 *             # print("Scoring features")
 *             while feat_iter.has_more():
 *                 eid = feat_iter.get_next_value()             # <<<<<<<<<<<<<<
 *                 if eid is None:
 *                     continue
 */

#line 492 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->get_next_value(__pyx_v_feat_iter, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 492, __pyx_L1_error)

#line 492 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_9);

#line 492 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_XDECREF_SET(__pyx_v_eid, ((PyObject*)__pyx_t_9));

#line 492 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":493
 *             while feat_iter.has_more():
 *                 eid = feat_iter.get_next_value()
 *                 if eid is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 counter += 1
 */

#line 493 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_4 = (__pyx_v_eid == ((PyObject*)Py_None));

#line 493 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_5 = (__pyx_t_4 != 0);

#line 493 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_t_5) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":494
 *                 eid = feat_iter.get_next_value()
 *                 if eid is None:
 *                     continue             # <<<<<<<<<<<<<<
 *                 counter += 1
 *                 conformer.acquire(eid, snapped_tid.peaklist)
 */

#line 494 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        goto __pyx_L9_continue;

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":493
 *             while feat_iter.has_more():
 *                 eid = feat_iter.get_next_value()
 *                 if eid is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 counter += 1
 */

#line 493 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":495
 *                 if eid is None:
 *                     continue
 *                 counter += 1             # <<<<<<<<<<<<<<
 *                 conformer.acquire(eid, snapped_tid.peaklist)
 *                 conformer.conform()
 */

#line 495 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_counter = (__pyx_v_counter + 1);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":496
 *                     continue
 *                 counter += 1
 *                 conformer.acquire(eid, snapped_tid.peaklist)             # <<<<<<<<<<<<<<
 *                 conformer.conform()
 *                 cleaned_eid = conformer.experimental
 */

#line 496 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = __pyx_v_snapped_tid->peaklist;

#line 496 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(__pyx_t_9);

#line 496 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_acquire(__pyx_v_conformer, __pyx_v_eid, ((PyObject*)__pyx_t_9));

#line 496 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":497
 *                 counter += 1
 *                 conformer.acquire(eid, snapped_tid.peaklist)
 *                 conformer.conform()             # <<<<<<<<<<<<<<
 *                 cleaned_eid = conformer.experimental
 *                 tid = conformer.theoretical
 */

#line 497 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform(__pyx_v_conformer, NULL);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":498
 *                 conformer.acquire(eid, snapped_tid.peaklist)
 *                 conformer.conform()
 *                 cleaned_eid = conformer.experimental             # <<<<<<<<<<<<<<
 *                 tid = conformer.theoretical
 *                 n_missing = conformer.n_missing
 */

#line 498 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = __pyx_v_conformer->experimental;

#line 498 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(__pyx_t_9);

#line 498 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_XDECREF_SET(__pyx_v_cleaned_eid, ((PyObject*)__pyx_t_9));

#line 498 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":499
 *                 conformer.conform()
 *                 cleaned_eid = conformer.experimental
 *                 tid = conformer.theoretical             # <<<<<<<<<<<<<<
 *                 n_missing = conformer.n_missing
 *                 if n_missing > max_missed_peaks:
 */

#line 499 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = __pyx_v_conformer->theoretical;

#line 499 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(__pyx_t_9);

#line 499 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_XDECREF_SET(__pyx_v_tid, ((PyObject*)__pyx_t_9));

#line 499 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":500
 *                 cleaned_eid = conformer.experimental
 *                 tid = conformer.theoretical
 *                 n_missing = conformer.n_missing             # <<<<<<<<<<<<<<
 *                 if n_missing > max_missed_peaks:
 *                     continue
 */

#line 500 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_13 = __pyx_v_conformer->n_missing;

#line 500 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_n_missing = __pyx_t_13;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":501
 *                 tid = conformer.theoretical
 *                 n_missing = conformer.n_missing
 *                 if n_missing > max_missed_peaks:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */

#line 501 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_5 = ((__pyx_v_n_missing > __pyx_v_max_missed_peaks) != 0);

#line 501 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_t_5) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":502
 *                 n_missing = conformer.n_missing
 *                 if n_missing > max_missed_peaks:
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 score = self.scorer._evaluate(None, cleaned_eid, tid)
 */

#line 502 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        goto __pyx_L9_continue;

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":501
 *                 tid = conformer.theoretical
 *                 n_missing = conformer.n_missing
 *                 if n_missing > max_missed_peaks:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */

#line 501 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":504
 *                     continue
 * 
 *                 score = self.scorer._evaluate(None, cleaned_eid, tid)             # <<<<<<<<<<<<<<
 *                 if npy_isnan(score):
 *                     continue
 */

#line 504 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_score = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->scorer->__pyx_vtab)->_evaluate(__pyx_v_self->scorer, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)Py_None), __pyx_v_cleaned_eid, __pyx_v_tid, 0);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":505
 * 
 *                 score = self.scorer._evaluate(None, cleaned_eid, tid)
 *                 if npy_isnan(score):             # <<<<<<<<<<<<<<
 *                     continue
 *                 score_acc += score
 */

#line 505 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_5 = (npy_isnan(__pyx_v_score) != 0);

#line 505 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_t_5) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":506
 *                 score = self.scorer._evaluate(None, cleaned_eid, tid)
 *                 if npy_isnan(score):
 *                     continue             # <<<<<<<<<<<<<<
 *                 score_acc += score
 *                 double_vector_append(score_vec, score)
 */

#line 506 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        goto __pyx_L9_continue;

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":505
 * 
 *                 score = self.scorer._evaluate(None, cleaned_eid, tid)
 *                 if npy_isnan(score):             # <<<<<<<<<<<<<<
 *                     continue
 *                 score_acc += score
 */

#line 505 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":507
 *                 if npy_isnan(score):
 *                     continue
 *                 score_acc += score             # <<<<<<<<<<<<<<
 *                 double_vector_append(score_vec, score)
 *                 double_vector_append(time_vec, feat_iter.get_current_time())
 */

#line 507 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_score_acc = (__pyx_v_score_acc + __pyx_v_score);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":508
 *                     continue
 *                 score_acc += score
 *                 double_vector_append(score_vec, score)             # <<<<<<<<<<<<<<
 *                 double_vector_append(time_vec, feat_iter.get_current_time())
 *             if score_vec.used == 0:
 */

#line 508 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_score_vec, __pyx_v_score));

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":509
 *                 score_acc += score
 *                 double_vector_append(score_vec, score)
 *                 double_vector_append(time_vec, feat_iter.get_current_time())             # <<<<<<<<<<<<<<
 *             if score_vec.used == 0:
 *                 free_double_vector(score_vec)
 */

#line 509 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_time_vec, ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_feat_iter->__pyx_vtab)->get_current_time(__pyx_v_feat_iter, 0)));

#line 509 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_L9_continue:;

#line 509 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":510
 *                 double_vector_append(score_vec, score)
 *                 double_vector_append(time_vec, feat_iter.get_current_time())
 *             if score_vec.used == 0:             # <<<<<<<<<<<<<<
 *                 free_double_vector(score_vec)
 *                 free_double_vector(time_vec)
 */

#line 510 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_5 = ((__pyx_v_score_vec->used == 0) != 0);

#line 510 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":511
 *                 double_vector_append(time_vec, feat_iter.get_current_time())
 *             if score_vec.used == 0:
 *                 free_double_vector(score_vec)             # <<<<<<<<<<<<<<
 *                 free_double_vector(time_vec)
 *                 continue
 */

#line 511 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_score_vec);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":512
 *             if score_vec.used == 0:
 *                 free_double_vector(score_vec)
 *                 free_double_vector(time_vec)             # <<<<<<<<<<<<<<
 *                 continue
 *             # print(f"Scored over {counter} points")
 */

#line 512 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_time_vec);

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":513
 *                 free_double_vector(score_vec)
 *                 free_double_vector(time_vec)
 *                 continue             # <<<<<<<<<<<<<<
 *             # print(f"Scored over {counter} points")
 *             # Compute feature score from score_vec
 */

#line 513 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":510
 *                 double_vector_append(score_vec, score)
 *                 double_vector_append(time_vec, feat_iter.get_current_time())
 *             if score_vec.used == 0:             # <<<<<<<<<<<<<<
 *                 free_double_vector(score_vec)
 *                 free_double_vector(time_vec)
 */

#line 510 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":516
 *             # print(f"Scored over {counter} points")
 *             # Compute feature score from score_vec
 *             final_score = self._find_thresholded_score(             # <<<<<<<<<<<<<<
 *                 score_vec, threshold_scale)
 *             # print(f"Final Score {final_score}")
 */

#line 516 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_final_score = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->_find_thresholded_score(__pyx_v_self, __pyx_v_score_vec, __pyx_v_threshold_scale);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":519
 *                 score_vec, threshold_scale)
 *             # print(f"Final Score {final_score}")
 *             missing_features = 0             # <<<<<<<<<<<<<<
 *             feat_n = PyList_GET_SIZE(features)
 *             for feat_i in range(feat_n):
 */

#line 519 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_missing_features = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":520
 *             # print(f"Final Score {final_score}")
 *             missing_features = 0
 *             feat_n = PyList_GET_SIZE(features)             # <<<<<<<<<<<<<<
 *             for feat_i in range(feat_n):
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 */

#line 520 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_feat_n = PyList_GET_SIZE(__pyx_v_features);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":521
 *             missing_features = 0
 *             feat_n = PyList_GET_SIZE(features)
 *             for feat_i in range(feat_n):             # <<<<<<<<<<<<<<
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *                 if f is None:
 */

#line 521 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_13 = __pyx_v_feat_n;

#line 521 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_14 = __pyx_t_13;

#line 521 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 521 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_feat_i = __pyx_t_15;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":522
 *             feat_n = PyList_GET_SIZE(features)
 *             for feat_i in range(feat_n):
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)             # <<<<<<<<<<<<<<
 *                 if f is None:
 *                     missing_features += 1
 */

#line 522 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_12 = PyList_GET_ITEM(__pyx_v_features, __pyx_v_feat_i);

#line 522 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = ((PyObject *)__pyx_t_12);

#line 522 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_INCREF(__pyx_t_9);

#line 522 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_9));

#line 522 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":523
 *             for feat_i in range(feat_n):
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *                 if f is None:             # <<<<<<<<<<<<<<
 *                     missing_features += 1
 *             f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 */

#line 523 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_5 = (((PyObject *)__pyx_v_f) == Py_None);

#line 523 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_4 = (__pyx_t_5 != 0);

#line 523 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_t_4) {

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":524
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *                 if f is None:
 *                     missing_features += 1             # <<<<<<<<<<<<<<
 *             f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 *             neutral_mass = calc_neutral_mass(
 */

#line 524 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_v_missing_features = (__pyx_v_missing_features + 1);

        /* "ms_deisotope/_c/feature_map/feature_processor.pyx":523
 *             for feat_i in range(feat_n):
 *                 f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
 *                 if f is None:             # <<<<<<<<<<<<<<
 *                     missing_features += 1
 *             f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 */

#line 523 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 523 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":525
 *                 if f is None:
 *                     missing_features += 1
 *             f = <LCMSFeature>PyList_GET_ITEM(features, 0)             # <<<<<<<<<<<<<<
 *             neutral_mass = calc_neutral_mass(
 *                     snapped_tid.get_monoisotopic_mz(), charge, charge_carrier)
 */

#line 525 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_12 = PyList_GET_ITEM(__pyx_v_features, 0);

#line 525 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_9 = ((PyObject *)__pyx_t_12);

#line 525 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_9);

#line 525 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_9));

#line 525 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_9 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":526
 *                     missing_features += 1
 *             f = <LCMSFeature>PyList_GET_ITEM(features, 0)
 *             neutral_mass = calc_neutral_mass(             # <<<<<<<<<<<<<<
 *                     snapped_tid.get_monoisotopic_mz(), charge, charge_carrier)
 *             # print("Preparing to hand off arrays")
 */

#line 526 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_16.__pyx_n = 1;

#line 526 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_16.charge_carrier = __pyx_v_charge_carrier;

#line 526 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_11 = __pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass(((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_snapped_tid->__pyx_vtab)->get_monoisotopic_mz(__pyx_v_snapped_tid), __pyx_v_charge, &__pyx_t_16); 

#line 526 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_neutral_mass = __pyx_t_11;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":529
 *                     snapped_tid.get_monoisotopic_mz(), charge, charge_carrier)
 *             # print("Preparing to hand off arrays")
 *             scores_array = double_vector_to_ndarray(score_vec)             # <<<<<<<<<<<<<<
 *             times_array = double_vector_to_ndarray(time_vec)
 *             # print("Creating fit")
 */

#line 529 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_9 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(__pyx_v_score_vec); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 529, __pyx_L1_error)

#line 529 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_9);

#line 529 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_scores_array, __pyx_t_9);

#line 529 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_9 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":530
 *             # print("Preparing to hand off arrays")
 *             scores_array = double_vector_to_ndarray(score_vec)
 *             times_array = double_vector_to_ndarray(time_vec)             # <<<<<<<<<<<<<<
 *             # print("Creating fit")
 * 
 */

#line 530 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_9 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_double_vector_to_ndarray(__pyx_v_time_vec); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 530, __pyx_L1_error)

#line 530 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_9);

#line 530 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_times_array, __pyx_t_9);

#line 530 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_9 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":536
 *                 features, snapped_tid,
 *                 final_score, charge,
 *                 missing_features, [],             # <<<<<<<<<<<<<<
 *                 None, neutral_mass,
 *                 counter, scores_array,
 */

#line 536 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 536, __pyx_L1_error)

#line 536 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_9);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":533
 *             # print("Creating fit")
 * 
 *             fit = LCMSFeatureSetFit._create(             # <<<<<<<<<<<<<<
 *                 features, snapped_tid,
 *                 final_score, charge,
 */

#line 533 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_10 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit->_create(__pyx_v_features, __pyx_v_snapped_tid, __pyx_v_final_score, __pyx_v_charge, __pyx_v_missing_features, ((PyObject*)__pyx_t_9), Py_None, __pyx_v_neutral_mass, __pyx_v_counter, __pyx_v_scores_array, __pyx_v_times_array)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 533, __pyx_L1_error)

#line 533 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_10);

#line 533 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 533 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_fit, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *)__pyx_t_10));

#line 533 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_10 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":542
 * 
 *             # print("Freeing arrays")
 *             free_double_vector(score_vec)             # <<<<<<<<<<<<<<
 *             free_double_vector(time_vec)
 *             if self.scorer.reject_score(fit.score):
 */

#line 542 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_score_vec);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":543
 *             # print("Freeing arrays")
 *             free_double_vector(score_vec)
 *             free_double_vector(time_vec)             # <<<<<<<<<<<<<<
 *             if self.scorer.reject_score(fit.score):
 *                 continue
 */

#line 543 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_time_vec);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":544
 *             free_double_vector(score_vec)
 *             free_double_vector(time_vec)
 *             if self.scorer.reject_score(fit.score):             # <<<<<<<<<<<<<<
 *                 continue
 *             feature_fits.append(fit)
 */

#line 544 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_4 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->scorer->__pyx_vtab)->reject_score(__pyx_v_self->scorer, __pyx_v_fit->score, 0) != 0);

#line 544 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_4) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":545
 *             free_double_vector(time_vec)
 *             if self.scorer.reject_score(fit.score):
 *                 continue             # <<<<<<<<<<<<<<
 *             feature_fits.append(fit)
 *         # print(f"Completed generating fits for {mz} @ {charge}: {combn_i}")
 */

#line 545 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":544
 *             free_double_vector(score_vec)
 *             free_double_vector(time_vec)
 *             if self.scorer.reject_score(fit.score):             # <<<<<<<<<<<<<<
 *                 continue
 *             feature_fits.append(fit)
 */

#line 544 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":546
 *             if self.scorer.reject_score(fit.score):
 *                 continue
 *             feature_fits.append(fit)             # <<<<<<<<<<<<<<
 *         # print(f"Completed generating fits for {mz} @ {charge}: {combn_i}")
 *         return feature_fits
 */

#line 546 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_feature_fits, ((PyObject *)__pyx_v_fit)); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 546, __pyx_L1_error)

#line 546 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_L3_continue:;

#line 546 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 546 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_break:;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":548
 *             feature_fits.append(fit)
 *         # print(f"Completed generating fits for {mz} @ {charge}: {combn_i}")
 *         return feature_fits             # <<<<<<<<<<<<<<
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,
 */

#line 548 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 548 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_feature_fits);

#line 548 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_feature_fits;

#line 548 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":428
 *         return fit
 * 
 *     cdef list _fit_theoretical_distribution_on_features(self, double mz, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                         TheoreticalIsotopicPattern base_tid,
 *                                                         double charge_carrier=PROTON,
 */

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_9);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_theoretical_distribution_on_features", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_feature_groups);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_feature_fits);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_features);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_eid);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_cleaned_eid);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_tid);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_scores_array);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_times_array);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_snapped_tid);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_feat_iter);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_f);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_combn_iter);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_conformer);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_fit);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 428 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":550
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_13_fit_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set *__pyx_optional_args) {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_v_left_search = ((int)1);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_v_right_search = ((int)1);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_charge_carrier = __pyx_k__7;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_truncate_after = ((double)0.8);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_v_max_missed_peaks = ((int)1);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_threshold_scale = ((double)0.3);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":552
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list feature_fits
 */

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
  PyObject *__pyx_v_feature_fits = 0
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_base_tid = 0
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_13 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_14;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_15 = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution __pyx_t_16;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features __pyx_t_17;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_fit_feature_set", 0);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_optional_args) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_left_search = __pyx_optional_args->left_search;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_v_right_search = __pyx_optional_args->right_search;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__pyx_optional_args->__pyx_n > 2) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (__pyx_optional_args->__pyx_n > 3) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __pyx_v_truncate_after = __pyx_optional_args->truncate_after;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            if (__pyx_optional_args->__pyx_n > 4) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
              __pyx_v_max_missed_peaks = __pyx_optional_args->max_missed_peaks;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
              if (__pyx_optional_args->__pyx_n > 5) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
                __pyx_v_threshold_scale = __pyx_optional_args->threshold_scale;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
                if (__pyx_optional_args->__pyx_n > 6) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
                  __pyx_v_feature = __pyx_optional_args->feature;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
                }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
              }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":550
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if called by wrapper */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if overridden in Python */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_feature_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_13_fit_feature_set)) {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_5);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_left_search); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_right_search); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_8 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_8);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_truncate_after); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_9);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_max_missed_peaks); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_10);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_11 = PyFloat_FromDouble(__pyx_v_threshold_scale); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_11);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_12 = __pyx_t_1; 
#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_t_13 = NULL;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_14 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (likely(__pyx_t_13)) {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(__pyx_t_13);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(function);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_DECREF_SET(__pyx_t_12, function);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __pyx_t_14 = 1;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCALL

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (PyFunction_Check(__pyx_t_12)) {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[11] = {__pyx_t_13, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, ((PyObject *)__pyx_v_feature)};

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_14, 10+__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCCALL

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[11] = {__pyx_t_13, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, ((PyObject *)__pyx_v_feature)};

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_14, 10+__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_15 = PyTuple_New(10+__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_15);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (__pyx_t_13) {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_13); __pyx_t_13 = NULL;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_14, __pyx_t_3);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_14, __pyx_t_4);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_5);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_15, 2+__pyx_t_14, __pyx_t_5);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_6);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_15, 3+__pyx_t_14, __pyx_t_6);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_7);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_15, 4+__pyx_t_14, __pyx_t_7);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_8);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_15, 5+__pyx_t_14, __pyx_t_8);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_9);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_15, 6+__pyx_t_14, __pyx_t_9);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_10);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_15, 7+__pyx_t_14, __pyx_t_10);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_11);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_15, 8+__pyx_t_14, __pyx_t_11);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_feature));

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_feature));

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_15, 9+__pyx_t_14, ((PyObject *)__pyx_v_feature));

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_3 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_4 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_5 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_6 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_7 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_8 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_9 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_10 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_11 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_2 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        goto __pyx_L0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #endif

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":556
 *             list feature_fits
 *             TheoreticalIsotopicPattern base_tid
 *         base_tid = self.create_theoretical_distribution(             # <<<<<<<<<<<<<<
 *             mz, charge, charge_carrier, truncate_after)
 *         # print(f"Fitting feature set for {mz} at charge {charge}")
 */

#line 556 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_16.__pyx_n = 2;

#line 556 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_16.charge_carrier = __pyx_v_charge_carrier;

#line 556 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_16.truncate_after = __pyx_v_truncate_after;

#line 556 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->create_theoretical_distribution(__pyx_v_self, __pyx_v_mz, __pyx_v_charge, 0, &__pyx_t_16)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)

#line 556 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 556 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_base_tid = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_1);

#line 556 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":559
 *             mz, charge, charge_carrier, truncate_after)
 *         # print(f"Fitting feature set for {mz} at charge {charge}")
 *         feature_fits = self._fit_theoretical_distribution_on_features(             # <<<<<<<<<<<<<<
 *             mz, error_tolerance, charge, base_tid,
 *             charge_carrier, truncate_after, max_missed_peaks, threshold_scale,
 */

#line 559 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_17.__pyx_n = 5;

#line 559 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_17.charge_carrier = __pyx_v_charge_carrier;

#line 559 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_17.truncate_after = __pyx_v_truncate_after;

#line 559 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_17.max_missed_peaks = __pyx_v_max_missed_peaks;

#line 559 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_17.threshold_scale = __pyx_v_threshold_scale;

#line 559 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_17.feature = __pyx_v_feature;

#line 559 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->_fit_theoretical_distribution_on_features(__pyx_v_self, __pyx_v_mz, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_base_tid, &__pyx_t_17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)

#line 559 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 559 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_feature_fits = ((PyObject*)__pyx_t_1);

#line 559 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":564
 *             feature)
 *         # print(f"Fitted {len(feature_fits)}")
 *         return feature_fits             # <<<<<<<<<<<<<<
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,
 */

#line 564 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 564 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_feature_fits);

#line 564 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_feature_fits;

#line 564 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":550
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_9);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_11);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_12);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_13);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_15);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_feature_fits);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_base_tid);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_13_fit_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_12_fit_feature_set[] = "LCMSFeatureProcessorBase._fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1, int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8, int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None) -> list";

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_13_fit_feature_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz
#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_error_tolerance
#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  int __pyx_v_charge
#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  int __pyx_v_left_search
#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  int __pyx_v_right_search
#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_charge_carrier
#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_truncate_after
#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  int __pyx_v_max_missed_peaks
#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_threshold_scale
#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = 0
#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_fit_feature_set (wrapper)", 0);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge,&__pyx_n_s_left_search,&__pyx_n_s_right_search,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_max_missed_peaks,&__pyx_n_s_threshold_scale,&__pyx_n_s_feature,0};

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":552
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list feature_fits
 */

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    values[9] = (PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_kwds)) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      Py_ssize_t kw_args;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {
        case 10: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[9] = PyTuple_GET_ITEM(__pyx_args, 9);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  9: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[8] = PyTuple_GET_ITEM(__pyx_args, 8);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  8: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[7] = PyTuple_GET_ITEM(__pyx_args, 7);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  7: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[6] = PyTuple_GET_ITEM(__pyx_args, 6);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  6: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0: break;
        default: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0:

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  1:

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_feature_set", 0, 3, 10, 1); 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 550, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  2:

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_feature_set", 0, 3, 10, 2); 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 550, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  3:

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_search);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[3] = value; kw_args--; }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  4:

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_search);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  5:

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[5] = value; kw_args--; }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  6:

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[6] = value; kw_args--; }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  7:

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_missed_peaks);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[7] = value; kw_args--; }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  8:

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_threshold_scale);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[8] = value; kw_args--; }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  9:

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feature);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[9] = value; kw_args--; }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(kw_args > 0)) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_fit_feature_set") < 0)) __PYX_ERR(0, 550, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 10: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[9] = PyTuple_GET_ITEM(__pyx_args, 9);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  9: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[8] = PyTuple_GET_ITEM(__pyx_args, 8);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  8: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[7] = PyTuple_GET_ITEM(__pyx_args, 7);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  7: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[6] = PyTuple_GET_ITEM(__pyx_args, 6);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  6: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        break;
        default: 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[3]) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_left_search = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_left_search == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_left_search = ((int)1);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[4]) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_right_search = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_right_search == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 551, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_right_search = ((int)1);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[5]) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 551, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_k__7;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[6]) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_truncate_after = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_truncate_after == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 551, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_truncate_after = ((double)0.8);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[7]) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_max_missed_peaks = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_max_missed_peaks == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 552, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_max_missed_peaks = ((int)1);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[8]) {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_threshold_scale = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_threshold_scale == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 552, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_threshold_scale = ((double)0.3);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_feature = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)values[9]);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_fit_feature_set", 0, 3, 10, PyTuple_GET_SIZE(__pyx_args)); 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 550, __pyx_L3_error)

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L3_error:;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return NULL;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_feature), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, 1, "feature", 0))) __PYX_ERR(0, 552, __pyx_L1_error)
  __pyx_r = 
#line 552 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_12_fit_feature_set(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_mz, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_left_search, __pyx_v_right_search, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_max_missed_peaks, __pyx_v_threshold_scale, __pyx_v_feature);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":550
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_12_fit_feature_set(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_v_left_search, int __pyx_v_right_search, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature) {

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set __pyx_t_2;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_fit_feature_set", 0);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.__pyx_n = 7;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.left_search = __pyx_v_left_search;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.right_search = __pyx_v_right_search;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.max_missed_peaks = __pyx_v_max_missed_peaks;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.threshold_scale = __pyx_v_threshold_scale;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.feature = __pyx_v_feature;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->_fit_feature_set(__pyx_v_self, __pyx_v_mz, __pyx_v_error_tolerance, __pyx_v_charge, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 550, __pyx_L1_error)

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_feature_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":566
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_15_fit_composition(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_composition, double __pyx_v_error_tolerance, int __pyx_v_charge, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition *__pyx_optional_args) {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_charge_carrier = __pyx_k__8;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_truncate_after = ((double)0.8);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_ignore_below = ((double)0.05);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_v_max_missed_peaks = ((int)1);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_v_threshold_scale = ((double)0.3);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":568
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list base_tid, kept_tid, feature_fits
 */

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);
  PyObject *__pyx_v_base_tid = 0
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_kept_tid = 0
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  PyObject *__pyx_v_feature_fits = 0
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_total
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_cumsum
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_mz
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_tp = 0
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_i
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  size_t __pyx_v_n
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_12;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_13 = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants __pyx_t_14;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_15;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_16;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  size_t __pyx_t_17;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_18;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_19;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_t_20;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  double __pyx_t_21;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features __pyx_t_22;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_fit_composition", 0);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_optional_args) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_v_truncate_after = __pyx_optional_args->truncate_after;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__pyx_optional_args->__pyx_n > 2) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_v_ignore_below = __pyx_optional_args->ignore_below;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (__pyx_optional_args->__pyx_n > 3) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __pyx_v_max_missed_peaks = __pyx_optional_args->max_missed_peaks;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            if (__pyx_optional_args->__pyx_n > 4) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
              __pyx_v_threshold_scale = __pyx_optional_args->threshold_scale;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
              if (__pyx_optional_args->__pyx_n > 5) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
                __pyx_v_feature = __pyx_optional_args->feature;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
              }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":566
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if called by wrapper */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* Check if overridden in Python */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fit_composition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_15_fit_composition)) {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_charge_carrier); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_5);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_truncate_after); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_ignore_below); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_max_missed_peaks); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_8);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_threshold_scale); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_GOTREF(__pyx_t_9);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_10 = __pyx_t_1; 
#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_t_11 = NULL;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_12 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (likely(__pyx_t_11)) {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(__pyx_t_11);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_INCREF(function);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_DECREF_SET(__pyx_t_10, function);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __pyx_t_12 = 1;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCALL

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (PyFunction_Check(__pyx_t_10)) {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[10] = {__pyx_t_11, __pyx_v_composition, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, ((PyObject *)__pyx_v_feature)};

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 9+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #if CYTHON_FAST_PYCCALL

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject *__pyx_temp[10] = {__pyx_t_11, __pyx_v_composition, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, ((PyObject *)__pyx_v_feature)};

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 9+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        } else

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        #endif

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_13 = PyTuple_New(9+__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_13);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (__pyx_t_11) {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
            __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_11); __pyx_t_11 = NULL;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          }

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_INCREF(__pyx_v_composition);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_v_composition);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_12, __pyx_v_composition);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_12, __pyx_t_3);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_12, __pyx_t_4);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_5);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_13, 3+__pyx_t_12, __pyx_t_5);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_6);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_13, 4+__pyx_t_12, __pyx_t_6);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_7);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_13, 5+__pyx_t_12, __pyx_t_7);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_8);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_13, 6+__pyx_t_12, __pyx_t_8);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(__pyx_t_9);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_13, 7+__pyx_t_12, __pyx_t_9);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_feature));

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_feature));

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyTuple_SET_ITEM(__pyx_t_13, 8+__pyx_t_12, ((PyObject *)__pyx_v_feature));

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_3 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_4 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_5 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_6 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_7 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_8 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_9 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_t_2 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        goto __pyx_L0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #endif

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    #endif

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":577
 *         # Can't refactor shared components with `create_theoretical_isotopic_distribution` because
 *         # this needs to include its own truncate-after component as well as ignore-under
 *         base_tid = _isotopic_variants(composition, npeaks=None, charge=charge, charge_carrier=charge_carrier)             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(base_tid)
 *         total = 0
 */

#line 577 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_14.__pyx_n = 3;

#line 577 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_14.npeaks = Py_None;

#line 577 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_14.charge = __pyx_v_charge;

#line 577 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_14.charge_carrier = __pyx_v_charge_carrier;

#line 577 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_f_7brainpy_2_c_21isotopic_distribution__isotopic_variants(__pyx_v_composition, 0, &__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)

#line 577 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 577 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_base_tid = ((PyObject*)__pyx_t_1);

#line 577 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":578
 *         # this needs to include its own truncate-after component as well as ignore-under
 *         base_tid = _isotopic_variants(composition, npeaks=None, charge=charge, charge_carrier=charge_carrier)
 *         n = PyList_GET_SIZE(base_tid)             # <<<<<<<<<<<<<<
 *         total = 0
 *         cumsum = 0
 */

#line 578 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_base_tid);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":579
 *         base_tid = _isotopic_variants(composition, npeaks=None, charge=charge, charge_carrier=charge_carrier)
 *         n = PyList_GET_SIZE(base_tid)
 *         total = 0             # <<<<<<<<<<<<<<
 *         cumsum = 0
 *         kept_tid = []
 */

#line 579 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_total = 0.0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":580
 *         n = PyList_GET_SIZE(base_tid)
 *         total = 0
 *         cumsum = 0             # <<<<<<<<<<<<<<
 *         kept_tid = []
 *         for i in range(n):
 */

#line 580 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_cumsum = 0.0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":581
 *         total = 0
 *         cumsum = 0
 *         kept_tid = []             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(base_tid, i)
 */

#line 581 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)

#line 581 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 581 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_kept_tid = ((PyObject*)__pyx_t_1);

#line 581 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":582
 *         cumsum = 0
 *         kept_tid = []
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(base_tid, i)
 *             cumsum += tp.intensity
 */

#line 582 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_15 = __pyx_v_n;

#line 582 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_16 = __pyx_t_15;

#line 582 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {

#line 582 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_17;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":583
 *         kept_tid = []
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(base_tid, i)             # <<<<<<<<<<<<<<
 *             cumsum += tp.intensity
 *             if (tp.intensity < ignore_below) and (i > 1):
 */

#line 583 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_18 = PyList_GET_ITEM(__pyx_v_base_tid, __pyx_v_i);

#line 583 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_18);

#line 583 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 583 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_tp, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 583 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":584
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(base_tid, i)
 *             cumsum += tp.intensity             # <<<<<<<<<<<<<<
 *             if (tp.intensity < ignore_below) and (i > 1):
 *                 pass
 */

#line 584 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_cumsum = (__pyx_v_cumsum + __pyx_v_tp->intensity);

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":585
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(base_tid, i)
 *             cumsum += tp.intensity
 *             if (tp.intensity < ignore_below) and (i > 1):             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_20 = ((__pyx_v_tp->intensity < __pyx_v_ignore_below) != 0);

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_20) {

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_19 = __pyx_t_20;

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L6_bool_binop_done;

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_20 = ((__pyx_v_i > 1) != 0);

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_19 = __pyx_t_20;

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_L6_bool_binop_done:;

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_19) {

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L5;

#line 585 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":588
 *                 pass
 *             else:
 *                 PyList_Append(kept_tid, tp)             # <<<<<<<<<<<<<<
 *                 total += tp.intensity
 *             if cumsum >= truncate_after:
 */

#line 588 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    /*else*/ {

#line 588 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_t_12 = PyList_Append(__pyx_v_kept_tid, ((PyObject *)__pyx_v_tp)); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 588, __pyx_L1_error)

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":589
 *             else:
 *                 PyList_Append(kept_tid, tp)
 *                 total += tp.intensity             # <<<<<<<<<<<<<<
 *             if cumsum >= truncate_after:
 *                 break
 */

#line 589 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_total = (__pyx_v_total + __pyx_v_tp->intensity);

#line 589 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 589 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_L5:;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":590
 *                 PyList_Append(kept_tid, tp)
 *                 total += tp.intensity
 *             if cumsum >= truncate_after:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */

#line 590 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_19 = ((__pyx_v_cumsum >= __pyx_v_truncate_after) != 0);

#line 590 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (__pyx_t_19) {

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":591
 *                 total += tp.intensity
 *             if cumsum >= truncate_after:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         n = PyList_GET_SIZE(kept_tid)
 */

#line 591 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      goto __pyx_L4_break;

      /* "ms_deisotope/_c/feature_map/feature_processor.pyx":590
 *                 PyList_Append(kept_tid, tp)
 *                 total += tp.intensity
 *             if cumsum >= truncate_after:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */

#line 590 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 590 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 590 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_break:;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":593
 *                 break
 * 
 *         n = PyList_GET_SIZE(kept_tid)             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, i)
 */

#line 593 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_kept_tid);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":594
 * 
 *         n = PyList_GET_SIZE(kept_tid)
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, i)
 *             tp.intensity /= total
 */

#line 594 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_15 = __pyx_v_n;

#line 594 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_16 = __pyx_t_15;

#line 594 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {

#line 594 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_i = __pyx_t_17;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":595
 *         n = PyList_GET_SIZE(kept_tid)
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, i)             # <<<<<<<<<<<<<<
 *             tp.intensity /= total
 * 
 */

#line 595 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_18 = PyList_GET_ITEM(__pyx_v_kept_tid, __pyx_v_i);

#line 595 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_18);

#line 595 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 595 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF_SET(__pyx_v_tp, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 595 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":596
 *         for i in range(n):
 *             tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, i)
 *             tp.intensity /= total             # <<<<<<<<<<<<<<
 * 
 *         if n == 0:
 */

#line 596 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_v_total == 0)) {

#line 596 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 596 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __PYX_ERR(0, 596, __pyx_L1_error)

#line 596 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 596 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_tp->intensity = (__pyx_v_tp->intensity / __pyx_v_total);

#line 596 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":598
 *             tp.intensity /= total
 * 
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return []
 * 
 */

#line 598 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_19 = ((__pyx_v_n == 0) != 0);

#line 598 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (__pyx_t_19) {

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":599
 * 
 *         if n == 0:
 *             return []             # <<<<<<<<<<<<<<
 * 
 *         tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, 0)
 */

#line 599 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 599 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 599, __pyx_L1_error)

#line 599 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 599 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_r = ((PyObject*)__pyx_t_1);

#line 599 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_t_1 = 0;

#line 599 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":598
 *             tp.intensity /= total
 * 
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return []
 * 
 */

#line 598 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":601
 *             return []
 * 
 *         tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, 0)             # <<<<<<<<<<<<<<
 *         mz = tp.mz
 * 
 */

#line 601 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_18 = PyList_GET_ITEM(__pyx_v_kept_tid, 0);

#line 601 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_t_18);

#line 601 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 601 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF_SET(__pyx_v_tp, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 601 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":602
 * 
 *         tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, 0)
 *         mz = tp.mz             # <<<<<<<<<<<<<<
 * 
 *         feature_fits = self._fit_theoretical_distribution_on_features(
 */

#line 602 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_21 = __pyx_v_tp->mz;

#line 602 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_mz = __pyx_t_21;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":605
 * 
 *         feature_fits = self._fit_theoretical_distribution_on_features(
 *             mz, error_tolerance, charge, kept_tid,             # <<<<<<<<<<<<<<
 *             charge_carrier, truncate_after, max_missed_peaks, threshold_scale,
 *             feature)
 */

#line 605 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (!(likely(__Pyx_TypeTest(__pyx_v_kept_tid, __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern)))) __PYX_ERR(0, 605, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":604
 *         mz = tp.mz
 * 
 *         feature_fits = self._fit_theoretical_distribution_on_features(             # <<<<<<<<<<<<<<
 *             mz, error_tolerance, charge, kept_tid,
 *             charge_carrier, truncate_after, max_missed_peaks, threshold_scale,
 */

#line 604 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_22.__pyx_n = 5;

#line 604 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_22.charge_carrier = __pyx_v_charge_carrier;

#line 604 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_22.truncate_after = __pyx_v_truncate_after;

#line 604 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_22.max_missed_peaks = __pyx_v_max_missed_peaks;

#line 604 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_22.threshold_scale = __pyx_v_threshold_scale;

#line 604 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_22.feature = __pyx_v_feature;

#line 604 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self->__pyx_vtab)->_fit_theoretical_distribution_on_features(__pyx_v_self, __pyx_v_mz, __pyx_v_error_tolerance, __pyx_v_charge, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_kept_tid), &__pyx_t_22); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L1_error)

#line 604 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 604 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_feature_fits = ((PyObject*)__pyx_t_1);

#line 604 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":608
 *             charge_carrier, truncate_after, max_missed_peaks, threshold_scale,
 *             feature)
 *         return feature_fits             # <<<<<<<<<<<<<<
 */

#line 608 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 608 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_v_feature_fits);

#line 608 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_v_feature_fits;

#line 608 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":566
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_9);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_11);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_13);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_base_tid);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_kept_tid);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_v_feature_fits);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_tp);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_15_fit_composition(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_14_fit_composition[] = "LCMSFeatureProcessorBase._fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON, double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None) -> list";

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_15_fit_composition(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_composition = 0
#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_error_tolerance
#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  int __pyx_v_charge
#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_charge_carrier
#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_truncate_after
#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_ignore_below
#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  int __pyx_v_max_missed_peaks
#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  double __pyx_v_threshold_scale
#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = 0
#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_fit_composition (wrapper)", 0);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_composition,&__pyx_n_s_error_tolerance,&__pyx_n_s_charge,&__pyx_n_s_charge_carrier,&__pyx_n_s_truncate_after,&__pyx_n_s_ignore_below,&__pyx_n_s_max_missed_peaks,&__pyx_n_s_threshold_scale,&__pyx_n_s_feature,0};

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};

    /* "ms_deisotope/_c/feature_map/feature_processor.pyx":568
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list base_tid, kept_tid, feature_fits
 */

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    values[8] = (PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (unlikely(__pyx_kwds)) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      Py_ssize_t kw_args;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {
        case  9: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[8] = PyTuple_GET_ITEM(__pyx_args, 8);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  8: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[7] = PyTuple_GET_ITEM(__pyx_args, 7);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  7: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[6] = PyTuple_GET_ITEM(__pyx_args, 6);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  6: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0: break;
        default: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (pos_args) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  0:

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_composition)) != 0)) kw_args--;
        else 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  1:

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_composition", 0, 3, 9, 1); 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 566, __pyx_L3_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  2:

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_fit_composition", 0, 3, 9, 2); 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 566, __pyx_L3_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  3:

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[3] = value; kw_args--; }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  4:

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_truncate_after);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  5:

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ignore_below);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[5] = value; kw_args--; }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  6:

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_missed_peaks);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[6] = value; kw_args--; }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  7:

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_threshold_scale);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[7] = value; kw_args--; }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        case  8:

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (kw_args > 0) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feature);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
          if (value) { values[8] = value; kw_args--; }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      if (unlikely(kw_args > 0)) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_fit_composition") < 0)) __PYX_ERR(0, 566, __pyx_L3_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[8] = PyTuple_GET_ITEM(__pyx_args, 8);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  8: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[7] = PyTuple_GET_ITEM(__pyx_args, 7);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  7: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[6] = PyTuple_GET_ITEM(__pyx_args, 6);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  6: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
        break;
        default: 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
goto __pyx_L5_argtuple_error;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_composition = ((PyObject*)values[0]);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 566, __pyx_L3_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 566, __pyx_L3_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[3]) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 566, __pyx_L3_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_charge_carrier = __pyx_k__8;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[4]) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_truncate_after = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_truncate_after == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 567, __pyx_L3_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_truncate_after = ((double)0.8);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[5]) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_ignore_below = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_ignore_below == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 567, __pyx_L3_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_ignore_below = ((double)0.05);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[6]) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_max_missed_peaks = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_max_missed_peaks == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 567, __pyx_L3_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_max_missed_peaks = ((int)1);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    if (values[7]) {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_threshold_scale = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_threshold_scale == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 568, __pyx_L3_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    } else {

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
      __pyx_v_threshold_scale = ((double)0.3);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
    __pyx_v_feature = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)values[8]);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  }

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_fit_composition", 0, 3, 9, PyTuple_GET_SIZE(__pyx_args)); 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__PYX_ERR(0, 566, __pyx_L3_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L3_error:;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return NULL;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_composition), (&PyDict_Type), 1, "composition", 1))) __PYX_ERR(0, 566, __pyx_L1_error)

#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_feature), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, 1, "feature", 0))) __PYX_ERR(0, 568, __pyx_L1_error)
  __pyx_r = 
#line 568 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_14_fit_composition(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), __pyx_v_composition, __pyx_v_error_tolerance, __pyx_v_charge, __pyx_v_charge_carrier, __pyx_v_truncate_after, __pyx_v_ignore_below, __pyx_v_max_missed_peaks, __pyx_v_threshold_scale, __pyx_v_feature);

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":566
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_14_fit_composition(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_composition, double __pyx_v_error_tolerance, int __pyx_v_charge, double __pyx_v_charge_carrier, double __pyx_v_truncate_after, double __pyx_v_ignore_below, int __pyx_v_max_missed_peaks, double __pyx_v_threshold_scale, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature) {

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition __pyx_t_2;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("_fit_composition", 0);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.__pyx_n = 6;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.truncate_after = __pyx_v_truncate_after;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.ignore_below = __pyx_v_ignore_below;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.max_missed_peaks = __pyx_v_max_missed_peaks;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.threshold_scale = __pyx_v_threshold_scale;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2.feature = __pyx_v_feature;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase->_fit_composition(__pyx_v_self, __pyx_v_composition, __pyx_v_error_tolerance, __pyx_v_charge, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = __pyx_t_1;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase._fit_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = NULL;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":265
 * cdef class LCMSFeatureProcessorBase(object):
 *     cdef:
 *         public LCMSFeatureMap feature_map             # <<<<<<<<<<<<<<
 *         public IsotopicFitterBase scorer
 *         public AveragineCache averagine
 */

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_1__get__(PyObject *__pyx_v_self) {

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->feature_map));

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = ((PyObject *)__pyx_v_self->feature_map);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap))))) __PYX_ERR(0, 265, __pyx_L1_error)

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_value;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->feature_map);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->feature_map));

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->feature_map = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_t_1);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.feature_map.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = -1;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_5__del__(PyObject *__pyx_v_self) {

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(Py_None);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(Py_None);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->feature_map);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->feature_map));

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->feature_map = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)Py_None);

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 265 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":266
 *     cdef:
 *         public LCMSFeatureMap feature_map
 *         public IsotopicFitterBase scorer             # <<<<<<<<<<<<<<
 *         public AveragineCache averagine
 * 
 */

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_1__get__(PyObject *__pyx_v_self) {

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->scorer));

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = ((PyObject *)__pyx_v_self->scorer);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase))))) __PYX_ERR(0, 266, __pyx_L1_error)

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_value;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->scorer);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scorer));

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_t_1);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.scorer.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = -1;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_5__del__(PyObject *__pyx_v_self) {

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(Py_None);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(Py_None);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->scorer);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scorer));

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)Py_None);

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 266 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_processor.pyx":267
 *         public LCMSFeatureMap feature_map
 *         public IsotopicFitterBase scorer
 *         public AveragineCache averagine             # <<<<<<<<<<<<<<
 * 
 *     cpdef TheoreticalIsotopicPattern create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8,
 */

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_1__get__(PyObject *__pyx_v_self) {

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = 0;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_r = NULL;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->averagine));

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = ((PyObject *)__pyx_v_self->averagine);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_lineno = 0;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  const char *__pyx_filename = NULL;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_clineno = 0;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache))))) __PYX_ERR(0, 267, __pyx_L1_error)

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __pyx_v_value;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->averagine);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->averagine));

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)__pyx_t_1);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = 0;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  goto __pyx_L0;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L1_error:;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.averagine.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = -1;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_L0:;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
/* Python wrapper */

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_5__del__(PyObject *__pyx_v_self) {

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  int __pyx_r;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyDeclarations

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_INCREF(Py_None);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(Py_None);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_v_self->averagine);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->averagine));

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_self->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)Py_None);

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"


#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  /* function exit code */

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_r = 0;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_RefNannyFinishContext();

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  return __pyx_r;

#line 267 "ms_deisotope/_c/feature_map/feature_processor.pyx"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_16__reduce_cython__[] = "LCMSFeatureProcessorBase.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_16__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_16__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.averagine, self.feature_map, self.scorer)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->averagine));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->averagine));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->averagine));

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->feature_map));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->feature_map));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->feature_map));

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->scorer));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->scorer));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->scorer));

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.averagine, self.feature_map, self.scorer)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_1;

#line 6 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.averagine, self.feature_map, self.scorer)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_2 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));

#line 8 "(tree fragment)"
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.averagine is not None or self.feature_map is not None or self.scorer is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.averagine, self.feature_map, self.scorer)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.averagine is not None or self.feature_map is not None or self.scorer is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_2 = (((PyObject *)__pyx_v_self->averagine) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_5 = (__pyx_t_2 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_5) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_3 = __pyx_t_5;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_5 = (((PyObject *)__pyx_v_self->feature_map) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_2 = (__pyx_t_5 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_2) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_3 = __pyx_t_2;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_2 = (((PyObject *)__pyx_v_self->scorer) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_5 = (__pyx_t_2 != 0);

#line 11 "(tree fragment)"
    __pyx_t_3 = __pyx_t_5;

#line 11 "(tree fragment)"
    __pyx_L4_bool_binop_done:;

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_3;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.averagine is not None or self.feature_map is not None or self.scorer is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_3 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.averagine is not None or self.feature_map is not None or self.scorer is not None
 *     if use_setstate:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_LCMSFeatureProces); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_177677261);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_177677261);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_177677261);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_6;

#line 13 "(tree fragment)"
    __pyx_t_6 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.averagine is not None or self.feature_map is not None or self.scorer is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, None), state
 *     else:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LCMSFeatureProcessorBase__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_LCMSFeatureProces); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 15 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_177677261);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_177677261);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_177677261);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_6);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);

#line 15 "(tree fragment)"
    __pyx_t_6 = 0;

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_4;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LCMSFeatureProcessorBase__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_18__setstate_cython__[] = "LCMSFeatureProcessorBase.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_18__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_18__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LCMSFeatureProcessorBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_LCMSFeatureProcessorBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LCMSFeatureProcessorBase, (type(self), 0xa9723cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LCMSFeatureProcessorBase__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_envelope_conformer(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_3__pyx_unpickle_envelope_conformer(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_2__pyx_unpickle_envelope_conformer[] = "__pyx_unpickle_envelope_conformer(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_11feature_map_17feature_processor_3__pyx_unpickle_envelope_conformer = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_envelope_conformer", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_3__pyx_unpickle_envelope_conformer, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_2__pyx_unpickle_envelope_conformer};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_3__pyx_unpickle_envelope_conformer(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_envelope_conformer (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_envelope_conformer", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_envelope_conformer", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_envelope_conformer") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_envelope_conformer", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_envelope_conformer", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_2__pyx_unpickle_envelope_conformer(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_2__pyx_unpickle_envelope_conformer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_envelope_conformer", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xba9b7dc, 0x30f7136, 0x9f9ee19):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xba9b7dc, 0x30f7136, 0x9f9ee19) = (experimental, n_missing, theoretical))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__9, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xba9b7dc, 0x30f7136, 0x9f9ee19):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xba9b7dc, 0x30f7136, 0x9f9ee19) = (experimental, n_missing, theoretical))" % __pyx_checksum)
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xba9b7dc, 0x30f7136, 0x9f9ee19):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xba9b7dc, 0x30f7136, 0x9f9ee19) = (experimental, n_missing, theoretical))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xba9b7dc, 0x30f7136, 0x9f9ee19):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xba9b7dc, 0x30f7136, 0x9f9ee19) = (experimental, n_missing, theoretical))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xba9b7dc, 0x30f7136, 0x9f9ee19) = (experimental, n_missing, theoretical))" % __pyx_checksum)
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xba9b7dc, 0x30f7136, 0x9f9ee19) = (experimental, n_missing, theoretical))" % __pyx_checksum)
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_envelope_conformer__set_state(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xba9b7dc, 0x30f7136, 0x9f9ee19) = (experimental, n_missing, theoretical))" % __pyx_checksum)
 *     __pyx_result = envelope_conformer.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_envelope_conformer(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_envelope_conformer", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_envelope_conformer__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  size_t __pyx_t_2;

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_envelope_conformer__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->experimental);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->experimental);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->experimental = ((PyObject*)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->n_missing = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->theoretical);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->theoretical);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->theoretical = ((PyObject*)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = ((__pyx_t_4 > 3) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_5) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_3 = __pyx_t_5;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 13 "(tree fragment)"
  __pyx_t_3 = __pyx_t_6;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_9 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {

#line 14 "(tree fragment)"
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_9)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_8, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_envelope_conformer__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_LCMSFeatureProcessorBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_5__pyx_unpickle_LCMSFeatureProcessorBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_4__pyx_unpickle_LCMSFeatureProcessorBase[] = "__pyx_unpickle_LCMSFeatureProcessorBase(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_11feature_map_17feature_processor_5__pyx_unpickle_LCMSFeatureProcessorBase = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_LCMSFeatureProcessorBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_5__pyx_unpickle_LCMSFeatureProcessorBase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_4__pyx_unpickle_LCMSFeatureProcessorBase};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_5__pyx_unpickle_LCMSFeatureProcessorBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_LCMSFeatureProcessorBase (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LCMSFeatureProcessorBase", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LCMSFeatureProcessorBase", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_LCMSFeatureProcessorBase") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LCMSFeatureProcessorBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_LCMSFeatureProcessorBase", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_4__pyx_unpickle_LCMSFeatureProcessorBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_17feature_processor_4__pyx_unpickle_LCMSFeatureProcessorBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_LCMSFeatureProcessorBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xa9723cd, 0xcb14c85, 0xe2c6adb):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xa9723cd, 0xcb14c85, 0xe2c6adb) = (averagine, feature_map, scorer))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__10, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xa9723cd, 0xcb14c85, 0xe2c6adb):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xa9723cd, 0xcb14c85, 0xe2c6adb) = (averagine, feature_map, scorer))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xa9723cd, 0xcb14c85, 0xe2c6adb):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xa9723cd, 0xcb14c85, 0xe2c6adb) = (averagine, feature_map, scorer))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xa9723cd, 0xcb14c85, 0xe2c6adb):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xa9723cd, 0xcb14c85, 0xe2c6adb) = (averagine, feature_map, scorer))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xa9723cd, 0xcb14c85, 0xe2c6adb) = (averagine, feature_map, scorer))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xa9723cd, 0xcb14c85, 0xe2c6adb) = (averagine, feature_map, scorer))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_LCMSFeatureProcessorBase__set_state(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xa9723cd, 0xcb14c85, 0xe2c6adb) = (averagine, feature_map, scorer))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureProcessorBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LCMSFeatureProcessorBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_LCMSFeatureProcessorBase", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor___pyx_unpickle_LCMSFeatureProcessorBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  int __pyx_t_2;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_3;

#line 11 "(tree fragment)"
  int __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_LCMSFeatureProcessorBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache))))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->averagine);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->averagine));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap))))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->feature_map);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->feature_map));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->feature_map = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase))))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->scorer);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->scorer));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_4 = ((__pyx_t_3 > 3) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_4) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_2 = __pyx_t_4;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = (__pyx_t_4 != 0);

#line 13 "(tree fragment)"
  __pyx_t_2 = __pyx_t_5;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 14 "(tree fragment)"
    __pyx_t_8 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {

#line 14 "(tree fragment)"
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_8)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_7, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_LCMSFeatureProcessorBase__set_state(<LCMSFeatureProcessorBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureProcessorBase__set_state(LCMSFeatureProcessorBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.averagine = __pyx_state[0]; __pyx_result.feature_map = __pyx_state[1]; __pyx_result.scorer = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_processor.__pyx_unpickle_LCMSFeatureProcessorBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 735, __pyx_L1_error)

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 738, __pyx_L1_error)

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 741, __pyx_L1_error)

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 744, __pyx_L1_error)

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 747, __pyx_L1_error)

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /*else*/ {

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(__pyx_empty_tuple);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = __pyx_empty_tuple;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":929
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */

#line 929 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_INCREF(__pyx_v_base);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":930
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */

#line 930 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base
#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":933
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */

#line 933 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":935
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":936
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_base));

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyObject *)__pyx_v_base);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":942
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */

#line 942 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 942, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":943
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 943, __pyx_L5_except_error)

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(2, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":948
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 948 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 948, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":949
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 949, __pyx_L5_except_error)

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(2, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":954
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 954 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 954, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":955
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 955, __pyx_L5_except_error)

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":956
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(2, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":978
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":993
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_datetime __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1003
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_timedelta __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1010
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1017
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->collections = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyObject *etype, *eval, *etb;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_Fetch(&etype, &eval, &etb);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_3__dealloc__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_Restore(etype, eval, etb);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->collections);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->collections) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->collections, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject* tmp;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->collections);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->collections = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_collections(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_collections(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11collections_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_size(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_size(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4size_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_done(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_done(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4done_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_total_combinations(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_total_combinations(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_18total_combinations_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_specialmethod___pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11__next__(self);}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"get_lengths", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_5get_lengths, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_4get_lengths},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"get_indices", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_7get_indices, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_6get_indices},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11__next__, METH_NOARGS|METH_COEXIST, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_13__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_12__reduce_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_15__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_14__setstate_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"collections", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_collections, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_collections, (char *)"collections: list", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"size", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_size, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_size, (char *)"size: 'size_t'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"done", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_done, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_done, (char *)"done: 'bool'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"total_combinations", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_total_combinations, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_total_combinations, (char *)"total_combinations: 'size_t'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.feature_processor.CartesianProductIterator", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "CartesianProductIterator(collections)", /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_9__iter__, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_11__next__, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_1__init__, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->experimental = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->theoretical = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->experimental);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->theoretical);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->experimental) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->experimental, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->theoretical) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->theoretical, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject* tmp;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->experimental);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->experimental = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->theoretical);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->theoretical = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_1__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer___reduce_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_3__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_2__setstate_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.feature_processor.envelope_conformer", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->feature_map = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_USE_TP_FINALIZE

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->feature_map);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->scorer);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->averagine);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->feature_map) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(((PyObject *)p->feature_map), a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->scorer) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(((PyObject *)p->scorer), a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->averagine) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(((PyObject *)p->averagine), a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject* tmp;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->feature_map);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->feature_map = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->scorer);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->averagine);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->averagine = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_feature_map(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_feature_map(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11feature_map_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_scorer(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_scorer(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6scorer_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_averagine(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_averagine(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9averagine_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"create_theoretical_distribution", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_1create_theoretical_distribution, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"find_all_features", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_3find_all_features, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_2find_all_features},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"find_features", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_5find_features, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_4find_features},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"match_theoretical_isotopic_distribution", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_7match_theoretical_isotopic_distribution, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_6match_theoretical_isotopic_distribution},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"conform_envelopes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_9conform_envelopes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_8conform_envelopes},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_fit_single_feature_set", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_11_fit_single_feature_set, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_10_fit_single_feature_set},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_fit_feature_set", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_13_fit_feature_set, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_12_fit_feature_set},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_fit_composition", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_15_fit_composition, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_14_fit_composition},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_17__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_16__reduce_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_19__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_18__setstate_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"feature_map", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_feature_map, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_feature_map, (char *)"feature_map: ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"scorer", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_scorer, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_scorer, (char *)"scorer: ms_deisotope._c.scoring.IsotopicFitterBase", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"averagine", __pyx_getprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_averagine, __pyx_setprop_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_averagine, (char *)"averagine: ms_deisotope._c.averagine.AveragineCache", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.feature_processor.LCMSFeatureProcessorBase", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"conform_envelopes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_17feature_processor_1conform_envelopes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_17feature_processor_conform_envelopes},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
#if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
#if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_pymod_exec_feature_processor(PyObject* module); /*proto*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyModuleDef_Slot __pyx_moduledef_slots[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {Py_mod_create, (void*)__pyx_pymod_create},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {Py_mod_exec, (void*)__pyx_pymod_exec_feature_processor},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, NULL}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
#endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyModuleDef __pyx_moduledef = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyModuleDef_HEAD_INIT,

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    "feature_processor",

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    0, /* m_doc */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    0, /* m_size */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    -1, /* m_size */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_methods /* m_methods */,

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_moduledef_slots, /* m_slots */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    NULL, /* m_reload */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    NULL, /* m_traverse */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    NULL, /* m_clear */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    NULL /* m_free */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
#endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_0_2f_0_2f, __pyx_k_0_2f_0_2f, sizeof(__pyx_k_0_2f_0_2f), 0, 0, 1, 0},
  {&__pyx_n_s_CartesianProductIterator, __pyx_k_CartesianProductIterator, sizeof(__pyx_k_CartesianProductIterator), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
  {&__pyx_n_s_LCMSFeatureProcessorBase, __pyx_k_LCMSFeatureProcessorBase, sizeof(__pyx_k_LCMSFeatureProcessorBase), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 0, 1, 0},
  {&__pyx_n_s_base_theoretical, __pyx_k_base_theoretical, sizeof(__pyx_k_base_theoretical), 0, 0, 1, 1},
  {&__pyx_n_s_base_tid, __pyx_k_base_tid, sizeof(__pyx_k_base_tid), 0, 0, 1, 1},
  {&__pyx_n_s_charge, __pyx_k_charge, sizeof(__pyx_k_charge), 0, 0, 1, 1},
  {&__pyx_n_s_charge_carrier, __pyx_k_charge_carrier, sizeof(__pyx_k_charge_carrier), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_composition, __pyx_k_composition, sizeof(__pyx_k_composition), 0, 0, 1, 1},
  {&__pyx_n_s_conform_envelopes, __pyx_k_conform_envelopes, sizeof(__pyx_k_conform_envelopes), 0, 0, 1, 1},
  {&__pyx_n_s_create_theoretical_distribution, __pyx_k_create_theoretical_distribution, sizeof(__pyx_k_create_theoretical_distribution), 0, 0, 1, 1},
  {&__pyx_kp_s_d_combinations_have_been_consid, __pyx_k_d_combinations_have_been_consid, sizeof(__pyx_k_d_combinations_have_been_consid), 0, 0, 1, 0},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_end_time, __pyx_k_end_time, sizeof(__pyx_k_end_time), 0, 0, 1, 1},
  {&__pyx_n_s_envelope_conformer, __pyx_k_envelope_conformer, sizeof(__pyx_k_envelope_conformer), 0, 0, 1, 1},
  {&__pyx_n_s_error_tolerance, __pyx_k_error_tolerance, sizeof(__pyx_k_error_tolerance), 0, 0, 1, 1},
  {&__pyx_n_s_experimental, __pyx_k_experimental, sizeof(__pyx_k_experimental), 0, 0, 1, 1},
  {&__pyx_n_s_feature, __pyx_k_feature, sizeof(__pyx_k_feature), 0, 0, 1, 1},
  {&__pyx_n_s_features, __pyx_k_features, sizeof(__pyx_k_features), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_find_all_features, __pyx_k_find_all_features, sizeof(__pyx_k_find_all_features), 0, 0, 1, 1},
  {&__pyx_n_s_find_features, __pyx_k_find_features, sizeof(__pyx_k_find_features), 0, 0, 1, 1},
  {&__pyx_n_s_fit_composition, __pyx_k_fit_composition, sizeof(__pyx_k_fit_composition), 0, 0, 1, 1},
  {&__pyx_n_s_fit_feature_set, __pyx_k_fit_feature_set, sizeof(__pyx_k_fit_feature_set), 0, 0, 1, 1},
  {&__pyx_n_s_fit_single_feature_set, __pyx_k_fit_single_feature_set, sizeof(__pyx_k_fit_single_feature_set), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_ignore_below, __pyx_k_ignore_below, sizeof(__pyx_k_ignore_below), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_interval, __pyx_k_interval, sizeof(__pyx_k_interval), 0, 0, 1, 1},
  {&__pyx_n_s_left_search, __pyx_k_left_search, sizeof(__pyx_k_left_search), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_match_theoretical_isotopic_distr, __pyx_k_match_theoretical_isotopic_distr, sizeof(__pyx_k_match_theoretical_isotopic_distr), 0, 0, 1, 1},
  {&__pyx_n_s_max_missed_peaks, __pyx_k_max_missed_peaks, sizeof(__pyx_k_max_missed_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_minimum_theoretical_abundance, __pyx_k_minimum_theoretical_abundance, sizeof(__pyx_k_minimum_theoretical_abundance), 0, 0, 1, 1},
  {&__pyx_n_s_ms_deisotope__c_feature_map_feat, __pyx_k_ms_deisotope__c_feature_map_feat, sizeof(__pyx_k_ms_deisotope__c_feature_map_feat), 0, 0, 1, 1},
  {&__pyx_n_s_mz, __pyx_k_mz, sizeof(__pyx_k_mz), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_LCMSFeatureProces, __pyx_k_pyx_unpickle_LCMSFeatureProces, sizeof(__pyx_k_pyx_unpickle_LCMSFeatureProces), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_envelope_conforme, __pyx_k_pyx_unpickle_envelope_conforme, sizeof(__pyx_k_pyx_unpickle_envelope_conforme), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_right_search, __pyx_k_right_search, sizeof(__pyx_k_right_search), 0, 0, 1, 1},
  {&__pyx_kp_s_self_indices_self_lengths_cannot, __pyx_k_self_indices_self_lengths_cannot, sizeof(__pyx_k_self_indices_self_lengths_cannot), 0, 0, 1, 0},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_start_time, __pyx_k_start_time, sizeof(__pyx_k_start_time), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_theoretical_distribution, __pyx_k_theoretical_distribution, sizeof(__pyx_k_theoretical_distribution), 0, 0, 1, 1},
  {&__pyx_n_s_threshold_scale, __pyx_k_threshold_scale, sizeof(__pyx_k_threshold_scale), 0, 0, 1, 1},
  {&__pyx_n_s_truncate_after, __pyx_k_truncate_after, sizeof(__pyx_k_truncate_after), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 56, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 83, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(0, 105, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 160, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 944, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")
 */

#line 2 "(tree fragment)"
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_self_indices_self_lengths_cannot); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple_);

#line 2 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.indices,self.lengths cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_self_indices_self_lengths_cannot); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__2);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__2);

#line 4 "(tree fragment)"
  __pyx_tuple__9 = PyTuple_Pack(3, __pyx_int_195672028, __pyx_int_51343670, __pyx_int_167374361); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__9);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__9);

#line 4 "(tree fragment)"
  __pyx_tuple__10 = PyTuple_Pack(3, __pyx_int_177677261, __pyx_int_212946053, __pyx_int_237791963); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__10);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(2, 944, __pyx_L1_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__11);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(2, 950, __pyx_L1_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__12);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "(tree fragment)":1
 * def __pyx_unpickle_envelope_conformer(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_tuple__13 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__13);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__13);

#line 1 "(tree fragment)"
  __pyx_codeobj__14 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__13, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_envelope_conforme, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__14)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __pyx_tuple__15 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__15);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__15);

#line 1 "(tree fragment)"
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__15, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_LCMSFeatureProces, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return 0;

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return -1;

#line 1 "(tree fragment)"
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_51343670 = PyInt_FromLong(51343670L); if (unlikely(!__pyx_int_51343670)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_167374361 = PyInt_FromLong(167374361L); if (unlikely(!__pyx_int_167374361)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_177677261 = PyInt_FromLong(177677261L); if (unlikely(!__pyx_int_177677261)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_195672028 = PyInt_FromLong(195672028L); if (unlikely(!__pyx_int_195672028)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_212946053 = PyInt_FromLong(212946053L); if (unlikely(!__pyx_int_212946053)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_237791963 = PyInt_FromLong(237791963L); if (unlikely(!__pyx_int_237791963)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator._initialize = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *, PyObject *))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator__initialize;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator._create = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *(*)(PyObject *))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator__create;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.has_more = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_has_more;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.compose_next_value = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_compose_next_value;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.advance = (void (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_advance;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.get_next_value = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator *))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24CartesianProductIterator_get_next_value;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CartesianProductIterator, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator = &__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_CartesianProductIterator;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer._create = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *(*)(void))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer__create;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.acquire = (void (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, PyObject *, PyObject *))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_acquire;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.conform = (void (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer *, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_18envelope_conformer_conform;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer) < 0) __PYX_ERR(0, 186, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer) < 0) __PYX_ERR(0, 186, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_envelope_conformer, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer) < 0) __PYX_ERR(0, 186, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer) < 0) __PYX_ERR(0, 186, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer = &__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_envelope_conformer;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.create_theoretical_distribution = (struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_create_theoretical_distribution;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.find_all_features = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_all_features;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.find_features = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_find_features;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.match_theoretical_isotopic_distribution = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_match_theoretical_isotopic_distribution;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.conform_envelopes = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase_conform_envelopes;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._conform_envelopes = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, size_t *))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__conform_envelopes;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._find_thresholded_score = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *, double))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__find_thresholded_score;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._fit_single_feature_set = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, PyObject *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_single_feature_set;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._fit_theoretical_distribution_on_features = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_theoretical_distribution_on_features;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._fit_feature_set = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, double, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_feature_set;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase._fit_composition = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase *, PyObject *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_17feature_processor_24LCMSFeatureProcessorBase__fit_composition;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LCMSFeatureProcessorBase, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase = &__pyx_type_12ms_deisotope_2_c_11feature_map_17feature_processor_LCMSFeatureProcessorBase;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakBase", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase) __PYX_ERR(6, 2, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase)) __PYX_ERR(6, 2, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "FittedPeak", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak)) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSet", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet) __PYX_ERR(6, 29, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet)) __PYX_ERR(6, 29, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSetIndexed", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) __PYX_ERR(6, 83, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed)) __PYX_ERR(6, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_11composition_PyComposition = __Pyx_ImportType(__pyx_t_1, "brainpy._c.composition", "PyComposition", sizeof(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_11composition_PyComposition) __PYX_ERR(7, 123, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition = (struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_11composition_PyComposition->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition)) __PYX_ERR(7, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_distribution"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = __Pyx_ImportType(__pyx_t_1, "brainpy._c.isotopic_distribution", "TheoreticalPeak", sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak) __PYX_ERR(8, 63, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = (struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak)) __PYX_ERR(8, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(2, 770, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(2, 772, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(2, 774, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(2, 776, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(2, 778, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(2, 780, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(2, 782, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(2, 784, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(2, 786, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(2, 788, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_index"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_index", "PeakIndex", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex) __PYX_ERR(9, 5, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex)) __PYX_ERR(9, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "Averagine", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine) __PYX_ERR(10, 12, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine)) __PYX_ERR(10, 12, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "TheoreticalIsotopicPattern", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern) __PYX_ERR(10, 23, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern)) __PYX_ERR(10, 23, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "AveragineCache", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache) __PYX_ERR(10, 79, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache)) __PYX_ERR(10, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.deconvoluter_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "DeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase) __PYX_ERR(11, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase)) __PYX_ERR(11, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "AveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase) __PYX_ERR(11, 57, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase)) __PYX_ERR(11, 57, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "MultiAveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase) __PYX_ERR(11, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase)) __PYX_ERR(11, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "ChargeIterator", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator) __PYX_ERR(11, 91, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator)) __PYX_ERR(11, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitRecord", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord) __PYX_ERR(12, 16, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord)) __PYX_ERR(12, 16, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FitSelectorBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase) __PYX_ERR(12, 38, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase)) __PYX_ERR(12, 38, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase) __PYX_ERR(12, 50, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase)) __PYX_ERR(12, 50, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "LeastSquaresFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter) __PYX_ERR(12, 62, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter)) __PYX_ERR(12, 62, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "GTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter) __PYX_ERR(12, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter)) __PYX_ERR(12, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledGTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter) __PYX_ERR(12, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter)) __PYX_ERR(12, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "MSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter) __PYX_ERR(12, 72, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter)) __PYX_ERR(12, 72, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "PenalizedMSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter) __PYX_ERR(12, 78, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter)) __PYX_ERR(12, 78, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FunctionScorer", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer) __PYX_ERR(12, 84, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer)) __PYX_ERR(12, 84, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "InterferenceDetection", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection) __PYX_ERR(12, 89, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection)) __PYX_ERR(12, 89, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DistinctPatternFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter) __PYX_ERR(12, 95, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter)) __PYX_ERR(12, 95, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledPenalizedMSDeconvFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter) __PYX_ERR(12, 102, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter)) __PYX_ERR(12, 102, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DotProductFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter) __PYX_ERR(12, 112, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter)) __PYX_ERR(12, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.feature_map.lcms_feature"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeatureTreeList", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList) __PYX_ERR(13, 4, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList)) __PYX_ERR(13, 4, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeatureTreeNodeBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase) __PYX_ERR(13, 23, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase)) __PYX_ERR(13, 23, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeatureTreeNode", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode) __PYX_ERR(13, 44, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode)) __PYX_ERR(13, 44, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "FeatureBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase) __PYX_ERR(13, 56, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase)) __PYX_ERR(13, 56, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeature", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature) __PYX_ERR(13, 73, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature)) __PYX_ERR(13, 73, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "EmptyFeature", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature) __PYX_ERR(13, 104, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature)) __PYX_ERR(13, 104, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "FeatureSetIterator", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator) __PYX_ERR(13, 109, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator)) __PYX_ERR(13, 109, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "RunningWeightedAverage", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage) __PYX_ERR(13, 137, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage)) __PYX_ERR(13, 137, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "RunningWeightedAverageNeutralMass", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass) __PYX_ERR(13, 154, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass)) __PYX_ERR(13, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.feature_map.feature_map"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.feature_map", "LCMSFeatureMap", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap) __PYX_ERR(14, 5, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap)) __PYX_ERR(14, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "_Index", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index) __PYX_ERR(15, 5, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index)) __PYX_ERR(15, 5, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "EnvelopePair", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair) __PYX_ERR(15, 14, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair)) __PYX_ERR(15, 14, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "Envelope", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope) __PYX_ERR(15, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope)) __PYX_ERR(15, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak) __PYX_ERR(15, 40, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak)) __PYX_ERR(15, 40, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "IonMobilityDeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak) __PYX_ERR(15, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak)) __PYX_ERR(15, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSolution", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution) __PYX_ERR(15, 71, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution)) __PYX_ERR(15, 71, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet) __PYX_ERR(15, 75, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet)) __PYX_ERR(15, 75, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSetIndexed", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed) __PYX_ERR(15, 104, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed)) __PYX_ERR(15, 104, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "_CPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet) __PYX_ERR(15, 161, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet)) __PYX_ERR(15, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.feature_map.feature_fit"); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.feature_fit", "map_coord", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord) __PYX_ERR(16, 8, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_map_coord)) __PYX_ERR(16, 8, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.feature_fit", "LCMSFeatureSetFit", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit) __PYX_ERR(16, 22, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_fit_LCMSFeatureSetFit)) __PYX_ERR(16, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "nist_mass", (void **)&__pyx_vp_7brainpy_2_c_11composition_nist_mass, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "PROTON", (void **)&__pyx_vp_7brainpy_2_c_11composition_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_ElementTable", (void **)&__pyx_vp_7brainpy_2_c_11composition__ElementTable, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_constants"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "DEFAULT_ISOTOPIC_CONSTANTS_SIZE", (void **)&__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "PROTON", (void **)&__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INFINITY", (void **)&__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.double_vector"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "make_double_vector", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_make_double_vector, "struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *(void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "double_vector_append", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_double_vector_append, "int (struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "free_double_vector", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_free_double_vector, "void (struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "double_vector_to_list", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_double_vector_to_list, "PyObject *(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_distribution"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "_isotopic_variants", (void (**)(void))&__pyx_f_7brainpy_2_c_21isotopic_distribution__isotopic_variants, "PyObject *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "neutral_mass", (void (**)(void))&__pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass, "double (double, int, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initfeature_processor(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initfeature_processor(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_feature_processor(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_feature_processor(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_feature_processor(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'feature_processor' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_feature_processor(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("feature_processor", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_deisotope___c__feature_map__feature_processor) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_deisotope._c.feature_map.feature_processor")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_deisotope._c.feature_map.feature_processor", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":35
 * 
 * cimport numpy as np
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 35 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)

#line 35 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 35 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 35, __pyx_L1_error)

#line 35 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":38
 * 
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 38 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_2 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 38, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":41
 * 
 * 
 * cdef object zeros = np.zeros             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 41 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 41 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 41 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 41 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_XGOTREF(__pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros);

#line 41 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF_SET(__pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_zeros, __pyx_t_3);

#line 41 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 41 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":44
 * 
 * 
 * cdef double INF = float('inf')             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 44 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_4 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_4 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 44, __pyx_L1_error)

#line 44 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_INF = __pyx_t_4;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":260
 * 
 * 
 * cdef size_t MAX_COMBINATIONS = 10000 // 2             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 260 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_v_12ms_deisotope_2_c_11feature_map_17feature_processor_MAX_COMBINATIONS = 0x1388;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":269
 *         public AveragineCache averagine
 * 
 *     cpdef TheoreticalIsotopicPattern create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8,             # <<<<<<<<<<<<<<
 *                                                                      double ignore_below=0.05):
 *         cdef:
 */

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_k__3 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

#line 269 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_k__3 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":373
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,
 *                                                     int charge, double charge_carrier=PROTON, int max_missed_peaks=1,             # <<<<<<<<<<<<<<
 *                                                     double threshold_scale=0.3):
 *         cdef:
 */

#line 373 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_k__4 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":372
 *         return acc / count
 * 
 *     cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,             # <<<<<<<<<<<<<<
 *                                                     int charge, double charge_carrier=PROTON, int max_missed_peaks=1,
 *                                                     double threshold_scale=0.3):
 */

#line 372 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_k__4 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":430
 *     cdef list _fit_theoretical_distribution_on_features(self, double mz, double error_tolerance, int charge,
 *                                                         TheoreticalIsotopicPattern base_tid,
 *                                                         double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                                         double truncate_after=0.8, int max_missed_peaks=1,
 *                                                         double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 430 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_k__5 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":551
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,             # <<<<<<<<<<<<<<
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 *         cdef:
 */

#line 551 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_k__7 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":550
 *         return feature_fits
 * 
 *     cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,             # <<<<<<<<<<<<<<
 *                                 int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
 *                                 int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 550 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_k__7 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":566
 *         return feature_fits
 * 
 *     cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,             # <<<<<<<<<<<<<<
 *                                 double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
 *                                 double threshold_scale=0.3, LCMSFeature feature=None):
 */

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_k__8 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

#line 566 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_k__8 = __pyx_v_12ms_deisotope_2_c_9averagine_PROTON;

  /* "(tree fragment)":1
 * def __pyx_unpickle_envelope_conformer(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_11feature_map_17feature_processor_3__pyx_unpickle_envelope_conformer, NULL, __pyx_n_s_ms_deisotope__c_feature_map_feat); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_envelope_conforme, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_envelope_conformer__set_state(<envelope_conformer> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_envelope_conformer__set_state(envelope_conformer __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.experimental = __pyx_state[0]; __pyx_result.n_missing = __pyx_state[1]; __pyx_result.theoretical = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_11feature_map_17feature_processor_5__pyx_unpickle_LCMSFeatureProcessorBase, NULL, __pyx_n_s_ms_deisotope__c_feature_map_feat); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 11 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_LCMSFeatureProces, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/feature_processor.pyx":1
 * # cython: embedsignature=True             # <<<<<<<<<<<<<<
 * #
 * 
 */

#line 1 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 1 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/feature_map/feature_processor.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /*--- Wrapped vars code ---*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_3);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_m) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_d) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("init ms_deisotope._c.feature_map.feature_processor", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    Py_CLEAR(__pyx_m);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else if (!PyErr_Occurred()) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_ImportError, "init ms_deisotope._c.feature_map.feature_processor");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return (__pyx_m != NULL) ? 0 : -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #elif PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_m;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1 "cython_utility"

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* BufferIndexError */
  static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseDoubleKeywords */
  static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
  static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
  static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyObjectCallNoArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* WriteUnraisableException */
  static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyErrExceptionMatches */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
  static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
  static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* PyDictVersioning */
  #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
  #if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* ArgTypeTest */
  static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* GetItemInt */
  static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* SetItemInt */
  static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetTopmostException */
  #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
  static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
  static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
  static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* pyobject_as_double */
  static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
    float_value = PyNumber_Float(obj);  if ((0)) goto bad;
#else
    PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
    if (likely(nb) && likely(nb->nb_float)) {
        float_value = nb->nb_float(obj);
        if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
            PyErr_Format(PyExc_TypeError,
                "__float__ returned non-float (type %.200s)",
                Py_TYPE(float_value)->tp_name);
            Py_DECREF(float_value);
            goto bad;
        }
    } else if (PyUnicode_CheckExact(obj) || PyBytes_CheckExact(obj)) {
#if PY_MAJOR_VERSION >= 3
        float_value = PyFloat_FromString(obj);
#else
        float_value = PyFloat_FromString(obj, 0);
#endif
    } else {
        PyObject* args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        PyTuple_SET_ITEM(args, 0, obj);
        float_value = PyObject_Call((PyObject*)&PyFloat_Type, args, 0);
        PyTuple_SET_ITEM(args, 0, 0);
        Py_DECREF(args);
    }
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
bad:
    return (double)-1;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* Print */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* PrintOne */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrImport */
  #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
  #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
