/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/brainpy/_c",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/ms_peak_picker/_c",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include"
        ],
        "name": "ms_deisotope._c.similarity_methods",
        "sources": [
            "src/ms_deisotope/_c/similarity_methods.pyx"
        ]
    },
    "module_name": "ms_deisotope._c.similarity_methods"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_deisotope___c__similarity_methods
#define __PYX_HAVE_API__ms_deisotope___c__similarity_methods
/* Early includes */
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "pythread.h"
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "src/ms_deisotope/_c/similarity_methods.pyx",
  "stringsource",
  "__init__.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "peak_set.pxd",
  "peak_index.pxd",
  "composition.pxd",
  "isotopic_distribution.pxd",
  "src/ms_deisotope/_c/averagine.pxd",
  "src/ms_deisotope/_c/deconvoluter_base.pxd",
  "src/ms_deisotope/_c/scoring.pxd",
  "src/ms_deisotope/_c/peak_set.pxd",
};
/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif


/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */

#line 689 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */

#line 690 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */

#line 691 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */

#line 692 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */

#line 696 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */

#line 697 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */

#line 698 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */

#line 699 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */

#line 703 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */

#line 704 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */

#line 713 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */

#line 714 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */

#line 715 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */

#line 717 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */

#line 718 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */

#line 719 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */

#line 721 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */

#line 722 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */

#line 724 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */

#line 725 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */

#line 726 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "brainpy/_c/composition.pxd":3
 * from brainpy.mass_dict import nist_mass as __nist_mass
 * 
 * ctypedef long count_type             # <<<<<<<<<<<<<<
 * 
 * cdef dict nist_mass
 */

#line 3 "brainpy/_c/composition.pxd"
typedef long __pyx_t_7brainpy_2_c_11composition_count_type;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex;
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition;
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet;
struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch;
struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment;
struct __pyx_obj___Pyx_EnumMeta;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */

#line 728 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */

#line 729 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */

#line 730 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_complex_t;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted;

/* "ms_peak_picker/_c/peak_set.pxd":42
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout)
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self)
 */

#line 42 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak {

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":50
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak getitem(self, size_t i)
 */

#line 50 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for {

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  double error_tolerance;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":64
 * 
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */

#line 64 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell {

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  double center_value;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t start;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t end;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":70
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */

#line 70 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list {

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell *index;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  size_t size;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double low;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double high;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":90
 *     cpdef _allocate_index(self, size_t interval_index_size)
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=*, size_t hint_start=*, size_t hint_end=*)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n)
 */

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted {

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_start;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_end;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
};

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for;

/* "ms_peak_picker/_c/peak_index.pxd":11
 *         public PeakSet peaks
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 */

#line 11 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak {

#line 11 "ms_peak_picker/_c/peak_index.pxd"
  int __pyx_n;

#line 11 "ms_peak_picker/_c/peak_index.pxd"
  double tolerance;

#line 11 "ms_peak_picker/_c/peak_index.pxd"
};

/* "ms_peak_picker/_c/peak_index.pxd":14
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self)
 */

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for {

#line 14 "ms_peak_picker/_c/peak_index.pxd"
  int __pyx_n;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
  double error_tolerance;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
};

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Isotope;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Element;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Composition;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop;

/* "brainpy/_c/composition.pxd":10
 * cdef double PROTON = nist_mass["H+"][0][0]
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 */

#line 10 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass {

#line 10 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 10 "brainpy/_c/composition.pxd"
  double charge_carrier;

#line 10 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":11
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 * 
 */

#line 11 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio {

#line 11 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 11 "brainpy/_c/composition.pxd"
  double charge_carrier;

#line 11 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":17
 * # Isotope and IsotopeMap Declarations
 * 
 * cdef struct Isotope:             # <<<<<<<<<<<<<<
 *     double mass
 *     double abundance
 */

#line 17 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Isotope {

#line 17 "brainpy/_c/composition.pxd"
  double mass;

#line 17 "brainpy/_c/composition.pxd"
  double abundance;

#line 17 "brainpy/_c/composition.pxd"
  int neutrons;

#line 17 "brainpy/_c/composition.pxd"
  int neutron_shift;

#line 17 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":23
 *     int neutron_shift
 * 
 * cdef struct IsotopeMap:             # <<<<<<<<<<<<<<
 *     Isotope* bins
 *     size_t size
 */

#line 23 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap {

#line 23 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *bins;

#line 23 "brainpy/_c/composition.pxd"
  size_t size;

#line 23 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":36
 * # Element Declarations
 * 
 * cdef struct Element:             # <<<<<<<<<<<<<<
 *     char* symbol
 *     IsotopeMap* isotopes
 */

#line 36 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Element {

#line 36 "brainpy/_c/composition.pxd"
  char *symbol;

#line 36 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *isotopes;

#line 36 "brainpy/_c/composition.pxd"
  int monoisotopic_isotope_index;

#line 36 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":56
 * # ElementHashTable and ElementHashBucket Declarations
 * 
 * cdef struct ElementHashBucket:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */

#line 56 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket {

#line 56 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;

#line 56 "brainpy/_c/composition.pxd"
  size_t used;

#line 56 "brainpy/_c/composition.pxd"
  size_t size;

#line 56 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":65
 * 
 * 
 * cdef struct ElementHashTable:             # <<<<<<<<<<<<<<
 *     ElementHashBucket* buckets
 *     size_t size
 */

#line 65 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable {

#line 65 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *buckets;

#line 65 "brainpy/_c/composition.pxd"
  size_t size;

#line 65 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":94
 * # Composition Declarations
 * 
 * cdef struct Composition:             # <<<<<<<<<<<<<<
 *     char** elements
 *     count_type* counts
 */

#line 94 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Composition {

#line 94 "brainpy/_c/composition.pxd"
  char **elements;

#line 94 "brainpy/_c/composition.pxd"
  __pyx_t_7brainpy_2_c_11composition_count_type *counts;

#line 94 "brainpy/_c/composition.pxd"
  size_t size;

#line 94 "brainpy/_c/composition.pxd"
  size_t used;

#line 94 "brainpy/_c/composition.pxd"
  int max_variants;

#line 94 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":130
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base)
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _initialize_from_formula(self, str formula)
 */

#line 130 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl {

#line 130 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 130 "brainpy/_c/composition.pxd"
  int free_existing;

#line 130 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":144
 *     cpdef list items(self)
 * 
 *     cpdef pop(self, str key, object default=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef count_type getitem(self, str key)
 */

#line 144 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop {

#line 144 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 144 "brainpy/_c/composition.pxd"
  PyObject *__pyx_default;

#line 144 "brainpy/_c/composition.pxd"
};

#line 144 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector;

/* "brainpy/_c/double_vector.pxd":1
 * cdef struct DoubleVector:             # <<<<<<<<<<<<<<
 *     double* v
 *     size_t used
 */

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector {

#line 1 "brainpy/_c/double_vector.pxd"
  double *v;

#line 1 "brainpy/_c/double_vector.pxd"
  size_t used;

#line 1 "brainpy/_c/double_vector.pxd"
  size_t size;

#line 1 "brainpy/_c/double_vector.pxd"
};

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters;

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants;

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants;

/* "brainpy/_c/isotopic_constants.pxd":6
 * from brainpy._c.double_vector cimport DoubleVector
 * 
 * ctypedef DoubleVector dvec             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 6 "brainpy/_c/isotopic_constants.pxd"
typedef struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector __pyx_t_7brainpy_2_c_18isotopic_constants_dvec;

/* "brainpy/_c/isotopic_constants.pxd":9
 * 
 * 
 * cdef struct PolynomialParameters:             # <<<<<<<<<<<<<<
 *     dvec* elementary_symmetric_polynomial
 *     dvec* power_sum
 */

#line 9 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters {

#line 9 "brainpy/_c/isotopic_constants.pxd"
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *elementary_symmetric_polynomial;

#line 9 "brainpy/_c/isotopic_constants.pxd"
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *power_sum;

#line 9 "brainpy/_c/isotopic_constants.pxd"
};

/* "brainpy/_c/isotopic_constants.pxd":14
 * 
 * 
 * cdef struct PhiConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     Element* element
 */

#line 14 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants {

#line 14 "brainpy/_c/isotopic_constants.pxd"
  int order;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element *element;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *element_coefficients;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *mass_coefficients;

#line 14 "brainpy/_c/isotopic_constants.pxd"
};

/* "brainpy/_c/isotopic_constants.pxd":21
 * 
 * 
 * cdef struct IsotopicConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     PhiConstants** constants
 */

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants {

#line 21 "brainpy/_c/isotopic_constants.pxd"
  int order;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants **constants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  size_t size;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  size_t used;

#line 21 "brainpy/_c/isotopic_constants.pxd"
};

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx;

/* "brainpy/_c/isotopic_distribution.pxd":7
 * 
 * 
 * cdef struct Peak:             # <<<<<<<<<<<<<<
 *     double mz
 *     double intensity
 */

#line 7 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak {

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  double mz;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  double intensity;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":13
 * 
 * 
 * cdef struct PeakList:             # <<<<<<<<<<<<<<
 *     Peak* peaks
 *     size_t used
 */

#line 13 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList {

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *peaks;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  size_t used;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  size_t size;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":35
 * 
 * 
 * cdef struct ElementCache:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */

#line 35 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache {

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  size_t used;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  size_t size;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *source;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":49
 * 
 * 
 * cdef struct IsotopicDistribution:             # <<<<<<<<<<<<<<
 *     Composition* composition
 *     IsotopicConstants* _isotopic_constants
 */

#line 49 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution {

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Composition *composition;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *_isotopic_constants;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  int order;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  double average_mass;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *monoisotopic_peak;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":58
 * 
 * 
 * cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil
 * 
 */

#line 58 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution {

#line 58 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 58 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 58 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":77
 * 
 * cdef int max_variants(Composition* composition, ElementCache* cache) nogil
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)
 */

#line 77 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks {

#line 77 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 77 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 77 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":79
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 79 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants {

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  PyObject *npeaks;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  double charge_carrier;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":82
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=*, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 82 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants {

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  double charge_carrier;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":85
 * 
 * 
 * cdef size_t max_variants_approx(double mass, double lambda_factor=*, size_t maxiter=*, double threshold=*) nogil             # <<<<<<<<<<<<<<
 */

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx {

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  double lambda_factor;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  size_t maxiter;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  double threshold;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
};

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate;

/* "ms_deisotope/_c/averagine.pxd":7
 * cdef double PROTON
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)
 * cdef void slide(double mz, list peaklist)
 */

#line 7 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio {

#line 7 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 7 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 7 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":8
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef void slide(double mz, list peaklist)
 * cdef dict scale_dict(dict data, double factor)
 */

#line 8 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass {

#line 8 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 8 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 8 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":17
 *         public dict base_composition
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 */

#line 17 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale {

#line 17 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 17 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 17 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 17 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":18
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */

#line 18 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster {

#line 18 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 18 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 18 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":19
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 19 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster {

#line 19 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 19 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 19 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":39
 *     cdef bint _eq_inst(self, TheoreticalIsotopicPattern other)
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 */

#line 39 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below {

#line 39 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 39 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 39 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":40
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)
 */

#line 40 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after {

#line 40 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 40 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 40 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":42
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern scale_raw(self, double scale_factor)
 *     cpdef double drop_last_peak(self)
 */

#line 42 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale {

#line 42 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 42 "ms_deisotope/_c/averagine.pxd"
  PyObject *method;

#line 42 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":53
 * 
 *     @cython.final
 *     cdef inline TheoreticalIsotopicPattern _scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 * 
 *     @cython.final
 */

#line 53 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale {

#line 53 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 53 "ms_deisotope/_c/averagine.pxd"
  PyObject *method;

#line 53 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":88
 *     cdef double _make_cache_key(self, double mz)
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */

#line 88 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair {

#line 88 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 88 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 88 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":89
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 89 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster {

#line 89 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 89 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 89 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":92
 * 
 * 
 * cpdef double isotopic_shift(int charge=*)             # <<<<<<<<<<<<<<
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)
 */

#line 92 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift {

#line 92 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 92 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 92 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":94
 * cpdef double isotopic_shift(int charge=*)
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)             # <<<<<<<<<<<<<<
 */

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate {

#line 94 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 94 "ms_deisotope/_c/averagine.pxd"
  double lambda_factor;

#line 94 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 94 "ms_deisotope/_c/averagine.pxd"
};

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs;

/* "ms_deisotope/_c/deconvoluter_base.pxd":42
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance)
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 */

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution {

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":45
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 *     cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound)
 * 
 */

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction {

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":50
 *     cpdef bint _check_fit(self, IsotopicFitRecord fit)
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 */

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak {

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":51
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 */

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak {

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":53
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 * 
 */

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace {

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":54
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace {

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":61
 *         public AveragineCache averagine
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution {

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":65
 *                                                          double ignore_below=*)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges {

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":73
 *         public list averagines
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution {

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":76
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges {

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":83
 * 
 * 
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=*,             # <<<<<<<<<<<<<<
 *                                      object charge_range=*,
 *                                      int left_search_limit=*, int right_search_limit=*,
 */

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs {

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *charge_range;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int left_search_limit;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int right_search_limit;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int recalculate_starting_peak;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int use_quick_charge;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference;

/* "ms_deisotope/_c/scoring.pxd":29
 * 
 *     @staticmethod
 *     cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,             # <<<<<<<<<<<<<<
 *                                    list experimental, object data=*, int missed_peaks=*)
 * 
 */

#line 29 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create {

#line 29 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 29 "ms_deisotope/_c/scoring.pxd"
  PyObject *data;

#line 29 "ms_deisotope/_c/scoring.pxd"
  int missed_peaks;

#line 29 "ms_deisotope/_c/scoring.pxd"
};

/* "ms_deisotope/_c/scoring.pxd":75
 *     cdef public double mass_error_tolerance
 * 
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=*, double minimum_signal_to_noise=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 75 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak {

#line 75 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 75 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 75 "ms_deisotope/_c/scoring.pxd"
  double minimum_signal_to_noise;

#line 75 "ms_deisotope/_c/scoring.pxd"
};

/* "ms_deisotope/_c/scoring.pxd":92
 *     cdef public PeakSet peaklist
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=*, double upper=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference {

#line 92 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 92 "ms_deisotope/_c/scoring.pxd"
  double lower;

#line 92 "ms_deisotope/_c/scoring.pxd"
  double upper;

#line 92 "ms_deisotope/_c/scoring.pxd"
};

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_all_peaks_for;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for;

/* "ms_deisotope/_c/peak_set.pxd":138
 * 
 * 
 * cdef enum CPeaksFlags:             # <<<<<<<<<<<<<<
 *     borrowing = 0
 *     owns_peaks = 1
 */

#line 138 "ms_deisotope/_c/peak_set.pxd"
enum __pyx_t_12ms_deisotope_2_c_8peak_set_CPeaksFlags {

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_borrowing = 0,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_peaks = 1,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_index = 2,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_peaks_and_index = 3

#line 138 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":87
 *     cpdef DeconvolutedPeakSet copy(self)
 * 
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 */

#line 87 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak {

#line 87 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 87 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 87 "ms_deisotope/_c/peak_set.pxd"
  int use_mz;

#line 87 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":89
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)
 */

#line 89 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak {

#line 89 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 89 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 89 "ms_deisotope/_c/peak_set.pxd"
  int use_mz;

#line 89 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":91
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef DeconvolutedPeak _get_nearest_peak(self, double neutral_mass, double* errout)
 */

#line 91 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for {

#line 91 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 91 "ms_deisotope/_c/peak_set.pxd"
  double tolerance;

#line 91 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":118
 * cdef size_t INTERVAL_INDEX_SIZE
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */

#line 118 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell {

#line 118 "ms_deisotope/_c/peak_set.pxd"
  double center_value;

#line 118 "ms_deisotope/_c/peak_set.pxd"
  size_t start;

#line 118 "ms_deisotope/_c/peak_set.pxd"
  size_t end;

#line 118 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":124
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */

#line 124 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list {

#line 124 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell *index;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  size_t size;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  double low;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  double high;

#line 124 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":131
 * 
 * 
 * cdef struct deconvoluted_peak_t:             # <<<<<<<<<<<<<<
 *     double neutral_mass
 *     double intensity
 */

#line 131 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t {

#line 131 "ms_deisotope/_c/peak_set.pxd"
  double neutral_mass;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  double intensity;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  int charge;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  unsigned int index;

#line 131 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":144
 *     owns_peaks_and_index = 3
 * 
 * cdef struct deconvoluted_peak_set_t:             # <<<<<<<<<<<<<<
 *     deconvoluted_peak_t* peaks
 *     double* mass_index
 */

#line 144 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t {

#line 144 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *peaks;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  double *mass_index;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  size_t size;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  enum __pyx_t_12ms_deisotope_2_c_8peak_set_CPeaksFlags flags;

#line 144 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":154
 * cdef int free_deconvoluted_peak_set_t(deconvoluted_peak_set_t* destination) nogil
 * 
 * cdef deconvoluted_peak_set_t deconvoluted_peak_set_all_peaks_for(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * cdef deconvoluted_peak_t* deconvoluted_peak_set_has_peak(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil
 * 
 */

#line 154 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_all_peaks_for {

#line 154 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 154 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 154 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":155
 * 
 * cdef deconvoluted_peak_set_t deconvoluted_peak_set_all_peaks_for(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil
 * cdef deconvoluted_peak_t* deconvoluted_peak_set_has_peak(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cdef int deconvoluted_peak_eq(deconvoluted_peak_t* self, deconvoluted_peak_t* other) nogil
 */

#line 155 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_has_peak {

#line 155 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 155 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 155 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":167
 *     cdef deconvoluted_peak_t* getitem(self, size_t i) nogil
 * 
 *     cdef deconvoluted_peak_t* _has_peak(self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil
 * 
 */

#line 167 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak {

#line 167 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 167 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 167 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":168
 * 
 *     cdef deconvoluted_peak_t* _has_peak(self, double neutral_mass, double error_tolerance=*) nogil
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)
 */

#line 168 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for {

#line 168 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 168 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 168 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":170
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double error_tolerance=*)
 */

#line 170 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak {

#line 170 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 170 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 170 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":171
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double error_tolerance=*)             # <<<<<<<<<<<<<<
 */

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for {

#line 171 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 171 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 171 "ms_deisotope/_c/peak_set.pxd"
};

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__create;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_calculate_dot_product;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_align;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_deconvoluted_peak_sets;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolved_dot_product;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_deconvoluted;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks;

/* "ms_deisotope/_c/similarity_methods.pyx":32
 * 
 * 
 * cpdef enum SimilarityMetrics:             # <<<<<<<<<<<<<<
 *     dot_product
 *     normalized_dot_product
 */

#line 32 "ms_deisotope/_c/similarity_methods.pyx"
enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics {

#line 32 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_e_12ms_deisotope_2_c_18similarity_methods_dot_product,

#line 32 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_e_12ms_deisotope_2_c_18similarity_methods_normalized_dot_product

#line 32 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":37
 * 
 * 
 * cpdef enum PeakType:             # <<<<<<<<<<<<<<
 *     Fitted
 *     Deconvoluted
 */

#line 37 "ms_deisotope/_c/similarity_methods.pyx"
enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType {

#line 37 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_e_12ms_deisotope_2_c_18similarity_methods_Fitted,

#line 37 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_e_12ms_deisotope_2_c_18similarity_methods_Deconvoluted

#line 37 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":81
 * 
 *     @staticmethod
 *     cdef PeakMatch _create(PeakBase peak_a, PeakBase peak_b, double score=0.0):             # <<<<<<<<<<<<<<
 *         cdef PeakMatch self = PeakMatch.__new__(PeakMatch)
 *         self.peak_a = peak_a
 */

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__create {

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  double score;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":329
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cdef double calculate_dot_product(cnp.ndarray[double, ndim=1] bin_a, cnp.ndarray[double, ndim=1] bin_b, bint normalize=True):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double n_a, n_b, n_ab, z, a, b
 */

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_calculate_dot_product {

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  int normalize;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":449
 *                 "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 * 
 *     cpdef align(self, double error_tolerance=2e-5, double shift=0.0):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list pairs
 */

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_align {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  double shift;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":480
 * 
 * 
 * cdef list convolve_peak_sets(PeakSet peak_set_a, PeakSet peak_set_b, double error_tolerance=2e-5, double shift=0.0):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n, j, m
 */

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets {

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  double shift;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":504
 * 
 * 
 * cdef list convolve_deconvoluted_peak_sets(DeconvolutedPeakSet peak_set_a, DeconvolutedPeakSet peak_set_b,             # <<<<<<<<<<<<<<
 *                                           double error_tolerance=2e-5, double shift=0.0):
 *     cdef:
 */

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_deconvoluted_peak_sets {

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  double shift;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":528
 * 
 * 
 * cpdef double convolved_dot_product(list peak_pairs, PeakType peak_type=PeakType.Fitted, double norm_a=1.0,             # <<<<<<<<<<<<<<
 *                                    double norm_b=1.0, bint sqrt_transform=False):
 *     cdef:
 */

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolved_dot_product {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType peak_type;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  double norm_a;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  double norm_b;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  int sqrt_transform;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":572
 * 
 * 
 * cdef double convolve_peak_sets2_fitted(PeakSet peak_set_a, PeakSet peak_set_b, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n, j, m
 */

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted {

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":593
 * 
 * 
 * cdef double convolve_peak_sets2_deconvoluted(DeconvolutedPeakSet peak_set_a, DeconvolutedPeakSet peak_set_b,             # <<<<<<<<<<<<<<
 *                                              double error_tolerance=2e-5):
 *     cdef:
 */

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_deconvoluted {

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":127
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cpdef double peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0) except -1:             # <<<<<<<<<<<<<<
 *     """Computes the normalized dot product, also called cosine similarity between
 *     two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).
 */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":258
 * 
 * @cython.boundscheck(False)
 * cpdef tuple bin_peaks(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double hi, scaler
 */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":614
 * 
 * 
 * cpdef double ppm_peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double ab, aa, bb
 */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_deisotope/_c/similarity_methods.pyx":635
 * 
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,             # <<<<<<<<<<<<<<
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double shift;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int normalize;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double shift;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int normalize;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double shift;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int normalize;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_n;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double error_tolerance;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double shift;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int normalize;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
};

/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtab;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double mz;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double intensity;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double area;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double signal_to_noise;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double full_width_at_half_max;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double left_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double right_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long peak_count;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long index;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtab;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *peaks;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int indexed;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  double *mz_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *interval_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

#line 5 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex {

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyObject_HEAD

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtab;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyArrayObject *mz_array;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyArrayObject *intensity_array;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaks;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
};

#line 5 "ms_peak_picker/_c/peak_index.pxd"


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

#line 123 "brainpy/_c/composition.pxd"
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition {

#line 123 "brainpy/_c/composition.pxd"
  PyObject_HEAD

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtab;

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Composition *impl;

#line 123 "brainpy/_c/composition.pxd"
  double cached_mass;

#line 123 "brainpy/_c/composition.pxd"
  int _clean;

#line 123 "brainpy/_c/composition.pxd"
};

#line 123 "brainpy/_c/composition.pxd"


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 63 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  PyObject_HEAD

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtab;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  double mz;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  double intensity;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
};

#line 63 "brainpy/_c/isotopic_distribution.pxd"


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

#line 12 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine {

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtab;

#line 12 "ms_deisotope/_c/averagine.pxd"
  double base_mass;

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject *base_composition;

#line 12 "ms_deisotope/_c/averagine.pxd"
};

#line 12 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

#line 23 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtab;

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *peaklist;

#line 23 "ms_deisotope/_c/averagine.pxd"
  double origin;

#line 23 "ms_deisotope/_c/averagine.pxd"
  double offset;

#line 23 "ms_deisotope/_c/averagine.pxd"
};

#line 23 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

#line 79 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache {

#line 79 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtab;

#line 79 "ms_deisotope/_c/averagine.pxd"
  PyObject *backend;

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *averagine;

#line 79 "ms_deisotope/_c/averagine.pxd"
  double cache_truncation;

#line 79 "ms_deisotope/_c/averagine.pxd"
  int enabled;

#line 79 "ms_deisotope/_c/averagine.pxd"
};

#line 79 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject_HEAD

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtab;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int use_subtraction;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *scale_method;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int merge_isobaric_peaks;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  double minimum_intensity;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *scorer;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int verbose;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *_slice_cache;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *averagine;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *averagines;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject_HEAD

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtab;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int lower;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int upper;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int sign;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int *values;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t size;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t index;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

#line 16 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtab;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *seed_peak;

#line 16 "ms_deisotope/_c/scoring.pxd"
  double score;

#line 16 "ms_deisotope/_c/scoring.pxd"
  int charge;

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject *experimental;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *theoretical;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *monoisotopic_peak;

#line 16 "ms_deisotope/_c/scoring.pxd"
  int missed_peaks;

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject *data;

#line 16 "ms_deisotope/_c/scoring.pxd"
  Py_hash_t _hash;

#line 16 "ms_deisotope/_c/scoring.pxd"
};

#line 16 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

#line 38 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase {

#line 38 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtab;

#line 38 "ms_deisotope/_c/scoring.pxd"
  double minimum_score;

#line 38 "ms_deisotope/_c/scoring.pxd"
};

#line 38 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

#line 50 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {

#line 50 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtab;

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *select;

#line 50 "ms_deisotope/_c/scoring.pxd"
};

#line 50 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 62 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {

#line 62 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 62 "ms_deisotope/_c/scoring.pxd"
};

#line 62 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 66 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter {

#line 66 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 66 "ms_deisotope/_c/scoring.pxd"
};

#line 66 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 69 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {

#line 69 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 69 "ms_deisotope/_c/scoring.pxd"
};

#line 69 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

#line 72 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter {

#line 72 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 72 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 72 "ms_deisotope/_c/scoring.pxd"
};

#line 72 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

#line 78 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {

#line 78 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 78 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 78 "ms_deisotope/_c/scoring.pxd"
  double penalty_factor;

#line 78 "ms_deisotope/_c/scoring.pxd"
};

#line 78 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

#line 84 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer {

#line 84 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 84 "ms_deisotope/_c/scoring.pxd"
  PyObject *function;

#line 84 "ms_deisotope/_c/scoring.pxd"
};

#line 84 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

#line 89 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection {

#line 89 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 89 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtab;

#line 89 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;

#line 89 "ms_deisotope/_c/scoring.pxd"
};

#line 89 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

#line 95 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *interference_detector;

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *g_test_scaled;

#line 95 "ms_deisotope/_c/scoring.pxd"
  double peak_count_scale;

#line 95 "ms_deisotope/_c/scoring.pxd"
  double domain_scale;

#line 95 "ms_deisotope/_c/scoring.pxd"
};

#line 95 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

#line 102 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 102 "ms_deisotope/_c/scoring.pxd"
  double scale_factor;

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *scorer;

#line 102 "ms_deisotope/_c/scoring.pxd"
};

#line 102 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

#line 112 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter {

#line 112 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 112 "ms_deisotope/_c/scoring.pxd"
};

#line 112 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */

#line 5 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index {

#line 5 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 5 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtab;

#line 5 "ms_deisotope/_c/peak_set.pxd"
  size_t neutral_mass;

#line 5 "ms_deisotope/_c/peak_set.pxd"
  size_t mz;

#line 5 "ms_deisotope/_c/peak_set.pxd"
};

#line 5 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 14 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair {

#line 14 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 14 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtab;

#line 14 "ms_deisotope/_c/peak_set.pxd"
  double mz;

#line 14 "ms_deisotope/_c/peak_set.pxd"
  double intensity;

#line 14 "ms_deisotope/_c/peak_set.pxd"
};

#line 14 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */

#line 25 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope {

#line 25 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtab;

#line 25 "ms_deisotope/_c/peak_set.pxd"
  PyObject *pairs;

#line 25 "ms_deisotope/_c/peak_set.pxd"
};

#line 25 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */

#line 40 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double neutral_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double signal_to_noise;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *_index;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double full_width_at_half_max;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int charge;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double a_to_a2_ratio;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double most_abundant_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double average_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double score;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *envelope;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *fit;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int chosen_for_msms;

#line 40 "ms_deisotope/_c/peak_set.pxd"
};

#line 40 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class IonMobilityDeconvolutedPeak(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double drift_time
 */

#line 66 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak {

#line 66 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 66 "ms_deisotope/_c/peak_set.pxd"
  double drift_time;

#line 66 "ms_deisotope/_c/peak_set.pxd"
};

#line 66 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":71
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */

#line 71 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {

#line 71 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 71 "ms_deisotope/_c/peak_set.pxd"
  PyObject *solution;

#line 71 "ms_deisotope/_c/peak_set.pxd"
};

#line 71 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":75
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 75 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtab;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *peaks;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *_mz_ordered;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  int indexed;

#line 75 "ms_deisotope/_c/peak_set.pxd"
};

#line 75 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":104
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */

#line 104 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  double *neutral_mass_array;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  double *mz_array;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list *interval_index;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  size_t _size;

#line 104 "ms_deisotope/_c/peak_set.pxd"
};

#line 104 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":161
 * 
 * 
 * cdef class _CPeakSet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         deconvoluted_peak_set_t* ptr
 */

#line 161 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet {

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet *__pyx_vtab;

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t *ptr;

#line 161 "ms_deisotope/_c/peak_set.pxd"
};

#line 161 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/similarity_methods.pyx":44
 * @cython.freelist(10000)
 * @cython.final
 * cdef class PeakMatch(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public PeakBase peak_a
 */

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch {

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject_HEAD

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_vtab;

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *peak_a;

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *peak_b;

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
  double score;

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 44 "ms_deisotope/_c/similarity_methods.pyx"


/* "ms_deisotope/_c/similarity_methods.pyx":353
 * 
 * 
 * cdef class SpectrumAlignment(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peak_pairs
 */

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment {

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject_HEAD

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_vtab;

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *peak_pairs;

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  double shift;

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  double score;

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *peak_set_a;

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *peak_set_b;

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType peak_type;

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  double norm_aa_cosine;

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  double norm_bb_cosine;

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  int normalize;

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  int sqrt_transform;

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 353 "ms_deisotope/_c/similarity_methods.pyx"


/* "EnumBase":15
 * 
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):             # <<<<<<<<<<<<<<
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)
 */

#line 15 "EnumBase"
struct __pyx_obj___Pyx_EnumMeta {

#line 15 "EnumBase"
  PyHeapTypeObject __pyx_base;

#line 15 "EnumBase"
};

#line 15 "EnumBase"


#line 15 "EnumBase"


/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"


#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"


#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_create)(double, double, double, double, double, double, long, long, double);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak;

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"


#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_create)(PyObject *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*reindex)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_get_nearest_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_between_bounds)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double, size_t *, size_t *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getitem)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"


#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*_allocate_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, size_t, int __pyx_skip_dispatch);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak_hinted)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

#line 5 "ms_peak_picker/_c/peak_index.pxd"


#line 5 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex {

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak *__pyx_optional_args);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, double);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for *__pyx_optional_args);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
};

#line 5 "ms_peak_picker/_c/peak_index.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex;

#line 5 "ms_peak_picker/_c/peak_index.pxd"


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

#line 123 "brainpy/_c/composition.pxd"


#line 123 "brainpy/_c/composition.pxd"
struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition {

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*_create)(struct __pyx_t_7brainpy_2_c_11composition_Composition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*_set_impl)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args);

#line 123 "brainpy/_c/composition.pxd"
  void (*_initialize_from_formula)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);

#line 123 "brainpy/_c/composition.pxd"
  double (*mass)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  int (*__pyx___equality_pycomposition)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  int (*__pyx___equality_dict)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*copy)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*update)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*keys)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*values)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*items)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*pop)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop *__pyx_optional_args);

#line 123 "brainpy/_c/composition.pxd"
  __pyx_t_7brainpy_2_c_11composition_count_type (*getitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);

#line 123 "brainpy/_c/composition.pxd"
  void (*setitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);

#line 123 "brainpy/_c/composition.pxd"
  void (*increment)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);

#line 123 "brainpy/_c/composition.pxd"
  void (*add_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*subtract_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*scale_by)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, long);

#line 123 "brainpy/_c/composition.pxd"
};

#line 123 "brainpy/_c/composition.pxd"
static struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition;

#line 123 "brainpy/_c/composition.pxd"


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 63 "brainpy/_c/isotopic_distribution.pxd"


#line 63 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*_create)(double, double, int);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  int (*_eq)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*clone)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
};

#line 63 "brainpy/_c/isotopic_distribution.pxd"
static struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;

#line 63 "brainpy/_c/isotopic_distribution.pxd"


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

#line 12 "ms_deisotope/_c/averagine.pxd"


#line 12 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine {

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
};

#line 12 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine;

#line 12 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

#line 23 "ms_deisotope/_c/averagine.pxd"


#line 23 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_create)(PyObject *, double, double);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  int (*_eq_inst)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*ignore_below)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*truncate_after)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale_raw)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*drop_last_peak)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*total)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*normalize)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*get)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, Py_ssize_t);

#line 23 "ms_deisotope/_c/averagine.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*get_monoisotopic_mz)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*get_processed_peaks)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  size_t (*basepeak_index)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_drop_last)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
};

#line 23 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;

#line 23 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

#line 79 "ms_deisotope/_c/averagine.pxd"


#line 79 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache {

#line 79 "ms_deisotope/_c/averagine.pxd"
  double (*_make_cache_key)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double);

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*has_mz_charge_pair)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair *__pyx_optional_args);

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster *__pyx_optional_args);

#line 79 "ms_deisotope/_c/averagine.pxd"
};

#line 79 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache;

#line 79 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*between)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*match_theoretical_isotopic_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*scale_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_evaluate_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*subtraction)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*fit_incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_check_fit)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_find_next_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_find_previous_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_find_next_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_find_previous_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*_create)(int, int);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_tuple)(PyObject *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_quickcharge)(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*sequence_from_quickcharge)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*make_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  void (*release_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  void (*set_bounds)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int, int);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

#line 16 "ms_deisotope/_c/scoring.pxd"


#line 16 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_create)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create *__pyx_optional_args);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_lt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_gt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
};

#line 16 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;

#line 16 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

#line 38 "ms_deisotope/_c/scoring.pxd"


#line 38 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase {

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*best)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_best_from_set)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, double, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
};

#line 38 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase;

#line 38 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

#line 50 "ms_deisotope/_c/scoring.pxd"


#line 50 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {

#line 50 "ms_deisotope/_c/scoring.pxd"
  double (*_evaluate)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, double, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *(*_configure)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
};

#line 50 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;

#line 50 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 62 "ms_deisotope/_c/scoring.pxd"


#line 62 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {

#line 62 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 62 "ms_deisotope/_c/scoring.pxd"
};

#line 62 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;

#line 62 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 66 "ms_deisotope/_c/scoring.pxd"


#line 66 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter {

#line 66 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 66 "ms_deisotope/_c/scoring.pxd"
};

#line 66 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter;

#line 66 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 69 "ms_deisotope/_c/scoring.pxd"


#line 69 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {

#line 69 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 69 "ms_deisotope/_c/scoring.pxd"
};

#line 69 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;

#line 69 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

#line 72 "ms_deisotope/_c/scoring.pxd"


#line 72 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter {

#line 72 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 72 "ms_deisotope/_c/scoring.pxd"
  double (*score_peak)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak *__pyx_optional_args);

#line 72 "ms_deisotope/_c/scoring.pxd"
};

#line 72 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter;

#line 72 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

#line 78 "ms_deisotope/_c/scoring.pxd"


#line 78 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {

#line 78 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 78 "ms_deisotope/_c/scoring.pxd"
};

#line 78 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;

#line 78 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

#line 84 "ms_deisotope/_c/scoring.pxd"


#line 84 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer {

#line 84 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 84 "ms_deisotope/_c/scoring.pxd"
};

#line 84 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer;

#line 84 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

#line 89 "ms_deisotope/_c/scoring.pxd"


#line 89 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection {

#line 89 "ms_deisotope/_c/scoring.pxd"
  double (*detect_interference)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference *__pyx_optional_args);

#line 89 "ms_deisotope/_c/scoring.pxd"
};

#line 89 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection;

#line 89 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

#line 95 "ms_deisotope/_c/scoring.pxd"


#line 95 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 95 "ms_deisotope/_c/scoring.pxd"
};

#line 95 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;

#line 95 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

#line 102 "ms_deisotope/_c/scoring.pxd"


#line 102 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 102 "ms_deisotope/_c/scoring.pxd"
  double (*_calculate_scale_factor)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 102 "ms_deisotope/_c/scoring.pxd"
  void (*scale_fitted_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);

#line 102 "ms_deisotope/_c/scoring.pxd"
  void (*scale_theoretical_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);

#line 102 "ms_deisotope/_c/scoring.pxd"
};

#line 102 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;

#line 102 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

#line 112 "ms_deisotope/_c/scoring.pxd"


#line 112 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter {

#line 112 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 112 "ms_deisotope/_c/scoring.pxd"
};

#line 112 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter;

#line 112 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */

#line 5 "ms_deisotope/_c/peak_set.pxd"


#line 5 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index {

#line 5 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *(*_create)(size_t, size_t);

#line 5 "ms_deisotope/_c/peak_set.pxd"
};

#line 5 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index;

#line 5 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 14 "ms_deisotope/_c/peak_set.pxd"


#line 14 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair {

#line 14 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, int __pyx_skip_dispatch);

#line 14 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*_create)(double, double);

#line 14 "ms_deisotope/_c/peak_set.pxd"
};

#line 14 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair;

#line 14 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */

#line 25 "ms_deisotope/_c/peak_set.pxd"


#line 25 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope {

#line 25 "ms_deisotope/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, size_t);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*_create)(PyObject *);

#line 25 "ms_deisotope/_c/peak_set.pxd"
};

#line 25 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope;

#line 25 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */

#line 40 "ms_deisotope/_c/peak_set.pxd"


#line 40 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, int __pyx_skip_dispatch);

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_create_simple)(double, double, int, double, double, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);

#line 40 "ms_deisotope/_c/peak_set.pxd"
};

#line 40 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;

#line 40 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class IonMobilityDeconvolutedPeak(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double drift_time
 */

#line 66 "ms_deisotope/_c/peak_set.pxd"


#line 66 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak {

#line 66 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 66 "ms_deisotope/_c/peak_set.pxd"
};

#line 66 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak;

#line 66 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":71
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */

#line 71 "ms_deisotope/_c/peak_set.pxd"


#line 71 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {

#line 71 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 71 "ms_deisotope/_c/peak_set.pxd"
};

#line 71 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;

#line 71 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":75
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 75 "ms_deisotope/_c/peak_set.pxd"


#line 75 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*reindex)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*copy)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_get_nearest_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, double *);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*getslice)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t, size_t);

#line 75 "ms_deisotope/_c/peak_set.pxd"
};

#line 75 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;

#line 75 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":104
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */

#line 104 "ms_deisotope/_c/peak_set.pxd"


#line 104 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  void (*_build_index_arrays)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *);

#line 104 "ms_deisotope/_c/peak_set.pxd"
  int (*_interval_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *, double, double, size_t *, size_t *);

#line 104 "ms_deisotope/_c/peak_set.pxd"
};

#line 104 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;

#line 104 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":161
 * 
 * 
 * cdef class _CPeakSet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         deconvoluted_peak_set_t* ptr
 */

#line 161 "ms_deisotope/_c/peak_set.pxd"


#line 161 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet {

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, size_t);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t (*_all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
};

#line 161 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet;

#line 161 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/similarity_methods.pyx":44
 * @cython.freelist(10000)
 * @cython.final
 * cdef class PeakMatch(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public PeakBase peak_a
 */

#line 44 "ms_deisotope/_c/similarity_methods.pyx"


#line 44 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_PeakMatch {

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
  int (*_lt)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, int __pyx_skip_dispatch);

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
  int (*_gt)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, int __pyx_skip_dispatch);

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, int __pyx_skip_dispatch);

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *(*_create)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__create *__pyx_optional_args);

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_vtabptr_12ms_deisotope_2_c_18similarity_methods_PeakMatch;

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__lt(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, int __pyx_skip_dispatch);

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__gt(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, int __pyx_skip_dispatch);

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__eq(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, int __pyx_skip_dispatch);

#line 44 "ms_deisotope/_c/similarity_methods.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__create(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__create *__pyx_optional_args);

#line 44 "ms_deisotope/_c/similarity_methods.pyx"


/* "ms_deisotope/_c/similarity_methods.pyx":353
 * 
 * 
 * cdef class SpectrumAlignment(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peak_pairs
 */

#line 353 "ms_deisotope/_c/similarity_methods.pyx"


#line 353 "ms_deisotope/_c/similarity_methods.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment {

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *(*calculate_normalization)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *, int __pyx_skip_dispatch);

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *(*_determine_peak_type)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *, int __pyx_skip_dispatch);

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *(*align)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_align *__pyx_optional_args);

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *(*_calculate_score)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *, int __pyx_skip_dispatch);

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
};

#line 353 "ms_deisotope/_c/similarity_methods.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_vtabptr_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* Profile.proto */
#ifndef CYTHON_PROFILE
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
  #define CYTHON_PROFILE 0
#else
  #define CYTHON_PROFILE 1
#endif
#endif
#ifndef CYTHON_TRACE_NOGIL
  #define CYTHON_TRACE_NOGIL 0
#else
  #if CYTHON_TRACE_NOGIL && !defined(CYTHON_TRACE)
    #define CYTHON_TRACE 1
  #endif
#endif
#ifndef CYTHON_TRACE
  #define CYTHON_TRACE 0
#endif
#if CYTHON_TRACE
  #undef CYTHON_PROFILE_REUSE_FRAME
#endif
#ifndef CYTHON_PROFILE_REUSE_FRAME
  #define CYTHON_PROFILE_REUSE_FRAME 0
#endif
#if CYTHON_PROFILE || CYTHON_TRACE
  #include "compile.h"
  #include "frameobject.h"
  #include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #if CYTHON_PROFILE_REUSE_FRAME
    #define CYTHON_FRAME_MODIFIER static
    #define CYTHON_FRAME_DEL(frame)
  #else
    #define CYTHON_FRAME_MODIFIER
    #define CYTHON_FRAME_DEL(frame) Py_CLEAR(frame)
  #endif
  #define __Pyx_TraceDeclarations\
      static PyCodeObject *__pyx_frame_code = NULL;\
      CYTHON_FRAME_MODIFIER PyFrameObject *__pyx_frame = NULL;\
      int __Pyx_use_tracing = 0;
  #define __Pyx_TraceFrameInit(codeobj)\
      if (codeobj) __pyx_frame_code = (PyCodeObject*) codeobj;
#if PY_VERSION_HEX >= 0x030b00a2
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     (unlikely((tstate)->cframe->use_tracing) &&\
         (!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #define __Pyx_EnterTracing(tstate) PyThreadState_EnterTracing(tstate)
  #define __Pyx_LeaveTracing(tstate) PyThreadState_LeaveTracing(tstate)
#elif PY_VERSION_HEX >= 0x030a00b1
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     (unlikely((tstate)->cframe->use_tracing) &&\
         (!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #define __Pyx_EnterTracing(tstate)\
      do { tstate->tracing++; tstate->cframe->use_tracing = 0; } while (0)
  #define __Pyx_LeaveTracing(tstate)\
      do {\
          tstate->tracing--;\
          tstate->cframe->use_tracing = ((CYTHON_TRACE && tstate->c_tracefunc != NULL)\
                                 || tstate->c_profilefunc != NULL);\
      } while (0)
#else
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     (unlikely((tstate)->use_tracing) &&\
         (!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #define __Pyx_EnterTracing(tstate)\
      do { tstate->tracing++; tstate->use_tracing = 0; } while (0)
  #define __Pyx_LeaveTracing(tstate)\
      do {\
          tstate->tracing--;\
          tstate->use_tracing = ((CYTHON_TRACE && tstate->c_tracefunc != NULL)\
                                         || tstate->c_profilefunc != NULL);\
      } while (0)
#endif
  #ifdef WITH_THREAD
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)\
  if (nogil) {\
      if (CYTHON_TRACE_NOGIL) {\
          PyThreadState *tstate;\
          PyGILState_STATE state = PyGILState_Ensure();\
          tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 1, 1)) {\
              __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          }\
          PyGILState_Release(state);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  } else {\
      PyThreadState* tstate = PyThreadState_GET();\
      if (__Pyx_IsTracing(tstate, 1, 1)) {\
          __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  }
  #else
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)\
  {   PyThreadState* tstate = PyThreadState_GET();\
      if (__Pyx_IsTracing(tstate, 1, 1)) {\
          __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  }
  #endif
  #define __Pyx_TraceException()\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (__Pyx_IsTracing(tstate, 0, 1)) {\
          __Pyx_EnterTracing(tstate);\
          PyObject *exc_info = __Pyx_GetExceptionTuple(tstate);\
          if (exc_info) {\
              if (CYTHON_TRACE && tstate->c_tracefunc)\
                  tstate->c_tracefunc(\
                      tstate->c_traceobj, __pyx_frame, PyTrace_EXCEPTION, exc_info);\
              tstate->c_profilefunc(\
                  tstate->c_profileobj, __pyx_frame, PyTrace_EXCEPTION, exc_info);\
              Py_DECREF(exc_info);\
          }\
          __Pyx_LeaveTracing(tstate);\
      }\
  }
  static void __Pyx_call_return_trace_func(PyThreadState *tstate, PyFrameObject *frame, PyObject *result) {
      PyObject *type, *value, *traceback;
      __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
      __Pyx_EnterTracing(tstate);
      if (CYTHON_TRACE && tstate->c_tracefunc)
          tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_RETURN, result);
      if (tstate->c_profilefunc)
          tstate->c_profilefunc(tstate->c_profileobj, frame, PyTrace_RETURN, result);
      CYTHON_FRAME_DEL(frame);
      __Pyx_LeaveTracing(tstate);
      __Pyx_ErrRestoreInState(tstate, type, value, traceback);
  }
  #ifdef WITH_THREAD
  #define __Pyx_TraceReturn(result, nogil)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (__Pyx_IsTracing(tstate, 0, 0)) {\
                  __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
              }\
              PyGILState_Release(state);\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 0, 0)) {\
              __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
          }\
      }\
  }
  #else
  #define __Pyx_TraceReturn(result, nogil)\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (__Pyx_IsTracing(tstate, 0, 0)) {\
          __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
      }\
  }
  #endif
  static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno);
  static int __Pyx_TraceSetupAndCall(PyCodeObject** code, PyFrameObject** frame, PyThreadState* tstate, const char *funcname, const char *srcfile, int firstlineno);
#else
  #define __Pyx_TraceDeclarations
  #define __Pyx_TraceFrameInit(codeobj)
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)   if ((1)); else goto_error;
  #define __Pyx_TraceException()
  #define __Pyx_TraceReturn(result, nogil)
#endif
#if CYTHON_TRACE
  static int __Pyx_call_line_trace_func(PyThreadState *tstate, PyFrameObject *frame, int lineno) {
      int ret;
      PyObject *type, *value, *traceback;
      __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
      __Pyx_PyFrame_SetLineNumber(frame, lineno);
      __Pyx_EnterTracing(tstate);
      ret = tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_LINE, NULL);
      __Pyx_LeaveTracing(tstate);
      if (likely(!ret)) {
          __Pyx_ErrRestoreInState(tstate, type, value, traceback);
      } else {
          Py_XDECREF(type);
          Py_XDECREF(value);
          Py_XDECREF(traceback);
      }
      return ret;
  }
  #ifdef WITH_THREAD
  #define __Pyx_TraceLine(lineno, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              int ret = 0;\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (__Pyx_IsTracing(tstate, 0, 0) && tstate->c_tracefunc && __pyx_frame->f_trace) {\
                  ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
              }\
              PyGILState_Release(state);\
              if (unlikely(ret)) goto_error;\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 0, 0) && tstate->c_tracefunc && __pyx_frame->f_trace) {\
              int ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
              if (unlikely(ret)) goto_error;\
          }\
      }\
  }
  #else
  #define __Pyx_TraceLine(lineno, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (__Pyx_IsTracing(tstate, 0, 0) && tstate->c_tracefunc && __pyx_frame->f_trace) {\
          int ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
          if (unlikely(ret)) goto_error;\
      }\
  }
  #endif
#else
  #define __Pyx_TraceLine(lineno, nogil, goto_error)   if ((1)); else goto_error;
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* FusedFunction.proto */
typedef struct {
    __pyx_CyFunctionObject func;
    PyObject *__signatures__;
    PyObject *type;
    PyObject *self;
} __pyx_FusedFunctionObject;
static PyObject *__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                                         PyObject *qualname, PyObject *closure,
                                         PyObject *module, PyObject *globals,
                                         PyObject *code);
static int __pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self);
static PyTypeObject *__pyx_FusedFunctionType = NULL;
static int __pyx_FusedFunction_init(void);
#define __Pyx_FusedFunction_USED

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* IntPow.proto */
static CYTHON_INLINE long __Pyx_pow_long(long, long);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType value);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType __Pyx_PyInt_As_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* ImportNumPyArray.proto */
static PyObject *__pyx_numpy_ndarray = NULL;
static PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static int __pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__lt(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__gt(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__eq(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__create(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak_b, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__create *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_calculate_normalization(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment__determine_peak_type(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_align(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_align *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment__calculate_score(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;

/* Module declarations from 'ms_peak_picker._c.peak_set' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = 0;
static size_t *__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'ms_peak_picker._c.peak_index' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = 0;

/* Module declarations from 'brainpy._c.composition' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition_PyComposition = 0;
static PyObject **__pyx_vp_7brainpy_2_c_11composition_nist_mass = 0;
#define __pyx_v_7brainpy_2_c_11composition_nist_mass (*__pyx_vp_7brainpy_2_c_11composition_nist_mass)
static double *__pyx_vp_7brainpy_2_c_11composition_PROTON = 0;
#define __pyx_v_7brainpy_2_c_11composition_PROTON (*__pyx_vp_7brainpy_2_c_11composition_PROTON)
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable **__pyx_vp_7brainpy_2_c_11composition__ElementTable = 0;
#define __pyx_v_7brainpy_2_c_11composition__ElementTable (*__pyx_vp_7brainpy_2_c_11composition__ElementTable)

/* Module declarations from 'brainpy._c.double_vector' */

/* Module declarations from 'brainpy._c.isotopic_constants' */
static size_t *__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE = 0;
#define __pyx_v_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE (*__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE)

/* Module declarations from 'brainpy._c.isotopic_distribution' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = 0;

/* Module declarations from 'ms_deisotope._c.averagine' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = 0;
static double *__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON = 0;
#define __pyx_v_12ms_deisotope_2_c_9averagine_PROTON (*__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON)

/* Module declarations from 'ms_deisotope._c.deconvoluter_base' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = 0;

/* Module declarations from 'ms_deisotope._c.scoring' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = 0;
static double *__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY = 0;
#define __pyx_v_12ms_deisotope_2_c_7scoring_INFINITY (*__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY)

/* Module declarations from 'ms_deisotope._c.peak_set' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet = 0;
static size_t *__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'ms_deisotope._c.similarity_methods' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_18similarity_methods_PeakMatch = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment = 0;
static PyTypeObject *__pyx_ptype___Pyx_EnumMeta = 0;
static PyObject *__pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros = 0;
static PyObject *__Pyx_OrderedDict = 0;
static PyObject *__Pyx_EnumBase = 0;
static PyObject *__Pyx_globals = 0;
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_deconvoluted_peak_sets(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_deconvoluted_peak_sets *__pyx_optional_args); /*proto*/
static double __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolved_dot_product(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolved_dot_product *__pyx_optional_args); /*proto*/
static double __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted *__pyx_optional_args); /*proto*/
static double __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_deconvoluted(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_deconvoluted *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods___pyx_unpickle_PeakMatch__set_state(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, PyObject *); /*proto*/
static PyObject **__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *); /*proto*/
static PyObject **__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *); /*proto*/
static PyObject **__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *); /*proto*/
static PyObject **__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(PyObject *); /*proto*/
static double __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch, struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity *__pyx_optional_args); /*proto*/
static double __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch, struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity *__pyx_optional_args); /*proto*/
static double __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch, struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity *__pyx_optional_args); /*proto*/
static double __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity *__pyx_optional_args); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch, struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks *__pyx_optional_args); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch, struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks *__pyx_optional_args); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch, struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks *__pyx_optional_args); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks *__pyx_optional_args); /*proto*/
static double __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch, struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity *__pyx_optional_args); /*proto*/
static double __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch, struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity *__pyx_optional_args); /*proto*/
static double __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch, struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity *__pyx_optional_args); /*proto*/
static double __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity *__pyx_optional_args); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch, struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks *__pyx_optional_args); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch, struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks *__pyx_optional_args); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch, struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks *__pyx_optional_args); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods___pyx_unpickle_SpectrumAlignment__set_state(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *, PyObject *); /*proto*/
static PyObject *__pyx_unpickle___Pyx_EnumMeta__set_state(struct __pyx_obj___Pyx_EnumMeta *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "ms_deisotope._c.similarity_methods"
extern int __pyx_module_is_main_ms_deisotope___c__similarity_methods;
int __pyx_module_is_main_ms_deisotope___c__similarity_methods = 0;

/* Implementation of 'ms_deisotope._c.similarity_methods' */
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin___import__;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_s[] = "s";
static const char __pyx_k_v[] = "v";
static const char __pyx_k__2[] = "";
static const char __pyx_k__3[] = "()";
static const char __pyx_k__4[] = "|";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_dct[] = "dct";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_s_s[] = "%s.%s";
static const char __pyx_k_str[] = "__str__";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_align[] = "align";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_s_s_d[] = "<%s.%s: %d>";
static const char __pyx_k_score[] = "score";
static const char __pyx_k_shift[] = "shift";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_Fitted[] = "Fitted";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_norm_a[] = "norm_a";
static const char __pyx_k_norm_b[] = "norm_b";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_peak_a[] = "peak_a";
static const char __pyx_k_peak_b[] = "peak_b";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_IntEnum[] = "IntEnum";
static const char __pyx_k_PeakSet[] = "PeakSet";
static const char __pyx_k_members[] = "__members__";
static const char __pyx_k_parents[] = "parents";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_EnumBase[] = "EnumBase";
static const char __pyx_k_EnumType[] = "EnumType";
static const char __pyx_k_PeakType[] = "PeakType";
static const char __pyx_k_defaults[] = "defaults";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_PeakIndex[] = "PeakIndex";
static const char __pyx_k_PeakMatch[] = "PeakMatch";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_bin_peaks[] = "bin_peaks";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_normalize[] = "normalize";
static const char __pyx_k_peak_type[] = "peak_type";
static const char __pyx_k_precision[] = "precision";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_peak_pairs[] = "peak_pairs";
static const char __pyx_k_peak_set_a[] = "peak_set_a";
static const char __pyx_k_peak_set_b[] = "peak_set_b";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_signatures[] = "signatures";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_OrderedDict[] = "OrderedDict";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_align_peaks[] = "align_peaks";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_dot_product[] = "dot_product";
static const char __pyx_k_Deconvoluted[] = "Deconvoluted";
static const char __pyx_k_Pyx_EnumBase[] = "__Pyx_EnumBase";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_sqrt_transform[] = "sqrt_transform";
static const char __pyx_k_calculate_score[] = "_calculate_score";
static const char __pyx_k_error_tolerance[] = "error_tolerance";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_SimilarityMetrics[] = "SimilarityMetrics";
static const char __pyx_k_SpectrumAlignment[] = "SpectrumAlignment";
static const char __pyx_k_Pyx_EnumBase___new[] = "__Pyx_EnumBase.__new__";
static const char __pyx_k_Pyx_EnumBase___str[] = "__Pyx_EnumBase.__str__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_DeconvolutedPeakSet[] = "DeconvolutedPeakSet";
static const char __pyx_k_Pyx_EnumBase___repr[] = "__Pyx_EnumBase.__repr__";
static const char __pyx_k_determine_peak_type[] = "_determine_peak_type";
static const char __pyx_k_peak_set_similarity[] = "peak_set_similarity";
static const char __pyx_k_pyx_fuse_0bin_peaks[] = "__pyx_fuse_0bin_peaks";
static const char __pyx_k_pyx_fuse_1bin_peaks[] = "__pyx_fuse_1bin_peaks";
static const char __pyx_k_pyx_fuse_2bin_peaks[] = "__pyx_fuse_2bin_peaks";
static const char __pyx_k_pyx_fuse_3bin_peaks[] = "__pyx_fuse_3bin_peaks";
static const char __pyx_k_Unknown_enum_value_s[] = "Unknown enum value: '%s'";
static const char __pyx_k_convolved_dot_product[] = "convolved_dot_product";
static const char __pyx_k_pyx_fuse_0align_peaks[] = "__pyx_fuse_0align_peaks";
static const char __pyx_k_pyx_fuse_1align_peaks[] = "__pyx_fuse_1align_peaks";
static const char __pyx_k_pyx_fuse_2align_peaks[] = "__pyx_fuse_2align_peaks";
static const char __pyx_k_pyx_fuse_3align_peaks[] = "__pyx_fuse_3align_peaks";
static const char __pyx_k_normalized_dot_product[] = "normalized_dot_product";
static const char __pyx_k_pyx_unpickle_PeakMatch[] = "__pyx_unpickle_PeakMatch";
static const char __pyx_k_calculate_normalization[] = "calculate_normalization";
static const char __pyx_k_ppm_peak_set_similarity[] = "ppm_peak_set_similarity";
static const char __pyx_k_ms_deisotope__c_peak_set[] = "ms_deisotope._c.peak_set";
static const char __pyx_k_ms_peak_picker__c_peak_set[] = "ms_peak_picker._c.peak_set";
static const char __pyx_k_No_matching_signature_found[] = "No matching signature found";
static const char __pyx_k_pyx_unpickle___Pyx_EnumMeta[] = "__pyx_unpickle___Pyx_EnumMeta";
static const char __pyx_k_ms_peak_picker__c_peak_index[] = "ms_peak_picker._c.peak_index";
static const char __pyx_k_pyx_fuse_0peak_set_similarity[] = "__pyx_fuse_0peak_set_similarity";
static const char __pyx_k_pyx_fuse_1peak_set_similarity[] = "__pyx_fuse_1peak_set_similarity";
static const char __pyx_k_pyx_fuse_2peak_set_similarity[] = "__pyx_fuse_2peak_set_similarity";
static const char __pyx_k_pyx_fuse_3peak_set_similarity[] = "__pyx_fuse_3peak_set_similarity";
static const char __pyx_k_pyx_fuse_0ppm_peak_set_similar[] = "__pyx_fuse_0ppm_peak_set_similarity";
static const char __pyx_k_pyx_fuse_1ppm_peak_set_similar[] = "__pyx_fuse_1ppm_peak_set_similarity";
static const char __pyx_k_pyx_fuse_2ppm_peak_set_similar[] = "__pyx_fuse_2ppm_peak_set_similarity";
static const char __pyx_k_pyx_fuse_3ppm_peak_set_similar[] = "__pyx_fuse_3ppm_peak_set_similarity";
static const char __pyx_k_pyx_unpickle_SpectrumAlignment[] = "__pyx_unpickle_SpectrumAlignment";
static const char __pyx_k_Cannot_handle_objects_of_type_s[] = "Cannot handle objects of type %s";
static const char __pyx_k_Peak_set_types_must_match_got_s[] = "Peak set types must match, got %s and %s";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_self___class_____name___self_pe[] = "{self.__class__.__name__}({self.peak_a}, {self.peak_b}, {self.score})";
static const char __pyx_k_Expected_at_least_d_argument_s_g[] = "Expected at least %d argument%s, got %d";
static const char __pyx_k_Function_call_with_ambiguous_arg[] = "Function call with ambiguous argument types";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x797b24e, 0x429ecd1, 0x7e0f493) = (peak_a, peak_b, score))";
static const char __pyx_k_ms_deisotope__c_similarity_metho[] = "ms_deisotope._c.similarity_methods";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_src_ms_deisotope__c_similarity_m[] = "src/ms_deisotope/_c/similarity_methods.pyx";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0x158795d, 0xf571677, 0x184e107) = (norm_aa_cosine, norm_bb_cosine, normalize, peak_pairs, peak_set_a, peak_set_b, peak_type, score, shift, sqrt_transform))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0xd41d8cd, 0xe3b0c44, 0xda39a3e) = ())";
static PyObject *__pyx_kp_s_Cannot_handle_objects_of_type_s;
static PyObject *__pyx_n_s_Deconvoluted;
static PyObject *__pyx_n_s_DeconvolutedPeakSet;
static PyObject *__pyx_n_s_EnumBase;
static PyObject *__pyx_n_s_EnumType;
static PyObject *__pyx_kp_s_Expected_at_least_d_argument_s_g;
static PyObject *__pyx_n_s_Fitted;
static PyObject *__pyx_kp_s_Function_call_with_ambiguous_arg;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_IntEnum;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_No_matching_signature_found;
static PyObject *__pyx_n_s_OrderedDict;
static PyObject *__pyx_n_s_PeakIndex;
static PyObject *__pyx_n_s_PeakMatch;
static PyObject *__pyx_n_s_PeakSet;
static PyObject *__pyx_n_s_PeakType;
static PyObject *__pyx_kp_s_Peak_set_types_must_match_got_s;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_Pyx_EnumBase;
static PyObject *__pyx_n_s_Pyx_EnumBase___new;
static PyObject *__pyx_n_s_Pyx_EnumBase___repr;
static PyObject *__pyx_n_s_Pyx_EnumBase___str;
static PyObject *__pyx_n_s_SimilarityMetrics;
static PyObject *__pyx_n_s_SpectrumAlignment;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unknown_enum_value_s;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s__2;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_kp_s__4;
static PyObject *__pyx_n_s_align;
static PyObject *__pyx_n_s_align_peaks;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_bin_peaks;
static PyObject *__pyx_n_s_calculate_normalization;
static PyObject *__pyx_n_s_calculate_score;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_cls;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_convolved_dot_product;
static PyObject *__pyx_n_s_dct;
static PyObject *__pyx_n_s_defaults;
static PyObject *__pyx_n_s_determine_peak_type;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dot_product;
static PyObject *__pyx_n_s_enum;
static PyObject *__pyx_n_s_error_tolerance;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_members;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_kp_s_ms_deisotope__c_peak_set;
static PyObject *__pyx_n_s_ms_deisotope__c_similarity_metho;
static PyObject *__pyx_kp_s_ms_peak_picker__c_peak_index;
static PyObject *__pyx_kp_s_ms_peak_picker__c_peak_set;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_norm_a;
static PyObject *__pyx_n_s_norm_b;
static PyObject *__pyx_n_s_normalize;
static PyObject *__pyx_n_s_normalized_dot_product;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_parents;
static PyObject *__pyx_n_s_peak_a;
static PyObject *__pyx_n_s_peak_b;
static PyObject *__pyx_n_s_peak_pairs;
static PyObject *__pyx_n_s_peak_set_a;
static PyObject *__pyx_n_s_peak_set_b;
static PyObject *__pyx_n_s_peak_set_similarity;
static PyObject *__pyx_n_s_peak_type;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_ppm_peak_set_similarity;
static PyObject *__pyx_n_s_precision;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_fuse_0align_peaks;
static PyObject *__pyx_n_s_pyx_fuse_0bin_peaks;
static PyObject *__pyx_n_s_pyx_fuse_0peak_set_similarity;
static PyObject *__pyx_n_s_pyx_fuse_0ppm_peak_set_similar;
static PyObject *__pyx_n_s_pyx_fuse_1align_peaks;
static PyObject *__pyx_n_s_pyx_fuse_1bin_peaks;
static PyObject *__pyx_n_s_pyx_fuse_1peak_set_similarity;
static PyObject *__pyx_n_s_pyx_fuse_1ppm_peak_set_similar;
static PyObject *__pyx_n_s_pyx_fuse_2align_peaks;
static PyObject *__pyx_n_s_pyx_fuse_2bin_peaks;
static PyObject *__pyx_n_s_pyx_fuse_2peak_set_similarity;
static PyObject *__pyx_n_s_pyx_fuse_2ppm_peak_set_similar;
static PyObject *__pyx_n_s_pyx_fuse_3align_peaks;
static PyObject *__pyx_n_s_pyx_fuse_3bin_peaks;
static PyObject *__pyx_n_s_pyx_fuse_3peak_set_similarity;
static PyObject *__pyx_n_s_pyx_fuse_3ppm_peak_set_similar;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_PeakMatch;
static PyObject *__pyx_n_s_pyx_unpickle_SpectrumAlignment;
static PyObject *__pyx_n_s_pyx_unpickle___Pyx_EnumMeta;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_kp_s_s_s;
static PyObject *__pyx_kp_s_s_s_d;
static PyObject *__pyx_n_s_score;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self___class_____name___self_pe;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shift;
static PyObject *__pyx_n_s_signatures;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_sqrt_transform;
static PyObject *__pyx_kp_s_src_ms_deisotope__c_similarity_m;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_zeros;
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch___init__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v_peak_a, PyObject *__pyx_v_peak_b, PyObject *__pyx_v_score); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_2_lt(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_4_gt(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6_eq(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_8__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other, int __pyx_v_code); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_10__getitem__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_12__len__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_14__repr__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_4__del__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_4__del__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_16__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_18__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_12__pyx_fuse_0peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, int __pyx_v_precision); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_14__pyx_fuse_1peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, int __pyx_v_precision); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_16__pyx_fuse_2peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, int __pyx_v_precision); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_18__pyx_fuse_3peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, int __pyx_v_precision); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_2bin_peaks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_22__pyx_fuse_0bin_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, int __pyx_v_precision); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_24__pyx_fuse_1bin_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, int __pyx_v_precision); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_26__pyx_fuse_2bin_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, int __pyx_v_precision); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_28__pyx_fuse_3bin_peaks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, int __pyx_v_precision); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment___init__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, double __pyx_v_error_tolerance, double __pyx_v_shift, int __pyx_v_normalize, int __pyx_v_sqrt_transform); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_2calculate_normalization(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_4_determine_peak_type(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_6align(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, double __pyx_v_error_tolerance, double __pyx_v_shift); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_8_calculate_score(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_4__del__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_4__del__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_4__del__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_12__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_4convolved_dot_product(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak_pairs, enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType __pyx_v_peak_type, double __pyx_v_norm_a, double __pyx_v_norm_b, int __pyx_v_sqrt_transform); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_6ppm_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_34__pyx_fuse_0ppm_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_36__pyx_fuse_1ppm_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_38__pyx_fuse_2ppm_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_40__pyx_fuse_3ppm_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_8align_peaks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_44__pyx_fuse_0align_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, double __pyx_v_error_tolerance, double __pyx_v_shift, int __pyx_v_normalize); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_46__pyx_fuse_1align_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, double __pyx_v_error_tolerance, double __pyx_v_shift, int __pyx_v_normalize); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_48__pyx_fuse_2align_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, double __pyx_v_error_tolerance, double __pyx_v_shift, int __pyx_v_normalize); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_50__pyx_fuse_3align_peaks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, double __pyx_v_error_tolerance, double __pyx_v_shift, int __pyx_v_normalize); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_10__pyx_unpickle_PeakMatch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_32__pyx_unpickle_SpectrumAlignment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name, PyObject *__pyx_v_parents, PyObject *__pyx_v_dct); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_value, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_18similarity_methods_PeakMatch(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___Pyx_EnumMeta(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_0_0;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_22575453;
static PyObject *__pyx_int_25485575;
static PyObject *__pyx_int_69856465;
static PyObject *__pyx_int_127382094;
static PyObject *__pyx_int_132183187;
static PyObject *__pyx_int_222419149;
static PyObject *__pyx_int_228825662;
static PyObject *__pyx_int_238750788;
static PyObject *__pyx_int_257365623;
static PyObject *__pyx_k_;
static int __pyx_k__7;
static int __pyx_k__8;
static int __pyx_k__9;
static int __pyx_k__10;
static PyObject *__pyx_k__11;
static int __pyx_k__12;
static int __pyx_k__13;
static int __pyx_k__14;
static int __pyx_k__15;
static enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType __pyx_k__16;
static PyObject *__pyx_k__17;
static double __pyx_k__18;
static double __pyx_k__19;
static double __pyx_k__20;
static double __pyx_k__21;
static PyObject *__pyx_k__22;
static PyObject *__pyx_k__23;
static double __pyx_k__24;
static double __pyx_k__25;
static int __pyx_k__26;
static double __pyx_k__27;
static double __pyx_k__28;
static int __pyx_k__29;
static double __pyx_k__30;
static double __pyx_k__31;
static int __pyx_k__32;
static double __pyx_k__33;
static double __pyx_k__34;
static int __pyx_k__35;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__45;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__52;
static PyObject *__pyx_codeobj__54;
/* Late includes */

/* "ms_deisotope/_c/similarity_methods.pyx":50
 *         public double score
 * 
 *     def __init__(self, peak_a, peak_b, score=0.0):             # <<<<<<<<<<<<<<
 *         self.peak_a = peak_a
 *         self.peak_b = peak_b
 */

#line 50 "ms_deisotope/_c/similarity_methods.pyx"


#line 50 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peak_a = 0
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_peak_b = 0
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_score = 0
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_a,&__pyx_n_s_peak_b,&__pyx_n_s_score,0};

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
    values[2] = ((PyObject *)__pyx_float_0_0);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_a)) != 0)) kw_args--;
        else 
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_b)) != 0)) kw_args--;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); 
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 50, __pyx_L3_error)

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_score);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 50, __pyx_L3_error)

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_a = values[0];

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_b = values[1];

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_score = values[2];

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 50, __pyx_L3_error)

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  return -1;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 50 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch___init__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self), __pyx_v_peak_a, __pyx_v_peak_b, __pyx_v_score);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"


#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 50 "ms_deisotope/_c/similarity_methods.pyx"


#line 50 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch___init__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v_peak_a, PyObject *__pyx_v_peak_b, PyObject *__pyx_v_score) {

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_2;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__init__", __pyx_f[0], 50, 0, __PYX_ERR(0, 50, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":51
 * 
 *     def __init__(self, peak_a, peak_b, score=0.0):
 *         self.peak_a = peak_a             # <<<<<<<<<<<<<<
 *         self.peak_b = peak_b
 *         self.score = score
 */

#line 51 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_v_peak_a) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_peak_a, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 51, __pyx_L1_error)

#line 51 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_peak_a;

#line 51 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 51 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 51 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_a);

#line 51 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peak_a));

#line 51 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_a = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 51 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":52
 *     def __init__(self, peak_a, peak_b, score=0.0):
 *         self.peak_a = peak_a
 *         self.peak_b = peak_b             # <<<<<<<<<<<<<<
 *         self.score = score
 * 
 */

#line 52 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_v_peak_b) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_peak_b, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 52, __pyx_L1_error)

#line 52 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_peak_b;

#line 52 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 52 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 52 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_b);

#line 52 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peak_b));

#line 52 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_b = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 52 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":53
 *         self.peak_a = peak_a
 *         self.peak_b = peak_b
 *         self.score = score             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _lt(self, PeakMatch other):
 */

#line 53 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_score); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 53, __pyx_L1_error)

#line 53 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->score = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":50
 *         public double score
 * 
 *     def __init__(self, peak_a, peak_b, score=0.0):             # <<<<<<<<<<<<<<
 *         self.peak_a = peak_a
 *         self.peak_b = peak_b
 */

#line 50 "ms_deisotope/_c/similarity_methods.pyx"


#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 50 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":55
 *         self.score = score
 * 
 *     cpdef bint _lt(self, PeakMatch other):             # <<<<<<<<<<<<<<
 *         return self.score < other.score
 * 
 */

#line 55 "ms_deisotope/_c/similarity_methods.pyx"


#line 55 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_3_lt(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__lt(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch) {

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_lt", 0);

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("_lt", __pyx_f[0], 55, 0, __PYX_ERR(0, 55, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":56
 * 
 *     cpdef bint _lt(self, PeakMatch other):
 *         return self.score < other.score             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _gt(self, PeakMatch other):
 */

#line 56 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = (__pyx_v_self->score < __pyx_v_other->score);

#line 56 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":55
 *         self.score = score
 * 
 *     cpdef bint _lt(self, PeakMatch other):             # <<<<<<<<<<<<<<
 *         return self.score < other.score
 * 
 */

#line 55 "ms_deisotope/_c/similarity_methods.pyx"


#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods.PeakMatch._lt", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 55 "ms_deisotope/_c/similarity_methods.pyx"


#line 55 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_3_lt(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_2_lt[] = "PeakMatch._lt(self, PeakMatch other) -> bool";

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_3_lt(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_lt (wrapper)", 0);

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_18similarity_methods_PeakMatch, 1, "other", 0))) __PYX_ERR(0, 55, __pyx_L1_error)
  __pyx_r = 
#line 55 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_2_lt(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_other));

#line 55 "ms_deisotope/_c/similarity_methods.pyx"


#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 55 "ms_deisotope/_c/similarity_methods.pyx"


#line 55 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_2_lt(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other) {

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_lt", 0);

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("_lt (wrapper)", __pyx_f[0], 55, 0, __PYX_ERR(0, 55, __pyx_L1_error));

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__lt(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"


#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch._lt", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 55 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":58
 *         return self.score < other.score
 * 
 *     cpdef bint _gt(self, PeakMatch other):             # <<<<<<<<<<<<<<
 *         return self.score > other.score
 * 
 */

#line 58 "ms_deisotope/_c/similarity_methods.pyx"


#line 58 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5_gt(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__gt(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch) {

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_gt", 0);

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("_gt", __pyx_f[0], 58, 0, __PYX_ERR(0, 58, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":59
 * 
 *     cpdef bint _gt(self, PeakMatch other):
 *         return self.score > other.score             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, PeakMatch other):
 */

#line 59 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = (__pyx_v_self->score > __pyx_v_other->score);

#line 59 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":58
 *         return self.score < other.score
 * 
 *     cpdef bint _gt(self, PeakMatch other):             # <<<<<<<<<<<<<<
 *         return self.score > other.score
 * 
 */

#line 58 "ms_deisotope/_c/similarity_methods.pyx"


#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods.PeakMatch._gt", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 58 "ms_deisotope/_c/similarity_methods.pyx"


#line 58 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5_gt(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_4_gt[] = "PeakMatch._gt(self, PeakMatch other) -> bool";

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5_gt(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_gt (wrapper)", 0);

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_18similarity_methods_PeakMatch, 1, "other", 0))) __PYX_ERR(0, 58, __pyx_L1_error)
  __pyx_r = 
#line 58 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_4_gt(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_other));

#line 58 "ms_deisotope/_c/similarity_methods.pyx"


#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 58 "ms_deisotope/_c/similarity_methods.pyx"


#line 58 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_4_gt(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other) {

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_gt", 0);

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("_gt (wrapper)", __pyx_f[0], 58, 0, __PYX_ERR(0, 58, __pyx_L1_error));

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__gt(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"


#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch._gt", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 58 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":61
 *         return self.score > other.score
 * 
 *     cpdef bint _eq(self, PeakMatch other):             # <<<<<<<<<<<<<<
 *         return self.peak_a == other.peak_a and self.peak_b == other.peak_b
 * 
 */

#line 61 "ms_deisotope/_c/similarity_methods.pyx"


#line 61 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_7_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__eq(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other, CYTHON_UNUSED int __pyx_skip_dispatch) {

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_1;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_3;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_eq", 0);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("_eq", __pyx_f[0], 61, 0, __PYX_ERR(0, 61, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":62
 * 
 *     cpdef bint _eq(self, PeakMatch other):
 *         return self.peak_a == other.peak_a and self.peak_b == other.peak_b             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, PeakMatch other, int code):
 */

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyObject_RichCompare(((PyObject *)__pyx_v_self->peak_a), ((PyObject *)__pyx_v_other->peak_a), Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 62, __pyx_L1_error)

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 62, __pyx_L1_error)

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_3) {

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_t_3;

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L3_bool_binop_done;

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyObject_RichCompare(((PyObject *)__pyx_v_self->peak_b), ((PyObject *)__pyx_v_other->peak_b), Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 62, __pyx_L1_error)

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 62, __pyx_L1_error)

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_t_3;

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_bool_binop_done:;

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 62 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":61
 *         return self.score > other.score
 * 
 *     cpdef bint _eq(self, PeakMatch other):             # <<<<<<<<<<<<<<
 *         return self.peak_a == other.peak_a and self.peak_b == other.peak_b
 * 
 */

#line 61 "ms_deisotope/_c/similarity_methods.pyx"


#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods.PeakMatch._eq", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 61 "ms_deisotope/_c/similarity_methods.pyx"


#line 61 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_7_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6_eq[] = "PeakMatch._eq(self, PeakMatch other) -> bool";

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_7_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_eq (wrapper)", 0);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_18similarity_methods_PeakMatch, 1, "other", 0))) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_r = 
#line 61 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6_eq(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_other));

#line 61 "ms_deisotope/_c/similarity_methods.pyx"


#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 61 "ms_deisotope/_c/similarity_methods.pyx"


#line 61 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6_eq(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other) {

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_eq", 0);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("_eq (wrapper)", __pyx_f[0], 61, 0, __PYX_ERR(0, 61, __pyx_L1_error));

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__eq(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"


#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch._eq", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 61 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":64
 *         return self.peak_a == other.peak_a and self.peak_b == other.peak_b
 * 
 *     def __richcmp__(self, PeakMatch other, int code):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             if code == 3:
 */

#line 64 "ms_deisotope/_c/similarity_methods.pyx"


#line 64 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /*proto*/

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_18similarity_methods_PeakMatch, 1, "other", 0))) __PYX_ERR(0, 64, __pyx_L1_error)
  __pyx_r = 
#line 64 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_8__richcmp__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_other), ((int)__pyx_v_code));

#line 64 "ms_deisotope/_c/similarity_methods.pyx"


#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 64 "ms_deisotope/_c/similarity_methods.pyx"


#line 64 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_8__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_other, int __pyx_v_code) {

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_1;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_2;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__richcmp__", 0);

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__richcmp__", __pyx_f[0], 64, 0, __PYX_ERR(0, 64, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":65
 * 
 *     def __richcmp__(self, PeakMatch other, int code):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             if code == 3:
 *                 return True
 */

#line 65 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = (((PyObject *)__pyx_v_other) == Py_None);

#line 65 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 65 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/similarity_methods.pyx":66
 *     def __richcmp__(self, PeakMatch other, int code):
 *         if other is None:
 *             if code == 3:             # <<<<<<<<<<<<<<
 *                 return True
 *             else:
 */

#line 66 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = ((__pyx_v_code == 3) != 0);

#line 66 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/similarity_methods.pyx":67
 *         if other is None:
 *             if code == 3:
 *                 return True             # <<<<<<<<<<<<<<
 *             else:
 *                 return False
 */

#line 67 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 67 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(Py_True);

#line 67 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_r = Py_True;

#line 67 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/similarity_methods.pyx":66
 *     def __richcmp__(self, PeakMatch other, int code):
 *         if other is None:
 *             if code == 3:             # <<<<<<<<<<<<<<
 *                 return True
 *             else:
 */

#line 66 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":69
 *                 return True
 *             else:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         if code == 0:
 */

#line 69 "ms_deisotope/_c/similarity_methods.pyx"
    /*else*/ {

#line 69 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 69 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(Py_False);

#line 69 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_r = Py_False;

#line 69 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L0;

#line 69 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":65
 * 
 *     def __richcmp__(self, PeakMatch other, int code):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             if code == 3:
 *                 return True
 */

#line 65 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":71
 *                 return False
 * 
 *         if code == 0:             # <<<<<<<<<<<<<<
 *             return self._lt(other)
 *         elif code == 2:
 */

#line 71 "ms_deisotope/_c/similarity_methods.pyx"
  switch (__pyx_v_code) {

#line 71 "ms_deisotope/_c/similarity_methods.pyx"
    case 0:

    /* "ms_deisotope/_c/similarity_methods.pyx":72
 * 
 *         if code == 0:
 *             return self._lt(other)             # <<<<<<<<<<<<<<
 *         elif code == 2:
 *             return self._eq(other)
 */

#line 72 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 72 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__lt(__pyx_v_self, __pyx_v_other, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 72, __pyx_L1_error)

#line 72 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 72 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = __pyx_t_3;

#line 72 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = 0;

#line 72 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/similarity_methods.pyx":71
 *                 return False
 * 
 *         if code == 0:             # <<<<<<<<<<<<<<
 *             return self._lt(other)
 *         elif code == 2:
 */

#line 71 "ms_deisotope/_c/similarity_methods.pyx"
    break;

#line 71 "ms_deisotope/_c/similarity_methods.pyx"
    case 2:

    /* "ms_deisotope/_c/similarity_methods.pyx":74
 *             return self._lt(other)
 *         elif code == 2:
 *             return self._eq(other)             # <<<<<<<<<<<<<<
 *         elif code == 3:
 *             return not (self._eq(other))
 */

#line 74 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 74 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__eq(__pyx_v_self, __pyx_v_other, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)

#line 74 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 74 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = __pyx_t_3;

#line 74 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = 0;

#line 74 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/similarity_methods.pyx":73
 *         if code == 0:
 *             return self._lt(other)
 *         elif code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */

#line 73 "ms_deisotope/_c/similarity_methods.pyx"
    break;

#line 73 "ms_deisotope/_c/similarity_methods.pyx"
    case 3:

    /* "ms_deisotope/_c/similarity_methods.pyx":76
 *             return self._eq(other)
 *         elif code == 3:
 *             return not (self._eq(other))             # <<<<<<<<<<<<<<
 *         elif code == 4:
 *             return self._gt(other)
 */

#line 76 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 76 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = __Pyx_PyBool_FromLong((!(__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__eq(__pyx_v_self, __pyx_v_other, 0) != 0))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 76, __pyx_L1_error)

#line 76 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 76 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = __pyx_t_3;

#line 76 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = 0;

#line 76 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/similarity_methods.pyx":75
 *         elif code == 2:
 *             return self._eq(other)
 *         elif code == 3:             # <<<<<<<<<<<<<<
 *             return not (self._eq(other))
 *         elif code == 4:
 */

#line 75 "ms_deisotope/_c/similarity_methods.pyx"
    break;

#line 75 "ms_deisotope/_c/similarity_methods.pyx"
    case 4:

    /* "ms_deisotope/_c/similarity_methods.pyx":78
 *             return not (self._eq(other))
 *         elif code == 4:
 *             return self._gt(other)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 78 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 78 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__gt(__pyx_v_self, __pyx_v_other, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 78, __pyx_L1_error)

#line 78 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 78 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = __pyx_t_3;

#line 78 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = 0;

#line 78 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/similarity_methods.pyx":77
 *         elif code == 3:
 *             return not (self._eq(other))
 *         elif code == 4:             # <<<<<<<<<<<<<<
 *             return self._gt(other)
 * 
 */

#line 77 "ms_deisotope/_c/similarity_methods.pyx"
    break;

#line 77 "ms_deisotope/_c/similarity_methods.pyx"
    default: break;

#line 77 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":64
 *         return self.peak_a == other.peak_a and self.peak_b == other.peak_b
 * 
 *     def __richcmp__(self, PeakMatch other, int code):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             if code == 3:
 */

#line 64 "ms_deisotope/_c/similarity_methods.pyx"


#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 64 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":81
 * 
 *     @staticmethod
 *     cdef PeakMatch _create(PeakBase peak_a, PeakBase peak_b, double score=0.0):             # <<<<<<<<<<<<<<
 *         cdef PeakMatch self = PeakMatch.__new__(PeakMatch)
 *         self.peak_a = peak_a
 */

#line 81 "ms_deisotope/_c/similarity_methods.pyx"


#line 81 "ms_deisotope/_c/similarity_methods.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__create(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak_b, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__create *__pyx_optional_args) {

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_score = ((double)0.0);
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self = 0
#line 81 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_r = NULL;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_create", 0);

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("_create", __pyx_f[0], 81, 0, __PYX_ERR(0, 81, __pyx_L1_error));

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_score = __pyx_optional_args->score;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":82
 *     @staticmethod
 *     cdef PeakMatch _create(PeakBase peak_a, PeakBase peak_b, double score=0.0):
 *         cdef PeakMatch self = PeakMatch.__new__(PeakMatch)             # <<<<<<<<<<<<<<
 *         self.peak_a = peak_a
 *         self.peak_b = peak_b
 */

#line 82 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_18similarity_methods_PeakMatch(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_18similarity_methods_PeakMatch), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)

#line 82 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

#line 82 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self = ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_t_1);

#line 82 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":83
 *     cdef PeakMatch _create(PeakBase peak_a, PeakBase peak_b, double score=0.0):
 *         cdef PeakMatch self = PeakMatch.__new__(PeakMatch)
 *         self.peak_a = peak_a             # <<<<<<<<<<<<<<
 *         self.peak_b = peak_b
 *         self.score = score
 */

#line 83 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_a));

#line 83 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak_a));

#line 83 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_a);

#line 83 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peak_a));

#line 83 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_a = __pyx_v_peak_a;

  /* "ms_deisotope/_c/similarity_methods.pyx":84
 *         cdef PeakMatch self = PeakMatch.__new__(PeakMatch)
 *         self.peak_a = peak_a
 *         self.peak_b = peak_b             # <<<<<<<<<<<<<<
 *         self.score = score
 *         return self
 */

#line 84 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_b));

#line 84 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak_b));

#line 84 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_b);

#line 84 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peak_b));

#line 84 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_b = __pyx_v_peak_b;

  /* "ms_deisotope/_c/similarity_methods.pyx":85
 *         self.peak_a = peak_a
 *         self.peak_b = peak_b
 *         self.score = score             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 85 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->score = __pyx_v_score;

  /* "ms_deisotope/_c/similarity_methods.pyx":86
 *         self.peak_b = peak_b
 *         self.score = score
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, i):
 */

#line 86 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 86 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 86 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_self;

#line 86 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":81
 * 
 *     @staticmethod
 *     cdef PeakMatch _create(PeakBase peak_a, PeakBase peak_b, double score=0.0):             # <<<<<<<<<<<<<<
 *         cdef PeakMatch self = PeakMatch.__new__(PeakMatch)
 *         self.peak_a = peak_a
 */

#line 81 "ms_deisotope/_c/similarity_methods.pyx"


#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch._create", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_self);

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 81 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":88
 *         return self
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         if i == 0:
 *             return self.peak_a
 */

#line 88 "ms_deisotope/_c/similarity_methods.pyx"


#line 88 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_11__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_11__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 88 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_10__getitem__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self), ((PyObject *)__pyx_v_i));

#line 88 "ms_deisotope/_c/similarity_methods.pyx"


#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 88 "ms_deisotope/_c/similarity_methods.pyx"


#line 88 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_10__getitem__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v_i) {

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_2;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__getitem__", 0);

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__getitem__", __pyx_f[0], 88, 0, __PYX_ERR(0, 88, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":89
 * 
 *     def __getitem__(self, i):
 *         if i == 0:             # <<<<<<<<<<<<<<
 *             return self.peak_a
 *         elif i == 1:
 */

#line 89 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_i, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)

#line 89 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 89 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 89, __pyx_L1_error)

#line 89 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 89 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/similarity_methods.pyx":90
 *     def __getitem__(self, i):
 *         if i == 0:
 *             return self.peak_a             # <<<<<<<<<<<<<<
 *         elif i == 1:
 *             return self.peak_b
 */

#line 90 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 90 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_self->peak_a));

#line 90 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = ((PyObject *)__pyx_v_self->peak_a);

#line 90 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/similarity_methods.pyx":89
 * 
 *     def __getitem__(self, i):
 *         if i == 0:             # <<<<<<<<<<<<<<
 *             return self.peak_a
 *         elif i == 1:
 */

#line 89 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":91
 *         if i == 0:
 *             return self.peak_a
 *         elif i == 1:             # <<<<<<<<<<<<<<
 *             return self.peak_b
 *         else:
 */

#line 91 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_i, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)

#line 91 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 91 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 91, __pyx_L1_error)

#line 91 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 91 "ms_deisotope/_c/similarity_methods.pyx"
  if (likely(__pyx_t_2)) {

    /* "ms_deisotope/_c/similarity_methods.pyx":92
 *             return self.peak_a
 *         elif i == 1:
 *             return self.peak_b             # <<<<<<<<<<<<<<
 *         else:
 *             raise IndexError(i)
 */

#line 92 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 92 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_self->peak_b));

#line 92 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = ((PyObject *)__pyx_v_self->peak_b);

#line 92 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/similarity_methods.pyx":91
 *         if i == 0:
 *             return self.peak_a
 *         elif i == 1:             # <<<<<<<<<<<<<<
 *             return self.peak_b
 *         else:
 */

#line 91 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":94
 *             return self.peak_b
 *         else:
 *             raise IndexError(i)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 94 "ms_deisotope/_c/similarity_methods.pyx"
  /*else*/ {

#line 94 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)

#line 94 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 94 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 94 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 94 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 94, __pyx_L1_error)

#line 94 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":88
 *         return self
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         if i == 0:
 *             return self.peak_a
 */

#line 88 "ms_deisotope/_c/similarity_methods.pyx"


#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 88 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":96
 *             raise IndexError(i)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return 2
 * 
 */

#line 96 "ms_deisotope/_c/similarity_methods.pyx"


#line 96 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_13__len__(PyObject *__pyx_v_self); /*proto*/

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_13__len__(PyObject *__pyx_v_self) {

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_r;

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 96 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_12__len__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self));

#line 96 "ms_deisotope/_c/similarity_methods.pyx"


#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 96 "ms_deisotope/_c/similarity_methods.pyx"


#line 96 "ms_deisotope/_c/similarity_methods.pyx"
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_12__len__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self) {

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_r;

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__len__", 0);

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__len__", __pyx_f[0], 96, 0, __PYX_ERR(0, 96, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":97
 * 
 *     def __len__(self):
 *         return 2             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 97 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 2;

#line 97 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":96
 *             raise IndexError(i)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return 2
 * 
 */

#line 96 "ms_deisotope/_c/similarity_methods.pyx"


#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 96 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":99
 *         return 2
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         template = "{self.__class__.__name__}({self.peak_a}, {self.peak_b}, {self.score})"
 *         return template.format(self=self)
 */

#line 99 "ms_deisotope/_c/similarity_methods.pyx"


#line 99 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_15__repr__(PyObject *__pyx_v_self); /*proto*/

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_15__repr__(PyObject *__pyx_v_self) {

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 99 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_14__repr__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self));

#line 99 "ms_deisotope/_c/similarity_methods.pyx"


#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 99 "ms_deisotope/_c/similarity_methods.pyx"


#line 99 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_14__repr__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self) {
  PyObject *__pyx_v_template = NULL
#line 99 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__repr__", __pyx_f[0], 99, 0, __PYX_ERR(0, 99, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":100
 * 
 *     def __repr__(self):
 *         template = "{self.__class__.__name__}({self.peak_a}, {self.peak_b}, {self.score})"             # <<<<<<<<<<<<<<
 *         return template.format(self=self)
 * 
 */

#line 100 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_kp_s_self___class_____name___self_pe);

#line 100 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_template = __pyx_kp_s_self___class_____name___self_pe;

  /* "ms_deisotope/_c/similarity_methods.pyx":101
 *     def __repr__(self):
 *         template = "{self.__class__.__name__}({self.peak_a}, {self.peak_b}, {self.score})"
 *         return template.format(self=self)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_template, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 101, __pyx_L1_error)

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 101, __pyx_L1_error)

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_3;

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 101 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":99
 *         return 2
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         template = "{self.__class__.__name__}({self.peak_a}, {self.peak_b}, {self.score})"
 *         return template.format(self=self)
 */

#line 99 "ms_deisotope/_c/similarity_methods.pyx"


#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_template);

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 99 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":46
 * cdef class PeakMatch(object):
 *     cdef:
 *         public PeakBase peak_a             # <<<<<<<<<<<<<<
 *         public PeakBase peak_b
 *         public double score
 */

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_1__get__(PyObject *__pyx_v_self) {

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 46 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self));

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self) {

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 46, 0, __PYX_ERR(0, 46, __pyx_L1_error));

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->peak_a));

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = ((PyObject *)__pyx_v_self->peak_a);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.peak_a.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 46 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v_value) {

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 46, 0, __PYX_ERR(0, 46, __pyx_L1_error));

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 46, __pyx_L1_error)

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_value;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_a);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peak_a));

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_a = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.peak_a.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_5__del__(PyObject *__pyx_v_self) {

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 46 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self));

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_4__del__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self) {

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__del__", __pyx_f[0], 46, 0, __PYX_ERR(0, 46, __pyx_L1_error));

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(Py_None);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(Py_None);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_a);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peak_a));

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_a = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"


#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.peak_a.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":47
 *     cdef:
 *         public PeakBase peak_a
 *         public PeakBase peak_b             # <<<<<<<<<<<<<<
 *         public double score
 * 
 */

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_1__get__(PyObject *__pyx_v_self) {

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 47 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self));

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self) {

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 47, 0, __PYX_ERR(0, 47, __pyx_L1_error));

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->peak_b));

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = ((PyObject *)__pyx_v_self->peak_b);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.peak_b.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 47 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v_value) {

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 47, 0, __PYX_ERR(0, 47, __pyx_L1_error));

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_value;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_b);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peak_b));

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_b = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.peak_b.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_5__del__(PyObject *__pyx_v_self) {

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 47 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self));

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_4__del__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self) {

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__del__", __pyx_f[0], 47, 0, __PYX_ERR(0, 47, __pyx_L1_error));

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(Py_None);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(Py_None);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_b);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peak_b));

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_b = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"


#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.peak_b.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 47 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":48
 *         public PeakBase peak_a
 *         public PeakBase peak_b
 *         public double score             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, peak_a, peak_b, score=0.0):
 */

#line 48 "ms_deisotope/_c/similarity_methods.pyx"


#line 48 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score_1__get__(PyObject *__pyx_v_self) {

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 48 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self));

#line 48 "ms_deisotope/_c/similarity_methods.pyx"


#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 48 "ms_deisotope/_c/similarity_methods.pyx"


#line 48 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self) {

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 48, 0, __PYX_ERR(0, 48, __pyx_L1_error));

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"


#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.score.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 48 "ms_deisotope/_c/similarity_methods.pyx"


#line 48 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 48 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 48 "ms_deisotope/_c/similarity_methods.pyx"


#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 48 "ms_deisotope/_c/similarity_methods.pyx"


#line 48 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v_value) {

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 48, 0, __PYX_ERR(0, 48, __pyx_L1_error));

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->score = __pyx_t_1;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"


#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.score.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 48 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_16__reduce_cython__[] = "PeakMatch.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_16__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_16__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_TraceDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  int __pyx_t_4;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

#line 1 "(tree fragment)"
  __Pyx_TraceCall("__reduce_cython__", __pyx_f[1], 1, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.peak_a, self.peak_b, self.score)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->peak_a));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->peak_a));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->peak_a));

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->peak_b));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->peak_b));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self->peak_b));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.peak_a, self.peak_b, self.score)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_2;

#line 6 "(tree fragment)"
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.peak_a, self.peak_b, self.score)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_3 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));

#line 8 "(tree fragment)"
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.peak_a is not None or self.peak_b is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.peak_a, self.peak_b, self.score)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.peak_a is not None or self.peak_b is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PeakMatch, (type(self), 0x797b24e, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_3 = (((PyObject *)__pyx_v_self->peak_a) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_5 = (__pyx_t_3 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_5) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_4 = __pyx_t_5;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_5 = (((PyObject *)__pyx_v_self->peak_b) != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_3 = (__pyx_t_5 != 0);

#line 11 "(tree fragment)"
    __pyx_t_4 = __pyx_t_3;

#line 11 "(tree fragment)"
    __pyx_L4_bool_binop_done:;

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_4;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.peak_a is not None or self.peak_b is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PeakMatch, (type(self), 0x797b24e, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_4 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_4) {

    /* "(tree fragment)":13
 *         use_setstate = self.peak_a is not None or self.peak_b is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PeakMatch, (type(self), 0x797b24e, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PeakMatch, (type(self), 0x797b24e, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_PeakMatch); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_127382094);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_127382094);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_127382094);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    __pyx_t_2 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_6;

#line 13 "(tree fragment)"
    __pyx_t_6 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.peak_a is not None or self.peak_b is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PeakMatch, (type(self), 0x797b24e, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PeakMatch, (type(self), 0x797b24e, None), state
 *     else:
 *         return __pyx_unpickle_PeakMatch, (type(self), 0x797b24e, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PeakMatch__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_PeakMatch); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 15 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_127382094);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_127382094);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_127382094);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_6);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);

#line 15 "(tree fragment)"
    __pyx_t_6 = 0;

#line 15 "(tree fragment)"
    __pyx_t_2 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_1;

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PeakMatch, (type(self), 0x797b24e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PeakMatch__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_18__setstate_cython__[] = "PeakMatch.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_18__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_18__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_TraceDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

#line 16 "(tree fragment)"
  __Pyx_TraceCall("__setstate_cython__", __pyx_f[1], 16, 0, __PYX_ERR(1, 16, __pyx_L1_error));

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PeakMatch, (type(self), 0x797b24e, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PeakMatch__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods___pyx_unpickle_PeakMatch__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PeakMatch, (type(self), 0x797b24e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PeakMatch__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.PeakMatch.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/similarity_methods.pyx":104
 * 
 * 
 * cdef PyObject** _make_fast_array(peak_collection peak_set):             # <<<<<<<<<<<<<<
 *     cdef:
 *         PyObject** items
 */

#line 104 "ms_deisotope/_c/similarity_methods.pyx"


#line 104 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject **__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set) {
  PyObject **__pyx_v_items
#line 104 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject **__pyx_r;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0_make_fast_array", 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_0_make_fast_array", __pyx_f[0], 104, 0, __PYX_ERR(0, 104, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":113
 *     elif peak_collection is PeakSet:
 *         items = PySequence_Fast_ITEMS(
 *             PySequence_Fast(peak_set.peaks, "Error, could not create fast array from PeakSet"))             # <<<<<<<<<<<<<<
 *     elif peak_collection is DeconvolutedPeakSet:
 *         items = PySequence_Fast_ITEMS(
 */

#line 113 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_peak_set->peaks;

#line 113 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 113 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PySequence_Fast(__pyx_t_1, ((char *)"Error, could not create fast array from PeakSet")); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 113, __pyx_L1_error)

#line 113 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 113 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":112
 *             PySequence_Fast(peak_set.peaks.peaks, "Error, could not create fast array from PeakIndex"))
 *     elif peak_collection is PeakSet:
 *         items = PySequence_Fast_ITEMS(             # <<<<<<<<<<<<<<
 *             PySequence_Fast(peak_set.peaks, "Error, could not create fast array from PeakSet"))
 *     elif peak_collection is DeconvolutedPeakSet:
 */

#line 112 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items = PySequence_Fast_ITEMS(__pyx_t_2);

#line 112 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":122
 *         items = PySequence_Fast_ITEMS(
 *             PySequence_Fast(peak_set, "Error, could not create fast array from object"))
 *     return items             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 122 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_items;

#line 122 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":104
 * 
 * 
 * cdef PyObject** _make_fast_array(peak_collection peak_set):             # <<<<<<<<<<<<<<
 *     cdef:
 *         PyObject** items
 */

#line 104 "ms_deisotope/_c/similarity_methods.pyx"


#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods._make_fast_array", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 104 "ms_deisotope/_c/similarity_methods.pyx"


#line 104 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject **__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set) {
  PyObject **__pyx_v_items
#line 104 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject **__pyx_r;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1_make_fast_array", 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_1_make_fast_array", __pyx_f[0], 104, 0, __PYX_ERR(0, 104, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":110
 *     if peak_collection is PeakIndex:
 *         items = PySequence_Fast_ITEMS(
 *             PySequence_Fast(peak_set.peaks.peaks, "Error, could not create fast array from PeakIndex"))             # <<<<<<<<<<<<<<
 *     elif peak_collection is PeakSet:
 *         items = PySequence_Fast_ITEMS(
 */

#line 110 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_peak_set->peaks->peaks;

#line 110 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 110 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PySequence_Fast(__pyx_t_1, ((char *)"Error, could not create fast array from PeakIndex")); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)

#line 110 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 110 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":109
 * 
 *     if peak_collection is PeakIndex:
 *         items = PySequence_Fast_ITEMS(             # <<<<<<<<<<<<<<
 *             PySequence_Fast(peak_set.peaks.peaks, "Error, could not create fast array from PeakIndex"))
 *     elif peak_collection is PeakSet:
 */

#line 109 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items = PySequence_Fast_ITEMS(__pyx_t_2);

#line 109 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":122
 *         items = PySequence_Fast_ITEMS(
 *             PySequence_Fast(peak_set, "Error, could not create fast array from object"))
 *     return items             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 122 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_items;

#line 122 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":104
 * 
 * 
 * cdef PyObject** _make_fast_array(peak_collection peak_set):             # <<<<<<<<<<<<<<
 *     cdef:
 *         PyObject** items
 */

#line 104 "ms_deisotope/_c/similarity_methods.pyx"


#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods._make_fast_array", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 104 "ms_deisotope/_c/similarity_methods.pyx"


#line 104 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject **__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set) {
  PyObject **__pyx_v_items
#line 104 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject **__pyx_r;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2_make_fast_array", 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_2_make_fast_array", __pyx_f[0], 104, 0, __PYX_ERR(0, 104, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":116
 *     elif peak_collection is DeconvolutedPeakSet:
 *         items = PySequence_Fast_ITEMS(
 *             PySequence_Fast(peak_set.peaks, "Error, could not create fast array from DeconvolutedPeakSet"))             # <<<<<<<<<<<<<<
 *     else:
 *         # This block is unsafe for any type that is not a list or a tuple, risking segfaults.
 */

#line 116 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_peak_set->peaks;

#line 116 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 116 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PySequence_Fast(__pyx_t_1, ((char *)"Error, could not create fast array from DeconvolutedPeakSet")); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 116, __pyx_L1_error)

#line 116 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 116 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":115
 *             PySequence_Fast(peak_set.peaks, "Error, could not create fast array from PeakSet"))
 *     elif peak_collection is DeconvolutedPeakSet:
 *         items = PySequence_Fast_ITEMS(             # <<<<<<<<<<<<<<
 *             PySequence_Fast(peak_set.peaks, "Error, could not create fast array from DeconvolutedPeakSet"))
 *     else:
 */

#line 115 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items = PySequence_Fast_ITEMS(__pyx_t_2);

#line 115 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":122
 *         items = PySequence_Fast_ITEMS(
 *             PySequence_Fast(peak_set, "Error, could not create fast array from object"))
 *     return items             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 122 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_items;

#line 122 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":104
 * 
 * 
 * cdef PyObject** _make_fast_array(peak_collection peak_set):             # <<<<<<<<<<<<<<
 *     cdef:
 *         PyObject** items
 */

#line 104 "ms_deisotope/_c/similarity_methods.pyx"


#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods._make_fast_array", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 104 "ms_deisotope/_c/similarity_methods.pyx"


#line 104 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject **__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(PyObject *__pyx_v_peak_set) {
  PyObject **__pyx_v_items
#line 104 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject **__pyx_r;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3_make_fast_array", 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_3_make_fast_array", __pyx_f[0], 104, 0, __PYX_ERR(0, 104, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":121
 *         # The list/tuple must have a lifespan exceeding the duration of this function call.
 *         items = PySequence_Fast_ITEMS(
 *             PySequence_Fast(peak_set, "Error, could not create fast array from object"))             # <<<<<<<<<<<<<<
 *     return items
 * 
 */

#line 121 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PySequence_Fast(__pyx_v_peak_set, ((char *)"Error, could not create fast array from object")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)

#line 121 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

  /* "ms_deisotope/_c/similarity_methods.pyx":120
 *         # This block is unsafe for any type that is not a list or a tuple, risking segfaults.
 *         # The list/tuple must have a lifespan exceeding the duration of this function call.
 *         items = PySequence_Fast_ITEMS(             # <<<<<<<<<<<<<<
 *             PySequence_Fast(peak_set, "Error, could not create fast array from object"))
 *     return items
 */

#line 120 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items = PySequence_Fast_ITEMS(__pyx_t_1);

#line 120 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":122
 *         items = PySequence_Fast_ITEMS(
 *             PySequence_Fast(peak_set, "Error, could not create fast array from object"))
 *     return items             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 122 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_items;

#line 122 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":104
 * 
 * 
 * cdef PyObject** _make_fast_array(peak_collection peak_set):             # <<<<<<<<<<<<<<
 *     cdef:
 *         PyObject** items
 */

#line 104 "ms_deisotope/_c/similarity_methods.pyx"


#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods._make_fast_array", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 104 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":127
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cpdef double peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0) except -1:             # <<<<<<<<<<<<<<
 *     """Computes the normalized dot product, also called cosine similarity between
 *     two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).
 */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_1peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity[] = "peak_set_similarity(signatures, args, kwargs, defaults)\nComputes the normalized dot product, also called cosine similarity between\n    two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).\n\n    Parameters\n    ----------\n    peak_set_a : Iterable of Peak-like\n    peak_set_b : Iterable of Peak-like\n        The two peak collections to compare. It is usually only useful to\n        compare the similarity of peaks of the same class, so the types\n        of the elements of `peak_set_a` and `peak_set_b` should match.\n    precision : int, optional\n        The precision of rounding to use when binning spectra. Defaults to 0\n\n    Returns\n    -------\n    float\n        The similarity between peak_set_a and peak_set_b. Between 0.0 and 1.0\n    ";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_18similarity_methods_1peak_set_similarity = 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
{"peak_set_similarity", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_1peak_set_similarity, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_1peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_signatures = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_args = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signatures,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[4] = {0,0,0,0};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    values[2] = __pyx_k_;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  4: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signatures)) != 0)) kw_args--;
        else 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 1); 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  3:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 3); 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_signatures = values[0];

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_args = values[1];

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_kwargs = values[2];

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_defaults = values[3];

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(__pyx_self, __pyx_v_signatures, __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  Py_ssize_t __pyx_v_i
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_arg = NULL
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_candidates = NULL
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_sig = NULL
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_match_found
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_src_sig = NULL
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_dst_type = NULL
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_2;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_3;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_4;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_5;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_7;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_8;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_9;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_13;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_14;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_15;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_16;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("peak_set_similarity", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("peak_set_similarity", __pyx_f[0], 127, 0, __PYX_ERR(0, 127, __pyx_L1_error));

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_kwargs);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_4) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L4_bool_binop_done;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((!__pyx_t_4) != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_bool_binop_done:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_itemsize = -1L;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_args == Py_None)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = ((0 < __pyx_t_5) != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L6;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_4) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L7_bool_binop_done;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_peak_set_a, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_t_4 != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L7_bool_binop_done:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_peak_set_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L6;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /*else*/ {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_int_2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_int_2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_n_s_s);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_n_s_s);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L6:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  while (1) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s__2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_peak_picker__c_peak_set);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_3) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s__2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_ms_peak_picker__c_peak_index);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PeakIndex); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = (__pyx_t_3 != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_2) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakIndex, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s__2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_deisotope__c_peak_set);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_DeconvolutedPeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_3) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_DeconvolutedPeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_object, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L10_break;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L10_break:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_candidates = ((PyObject*)__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_signatures == Py_None)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_t_6;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  while (1) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_7, &__pyx_t_5, &__pyx_t_6, NULL, NULL, __pyx_t_8);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 == 0)) break;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_match_found = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_12 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(__pyx_t_12)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_12);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(function);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__3) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_10);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(__pyx_t_10)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_10);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(function);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__4) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__4);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_14 = __pyx_t_13;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_i = __pyx_t_15;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_2 = (__pyx_t_3 != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_2) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_6, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_match_found = 1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L19;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        /*else*/ {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_match_found = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L17_break;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_L19:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_L17_break:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = (__pyx_v_match_found != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_2) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((!__pyx_t_2) != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_3) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((__pyx_t_7 > 1) != 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_3) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /*else*/ {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_signatures == Py_None)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = __pyx_t_1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_11);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_12);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_dest_sig);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_ndarray);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_arg);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_candidates);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_sig);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_src_sig);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_dst_type);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_13__pyx_fuse_0peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_1peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity *__pyx_optional_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_precision = __pyx_k__7;
  double __pyx_v_n_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_n_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_n_ab
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_z
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_hi
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_scaler
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double *__pyx_v_bin_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double *__pyx_v_bin_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_i
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED long __pyx_v_n
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_k
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_index
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_n_peaks_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_n_peaks_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject **__pyx_v_items_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject **__pyx_v_items_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_p_peak
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_2;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_3;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_4;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_5;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_6;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_7;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_8;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_9;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_10;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0peak_set_similarity", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_0peak_set_similarity", __pyx_f[0], 127, 0, __PYX_ERR(0, 127, __pyx_L1_error));

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_optional_args->precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":160
 *         PyObject* p_peak
 * 
 *     n_a = 0             # <<<<<<<<<<<<<<
 *     n_b = 0
 *     a = 0
 */

#line 160 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_a = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":161
 * 
 *     n_a = 0
 *     n_b = 0             # <<<<<<<<<<<<<<
 *     a = 0
 *     b = 0
 */

#line 161 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_b = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":162
 *     n_a = 0
 *     n_b = 0
 *     a = 0             # <<<<<<<<<<<<<<
 *     b = 0
 *     z = 0
 */

#line 162 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_a = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":163
 *     n_b = 0
 *     a = 0
 *     b = 0             # <<<<<<<<<<<<<<
 *     z = 0
 * 
 */

#line 163 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_b = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":164
 *     a = 0
 *     b = 0
 *     z = 0             # <<<<<<<<<<<<<<
 * 
 *     # Determine maximum value
 */

#line 164 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_z = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":167
 * 
 *     # Determine maximum value
 *     hi = 0             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakIndex:
 *         peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
 */

#line 167 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":184
 *                 hi = peak.mz
 *     elif peak_collection is PeakSet:
 *         peak = peak_set_a.getitem(peak_set_a.get_size() - 1)             # <<<<<<<<<<<<<<
 *         hi = peak.mz
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
 */

#line 184 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->getitem(__pyx_v_peak_set_a, (((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a) - 1))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)

#line 184 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 184 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 184 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":185
 *     elif peak_collection is PeakSet:
 *         peak = peak_set_a.getitem(peak_set_a.get_size() - 1)
 *         hi = peak.mz             # <<<<<<<<<<<<<<
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)
 */

#line 185 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_peak->mz;

#line 185 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":186
 *         peak = peak_set_a.getitem(peak_set_a.get_size() - 1)
 *         hi = peak.mz
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)             # <<<<<<<<<<<<<<
 *         hi = max(peak.mz, hi)
 *     elif peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 186 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->getitem(__pyx_v_peak_set_b, (((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->get_size(__pyx_v_peak_set_b) - 1))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)

#line 186 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 186 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 186 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":187
 *         hi = peak.mz
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)             # <<<<<<<<<<<<<<
 *     elif peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         peak = peak_set_a._mz_ordered[-1]
 */

#line 187 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_hi;

#line 187 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_v_peak->mz;

#line 187 "ms_deisotope/_c/similarity_methods.pyx"
  if (((__pyx_t_2 > __pyx_t_3) != 0)) {

#line 187 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_2;

#line 187 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 187 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_3;

#line 187 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 187 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_4;

  /* "ms_deisotope/_c/similarity_methods.pyx":195
 * 
 *     # Calculate bin array size and allocate memory
 *     scaler = 10 ** precision             # <<<<<<<<<<<<<<
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))
 */

#line 195 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_scaler = __Pyx_pow_long(10, ((long)__pyx_v_precision));

  /* "ms_deisotope/_c/similarity_methods.pyx":196
 *     # Calculate bin array size and allocate memory
 *     scaler = 10 ** precision
 *     k = <long>((hi * scaler) + 1)             # <<<<<<<<<<<<<<
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 */

#line 196 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_k = ((long)((__pyx_v_hi * __pyx_v_scaler) + 1.0));

  /* "ms_deisotope/_c/similarity_methods.pyx":197
 *     scaler = 10 ** precision
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))             # <<<<<<<<<<<<<<
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 */

#line 197 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_a = ((double *)malloc((__pyx_v_k * (sizeof(double)))));

  /* "ms_deisotope/_c/similarity_methods.pyx":198
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))             # <<<<<<<<<<<<<<
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:
 */

#line 198 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_b = ((double *)malloc((__pyx_v_k * (sizeof(double)))));

  /* "ms_deisotope/_c/similarity_methods.pyx":199
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:             # <<<<<<<<<<<<<<
 *         if bin_a != NULL:
 *             free(bin_a)
 */

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = ((__pyx_v_bin_a == NULL) != 0);

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  if (!__pyx_t_6) {

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L4_bool_binop_done;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = ((__pyx_v_bin_b == NULL) != 0);

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = __pyx_t_6;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_bool_binop_done:;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/similarity_methods.pyx":200
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_a)
 *         if bin_b != NULL:
 */

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_bin_a != NULL) != 0);

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":201
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:
 *             free(bin_a)             # <<<<<<<<<<<<<<
 *         if bin_b != NULL:
 *             free(bin_b)
 */

#line 201 "ms_deisotope/_c/similarity_methods.pyx"
      free(__pyx_v_bin_a);

      /* "ms_deisotope/_c/similarity_methods.pyx":200
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_a)
 *         if bin_b != NULL:
 */

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":202
 *         if bin_a != NULL:
 *             free(bin_a)
 *         if bin_b != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 */

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_bin_b != NULL) != 0);

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":203
 *             free(bin_a)
 *         if bin_b != NULL:
 *             free(bin_b)             # <<<<<<<<<<<<<<
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1
 */

#line 203 "ms_deisotope/_c/similarity_methods.pyx"
      free(__pyx_v_bin_b);

      /* "ms_deisotope/_c/similarity_methods.pyx":202
 *         if bin_a != NULL:
 *             free(bin_a)
 *         if bin_b != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 */

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":204
 *         if bin_b != NULL:
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")             # <<<<<<<<<<<<<<
 *         return -1
 *     for i in range(k):
 */

#line 204 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(__pyx_builtin_MemoryError, ((char *)"Could not allocate bins"));

    /* "ms_deisotope/_c/similarity_methods.pyx":205
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1             # <<<<<<<<<<<<<<
 *     for i in range(k):
 *         bin_a[i] = bin_b[i] = 0
 */

#line 205 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = -1.0;

#line 205 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/similarity_methods.pyx":199
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:             # <<<<<<<<<<<<<<
 *         if bin_a != NULL:
 *             free(bin_a)
 */

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":206
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1
 *     for i in range(k):             # <<<<<<<<<<<<<<
 *         bin_a[i] = bin_b[i] = 0
 * 
 */

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_7 = __pyx_v_k;

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = __pyx_t_7;

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_9;

    /* "ms_deisotope/_c/similarity_methods.pyx":207
 *         return -1
 *     for i in range(k):
 *         bin_a[i] = bin_b[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     # Fill bins
 */

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
    (__pyx_v_bin_a[__pyx_v_i]) = 0.0;

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
    (__pyx_v_bin_b[__pyx_v_i]) = 0.0;

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":214
 *         n_peaks_a = len(_hold_a)
 *     else:
 *         n_peaks_a = peak_set_a.get_size()             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 214 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_peaks_a = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":220
 *         n_peaks_b = len(_hold_b)
 *     else:
 *         n_peaks_b = peak_set_b.get_size()             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 220 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_peaks_b = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->get_size(__pyx_v_peak_set_b);

  /* "ms_deisotope/_c/similarity_methods.pyx":226
 *         items_b = _make_fast_array(_hold_b)
 *     else:
 *         items_a = _make_fast_array(peak_set_a)             # <<<<<<<<<<<<<<
 *         items_b = _make_fast_array(peak_set_b)
 *     with nogil:
 */

#line 226 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items_a = __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":227
 *     else:
 *         items_a = _make_fast_array(peak_set_a)
 *         items_b = _make_fast_array(peak_set_b)             # <<<<<<<<<<<<<<
 *     with nogil:
 *         # Fill bins
 */

#line 227 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items_b = __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(__pyx_v_peak_set_b);

  /* "ms_deisotope/_c/similarity_methods.pyx":228
 *         items_a = _make_fast_array(peak_set_a)
 *         items_b = _make_fast_array(peak_set_b)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # Fill bins
 *         for i in range(n_peaks_a):
 */

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      #ifdef WITH_THREAD

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      PyThreadState *_save;

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      Py_UNBLOCK_THREADS

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_FastGIL_Remember();

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      /*try:*/ {

        /* "ms_deisotope/_c/similarity_methods.pyx":230
 *     with nogil:
 *         # Fill bins
 *         for i in range(n_peaks_a):             # <<<<<<<<<<<<<<
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 */

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_7 = __pyx_v_n_peaks_a;

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_8 = __pyx_t_7;

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_i = __pyx_t_9;

          /* "ms_deisotope/_c/similarity_methods.pyx":231
 *         # Fill bins
 *         for i in range(n_peaks_a):
 *             p_peak = items_a[i]             # <<<<<<<<<<<<<<
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 */

#line 231 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_p_peak = (__pyx_v_items_a[__pyx_v_i]);

          /* "ms_deisotope/_c/similarity_methods.pyx":232
 *         for i in range(n_peaks_a):
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)             # <<<<<<<<<<<<<<
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 * 
 */

#line 232 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_index = ((long)(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->mz * __pyx_v_scaler));

          /* "ms_deisotope/_c/similarity_methods.pyx":233
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_a[index] += (<PeakBase>p_peak).intensity             # <<<<<<<<<<<<<<
 * 
 *         for i in range(n_peaks_b):
 */

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_10 = __pyx_v_index;

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
          (__pyx_v_bin_a[__pyx_t_10]) = ((__pyx_v_bin_a[__pyx_t_10]) + ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->intensity);

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":235
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 * 
 *         for i in range(n_peaks_b):             # <<<<<<<<<<<<<<
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 */

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_7 = __pyx_v_n_peaks_b;

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_8 = __pyx_t_7;

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_i = __pyx_t_9;

          /* "ms_deisotope/_c/similarity_methods.pyx":236
 * 
 *         for i in range(n_peaks_b):
 *             p_peak = items_b[i]             # <<<<<<<<<<<<<<
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_b[index] += (<PeakBase>p_peak).intensity
 */

#line 236 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_p_peak = (__pyx_v_items_b[__pyx_v_i]);

          /* "ms_deisotope/_c/similarity_methods.pyx":237
 *         for i in range(n_peaks_b):
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)             # <<<<<<<<<<<<<<
 *             bin_b[index] += (<PeakBase>p_peak).intensity
 * 
 */

#line 237 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_index = ((long)(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->mz * __pyx_v_scaler));

          /* "ms_deisotope/_c/similarity_methods.pyx":238
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_b[index] += (<PeakBase>p_peak).intensity             # <<<<<<<<<<<<<<
 * 
 *         # Calculate dot product and normalizers in parallel
 */

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_10 = __pyx_v_index;

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
          (__pyx_v_bin_b[__pyx_t_10]) = ((__pyx_v_bin_b[__pyx_t_10]) + ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->intensity);

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":241
 * 
 *         # Calculate dot product and normalizers in parallel
 *         n = k             # <<<<<<<<<<<<<<
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]
 */

#line 241 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_n = __pyx_v_k;

        /* "ms_deisotope/_c/similarity_methods.pyx":242
 *         # Calculate dot product and normalizers in parallel
 *         n = k
 *         for i in prange(n, schedule='static', num_threads=4):             # <<<<<<<<<<<<<<
 *             a = bin_a[i]
 *             b = bin_b[i]
 */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_7 = __pyx_v_n;

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        if ((1 == 0)) abort();

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #undef likely

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #undef unlikely

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #define likely(x)   (x)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #define unlikely(x) (x)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            #endif

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            __pyx_t_9 = (__pyx_t_7 - 0 + 1 - 1/abs(1)) / 1;

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            if (__pyx_t_9 > 0)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #ifdef _OPENMP
                #pragma omp parallel reduction(+:__pyx_v_n_a) reduction(+:__pyx_v_n_b) reduction(+:__pyx_v_z) num_threads(4)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"


#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #endif /* _OPENMP */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_a) lastprivate(__pyx_v_b) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) schedule(static)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"


#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                    #endif /* _OPENMP */
                    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_9; __pyx_t_8++)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"
{

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                        {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_i = (long)(0 + 1 * __pyx_t_8);

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            /* Initialize private variables to invalid values */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_a = ((double)__PYX_NAN());

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_b = ((double)__PYX_NAN());

                            /* "ms_deisotope/_c/similarity_methods.pyx":243
 *         n = k
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]             # <<<<<<<<<<<<<<
 *             b = bin_b[i]
 *             n_a += a * a
 */

#line 243 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_a = (__pyx_v_bin_a[__pyx_v_i]);

                            /* "ms_deisotope/_c/similarity_methods.pyx":244
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]
 *             b = bin_b[i]             # <<<<<<<<<<<<<<
 *             n_a += a * a
 *             n_b += b * b
 */

#line 244 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_b = (__pyx_v_bin_b[__pyx_v_i]);

                            /* "ms_deisotope/_c/similarity_methods.pyx":245
 *             a = bin_a[i]
 *             b = bin_b[i]
 *             n_a += a * a             # <<<<<<<<<<<<<<
 *             n_b += b * b
 *             z += a * b
 */

#line 245 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_n_a = (__pyx_v_n_a + (__pyx_v_a * __pyx_v_a));

                            /* "ms_deisotope/_c/similarity_methods.pyx":246
 *             b = bin_b[i]
 *             n_a += a * a
 *             n_b += b * b             # <<<<<<<<<<<<<<
 *             z += a * b
 * 
 */

#line 246 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_n_b = (__pyx_v_n_b + (__pyx_v_b * __pyx_v_b));

                            /* "ms_deisotope/_c/similarity_methods.pyx":247
 *             n_a += a * a
 *             n_b += b * b
 *             z += a * b             # <<<<<<<<<<<<<<
 * 
 *         free(bin_a)
 */

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_z = (__pyx_v_z + (__pyx_v_a * __pyx_v_b));

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                        }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                    }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #undef likely

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #undef unlikely

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #define likely(x)   __builtin_expect(!!(x), 1)

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #define unlikely(x) __builtin_expect(!!(x), 0)

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

        /* "ms_deisotope/_c/similarity_methods.pyx":249
 *             z += a * b
 * 
 *         free(bin_a)             # <<<<<<<<<<<<<<
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 */

#line 249 "ms_deisotope/_c/similarity_methods.pyx"
        free(__pyx_v_bin_a);

        /* "ms_deisotope/_c/similarity_methods.pyx":250
 * 
 *         free(bin_a)
 *         free(bin_b)             # <<<<<<<<<<<<<<
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:
 */

#line 250 "ms_deisotope/_c/similarity_methods.pyx"
        free(__pyx_v_bin_b);

        /* "ms_deisotope/_c/similarity_methods.pyx":251
 *         free(bin_a)
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)             # <<<<<<<<<<<<<<
 *         if n_ab == 0:
 *             return 0.0
 */

#line 251 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_n_ab = (sqrt(__pyx_v_n_a) * sqrt(__pyx_v_n_b));

        /* "ms_deisotope/_c/similarity_methods.pyx":252
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 *         return z / n_ab
 */

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_5 = ((__pyx_v_n_ab == 0.0) != 0);

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_5) {

          /* "ms_deisotope/_c/similarity_methods.pyx":253
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:
 *             return 0.0             # <<<<<<<<<<<<<<
 *         return z / n_ab
 * 
 */

#line 253 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_r = 0.0;

#line 253 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L10_return;

          /* "ms_deisotope/_c/similarity_methods.pyx":252
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 *         return z / n_ab
 */

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":254
 *         if n_ab == 0:
 *             return 0.0
 *         return z / n_ab             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_r = (__pyx_v_z / __pyx_v_n_ab);

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L10_return;

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":228
 *         items_a = _make_fast_array(peak_set_a)
 *         items_b = _make_fast_array(peak_set_b)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # Fill bins
 *         for i in range(n_peaks_a):
 */

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      /*finally:*/ {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_L10_return: {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          #ifdef WITH_THREAD

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_FastGIL_Forget();

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          Py_BLOCK_THREADS

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          #endif

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L0;

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":127
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cpdef double peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0) except -1:             # <<<<<<<<<<<<<<
 *     """Computes the normalized dot product, also called cosine similarity between
 *     two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).
 */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1.0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_13__pyx_fuse_0peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_13__pyx_fuse_0peak_set_similarity = 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_0peak_set_similarity", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_13__pyx_fuse_0peak_set_similarity, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_13__pyx_fuse_0peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_precision
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0peak_set_similarity (wrapper)", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_precision,0};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0peak_set_similarity", 0, 2, 3, 1); 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_precision);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0peak_set_similarity") < 0)) __PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[1]);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_precision == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_k__7;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0peak_set_similarity", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_0peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peak_set_a", 0))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peak_set_b", 0))) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_r = 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_12__pyx_fuse_0peak_set_similarity(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_precision);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_12__pyx_fuse_0peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, int __pyx_v_precision) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity __pyx_t_2;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0peak_set_similarity", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_0peak_set_similarity (wrapper)", __pyx_f[0], 127, 0, __PYX_ERR(0, 127, __pyx_L1_error));

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.precision = __pyx_v_precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_3;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_0peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_15__pyx_fuse_1peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_1peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity *__pyx_optional_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_precision = __pyx_k__8;
  double __pyx_v_n_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_n_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_n_ab
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_z
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_hi
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_scaler
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double *__pyx_v_bin_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double *__pyx_v_bin_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_i
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED long __pyx_v_n
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_k
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_index
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_n_peaks_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_n_peaks_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject **__pyx_v_items_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject **__pyx_v_items_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_p_peak
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_2;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_3;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_4;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_5;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_6;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_7;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_8;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_9;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_10;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1peak_set_similarity", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_1peak_set_similarity", __pyx_f[0], 127, 0, __PYX_ERR(0, 127, __pyx_L1_error));

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_optional_args->precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":160
 *         PyObject* p_peak
 * 
 *     n_a = 0             # <<<<<<<<<<<<<<
 *     n_b = 0
 *     a = 0
 */

#line 160 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_a = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":161
 * 
 *     n_a = 0
 *     n_b = 0             # <<<<<<<<<<<<<<
 *     a = 0
 *     b = 0
 */

#line 161 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_b = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":162
 *     n_a = 0
 *     n_b = 0
 *     a = 0             # <<<<<<<<<<<<<<
 *     b = 0
 *     z = 0
 */

#line 162 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_a = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":163
 *     n_b = 0
 *     a = 0
 *     b = 0             # <<<<<<<<<<<<<<
 *     z = 0
 * 
 */

#line 163 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_b = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":164
 *     a = 0
 *     b = 0
 *     z = 0             # <<<<<<<<<<<<<<
 * 
 *     # Determine maximum value
 */

#line 164 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_z = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":167
 * 
 *     # Determine maximum value
 *     hi = 0             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakIndex:
 *         peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
 */

#line 167 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":169
 *     hi = 0
 *     if peak_collection is PeakIndex:
 *         peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)             # <<<<<<<<<<<<<<
 *         hi = peak.mz
 *         peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)
 */

#line 169 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->peaks->__pyx_vtab)->getitem(__pyx_v_peak_set_a->peaks, (((struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a) - 1))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)

#line 169 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 169 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 169 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":170
 *     if peak_collection is PeakIndex:
 *         peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
 *         hi = peak.mz             # <<<<<<<<<<<<<<
 *         peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)
 */

#line 170 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_peak->mz;

#line 170 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":171
 *         peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
 *         hi = peak.mz
 *         peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)             # <<<<<<<<<<<<<<
 *         hi = max(peak.mz, hi)
 *     elif peak_collection is object:
 */

#line 171 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->peaks->__pyx_vtab)->getitem(__pyx_v_peak_set_b->peaks, (((struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *)__pyx_v_peak_set_b->__pyx_vtab)->get_size(__pyx_v_peak_set_b) - 1))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 171 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 171 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":172
 *         hi = peak.mz
 *         peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)             # <<<<<<<<<<<<<<
 *     elif peak_collection is object:
 *         hi = 0
 */

#line 172 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_hi;

#line 172 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_v_peak->mz;

#line 172 "ms_deisotope/_c/similarity_methods.pyx"
  if (((__pyx_t_2 > __pyx_t_3) != 0)) {

#line 172 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_2;

#line 172 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 172 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_3;

#line 172 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 172 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_4;

  /* "ms_deisotope/_c/similarity_methods.pyx":195
 * 
 *     # Calculate bin array size and allocate memory
 *     scaler = 10 ** precision             # <<<<<<<<<<<<<<
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))
 */

#line 195 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_scaler = __Pyx_pow_long(10, ((long)__pyx_v_precision));

  /* "ms_deisotope/_c/similarity_methods.pyx":196
 *     # Calculate bin array size and allocate memory
 *     scaler = 10 ** precision
 *     k = <long>((hi * scaler) + 1)             # <<<<<<<<<<<<<<
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 */

#line 196 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_k = ((long)((__pyx_v_hi * __pyx_v_scaler) + 1.0));

  /* "ms_deisotope/_c/similarity_methods.pyx":197
 *     scaler = 10 ** precision
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))             # <<<<<<<<<<<<<<
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 */

#line 197 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_a = ((double *)malloc((__pyx_v_k * (sizeof(double)))));

  /* "ms_deisotope/_c/similarity_methods.pyx":198
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))             # <<<<<<<<<<<<<<
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:
 */

#line 198 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_b = ((double *)malloc((__pyx_v_k * (sizeof(double)))));

  /* "ms_deisotope/_c/similarity_methods.pyx":199
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:             # <<<<<<<<<<<<<<
 *         if bin_a != NULL:
 *             free(bin_a)
 */

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = ((__pyx_v_bin_a == NULL) != 0);

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  if (!__pyx_t_6) {

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L4_bool_binop_done;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = ((__pyx_v_bin_b == NULL) != 0);

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = __pyx_t_6;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_bool_binop_done:;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/similarity_methods.pyx":200
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_a)
 *         if bin_b != NULL:
 */

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_bin_a != NULL) != 0);

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":201
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:
 *             free(bin_a)             # <<<<<<<<<<<<<<
 *         if bin_b != NULL:
 *             free(bin_b)
 */

#line 201 "ms_deisotope/_c/similarity_methods.pyx"
      free(__pyx_v_bin_a);

      /* "ms_deisotope/_c/similarity_methods.pyx":200
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_a)
 *         if bin_b != NULL:
 */

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":202
 *         if bin_a != NULL:
 *             free(bin_a)
 *         if bin_b != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 */

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_bin_b != NULL) != 0);

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":203
 *             free(bin_a)
 *         if bin_b != NULL:
 *             free(bin_b)             # <<<<<<<<<<<<<<
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1
 */

#line 203 "ms_deisotope/_c/similarity_methods.pyx"
      free(__pyx_v_bin_b);

      /* "ms_deisotope/_c/similarity_methods.pyx":202
 *         if bin_a != NULL:
 *             free(bin_a)
 *         if bin_b != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 */

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":204
 *         if bin_b != NULL:
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")             # <<<<<<<<<<<<<<
 *         return -1
 *     for i in range(k):
 */

#line 204 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(__pyx_builtin_MemoryError, ((char *)"Could not allocate bins"));

    /* "ms_deisotope/_c/similarity_methods.pyx":205
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1             # <<<<<<<<<<<<<<
 *     for i in range(k):
 *         bin_a[i] = bin_b[i] = 0
 */

#line 205 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = -1.0;

#line 205 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/similarity_methods.pyx":199
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:             # <<<<<<<<<<<<<<
 *         if bin_a != NULL:
 *             free(bin_a)
 */

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":206
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1
 *     for i in range(k):             # <<<<<<<<<<<<<<
 *         bin_a[i] = bin_b[i] = 0
 * 
 */

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_7 = __pyx_v_k;

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = __pyx_t_7;

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_9;

    /* "ms_deisotope/_c/similarity_methods.pyx":207
 *         return -1
 *     for i in range(k):
 *         bin_a[i] = bin_b[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     # Fill bins
 */

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
    (__pyx_v_bin_a[__pyx_v_i]) = 0.0;

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
    (__pyx_v_bin_b[__pyx_v_i]) = 0.0;

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":214
 *         n_peaks_a = len(_hold_a)
 *     else:
 *         n_peaks_a = peak_set_a.get_size()             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 214 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_peaks_a = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":220
 *         n_peaks_b = len(_hold_b)
 *     else:
 *         n_peaks_b = peak_set_b.get_size()             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 220 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_peaks_b = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *)__pyx_v_peak_set_b->__pyx_vtab)->get_size(__pyx_v_peak_set_b);

  /* "ms_deisotope/_c/similarity_methods.pyx":226
 *         items_b = _make_fast_array(_hold_b)
 *     else:
 *         items_a = _make_fast_array(peak_set_a)             # <<<<<<<<<<<<<<
 *         items_b = _make_fast_array(peak_set_b)
 *     with nogil:
 */

#line 226 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items_a = __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":227
 *     else:
 *         items_a = _make_fast_array(peak_set_a)
 *         items_b = _make_fast_array(peak_set_b)             # <<<<<<<<<<<<<<
 *     with nogil:
 *         # Fill bins
 */

#line 227 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items_b = __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(__pyx_v_peak_set_b);

  /* "ms_deisotope/_c/similarity_methods.pyx":228
 *         items_a = _make_fast_array(peak_set_a)
 *         items_b = _make_fast_array(peak_set_b)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # Fill bins
 *         for i in range(n_peaks_a):
 */

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      #ifdef WITH_THREAD

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      PyThreadState *_save;

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      Py_UNBLOCK_THREADS

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_FastGIL_Remember();

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      /*try:*/ {

        /* "ms_deisotope/_c/similarity_methods.pyx":230
 *     with nogil:
 *         # Fill bins
 *         for i in range(n_peaks_a):             # <<<<<<<<<<<<<<
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 */

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_7 = __pyx_v_n_peaks_a;

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_8 = __pyx_t_7;

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_i = __pyx_t_9;

          /* "ms_deisotope/_c/similarity_methods.pyx":231
 *         # Fill bins
 *         for i in range(n_peaks_a):
 *             p_peak = items_a[i]             # <<<<<<<<<<<<<<
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 */

#line 231 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_p_peak = (__pyx_v_items_a[__pyx_v_i]);

          /* "ms_deisotope/_c/similarity_methods.pyx":232
 *         for i in range(n_peaks_a):
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)             # <<<<<<<<<<<<<<
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 * 
 */

#line 232 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_index = ((long)(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->mz * __pyx_v_scaler));

          /* "ms_deisotope/_c/similarity_methods.pyx":233
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_a[index] += (<PeakBase>p_peak).intensity             # <<<<<<<<<<<<<<
 * 
 *         for i in range(n_peaks_b):
 */

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_10 = __pyx_v_index;

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
          (__pyx_v_bin_a[__pyx_t_10]) = ((__pyx_v_bin_a[__pyx_t_10]) + ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->intensity);

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":235
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 * 
 *         for i in range(n_peaks_b):             # <<<<<<<<<<<<<<
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 */

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_7 = __pyx_v_n_peaks_b;

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_8 = __pyx_t_7;

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_i = __pyx_t_9;

          /* "ms_deisotope/_c/similarity_methods.pyx":236
 * 
 *         for i in range(n_peaks_b):
 *             p_peak = items_b[i]             # <<<<<<<<<<<<<<
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_b[index] += (<PeakBase>p_peak).intensity
 */

#line 236 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_p_peak = (__pyx_v_items_b[__pyx_v_i]);

          /* "ms_deisotope/_c/similarity_methods.pyx":237
 *         for i in range(n_peaks_b):
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)             # <<<<<<<<<<<<<<
 *             bin_b[index] += (<PeakBase>p_peak).intensity
 * 
 */

#line 237 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_index = ((long)(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->mz * __pyx_v_scaler));

          /* "ms_deisotope/_c/similarity_methods.pyx":238
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_b[index] += (<PeakBase>p_peak).intensity             # <<<<<<<<<<<<<<
 * 
 *         # Calculate dot product and normalizers in parallel
 */

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_10 = __pyx_v_index;

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
          (__pyx_v_bin_b[__pyx_t_10]) = ((__pyx_v_bin_b[__pyx_t_10]) + ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->intensity);

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":241
 * 
 *         # Calculate dot product and normalizers in parallel
 *         n = k             # <<<<<<<<<<<<<<
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]
 */

#line 241 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_n = __pyx_v_k;

        /* "ms_deisotope/_c/similarity_methods.pyx":242
 *         # Calculate dot product and normalizers in parallel
 *         n = k
 *         for i in prange(n, schedule='static', num_threads=4):             # <<<<<<<<<<<<<<
 *             a = bin_a[i]
 *             b = bin_b[i]
 */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_7 = __pyx_v_n;

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        if ((1 == 0)) abort();

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #undef likely

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #undef unlikely

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #define likely(x)   (x)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #define unlikely(x) (x)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            #endif

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            __pyx_t_9 = (__pyx_t_7 - 0 + 1 - 1/abs(1)) / 1;

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            if (__pyx_t_9 > 0)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #ifdef _OPENMP
                #pragma omp parallel reduction(+:__pyx_v_n_a) reduction(+:__pyx_v_n_b) reduction(+:__pyx_v_z) num_threads(4)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"


#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #endif /* _OPENMP */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_a) lastprivate(__pyx_v_b) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) schedule(static)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"


#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                    #endif /* _OPENMP */
                    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_9; __pyx_t_8++)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"
{

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                        {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_i = (long)(0 + 1 * __pyx_t_8);

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            /* Initialize private variables to invalid values */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_a = ((double)__PYX_NAN());

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_b = ((double)__PYX_NAN());

                            /* "ms_deisotope/_c/similarity_methods.pyx":243
 *         n = k
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]             # <<<<<<<<<<<<<<
 *             b = bin_b[i]
 *             n_a += a * a
 */

#line 243 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_a = (__pyx_v_bin_a[__pyx_v_i]);

                            /* "ms_deisotope/_c/similarity_methods.pyx":244
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]
 *             b = bin_b[i]             # <<<<<<<<<<<<<<
 *             n_a += a * a
 *             n_b += b * b
 */

#line 244 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_b = (__pyx_v_bin_b[__pyx_v_i]);

                            /* "ms_deisotope/_c/similarity_methods.pyx":245
 *             a = bin_a[i]
 *             b = bin_b[i]
 *             n_a += a * a             # <<<<<<<<<<<<<<
 *             n_b += b * b
 *             z += a * b
 */

#line 245 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_n_a = (__pyx_v_n_a + (__pyx_v_a * __pyx_v_a));

                            /* "ms_deisotope/_c/similarity_methods.pyx":246
 *             b = bin_b[i]
 *             n_a += a * a
 *             n_b += b * b             # <<<<<<<<<<<<<<
 *             z += a * b
 * 
 */

#line 246 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_n_b = (__pyx_v_n_b + (__pyx_v_b * __pyx_v_b));

                            /* "ms_deisotope/_c/similarity_methods.pyx":247
 *             n_a += a * a
 *             n_b += b * b
 *             z += a * b             # <<<<<<<<<<<<<<
 * 
 *         free(bin_a)
 */

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_z = (__pyx_v_z + (__pyx_v_a * __pyx_v_b));

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                        }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                    }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #undef likely

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #undef unlikely

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #define likely(x)   __builtin_expect(!!(x), 1)

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #define unlikely(x) __builtin_expect(!!(x), 0)

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

        /* "ms_deisotope/_c/similarity_methods.pyx":249
 *             z += a * b
 * 
 *         free(bin_a)             # <<<<<<<<<<<<<<
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 */

#line 249 "ms_deisotope/_c/similarity_methods.pyx"
        free(__pyx_v_bin_a);

        /* "ms_deisotope/_c/similarity_methods.pyx":250
 * 
 *         free(bin_a)
 *         free(bin_b)             # <<<<<<<<<<<<<<
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:
 */

#line 250 "ms_deisotope/_c/similarity_methods.pyx"
        free(__pyx_v_bin_b);

        /* "ms_deisotope/_c/similarity_methods.pyx":251
 *         free(bin_a)
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)             # <<<<<<<<<<<<<<
 *         if n_ab == 0:
 *             return 0.0
 */

#line 251 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_n_ab = (sqrt(__pyx_v_n_a) * sqrt(__pyx_v_n_b));

        /* "ms_deisotope/_c/similarity_methods.pyx":252
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 *         return z / n_ab
 */

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_5 = ((__pyx_v_n_ab == 0.0) != 0);

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_5) {

          /* "ms_deisotope/_c/similarity_methods.pyx":253
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:
 *             return 0.0             # <<<<<<<<<<<<<<
 *         return z / n_ab
 * 
 */

#line 253 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_r = 0.0;

#line 253 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L10_return;

          /* "ms_deisotope/_c/similarity_methods.pyx":252
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 *         return z / n_ab
 */

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":254
 *         if n_ab == 0:
 *             return 0.0
 *         return z / n_ab             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_r = (__pyx_v_z / __pyx_v_n_ab);

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L10_return;

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":228
 *         items_a = _make_fast_array(peak_set_a)
 *         items_b = _make_fast_array(peak_set_b)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # Fill bins
 *         for i in range(n_peaks_a):
 */

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      /*finally:*/ {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_L10_return: {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          #ifdef WITH_THREAD

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_FastGIL_Forget();

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          Py_BLOCK_THREADS

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          #endif

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L0;

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":127
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cpdef double peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0) except -1:             # <<<<<<<<<<<<<<
 *     """Computes the normalized dot product, also called cosine similarity between
 *     two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).
 */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1.0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_15__pyx_fuse_1peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_15__pyx_fuse_1peak_set_similarity = 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_1peak_set_similarity", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_15__pyx_fuse_1peak_set_similarity, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_15__pyx_fuse_1peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_precision
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1peak_set_similarity (wrapper)", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_precision,0};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1peak_set_similarity", 0, 2, 3, 1); 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_precision);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1peak_set_similarity") < 0)) __PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[0]);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[1]);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_precision == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_k__8;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1peak_set_similarity", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_1peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peak_set_a", 0))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peak_set_b", 0))) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_r = 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_14__pyx_fuse_1peak_set_similarity(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_precision);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_14__pyx_fuse_1peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, int __pyx_v_precision) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity __pyx_t_2;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1peak_set_similarity", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_1peak_set_similarity (wrapper)", __pyx_f[0], 127, 0, __PYX_ERR(0, 127, __pyx_L1_error));

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.precision = __pyx_v_precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_3;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_1peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17__pyx_fuse_2peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_1peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity *__pyx_optional_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_precision = __pyx_k__9;
  double __pyx_v_n_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_n_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_n_ab
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_z
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_hi
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_scaler
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double *__pyx_v_bin_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double *__pyx_v_bin_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_i
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED long __pyx_v_n
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_k
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_index
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_n_peaks_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_n_peaks_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject **__pyx_v_items_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject **__pyx_v_items_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_p_peak
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_2;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_3;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_4;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_5;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_6;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_7;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_8;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_9;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_10;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2peak_set_similarity", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_2peak_set_similarity", __pyx_f[0], 127, 0, __PYX_ERR(0, 127, __pyx_L1_error));

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_optional_args->precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":160
 *         PyObject* p_peak
 * 
 *     n_a = 0             # <<<<<<<<<<<<<<
 *     n_b = 0
 *     a = 0
 */

#line 160 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_a = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":161
 * 
 *     n_a = 0
 *     n_b = 0             # <<<<<<<<<<<<<<
 *     a = 0
 *     b = 0
 */

#line 161 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_b = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":162
 *     n_a = 0
 *     n_b = 0
 *     a = 0             # <<<<<<<<<<<<<<
 *     b = 0
 *     z = 0
 */

#line 162 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_a = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":163
 *     n_b = 0
 *     a = 0
 *     b = 0             # <<<<<<<<<<<<<<
 *     z = 0
 * 
 */

#line 163 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_b = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":164
 *     a = 0
 *     b = 0
 *     z = 0             # <<<<<<<<<<<<<<
 * 
 *     # Determine maximum value
 */

#line 164 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_z = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":167
 * 
 *     # Determine maximum value
 *     hi = 0             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakIndex:
 *         peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
 */

#line 167 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":189
 *         hi = max(peak.mz, hi)
 *     elif peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         peak = peak_set_a._mz_ordered[-1]             # <<<<<<<<<<<<<<
 *         hi = peak.mz
 *         peak = peak_set_b._mz_ordered[-1]
 */

#line 189 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_peak_set_a->_mz_ordered == Py_None)) {

#line 189 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 189 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 189, __pyx_L1_error)

#line 189 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 189 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_peak_set_a->_mz_ordered, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)

#line 189 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 189 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 189, __pyx_L1_error)

#line 189 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 189 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":190
 *     elif peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         peak = peak_set_a._mz_ordered[-1]
 *         hi = peak.mz             # <<<<<<<<<<<<<<
 *         peak = peak_set_b._mz_ordered[-1]
 *         hi = max(peak.mz, hi)
 */

#line 190 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_peak->mz;

#line 190 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":191
 *         peak = peak_set_a._mz_ordered[-1]
 *         hi = peak.mz
 *         peak = peak_set_b._mz_ordered[-1]             # <<<<<<<<<<<<<<
 *         hi = max(peak.mz, hi)
 * 
 */

#line 191 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_peak_set_b->_mz_ordered == Py_None)) {

#line 191 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 191 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 191, __pyx_L1_error)

#line 191 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 191 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_peak_set_b->_mz_ordered, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)

#line 191 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 191 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 191, __pyx_L1_error)

#line 191 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 191 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":192
 *         hi = peak.mz
 *         peak = peak_set_b._mz_ordered[-1]
 *         hi = max(peak.mz, hi)             # <<<<<<<<<<<<<<
 * 
 *     # Calculate bin array size and allocate memory
 */

#line 192 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_hi;

#line 192 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_v_peak->mz;

#line 192 "ms_deisotope/_c/similarity_methods.pyx"
  if (((__pyx_t_2 > __pyx_t_3) != 0)) {

#line 192 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_2;

#line 192 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 192 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_3;

#line 192 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 192 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_4;

  /* "ms_deisotope/_c/similarity_methods.pyx":195
 * 
 *     # Calculate bin array size and allocate memory
 *     scaler = 10 ** precision             # <<<<<<<<<<<<<<
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))
 */

#line 195 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_scaler = __Pyx_pow_long(10, ((long)__pyx_v_precision));

  /* "ms_deisotope/_c/similarity_methods.pyx":196
 *     # Calculate bin array size and allocate memory
 *     scaler = 10 ** precision
 *     k = <long>((hi * scaler) + 1)             # <<<<<<<<<<<<<<
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 */

#line 196 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_k = ((long)((__pyx_v_hi * __pyx_v_scaler) + 1.0));

  /* "ms_deisotope/_c/similarity_methods.pyx":197
 *     scaler = 10 ** precision
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))             # <<<<<<<<<<<<<<
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 */

#line 197 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_a = ((double *)malloc((__pyx_v_k * (sizeof(double)))));

  /* "ms_deisotope/_c/similarity_methods.pyx":198
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))             # <<<<<<<<<<<<<<
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:
 */

#line 198 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_b = ((double *)malloc((__pyx_v_k * (sizeof(double)))));

  /* "ms_deisotope/_c/similarity_methods.pyx":199
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:             # <<<<<<<<<<<<<<
 *         if bin_a != NULL:
 *             free(bin_a)
 */

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = ((__pyx_v_bin_a == NULL) != 0);

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  if (!__pyx_t_6) {

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L4_bool_binop_done;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = ((__pyx_v_bin_b == NULL) != 0);

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = __pyx_t_6;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_bool_binop_done:;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/similarity_methods.pyx":200
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_a)
 *         if bin_b != NULL:
 */

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_bin_a != NULL) != 0);

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":201
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:
 *             free(bin_a)             # <<<<<<<<<<<<<<
 *         if bin_b != NULL:
 *             free(bin_b)
 */

#line 201 "ms_deisotope/_c/similarity_methods.pyx"
      free(__pyx_v_bin_a);

      /* "ms_deisotope/_c/similarity_methods.pyx":200
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_a)
 *         if bin_b != NULL:
 */

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":202
 *         if bin_a != NULL:
 *             free(bin_a)
 *         if bin_b != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 */

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_bin_b != NULL) != 0);

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":203
 *             free(bin_a)
 *         if bin_b != NULL:
 *             free(bin_b)             # <<<<<<<<<<<<<<
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1
 */

#line 203 "ms_deisotope/_c/similarity_methods.pyx"
      free(__pyx_v_bin_b);

      /* "ms_deisotope/_c/similarity_methods.pyx":202
 *         if bin_a != NULL:
 *             free(bin_a)
 *         if bin_b != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 */

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":204
 *         if bin_b != NULL:
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")             # <<<<<<<<<<<<<<
 *         return -1
 *     for i in range(k):
 */

#line 204 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(__pyx_builtin_MemoryError, ((char *)"Could not allocate bins"));

    /* "ms_deisotope/_c/similarity_methods.pyx":205
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1             # <<<<<<<<<<<<<<
 *     for i in range(k):
 *         bin_a[i] = bin_b[i] = 0
 */

#line 205 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = -1.0;

#line 205 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/similarity_methods.pyx":199
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:             # <<<<<<<<<<<<<<
 *         if bin_a != NULL:
 *             free(bin_a)
 */

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":206
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1
 *     for i in range(k):             # <<<<<<<<<<<<<<
 *         bin_a[i] = bin_b[i] = 0
 * 
 */

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_7 = __pyx_v_k;

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = __pyx_t_7;

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_9;

    /* "ms_deisotope/_c/similarity_methods.pyx":207
 *         return -1
 *     for i in range(k):
 *         bin_a[i] = bin_b[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     # Fill bins
 */

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
    (__pyx_v_bin_a[__pyx_v_i]) = 0.0;

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
    (__pyx_v_bin_b[__pyx_v_i]) = 0.0;

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":214
 *         n_peaks_a = len(_hold_a)
 *     else:
 *         n_peaks_a = peak_set_a.get_size()             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 214 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_peaks_a = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":220
 *         n_peaks_b = len(_hold_b)
 *     else:
 *         n_peaks_b = peak_set_b.get_size()             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 220 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_peaks_b = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->get_size(__pyx_v_peak_set_b);

  /* "ms_deisotope/_c/similarity_methods.pyx":226
 *         items_b = _make_fast_array(_hold_b)
 *     else:
 *         items_a = _make_fast_array(peak_set_a)             # <<<<<<<<<<<<<<
 *         items_b = _make_fast_array(peak_set_b)
 *     with nogil:
 */

#line 226 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items_a = __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":227
 *     else:
 *         items_a = _make_fast_array(peak_set_a)
 *         items_b = _make_fast_array(peak_set_b)             # <<<<<<<<<<<<<<
 *     with nogil:
 *         # Fill bins
 */

#line 227 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items_b = __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(__pyx_v_peak_set_b);

  /* "ms_deisotope/_c/similarity_methods.pyx":228
 *         items_a = _make_fast_array(peak_set_a)
 *         items_b = _make_fast_array(peak_set_b)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # Fill bins
 *         for i in range(n_peaks_a):
 */

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      #ifdef WITH_THREAD

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      PyThreadState *_save;

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      Py_UNBLOCK_THREADS

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_FastGIL_Remember();

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      /*try:*/ {

        /* "ms_deisotope/_c/similarity_methods.pyx":230
 *     with nogil:
 *         # Fill bins
 *         for i in range(n_peaks_a):             # <<<<<<<<<<<<<<
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 */

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_7 = __pyx_v_n_peaks_a;

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_8 = __pyx_t_7;

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_i = __pyx_t_9;

          /* "ms_deisotope/_c/similarity_methods.pyx":231
 *         # Fill bins
 *         for i in range(n_peaks_a):
 *             p_peak = items_a[i]             # <<<<<<<<<<<<<<
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 */

#line 231 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_p_peak = (__pyx_v_items_a[__pyx_v_i]);

          /* "ms_deisotope/_c/similarity_methods.pyx":232
 *         for i in range(n_peaks_a):
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)             # <<<<<<<<<<<<<<
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 * 
 */

#line 232 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_index = ((long)(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->mz * __pyx_v_scaler));

          /* "ms_deisotope/_c/similarity_methods.pyx":233
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_a[index] += (<PeakBase>p_peak).intensity             # <<<<<<<<<<<<<<
 * 
 *         for i in range(n_peaks_b):
 */

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_10 = __pyx_v_index;

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
          (__pyx_v_bin_a[__pyx_t_10]) = ((__pyx_v_bin_a[__pyx_t_10]) + ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->intensity);

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":235
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 * 
 *         for i in range(n_peaks_b):             # <<<<<<<<<<<<<<
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 */

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_7 = __pyx_v_n_peaks_b;

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_8 = __pyx_t_7;

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_i = __pyx_t_9;

          /* "ms_deisotope/_c/similarity_methods.pyx":236
 * 
 *         for i in range(n_peaks_b):
 *             p_peak = items_b[i]             # <<<<<<<<<<<<<<
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_b[index] += (<PeakBase>p_peak).intensity
 */

#line 236 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_p_peak = (__pyx_v_items_b[__pyx_v_i]);

          /* "ms_deisotope/_c/similarity_methods.pyx":237
 *         for i in range(n_peaks_b):
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)             # <<<<<<<<<<<<<<
 *             bin_b[index] += (<PeakBase>p_peak).intensity
 * 
 */

#line 237 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_index = ((long)(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->mz * __pyx_v_scaler));

          /* "ms_deisotope/_c/similarity_methods.pyx":238
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_b[index] += (<PeakBase>p_peak).intensity             # <<<<<<<<<<<<<<
 * 
 *         # Calculate dot product and normalizers in parallel
 */

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_10 = __pyx_v_index;

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
          (__pyx_v_bin_b[__pyx_t_10]) = ((__pyx_v_bin_b[__pyx_t_10]) + ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->intensity);

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":241
 * 
 *         # Calculate dot product and normalizers in parallel
 *         n = k             # <<<<<<<<<<<<<<
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]
 */

#line 241 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_n = __pyx_v_k;

        /* "ms_deisotope/_c/similarity_methods.pyx":242
 *         # Calculate dot product and normalizers in parallel
 *         n = k
 *         for i in prange(n, schedule='static', num_threads=4):             # <<<<<<<<<<<<<<
 *             a = bin_a[i]
 *             b = bin_b[i]
 */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_7 = __pyx_v_n;

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        if ((1 == 0)) abort();

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #undef likely

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #undef unlikely

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #define likely(x)   (x)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #define unlikely(x) (x)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            #endif

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            __pyx_t_9 = (__pyx_t_7 - 0 + 1 - 1/abs(1)) / 1;

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            if (__pyx_t_9 > 0)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #ifdef _OPENMP
                #pragma omp parallel reduction(+:__pyx_v_n_a) reduction(+:__pyx_v_n_b) reduction(+:__pyx_v_z) num_threads(4)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"


#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #endif /* _OPENMP */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_a) lastprivate(__pyx_v_b) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) schedule(static)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"


#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                    #endif /* _OPENMP */
                    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_9; __pyx_t_8++)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"
{

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                        {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_i = (long)(0 + 1 * __pyx_t_8);

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            /* Initialize private variables to invalid values */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_a = ((double)__PYX_NAN());

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_b = ((double)__PYX_NAN());

                            /* "ms_deisotope/_c/similarity_methods.pyx":243
 *         n = k
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]             # <<<<<<<<<<<<<<
 *             b = bin_b[i]
 *             n_a += a * a
 */

#line 243 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_a = (__pyx_v_bin_a[__pyx_v_i]);

                            /* "ms_deisotope/_c/similarity_methods.pyx":244
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]
 *             b = bin_b[i]             # <<<<<<<<<<<<<<
 *             n_a += a * a
 *             n_b += b * b
 */

#line 244 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_b = (__pyx_v_bin_b[__pyx_v_i]);

                            /* "ms_deisotope/_c/similarity_methods.pyx":245
 *             a = bin_a[i]
 *             b = bin_b[i]
 *             n_a += a * a             # <<<<<<<<<<<<<<
 *             n_b += b * b
 *             z += a * b
 */

#line 245 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_n_a = (__pyx_v_n_a + (__pyx_v_a * __pyx_v_a));

                            /* "ms_deisotope/_c/similarity_methods.pyx":246
 *             b = bin_b[i]
 *             n_a += a * a
 *             n_b += b * b             # <<<<<<<<<<<<<<
 *             z += a * b
 * 
 */

#line 246 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_n_b = (__pyx_v_n_b + (__pyx_v_b * __pyx_v_b));

                            /* "ms_deisotope/_c/similarity_methods.pyx":247
 *             n_a += a * a
 *             n_b += b * b
 *             z += a * b             # <<<<<<<<<<<<<<
 * 
 *         free(bin_a)
 */

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_z = (__pyx_v_z + (__pyx_v_a * __pyx_v_b));

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                        }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                    }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #undef likely

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #undef unlikely

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #define likely(x)   __builtin_expect(!!(x), 1)

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #define unlikely(x) __builtin_expect(!!(x), 0)

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

        /* "ms_deisotope/_c/similarity_methods.pyx":249
 *             z += a * b
 * 
 *         free(bin_a)             # <<<<<<<<<<<<<<
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 */

#line 249 "ms_deisotope/_c/similarity_methods.pyx"
        free(__pyx_v_bin_a);

        /* "ms_deisotope/_c/similarity_methods.pyx":250
 * 
 *         free(bin_a)
 *         free(bin_b)             # <<<<<<<<<<<<<<
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:
 */

#line 250 "ms_deisotope/_c/similarity_methods.pyx"
        free(__pyx_v_bin_b);

        /* "ms_deisotope/_c/similarity_methods.pyx":251
 *         free(bin_a)
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)             # <<<<<<<<<<<<<<
 *         if n_ab == 0:
 *             return 0.0
 */

#line 251 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_n_ab = (sqrt(__pyx_v_n_a) * sqrt(__pyx_v_n_b));

        /* "ms_deisotope/_c/similarity_methods.pyx":252
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 *         return z / n_ab
 */

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_5 = ((__pyx_v_n_ab == 0.0) != 0);

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_5) {

          /* "ms_deisotope/_c/similarity_methods.pyx":253
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:
 *             return 0.0             # <<<<<<<<<<<<<<
 *         return z / n_ab
 * 
 */

#line 253 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_r = 0.0;

#line 253 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L10_return;

          /* "ms_deisotope/_c/similarity_methods.pyx":252
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 *         return z / n_ab
 */

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":254
 *         if n_ab == 0:
 *             return 0.0
 *         return z / n_ab             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_r = (__pyx_v_z / __pyx_v_n_ab);

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L10_return;

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":228
 *         items_a = _make_fast_array(peak_set_a)
 *         items_b = _make_fast_array(peak_set_b)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # Fill bins
 *         for i in range(n_peaks_a):
 */

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      /*finally:*/ {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_L10_return: {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          #ifdef WITH_THREAD

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_FastGIL_Forget();

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          Py_BLOCK_THREADS

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          #endif

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L0;

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":127
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cpdef double peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0) except -1:             # <<<<<<<<<<<<<<
 *     """Computes the normalized dot product, also called cosine similarity between
 *     two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).
 */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1.0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17__pyx_fuse_2peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_17__pyx_fuse_2peak_set_similarity = 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_2peak_set_similarity", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17__pyx_fuse_2peak_set_similarity, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17__pyx_fuse_2peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_precision
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2peak_set_similarity (wrapper)", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_precision,0};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2peak_set_similarity", 0, 2, 3, 1); 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_precision);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2peak_set_similarity") < 0)) __PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)values[0]);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)values[1]);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_precision == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_k__9;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2peak_set_similarity", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_2peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet, 1, "peak_set_a", 0))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet, 1, "peak_set_b", 0))) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_r = 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_16__pyx_fuse_2peak_set_similarity(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_precision);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_16__pyx_fuse_2peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, int __pyx_v_precision) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity __pyx_t_2;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2peak_set_similarity", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_2peak_set_similarity (wrapper)", __pyx_f[0], 127, 0, __PYX_ERR(0, 127, __pyx_L1_error));

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.precision = __pyx_v_precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_3;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_2peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_19__pyx_fuse_3peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_1peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity *__pyx_optional_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_precision = __pyx_k__10;
  PyObject *__pyx_v__hold_a = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v__hold_b = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_n_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_n_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_n_ab
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_z
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_hi
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_scaler
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double *__pyx_v_bin_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  double *__pyx_v_bin_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_i
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED long __pyx_v_n
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_k
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_index
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_n_peaks_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_n_peaks_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject **__pyx_v_items_a
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject **__pyx_v_items_b
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_p_peak
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_obj = NULL
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_2;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *(*__pyx_t_3)(PyObject *);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_5;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_6;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_7;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_8;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_9;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_10;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_11;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3peak_set_similarity", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_3peak_set_similarity", __pyx_f[0], 127, 0, __PYX_ERR(0, 127, __pyx_L1_error));

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_optional_args->precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":160
 *         PyObject* p_peak
 * 
 *     n_a = 0             # <<<<<<<<<<<<<<
 *     n_b = 0
 *     a = 0
 */

#line 160 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_a = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":161
 * 
 *     n_a = 0
 *     n_b = 0             # <<<<<<<<<<<<<<
 *     a = 0
 *     b = 0
 */

#line 161 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_b = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":162
 *     n_a = 0
 *     n_b = 0
 *     a = 0             # <<<<<<<<<<<<<<
 *     b = 0
 *     z = 0
 */

#line 162 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_a = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":163
 *     n_b = 0
 *     a = 0
 *     b = 0             # <<<<<<<<<<<<<<
 *     z = 0
 * 
 */

#line 163 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_b = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":164
 *     a = 0
 *     b = 0
 *     z = 0             # <<<<<<<<<<<<<<
 * 
 *     # Determine maximum value
 */

#line 164 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_z = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":167
 * 
 *     # Determine maximum value
 *     hi = 0             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakIndex:
 *         peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
 */

#line 167 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":174
 *         hi = max(peak.mz, hi)
 *     elif peak_collection is object:
 *         hi = 0             # <<<<<<<<<<<<<<
 *         for obj in peak_set_a:
 *             peak = <PeakBase>obj
 */

#line 174 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":175
 *     elif peak_collection is object:
 *         hi = 0
 *         for obj in peak_set_a:             # <<<<<<<<<<<<<<
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 */

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
  if (likely(PyList_CheckExact(__pyx_v_peak_set_a)) || PyTuple_CheckExact(__pyx_v_peak_set_a)) {

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_peak_set_a; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = NULL;

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
  } else {
    __pyx_t_2 = -1; 
#line 175 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_1 = PyObject_GetIter(__pyx_v_peak_set_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
  for (;;) {

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(!__pyx_t_3)) {

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(PyList_CheckExact(__pyx_t_1))) {

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        #else

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        #else

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
      }
    } else 
#line 175 "ms_deisotope/_c/similarity_methods.pyx"
{

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(!__pyx_t_4)) {

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        if (exc_type) {

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
          else __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
        break;

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_4);

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":176
 *         hi = 0
 *         for obj in peak_set_a:
 *             peak = <PeakBase>obj             # <<<<<<<<<<<<<<
 *             if peak.mz > hi:
 *                 hi = peak.mz
 */

#line 176 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_v_obj;

#line 176 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_4);

#line 176 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 176 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":177
 *         for obj in peak_set_a:
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:             # <<<<<<<<<<<<<<
 *                 hi = peak.mz
 *         for obj in peak_set_b:
 */

#line 177 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_peak->mz > __pyx_v_hi) != 0);

#line 177 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":178
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 *                 hi = peak.mz             # <<<<<<<<<<<<<<
 *         for obj in peak_set_b:
 *             peak = <PeakBase>obj
 */

#line 178 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = __pyx_v_peak->mz;

#line 178 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_hi = __pyx_t_6;

      /* "ms_deisotope/_c/similarity_methods.pyx":177
 *         for obj in peak_set_a:
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:             # <<<<<<<<<<<<<<
 *                 hi = peak.mz
 *         for obj in peak_set_b:
 */

#line 177 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":175
 *     elif peak_collection is object:
 *         hi = 0
 *         for obj in peak_set_a:             # <<<<<<<<<<<<<<
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 */

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 175 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":179
 *             if peak.mz > hi:
 *                 hi = peak.mz
 *         for obj in peak_set_b:             # <<<<<<<<<<<<<<
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 */

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
  if (likely(PyList_CheckExact(__pyx_v_peak_set_b)) || PyTuple_CheckExact(__pyx_v_peak_set_b)) {

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_peak_set_b; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = NULL;

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
  } else {
    __pyx_t_2 = -1; 
#line 179 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_1 = PyObject_GetIter(__pyx_v_peak_set_b); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
  for (;;) {

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(!__pyx_t_3)) {

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(PyList_CheckExact(__pyx_t_1))) {

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        #else

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        #else

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
      }
    } else 
#line 179 "ms_deisotope/_c/similarity_methods.pyx"
{

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(!__pyx_t_4)) {

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        if (exc_type) {

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
          else __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
        break;

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_4);

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":180
 *                 hi = peak.mz
 *         for obj in peak_set_b:
 *             peak = <PeakBase>obj             # <<<<<<<<<<<<<<
 *             if peak.mz > hi:
 *                 hi = peak.mz
 */

#line 180 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_v_obj;

#line 180 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_4);

#line 180 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 180 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":181
 *         for obj in peak_set_b:
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:             # <<<<<<<<<<<<<<
 *                 hi = peak.mz
 *     elif peak_collection is PeakSet:
 */

#line 181 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_peak->mz > __pyx_v_hi) != 0);

#line 181 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":182
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 *                 hi = peak.mz             # <<<<<<<<<<<<<<
 *     elif peak_collection is PeakSet:
 *         peak = peak_set_a.getitem(peak_set_a.get_size() - 1)
 */

#line 182 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = __pyx_v_peak->mz;

#line 182 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_hi = __pyx_t_6;

      /* "ms_deisotope/_c/similarity_methods.pyx":181
 *         for obj in peak_set_b:
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:             # <<<<<<<<<<<<<<
 *                 hi = peak.mz
 *     elif peak_collection is PeakSet:
 */

#line 181 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":179
 *             if peak.mz > hi:
 *                 hi = peak.mz
 *         for obj in peak_set_b:             # <<<<<<<<<<<<<<
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 */

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 179 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":195
 * 
 *     # Calculate bin array size and allocate memory
 *     scaler = 10 ** precision             # <<<<<<<<<<<<<<
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))
 */

#line 195 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_scaler = __Pyx_pow_long(10, ((long)__pyx_v_precision));

  /* "ms_deisotope/_c/similarity_methods.pyx":196
 *     # Calculate bin array size and allocate memory
 *     scaler = 10 ** precision
 *     k = <long>((hi * scaler) + 1)             # <<<<<<<<<<<<<<
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 */

#line 196 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_k = ((long)((__pyx_v_hi * __pyx_v_scaler) + 1.0));

  /* "ms_deisotope/_c/similarity_methods.pyx":197
 *     scaler = 10 ** precision
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))             # <<<<<<<<<<<<<<
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 */

#line 197 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_a = ((double *)malloc((__pyx_v_k * (sizeof(double)))));

  /* "ms_deisotope/_c/similarity_methods.pyx":198
 *     k = <long>((hi * scaler) + 1)
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))             # <<<<<<<<<<<<<<
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:
 */

#line 198 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_b = ((double *)malloc((__pyx_v_k * (sizeof(double)))));

  /* "ms_deisotope/_c/similarity_methods.pyx":199
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:             # <<<<<<<<<<<<<<
 *         if bin_a != NULL:
 *             free(bin_a)
 */

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_7 = ((__pyx_v_bin_a == NULL) != 0);

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  if (!__pyx_t_7) {

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = __pyx_t_7;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L10_bool_binop_done;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_7 = ((__pyx_v_bin_b == NULL) != 0);

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = __pyx_t_7;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L10_bool_binop_done:;

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/similarity_methods.pyx":200
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_a)
 *         if bin_b != NULL:
 */

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_bin_a != NULL) != 0);

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":201
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:
 *             free(bin_a)             # <<<<<<<<<<<<<<
 *         if bin_b != NULL:
 *             free(bin_b)
 */

#line 201 "ms_deisotope/_c/similarity_methods.pyx"
      free(__pyx_v_bin_a);

      /* "ms_deisotope/_c/similarity_methods.pyx":200
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:
 *         if bin_a != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_a)
 *         if bin_b != NULL:
 */

#line 200 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":202
 *         if bin_a != NULL:
 *             free(bin_a)
 *         if bin_b != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 */

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_bin_b != NULL) != 0);

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":203
 *             free(bin_a)
 *         if bin_b != NULL:
 *             free(bin_b)             # <<<<<<<<<<<<<<
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1
 */

#line 203 "ms_deisotope/_c/similarity_methods.pyx"
      free(__pyx_v_bin_b);

      /* "ms_deisotope/_c/similarity_methods.pyx":202
 *         if bin_a != NULL:
 *             free(bin_a)
 *         if bin_b != NULL:             # <<<<<<<<<<<<<<
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 */

#line 202 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":204
 *         if bin_b != NULL:
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")             # <<<<<<<<<<<<<<
 *         return -1
 *     for i in range(k):
 */

#line 204 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(__pyx_builtin_MemoryError, ((char *)"Could not allocate bins"));

    /* "ms_deisotope/_c/similarity_methods.pyx":205
 *             free(bin_b)
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1             # <<<<<<<<<<<<<<
 *     for i in range(k):
 *         bin_a[i] = bin_b[i] = 0
 */

#line 205 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = -1.0;

#line 205 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/similarity_methods.pyx":199
 *     bin_a = <double*>malloc(k * sizeof(double))
 *     bin_b = <double*>malloc(k * sizeof(double))
 *     if bin_a == NULL or bin_b == NULL:             # <<<<<<<<<<<<<<
 *         if bin_a != NULL:
 *             free(bin_a)
 */

#line 199 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":206
 *         PyErr_SetString(MemoryError, "Could not allocate bins")
 *         return -1
 *     for i in range(k):             # <<<<<<<<<<<<<<
 *         bin_a[i] = bin_b[i] = 0
 * 
 */

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = __pyx_v_k;

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_9 = __pyx_t_8;

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 206 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/similarity_methods.pyx":207
 *         return -1
 *     for i in range(k):
 *         bin_a[i] = bin_b[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     # Fill bins
 */

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
    (__pyx_v_bin_a[__pyx_v_i]) = 0.0;

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
    (__pyx_v_bin_b[__pyx_v_i]) = 0.0;

#line 207 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":211
 *     # Fill bins
 *     if peak_collection is object:
 *         _hold_a = list(peak_set_a)             # <<<<<<<<<<<<<<
 *         n_peaks_a = len(_hold_a)
 *     else:
 */

#line 211 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PySequence_List(__pyx_v_peak_set_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 211, __pyx_L1_error)

#line 211 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 211 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v__hold_a = ((PyObject*)__pyx_t_1);

#line 211 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":212
 *     if peak_collection is object:
 *         _hold_a = list(peak_set_a)
 *         n_peaks_a = len(_hold_a)             # <<<<<<<<<<<<<<
 *     else:
 *         n_peaks_a = peak_set_a.get_size()
 */

#line 212 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyList_GET_SIZE(__pyx_v__hold_a); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 212, __pyx_L1_error)

#line 212 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_peaks_a = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":217
 * 
 *     if peak_collection is object:
 *         _hold_b = list(peak_set_b)             # <<<<<<<<<<<<<<
 *         n_peaks_b = len(_hold_b)
 *     else:
 */

#line 217 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PySequence_List(__pyx_v_peak_set_b); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)

#line 217 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 217 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v__hold_b = ((PyObject*)__pyx_t_1);

#line 217 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":218
 *     if peak_collection is object:
 *         _hold_b = list(peak_set_b)
 *         n_peaks_b = len(_hold_b)             # <<<<<<<<<<<<<<
 *     else:
 *         n_peaks_b = peak_set_b.get_size()
 */

#line 218 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyList_GET_SIZE(__pyx_v__hold_b); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 218, __pyx_L1_error)

#line 218 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n_peaks_b = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":223
 * 
 *     if peak_collection is object:
 *         items_a = _make_fast_array(_hold_a)             # <<<<<<<<<<<<<<
 *         items_b = _make_fast_array(_hold_b)
 *     else:
 */

#line 223 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items_a = __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(__pyx_v__hold_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":224
 *     if peak_collection is object:
 *         items_a = _make_fast_array(_hold_a)
 *         items_b = _make_fast_array(_hold_b)             # <<<<<<<<<<<<<<
 *     else:
 *         items_a = _make_fast_array(peak_set_a)
 */

#line 224 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_items_b = __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods__make_fast_array(__pyx_v__hold_b);

  /* "ms_deisotope/_c/similarity_methods.pyx":228
 *         items_a = _make_fast_array(peak_set_a)
 *         items_b = _make_fast_array(peak_set_b)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # Fill bins
 *         for i in range(n_peaks_a):
 */

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      #ifdef WITH_THREAD

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      PyThreadState *_save;

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      Py_UNBLOCK_THREADS

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_FastGIL_Remember();

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      /*try:*/ {

        /* "ms_deisotope/_c/similarity_methods.pyx":230
 *     with nogil:
 *         # Fill bins
 *         for i in range(n_peaks_a):             # <<<<<<<<<<<<<<
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 */

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_8 = __pyx_v_n_peaks_a;

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_9 = __pyx_t_8;

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
        for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 230 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_i = __pyx_t_10;

          /* "ms_deisotope/_c/similarity_methods.pyx":231
 *         # Fill bins
 *         for i in range(n_peaks_a):
 *             p_peak = items_a[i]             # <<<<<<<<<<<<<<
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 */

#line 231 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_p_peak = (__pyx_v_items_a[__pyx_v_i]);

          /* "ms_deisotope/_c/similarity_methods.pyx":232
 *         for i in range(n_peaks_a):
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)             # <<<<<<<<<<<<<<
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 * 
 */

#line 232 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_index = ((long)(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->mz * __pyx_v_scaler));

          /* "ms_deisotope/_c/similarity_methods.pyx":233
 *             p_peak = items_a[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_a[index] += (<PeakBase>p_peak).intensity             # <<<<<<<<<<<<<<
 * 
 *         for i in range(n_peaks_b):
 */

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_11 = __pyx_v_index;

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
          (__pyx_v_bin_a[__pyx_t_11]) = ((__pyx_v_bin_a[__pyx_t_11]) + ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->intensity);

#line 233 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":235
 *             bin_a[index] += (<PeakBase>p_peak).intensity
 * 
 *         for i in range(n_peaks_b):             # <<<<<<<<<<<<<<
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 */

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_8 = __pyx_v_n_peaks_b;

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_9 = __pyx_t_8;

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
        for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 235 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_i = __pyx_t_10;

          /* "ms_deisotope/_c/similarity_methods.pyx":236
 * 
 *         for i in range(n_peaks_b):
 *             p_peak = items_b[i]             # <<<<<<<<<<<<<<
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_b[index] += (<PeakBase>p_peak).intensity
 */

#line 236 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_p_peak = (__pyx_v_items_b[__pyx_v_i]);

          /* "ms_deisotope/_c/similarity_methods.pyx":237
 *         for i in range(n_peaks_b):
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)             # <<<<<<<<<<<<<<
 *             bin_b[index] += (<PeakBase>p_peak).intensity
 * 
 */

#line 237 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_index = ((long)(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->mz * __pyx_v_scaler));

          /* "ms_deisotope/_c/similarity_methods.pyx":238
 *             p_peak = items_b[i]
 *             index = int((<PeakBase>p_peak).mz * scaler)
 *             bin_b[index] += (<PeakBase>p_peak).intensity             # <<<<<<<<<<<<<<
 * 
 *         # Calculate dot product and normalizers in parallel
 */

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_11 = __pyx_v_index;

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
          (__pyx_v_bin_b[__pyx_t_11]) = ((__pyx_v_bin_b[__pyx_t_11]) + ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_p_peak)->intensity);

#line 238 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":241
 * 
 *         # Calculate dot product and normalizers in parallel
 *         n = k             # <<<<<<<<<<<<<<
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]
 */

#line 241 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_n = __pyx_v_k;

        /* "ms_deisotope/_c/similarity_methods.pyx":242
 *         # Calculate dot product and normalizers in parallel
 *         n = k
 *         for i in prange(n, schedule='static', num_threads=4):             # <<<<<<<<<<<<<<
 *             a = bin_a[i]
 *             b = bin_b[i]
 */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_8 = __pyx_v_n;

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        if ((1 == 0)) abort();

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
        {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #undef likely

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #undef unlikely

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #define likely(x)   (x)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #define unlikely(x) (x)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            #endif

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            __pyx_t_10 = (__pyx_t_8 - 0 + 1 - 1/abs(1)) / 1;

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            if (__pyx_t_10 > 0)

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
            {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #ifdef _OPENMP
                #pragma omp parallel reduction(+:__pyx_v_n_a) reduction(+:__pyx_v_n_b) reduction(+:__pyx_v_z) num_threads(4)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"


#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                #endif /* _OPENMP */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_a) lastprivate(__pyx_v_b) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) schedule(static)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"


#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                    #endif /* _OPENMP */
                    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_10; __pyx_t_9++)
#line 242 "ms_deisotope/_c/similarity_methods.pyx"
{

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                        {

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_i = (long)(0 + 1 * __pyx_t_9);

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            /* Initialize private variables to invalid values */

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_a = ((double)__PYX_NAN());

#line 242 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_b = ((double)__PYX_NAN());

                            /* "ms_deisotope/_c/similarity_methods.pyx":243
 *         n = k
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]             # <<<<<<<<<<<<<<
 *             b = bin_b[i]
 *             n_a += a * a
 */

#line 243 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_a = (__pyx_v_bin_a[__pyx_v_i]);

                            /* "ms_deisotope/_c/similarity_methods.pyx":244
 *         for i in prange(n, schedule='static', num_threads=4):
 *             a = bin_a[i]
 *             b = bin_b[i]             # <<<<<<<<<<<<<<
 *             n_a += a * a
 *             n_b += b * b
 */

#line 244 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_b = (__pyx_v_bin_b[__pyx_v_i]);

                            /* "ms_deisotope/_c/similarity_methods.pyx":245
 *             a = bin_a[i]
 *             b = bin_b[i]
 *             n_a += a * a             # <<<<<<<<<<<<<<
 *             n_b += b * b
 *             z += a * b
 */

#line 245 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_n_a = (__pyx_v_n_a + (__pyx_v_a * __pyx_v_a));

                            /* "ms_deisotope/_c/similarity_methods.pyx":246
 *             b = bin_b[i]
 *             n_a += a * a
 *             n_b += b * b             # <<<<<<<<<<<<<<
 *             z += a * b
 * 
 */

#line 246 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_n_b = (__pyx_v_n_b + (__pyx_v_b * __pyx_v_b));

                            /* "ms_deisotope/_c/similarity_methods.pyx":247
 *             n_a += a * a
 *             n_b += b * b
 *             z += a * b             # <<<<<<<<<<<<<<
 * 
 *         free(bin_a)
 */

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_z = (__pyx_v_z + (__pyx_v_a * __pyx_v_b));

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                        }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                    }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
                }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #undef likely

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #undef unlikely

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #define likely(x)   __builtin_expect(!!(x), 1)

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
            #define unlikely(x) __builtin_expect(!!(x), 0)

#line 247 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

        /* "ms_deisotope/_c/similarity_methods.pyx":249
 *             z += a * b
 * 
 *         free(bin_a)             # <<<<<<<<<<<<<<
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 */

#line 249 "ms_deisotope/_c/similarity_methods.pyx"
        free(__pyx_v_bin_a);

        /* "ms_deisotope/_c/similarity_methods.pyx":250
 * 
 *         free(bin_a)
 *         free(bin_b)             # <<<<<<<<<<<<<<
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:
 */

#line 250 "ms_deisotope/_c/similarity_methods.pyx"
        free(__pyx_v_bin_b);

        /* "ms_deisotope/_c/similarity_methods.pyx":251
 *         free(bin_a)
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)             # <<<<<<<<<<<<<<
 *         if n_ab == 0:
 *             return 0.0
 */

#line 251 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_n_ab = (sqrt(__pyx_v_n_a) * sqrt(__pyx_v_n_b));

        /* "ms_deisotope/_c/similarity_methods.pyx":252
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 *         return z / n_ab
 */

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_5 = ((__pyx_v_n_ab == 0.0) != 0);

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_5) {

          /* "ms_deisotope/_c/similarity_methods.pyx":253
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:
 *             return 0.0             # <<<<<<<<<<<<<<
 *         return z / n_ab
 * 
 */

#line 253 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_r = 0.0;

#line 253 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L16_return;

          /* "ms_deisotope/_c/similarity_methods.pyx":252
 *         free(bin_b)
 *         n_ab = sqrt(n_a) * sqrt(n_b)
 *         if n_ab == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 *         return z / n_ab
 */

#line 252 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":254
 *         if n_ab == 0:
 *             return 0.0
 *         return z / n_ab             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_r = (__pyx_v_z / __pyx_v_n_ab);

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L16_return;

#line 254 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":228
 *         items_a = _make_fast_array(peak_set_a)
 *         items_b = _make_fast_array(peak_set_b)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # Fill bins
 *         for i in range(n_peaks_a):
 */

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      /*finally:*/ {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_L16_return: {

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          #ifdef WITH_THREAD

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_FastGIL_Forget();

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          Py_BLOCK_THREADS

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          #endif

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L0;

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 228 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":127
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cpdef double peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0) except -1:             # <<<<<<<<<<<<<<
 *     """Computes the normalized dot product, also called cosine similarity between
 *     two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).
 */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1.0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v__hold_a);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v__hold_b);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_obj);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_19__pyx_fuse_3peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_19__pyx_fuse_3peak_set_similarity = 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_3peak_set_similarity", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_19__pyx_fuse_3peak_set_similarity, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_19__pyx_fuse_3peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peak_set_a = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_peak_set_b = 0
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_precision
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3peak_set_similarity (wrapper)", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_precision,0};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3peak_set_similarity", 0, 2, 3, 1); 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_precision);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3peak_set_similarity") < 0)) __PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = values[0];

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = values[1];

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_precision == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_k__10;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3peak_set_similarity", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 127, __pyx_L3_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_3peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 127 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_18__pyx_fuse_3peak_set_similarity(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_precision);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_18__pyx_fuse_3peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, int __pyx_v_precision) {

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity __pyx_t_2;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3peak_set_similarity", 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_3peak_set_similarity (wrapper)", __pyx_f[0], 127, 0, __PYX_ERR(0, 127, __pyx_L1_error));

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.precision = __pyx_v_precision;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_peak_set_similarity(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_3;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"


#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_3peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":258
 * 
 * @cython.boundscheck(False)
 * cpdef tuple bin_peaks(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double hi, scaler
 */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_3bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_2bin_peaks[] = "bin_peaks(signatures, args, kwargs, defaults)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_18similarity_methods_3bin_peaks = 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
{"bin_peaks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_3bin_peaks, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_2bin_peaks};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_3bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_signatures = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_args = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signatures,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[4] = {0,0,0,0};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    values[2] = __pyx_k__11;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  4: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signatures)) != 0)) kw_args--;
        else 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 1); 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  3:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 3); 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_signatures = values[0];

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_args = values[1];

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_kwargs = values[2];

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_defaults = values[3];

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_2bin_peaks(__pyx_self, __pyx_v_signatures, __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_2bin_peaks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  Py_ssize_t __pyx_v_i
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_arg = NULL
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_candidates = NULL
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_sig = NULL
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_match_found
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_src_sig = NULL
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_dst_type = NULL
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_3;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_4;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_5;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_7;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_8;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_9;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_13;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_14;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_15;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_16;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("bin_peaks", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("bin_peaks", __pyx_f[0], 258, 0, __PYX_ERR(0, 258, __pyx_L1_error));

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_kwargs);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_4) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L4_bool_binop_done;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((!__pyx_t_4) != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_bool_binop_done:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_itemsize = -1L;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_args == Py_None)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = ((0 < __pyx_t_5) != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L6;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_4) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L7_bool_binop_done;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_peak_set_a, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_t_4 != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L7_bool_binop_done:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_peak_set_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L6;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /*else*/ {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_int_2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_int_2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_n_s_s);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_n_s_s);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L6:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  while (1) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s__2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_peak_picker__c_peak_set);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_3) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s__2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_ms_peak_picker__c_peak_index);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PeakIndex); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = (__pyx_t_3 != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_2) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakIndex, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s__2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_deisotope__c_peak_set);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_DeconvolutedPeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_3) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_DeconvolutedPeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_object, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L10_break;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L10_break:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_candidates = ((PyObject*)__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_signatures == Py_None)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_t_6;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  while (1) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_7, &__pyx_t_5, &__pyx_t_6, NULL, NULL, __pyx_t_8);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 == 0)) break;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_match_found = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_12 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(__pyx_t_12)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_12);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(function);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__3) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__3);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_10);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(__pyx_t_10)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_10);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(function);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__4) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__4);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_14 = __pyx_t_13;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_i = __pyx_t_15;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_2 = (__pyx_t_3 != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_2) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_6, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_match_found = 1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L19;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        /*else*/ {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_match_found = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L17_break;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_L19:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_L17_break:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = (__pyx_v_match_found != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_2) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((!__pyx_t_2) != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_3) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((__pyx_t_7 > 1) != 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_3) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /*else*/ {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_signatures == Py_None)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = __pyx_t_1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_11);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_12);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_dest_sig);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_ndarray);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_arg);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_candidates);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_sig);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_src_sig);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_dst_type);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_23__pyx_fuse_0bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_3bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks *__pyx_optional_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_precision = __pyx_k__12;
  double __pyx_v_hi
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_scaler
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyArrayObject *__pyx_v_bin_a = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyArrayObject *__pyx_v_bin_b = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_i
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_n
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_index
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bin_a
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_bin_a
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bin_b
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_bin_b
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_3;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_4;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyArrayObject *__pyx_t_8 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_9;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_13;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_14;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_15;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_16;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0bin_peaks", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_0bin_peaks", __pyx_f[0], 258, 0, __PYX_ERR(0, 258, __pyx_L1_error));

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_optional_args->precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_a.pybuffer.buf = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_a.refcount = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_a.data = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_a.rcbuffer = &__pyx_pybuffer_bin_a;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_b.pybuffer.buf = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_b.refcount = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_b.data = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_b.rcbuffer = &__pyx_pybuffer_bin_b;

  /* "ms_deisotope/_c/similarity_methods.pyx":287
 *         hi = max(peak.mz, hi)
 *     else:
 *         hi = 0             # <<<<<<<<<<<<<<
 *         peak = peak_set_a.getitem(peak_set_a.get_size() - 1)
 *         hi = peak.mz
 */

#line 287 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":288
 *     else:
 *         hi = 0
 *         peak = peak_set_a.getitem(peak_set_a.get_size() - 1)             # <<<<<<<<<<<<<<
 *         hi = peak.mz
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
 */

#line 288 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->getitem(__pyx_v_peak_set_a, (((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a) - 1))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)

#line 288 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 288 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 288 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":289
 *         hi = 0
 *         peak = peak_set_a.getitem(peak_set_a.get_size() - 1)
 *         hi = peak.mz             # <<<<<<<<<<<<<<
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)
 */

#line 289 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_peak->mz;

#line 289 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":290
 *         peak = peak_set_a.getitem(peak_set_a.get_size() - 1)
 *         hi = peak.mz
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)             # <<<<<<<<<<<<<<
 *         hi = max(peak.mz, hi)
 *     scaler = 10 ** precision
 */

#line 290 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->getitem(__pyx_v_peak_set_b, (((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->get_size(__pyx_v_peak_set_b) - 1))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)

#line 290 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 290 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 290 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":291
 *         hi = peak.mz
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)             # <<<<<<<<<<<<<<
 *     scaler = 10 ** precision
 *     hi = (hi * scaler) + 1
 */

#line 291 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_hi;

#line 291 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_v_peak->mz;

#line 291 "ms_deisotope/_c/similarity_methods.pyx"
  if (((__pyx_t_2 > __pyx_t_3) != 0)) {

#line 291 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_2;

#line 291 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 291 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_3;

#line 291 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 291 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_4;

  /* "ms_deisotope/_c/similarity_methods.pyx":292
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)
 *     scaler = 10 ** precision             # <<<<<<<<<<<<<<
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)
 */

#line 292 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_scaler = __Pyx_pow_long(10, ((long)__pyx_v_precision));

  /* "ms_deisotope/_c/similarity_methods.pyx":293
 *         hi = max(peak.mz, hi)
 *     scaler = 10 ** precision
 *     hi = (hi * scaler) + 1             # <<<<<<<<<<<<<<
 *     bin_a = zeros(<int>hi)
 *     bin_b = zeros(<int>hi)
 */

#line 293 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = ((__pyx_v_hi * __pyx_v_scaler) + 1.0);

  /* "ms_deisotope/_c/similarity_methods.pyx":294
 *     scaler = 10 ** precision
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)             # <<<<<<<<<<<<<<
 *     bin_b = zeros(<int>hi)
 * 
 */

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = __Pyx_PyInt_From_int(((int)__pyx_v_hi)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros);
  __pyx_t_6 = __pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros; 
#line 294 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_7 = NULL;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(__pyx_t_7)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_7);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(function);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF_SET(__pyx_t_6, function);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer, (PyObject*)__pyx_v_bin_a, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_12);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_Restore(__pyx_t_10, __pyx_t_11, __pyx_t_12);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_10 = __pyx_t_11 = __pyx_t_12 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_pybuffernd_bin_a.diminfo[0].strides = __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bin_a.diminfo[0].shape = __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.shape[0];

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_a = ((PyArrayObject *)__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":295
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)
 *     bin_b = zeros(<int>hi)             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = __Pyx_PyInt_From_int(((int)__pyx_v_hi)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros);
  __pyx_t_5 = __pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros; 
#line 295 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_7 = NULL;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(__pyx_t_7)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_7);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(function);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF_SET(__pyx_t_5, function);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_Fetch(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer, (PyObject*)__pyx_v_bin_b, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        Py_XDECREF(__pyx_t_12); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_Restore(__pyx_t_12, __pyx_t_11, __pyx_t_10);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_12 = __pyx_t_11 = __pyx_t_10 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_pybuffernd_bin_b.diminfo[0].strides = __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bin_b.diminfo[0].shape = __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.shape[0];

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_b = ((PyArrayObject *)__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":300
 *         n = len(peak_set_a)
 *     else:
 *         n = peak_set_a.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         if peak_collection is PeakIndex:
 */

#line 300 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":301
 *     else:
 *         n = peak_set_a.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakIndex:
 *             peak = peak_set_a.peaks.getitem(i)
 */

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_13 = __pyx_v_n;

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_14 = __pyx_t_13;

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_15;

    /* "ms_deisotope/_c/similarity_methods.pyx":307
 *             peak = peak_set_a[i]
 *         else:
 *             peak = peak_set_a.getitem(i)             # <<<<<<<<<<<<<<
 *         index = int(peak.mz * scaler)
 *         bin_a[index] += peak.intensity
 */

#line 307 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->getitem(__pyx_v_peak_set_a, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)

#line 307 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 307 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 307 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":308
 *         else:
 *             peak = peak_set_a.getitem(i)
 *         index = int(peak.mz * scaler)             # <<<<<<<<<<<<<<
 *         bin_a[index] += peak.intensity
 * 
 */

#line 308 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_index = ((size_t)(__pyx_v_peak->mz * __pyx_v_scaler));

    /* "ms_deisotope/_c/similarity_methods.pyx":309
 *             peak = peak_set_a.getitem(i)
 *         index = int(peak.mz * scaler)
 *         bin_a[index] += peak.intensity             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_16 = __pyx_v_index;

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
    *__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_bin_a.diminfo[0].strides) += __pyx_v_peak->intensity;

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":314
 *         n = len(peak_set_b)
 *     else:
 *         n = peak_set_b.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         if peak_collection is PeakIndex:
 */

#line 314 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->get_size(__pyx_v_peak_set_b);

  /* "ms_deisotope/_c/similarity_methods.pyx":315
 *     else:
 *         n = peak_set_b.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakIndex:
 *             peak = peak_set_b.peaks.getitem(i)
 */

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_13 = __pyx_v_n;

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_14 = __pyx_t_13;

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_15;

    /* "ms_deisotope/_c/similarity_methods.pyx":321
 *             peak = peak_set_b[i]
 *         else:
 *             peak = peak_set_b.getitem(i)             # <<<<<<<<<<<<<<
 *         index = int(peak.mz * scaler)
 *         bin_b[index] += peak.intensity
 */

#line 321 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->getitem(__pyx_v_peak_set_b, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)

#line 321 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 321 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 321 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":322
 *         else:
 *             peak = peak_set_b.getitem(i)
 *         index = int(peak.mz * scaler)             # <<<<<<<<<<<<<<
 *         bin_b[index] += peak.intensity
 *     return (bin_a, bin_b)
 */

#line 322 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_index = ((size_t)(__pyx_v_peak->mz * __pyx_v_scaler));

    /* "ms_deisotope/_c/similarity_methods.pyx":323
 *             peak = peak_set_b.getitem(i)
 *         index = int(peak.mz * scaler)
 *         bin_b[index] += peak.intensity             # <<<<<<<<<<<<<<
 *     return (bin_a, bin_b)
 * 
 */

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_16 = __pyx_v_index;

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
    *__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_bin_b.diminfo[0].strides) += __pyx_v_peak->intensity;

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":324
 *         index = int(peak.mz * scaler)
 *         bin_b[index] += peak.intensity
 *     return (bin_a, bin_b)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = ((PyObject*)__pyx_t_1);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":258
 * 
 * @cython.boundscheck(False)
 * cpdef tuple bin_peaks(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double hi, scaler
 */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_PyThreadState_declare

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_PyThreadState_assign

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L2:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_bin_a);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_bin_b);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_23__pyx_fuse_0bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_23__pyx_fuse_0bin_peaks = 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_0bin_peaks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_23__pyx_fuse_0bin_peaks, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_2bin_peaks};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_23__pyx_fuse_0bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_precision
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0bin_peaks (wrapper)", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_precision,0};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0bin_peaks", 0, 2, 3, 1); 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_precision);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0bin_peaks") < 0)) __PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[1]);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_precision == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_k__12;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0bin_peaks", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_0bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peak_set_a", 0))) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peak_set_b", 0))) __PYX_ERR(0, 258, __pyx_L1_error)
  __pyx_r = 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_22__pyx_fuse_0bin_peaks(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_precision);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_22__pyx_fuse_0bin_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, int __pyx_v_precision) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks __pyx_t_2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0bin_peaks", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_0bin_peaks (wrapper)", __pyx_f[0], 258, 0, __PYX_ERR(0, 258, __pyx_L1_error));

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.precision = __pyx_v_precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_0bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_25__pyx_fuse_1bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_3bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks *__pyx_optional_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_precision = __pyx_k__13;
  double __pyx_v_hi
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_scaler
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyArrayObject *__pyx_v_bin_a = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyArrayObject *__pyx_v_bin_b = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_i
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_n
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_index
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bin_a
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_bin_a
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bin_b
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_bin_b
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_3;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_4;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyArrayObject *__pyx_t_8 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_9;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_13;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_14;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_15;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_16;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1bin_peaks", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_1bin_peaks", __pyx_f[0], 258, 0, __PYX_ERR(0, 258, __pyx_L1_error));

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_optional_args->precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_a.pybuffer.buf = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_a.refcount = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_a.data = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_a.rcbuffer = &__pyx_pybuffer_bin_a;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_b.pybuffer.buf = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_b.refcount = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_b.data = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_b.rcbuffer = &__pyx_pybuffer_bin_b;

  /* "ms_deisotope/_c/similarity_methods.pyx":266
 * 
 *     if peak_collection is PeakIndex:
 *         hi = 0             # <<<<<<<<<<<<<<
 *         peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
 *         hi = peak.mz
 */

#line 266 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":267
 *     if peak_collection is PeakIndex:
 *         hi = 0
 *         peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)             # <<<<<<<<<<<<<<
 *         hi = peak.mz
 *         peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)
 */

#line 267 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->peaks->__pyx_vtab)->getitem(__pyx_v_peak_set_a->peaks, (((struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a) - 1))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error)

#line 267 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 267 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 267 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":268
 *         hi = 0
 *         peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
 *         hi = peak.mz             # <<<<<<<<<<<<<<
 *         peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)
 */

#line 268 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_peak->mz;

#line 268 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":269
 *         peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
 *         hi = peak.mz
 *         peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)             # <<<<<<<<<<<<<<
 *         hi = max(peak.mz, hi)
 *     elif peak_collection is object:
 */

#line 269 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->peaks->__pyx_vtab)->getitem(__pyx_v_peak_set_b->peaks, (((struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *)__pyx_v_peak_set_b->__pyx_vtab)->get_size(__pyx_v_peak_set_b) - 1))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)

#line 269 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 269 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 269 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":270
 *         hi = peak.mz
 *         peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)             # <<<<<<<<<<<<<<
 *     elif peak_collection is object:
 *         hi = 0
 */

#line 270 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_hi;

#line 270 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_v_peak->mz;

#line 270 "ms_deisotope/_c/similarity_methods.pyx"
  if (((__pyx_t_2 > __pyx_t_3) != 0)) {

#line 270 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_2;

#line 270 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 270 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_3;

#line 270 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 270 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_4;

  /* "ms_deisotope/_c/similarity_methods.pyx":292
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)
 *     scaler = 10 ** precision             # <<<<<<<<<<<<<<
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)
 */

#line 292 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_scaler = __Pyx_pow_long(10, ((long)__pyx_v_precision));

  /* "ms_deisotope/_c/similarity_methods.pyx":293
 *         hi = max(peak.mz, hi)
 *     scaler = 10 ** precision
 *     hi = (hi * scaler) + 1             # <<<<<<<<<<<<<<
 *     bin_a = zeros(<int>hi)
 *     bin_b = zeros(<int>hi)
 */

#line 293 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = ((__pyx_v_hi * __pyx_v_scaler) + 1.0);

  /* "ms_deisotope/_c/similarity_methods.pyx":294
 *     scaler = 10 ** precision
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)             # <<<<<<<<<<<<<<
 *     bin_b = zeros(<int>hi)
 * 
 */

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = __Pyx_PyInt_From_int(((int)__pyx_v_hi)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros);
  __pyx_t_6 = __pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros; 
#line 294 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_7 = NULL;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(__pyx_t_7)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_7);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(function);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF_SET(__pyx_t_6, function);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer, (PyObject*)__pyx_v_bin_a, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_12);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_Restore(__pyx_t_10, __pyx_t_11, __pyx_t_12);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_10 = __pyx_t_11 = __pyx_t_12 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_pybuffernd_bin_a.diminfo[0].strides = __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bin_a.diminfo[0].shape = __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.shape[0];

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_a = ((PyArrayObject *)__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":295
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)
 *     bin_b = zeros(<int>hi)             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = __Pyx_PyInt_From_int(((int)__pyx_v_hi)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros);
  __pyx_t_5 = __pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros; 
#line 295 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_7 = NULL;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(__pyx_t_7)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_7);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(function);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF_SET(__pyx_t_5, function);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_Fetch(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer, (PyObject*)__pyx_v_bin_b, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        Py_XDECREF(__pyx_t_12); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_Restore(__pyx_t_12, __pyx_t_11, __pyx_t_10);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_12 = __pyx_t_11 = __pyx_t_10 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_pybuffernd_bin_b.diminfo[0].strides = __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bin_b.diminfo[0].shape = __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.shape[0];

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_b = ((PyArrayObject *)__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":300
 *         n = len(peak_set_a)
 *     else:
 *         n = peak_set_a.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         if peak_collection is PeakIndex:
 */

#line 300 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":301
 *     else:
 *         n = peak_set_a.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakIndex:
 *             peak = peak_set_a.peaks.getitem(i)
 */

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_13 = __pyx_v_n;

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_14 = __pyx_t_13;

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_15;

    /* "ms_deisotope/_c/similarity_methods.pyx":303
 *     for i in range(n):
 *         if peak_collection is PeakIndex:
 *             peak = peak_set_a.peaks.getitem(i)             # <<<<<<<<<<<<<<
 *         elif peak_collection is object:
 *             peak = peak_set_a[i]
 */

#line 303 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->peaks->__pyx_vtab)->getitem(__pyx_v_peak_set_a->peaks, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L1_error)

#line 303 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 303 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 303 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":308
 *         else:
 *             peak = peak_set_a.getitem(i)
 *         index = int(peak.mz * scaler)             # <<<<<<<<<<<<<<
 *         bin_a[index] += peak.intensity
 * 
 */

#line 308 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_index = ((size_t)(__pyx_v_peak->mz * __pyx_v_scaler));

    /* "ms_deisotope/_c/similarity_methods.pyx":309
 *             peak = peak_set_a.getitem(i)
 *         index = int(peak.mz * scaler)
 *         bin_a[index] += peak.intensity             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_16 = __pyx_v_index;

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
    *__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_bin_a.diminfo[0].strides) += __pyx_v_peak->intensity;

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":314
 *         n = len(peak_set_b)
 *     else:
 *         n = peak_set_b.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         if peak_collection is PeakIndex:
 */

#line 314 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *)__pyx_v_peak_set_b->__pyx_vtab)->get_size(__pyx_v_peak_set_b);

  /* "ms_deisotope/_c/similarity_methods.pyx":315
 *     else:
 *         n = peak_set_b.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakIndex:
 *             peak = peak_set_b.peaks.getitem(i)
 */

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_13 = __pyx_v_n;

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_14 = __pyx_t_13;

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_15;

    /* "ms_deisotope/_c/similarity_methods.pyx":317
 *     for i in range(n):
 *         if peak_collection is PeakIndex:
 *             peak = peak_set_b.peaks.getitem(i)             # <<<<<<<<<<<<<<
 *         elif peak_collection is object:
 *             peak = peak_set_b[i]
 */

#line 317 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->peaks->__pyx_vtab)->getitem(__pyx_v_peak_set_b->peaks, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)

#line 317 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 317 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 317 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":322
 *         else:
 *             peak = peak_set_b.getitem(i)
 *         index = int(peak.mz * scaler)             # <<<<<<<<<<<<<<
 *         bin_b[index] += peak.intensity
 *     return (bin_a, bin_b)
 */

#line 322 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_index = ((size_t)(__pyx_v_peak->mz * __pyx_v_scaler));

    /* "ms_deisotope/_c/similarity_methods.pyx":323
 *             peak = peak_set_b.getitem(i)
 *         index = int(peak.mz * scaler)
 *         bin_b[index] += peak.intensity             # <<<<<<<<<<<<<<
 *     return (bin_a, bin_b)
 * 
 */

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_16 = __pyx_v_index;

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
    *__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_bin_b.diminfo[0].strides) += __pyx_v_peak->intensity;

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":324
 *         index = int(peak.mz * scaler)
 *         bin_b[index] += peak.intensity
 *     return (bin_a, bin_b)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = ((PyObject*)__pyx_t_1);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":258
 * 
 * @cython.boundscheck(False)
 * cpdef tuple bin_peaks(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double hi, scaler
 */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_PyThreadState_declare

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_PyThreadState_assign

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L2:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_bin_a);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_bin_b);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_25__pyx_fuse_1bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_25__pyx_fuse_1bin_peaks = 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_1bin_peaks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_25__pyx_fuse_1bin_peaks, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_2bin_peaks};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_25__pyx_fuse_1bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_precision
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1bin_peaks (wrapper)", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_precision,0};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1bin_peaks", 0, 2, 3, 1); 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_precision);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1bin_peaks") < 0)) __PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[0]);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[1]);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_precision == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_k__13;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1bin_peaks", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_1bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peak_set_a", 0))) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peak_set_b", 0))) __PYX_ERR(0, 258, __pyx_L1_error)
  __pyx_r = 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_24__pyx_fuse_1bin_peaks(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_precision);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_24__pyx_fuse_1bin_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, int __pyx_v_precision) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks __pyx_t_2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1bin_peaks", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_1bin_peaks (wrapper)", __pyx_f[0], 258, 0, __PYX_ERR(0, 258, __pyx_L1_error));

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.precision = __pyx_v_precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_1bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_27__pyx_fuse_2bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_3bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks *__pyx_optional_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_precision = __pyx_k__14;
  double __pyx_v_hi
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_scaler
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyArrayObject *__pyx_v_bin_a = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyArrayObject *__pyx_v_bin_b = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_i
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_n
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_index
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bin_a
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_bin_a
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bin_b
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_bin_b
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_3;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_4;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyArrayObject *__pyx_t_8 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_9;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_13;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_14;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_15;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_16;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2bin_peaks", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_2bin_peaks", __pyx_f[0], 258, 0, __PYX_ERR(0, 258, __pyx_L1_error));

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_optional_args->precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_a.pybuffer.buf = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_a.refcount = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_a.data = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_a.rcbuffer = &__pyx_pybuffer_bin_a;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_b.pybuffer.buf = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_b.refcount = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_b.data = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_b.rcbuffer = &__pyx_pybuffer_bin_b;

  /* "ms_deisotope/_c/similarity_methods.pyx":282
 *                 hi = peak.mz
 *     elif peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         peak = peak_set_a._mz_ordered[-1]             # <<<<<<<<<<<<<<
 *         hi = peak.mz
 *         peak = peak_set_b._mz_ordered[-1]
 */

#line 282 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_peak_set_a->_mz_ordered == Py_None)) {

#line 282 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 282 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 282, __pyx_L1_error)

#line 282 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 282 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_peak_set_a->_mz_ordered, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 282, __pyx_L1_error)

#line 282 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 282 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 282, __pyx_L1_error)

#line 282 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 282 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":283
 *     elif peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         peak = peak_set_a._mz_ordered[-1]
 *         hi = peak.mz             # <<<<<<<<<<<<<<
 *         peak = peak_set_b._mz_ordered[-1]
 *         hi = max(peak.mz, hi)
 */

#line 283 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_peak->mz;

#line 283 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":284
 *         peak = peak_set_a._mz_ordered[-1]
 *         hi = peak.mz
 *         peak = peak_set_b._mz_ordered[-1]             # <<<<<<<<<<<<<<
 *         hi = max(peak.mz, hi)
 *     else:
 */

#line 284 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_peak_set_b->_mz_ordered == Py_None)) {

#line 284 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 284 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 284, __pyx_L1_error)

#line 284 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 284 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_peak_set_b->_mz_ordered, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)

#line 284 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 284 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 284, __pyx_L1_error)

#line 284 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 284 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":285
 *         hi = peak.mz
 *         peak = peak_set_b._mz_ordered[-1]
 *         hi = max(peak.mz, hi)             # <<<<<<<<<<<<<<
 *     else:
 *         hi = 0
 */

#line 285 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_hi;

#line 285 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_v_peak->mz;

#line 285 "ms_deisotope/_c/similarity_methods.pyx"
  if (((__pyx_t_2 > __pyx_t_3) != 0)) {

#line 285 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_2;

#line 285 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 285 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_t_3;

#line 285 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 285 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = __pyx_t_4;

  /* "ms_deisotope/_c/similarity_methods.pyx":292
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)
 *     scaler = 10 ** precision             # <<<<<<<<<<<<<<
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)
 */

#line 292 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_scaler = __Pyx_pow_long(10, ((long)__pyx_v_precision));

  /* "ms_deisotope/_c/similarity_methods.pyx":293
 *         hi = max(peak.mz, hi)
 *     scaler = 10 ** precision
 *     hi = (hi * scaler) + 1             # <<<<<<<<<<<<<<
 *     bin_a = zeros(<int>hi)
 *     bin_b = zeros(<int>hi)
 */

#line 293 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = ((__pyx_v_hi * __pyx_v_scaler) + 1.0);

  /* "ms_deisotope/_c/similarity_methods.pyx":294
 *     scaler = 10 ** precision
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)             # <<<<<<<<<<<<<<
 *     bin_b = zeros(<int>hi)
 * 
 */

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = __Pyx_PyInt_From_int(((int)__pyx_v_hi)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros);
  __pyx_t_6 = __pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros; 
#line 294 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_7 = NULL;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(__pyx_t_7)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_7);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(function);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF_SET(__pyx_t_6, function);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer, (PyObject*)__pyx_v_bin_a, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_12);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_Restore(__pyx_t_10, __pyx_t_11, __pyx_t_12);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_10 = __pyx_t_11 = __pyx_t_12 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_pybuffernd_bin_a.diminfo[0].strides = __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bin_a.diminfo[0].shape = __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.shape[0];

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_a = ((PyArrayObject *)__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":295
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)
 *     bin_b = zeros(<int>hi)             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = __Pyx_PyInt_From_int(((int)__pyx_v_hi)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros);
  __pyx_t_5 = __pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros; 
#line 295 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_7 = NULL;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(__pyx_t_7)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_7);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(function);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF_SET(__pyx_t_5, function);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_Fetch(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer, (PyObject*)__pyx_v_bin_b, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        Py_XDECREF(__pyx_t_12); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_Restore(__pyx_t_12, __pyx_t_11, __pyx_t_10);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_12 = __pyx_t_11 = __pyx_t_10 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_pybuffernd_bin_b.diminfo[0].strides = __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bin_b.diminfo[0].shape = __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.shape[0];

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_8 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_b = ((PyArrayObject *)__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":300
 *         n = len(peak_set_a)
 *     else:
 *         n = peak_set_a.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         if peak_collection is PeakIndex:
 */

#line 300 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":301
 *     else:
 *         n = peak_set_a.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakIndex:
 *             peak = peak_set_a.peaks.getitem(i)
 */

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_13 = __pyx_v_n;

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_14 = __pyx_t_13;

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_15;

    /* "ms_deisotope/_c/similarity_methods.pyx":307
 *             peak = peak_set_a[i]
 *         else:
 *             peak = peak_set_a.getitem(i)             # <<<<<<<<<<<<<<
 *         index = int(peak.mz * scaler)
 *         bin_a[index] += peak.intensity
 */

#line 307 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->getitem(__pyx_v_peak_set_a, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)

#line 307 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 307 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 307 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":308
 *         else:
 *             peak = peak_set_a.getitem(i)
 *         index = int(peak.mz * scaler)             # <<<<<<<<<<<<<<
 *         bin_a[index] += peak.intensity
 * 
 */

#line 308 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_index = ((size_t)(__pyx_v_peak->mz * __pyx_v_scaler));

    /* "ms_deisotope/_c/similarity_methods.pyx":309
 *             peak = peak_set_a.getitem(i)
 *         index = int(peak.mz * scaler)
 *         bin_a[index] += peak.intensity             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_16 = __pyx_v_index;

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
    *__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_bin_a.diminfo[0].strides) += __pyx_v_peak->intensity;

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":314
 *         n = len(peak_set_b)
 *     else:
 *         n = peak_set_b.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         if peak_collection is PeakIndex:
 */

#line 314 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->get_size(__pyx_v_peak_set_b);

  /* "ms_deisotope/_c/similarity_methods.pyx":315
 *     else:
 *         n = peak_set_b.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakIndex:
 *             peak = peak_set_b.peaks.getitem(i)
 */

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_13 = __pyx_v_n;

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_14 = __pyx_t_13;

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_15;

    /* "ms_deisotope/_c/similarity_methods.pyx":321
 *             peak = peak_set_b[i]
 *         else:
 *             peak = peak_set_b.getitem(i)             # <<<<<<<<<<<<<<
 *         index = int(peak.mz * scaler)
 *         bin_b[index] += peak.intensity
 */

#line 321 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->getitem(__pyx_v_peak_set_b, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)

#line 321 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 321 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 321 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":322
 *         else:
 *             peak = peak_set_b.getitem(i)
 *         index = int(peak.mz * scaler)             # <<<<<<<<<<<<<<
 *         bin_b[index] += peak.intensity
 *     return (bin_a, bin_b)
 */

#line 322 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_index = ((size_t)(__pyx_v_peak->mz * __pyx_v_scaler));

    /* "ms_deisotope/_c/similarity_methods.pyx":323
 *             peak = peak_set_b.getitem(i)
 *         index = int(peak.mz * scaler)
 *         bin_b[index] += peak.intensity             # <<<<<<<<<<<<<<
 *     return (bin_a, bin_b)
 * 
 */

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_16 = __pyx_v_index;

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
    *__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_bin_b.diminfo[0].strides) += __pyx_v_peak->intensity;

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":324
 *         index = int(peak.mz * scaler)
 *         bin_b[index] += peak.intensity
 *     return (bin_a, bin_b)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = ((PyObject*)__pyx_t_1);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":258
 * 
 * @cython.boundscheck(False)
 * cpdef tuple bin_peaks(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double hi, scaler
 */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_PyThreadState_declare

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_PyThreadState_assign

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L2:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_bin_a);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_bin_b);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_27__pyx_fuse_2bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_27__pyx_fuse_2bin_peaks = 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_2bin_peaks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_27__pyx_fuse_2bin_peaks, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_2bin_peaks};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_27__pyx_fuse_2bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_precision
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2bin_peaks (wrapper)", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_precision,0};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2bin_peaks", 0, 2, 3, 1); 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_precision);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2bin_peaks") < 0)) __PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)values[0]);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)values[1]);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_precision == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_k__14;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2bin_peaks", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_2bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet, 1, "peak_set_a", 0))) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet, 1, "peak_set_b", 0))) __PYX_ERR(0, 258, __pyx_L1_error)
  __pyx_r = 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_26__pyx_fuse_2bin_peaks(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_precision);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_26__pyx_fuse_2bin_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, int __pyx_v_precision) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks __pyx_t_2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2bin_peaks", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_2bin_peaks (wrapper)", __pyx_f[0], 258, 0, __PYX_ERR(0, 258, __pyx_L1_error));

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.precision = __pyx_v_precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_2bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_29__pyx_fuse_3bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_3bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks *__pyx_optional_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_precision = __pyx_k__15;
  double __pyx_v_hi
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_scaler
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyArrayObject *__pyx_v_bin_a = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyArrayObject *__pyx_v_bin_b = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_i
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_n
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_index
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_obj = NULL
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bin_a
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_bin_a
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bin_b
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_bin_b
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *(*__pyx_t_3)(PyObject *);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_5;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_6;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyArrayObject *__pyx_t_9 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_10;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_13 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_14;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_15;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_16;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_17;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3bin_peaks", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_3bin_peaks", __pyx_f[0], 258, 0, __PYX_ERR(0, 258, __pyx_L1_error));

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_optional_args->precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_a.pybuffer.buf = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_a.refcount = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_a.data = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_a.rcbuffer = &__pyx_pybuffer_bin_a;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_b.pybuffer.buf = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_b.refcount = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_b.data = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_b.rcbuffer = &__pyx_pybuffer_bin_b;

  /* "ms_deisotope/_c/similarity_methods.pyx":272
 *         hi = max(peak.mz, hi)
 *     elif peak_collection is object:
 *         hi = 0             # <<<<<<<<<<<<<<
 *         for obj in peak_set_a:
 *             peak = <PeakBase>obj
 */

#line 272 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":273
 *     elif peak_collection is object:
 *         hi = 0
 *         for obj in peak_set_a:             # <<<<<<<<<<<<<<
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 */

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
  if (likely(PyList_CheckExact(__pyx_v_peak_set_a)) || PyTuple_CheckExact(__pyx_v_peak_set_a)) {

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_peak_set_a; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = NULL;

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
  } else {
    __pyx_t_2 = -1; 
#line 273 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_1 = PyObject_GetIter(__pyx_v_peak_set_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
  for (;;) {

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(!__pyx_t_3)) {

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(PyList_CheckExact(__pyx_t_1))) {

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        #else

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        #else

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
      }
    } else 
#line 273 "ms_deisotope/_c/similarity_methods.pyx"
{

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(!__pyx_t_4)) {

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        if (exc_type) {

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
          else __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
        break;

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_4);

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":274
 *         hi = 0
 *         for obj in peak_set_a:
 *             peak = <PeakBase>obj             # <<<<<<<<<<<<<<
 *             if peak.mz > hi:
 *                 hi = peak.mz
 */

#line 274 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_v_obj;

#line 274 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_4);

#line 274 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 274 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":275
 *         for obj in peak_set_a:
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:             # <<<<<<<<<<<<<<
 *                 hi = peak.mz
 *         for obj in peak_set_b:
 */

#line 275 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_peak->mz > __pyx_v_hi) != 0);

#line 275 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":276
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 *                 hi = peak.mz             # <<<<<<<<<<<<<<
 *         for obj in peak_set_b:
 *             peak = <PeakBase>obj
 */

#line 276 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = __pyx_v_peak->mz;

#line 276 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_hi = __pyx_t_6;

      /* "ms_deisotope/_c/similarity_methods.pyx":275
 *         for obj in peak_set_a:
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:             # <<<<<<<<<<<<<<
 *                 hi = peak.mz
 *         for obj in peak_set_b:
 */

#line 275 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":273
 *     elif peak_collection is object:
 *         hi = 0
 *         for obj in peak_set_a:             # <<<<<<<<<<<<<<
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 */

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 273 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":277
 *             if peak.mz > hi:
 *                 hi = peak.mz
 *         for obj in peak_set_b:             # <<<<<<<<<<<<<<
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 */

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
  if (likely(PyList_CheckExact(__pyx_v_peak_set_b)) || PyTuple_CheckExact(__pyx_v_peak_set_b)) {

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_peak_set_b; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = NULL;

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
  } else {
    __pyx_t_2 = -1; 
#line 277 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_1 = PyObject_GetIter(__pyx_v_peak_set_b); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 277, __pyx_L1_error)

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
  for (;;) {

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(!__pyx_t_3)) {

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(PyList_CheckExact(__pyx_t_1))) {

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 277, __pyx_L1_error)

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        #else

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 277, __pyx_L1_error)

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 277, __pyx_L1_error)

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        #else

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 277, __pyx_L1_error)

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
      }
    } else 
#line 277 "ms_deisotope/_c/similarity_methods.pyx"
{

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(!__pyx_t_4)) {

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        if (exc_type) {

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
          else __PYX_ERR(0, 277, __pyx_L1_error)

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
        break;

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_4);

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":278
 *                 hi = peak.mz
 *         for obj in peak_set_b:
 *             peak = <PeakBase>obj             # <<<<<<<<<<<<<<
 *             if peak.mz > hi:
 *                 hi = peak.mz
 */

#line 278 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = __pyx_v_obj;

#line 278 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_4);

#line 278 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 278 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":279
 *         for obj in peak_set_b:
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:             # <<<<<<<<<<<<<<
 *                 hi = peak.mz
 *     elif peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 279 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = ((__pyx_v_peak->mz > __pyx_v_hi) != 0);

#line 279 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/similarity_methods.pyx":280
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 *                 hi = peak.mz             # <<<<<<<<<<<<<<
 *     elif peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         peak = peak_set_a._mz_ordered[-1]
 */

#line 280 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = __pyx_v_peak->mz;

#line 280 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_hi = __pyx_t_6;

      /* "ms_deisotope/_c/similarity_methods.pyx":279
 *         for obj in peak_set_b:
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:             # <<<<<<<<<<<<<<
 *                 hi = peak.mz
 *     elif peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 279 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":277
 *             if peak.mz > hi:
 *                 hi = peak.mz
 *         for obj in peak_set_b:             # <<<<<<<<<<<<<<
 *             peak = <PeakBase>obj
 *             if peak.mz > hi:
 */

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 277 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":292
 *         peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
 *         hi = max(peak.mz, hi)
 *     scaler = 10 ** precision             # <<<<<<<<<<<<<<
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)
 */

#line 292 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_scaler = __Pyx_pow_long(10, ((long)__pyx_v_precision));

  /* "ms_deisotope/_c/similarity_methods.pyx":293
 *         hi = max(peak.mz, hi)
 *     scaler = 10 ** precision
 *     hi = (hi * scaler) + 1             # <<<<<<<<<<<<<<
 *     bin_a = zeros(<int>hi)
 *     bin_b = zeros(<int>hi)
 */

#line 293 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_hi = ((__pyx_v_hi * __pyx_v_scaler) + 1.0);

  /* "ms_deisotope/_c/similarity_methods.pyx":294
 *     scaler = 10 ** precision
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)             # <<<<<<<<<<<<<<
 *     bin_b = zeros(<int>hi)
 * 
 */

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = __Pyx_PyInt_From_int(((int)__pyx_v_hi)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros);
  __pyx_t_7 = __pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros; 
#line 294 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_8 = NULL;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(__pyx_t_8)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_8);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(function);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF_SET(__pyx_t_7, function);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_9 = ((PyArrayObject *)__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer, (PyObject*)__pyx_t_9, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_10 < 0)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer, (PyObject*)__pyx_v_bin_a, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_12); Py_XDECREF(__pyx_t_13);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_Restore(__pyx_t_11, __pyx_t_12, __pyx_t_13);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_11 = __pyx_t_12 = __pyx_t_13 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_pybuffernd_bin_a.diminfo[0].strides = __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bin_a.diminfo[0].shape = __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.shape[0];

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_9 = 0;

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_a = ((PyArrayObject *)__pyx_t_1);

#line 294 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":295
 *     hi = (hi * scaler) + 1
 *     bin_a = zeros(<int>hi)
 *     bin_b = zeros(<int>hi)             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_7 = __Pyx_PyInt_From_int(((int)__pyx_v_hi)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_7);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros);
  __pyx_t_4 = __pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros; 
#line 295 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_8 = NULL;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(__pyx_t_8)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_8);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(function);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF_SET(__pyx_t_4, function);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_8, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_9 = ((PyArrayObject *)__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer, (PyObject*)__pyx_t_9, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_10 < 0)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_Fetch(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer, (PyObject*)__pyx_v_bin_b, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        Py_XDECREF(__pyx_t_13); Py_XDECREF(__pyx_t_12); Py_XDECREF(__pyx_t_11);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      } else {

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_Restore(__pyx_t_13, __pyx_t_12, __pyx_t_11);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_13 = __pyx_t_12 = __pyx_t_11 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_pybuffernd_bin_b.diminfo[0].strides = __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bin_b.diminfo[0].shape = __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.shape[0];

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_9 = 0;

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bin_b = ((PyArrayObject *)__pyx_t_1);

#line 295 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":298
 * 
 *     if peak_collection is object:
 *         n = len(peak_set_a)             # <<<<<<<<<<<<<<
 *     else:
 *         n = peak_set_a.get_size()
 */

#line 298 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyObject_Length(__pyx_v_peak_set_a); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 298, __pyx_L1_error)

#line 298 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":301
 *     else:
 *         n = peak_set_a.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakIndex:
 *             peak = peak_set_a.peaks.getitem(i)
 */

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_14 = __pyx_v_n;

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_15 = __pyx_t_14;

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {

#line 301 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_16;

    /* "ms_deisotope/_c/similarity_methods.pyx":305
 *             peak = peak_set_a.peaks.getitem(i)
 *         elif peak_collection is object:
 *             peak = peak_set_a[i]             # <<<<<<<<<<<<<<
 *         else:
 *             peak = peak_set_a.getitem(i)
 */

#line 305 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_peak_set_a, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)

#line 305 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 305 "ms_deisotope/_c/similarity_methods.pyx"
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 305, __pyx_L1_error)

#line 305 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 305 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":308
 *         else:
 *             peak = peak_set_a.getitem(i)
 *         index = int(peak.mz * scaler)             # <<<<<<<<<<<<<<
 *         bin_a[index] += peak.intensity
 * 
 */

#line 308 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_index = ((size_t)(__pyx_v_peak->mz * __pyx_v_scaler));

    /* "ms_deisotope/_c/similarity_methods.pyx":309
 *             peak = peak_set_a.getitem(i)
 *         index = int(peak.mz * scaler)
 *         bin_a[index] += peak.intensity             # <<<<<<<<<<<<<<
 * 
 *     if peak_collection is object:
 */

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_17 = __pyx_v_index;

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
    *__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_bin_a.diminfo[0].strides) += __pyx_v_peak->intensity;

#line 309 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":312
 * 
 *     if peak_collection is object:
 *         n = len(peak_set_b)             # <<<<<<<<<<<<<<
 *     else:
 *         n = peak_set_b.get_size()
 */

#line 312 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyObject_Length(__pyx_v_peak_set_b); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 312, __pyx_L1_error)

#line 312 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":315
 *     else:
 *         n = peak_set_b.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakIndex:
 *             peak = peak_set_b.peaks.getitem(i)
 */

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_14 = __pyx_v_n;

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_15 = __pyx_t_14;

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {

#line 315 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_16;

    /* "ms_deisotope/_c/similarity_methods.pyx":319
 *             peak = peak_set_b.peaks.getitem(i)
 *         elif peak_collection is object:
 *             peak = peak_set_b[i]             # <<<<<<<<<<<<<<
 *         else:
 *             peak = peak_set_b.getitem(i)
 */

#line 319 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_peak_set_b, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L1_error)

#line 319 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 319 "ms_deisotope/_c/similarity_methods.pyx"
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 319, __pyx_L1_error)

#line 319 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 319 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":322
 *         else:
 *             peak = peak_set_b.getitem(i)
 *         index = int(peak.mz * scaler)             # <<<<<<<<<<<<<<
 *         bin_b[index] += peak.intensity
 *     return (bin_a, bin_b)
 */

#line 322 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_index = ((size_t)(__pyx_v_peak->mz * __pyx_v_scaler));

    /* "ms_deisotope/_c/similarity_methods.pyx":323
 *             peak = peak_set_b.getitem(i)
 *         index = int(peak.mz * scaler)
 *         bin_b[index] += peak.intensity             # <<<<<<<<<<<<<<
 *     return (bin_a, bin_b)
 * 
 */

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_17 = __pyx_v_index;

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
    *__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_bin_b.diminfo[0].strides) += __pyx_v_peak->intensity;

#line 323 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":324
 *         index = int(peak.mz * scaler)
 *         bin_b[index] += peak.intensity
 *     return (bin_a, bin_b)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_bin_a));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_bin_b));

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = ((PyObject*)__pyx_t_1);

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 324 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":258
 * 
 * @cython.boundscheck(False)
 * cpdef tuple bin_peaks(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double hi, scaler
 */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_PyThreadState_declare

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_PyThreadState_assign

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L2:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_bin_a);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_bin_b);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_obj);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_29__pyx_fuse_3bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_29__pyx_fuse_3bin_peaks = 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_3bin_peaks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_29__pyx_fuse_3bin_peaks, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_2bin_peaks};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_29__pyx_fuse_3bin_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peak_set_a = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_peak_set_b = 0
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_precision
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3bin_peaks (wrapper)", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_precision,0};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3bin_peaks", 0, 2, 3, 1); 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_precision);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3bin_peaks") < 0)) __PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = values[0];

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = values[1];

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_precision == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_precision = __pyx_k__15;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3bin_peaks", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 258, __pyx_L3_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_3bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 258 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_28__pyx_fuse_3bin_peaks(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_precision);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_28__pyx_fuse_3bin_peaks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, int __pyx_v_precision) {

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_bin_peaks __pyx_t_2;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3bin_peaks", 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_3bin_peaks (wrapper)", __pyx_f[0], 258, 0, __PYX_ERR(0, 258, __pyx_L1_error));

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.precision = __pyx_v_precision;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_bin_peaks(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"


#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_3bin_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":329
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cdef double calculate_dot_product(cnp.ndarray[double, ndim=1] bin_a, cnp.ndarray[double, ndim=1] bin_b, bint normalize=True):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double n_a, n_b, n_ab, z, a, b
 */

#line 329 "ms_deisotope/_c/similarity_methods.pyx"


#line 329 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_f_12ms_deisotope_2_c_18similarity_methods_calculate_dot_product(PyArrayObject *__pyx_v_bin_a, PyArrayObject *__pyx_v_bin_b, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_calculate_dot_product *__pyx_optional_args) {

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_normalize = ((int)1);
  double __pyx_v_n_a
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_n_b
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_n_ab
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_z
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_a
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_b
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;
  long __pyx_v_i
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED long __pyx_v_n
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bin_a
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_bin_a
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bin_b
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_bin_b
#line 329 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_1;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_2;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  long __pyx_t_3;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_4;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_5;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("calculate_dot_product", 0);

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("calculate_dot_product", __pyx_f[0], 329, 0, __PYX_ERR(0, 329, __pyx_L1_error));

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_normalize = __pyx_optional_args->normalize;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_a.pybuffer.buf = NULL;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_a.refcount = 0;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_a.data = NULL;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_a.rcbuffer = &__pyx_pybuffer_bin_a;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_b.pybuffer.buf = NULL;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffer_bin_b.refcount = 0;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_b.data = NULL;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_b.rcbuffer = &__pyx_pybuffer_bin_b;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer, (PyObject*)__pyx_v_bin_a, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 329, __pyx_L1_error)

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_a.diminfo[0].strides = __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bin_a.diminfo[0].shape = __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.shape[0];

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer, (PyObject*)__pyx_v_bin_b, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 329, __pyx_L1_error)

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_pybuffernd_bin_b.diminfo[0].strides = __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bin_b.diminfo[0].shape = __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.shape[0];

  /* "ms_deisotope/_c/similarity_methods.pyx":333
 *         double n_a, n_b, n_ab, z, a, b
 *         long i, n
 *     with nogil:             # <<<<<<<<<<<<<<
 *         n_a = 0
 *         n_b = 0
 */

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
      #ifdef WITH_THREAD

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
      PyThreadState *_save;

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
      Py_UNBLOCK_THREADS

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_FastGIL_Remember();

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
      /*try:*/ {

        /* "ms_deisotope/_c/similarity_methods.pyx":334
 *         long i, n
 *     with nogil:
 *         n_a = 0             # <<<<<<<<<<<<<<
 *         n_b = 0
 *         z = 0
 */

#line 334 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_n_a = 0.0;

        /* "ms_deisotope/_c/similarity_methods.pyx":335
 *     with nogil:
 *         n_a = 0
 *         n_b = 0             # <<<<<<<<<<<<<<
 *         z = 0
 *         n = bin_a.shape[0]
 */

#line 335 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_n_b = 0.0;

        /* "ms_deisotope/_c/similarity_methods.pyx":336
 *         n_a = 0
 *         n_b = 0
 *         z = 0             # <<<<<<<<<<<<<<
 *         n = bin_a.shape[0]
 *         for i in prange(n, schedule='static'):
 */

#line 336 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_z = 0.0;

        /* "ms_deisotope/_c/similarity_methods.pyx":337
 *         n_b = 0
 *         z = 0
 *         n = bin_a.shape[0]             # <<<<<<<<<<<<<<
 *         for i in prange(n, schedule='static'):
 *             a = bin_a[i]
 */

#line 337 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_n = (__pyx_v_bin_a->dimensions[0]);

        /* "ms_deisotope/_c/similarity_methods.pyx":338
 *         z = 0
 *         n = bin_a.shape[0]
 *         for i in prange(n, schedule='static'):             # <<<<<<<<<<<<<<
 *             a = bin_a[i]
 *             b = bin_b[i]
 */

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_1 = __pyx_v_n;

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
        if ((1 == 0)) abort();

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
        {

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                #undef likely

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                #undef unlikely

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                #define likely(x)   (x)

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                #define unlikely(x) (x)

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
            #endif

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
            __pyx_t_3 = (__pyx_t_1 - 0 + 1 - 1/abs(1)) / 1;

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
            if (__pyx_t_3 > 0)

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
            {

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                #ifdef _OPENMP
                #pragma omp parallel reduction(+:__pyx_v_n_a) reduction(+:__pyx_v_n_b) reduction(+:__pyx_v_z) private(__pyx_t_4)
#line 338 "ms_deisotope/_c/similarity_methods.pyx"


#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                #endif /* _OPENMP */

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                {

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_a) lastprivate(__pyx_v_b) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) schedule(static)
#line 338 "ms_deisotope/_c/similarity_methods.pyx"


#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                    #endif /* _OPENMP */
                    for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_3; __pyx_t_2++)
#line 338 "ms_deisotope/_c/similarity_methods.pyx"
{

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                        {

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_i = (long)(0 + 1 * __pyx_t_2);

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                            /* Initialize private variables to invalid values */

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_a = ((double)__PYX_NAN());

#line 338 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_b = ((double)__PYX_NAN());

                            /* "ms_deisotope/_c/similarity_methods.pyx":339
 *         n = bin_a.shape[0]
 *         for i in prange(n, schedule='static'):
 *             a = bin_a[i]             # <<<<<<<<<<<<<<
 *             b = bin_b[i]
 *             n_a += a * a
 */

#line 339 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_t_4 = __pyx_v_i;

#line 339 "ms_deisotope/_c/similarity_methods.pyx"
                            if (__pyx_t_4 < 0) __pyx_t_4 += __pyx_pybuffernd_bin_a.diminfo[0].shape;

#line 339 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_a = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_bin_a.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_bin_a.diminfo[0].strides));

                            /* "ms_deisotope/_c/similarity_methods.pyx":340
 *         for i in prange(n, schedule='static'):
 *             a = bin_a[i]
 *             b = bin_b[i]             # <<<<<<<<<<<<<<
 *             n_a += a * a
 *             n_b += b * b
 */

#line 340 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_t_4 = __pyx_v_i;

#line 340 "ms_deisotope/_c/similarity_methods.pyx"
                            if (__pyx_t_4 < 0) __pyx_t_4 += __pyx_pybuffernd_bin_b.diminfo[0].shape;

#line 340 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_b = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_bin_b.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_bin_b.diminfo[0].strides));

                            /* "ms_deisotope/_c/similarity_methods.pyx":341
 *             a = bin_a[i]
 *             b = bin_b[i]
 *             n_a += a * a             # <<<<<<<<<<<<<<
 *             n_b += b * b
 *             z += a * b
 */

#line 341 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_n_a = (__pyx_v_n_a + (__pyx_v_a * __pyx_v_a));

                            /* "ms_deisotope/_c/similarity_methods.pyx":342
 *             b = bin_b[i]
 *             n_a += a * a
 *             n_b += b * b             # <<<<<<<<<<<<<<
 *             z += a * b
 *         if normalize:
 */

#line 342 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_n_b = (__pyx_v_n_b + (__pyx_v_b * __pyx_v_b));

                            /* "ms_deisotope/_c/similarity_methods.pyx":343
 *             n_a += a * a
 *             n_b += b * b
 *             z += a * b             # <<<<<<<<<<<<<<
 *         if normalize:
 *             n_ab = sqrt(n_a) * sqrt(n_b)
 */

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
                            __pyx_v_z = (__pyx_v_z + (__pyx_v_a * __pyx_v_b));

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
                        }

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
                    }

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
                }

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
            }

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
            #undef likely

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
            #undef unlikely

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
            #define likely(x)   __builtin_expect(!!(x), 1)

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
            #define unlikely(x) __builtin_expect(!!(x), 0)

#line 343 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

        /* "ms_deisotope/_c/similarity_methods.pyx":344
 *             n_b += b * b
 *             z += a * b
 *         if normalize:             # <<<<<<<<<<<<<<
 *             n_ab = sqrt(n_a) * sqrt(n_b)
 *             if n_ab == 0:
 */

#line 344 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_5 = (__pyx_v_normalize != 0);

#line 344 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_5) {

          /* "ms_deisotope/_c/similarity_methods.pyx":345
 *             z += a * b
 *         if normalize:
 *             n_ab = sqrt(n_a) * sqrt(n_b)             # <<<<<<<<<<<<<<
 *             if n_ab == 0:
 *                 return 0.0
 */

#line 345 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_n_ab = (sqrt(__pyx_v_n_a) * sqrt(__pyx_v_n_b));

          /* "ms_deisotope/_c/similarity_methods.pyx":346
 *         if normalize:
 *             n_ab = sqrt(n_a) * sqrt(n_b)
 *             if n_ab == 0:             # <<<<<<<<<<<<<<
 *                 return 0.0
 *             return z / n_ab
 */

#line 346 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_5 = ((__pyx_v_n_ab == 0.0) != 0);

#line 346 "ms_deisotope/_c/similarity_methods.pyx"
          if (__pyx_t_5) {

            /* "ms_deisotope/_c/similarity_methods.pyx":347
 *             n_ab = sqrt(n_a) * sqrt(n_b)
 *             if n_ab == 0:
 *                 return 0.0             # <<<<<<<<<<<<<<
 *             return z / n_ab
 *         return z
 */

#line 347 "ms_deisotope/_c/similarity_methods.pyx"
            __pyx_r = 0.0;

#line 347 "ms_deisotope/_c/similarity_methods.pyx"
            goto __pyx_L3_return;

            /* "ms_deisotope/_c/similarity_methods.pyx":346
 *         if normalize:
 *             n_ab = sqrt(n_a) * sqrt(n_b)
 *             if n_ab == 0:             # <<<<<<<<<<<<<<
 *                 return 0.0
 *             return z / n_ab
 */

#line 346 "ms_deisotope/_c/similarity_methods.pyx"
          }

          /* "ms_deisotope/_c/similarity_methods.pyx":348
 *             if n_ab == 0:
 *                 return 0.0
 *             return z / n_ab             # <<<<<<<<<<<<<<
 *         return z
 * 
 */

#line 348 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_r = (__pyx_v_z / __pyx_v_n_ab);

#line 348 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L3_return;

          /* "ms_deisotope/_c/similarity_methods.pyx":344
 *             n_b += b * b
 *             z += a * b
 *         if normalize:             # <<<<<<<<<<<<<<
 *             n_ab = sqrt(n_a) * sqrt(n_b)
 *             if n_ab == 0:
 */

#line 344 "ms_deisotope/_c/similarity_methods.pyx"
        }

        /* "ms_deisotope/_c/similarity_methods.pyx":349
 *                 return 0.0
 *             return z / n_ab
 *         return z             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 349 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_r = __pyx_v_z;

#line 349 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L3_return;

#line 349 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":333
 *         double n_a, n_b, n_ab, z, a, b
 *         long i, n
 *     with nogil:             # <<<<<<<<<<<<<<
 *         n_a = 0
 *         n_b = 0
 */

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
      /*finally:*/ {

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_L3_return: {

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
          #ifdef WITH_THREAD

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_FastGIL_Forget();

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
          Py_BLOCK_THREADS

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
          #endif

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L0;

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 333 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":329
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cdef double calculate_dot_product(cnp.ndarray[double, ndim=1] bin_a, cnp.ndarray[double, ndim=1] bin_b, bint normalize=True):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double n_a, n_b, n_ab, z, a, b
 */

#line 329 "ms_deisotope/_c/similarity_methods.pyx"


#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_PyThreadState_declare

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_PyThreadState_assign

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods.calculate_dot_product", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L2;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_a.rcbuffer->pybuffer);

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bin_b.rcbuffer->pybuffer);

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L2:;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 329 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":366
 *         public bint sqrt_transform
 * 
 *     def __init__(self, peak_set_a, peak_set_b, double error_tolerance=2e-5, double shift=0.0,             # <<<<<<<<<<<<<<
 *                  bint normalize=True, bint sqrt_transform=False):
 *         self.peak_set_a = peak_set_a
 */

#line 366 "ms_deisotope/_c/similarity_methods.pyx"


#line 366 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peak_set_a = 0
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_peak_set_b = 0
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_error_tolerance
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_shift
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_normalize
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_sqrt_transform
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_error_tolerance,&__pyx_n_s_shift,&__pyx_n_s_normalize,&__pyx_n_s_sqrt_transform,0};

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[6] = {0,0,0,0,0,0};

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  6: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 6, 1); 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 366, __pyx_L3_error)

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        case  3:

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shift);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[3] = value; kw_args--; }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        case  4:

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_normalize);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        case  5:

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sqrt_transform);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[5] = value; kw_args--; }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 366, __pyx_L3_error)

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 366 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = values[0];

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = values[1];

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 366, __pyx_L3_error)

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = ((double)2e-5);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[3]) {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_shift == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 366, __pyx_L3_error)

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = ((double)0.0);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[4]) {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_normalize = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_normalize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 367, __pyx_L3_error)

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

      /* "ms_deisotope/_c/similarity_methods.pyx":367
 * 
 *     def __init__(self, peak_set_a, peak_set_b, double error_tolerance=2e-5, double shift=0.0,
 *                  bint normalize=True, bint sqrt_transform=False):             # <<<<<<<<<<<<<<
 *         self.peak_set_a = peak_set_a
 *         self.peak_set_b = peak_set_b
 */

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_normalize = ((int)1);

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[5]) {

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_sqrt_transform = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_sqrt_transform == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 367, __pyx_L3_error)

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_sqrt_transform = ((int)0);

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 6, PyTuple_GET_SIZE(__pyx_args)); 
#line 367 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 366, __pyx_L3_error)

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
  return -1;

#line 367 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 367 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment___init__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_error_tolerance, __pyx_v_shift, __pyx_v_normalize, __pyx_v_sqrt_transform);

  /* "ms_deisotope/_c/similarity_methods.pyx":366
 *         public bint sqrt_transform
 * 
 *     def __init__(self, peak_set_a, peak_set_b, double error_tolerance=2e-5, double shift=0.0,             # <<<<<<<<<<<<<<
 *                  bint normalize=True, bint sqrt_transform=False):
 *         self.peak_set_a = peak_set_a
 */

#line 366 "ms_deisotope/_c/similarity_methods.pyx"


#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 366 "ms_deisotope/_c/similarity_methods.pyx"


#line 366 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment___init__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, double __pyx_v_error_tolerance, double __pyx_v_shift, int __pyx_v_normalize, int __pyx_v_sqrt_transform) {

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_align __pyx_t_2;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__init__", __pyx_f[0], 366, 0, __PYX_ERR(0, 366, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":368
 *     def __init__(self, peak_set_a, peak_set_b, double error_tolerance=2e-5, double shift=0.0,
 *                  bint normalize=True, bint sqrt_transform=False):
 *         self.peak_set_a = peak_set_a             # <<<<<<<<<<<<<<
 *         self.peak_set_b = peak_set_b
 *         self.peak_pairs = []
 */

#line 368 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_peak_set_a);

#line 368 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_v_peak_set_a);

#line 368 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_set_a);

#line 368 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_v_self->peak_set_a);

#line 368 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_set_a = __pyx_v_peak_set_a;

  /* "ms_deisotope/_c/similarity_methods.pyx":369
 *                  bint normalize=True, bint sqrt_transform=False):
 *         self.peak_set_a = peak_set_a
 *         self.peak_set_b = peak_set_b             # <<<<<<<<<<<<<<
 *         self.peak_pairs = []
 *         self.shift = shift
 */

#line 369 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_peak_set_b);

#line 369 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_v_peak_set_b);

#line 369 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_set_b);

#line 369 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_v_self->peak_set_b);

#line 369 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_set_b = __pyx_v_peak_set_b;

  /* "ms_deisotope/_c/similarity_methods.pyx":370
 *         self.peak_set_a = peak_set_a
 *         self.peak_set_b = peak_set_b
 *         self.peak_pairs = []             # <<<<<<<<<<<<<<
 *         self.shift = shift
 * 
 */

#line 370 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L1_error)

#line 370 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 370 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 370 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_pairs);

#line 370 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_v_self->peak_pairs);

#line 370 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_pairs = ((PyObject*)__pyx_t_1);

#line 370 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":371
 *         self.peak_set_b = peak_set_b
 *         self.peak_pairs = []
 *         self.shift = shift             # <<<<<<<<<<<<<<
 * 
 *         self.score = 0
 */

#line 371 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->shift = __pyx_v_shift;

  /* "ms_deisotope/_c/similarity_methods.pyx":373
 *         self.shift = shift
 * 
 *         self.score = 0             # <<<<<<<<<<<<<<
 *         self.norm_aa_cosine = 1.0
 *         self.norm_bb_cosine = 1.0
 */

#line 373 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->score = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":374
 * 
 *         self.score = 0
 *         self.norm_aa_cosine = 1.0             # <<<<<<<<<<<<<<
 *         self.norm_bb_cosine = 1.0
 * 
 */

#line 374 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->norm_aa_cosine = 1.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":375
 *         self.score = 0
 *         self.norm_aa_cosine = 1.0
 *         self.norm_bb_cosine = 1.0             # <<<<<<<<<<<<<<
 * 
 *         self.normalize = normalize
 */

#line 375 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->norm_bb_cosine = 1.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":377
 *         self.norm_bb_cosine = 1.0
 * 
 *         self.normalize = normalize             # <<<<<<<<<<<<<<
 *         self.sqrt_transform = sqrt_transform
 * 
 */

#line 377 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->normalize = __pyx_v_normalize;

  /* "ms_deisotope/_c/similarity_methods.pyx":378
 * 
 *         self.normalize = normalize
 *         self.sqrt_transform = sqrt_transform             # <<<<<<<<<<<<<<
 * 
 *         self._determine_peak_type()
 */

#line 378 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->sqrt_transform = __pyx_v_sqrt_transform;

  /* "ms_deisotope/_c/similarity_methods.pyx":380
 *         self.sqrt_transform = sqrt_transform
 * 
 *         self._determine_peak_type()             # <<<<<<<<<<<<<<
 *         self.calculate_normalization()
 *         self.align(error_tolerance, shift)
 */

#line 380 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self->__pyx_vtab)->_determine_peak_type(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)

#line 380 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 380 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":381
 * 
 *         self._determine_peak_type()
 *         self.calculate_normalization()             # <<<<<<<<<<<<<<
 *         self.align(error_tolerance, shift)
 * 
 */

#line 381 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self->__pyx_vtab)->calculate_normalization(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)

#line 381 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 381 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":382
 *         self._determine_peak_type()
 *         self.calculate_normalization()
 *         self.align(error_tolerance, shift)             # <<<<<<<<<<<<<<
 * 
 *     cpdef calculate_normalization(self):
 */

#line 382 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 2;

#line 382 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 382 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.shift = __pyx_v_shift;

#line 382 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self->__pyx_vtab)->align(__pyx_v_self, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)

#line 382 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 382 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":366
 *         public bint sqrt_transform
 * 
 *     def __init__(self, peak_set_a, peak_set_b, double error_tolerance=2e-5, double shift=0.0,             # <<<<<<<<<<<<<<
 *                  bint normalize=True, bint sqrt_transform=False):
 *         self.peak_set_a = peak_set_a
 */

#line 366 "ms_deisotope/_c/similarity_methods.pyx"


#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 366 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":384
 *         self.align(error_tolerance, shift)
 * 
 *     cpdef calculate_normalization(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 384 "ms_deisotope/_c/similarity_methods.pyx"


#line 384 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_3calculate_normalization(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_calculate_normalization(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 384 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_n
#line 384 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set = 0
#line 384 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_dpeak_set = 0
#line 384 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_intensity
#line 384 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_acc_cosine
#line 384 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_sqrt_transform
#line 384 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_5;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_6;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_7;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_8;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_9;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("calculate_normalization", 0);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("calculate_normalization", __pyx_f[0], 384, 0, __PYX_ERR(0, 384, __pyx_L1_error));

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  /* Check if called by wrapper */

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  /* Check if overridden in Python */

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_calculate_normalization); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_3calculate_normalization)) {

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 384 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_4 = NULL;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
          if (likely(__pyx_t_4)) {

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_INCREF(function);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
          }

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 384, __pyx_L1_error)

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_r = __pyx_t_2;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_2 = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
    #endif

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":392
 *             bint sqrt_transform
 * 
 *         sqrt_transform = self.sqrt_transform             # <<<<<<<<<<<<<<
 *         if self.peak_type == PeakType.Fitted:
 *             peak_set = <PeakSet>self.peak_set_a
 */

#line 392 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = __pyx_v_self->sqrt_transform;

#line 392 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_sqrt_transform = __pyx_t_5;

  /* "ms_deisotope/_c/similarity_methods.pyx":393
 * 
 *         sqrt_transform = self.sqrt_transform
 *         if self.peak_type == PeakType.Fitted:             # <<<<<<<<<<<<<<
 *             peak_set = <PeakSet>self.peak_set_a
 *             n = peak_set.get_size()
 */

#line 393 "ms_deisotope/_c/similarity_methods.pyx"
  switch (__pyx_v_self->peak_type) {

#line 393 "ms_deisotope/_c/similarity_methods.pyx"
    case __pyx_e_12ms_deisotope_2_c_18similarity_methods_Fitted:

    /* "ms_deisotope/_c/similarity_methods.pyx":394
 *         sqrt_transform = self.sqrt_transform
 *         if self.peak_type == PeakType.Fitted:
 *             peak_set = <PeakSet>self.peak_set_a             # <<<<<<<<<<<<<<
 *             n = peak_set.get_size()
 *             acc_cosine = 0.0
 */

#line 394 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_self->peak_set_a;

#line 394 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 394 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);

#line 394 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":395
 *         if self.peak_type == PeakType.Fitted:
 *             peak_set = <PeakSet>self.peak_set_a
 *             n = peak_set.get_size()             # <<<<<<<<<<<<<<
 *             acc_cosine = 0.0
 *             for i in range(n):
 */

#line 395 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set->__pyx_vtab)->get_size(__pyx_v_peak_set);

    /* "ms_deisotope/_c/similarity_methods.pyx":396
 *             peak_set = <PeakSet>self.peak_set_a
 *             n = peak_set.get_size()
 *             acc_cosine = 0.0             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 intensity = peak_set.getitem(i).intensity
 */

#line 396 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_acc_cosine = 0.0;

    /* "ms_deisotope/_c/similarity_methods.pyx":397
 *             n = peak_set.get_size()
 *             acc_cosine = 0.0
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 intensity = peak_set.getitem(i).intensity
 *                 if sqrt_transform:
 */

#line 397 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __pyx_v_n;

#line 397 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = __pyx_t_6;

#line 397 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 397 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_i = __pyx_t_8;

      /* "ms_deisotope/_c/similarity_methods.pyx":398
 *             acc_cosine = 0.0
 *             for i in range(n):
 *                 intensity = peak_set.getitem(i).intensity             # <<<<<<<<<<<<<<
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)
 */

#line 398 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set->__pyx_vtab)->getitem(__pyx_v_peak_set, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)

#line 398 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 398 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_9 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1)->__pyx_base.intensity;

#line 398 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 398 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_intensity = __pyx_t_9;

      /* "ms_deisotope/_c/similarity_methods.pyx":399
 *             for i in range(n):
 *                 intensity = peak_set.getitem(i).intensity
 *                 if sqrt_transform:             # <<<<<<<<<<<<<<
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 */

#line 399 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_5 = (__pyx_v_sqrt_transform != 0);

#line 399 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_5) {

        /* "ms_deisotope/_c/similarity_methods.pyx":400
 *                 intensity = peak_set.getitem(i).intensity
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)             # <<<<<<<<<<<<<<
 *                 acc_cosine += intensity ** 2
 *             self.norm_aa_cosine = sqrt(acc_cosine)
 */

#line 400 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_intensity = sqrt(__pyx_v_intensity);

        /* "ms_deisotope/_c/similarity_methods.pyx":399
 *             for i in range(n):
 *                 intensity = peak_set.getitem(i).intensity
 *                 if sqrt_transform:             # <<<<<<<<<<<<<<
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 */

#line 399 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":401
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2             # <<<<<<<<<<<<<<
 *             self.norm_aa_cosine = sqrt(acc_cosine)
 *             peak_set = <PeakSet>self.peak_set_b
 */

#line 401 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_acc_cosine = (__pyx_v_acc_cosine + pow(__pyx_v_intensity, 2.0));

#line 401 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":402
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 *             self.norm_aa_cosine = sqrt(acc_cosine)             # <<<<<<<<<<<<<<
 *             peak_set = <PeakSet>self.peak_set_b
 *             n = peak_set.get_size()
 */

#line 402 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_self->norm_aa_cosine = sqrt(__pyx_v_acc_cosine);

    /* "ms_deisotope/_c/similarity_methods.pyx":403
 *                 acc_cosine += intensity ** 2
 *             self.norm_aa_cosine = sqrt(acc_cosine)
 *             peak_set = <PeakSet>self.peak_set_b             # <<<<<<<<<<<<<<
 *             n = peak_set.get_size()
 *             acc_cosine = 0.0
 */

#line 403 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_self->peak_set_b;

#line 403 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 403 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_peak_set, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1));

#line 403 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":404
 *             self.norm_aa_cosine = sqrt(acc_cosine)
 *             peak_set = <PeakSet>self.peak_set_b
 *             n = peak_set.get_size()             # <<<<<<<<<<<<<<
 *             acc_cosine = 0.0
 *             for i in range(n):
 */

#line 404 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set->__pyx_vtab)->get_size(__pyx_v_peak_set);

    /* "ms_deisotope/_c/similarity_methods.pyx":405
 *             peak_set = <PeakSet>self.peak_set_b
 *             n = peak_set.get_size()
 *             acc_cosine = 0.0             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 intensity = peak_set.getitem(i).intensity
 */

#line 405 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_acc_cosine = 0.0;

    /* "ms_deisotope/_c/similarity_methods.pyx":406
 *             n = peak_set.get_size()
 *             acc_cosine = 0.0
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 intensity = peak_set.getitem(i).intensity
 *                 if sqrt_transform:
 */

#line 406 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __pyx_v_n;

#line 406 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = __pyx_t_6;

#line 406 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 406 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_i = __pyx_t_8;

      /* "ms_deisotope/_c/similarity_methods.pyx":407
 *             acc_cosine = 0.0
 *             for i in range(n):
 *                 intensity = peak_set.getitem(i).intensity             # <<<<<<<<<<<<<<
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)
 */

#line 407 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set->__pyx_vtab)->getitem(__pyx_v_peak_set, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)

#line 407 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 407 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_9 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1)->__pyx_base.intensity;

#line 407 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 407 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_intensity = __pyx_t_9;

      /* "ms_deisotope/_c/similarity_methods.pyx":408
 *             for i in range(n):
 *                 intensity = peak_set.getitem(i).intensity
 *                 if sqrt_transform:             # <<<<<<<<<<<<<<
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 */

#line 408 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_5 = (__pyx_v_sqrt_transform != 0);

#line 408 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_5) {

        /* "ms_deisotope/_c/similarity_methods.pyx":409
 *                 intensity = peak_set.getitem(i).intensity
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)             # <<<<<<<<<<<<<<
 *                 acc_cosine += intensity ** 2
 *             self.norm_bb_cosine = sqrt(acc_cosine)
 */

#line 409 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_intensity = sqrt(__pyx_v_intensity);

        /* "ms_deisotope/_c/similarity_methods.pyx":408
 *             for i in range(n):
 *                 intensity = peak_set.getitem(i).intensity
 *                 if sqrt_transform:             # <<<<<<<<<<<<<<
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 */

#line 408 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":410
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2             # <<<<<<<<<<<<<<
 *             self.norm_bb_cosine = sqrt(acc_cosine)
 *         elif self.peak_type == PeakType.Deconvoluted:
 */

#line 410 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_acc_cosine = (__pyx_v_acc_cosine + pow(__pyx_v_intensity, 2.0));

#line 410 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":411
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 *             self.norm_bb_cosine = sqrt(acc_cosine)             # <<<<<<<<<<<<<<
 *         elif self.peak_type == PeakType.Deconvoluted:
 *             dpeak_set = <DeconvolutedPeakSet>self.peak_set_a
 */

#line 411 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_self->norm_bb_cosine = sqrt(__pyx_v_acc_cosine);

    /* "ms_deisotope/_c/similarity_methods.pyx":393
 * 
 *         sqrt_transform = self.sqrt_transform
 *         if self.peak_type == PeakType.Fitted:             # <<<<<<<<<<<<<<
 *             peak_set = <PeakSet>self.peak_set_a
 *             n = peak_set.get_size()
 */

#line 393 "ms_deisotope/_c/similarity_methods.pyx"
    break;

#line 393 "ms_deisotope/_c/similarity_methods.pyx"
    case __pyx_e_12ms_deisotope_2_c_18similarity_methods_Deconvoluted:

    /* "ms_deisotope/_c/similarity_methods.pyx":413
 *             self.norm_bb_cosine = sqrt(acc_cosine)
 *         elif self.peak_type == PeakType.Deconvoluted:
 *             dpeak_set = <DeconvolutedPeakSet>self.peak_set_a             # <<<<<<<<<<<<<<
 *             n = dpeak_set.get_size()
 *             acc_cosine = 0.0
 */

#line 413 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_self->peak_set_a;

#line 413 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 413 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_dpeak_set = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_t_1);

#line 413 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":414
 *         elif self.peak_type == PeakType.Deconvoluted:
 *             dpeak_set = <DeconvolutedPeakSet>self.peak_set_a
 *             n = dpeak_set.get_size()             # <<<<<<<<<<<<<<
 *             acc_cosine = 0.0
 *             for i in range(n):
 */

#line 414 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_dpeak_set->__pyx_vtab)->get_size(__pyx_v_dpeak_set);

    /* "ms_deisotope/_c/similarity_methods.pyx":415
 *             dpeak_set = <DeconvolutedPeakSet>self.peak_set_a
 *             n = dpeak_set.get_size()
 *             acc_cosine = 0.0             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 intensity = dpeak_set.getitem(i).intensity
 */

#line 415 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_acc_cosine = 0.0;

    /* "ms_deisotope/_c/similarity_methods.pyx":416
 *             n = dpeak_set.get_size()
 *             acc_cosine = 0.0
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 intensity = dpeak_set.getitem(i).intensity
 *                 if sqrt_transform:
 */

#line 416 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __pyx_v_n;

#line 416 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = __pyx_t_6;

#line 416 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 416 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_i = __pyx_t_8;

      /* "ms_deisotope/_c/similarity_methods.pyx":417
 *             acc_cosine = 0.0
 *             for i in range(n):
 *                 intensity = dpeak_set.getitem(i).intensity             # <<<<<<<<<<<<<<
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)
 */

#line 417 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_dpeak_set->__pyx_vtab)->getitem(__pyx_v_dpeak_set, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)

#line 417 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 417 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_9 = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_1)->__pyx_base.intensity;

#line 417 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 417 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_intensity = __pyx_t_9;

      /* "ms_deisotope/_c/similarity_methods.pyx":418
 *             for i in range(n):
 *                 intensity = dpeak_set.getitem(i).intensity
 *                 if sqrt_transform:             # <<<<<<<<<<<<<<
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 */

#line 418 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_5 = (__pyx_v_sqrt_transform != 0);

#line 418 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_5) {

        /* "ms_deisotope/_c/similarity_methods.pyx":419
 *                 intensity = dpeak_set.getitem(i).intensity
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)             # <<<<<<<<<<<<<<
 *                 acc_cosine += intensity ** 2
 *             self.norm_aa_cosine = sqrt(acc_cosine)
 */

#line 419 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_intensity = sqrt(__pyx_v_intensity);

        /* "ms_deisotope/_c/similarity_methods.pyx":418
 *             for i in range(n):
 *                 intensity = dpeak_set.getitem(i).intensity
 *                 if sqrt_transform:             # <<<<<<<<<<<<<<
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 */

#line 418 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":420
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2             # <<<<<<<<<<<<<<
 *             self.norm_aa_cosine = sqrt(acc_cosine)
 *             dpeak_set = <DeconvolutedPeakSet>self.peak_set_b
 */

#line 420 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_acc_cosine = (__pyx_v_acc_cosine + pow(__pyx_v_intensity, 2.0));

#line 420 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":421
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 *             self.norm_aa_cosine = sqrt(acc_cosine)             # <<<<<<<<<<<<<<
 *             dpeak_set = <DeconvolutedPeakSet>self.peak_set_b
 *             n = dpeak_set.get_size()
 */

#line 421 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_self->norm_aa_cosine = sqrt(__pyx_v_acc_cosine);

    /* "ms_deisotope/_c/similarity_methods.pyx":422
 *                 acc_cosine += intensity ** 2
 *             self.norm_aa_cosine = sqrt(acc_cosine)
 *             dpeak_set = <DeconvolutedPeakSet>self.peak_set_b             # <<<<<<<<<<<<<<
 *             n = dpeak_set.get_size()
 *             acc_cosine = 0.0
 */

#line 422 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_self->peak_set_b;

#line 422 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 422 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_dpeak_set, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_t_1));

#line 422 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":423
 *             self.norm_aa_cosine = sqrt(acc_cosine)
 *             dpeak_set = <DeconvolutedPeakSet>self.peak_set_b
 *             n = dpeak_set.get_size()             # <<<<<<<<<<<<<<
 *             acc_cosine = 0.0
 *             for i in range(n):
 */

#line 423 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_dpeak_set->__pyx_vtab)->get_size(__pyx_v_dpeak_set);

    /* "ms_deisotope/_c/similarity_methods.pyx":424
 *             dpeak_set = <DeconvolutedPeakSet>self.peak_set_b
 *             n = dpeak_set.get_size()
 *             acc_cosine = 0.0             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 intensity = dpeak_set.getitem(i).intensity
 */

#line 424 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_acc_cosine = 0.0;

    /* "ms_deisotope/_c/similarity_methods.pyx":425
 *             n = dpeak_set.get_size()
 *             acc_cosine = 0.0
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 intensity = dpeak_set.getitem(i).intensity
 *                 if sqrt_transform:
 */

#line 425 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __pyx_v_n;

#line 425 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = __pyx_t_6;

#line 425 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 425 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_i = __pyx_t_8;

      /* "ms_deisotope/_c/similarity_methods.pyx":426
 *             acc_cosine = 0.0
 *             for i in range(n):
 *                 intensity = dpeak_set.getitem(i).intensity             # <<<<<<<<<<<<<<
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)
 */

#line 426 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_dpeak_set->__pyx_vtab)->getitem(__pyx_v_dpeak_set, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)

#line 426 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 426 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_9 = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_1)->__pyx_base.intensity;

#line 426 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 426 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_intensity = __pyx_t_9;

      /* "ms_deisotope/_c/similarity_methods.pyx":427
 *             for i in range(n):
 *                 intensity = dpeak_set.getitem(i).intensity
 *                 if sqrt_transform:             # <<<<<<<<<<<<<<
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 */

#line 427 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_5 = (__pyx_v_sqrt_transform != 0);

#line 427 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_5) {

        /* "ms_deisotope/_c/similarity_methods.pyx":428
 *                 intensity = dpeak_set.getitem(i).intensity
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)             # <<<<<<<<<<<<<<
 *                 acc_cosine += intensity ** 2
 *             self.norm_bb_cosine = sqrt(acc_cosine)
 */

#line 428 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_intensity = sqrt(__pyx_v_intensity);

        /* "ms_deisotope/_c/similarity_methods.pyx":427
 *             for i in range(n):
 *                 intensity = dpeak_set.getitem(i).intensity
 *                 if sqrt_transform:             # <<<<<<<<<<<<<<
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 */

#line 427 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":429
 *                 if sqrt_transform:
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2             # <<<<<<<<<<<<<<
 *             self.norm_bb_cosine = sqrt(acc_cosine)
 * 
 */

#line 429 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_acc_cosine = (__pyx_v_acc_cosine + pow(__pyx_v_intensity, 2.0));

#line 429 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":430
 *                     intensity = sqrt(intensity)
 *                 acc_cosine += intensity ** 2
 *             self.norm_bb_cosine = sqrt(acc_cosine)             # <<<<<<<<<<<<<<
 * 
 *     cpdef _determine_peak_type(self):
 */

#line 430 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_self->norm_bb_cosine = sqrt(__pyx_v_acc_cosine);

    /* "ms_deisotope/_c/similarity_methods.pyx":412
 *                 acc_cosine += intensity ** 2
 *             self.norm_bb_cosine = sqrt(acc_cosine)
 *         elif self.peak_type == PeakType.Deconvoluted:             # <<<<<<<<<<<<<<
 *             dpeak_set = <DeconvolutedPeakSet>self.peak_set_a
 *             n = dpeak_set.get_size()
 */

#line 412 "ms_deisotope/_c/similarity_methods.pyx"
    break;

#line 412 "ms_deisotope/_c/similarity_methods.pyx"
    default: break;

#line 412 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":384
 *         self.align(error_tolerance, shift)
 * 
 *     cpdef calculate_normalization(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 384 "ms_deisotope/_c/similarity_methods.pyx"


#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.calculate_normalization", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak_set);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_dpeak_set);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 384 "ms_deisotope/_c/similarity_methods.pyx"


#line 384 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_3calculate_normalization(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_2calculate_normalization[] = "SpectrumAlignment.calculate_normalization(self)";

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_3calculate_normalization(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("calculate_normalization (wrapper)", 0);
  __pyx_r = 
#line 384 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_2calculate_normalization(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 384 "ms_deisotope/_c/similarity_methods.pyx"


#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 384 "ms_deisotope/_c/similarity_methods.pyx"


#line 384 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_2calculate_normalization(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("calculate_normalization", 0);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("calculate_normalization (wrapper)", __pyx_f[0], 384, 0, __PYX_ERR(0, 384, __pyx_L1_error));

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_calculate_normalization(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"


#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.calculate_normalization", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 384 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":432
 *             self.norm_bb_cosine = sqrt(acc_cosine)
 * 
 *     cpdef _determine_peak_type(self):             # <<<<<<<<<<<<<<
 *         if isinstance(self.peak_set_a, DeconvolutedPeakSet):
 *             if isinstance(self.peak_set_b, DeconvolutedPeakSet):
 */

#line 432 "ms_deisotope/_c/similarity_methods.pyx"


#line 432 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5_determine_peak_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment__determine_peak_type(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, int __pyx_skip_dispatch) {

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_5;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_6;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_determine_peak_type", 0);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("_determine_peak_type", __pyx_f[0], 432, 0, __PYX_ERR(0, 432, __pyx_L1_error));

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  /* Check if called by wrapper */

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  /* Check if overridden in Python */

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_determine_peak_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5_determine_peak_type)) {

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 432 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_4 = NULL;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
          if (likely(__pyx_t_4)) {

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_INCREF(function);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
          }

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_r = __pyx_t_2;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_2 = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
    #endif

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":433
 * 
 *     cpdef _determine_peak_type(self):
 *         if isinstance(self.peak_set_a, DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *             if isinstance(self.peak_set_b, DeconvolutedPeakSet):
 *                 self.peak_type = PeakType.Deconvoluted
 */

#line 433 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_self->peak_set_a;

#line 433 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 433 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet); 

#line 433 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 433 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 433 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/similarity_methods.pyx":434
 *     cpdef _determine_peak_type(self):
 *         if isinstance(self.peak_set_a, DeconvolutedPeakSet):
 *             if isinstance(self.peak_set_b, DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *                 self.peak_type = PeakType.Deconvoluted
 *             else:
 */

#line 434 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_self->peak_set_b;

#line 434 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 434 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet); 

#line 434 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 434 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = (__pyx_t_6 != 0);

#line 434 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(__pyx_t_5)) {

      /* "ms_deisotope/_c/similarity_methods.pyx":435
 *         if isinstance(self.peak_set_a, DeconvolutedPeakSet):
 *             if isinstance(self.peak_set_b, DeconvolutedPeakSet):
 *                 self.peak_type = PeakType.Deconvoluted             # <<<<<<<<<<<<<<
 *             else:
 *                 raise TypeError(
 */

#line 435 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_self->peak_type = __pyx_e_12ms_deisotope_2_c_18similarity_methods_Deconvoluted;

      /* "ms_deisotope/_c/similarity_methods.pyx":434
 *     cpdef _determine_peak_type(self):
 *         if isinstance(self.peak_set_a, DeconvolutedPeakSet):
 *             if isinstance(self.peak_set_b, DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *                 self.peak_type = PeakType.Deconvoluted
 *             else:
 */

#line 434 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L4;

#line 434 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":437
 *                 self.peak_type = PeakType.Deconvoluted
 *             else:
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 *         elif isinstance(self.peak_set_a, PeakSet):
 */

#line 437 "ms_deisotope/_c/similarity_methods.pyx"
    /*else*/ {

      /* "ms_deisotope/_c/similarity_methods.pyx":438
 *             else:
 *                 raise TypeError(
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))             # <<<<<<<<<<<<<<
 *         elif isinstance(self.peak_set_a, PeakSet):
 *             if isinstance(self.peak_set_b, PeakSet):
 */

#line 438 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 438, __pyx_L1_error)

#line 438 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 438 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_a)));

#line 438 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_a)));

#line 438 "ms_deisotope/_c/similarity_methods.pyx"
      PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(__pyx_v_self->peak_set_a)));

#line 438 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_b)));

#line 438 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_b)));

#line 438 "ms_deisotope/_c/similarity_methods.pyx"
      PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)Py_TYPE(__pyx_v_self->peak_set_b)));

#line 438 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Peak_set_types_must_match_got_s, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)

#line 438 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 438 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":437
 *                 self.peak_type = PeakType.Deconvoluted
 *             else:
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 *         elif isinstance(self.peak_set_a, PeakSet):
 */

#line 437 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L1_error)

#line 437 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 437 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 437 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 437 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 437 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 437, __pyx_L1_error)

#line 437 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 437 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_L4:;

    /* "ms_deisotope/_c/similarity_methods.pyx":433
 * 
 *     cpdef _determine_peak_type(self):
 *         if isinstance(self.peak_set_a, DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *             if isinstance(self.peak_set_b, DeconvolutedPeakSet):
 *                 self.peak_type = PeakType.Deconvoluted
 */

#line 433 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L3;

#line 433 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":439
 *                 raise TypeError(
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 *         elif isinstance(self.peak_set_a, PeakSet):             # <<<<<<<<<<<<<<
 *             if isinstance(self.peak_set_b, PeakSet):
 *                 self.peak_type = PeakType.Fitted
 */

#line 439 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_self->peak_set_a;

#line 439 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 439 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet); 

#line 439 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 439 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 439 "ms_deisotope/_c/similarity_methods.pyx"
  if (likely(__pyx_t_6)) {

    /* "ms_deisotope/_c/similarity_methods.pyx":440
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 *         elif isinstance(self.peak_set_a, PeakSet):
 *             if isinstance(self.peak_set_b, PeakSet):             # <<<<<<<<<<<<<<
 *                 self.peak_type = PeakType.Fitted
 *             else:
 */

#line 440 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_self->peak_set_b;

#line 440 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 440 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet); 

#line 440 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 440 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = (__pyx_t_6 != 0);

#line 440 "ms_deisotope/_c/similarity_methods.pyx"
    if (likely(__pyx_t_5)) {

      /* "ms_deisotope/_c/similarity_methods.pyx":441
 *         elif isinstance(self.peak_set_a, PeakSet):
 *             if isinstance(self.peak_set_b, PeakSet):
 *                 self.peak_type = PeakType.Fitted             # <<<<<<<<<<<<<<
 *             else:
 *                 raise TypeError(
 */

#line 441 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_self->peak_type = __pyx_e_12ms_deisotope_2_c_18similarity_methods_Fitted;

      /* "ms_deisotope/_c/similarity_methods.pyx":440
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 *         elif isinstance(self.peak_set_a, PeakSet):
 *             if isinstance(self.peak_set_b, PeakSet):             # <<<<<<<<<<<<<<
 *                 self.peak_type = PeakType.Fitted
 *             else:
 */

#line 440 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L5;

#line 440 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":443
 *                 self.peak_type = PeakType.Fitted
 *             else:
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 *         else:
 */

#line 443 "ms_deisotope/_c/similarity_methods.pyx"
    /*else*/ {

      /* "ms_deisotope/_c/similarity_methods.pyx":444
 *             else:
 *                 raise TypeError(
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError(
 */

#line 444 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error)

#line 444 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 444 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_a)));

#line 444 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_a)));

#line 444 "ms_deisotope/_c/similarity_methods.pyx"
      PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(__pyx_v_self->peak_set_a)));

#line 444 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_b)));

#line 444 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_b)));

#line 444 "ms_deisotope/_c/similarity_methods.pyx"
      PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)Py_TYPE(__pyx_v_self->peak_set_b)));

#line 444 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Peak_set_types_must_match_got_s, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 444, __pyx_L1_error)

#line 444 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 444 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":443
 *                 self.peak_type = PeakType.Fitted
 *             else:
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 *         else:
 */

#line 443 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 443, __pyx_L1_error)

#line 443 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 443 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 443 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 443 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 443 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 443, __pyx_L1_error)

#line 443 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 443 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_L5:;

    /* "ms_deisotope/_c/similarity_methods.pyx":439
 *                 raise TypeError(
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 *         elif isinstance(self.peak_set_a, PeakSet):             # <<<<<<<<<<<<<<
 *             if isinstance(self.peak_set_b, PeakSet):
 *                 self.peak_type = PeakType.Fitted
 */

#line 439 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L3;

#line 439 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":446
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 * 
 */

#line 446 "ms_deisotope/_c/similarity_methods.pyx"
  /*else*/ {

    /* "ms_deisotope/_c/similarity_methods.pyx":447
 *         else:
 *             raise TypeError(
 *                 "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))             # <<<<<<<<<<<<<<
 * 
 *     cpdef align(self, double error_tolerance=2e-5, double shift=0.0):
 */

#line 447 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)

#line 447 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 447 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_a)));

#line 447 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_a)));

#line 447 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(__pyx_v_self->peak_set_a)));

#line 447 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_b)));

#line 447 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_self->peak_set_b)));

#line 447 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)Py_TYPE(__pyx_v_self->peak_set_b)));

#line 447 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Peak_set_types_must_match_got_s, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)

#line 447 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 447 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":446
 *                     "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 * 
 */

#line 446 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)

#line 446 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 446 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 446 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 446 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 446 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 446, __pyx_L1_error)

#line 446 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 446 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3:;

  /* "ms_deisotope/_c/similarity_methods.pyx":432
 *             self.norm_bb_cosine = sqrt(acc_cosine)
 * 
 *     cpdef _determine_peak_type(self):             # <<<<<<<<<<<<<<
 *         if isinstance(self.peak_set_a, DeconvolutedPeakSet):
 *             if isinstance(self.peak_set_b, DeconvolutedPeakSet):
 */

#line 432 "ms_deisotope/_c/similarity_methods.pyx"


#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment._determine_peak_type", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 432 "ms_deisotope/_c/similarity_methods.pyx"


#line 432 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5_determine_peak_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_4_determine_peak_type[] = "SpectrumAlignment._determine_peak_type(self)";

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5_determine_peak_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_determine_peak_type (wrapper)", 0);
  __pyx_r = 
#line 432 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_4_determine_peak_type(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 432 "ms_deisotope/_c/similarity_methods.pyx"


#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 432 "ms_deisotope/_c/similarity_methods.pyx"


#line 432 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_4_determine_peak_type(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_determine_peak_type", 0);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("_determine_peak_type (wrapper)", __pyx_f[0], 432, 0, __PYX_ERR(0, 432, __pyx_L1_error));

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment__determine_peak_type(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"


#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment._determine_peak_type", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 432 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":449
 *                 "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 * 
 *     cpdef align(self, double error_tolerance=2e-5, double shift=0.0):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list pairs
 */

#line 449 "ms_deisotope/_c/similarity_methods.pyx"


#line 449 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_7align(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_align(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_align *__pyx_optional_args) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_error_tolerance = ((double)2e-5);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_shift = ((double)0.0);
  PyObject *__pyx_v_pairs = 0
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_7;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets __pyx_t_9;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_10;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_11;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_deconvoluted_peak_sets __pyx_t_12;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("align", 0);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("align", __pyx_f[0], 449, 0, __PYX_ERR(0, 449, __pyx_L1_error));

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_shift = __pyx_optional_args->shift;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  /* Check if called by wrapper */

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  /* Check if overridden in Python */

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_align); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_7align)) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 449, __pyx_L1_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_shift); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 449, __pyx_L1_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; 
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_6 = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_7 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          if (likely(__pyx_t_6)) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_INCREF(__pyx_t_6);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_INCREF(function);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_DECREF_SET(__pyx_t_5, function);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
            __pyx_t_7 = 1;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        #if CYTHON_FAST_PYCALL

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        if (PyFunction_Check(__pyx_t_5)) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 449, __pyx_L1_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        } else

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        #if CYTHON_FAST_PYCCALL

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 449, __pyx_L1_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        } else

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        #endif

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 449, __pyx_L1_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_GOTREF(__pyx_t_8);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          if (__pyx_t_6) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_3 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_4 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 449, __pyx_L1_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_r = __pyx_t_2;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_2 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    #endif

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":452
 *         cdef:
 *             list pairs
 *         pairs = []             # <<<<<<<<<<<<<<
 *         if self.peak_type == PeakType.Fitted:
 *             pairs = convolve_peak_sets(<PeakSet>self.peak_set_a, <PeakSet>self.peak_set_b, error_tolerance)
 */

#line 452 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)

#line 452 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 452 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_pairs = ((PyObject*)__pyx_t_1);

#line 452 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":453
 *             list pairs
 *         pairs = []
 *         if self.peak_type == PeakType.Fitted:             # <<<<<<<<<<<<<<
 *             pairs = convolve_peak_sets(<PeakSet>self.peak_set_a, <PeakSet>self.peak_set_b, error_tolerance)
 *             if shift != 0:
 */

#line 453 "ms_deisotope/_c/similarity_methods.pyx"
  switch (__pyx_v_self->peak_type) {

#line 453 "ms_deisotope/_c/similarity_methods.pyx"
    case __pyx_e_12ms_deisotope_2_c_18similarity_methods_Fitted:

    /* "ms_deisotope/_c/similarity_methods.pyx":454
 *         pairs = []
 *         if self.peak_type == PeakType.Fitted:
 *             pairs = convolve_peak_sets(<PeakSet>self.peak_set_a, <PeakSet>self.peak_set_b, error_tolerance)             # <<<<<<<<<<<<<<
 *             if shift != 0:
 *                 pairs.extend(
 */

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_self->peak_set_a;

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = __pyx_v_self->peak_set_b;

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9.__pyx_n = 1;

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9.error_tolerance = __pyx_v_error_tolerance;

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2), &__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 454, __pyx_L1_error)

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_pairs, ((PyObject*)__pyx_t_5));

#line 454 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":455
 *         if self.peak_type == PeakType.Fitted:
 *             pairs = convolve_peak_sets(<PeakSet>self.peak_set_a, <PeakSet>self.peak_set_b, error_tolerance)
 *             if shift != 0:             # <<<<<<<<<<<<<<
 *                 pairs.extend(
 *                     convolve_peak_sets(
 */

#line 455 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = ((__pyx_v_shift != 0.0) != 0);

#line 455 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/similarity_methods.pyx":456
 *             pairs = convolve_peak_sets(<PeakSet>self.peak_set_a, <PeakSet>self.peak_set_b, error_tolerance)
 *             if shift != 0:
 *                 pairs.extend(             # <<<<<<<<<<<<<<
 *                     convolve_peak_sets(
 *                         <PeakSet>self.peak_set_a,
 */

#line 456 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__pyx_v_pairs == Py_None)) {

#line 456 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");

#line 456 "ms_deisotope/_c/similarity_methods.pyx"
        __PYX_ERR(0, 456, __pyx_L1_error)

#line 456 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":458
 *                 pairs.extend(
 *                     convolve_peak_sets(
 *                         <PeakSet>self.peak_set_a,             # <<<<<<<<<<<<<<
 *                         <PeakSet>self.peak_set_b, error_tolerance, shift))
 *         elif self.peak_type == PeakType.Deconvoluted:
 */

#line 458 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_5 = __pyx_v_self->peak_set_a;

#line 458 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_5);

      /* "ms_deisotope/_c/similarity_methods.pyx":459
 *                     convolve_peak_sets(
 *                         <PeakSet>self.peak_set_a,
 *                         <PeakSet>self.peak_set_b, error_tolerance, shift))             # <<<<<<<<<<<<<<
 *         elif self.peak_type == PeakType.Deconvoluted:
 *             pairs = convolve_deconvoluted_peak_sets(
 */

#line 459 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_2 = __pyx_v_self->peak_set_b;

#line 459 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_2);

      /* "ms_deisotope/_c/similarity_methods.pyx":457
 *             if shift != 0:
 *                 pairs.extend(
 *                     convolve_peak_sets(             # <<<<<<<<<<<<<<
 *                         <PeakSet>self.peak_set_a,
 *                         <PeakSet>self.peak_set_b, error_tolerance, shift))
 */

#line 457 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_9.__pyx_n = 2;

#line 457 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_9.error_tolerance = __pyx_v_error_tolerance;

#line 457 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_9.shift = __pyx_v_shift;

#line 457 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_5), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2), &__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)

#line 457 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 457 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 457 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":456
 *             pairs = convolve_peak_sets(<PeakSet>self.peak_set_a, <PeakSet>self.peak_set_b, error_tolerance)
 *             if shift != 0:
 *                 pairs.extend(             # <<<<<<<<<<<<<<
 *                     convolve_peak_sets(
 *                         <PeakSet>self.peak_set_a,
 */

#line 456 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_pairs, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 456, __pyx_L1_error)

#line 456 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":455
 *         if self.peak_type == PeakType.Fitted:
 *             pairs = convolve_peak_sets(<PeakSet>self.peak_set_a, <PeakSet>self.peak_set_b, error_tolerance)
 *             if shift != 0:             # <<<<<<<<<<<<<<
 *                 pairs.extend(
 *                     convolve_peak_sets(
 */

#line 455 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":453
 *             list pairs
 *         pairs = []
 *         if self.peak_type == PeakType.Fitted:             # <<<<<<<<<<<<<<
 *             pairs = convolve_peak_sets(<PeakSet>self.peak_set_a, <PeakSet>self.peak_set_b, error_tolerance)
 *             if shift != 0:
 */

#line 453 "ms_deisotope/_c/similarity_methods.pyx"
    break;

#line 453 "ms_deisotope/_c/similarity_methods.pyx"
    case __pyx_e_12ms_deisotope_2_c_18similarity_methods_Deconvoluted:

    /* "ms_deisotope/_c/similarity_methods.pyx":462
 *         elif self.peak_type == PeakType.Deconvoluted:
 *             pairs = convolve_deconvoluted_peak_sets(
 *                 <DeconvolutedPeakSet>self.peak_set_a,             # <<<<<<<<<<<<<<
 *                 <DeconvolutedPeakSet>self.peak_set_b, error_tolerance)
 *             if shift != 0:
 */

#line 462 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __pyx_v_self->peak_set_a;

#line 462 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

    /* "ms_deisotope/_c/similarity_methods.pyx":463
 *             pairs = convolve_deconvoluted_peak_sets(
 *                 <DeconvolutedPeakSet>self.peak_set_a,
 *                 <DeconvolutedPeakSet>self.peak_set_b, error_tolerance)             # <<<<<<<<<<<<<<
 *             if shift != 0:
 *                 pairs.extend(
 */

#line 463 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = __pyx_v_self->peak_set_b;

#line 463 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_2);

    /* "ms_deisotope/_c/similarity_methods.pyx":461
 *                         <PeakSet>self.peak_set_b, error_tolerance, shift))
 *         elif self.peak_type == PeakType.Deconvoluted:
 *             pairs = convolve_deconvoluted_peak_sets(             # <<<<<<<<<<<<<<
 *                 <DeconvolutedPeakSet>self.peak_set_a,
 *                 <DeconvolutedPeakSet>self.peak_set_b, error_tolerance)
 */

#line 461 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_12.__pyx_n = 1;

#line 461 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_12.error_tolerance = __pyx_v_error_tolerance;

#line 461 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_deconvoluted_peak_sets(((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_t_1), ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_t_2), &__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 461, __pyx_L1_error)

#line 461 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 461 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 461 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 461 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_pairs, ((PyObject*)__pyx_t_5));

#line 461 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":464
 *                 <DeconvolutedPeakSet>self.peak_set_a,
 *                 <DeconvolutedPeakSet>self.peak_set_b, error_tolerance)
 *             if shift != 0:             # <<<<<<<<<<<<<<
 *                 pairs.extend(
 *                     convolve_deconvoluted_peak_sets(
 */

#line 464 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = ((__pyx_v_shift != 0.0) != 0);

#line 464 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/similarity_methods.pyx":465
 *                 <DeconvolutedPeakSet>self.peak_set_b, error_tolerance)
 *             if shift != 0:
 *                 pairs.extend(             # <<<<<<<<<<<<<<
 *                     convolve_deconvoluted_peak_sets(
 *                         <DeconvolutedPeakSet>self.peak_set_a,
 */

#line 465 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__pyx_v_pairs == Py_None)) {

#line 465 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");

#line 465 "ms_deisotope/_c/similarity_methods.pyx"
        __PYX_ERR(0, 465, __pyx_L1_error)

#line 465 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":467
 *                 pairs.extend(
 *                     convolve_deconvoluted_peak_sets(
 *                         <DeconvolutedPeakSet>self.peak_set_a,             # <<<<<<<<<<<<<<
 *                         <DeconvolutedPeakSet>self.peak_set_b, error_tolerance, shift))
 *         self.peak_pairs = pairs
 */

#line 467 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_5 = __pyx_v_self->peak_set_a;

#line 467 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_5);

      /* "ms_deisotope/_c/similarity_methods.pyx":468
 *                     convolve_deconvoluted_peak_sets(
 *                         <DeconvolutedPeakSet>self.peak_set_a,
 *                         <DeconvolutedPeakSet>self.peak_set_b, error_tolerance, shift))             # <<<<<<<<<<<<<<
 *         self.peak_pairs = pairs
 *         self._calculate_score()
 */

#line 468 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_2 = __pyx_v_self->peak_set_b;

#line 468 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_2);

      /* "ms_deisotope/_c/similarity_methods.pyx":466
 *             if shift != 0:
 *                 pairs.extend(
 *                     convolve_deconvoluted_peak_sets(             # <<<<<<<<<<<<<<
 *                         <DeconvolutedPeakSet>self.peak_set_a,
 *                         <DeconvolutedPeakSet>self.peak_set_b, error_tolerance, shift))
 */

#line 466 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_12.__pyx_n = 2;

#line 466 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_12.error_tolerance = __pyx_v_error_tolerance;

#line 466 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_12.shift = __pyx_v_shift;

#line 466 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_deconvoluted_peak_sets(((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_t_5), ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_t_2), &__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)

#line 466 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 466 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 466 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":465
 *                 <DeconvolutedPeakSet>self.peak_set_b, error_tolerance)
 *             if shift != 0:
 *                 pairs.extend(             # <<<<<<<<<<<<<<
 *                     convolve_deconvoluted_peak_sets(
 *                         <DeconvolutedPeakSet>self.peak_set_a,
 */

#line 465 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_pairs, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 465, __pyx_L1_error)

#line 465 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":464
 *                 <DeconvolutedPeakSet>self.peak_set_a,
 *                 <DeconvolutedPeakSet>self.peak_set_b, error_tolerance)
 *             if shift != 0:             # <<<<<<<<<<<<<<
 *                 pairs.extend(
 *                     convolve_deconvoluted_peak_sets(
 */

#line 464 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":460
 *                         <PeakSet>self.peak_set_a,
 *                         <PeakSet>self.peak_set_b, error_tolerance, shift))
 *         elif self.peak_type == PeakType.Deconvoluted:             # <<<<<<<<<<<<<<
 *             pairs = convolve_deconvoluted_peak_sets(
 *                 <DeconvolutedPeakSet>self.peak_set_a,
 */

#line 460 "ms_deisotope/_c/similarity_methods.pyx"
    break;

#line 460 "ms_deisotope/_c/similarity_methods.pyx"
    default: break;

#line 460 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":469
 *                         <DeconvolutedPeakSet>self.peak_set_a,
 *                         <DeconvolutedPeakSet>self.peak_set_b, error_tolerance, shift))
 *         self.peak_pairs = pairs             # <<<<<<<<<<<<<<
 *         self._calculate_score()
 * 
 */

#line 469 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_pairs);

#line 469 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_v_pairs);

#line 469 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_pairs);

#line 469 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_v_self->peak_pairs);

#line 469 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_pairs = __pyx_v_pairs;

  /* "ms_deisotope/_c/similarity_methods.pyx":470
 *                         <DeconvolutedPeakSet>self.peak_set_b, error_tolerance, shift))
 *         self.peak_pairs = pairs
 *         self._calculate_score()             # <<<<<<<<<<<<<<
 * 
 *     cpdef _calculate_score(self):
 */

#line 470 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self->__pyx_vtab)->_calculate_score(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 470, __pyx_L1_error)

#line 470 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 470 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":449
 *                 "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
 * 
 *     cpdef align(self, double error_tolerance=2e-5, double shift=0.0):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list pairs
 */

#line 449 "ms_deisotope/_c/similarity_methods.pyx"


#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.align", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_pairs);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 449 "ms_deisotope/_c/similarity_methods.pyx"


#line 449 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_7align(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_6align[] = "SpectrumAlignment.align(self, double error_tolerance=2e-5, double shift=0.0)";

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_7align(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_error_tolerance
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_shift
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("align (wrapper)", 0);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_error_tolerance,&__pyx_n_s_shift,0};

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[2] = {0,0};

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  2: 
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shift);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "align") < 0)) __PYX_ERR(0, 449, __pyx_L3_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: 
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
break;
        default: 
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[0]) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L3_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = ((double)2e-5);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[1]) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_shift == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L3_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = ((double)0.0);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("align", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 449, __pyx_L3_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.align", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 449 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_6align(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), __pyx_v_error_tolerance, __pyx_v_shift);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"


#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 449 "ms_deisotope/_c/similarity_methods.pyx"


#line 449 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_6align(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, double __pyx_v_error_tolerance, double __pyx_v_shift) {

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_align __pyx_t_2;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("align", 0);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("align (wrapper)", __pyx_f[0], 449, 0, __PYX_ERR(0, 449, __pyx_L1_error));

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 2;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.shift = __pyx_v_shift;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment->align(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"


#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.align", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 449 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":472
 *         self._calculate_score()
 * 
 *     cpdef _calculate_score(self):             # <<<<<<<<<<<<<<
 *         self.score = convolved_dot_product(
 *             self.peak_pairs, self.peak_type, sqrt_transform=self.sqrt_transform)
 */

#line 472 "ms_deisotope/_c/similarity_methods.pyx"


#line 472 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9_calculate_score(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment__calculate_score(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, int __pyx_skip_dispatch) {

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_5;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_6;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_calculate_score", 0);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("_calculate_score", __pyx_f[0], 472, 0, __PYX_ERR(0, 472, __pyx_L1_error));

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  /* Check if called by wrapper */

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  /* Check if overridden in Python */

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_calculate_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9_calculate_score)) {

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 472 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_t_4 = NULL;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
          if (likely(__pyx_t_4)) {

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_INCREF(function);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
          }

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_r = __pyx_t_2;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_2 = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      #endif

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
    #endif

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":473
 * 
 *     cpdef _calculate_score(self):
 *         self.score = convolved_dot_product(             # <<<<<<<<<<<<<<
 *             self.peak_pairs, self.peak_type, sqrt_transform=self.sqrt_transform)
 *         if self.normalize:
 */

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_convolved_dot_product); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 473, __pyx_L1_error)

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

  /* "ms_deisotope/_c/similarity_methods.pyx":474
 *     cpdef _calculate_score(self):
 *         self.score = convolved_dot_product(
 *             self.peak_pairs, self.peak_type, sqrt_transform=self.sqrt_transform)             # <<<<<<<<<<<<<<
 *         if self.normalize:
 *             self.score /= (self.norm_aa_cosine * self.norm_bb_cosine)
 */

#line 474 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(__pyx_v_self->peak_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 474, __pyx_L1_error)

#line 474 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

  /* "ms_deisotope/_c/similarity_methods.pyx":473
 * 
 *     cpdef _calculate_score(self):
 *         self.score = convolved_dot_product(             # <<<<<<<<<<<<<<
 *             self.peak_pairs, self.peak_type, sqrt_transform=self.sqrt_transform)
 *         if self.normalize:
 */

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 473, __pyx_L1_error)

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_self->peak_pairs);

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_v_self->peak_pairs);

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->peak_pairs);

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":474
 *     cpdef _calculate_score(self):
 *         self.score = convolved_dot_product(
 *             self.peak_pairs, self.peak_type, sqrt_transform=self.sqrt_transform)             # <<<<<<<<<<<<<<
 *         if self.normalize:
 *             self.score /= (self.norm_aa_cosine * self.norm_bb_cosine)
 */

#line 474 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 474, __pyx_L1_error)

#line 474 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 474 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->sqrt_transform); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 474, __pyx_L1_error)

#line 474 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 474 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_sqrt_transform, __pyx_t_4) < 0) __PYX_ERR(0, 474, __pyx_L1_error)

#line 474 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":473
 * 
 *     cpdef _calculate_score(self):
 *         self.score = convolved_dot_product(             # <<<<<<<<<<<<<<
 *             self.peak_pairs, self.peak_type, sqrt_transform=self.sqrt_transform)
 *         if self.normalize:
 */

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 473, __pyx_L1_error)

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 473, __pyx_L1_error)

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 473 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->score = __pyx_t_5;

  /* "ms_deisotope/_c/similarity_methods.pyx":475
 *         self.score = convolved_dot_product(
 *             self.peak_pairs, self.peak_type, sqrt_transform=self.sqrt_transform)
 *         if self.normalize:             # <<<<<<<<<<<<<<
 *             self.score /= (self.norm_aa_cosine * self.norm_bb_cosine)
 * 
 */

#line 475 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = (__pyx_v_self->normalize != 0);

#line 475 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/similarity_methods.pyx":476
 *             self.peak_pairs, self.peak_type, sqrt_transform=self.sqrt_transform)
 *         if self.normalize:
 *             self.score /= (self.norm_aa_cosine * self.norm_bb_cosine)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 476 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = (__pyx_v_self->norm_aa_cosine * __pyx_v_self->norm_bb_cosine);

#line 476 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_5 == 0)) {

#line 476 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 476 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 476, __pyx_L1_error)

#line 476 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 476 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_self->score = (__pyx_v_self->score / __pyx_t_5);

    /* "ms_deisotope/_c/similarity_methods.pyx":475
 *         self.score = convolved_dot_product(
 *             self.peak_pairs, self.peak_type, sqrt_transform=self.sqrt_transform)
 *         if self.normalize:             # <<<<<<<<<<<<<<
 *             self.score /= (self.norm_aa_cosine * self.norm_bb_cosine)
 * 
 */

#line 475 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":472
 *         self._calculate_score()
 * 
 *     cpdef _calculate_score(self):             # <<<<<<<<<<<<<<
 *         self.score = convolved_dot_product(
 *             self.peak_pairs, self.peak_type, sqrt_transform=self.sqrt_transform)
 */

#line 472 "ms_deisotope/_c/similarity_methods.pyx"


#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment._calculate_score", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 472 "ms_deisotope/_c/similarity_methods.pyx"


#line 472 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9_calculate_score(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_8_calculate_score[] = "SpectrumAlignment._calculate_score(self)";

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9_calculate_score(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_calculate_score (wrapper)", 0);
  __pyx_r = 
#line 472 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_8_calculate_score(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 472 "ms_deisotope/_c/similarity_methods.pyx"


#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 472 "ms_deisotope/_c/similarity_methods.pyx"


#line 472 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_8_calculate_score(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("_calculate_score", 0);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("_calculate_score (wrapper)", __pyx_f[0], 472, 0, __PYX_ERR(0, 472, __pyx_L1_error));

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment__calculate_score(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"


#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment._calculate_score", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 472 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":355
 * cdef class SpectrumAlignment(object):
 *     cdef:
 *         public list peak_pairs             # <<<<<<<<<<<<<<
 *         public double shift
 *         public double score
 */

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_1__get__(PyObject *__pyx_v_self) {

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 355 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 355, 0, __PYX_ERR(0, 355, __pyx_L1_error));

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_self->peak_pairs);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_self->peak_pairs;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.peak_pairs.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 355 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value) {

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 355, 0, __PYX_ERR(0, 355, __pyx_L1_error));

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 355, __pyx_L1_error)

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_value;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_pairs);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_v_self->peak_pairs);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_pairs = ((PyObject*)__pyx_t_1);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.peak_pairs.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_5__del__(PyObject *__pyx_v_self) {

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 355 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_4__del__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__del__", __pyx_f[0], 355, 0, __PYX_ERR(0, 355, __pyx_L1_error));

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(Py_None);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(Py_None);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_pairs);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_v_self->peak_pairs);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_pairs = ((PyObject*)Py_None);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"


#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.peak_pairs.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 355 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":356
 *     cdef:
 *         public list peak_pairs
 *         public double shift             # <<<<<<<<<<<<<<
 *         public double score
 *         public object peak_set_a
 */

#line 356 "ms_deisotope/_c/similarity_methods.pyx"


#line 356 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift_1__get__(PyObject *__pyx_v_self) {

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 356 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 356 "ms_deisotope/_c/similarity_methods.pyx"


#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 356 "ms_deisotope/_c/similarity_methods.pyx"


#line 356 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 356, 0, __PYX_ERR(0, 356, __pyx_L1_error));

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->shift); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 356, __pyx_L1_error)

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"


#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.shift.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 356 "ms_deisotope/_c/similarity_methods.pyx"


#line 356 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 356 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 356 "ms_deisotope/_c/similarity_methods.pyx"


#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 356 "ms_deisotope/_c/similarity_methods.pyx"


#line 356 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value) {

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 356, 0, __PYX_ERR(0, 356, __pyx_L1_error));

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 356, __pyx_L1_error)

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->shift = __pyx_t_1;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"


#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.shift.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 356 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":357
 *         public list peak_pairs
 *         public double shift
 *         public double score             # <<<<<<<<<<<<<<
 *         public object peak_set_a
 *         public object peak_set_b
 */

#line 357 "ms_deisotope/_c/similarity_methods.pyx"


#line 357 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score_1__get__(PyObject *__pyx_v_self) {

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 357 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 357 "ms_deisotope/_c/similarity_methods.pyx"


#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 357 "ms_deisotope/_c/similarity_methods.pyx"


#line 357 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 357, 0, __PYX_ERR(0, 357, __pyx_L1_error));

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"


#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.score.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 357 "ms_deisotope/_c/similarity_methods.pyx"


#line 357 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 357 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 357 "ms_deisotope/_c/similarity_methods.pyx"


#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 357 "ms_deisotope/_c/similarity_methods.pyx"


#line 357 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value) {

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 357, 0, __PYX_ERR(0, 357, __pyx_L1_error));

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L1_error)

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->score = __pyx_t_1;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"


#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.score.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 357 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":358
 *         public double shift
 *         public double score
 *         public object peak_set_a             # <<<<<<<<<<<<<<
 *         public object peak_set_b
 *         public PeakType peak_type
 */

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_1__get__(PyObject *__pyx_v_self) {

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 358 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 358, 0, __PYX_ERR(0, 358, __pyx_L1_error));

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_self->peak_set_a);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_self->peak_set_a;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.peak_set_a.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 358 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value) {

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 358, 0, __PYX_ERR(0, 358, __pyx_L1_error));

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_value);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_v_value);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_set_a);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_v_self->peak_set_a);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_set_a = __pyx_v_value;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.peak_set_a.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_5__del__(PyObject *__pyx_v_self) {

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 358 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_4__del__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__del__", __pyx_f[0], 358, 0, __PYX_ERR(0, 358, __pyx_L1_error));

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(Py_None);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(Py_None);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_set_a);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_v_self->peak_set_a);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_set_a = Py_None;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"


#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.peak_set_a.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 358 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":359
 *         public double score
 *         public object peak_set_a
 *         public object peak_set_b             # <<<<<<<<<<<<<<
 *         public PeakType peak_type
 *         public double norm_aa_cosine
 */

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_1__get__(PyObject *__pyx_v_self) {

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 359 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 359, 0, __PYX_ERR(0, 359, __pyx_L1_error));

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_self->peak_set_b);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_self->peak_set_b;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.peak_set_b.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 359 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value) {

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 359, 0, __PYX_ERR(0, 359, __pyx_L1_error));

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_value);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_v_value);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_set_b);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_v_self->peak_set_b);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_set_b = __pyx_v_value;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.peak_set_b.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_5__del__(PyObject *__pyx_v_self) {

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 359 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_4__del__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__del__", __pyx_f[0], 359, 0, __PYX_ERR(0, 359, __pyx_L1_error));

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(Py_None);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(Py_None);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_v_self->peak_set_b);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_v_self->peak_set_b);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_set_b = Py_None;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"


#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.peak_set_b.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 359 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":360
 *         public object peak_set_a
 *         public object peak_set_b
 *         public PeakType peak_type             # <<<<<<<<<<<<<<
 *         public double norm_aa_cosine
 *         public double norm_bb_cosine
 */

#line 360 "ms_deisotope/_c/similarity_methods.pyx"


#line 360 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type_1__get__(PyObject *__pyx_v_self) {

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 360 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 360 "ms_deisotope/_c/similarity_methods.pyx"


#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 360 "ms_deisotope/_c/similarity_methods.pyx"


#line 360 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 360, 0, __PYX_ERR(0, 360, __pyx_L1_error));

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(__pyx_v_self->peak_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"


#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.peak_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 360 "ms_deisotope/_c/similarity_methods.pyx"


#line 360 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 360 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 360 "ms_deisotope/_c/similarity_methods.pyx"


#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 360 "ms_deisotope/_c/similarity_methods.pyx"


#line 360 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value) {

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType __pyx_t_1;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 360, 0, __PYX_ERR(0, 360, __pyx_L1_error));

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)__Pyx_PyInt_As_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(__pyx_v_value)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 360, __pyx_L1_error)

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->peak_type = __pyx_t_1;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"


#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.peak_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 360 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":361
 *         public object peak_set_b
 *         public PeakType peak_type
 *         public double norm_aa_cosine             # <<<<<<<<<<<<<<
 *         public double norm_bb_cosine
 *         public bint normalize
 */

#line 361 "ms_deisotope/_c/similarity_methods.pyx"


#line 361 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine_1__get__(PyObject *__pyx_v_self) {

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 361 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 361 "ms_deisotope/_c/similarity_methods.pyx"


#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 361 "ms_deisotope/_c/similarity_methods.pyx"


#line 361 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 361, 0, __PYX_ERR(0, 361, __pyx_L1_error));

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->norm_aa_cosine); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L1_error)

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"


#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.norm_aa_cosine.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 361 "ms_deisotope/_c/similarity_methods.pyx"


#line 361 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 361 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 361 "ms_deisotope/_c/similarity_methods.pyx"


#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 361 "ms_deisotope/_c/similarity_methods.pyx"


#line 361 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value) {

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 361, 0, __PYX_ERR(0, 361, __pyx_L1_error));

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 361, __pyx_L1_error)

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->norm_aa_cosine = __pyx_t_1;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"


#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.norm_aa_cosine.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 361 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":362
 *         public PeakType peak_type
 *         public double norm_aa_cosine
 *         public double norm_bb_cosine             # <<<<<<<<<<<<<<
 *         public bint normalize
 *         public bint sqrt_transform
 */

#line 362 "ms_deisotope/_c/similarity_methods.pyx"


#line 362 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine_1__get__(PyObject *__pyx_v_self) {

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 362 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 362 "ms_deisotope/_c/similarity_methods.pyx"


#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 362 "ms_deisotope/_c/similarity_methods.pyx"


#line 362 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 362, 0, __PYX_ERR(0, 362, __pyx_L1_error));

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->norm_bb_cosine); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"


#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.norm_bb_cosine.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 362 "ms_deisotope/_c/similarity_methods.pyx"


#line 362 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 362 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 362 "ms_deisotope/_c/similarity_methods.pyx"


#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 362 "ms_deisotope/_c/similarity_methods.pyx"


#line 362 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value) {

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 362, 0, __PYX_ERR(0, 362, __pyx_L1_error));

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 362, __pyx_L1_error)

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->norm_bb_cosine = __pyx_t_1;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"


#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.norm_bb_cosine.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 362 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":363
 *         public double norm_aa_cosine
 *         public double norm_bb_cosine
 *         public bint normalize             # <<<<<<<<<<<<<<
 *         public bint sqrt_transform
 * 
 */

#line 363 "ms_deisotope/_c/similarity_methods.pyx"


#line 363 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize_1__get__(PyObject *__pyx_v_self) {

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 363 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 363 "ms_deisotope/_c/similarity_methods.pyx"


#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 363 "ms_deisotope/_c/similarity_methods.pyx"


#line 363 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 363, 0, __PYX_ERR(0, 363, __pyx_L1_error));

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->normalize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"


#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.normalize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 363 "ms_deisotope/_c/similarity_methods.pyx"


#line 363 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 363 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 363 "ms_deisotope/_c/similarity_methods.pyx"


#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 363 "ms_deisotope/_c/similarity_methods.pyx"


#line 363 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value) {

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_1;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 363, 0, __PYX_ERR(0, 363, __pyx_L1_error));

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 363, __pyx_L1_error)

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->normalize = __pyx_t_1;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"


#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.normalize.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 363 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":364
 *         public double norm_bb_cosine
 *         public bint normalize
 *         public bint sqrt_transform             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, peak_set_a, peak_set_b, double error_tolerance=2e-5, double shift=0.0,
 */

#line 364 "ms_deisotope/_c/similarity_methods.pyx"


#line 364 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform_1__get__(PyObject *__pyx_v_self) {

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 364 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform___get__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 364 "ms_deisotope/_c/similarity_methods.pyx"


#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 364 "ms_deisotope/_c/similarity_methods.pyx"


#line 364 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform___get__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__get__", __pyx_f[0], 364, 0, __PYX_ERR(0, 364, __pyx_L1_error));

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->sqrt_transform); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"


#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.sqrt_transform.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 364 "ms_deisotope/_c/similarity_methods.pyx"


#line 364 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 364 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 364 "ms_deisotope/_c/similarity_methods.pyx"


#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 364 "ms_deisotope/_c/similarity_methods.pyx"


#line 364 "ms_deisotope/_c/similarity_methods.pyx"
static int __pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform_2__set__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v_value) {

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_r;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_1;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__set__", __pyx_f[0], 364, 0, __PYX_ERR(0, 364, __pyx_L1_error));

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 364, __pyx_L1_error)

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_self->sqrt_transform = __pyx_t_1;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"


#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.sqrt_transform.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = -1;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 364 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10__reduce_cython__[] = "SpectrumAlignment.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_TraceDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_9;

#line 1 "(tree fragment)"
  int __pyx_t_10;

#line 1 "(tree fragment)"
  int __pyx_t_11;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

#line 1 "(tree fragment)"
  __Pyx_TraceCall("__reduce_cython__", __pyx_f[1], 1, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.norm_aa_cosine, self.norm_bb_cosine, self.normalize, self.peak_pairs, self.peak_set_a, self.peak_set_b, self.peak_type, self.score, self.shift, self.sqrt_transform)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->norm_aa_cosine); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->norm_bb_cosine); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->normalize); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(__pyx_v_self->peak_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_5);

#line 5 "(tree fragment)"
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->shift); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_6);

#line 5 "(tree fragment)"
  __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->sqrt_transform); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_7);

#line 5 "(tree fragment)"
  __pyx_t_8 = PyTuple_New(10); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_8);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_3);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_3);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->peak_pairs);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->peak_pairs);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_v_self->peak_pairs);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->peak_set_a);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->peak_set_a);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_v_self->peak_set_a);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->peak_set_b);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->peak_set_b);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_8, 5, __pyx_v_self->peak_set_b);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_4);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_8, 6, __pyx_t_4);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_5);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_8, 7, __pyx_t_5);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_6);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_8, 8, __pyx_t_6);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_7);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_8, 9, __pyx_t_7);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

#line 5 "(tree fragment)"
  __pyx_t_4 = 0;

#line 5 "(tree fragment)"
  __pyx_t_5 = 0;

#line 5 "(tree fragment)"
  __pyx_t_6 = 0;

#line 5 "(tree fragment)"
  __pyx_t_7 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_8);

#line 5 "(tree fragment)"
  __pyx_t_8 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.norm_aa_cosine, self.norm_bb_cosine, self.normalize, self.peak_pairs, self.peak_set_a, self.peak_set_b, self.peak_type, self.score, self.shift, self.sqrt_transform)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_8 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_8);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_8;

#line 6 "(tree fragment)"
  __pyx_t_8 = 0;

  /* "(tree fragment)":7
 *     state = (self.norm_aa_cosine, self.norm_bb_cosine, self.normalize, self.peak_pairs, self.peak_set_a, self.peak_set_b, self.peak_type, self.score, self.shift, self.sqrt_transform)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_9 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_10 = (__pyx_t_9 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_7));

#line 8 "(tree fragment)"
    __pyx_t_7 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.peak_pairs is not None or self.peak_set_a is not None or self.peak_set_b is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.norm_aa_cosine, self.norm_bb_cosine, self.normalize, self.peak_pairs, self.peak_set_a, self.peak_set_b, self.peak_type, self.score, self.shift, self.sqrt_transform)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.peak_pairs is not None or self.peak_set_a is not None or self.peak_set_b is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_SpectrumAlignment, (type(self), 0x158795d, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_9 = (__pyx_v_self->peak_pairs != ((PyObject*)Py_None));

#line 11 "(tree fragment)"
    __pyx_t_11 = (__pyx_t_9 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_11) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_10 = __pyx_t_11;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_11 = (__pyx_v_self->peak_set_a != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_9 = (__pyx_t_11 != 0);

#line 11 "(tree fragment)"
    if (!__pyx_t_9) {

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      __pyx_t_10 = __pyx_t_9;

#line 11 "(tree fragment)"
      goto __pyx_L4_bool_binop_done;

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    __pyx_t_9 = (__pyx_v_self->peak_set_b != Py_None);

#line 11 "(tree fragment)"
    __pyx_t_11 = (__pyx_t_9 != 0);

#line 11 "(tree fragment)"
    __pyx_t_10 = __pyx_t_11;

#line 11 "(tree fragment)"
    __pyx_L4_bool_binop_done:;

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_10;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.peak_pairs is not None or self.peak_set_a is not None or self.peak_set_b is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SpectrumAlignment, (type(self), 0x158795d, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_10 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_10) {

    /* "(tree fragment)":13
 *         use_setstate = self.peak_pairs is not None or self.peak_set_a is not None or self.peak_set_b is not None
 *     if use_setstate:
 *         return __pyx_unpickle_SpectrumAlignment, (type(self), 0x158795d, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_SpectrumAlignment, (type(self), 0x158795d, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_SpectrumAlignment); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 13 "(tree fragment)"
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_22575453);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_22575453);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_int_22575453);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_8, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_7);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_8);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_8);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_7 = 0;

#line 13 "(tree fragment)"
    __pyx_t_8 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_6;

#line 13 "(tree fragment)"
    __pyx_t_6 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.peak_pairs is not None or self.peak_set_a is not None or self.peak_set_b is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SpectrumAlignment, (type(self), 0x158795d, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_SpectrumAlignment, (type(self), 0x158795d, None), state
 *     else:
 *         return __pyx_unpickle_SpectrumAlignment, (type(self), 0x158795d, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SpectrumAlignment__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_SpectrumAlignment); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 15 "(tree fragment)"
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_22575453);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_22575453);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_int_22575453);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_6);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_8);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8);

#line 15 "(tree fragment)"
    __pyx_t_6 = 0;

#line 15 "(tree fragment)"
    __pyx_t_8 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_7;

#line 15 "(tree fragment)"
    __pyx_t_7 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SpectrumAlignment, (type(self), 0x158795d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SpectrumAlignment__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_12__setstate_cython__[] = "SpectrumAlignment.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_12__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_12__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_TraceDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

#line 16 "(tree fragment)"
  __Pyx_TraceCall("__setstate_cython__", __pyx_f[1], 16, 0, __PYX_ERR(1, 16, __pyx_L1_error));

  /* "(tree fragment)":17
 *         return __pyx_unpickle_SpectrumAlignment, (type(self), 0x158795d, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SpectrumAlignment__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods___pyx_unpickle_SpectrumAlignment__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SpectrumAlignment, (type(self), 0x158795d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SpectrumAlignment__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.SpectrumAlignment.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/similarity_methods.pyx":480
 * 
 * 
 * cdef list convolve_peak_sets(PeakSet peak_set_a, PeakSet peak_set_b, double error_tolerance=2e-5, double shift=0.0):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n, j, m
 */

#line 480 "ms_deisotope/_c/similarity_methods.pyx"


#line 480 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets *__pyx_optional_args) {

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_error_tolerance = ((double)2e-5);

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_shift = ((double)0.0);
  size_t __pyx_v_i
#line 480 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_n
#line 480 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_j
#line 480 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_m
#line 480 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0
#line 480 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_other = 0
#line 480 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_peaks_slice = 0
#line 480 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_peak_pairs = 0
#line 480 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED PyObject *__pyx_v_pairs_for_peak = 0
#line 480 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_pt
#line 480 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_2;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_3;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_4;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for __pyx_t_5;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_6;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_7;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_8;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_9;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_10;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_11;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("convolve_peak_sets", 0);

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("convolve_peak_sets", __pyx_f[0], 480, 0, __PYX_ERR(0, 480, __pyx_L1_error));

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_shift = __pyx_optional_args->shift;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":490
 *         double pt
 * 
 *     peak_pairs = []             # <<<<<<<<<<<<<<
 *     n = peak_set_a.get_size()
 *     for i in range(n):
 */

#line 490 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)

#line 490 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 490 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_peak_pairs = ((PyObject*)__pyx_t_1);

#line 490 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":491
 * 
 *     peak_pairs = []
 *     n = peak_set_a.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         pairs_for_peak = []
 */

#line 491 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":492
 *     peak_pairs = []
 *     n = peak_set_a.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         pairs_for_peak = []
 *         peak = peak_set_a.getitem(i)
 */

#line 492 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_n;

#line 492 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 492 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 492 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/similarity_methods.pyx":493
 *     n = peak_set_a.get_size()
 *     for i in range(n):
 *         pairs_for_peak = []             # <<<<<<<<<<<<<<
 *         peak = peak_set_a.getitem(i)
 *         pt = peak.mz + shift
 */

#line 493 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L1_error)

#line 493 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 493 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_pairs_for_peak, ((PyObject*)__pyx_t_1));

#line 493 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":494
 *     for i in range(n):
 *         pairs_for_peak = []
 *         peak = peak_set_a.getitem(i)             # <<<<<<<<<<<<<<
 *         pt = peak.mz + shift
 *         peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)
 */

#line 494 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->getitem(__pyx_v_peak_set_a, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L1_error)

#line 494 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 494 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 494 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":495
 *         pairs_for_peak = []
 *         peak = peak_set_a.getitem(i)
 *         pt = peak.mz + shift             # <<<<<<<<<<<<<<
 *         peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)
 *         m = PyTuple_Size(peaks_slice)
 */

#line 495 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_pt = (__pyx_v_peak->__pyx_base.mz + __pyx_v_shift);

    /* "ms_deisotope/_c/similarity_methods.pyx":496
 *         peak = peak_set_a.getitem(i)
 *         pt = peak.mz + shift
 *         peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)             # <<<<<<<<<<<<<<
 *         m = PyTuple_Size(peaks_slice)
 *         for j in range(m):
 */

#line 496 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5.__pyx_n = 1;

#line 496 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5.error_tolerance = __pyx_v_error_tolerance;

#line 496 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->all_peaks_for(__pyx_v_peak_set_b, __pyx_v_pt, 0, &__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 496, __pyx_L1_error)

#line 496 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 496 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peaks_slice, ((PyObject*)__pyx_t_1));

#line 496 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":497
 *         pt = peak.mz + shift
 *         peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)
 *         m = PyTuple_Size(peaks_slice)             # <<<<<<<<<<<<<<
 *         for j in range(m):
 *             other = <FittedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 */

#line 497 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyTuple_Size(__pyx_v_peaks_slice); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 497, __pyx_L1_error)

#line 497 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_m = __pyx_t_6;

    /* "ms_deisotope/_c/similarity_methods.pyx":498
 *         peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)
 *         m = PyTuple_Size(peaks_slice)
 *         for j in range(m):             # <<<<<<<<<<<<<<
 *             other = <FittedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 *             peak_pairs.append(PeakMatch._create(peak, other))
 */

#line 498 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = __pyx_v_m;

#line 498 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_8 = __pyx_t_7;

#line 498 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 498 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_j = __pyx_t_9;

      /* "ms_deisotope/_c/similarity_methods.pyx":499
 *         m = PyTuple_Size(peaks_slice)
 *         for j in range(m):
 *             other = <FittedPeak>PyTuple_GET_ITEM(peaks_slice, j)             # <<<<<<<<<<<<<<
 *             peak_pairs.append(PeakMatch._create(peak, other))
 *     return peak_pairs
 */

#line 499 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_10 = PyTuple_GET_ITEM(__pyx_v_peaks_slice, __pyx_v_j);

#line 499 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_t_10);

#line 499 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 499 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_other, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 499 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":500
 *         for j in range(m):
 *             other = <FittedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 *             peak_pairs.append(PeakMatch._create(peak, other))             # <<<<<<<<<<<<<<
 *     return peak_pairs
 * 
 */

#line 500 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_18similarity_methods_PeakMatch->_create(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_peak), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_other), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)

#line 500 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 500 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_peak_pairs, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 500, __pyx_L1_error)

#line 500 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 500 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 500 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":501
 *             other = <FittedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 *             peak_pairs.append(PeakMatch._create(peak, other))
 *     return peak_pairs             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 501 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 501 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_peak_pairs);

#line 501 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_peak_pairs;

#line 501 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":480
 * 
 * 
 * cdef list convolve_peak_sets(PeakSet peak_set_a, PeakSet peak_set_b, double error_tolerance=2e-5, double shift=0.0):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n, j, m
 */

#line 480 "ms_deisotope/_c/similarity_methods.pyx"


#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.convolve_peak_sets", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_other);

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_peaks_slice);

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_peak_pairs);

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_pairs_for_peak);

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 480 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":504
 * 
 * 
 * cdef list convolve_deconvoluted_peak_sets(DeconvolutedPeakSet peak_set_a, DeconvolutedPeakSet peak_set_b,             # <<<<<<<<<<<<<<
 *                                           double error_tolerance=2e-5, double shift=0.0):
 *     cdef:
 */

#line 504 "ms_deisotope/_c/similarity_methods.pyx"


#line 504 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_deconvoluted_peak_sets(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_deconvoluted_peak_sets *__pyx_optional_args) {

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_error_tolerance = ((double)2e-5);

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_shift = ((double)0.0);
  size_t __pyx_v_i
#line 504 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_n
#line 504 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_j
#line 504 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_m
#line 504 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_peak = 0
#line 504 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_other = 0
#line 504 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_peaks_slice = 0
#line 504 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_peak_pairs = 0
#line 504 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_pt
#line 504 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_2;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_3;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_4;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for __pyx_t_5;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_6;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_7;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_8;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_9;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_10;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_11;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("convolve_deconvoluted_peak_sets", 0);

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("convolve_deconvoluted_peak_sets", __pyx_f[0], 504, 0, __PYX_ERR(0, 504, __pyx_L1_error));

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_shift = __pyx_optional_args->shift;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":515
 *         double pt
 * 
 *     peak_pairs = []             # <<<<<<<<<<<<<<
 *     n = peak_set_a.get_size()
 *     for i in range(n):
 */

#line 515 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 515, __pyx_L1_error)

#line 515 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 515 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_peak_pairs = ((PyObject*)__pyx_t_1);

#line 515 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":516
 * 
 *     peak_pairs = []
 *     n = peak_set_a.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         peak = peak_set_a.getitem(i)
 */

#line 516 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":517
 *     peak_pairs = []
 *     n = peak_set_a.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         peak = peak_set_a.getitem(i)
 *         pt = peak.neutral_mass + shift
 */

#line 517 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_v_n;

#line 517 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 517 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 517 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/similarity_methods.pyx":518
 *     n = peak_set_a.get_size()
 *     for i in range(n):
 *         peak = peak_set_a.getitem(i)             # <<<<<<<<<<<<<<
 *         pt = peak.neutral_mass + shift
 *         peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)
 */

#line 518 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->getitem(__pyx_v_peak_set_a, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L1_error)

#line 518 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 518 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_1));

#line 518 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":519
 *     for i in range(n):
 *         peak = peak_set_a.getitem(i)
 *         pt = peak.neutral_mass + shift             # <<<<<<<<<<<<<<
 *         peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)
 *         m = PyTuple_Size(peaks_slice)
 */

#line 519 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_pt = (__pyx_v_peak->neutral_mass + __pyx_v_shift);

    /* "ms_deisotope/_c/similarity_methods.pyx":520
 *         peak = peak_set_a.getitem(i)
 *         pt = peak.neutral_mass + shift
 *         peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)             # <<<<<<<<<<<<<<
 *         m = PyTuple_Size(peaks_slice)
 *         for j in range(m):
 */

#line 520 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5.__pyx_n = 1;

#line 520 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5.tolerance = __pyx_v_error_tolerance;

#line 520 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->all_peaks_for(__pyx_v_peak_set_b, __pyx_v_pt, 0, &__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L1_error)

#line 520 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 520 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peaks_slice, ((PyObject*)__pyx_t_1));

#line 520 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":521
 *         pt = peak.neutral_mass + shift
 *         peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)
 *         m = PyTuple_Size(peaks_slice)             # <<<<<<<<<<<<<<
 *         for j in range(m):
 *             other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 */

#line 521 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyTuple_Size(__pyx_v_peaks_slice); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 521, __pyx_L1_error)

#line 521 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_m = __pyx_t_6;

    /* "ms_deisotope/_c/similarity_methods.pyx":522
 *         peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)
 *         m = PyTuple_Size(peaks_slice)
 *         for j in range(m):             # <<<<<<<<<<<<<<
 *             other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 *             peak_pairs.append(PeakMatch._create(peak, other))
 */

#line 522 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = __pyx_v_m;

#line 522 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_8 = __pyx_t_7;

#line 522 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 522 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_j = __pyx_t_9;

      /* "ms_deisotope/_c/similarity_methods.pyx":523
 *         m = PyTuple_Size(peaks_slice)
 *         for j in range(m):
 *             other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)             # <<<<<<<<<<<<<<
 *             peak_pairs.append(PeakMatch._create(peak, other))
 *     return peak_pairs
 */

#line 523 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_10 = PyTuple_GET_ITEM(__pyx_v_peaks_slice, __pyx_v_j);

#line 523 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_t_10);

#line 523 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 523 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_other, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_1));

#line 523 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":524
 *         for j in range(m):
 *             other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 *             peak_pairs.append(PeakMatch._create(peak, other))             # <<<<<<<<<<<<<<
 *     return peak_pairs
 * 
 */

#line 524 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_18similarity_methods_PeakMatch->_create(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_peak), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_other), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 524, __pyx_L1_error)

#line 524 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 524 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_peak_pairs, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 524, __pyx_L1_error)

#line 524 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 524 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 524 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":525
 *             other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 *             peak_pairs.append(PeakMatch._create(peak, other))
 *     return peak_pairs             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 525 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 525 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_peak_pairs);

#line 525 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_peak_pairs;

#line 525 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":504
 * 
 * 
 * cdef list convolve_deconvoluted_peak_sets(DeconvolutedPeakSet peak_set_a, DeconvolutedPeakSet peak_set_b,             # <<<<<<<<<<<<<<
 *                                           double error_tolerance=2e-5, double shift=0.0):
 *     cdef:
 */

#line 504 "ms_deisotope/_c/similarity_methods.pyx"


#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.convolve_deconvoluted_peak_sets", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_other);

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_peaks_slice);

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_peak_pairs);

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 504 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":528
 * 
 * 
 * cpdef double convolved_dot_product(list peak_pairs, PeakType peak_type=PeakType.Fitted, double norm_a=1.0,             # <<<<<<<<<<<<<<
 *                                    double norm_b=1.0, bint sqrt_transform=False):
 *     cdef:
 */

#line 528 "ms_deisotope/_c/similarity_methods.pyx"


#line 528 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_5convolved_dot_product(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolved_dot_product(PyObject *__pyx_v_peak_pairs, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolved_dot_product *__pyx_optional_args) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType __pyx_v_peak_type = __pyx_k__16;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_norm_a = ((double)1.0);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_norm_b = ((double)1.0);

  /* "ms_deisotope/_c/similarity_methods.pyx":529
 * 
 * cpdef double convolved_dot_product(list peak_pairs, PeakType peak_type=PeakType.Fitted, double norm_a=1.0,
 *                                    double norm_b=1.0, bint sqrt_transform=False):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_sqrt_transform = ((int)0);
  size_t __pyx_v_i
#line 529 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_n
#line 529 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v_pair = 0
#line 529 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_d
#line 529 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_seen1 = 0
#line 529 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_seen2 = 0
#line 529 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_key1 = 0
#line 529 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_key2 = 0
#line 529 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_2;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_3;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_4;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_5;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_6;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_7;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_8;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_9;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_10;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_11;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_12;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("convolved_dot_product", 0);

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("convolved_dot_product", __pyx_f[0], 528, 0, __PYX_ERR(0, 528, __pyx_L1_error));

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_peak_type = __pyx_optional_args->peak_type;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_norm_a = __pyx_optional_args->norm_a;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_optional_args->__pyx_n > 2) {

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_norm_b = __pyx_optional_args->norm_b;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
          if (__pyx_optional_args->__pyx_n > 3) {

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
            __pyx_v_sqrt_transform = __pyx_optional_args->sqrt_transform;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
          }

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":538
 *         set seen1, seen2
 *         object key1, key2
 *     seen1 = set()             # <<<<<<<<<<<<<<
 *     seen2 = set()
 *     d = 0.0
 */

#line 538 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)

#line 538 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 538 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_seen1 = ((PyObject*)__pyx_t_1);

#line 538 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":539
 *         object key1, key2
 *     seen1 = set()
 *     seen2 = set()             # <<<<<<<<<<<<<<
 *     d = 0.0
 *     n = PyList_Size(peak_pairs)
 */

#line 539 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)

#line 539 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 539 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_seen2 = ((PyObject*)__pyx_t_1);

#line 539 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":540
 *     seen1 = set()
 *     seen2 = set()
 *     d = 0.0             # <<<<<<<<<<<<<<
 *     n = PyList_Size(peak_pairs)
 *     for i in range(n):
 */

#line 540 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_d = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":541
 *     seen2 = set()
 *     d = 0.0
 *     n = PyList_Size(peak_pairs)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)
 */

#line 541 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyList_Size(__pyx_v_peak_pairs); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 541, __pyx_L1_error)

#line 541 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = __pyx_t_2;

  /* "ms_deisotope/_c/similarity_methods.pyx":542
 *     d = 0.0
 *     n = PyList_Size(peak_pairs)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)
 *         if sqrt_transform:
 */

#line 542 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_v_n;

#line 542 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = __pyx_t_3;

#line 542 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {

#line 542 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_5;

    /* "ms_deisotope/_c/similarity_methods.pyx":543
 *     n = PyList_Size(peak_pairs)
 *     for i in range(n):
 *         pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)             # <<<<<<<<<<<<<<
 *         if sqrt_transform:
 *             pair.score = (sqrt(pair.peak_a.intensity) / norm_a) * (sqrt(pair.peak_b.intensity) / norm_b)
 */

#line 543 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyList_GET_ITEM(__pyx_v_peak_pairs, __pyx_v_i);

#line 543 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_6);

#line 543 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 543 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_pair, ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_t_1));

#line 543 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":544
 *     for i in range(n):
 *         pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)
 *         if sqrt_transform:             # <<<<<<<<<<<<<<
 *             pair.score = (sqrt(pair.peak_a.intensity) / norm_a) * (sqrt(pair.peak_b.intensity) / norm_b)
 *         else:
 */

#line 544 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = (__pyx_v_sqrt_transform != 0);

#line 544 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/similarity_methods.pyx":545
 *         pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)
 *         if sqrt_transform:
 *             pair.score = (sqrt(pair.peak_a.intensity) / norm_a) * (sqrt(pair.peak_b.intensity) / norm_b)             # <<<<<<<<<<<<<<
 *         else:
 *             pair.score = (pair.peak_a.intensity / norm_a) * (pair.peak_b.intensity / norm_b)
 */

#line 545 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_8 = sqrt(__pyx_v_pair->peak_a->intensity);

#line 545 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__pyx_v_norm_a == 0)) {

#line 545 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 545 "ms_deisotope/_c/similarity_methods.pyx"
        __PYX_ERR(0, 545, __pyx_L1_error)

#line 545 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 545 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_9 = sqrt(__pyx_v_pair->peak_b->intensity);

#line 545 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__pyx_v_norm_b == 0)) {

#line 545 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 545 "ms_deisotope/_c/similarity_methods.pyx"
        __PYX_ERR(0, 545, __pyx_L1_error)

#line 545 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 545 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_pair->score = ((__pyx_t_8 / __pyx_v_norm_a) * (__pyx_t_9 / __pyx_v_norm_b));

      /* "ms_deisotope/_c/similarity_methods.pyx":544
 *     for i in range(n):
 *         pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)
 *         if sqrt_transform:             # <<<<<<<<<<<<<<
 *             pair.score = (sqrt(pair.peak_a.intensity) / norm_a) * (sqrt(pair.peak_b.intensity) / norm_b)
 *         else:
 */

#line 544 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L5;

#line 544 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":547
 *             pair.score = (sqrt(pair.peak_a.intensity) / norm_a) * (sqrt(pair.peak_b.intensity) / norm_b)
 *         else:
 *             pair.score = (pair.peak_a.intensity / norm_a) * (pair.peak_b.intensity / norm_b)             # <<<<<<<<<<<<<<
 *     PyList_Sort(peak_pairs)
 * 
 */

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
    /*else*/ {

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__pyx_v_norm_a == 0)) {

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
        __PYX_ERR(0, 547, __pyx_L1_error)

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__pyx_v_norm_b == 0)) {

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
        __PYX_ERR(0, 547, __pyx_L1_error)

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_pair->score = ((__pyx_v_pair->peak_a->intensity / __pyx_v_norm_a) * (__pyx_v_pair->peak_b->intensity / __pyx_v_norm_b));

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_L5:;

#line 547 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":548
 *         else:
 *             pair.score = (pair.peak_a.intensity / norm_a) * (pair.peak_b.intensity / norm_b)
 *     PyList_Sort(peak_pairs)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(n - 1, -1, -1):
 */

#line 548 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_10 = PyList_Sort(__pyx_v_peak_pairs); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 548, __pyx_L1_error)

  /* "ms_deisotope/_c/similarity_methods.pyx":550
 *     PyList_Sort(peak_pairs)
 * 
 *     for i in range(n - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)
 *         if peak_type == PeakType.Fitted:
 */

#line 550 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_3 = (__pyx_v_n - 1) + 1; __pyx_t_3 > -1L + 1; ) { __pyx_t_3-=1;

#line 550 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/similarity_methods.pyx":551
 * 
 *     for i in range(n - 1, -1, -1):
 *         pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)             # <<<<<<<<<<<<<<
 *         if peak_type == PeakType.Fitted:
 *             key1 = (<FittedPeak>pair.peak_a).peak_count
 */

#line 551 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyList_GET_ITEM(__pyx_v_peak_pairs, __pyx_v_i);

#line 551 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_6);

#line 551 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 551 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_pair, ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_t_1));

#line 551 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":552
 *     for i in range(n - 1, -1, -1):
 *         pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)
 *         if peak_type == PeakType.Fitted:             # <<<<<<<<<<<<<<
 *             key1 = (<FittedPeak>pair.peak_a).peak_count
 *             if key1 in seen1:
 */

#line 552 "ms_deisotope/_c/similarity_methods.pyx"
    switch (__pyx_v_peak_type) {

#line 552 "ms_deisotope/_c/similarity_methods.pyx"
      case __pyx_e_12ms_deisotope_2_c_18similarity_methods_Fitted:

      /* "ms_deisotope/_c/similarity_methods.pyx":553
 *         pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)
 *         if peak_type == PeakType.Fitted:
 *             key1 = (<FittedPeak>pair.peak_a).peak_count             # <<<<<<<<<<<<<<
 *             if key1 in seen1:
 *                 continue
 */

#line 553 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __Pyx_PyInt_From_long(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_pair->peak_a)->peak_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 553, __pyx_L1_error)

#line 553 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 553 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_key1, __pyx_t_1);

#line 553 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":554
 *         if peak_type == PeakType.Fitted:
 *             key1 = (<FittedPeak>pair.peak_a).peak_count
 *             if key1 in seen1:             # <<<<<<<<<<<<<<
 *                 continue
 *             key2 = (<FittedPeak>pair.peak_b).peak_count
 */

#line 554 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_7 = (__Pyx_PySet_ContainsTF(__pyx_v_key1, __pyx_v_seen1, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 554, __pyx_L1_error)

#line 554 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_11 = (__pyx_t_7 != 0);

#line 554 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_11) {

        /* "ms_deisotope/_c/similarity_methods.pyx":555
 *             key1 = (<FittedPeak>pair.peak_a).peak_count
 *             if key1 in seen1:
 *                 continue             # <<<<<<<<<<<<<<
 *             key2 = (<FittedPeak>pair.peak_b).peak_count
 *             if key2 in seen2:
 */

#line 555 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L6_continue;

        /* "ms_deisotope/_c/similarity_methods.pyx":554
 *         if peak_type == PeakType.Fitted:
 *             key1 = (<FittedPeak>pair.peak_a).peak_count
 *             if key1 in seen1:             # <<<<<<<<<<<<<<
 *                 continue
 *             key2 = (<FittedPeak>pair.peak_b).peak_count
 */

#line 554 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":556
 *             if key1 in seen1:
 *                 continue
 *             key2 = (<FittedPeak>pair.peak_b).peak_count             # <<<<<<<<<<<<<<
 *             if key2 in seen2:
 *                 continue
 */

#line 556 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __Pyx_PyInt_From_long(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_pair->peak_b)->peak_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)

#line 556 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 556 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_key2, __pyx_t_1);

#line 556 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":557
 *                 continue
 *             key2 = (<FittedPeak>pair.peak_b).peak_count
 *             if key2 in seen2:             # <<<<<<<<<<<<<<
 *                 continue
 *         elif peak_type == PeakType.Deconvoluted:
 */

#line 557 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_11 = (__Pyx_PySet_ContainsTF(__pyx_v_key2, __pyx_v_seen2, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 557, __pyx_L1_error)

#line 557 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_7 = (__pyx_t_11 != 0);

#line 557 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_7) {

        /* "ms_deisotope/_c/similarity_methods.pyx":558
 *             key2 = (<FittedPeak>pair.peak_b).peak_count
 *             if key2 in seen2:
 *                 continue             # <<<<<<<<<<<<<<
 *         elif peak_type == PeakType.Deconvoluted:
 *             key1 = (<DeconvolutedPeak>pair.peak_a)._index.neutral_mass
 */

#line 558 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L6_continue;

        /* "ms_deisotope/_c/similarity_methods.pyx":557
 *                 continue
 *             key2 = (<FittedPeak>pair.peak_b).peak_count
 *             if key2 in seen2:             # <<<<<<<<<<<<<<
 *                 continue
 *         elif peak_type == PeakType.Deconvoluted:
 */

#line 557 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":552
 *     for i in range(n - 1, -1, -1):
 *         pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)
 *         if peak_type == PeakType.Fitted:             # <<<<<<<<<<<<<<
 *             key1 = (<FittedPeak>pair.peak_a).peak_count
 *             if key1 in seen1:
 */

#line 552 "ms_deisotope/_c/similarity_methods.pyx"
      break;

#line 552 "ms_deisotope/_c/similarity_methods.pyx"
      case __pyx_e_12ms_deisotope_2_c_18similarity_methods_Deconvoluted:

      /* "ms_deisotope/_c/similarity_methods.pyx":560
 *                 continue
 *         elif peak_type == PeakType.Deconvoluted:
 *             key1 = (<DeconvolutedPeak>pair.peak_a)._index.neutral_mass             # <<<<<<<<<<<<<<
 *             if key1 in seen1:
 *                 continue
 */

#line 560 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __Pyx_PyInt_FromSize_t(((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_v_pair->peak_a)->_index->neutral_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 560, __pyx_L1_error)

#line 560 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 560 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_key1, __pyx_t_1);

#line 560 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":561
 *         elif peak_type == PeakType.Deconvoluted:
 *             key1 = (<DeconvolutedPeak>pair.peak_a)._index.neutral_mass
 *             if key1 in seen1:             # <<<<<<<<<<<<<<
 *                 continue
 *             key2 = (<DeconvolutedPeak>pair.peak_b)._index.neutral_mass
 */

#line 561 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_7 = (__Pyx_PySet_ContainsTF(__pyx_v_key1, __pyx_v_seen1, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 561, __pyx_L1_error)

#line 561 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_11 = (__pyx_t_7 != 0);

#line 561 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_11) {

        /* "ms_deisotope/_c/similarity_methods.pyx":562
 *             key1 = (<DeconvolutedPeak>pair.peak_a)._index.neutral_mass
 *             if key1 in seen1:
 *                 continue             # <<<<<<<<<<<<<<
 *             key2 = (<DeconvolutedPeak>pair.peak_b)._index.neutral_mass
 *             if key2 in seen2:
 */

#line 562 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L6_continue;

        /* "ms_deisotope/_c/similarity_methods.pyx":561
 *         elif peak_type == PeakType.Deconvoluted:
 *             key1 = (<DeconvolutedPeak>pair.peak_a)._index.neutral_mass
 *             if key1 in seen1:             # <<<<<<<<<<<<<<
 *                 continue
 *             key2 = (<DeconvolutedPeak>pair.peak_b)._index.neutral_mass
 */

#line 561 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":563
 *             if key1 in seen1:
 *                 continue
 *             key2 = (<DeconvolutedPeak>pair.peak_b)._index.neutral_mass             # <<<<<<<<<<<<<<
 *             if key2 in seen2:
 *                 continue
 */

#line 563 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = __Pyx_PyInt_FromSize_t(((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_v_pair->peak_b)->_index->neutral_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)

#line 563 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 563 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_key2, __pyx_t_1);

#line 563 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":564
 *                 continue
 *             key2 = (<DeconvolutedPeak>pair.peak_b)._index.neutral_mass
 *             if key2 in seen2:             # <<<<<<<<<<<<<<
 *                 continue
 *         seen1.add(key1)
 */

#line 564 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_11 = (__Pyx_PySet_ContainsTF(__pyx_v_key2, __pyx_v_seen2, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 564, __pyx_L1_error)

#line 564 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_7 = (__pyx_t_11 != 0);

#line 564 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_7) {

        /* "ms_deisotope/_c/similarity_methods.pyx":565
 *             key2 = (<DeconvolutedPeak>pair.peak_b)._index.neutral_mass
 *             if key2 in seen2:
 *                 continue             # <<<<<<<<<<<<<<
 *         seen1.add(key1)
 *         seen2.add(key2)
 */

#line 565 "ms_deisotope/_c/similarity_methods.pyx"
        goto __pyx_L6_continue;

        /* "ms_deisotope/_c/similarity_methods.pyx":564
 *                 continue
 *             key2 = (<DeconvolutedPeak>pair.peak_b)._index.neutral_mass
 *             if key2 in seen2:             # <<<<<<<<<<<<<<
 *                 continue
 *         seen1.add(key1)
 */

#line 564 "ms_deisotope/_c/similarity_methods.pyx"
      }

      /* "ms_deisotope/_c/similarity_methods.pyx":559
 *             if key2 in seen2:
 *                 continue
 *         elif peak_type == PeakType.Deconvoluted:             # <<<<<<<<<<<<<<
 *             key1 = (<DeconvolutedPeak>pair.peak_a)._index.neutral_mass
 *             if key1 in seen1:
 */

#line 559 "ms_deisotope/_c/similarity_methods.pyx"
      break;

#line 559 "ms_deisotope/_c/similarity_methods.pyx"
      default: break;

#line 559 "ms_deisotope/_c/similarity_methods.pyx"
    }

    /* "ms_deisotope/_c/similarity_methods.pyx":566
 *             if key2 in seen2:
 *                 continue
 *         seen1.add(key1)             # <<<<<<<<<<<<<<
 *         seen2.add(key2)
 *         d += pair.score
 */

#line 566 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__pyx_v_key1)) { __Pyx_RaiseUnboundLocalError("key1"); __PYX_ERR(0, 566, __pyx_L1_error) }

#line 566 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_12 = PySet_Add(__pyx_v_seen1, __pyx_v_key1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 566, __pyx_L1_error)

    /* "ms_deisotope/_c/similarity_methods.pyx":567
 *                 continue
 *         seen1.add(key1)
 *         seen2.add(key2)             # <<<<<<<<<<<<<<
 *         d += pair.score
 *     return d
 */

#line 567 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__pyx_v_key2)) { __Pyx_RaiseUnboundLocalError("key2"); __PYX_ERR(0, 567, __pyx_L1_error) }

#line 567 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_12 = PySet_Add(__pyx_v_seen2, __pyx_v_key2); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 567, __pyx_L1_error)

    /* "ms_deisotope/_c/similarity_methods.pyx":568
 *         seen1.add(key1)
 *         seen2.add(key2)
 *         d += pair.score             # <<<<<<<<<<<<<<
 *     return d
 * 
 */

#line 568 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_d = (__pyx_v_d + __pyx_v_pair->score);

#line 568 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_L6_continue:;

#line 568 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":569
 *         seen2.add(key2)
 *         d += pair.score
 *     return d             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 569 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_d;

#line 569 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":528
 * 
 * 
 * cpdef double convolved_dot_product(list peak_pairs, PeakType peak_type=PeakType.Fitted, double norm_a=1.0,             # <<<<<<<<<<<<<<
 *                                    double norm_b=1.0, bint sqrt_transform=False):
 *     cdef:
 */

#line 528 "ms_deisotope/_c/similarity_methods.pyx"


#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods.convolved_dot_product", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_pair);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_seen1);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_seen2);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_key1);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_key2);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 528 "ms_deisotope/_c/similarity_methods.pyx"


#line 528 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_5convolved_dot_product(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_4convolved_dot_product[] = "convolved_dot_product(list peak_pairs, PeakType peak_type=PeakType.Fitted, double norm_a=1.0, double norm_b=1.0, bool sqrt_transform=False) -> double";

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_5convolved_dot_product(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peak_pairs = 0
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
;
  enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType __pyx_v_peak_type
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_norm_a
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_norm_b
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_sqrt_transform
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("convolved_dot_product (wrapper)", 0);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_pairs,&__pyx_n_s_peak_type,&__pyx_n_s_norm_a,&__pyx_n_s_norm_b,&__pyx_n_s_sqrt_transform,0};

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[5] = {0,0,0,0,0};

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  5: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_pairs)) != 0)) kw_args--;
        else 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_type);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_norm_a);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        case  3:

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_norm_b);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[3] = value; kw_args--; }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        case  4:

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sqrt_transform);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "convolved_dot_product") < 0)) __PYX_ERR(0, 528, __pyx_L3_error)

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 528 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_pairs = ((PyObject*)values[0]);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[1]) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_peak_type = ((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)__Pyx_PyInt_As_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(values[1])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L3_error)

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_peak_type = __pyx_k__16;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_norm_a = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_norm_a == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L3_error)

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_norm_a = ((double)1.0);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[3]) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_norm_b = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_norm_b == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 529, __pyx_L3_error)

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_norm_b = ((double)1.0);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[4]) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_sqrt_transform = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_sqrt_transform == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 529, __pyx_L3_error)

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

      /* "ms_deisotope/_c/similarity_methods.pyx":529
 * 
 * cpdef double convolved_dot_product(list peak_pairs, PeakType peak_type=PeakType.Fitted, double norm_a=1.0,
 *                                    double norm_b=1.0, bint sqrt_transform=False):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_sqrt_transform = ((int)0);

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convolved_dot_product", 0, 1, 5, PyTuple_GET_SIZE(__pyx_args)); 
#line 529 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 528, __pyx_L3_error)

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.convolved_dot_product", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 529 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_pairs), (&PyList_Type), 1, "peak_pairs", 1))) __PYX_ERR(0, 528, __pyx_L1_error)
  __pyx_r = 
#line 529 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_4convolved_dot_product(__pyx_self, __pyx_v_peak_pairs, __pyx_v_peak_type, __pyx_v_norm_a, __pyx_v_norm_b, __pyx_v_sqrt_transform);

  /* "ms_deisotope/_c/similarity_methods.pyx":528
 * 
 * 
 * cpdef double convolved_dot_product(list peak_pairs, PeakType peak_type=PeakType.Fitted, double norm_a=1.0,             # <<<<<<<<<<<<<<
 *                                    double norm_b=1.0, bint sqrt_transform=False):
 *     cdef:
 */

#line 528 "ms_deisotope/_c/similarity_methods.pyx"


#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 528 "ms_deisotope/_c/similarity_methods.pyx"


#line 528 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_4convolved_dot_product(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak_pairs, enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType __pyx_v_peak_type, double __pyx_v_norm_a, double __pyx_v_norm_b, int __pyx_v_sqrt_transform) {

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolved_dot_product __pyx_t_2;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("convolved_dot_product", 0);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("convolved_dot_product (wrapper)", __pyx_f[0], 528, 0, __PYX_ERR(0, 528, __pyx_L1_error));

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 4;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.peak_type = __pyx_v_peak_type;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.norm_a = __pyx_v_norm_a;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.norm_b = __pyx_v_norm_b;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.sqrt_transform = __pyx_v_sqrt_transform;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolved_dot_product(__pyx_v_peak_pairs, 0, &__pyx_t_2); 

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 528, __pyx_L1_error)

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_3;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"


#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.convolved_dot_product", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":572
 * 
 * 
 * cdef double convolve_peak_sets2_fitted(PeakSet peak_set_a, PeakSet peak_set_b, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n, j, m
 */

#line 572 "ms_deisotope/_c/similarity_methods.pyx"


#line 572 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted *__pyx_optional_args) {

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_error_tolerance = ((double)2e-5);
  size_t __pyx_v_i
#line 572 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_n
#line 572 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_j
#line 572 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_m
#line 572 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0
#line 572 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_other = 0
#line 572 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks_slice = 0
#line 572 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_acc
#line 572 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_1;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_2;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_3;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_5;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_6;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_7;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("convolve_peak_sets2_fitted", 0);

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("convolve_peak_sets2_fitted", __pyx_f[0], 572, 0, __PYX_ERR(0, 572, __pyx_L1_error));

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":580
 *         double acc
 * 
 *     acc = 0.0             # <<<<<<<<<<<<<<
 *     n = peak_set_a.get_size()
 *     for i in range(n):
 */

#line 580 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_acc = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":581
 * 
 *     acc = 0.0
 *     n = peak_set_a.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         peak = peak_set_a.getitem(i)
 */

#line 581 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":582
 *     acc = 0.0
 *     n = peak_set_a.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         peak = peak_set_a.getitem(i)
 *         peaks_slice = peak_set_b._between(
 */

#line 582 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 582 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 582 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 582 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/similarity_methods.pyx":583
 *     n = peak_set_a.get_size()
 *     for i in range(n):
 *         peak = peak_set_a.getitem(i)             # <<<<<<<<<<<<<<
 *         peaks_slice = peak_set_b._between(
 *             peak.mz - peak.mz * error_tolerance, peak.mz + peak.mz * error_tolerance)
 */

#line 583 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->getitem(__pyx_v_peak_set_a, __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 583, __pyx_L1_error)

#line 583 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 583 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_4));

#line 583 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":584
 *     for i in range(n):
 *         peak = peak_set_a.getitem(i)
 *         peaks_slice = peak_set_b._between(             # <<<<<<<<<<<<<<
 *             peak.mz - peak.mz * error_tolerance, peak.mz + peak.mz * error_tolerance)
 *         m = peaks_slice.get_size()
 */

#line 584 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->_between(__pyx_v_peak_set_b, (__pyx_v_peak->__pyx_base.mz - (__pyx_v_peak->__pyx_base.mz * __pyx_v_error_tolerance)), (__pyx_v_peak->__pyx_base.mz + (__pyx_v_peak->__pyx_base.mz * __pyx_v_error_tolerance)))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 584, __pyx_L1_error)

#line 584 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 584 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peaks_slice, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_4));

#line 584 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":586
 *         peaks_slice = peak_set_b._between(
 *             peak.mz - peak.mz * error_tolerance, peak.mz + peak.mz * error_tolerance)
 *         m = peaks_slice.get_size()             # <<<<<<<<<<<<<<
 *         for j in range(m):
 *             other = peaks_slice.getitem(j)
 */

#line 586 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_m = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks_slice->__pyx_vtab)->get_size(__pyx_v_peaks_slice);

    /* "ms_deisotope/_c/similarity_methods.pyx":587
 *             peak.mz - peak.mz * error_tolerance, peak.mz + peak.mz * error_tolerance)
 *         m = peaks_slice.get_size()
 *         for j in range(m):             # <<<<<<<<<<<<<<
 *             other = peaks_slice.getitem(j)
 *             acc += peak.intensity * other.intensity
 */

#line 587 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = __pyx_v_m;

#line 587 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __pyx_t_5;

#line 587 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {

#line 587 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_j = __pyx_t_7;

      /* "ms_deisotope/_c/similarity_methods.pyx":588
 *         m = peaks_slice.get_size()
 *         for j in range(m):
 *             other = peaks_slice.getitem(j)             # <<<<<<<<<<<<<<
 *             acc += peak.intensity * other.intensity
 *     return acc
 */

#line 588 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks_slice->__pyx_vtab)->getitem(__pyx_v_peaks_slice, __pyx_v_j)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 588, __pyx_L1_error)

#line 588 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 588 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_other, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_4));

#line 588 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_4 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":589
 *         for j in range(m):
 *             other = peaks_slice.getitem(j)
 *             acc += peak.intensity * other.intensity             # <<<<<<<<<<<<<<
 *     return acc
 * 
 */

#line 589 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_acc = (__pyx_v_acc + (__pyx_v_peak->__pyx_base.intensity * __pyx_v_other->__pyx_base.intensity));

#line 589 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 589 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":590
 *             other = peaks_slice.getitem(j)
 *             acc += peak.intensity * other.intensity
 *     return acc             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 590 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_acc;

#line 590 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":572
 * 
 * 
 * cdef double convolve_peak_sets2_fitted(PeakSet peak_set_a, PeakSet peak_set_b, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n, j, m
 */

#line 572 "ms_deisotope/_c/similarity_methods.pyx"


#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods.convolve_peak_sets2_fitted", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_other);

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peaks_slice);

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 572 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":593
 * 
 * 
 * cdef double convolve_peak_sets2_deconvoluted(DeconvolutedPeakSet peak_set_a, DeconvolutedPeakSet peak_set_b,             # <<<<<<<<<<<<<<
 *                                              double error_tolerance=2e-5):
 *     cdef:
 */

#line 593 "ms_deisotope/_c/similarity_methods.pyx"


#line 593 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_deconvoluted(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_deconvoluted *__pyx_optional_args) {

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_error_tolerance = ((double)2e-5);
  size_t __pyx_v_i
#line 593 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_n
#line 593 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_j
#line 593 "ms_deisotope/_c/similarity_methods.pyx"
;
  size_t __pyx_v_m
#line 593 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_peak = 0
#line 593 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_other = 0
#line 593 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_peaks_slice = 0
#line 593 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_acc
#line 593 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_1;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_2;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_3;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for __pyx_t_5;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_6;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_7;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_8;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  size_t __pyx_t_9;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_10;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("convolve_peak_sets2_deconvoluted", 0);

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("convolve_peak_sets2_deconvoluted", __pyx_f[0], 593, 0, __PYX_ERR(0, 593, __pyx_L1_error));

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":602
 *         double acc
 * 
 *     acc = 0.0             # <<<<<<<<<<<<<<
 *     n = peak_set_a.get_size()
 *     for i in range(n):
 */

#line 602 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_acc = 0.0;

  /* "ms_deisotope/_c/similarity_methods.pyx":603
 * 
 *     acc = 0.0
 *     n = peak_set_a.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         peak = peak_set_a.getitem(i)
 */

#line 603 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->get_size(__pyx_v_peak_set_a);

  /* "ms_deisotope/_c/similarity_methods.pyx":604
 *     acc = 0.0
 *     n = peak_set_a.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         peak = peak_set_a.getitem(i)
 *         peaks_slice = peak_set_b.all_peaks_for(peak.neutral_mass, error_tolerance)
 */

#line 604 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 604 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 604 "ms_deisotope/_c/similarity_methods.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 604 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/similarity_methods.pyx":605
 *     n = peak_set_a.get_size()
 *     for i in range(n):
 *         peak = peak_set_a.getitem(i)             # <<<<<<<<<<<<<<
 *         peaks_slice = peak_set_b.all_peaks_for(peak.neutral_mass, error_tolerance)
 *         m = PyTuple_Size(peaks_slice)
 */

#line 605 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_a->__pyx_vtab)->getitem(__pyx_v_peak_set_a, __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L1_error)

#line 605 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 605 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_4));

#line 605 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":606
 *     for i in range(n):
 *         peak = peak_set_a.getitem(i)
 *         peaks_slice = peak_set_b.all_peaks_for(peak.neutral_mass, error_tolerance)             # <<<<<<<<<<<<<<
 *         m = PyTuple_Size(peaks_slice)
 *         for j in range(m):
 */

#line 606 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5.__pyx_n = 1;

#line 606 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5.tolerance = __pyx_v_error_tolerance;

#line 606 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set_b->__pyx_vtab)->all_peaks_for(__pyx_v_peak_set_b, __pyx_v_peak->neutral_mass, 0, &__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 606, __pyx_L1_error)

#line 606 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 606 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peaks_slice, ((PyObject*)__pyx_t_4));

#line 606 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/similarity_methods.pyx":607
 *         peak = peak_set_a.getitem(i)
 *         peaks_slice = peak_set_b.all_peaks_for(peak.neutral_mass, error_tolerance)
 *         m = PyTuple_Size(peaks_slice)             # <<<<<<<<<<<<<<
 *         for j in range(m):
 *             other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 */

#line 607 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyTuple_Size(__pyx_v_peaks_slice); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 607, __pyx_L1_error)

#line 607 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_m = __pyx_t_6;

    /* "ms_deisotope/_c/similarity_methods.pyx":608
 *         peaks_slice = peak_set_b.all_peaks_for(peak.neutral_mass, error_tolerance)
 *         m = PyTuple_Size(peaks_slice)
 *         for j in range(m):             # <<<<<<<<<<<<<<
 *             other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 *             acc += peak.intensity * other.intensity
 */

#line 608 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_7 = __pyx_v_m;

#line 608 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_8 = __pyx_t_7;

#line 608 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {

#line 608 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_j = __pyx_t_9;

      /* "ms_deisotope/_c/similarity_methods.pyx":609
 *         m = PyTuple_Size(peaks_slice)
 *         for j in range(m):
 *             other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)             # <<<<<<<<<<<<<<
 *             acc += peak.intensity * other.intensity
 *     return acc
 */

#line 609 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_10 = PyTuple_GET_ITEM(__pyx_v_peaks_slice, __pyx_v_j);

#line 609 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_4 = ((PyObject *)__pyx_t_10);

#line 609 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 609 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_other, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_4));

#line 609 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_4 = 0;

      /* "ms_deisotope/_c/similarity_methods.pyx":610
 *         for j in range(m):
 *             other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 *             acc += peak.intensity * other.intensity             # <<<<<<<<<<<<<<
 *     return acc
 * 
 */

#line 610 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_acc = (__pyx_v_acc + (__pyx_v_peak->__pyx_base.intensity * __pyx_v_other->__pyx_base.intensity));

#line 610 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 610 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":611
 *             other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)
 *             acc += peak.intensity * other.intensity
 *     return acc             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 611 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_v_acc;

#line 611 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":593
 * 
 * 
 * cdef double convolve_peak_sets2_deconvoluted(DeconvolutedPeakSet peak_set_a, DeconvolutedPeakSet peak_set_b,             # <<<<<<<<<<<<<<
 *                                              double error_tolerance=2e-5):
 *     cdef:
 */

#line 593 "ms_deisotope/_c/similarity_methods.pyx"


#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods.convolve_peak_sets2_deconvoluted", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_other);

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_peaks_slice);

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 593 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":614
 * 
 * 
 * cpdef double ppm_peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double ab, aa, bb
 */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_7ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_6ppm_peak_set_similarity[] = "ppm_peak_set_similarity(signatures, args, kwargs, defaults)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_18similarity_methods_7ppm_peak_set_similarity = 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
{"ppm_peak_set_similarity", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_7ppm_peak_set_similarity, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_6ppm_peak_set_similarity};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_7ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_signatures = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_args = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signatures,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[4] = {0,0,0,0};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    values[2] = __pyx_k__17;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  4: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signatures)) != 0)) kw_args--;
        else 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 1); 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  3:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 3); 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_signatures = values[0];

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_args = values[1];

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_kwargs = values[2];

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_defaults = values[3];

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_6ppm_peak_set_similarity(__pyx_self, __pyx_v_signatures, __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_6ppm_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  Py_ssize_t __pyx_v_i
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_arg = NULL
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_candidates = NULL
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_sig = NULL
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_match_found
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_src_sig = NULL
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_dst_type = NULL
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_2;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_3;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_4;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_5;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_7;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_8;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_9;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_13;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_14;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_15;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_16;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("ppm_peak_set_similarity", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("ppm_peak_set_similarity", __pyx_f[0], 614, 0, __PYX_ERR(0, 614, __pyx_L1_error));

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_kwargs);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_4) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L4_bool_binop_done;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((!__pyx_t_4) != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_bool_binop_done:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_itemsize = -1L;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_args == Py_None)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = ((0 < __pyx_t_5) != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(((PyObject*)__pyx_v_args), 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L6;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_4) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L7_bool_binop_done;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_peak_set_a, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_t_4 != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L7_bool_binop_done:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_peak_set_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L6;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /*else*/ {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_int_2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_int_2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_n_s_s);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_n_s_s);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L6:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  while (1) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s__2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_peak_picker__c_peak_set);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_3) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s__2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_ms_peak_picker__c_peak_index);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PeakIndex); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = (__pyx_t_3 != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_2) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakIndex, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s__2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_deisotope__c_peak_set);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_DeconvolutedPeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_3) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_DeconvolutedPeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_object, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L10_break;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L10_break:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_candidates = ((PyObject*)__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_signatures == Py_None)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_t_6;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  while (1) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_7, &__pyx_t_5, &__pyx_t_6, NULL, NULL, __pyx_t_8);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 == 0)) break;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_match_found = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_12 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(__pyx_t_12)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_12);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(function);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__3) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_10);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(__pyx_t_10)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_10);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(function);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__4) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__4);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_14 = __pyx_t_13;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_i = __pyx_t_15;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_2 = (__pyx_t_3 != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_2) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_6, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_match_found = 1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L19;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        /*else*/ {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_match_found = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L17_break;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_L19:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_L17_break:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = (__pyx_v_match_found != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_2) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((!__pyx_t_2) != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_3) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((__pyx_t_7 > 1) != 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_3) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /*else*/ {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_signatures == Py_None)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_candidates, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = __pyx_t_11;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_11);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_12);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_dest_sig);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_ndarray);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_arg);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_candidates);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_sig);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_src_sig);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_dst_type);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_35__pyx_fuse_0ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_7ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity *__pyx_optional_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_error_tolerance = __pyx_k__18;
  double __pyx_v_ab
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_aa
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_bb
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted __pyx_t_2;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0ppm_peak_set_similarity", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_0ppm_peak_set_similarity", __pyx_f[0], 614, 0, __PYX_ERR(0, 614, __pyx_L1_error));

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":618
 *         double ab, aa, bb
 *     if peak_collection is PeakSet:
 *         ab = convolve_peak_sets2_fitted(peak_set_a, peak_set_b, error_tolerance)             # <<<<<<<<<<<<<<
 *         aa = convolve_peak_sets2_fitted(peak_set_a, peak_set_a, error_tolerance)
 *         bb = convolve_peak_sets2_fitted(peak_set_b, peak_set_b, error_tolerance)
 */

#line 618 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 618 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 618 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted(__pyx_v_peak_set_a, __pyx_v_peak_set_b, &__pyx_t_2); 

#line 618 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_ab = __pyx_t_1;

  /* "ms_deisotope/_c/similarity_methods.pyx":619
 *     if peak_collection is PeakSet:
 *         ab = convolve_peak_sets2_fitted(peak_set_a, peak_set_b, error_tolerance)
 *         aa = convolve_peak_sets2_fitted(peak_set_a, peak_set_a, error_tolerance)             # <<<<<<<<<<<<<<
 *         bb = convolve_peak_sets2_fitted(peak_set_b, peak_set_b, error_tolerance)
 *     elif peak_collection is PeakIndex:
 */

#line 619 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 619 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 619 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted(__pyx_v_peak_set_a, __pyx_v_peak_set_a, &__pyx_t_2); 

#line 619 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_aa = __pyx_t_1;

  /* "ms_deisotope/_c/similarity_methods.pyx":620
 *         ab = convolve_peak_sets2_fitted(peak_set_a, peak_set_b, error_tolerance)
 *         aa = convolve_peak_sets2_fitted(peak_set_a, peak_set_a, error_tolerance)
 *         bb = convolve_peak_sets2_fitted(peak_set_b, peak_set_b, error_tolerance)             # <<<<<<<<<<<<<<
 *     elif peak_collection is PeakIndex:
 *         ab = convolve_peak_sets2_fitted(peak_set_a.peaks, peak_set_b.peaks, error_tolerance)
 */

#line 620 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 620 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 620 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted(__pyx_v_peak_set_b, __pyx_v_peak_set_b, &__pyx_t_2); 

#line 620 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bb = __pyx_t_1;

  /* "ms_deisotope/_c/similarity_methods.pyx":632
 *         raise TypeError("Cannot handle objects of type %s" % type(peak_set_a))
 * 
 *     return (ab) / (sqrt((aa)) * sqrt((bb)))             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = (sqrt(__pyx_v_aa) * sqrt(__pyx_v_bb));

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_t_1 == 0)) {

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 632, __pyx_L1_error)

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = (__pyx_v_ab / __pyx_t_1);

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":614
 * 
 * 
 * cpdef double ppm_peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double ab, aa, bb
 */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods.ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_35__pyx_fuse_0ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_35__pyx_fuse_0ppm_peak_set_similarity = 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_0ppm_peak_set_similarity", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_35__pyx_fuse_0ppm_peak_set_similarity, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_6ppm_peak_set_similarity};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_35__pyx_fuse_0ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_error_tolerance
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0ppm_peak_set_similarity (wrapper)", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_error_tolerance,0};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0ppm_peak_set_similarity", 0, 2, 3, 1); 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0ppm_peak_set_similarity") < 0)) __PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[1]);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_k__18;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0ppm_peak_set_similarity", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_0ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peak_set_a", 0))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peak_set_b", 0))) __PYX_ERR(0, 614, __pyx_L1_error)
  __pyx_r = 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_34__pyx_fuse_0ppm_peak_set_similarity(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_error_tolerance);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_34__pyx_fuse_0ppm_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, double __pyx_v_error_tolerance) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity __pyx_t_2;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0ppm_peak_set_similarity", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_0ppm_peak_set_similarity (wrapper)", __pyx_f[0], 614, 0, __PYX_ERR(0, 614, __pyx_L1_error));

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); 

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_3;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_0ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_37__pyx_fuse_1ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_7ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity *__pyx_optional_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_error_tolerance = __pyx_k__19;
  double __pyx_v_ab
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_aa
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_bb
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_3;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted __pyx_t_4;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1ppm_peak_set_similarity", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_1ppm_peak_set_similarity", __pyx_f[0], 614, 0, __PYX_ERR(0, 614, __pyx_L1_error));

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":622
 *         bb = convolve_peak_sets2_fitted(peak_set_b, peak_set_b, error_tolerance)
 *     elif peak_collection is PeakIndex:
 *         ab = convolve_peak_sets2_fitted(peak_set_a.peaks, peak_set_b.peaks, error_tolerance)             # <<<<<<<<<<<<<<
 *         aa = convolve_peak_sets2_fitted(peak_set_a.peaks, peak_set_a.peaks, error_tolerance)
 *         bb = convolve_peak_sets2_fitted(peak_set_b.peaks, peak_set_b.peaks, error_tolerance)
 */

#line 622 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_v_peak_set_a->peaks);

#line 622 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 622 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = ((PyObject *)__pyx_v_peak_set_b->peaks);

#line 622 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_2);

#line 622 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4.__pyx_n = 1;

#line 622 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4.error_tolerance = __pyx_v_error_tolerance;

#line 622 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2), &__pyx_t_4); 

#line 622 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 622 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 622 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_ab = __pyx_t_3;

  /* "ms_deisotope/_c/similarity_methods.pyx":623
 *     elif peak_collection is PeakIndex:
 *         ab = convolve_peak_sets2_fitted(peak_set_a.peaks, peak_set_b.peaks, error_tolerance)
 *         aa = convolve_peak_sets2_fitted(peak_set_a.peaks, peak_set_a.peaks, error_tolerance)             # <<<<<<<<<<<<<<
 *         bb = convolve_peak_sets2_fitted(peak_set_b.peaks, peak_set_b.peaks, error_tolerance)
 *     elif peak_collection is DeconvolutedPeakSet:
 */

#line 623 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = ((PyObject *)__pyx_v_peak_set_a->peaks);

#line 623 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_2);

#line 623 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_v_peak_set_a->peaks);

#line 623 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 623 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4.__pyx_n = 1;

#line 623 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4.error_tolerance = __pyx_v_error_tolerance;

#line 623 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1), &__pyx_t_4); 

#line 623 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 623 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 623 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_aa = __pyx_t_3;

  /* "ms_deisotope/_c/similarity_methods.pyx":624
 *         ab = convolve_peak_sets2_fitted(peak_set_a.peaks, peak_set_b.peaks, error_tolerance)
 *         aa = convolve_peak_sets2_fitted(peak_set_a.peaks, peak_set_a.peaks, error_tolerance)
 *         bb = convolve_peak_sets2_fitted(peak_set_b.peaks, peak_set_b.peaks, error_tolerance)             # <<<<<<<<<<<<<<
 *     elif peak_collection is DeconvolutedPeakSet:
 *         ab = convolve_peak_sets2_deconvoluted(peak_set_a, peak_set_b, error_tolerance)
 */

#line 624 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_v_peak_set_b->peaks);

#line 624 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 624 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = ((PyObject *)__pyx_v_peak_set_b->peaks);

#line 624 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_t_2);

#line 624 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4.__pyx_n = 1;

#line 624 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4.error_tolerance = __pyx_v_error_tolerance;

#line 624 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_fitted(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2), &__pyx_t_4); 

#line 624 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 624 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 624 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bb = __pyx_t_3;

  /* "ms_deisotope/_c/similarity_methods.pyx":632
 *         raise TypeError("Cannot handle objects of type %s" % type(peak_set_a))
 * 
 *     return (ab) / (sqrt((aa)) * sqrt((bb)))             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (sqrt(__pyx_v_aa) * sqrt(__pyx_v_bb));

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_t_3 == 0)) {

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 632, __pyx_L1_error)

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = (__pyx_v_ab / __pyx_t_3);

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":614
 * 
 * 
 * cpdef double ppm_peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double ab, aa, bb
 */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods.ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_37__pyx_fuse_1ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_37__pyx_fuse_1ppm_peak_set_similarity = 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_1ppm_peak_set_similarity", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_37__pyx_fuse_1ppm_peak_set_similarity, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_6ppm_peak_set_similarity};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_37__pyx_fuse_1ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_error_tolerance
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1ppm_peak_set_similarity (wrapper)", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_error_tolerance,0};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1ppm_peak_set_similarity", 0, 2, 3, 1); 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1ppm_peak_set_similarity") < 0)) __PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[0]);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[1]);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_k__19;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1ppm_peak_set_similarity", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_1ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peak_set_a", 0))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peak_set_b", 0))) __PYX_ERR(0, 614, __pyx_L1_error)
  __pyx_r = 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_36__pyx_fuse_1ppm_peak_set_similarity(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_error_tolerance);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_36__pyx_fuse_1ppm_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, double __pyx_v_error_tolerance) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity __pyx_t_2;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1ppm_peak_set_similarity", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_1ppm_peak_set_similarity (wrapper)", __pyx_f[0], 614, 0, __PYX_ERR(0, 614, __pyx_L1_error));

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); 

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_3;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_1ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_39__pyx_fuse_2ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_7ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity *__pyx_optional_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_error_tolerance = __pyx_k__20;
  double __pyx_v_ab
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_aa
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_bb
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_deconvoluted __pyx_t_2;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2ppm_peak_set_similarity", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_2ppm_peak_set_similarity", __pyx_f[0], 614, 0, __PYX_ERR(0, 614, __pyx_L1_error));

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":626
 *         bb = convolve_peak_sets2_fitted(peak_set_b.peaks, peak_set_b.peaks, error_tolerance)
 *     elif peak_collection is DeconvolutedPeakSet:
 *         ab = convolve_peak_sets2_deconvoluted(peak_set_a, peak_set_b, error_tolerance)             # <<<<<<<<<<<<<<
 *         aa = convolve_peak_sets2_deconvoluted(peak_set_a, peak_set_a, error_tolerance)
 *         bb = convolve_peak_sets2_deconvoluted(peak_set_b, peak_set_b, error_tolerance)
 */

#line 626 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 626 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 626 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_deconvoluted(__pyx_v_peak_set_a, __pyx_v_peak_set_b, &__pyx_t_2); 

#line 626 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_ab = __pyx_t_1;

  /* "ms_deisotope/_c/similarity_methods.pyx":627
 *     elif peak_collection is DeconvolutedPeakSet:
 *         ab = convolve_peak_sets2_deconvoluted(peak_set_a, peak_set_b, error_tolerance)
 *         aa = convolve_peak_sets2_deconvoluted(peak_set_a, peak_set_a, error_tolerance)             # <<<<<<<<<<<<<<
 *         bb = convolve_peak_sets2_deconvoluted(peak_set_b, peak_set_b, error_tolerance)
 *     elif peak_collection is object:
 */

#line 627 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 627 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 627 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_deconvoluted(__pyx_v_peak_set_a, __pyx_v_peak_set_a, &__pyx_t_2); 

#line 627 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_aa = __pyx_t_1;

  /* "ms_deisotope/_c/similarity_methods.pyx":628
 *         ab = convolve_peak_sets2_deconvoluted(peak_set_a, peak_set_b, error_tolerance)
 *         aa = convolve_peak_sets2_deconvoluted(peak_set_a, peak_set_a, error_tolerance)
 *         bb = convolve_peak_sets2_deconvoluted(peak_set_b, peak_set_b, error_tolerance)             # <<<<<<<<<<<<<<
 *     elif peak_collection is object:
 *         raise TypeError("Cannot handle objects of type %s" % type(peak_set_a))
 */

#line 628 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 628 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 628 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_18similarity_methods_convolve_peak_sets2_deconvoluted(__pyx_v_peak_set_b, __pyx_v_peak_set_b, &__pyx_t_2); 

#line 628 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_bb = __pyx_t_1;

  /* "ms_deisotope/_c/similarity_methods.pyx":632
 *         raise TypeError("Cannot handle objects of type %s" % type(peak_set_a))
 * 
 *     return (ab) / (sqrt((aa)) * sqrt((bb)))             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = (sqrt(__pyx_v_aa) * sqrt(__pyx_v_bb));

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_t_1 == 0)) {

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 632, __pyx_L1_error)

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = (__pyx_v_ab / __pyx_t_1);

#line 632 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":614
 * 
 * 
 * cpdef double ppm_peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double ab, aa, bb
 */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods.ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_39__pyx_fuse_2ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_39__pyx_fuse_2ppm_peak_set_similarity = 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_2ppm_peak_set_similarity", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_39__pyx_fuse_2ppm_peak_set_similarity, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_6ppm_peak_set_similarity};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_39__pyx_fuse_2ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_error_tolerance
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2ppm_peak_set_similarity (wrapper)", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_error_tolerance,0};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2ppm_peak_set_similarity", 0, 2, 3, 1); 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2ppm_peak_set_similarity") < 0)) __PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)values[0]);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)values[1]);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_k__20;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2ppm_peak_set_similarity", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_2ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet, 1, "peak_set_a", 0))) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet, 1, "peak_set_b", 0))) __PYX_ERR(0, 614, __pyx_L1_error)
  __pyx_r = 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_38__pyx_fuse_2ppm_peak_set_similarity(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_error_tolerance);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_38__pyx_fuse_2ppm_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, double __pyx_v_error_tolerance) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity __pyx_t_2;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2ppm_peak_set_similarity", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_2ppm_peak_set_similarity (wrapper)", __pyx_f[0], 614, 0, __PYX_ERR(0, 614, __pyx_L1_error));

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); 

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_3;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_2ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_41__pyx_fuse_3ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_7ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static double __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(PyObject *__pyx_v_peak_set_a, CYTHON_UNUSED PyObject *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity *__pyx_optional_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3ppm_peak_set_similarity", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_3ppm_peak_set_similarity", __pyx_f[0], 614, 0, __PYX_ERR(0, 614, __pyx_L1_error));

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":630
 *         bb = convolve_peak_sets2_deconvoluted(peak_set_b, peak_set_b, error_tolerance)
 *     elif peak_collection is object:
 *         raise TypeError("Cannot handle objects of type %s" % type(peak_set_a))             # <<<<<<<<<<<<<<
 * 
 *     return (ab) / (sqrt((aa)) * sqrt((bb)))
 */

#line 630 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_handle_objects_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_peak_set_a))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)

#line 630 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 630 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)

#line 630 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 630 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 630 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 630 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 630 "ms_deisotope/_c/similarity_methods.pyx"
  __PYX_ERR(0, 630, __pyx_L1_error)

  /* "ms_deisotope/_c/similarity_methods.pyx":614
 * 
 * 
 * cpdef double ppm_peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double ab, aa, bb
 */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.similarity_methods.ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(Py_None, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_41__pyx_fuse_3ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_41__pyx_fuse_3ppm_peak_set_similarity = 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_3ppm_peak_set_similarity", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_41__pyx_fuse_3ppm_peak_set_similarity, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_6ppm_peak_set_similarity};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_41__pyx_fuse_3ppm_peak_set_similarity(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peak_set_a = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_peak_set_b = 0
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_error_tolerance
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3ppm_peak_set_similarity (wrapper)", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_error_tolerance,0};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[3] = {0,0,0};

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  3: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3ppm_peak_set_similarity", 0, 2, 3, 1); 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3ppm_peak_set_similarity") < 0)) __PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = values[0];

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = values[1];

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_k__21;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3ppm_peak_set_similarity", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 614, __pyx_L3_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_3ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 614 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_40__pyx_fuse_3ppm_peak_set_similarity(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_error_tolerance);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_40__pyx_fuse_3ppm_peak_set_similarity(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, double __pyx_v_error_tolerance) {

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_t_1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity __pyx_t_2;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3ppm_peak_set_similarity", 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_3ppm_peak_set_similarity (wrapper)", __pyx_f[0], 614, 0, __PYX_ERR(0, 614, __pyx_L1_error));

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_ppm_peak_set_similarity(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2); 

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_3;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"


#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_3ppm_peak_set_similarity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "ms_deisotope/_c/similarity_methods.pyx":635
 * 
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,             # <<<<<<<<<<<<<<
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_8align_peaks[] = "align_peaks(signatures, args, kwargs, defaults)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_18similarity_methods_9align_peaks = 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
{"align_peaks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9align_peaks, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_8align_peaks};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_signatures = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_args = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signatures,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[4] = {0,0,0,0};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    values[2] = __pyx_k__22;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    values[3] = __pyx_k__23;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  4: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signatures)) != 0)) kw_args--;
        else 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 1); 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  3:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[3] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_signatures = values[0];

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_args = values[1];

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_kwargs = values[2];

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_defaults = values[3];

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_8align_peaks(__pyx_self, __pyx_v_signatures, __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_8align_peaks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  Py_ssize_t __pyx_v_i
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_arg = NULL
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_candidates = NULL
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_sig = NULL
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_match_found
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_src_sig = NULL
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_dst_type = NULL
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_2;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_3;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_4;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_5;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_7;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_8;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_9;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_13;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_14;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  Py_ssize_t __pyx_t_15;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_t_16;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("align_peaks", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("align_peaks", __pyx_f[0], 635, 0, __PYX_ERR(0, 635, __pyx_L1_error));

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(__pyx_v_kwargs);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_4) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L4_bool_binop_done;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((!__pyx_t_4) != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_bool_binop_done:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_itemsize = -1L;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_args == Py_None)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = ((0 < __pyx_t_5) != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(((PyObject*)__pyx_v_args), 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L6;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_4) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L7_bool_binop_done;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_peak_set_a, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = (__pyx_t_4 != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L7_bool_binop_done:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_2) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_peak_set_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L6;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /*else*/ {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_int_2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_int_2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_n_s_s);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_n_s_s);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L6:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  while (1) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s__2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_peak_picker__c_peak_set);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_3) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s__2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_ms_peak_picker__c_peak_index);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PeakIndex); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = (__pyx_t_3 != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_2) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakIndex, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s__2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s__2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s__2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_deisotope__c_peak_set);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_INCREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_DeconvolutedPeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_3) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_DeconvolutedPeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      goto __pyx_L10_break;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_object, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L10_break;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L10_break:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_v_candidates = ((PyObject*)__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_5 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(__pyx_v_signatures == Py_None)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_t_6;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  while (1) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_7, &__pyx_t_5, &__pyx_t_6, NULL, NULL, __pyx_t_8);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 == 0)) break;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_match_found = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_12 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(__pyx_t_12)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_12);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(function);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__3) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_10);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_10 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (likely(__pyx_t_10)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(__pyx_t_10);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_INCREF(function);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__4) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_14 = __pyx_t_13;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_i = __pyx_t_15;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_2 = (__pyx_t_3 != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_t_2) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_6, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_t_2) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_match_found = 1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L19;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        /*else*/ {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_match_found = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          goto __pyx_L17_break;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_L19:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_L17_break:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_2 = (__pyx_v_match_found != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_t_2) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((!__pyx_t_2) != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_3) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = ((__pyx_t_7 > 1) != 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_t_3) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /*else*/ {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_v_signatures == Py_None)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_candidates, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_r = __pyx_t_11;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_t_11 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    goto __pyx_L0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_11);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_12);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_dest_sig);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_ndarray);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_arg);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_candidates);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_sig);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_src_sig);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_dst_type);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_45__pyx_fuse_0align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks *__pyx_optional_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_error_tolerance = __pyx_k__24;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_shift = __pyx_k__25;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_normalize = __pyx_k__26;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0align_peaks", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_0align_peaks", __pyx_f[0], 635, 0, __PYX_ERR(0, 635, __pyx_L1_error));

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_shift = __pyx_optional_args->shift;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_optional_args->__pyx_n > 2) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_normalize = __pyx_optional_args->normalize;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":638
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         return SpectrumAlignment(peak_set_a, peak_set_b, error_tolerance, shift, normalize, sqrt_transform=False)             # <<<<<<<<<<<<<<
 *     elif peak_collection is PeakIndex:
 *         return SpectrumAlignment(peak_set_a.peaks, peak_set_b.peaks, error_tolerance, shift, normalize, sqrt_transform=False)
 */

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_shift); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_normalize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set_a));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak_set_a));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_peak_set_a));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set_b));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak_set_b));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_peak_set_b));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_2);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_3);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_sqrt_transform, Py_False) < 0) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment), __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_t_2);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":635
 * 
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,             # <<<<<<<<<<<<<<
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_45__pyx_fuse_0align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_45__pyx_fuse_0align_peaks = 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_0align_peaks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_45__pyx_fuse_0align_peaks, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_8align_peaks};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_45__pyx_fuse_0align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_error_tolerance
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_shift
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_normalize
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0align_peaks (wrapper)", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_error_tolerance,&__pyx_n_s_shift,&__pyx_n_s_normalize,0};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[5] = {0,0,0,0,0};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  5: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0align_peaks", 0, 2, 5, 1); 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  3:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shift);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[3] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  4:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_normalize);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0align_peaks") < 0)) __PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[1]);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_k__24;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[3]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_shift == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = __pyx_k__25;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[4]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_normalize = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_normalize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_normalize = __pyx_k__26;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0align_peaks", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_0align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peak_set_a", 0))) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peak_set_b", 0))) __PYX_ERR(0, 635, __pyx_L1_error)
  __pyx_r = 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_44__pyx_fuse_0align_peaks(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_error_tolerance, __pyx_v_shift, __pyx_v_normalize);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_44__pyx_fuse_0align_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peak_set_b, double __pyx_v_error_tolerance, double __pyx_v_shift, int __pyx_v_normalize) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_0__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks __pyx_t_2;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0align_peaks", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_0align_peaks (wrapper)", __pyx_f[0], 635, 0, __PYX_ERR(0, 635, __pyx_L1_error));

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 3;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.shift = __pyx_v_shift;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.normalize = __pyx_v_normalize;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_0align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_47__pyx_fuse_1align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks *__pyx_optional_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_error_tolerance = __pyx_k__27;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_shift = __pyx_k__28;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_normalize = __pyx_k__29;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1align_peaks", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_1align_peaks", __pyx_f[0], 635, 0, __PYX_ERR(0, 635, __pyx_L1_error));

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_shift = __pyx_optional_args->shift;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_optional_args->__pyx_n > 2) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_normalize = __pyx_optional_args->normalize;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":640
 *         return SpectrumAlignment(peak_set_a, peak_set_b, error_tolerance, shift, normalize, sqrt_transform=False)
 *     elif peak_collection is PeakIndex:
 *         return SpectrumAlignment(peak_set_a.peaks, peak_set_b.peaks, error_tolerance, shift, normalize, sqrt_transform=False)             # <<<<<<<<<<<<<<
 *     elif peak_collection is object:
 *         raise TypeError("Cannot handle objects of type %s" % type(peak_set_a))
 */

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_shift); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 640, __pyx_L1_error)

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_normalize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 640, __pyx_L1_error)

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 640, __pyx_L1_error)

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set_a->peaks));

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak_set_a->peaks));

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_peak_set_a->peaks));

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set_b->peaks));

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak_set_b->peaks));

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_peak_set_b->peaks));

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_2);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_3);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 640, __pyx_L1_error)

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_sqrt_transform, Py_False) < 0) __PYX_ERR(0, 640, __pyx_L1_error)

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment), __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 640, __pyx_L1_error)

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_t_2);

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

#line 640 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":635
 * 
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,             # <<<<<<<<<<<<<<
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_47__pyx_fuse_1align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_47__pyx_fuse_1align_peaks = 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_1align_peaks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_47__pyx_fuse_1align_peaks, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_8align_peaks};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_47__pyx_fuse_1align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_error_tolerance
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_shift
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_normalize
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1align_peaks (wrapper)", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_error_tolerance,&__pyx_n_s_shift,&__pyx_n_s_normalize,0};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[5] = {0,0,0,0,0};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  5: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1align_peaks", 0, 2, 5, 1); 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  3:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shift);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[3] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  4:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_normalize);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1align_peaks") < 0)) __PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[0]);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[1]);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_k__27;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[3]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_shift == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = __pyx_k__28;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[4]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_normalize = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_normalize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_normalize = __pyx_k__29;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1align_peaks", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_1align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peak_set_a", 0))) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peak_set_b", 0))) __PYX_ERR(0, 635, __pyx_L1_error)
  __pyx_r = 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_46__pyx_fuse_1align_peaks(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_error_tolerance, __pyx_v_shift, __pyx_v_normalize);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_46__pyx_fuse_1align_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_a, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peak_set_b, double __pyx_v_error_tolerance, double __pyx_v_shift, int __pyx_v_normalize) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_1__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks __pyx_t_2;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1align_peaks", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_1align_peaks (wrapper)", __pyx_f[0], 635, 0, __PYX_ERR(0, 635, __pyx_L1_error));

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 3;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.shift = __pyx_v_shift;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.normalize = __pyx_v_normalize;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_1align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_49__pyx_fuse_2align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks *__pyx_optional_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_error_tolerance = __pyx_k__30;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  double __pyx_v_shift = __pyx_k__31;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_v_normalize = __pyx_k__32;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2align_peaks", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_2align_peaks", __pyx_f[0], 635, 0, __PYX_ERR(0, 635, __pyx_L1_error));

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        __pyx_v_shift = __pyx_optional_args->shift;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (__pyx_optional_args->__pyx_n > 2) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          __pyx_v_normalize = __pyx_optional_args->normalize;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":638
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         return SpectrumAlignment(peak_set_a, peak_set_b, error_tolerance, shift, normalize, sqrt_transform=False)             # <<<<<<<<<<<<<<
 *     elif peak_collection is PeakIndex:
 *         return SpectrumAlignment(peak_set_a.peaks, peak_set_b.peaks, error_tolerance, shift, normalize, sqrt_transform=False)
 */

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_shift); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_normalize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set_a));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak_set_a));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_peak_set_a));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set_b));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak_set_b));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_peak_set_b));

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_2);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_3);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_sqrt_transform, Py_False) < 0) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment), __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 638, __pyx_L1_error)

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_t_2);

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

#line 638 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/similarity_methods.pyx":635
 * 
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,             # <<<<<<<<<<<<<<
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_49__pyx_fuse_2align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_49__pyx_fuse_2align_peaks = 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_2align_peaks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_49__pyx_fuse_2align_peaks, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_8align_peaks};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_49__pyx_fuse_2align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_error_tolerance
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_shift
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_normalize
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2align_peaks (wrapper)", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_error_tolerance,&__pyx_n_s_shift,&__pyx_n_s_normalize,0};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[5] = {0,0,0,0,0};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  5: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2align_peaks", 0, 2, 5, 1); 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  3:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shift);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[3] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  4:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_normalize);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2align_peaks") < 0)) __PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)values[0]);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)values[1]);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_k__30;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[3]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_shift == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = __pyx_k__31;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[4]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_normalize = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_normalize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_normalize = __pyx_k__32;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2align_peaks", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_2align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_a), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet, 1, "peak_set_a", 0))) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak_set_b), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet, 1, "peak_set_b", 0))) __PYX_ERR(0, 635, __pyx_L1_error)
  __pyx_r = 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_48__pyx_fuse_2align_peaks(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_error_tolerance, __pyx_v_shift, __pyx_v_normalize);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_48__pyx_fuse_2align_peaks(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_a, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peak_set_b, double __pyx_v_error_tolerance, double __pyx_v_shift, int __pyx_v_normalize) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_2__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks __pyx_t_2;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2align_peaks", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_2align_peaks (wrapper)", __pyx_f[0], 635, 0, __PYX_ERR(0, 635, __pyx_L1_error));

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 3;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.shift = __pyx_v_shift;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.normalize = __pyx_v_normalize;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_2align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_51__pyx_fuse_3align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(PyObject *__pyx_v_peak_set_a, CYTHON_UNUSED PyObject *__pyx_v_peak_set_b, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks *__pyx_optional_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3align_peaks", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_3align_peaks", __pyx_f[0], 635, 0, __PYX_ERR(0, 635, __pyx_L1_error));

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (__pyx_optional_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

  /* "ms_deisotope/_c/similarity_methods.pyx":642
 *         return SpectrumAlignment(peak_set_a.peaks, peak_set_b.peaks, error_tolerance, shift, normalize, sqrt_transform=False)
 *     elif peak_collection is object:
 *         raise TypeError("Cannot handle objects of type %s" % type(peak_set_a))             # <<<<<<<<<<<<<<
 */

#line 642 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_handle_objects_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_peak_set_a))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)

#line 642 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 642 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 642, __pyx_L1_error)

#line 642 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 642 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 642 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 642 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 642 "ms_deisotope/_c/similarity_methods.pyx"
  __PYX_ERR(0, 642, __pyx_L1_error)

  /* "ms_deisotope/_c/similarity_methods.pyx":635
 * 
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,             # <<<<<<<<<<<<<<
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
/* Python wrapper */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_51__pyx_fuse_3align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_51__pyx_fuse_3align_peaks = 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
{"__pyx_fuse_3align_peaks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_51__pyx_fuse_3align_peaks, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_8align_peaks};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_51__pyx_fuse_3align_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peak_set_a = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  PyObject *__pyx_v_peak_set_b = 0
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_error_tolerance
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  double __pyx_v_shift
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;
  int __pyx_v_normalize
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3align_peaks (wrapper)", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak_set_a,&__pyx_n_s_peak_set_b,&__pyx_n_s_error_tolerance,&__pyx_n_s_shift,&__pyx_n_s_normalize,0};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    PyObject* values[5] = {0,0,0,0,0};

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (unlikely(__pyx_kwds)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      Py_ssize_t kw_args;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {
        case  5: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  0: break;
        default: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (pos_args) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  0:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_a)) != 0)) kw_args--;
        else 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  1:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_set_b)) != 0)) kw_args--;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3align_peaks", 0, 2, 5, 1); 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  2:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  3:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shift);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[3] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        case  4:

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (kw_args > 0) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_normalize);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
          if (value) { values[4] = value; kw_args--; }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      if (unlikely(kw_args > 0)) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3align_peaks") < 0)) __PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
        break;
        default: 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
goto __pyx_L5_argtuple_error;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_a = values[0];

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    __pyx_v_peak_set_b = values[1];

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[2]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_error_tolerance = __pyx_k__33;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[3]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_shift == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_shift = __pyx_k__34;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    if (values[4]) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_normalize = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_normalize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    } else {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
      __pyx_v_normalize = __pyx_k__35;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
    }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  }

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3align_peaks", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__PYX_ERR(0, 635, __pyx_L3_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L3_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_3align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 635 "ms_deisotope/_c/similarity_methods.pyx"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_50__pyx_fuse_3align_peaks(__pyx_self, __pyx_v_peak_set_a, __pyx_v_peak_set_b, __pyx_v_error_tolerance, __pyx_v_shift, __pyx_v_normalize);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_50__pyx_fuse_3align_peaks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak_set_a, PyObject *__pyx_v_peak_set_b, double __pyx_v_error_tolerance, double __pyx_v_shift, int __pyx_v_normalize) {

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyDeclarations

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  struct __pyx_fuse_3__pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_align_peaks __pyx_t_2;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_lineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  const char *__pyx_filename = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  int __pyx_clineno = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3align_peaks", 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceCall("__pyx_fuse_3align_peaks (wrapper)", __pyx_f[0], 635, 0, __PYX_ERR(0, 635, __pyx_L1_error));

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.__pyx_n = 3;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.shift = __pyx_v_shift;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2.normalize = __pyx_v_normalize;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_18similarity_methods_align_peaks(__pyx_v_peak_set_a, __pyx_v_peak_set_b, 0, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = __pyx_t_1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  goto __pyx_L0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"


#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  /* function exit code */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L1_error:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_fuse_3align_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_r = NULL;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_L0:;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_RefNannyFinishContext();

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  return __pyx_r;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PeakMatch(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_11__pyx_unpickle_PeakMatch(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_10__pyx_unpickle_PeakMatch[] = "__pyx_unpickle_PeakMatch(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_18similarity_methods_11__pyx_unpickle_PeakMatch = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_PeakMatch", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_11__pyx_unpickle_PeakMatch, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_10__pyx_unpickle_PeakMatch};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_11__pyx_unpickle_PeakMatch(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_PeakMatch (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PeakMatch", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PeakMatch", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_PeakMatch") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PeakMatch", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_unpickle_PeakMatch", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_10__pyx_unpickle_PeakMatch(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_10__pyx_unpickle_PeakMatch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_TraceDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_TraceFrameInit(__pyx_codeobj__36)

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_PeakMatch", 0);

#line 1 "(tree fragment)"
  __Pyx_TraceCall("__pyx_unpickle_PeakMatch", __pyx_f[1], 1, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x797b24e, 0x429ecd1, 0x7e0f493):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x797b24e, 0x429ecd1, 0x7e0f493) = (peak_a, peak_b, score))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__37, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x797b24e, 0x429ecd1, 0x7e0f493):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x797b24e, 0x429ecd1, 0x7e0f493) = (peak_a, peak_b, score))" % __pyx_checksum)
 *     __pyx_result = PeakMatch.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x797b24e, 0x429ecd1, 0x7e0f493):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x797b24e, 0x429ecd1, 0x7e0f493) = (peak_a, peak_b, score))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = PeakMatch.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x797b24e, 0x429ecd1, 0x7e0f493):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x797b24e, 0x429ecd1, 0x7e0f493) = (peak_a, peak_b, score))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x797b24e, 0x429ecd1, 0x7e0f493) = (peak_a, peak_b, score))" % __pyx_checksum)
 *     __pyx_result = PeakMatch.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PeakMatch__set_state(<PeakMatch> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_18similarity_methods_PeakMatch), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x797b24e, 0x429ecd1, 0x7e0f493) = (peak_a, peak_b, score))" % __pyx_checksum)
 *     __pyx_result = PeakMatch.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PeakMatch__set_state(<PeakMatch> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PeakMatch.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PeakMatch__set_state(<PeakMatch> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PeakMatch__set_state(PeakMatch __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_18similarity_methods___pyx_unpickle_PeakMatch__set_state(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x797b24e, 0x429ecd1, 0x7e0f493) = (peak_a, peak_b, score))" % __pyx_checksum)
 *     __pyx_result = PeakMatch.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PeakMatch__set_state(<PeakMatch> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PeakMatch__set_state(<PeakMatch> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PeakMatch__set_state(PeakMatch __pyx_result, tuple __pyx_state):
 *     __pyx_result.peak_a = __pyx_state[0]; __pyx_result.peak_b = __pyx_state[1]; __pyx_result.score = __pyx_state[2]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PeakMatch(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_unpickle_PeakMatch", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PeakMatch__set_state(<PeakMatch> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PeakMatch__set_state(PeakMatch __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.peak_a = __pyx_state[0]; __pyx_result.peak_b = __pyx_state[1]; __pyx_result.score = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods___pyx_unpickle_PeakMatch__set_state(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_TraceDeclarations

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  double __pyx_t_2;

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_PeakMatch__set_state", 0);

#line 11 "(tree fragment)"
  __Pyx_TraceCall("__pyx_unpickle_PeakMatch__set_state", __pyx_f[1], 11, 0, __PYX_ERR(1, 11, __pyx_L1_error));

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PeakMatch__set_state(PeakMatch __pyx_result, tuple __pyx_state):
 *     __pyx_result.peak_a = __pyx_state[0]; __pyx_result.peak_b = __pyx_state[1]; __pyx_result.score = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->peak_a);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->peak_a));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->peak_a = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->peak_b);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->peak_b));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->peak_b = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->score = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PeakMatch__set_state(PeakMatch __pyx_result, tuple __pyx_state):
 *     __pyx_result.peak_a = __pyx_state[0]; __pyx_result.peak_b = __pyx_state[1]; __pyx_result.score = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = ((__pyx_t_4 > 3) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_5) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_3 = __pyx_t_5;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 13 "(tree fragment)"
  __pyx_t_3 = __pyx_t_6;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.peak_a = __pyx_state[0]; __pyx_result.peak_b = __pyx_state[1]; __pyx_result.score = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_9 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {

#line 14 "(tree fragment)"
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_9)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_8, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PeakMatch__set_state(PeakMatch __pyx_result, tuple __pyx_state):
 *     __pyx_result.peak_a = __pyx_state[0]; __pyx_result.peak_b = __pyx_state[1]; __pyx_result.score = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PeakMatch__set_state(<PeakMatch> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PeakMatch__set_state(PeakMatch __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.peak_a = __pyx_state[0]; __pyx_result.peak_b = __pyx_state[1]; __pyx_result.score = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_unpickle_PeakMatch__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_SpectrumAlignment(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_33__pyx_unpickle_SpectrumAlignment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_18similarity_methods_32__pyx_unpickle_SpectrumAlignment[] = "__pyx_unpickle_SpectrumAlignment(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_18similarity_methods_33__pyx_unpickle_SpectrumAlignment = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_SpectrumAlignment", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_33__pyx_unpickle_SpectrumAlignment, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_32__pyx_unpickle_SpectrumAlignment};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_18similarity_methods_33__pyx_unpickle_SpectrumAlignment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpectrumAlignment (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpectrumAlignment", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpectrumAlignment", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_SpectrumAlignment") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SpectrumAlignment", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_unpickle_SpectrumAlignment", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_18similarity_methods_32__pyx_unpickle_SpectrumAlignment(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_18similarity_methods_32__pyx_unpickle_SpectrumAlignment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_TraceDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_TraceFrameInit(__pyx_codeobj__38)

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpectrumAlignment", 0);

#line 1 "(tree fragment)"
  __Pyx_TraceCall("__pyx_unpickle_SpectrumAlignment", __pyx_f[1], 1, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x158795d, 0xf571677, 0x184e107):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x158795d, 0xf571677, 0x184e107) = (norm_aa_cosine, norm_bb_cosine, normalize, peak_pairs, peak_set_a, peak_set_b, peak_type, score, shift, sqrt_transform))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__39, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x158795d, 0xf571677, 0x184e107):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x158795d, 0xf571677, 0x184e107) = (norm_aa_cosine, norm_bb_cosine, normalize, peak_pairs, peak_set_a, peak_set_b, peak_type, score, shift, sqrt_transform))" % __pyx_checksum)
 *     __pyx_result = SpectrumAlignment.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x158795d, 0xf571677, 0x184e107):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x158795d, 0xf571677, 0x184e107) = (norm_aa_cosine, norm_bb_cosine, normalize, peak_pairs, peak_set_a, peak_set_b, peak_type, score, shift, sqrt_transform))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = SpectrumAlignment.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x158795d, 0xf571677, 0x184e107):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x158795d, 0xf571677, 0x184e107) = (norm_aa_cosine, norm_bb_cosine, normalize, peak_pairs, peak_set_a, peak_set_b, peak_type, score, shift, sqrt_transform))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x158795d, 0xf571677, 0x184e107) = (norm_aa_cosine, norm_bb_cosine, normalize, peak_pairs, peak_set_a, peak_set_b, peak_type, score, shift, sqrt_transform))" % __pyx_checksum)
 *     __pyx_result = SpectrumAlignment.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpectrumAlignment__set_state(<SpectrumAlignment> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x158795d, 0xf571677, 0x184e107) = (norm_aa_cosine, norm_bb_cosine, normalize, peak_pairs, peak_set_a, peak_set_b, peak_type, score, shift, sqrt_transform))" % __pyx_checksum)
 *     __pyx_result = SpectrumAlignment.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SpectrumAlignment__set_state(<SpectrumAlignment> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = SpectrumAlignment.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpectrumAlignment__set_state(<SpectrumAlignment> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_SpectrumAlignment__set_state(SpectrumAlignment __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_18similarity_methods___pyx_unpickle_SpectrumAlignment__set_state(((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x158795d, 0xf571677, 0x184e107) = (norm_aa_cosine, norm_bb_cosine, normalize, peak_pairs, peak_set_a, peak_set_b, peak_type, score, shift, sqrt_transform))" % __pyx_checksum)
 *     __pyx_result = SpectrumAlignment.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SpectrumAlignment__set_state(<SpectrumAlignment> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SpectrumAlignment__set_state(<SpectrumAlignment> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_SpectrumAlignment__set_state(SpectrumAlignment __pyx_result, tuple __pyx_state):
 *     __pyx_result.norm_aa_cosine = __pyx_state[0]; __pyx_result.norm_bb_cosine = __pyx_state[1]; __pyx_result.normalize = __pyx_state[2]; __pyx_result.peak_pairs = __pyx_state[3]; __pyx_result.peak_set_a = __pyx_state[4]; __pyx_result.peak_set_b = __pyx_state[5]; __pyx_result.peak_type = __pyx_state[6]; __pyx_result.score = __pyx_state[7]; __pyx_result.shift = __pyx_state[8]; __pyx_result.sqrt_transform = __pyx_state[9]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_SpectrumAlignment(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_unpickle_SpectrumAlignment", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_SpectrumAlignment__set_state(<SpectrumAlignment> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SpectrumAlignment__set_state(SpectrumAlignment __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.norm_aa_cosine = __pyx_state[0]; __pyx_result.norm_bb_cosine = __pyx_state[1]; __pyx_result.normalize = __pyx_state[2]; __pyx_result.peak_pairs = __pyx_state[3]; __pyx_result.peak_set_a = __pyx_state[4]; __pyx_result.peak_set_b = __pyx_state[5]; __pyx_result.peak_type = __pyx_state[6]; __pyx_result.score = __pyx_state[7]; __pyx_result.shift = __pyx_state[8]; __pyx_result.sqrt_transform = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_18similarity_methods___pyx_unpickle_SpectrumAlignment__set_state(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_TraceDeclarations

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  double __pyx_t_2;

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType __pyx_t_4;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  int __pyx_t_7;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_10 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_SpectrumAlignment__set_state", 0);

#line 11 "(tree fragment)"
  __Pyx_TraceCall("__pyx_unpickle_SpectrumAlignment__set_state", __pyx_f[1], 11, 0, __PYX_ERR(1, 11, __pyx_L1_error));

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_SpectrumAlignment__set_state(SpectrumAlignment __pyx_result, tuple __pyx_state):
 *     __pyx_result.norm_aa_cosine = __pyx_state[0]; __pyx_result.norm_bb_cosine = __pyx_state[1]; __pyx_result.normalize = __pyx_state[2]; __pyx_result.peak_pairs = __pyx_state[3]; __pyx_result.peak_set_a = __pyx_state[4]; __pyx_result.peak_set_b = __pyx_state[5]; __pyx_result.peak_type = __pyx_state[6]; __pyx_result.score = __pyx_state[7]; __pyx_result.shift = __pyx_state[8]; __pyx_result.sqrt_transform = __pyx_state[9]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->norm_aa_cosine = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->norm_bb_cosine = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->normalize = __pyx_t_3;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->peak_pairs);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->peak_pairs);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->peak_pairs = ((PyObject*)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->peak_set_a);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->peak_set_a);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->peak_set_a = __pyx_t_1;

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->peak_set_b);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->peak_set_b);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->peak_set_b = __pyx_t_1;

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_4 = ((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)__Pyx_PyInt_As_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->peak_type = __pyx_t_4;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->score = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->shift = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->sqrt_transform = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_SpectrumAlignment__set_state(SpectrumAlignment __pyx_result, tuple __pyx_state):
 *     __pyx_result.norm_aa_cosine = __pyx_state[0]; __pyx_result.norm_bb_cosine = __pyx_state[1]; __pyx_result.normalize = __pyx_state[2]; __pyx_result.peak_pairs = __pyx_state[3]; __pyx_result.peak_set_a = __pyx_state[4]; __pyx_result.peak_set_b = __pyx_state[5]; __pyx_result.peak_type = __pyx_state[6]; __pyx_result.score = __pyx_state[7]; __pyx_result.shift = __pyx_state[8]; __pyx_result.sqrt_transform = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = ((__pyx_t_5 > 10) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_6) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_3 = __pyx_t_6;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_7 = (__pyx_t_6 != 0);

#line 13 "(tree fragment)"
  __pyx_t_3 = __pyx_t_7;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.norm_aa_cosine = __pyx_state[0]; __pyx_result.norm_bb_cosine = __pyx_state[1]; __pyx_result.normalize = __pyx_state[2]; __pyx_result.peak_pairs = __pyx_state[3]; __pyx_result.peak_set_a = __pyx_state[4]; __pyx_result.peak_set_b = __pyx_state[5]; __pyx_result.peak_type = __pyx_state[6]; __pyx_result.score = __pyx_state[7]; __pyx_result.shift = __pyx_state[8]; __pyx_result.sqrt_transform = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[10])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_9);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __pyx_t_10 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {

#line 14 "(tree fragment)"
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_10)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_10);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_9, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_SpectrumAlignment__set_state(SpectrumAlignment __pyx_result, tuple __pyx_state):
 *     __pyx_result.norm_aa_cosine = __pyx_state[0]; __pyx_result.norm_bb_cosine = __pyx_state[1]; __pyx_result.normalize = __pyx_state[2]; __pyx_result.peak_pairs = __pyx_state[3]; __pyx_result.peak_set_a = __pyx_state[4]; __pyx_result.peak_set_b = __pyx_state[5]; __pyx_result.peak_type = __pyx_state[6]; __pyx_result.score = __pyx_state[7]; __pyx_result.shift = __pyx_state[8]; __pyx_result.sqrt_transform = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_SpectrumAlignment__set_state(<SpectrumAlignment> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SpectrumAlignment__set_state(SpectrumAlignment __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.norm_aa_cosine = __pyx_state[0]; __pyx_result.norm_bb_cosine = __pyx_state[1]; __pyx_result.normalize = __pyx_state[2]; __pyx_result.peak_pairs = __pyx_state[3]; __pyx_result.peak_set_a = __pyx_state[4]; __pyx_result.peak_set_b = __pyx_state[5]; __pyx_result.peak_type = __pyx_state[6]; __pyx_result.score = __pyx_state[7]; __pyx_result.shift = __pyx_state[8]; __pyx_result.sqrt_transform = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_10);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.similarity_methods.__pyx_unpickle_SpectrumAlignment__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("PyArray_MultiIterNew1", __pyx_f[2], 734, 0, __PYX_ERR(2, 734, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 735, __pyx_L1_error)

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("PyArray_MultiIterNew2", __pyx_f[2], 737, 0, __PYX_ERR(2, 737, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 738, __pyx_L1_error)

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("PyArray_MultiIterNew3", __pyx_f[2], 740, 0, __PYX_ERR(2, 740, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 741, __pyx_L1_error)

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("PyArray_MultiIterNew4", __pyx_f[2], 743, 0, __PYX_ERR(2, 743, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 744, __pyx_L1_error)

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("PyArray_MultiIterNew5", __pyx_f[2], 746, 0, __PYX_ERR(2, 746, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 747, __pyx_L1_error)

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("PyDataType_SHAPE", __pyx_f[2], 749, 0, __PYX_ERR(2, 749, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /*else*/ {

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(__pyx_empty_tuple);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = __pyx_empty_tuple;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyDataType_SHAPE", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("set_array_base", 0);

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("set_array_base", __pyx_f[2], 928, 0, __PYX_ERR(2, 928, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":929
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */

#line 929 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_INCREF(__pyx_v_base);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":930
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */

#line 930 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_WriteUnraisable("numpy.set_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(Py_None, 0);

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base
#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("get_array_base", 0);

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("get_array_base", __pyx_f[2], 932, 0, __PYX_ERR(2, 932, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":933
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */

#line 933 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":935
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":936
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_base));

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyObject *)__pyx_v_base);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.get_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_array", 0);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("import_array", __pyx_f[2], 940, 0, __PYX_ERR(2, 940, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":942
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */

#line 942 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 942, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":943
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 943, __pyx_L5_except_error)

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(2, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(Py_None, 0);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_umath", 0);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("import_umath", __pyx_f[2], 946, 0, __PYX_ERR(2, 946, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":948
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 948 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 948, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":949
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 949, __pyx_L5_except_error)

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__41, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(2, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(Py_None, 0);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_ufunc", 0);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("import_ufunc", __pyx_f[2], 952, 0, __PYX_ERR(2, 952, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":954
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 954 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 954, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":955
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 955, __pyx_L5_except_error)

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":956
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__41, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(2, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(Py_None, 0);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("is_timedelta64_object", __pyx_f[2], 966, 0, __PYX_ERR(2, 966, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":978
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_WriteUnraisable("numpy.is_timedelta64_object", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(Py_None, 0);

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("is_datetime64_object", __pyx_f[2], 981, 0, __PYX_ERR(2, 981, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":993
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_WriteUnraisable("numpy.is_datetime64_object", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(Py_None, 0);

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_datetime __pyx_r;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("get_datetime64_value", __pyx_f[2], 996, 1, __PYX_ERR(2, 996, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1003
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_WriteUnraisable("numpy.get_datetime64_value", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(Py_None, 1);

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_timedelta __pyx_r;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("get_timedelta64_value", __pyx_f[2], 1006, 1, __PYX_ERR(2, 1006, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1010
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_WriteUnraisable("numpy.get_timedelta64_value", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(Py_None, 1);

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  NPY_DATETIMEUNIT __pyx_r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceDeclarations

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceCall("get_datetime64_unit", __pyx_f[2], 1013, 1, __PYX_ERR(2, 1013, __pyx_L1_error));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1017
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_WriteUnraisable("numpy.get_datetime64_unit", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = (NPY_DATETIMEUNIT) 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_TraceReturn(Py_None, 1);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "EnumBase":16
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):             # <<<<<<<<<<<<<<
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 */

#line 16 "EnumBase"


#line 16 "EnumBase"
/* Python wrapper */

#line 16 "EnumBase"
static int __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 16 "EnumBase"
static int __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0
#line 16 "EnumBase"
;
  PyObject *__pyx_v_parents = 0
#line 16 "EnumBase"
;
  PyObject *__pyx_v_dct = 0
#line 16 "EnumBase"
;

#line 16 "EnumBase"
  int __pyx_lineno = 0;

#line 16 "EnumBase"
  const char *__pyx_filename = NULL;

#line 16 "EnumBase"
  int __pyx_clineno = 0;

#line 16 "EnumBase"
  int __pyx_r;

#line 16 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 16 "EnumBase"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 16 "EnumBase"
  {

#line 16 "EnumBase"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_parents,&__pyx_n_s_dct,0};

#line 16 "EnumBase"
    PyObject* values[3] = {0,0,0};

#line 16 "EnumBase"
    if (unlikely(__pyx_kwds)) {

#line 16 "EnumBase"
      Py_ssize_t kw_args;

#line 16 "EnumBase"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 16 "EnumBase"
      switch (pos_args) {
        case  3: 
#line 16 "EnumBase"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 16 "EnumBase"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 16 "EnumBase"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 16 "EnumBase"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 16 "EnumBase"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 16 "EnumBase"
        CYTHON_FALLTHROUGH;

#line 16 "EnumBase"
        case  0: break;
        default: 
#line 16 "EnumBase"
goto __pyx_L5_argtuple_error;

#line 16 "EnumBase"
      }

#line 16 "EnumBase"
      kw_args = PyDict_Size(__pyx_kwds);

#line 16 "EnumBase"
      switch (pos_args) {

#line 16 "EnumBase"
        case  0:

#line 16 "EnumBase"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else 
#line 16 "EnumBase"
goto __pyx_L5_argtuple_error;

#line 16 "EnumBase"
        CYTHON_FALLTHROUGH;

#line 16 "EnumBase"
        case  1:

#line 16 "EnumBase"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parents)) != 0)) kw_args--;

#line 16 "EnumBase"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); 
#line 16 "EnumBase"
__PYX_ERR(1, 16, __pyx_L3_error)

#line 16 "EnumBase"
        }

#line 16 "EnumBase"
        CYTHON_FALLTHROUGH;

#line 16 "EnumBase"
        case  2:

#line 16 "EnumBase"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dct)) != 0)) kw_args--;

#line 16 "EnumBase"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); 
#line 16 "EnumBase"
__PYX_ERR(1, 16, __pyx_L3_error)

#line 16 "EnumBase"
        }

#line 16 "EnumBase"
      }

#line 16 "EnumBase"
      if (unlikely(kw_args > 0)) {

#line 16 "EnumBase"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)

#line 16 "EnumBase"
      }

#line 16 "EnumBase"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 16 "EnumBase"
      goto __pyx_L5_argtuple_error;

#line 16 "EnumBase"
    } else {

#line 16 "EnumBase"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 16 "EnumBase"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 16 "EnumBase"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 16 "EnumBase"
    }

#line 16 "EnumBase"
    __pyx_v_name = values[0];

#line 16 "EnumBase"
    __pyx_v_parents = values[1];

#line 16 "EnumBase"
    __pyx_v_dct = values[2];

#line 16 "EnumBase"
  }

#line 16 "EnumBase"
  goto __pyx_L4_argument_unpacking_done;

#line 16 "EnumBase"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 16 "EnumBase"
__PYX_ERR(1, 16, __pyx_L3_error)

#line 16 "EnumBase"
  __pyx_L3_error:;

#line 16 "EnumBase"
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 16 "EnumBase"
  return -1;

#line 16 "EnumBase"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 16 "EnumBase"
__pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct);

#line 16 "EnumBase"


#line 16 "EnumBase"
  /* function exit code */

#line 16 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 16 "EnumBase"
  return __pyx_r;

#line 16 "EnumBase"
}

#line 16 "EnumBase"


#line 16 "EnumBase"
static int __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name, PyObject *__pyx_v_parents, PyObject *__pyx_v_dct) {

#line 16 "EnumBase"
  int __pyx_r;

#line 16 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 16 "EnumBase"
  PyObject *__pyx_t_1 = NULL;

#line 16 "EnumBase"
  PyObject *__pyx_t_2 = NULL;

#line 16 "EnumBase"
  PyObject *__pyx_t_3 = NULL;

#line 16 "EnumBase"
  int __pyx_t_4;

#line 16 "EnumBase"
  PyObject *__pyx_t_5 = NULL;

#line 16 "EnumBase"
  int __pyx_lineno = 0;

#line 16 "EnumBase"
  const char *__pyx_filename = NULL;

#line 16 "EnumBase"
  int __pyx_clineno = 0;

#line 16 "EnumBase"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "EnumBase":17
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)             # <<<<<<<<<<<<<<
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):
 */

#line 17 "EnumBase"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyType_Type)), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "EnumBase"
  __Pyx_GOTREF(__pyx_t_2);

#line 17 "EnumBase"
  __pyx_t_3 = NULL;

#line 17 "EnumBase"
  __pyx_t_4 = 0;

#line 17 "EnumBase"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 17 "EnumBase"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 17 "EnumBase"
    if (likely(__pyx_t_3)) {

#line 17 "EnumBase"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 17 "EnumBase"
      __Pyx_INCREF(__pyx_t_3);

#line 17 "EnumBase"
      __Pyx_INCREF(function);

#line 17 "EnumBase"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 17 "EnumBase"
      __pyx_t_4 = 1;

#line 17 "EnumBase"
    }

#line 17 "EnumBase"
  }

#line 17 "EnumBase"
  #if CYTHON_FAST_PYCALL

#line 17 "EnumBase"
  if (PyFunction_Check(__pyx_t_2)) {

#line 17 "EnumBase"
    PyObject *__pyx_temp[5] = {__pyx_t_3, ((PyObject *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct};

#line 17 "EnumBase"
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "EnumBase"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 17 "EnumBase"
    __Pyx_GOTREF(__pyx_t_1);

#line 17 "EnumBase"
  } else

#line 17 "EnumBase"
  #endif

#line 17 "EnumBase"
  #if CYTHON_FAST_PYCCALL

#line 17 "EnumBase"
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {

#line 17 "EnumBase"
    PyObject *__pyx_temp[5] = {__pyx_t_3, ((PyObject *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct};

#line 17 "EnumBase"
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "EnumBase"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 17 "EnumBase"
    __Pyx_GOTREF(__pyx_t_1);

#line 17 "EnumBase"
  } else

#line 17 "EnumBase"
  #endif

#line 17 "EnumBase"
  {

#line 17 "EnumBase"
    __pyx_t_5 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "EnumBase"
    __Pyx_GOTREF(__pyx_t_5);

#line 17 "EnumBase"
    if (__pyx_t_3) {

#line 17 "EnumBase"
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;

#line 17 "EnumBase"
    }

#line 17 "EnumBase"
    __Pyx_INCREF(((PyObject *)__pyx_v_cls));

#line 17 "EnumBase"
    __Pyx_GIVEREF(((PyObject *)__pyx_v_cls));

#line 17 "EnumBase"
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, ((PyObject *)__pyx_v_cls));

#line 17 "EnumBase"
    __Pyx_INCREF(__pyx_v_name);

#line 17 "EnumBase"
    __Pyx_GIVEREF(__pyx_v_name);

#line 17 "EnumBase"
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_name);

#line 17 "EnumBase"
    __Pyx_INCREF(__pyx_v_parents);

#line 17 "EnumBase"
    __Pyx_GIVEREF(__pyx_v_parents);

#line 17 "EnumBase"
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_parents);

#line 17 "EnumBase"
    __Pyx_INCREF(__pyx_v_dct);

#line 17 "EnumBase"
    __Pyx_GIVEREF(__pyx_v_dct);

#line 17 "EnumBase"
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_4, __pyx_v_dct);

#line 17 "EnumBase"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "EnumBase"
    __Pyx_GOTREF(__pyx_t_1);

#line 17 "EnumBase"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 17 "EnumBase"
  }

#line 17 "EnumBase"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 17 "EnumBase"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":18
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()             # <<<<<<<<<<<<<<
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 */

#line 18 "EnumBase"
  __Pyx_INCREF(__Pyx_OrderedDict);
  __pyx_t_2 = __Pyx_OrderedDict; 
#line 18 "EnumBase"
__pyx_t_5 = NULL;

#line 18 "EnumBase"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {

#line 18 "EnumBase"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);

#line 18 "EnumBase"
    if (likely(__pyx_t_5)) {

#line 18 "EnumBase"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 18 "EnumBase"
      __Pyx_INCREF(__pyx_t_5);

#line 18 "EnumBase"
      __Pyx_INCREF(function);

#line 18 "EnumBase"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 18 "EnumBase"
    }

#line 18 "EnumBase"
  }

#line 18 "EnumBase"
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);

#line 18 "EnumBase"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 18 "EnumBase"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 18, __pyx_L1_error)

#line 18 "EnumBase"
  __Pyx_GOTREF(__pyx_t_1);

#line 18 "EnumBase"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 18 "EnumBase"
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members, __pyx_t_1) < 0) __PYX_ERR(1, 18, __pyx_L1_error)

#line 18 "EnumBase"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":16
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):             # <<<<<<<<<<<<<<
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 */

#line 16 "EnumBase"


#line 16 "EnumBase"
  /* function exit code */

#line 16 "EnumBase"
  __pyx_r = 0;

#line 16 "EnumBase"
  goto __pyx_L0;

#line 16 "EnumBase"
  __pyx_L1_error:;

#line 16 "EnumBase"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "EnumBase"
  __Pyx_XDECREF(__pyx_t_2);

#line 16 "EnumBase"
  __Pyx_XDECREF(__pyx_t_3);

#line 16 "EnumBase"
  __Pyx_XDECREF(__pyx_t_5);

#line 16 "EnumBase"
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "EnumBase"
  __pyx_r = -1;

#line 16 "EnumBase"
  __pyx_L0:;

#line 16 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 16 "EnumBase"
  return __pyx_r;

#line 16 "EnumBase"
}

/* "EnumBase":19
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):             # <<<<<<<<<<<<<<
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 */

#line 19 "EnumBase"


#line 19 "EnumBase"
/* Python wrapper */

#line 19 "EnumBase"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__(PyObject *__pyx_v_cls); /*proto*/

#line 19 "EnumBase"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__(PyObject *__pyx_v_cls) {

#line 19 "EnumBase"
  PyObject *__pyx_r = 0;

#line 19 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 19 "EnumBase"
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = 
#line 19 "EnumBase"
__pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls));

#line 19 "EnumBase"


#line 19 "EnumBase"
  /* function exit code */

#line 19 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 19 "EnumBase"
  return __pyx_r;

#line 19 "EnumBase"
}

#line 19 "EnumBase"


#line 19 "EnumBase"
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls) {

#line 19 "EnumBase"
  PyObject *__pyx_r = NULL;

#line 19 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 19 "EnumBase"
  PyObject *__pyx_t_1 = NULL;

#line 19 "EnumBase"
  PyObject *__pyx_t_2 = NULL;

#line 19 "EnumBase"
  PyObject *__pyx_t_3 = NULL;

#line 19 "EnumBase"
  int __pyx_lineno = 0;

#line 19 "EnumBase"
  const char *__pyx_filename = NULL;

#line 19 "EnumBase"
  int __pyx_clineno = 0;

#line 19 "EnumBase"
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "EnumBase":20
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())             # <<<<<<<<<<<<<<
 *     def __getitem__(cls, name):
 *         return cls.__members__[name]
 */

#line 20 "EnumBase"
  __Pyx_XDECREF(__pyx_r);

#line 20 "EnumBase"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 20, __pyx_L1_error)

#line 20 "EnumBase"
  __Pyx_GOTREF(__pyx_t_2);

#line 20 "EnumBase"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 20, __pyx_L1_error)

#line 20 "EnumBase"
  __Pyx_GOTREF(__pyx_t_3);

#line 20 "EnumBase"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 20 "EnumBase"
  __pyx_t_2 = NULL;

#line 20 "EnumBase"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {

#line 20 "EnumBase"
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);

#line 20 "EnumBase"
    if (likely(__pyx_t_2)) {

#line 20 "EnumBase"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 20 "EnumBase"
      __Pyx_INCREF(__pyx_t_2);

#line 20 "EnumBase"
      __Pyx_INCREF(function);

#line 20 "EnumBase"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 20 "EnumBase"
    }

#line 20 "EnumBase"
  }

#line 20 "EnumBase"
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 20 "EnumBase"
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 20 "EnumBase"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 20, __pyx_L1_error)

#line 20 "EnumBase"
  __Pyx_GOTREF(__pyx_t_1);

#line 20 "EnumBase"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 20 "EnumBase"
  __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 20, __pyx_L1_error)

#line 20 "EnumBase"
  __Pyx_GOTREF(__pyx_t_3);

#line 20 "EnumBase"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 20 "EnumBase"
  __pyx_r = __pyx_t_3;

#line 20 "EnumBase"
  __pyx_t_3 = 0;

#line 20 "EnumBase"
  goto __pyx_L0;

  /* "EnumBase":19
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):             # <<<<<<<<<<<<<<
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 */

#line 19 "EnumBase"


#line 19 "EnumBase"
  /* function exit code */

#line 19 "EnumBase"
  __pyx_L1_error:;

#line 19 "EnumBase"
  __Pyx_XDECREF(__pyx_t_1);

#line 19 "EnumBase"
  __Pyx_XDECREF(__pyx_t_2);

#line 19 "EnumBase"
  __Pyx_XDECREF(__pyx_t_3);

#line 19 "EnumBase"
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 19 "EnumBase"
  __pyx_r = NULL;

#line 19 "EnumBase"
  __pyx_L0:;

#line 19 "EnumBase"
  __Pyx_XGIVEREF(__pyx_r);

#line 19 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 19 "EnumBase"
  return __pyx_r;

#line 19 "EnumBase"
}

/* "EnumBase":21
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):             # <<<<<<<<<<<<<<
 *         return cls.__members__[name]
 * 
 */

#line 21 "EnumBase"


#line 21 "EnumBase"
/* Python wrapper */

#line 21 "EnumBase"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__(PyObject *__pyx_v_cls, PyObject *__pyx_v_name); /*proto*/

#line 21 "EnumBase"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__(PyObject *__pyx_v_cls, PyObject *__pyx_v_name) {

#line 21 "EnumBase"
  PyObject *__pyx_r = 0;

#line 21 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 21 "EnumBase"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 21 "EnumBase"
__pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls), ((PyObject *)__pyx_v_name));

#line 21 "EnumBase"


#line 21 "EnumBase"
  /* function exit code */

#line 21 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 21 "EnumBase"
  return __pyx_r;

#line 21 "EnumBase"
}

#line 21 "EnumBase"


#line 21 "EnumBase"
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name) {

#line 21 "EnumBase"
  PyObject *__pyx_r = NULL;

#line 21 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 21 "EnumBase"
  PyObject *__pyx_t_1 = NULL;

#line 21 "EnumBase"
  PyObject *__pyx_t_2 = NULL;

#line 21 "EnumBase"
  int __pyx_lineno = 0;

#line 21 "EnumBase"
  const char *__pyx_filename = NULL;

#line 21 "EnumBase"
  int __pyx_clineno = 0;

#line 21 "EnumBase"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "EnumBase":22
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 *         return cls.__members__[name]             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 22 "EnumBase"
  __Pyx_XDECREF(__pyx_r);

#line 22 "EnumBase"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 22, __pyx_L1_error)

#line 22 "EnumBase"
  __Pyx_GOTREF(__pyx_t_1);

#line 22 "EnumBase"
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 22, __pyx_L1_error)

#line 22 "EnumBase"
  __Pyx_GOTREF(__pyx_t_2);

#line 22 "EnumBase"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 22 "EnumBase"
  __pyx_r = __pyx_t_2;

#line 22 "EnumBase"
  __pyx_t_2 = 0;

#line 22 "EnumBase"
  goto __pyx_L0;

  /* "EnumBase":21
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):             # <<<<<<<<<<<<<<
 *         return cls.__members__[name]
 * 
 */

#line 21 "EnumBase"


#line 21 "EnumBase"
  /* function exit code */

#line 21 "EnumBase"
  __pyx_L1_error:;

#line 21 "EnumBase"
  __Pyx_XDECREF(__pyx_t_1);

#line 21 "EnumBase"
  __Pyx_XDECREF(__pyx_t_2);

#line 21 "EnumBase"
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 21 "EnumBase"
  __pyx_r = NULL;

#line 21 "EnumBase"
  __pyx_L0:;

#line 21 "EnumBase"
  __Pyx_XGIVEREF(__pyx_r);

#line 21 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 21 "EnumBase"
  return __pyx_r;

#line 21 "EnumBase"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_empty_tuple);

#line 5 "(tree fragment)"
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_1;

#line 6 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_2 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));

#line 8 "(tree fragment)"
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = 0;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_3 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_222419149);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_222419149);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_5;

#line 13 "(tree fragment)"
    __pyx_t_5 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 15 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_222419149);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_222419149);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_5);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);

#line 15 "(tree fragment)"
    __pyx_t_5 = 0;

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_4;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_unpickle___Pyx_EnumMeta__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

#line 28 "EnumBase"


#line 28 "EnumBase"
/* Python wrapper */

#line 28 "EnumBase"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_1__new__ = 
#line 28 "EnumBase"
{"__new__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__, METH_VARARGS|METH_KEYWORDS, 0};

#line 28 "EnumBase"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0
#line 28 "EnumBase"
;
  PyObject *__pyx_v_value = 0
#line 28 "EnumBase"
;
  PyObject *__pyx_v_name = 0
#line 28 "EnumBase"
;

#line 28 "EnumBase"
  int __pyx_lineno = 0;

#line 28 "EnumBase"
  const char *__pyx_filename = NULL;

#line 28 "EnumBase"
  int __pyx_clineno = 0;

#line 28 "EnumBase"
  PyObject *__pyx_r = 0;

#line 28 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 28 "EnumBase"
  __Pyx_RefNannySetupContext("__new__ (wrapper)", 0);

#line 28 "EnumBase"
  {

#line 28 "EnumBase"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_value,&__pyx_n_s_name,0};

#line 28 "EnumBase"
    PyObject* values[3] = {0,0,0};

#line 28 "EnumBase"
    values[2] = ((PyObject *)((PyObject *)Py_None));

#line 28 "EnumBase"
    if (unlikely(__pyx_kwds)) {

#line 28 "EnumBase"
      Py_ssize_t kw_args;

#line 28 "EnumBase"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 28 "EnumBase"
      switch (pos_args) {
        case  3: 
#line 28 "EnumBase"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 28 "EnumBase"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 28 "EnumBase"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 28 "EnumBase"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 28 "EnumBase"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 28 "EnumBase"
        CYTHON_FALLTHROUGH;

#line 28 "EnumBase"
        case  0: break;
        default: 
#line 28 "EnumBase"
goto __pyx_L5_argtuple_error;

#line 28 "EnumBase"
      }

#line 28 "EnumBase"
      kw_args = PyDict_Size(__pyx_kwds);

#line 28 "EnumBase"
      switch (pos_args) {

#line 28 "EnumBase"
        case  0:

#line 28 "EnumBase"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else 
#line 28 "EnumBase"
goto __pyx_L5_argtuple_error;

#line 28 "EnumBase"
        CYTHON_FALLTHROUGH;

#line 28 "EnumBase"
        case  1:

#line 28 "EnumBase"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;

#line 28 "EnumBase"
        else {
          __Pyx_RaiseArgtupleInvalid("__new__", 0, 2, 3, 1); 
#line 28 "EnumBase"
__PYX_ERR(1, 28, __pyx_L3_error)

#line 28 "EnumBase"
        }

#line 28 "EnumBase"
        CYTHON_FALLTHROUGH;

#line 28 "EnumBase"
        case  2:

#line 28 "EnumBase"
        if (kw_args > 0) {

#line 28 "EnumBase"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name);

#line 28 "EnumBase"
          if (value) { values[2] = value; kw_args--; }

#line 28 "EnumBase"
        }

#line 28 "EnumBase"
      }

#line 28 "EnumBase"
      if (unlikely(kw_args > 0)) {

#line 28 "EnumBase"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__new__") < 0)) __PYX_ERR(1, 28, __pyx_L3_error)

#line 28 "EnumBase"
      }

#line 28 "EnumBase"
    } else {

#line 28 "EnumBase"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 28 "EnumBase"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 28 "EnumBase"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 28 "EnumBase"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 28 "EnumBase"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 28 "EnumBase"
        break;
        default: 
#line 28 "EnumBase"
goto __pyx_L5_argtuple_error;

#line 28 "EnumBase"
      }

#line 28 "EnumBase"
    }

#line 28 "EnumBase"
    __pyx_v_cls = values[0];

#line 28 "EnumBase"
    __pyx_v_value = values[1];

#line 28 "EnumBase"
    __pyx_v_name = values[2];

#line 28 "EnumBase"
  }

#line 28 "EnumBase"
  goto __pyx_L4_argument_unpacking_done;

#line 28 "EnumBase"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__new__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 28 "EnumBase"
__PYX_ERR(1, 28, __pyx_L3_error)

#line 28 "EnumBase"
  __pyx_L3_error:;

#line 28 "EnumBase"
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 28 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 28 "EnumBase"
  return NULL;

#line 28 "EnumBase"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 28 "EnumBase"
__pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(__pyx_self, __pyx_v_cls, __pyx_v_value, __pyx_v_name);

#line 28 "EnumBase"


#line 28 "EnumBase"
  /* function exit code */

#line 28 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 28 "EnumBase"
  return __pyx_r;

#line 28 "EnumBase"
}

#line 28 "EnumBase"


#line 28 "EnumBase"
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_value, PyObject *__pyx_v_name) {
  PyObject *__pyx_v_v = NULL
#line 28 "EnumBase"
;
  PyObject *__pyx_v_res = NULL
#line 28 "EnumBase"
;

#line 28 "EnumBase"
  PyObject *__pyx_r = NULL;

#line 28 "EnumBase"
  __Pyx_TraceDeclarations

#line 28 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 28 "EnumBase"
  PyObject *__pyx_t_1 = NULL;

#line 28 "EnumBase"
  Py_ssize_t __pyx_t_2;

#line 28 "EnumBase"
  PyObject *(*__pyx_t_3)(PyObject *);

#line 28 "EnumBase"
  PyObject *__pyx_t_4 = NULL;

#line 28 "EnumBase"
  int __pyx_t_5;

#line 28 "EnumBase"
  int __pyx_t_6;

#line 28 "EnumBase"
  PyObject *__pyx_t_7 = NULL;

#line 28 "EnumBase"
  int __pyx_t_8;

#line 28 "EnumBase"
  PyObject *__pyx_t_9 = NULL;

#line 28 "EnumBase"
  int __pyx_t_10;

#line 28 "EnumBase"
  int __pyx_lineno = 0;

#line 28 "EnumBase"
  const char *__pyx_filename = NULL;

#line 28 "EnumBase"
  int __pyx_clineno = 0;

#line 28 "EnumBase"
  __Pyx_TraceFrameInit(__pyx_codeobj__42)

#line 28 "EnumBase"
  __Pyx_RefNannySetupContext("__new__", 0);

#line 28 "EnumBase"
  __Pyx_TraceCall("__new__", __pyx_f[1], 28, 0, __PYX_ERR(1, 28, __pyx_L1_error));

  /* "EnumBase":29
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 *         for v in cls:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return v
 */

#line 29 "EnumBase"
  if (likely(PyList_CheckExact(__pyx_v_cls)) || PyTuple_CheckExact(__pyx_v_cls)) {

#line 29 "EnumBase"
    __pyx_t_1 = __pyx_v_cls; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;

#line 29 "EnumBase"
    __pyx_t_3 = NULL;

#line 29 "EnumBase"
  } else {
    __pyx_t_2 = -1; 
#line 29 "EnumBase"
__pyx_t_1 = PyObject_GetIter(__pyx_v_cls); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 29, __pyx_L1_error)

#line 29 "EnumBase"
    __Pyx_GOTREF(__pyx_t_1);

#line 29 "EnumBase"
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 29, __pyx_L1_error)

#line 29 "EnumBase"
  }

#line 29 "EnumBase"
  for (;;) {

#line 29 "EnumBase"
    if (likely(!__pyx_t_3)) {

#line 29 "EnumBase"
      if (likely(PyList_CheckExact(__pyx_t_1))) {

#line 29 "EnumBase"
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 29 "EnumBase"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 29 "EnumBase"
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 29, __pyx_L1_error)

#line 29 "EnumBase"
        #else

#line 29 "EnumBase"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)

#line 29 "EnumBase"
        __Pyx_GOTREF(__pyx_t_4);

#line 29 "EnumBase"
        #endif

#line 29 "EnumBase"
      } else {

#line 29 "EnumBase"
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;

#line 29 "EnumBase"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 29 "EnumBase"
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 29, __pyx_L1_error)

#line 29 "EnumBase"
        #else

#line 29 "EnumBase"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)

#line 29 "EnumBase"
        __Pyx_GOTREF(__pyx_t_4);

#line 29 "EnumBase"
        #endif

#line 29 "EnumBase"
      }
    } else 
#line 29 "EnumBase"
{

#line 29 "EnumBase"
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);

#line 29 "EnumBase"
      if (unlikely(!__pyx_t_4)) {

#line 29 "EnumBase"
        PyObject* exc_type = PyErr_Occurred();

#line 29 "EnumBase"
        if (exc_type) {

#line 29 "EnumBase"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 29 "EnumBase"
          else __PYX_ERR(1, 29, __pyx_L1_error)

#line 29 "EnumBase"
        }

#line 29 "EnumBase"
        break;

#line 29 "EnumBase"
      }

#line 29 "EnumBase"
      __Pyx_GOTREF(__pyx_t_4);

#line 29 "EnumBase"
    }

#line 29 "EnumBase"
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);

#line 29 "EnumBase"
    __pyx_t_4 = 0;

    /* "EnumBase":30
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return v
 *         if name is None:
 */

#line 30 "EnumBase"
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_v, __pyx_v_value, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 30, __pyx_L1_error)

#line 30 "EnumBase"
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 30, __pyx_L1_error)

#line 30 "EnumBase"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 30 "EnumBase"
    if (__pyx_t_5) {

      /* "EnumBase":31
 *         for v in cls:
 *             if v == value:
 *                 return v             # <<<<<<<<<<<<<<
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)
 */

#line 31 "EnumBase"
      __Pyx_XDECREF(__pyx_r);

#line 31 "EnumBase"
      __Pyx_INCREF(__pyx_v_v);

#line 31 "EnumBase"
      __pyx_r = __pyx_v_v;

#line 31 "EnumBase"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 31 "EnumBase"
      goto __pyx_L0;

      /* "EnumBase":30
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return v
 *         if name is None:
 */

#line 30 "EnumBase"
    }

    /* "EnumBase":29
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 *         for v in cls:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return v
 */

#line 29 "EnumBase"
  }

#line 29 "EnumBase"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":32
 *             if v == value:
 *                 return v
 *         if name is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 */

#line 32 "EnumBase"
  __pyx_t_5 = (__pyx_v_name == Py_None);

#line 32 "EnumBase"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 32 "EnumBase"
  if (unlikely(__pyx_t_6)) {

    /* "EnumBase":33
 *                 return v
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)             # <<<<<<<<<<<<<<
 *         res = int.__new__(cls, value)
 *         res.name = name
 */

#line 33 "EnumBase"
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Unknown_enum_value_s, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 33, __pyx_L1_error)

#line 33 "EnumBase"
    __Pyx_GOTREF(__pyx_t_1);

#line 33 "EnumBase"
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 33, __pyx_L1_error)

#line 33 "EnumBase"
    __Pyx_GOTREF(__pyx_t_4);

#line 33 "EnumBase"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 33 "EnumBase"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 33 "EnumBase"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 33 "EnumBase"
    __PYX_ERR(1, 33, __pyx_L1_error)

    /* "EnumBase":32
 *             if v == value:
 *                 return v
 *         if name is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 */

#line 32 "EnumBase"
  }

  /* "EnumBase":34
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)             # <<<<<<<<<<<<<<
 *         res.name = name
 *         setattr(cls, name, res)
 */

#line 34 "EnumBase"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyInt_Type)), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 34, __pyx_L1_error)

#line 34 "EnumBase"
  __Pyx_GOTREF(__pyx_t_1);

#line 34 "EnumBase"
  __pyx_t_7 = NULL;

#line 34 "EnumBase"
  __pyx_t_8 = 0;

#line 34 "EnumBase"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 34 "EnumBase"
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);

#line 34 "EnumBase"
    if (likely(__pyx_t_7)) {

#line 34 "EnumBase"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 34 "EnumBase"
      __Pyx_INCREF(__pyx_t_7);

#line 34 "EnumBase"
      __Pyx_INCREF(function);

#line 34 "EnumBase"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 34 "EnumBase"
      __pyx_t_8 = 1;

#line 34 "EnumBase"
    }

#line 34 "EnumBase"
  }

#line 34 "EnumBase"
  #if CYTHON_FAST_PYCALL

#line 34 "EnumBase"
  if (PyFunction_Check(__pyx_t_1)) {

#line 34 "EnumBase"
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_cls, __pyx_v_value};

#line 34 "EnumBase"
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)

#line 34 "EnumBase"
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 34 "EnumBase"
    __Pyx_GOTREF(__pyx_t_4);

#line 34 "EnumBase"
  } else

#line 34 "EnumBase"
  #endif

#line 34 "EnumBase"
  #if CYTHON_FAST_PYCCALL

#line 34 "EnumBase"
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {

#line 34 "EnumBase"
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_cls, __pyx_v_value};

#line 34 "EnumBase"
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)

#line 34 "EnumBase"
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 34 "EnumBase"
    __Pyx_GOTREF(__pyx_t_4);

#line 34 "EnumBase"
  } else

#line 34 "EnumBase"
  #endif

#line 34 "EnumBase"
  {

#line 34 "EnumBase"
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 34, __pyx_L1_error)

#line 34 "EnumBase"
    __Pyx_GOTREF(__pyx_t_9);

#line 34 "EnumBase"
    if (__pyx_t_7) {

#line 34 "EnumBase"
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;

#line 34 "EnumBase"
    }

#line 34 "EnumBase"
    __Pyx_INCREF(__pyx_v_cls);

#line 34 "EnumBase"
    __Pyx_GIVEREF(__pyx_v_cls);

#line 34 "EnumBase"
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_cls);

#line 34 "EnumBase"
    __Pyx_INCREF(__pyx_v_value);

#line 34 "EnumBase"
    __Pyx_GIVEREF(__pyx_v_value);

#line 34 "EnumBase"
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_value);

#line 34 "EnumBase"
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)

#line 34 "EnumBase"
    __Pyx_GOTREF(__pyx_t_4);

#line 34 "EnumBase"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 34 "EnumBase"
  }

#line 34 "EnumBase"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 34 "EnumBase"
  __pyx_v_res = __pyx_t_4;

#line 34 "EnumBase"
  __pyx_t_4 = 0;

  /* "EnumBase":35
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 *         res.name = name             # <<<<<<<<<<<<<<
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res
 */

#line 35 "EnumBase"
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_res, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(1, 35, __pyx_L1_error)

  /* "EnumBase":36
 *         res = int.__new__(cls, value)
 *         res.name = name
 *         setattr(cls, name, res)             # <<<<<<<<<<<<<<
 *         cls.__members__[name] = res
 *         return res
 */

#line 36 "EnumBase"
  __pyx_t_10 = PyObject_SetAttr(__pyx_v_cls, __pyx_v_name, __pyx_v_res); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(1, 36, __pyx_L1_error)

  /* "EnumBase":37
 *         res.name = name
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res             # <<<<<<<<<<<<<<
 *         return res
 *     def __repr__(self):
 */

#line 37 "EnumBase"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_members); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 37, __pyx_L1_error)

#line 37 "EnumBase"
  __Pyx_GOTREF(__pyx_t_4);

#line 37 "EnumBase"
  if (unlikely(PyObject_SetItem(__pyx_t_4, __pyx_v_name, __pyx_v_res) < 0)) __PYX_ERR(1, 37, __pyx_L1_error)

#line 37 "EnumBase"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumBase":38
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res
 *         return res             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 */

#line 38 "EnumBase"
  __Pyx_XDECREF(__pyx_r);

#line 38 "EnumBase"
  __Pyx_INCREF(__pyx_v_res);

#line 38 "EnumBase"
  __pyx_r = __pyx_v_res;

#line 38 "EnumBase"
  goto __pyx_L0;

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

#line 28 "EnumBase"


#line 28 "EnumBase"
  /* function exit code */

#line 28 "EnumBase"
  __pyx_L1_error:;

#line 28 "EnumBase"
  __Pyx_XDECREF(__pyx_t_1);

#line 28 "EnumBase"
  __Pyx_XDECREF(__pyx_t_4);

#line 28 "EnumBase"
  __Pyx_XDECREF(__pyx_t_7);

#line 28 "EnumBase"
  __Pyx_XDECREF(__pyx_t_9);

#line 28 "EnumBase"
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 28 "EnumBase"
  __pyx_r = NULL;

#line 28 "EnumBase"
  __pyx_L0:;

#line 28 "EnumBase"
  __Pyx_XDECREF(__pyx_v_v);

#line 28 "EnumBase"
  __Pyx_XDECREF(__pyx_v_res);

#line 28 "EnumBase"
  __Pyx_XGIVEREF(__pyx_r);

#line 28 "EnumBase"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 28 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 28 "EnumBase"
  return __pyx_r;

#line 28 "EnumBase"
}

/* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

#line 39 "EnumBase"


#line 39 "EnumBase"
/* Python wrapper */

#line 39 "EnumBase"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_3__repr__ = 
#line 39 "EnumBase"
{"__repr__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__, METH_O, 0};

#line 39 "EnumBase"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {

#line 39 "EnumBase"
  PyObject *__pyx_r = 0;

#line 39 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 39 "EnumBase"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 39 "EnumBase"
__pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

#line 39 "EnumBase"


#line 39 "EnumBase"
  /* function exit code */

#line 39 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 39 "EnumBase"
  return __pyx_r;

#line 39 "EnumBase"
}

#line 39 "EnumBase"


#line 39 "EnumBase"
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {

#line 39 "EnumBase"
  PyObject *__pyx_r = NULL;

#line 39 "EnumBase"
  __Pyx_TraceDeclarations

#line 39 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 39 "EnumBase"
  PyObject *__pyx_t_1 = NULL;

#line 39 "EnumBase"
  PyObject *__pyx_t_2 = NULL;

#line 39 "EnumBase"
  PyObject *__pyx_t_3 = NULL;

#line 39 "EnumBase"
  int __pyx_lineno = 0;

#line 39 "EnumBase"
  const char *__pyx_filename = NULL;

#line 39 "EnumBase"
  int __pyx_clineno = 0;

#line 39 "EnumBase"
  __Pyx_TraceFrameInit(__pyx_codeobj__43)

#line 39 "EnumBase"
  __Pyx_RefNannySetupContext("__repr__", 0);

#line 39 "EnumBase"
  __Pyx_TraceCall("__repr__", __pyx_f[1], 39, 0, __PYX_ERR(1, 39, __pyx_L1_error));

  /* "EnumBase":40
 *         return res
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)             # <<<<<<<<<<<<<<
 *     def __str__(self):
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 */

#line 40 "EnumBase"
  __Pyx_XDECREF(__pyx_r);

#line 40 "EnumBase"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)

#line 40 "EnumBase"
  __Pyx_GOTREF(__pyx_t_1);

#line 40 "EnumBase"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 40, __pyx_L1_error)

#line 40 "EnumBase"
  __Pyx_GOTREF(__pyx_t_2);

#line 40 "EnumBase"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 40 "EnumBase"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)

#line 40 "EnumBase"
  __Pyx_GOTREF(__pyx_t_1);

#line 40 "EnumBase"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 40, __pyx_L1_error)

#line 40 "EnumBase"
  __Pyx_GOTREF(__pyx_t_3);

#line 40 "EnumBase"
  __Pyx_GIVEREF(__pyx_t_2);

#line 40 "EnumBase"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);

#line 40 "EnumBase"
  __Pyx_GIVEREF(__pyx_t_1);

#line 40 "EnumBase"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);

#line 40 "EnumBase"
  __Pyx_INCREF(__pyx_v_self);

#line 40 "EnumBase"
  __Pyx_GIVEREF(__pyx_v_self);

#line 40 "EnumBase"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self);

#line 40 "EnumBase"
  __pyx_t_2 = 0;

#line 40 "EnumBase"
  __pyx_t_1 = 0;

#line 40 "EnumBase"
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s_d, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)

#line 40 "EnumBase"
  __Pyx_GOTREF(__pyx_t_1);

#line 40 "EnumBase"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 40 "EnumBase"
  __pyx_r = __pyx_t_1;

#line 40 "EnumBase"
  __pyx_t_1 = 0;

#line 40 "EnumBase"
  goto __pyx_L0;

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

#line 39 "EnumBase"


#line 39 "EnumBase"
  /* function exit code */

#line 39 "EnumBase"
  __pyx_L1_error:;

#line 39 "EnumBase"
  __Pyx_XDECREF(__pyx_t_1);

#line 39 "EnumBase"
  __Pyx_XDECREF(__pyx_t_2);

#line 39 "EnumBase"
  __Pyx_XDECREF(__pyx_t_3);

#line 39 "EnumBase"
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 39 "EnumBase"
  __pyx_r = NULL;

#line 39 "EnumBase"
  __pyx_L0:;

#line 39 "EnumBase"
  __Pyx_XGIVEREF(__pyx_r);

#line 39 "EnumBase"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 39 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 39 "EnumBase"
  return __pyx_r;

#line 39 "EnumBase"
}

/* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

#line 41 "EnumBase"


#line 41 "EnumBase"
/* Python wrapper */

#line 41 "EnumBase"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_5__str__ = 
#line 41 "EnumBase"
{"__str__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__, METH_O, 0};

#line 41 "EnumBase"
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {

#line 41 "EnumBase"
  PyObject *__pyx_r = 0;

#line 41 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 41 "EnumBase"
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = 
#line 41 "EnumBase"
__pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(__pyx_self, ((PyObject *)__pyx_v_self));

#line 41 "EnumBase"


#line 41 "EnumBase"
  /* function exit code */

#line 41 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 41 "EnumBase"
  return __pyx_r;

#line 41 "EnumBase"
}

#line 41 "EnumBase"


#line 41 "EnumBase"
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {

#line 41 "EnumBase"
  PyObject *__pyx_r = NULL;

#line 41 "EnumBase"
  __Pyx_TraceDeclarations

#line 41 "EnumBase"
  __Pyx_RefNannyDeclarations

#line 41 "EnumBase"
  PyObject *__pyx_t_1 = NULL;

#line 41 "EnumBase"
  PyObject *__pyx_t_2 = NULL;

#line 41 "EnumBase"
  PyObject *__pyx_t_3 = NULL;

#line 41 "EnumBase"
  int __pyx_lineno = 0;

#line 41 "EnumBase"
  const char *__pyx_filename = NULL;

#line 41 "EnumBase"
  int __pyx_clineno = 0;

#line 41 "EnumBase"
  __Pyx_TraceFrameInit(__pyx_codeobj__44)

#line 41 "EnumBase"
  __Pyx_RefNannySetupContext("__str__", 0);

#line 41 "EnumBase"
  __Pyx_TraceCall("__str__", __pyx_f[1], 41, 0, __PYX_ERR(1, 41, __pyx_L1_error));

  /* "EnumBase":42
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 *         return "%s.%s" % (self.__class__.__name__, self.name)             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */

#line 42 "EnumBase"
  __Pyx_XDECREF(__pyx_r);

#line 42 "EnumBase"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)

#line 42 "EnumBase"
  __Pyx_GOTREF(__pyx_t_1);

#line 42 "EnumBase"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 42, __pyx_L1_error)

#line 42 "EnumBase"
  __Pyx_GOTREF(__pyx_t_2);

#line 42 "EnumBase"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 42 "EnumBase"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)

#line 42 "EnumBase"
  __Pyx_GOTREF(__pyx_t_1);

#line 42 "EnumBase"
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 42, __pyx_L1_error)

#line 42 "EnumBase"
  __Pyx_GOTREF(__pyx_t_3);

#line 42 "EnumBase"
  __Pyx_GIVEREF(__pyx_t_2);

#line 42 "EnumBase"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);

#line 42 "EnumBase"
  __Pyx_GIVEREF(__pyx_t_1);

#line 42 "EnumBase"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);

#line 42 "EnumBase"
  __pyx_t_2 = 0;

#line 42 "EnumBase"
  __pyx_t_1 = 0;

#line 42 "EnumBase"
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)

#line 42 "EnumBase"
  __Pyx_GOTREF(__pyx_t_1);

#line 42 "EnumBase"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 42 "EnumBase"
  __pyx_r = __pyx_t_1;

#line 42 "EnumBase"
  __pyx_t_1 = 0;

#line 42 "EnumBase"
  goto __pyx_L0;

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

#line 41 "EnumBase"


#line 41 "EnumBase"
  /* function exit code */

#line 41 "EnumBase"
  __pyx_L1_error:;

#line 41 "EnumBase"
  __Pyx_XDECREF(__pyx_t_1);

#line 41 "EnumBase"
  __Pyx_XDECREF(__pyx_t_2);

#line 41 "EnumBase"
  __Pyx_XDECREF(__pyx_t_3);

#line 41 "EnumBase"
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 41 "EnumBase"
  __pyx_r = NULL;

#line 41 "EnumBase"
  __pyx_L0:;

#line 41 "EnumBase"
  __Pyx_XGIVEREF(__pyx_r);

#line 41 "EnumBase"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 41 "EnumBase"
  __Pyx_RefNannyFinishContext();

#line 41 "EnumBase"
  return __pyx_r;

#line 41 "EnumBase"
}

/* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta = 
#line 1 "(tree fragment)"
{"__pyx_unpickle___Pyx_EnumMeta", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta, METH_VARARGS|METH_KEYWORDS, 0};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle___Pyx_EnumMeta") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_TraceDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_TraceFrameInit(__pyx_codeobj__45)

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta", 0);

#line 1 "(tree fragment)"
  __Pyx_TraceCall("__pyx_unpickle___Pyx_EnumMeta", __pyx_f[1], 1, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd41d8cd, 0xe3b0c44, 0xda39a3e):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd41d8cd, 0xe3b0c44, 0xda39a3e) = ())" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__46, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd41d8cd, 0xe3b0c44, 0xda39a3e):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd41d8cd, 0xe3b0c44, 0xda39a3e) = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xd41d8cd, 0xe3b0c44, 0xda39a3e):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd41d8cd, 0xe3b0c44, 0xda39a3e) = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd41d8cd, 0xe3b0c44, 0xda39a3e):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd41d8cd, 0xe3b0c44, 0xda39a3e) = ())" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd41d8cd, 0xe3b0c44, 0xda39a3e) = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype___Pyx_EnumMeta), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd41d8cd, 0xe3b0c44, 0xda39a3e) = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_unpickle___Pyx_EnumMeta__set_state(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd41d8cd, 0xe3b0c44, 0xda39a3e) = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_unpickle___Pyx_EnumMeta__set_state(struct __pyx_obj___Pyx_EnumMeta *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_TraceDeclarations

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  int __pyx_t_1;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_2;

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  int __pyx_t_4;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta__set_state", 0);

#line 11 "(tree fragment)"
  __Pyx_TraceCall("__pyx_unpickle___Pyx_EnumMeta__set_state", __pyx_f[1], 11, 0, __PYX_ERR(1, 11, __pyx_L1_error));

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_3) {

#line 12 "(tree fragment)"
  } else {

#line 12 "(tree fragment)"
    __pyx_t_1 = __pyx_t_3;

#line 12 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 12 "(tree fragment)"
  __pyx_t_1 = __pyx_t_4;

#line 12 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 12 "(tree fragment)"
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */

#line 13 "(tree fragment)"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 13 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 13 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 13 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 13 "(tree fragment)"
      __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    }

#line 13 "(tree fragment)"
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 13 "(tree fragment)"
    __pyx_t_8 = NULL;

#line 13 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {

#line 13 "(tree fragment)"
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);

#line 13 "(tree fragment)"
      if (likely(__pyx_t_8)) {

#line 13 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);

#line 13 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_8);

#line 13 "(tree fragment)"
        __Pyx_INCREF(function);

#line 13 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_7, function);

#line 13 "(tree fragment)"
      }

#line 13 "(tree fragment)"
    }

#line 13 "(tree fragment)"
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 13 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 13 "(tree fragment)"
    if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 13 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 13 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_TraceReturn(__pyx_r, 0);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_PeakMatch __pyx_vtable_12ms_deisotope_2_c_18similarity_methods_PeakMatch;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *__pyx_freelist_12ms_deisotope_2_c_18similarity_methods_PeakMatch[10000];

#line 11 "(tree fragment)"
static int __pyx_freecount_12ms_deisotope_2_c_18similarity_methods_PeakMatch = 0;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_18similarity_methods_PeakMatch(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12ms_deisotope_2_c_18similarity_methods_PeakMatch > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch)))) {

#line 11 "(tree fragment)"
    o = (PyObject*)__pyx_freelist_12ms_deisotope_2_c_18similarity_methods_PeakMatch[--__pyx_freecount_12ms_deisotope_2_c_18similarity_methods_PeakMatch];

#line 11 "(tree fragment)"
    memset(o, 0, sizeof(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch));

#line 11 "(tree fragment)"
    (void) PyObject_INIT(o, t);

#line 11 "(tree fragment)"
    PyObject_GC_Track(o);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
    if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)o);

#line 11 "(tree fragment)"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_18similarity_methods_PeakMatch;

#line 11 "(tree fragment)"
  p->peak_a = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->peak_b = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_18similarity_methods_PeakMatch(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *p = (struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)o;

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  Py_CLEAR(p->peak_a);

#line 11 "(tree fragment)"
  Py_CLEAR(p->peak_b);

#line 11 "(tree fragment)"
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12ms_deisotope_2_c_18similarity_methods_PeakMatch < 10000) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch)))) {

#line 11 "(tree fragment)"
    __pyx_freelist_12ms_deisotope_2_c_18similarity_methods_PeakMatch[__pyx_freecount_12ms_deisotope_2_c_18similarity_methods_PeakMatch++] = ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)o);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_12ms_deisotope_2_c_18similarity_methods_PeakMatch(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *p = (struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)o;

#line 11 "(tree fragment)"
  if (p->peak_a) {

#line 11 "(tree fragment)"
    e = (*v)(((PyObject *)p->peak_a), a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->peak_b) {

#line 11 "(tree fragment)"
    e = (*v)(((PyObject *)p->peak_b), a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_12ms_deisotope_2_c_18similarity_methods_PeakMatch(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *p = (struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *)o;

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->peak_a);

#line 11 "(tree fragment)"
  p->peak_a = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->peak_b);

#line 11 "(tree fragment)"
  p->peak_b = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"
static PyObject *__pyx_sq_item_12ms_deisotope_2_c_18similarity_methods_PeakMatch(PyObject *o, Py_ssize_t i) {

#line 11 "(tree fragment)"
  PyObject *r;

#line 11 "(tree fragment)"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 11 "(tree fragment)"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 11 "(tree fragment)"
  Py_DECREF(x);

#line 11 "(tree fragment)"
  return r;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_peak_a(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_peak_a(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_a_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_peak_b(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_peak_b(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6peak_b_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_score(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_score(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5score_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_18similarity_methods_PeakMatch[] = {

#line 11 "(tree fragment)"
  {"_lt", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_3_lt, METH_O, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_2_lt},

#line 11 "(tree fragment)"
  {"_gt", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_5_gt, METH_O, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_4_gt},

#line 11 "(tree fragment)"
  {"_eq", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_7_eq, METH_O, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_6_eq},

#line 11 "(tree fragment)"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_17__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_16__reduce_cython__},

#line 11 "(tree fragment)"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_19__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_18__setstate_cython__},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_18similarity_methods_PeakMatch[] = {

#line 11 "(tree fragment)"
  {(char *)"peak_a", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_peak_a, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_peak_a, (char *)"peak_a: ms_peak_picker._c.peak_set.PeakBase", 0},

#line 11 "(tree fragment)"
  {(char *)"peak_b", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_peak_b, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_peak_b, (char *)"peak_b: ms_peak_picker._c.peak_set.PeakBase", 0},

#line 11 "(tree fragment)"
  {(char *)"score", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_score, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_score, (char *)"score: 'double'", 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PySequenceMethods __pyx_tp_as_sequence_PeakMatch = {

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_13__len__, /*sq_length*/

#line 11 "(tree fragment)"
  0, /*sq_concat*/

#line 11 "(tree fragment)"
  0, /*sq_repeat*/

#line 11 "(tree fragment)"
  __pyx_sq_item_12ms_deisotope_2_c_18similarity_methods_PeakMatch, /*sq_item*/

#line 11 "(tree fragment)"
  0, /*sq_slice*/

#line 11 "(tree fragment)"
  0, /*sq_ass_item*/

#line 11 "(tree fragment)"
  0, /*sq_ass_slice*/

#line 11 "(tree fragment)"
  0, /*sq_contains*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_concat*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_repeat*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMappingMethods __pyx_tp_as_mapping_PeakMatch = {

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_13__len__, /*mp_length*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_11__getitem__, /*mp_subscript*/

#line 11 "(tree fragment)"
  0, /*mp_ass_subscript*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_18similarity_methods_PeakMatch = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.similarity_methods.PeakMatch", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_18similarity_methods_PeakMatch, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_15__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_sequence_PeakMatch, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_mapping_PeakMatch, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "PeakMatch(peak_a, peak_b, score=0.0)", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_18similarity_methods_PeakMatch, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_18similarity_methods_PeakMatch, /*tp_clear*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_9__richcmp__, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_18similarity_methods_PeakMatch, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_12ms_deisotope_2_c_18similarity_methods_PeakMatch, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_18similarity_methods_9PeakMatch_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_18similarity_methods_PeakMatch, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment __pyx_vtable_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)o);

#line 11 "(tree fragment)"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment;

#line 11 "(tree fragment)"
  p->peak_pairs = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->peak_set_a = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->peak_set_b = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *p = (struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  Py_CLEAR(p->peak_pairs);

#line 11 "(tree fragment)"
  Py_CLEAR(p->peak_set_a);

#line 11 "(tree fragment)"
  Py_CLEAR(p->peak_set_b);

#line 11 "(tree fragment)"
  (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *p = (struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)o;

#line 11 "(tree fragment)"
  if (p->peak_pairs) {

#line 11 "(tree fragment)"
    e = (*v)(p->peak_pairs, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->peak_set_a) {

#line 11 "(tree fragment)"
    e = (*v)(p->peak_set_a, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->peak_set_b) {

#line 11 "(tree fragment)"
    e = (*v)(p->peak_set_b, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *p = (struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *)o;

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->peak_pairs);

#line 11 "(tree fragment)"
  p->peak_pairs = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->peak_set_a);

#line 11 "(tree fragment)"
  p->peak_set_a = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->peak_set_b);

#line 11 "(tree fragment)"
  p->peak_set_b = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_pairs(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_pairs(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_pairs_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_shift(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_shift(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5shift_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_score(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_score(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5score_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_set_a(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_set_a(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_a_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_set_b(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_set_b(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10peak_set_b_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_type(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9peak_type_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_norm_aa_cosine(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_norm_aa_cosine(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_aa_cosine_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_norm_bb_cosine(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_norm_bb_cosine(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14norm_bb_cosine_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_normalize(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_normalize(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9normalize_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_sqrt_transform(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_sqrt_transform(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_14sqrt_transform_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment[] = {

#line 11 "(tree fragment)"
  {"calculate_normalization", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_3calculate_normalization, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_2calculate_normalization},

#line 11 "(tree fragment)"
  {"_determine_peak_type", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_5_determine_peak_type, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_4_determine_peak_type},

#line 11 "(tree fragment)"
  {"align", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_7align, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_6align},

#line 11 "(tree fragment)"
  {"_calculate_score", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_9_calculate_score, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_8_calculate_score},

#line 11 "(tree fragment)"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_11__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_10__reduce_cython__},

#line 11 "(tree fragment)"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_13__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_12__setstate_cython__},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment[] = {

#line 11 "(tree fragment)"
  {(char *)"peak_pairs", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_pairs, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_pairs, (char *)"peak_pairs: list", 0},

#line 11 "(tree fragment)"
  {(char *)"shift", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_shift, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_shift, (char *)"shift: 'double'", 0},

#line 11 "(tree fragment)"
  {(char *)"score", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_score, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_score, (char *)"score: 'double'", 0},

#line 11 "(tree fragment)"
  {(char *)"peak_set_a", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_set_a, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_set_a, (char *)"peak_set_a: object", 0},

#line 11 "(tree fragment)"
  {(char *)"peak_set_b", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_set_b, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_set_b, (char *)"peak_set_b: object", 0},

#line 11 "(tree fragment)"
  {(char *)"peak_type", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_type, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_peak_type, (char *)"peak_type: 'PeakType'", 0},

#line 11 "(tree fragment)"
  {(char *)"norm_aa_cosine", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_norm_aa_cosine, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_norm_aa_cosine, (char *)"norm_aa_cosine: 'double'", 0},

#line 11 "(tree fragment)"
  {(char *)"norm_bb_cosine", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_norm_bb_cosine, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_norm_bb_cosine, (char *)"norm_bb_cosine: 'double'", 0},

#line 11 "(tree fragment)"
  {(char *)"normalize", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_normalize, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_normalize, (char *)"normalize: 'bool'", 0},

#line 11 "(tree fragment)"
  {(char *)"sqrt_transform", __pyx_getprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_sqrt_transform, __pyx_setprop_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_sqrt_transform, (char *)"sqrt_transform: 'bool'", 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.similarity_methods.SpectrumAlignment", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "SpectrumAlignment(peak_set_a, peak_set_b, double error_tolerance=2e-5, double shift=0.0, bool normalize=True, bool sqrt_transform=False)", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new___Pyx_EnumMeta(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  PyObject *o = (&PyType_Type)->tp_new(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc___Pyx_EnumMeta(PyObject *o) {

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  PyObject_GC_Track(o);

#line 11 "(tree fragment)"
  (&PyType_Type)->tp_dealloc(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse___Pyx_EnumMeta(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  if (!(&PyType_Type)->tp_traverse); else { e = (&PyType_Type)->tp_traverse(o,v,a); if (e) return e; }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear___Pyx_EnumMeta(PyObject *o) {

#line 11 "(tree fragment)"
  if (!(&PyType_Type)->tp_clear); else (&PyType_Type)->tp_clear(o);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"
static PyObject *__pyx_sq_item___Pyx_EnumMeta(PyObject *o, Py_ssize_t i) {

#line 11 "(tree fragment)"
  PyObject *r;

#line 11 "(tree fragment)"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 11 "(tree fragment)"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 11 "(tree fragment)"
  Py_DECREF(x);

#line 11 "(tree fragment)"
  return r;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods___Pyx_EnumMeta[] = {

#line 11 "(tree fragment)"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__, METH_O, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PySequenceMethods __pyx_tp_as_sequence___Pyx_EnumMeta = {

#line 11 "(tree fragment)"
  0, /*sq_length*/

#line 11 "(tree fragment)"
  0, /*sq_concat*/

#line 11 "(tree fragment)"
  0, /*sq_repeat*/

#line 11 "(tree fragment)"
  __pyx_sq_item___Pyx_EnumMeta, /*sq_item*/

#line 11 "(tree fragment)"
  0, /*sq_slice*/

#line 11 "(tree fragment)"
  0, /*sq_ass_item*/

#line 11 "(tree fragment)"
  0, /*sq_ass_slice*/

#line 11 "(tree fragment)"
  0, /*sq_contains*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_concat*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_repeat*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMappingMethods __pyx_tp_as_mapping___Pyx_EnumMeta = {

#line 11 "(tree fragment)"
  0, /*mp_length*/

#line 11 "(tree fragment)"
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__, /*mp_subscript*/

#line 11 "(tree fragment)"
  0, /*mp_ass_subscript*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __Pyx_EnumMeta = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.similarity_methods.__Pyx_EnumMeta", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj___Pyx_EnumMeta), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc___Pyx_EnumMeta, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_sequence___Pyx_EnumMeta, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_mapping___Pyx_EnumMeta, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  0, /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse___Pyx_EnumMeta, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear___Pyx_EnumMeta, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods___Pyx_EnumMeta, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new___Pyx_EnumMeta, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods[] = {

#line 11 "(tree fragment)"
  {"convolved_dot_product", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_18similarity_methods_5convolved_dot_product, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_18similarity_methods_4convolved_dot_product},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
#if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
#if CYTHON_PEP489_MULTI_PHASE_INIT

#line 11 "(tree fragment)"
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/

#line 11 "(tree fragment)"
static int __pyx_pymod_exec_similarity_methods(PyObject* module); /*proto*/

#line 11 "(tree fragment)"
static PyModuleDef_Slot __pyx_moduledef_slots[] = {

#line 11 "(tree fragment)"
  {Py_mod_create, (void*)__pyx_pymod_create},

#line 11 "(tree fragment)"
  {Py_mod_exec, (void*)__pyx_pymod_exec_similarity_methods},

#line 11 "(tree fragment)"
  {0, NULL}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
#endif

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyModuleDef __pyx_moduledef = {

#line 11 "(tree fragment)"
    PyModuleDef_HEAD_INIT,

#line 11 "(tree fragment)"
    "similarity_methods",

#line 11 "(tree fragment)"
    0, /* m_doc */

#line 11 "(tree fragment)"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 11 "(tree fragment)"
    0, /* m_size */

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
    -1, /* m_size */

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
    __pyx_methods /* m_methods */,

#line 11 "(tree fragment)"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 11 "(tree fragment)"
    __pyx_moduledef_slots, /* m_slots */

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
    NULL, /* m_reload */

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
    NULL, /* m_traverse */

#line 11 "(tree fragment)"
    NULL, /* m_clear */

#line 11 "(tree fragment)"
    NULL /* m_free */

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
#endif

#line 11 "(tree fragment)"
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_Cannot_handle_objects_of_type_s, __pyx_k_Cannot_handle_objects_of_type_s, sizeof(__pyx_k_Cannot_handle_objects_of_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_Deconvoluted, __pyx_k_Deconvoluted, sizeof(__pyx_k_Deconvoluted), 0, 0, 1, 1},
  {&__pyx_n_s_DeconvolutedPeakSet, __pyx_k_DeconvolutedPeakSet, sizeof(__pyx_k_DeconvolutedPeakSet), 0, 0, 1, 1},
  {&__pyx_n_s_EnumBase, __pyx_k_EnumBase, sizeof(__pyx_k_EnumBase), 0, 0, 1, 1},
  {&__pyx_n_s_EnumType, __pyx_k_EnumType, sizeof(__pyx_k_EnumType), 0, 0, 1, 1},
  {&__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_k_Expected_at_least_d_argument_s_g, sizeof(__pyx_k_Expected_at_least_d_argument_s_g), 0, 0, 1, 0},
  {&__pyx_n_s_Fitted, __pyx_k_Fitted, sizeof(__pyx_k_Fitted), 0, 0, 1, 1},
  {&__pyx_kp_s_Function_call_with_ambiguous_arg, __pyx_k_Function_call_with_ambiguous_arg, sizeof(__pyx_k_Function_call_with_ambiguous_arg), 0, 0, 1, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_IntEnum, __pyx_k_IntEnum, sizeof(__pyx_k_IntEnum), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_No_matching_signature_found, __pyx_k_No_matching_signature_found, sizeof(__pyx_k_No_matching_signature_found), 0, 0, 1, 0},
  {&__pyx_n_s_OrderedDict, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
  {&__pyx_n_s_PeakIndex, __pyx_k_PeakIndex, sizeof(__pyx_k_PeakIndex), 0, 0, 1, 1},
  {&__pyx_n_s_PeakMatch, __pyx_k_PeakMatch, sizeof(__pyx_k_PeakMatch), 0, 0, 1, 1},
  {&__pyx_n_s_PeakSet, __pyx_k_PeakSet, sizeof(__pyx_k_PeakSet), 0, 0, 1, 1},
  {&__pyx_n_s_PeakType, __pyx_k_PeakType, sizeof(__pyx_k_PeakType), 0, 0, 1, 1},
  {&__pyx_kp_s_Peak_set_types_must_match_got_s, __pyx_k_Peak_set_types_must_match_got_s, sizeof(__pyx_k_Peak_set_types_must_match_got_s), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase, __pyx_k_Pyx_EnumBase, sizeof(__pyx_k_Pyx_EnumBase), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___new, __pyx_k_Pyx_EnumBase___new, sizeof(__pyx_k_Pyx_EnumBase___new), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___repr, __pyx_k_Pyx_EnumBase___repr, sizeof(__pyx_k_Pyx_EnumBase___repr), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___str, __pyx_k_Pyx_EnumBase___str, sizeof(__pyx_k_Pyx_EnumBase___str), 0, 0, 1, 1},
  {&__pyx_n_s_SimilarityMetrics, __pyx_k_SimilarityMetrics, sizeof(__pyx_k_SimilarityMetrics), 0, 0, 1, 1},
  {&__pyx_n_s_SpectrumAlignment, __pyx_k_SpectrumAlignment, sizeof(__pyx_k_SpectrumAlignment), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unknown_enum_value_s, __pyx_k_Unknown_enum_value_s, sizeof(__pyx_k_Unknown_enum_value_s), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 0},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_kp_s__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 0},
  {&__pyx_n_s_align, __pyx_k_align, sizeof(__pyx_k_align), 0, 0, 1, 1},
  {&__pyx_n_s_align_peaks, __pyx_k_align_peaks, sizeof(__pyx_k_align_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_bin_peaks, __pyx_k_bin_peaks, sizeof(__pyx_k_bin_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_calculate_normalization, __pyx_k_calculate_normalization, sizeof(__pyx_k_calculate_normalization), 0, 0, 1, 1},
  {&__pyx_n_s_calculate_score, __pyx_k_calculate_score, sizeof(__pyx_k_calculate_score), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_convolved_dot_product, __pyx_k_convolved_dot_product, sizeof(__pyx_k_convolved_dot_product), 0, 0, 1, 1},
  {&__pyx_n_s_dct, __pyx_k_dct, sizeof(__pyx_k_dct), 0, 0, 1, 1},
  {&__pyx_n_s_defaults, __pyx_k_defaults, sizeof(__pyx_k_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_determine_peak_type, __pyx_k_determine_peak_type, sizeof(__pyx_k_determine_peak_type), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dot_product, __pyx_k_dot_product, sizeof(__pyx_k_dot_product), 0, 0, 1, 1},
  {&__pyx_n_s_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
  {&__pyx_n_s_error_tolerance, __pyx_k_error_tolerance, sizeof(__pyx_k_error_tolerance), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_members, __pyx_k_members, sizeof(__pyx_k_members), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_kp_s_ms_deisotope__c_peak_set, __pyx_k_ms_deisotope__c_peak_set, sizeof(__pyx_k_ms_deisotope__c_peak_set), 0, 0, 1, 0},
  {&__pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_k_ms_deisotope__c_similarity_metho, sizeof(__pyx_k_ms_deisotope__c_similarity_metho), 0, 0, 1, 1},
  {&__pyx_kp_s_ms_peak_picker__c_peak_index, __pyx_k_ms_peak_picker__c_peak_index, sizeof(__pyx_k_ms_peak_picker__c_peak_index), 0, 0, 1, 0},
  {&__pyx_kp_s_ms_peak_picker__c_peak_set, __pyx_k_ms_peak_picker__c_peak_set, sizeof(__pyx_k_ms_peak_picker__c_peak_set), 0, 0, 1, 0},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_norm_a, __pyx_k_norm_a, sizeof(__pyx_k_norm_a), 0, 0, 1, 1},
  {&__pyx_n_s_norm_b, __pyx_k_norm_b, sizeof(__pyx_k_norm_b), 0, 0, 1, 1},
  {&__pyx_n_s_normalize, __pyx_k_normalize, sizeof(__pyx_k_normalize), 0, 0, 1, 1},
  {&__pyx_n_s_normalized_dot_product, __pyx_k_normalized_dot_product, sizeof(__pyx_k_normalized_dot_product), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_parents, __pyx_k_parents, sizeof(__pyx_k_parents), 0, 0, 1, 1},
  {&__pyx_n_s_peak_a, __pyx_k_peak_a, sizeof(__pyx_k_peak_a), 0, 0, 1, 1},
  {&__pyx_n_s_peak_b, __pyx_k_peak_b, sizeof(__pyx_k_peak_b), 0, 0, 1, 1},
  {&__pyx_n_s_peak_pairs, __pyx_k_peak_pairs, sizeof(__pyx_k_peak_pairs), 0, 0, 1, 1},
  {&__pyx_n_s_peak_set_a, __pyx_k_peak_set_a, sizeof(__pyx_k_peak_set_a), 0, 0, 1, 1},
  {&__pyx_n_s_peak_set_b, __pyx_k_peak_set_b, sizeof(__pyx_k_peak_set_b), 0, 0, 1, 1},
  {&__pyx_n_s_peak_set_similarity, __pyx_k_peak_set_similarity, sizeof(__pyx_k_peak_set_similarity), 0, 0, 1, 1},
  {&__pyx_n_s_peak_type, __pyx_k_peak_type, sizeof(__pyx_k_peak_type), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_ppm_peak_set_similarity, __pyx_k_ppm_peak_set_similarity, sizeof(__pyx_k_ppm_peak_set_similarity), 0, 0, 1, 1},
  {&__pyx_n_s_precision, __pyx_k_precision, sizeof(__pyx_k_precision), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0align_peaks, __pyx_k_pyx_fuse_0align_peaks, sizeof(__pyx_k_pyx_fuse_0align_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0bin_peaks, __pyx_k_pyx_fuse_0bin_peaks, sizeof(__pyx_k_pyx_fuse_0bin_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0peak_set_similarity, __pyx_k_pyx_fuse_0peak_set_similarity, sizeof(__pyx_k_pyx_fuse_0peak_set_similarity), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0ppm_peak_set_similar, __pyx_k_pyx_fuse_0ppm_peak_set_similar, sizeof(__pyx_k_pyx_fuse_0ppm_peak_set_similar), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1align_peaks, __pyx_k_pyx_fuse_1align_peaks, sizeof(__pyx_k_pyx_fuse_1align_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1bin_peaks, __pyx_k_pyx_fuse_1bin_peaks, sizeof(__pyx_k_pyx_fuse_1bin_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1peak_set_similarity, __pyx_k_pyx_fuse_1peak_set_similarity, sizeof(__pyx_k_pyx_fuse_1peak_set_similarity), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1ppm_peak_set_similar, __pyx_k_pyx_fuse_1ppm_peak_set_similar, sizeof(__pyx_k_pyx_fuse_1ppm_peak_set_similar), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_2align_peaks, __pyx_k_pyx_fuse_2align_peaks, sizeof(__pyx_k_pyx_fuse_2align_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_2bin_peaks, __pyx_k_pyx_fuse_2bin_peaks, sizeof(__pyx_k_pyx_fuse_2bin_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_2peak_set_similarity, __pyx_k_pyx_fuse_2peak_set_similarity, sizeof(__pyx_k_pyx_fuse_2peak_set_similarity), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_2ppm_peak_set_similar, __pyx_k_pyx_fuse_2ppm_peak_set_similar, sizeof(__pyx_k_pyx_fuse_2ppm_peak_set_similar), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_3align_peaks, __pyx_k_pyx_fuse_3align_peaks, sizeof(__pyx_k_pyx_fuse_3align_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_3bin_peaks, __pyx_k_pyx_fuse_3bin_peaks, sizeof(__pyx_k_pyx_fuse_3bin_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_3peak_set_similarity, __pyx_k_pyx_fuse_3peak_set_similarity, sizeof(__pyx_k_pyx_fuse_3peak_set_similarity), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_3ppm_peak_set_similar, __pyx_k_pyx_fuse_3ppm_peak_set_similar, sizeof(__pyx_k_pyx_fuse_3ppm_peak_set_similar), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_PeakMatch, __pyx_k_pyx_unpickle_PeakMatch, sizeof(__pyx_k_pyx_unpickle_PeakMatch), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_SpectrumAlignment, __pyx_k_pyx_unpickle_SpectrumAlignment, sizeof(__pyx_k_pyx_unpickle_SpectrumAlignment), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle___Pyx_EnumMeta, __pyx_k_pyx_unpickle___Pyx_EnumMeta, sizeof(__pyx_k_pyx_unpickle___Pyx_EnumMeta), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_kp_s_s_s, __pyx_k_s_s, sizeof(__pyx_k_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_d, __pyx_k_s_s_d, sizeof(__pyx_k_s_s_d), 0, 0, 1, 0},
  {&__pyx_n_s_score, __pyx_k_score, sizeof(__pyx_k_score), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self___class_____name___self_pe, __pyx_k_self___class_____name___self_pe, sizeof(__pyx_k_self___class_____name___self_pe), 0, 0, 1, 0},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shift, __pyx_k_shift, sizeof(__pyx_k_shift), 0, 0, 1, 1},
  {&__pyx_n_s_signatures, __pyx_k_signatures, sizeof(__pyx_k_signatures), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_sqrt_transform, __pyx_k_sqrt_transform, sizeof(__pyx_k_sqrt_transform), 0, 0, 1, 1},
  {&__pyx_kp_s_src_ms_deisotope__c_similarity_m, __pyx_k_src_ms_deisotope__c_similarity_m, sizeof(__pyx_k_src_ms_deisotope__c_similarity_m), 0, 0, 1, 0},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 94, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_builtin___import__ = __Pyx_GetBuiltinName(__pyx_n_s_import); if (!__pyx_builtin___import__) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 204, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 944, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(1, 33, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "ms_deisotope/_c/similarity_methods.pyx":127
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cpdef double peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0) except -1:             # <<<<<<<<<<<<<<
 *     """Computes the normalized dot product, also called cosine similarity between
 *     two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).
 */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_No_matching_signature_found); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_tuple__5);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_tuple__5);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_Function_call_with_ambiguous_arg); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_tuple__6);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x797b24e, 0x429ecd1, 0x7e0f493):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x797b24e, 0x429ecd1, 0x7e0f493) = (peak_a, peak_b, score))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_tuple__37 = PyTuple_Pack(3, __pyx_int_127382094, __pyx_int_69856465, __pyx_int_132183187); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__37);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__37);

#line 4 "(tree fragment)"
  __pyx_tuple__39 = PyTuple_Pack(3, __pyx_int_22575453, __pyx_int_257365623, __pyx_int_25485575); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__39);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(2, 944, __pyx_L1_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__40);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(2, 950, __pyx_L1_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__41);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd41d8cd, 0xe3b0c44, 0xda39a3e):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd41d8cd, 0xe3b0c44, 0xda39a3e) = ())" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_tuple__46 = PyTuple_Pack(3, __pyx_int_222419149, __pyx_int_238750788, __pyx_int_228825662); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__46);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "ms_deisotope/_c/similarity_methods.pyx":127
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cpdef double peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0) except -1:             # <<<<<<<<<<<<<<
 *     """Computes the normalized dot product, also called cosine similarity between
 *     two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).
 */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_tuple__47 = PyTuple_Pack(3, __pyx_n_s_peak_set_a, __pyx_n_s_peak_set_b, __pyx_n_s_precision); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_tuple__47);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_tuple__47);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_ms_deisotope__c_similarity_m, __pyx_n_s_pyx_fuse_0peak_set_similarity, 127, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 127, __pyx_L1_error)

  /* "ms_deisotope/_c/similarity_methods.pyx":258
 * 
 * @cython.boundscheck(False)
 * cpdef tuple bin_peaks(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double hi, scaler
 */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_tuple__49 = PyTuple_Pack(3, __pyx_n_s_peak_set_a, __pyx_n_s_peak_set_b, __pyx_n_s_precision); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_tuple__49);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_tuple__49);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_ms_deisotope__c_similarity_m, __pyx_n_s_pyx_fuse_0bin_peaks, 258, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 258, __pyx_L1_error)

  /* "ms_deisotope/_c/similarity_methods.pyx":614
 * 
 * 
 * cpdef double ppm_peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double ab, aa, bb
 */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_tuple__51 = PyTuple_Pack(3, __pyx_n_s_peak_set_a, __pyx_n_s_peak_set_b, __pyx_n_s_error_tolerance); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_tuple__51);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_tuple__51);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__51, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_ms_deisotope__c_similarity_m, __pyx_n_s_pyx_fuse_0ppm_peak_set_similar, 614, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 614, __pyx_L1_error)

  /* "ms_deisotope/_c/similarity_methods.pyx":635
 * 
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,             # <<<<<<<<<<<<<<
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_tuple__53 = PyTuple_Pack(5, __pyx_n_s_peak_set_a, __pyx_n_s_peak_set_b, __pyx_n_s_error_tolerance, __pyx_n_s_shift, __pyx_n_s_normalize); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_tuple__53);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_tuple__53);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(5, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_ms_deisotope__c_similarity_m, __pyx_n_s_pyx_fuse_0align_peaks, 635, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 635, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_PeakMatch(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_tuple__55 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__55);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__55);

#line 1 "(tree fragment)"
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PeakMatch, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __pyx_tuple__56 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__56);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__56);

#line 1 "(tree fragment)"
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__56, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_SpectrumAlignment, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

#line 28 "EnumBase"
  __pyx_tuple__57 = PyTuple_Pack(5, __pyx_n_s_cls, __pyx_n_s_value, __pyx_n_s_name, __pyx_n_s_v, __pyx_n_s_res); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(1, 28, __pyx_L1_error)

#line 28 "EnumBase"
  __Pyx_GOTREF(__pyx_tuple__57);

#line 28 "EnumBase"
  __Pyx_GIVEREF(__pyx_tuple__57);

#line 28 "EnumBase"
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_new, 28, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(1, 28, __pyx_L1_error)

#line 28 "EnumBase"
  __pyx_tuple__58 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(1, 28, __pyx_L1_error)

#line 28 "EnumBase"
  __Pyx_GOTREF(__pyx_tuple__58);

#line 28 "EnumBase"
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

#line 39 "EnumBase"
  __pyx_tuple__59 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(1, 39, __pyx_L1_error)

#line 39 "EnumBase"
  __Pyx_GOTREF(__pyx_tuple__59);

#line 39 "EnumBase"
  __Pyx_GIVEREF(__pyx_tuple__59);

#line 39 "EnumBase"
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_repr, 39, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(1, 39, __pyx_L1_error)

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

#line 41 "EnumBase"
  __pyx_tuple__60 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(1, 41, __pyx_L1_error)

#line 41 "EnumBase"
  __Pyx_GOTREF(__pyx_tuple__60);

#line 41 "EnumBase"
  __Pyx_GIVEREF(__pyx_tuple__60);

#line 41 "EnumBase"
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_str, 41, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(1, 41, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_tuple__61 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__61);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__61);

#line 1 "(tree fragment)"
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return 0;

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return -1;

#line 1 "(tree fragment)"
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* InitThreads.init */
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0
PyEval_InitThreads();
#endif

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_22575453 = PyInt_FromLong(22575453L); if (unlikely(!__pyx_int_22575453)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_25485575 = PyInt_FromLong(25485575L); if (unlikely(!__pyx_int_25485575)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_69856465 = PyInt_FromLong(69856465L); if (unlikely(!__pyx_int_69856465)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_127382094 = PyInt_FromLong(127382094L); if (unlikely(!__pyx_int_127382094)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_132183187 = PyInt_FromLong(132183187L); if (unlikely(!__pyx_int_132183187)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_228825662 = PyInt_FromLong(228825662L); if (unlikely(!__pyx_int_228825662)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_238750788 = PyInt_FromLong(238750788L); if (unlikely(!__pyx_int_238750788)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_257365623 = PyInt_FromLong(257365623L); if (unlikely(!__pyx_int_257365623)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros = Py_None; Py_INCREF(Py_None);
  __Pyx_OrderedDict = Py_None; Py_INCREF(Py_None);
  __Pyx_EnumBase = Py_None; Py_INCREF(Py_None);
  __Pyx_globals = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_12ms_deisotope_2_c_18similarity_methods_PeakMatch = &__pyx_vtable_12ms_deisotope_2_c_18similarity_methods_PeakMatch;
  __pyx_vtable_12ms_deisotope_2_c_18similarity_methods_PeakMatch._lt = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__lt;
  __pyx_vtable_12ms_deisotope_2_c_18similarity_methods_PeakMatch._gt = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__gt;
  __pyx_vtable_12ms_deisotope_2_c_18similarity_methods_PeakMatch._eq = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__eq;
  __pyx_vtable_12ms_deisotope_2_c_18similarity_methods_PeakMatch._create = (struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_PeakMatch *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__create *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_18similarity_methods_9PeakMatch__create;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_18similarity_methods_PeakMatch) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_18similarity_methods_PeakMatch.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_18similarity_methods_PeakMatch.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_18similarity_methods_PeakMatch.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_18similarity_methods_PeakMatch.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_18similarity_methods_PeakMatch.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_18similarity_methods_PeakMatch) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PeakMatch, (PyObject *)&__pyx_type_12ms_deisotope_2_c_18similarity_methods_PeakMatch) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_18similarity_methods_PeakMatch) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_18similarity_methods_PeakMatch = &__pyx_type_12ms_deisotope_2_c_18similarity_methods_PeakMatch;
  __pyx_vtabptr_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment = &__pyx_vtable_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment;
  __pyx_vtable_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment.calculate_normalization = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_calculate_normalization;
  __pyx_vtable_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment._determine_peak_type = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment__determine_peak_type;
  __pyx_vtable_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment.align = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_align *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment_align;
  __pyx_vtable_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment._calculate_score = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_18similarity_methods_17SpectrumAlignment__calculate_score;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment) < 0) __PYX_ERR(0, 353, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment) < 0) __PYX_ERR(0, 353, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SpectrumAlignment, (PyObject *)&__pyx_type_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment) < 0) __PYX_ERR(0, 353, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment) < 0) __PYX_ERR(0, 353, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment = &__pyx_type_12ms_deisotope_2_c_18similarity_methods_SpectrumAlignment;
  __Pyx_EnumMeta.tp_base = (&PyType_Type);
  if (PyType_Ready(&__Pyx_EnumMeta) < 0) __PYX_ERR(1, 15, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __Pyx_EnumMeta.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__Pyx_EnumMeta.tp_dictoffset && __Pyx_EnumMeta.tp_getattro == PyObject_GenericGetAttr)) {
    __Pyx_EnumMeta.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__Pyx_EnumMeta) < 0) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_ptype___Pyx_EnumMeta = &__Pyx_EnumMeta;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(2, 770, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(2, 772, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(2, 774, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(2, 776, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(2, 778, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(2, 780, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(2, 782, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(2, 784, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(2, 786, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(2, 788, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakBase", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase) __PYX_ERR(6, 2, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase)) __PYX_ERR(6, 2, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "FittedPeak", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak)) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSet", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet) __PYX_ERR(6, 29, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet)) __PYX_ERR(6, 29, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSetIndexed", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) __PYX_ERR(6, 83, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed)) __PYX_ERR(6, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_index"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_index", "PeakIndex", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex) __PYX_ERR(7, 5, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex)) __PYX_ERR(7, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_11composition_PyComposition = __Pyx_ImportType(__pyx_t_1, "brainpy._c.composition", "PyComposition", sizeof(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_11composition_PyComposition) __PYX_ERR(8, 123, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition = (struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_11composition_PyComposition->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition)) __PYX_ERR(8, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_distribution"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = __Pyx_ImportType(__pyx_t_1, "brainpy._c.isotopic_distribution", "TheoreticalPeak", sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak) __PYX_ERR(9, 63, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = (struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak)) __PYX_ERR(9, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "Averagine", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine) __PYX_ERR(10, 12, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine)) __PYX_ERR(10, 12, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "TheoreticalIsotopicPattern", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern) __PYX_ERR(10, 23, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern)) __PYX_ERR(10, 23, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "AveragineCache", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache) __PYX_ERR(10, 79, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache)) __PYX_ERR(10, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.deconvoluter_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "DeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase) __PYX_ERR(11, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase)) __PYX_ERR(11, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "AveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase) __PYX_ERR(11, 57, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase)) __PYX_ERR(11, 57, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "MultiAveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase) __PYX_ERR(11, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase)) __PYX_ERR(11, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "ChargeIterator", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator) __PYX_ERR(11, 91, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator)) __PYX_ERR(11, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitRecord", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord) __PYX_ERR(12, 16, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord)) __PYX_ERR(12, 16, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FitSelectorBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase) __PYX_ERR(12, 38, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase)) __PYX_ERR(12, 38, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase) __PYX_ERR(12, 50, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase)) __PYX_ERR(12, 50, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "LeastSquaresFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter) __PYX_ERR(12, 62, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter)) __PYX_ERR(12, 62, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "GTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter) __PYX_ERR(12, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter)) __PYX_ERR(12, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledGTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter) __PYX_ERR(12, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter)) __PYX_ERR(12, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "MSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter) __PYX_ERR(12, 72, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter)) __PYX_ERR(12, 72, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "PenalizedMSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter) __PYX_ERR(12, 78, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter)) __PYX_ERR(12, 78, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FunctionScorer", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer) __PYX_ERR(12, 84, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer)) __PYX_ERR(12, 84, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "InterferenceDetection", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection) __PYX_ERR(12, 89, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection)) __PYX_ERR(12, 89, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DistinctPatternFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter) __PYX_ERR(12, 95, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter)) __PYX_ERR(12, 95, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledPenalizedMSDeconvFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter) __PYX_ERR(12, 102, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter)) __PYX_ERR(12, 102, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DotProductFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter) __PYX_ERR(12, 112, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter)) __PYX_ERR(12, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "_Index", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index) __PYX_ERR(13, 5, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index)) __PYX_ERR(13, 5, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "EnvelopePair", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair) __PYX_ERR(13, 14, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair)) __PYX_ERR(13, 14, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "Envelope", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope) __PYX_ERR(13, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope)) __PYX_ERR(13, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak) __PYX_ERR(13, 40, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak)) __PYX_ERR(13, 40, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "IonMobilityDeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak) __PYX_ERR(13, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak)) __PYX_ERR(13, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSolution", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution) __PYX_ERR(13, 71, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution)) __PYX_ERR(13, 71, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet) __PYX_ERR(13, 75, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet)) __PYX_ERR(13, 75, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSetIndexed", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed) __PYX_ERR(13, 104, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed)) __PYX_ERR(13, 104, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "_CPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet) __PYX_ERR(13, 161, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet)) __PYX_ERR(13, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "nist_mass", (void **)&__pyx_vp_7brainpy_2_c_11composition_nist_mass, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "PROTON", (void **)&__pyx_vp_7brainpy_2_c_11composition_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_ElementTable", (void **)&__pyx_vp_7brainpy_2_c_11composition__ElementTable, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_constants"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "DEFAULT_ISOTOPIC_CONSTANTS_SIZE", (void **)&__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "PROTON", (void **)&__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INFINITY", (void **)&__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initsimilarity_methods(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initsimilarity_methods(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_similarity_methods(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_similarity_methods(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_similarity_methods(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  __Pyx_TraceDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'similarity_methods' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_similarity_methods(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("similarity_methods", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_deisotope___c__similarity_methods) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_deisotope._c.similarity_methods")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_deisotope._c.similarity_methods", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __Pyx_TraceCall("__Pyx_PyMODINIT_FUNC PyInit_similarity_methods(void)", __pyx_f[0], 1, 0, __PYX_ERR(0, 1, __pyx_L1_error));

  /* "ms_deisotope/_c/similarity_methods.pyx":20
 * 
 * cimport numpy as cnp
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * cdef object zeros = np.zeros
 */

#line 20 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 20 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":22
 * import numpy as np
 * 
 * cdef object zeros = np.zeros             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 22 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 22 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 22 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 22 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_XGOTREF(__pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros);

#line 22 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF_SET(__pyx_v_12ms_deisotope_2_c_18similarity_methods_zeros, __pyx_t_2);

#line 22 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 22 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":127
 * @cython.boundscheck(False)
 * @cython.cdivision(True)
 * cpdef double peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0) except -1:             # <<<<<<<<<<<<<<
 *     """Computes the normalized dot product, also called cosine similarity between
 *     two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).
 */

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyInt_From_long(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyInt_From_long(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k_ = __pyx_t_2;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__7 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__7 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__8 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__8 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__9 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__9 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__10 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__10 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_13__pyx_fuse_0peak_set_similarity, 0, __pyx_n_s_pyx_fuse_0peak_set_similarity, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_PeakSet, __pyx_t_3) < 0) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_15__pyx_fuse_1peak_set_similarity, 0, __pyx_n_s_pyx_fuse_1peak_set_similarity, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_PeakIndex, __pyx_t_3) < 0) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_17__pyx_fuse_2peak_set_similarity, 0, __pyx_n_s_pyx_fuse_2peak_set_similarity, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_DeconvolutedPeakSet, __pyx_t_3) < 0) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_19__pyx_fuse_3peak_set_similarity, 0, __pyx_n_s_pyx_fuse_3peak_set_similarity, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_object, __pyx_t_3) < 0) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_1peak_set_similarity, 0, __pyx_n_s_peak_set_similarity, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_1);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  ((__pyx_FusedFunctionObject *) __pyx_t_3)->__signatures__ = __pyx_t_2;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_peak_set_similarity, __pyx_t_3) < 0) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 127 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":258
 * 
 * @cython.boundscheck(False)
 * cpdef tuple bin_peaks(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double hi, scaler
 */

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyInt_From_long(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyInt_From_long(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__11 = __pyx_t_1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__12 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__12 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__13 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__13 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__14 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__14 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__15 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__15 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_23__pyx_fuse_0bin_peaks, 0, __pyx_n_s_pyx_fuse_0bin_peaks, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_3);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_PeakSet, __pyx_t_2) < 0) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_25__pyx_fuse_1bin_peaks, 0, __pyx_n_s_pyx_fuse_1bin_peaks, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_3);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_PeakIndex, __pyx_t_2) < 0) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_27__pyx_fuse_2bin_peaks, 0, __pyx_n_s_pyx_fuse_2bin_peaks, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_3);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_DeconvolutedPeakSet, __pyx_t_2) < 0) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_29__pyx_fuse_3bin_peaks, 0, __pyx_n_s_pyx_fuse_3bin_peaks, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_3);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_object, __pyx_t_2) < 0) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_3bin_peaks, 0, __pyx_n_s_bin_peaks, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_3);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  ((__pyx_FusedFunctionObject *) __pyx_t_2)->__signatures__ = __pyx_t_1;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bin_peaks, __pyx_t_2) < 0) __PYX_ERR(0, 258, __pyx_L1_error)

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 258 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":528
 * 
 * 
 * cpdef double convolved_dot_product(list peak_pairs, PeakType peak_type=PeakType.Fitted, double norm_a=1.0,             # <<<<<<<<<<<<<<
 *                                    double norm_b=1.0, bint sqrt_transform=False):
 *     cdef:
 */

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__16 = __pyx_e_12ms_deisotope_2_c_18similarity_methods_Fitted;

#line 528 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__16 = __pyx_e_12ms_deisotope_2_c_18similarity_methods_Fitted;

  /* "ms_deisotope/_c/similarity_methods.pyx":614
 * 
 * 
 * cpdef double ppm_peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double ab, aa, bb
 */

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(2e-5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(2e-5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__17 = __pyx_t_3;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__18 = 2e-5;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__18 = 2e-5;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__19 = 2e-5;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__19 = 2e-5;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__20 = 2e-5;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__20 = 2e-5;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__21 = 2e-5;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__21 = 2e-5;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_35__pyx_fuse_0ppm_peak_set_similarity, 0, __pyx_n_s_pyx_fuse_0ppm_peak_set_similar, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_t_2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_PeakSet, __pyx_t_1) < 0) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_37__pyx_fuse_1ppm_peak_set_similarity, 0, __pyx_n_s_pyx_fuse_1ppm_peak_set_similar, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_t_2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_PeakIndex, __pyx_t_1) < 0) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_39__pyx_fuse_2ppm_peak_set_similarity, 0, __pyx_n_s_pyx_fuse_2ppm_peak_set_similar, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_t_2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_DeconvolutedPeakSet, __pyx_t_1) < 0) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_41__pyx_fuse_3ppm_peak_set_similarity, 0, __pyx_n_s_pyx_fuse_3ppm_peak_set_similar, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_t_2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_object, __pyx_t_1) < 0) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_7ppm_peak_set_similarity, 0, __pyx_n_s_ppm_peak_set_similarity, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_t_2);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  ((__pyx_FusedFunctionObject *) __pyx_t_1)->__signatures__ = __pyx_t_3;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ppm_peak_set_similarity, __pyx_t_1) < 0) __PYX_ERR(0, 614, __pyx_L1_error)

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":635
 * 
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,             # <<<<<<<<<<<<<<
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = PyFloat_FromDouble(2e-5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_2);

  /* "ms_deisotope/_c/similarity_methods.pyx":636
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,
 *                                     double shift=0.0, bint normalize=True):             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         return SpectrumAlignment(peak_set_a, peak_set_b, error_tolerance, shift, normalize, sqrt_transform=False)
 */

#line 636 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)

#line 636 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 636 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyBool_FromLong(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 636, __pyx_L1_error)

#line 636 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

  /* "ms_deisotope/_c/similarity_methods.pyx":635
 * 
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,             # <<<<<<<<<<<<<<
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_2 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(2e-5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__22 = __pyx_t_3;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":636
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,
 *                                     double shift=0.0, bint normalize=True):             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         return SpectrumAlignment(peak_set_a, peak_set_b, error_tolerance, shift, normalize, sqrt_transform=False)
 */

#line 636 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 636, __pyx_L1_error)

#line 636 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 636 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__23 = __pyx_t_3;

#line 636 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 636 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":635
 * 
 * 
 * cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,             # <<<<<<<<<<<<<<
 *                                     double shift=0.0, bint normalize=True):
 *     if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__24 = 2e-5;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__25 = 0.0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__26 = 1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__24 = 2e-5;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__25 = 0.0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__26 = 1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__27 = 2e-5;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__28 = 0.0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__29 = 1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__27 = 2e-5;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__28 = 0.0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__29 = 1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__30 = 2e-5;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__31 = 0.0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__32 = 1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__30 = 2e-5;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__31 = 0.0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__32 = 1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__33 = 2e-5;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__34 = 0.0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__35 = 1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__33 = 2e-5;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__34 = 0.0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_k__35 = 1;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_45__pyx_fuse_0align_peaks, 0, __pyx_n_s_pyx_fuse_0align_peaks, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_t_4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_PeakSet, __pyx_t_1) < 0) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_47__pyx_fuse_1align_peaks, 0, __pyx_n_s_pyx_fuse_1align_peaks, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_t_4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_PeakIndex, __pyx_t_1) < 0) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_49__pyx_fuse_2align_peaks, 0, __pyx_n_s_pyx_fuse_2align_peaks, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_t_4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_DeconvolutedPeakSet, __pyx_t_1) < 0) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_51__pyx_fuse_3align_peaks, 0, __pyx_n_s_pyx_fuse_3align_peaks, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_t_4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_object, __pyx_t_1) < 0) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_9align_peaks, 0, __pyx_n_s_align_peaks, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_t_4);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  ((__pyx_FusedFunctionObject *) __pyx_t_1)->__signatures__ = __pyx_t_3;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_3 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_align_peaks, __pyx_t_1) < 0) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 635 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PeakMatch(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_4 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_11__pyx_unpickle_PeakMatch, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PeakMatch, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_PeakMatch__set_state(<PeakMatch> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PeakMatch__set_state(PeakMatch __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.peak_a = __pyx_state[0]; __pyx_result.peak_b = __pyx_state[1]; __pyx_result.score = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"
  __pyx_t_4 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_18similarity_methods_33__pyx_unpickle_SpectrumAlignment, NULL, __pyx_n_s_ms_deisotope__c_similarity_metho); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 11 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_SpectrumAlignment, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "ms_deisotope/_c/similarity_methods.pyx":1
 * # cython: profile=True             # <<<<<<<<<<<<<<
 * cimport cython
 * from cython.parallel cimport prange
 */

#line 1 "ms_deisotope/_c/similarity_methods.pyx"
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 1 "ms_deisotope/_c/similarity_methods.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/similarity_methods.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumBase":9
 * 
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:             # <<<<<<<<<<<<<<
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 */

#line 9 "EnumBase"
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x02070000) != 0);

#line 9 "EnumBase"
  if (__pyx_t_5) {

    /* "EnumBase":10
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:
 *     from collections import OrderedDict as __Pyx_OrderedDict             # <<<<<<<<<<<<<<
 * else:
 *     __Pyx_OrderedDict = dict
 */

#line 10 "EnumBase"
    __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 10, __pyx_L1_error)

#line 10 "EnumBase"
    __Pyx_GOTREF(__pyx_t_4);

#line 10 "EnumBase"
    __Pyx_INCREF(__pyx_n_s_OrderedDict);

#line 10 "EnumBase"
    __Pyx_GIVEREF(__pyx_n_s_OrderedDict);

#line 10 "EnumBase"
    PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_OrderedDict);

#line 10 "EnumBase"
    __pyx_t_1 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_4, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)

#line 10 "EnumBase"
    __Pyx_GOTREF(__pyx_t_1);

#line 10 "EnumBase"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 10 "EnumBase"
    __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 10, __pyx_L1_error)

#line 10 "EnumBase"
    __Pyx_GOTREF(__pyx_t_4);

#line 10 "EnumBase"
    __Pyx_INCREF(__pyx_t_4);

#line 10 "EnumBase"
    __Pyx_XGOTREF(__Pyx_OrderedDict);

#line 10 "EnumBase"
    __Pyx_DECREF_SET(__Pyx_OrderedDict, __pyx_t_4);

#line 10 "EnumBase"
    __Pyx_GIVEREF(__pyx_t_4);

#line 10 "EnumBase"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 10 "EnumBase"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "EnumBase":9
 * 
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:             # <<<<<<<<<<<<<<
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 */

#line 9 "EnumBase"
    goto __pyx_L2;

#line 9 "EnumBase"
  }

  /* "EnumBase":12
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 *     __Pyx_OrderedDict = dict             # <<<<<<<<<<<<<<
 * 
 * @cython.internal
 */

#line 12 "EnumBase"
  /*else*/ {

#line 12 "EnumBase"
    __Pyx_INCREF(((PyObject *)(&PyDict_Type)));

#line 12 "EnumBase"
    __Pyx_XGOTREF(__Pyx_OrderedDict);

#line 12 "EnumBase"
    __Pyx_DECREF_SET(__Pyx_OrderedDict, ((PyObject *)(&PyDict_Type)));

#line 12 "EnumBase"
    __Pyx_GIVEREF(((PyObject *)(&PyDict_Type)));

#line 12 "EnumBase"
  }

#line 12 "EnumBase"
  __pyx_L2:;

  /* "EnumBase":26
 * 
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):             # <<<<<<<<<<<<<<
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 */

#line 26 "EnumBase"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 26, __pyx_L1_error)

#line 26 "EnumBase"
  __Pyx_GOTREF(__pyx_t_1);

#line 26 "EnumBase"
  __Pyx_INCREF(((PyObject *)(&PyInt_Type)));

#line 26 "EnumBase"
  __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));

#line 26 "EnumBase"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&PyInt_Type)));

#line 26 "EnumBase"
  __pyx_t_4 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)

#line 26 "EnumBase"
  __Pyx_GOTREF(__pyx_t_4);

#line 26 "EnumBase"
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_4, __pyx_t_1, __pyx_n_s_Pyx_EnumBase, __pyx_n_s_Pyx_EnumBase, (PyObject *) NULL, __pyx_n_s_EnumBase, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 26, __pyx_L1_error)

#line 26 "EnumBase"
  __Pyx_GOTREF(__pyx_t_3);

  /* "EnumBase":27
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta             # <<<<<<<<<<<<<<
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 */

#line 27 "EnumBase"
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_metaclass, ((PyObject *)__pyx_ptype___Pyx_EnumMeta)) < 0) __PYX_ERR(1, 27, __pyx_L1_error)

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

#line 28 "EnumBase"
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_1__new__, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_Pyx_EnumBase___new, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 28, __pyx_L1_error)

#line 28 "EnumBase"
  __Pyx_GOTREF(__pyx_t_2);

#line 28 "EnumBase"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__58);

#line 28 "EnumBase"
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_new, __pyx_t_2) < 0) __PYX_ERR(1, 28, __pyx_L1_error)

#line 28 "EnumBase"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

#line 39 "EnumBase"
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_3__repr__, 0, __pyx_n_s_Pyx_EnumBase___repr, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 39, __pyx_L1_error)

#line 39 "EnumBase"
  __Pyx_GOTREF(__pyx_t_2);

#line 39 "EnumBase"
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_repr, __pyx_t_2) < 0) __PYX_ERR(1, 39, __pyx_L1_error)

#line 39 "EnumBase"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

#line 41 "EnumBase"
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_5__str__, 0, __pyx_n_s_Pyx_EnumBase___str, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 41, __pyx_L1_error)

#line 41 "EnumBase"
  __Pyx_GOTREF(__pyx_t_2);

#line 41 "EnumBase"
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_str, __pyx_t_2) < 0) __PYX_ERR(1, 41, __pyx_L1_error)

#line 41 "EnumBase"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":26
 * 
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):             # <<<<<<<<<<<<<<
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 */

#line 26 "EnumBase"
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_4, __pyx_n_s_Pyx_EnumBase, __pyx_t_1, __pyx_t_3, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 26, __pyx_L1_error)

#line 26 "EnumBase"
  __Pyx_GOTREF(__pyx_t_2);

#line 26 "EnumBase"
  __Pyx_XGOTREF(__Pyx_EnumBase);

#line 26 "EnumBase"
  __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_2);

#line 26 "EnumBase"
  __Pyx_GIVEREF(__pyx_t_2);

#line 26 "EnumBase"
  __pyx_t_2 = 0;

#line 26 "EnumBase"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 26 "EnumBase"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 26 "EnumBase"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":44
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 *     from enum import IntEnum as __Pyx_EnumBase
 * 
 */

#line 44 "EnumBase"
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);

#line 44 "EnumBase"
  if (__pyx_t_5) {

    /* "EnumBase":45
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 *     from enum import IntEnum as __Pyx_EnumBase             # <<<<<<<<<<<<<<
 * 
 */

#line 45 "EnumBase"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 45, __pyx_L1_error)

#line 45 "EnumBase"
    __Pyx_GOTREF(__pyx_t_1);

#line 45 "EnumBase"
    __Pyx_INCREF(__pyx_n_s_IntEnum);

#line 45 "EnumBase"
    __Pyx_GIVEREF(__pyx_n_s_IntEnum);

#line 45 "EnumBase"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_IntEnum);

#line 45 "EnumBase"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_enum, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 45, __pyx_L1_error)

#line 45 "EnumBase"
    __Pyx_GOTREF(__pyx_t_4);

#line 45 "EnumBase"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 45 "EnumBase"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 45, __pyx_L1_error)

#line 45 "EnumBase"
    __Pyx_GOTREF(__pyx_t_1);

#line 45 "EnumBase"
    __Pyx_INCREF(__pyx_t_1);

#line 45 "EnumBase"
    __Pyx_XGOTREF(__Pyx_EnumBase);

#line 45 "EnumBase"
    __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_1);

#line 45 "EnumBase"
    __Pyx_GIVEREF(__pyx_t_1);

#line 45 "EnumBase"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 45 "EnumBase"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumBase":44
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 *     from enum import IntEnum as __Pyx_EnumBase
 * 
 */

#line 44 "EnumBase"
  }

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_4 = PyCFunction_NewEx(&__pyx_mdef_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta, NULL, __pyx_n_s_EnumBase); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */

#line 50 "EnumType"
  __pyx_t_4 = __Pyx_Globals(); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 50, __pyx_L1_error)

#line 50 "EnumType"
  __Pyx_GOTREF(__pyx_t_4);

#line 50 "EnumType"
  if (!(likely(PyDict_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 50, __pyx_L1_error)

#line 50 "EnumType"
  __Pyx_XGOTREF(__Pyx_globals);

#line 50 "EnumType"
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_4));

#line 50 "EnumType"
  __Pyx_GIVEREF(__pyx_t_4);

#line 50 "EnumType"
  __pyx_t_4 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     PeakType = __Pyx_EnumBase('PeakType', __Pyx_OrderedDict([
 */

#line 51 "EnumType"
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);

#line 51 "EnumType"
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     PeakType = __Pyx_EnumBase('PeakType', __Pyx_OrderedDict([
 *         ('Fitted', Fitted),             # <<<<<<<<<<<<<<
 *         ('Deconvoluted', Deconvoluted),
 *     ]))
 */

#line 54 "EnumType"
    __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(__pyx_e_12ms_deisotope_2_c_18similarity_methods_Fitted); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 54, __pyx_L1_error)

#line 54 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 54 "EnumType"
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 54, __pyx_L1_error)

#line 54 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 54 "EnumType"
    __Pyx_INCREF(__pyx_n_s_Fitted);

#line 54 "EnumType"
    __Pyx_GIVEREF(__pyx_n_s_Fitted);

#line 54 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Fitted);

#line 54 "EnumType"
    __Pyx_GIVEREF(__pyx_t_4);

#line 54 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);

#line 54 "EnumType"
    __pyx_t_4 = 0;

    /* "EnumType":55
 *     PeakType = __Pyx_EnumBase('PeakType', __Pyx_OrderedDict([
 *         ('Fitted', Fitted),
 *         ('Deconvoluted', Deconvoluted),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['Fitted'] = PeakType.Fitted
 */

#line 55 "EnumType"
    __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(__pyx_e_12ms_deisotope_2_c_18similarity_methods_Deconvoluted); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 55, __pyx_L1_error)

#line 55 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 55 "EnumType"
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 55, __pyx_L1_error)

#line 55 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 55 "EnumType"
    __Pyx_INCREF(__pyx_n_s_Deconvoluted);

#line 55 "EnumType"
    __Pyx_GIVEREF(__pyx_n_s_Deconvoluted);

#line 55 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Deconvoluted);

#line 55 "EnumType"
    __Pyx_GIVEREF(__pyx_t_4);

#line 55 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);

#line 55 "EnumType"
    __pyx_t_4 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     PeakType = __Pyx_EnumBase('PeakType', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('Fitted', Fitted),
 *         ('Deconvoluted', Deconvoluted),
 */

#line 53 "EnumType"
    __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 53, __pyx_L1_error)

#line 53 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 53 "EnumType"
    __Pyx_GIVEREF(__pyx_t_1);

#line 53 "EnumType"
    PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);

#line 53 "EnumType"
    __Pyx_GIVEREF(__pyx_t_3);

#line 53 "EnumType"
    PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);

#line 53 "EnumType"
    __pyx_t_1 = 0;

#line 53 "EnumType"
    __pyx_t_3 = 0;

#line 53 "EnumType"
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 53, __pyx_L1_error)

#line 53 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 53 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 53 "EnumType"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 53, __pyx_L1_error)

#line 53 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 53 "EnumType"
    __Pyx_INCREF(__pyx_n_s_PeakType);

#line 53 "EnumType"
    __Pyx_GIVEREF(__pyx_n_s_PeakType);

#line 53 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_PeakType);

#line 53 "EnumType"
    __Pyx_GIVEREF(__pyx_t_3);

#line 53 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);

#line 53 "EnumType"
    __pyx_t_3 = 0;

#line 53 "EnumType"
    __pyx_t_3 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 53, __pyx_L1_error)

#line 53 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 53 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 53 "EnumType"
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_PeakType, __pyx_t_3) < 0) __PYX_ERR(1, 53, __pyx_L1_error)

#line 53 "EnumType"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "EnumType":57
 *         ('Deconvoluted', Deconvoluted),
 *     ]))
 *     __Pyx_globals['Fitted'] = PeakType.Fitted             # <<<<<<<<<<<<<<
 *     __Pyx_globals['Deconvoluted'] = PeakType.Deconvoluted
 * else:
 */

#line 57 "EnumType"
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_PeakType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 57, __pyx_L1_error)

#line 57 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 57 "EnumType"
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Fitted); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 57, __pyx_L1_error)

#line 57 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 57 "EnumType"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 57 "EnumType"
    if (unlikely(__Pyx_globals == Py_None)) {

#line 57 "EnumType"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 57 "EnumType"
      __PYX_ERR(1, 57, __pyx_L1_error)

#line 57 "EnumType"
    }

#line 57 "EnumType"
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_Fitted, __pyx_t_4) < 0)) __PYX_ERR(1, 57, __pyx_L1_error)

#line 57 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":58
 *     ]))
 *     __Pyx_globals['Fitted'] = PeakType.Fitted
 *     __Pyx_globals['Deconvoluted'] = PeakType.Deconvoluted             # <<<<<<<<<<<<<<
 * else:
 *     class PeakType(__Pyx_EnumBase):
 */

#line 58 "EnumType"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_PeakType); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 58 "EnumType"
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Deconvoluted); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 58 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 58 "EnumType"
    if (unlikely(__Pyx_globals == Py_None)) {

#line 58 "EnumType"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 58 "EnumType"
      __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "EnumType"
    }

#line 58 "EnumType"
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_Deconvoluted, __pyx_t_3) < 0)) __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "EnumType"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     PeakType = __Pyx_EnumBase('PeakType', __Pyx_OrderedDict([
 */

#line 51 "EnumType"
    goto __pyx_L4;

#line 51 "EnumType"
  }

  /* "EnumType":60
 *     __Pyx_globals['Deconvoluted'] = PeakType.Deconvoluted
 * else:
 *     class PeakType(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['Fitted'] = PeakType(Fitted, 'Fitted')
 */

#line 60 "EnumType"
  /*else*/ {

#line 60 "EnumType"
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 60 "EnumType"
    __Pyx_INCREF(__Pyx_EnumBase);

#line 60 "EnumType"
    __Pyx_GIVEREF(__Pyx_EnumBase);

#line 60 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __Pyx_EnumBase);

#line 60 "EnumType"
    __pyx_t_4 = __Pyx_CalculateMetaclass(NULL, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 60 "EnumType"
    __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_4, __pyx_t_3, __pyx_n_s_PeakType, __pyx_n_s_PeakType, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 60 "EnumType"
    __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_4, __pyx_n_s_PeakType, __pyx_t_3, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "EnumType"
    __Pyx_GOTREF(__pyx_t_2);

#line 60 "EnumType"
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_PeakType, __pyx_t_2) < 0) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "EnumType"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 60 "EnumType"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 60 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 60 "EnumType"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "EnumType":62
 *     class PeakType(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['Fitted'] = PeakType(Fitted, 'Fitted')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['Deconvoluted'] = PeakType(Deconvoluted, 'Deconvoluted')
 * 
 */

#line 62 "EnumType"
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_PeakType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 62 "EnumType"
    __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(__pyx_e_12ms_deisotope_2_c_18similarity_methods_Fitted); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 62 "EnumType"
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 62 "EnumType"
    __Pyx_GIVEREF(__pyx_t_4);

#line 62 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);

#line 62 "EnumType"
    __Pyx_INCREF(__pyx_n_s_Fitted);

#line 62 "EnumType"
    __Pyx_GIVEREF(__pyx_n_s_Fitted);

#line 62 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_Fitted);

#line 62 "EnumType"
    __pyx_t_4 = 0;

#line 62 "EnumType"
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 62 "EnumType"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 62 "EnumType"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 62 "EnumType"
    if (unlikely(__Pyx_globals == Py_None)) {

#line 62 "EnumType"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 62 "EnumType"
      __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    }

#line 62 "EnumType"
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_Fitted, __pyx_t_4) < 0)) __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":63
 *         pass
 *     __Pyx_globals['Fitted'] = PeakType(Fitted, 'Fitted')
 *     __Pyx_globals['Deconvoluted'] = PeakType(Deconvoluted, 'Deconvoluted')             # <<<<<<<<<<<<<<
 * 
 */

#line 63 "EnumType"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_PeakType); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 63 "EnumType"
    __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(__pyx_e_12ms_deisotope_2_c_18similarity_methods_Deconvoluted); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 63 "EnumType"
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 63 "EnumType"
    __Pyx_GIVEREF(__pyx_t_1);

#line 63 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 63 "EnumType"
    __Pyx_INCREF(__pyx_n_s_Deconvoluted);

#line 63 "EnumType"
    __Pyx_GIVEREF(__pyx_n_s_Deconvoluted);

#line 63 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_Deconvoluted);

#line 63 "EnumType"
    __pyx_t_1 = 0;

#line 63 "EnumType"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 63 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 63 "EnumType"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 63 "EnumType"
    if (unlikely(__Pyx_globals == Py_None)) {

#line 63 "EnumType"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 63 "EnumType"
      __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    }

#line 63 "EnumType"
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_Deconvoluted, __pyx_t_1) < 0)) __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 63 "EnumType"
  }

#line 63 "EnumType"
  __pyx_L4:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */

#line 50 "EnumType"
  __pyx_t_1 = __Pyx_Globals(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 50, __pyx_L1_error)

#line 50 "EnumType"
  __Pyx_GOTREF(__pyx_t_1);

#line 50 "EnumType"
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 50, __pyx_L1_error)

#line 50 "EnumType"
  __Pyx_XGOTREF(__Pyx_globals);

#line 50 "EnumType"
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_1));

#line 50 "EnumType"
  __Pyx_GIVEREF(__pyx_t_1);

#line 50 "EnumType"
  __pyx_t_1 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     SimilarityMetrics = __Pyx_EnumBase('SimilarityMetrics', __Pyx_OrderedDict([
 */

#line 51 "EnumType"
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);

#line 51 "EnumType"
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     SimilarityMetrics = __Pyx_EnumBase('SimilarityMetrics', __Pyx_OrderedDict([
 *         ('dot_product', dot_product),             # <<<<<<<<<<<<<<
 *         ('normalized_dot_product', normalized_dot_product),
 *     ]))
 */

#line 54 "EnumType"
    __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics(__pyx_e_12ms_deisotope_2_c_18similarity_methods_dot_product); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 54, __pyx_L1_error)

#line 54 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 54 "EnumType"
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 54, __pyx_L1_error)

#line 54 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 54 "EnumType"
    __Pyx_INCREF(__pyx_n_s_dot_product);

#line 54 "EnumType"
    __Pyx_GIVEREF(__pyx_n_s_dot_product);

#line 54 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_dot_product);

#line 54 "EnumType"
    __Pyx_GIVEREF(__pyx_t_1);

#line 54 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);

#line 54 "EnumType"
    __pyx_t_1 = 0;

    /* "EnumType":55
 *     SimilarityMetrics = __Pyx_EnumBase('SimilarityMetrics', __Pyx_OrderedDict([
 *         ('dot_product', dot_product),
 *         ('normalized_dot_product', normalized_dot_product),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['dot_product'] = SimilarityMetrics.dot_product
 */

#line 55 "EnumType"
    __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics(__pyx_e_12ms_deisotope_2_c_18similarity_methods_normalized_dot_product); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 55, __pyx_L1_error)

#line 55 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 55 "EnumType"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 55, __pyx_L1_error)

#line 55 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 55 "EnumType"
    __Pyx_INCREF(__pyx_n_s_normalized_dot_product);

#line 55 "EnumType"
    __Pyx_GIVEREF(__pyx_n_s_normalized_dot_product);

#line 55 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_normalized_dot_product);

#line 55 "EnumType"
    __Pyx_GIVEREF(__pyx_t_1);

#line 55 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);

#line 55 "EnumType"
    __pyx_t_1 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     SimilarityMetrics = __Pyx_EnumBase('SimilarityMetrics', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('dot_product', dot_product),
 *         ('normalized_dot_product', normalized_dot_product),
 */

#line 53 "EnumType"
    __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 53, __pyx_L1_error)

#line 53 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 53 "EnumType"
    __Pyx_GIVEREF(__pyx_t_3);

#line 53 "EnumType"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);

#line 53 "EnumType"
    __Pyx_GIVEREF(__pyx_t_4);

#line 53 "EnumType"
    PyList_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);

#line 53 "EnumType"
    __pyx_t_3 = 0;

#line 53 "EnumType"
    __pyx_t_4 = 0;

#line 53 "EnumType"
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 53, __pyx_L1_error)

#line 53 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 53 "EnumType"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 53 "EnumType"
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 53, __pyx_L1_error)

#line 53 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 53 "EnumType"
    __Pyx_INCREF(__pyx_n_s_SimilarityMetrics);

#line 53 "EnumType"
    __Pyx_GIVEREF(__pyx_n_s_SimilarityMetrics);

#line 53 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_SimilarityMetrics);

#line 53 "EnumType"
    __Pyx_GIVEREF(__pyx_t_4);

#line 53 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);

#line 53 "EnumType"
    __pyx_t_4 = 0;

#line 53 "EnumType"
    __pyx_t_4 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 53, __pyx_L1_error)

#line 53 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 53 "EnumType"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 53 "EnumType"
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_SimilarityMetrics, __pyx_t_4) < 0) __PYX_ERR(1, 53, __pyx_L1_error)

#line 53 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":57
 *         ('normalized_dot_product', normalized_dot_product),
 *     ]))
 *     __Pyx_globals['dot_product'] = SimilarityMetrics.dot_product             # <<<<<<<<<<<<<<
 *     __Pyx_globals['normalized_dot_product'] = SimilarityMetrics.normalized_dot_product
 * else:
 */

#line 57 "EnumType"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SimilarityMetrics); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 57, __pyx_L1_error)

#line 57 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 57 "EnumType"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_dot_product); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 57, __pyx_L1_error)

#line 57 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 57 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 57 "EnumType"
    if (unlikely(__Pyx_globals == Py_None)) {

#line 57 "EnumType"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 57 "EnumType"
      __PYX_ERR(1, 57, __pyx_L1_error)

#line 57 "EnumType"
    }

#line 57 "EnumType"
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_dot_product, __pyx_t_1) < 0)) __PYX_ERR(1, 57, __pyx_L1_error)

#line 57 "EnumType"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "EnumType":58
 *     ]))
 *     __Pyx_globals['dot_product'] = SimilarityMetrics.dot_product
 *     __Pyx_globals['normalized_dot_product'] = SimilarityMetrics.normalized_dot_product             # <<<<<<<<<<<<<<
 * else:
 *     class SimilarityMetrics(__Pyx_EnumBase):
 */

#line 58 "EnumType"
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SimilarityMetrics); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 58 "EnumType"
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_normalized_dot_product); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 58 "EnumType"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 58 "EnumType"
    if (unlikely(__Pyx_globals == Py_None)) {

#line 58 "EnumType"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 58 "EnumType"
      __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "EnumType"
    }

#line 58 "EnumType"
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_normalized_dot_product, __pyx_t_4) < 0)) __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     SimilarityMetrics = __Pyx_EnumBase('SimilarityMetrics', __Pyx_OrderedDict([
 */

#line 51 "EnumType"
    goto __pyx_L5;

#line 51 "EnumType"
  }

  /* "EnumType":60
 *     __Pyx_globals['normalized_dot_product'] = SimilarityMetrics.normalized_dot_product
 * else:
 *     class SimilarityMetrics(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['dot_product'] = SimilarityMetrics(dot_product, 'dot_product')
 */

#line 60 "EnumType"
  /*else*/ {

#line 60 "EnumType"
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 60 "EnumType"
    __Pyx_INCREF(__Pyx_EnumBase);

#line 60 "EnumType"
    __Pyx_GIVEREF(__Pyx_EnumBase);

#line 60 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __Pyx_EnumBase);

#line 60 "EnumType"
    __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 60 "EnumType"
    __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_4, __pyx_n_s_SimilarityMetrics, __pyx_n_s_SimilarityMetrics, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 60 "EnumType"
    __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_SimilarityMetrics, __pyx_t_4, __pyx_t_3, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "EnumType"
    __Pyx_GOTREF(__pyx_t_2);

#line 60 "EnumType"
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_SimilarityMetrics, __pyx_t_2) < 0) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "EnumType"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 60 "EnumType"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 60 "EnumType"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 60 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":62
 *     class SimilarityMetrics(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['dot_product'] = SimilarityMetrics(dot_product, 'dot_product')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['normalized_dot_product'] = SimilarityMetrics(normalized_dot_product, 'normalized_dot_product')
 * 
 */

#line 62 "EnumType"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SimilarityMetrics); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 62 "EnumType"
    __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics(__pyx_e_12ms_deisotope_2_c_18similarity_methods_dot_product); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 62 "EnumType"
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 62 "EnumType"
    __Pyx_GIVEREF(__pyx_t_1);

#line 62 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 62 "EnumType"
    __Pyx_INCREF(__pyx_n_s_dot_product);

#line 62 "EnumType"
    __Pyx_GIVEREF(__pyx_n_s_dot_product);

#line 62 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_dot_product);

#line 62 "EnumType"
    __pyx_t_1 = 0;

#line 62 "EnumType"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 62 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 62 "EnumType"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 62 "EnumType"
    if (unlikely(__Pyx_globals == Py_None)) {

#line 62 "EnumType"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 62 "EnumType"
      __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    }

#line 62 "EnumType"
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_dot_product, __pyx_t_1) < 0)) __PYX_ERR(1, 62, __pyx_L1_error)

#line 62 "EnumType"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "EnumType":63
 *         pass
 *     __Pyx_globals['dot_product'] = SimilarityMetrics(dot_product, 'dot_product')
 *     __Pyx_globals['normalized_dot_product'] = SimilarityMetrics(normalized_dot_product, 'normalized_dot_product')             # <<<<<<<<<<<<<<
 * 
 */

#line 63 "EnumType"
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SimilarityMetrics); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    __Pyx_GOTREF(__pyx_t_1);

#line 63 "EnumType"
    __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics(__pyx_e_12ms_deisotope_2_c_18similarity_methods_normalized_dot_product); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 63 "EnumType"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    __Pyx_GOTREF(__pyx_t_4);

#line 63 "EnumType"
    __Pyx_GIVEREF(__pyx_t_3);

#line 63 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);

#line 63 "EnumType"
    __Pyx_INCREF(__pyx_n_s_normalized_dot_product);

#line 63 "EnumType"
    __Pyx_GIVEREF(__pyx_n_s_normalized_dot_product);

#line 63 "EnumType"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_n_s_normalized_dot_product);

#line 63 "EnumType"
    __pyx_t_3 = 0;

#line 63 "EnumType"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    __Pyx_GOTREF(__pyx_t_3);

#line 63 "EnumType"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 63 "EnumType"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 63 "EnumType"
    if (unlikely(__Pyx_globals == Py_None)) {

#line 63 "EnumType"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 63 "EnumType"
      __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    }

#line 63 "EnumType"
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_normalized_dot_product, __pyx_t_3) < 0)) __PYX_ERR(1, 63, __pyx_L1_error)

#line 63 "EnumType"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 63 "EnumType"
  }

#line 63 "EnumType"
  __pyx_L5:;

#line 63 "EnumType"
  __Pyx_TraceReturn(Py_None, 0);

#line 63 "EnumType"


#line 63 "EnumType"
  /*--- Wrapped vars code ---*/

#line 63 "EnumType"


#line 63 "EnumType"
  goto __pyx_L0;

#line 63 "EnumType"
  __pyx_L1_error:;

#line 63 "EnumType"
  __Pyx_XDECREF(__pyx_t_1);

#line 63 "EnumType"
  __Pyx_XDECREF(__pyx_t_2);

#line 63 "EnumType"
  __Pyx_XDECREF(__pyx_t_3);

#line 63 "EnumType"
  __Pyx_XDECREF(__pyx_t_4);

#line 63 "EnumType"
  if (__pyx_m) {

#line 63 "EnumType"
    if (__pyx_d) {

#line 63 "EnumType"
      __Pyx_AddTraceback("init ms_deisotope._c.similarity_methods", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 63 "EnumType"
    }

#line 63 "EnumType"
    Py_CLEAR(__pyx_m);

#line 63 "EnumType"
  } else if (!PyErr_Occurred()) {

#line 63 "EnumType"
    PyErr_SetString(PyExc_ImportError, "init ms_deisotope._c.similarity_methods");

#line 63 "EnumType"
  }

#line 63 "EnumType"
  __pyx_L0:;

#line 63 "EnumType"
  __Pyx_RefNannyFinishContext();

#line 63 "EnumType"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 63 "EnumType"
  return (__pyx_m != NULL) ? 0 : -1;

#line 63 "EnumType"
  #elif PY_MAJOR_VERSION >= 3

#line 63 "EnumType"
  return __pyx_m;

#line 63 "EnumType"
  #else

#line 63 "EnumType"
  return;

#line 63 "EnumType"
  #endif

#line 63 "EnumType"
}

#line 1 "cython_utility"

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* Profile */
#if CYTHON_PROFILE
static int __Pyx_TraceSetupAndCall(PyCodeObject** code,
                                   PyFrameObject** frame,
                                   PyThreadState* tstate,
                                   const char *funcname,
                                   const char *srcfile,
                                   int firstlineno) {
    PyObject *type, *value, *traceback;
    int retval;
    if (*frame == NULL || !CYTHON_PROFILE_REUSE_FRAME) {
        if (*code == NULL) {
            *code = __Pyx_createFrameCodeObject(funcname, srcfile, firstlineno);
            if (*code == NULL) return 0;
        }
        *frame = PyFrame_New(
            tstate,                          /*PyThreadState *tstate*/
            *code,                           /*PyCodeObject *code*/
            __pyx_d,                  /*PyObject *globals*/
            0                                /*PyObject *locals*/
        );
        if (*frame == NULL) return 0;
        if (CYTHON_TRACE && (*frame)->f_trace == NULL) {
            Py_INCREF(Py_None);
            (*frame)->f_trace = Py_None;
        }
#if PY_VERSION_HEX < 0x030400B1
    } else {
        (*frame)->f_tstate = tstate;
#endif
    }
    __Pyx_PyFrame_SetLineNumber(*frame, firstlineno);
    retval = 1;
    __Pyx_EnterTracing(tstate);
    __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
    #if CYTHON_TRACE
    if (tstate->c_tracefunc)
        retval = tstate->c_tracefunc(tstate->c_traceobj, *frame, PyTrace_CALL, NULL) == 0;
    if (retval && tstate->c_profilefunc)
    #endif
        retval = tstate->c_profilefunc(tstate->c_profileobj, *frame, PyTrace_CALL, NULL) == 0;
    __Pyx_LeaveTracing(tstate);
    if (retval) {
        __Pyx_ErrRestoreInState(tstate, type, value, traceback);
        return __Pyx_IsTracing(tstate, 0, 0) && retval;
    } else {
        Py_XDECREF(type);
        Py_XDECREF(value);
        Py_XDECREF(traceback);
        return -1;
    }
}
static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno) {
    PyCodeObject *py_code = 0;
#if PY_MAJOR_VERSION >= 3
    py_code = PyCode_NewEmpty(srcfile, funcname, firstlineno);
    if (likely(py_code)) {
        py_code->co_flags |= CO_OPTIMIZED | CO_NEWLOCALS;
    }
#else
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    py_funcname = PyString_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    py_srcfile = PyString_FromString(srcfile);
    if (unlikely(!py_srcfile)) goto bad;
    py_code = PyCode_New(
        0,
        0,
        0,
        CO_OPTIMIZED | CO_NEWLOCALS,
        __pyx_empty_bytes,     /*PyObject *code,*/
        __pyx_empty_tuple,     /*PyObject *consts,*/
        __pyx_empty_tuple,     /*PyObject *names,*/
        __pyx_empty_tuple,     /*PyObject *varnames,*/
        __pyx_empty_tuple,     /*PyObject *freevars,*/
        __pyx_empty_tuple,     /*PyObject *cellvars,*/
        py_srcfile,       /*PyObject *filename,*/
        py_funcname,      /*PyObject *name,*/
        firstlineno,
        __pyx_empty_bytes      /*PyObject *lnotab*/
    );
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
#endif
    return py_code;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* pyfrozenset_new */
  static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
  static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* None */
  static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetTopmostException */
  #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectSetAttrStr */
  #if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* ObjectGetItem */
  #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
  static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
  static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
  static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* FetchCommonType */
  static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
  #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* FusedFunction */
  static PyObject *
__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                        PyObject *qualname, PyObject *closure,
                        PyObject *module, PyObject *globals,
                        PyObject *code)
{
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_FusedFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        __pyx_FusedFunctionObject *fusedfunc = (__pyx_FusedFunctionObject *) op;
        fusedfunc->__signatures__ = NULL;
        fusedfunc->type = NULL;
        fusedfunc->self = NULL;
        PyObject_GC_Track(op);
    }
    return op;
}
static void
__pyx_FusedFunction_dealloc(__pyx_FusedFunctionObject *self)
{
    PyObject_GC_UnTrack(self);
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    __Pyx__CyFunction_dealloc((__pyx_CyFunctionObject *) self);
}
static int
__pyx_FusedFunction_traverse(__pyx_FusedFunctionObject *self,
                             visitproc visit,
                             void *arg)
{
    Py_VISIT(self->self);
    Py_VISIT(self->type);
    Py_VISIT(self->__signatures__);
    return __Pyx_CyFunction_traverse((__pyx_CyFunctionObject *) self, visit, arg);
}
static int
__pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self)
{
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    return __Pyx_CyFunction_clear((__pyx_CyFunctionObject *) self);
}
static PyObject *
__pyx_FusedFunction_descr_get(PyObject *self, PyObject *obj, PyObject *type)
{
    __pyx_FusedFunctionObject *func, *meth;
    func = (__pyx_FusedFunctionObject *) self;
    if (func->self || func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(self);
        return self;
    }
    if (obj == Py_None)
        obj = NULL;
    meth = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_New(
                    ((PyCFunctionObject *) func)->m_ml,
                    ((__pyx_CyFunctionObject *) func)->flags,
                    ((__pyx_CyFunctionObject *) func)->func_qualname,
                    ((__pyx_CyFunctionObject *) func)->func_closure,
                    ((PyCFunctionObject *) func)->m_module,
                    ((__pyx_CyFunctionObject *) func)->func_globals,
                    ((__pyx_CyFunctionObject *) func)->func_code);
    if (!meth)
        return NULL;
    if (func->func.defaults) {
        PyObject **pydefaults;
        int i;
        if (!__Pyx_CyFunction_InitDefaults((PyObject*)meth,
                                      func->func.defaults_size,
                                      func->func.defaults_pyobjects)) {
            Py_XDECREF((PyObject*)meth);
            return NULL;
        }
        memcpy(meth->func.defaults, func->func.defaults, func->func.defaults_size);
        pydefaults = __Pyx_CyFunction_Defaults(PyObject *, meth);
        for (i = 0; i < meth->func.defaults_pyobjects; i++)
            Py_XINCREF(pydefaults[i]);
    }
    Py_XINCREF(func->func.func_classobj);
    meth->func.func_classobj = func->func.func_classobj;
    Py_XINCREF(func->__signatures__);
    meth->__signatures__ = func->__signatures__;
    Py_XINCREF(type);
    meth->type = type;
    Py_XINCREF(func->func.defaults_tuple);
    meth->func.defaults_tuple = func->func.defaults_tuple;
    if (func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD)
        obj = type;
    Py_XINCREF(obj);
    meth->self = obj;
    return (PyObject *) meth;
}
static PyObject *
_obj_to_str(PyObject *obj)
{
    if (PyType_Check(obj))
        return PyObject_GetAttr(obj, __pyx_n_s_name_2);
    else
        return PyObject_Str(obj);
}
static PyObject *
__pyx_FusedFunction_getitem(__pyx_FusedFunctionObject *self, PyObject *idx)
{
    PyObject *signature = NULL;
    PyObject *unbound_result_func;
    PyObject *result_func = NULL;
    if (self->__signatures__ == NULL) {
        PyErr_SetString(PyExc_TypeError, "Function is not fused");
        return NULL;
    }
    if (PyTuple_Check(idx)) {
        PyObject *list = PyList_New(0);
        Py_ssize_t n = PyTuple_GET_SIZE(idx);
        PyObject *sep = NULL;
        int i;
        if (unlikely(!list))
            return NULL;
        for (i = 0; i < n; i++) {
            int ret;
            PyObject *string;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(idx, i);
#else
            PyObject *item = PySequence_ITEM(idx, i);  if (unlikely(!item)) goto __pyx_err;
#endif
            string = _obj_to_str(item);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
            Py_DECREF(item);
#endif
            if (unlikely(!string)) goto __pyx_err;
            ret = PyList_Append(list, string);
            Py_DECREF(string);
            if (unlikely(ret < 0)) goto __pyx_err;
        }
        sep = PyUnicode_FromString("|");
        if (likely(sep))
            signature = PyUnicode_Join(sep, list);
__pyx_err:
;
        Py_DECREF(list);
        Py_XDECREF(sep);
    } else {
        signature = _obj_to_str(idx);
    }
    if (!signature)
        return NULL;
    unbound_result_func = PyObject_GetItem(self->__signatures__, signature);
    if (unbound_result_func) {
        if (self->self || self->type) {
            __pyx_FusedFunctionObject *unbound = (__pyx_FusedFunctionObject *) unbound_result_func;
            Py_CLEAR(unbound->func.func_classobj);
            Py_XINCREF(self->func.func_classobj);
            unbound->func.func_classobj = self->func.func_classobj;
            result_func = __pyx_FusedFunction_descr_get(unbound_result_func,
                                                        self->self, self->type);
        } else {
            result_func = unbound_result_func;
            Py_INCREF(result_func);
        }
    }
    Py_DECREF(signature);
    Py_XDECREF(unbound_result_func);
    return result_func;
}
static PyObject *
__pyx_FusedFunction_callfunction(PyObject *func, PyObject *args, PyObject *kw)
{
     __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    int static_specialized = (cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD &&
                              !((__pyx_FusedFunctionObject *) func)->__signatures__);
    if (cyfunc->flags & __Pyx_CYFUNCTION_CCLASS && !static_specialized) {
        return __Pyx_CyFunction_CallAsMethod(func, args, kw);
    } else {
        return __Pyx_CyFunction_Call(func, args, kw);
    }
}
static PyObject *
__pyx_FusedFunction_call(PyObject *func, PyObject *args, PyObject *kw)
{
    __pyx_FusedFunctionObject *binding_func = (__pyx_FusedFunctionObject *) func;
    Py_ssize_t argc = PyTuple_GET_SIZE(args);
    PyObject *new_args = NULL;
    __pyx_FusedFunctionObject *new_func = NULL;
    PyObject *result = NULL;
    PyObject *self = NULL;
    int is_staticmethod = binding_func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD;
    int is_classmethod = binding_func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD;
    if (binding_func->self) {
        Py_ssize_t i;
        new_args = PyTuple_New(argc + 1);
        if (!new_args)
            return NULL;
        self = binding_func->self;
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_INCREF(self);
#endif
        Py_INCREF(self);
        PyTuple_SET_ITEM(new_args, 0, self);
        for (i = 0; i < argc; i++) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(args, i);
            Py_INCREF(item);
#else
            PyObject *item = PySequence_ITEM(args, i);  if (unlikely(!item)) goto bad;
#endif
            PyTuple_SET_ITEM(new_args, i + 1, item);
        }
        args = new_args;
    } else if (binding_func->type) {
        if (argc < 1) {
            PyErr_SetString(PyExc_TypeError, "Need at least one argument, 0 given.");
            return NULL;
        }
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        self = PyTuple_GET_ITEM(args, 0);
#else
        self = PySequence_ITEM(args, 0);  if (unlikely(!self)) return NULL;
#endif
    }
    if (self && !is_classmethod && !is_staticmethod) {
        int is_instance = PyObject_IsInstance(self, binding_func->type);
        if (unlikely(!is_instance)) {
            PyErr_Format(PyExc_TypeError,
                         "First argument should be of type %.200s, got %.200s.",
                         ((PyTypeObject *) binding_func->type)->tp_name,
                         Py_TYPE(self)->tp_name);
            goto bad;
        } else if (unlikely(is_instance == -1)) {
            goto bad;
        }
    }
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
    self = NULL;
#endif
    if (binding_func->__signatures__) {
        PyObject *tup;
        if (is_staticmethod && binding_func->func.flags & __Pyx_CYFUNCTION_CCLASS) {
            tup = PyTuple_Pack(3, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __Pyx_CyFunction_CallMethod(
                func, binding_func->__signatures__, tup, NULL);
        } else {
            tup = PyTuple_Pack(4, binding_func->__signatures__, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_callfunction(func, tup, NULL);
        }
        Py_DECREF(tup);
        if (unlikely(!new_func))
            goto bad;
        Py_XINCREF(binding_func->func.func_classobj);
        Py_CLEAR(new_func->func.func_classobj);
        new_func->func.func_classobj = binding_func->func.func_classobj;
        func = (PyObject *) new_func;
    }
    result = __pyx_FusedFunction_callfunction(func, args, kw);
bad:
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
#endif
    Py_XDECREF(new_args);
    Py_XDECREF((PyObject *) new_func);
    return result;
}
static PyMemberDef __pyx_FusedFunction_members[] = {
    {(char *) "__signatures__",
     T_OBJECT,
     offsetof(__pyx_FusedFunctionObject, __signatures__),
     READONLY,
     0},
    {0, 0, 0, 0, 0},
};
static PyMappingMethods __pyx_FusedFunction_mapping_methods = {
    0,
    (binaryfunc) __pyx_FusedFunction_getitem,
    0,
};
static PyTypeObject __pyx_FusedFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "fused_cython_function",
    sizeof(__pyx_FusedFunctionObject),
    0,
    (destructor) __pyx_FusedFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    0,
    0,
    0,
    &__pyx_FusedFunction_mapping_methods,
    0,
    (ternaryfunc) __pyx_FusedFunction_call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __pyx_FusedFunction_traverse,
    (inquiry) __pyx_FusedFunction_clear,
    0,
    0,
    0,
    0,
    0,
    __pyx_FusedFunction_members,
    __pyx_CyFunction_getsets,
    &__pyx_CyFunctionType_type,
    0,
    __pyx_FusedFunction_descr_get,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_FusedFunction_init(void) {
    __pyx_FusedFunctionType_type.tp_base = __pyx_CyFunctionType;
    __pyx_FusedFunctionType = __Pyx_FetchCommonType(&__pyx_FusedFunctionType_type);
    if (__pyx_FusedFunctionType == NULL) {
        return -1;
    }
    return 0;
}

/* CalculateMetaclass */
  static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* CythonFunction */
  static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* Py3ClassCreate */
  static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* Globals */
  static PyObject* __Pyx_Globals(void) {
    Py_ssize_t i;
    PyObject *names;
    PyObject *globals = __pyx_d;
    Py_INCREF(globals);
    names = PyObject_Dir(__pyx_m);
    if (!names)
        goto bad;
    for (i = PyList_GET_SIZE(names)-1; i >= 0; i--) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject* name = PySequence_ITEM(names, i);
        if (!name)
            goto bad;
#else
        PyObject* name = PyList_GET_ITEM(names, i);
#endif
        if (!PyDict_Contains(globals, name)) {
            PyObject* value = __Pyx_GetAttr(__pyx_m, name);
            if (!value) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                goto bad;
            }
            if (PyDict_SetItem(globals, name, value) < 0) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                Py_DECREF(value);
                goto bad;
            }
        }
#if CYTHON_COMPILING_IN_PYPY
        Py_DECREF(name);
#endif
    }
    Py_DECREF(names);
    return globals;
bad:
    Py_XDECREF(names);
    Py_XDECREF(globals);
    return NULL;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* IntPow */
  static CYTHON_INLINE long __Pyx_pow_long(long b, long e) {
    long t = b;
    switch (e) {
        case 3:
            t *= b;
        CYTHON_FALLTHROUGH;
        case 2:
            t *= b;
        CYTHON_FALLTHROUGH;
        case 1:
            return t;
        case 0:
            return 1;
    }
    #if 1
    if (unlikely(e<0)) return 0;
    #endif
    t = 1;
    while (likely(e)) {
        t *= (b * (e&1)) | ((~e)&1);
        b *= b;
        e >>= 1;
    }
    return t;
}

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType neg_one = (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) -1, const_zero = (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType __Pyx_PyInt_As_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType neg_one = (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) -1, const_zero = (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) (((((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) (((((((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) (((((((((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) (((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)-1)*(((((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) ((((((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) (((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)-1)*(((((((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) ((((((((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) (((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)-1)*(((((((((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) ((((((((((enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) -1;
        }
    } else {
        enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType");
    return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType");
    return (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_PeakType) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* ImportNumPyArray */
  static PyObject* __Pyx__ImportNumPyArray(void) {
    PyObject *numpy_module, *ndarray_object = NULL;
    numpy_module = __Pyx_Import(__pyx_n_s_numpy, NULL, 0);
    if (likely(numpy_module)) {
        ndarray_object = PyObject_GetAttrString(numpy_module, "ndarray");
        Py_DECREF(numpy_module);
    }
    if (unlikely(!ndarray_object)) {
        PyErr_Clear();
    }
    if (unlikely(!ndarray_object || !PyObject_TypeCheck(ndarray_object, &PyType_Type))) {
        Py_XDECREF(ndarray_object);
        Py_INCREF(Py_None);
        ndarray_object = Py_None;
    }
    return ndarray_object;
}
static CYTHON_INLINE PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void) {
    if (unlikely(!__pyx_numpy_ndarray)) {
        __pyx_numpy_ndarray = __Pyx__ImportNumPyArray();
    }
    Py_INCREF(__pyx_numpy_ndarray);
    return __pyx_numpy_ndarray;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics neg_one = (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics) -1, const_zero = (enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_12ms_deisotope_2_c_18similarity_methods_SimilarityMetrics),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrImport */
  #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
