/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/npy_math.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/brainpy/_c",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/ms_peak_picker/_c",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include"
        ],
        "name": "ms_deisotope._c.utils",
        "sources": [
            "src/ms_deisotope/_c/utils.pyx"
        ]
    },
    "module_name": "ms_deisotope._c.utils"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_deisotope___c__utils
#define __PYX_HAVE_API__ms_deisotope___c__utils
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include <math.h>
#include "numpy/npy_math.h"
#include "pythread.h"
#include <stdlib.h>
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "src/ms_deisotope/_c/utils.pyx",
  "__init__.pxd",
  "stringsource",
  "type.pxd",
  "peak_set.pxd",
  "peak_index.pxd",
  "composition.pxd",
  "isotopic_distribution.pxd",
  "src/ms_deisotope/_c/averagine.pxd",
  "src/ms_deisotope/_c/deconvoluter_base.pxd",
  "src/ms_deisotope/_c/scoring.pxd",
  "src/ms_deisotope/_c/peak_set.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_incr_aligned(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && CYTHON_COMPILING_IN_NOGIL
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #pragma intrinsic (_InterlockedExchangeAdd)
    #define __pyx_atomic_incr_aligned(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_aligned(value) _InterlockedExchangeAdd(value, -1)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview))
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview))
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()


/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */

#line 689 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */

#line 690 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */

#line 691 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */

#line 692 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */

#line 696 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */

#line 697 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */

#line 698 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */

#line 699 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */

#line 703 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */

#line 704 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */

#line 713 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */

#line 714 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */

#line 715 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */

#line 717 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */

#line 718 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */

#line 719 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */

#line 721 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */

#line 722 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */

#line 724 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */

#line 725 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */

#line 726 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "brainpy/_c/composition.pxd":3
 * from brainpy.mass_dict import nist_mass as __nist_mass
 * 
 * ctypedef long count_type             # <<<<<<<<<<<<<<
 * 
 * cdef dict nist_mass
 */

#line 3 "brainpy/_c/composition.pxd"
typedef long __pyx_t_7brainpy_2_c_11composition_count_type;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex;
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition;
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */

#line 728 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */

#line 729 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */

#line 730 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_complex_t;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted;

/* "ms_peak_picker/_c/peak_set.pxd":42
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout)
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self)
 */

#line 42 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak {

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":50
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak getitem(self, size_t i)
 */

#line 50 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for {

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  double error_tolerance;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":64
 * 
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */

#line 64 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell {

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  double center_value;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t start;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t end;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":70
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */

#line 70 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list {

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell *index;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  size_t size;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double low;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double high;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":90
 *     cpdef _allocate_index(self, size_t interval_index_size)
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=*, size_t hint_start=*, size_t hint_end=*)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n)
 */

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted {

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_start;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_end;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
};

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for;

/* "ms_peak_picker/_c/peak_index.pxd":11
 *         public PeakSet peaks
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 */

#line 11 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak {

#line 11 "ms_peak_picker/_c/peak_index.pxd"
  int __pyx_n;

#line 11 "ms_peak_picker/_c/peak_index.pxd"
  double tolerance;

#line 11 "ms_peak_picker/_c/peak_index.pxd"
};

/* "ms_peak_picker/_c/peak_index.pxd":14
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self)
 */

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for {

#line 14 "ms_peak_picker/_c/peak_index.pxd"
  int __pyx_n;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
  double error_tolerance;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
};

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Isotope;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Element;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Composition;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop;

/* "brainpy/_c/composition.pxd":10
 * cdef double PROTON = nist_mass["H+"][0][0]
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 */

#line 10 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass {

#line 10 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 10 "brainpy/_c/composition.pxd"
  double charge_carrier;

#line 10 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":11
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 * 
 */

#line 11 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio {

#line 11 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 11 "brainpy/_c/composition.pxd"
  double charge_carrier;

#line 11 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":17
 * # Isotope and IsotopeMap Declarations
 * 
 * cdef struct Isotope:             # <<<<<<<<<<<<<<
 *     double mass
 *     double abundance
 */

#line 17 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Isotope {

#line 17 "brainpy/_c/composition.pxd"
  double mass;

#line 17 "brainpy/_c/composition.pxd"
  double abundance;

#line 17 "brainpy/_c/composition.pxd"
  int neutrons;

#line 17 "brainpy/_c/composition.pxd"
  int neutron_shift;

#line 17 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":23
 *     int neutron_shift
 * 
 * cdef struct IsotopeMap:             # <<<<<<<<<<<<<<
 *     Isotope* bins
 *     size_t size
 */

#line 23 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap {

#line 23 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *bins;

#line 23 "brainpy/_c/composition.pxd"
  size_t size;

#line 23 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":36
 * # Element Declarations
 * 
 * cdef struct Element:             # <<<<<<<<<<<<<<
 *     char* symbol
 *     IsotopeMap* isotopes
 */

#line 36 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Element {

#line 36 "brainpy/_c/composition.pxd"
  char *symbol;

#line 36 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *isotopes;

#line 36 "brainpy/_c/composition.pxd"
  int monoisotopic_isotope_index;

#line 36 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":56
 * # ElementHashTable and ElementHashBucket Declarations
 * 
 * cdef struct ElementHashBucket:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */

#line 56 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket {

#line 56 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;

#line 56 "brainpy/_c/composition.pxd"
  size_t used;

#line 56 "brainpy/_c/composition.pxd"
  size_t size;

#line 56 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":65
 * 
 * 
 * cdef struct ElementHashTable:             # <<<<<<<<<<<<<<
 *     ElementHashBucket* buckets
 *     size_t size
 */

#line 65 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable {

#line 65 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *buckets;

#line 65 "brainpy/_c/composition.pxd"
  size_t size;

#line 65 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":94
 * # Composition Declarations
 * 
 * cdef struct Composition:             # <<<<<<<<<<<<<<
 *     char** elements
 *     count_type* counts
 */

#line 94 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Composition {

#line 94 "brainpy/_c/composition.pxd"
  char **elements;

#line 94 "brainpy/_c/composition.pxd"
  __pyx_t_7brainpy_2_c_11composition_count_type *counts;

#line 94 "brainpy/_c/composition.pxd"
  size_t size;

#line 94 "brainpy/_c/composition.pxd"
  size_t used;

#line 94 "brainpy/_c/composition.pxd"
  int max_variants;

#line 94 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":130
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base)
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _initialize_from_formula(self, str formula)
 */

#line 130 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl {

#line 130 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 130 "brainpy/_c/composition.pxd"
  int free_existing;

#line 130 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":144
 *     cpdef list items(self)
 * 
 *     cpdef pop(self, str key, object default=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef count_type getitem(self, str key)
 */

#line 144 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop {

#line 144 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 144 "brainpy/_c/composition.pxd"
  PyObject *__pyx_default;

#line 144 "brainpy/_c/composition.pxd"
};

#line 144 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector;

/* "brainpy/_c/double_vector.pxd":1
 * cdef struct DoubleVector:             # <<<<<<<<<<<<<<
 *     double* v
 *     size_t used
 */

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector {

#line 1 "brainpy/_c/double_vector.pxd"
  double *v;

#line 1 "brainpy/_c/double_vector.pxd"
  size_t used;

#line 1 "brainpy/_c/double_vector.pxd"
  size_t size;

#line 1 "brainpy/_c/double_vector.pxd"
};

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters;

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants;

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants;

/* "brainpy/_c/isotopic_constants.pxd":6
 * from brainpy._c.double_vector cimport DoubleVector
 * 
 * ctypedef DoubleVector dvec             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 6 "brainpy/_c/isotopic_constants.pxd"
typedef struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector __pyx_t_7brainpy_2_c_18isotopic_constants_dvec;

/* "brainpy/_c/isotopic_constants.pxd":9
 * 
 * 
 * cdef struct PolynomialParameters:             # <<<<<<<<<<<<<<
 *     dvec* elementary_symmetric_polynomial
 *     dvec* power_sum
 */

#line 9 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters {

#line 9 "brainpy/_c/isotopic_constants.pxd"
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *elementary_symmetric_polynomial;

#line 9 "brainpy/_c/isotopic_constants.pxd"
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *power_sum;

#line 9 "brainpy/_c/isotopic_constants.pxd"
};

/* "brainpy/_c/isotopic_constants.pxd":14
 * 
 * 
 * cdef struct PhiConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     Element* element
 */

#line 14 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants {

#line 14 "brainpy/_c/isotopic_constants.pxd"
  int order;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element *element;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *element_coefficients;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *mass_coefficients;

#line 14 "brainpy/_c/isotopic_constants.pxd"
};

/* "brainpy/_c/isotopic_constants.pxd":21
 * 
 * 
 * cdef struct IsotopicConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     PhiConstants** constants
 */

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants {

#line 21 "brainpy/_c/isotopic_constants.pxd"
  int order;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants **constants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  size_t size;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  size_t used;

#line 21 "brainpy/_c/isotopic_constants.pxd"
};

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx;

/* "brainpy/_c/isotopic_distribution.pxd":7
 * 
 * 
 * cdef struct Peak:             # <<<<<<<<<<<<<<
 *     double mz
 *     double intensity
 */

#line 7 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak {

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  double mz;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  double intensity;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":13
 * 
 * 
 * cdef struct PeakList:             # <<<<<<<<<<<<<<
 *     Peak* peaks
 *     size_t used
 */

#line 13 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList {

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *peaks;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  size_t used;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  size_t size;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":35
 * 
 * 
 * cdef struct ElementCache:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */

#line 35 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache {

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  size_t used;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  size_t size;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *source;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":49
 * 
 * 
 * cdef struct IsotopicDistribution:             # <<<<<<<<<<<<<<
 *     Composition* composition
 *     IsotopicConstants* _isotopic_constants
 */

#line 49 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution {

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Composition *composition;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *_isotopic_constants;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  int order;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  double average_mass;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *monoisotopic_peak;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":58
 * 
 * 
 * cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil
 * 
 */

#line 58 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution {

#line 58 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 58 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 58 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":77
 * 
 * cdef int max_variants(Composition* composition, ElementCache* cache) nogil
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)
 */

#line 77 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks {

#line 77 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 77 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 77 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":79
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 79 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants {

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  PyObject *npeaks;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  double charge_carrier;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":82
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=*, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 82 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants {

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  double charge_carrier;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":85
 * 
 * 
 * cdef size_t max_variants_approx(double mass, double lambda_factor=*, size_t maxiter=*, double threshold=*) nogil             # <<<<<<<<<<<<<<
 */

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx {

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  double lambda_factor;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  size_t maxiter;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  double threshold;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
};

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate;

/* "ms_deisotope/_c/averagine.pxd":7
 * cdef double PROTON
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)
 * cdef void slide(double mz, list peaklist)
 */

#line 7 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio {

#line 7 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 7 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 7 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":8
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef void slide(double mz, list peaklist)
 * cdef dict scale_dict(dict data, double factor)
 */

#line 8 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass {

#line 8 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 8 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 8 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":17
 *         public dict base_composition
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 */

#line 17 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale {

#line 17 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 17 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 17 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 17 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":18
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */

#line 18 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster {

#line 18 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 18 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 18 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":19
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 19 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster {

#line 19 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 19 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 19 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":39
 *     cdef bint _eq_inst(self, TheoreticalIsotopicPattern other)
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 */

#line 39 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below {

#line 39 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 39 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 39 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":40
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)
 */

#line 40 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after {

#line 40 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 40 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 40 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":42
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern scale_raw(self, double scale_factor)
 *     cpdef double drop_last_peak(self)
 */

#line 42 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale {

#line 42 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 42 "ms_deisotope/_c/averagine.pxd"
  PyObject *method;

#line 42 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":53
 * 
 *     @cython.final
 *     cdef inline TheoreticalIsotopicPattern _scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 * 
 *     @cython.final
 */

#line 53 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale {

#line 53 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 53 "ms_deisotope/_c/averagine.pxd"
  PyObject *method;

#line 53 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":88
 *     cdef double _make_cache_key(self, double mz)
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */

#line 88 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair {

#line 88 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 88 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 88 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":89
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 89 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster {

#line 89 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 89 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 89 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":92
 * 
 * 
 * cpdef double isotopic_shift(int charge=*)             # <<<<<<<<<<<<<<
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)
 */

#line 92 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift {

#line 92 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 92 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 92 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":94
 * cpdef double isotopic_shift(int charge=*)
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)             # <<<<<<<<<<<<<<
 */

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate {

#line 94 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 94 "ms_deisotope/_c/averagine.pxd"
  double lambda_factor;

#line 94 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 94 "ms_deisotope/_c/averagine.pxd"
};

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs;

/* "ms_deisotope/_c/deconvoluter_base.pxd":42
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance)
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 */

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution {

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":45
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 *     cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound)
 * 
 */

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction {

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":50
 *     cpdef bint _check_fit(self, IsotopicFitRecord fit)
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 */

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak {

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":51
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 */

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak {

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":53
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 * 
 */

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace {

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":54
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace {

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":61
 *         public AveragineCache averagine
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution {

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":65
 *                                                          double ignore_below=*)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges {

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":73
 *         public list averagines
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution {

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":76
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges {

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":83
 * 
 * 
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=*,             # <<<<<<<<<<<<<<
 *                                      object charge_range=*,
 *                                      int left_search_limit=*, int right_search_limit=*,
 */

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs {

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *charge_range;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int left_search_limit;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int right_search_limit;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int recalculate_starting_peak;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int use_quick_charge;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference;

/* "ms_deisotope/_c/scoring.pxd":29
 * 
 *     @staticmethod
 *     cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,             # <<<<<<<<<<<<<<
 *                                    list experimental, object data=*, int missed_peaks=*)
 * 
 */

#line 29 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create {

#line 29 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 29 "ms_deisotope/_c/scoring.pxd"
  PyObject *data;

#line 29 "ms_deisotope/_c/scoring.pxd"
  int missed_peaks;

#line 29 "ms_deisotope/_c/scoring.pxd"
};

/* "ms_deisotope/_c/scoring.pxd":75
 *     cdef public double mass_error_tolerance
 * 
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=*, double minimum_signal_to_noise=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 75 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak {

#line 75 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 75 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 75 "ms_deisotope/_c/scoring.pxd"
  double minimum_signal_to_noise;

#line 75 "ms_deisotope/_c/scoring.pxd"
};

/* "ms_deisotope/_c/scoring.pxd":92
 *     cdef public PeakSet peaklist
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=*, double upper=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference {

#line 92 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 92 "ms_deisotope/_c/scoring.pxd"
  double lower;

#line 92 "ms_deisotope/_c/scoring.pxd"
  double upper;

#line 92 "ms_deisotope/_c/scoring.pxd"
};

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_all_peaks_for;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for;

/* "ms_deisotope/_c/peak_set.pxd":138
 * 
 * 
 * cdef enum CPeaksFlags:             # <<<<<<<<<<<<<<
 *     borrowing = 0
 *     owns_peaks = 1
 */

#line 138 "ms_deisotope/_c/peak_set.pxd"
enum __pyx_t_12ms_deisotope_2_c_8peak_set_CPeaksFlags {

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_borrowing = 0,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_peaks = 1,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_index = 2,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_peaks_and_index = 3

#line 138 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":87
 *     cpdef DeconvolutedPeakSet copy(self)
 * 
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 */

#line 87 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak {

#line 87 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 87 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 87 "ms_deisotope/_c/peak_set.pxd"
  int use_mz;

#line 87 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":89
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)
 */

#line 89 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak {

#line 89 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 89 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 89 "ms_deisotope/_c/peak_set.pxd"
  int use_mz;

#line 89 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":91
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef DeconvolutedPeak _get_nearest_peak(self, double neutral_mass, double* errout)
 */

#line 91 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for {

#line 91 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 91 "ms_deisotope/_c/peak_set.pxd"
  double tolerance;

#line 91 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":118
 * cdef size_t INTERVAL_INDEX_SIZE
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */

#line 118 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell {

#line 118 "ms_deisotope/_c/peak_set.pxd"
  double center_value;

#line 118 "ms_deisotope/_c/peak_set.pxd"
  size_t start;

#line 118 "ms_deisotope/_c/peak_set.pxd"
  size_t end;

#line 118 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":124
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */

#line 124 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list {

#line 124 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell *index;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  size_t size;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  double low;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  double high;

#line 124 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":131
 * 
 * 
 * cdef struct deconvoluted_peak_t:             # <<<<<<<<<<<<<<
 *     double neutral_mass
 *     double intensity
 */

#line 131 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t {

#line 131 "ms_deisotope/_c/peak_set.pxd"
  double neutral_mass;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  double intensity;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  int charge;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  unsigned int index;

#line 131 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":144
 *     owns_peaks_and_index = 3
 * 
 * cdef struct deconvoluted_peak_set_t:             # <<<<<<<<<<<<<<
 *     deconvoluted_peak_t* peaks
 *     double* mass_index
 */

#line 144 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t {

#line 144 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *peaks;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  double *mass_index;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  size_t size;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  enum __pyx_t_12ms_deisotope_2_c_8peak_set_CPeaksFlags flags;

#line 144 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":154
 * cdef int free_deconvoluted_peak_set_t(deconvoluted_peak_set_t* destination) nogil
 * 
 * cdef deconvoluted_peak_set_t deconvoluted_peak_set_all_peaks_for(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * cdef deconvoluted_peak_t* deconvoluted_peak_set_has_peak(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil
 * 
 */

#line 154 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_all_peaks_for {

#line 154 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 154 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 154 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":155
 * 
 * cdef deconvoluted_peak_set_t deconvoluted_peak_set_all_peaks_for(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil
 * cdef deconvoluted_peak_t* deconvoluted_peak_set_has_peak(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cdef int deconvoluted_peak_eq(deconvoluted_peak_t* self, deconvoluted_peak_t* other) nogil
 */

#line 155 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_has_peak {

#line 155 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 155 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 155 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":167
 *     cdef deconvoluted_peak_t* getitem(self, size_t i) nogil
 * 
 *     cdef deconvoluted_peak_t* _has_peak(self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil
 * 
 */

#line 167 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak {

#line 167 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 167 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 167 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":168
 * 
 *     cdef deconvoluted_peak_t* _has_peak(self, double neutral_mass, double error_tolerance=*) nogil
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)
 */

#line 168 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for {

#line 168 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 168 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 168 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":170
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double error_tolerance=*)
 */

#line 170 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak {

#line 170 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 170 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 170 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":171
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double error_tolerance=*)             # <<<<<<<<<<<<<<
 */

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for {

#line 171 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 171 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 171 "ms_deisotope/_c/peak_set.pxd"
};

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_5utils_deserialize_deconvoluted_peak_set;

/* "ms_deisotope/_c/utils.pxd":22
 * 
 * cpdef list decode_envelopes(np.ndarray[np.float32_t, ndim=1] array)
 * cpdef DeconvolutedPeakSetIndexed deserialize_deconvoluted_peak_set(dict scan_dict, bint include_envelopes=*)             # <<<<<<<<<<<<<<
 * cpdef DeconvolutedPeakSetIndexed build_deconvoluted_peak_set_from_arrays(np.ndarray[double, ndim=1] mz_array,
 *                                                                          np.ndarray[double, ndim=1] intensity_array,
 */

#line 22 "ms_deisotope/_c/utils.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_5utils_deserialize_deconvoluted_peak_set {

#line 22 "ms_deisotope/_c/utils.pxd"
  int __pyx_n;

#line 22 "ms_deisotope/_c/utils.pxd"
  int include_envelopes;

#line 22 "ms_deisotope/_c/utils.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtab;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double mz;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double intensity;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double area;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double signal_to_noise;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double full_width_at_half_max;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double left_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double right_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long peak_count;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long index;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtab;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *peaks;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int indexed;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  double *mz_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *interval_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

#line 5 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex {

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyObject_HEAD

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtab;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyArrayObject *mz_array;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyArrayObject *intensity_array;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaks;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
};

#line 5 "ms_peak_picker/_c/peak_index.pxd"


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

#line 123 "brainpy/_c/composition.pxd"
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition {

#line 123 "brainpy/_c/composition.pxd"
  PyObject_HEAD

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtab;

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Composition *impl;

#line 123 "brainpy/_c/composition.pxd"
  double cached_mass;

#line 123 "brainpy/_c/composition.pxd"
  int _clean;

#line 123 "brainpy/_c/composition.pxd"
};

#line 123 "brainpy/_c/composition.pxd"


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 63 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  PyObject_HEAD

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtab;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  double mz;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  double intensity;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
};

#line 63 "brainpy/_c/isotopic_distribution.pxd"


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

#line 12 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine {

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtab;

#line 12 "ms_deisotope/_c/averagine.pxd"
  double base_mass;

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject *base_composition;

#line 12 "ms_deisotope/_c/averagine.pxd"
};

#line 12 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

#line 23 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtab;

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *peaklist;

#line 23 "ms_deisotope/_c/averagine.pxd"
  double origin;

#line 23 "ms_deisotope/_c/averagine.pxd"
  double offset;

#line 23 "ms_deisotope/_c/averagine.pxd"
};

#line 23 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

#line 79 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache {

#line 79 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtab;

#line 79 "ms_deisotope/_c/averagine.pxd"
  PyObject *backend;

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *averagine;

#line 79 "ms_deisotope/_c/averagine.pxd"
  double cache_truncation;

#line 79 "ms_deisotope/_c/averagine.pxd"
  int enabled;

#line 79 "ms_deisotope/_c/averagine.pxd"
};

#line 79 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject_HEAD

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtab;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int use_subtraction;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *scale_method;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int merge_isobaric_peaks;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  double minimum_intensity;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *scorer;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int verbose;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *_slice_cache;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *averagine;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *averagines;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject_HEAD

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtab;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int lower;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int upper;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int sign;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int *values;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t size;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t index;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

#line 16 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtab;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *seed_peak;

#line 16 "ms_deisotope/_c/scoring.pxd"
  double score;

#line 16 "ms_deisotope/_c/scoring.pxd"
  int charge;

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject *experimental;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *theoretical;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *monoisotopic_peak;

#line 16 "ms_deisotope/_c/scoring.pxd"
  int missed_peaks;

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject *data;

#line 16 "ms_deisotope/_c/scoring.pxd"
  Py_hash_t _hash;

#line 16 "ms_deisotope/_c/scoring.pxd"
};

#line 16 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

#line 38 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase {

#line 38 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtab;

#line 38 "ms_deisotope/_c/scoring.pxd"
  double minimum_score;

#line 38 "ms_deisotope/_c/scoring.pxd"
};

#line 38 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

#line 50 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {

#line 50 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtab;

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *select;

#line 50 "ms_deisotope/_c/scoring.pxd"
};

#line 50 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 62 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {

#line 62 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 62 "ms_deisotope/_c/scoring.pxd"
};

#line 62 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 66 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter {

#line 66 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 66 "ms_deisotope/_c/scoring.pxd"
};

#line 66 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 69 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {

#line 69 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 69 "ms_deisotope/_c/scoring.pxd"
};

#line 69 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

#line 72 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter {

#line 72 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 72 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 72 "ms_deisotope/_c/scoring.pxd"
};

#line 72 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

#line 78 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {

#line 78 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 78 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 78 "ms_deisotope/_c/scoring.pxd"
  double penalty_factor;

#line 78 "ms_deisotope/_c/scoring.pxd"
};

#line 78 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

#line 84 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer {

#line 84 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 84 "ms_deisotope/_c/scoring.pxd"
  PyObject *function;

#line 84 "ms_deisotope/_c/scoring.pxd"
};

#line 84 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

#line 89 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection {

#line 89 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 89 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtab;

#line 89 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;

#line 89 "ms_deisotope/_c/scoring.pxd"
};

#line 89 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

#line 95 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *interference_detector;

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *g_test_scaled;

#line 95 "ms_deisotope/_c/scoring.pxd"
  double peak_count_scale;

#line 95 "ms_deisotope/_c/scoring.pxd"
  double domain_scale;

#line 95 "ms_deisotope/_c/scoring.pxd"
};

#line 95 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

#line 102 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 102 "ms_deisotope/_c/scoring.pxd"
  double scale_factor;

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *scorer;

#line 102 "ms_deisotope/_c/scoring.pxd"
};

#line 102 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

#line 112 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter {

#line 112 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 112 "ms_deisotope/_c/scoring.pxd"
};

#line 112 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */

#line 5 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index {

#line 5 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 5 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtab;

#line 5 "ms_deisotope/_c/peak_set.pxd"
  size_t neutral_mass;

#line 5 "ms_deisotope/_c/peak_set.pxd"
  size_t mz;

#line 5 "ms_deisotope/_c/peak_set.pxd"
};

#line 5 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 14 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair {

#line 14 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 14 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtab;

#line 14 "ms_deisotope/_c/peak_set.pxd"
  double mz;

#line 14 "ms_deisotope/_c/peak_set.pxd"
  double intensity;

#line 14 "ms_deisotope/_c/peak_set.pxd"
};

#line 14 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */

#line 25 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope {

#line 25 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtab;

#line 25 "ms_deisotope/_c/peak_set.pxd"
  PyObject *pairs;

#line 25 "ms_deisotope/_c/peak_set.pxd"
};

#line 25 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */

#line 40 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double neutral_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double signal_to_noise;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *_index;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double full_width_at_half_max;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int charge;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double a_to_a2_ratio;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double most_abundant_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double average_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double score;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *envelope;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *fit;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int chosen_for_msms;

#line 40 "ms_deisotope/_c/peak_set.pxd"
};

#line 40 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class IonMobilityDeconvolutedPeak(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double drift_time
 */

#line 66 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak {

#line 66 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 66 "ms_deisotope/_c/peak_set.pxd"
  double drift_time;

#line 66 "ms_deisotope/_c/peak_set.pxd"
};

#line 66 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":71
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */

#line 71 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {

#line 71 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 71 "ms_deisotope/_c/peak_set.pxd"
  PyObject *solution;

#line 71 "ms_deisotope/_c/peak_set.pxd"
};

#line 71 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":75
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 75 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtab;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *peaks;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *_mz_ordered;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  int indexed;

#line 75 "ms_deisotope/_c/peak_set.pxd"
};

#line 75 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":104
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */

#line 104 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  double *neutral_mass_array;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  double *mz_array;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list *interval_index;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  size_t _size;

#line 104 "ms_deisotope/_c/peak_set.pxd"
};

#line 104 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":161
 * 
 * 
 * cdef class _CPeakSet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         deconvoluted_peak_set_t* ptr
 */

#line 161 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet {

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet *__pyx_vtab;

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t *ptr;

#line 161 "ms_deisotope/_c/peak_set.pxd"
};

#line 161 "ms_deisotope/_c/peak_set.pxd"


/* "View.MemoryView":106
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

#line 106 "View.MemoryView"
struct __pyx_array_obj {

#line 106 "View.MemoryView"
  PyObject_HEAD

#line 106 "View.MemoryView"
  struct __pyx_vtabstruct_array *__pyx_vtab;

#line 106 "View.MemoryView"
  char *data;

#line 106 "View.MemoryView"
  Py_ssize_t len;

#line 106 "View.MemoryView"
  char *format;

#line 106 "View.MemoryView"
  int ndim;

#line 106 "View.MemoryView"
  Py_ssize_t *_shape;

#line 106 "View.MemoryView"
  Py_ssize_t *_strides;

#line 106 "View.MemoryView"
  Py_ssize_t itemsize;

#line 106 "View.MemoryView"
  PyObject *mode;

#line 106 "View.MemoryView"
  PyObject *_format;

#line 106 "View.MemoryView"
  void (*callback_free_data)(void *);

#line 106 "View.MemoryView"
  int free_data;

#line 106 "View.MemoryView"
  int dtype_is_object;

#line 106 "View.MemoryView"
};

#line 106 "View.MemoryView"


/* "View.MemoryView":280
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */

#line 280 "View.MemoryView"
struct __pyx_MemviewEnum_obj {

#line 280 "View.MemoryView"
  PyObject_HEAD

#line 280 "View.MemoryView"
  PyObject *name;

#line 280 "View.MemoryView"
};

#line 280 "View.MemoryView"


/* "View.MemoryView":331
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

#line 331 "View.MemoryView"
struct __pyx_memoryview_obj {

#line 331 "View.MemoryView"
  PyObject_HEAD

#line 331 "View.MemoryView"
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;

#line 331 "View.MemoryView"
  PyObject *obj;

#line 331 "View.MemoryView"
  PyObject *_size;

#line 331 "View.MemoryView"
  PyObject *_array_interface;

#line 331 "View.MemoryView"
  PyThread_type_lock lock;

#line 331 "View.MemoryView"
  __pyx_atomic_int acquisition_count[2];

#line 331 "View.MemoryView"
  __pyx_atomic_int *acquisition_count_aligned_p;

#line 331 "View.MemoryView"
  Py_buffer view;

#line 331 "View.MemoryView"
  int flags;

#line 331 "View.MemoryView"
  int dtype_is_object;

#line 331 "View.MemoryView"
  __Pyx_TypeInfo *typeinfo;

#line 331 "View.MemoryView"
};

#line 331 "View.MemoryView"


/* "View.MemoryView":967
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

#line 967 "View.MemoryView"
struct __pyx_memoryviewslice_obj {

#line 967 "View.MemoryView"
  struct __pyx_memoryview_obj __pyx_base;

#line 967 "View.MemoryView"
  __Pyx_memviewslice from_slice;

#line 967 "View.MemoryView"
  PyObject *from_object;

#line 967 "View.MemoryView"
  PyObject *(*to_object_func)(char *);

#line 967 "View.MemoryView"
  int (*to_dtype_func)(char *, PyObject *);

#line 967 "View.MemoryView"
};

#line 967 "View.MemoryView"


#line 967 "View.MemoryView"


/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"


#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"


#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_create)(double, double, double, double, double, double, long, long, double);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak;

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"


#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_create)(PyObject *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*reindex)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_get_nearest_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_between_bounds)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double, size_t *, size_t *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getitem)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"


#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*_allocate_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, size_t, int __pyx_skip_dispatch);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak_hinted)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

#line 5 "ms_peak_picker/_c/peak_index.pxd"


#line 5 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex {

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak *__pyx_optional_args);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, double);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for *__pyx_optional_args);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
};

#line 5 "ms_peak_picker/_c/peak_index.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex;

#line 5 "ms_peak_picker/_c/peak_index.pxd"


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

#line 123 "brainpy/_c/composition.pxd"


#line 123 "brainpy/_c/composition.pxd"
struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition {

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*_create)(struct __pyx_t_7brainpy_2_c_11composition_Composition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*_set_impl)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args);

#line 123 "brainpy/_c/composition.pxd"
  void (*_initialize_from_formula)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);

#line 123 "brainpy/_c/composition.pxd"
  double (*mass)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  int (*__pyx___equality_pycomposition)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  int (*__pyx___equality_dict)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*copy)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*update)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*keys)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*values)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*items)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*pop)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop *__pyx_optional_args);

#line 123 "brainpy/_c/composition.pxd"
  __pyx_t_7brainpy_2_c_11composition_count_type (*getitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);

#line 123 "brainpy/_c/composition.pxd"
  void (*setitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);

#line 123 "brainpy/_c/composition.pxd"
  void (*increment)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);

#line 123 "brainpy/_c/composition.pxd"
  void (*add_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*subtract_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*scale_by)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, long);

#line 123 "brainpy/_c/composition.pxd"
};

#line 123 "brainpy/_c/composition.pxd"
static struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition;

#line 123 "brainpy/_c/composition.pxd"


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 63 "brainpy/_c/isotopic_distribution.pxd"


#line 63 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*_create)(double, double, int);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  int (*_eq)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*clone)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
};

#line 63 "brainpy/_c/isotopic_distribution.pxd"
static struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;

#line 63 "brainpy/_c/isotopic_distribution.pxd"


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

#line 12 "ms_deisotope/_c/averagine.pxd"


#line 12 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine {

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
};

#line 12 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine;

#line 12 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

#line 23 "ms_deisotope/_c/averagine.pxd"


#line 23 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_create)(PyObject *, double, double);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  int (*_eq_inst)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*ignore_below)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*truncate_after)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale_raw)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*drop_last_peak)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*total)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*normalize)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*get)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, Py_ssize_t);

#line 23 "ms_deisotope/_c/averagine.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*get_monoisotopic_mz)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*get_processed_peaks)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  size_t (*basepeak_index)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_drop_last)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
};

#line 23 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;

#line 23 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

#line 79 "ms_deisotope/_c/averagine.pxd"


#line 79 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache {

#line 79 "ms_deisotope/_c/averagine.pxd"
  double (*_make_cache_key)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double);

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*has_mz_charge_pair)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair *__pyx_optional_args);

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster *__pyx_optional_args);

#line 79 "ms_deisotope/_c/averagine.pxd"
};

#line 79 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache;

#line 79 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*between)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*match_theoretical_isotopic_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*scale_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_evaluate_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*subtraction)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*fit_incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_check_fit)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_find_next_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_find_previous_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_find_next_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_find_previous_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*_create)(int, int);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_tuple)(PyObject *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_quickcharge)(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*sequence_from_quickcharge)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*make_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  void (*release_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  void (*set_bounds)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int, int);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

#line 16 "ms_deisotope/_c/scoring.pxd"


#line 16 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_create)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create *__pyx_optional_args);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_lt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_gt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
};

#line 16 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;

#line 16 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

#line 38 "ms_deisotope/_c/scoring.pxd"


#line 38 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase {

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*best)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_best_from_set)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, double, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
};

#line 38 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase;

#line 38 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

#line 50 "ms_deisotope/_c/scoring.pxd"


#line 50 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {

#line 50 "ms_deisotope/_c/scoring.pxd"
  double (*_evaluate)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, double, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *(*_configure)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
};

#line 50 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;

#line 50 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 62 "ms_deisotope/_c/scoring.pxd"


#line 62 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {

#line 62 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 62 "ms_deisotope/_c/scoring.pxd"
};

#line 62 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;

#line 62 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 66 "ms_deisotope/_c/scoring.pxd"


#line 66 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter {

#line 66 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 66 "ms_deisotope/_c/scoring.pxd"
};

#line 66 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter;

#line 66 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 69 "ms_deisotope/_c/scoring.pxd"


#line 69 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {

#line 69 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 69 "ms_deisotope/_c/scoring.pxd"
};

#line 69 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;

#line 69 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

#line 72 "ms_deisotope/_c/scoring.pxd"


#line 72 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter {

#line 72 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 72 "ms_deisotope/_c/scoring.pxd"
  double (*score_peak)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak *__pyx_optional_args);

#line 72 "ms_deisotope/_c/scoring.pxd"
};

#line 72 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter;

#line 72 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

#line 78 "ms_deisotope/_c/scoring.pxd"


#line 78 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {

#line 78 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 78 "ms_deisotope/_c/scoring.pxd"
};

#line 78 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;

#line 78 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

#line 84 "ms_deisotope/_c/scoring.pxd"


#line 84 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer {

#line 84 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 84 "ms_deisotope/_c/scoring.pxd"
};

#line 84 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer;

#line 84 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

#line 89 "ms_deisotope/_c/scoring.pxd"


#line 89 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection {

#line 89 "ms_deisotope/_c/scoring.pxd"
  double (*detect_interference)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference *__pyx_optional_args);

#line 89 "ms_deisotope/_c/scoring.pxd"
};

#line 89 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection;

#line 89 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

#line 95 "ms_deisotope/_c/scoring.pxd"


#line 95 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 95 "ms_deisotope/_c/scoring.pxd"
};

#line 95 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;

#line 95 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

#line 102 "ms_deisotope/_c/scoring.pxd"


#line 102 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 102 "ms_deisotope/_c/scoring.pxd"
  double (*_calculate_scale_factor)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 102 "ms_deisotope/_c/scoring.pxd"
  void (*scale_fitted_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);

#line 102 "ms_deisotope/_c/scoring.pxd"
  void (*scale_theoretical_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);

#line 102 "ms_deisotope/_c/scoring.pxd"
};

#line 102 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;

#line 102 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

#line 112 "ms_deisotope/_c/scoring.pxd"


#line 112 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter {

#line 112 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 112 "ms_deisotope/_c/scoring.pxd"
};

#line 112 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter;

#line 112 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */

#line 5 "ms_deisotope/_c/peak_set.pxd"


#line 5 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index {

#line 5 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *(*_create)(size_t, size_t);

#line 5 "ms_deisotope/_c/peak_set.pxd"
};

#line 5 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index;

#line 5 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 14 "ms_deisotope/_c/peak_set.pxd"


#line 14 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair {

#line 14 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, int __pyx_skip_dispatch);

#line 14 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*_create)(double, double);

#line 14 "ms_deisotope/_c/peak_set.pxd"
};

#line 14 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair;

#line 14 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */

#line 25 "ms_deisotope/_c/peak_set.pxd"


#line 25 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope {

#line 25 "ms_deisotope/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, size_t);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*_create)(PyObject *);

#line 25 "ms_deisotope/_c/peak_set.pxd"
};

#line 25 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope;

#line 25 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */

#line 40 "ms_deisotope/_c/peak_set.pxd"


#line 40 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, int __pyx_skip_dispatch);

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_create_simple)(double, double, int, double, double, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);

#line 40 "ms_deisotope/_c/peak_set.pxd"
};

#line 40 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;

#line 40 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class IonMobilityDeconvolutedPeak(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double drift_time
 */

#line 66 "ms_deisotope/_c/peak_set.pxd"


#line 66 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak {

#line 66 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 66 "ms_deisotope/_c/peak_set.pxd"
};

#line 66 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak;

#line 66 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":71
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */

#line 71 "ms_deisotope/_c/peak_set.pxd"


#line 71 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {

#line 71 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 71 "ms_deisotope/_c/peak_set.pxd"
};

#line 71 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;

#line 71 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":75
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 75 "ms_deisotope/_c/peak_set.pxd"


#line 75 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*reindex)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*copy)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_get_nearest_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, double *);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*getslice)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t, size_t);

#line 75 "ms_deisotope/_c/peak_set.pxd"
};

#line 75 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;

#line 75 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":104
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */

#line 104 "ms_deisotope/_c/peak_set.pxd"


#line 104 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  void (*_build_index_arrays)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *);

#line 104 "ms_deisotope/_c/peak_set.pxd"
  int (*_interval_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *, double, double, size_t *, size_t *);

#line 104 "ms_deisotope/_c/peak_set.pxd"
};

#line 104 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;

#line 104 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":161
 * 
 * 
 * cdef class _CPeakSet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         deconvoluted_peak_set_t* ptr
 */

#line 161 "ms_deisotope/_c/peak_set.pxd"


#line 161 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet {

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, size_t);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t (*_all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
};

#line 161 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet;

#line 161 "ms_deisotope/_c/peak_set.pxd"


/* "View.MemoryView":106
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

#line 106 "View.MemoryView"


#line 106 "View.MemoryView"
struct __pyx_vtabstruct_array {

#line 106 "View.MemoryView"
  PyObject *(*get_memview)(struct __pyx_array_obj *);

#line 106 "View.MemoryView"
};

#line 106 "View.MemoryView"
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;

#line 106 "View.MemoryView"


/* "View.MemoryView":331
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

#line 331 "View.MemoryView"


#line 331 "View.MemoryView"
struct __pyx_vtabstruct_memoryview {

#line 331 "View.MemoryView"
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);

#line 331 "View.MemoryView"
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);

#line 331 "View.MemoryView"
};

#line 331 "View.MemoryView"
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;

#line 331 "View.MemoryView"


/* "View.MemoryView":967
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

#line 967 "View.MemoryView"


#line 967 "View.MemoryView"
struct __pyx_vtabstruct__memoryviewslice {

#line 967 "View.MemoryView"
  struct __pyx_vtabstruct_memoryview __pyx_base;

#line 967 "View.MemoryView"
};

#line 967 "View.MemoryView"
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* UnicodeAsUCS4.proto */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject*);

/* object_ord.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Ord(c)\
    (likely(PyUnicode_Check(c)) ? (long)__Pyx_PyUnicode_AsPy_UCS4(c) : __Pyx__PyObject_Ord(c))
#else
#define __Pyx_PyObject_Ord(c) __Pyx__PyObject_Ord(c)
#endif
static long __Pyx__PyObject_Ord(PyObject* c);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* DivInt[Py_ssize_t].proto */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t, Py_ssize_t);

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* FusedFunction.proto */
typedef struct {
    __pyx_CyFunctionObject func;
    PyObject *__signatures__;
    PyObject *type;
    PyObject *self;
} __pyx_FusedFunctionObject;
static PyObject *__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                                         PyObject *qualname, PyObject *closure,
                                         PyObject *module, PyObject *globals,
                                         PyObject *code);
static int __pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self);
static PyTypeObject *__pyx_FusedFunctionType = NULL;
static int __pyx_FusedFunction_init(void);
#define __Pyx_FusedFunction_USED

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_float(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double(PyObject *, int writable_flag);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_float(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_float(const char *itemp, PyObject *obj);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_double(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_double(const char *itemp, PyObject *obj);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* BytesContains.proto */
static CYTHON_INLINE int __Pyx_BytesContains(PyObject* bytes, char character);

/* ImportNumPyArray.proto */
static PyObject *__pyx_numpy_ndarray = NULL;
static PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'cython.view' */

/* Module declarations from 'cython' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'ms_peak_picker._c.peak_set' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = 0;
static size_t *__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'ms_peak_picker._c.peak_index' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = 0;

/* Module declarations from 'brainpy._c.composition' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition_PyComposition = 0;
static PyObject **__pyx_vp_7brainpy_2_c_11composition_nist_mass = 0;
#define __pyx_v_7brainpy_2_c_11composition_nist_mass (*__pyx_vp_7brainpy_2_c_11composition_nist_mass)
static double *__pyx_vp_7brainpy_2_c_11composition_PROTON = 0;
#define __pyx_v_7brainpy_2_c_11composition_PROTON (*__pyx_vp_7brainpy_2_c_11composition_PROTON)
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable **__pyx_vp_7brainpy_2_c_11composition__ElementTable = 0;
#define __pyx_v_7brainpy_2_c_11composition__ElementTable (*__pyx_vp_7brainpy_2_c_11composition__ElementTable)

/* Module declarations from 'brainpy._c.double_vector' */

/* Module declarations from 'brainpy._c.isotopic_constants' */
static size_t *__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE = 0;
#define __pyx_v_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE (*__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE)

/* Module declarations from 'brainpy._c.isotopic_distribution' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = 0;

/* Module declarations from 'ms_deisotope._c.averagine' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = 0;
static double *__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON = 0;
#define __pyx_v_12ms_deisotope_2_c_9averagine_PROTON (*__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON)
static double (*__pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass)(double, int, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass *__pyx_optional_args); /*proto*/

/* Module declarations from 'ms_deisotope._c.deconvoluter_base' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = 0;

/* Module declarations from 'ms_deisotope._c.scoring' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = 0;
static double *__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY = 0;
#define __pyx_v_12ms_deisotope_2_c_7scoring_INFINITY (*__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY)

/* Module declarations from 'ms_deisotope._c.peak_set' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet = 0;
static size_t *__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'cpython.list' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'numpy.math' */

/* Module declarations from 'ms_deisotope._c.utils' */
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static PyObject *__pyx_f_12ms_deisotope_2_c_5utils_intensity_getter(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_5utils_mz_getter(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_5utils_mass_getter(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_5utils_decode_envelopes(PyArrayObject *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_f_12ms_deisotope_2_c_5utils_deserialize_deconvoluted_peak_set(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_5utils_deserialize_deconvoluted_peak_set *__pyx_optional_args); /*proto*/
static struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_f_12ms_deisotope_2_c_5utils_build_deconvoluted_peak_set_from_arrays(PyArrayObject *, PyArrayObject *, PyArrayObject *, int __pyx_skip_dispatch); /*proto*/
static PyArrayObject *__pyx_f_12ms_deisotope_2_c_5utils_envelopes_to_array(PyObject *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_f_12ms_deisotope_2_c_5utils_deserialize_peak_set(PyObject *, int __pyx_skip_dispatch); /*proto*/
static double __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch); /*proto*/
static double __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch); /*proto*/
static double __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(PyObject *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch); /*proto*/
static double __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(PyObject *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *, int __pyx_skip_dispatch); /*proto*/
static double __pyx_fuse_4__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(PyObject *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(PyObject *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_fuse_4__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static double __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils_correlation(__Pyx_memviewslice, __Pyx_memviewslice, int __pyx_skip_dispatch); /*proto*/
static double __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils_correlation(__Pyx_memviewslice, __Pyx_memviewslice, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t = { "float32_t", NULL, sizeof(__pyx_t_5numpy_float32_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t = { "float64_t", NULL, sizeof(__pyx_t_5numpy_float64_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int8_t = { "int8_t", NULL, sizeof(__pyx_t_5numpy_int8_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int8_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int8_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_long = { "long", NULL, sizeof(long), { 0 }, 0, IS_UNSIGNED(long) ? 'U' : 'I', IS_UNSIGNED(long), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_float = { "float", NULL, sizeof(float), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "ms_deisotope._c.utils"
extern int __pyx_module_is_main_ms_deisotope___c__utils;
int __pyx_module_is_main_ms_deisotope___c__utils = 0;

/* Implementation of 'ms_deisotope._c.utils' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin___import__;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_IndexError;
static const char __pyx_k_[] = "";
static const char __pyx_k_O[] = "O";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k__2[] = "()";
static const char __pyx_k__3[] = "|";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_int8[] = "int8";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_peaks[] = "peaks";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_double[] = "double";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_PeakSet[] = "PeakSet";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_defaults[] = "defaults";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_mz_array[] = "mz_array";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_PeakIndex[] = "PeakIndex";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_m_z_array[] = "m/z array";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_scan_dict[] = "scan_dict";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_signatures[] = "signatures";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_correlation[] = "correlation";
static const char __pyx_k_charge_array[] = "charge array";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_charge_array_2[] = "charge_array";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_intensity_array[] = "intensity array";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_peak_sequence_bp[] = "_peak_sequence_bp";
static const char __pyx_k_include_envelopes[] = "include_envelopes";
static const char __pyx_k_intensity_array_2[] = "intensity_array";
static const char __pyx_k_peak_sequence_tic[] = "_peak_sequence_tic";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_DeconvolutedPeakSet[] = "DeconvolutedPeakSet";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_ms_deisotope__c_utils[] = "ms_deisotope._c.utils";
static const char __pyx_k_pyx_fuse_0correlation[] = "__pyx_fuse_0correlation";
static const char __pyx_k_pyx_fuse_1correlation[] = "__pyx_fuse_1correlation";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_isotopic_envelopes_array[] = "isotopic envelopes array";
static const char __pyx_k_ms_deisotope__c_peak_set[] = "ms_deisotope._c.peak_set";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_deconvolution_score_array[] = "deconvolution score array";
static const char __pyx_k_DeconvolutedPeakSetIndexed[] = "DeconvolutedPeakSetIndexed";
static const char __pyx_k_ms_peak_picker__c_peak_set[] = "ms_peak_picker._c.peak_set";
static const char __pyx_k_No_matching_signature_found[] = "No matching signature found";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_pyx_fuse_0_peak_sequence_bp[] = "__pyx_fuse_0_peak_sequence_bp";
static const char __pyx_k_pyx_fuse_1_peak_sequence_bp[] = "__pyx_fuse_1_peak_sequence_bp";
static const char __pyx_k_pyx_fuse_2_peak_sequence_bp[] = "__pyx_fuse_2_peak_sequence_bp";
static const char __pyx_k_pyx_fuse_3_peak_sequence_bp[] = "__pyx_fuse_3_peak_sequence_bp";
static const char __pyx_k_pyx_fuse_4_peak_sequence_bp[] = "__pyx_fuse_4_peak_sequence_bp";
static const char __pyx_k_ms_peak_picker__c_peak_index[] = "ms_peak_picker._c.peak_index";
static const char __pyx_k_pyx_fuse_0_peak_sequence_tic[] = "__pyx_fuse_0_peak_sequence_tic";
static const char __pyx_k_pyx_fuse_1_peak_sequence_tic[] = "__pyx_fuse_1_peak_sequence_tic";
static const char __pyx_k_pyx_fuse_2_peak_sequence_tic[] = "__pyx_fuse_2_peak_sequence_tic";
static const char __pyx_k_pyx_fuse_3_peak_sequence_tic[] = "__pyx_fuse_3_peak_sequence_tic";
static const char __pyx_k_pyx_fuse_4_peak_sequence_tic[] = "__pyx_fuse_4_peak_sequence_tic";
static const char __pyx_k_src_ms_deisotope__c_utils_pyx[] = "src/ms_deisotope/_c/utils.pyx";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Cannot_interpret_r_as_a_PeakBase[] = "Cannot interpret %r as a PeakBase object";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Expected_at_least_d_argument_s_g[] = "Expected at least %d argument%s, got %d";
static const char __pyx_k_Function_call_with_ambiguous_arg[] = "Function call with ambiguous argument types";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static PyObject *__pyx_kp_s_;
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_kp_s_Cannot_interpret_r_as_a_PeakBase;
static PyObject *__pyx_n_s_DeconvolutedPeakSet;
static PyObject *__pyx_n_s_DeconvolutedPeakSetIndexed;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_kp_s_Expected_at_least_d_argument_s_g;
static PyObject *__pyx_kp_s_Function_call_with_ambiguous_arg;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_kp_s_No_matching_signature_found;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_PeakIndex;
static PyObject *__pyx_n_s_PeakSet;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_kp_s__2;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_astype;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_kp_s_charge_array;
static PyObject *__pyx_n_s_charge_array_2;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_correlation;
static PyObject *__pyx_kp_s_deconvolution_score_array;
static PyObject *__pyx_n_s_defaults;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_double;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_float;
static PyObject *__pyx_n_s_float32;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_include_envelopes;
static PyObject *__pyx_n_s_int8;
static PyObject *__pyx_kp_s_intensity_array;
static PyObject *__pyx_n_s_intensity_array_2;
static PyObject *__pyx_kp_s_isotopic_envelopes_array;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_kp_s_m_z_array;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_kp_s_ms_deisotope__c_peak_set;
static PyObject *__pyx_n_s_ms_deisotope__c_utils;
static PyObject *__pyx_kp_s_ms_peak_picker__c_peak_index;
static PyObject *__pyx_kp_s_ms_peak_picker__c_peak_set;
static PyObject *__pyx_n_s_mz_array;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_peak_sequence_bp;
static PyObject *__pyx_n_s_peak_sequence_tic;
static PyObject *__pyx_n_s_peaks;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_fuse_0_peak_sequence_bp;
static PyObject *__pyx_n_s_pyx_fuse_0_peak_sequence_tic;
static PyObject *__pyx_n_s_pyx_fuse_0correlation;
static PyObject *__pyx_n_s_pyx_fuse_1_peak_sequence_bp;
static PyObject *__pyx_n_s_pyx_fuse_1_peak_sequence_tic;
static PyObject *__pyx_n_s_pyx_fuse_1correlation;
static PyObject *__pyx_n_s_pyx_fuse_2_peak_sequence_bp;
static PyObject *__pyx_n_s_pyx_fuse_2_peak_sequence_tic;
static PyObject *__pyx_n_s_pyx_fuse_3_peak_sequence_bp;
static PyObject *__pyx_n_s_pyx_fuse_3_peak_sequence_tic;
static PyObject *__pyx_n_s_pyx_fuse_4_peak_sequence_bp;
static PyObject *__pyx_n_s_pyx_fuse_4_peak_sequence_tic;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_scan_dict;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_signatures;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_kp_s_src_ms_deisotope__c_utils_pyx;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_intensity_getter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_2mz_getter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_4mass_getter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_6decode_envelopes(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_array); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_8envelopes_to_array(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_envelope_list); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_10deserialize_deconvoluted_peak_set(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_scan_dict, int __pyx_v_include_envelopes); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_12deserialize_peak_set(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_scan_dict); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_14build_deconvoluted_peak_set_from_arrays(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_mz_array, PyArrayObject *__pyx_v_intensity_array, PyArrayObject *__pyx_v_charge_array); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_16_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_22__pyx_fuse_0_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_24__pyx_fuse_1_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_26__pyx_fuse_2_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_28__pyx_fuse_3_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_30__pyx_fuse_4_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_18_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_34__pyx_fuse_0_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_36__pyx_fuse_1_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_38__pyx_fuse_2_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_40__pyx_fuse_3_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_42__pyx_fuse_4_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_20correlation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_46__pyx_fuse_0correlation(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_x, __Pyx_memviewslice __pyx_v_y); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_48__pyx_fuse_1correlation(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_x, __Pyx_memviewslice __pyx_v_y); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_112105877;
static PyObject *__pyx_int_136983863;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__22;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__39;
/* Late includes */

/* "ms_deisotope/_c/utils.pyx":27
 * 
 * @cython.cdivision(True)
 * cdef double ppm_error(double x, double y):             # <<<<<<<<<<<<<<
 *     return (x - y) / y
 * 
 */

#line 27 "ms_deisotope/_c/utils.pyx"


#line 27 "ms_deisotope/_c/utils.pyx"
static double __pyx_f_12ms_deisotope_2_c_5utils_ppm_error(double __pyx_v_x, double __pyx_v_y) {

#line 27 "ms_deisotope/_c/utils.pyx"
  double __pyx_r;

#line 27 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 27 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("ppm_error", 0);

  /* "ms_deisotope/_c/utils.pyx":28
 * @cython.cdivision(True)
 * cdef double ppm_error(double x, double y):
 *     return (x - y) / y             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 28 "ms_deisotope/_c/utils.pyx"
  __pyx_r = ((__pyx_v_x - __pyx_v_y) / __pyx_v_y);

#line 28 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":27
 * 
 * @cython.cdivision(True)
 * cdef double ppm_error(double x, double y):             # <<<<<<<<<<<<<<
 *     return (x - y) / y
 * 
 */

#line 27 "ms_deisotope/_c/utils.pyx"


#line 27 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 27 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 27 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 27 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 27 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":31
 * 
 * 
 * cdef double ppm2da(double mass, double error_tolerance):             # <<<<<<<<<<<<<<
 *     x = mass + mass * error_tolerance
 *     return x - mass
 */

#line 31 "ms_deisotope/_c/utils.pyx"


#line 31 "ms_deisotope/_c/utils.pyx"
static double __pyx_f_12ms_deisotope_2_c_5utils_ppm2da(double __pyx_v_mass, double __pyx_v_error_tolerance) {
  double __pyx_v_x
#line 31 "ms_deisotope/_c/utils.pyx"
;

#line 31 "ms_deisotope/_c/utils.pyx"
  double __pyx_r;

#line 31 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 31 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("ppm2da", 0);

  /* "ms_deisotope/_c/utils.pyx":32
 * 
 * cdef double ppm2da(double mass, double error_tolerance):
 *     x = mass + mass * error_tolerance             # <<<<<<<<<<<<<<
 *     return x - mass
 * 
 */

#line 32 "ms_deisotope/_c/utils.pyx"
  __pyx_v_x = (__pyx_v_mass + (__pyx_v_mass * __pyx_v_error_tolerance));

  /* "ms_deisotope/_c/utils.pyx":33
 * cdef double ppm2da(double mass, double error_tolerance):
 *     x = mass + mass * error_tolerance
 *     return x - mass             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 33 "ms_deisotope/_c/utils.pyx"
  __pyx_r = (__pyx_v_x - __pyx_v_mass);

#line 33 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":31
 * 
 * 
 * cdef double ppm2da(double mass, double error_tolerance):             # <<<<<<<<<<<<<<
 *     x = mass + mass * error_tolerance
 *     return x - mass
 */

#line 31 "ms_deisotope/_c/utils.pyx"


#line 31 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 31 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 31 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 31 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 31 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":37
 * 
 * @cython.cdivision(True)
 * cdef double da2ppm(double mass, double error_tolerance):             # <<<<<<<<<<<<<<
 *     x = mass + error_tolerance
 *     return (x - mass) / mass
 */

#line 37 "ms_deisotope/_c/utils.pyx"


#line 37 "ms_deisotope/_c/utils.pyx"
static double __pyx_f_12ms_deisotope_2_c_5utils_da2ppm(double __pyx_v_mass, double __pyx_v_error_tolerance) {
  double __pyx_v_x
#line 37 "ms_deisotope/_c/utils.pyx"
;

#line 37 "ms_deisotope/_c/utils.pyx"
  double __pyx_r;

#line 37 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 37 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("da2ppm", 0);

  /* "ms_deisotope/_c/utils.pyx":38
 * @cython.cdivision(True)
 * cdef double da2ppm(double mass, double error_tolerance):
 *     x = mass + error_tolerance             # <<<<<<<<<<<<<<
 *     return (x - mass) / mass
 * 
 */

#line 38 "ms_deisotope/_c/utils.pyx"
  __pyx_v_x = (__pyx_v_mass + __pyx_v_error_tolerance);

  /* "ms_deisotope/_c/utils.pyx":39
 * cdef double da2ppm(double mass, double error_tolerance):
 *     x = mass + error_tolerance
 *     return (x - mass) / mass             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 39 "ms_deisotope/_c/utils.pyx"
  __pyx_r = ((__pyx_v_x - __pyx_v_mass) / __pyx_v_mass);

#line 39 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":37
 * 
 * @cython.cdivision(True)
 * cdef double da2ppm(double mass, double error_tolerance):             # <<<<<<<<<<<<<<
 *     x = mass + error_tolerance
 *     return (x - mass) / mass
 */

#line 37 "ms_deisotope/_c/utils.pyx"


#line 37 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 37 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 37 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 37 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 37 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":42
 * 
 * 
 * cpdef intensity_getter(peak):             # <<<<<<<<<<<<<<
 *     return (<PeakBase>peak).intensity
 * 
 */

#line 42 "ms_deisotope/_c/utils.pyx"


#line 42 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_1intensity_getter(PyObject *__pyx_self, PyObject *__pyx_v_peak); /*proto*/

#line 42 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_5utils_intensity_getter(PyObject *__pyx_v_peak, CYTHON_UNUSED int __pyx_skip_dispatch) {

#line 42 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 42 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 42 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 42 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 42 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("intensity_getter", 0);

  /* "ms_deisotope/_c/utils.pyx":43
 * 
 * cpdef intensity_getter(peak):
 *     return (<PeakBase>peak).intensity             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 43 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 43 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyFloat_FromDouble(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_peak)->intensity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)

#line 43 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 43 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 43 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 43 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":42
 * 
 * 
 * cpdef intensity_getter(peak):             # <<<<<<<<<<<<<<
 *     return (<PeakBase>peak).intensity
 * 
 */

#line 42 "ms_deisotope/_c/utils.pyx"


#line 42 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 42 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.intensity_getter", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 42 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 42 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 42 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 42 "ms_deisotope/_c/utils.pyx"
}

#line 42 "ms_deisotope/_c/utils.pyx"


#line 42 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 42 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_1intensity_getter(PyObject *__pyx_self, PyObject *__pyx_v_peak); /*proto*/

#line 42 "ms_deisotope/_c/utils.pyx"
static char __pyx_doc_12ms_deisotope_2_c_5utils_intensity_getter[] = "intensity_getter(peak)";

#line 42 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_1intensity_getter(PyObject *__pyx_self, PyObject *__pyx_v_peak) {

#line 42 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("intensity_getter (wrapper)", 0);
  __pyx_r = 
#line 42 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_intensity_getter(__pyx_self, ((PyObject *)__pyx_v_peak));

#line 42 "ms_deisotope/_c/utils.pyx"


#line 42 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 42 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 42 "ms_deisotope/_c/utils.pyx"
}

#line 42 "ms_deisotope/_c/utils.pyx"


#line 42 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_intensity_getter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak) {

#line 42 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 42 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 42 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 42 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 42 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("intensity_getter", 0);

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 42 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_5utils_intensity_getter(__pyx_v_peak, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 42 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 42 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 42 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 42 "ms_deisotope/_c/utils.pyx"


#line 42 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 42 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.intensity_getter", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 42 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 42 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 42 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 42 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 42 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":46
 * 
 * 
 * cpdef mz_getter(peak):             # <<<<<<<<<<<<<<
 *     return (<PeakBase>peak).mz
 * 
 */

#line 46 "ms_deisotope/_c/utils.pyx"


#line 46 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_3mz_getter(PyObject *__pyx_self, PyObject *__pyx_v_peak); /*proto*/

#line 46 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_5utils_mz_getter(PyObject *__pyx_v_peak, CYTHON_UNUSED int __pyx_skip_dispatch) {

#line 46 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 46 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 46 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 46 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("mz_getter", 0);

  /* "ms_deisotope/_c/utils.pyx":47
 * 
 * cpdef mz_getter(peak):
 *     return (<PeakBase>peak).mz             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 47 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 47 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyFloat_FromDouble(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_peak)->mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 47 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 47 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":46
 * 
 * 
 * cpdef mz_getter(peak):             # <<<<<<<<<<<<<<
 *     return (<PeakBase>peak).mz
 * 
 */

#line 46 "ms_deisotope/_c/utils.pyx"


#line 46 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.mz_getter", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 46 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 46 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/utils.pyx"
}

#line 46 "ms_deisotope/_c/utils.pyx"


#line 46 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 46 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_3mz_getter(PyObject *__pyx_self, PyObject *__pyx_v_peak); /*proto*/

#line 46 "ms_deisotope/_c/utils.pyx"
static char __pyx_doc_12ms_deisotope_2_c_5utils_2mz_getter[] = "mz_getter(peak)";

#line 46 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_3mz_getter(PyObject *__pyx_self, PyObject *__pyx_v_peak) {

#line 46 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("mz_getter (wrapper)", 0);
  __pyx_r = 
#line 46 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_2mz_getter(__pyx_self, ((PyObject *)__pyx_v_peak));

#line 46 "ms_deisotope/_c/utils.pyx"


#line 46 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/utils.pyx"
}

#line 46 "ms_deisotope/_c/utils.pyx"


#line 46 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_2mz_getter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak) {

#line 46 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 46 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 46 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 46 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("mz_getter", 0);

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 46 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_5utils_mz_getter(__pyx_v_peak, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 46 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 46 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 46 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 46 "ms_deisotope/_c/utils.pyx"


#line 46 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.mz_getter", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 46 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 46 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 46 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":50
 * 
 * 
 * cpdef mass_getter(peak):             # <<<<<<<<<<<<<<
 *     return (<DeconvolutedPeak>peak).neutral_mass
 * 
 */

#line 50 "ms_deisotope/_c/utils.pyx"


#line 50 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_5mass_getter(PyObject *__pyx_self, PyObject *__pyx_v_peak); /*proto*/

#line 50 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_5utils_mass_getter(PyObject *__pyx_v_peak, CYTHON_UNUSED int __pyx_skip_dispatch) {

#line 50 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 50 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 50 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 50 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 50 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("mass_getter", 0);

  /* "ms_deisotope/_c/utils.pyx":51
 * 
 * cpdef mass_getter(peak):
 *     return (<DeconvolutedPeak>peak).neutral_mass             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 51 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 51 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyFloat_FromDouble(((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_v_peak)->neutral_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)

#line 51 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 51 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 51 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 51 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":50
 * 
 * 
 * cpdef mass_getter(peak):             # <<<<<<<<<<<<<<
 *     return (<DeconvolutedPeak>peak).neutral_mass
 * 
 */

#line 50 "ms_deisotope/_c/utils.pyx"


#line 50 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 50 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.mass_getter", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 50 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 50 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 50 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 50 "ms_deisotope/_c/utils.pyx"
}

#line 50 "ms_deisotope/_c/utils.pyx"


#line 50 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 50 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_5mass_getter(PyObject *__pyx_self, PyObject *__pyx_v_peak); /*proto*/

#line 50 "ms_deisotope/_c/utils.pyx"
static char __pyx_doc_12ms_deisotope_2_c_5utils_4mass_getter[] = "mass_getter(peak)";

#line 50 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_5mass_getter(PyObject *__pyx_self, PyObject *__pyx_v_peak) {

#line 50 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("mass_getter (wrapper)", 0);
  __pyx_r = 
#line 50 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_4mass_getter(__pyx_self, ((PyObject *)__pyx_v_peak));

#line 50 "ms_deisotope/_c/utils.pyx"


#line 50 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 50 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 50 "ms_deisotope/_c/utils.pyx"
}

#line 50 "ms_deisotope/_c/utils.pyx"


#line 50 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_4mass_getter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_peak) {

#line 50 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 50 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 50 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 50 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 50 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("mass_getter", 0);

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 50 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_5utils_mass_getter(__pyx_v_peak, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 50 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 50 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 50 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 50 "ms_deisotope/_c/utils.pyx"


#line 50 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 50 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.mass_getter", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 50 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 50 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 50 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 50 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 50 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":56
 * 
 * @cython.boundscheck(False)
 * cpdef list decode_envelopes(np.ndarray[np.float32_t, ndim=1] array):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list envelope_list, current_envelope
 */

#line 56 "ms_deisotope/_c/utils.pyx"


#line 56 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_7decode_envelopes(PyObject *__pyx_self, PyObject *__pyx_v_array); /*proto*/

#line 56 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_5utils_decode_envelopes(PyArrayObject *__pyx_v_array, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_envelope_list = 0
#line 56 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_current_envelope = 0
#line 56 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_current_envelope_tuple = 0
#line 56 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_i
#line 56 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 56 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n_members
#line 56 "ms_deisotope/_c/utils.pyx"
;
  __pyx_t_5numpy_float32_t __pyx_v_a
#line 56 "ms_deisotope/_c/utils.pyx"
;
  __pyx_t_5numpy_float32_t __pyx_v_b
#line 56 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_array
#line 56 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_array
#line 56 "ms_deisotope/_c/utils.pyx"
;

#line 56 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 56 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_2;

#line 56 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 56 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_4;

#line 56 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_5;

#line 56 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_6;

#line 56 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 56 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("decode_envelopes", 0);

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_array.pybuffer.buf = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_array.refcount = 0;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_array.data = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_array.rcbuffer = &__pyx_pybuffer_array;

#line 56 "ms_deisotope/_c/utils.pyx"
  {

#line 56 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 56 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_array, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 56, __pyx_L1_error)

#line 56 "ms_deisotope/_c/utils.pyx"
  }

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_array.diminfo[0].strides = __pyx_pybuffernd_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_array.diminfo[0].shape = __pyx_pybuffernd_array.rcbuffer->pybuffer.shape[0];

  /* "ms_deisotope/_c/utils.pyx":62
 *         size_t i, n, n_members
 *         np.float32_t a, b
 *     envelope_list = []             # <<<<<<<<<<<<<<
 *     current_envelope = []
 *     i = 0
 */

#line 62 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)

#line 62 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 62 "ms_deisotope/_c/utils.pyx"
  __pyx_v_envelope_list = ((PyObject*)__pyx_t_1);

#line 62 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":63
 *         np.float32_t a, b
 *     envelope_list = []
 *     current_envelope = []             # <<<<<<<<<<<<<<
 *     i = 0
 *     n = array.shape[0]
 */

#line 63 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)

#line 63 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 63 "ms_deisotope/_c/utils.pyx"
  __pyx_v_current_envelope = ((PyObject*)__pyx_t_1);

#line 63 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":64
 *     envelope_list = []
 *     current_envelope = []
 *     i = 0             # <<<<<<<<<<<<<<
 *     n = array.shape[0]
 *     n_members = 0
 */

#line 64 "ms_deisotope/_c/utils.pyx"
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/utils.pyx":65
 *     current_envelope = []
 *     i = 0
 *     n = array.shape[0]             # <<<<<<<<<<<<<<
 *     n_members = 0
 *     while i < n:
 */

#line 65 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = (__pyx_v_array->dimensions[0]);

  /* "ms_deisotope/_c/utils.pyx":66
 *     i = 0
 *     n = array.shape[0]
 *     n_members = 0             # <<<<<<<<<<<<<<
 *     while i < n:
 *         a = array[i]
 */

#line 66 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n_members = 0;

  /* "ms_deisotope/_c/utils.pyx":67
 *     n = array.shape[0]
 *     n_members = 0
 *     while i < n:             # <<<<<<<<<<<<<<
 *         a = array[i]
 *         b = array[i + 1]
 */

#line 67 "ms_deisotope/_c/utils.pyx"
  while (1) {

#line 67 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = ((__pyx_v_i < __pyx_v_n) != 0);

#line 67 "ms_deisotope/_c/utils.pyx"
    if (!__pyx_t_2) break;

    /* "ms_deisotope/_c/utils.pyx":68
 *     n_members = 0
 *     while i < n:
 *         a = array[i]             # <<<<<<<<<<<<<<
 *         b = array[i + 1]
 *         i += 2
 */

#line 68 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = __pyx_v_i;

#line 68 "ms_deisotope/_c/utils.pyx"
    __pyx_v_a = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float32_t *, __pyx_pybuffernd_array.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_array.diminfo[0].strides));

    /* "ms_deisotope/_c/utils.pyx":69
 *     while i < n:
 *         a = array[i]
 *         b = array[i + 1]             # <<<<<<<<<<<<<<
 *         i += 2
 *         if a == 0 and b == 0:
 */

#line 69 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = (__pyx_v_i + 1);

#line 69 "ms_deisotope/_c/utils.pyx"
    __pyx_v_b = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float32_t *, __pyx_pybuffernd_array.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_array.diminfo[0].strides));

    /* "ms_deisotope/_c/utils.pyx":70
 *         a = array[i]
 *         b = array[i + 1]
 *         i += 2             # <<<<<<<<<<<<<<
 *         if a == 0 and b == 0:
 *             if current_envelope is not None:
 */

#line 70 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = (__pyx_v_i + 2);

    /* "ms_deisotope/_c/utils.pyx":71
 *         b = array[i + 1]
 *         i += 2
 *         if a == 0 and b == 0:             # <<<<<<<<<<<<<<
 *             if current_envelope is not None:
 *                 if n_members > 0:
 */

#line 71 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((__pyx_v_a == 0.0) != 0);

#line 71 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_4) {

#line 71 "ms_deisotope/_c/utils.pyx"
    } else {

#line 71 "ms_deisotope/_c/utils.pyx"
      __pyx_t_2 = __pyx_t_4;

#line 71 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L6_bool_binop_done;

#line 71 "ms_deisotope/_c/utils.pyx"
    }

#line 71 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((__pyx_v_b == 0.0) != 0);

#line 71 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 71 "ms_deisotope/_c/utils.pyx"
    __pyx_L6_bool_binop_done:;

#line 71 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/utils.pyx":72
 *         i += 2
 *         if a == 0 and b == 0:
 *             if current_envelope is not None:             # <<<<<<<<<<<<<<
 *                 if n_members > 0:
 *                     current_envelope_tuple = tuple(current_envelope)
 */

#line 72 "ms_deisotope/_c/utils.pyx"
      __pyx_t_2 = (__pyx_v_current_envelope != ((PyObject*)Py_None));

#line 72 "ms_deisotope/_c/utils.pyx"
      __pyx_t_4 = (__pyx_t_2 != 0);

#line 72 "ms_deisotope/_c/utils.pyx"
      if (__pyx_t_4) {

        /* "ms_deisotope/_c/utils.pyx":73
 *         if a == 0 and b == 0:
 *             if current_envelope is not None:
 *                 if n_members > 0:             # <<<<<<<<<<<<<<
 *                     current_envelope_tuple = tuple(current_envelope)
 *                     PyList_Append(envelope_list, Envelope._create(current_envelope_tuple))
 */

#line 73 "ms_deisotope/_c/utils.pyx"
        __pyx_t_4 = ((__pyx_v_n_members > 0) != 0);

#line 73 "ms_deisotope/_c/utils.pyx"
        if (__pyx_t_4) {

          /* "ms_deisotope/_c/utils.pyx":74
 *             if current_envelope is not None:
 *                 if n_members > 0:
 *                     current_envelope_tuple = tuple(current_envelope)             # <<<<<<<<<<<<<<
 *                     PyList_Append(envelope_list, Envelope._create(current_envelope_tuple))
 *                 elif i > 2:
 */

#line 74 "ms_deisotope/_c/utils.pyx"
          __pyx_t_1 = PyList_AsTuple(__pyx_v_current_envelope); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)

#line 74 "ms_deisotope/_c/utils.pyx"
          __Pyx_GOTREF(__pyx_t_1);

#line 74 "ms_deisotope/_c/utils.pyx"
          __Pyx_XDECREF_SET(__pyx_v_current_envelope_tuple, ((PyObject*)__pyx_t_1));

#line 74 "ms_deisotope/_c/utils.pyx"
          __pyx_t_1 = 0;

          /* "ms_deisotope/_c/utils.pyx":75
 *                 if n_members > 0:
 *                     current_envelope_tuple = tuple(current_envelope)
 *                     PyList_Append(envelope_list, Envelope._create(current_envelope_tuple))             # <<<<<<<<<<<<<<
 *                 elif i > 2:
 *                     PyList_Append(envelope_list, Envelope._create(
 */

#line 75 "ms_deisotope/_c/utils.pyx"
          __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope->_create(__pyx_v_current_envelope_tuple)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L1_error)

#line 75 "ms_deisotope/_c/utils.pyx"
          __Pyx_GOTREF(__pyx_t_1);

#line 75 "ms_deisotope/_c/utils.pyx"
          __pyx_t_5 = PyList_Append(__pyx_v_envelope_list, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 75, __pyx_L1_error)

#line 75 "ms_deisotope/_c/utils.pyx"
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "ms_deisotope/_c/utils.pyx":73
 *         if a == 0 and b == 0:
 *             if current_envelope is not None:
 *                 if n_members > 0:             # <<<<<<<<<<<<<<
 *                     current_envelope_tuple = tuple(current_envelope)
 *                     PyList_Append(envelope_list, Envelope._create(current_envelope_tuple))
 */

#line 73 "ms_deisotope/_c/utils.pyx"
          goto __pyx_L9;

#line 73 "ms_deisotope/_c/utils.pyx"
        }

        /* "ms_deisotope/_c/utils.pyx":76
 *                     current_envelope_tuple = tuple(current_envelope)
 *                     PyList_Append(envelope_list, Envelope._create(current_envelope_tuple))
 *                 elif i > 2:             # <<<<<<<<<<<<<<
 *                     PyList_Append(envelope_list, Envelope._create(
 *                             ()
 */

#line 76 "ms_deisotope/_c/utils.pyx"
        __pyx_t_4 = ((__pyx_v_i > 2) != 0);

#line 76 "ms_deisotope/_c/utils.pyx"
        if (__pyx_t_4) {

          /* "ms_deisotope/_c/utils.pyx":77
 *                     PyList_Append(envelope_list, Envelope._create(current_envelope_tuple))
 *                 elif i > 2:
 *                     PyList_Append(envelope_list, Envelope._create(             # <<<<<<<<<<<<<<
 *                             ()
 *                         )
 */

#line 77 "ms_deisotope/_c/utils.pyx"
          __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope->_create(__pyx_empty_tuple)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)

#line 77 "ms_deisotope/_c/utils.pyx"
          __Pyx_GOTREF(__pyx_t_1);

#line 77 "ms_deisotope/_c/utils.pyx"
          __pyx_t_5 = PyList_Append(__pyx_v_envelope_list, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 77, __pyx_L1_error)

#line 77 "ms_deisotope/_c/utils.pyx"
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "ms_deisotope/_c/utils.pyx":76
 *                     current_envelope_tuple = tuple(current_envelope)
 *                     PyList_Append(envelope_list, Envelope._create(current_envelope_tuple))
 *                 elif i > 2:             # <<<<<<<<<<<<<<
 *                     PyList_Append(envelope_list, Envelope._create(
 *                             ()
 */

#line 76 "ms_deisotope/_c/utils.pyx"
        }

#line 76 "ms_deisotope/_c/utils.pyx"
        __pyx_L9:;

        /* "ms_deisotope/_c/utils.pyx":81
 *                         )
 *                     )
 *                 current_envelope = []             # <<<<<<<<<<<<<<
 *                 n_members = 0
 *         else:
 */

#line 81 "ms_deisotope/_c/utils.pyx"
        __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "ms_deisotope/_c/utils.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 81 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF_SET(__pyx_v_current_envelope, ((PyObject*)__pyx_t_1));

#line 81 "ms_deisotope/_c/utils.pyx"
        __pyx_t_1 = 0;

        /* "ms_deisotope/_c/utils.pyx":82
 *                     )
 *                 current_envelope = []
 *                 n_members = 0             # <<<<<<<<<<<<<<
 *         else:
 *             PyList_Append(current_envelope, EnvelopePair._create(a, b))
 */

#line 82 "ms_deisotope/_c/utils.pyx"
        __pyx_v_n_members = 0;

        /* "ms_deisotope/_c/utils.pyx":72
 *         i += 2
 *         if a == 0 and b == 0:
 *             if current_envelope is not None:             # <<<<<<<<<<<<<<
 *                 if n_members > 0:
 *                     current_envelope_tuple = tuple(current_envelope)
 */

#line 72 "ms_deisotope/_c/utils.pyx"
      }

      /* "ms_deisotope/_c/utils.pyx":71
 *         b = array[i + 1]
 *         i += 2
 *         if a == 0 and b == 0:             # <<<<<<<<<<<<<<
 *             if current_envelope is not None:
 *                 if n_members > 0:
 */

#line 71 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5;

#line 71 "ms_deisotope/_c/utils.pyx"
    }

    /* "ms_deisotope/_c/utils.pyx":84
 *                 n_members = 0
 *         else:
 *             PyList_Append(current_envelope, EnvelopePair._create(a, b))             # <<<<<<<<<<<<<<
 *             n_members += 1
 *     current_envelope_tuple = tuple(current_envelope)
 */

#line 84 "ms_deisotope/_c/utils.pyx"
    /*else*/ {

#line 84 "ms_deisotope/_c/utils.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair->_create(__pyx_v_a, __pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)

#line 84 "ms_deisotope/_c/utils.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 84 "ms_deisotope/_c/utils.pyx"
      __pyx_t_5 = PyList_Append(__pyx_v_current_envelope, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 84, __pyx_L1_error)

#line 84 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/utils.pyx":85
 *         else:
 *             PyList_Append(current_envelope, EnvelopePair._create(a, b))
 *             n_members += 1             # <<<<<<<<<<<<<<
 *     current_envelope_tuple = tuple(current_envelope)
 *     envelope_list.append(Envelope._create(current_envelope_tuple))
 */

#line 85 "ms_deisotope/_c/utils.pyx"
      __pyx_v_n_members = (__pyx_v_n_members + 1);

#line 85 "ms_deisotope/_c/utils.pyx"
    }

#line 85 "ms_deisotope/_c/utils.pyx"
    __pyx_L5:;

#line 85 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":86
 *             PyList_Append(current_envelope, EnvelopePair._create(a, b))
 *             n_members += 1
 *     current_envelope_tuple = tuple(current_envelope)             # <<<<<<<<<<<<<<
 *     envelope_list.append(Envelope._create(current_envelope_tuple))
 *     return envelope_list
 */

#line 86 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyList_AsTuple(__pyx_v_current_envelope); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 86 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF_SET(__pyx_v_current_envelope_tuple, ((PyObject*)__pyx_t_1));

#line 86 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":87
 *             n_members += 1
 *     current_envelope_tuple = tuple(current_envelope)
 *     envelope_list.append(Envelope._create(current_envelope_tuple))             # <<<<<<<<<<<<<<
 *     return envelope_list
 * 
 */

#line 87 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope->_create(__pyx_v_current_envelope_tuple)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)

#line 87 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 87 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_envelope_list, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 87, __pyx_L1_error)

#line 87 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":88
 *     current_envelope_tuple = tuple(current_envelope)
 *     envelope_list.append(Envelope._create(current_envelope_tuple))
 *     return envelope_list             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 88 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 88 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(__pyx_v_envelope_list);

#line 88 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_envelope_list;

#line 88 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":56
 * 
 * @cython.boundscheck(False)
 * cpdef list decode_envelopes(np.ndarray[np.float32_t, ndim=1] array):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list envelope_list, current_envelope
 */

#line 56 "ms_deisotope/_c/utils.pyx"


#line 56 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 56 "ms_deisotope/_c/utils.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 56 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_declare

#line 56 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_assign

#line 56 "ms_deisotope/_c/utils.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 56 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_array.rcbuffer->pybuffer);

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.decode_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 56 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L2;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_array.rcbuffer->pybuffer);

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_L2:;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_envelope_list);

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_current_envelope);

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_current_envelope_tuple);

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 56 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 56 "ms_deisotope/_c/utils.pyx"
}

#line 56 "ms_deisotope/_c/utils.pyx"


#line 56 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 56 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_7decode_envelopes(PyObject *__pyx_self, PyObject *__pyx_v_array); /*proto*/

#line 56 "ms_deisotope/_c/utils.pyx"
static char __pyx_doc_12ms_deisotope_2_c_5utils_6decode_envelopes[] = "decode_envelopes(ndarray array) -> list";

#line 56 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_7decode_envelopes(PyObject *__pyx_self, PyObject *__pyx_v_array) {

#line 56 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 56 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 56 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("decode_envelopes (wrapper)", 0);

#line 56 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_array), __pyx_ptype_5numpy_ndarray, 1, "array", 0))) __PYX_ERR(0, 56, __pyx_L1_error)
  __pyx_r = 
#line 56 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_6decode_envelopes(__pyx_self, ((PyArrayObject *)__pyx_v_array));

#line 56 "ms_deisotope/_c/utils.pyx"


#line 56 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 56 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 56 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 56 "ms_deisotope/_c/utils.pyx"
}

#line 56 "ms_deisotope/_c/utils.pyx"


#line 56 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_6decode_envelopes(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_array) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_array
#line 56 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_array
#line 56 "ms_deisotope/_c/utils.pyx"
;

#line 56 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 56 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 56 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("decode_envelopes", 0);

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_array.pybuffer.buf = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_array.refcount = 0;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_array.data = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_array.rcbuffer = &__pyx_pybuffer_array;

#line 56 "ms_deisotope/_c/utils.pyx"
  {

#line 56 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 56 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_array, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 56, __pyx_L1_error)

#line 56 "ms_deisotope/_c/utils.pyx"
  }

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_array.diminfo[0].strides = __pyx_pybuffernd_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_array.diminfo[0].shape = __pyx_pybuffernd_array.rcbuffer->pybuffer.shape[0];

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_5utils_decode_envelopes(((PyArrayObject *)__pyx_v_array), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 56 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 56 "ms_deisotope/_c/utils.pyx"


#line 56 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 56 "ms_deisotope/_c/utils.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 56 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_declare

#line 56 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_assign

#line 56 "ms_deisotope/_c/utils.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 56 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_array.rcbuffer->pybuffer);

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.decode_envelopes", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 56 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L2;

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_array.rcbuffer->pybuffer);

#line 56 "ms_deisotope/_c/utils.pyx"
  __pyx_L2:;

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 56 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 56 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 56 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":91
 * 
 * 
 * cpdef np.ndarray[np.float32_t] envelopes_to_array(list envelope_list):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t n, m, i, j, point_count, k
 */

#line 91 "ms_deisotope/_c/utils.pyx"


#line 91 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_9envelopes_to_array(PyObject *__pyx_self, PyObject *__pyx_v_envelope_list); /*proto*/

#line 91 "ms_deisotope/_c/utils.pyx"
static PyArrayObject *__pyx_f_12ms_deisotope_2_c_5utils_envelopes_to_array(PyObject *__pyx_v_envelope_list, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_n
#line 91 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_m
#line 91 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_i
#line 91 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_j
#line 91 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_point_count
#line 91 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_k
#line 91 "ms_deisotope/_c/utils.pyx"
;
  PyArrayObject *__pyx_v_collection = 0
#line 91 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_v_envelope = 0
#line 91 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_v_pair = 0
#line 91 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_collection
#line 91 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_collection
#line 91 "ms_deisotope/_c/utils.pyx"
;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyArrayObject *__pyx_r = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 91 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_1;

#line 91 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_2;

#line 91 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_4;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyArrayObject *__pyx_t_10 = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_11;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_13 = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_14 = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_15;

#line 91 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_16;

#line 91 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_17;

#line 91 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_18;

#line 91 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_19;

#line 91 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 91 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("envelopes_to_array", 0);

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_collection.pybuffer.buf = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_collection.refcount = 0;

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_collection.data = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_collection.rcbuffer = &__pyx_pybuffer_collection;

  /* "ms_deisotope/_c/utils.pyx":97
 *         Envelope envelope
 *         EnvelopePair pair
 *     n = PyList_GET_SIZE(envelope_list)             # <<<<<<<<<<<<<<
 *     point_count = 0
 *     for i in range(n):
 */

#line 97 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_envelope_list);

  /* "ms_deisotope/_c/utils.pyx":98
 *         EnvelopePair pair
 *     n = PyList_GET_SIZE(envelope_list)
 *     point_count = 0             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         envelope = <Envelope>PyList_GET_ITEM(envelope_list, i)
 */

#line 98 "ms_deisotope/_c/utils.pyx"
  __pyx_v_point_count = 0;

  /* "ms_deisotope/_c/utils.pyx":99
 *     n = PyList_GET_SIZE(envelope_list)
 *     point_count = 0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         envelope = <Envelope>PyList_GET_ITEM(envelope_list, i)
 *         point_count += (1 + envelope.get_size()) * 2
 */

#line 99 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 99 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 99 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 99 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/utils.pyx":100
 *     point_count = 0
 *     for i in range(n):
 *         envelope = <Envelope>PyList_GET_ITEM(envelope_list, i)             # <<<<<<<<<<<<<<
 *         point_count += (1 + envelope.get_size()) * 2
 *     collection = np.zeros(point_count, dtype=np.float32)
 */

#line 100 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_envelope_list, __pyx_v_i);

#line 100 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = ((PyObject *)__pyx_t_4);

#line 100 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 100 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_envelope, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *)__pyx_t_5));

#line 100 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/utils.pyx":101
 *     for i in range(n):
 *         envelope = <Envelope>PyList_GET_ITEM(envelope_list, i)
 *         point_count += (1 + envelope.get_size()) * 2             # <<<<<<<<<<<<<<
 *     collection = np.zeros(point_count, dtype=np.float32)
 *     k = 0
 */

#line 101 "ms_deisotope/_c/utils.pyx"
    __pyx_v_point_count = (__pyx_v_point_count + ((1 + ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *)__pyx_v_envelope->__pyx_vtab)->get_size(__pyx_v_envelope)) * 2));

#line 101 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":102
 *         envelope = <Envelope>PyList_GET_ITEM(envelope_list, i)
 *         point_count += (1 + envelope.get_size()) * 2
 *     collection = np.zeros(point_count, dtype=np.float32)             # <<<<<<<<<<<<<<
 *     k = 0
 *     for i in range(n):
 */

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 102 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 102, __pyx_L1_error)

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 102 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_point_count); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 102 "ms_deisotope/_c/utils.pyx"
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 102, __pyx_L1_error)

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_7);

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(__pyx_t_5);

#line 102 "ms_deisotope/_c/utils.pyx"
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);

#line 102 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = 0;

#line 102 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 102, __pyx_L1_error)

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_8);

#line 102 "ms_deisotope/_c/utils.pyx"
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_float32); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 102, __pyx_L1_error)

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_9);

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 102 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 102, __pyx_L1_error)

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 102 "ms_deisotope/_c/utils.pyx"
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 102, __pyx_L1_error)

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_9);

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 102 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 102 "ms_deisotope/_c/utils.pyx"
  if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 102, __pyx_L1_error)

#line 102 "ms_deisotope/_c/utils.pyx"
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_9);

#line 102 "ms_deisotope/_c/utils.pyx"
  {

#line 102 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 102 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_collection.rcbuffer->pybuffer);

#line 102 "ms_deisotope/_c/utils.pyx"
    __pyx_t_11 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_collection.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);

#line 102 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_11 < 0)) {

#line 102 "ms_deisotope/_c/utils.pyx"
      PyErr_Fetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);

#line 102 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_collection.rcbuffer->pybuffer, (PyObject*)__pyx_v_collection, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {

#line 102 "ms_deisotope/_c/utils.pyx"
        Py_XDECREF(__pyx_t_12); Py_XDECREF(__pyx_t_13); Py_XDECREF(__pyx_t_14);

#line 102 "ms_deisotope/_c/utils.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 102 "ms_deisotope/_c/utils.pyx"
      } else {

#line 102 "ms_deisotope/_c/utils.pyx"
        PyErr_Restore(__pyx_t_12, __pyx_t_13, __pyx_t_14);

#line 102 "ms_deisotope/_c/utils.pyx"
      }

#line 102 "ms_deisotope/_c/utils.pyx"
      __pyx_t_12 = __pyx_t_13 = __pyx_t_14 = 0;

#line 102 "ms_deisotope/_c/utils.pyx"
    }

#line 102 "ms_deisotope/_c/utils.pyx"
    __pyx_pybuffernd_collection.diminfo[0].strides = __pyx_pybuffernd_collection.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_collection.diminfo[0].shape = __pyx_pybuffernd_collection.rcbuffer->pybuffer.shape[0];

#line 102 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 102, __pyx_L1_error)

#line 102 "ms_deisotope/_c/utils.pyx"
  }

#line 102 "ms_deisotope/_c/utils.pyx"
  __pyx_t_10 = 0;

#line 102 "ms_deisotope/_c/utils.pyx"
  __pyx_v_collection = ((PyArrayObject *)__pyx_t_9);

#line 102 "ms_deisotope/_c/utils.pyx"
  __pyx_t_9 = 0;

  /* "ms_deisotope/_c/utils.pyx":103
 *         point_count += (1 + envelope.get_size()) * 2
 *     collection = np.zeros(point_count, dtype=np.float32)
 *     k = 0             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         k += 2
 */

#line 103 "ms_deisotope/_c/utils.pyx"
  __pyx_v_k = 0;

  /* "ms_deisotope/_c/utils.pyx":104
 *     collection = np.zeros(point_count, dtype=np.float32)
 *     k = 0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         k += 2
 *         envelope = <Envelope>PyList_GET_ITEM(envelope_list, i)
 */

#line 104 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 104 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 104 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 104 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/utils.pyx":105
 *     k = 0
 *     for i in range(n):
 *         k += 2             # <<<<<<<<<<<<<<
 *         envelope = <Envelope>PyList_GET_ITEM(envelope_list, i)
 *         m = envelope.get_size()
 */

#line 105 "ms_deisotope/_c/utils.pyx"
    __pyx_v_k = (__pyx_v_k + 2);

    /* "ms_deisotope/_c/utils.pyx":106
 *     for i in range(n):
 *         k += 2
 *         envelope = <Envelope>PyList_GET_ITEM(envelope_list, i)             # <<<<<<<<<<<<<<
 *         m = envelope.get_size()
 *         for j in  range(m):
 */

#line 106 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_envelope_list, __pyx_v_i);

#line 106 "ms_deisotope/_c/utils.pyx"
    __pyx_t_9 = ((PyObject *)__pyx_t_4);

#line 106 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_t_9);

#line 106 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_envelope, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *)__pyx_t_9));

#line 106 "ms_deisotope/_c/utils.pyx"
    __pyx_t_9 = 0;

    /* "ms_deisotope/_c/utils.pyx":107
 *         k += 2
 *         envelope = <Envelope>PyList_GET_ITEM(envelope_list, i)
 *         m = envelope.get_size()             # <<<<<<<<<<<<<<
 *         for j in  range(m):
 *             pair = envelope.getitem(j)
 */

#line 107 "ms_deisotope/_c/utils.pyx"
    __pyx_v_m = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *)__pyx_v_envelope->__pyx_vtab)->get_size(__pyx_v_envelope);

    /* "ms_deisotope/_c/utils.pyx":108
 *         envelope = <Envelope>PyList_GET_ITEM(envelope_list, i)
 *         m = envelope.get_size()
 *         for j in  range(m):             # <<<<<<<<<<<<<<
 *             pair = envelope.getitem(j)
 *             collection[k] = pair.mz
 */

#line 108 "ms_deisotope/_c/utils.pyx"
    __pyx_t_15 = __pyx_v_m;

#line 108 "ms_deisotope/_c/utils.pyx"
    __pyx_t_16 = __pyx_t_15;

#line 108 "ms_deisotope/_c/utils.pyx"
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {

#line 108 "ms_deisotope/_c/utils.pyx"
      __pyx_v_j = __pyx_t_17;

      /* "ms_deisotope/_c/utils.pyx":109
 *         m = envelope.get_size()
 *         for j in  range(m):
 *             pair = envelope.getitem(j)             # <<<<<<<<<<<<<<
 *             collection[k] = pair.mz
 *             k += 1
 */

#line 109 "ms_deisotope/_c/utils.pyx"
      __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *)__pyx_v_envelope->__pyx_vtab)->getitem(__pyx_v_envelope, __pyx_v_j)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 109, __pyx_L1_error)

#line 109 "ms_deisotope/_c/utils.pyx"
      __Pyx_GOTREF(__pyx_t_9);

#line 109 "ms_deisotope/_c/utils.pyx"
      __Pyx_XDECREF_SET(__pyx_v_pair, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *)__pyx_t_9));

#line 109 "ms_deisotope/_c/utils.pyx"
      __pyx_t_9 = 0;

      /* "ms_deisotope/_c/utils.pyx":110
 *         for j in  range(m):
 *             pair = envelope.getitem(j)
 *             collection[k] = pair.mz             # <<<<<<<<<<<<<<
 *             k += 1
 *             collection[k] = pair.intensity
 */

#line 110 "ms_deisotope/_c/utils.pyx"
      __pyx_t_18 = __pyx_v_pair->mz;

#line 110 "ms_deisotope/_c/utils.pyx"
      __pyx_t_19 = __pyx_v_k;

#line 110 "ms_deisotope/_c/utils.pyx"
      __pyx_t_11 = -1;

#line 110 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__pyx_t_19 >= (size_t)__pyx_pybuffernd_collection.diminfo[0].shape)) __pyx_t_11 = 0;

#line 110 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__pyx_t_11 != -1)) {

#line 110 "ms_deisotope/_c/utils.pyx"
        __Pyx_RaiseBufferIndexError(__pyx_t_11);

#line 110 "ms_deisotope/_c/utils.pyx"
        __PYX_ERR(0, 110, __pyx_L1_error)

#line 110 "ms_deisotope/_c/utils.pyx"
      }

#line 110 "ms_deisotope/_c/utils.pyx"
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float32_t *, __pyx_pybuffernd_collection.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_collection.diminfo[0].strides) = __pyx_t_18;

      /* "ms_deisotope/_c/utils.pyx":111
 *             pair = envelope.getitem(j)
 *             collection[k] = pair.mz
 *             k += 1             # <<<<<<<<<<<<<<
 *             collection[k] = pair.intensity
 *             k += 1
 */

#line 111 "ms_deisotope/_c/utils.pyx"
      __pyx_v_k = (__pyx_v_k + 1);

      /* "ms_deisotope/_c/utils.pyx":112
 *             collection[k] = pair.mz
 *             k += 1
 *             collection[k] = pair.intensity             # <<<<<<<<<<<<<<
 *             k += 1
 *     return collection
 */

#line 112 "ms_deisotope/_c/utils.pyx"
      __pyx_t_18 = __pyx_v_pair->intensity;

#line 112 "ms_deisotope/_c/utils.pyx"
      __pyx_t_19 = __pyx_v_k;

#line 112 "ms_deisotope/_c/utils.pyx"
      __pyx_t_11 = -1;

#line 112 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__pyx_t_19 >= (size_t)__pyx_pybuffernd_collection.diminfo[0].shape)) __pyx_t_11 = 0;

#line 112 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__pyx_t_11 != -1)) {

#line 112 "ms_deisotope/_c/utils.pyx"
        __Pyx_RaiseBufferIndexError(__pyx_t_11);

#line 112 "ms_deisotope/_c/utils.pyx"
        __PYX_ERR(0, 112, __pyx_L1_error)

#line 112 "ms_deisotope/_c/utils.pyx"
      }

#line 112 "ms_deisotope/_c/utils.pyx"
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float32_t *, __pyx_pybuffernd_collection.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_collection.diminfo[0].strides) = __pyx_t_18;

      /* "ms_deisotope/_c/utils.pyx":113
 *             k += 1
 *             collection[k] = pair.intensity
 *             k += 1             # <<<<<<<<<<<<<<
 *     return collection
 * 
 */

#line 113 "ms_deisotope/_c/utils.pyx"
      __pyx_v_k = (__pyx_v_k + 1);

#line 113 "ms_deisotope/_c/utils.pyx"
    }

#line 113 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":114
 *             collection[k] = pair.intensity
 *             k += 1
 *     return collection             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 114 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 114 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_collection));

#line 114 "ms_deisotope/_c/utils.pyx"
  __pyx_r = ((PyArrayObject *)__pyx_v_collection);

#line 114 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":91
 * 
 * 
 * cpdef np.ndarray[np.float32_t] envelopes_to_array(list envelope_list):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t n, m, i, j, point_count, k
 */

#line 91 "ms_deisotope/_c/utils.pyx"


#line 91 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_9);

#line 91 "ms_deisotope/_c/utils.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 91 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_declare

#line 91 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_assign

#line 91 "ms_deisotope/_c/utils.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 91 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_collection.rcbuffer->pybuffer);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.envelopes_to_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 91 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L2;

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_collection.rcbuffer->pybuffer);

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_L2:;

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_collection);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_envelope);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_pair);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 91 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 91 "ms_deisotope/_c/utils.pyx"
}

#line 91 "ms_deisotope/_c/utils.pyx"


#line 91 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 91 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_9envelopes_to_array(PyObject *__pyx_self, PyObject *__pyx_v_envelope_list); /*proto*/

#line 91 "ms_deisotope/_c/utils.pyx"
static char __pyx_doc_12ms_deisotope_2_c_5utils_8envelopes_to_array[] = "envelopes_to_array(list envelope_list) -> ndarray";

#line 91 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_9envelopes_to_array(PyObject *__pyx_self, PyObject *__pyx_v_envelope_list) {

#line 91 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 91 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("envelopes_to_array (wrapper)", 0);

#line 91 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_envelope_list), (&PyList_Type), 1, "envelope_list", 1))) __PYX_ERR(0, 91, __pyx_L1_error)
  __pyx_r = 
#line 91 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_8envelopes_to_array(__pyx_self, ((PyObject*)__pyx_v_envelope_list));

#line 91 "ms_deisotope/_c/utils.pyx"


#line 91 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 91 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 91 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 91 "ms_deisotope/_c/utils.pyx"
}

#line 91 "ms_deisotope/_c/utils.pyx"


#line 91 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_8envelopes_to_array(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_envelope_list) {

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 91 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 91 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("envelopes_to_array", 0);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_5utils_envelopes_to_array(__pyx_v_envelope_list, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 91 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 91 "ms_deisotope/_c/utils.pyx"


#line 91 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.envelopes_to_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 91 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 91 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 91 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 91 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":118
 * 
 * @cython.boundscheck(False)
 * cpdef DeconvolutedPeakSetIndexed deserialize_deconvoluted_peak_set(dict scan_dict, bint include_envelopes=True):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list envelopes, peaks
 */

#line 118 "ms_deisotope/_c/utils.pyx"


#line 118 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_11deserialize_deconvoluted_peak_set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 118 "ms_deisotope/_c/utils.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_f_12ms_deisotope_2_c_5utils_deserialize_deconvoluted_peak_set(PyObject *__pyx_v_scan_dict, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_5utils_deserialize_deconvoluted_peak_set *__pyx_optional_args) {

#line 118 "ms_deisotope/_c/utils.pyx"
  int __pyx_v_include_envelopes = ((int)1);
  PyObject *__pyx_v_envelopes = 0
#line 118 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_peaks = 0
#line 118 "ms_deisotope/_c/utils.pyx"
;
  PyArrayObject *__pyx_v_mz_array = 0
#line 118 "ms_deisotope/_c/utils.pyx"
;
  PyArrayObject *__pyx_v_intensity_array = 0
#line 118 "ms_deisotope/_c/utils.pyx"
;
  PyArrayObject *__pyx_v_charge_array = 0
#line 118 "ms_deisotope/_c/utils.pyx"
;
  PyArrayObject *__pyx_v_score_array = 0
#line 118 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 118 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_i
#line 118 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_mz
#line 118 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_peak_neutral_mass
#line 118 "ms_deisotope/_c/utils.pyx"
;
  int __pyx_v_charge
#line 118 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_peak = 0
#line 118 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_v_peak_set = 0
#line 118 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_charge_array
#line 118 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_charge_array
#line 118 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_intensity_array
#line 118 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_intensity_array
#line 118 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mz_array
#line 118 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_mz_array
#line 118 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_score_array
#line 118 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_score_array
#line 118 "ms_deisotope/_c/utils.pyx"
;

#line 118 "ms_deisotope/_c/utils.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_r = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 118 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_2;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyArrayObject *__pyx_t_6 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_7;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyArrayObject *__pyx_t_11 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyArrayObject *__pyx_t_12 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyArrayObject *__pyx_t_13 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_14;

#line 118 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_15;

#line 118 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_16;

#line 118 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_17;

#line 118 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_18;

#line 118 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_19;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_20;

#line 118 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_21;

#line 118 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("deserialize_deconvoluted_peak_set", 0);

#line 118 "ms_deisotope/_c/utils.pyx"
  if (__pyx_optional_args) {

#line 118 "ms_deisotope/_c/utils.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 118 "ms_deisotope/_c/utils.pyx"
      __pyx_v_include_envelopes = __pyx_optional_args->include_envelopes;

#line 118 "ms_deisotope/_c/utils.pyx"
    }

#line 118 "ms_deisotope/_c/utils.pyx"
  }

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_mz_array.pybuffer.buf = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_mz_array.refcount = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_mz_array.data = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_mz_array.rcbuffer = &__pyx_pybuffer_mz_array;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_intensity_array.pybuffer.buf = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_intensity_array.refcount = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_intensity_array.data = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_intensity_array.rcbuffer = &__pyx_pybuffer_intensity_array;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_charge_array.pybuffer.buf = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_charge_array.refcount = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_charge_array.data = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_charge_array.rcbuffer = &__pyx_pybuffer_charge_array;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_score_array.pybuffer.buf = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_score_array.refcount = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_score_array.data = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_score_array.rcbuffer = &__pyx_pybuffer_score_array;

  /* "ms_deisotope/_c/utils.pyx":131
 *         DeconvolutedPeakSetIndexed peak_set
 * 
 *     peaks = []             # <<<<<<<<<<<<<<
 *     if include_envelopes:
 *         envelopes = decode_envelopes(scan_dict["isotopic envelopes array"])
 */

#line 131 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)

#line 131 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 131 "ms_deisotope/_c/utils.pyx"
  __pyx_v_peaks = ((PyObject*)__pyx_t_1);

#line 131 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":132
 * 
 *     peaks = []
 *     if include_envelopes:             # <<<<<<<<<<<<<<
 *         envelopes = decode_envelopes(scan_dict["isotopic envelopes array"])
 *     else:
 */

#line 132 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = (__pyx_v_include_envelopes != 0);

#line 132 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/utils.pyx":133
 *     peaks = []
 *     if include_envelopes:
 *         envelopes = decode_envelopes(scan_dict["isotopic envelopes array"])             # <<<<<<<<<<<<<<
 *     else:
 *         envelopes = None
 */

#line 133 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_scan_dict == Py_None)) {

#line 133 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 133 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "ms_deisotope/_c/utils.pyx"
    }

#line 133 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_scan_dict, __pyx_kp_s_isotopic_envelopes_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 133 "ms_deisotope/_c/utils.pyx"
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_5utils_decode_envelopes(((PyArrayObject *)__pyx_t_1), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 133 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 133 "ms_deisotope/_c/utils.pyx"
    __pyx_v_envelopes = ((PyObject*)__pyx_t_3);

#line 133 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/utils.pyx":132
 * 
 *     peaks = []
 *     if include_envelopes:             # <<<<<<<<<<<<<<
 *         envelopes = decode_envelopes(scan_dict["isotopic envelopes array"])
 *     else:
 */

#line 132 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L3;

#line 132 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":135
 *         envelopes = decode_envelopes(scan_dict["isotopic envelopes array"])
 *     else:
 *         envelopes = None             # <<<<<<<<<<<<<<
 *     mz_array = scan_dict['m/z array'].astype(np.float64)
 *     intensity_array = scan_dict['intensity array'].astype(np.float64)
 */

#line 135 "ms_deisotope/_c/utils.pyx"
  /*else*/ {

#line 135 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 135 "ms_deisotope/_c/utils.pyx"
    __pyx_v_envelopes = ((PyObject*)Py_None);

#line 135 "ms_deisotope/_c/utils.pyx"
  }

#line 135 "ms_deisotope/_c/utils.pyx"
  __pyx_L3:;

  /* "ms_deisotope/_c/utils.pyx":136
 *     else:
 *         envelopes = None
 *     mz_array = scan_dict['m/z array'].astype(np.float64)             # <<<<<<<<<<<<<<
 *     intensity_array = scan_dict['intensity array'].astype(np.float64)
 *     charge_array = scan_dict['charge array'].astype(np.int8)
 */

#line 136 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_scan_dict == Py_None)) {

#line 136 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 136 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 136, __pyx_L1_error)

#line 136 "ms_deisotope/_c/utils.pyx"
  }

#line 136 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_scan_dict, __pyx_kp_s_m_z_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)

#line 136 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 136 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_astype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 136, __pyx_L1_error)

#line 136 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 136 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 136 "ms_deisotope/_c/utils.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)

#line 136 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 136 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 136, __pyx_L1_error)

#line 136 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 136 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 136 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = NULL;

#line 136 "ms_deisotope/_c/utils.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {

#line 136 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);

#line 136 "ms_deisotope/_c/utils.pyx"
    if (likely(__pyx_t_1)) {

#line 136 "ms_deisotope/_c/utils.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 136 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 136 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(function);

#line 136 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_t_4, function);

#line 136 "ms_deisotope/_c/utils.pyx"
    }

#line 136 "ms_deisotope/_c/utils.pyx"
  }

#line 136 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);

#line 136 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 136 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 136 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 136, __pyx_L1_error)

#line 136 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 136 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 136 "ms_deisotope/_c/utils.pyx"
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 136, __pyx_L1_error)

#line 136 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_3);

#line 136 "ms_deisotope/_c/utils.pyx"
  {

#line 136 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 136 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer);

#line 136 "ms_deisotope/_c/utils.pyx"
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);

#line 136 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 < 0)) {

#line 136 "ms_deisotope/_c/utils.pyx"
      PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);

#line 136 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_mz_array, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {

#line 136 "ms_deisotope/_c/utils.pyx"
        Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);

#line 136 "ms_deisotope/_c/utils.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 136 "ms_deisotope/_c/utils.pyx"
      } else {

#line 136 "ms_deisotope/_c/utils.pyx"
        PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);

#line 136 "ms_deisotope/_c/utils.pyx"
      }

#line 136 "ms_deisotope/_c/utils.pyx"
      __pyx_t_8 = __pyx_t_9 = __pyx_t_10 = 0;

#line 136 "ms_deisotope/_c/utils.pyx"
    }

#line 136 "ms_deisotope/_c/utils.pyx"
    __pyx_pybuffernd_mz_array.diminfo[0].strides = __pyx_pybuffernd_mz_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mz_array.diminfo[0].shape = __pyx_pybuffernd_mz_array.rcbuffer->pybuffer.shape[0];

#line 136 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 136, __pyx_L1_error)

#line 136 "ms_deisotope/_c/utils.pyx"
  }

#line 136 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = 0;

#line 136 "ms_deisotope/_c/utils.pyx"
  __pyx_v_mz_array = ((PyArrayObject *)__pyx_t_3);

#line 136 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/utils.pyx":137
 *         envelopes = None
 *     mz_array = scan_dict['m/z array'].astype(np.float64)
 *     intensity_array = scan_dict['intensity array'].astype(np.float64)             # <<<<<<<<<<<<<<
 *     charge_array = scan_dict['charge array'].astype(np.int8)
 *     score_array = scan_dict['deconvolution score array'].astype(np.float64)
 */

#line 137 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_scan_dict == Py_None)) {

#line 137 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 137 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 137, __pyx_L1_error)

#line 137 "ms_deisotope/_c/utils.pyx"
  }

#line 137 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_scan_dict, __pyx_kp_s_intensity_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 137, __pyx_L1_error)

#line 137 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 137 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_astype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 137, __pyx_L1_error)

#line 137 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 137 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 137 "ms_deisotope/_c/utils.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 137, __pyx_L1_error)

#line 137 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 137 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)

#line 137 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 137 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 137 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = NULL;

#line 137 "ms_deisotope/_c/utils.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {

#line 137 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);

#line 137 "ms_deisotope/_c/utils.pyx"
    if (likely(__pyx_t_4)) {

#line 137 "ms_deisotope/_c/utils.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 137 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 137 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(function);

#line 137 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_t_5, function);

#line 137 "ms_deisotope/_c/utils.pyx"
    }

#line 137 "ms_deisotope/_c/utils.pyx"
  }

#line 137 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1);

#line 137 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 137 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 137 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)

#line 137 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 137 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 137 "ms_deisotope/_c/utils.pyx"
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 137, __pyx_L1_error)

#line 137 "ms_deisotope/_c/utils.pyx"
  __pyx_t_11 = ((PyArrayObject *)__pyx_t_3);

#line 137 "ms_deisotope/_c/utils.pyx"
  {

#line 137 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 137 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer);

#line 137 "ms_deisotope/_c/utils.pyx"
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);

#line 137 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 < 0)) {

#line 137 "ms_deisotope/_c/utils.pyx"
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);

#line 137 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_intensity_array, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {

#line 137 "ms_deisotope/_c/utils.pyx"
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8);

#line 137 "ms_deisotope/_c/utils.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 137 "ms_deisotope/_c/utils.pyx"
      } else {

#line 137 "ms_deisotope/_c/utils.pyx"
        PyErr_Restore(__pyx_t_10, __pyx_t_9, __pyx_t_8);

#line 137 "ms_deisotope/_c/utils.pyx"
      }

#line 137 "ms_deisotope/_c/utils.pyx"
      __pyx_t_10 = __pyx_t_9 = __pyx_t_8 = 0;

#line 137 "ms_deisotope/_c/utils.pyx"
    }

#line 137 "ms_deisotope/_c/utils.pyx"
    __pyx_pybuffernd_intensity_array.diminfo[0].strides = __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_intensity_array.diminfo[0].shape = __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.shape[0];

#line 137 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 137, __pyx_L1_error)

#line 137 "ms_deisotope/_c/utils.pyx"
  }

#line 137 "ms_deisotope/_c/utils.pyx"
  __pyx_t_11 = 0;

#line 137 "ms_deisotope/_c/utils.pyx"
  __pyx_v_intensity_array = ((PyArrayObject *)__pyx_t_3);

#line 137 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/utils.pyx":138
 *     mz_array = scan_dict['m/z array'].astype(np.float64)
 *     intensity_array = scan_dict['intensity array'].astype(np.float64)
 *     charge_array = scan_dict['charge array'].astype(np.int8)             # <<<<<<<<<<<<<<
 *     score_array = scan_dict['deconvolution score array'].astype(np.float64)
 *     n = mz_array.shape[0]
 */

#line 138 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_scan_dict == Py_None)) {

#line 138 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 138 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 138, __pyx_L1_error)

#line 138 "ms_deisotope/_c/utils.pyx"
  }

#line 138 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_scan_dict, __pyx_kp_s_charge_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 138, __pyx_L1_error)

#line 138 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 138 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 138, __pyx_L1_error)

#line 138 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 138 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 138 "ms_deisotope/_c/utils.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 138, __pyx_L1_error)

#line 138 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 138 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 138, __pyx_L1_error)

#line 138 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 138 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 138 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = NULL;

#line 138 "ms_deisotope/_c/utils.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 138 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 138 "ms_deisotope/_c/utils.pyx"
    if (likely(__pyx_t_5)) {

#line 138 "ms_deisotope/_c/utils.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 138 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(__pyx_t_5);

#line 138 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(function);

#line 138 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 138 "ms_deisotope/_c/utils.pyx"
    }

#line 138 "ms_deisotope/_c/utils.pyx"
  }

#line 138 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4);

#line 138 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 138 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 138 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 138, __pyx_L1_error)

#line 138 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 138 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 138 "ms_deisotope/_c/utils.pyx"
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 138, __pyx_L1_error)

#line 138 "ms_deisotope/_c/utils.pyx"
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_3);

#line 138 "ms_deisotope/_c/utils.pyx"
  {

#line 138 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 138 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_charge_array.rcbuffer->pybuffer);

#line 138 "ms_deisotope/_c/utils.pyx"
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_charge_array.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);

#line 138 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 < 0)) {

#line 138 "ms_deisotope/_c/utils.pyx"
      PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);

#line 138 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_charge_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_charge_array, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {

#line 138 "ms_deisotope/_c/utils.pyx"
        Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);

#line 138 "ms_deisotope/_c/utils.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 138 "ms_deisotope/_c/utils.pyx"
      } else {

#line 138 "ms_deisotope/_c/utils.pyx"
        PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);

#line 138 "ms_deisotope/_c/utils.pyx"
      }

#line 138 "ms_deisotope/_c/utils.pyx"
      __pyx_t_8 = __pyx_t_9 = __pyx_t_10 = 0;

#line 138 "ms_deisotope/_c/utils.pyx"
    }

#line 138 "ms_deisotope/_c/utils.pyx"
    __pyx_pybuffernd_charge_array.diminfo[0].strides = __pyx_pybuffernd_charge_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_charge_array.diminfo[0].shape = __pyx_pybuffernd_charge_array.rcbuffer->pybuffer.shape[0];

#line 138 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 138, __pyx_L1_error)

#line 138 "ms_deisotope/_c/utils.pyx"
  }

#line 138 "ms_deisotope/_c/utils.pyx"
  __pyx_t_12 = 0;

#line 138 "ms_deisotope/_c/utils.pyx"
  __pyx_v_charge_array = ((PyArrayObject *)__pyx_t_3);

#line 138 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/utils.pyx":139
 *     intensity_array = scan_dict['intensity array'].astype(np.float64)
 *     charge_array = scan_dict['charge array'].astype(np.int8)
 *     score_array = scan_dict['deconvolution score array'].astype(np.float64)             # <<<<<<<<<<<<<<
 *     n = mz_array.shape[0]
 *     i = 0
 */

#line 139 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_scan_dict == Py_None)) {

#line 139 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 139 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/utils.pyx"
  }

#line 139 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_scan_dict, __pyx_kp_s_deconvolution_score_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 139 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_astype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 139 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 139 "ms_deisotope/_c/utils.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 139 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 139 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 139 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = NULL;

#line 139 "ms_deisotope/_c/utils.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {

#line 139 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);

#line 139 "ms_deisotope/_c/utils.pyx"
    if (likely(__pyx_t_1)) {

#line 139 "ms_deisotope/_c/utils.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 139 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 139 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(function);

#line 139 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_t_4, function);

#line 139 "ms_deisotope/_c/utils.pyx"
    }

#line 139 "ms_deisotope/_c/utils.pyx"
  }

#line 139 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);

#line 139 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 139 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 139 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 139 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 139 "ms_deisotope/_c/utils.pyx"
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/utils.pyx"
  __pyx_t_13 = ((PyArrayObject *)__pyx_t_3);

#line 139 "ms_deisotope/_c/utils.pyx"
  {

#line 139 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 139 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_score_array.rcbuffer->pybuffer);

#line 139 "ms_deisotope/_c/utils.pyx"
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_score_array.rcbuffer->pybuffer, (PyObject*)__pyx_t_13, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);

#line 139 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 < 0)) {

#line 139 "ms_deisotope/_c/utils.pyx"
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);

#line 139 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_score_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_score_array, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {

#line 139 "ms_deisotope/_c/utils.pyx"
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8);

#line 139 "ms_deisotope/_c/utils.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 139 "ms_deisotope/_c/utils.pyx"
      } else {

#line 139 "ms_deisotope/_c/utils.pyx"
        PyErr_Restore(__pyx_t_10, __pyx_t_9, __pyx_t_8);

#line 139 "ms_deisotope/_c/utils.pyx"
      }

#line 139 "ms_deisotope/_c/utils.pyx"
      __pyx_t_10 = __pyx_t_9 = __pyx_t_8 = 0;

#line 139 "ms_deisotope/_c/utils.pyx"
    }

#line 139 "ms_deisotope/_c/utils.pyx"
    __pyx_pybuffernd_score_array.diminfo[0].strides = __pyx_pybuffernd_score_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_score_array.diminfo[0].shape = __pyx_pybuffernd_score_array.rcbuffer->pybuffer.shape[0];

#line 139 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/utils.pyx"
  }

#line 139 "ms_deisotope/_c/utils.pyx"
  __pyx_t_13 = 0;

#line 139 "ms_deisotope/_c/utils.pyx"
  __pyx_v_score_array = ((PyArrayObject *)__pyx_t_3);

#line 139 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/utils.pyx":140
 *     charge_array = scan_dict['charge array'].astype(np.int8)
 *     score_array = scan_dict['deconvolution score array'].astype(np.float64)
 *     n = mz_array.shape[0]             # <<<<<<<<<<<<<<
 *     i = 0
 *     for i in range(n):
 */

#line 140 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = (__pyx_v_mz_array->dimensions[0]);

  /* "ms_deisotope/_c/utils.pyx":141
 *     score_array = scan_dict['deconvolution score array'].astype(np.float64)
 *     n = mz_array.shape[0]
 *     i = 0             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         mz = mz_array[i]
 */

#line 141 "ms_deisotope/_c/utils.pyx"
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/utils.pyx":142
 *     n = mz_array.shape[0]
 *     i = 0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         mz = mz_array[i]
 *         charge = charge_array[i]
 */

#line 142 "ms_deisotope/_c/utils.pyx"
  __pyx_t_14 = __pyx_v_n;

#line 142 "ms_deisotope/_c/utils.pyx"
  __pyx_t_15 = __pyx_t_14;

#line 142 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {

#line 142 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_16;

    /* "ms_deisotope/_c/utils.pyx":143
 *     i = 0
 *     for i in range(n):
 *         mz = mz_array[i]             # <<<<<<<<<<<<<<
 *         charge = charge_array[i]
 *         peak_neutral_mass = neutral_mass(mz, charge)
 */

#line 143 "ms_deisotope/_c/utils.pyx"
    __pyx_t_17 = __pyx_v_i;

#line 143 "ms_deisotope/_c/utils.pyx"
    __pyx_v_mz = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mz_array.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_mz_array.diminfo[0].strides));

    /* "ms_deisotope/_c/utils.pyx":144
 *     for i in range(n):
 *         mz = mz_array[i]
 *         charge = charge_array[i]             # <<<<<<<<<<<<<<
 *         peak_neutral_mass = neutral_mass(mz, charge)
 * 
 */

#line 144 "ms_deisotope/_c/utils.pyx"
    __pyx_t_17 = __pyx_v_i;

#line 144 "ms_deisotope/_c/utils.pyx"
    __pyx_v_charge = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int8_t *, __pyx_pybuffernd_charge_array.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_charge_array.diminfo[0].strides));

    /* "ms_deisotope/_c/utils.pyx":145
 *         mz = mz_array[i]
 *         charge = charge_array[i]
 *         peak_neutral_mass = neutral_mass(mz, charge)             # <<<<<<<<<<<<<<
 * 
 *         peak = DeconvolutedPeak._create_simple(
 */

#line 145 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peak_neutral_mass = __pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass(__pyx_v_mz, __pyx_v_charge, NULL);

    /* "ms_deisotope/_c/utils.pyx":149
 *         peak = DeconvolutedPeak._create_simple(
 *             peak_neutral_mass,
 *             intensity_array[i],             # <<<<<<<<<<<<<<
 *             charge_array[i],
 *             score_array[i],
 */

#line 149 "ms_deisotope/_c/utils.pyx"
    __pyx_t_17 = __pyx_v_i;

    /* "ms_deisotope/_c/utils.pyx":150
 *             peak_neutral_mass,
 *             intensity_array[i],
 *             charge_array[i],             # <<<<<<<<<<<<<<
 *             score_array[i],
 *             mz,
 */

#line 150 "ms_deisotope/_c/utils.pyx"
    __pyx_t_18 = __pyx_v_i;

    /* "ms_deisotope/_c/utils.pyx":151
 *             intensity_array[i],
 *             charge_array[i],
 *             score_array[i],             # <<<<<<<<<<<<<<
 *             mz,
 *             <Envelope>PyList_GET_ITEM(envelopes, i) if include_envelopes else None
 */

#line 151 "ms_deisotope/_c/utils.pyx"
    __pyx_t_19 = __pyx_v_i;

    /* "ms_deisotope/_c/utils.pyx":153
 *             score_array[i],
 *             mz,
 *             <Envelope>PyList_GET_ITEM(envelopes, i) if include_envelopes else None             # <<<<<<<<<<<<<<
 *         )
 *         peaks.append(peak)
 */

#line 153 "ms_deisotope/_c/utils.pyx"
    if ((__pyx_v_include_envelopes != 0)) {

#line 153 "ms_deisotope/_c/utils.pyx"
      __pyx_t_20 = PyList_GET_ITEM(__pyx_v_envelopes, __pyx_v_i);

#line 153 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(((PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *)__pyx_t_20)));

#line 153 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = ((PyObject *)__pyx_t_20);

#line 153 "ms_deisotope/_c/utils.pyx"
    } else {

#line 153 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(Py_None);

#line 153 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = Py_None;

#line 153 "ms_deisotope/_c/utils.pyx"
    }

    /* "ms_deisotope/_c/utils.pyx":147
 *         peak_neutral_mass = neutral_mass(mz, charge)
 * 
 *         peak = DeconvolutedPeak._create_simple(             # <<<<<<<<<<<<<<
 *             peak_neutral_mass,
 *             intensity_array[i],
 */

#line 147 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak->_create_simple(__pyx_v_peak_neutral_mass, (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_intensity_array.diminfo[0].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int8_t *, __pyx_pybuffernd_charge_array.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_charge_array.diminfo[0].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_score_array.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_score_array.diminfo[0].strides)), __pyx_v_mz, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *)__pyx_t_3))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)

#line 147 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 147 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 147 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_4));

#line 147 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/utils.pyx":155
 *             <Envelope>PyList_GET_ITEM(envelopes, i) if include_envelopes else None
 *         )
 *         peaks.append(peak)             # <<<<<<<<<<<<<<
 *     peak_set = DeconvolutedPeakSetIndexed(peaks)
 *     peak_set.reindex()
 */

#line 155 "ms_deisotope/_c/utils.pyx"
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_peaks, ((PyObject *)__pyx_v_peak)); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 155, __pyx_L1_error)

#line 155 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":156
 *         )
 *         peaks.append(peak)
 *     peak_set = DeconvolutedPeakSetIndexed(peaks)             # <<<<<<<<<<<<<<
 *     peak_set.reindex()
 *     return peak_set
 */

#line 156 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed), __pyx_v_peaks); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 156, __pyx_L1_error)

#line 156 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 156 "ms_deisotope/_c/utils.pyx"
  __pyx_v_peak_set = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *)__pyx_t_4);

#line 156 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/utils.pyx":157
 *         peaks.append(peak)
 *     peak_set = DeconvolutedPeakSetIndexed(peaks)
 *     peak_set.reindex()             # <<<<<<<<<<<<<<
 *     return peak_set
 * 
 */

#line 157 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *)__pyx_v_peak_set->__pyx_base.__pyx_vtab)->__pyx_base.reindex(((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set), 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 157, __pyx_L1_error)

#line 157 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 157 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "ms_deisotope/_c/utils.pyx":158
 *     peak_set = DeconvolutedPeakSetIndexed(peaks)
 *     peak_set.reindex()
 *     return peak_set             # <<<<<<<<<<<<<<
 * 
 * cpdef PeakIndex deserialize_peak_set(dict scan_dict):
 */

#line 158 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 158 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set));

#line 158 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_peak_set;

#line 158 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":118
 * 
 * @cython.boundscheck(False)
 * cpdef DeconvolutedPeakSetIndexed deserialize_deconvoluted_peak_set(dict scan_dict, bint include_envelopes=True):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list envelopes, peaks
 */

#line 118 "ms_deisotope/_c/utils.pyx"


#line 118 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 118 "ms_deisotope/_c/utils.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 118 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_declare

#line 118 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_assign

#line 118 "ms_deisotope/_c/utils.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 118 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_charge_array.rcbuffer->pybuffer);

#line 118 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer);

#line 118 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer);

#line 118 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_score_array.rcbuffer->pybuffer);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.deserialize_deconvoluted_peak_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L2;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_charge_array.rcbuffer->pybuffer);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_score_array.rcbuffer->pybuffer);

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_L2:;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_envelopes);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_peaks);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_mz_array);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_intensity_array);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_charge_array);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_score_array);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak_set);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 118 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 118 "ms_deisotope/_c/utils.pyx"
}

#line 118 "ms_deisotope/_c/utils.pyx"


#line 118 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 118 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_11deserialize_deconvoluted_peak_set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 118 "ms_deisotope/_c/utils.pyx"
static char __pyx_doc_12ms_deisotope_2_c_5utils_10deserialize_deconvoluted_peak_set[] = "deserialize_deconvoluted_peak_set(dict scan_dict, bool include_envelopes=True) -> DeconvolutedPeakSetIndexed";

#line 118 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_11deserialize_deconvoluted_peak_set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_scan_dict = 0
#line 118 "ms_deisotope/_c/utils.pyx"
;
  int __pyx_v_include_envelopes
#line 118 "ms_deisotope/_c/utils.pyx"
;

#line 118 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("deserialize_deconvoluted_peak_set (wrapper)", 0);

#line 118 "ms_deisotope/_c/utils.pyx"
  {

#line 118 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_scan_dict,&__pyx_n_s_include_envelopes,0};

#line 118 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 118 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 118 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 118 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 118 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 118 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 118 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 118 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 118 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 118 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 118 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 118 "ms_deisotope/_c/utils.pyx"
      }

#line 118 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 118 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 118 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 118 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scan_dict)) != 0)) kw_args--;
        else 
#line 118 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 118 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 118 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 118 "ms_deisotope/_c/utils.pyx"
        if (kw_args > 0) {

#line 118 "ms_deisotope/_c/utils.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_include_envelopes);

#line 118 "ms_deisotope/_c/utils.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 118 "ms_deisotope/_c/utils.pyx"
        }

#line 118 "ms_deisotope/_c/utils.pyx"
      }

#line 118 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 118 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "deserialize_deconvoluted_peak_set") < 0)) __PYX_ERR(0, 118, __pyx_L3_error)

#line 118 "ms_deisotope/_c/utils.pyx"
      }

#line 118 "ms_deisotope/_c/utils.pyx"
    } else {

#line 118 "ms_deisotope/_c/utils.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: 
#line 118 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 118 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 118 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 118 "ms_deisotope/_c/utils.pyx"
        break;
        default: 
#line 118 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 118 "ms_deisotope/_c/utils.pyx"
      }

#line 118 "ms_deisotope/_c/utils.pyx"
    }

#line 118 "ms_deisotope/_c/utils.pyx"
    __pyx_v_scan_dict = ((PyObject*)values[0]);

#line 118 "ms_deisotope/_c/utils.pyx"
    if (values[1]) {

#line 118 "ms_deisotope/_c/utils.pyx"
      __pyx_v_include_envelopes = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_include_envelopes == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 118, __pyx_L3_error)

#line 118 "ms_deisotope/_c/utils.pyx"
    } else {

#line 118 "ms_deisotope/_c/utils.pyx"
      __pyx_v_include_envelopes = ((int)1);

#line 118 "ms_deisotope/_c/utils.pyx"
    }

#line 118 "ms_deisotope/_c/utils.pyx"
  }

#line 118 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("deserialize_deconvoluted_peak_set", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 118 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 118, __pyx_L3_error)

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.deserialize_deconvoluted_peak_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 118 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 118 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_scan_dict), (&PyDict_Type), 1, "scan_dict", 1))) __PYX_ERR(0, 118, __pyx_L1_error)
  __pyx_r = 
#line 118 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_10deserialize_deconvoluted_peak_set(__pyx_self, __pyx_v_scan_dict, __pyx_v_include_envelopes);

#line 118 "ms_deisotope/_c/utils.pyx"


#line 118 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 118 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 118 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 118 "ms_deisotope/_c/utils.pyx"
}

#line 118 "ms_deisotope/_c/utils.pyx"


#line 118 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_10deserialize_deconvoluted_peak_set(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_scan_dict, int __pyx_v_include_envelopes) {

#line 118 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 118 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_5utils_deserialize_deconvoluted_peak_set __pyx_t_2;

#line 118 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("deserialize_deconvoluted_peak_set", 0);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2.include_envelopes = __pyx_v_include_envelopes;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_5utils_deserialize_deconvoluted_peak_set(__pyx_v_scan_dict, 0, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 118 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 118 "ms_deisotope/_c/utils.pyx"


#line 118 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.deserialize_deconvoluted_peak_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 118 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 118 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 118 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 118 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":160
 *     return peak_set
 * 
 * cpdef PeakIndex deserialize_peak_set(dict scan_dict):             # <<<<<<<<<<<<<<
 *     cdef:
 *         np.ndarray[np.float64_t] mz_array
 */

#line 160 "ms_deisotope/_c/utils.pyx"


#line 160 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_13deserialize_peak_set(PyObject *__pyx_self, PyObject *__pyx_v_scan_dict); /*proto*/

#line 160 "ms_deisotope/_c/utils.pyx"
static struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_f_12ms_deisotope_2_c_5utils_deserialize_peak_set(PyObject *__pyx_v_scan_dict, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyArrayObject *__pyx_v_mz_array = 0
#line 160 "ms_deisotope/_c/utils.pyx"
;
  PyArrayObject *__pyx_v_intensity_array = 0
#line 160 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 160 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_i
#line 160 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0
#line 160 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_peaks = 0
#line 160 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_peak_set = NULL
#line 160 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_intensity_array
#line 160 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_intensity_array
#line 160 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mz_array
#line 160 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_mz_array
#line 160 "ms_deisotope/_c/utils.pyx"
;

#line 160 "ms_deisotope/_c/utils.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_r = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 160 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  PyArrayObject *__pyx_t_5 = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_6;

#line 160 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  PyArrayObject *__pyx_t_10 = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_11;

#line 160 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_12;

#line 160 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_13;

#line 160 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_14;

#line 160 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_15;

#line 160 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_16;

#line 160 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_17;

#line 160 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_18;

#line 160 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_19;

#line 160 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_20 = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 160 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("deserialize_peak_set", 0);

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_mz_array.pybuffer.buf = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_mz_array.refcount = 0;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_mz_array.data = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_mz_array.rcbuffer = &__pyx_pybuffer_mz_array;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_intensity_array.pybuffer.buf = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_intensity_array.refcount = 0;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_intensity_array.data = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_intensity_array.rcbuffer = &__pyx_pybuffer_intensity_array;

  /* "ms_deisotope/_c/utils.pyx":167
 *         FittedPeak peak
 *         list peaks
 *     mz_array = scan_dict['m/z array'].astype(np.float64)             # <<<<<<<<<<<<<<
 *     intensity_array = scan_dict['intensity array'].astype(np.float64)
 *     n = len(mz_array)
 */

#line 167 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_scan_dict == Py_None)) {

#line 167 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 167 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "ms_deisotope/_c/utils.pyx"
  }

#line 167 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_scan_dict, __pyx_kp_s_m_z_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 167 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_astype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 167 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 167 "ms_deisotope/_c/utils.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 167 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 167 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 167 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = NULL;

#line 167 "ms_deisotope/_c/utils.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {

#line 167 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);

#line 167 "ms_deisotope/_c/utils.pyx"
    if (likely(__pyx_t_2)) {

#line 167 "ms_deisotope/_c/utils.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 167 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(__pyx_t_2);

#line 167 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(function);

#line 167 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 167 "ms_deisotope/_c/utils.pyx"
    }

#line 167 "ms_deisotope/_c/utils.pyx"
  }

#line 167 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);

#line 167 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 167 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 167 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 167 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 167 "ms_deisotope/_c/utils.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_1);

#line 167 "ms_deisotope/_c/utils.pyx"
  {

#line 167 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 167 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer);

#line 167 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);

#line 167 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 < 0)) {

#line 167 "ms_deisotope/_c/utils.pyx"
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);

#line 167 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_mz_array, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {

#line 167 "ms_deisotope/_c/utils.pyx"
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);

#line 167 "ms_deisotope/_c/utils.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 167 "ms_deisotope/_c/utils.pyx"
      } else {

#line 167 "ms_deisotope/_c/utils.pyx"
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);

#line 167 "ms_deisotope/_c/utils.pyx"
      }

#line 167 "ms_deisotope/_c/utils.pyx"
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;

#line 167 "ms_deisotope/_c/utils.pyx"
    }

#line 167 "ms_deisotope/_c/utils.pyx"
    __pyx_pybuffernd_mz_array.diminfo[0].strides = __pyx_pybuffernd_mz_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mz_array.diminfo[0].shape = __pyx_pybuffernd_mz_array.rcbuffer->pybuffer.shape[0];

#line 167 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "ms_deisotope/_c/utils.pyx"
  }

#line 167 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = 0;

#line 167 "ms_deisotope/_c/utils.pyx"
  __pyx_v_mz_array = ((PyArrayObject *)__pyx_t_1);

#line 167 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":168
 *         list peaks
 *     mz_array = scan_dict['m/z array'].astype(np.float64)
 *     intensity_array = scan_dict['intensity array'].astype(np.float64)             # <<<<<<<<<<<<<<
 *     n = len(mz_array)
 *     peaks = []
 */

#line 168 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_scan_dict == Py_None)) {

#line 168 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 168 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/utils.pyx"
  }

#line 168 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_scan_dict, __pyx_kp_s_intensity_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 168 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_astype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 168 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 168 "ms_deisotope/_c/utils.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 168 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 168 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 168 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = NULL;

#line 168 "ms_deisotope/_c/utils.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {

#line 168 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);

#line 168 "ms_deisotope/_c/utils.pyx"
    if (likely(__pyx_t_3)) {

#line 168 "ms_deisotope/_c/utils.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 168 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 168 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(function);

#line 168 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_t_4, function);

#line 168 "ms_deisotope/_c/utils.pyx"
    }

#line 168 "ms_deisotope/_c/utils.pyx"
  }

#line 168 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);

#line 168 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 168 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 168 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 168 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 168 "ms_deisotope/_c/utils.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/utils.pyx"
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_1);

#line 168 "ms_deisotope/_c/utils.pyx"
  {

#line 168 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 168 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer);

#line 168 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);

#line 168 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 < 0)) {

#line 168 "ms_deisotope/_c/utils.pyx"
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);

#line 168 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_intensity_array, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {

#line 168 "ms_deisotope/_c/utils.pyx"
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_7);

#line 168 "ms_deisotope/_c/utils.pyx"
        __Pyx_RaiseBufferFallbackError();

#line 168 "ms_deisotope/_c/utils.pyx"
      } else {

#line 168 "ms_deisotope/_c/utils.pyx"
        PyErr_Restore(__pyx_t_9, __pyx_t_8, __pyx_t_7);

#line 168 "ms_deisotope/_c/utils.pyx"
      }

#line 168 "ms_deisotope/_c/utils.pyx"
      __pyx_t_9 = __pyx_t_8 = __pyx_t_7 = 0;

#line 168 "ms_deisotope/_c/utils.pyx"
    }

#line 168 "ms_deisotope/_c/utils.pyx"
    __pyx_pybuffernd_intensity_array.diminfo[0].strides = __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_intensity_array.diminfo[0].shape = __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.shape[0];

#line 168 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/utils.pyx"
  }

#line 168 "ms_deisotope/_c/utils.pyx"
  __pyx_t_10 = 0;

#line 168 "ms_deisotope/_c/utils.pyx"
  __pyx_v_intensity_array = ((PyArrayObject *)__pyx_t_1);

#line 168 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":169
 *     mz_array = scan_dict['m/z array'].astype(np.float64)
 *     intensity_array = scan_dict['intensity array'].astype(np.float64)
 *     n = len(mz_array)             # <<<<<<<<<<<<<<
 *     peaks = []
 *     for i in range(n):
 */

#line 169 "ms_deisotope/_c/utils.pyx"
  __pyx_t_11 = PyObject_Length(((PyObject *)__pyx_v_mz_array)); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 169, __pyx_L1_error)

#line 169 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = __pyx_t_11;

  /* "ms_deisotope/_c/utils.pyx":170
 *     intensity_array = scan_dict['intensity array'].astype(np.float64)
 *     n = len(mz_array)
 *     peaks = []             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         peak = FittedPeak._create(
 */

#line 170 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)

#line 170 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 170 "ms_deisotope/_c/utils.pyx"
  __pyx_v_peaks = ((PyObject*)__pyx_t_1);

#line 170 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":171
 *     n = len(mz_array)
 *     peaks = []
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         peak = FittedPeak._create(
 *             mz_array[i], intensity_array[i], intensity_array[i],
 */

#line 171 "ms_deisotope/_c/utils.pyx"
  __pyx_t_12 = __pyx_v_n;

#line 171 "ms_deisotope/_c/utils.pyx"
  __pyx_t_13 = __pyx_t_12;

#line 171 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {

#line 171 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_14;

    /* "ms_deisotope/_c/utils.pyx":173
 *     for i in range(n):
 *         peak = FittedPeak._create(
 *             mz_array[i], intensity_array[i], intensity_array[i],             # <<<<<<<<<<<<<<
 *             0.01, 0.01, 0.1, 0, 0,
 *             intensity_array[i])
 */

#line 173 "ms_deisotope/_c/utils.pyx"
    __pyx_t_15 = __pyx_v_i;

#line 173 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 173 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_15 >= (size_t)__pyx_pybuffernd_mz_array.diminfo[0].shape)) __pyx_t_6 = 0;

#line 173 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 173 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 173 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "ms_deisotope/_c/utils.pyx"
    }

#line 173 "ms_deisotope/_c/utils.pyx"
    __pyx_t_16 = __pyx_v_i;

#line 173 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 173 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_16 >= (size_t)__pyx_pybuffernd_intensity_array.diminfo[0].shape)) __pyx_t_6 = 0;

#line 173 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 173 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 173 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "ms_deisotope/_c/utils.pyx"
    }

#line 173 "ms_deisotope/_c/utils.pyx"
    __pyx_t_17 = __pyx_v_i;

#line 173 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 173 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_17 >= (size_t)__pyx_pybuffernd_intensity_array.diminfo[0].shape)) __pyx_t_6 = 0;

#line 173 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 173 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 173 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "ms_deisotope/_c/utils.pyx"
    }

    /* "ms_deisotope/_c/utils.pyx":175
 *             mz_array[i], intensity_array[i], intensity_array[i],
 *             0.01, 0.01, 0.1, 0, 0,
 *             intensity_array[i])             # <<<<<<<<<<<<<<
 *         peaks.append(peak)
 *     peak_set = PeakSetIndexed(peaks)
 */

#line 175 "ms_deisotope/_c/utils.pyx"
    __pyx_t_18 = __pyx_v_i;

#line 175 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 175 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_18 >= (size_t)__pyx_pybuffernd_intensity_array.diminfo[0].shape)) __pyx_t_6 = 0;

#line 175 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 175 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 175 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/utils.pyx"
    }

    /* "ms_deisotope/_c/utils.pyx":172
 *     peaks = []
 *     for i in range(n):
 *         peak = FittedPeak._create(             # <<<<<<<<<<<<<<
 *             mz_array[i], intensity_array[i], intensity_array[i],
 *             0.01, 0.01, 0.1, 0, 0,
 */

#line 172 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak->_create((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_mz_array.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_mz_array.diminfo[0].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_intensity_array.diminfo[0].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_intensity_array.diminfo[0].strides)), 0.01, 0.01, 0.1, 0, 0, (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_intensity_array.diminfo[0].strides)))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 172 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 172 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/utils.pyx":176
 *             0.01, 0.01, 0.1, 0, 0,
 *             intensity_array[i])
 *         peaks.append(peak)             # <<<<<<<<<<<<<<
 *     peak_set = PeakSetIndexed(peaks)
 *     peak_set.reindex()
 */

#line 176 "ms_deisotope/_c/utils.pyx"
    __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_peaks, ((PyObject *)__pyx_v_peak)); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 176, __pyx_L1_error)

#line 176 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":177
 *             intensity_array[i])
 *         peaks.append(peak)
 *     peak_set = PeakSetIndexed(peaks)             # <<<<<<<<<<<<<<
 *     peak_set.reindex()
 *     return PeakIndex(np.array([]), np.array([]), peak_set)
 */

#line 177 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __pyx_v_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)

#line 177 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 177 "ms_deisotope/_c/utils.pyx"
  __pyx_v_peak_set = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_t_1);

#line 177 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":178
 *         peaks.append(peak)
 *     peak_set = PeakSetIndexed(peaks)
 *     peak_set.reindex()             # <<<<<<<<<<<<<<
 *     return PeakIndex(np.array([]), np.array([]), peak_set)
 * 
 */

#line 178 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_peak_set->__pyx_base.__pyx_vtab)->__pyx_base.reindex(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peak_set), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)

#line 178 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 178 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":179
 *     peak_set = PeakSetIndexed(peaks)
 *     peak_set.reindex()
 *     return PeakIndex(np.array([]), np.array([]), peak_set)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = NULL;

#line 179 "ms_deisotope/_c/utils.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {

#line 179 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 179 "ms_deisotope/_c/utils.pyx"
    if (likely(__pyx_t_3)) {

#line 179 "ms_deisotope/_c/utils.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 179 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 179 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(function);

#line 179 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 179 "ms_deisotope/_c/utils.pyx"
    }

#line 179 "ms_deisotope/_c/utils.pyx"
  }

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_20 = NULL;

#line 179 "ms_deisotope/_c/utils.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 179 "ms_deisotope/_c/utils.pyx"
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_3);

#line 179 "ms_deisotope/_c/utils.pyx"
    if (likely(__pyx_t_20)) {

#line 179 "ms_deisotope/_c/utils.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 179 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(__pyx_t_20);

#line 179 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(function);

#line 179 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 179 "ms_deisotope/_c/utils.pyx"
    }

#line 179 "ms_deisotope/_c/utils.pyx"
  }

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_20, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 179 "ms_deisotope/_c/utils.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 179 "ms_deisotope/_c/utils.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set));

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak_set));

#line 179 "ms_deisotope/_c/utils.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)__pyx_v_peak_set));

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 179 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)__pyx_t_2);

#line 179 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = 0;

#line 179 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":160
 *     return peak_set
 * 
 * cpdef PeakIndex deserialize_peak_set(dict scan_dict):             # <<<<<<<<<<<<<<
 *     cdef:
 *         np.ndarray[np.float64_t] mz_array
 */

#line 160 "ms_deisotope/_c/utils.pyx"


#line 160 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_20);

#line 160 "ms_deisotope/_c/utils.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 160 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_declare

#line 160 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_assign

#line 160 "ms_deisotope/_c/utils.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 160 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer);

#line 160 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.deserialize_peak_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 160 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L2;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer);

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_L2:;

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_mz_array);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_intensity_array);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_peaks);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak_set);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 160 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 160 "ms_deisotope/_c/utils.pyx"
}

#line 160 "ms_deisotope/_c/utils.pyx"


#line 160 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 160 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_13deserialize_peak_set(PyObject *__pyx_self, PyObject *__pyx_v_scan_dict); /*proto*/

#line 160 "ms_deisotope/_c/utils.pyx"
static char __pyx_doc_12ms_deisotope_2_c_5utils_12deserialize_peak_set[] = "deserialize_peak_set(dict scan_dict) -> PeakIndex";

#line 160 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_13deserialize_peak_set(PyObject *__pyx_self, PyObject *__pyx_v_scan_dict) {

#line 160 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 160 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 160 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("deserialize_peak_set (wrapper)", 0);

#line 160 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_scan_dict), (&PyDict_Type), 1, "scan_dict", 1))) __PYX_ERR(0, 160, __pyx_L1_error)
  __pyx_r = 
#line 160 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_12deserialize_peak_set(__pyx_self, ((PyObject*)__pyx_v_scan_dict));

#line 160 "ms_deisotope/_c/utils.pyx"


#line 160 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 160 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 160 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 160 "ms_deisotope/_c/utils.pyx"
}

#line 160 "ms_deisotope/_c/utils.pyx"


#line 160 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_12deserialize_peak_set(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_scan_dict) {

#line 160 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 160 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 160 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("deserialize_peak_set", 0);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_5utils_deserialize_peak_set(__pyx_v_scan_dict, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 160 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 160 "ms_deisotope/_c/utils.pyx"


#line 160 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.deserialize_peak_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 160 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 160 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 160 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 160 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":183
 * 
 * @cython.boundscheck(False)
 * cpdef DeconvolutedPeakSetIndexed build_deconvoluted_peak_set_from_arrays(np.ndarray[double, ndim=1] mz_array,             # <<<<<<<<<<<<<<
 *                                                                          np.ndarray[double, ndim=1] intensity_array,
 *                                                                          np.ndarray[long, ndim=1] charge_array):
 */

#line 183 "ms_deisotope/_c/utils.pyx"


#line 183 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_15build_deconvoluted_peak_set_from_arrays(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 183 "ms_deisotope/_c/utils.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_f_12ms_deisotope_2_c_5utils_build_deconvoluted_peak_set_from_arrays(PyArrayObject *__pyx_v_mz_array, PyArrayObject *__pyx_v_intensity_array, PyArrayObject *__pyx_v_charge_array, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_peaks = 0
#line 183 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 183 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_i
#line 183 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_mz
#line 183 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_peak_neutral_mass
#line 183 "ms_deisotope/_c/utils.pyx"
;
  int __pyx_v_charge
#line 183 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_v_peak = 0
#line 183 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_v_peak_set = 0
#line 183 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_charge_array
#line 183 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_charge_array
#line 183 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_intensity_array
#line 183 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_intensity_array
#line 183 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mz_array
#line 183 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_mz_array
#line 183 "ms_deisotope/_c/utils.pyx"
;

#line 183 "ms_deisotope/_c/utils.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_r = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 183 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_2;

#line 183 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 183 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_4;

#line 183 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_5;

#line 183 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_6;

#line 183 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_7;

#line 183 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_8;

#line 183 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("build_deconvoluted_peak_set_from_arrays", 0);

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_mz_array.pybuffer.buf = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_mz_array.refcount = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_mz_array.data = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_mz_array.rcbuffer = &__pyx_pybuffer_mz_array;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_intensity_array.pybuffer.buf = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_intensity_array.refcount = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_intensity_array.data = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_intensity_array.rcbuffer = &__pyx_pybuffer_intensity_array;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_charge_array.pybuffer.buf = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_charge_array.refcount = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_charge_array.data = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_charge_array.rcbuffer = &__pyx_pybuffer_charge_array;

#line 183 "ms_deisotope/_c/utils.pyx"
  {

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 183 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_mz_array, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 183, __pyx_L1_error)

#line 183 "ms_deisotope/_c/utils.pyx"
  }

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_mz_array.diminfo[0].strides = __pyx_pybuffernd_mz_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mz_array.diminfo[0].shape = __pyx_pybuffernd_mz_array.rcbuffer->pybuffer.shape[0];

#line 183 "ms_deisotope/_c/utils.pyx"
  {

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 183 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_intensity_array, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 183, __pyx_L1_error)

#line 183 "ms_deisotope/_c/utils.pyx"
  }

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_intensity_array.diminfo[0].strides = __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_intensity_array.diminfo[0].shape = __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.shape[0];

#line 183 "ms_deisotope/_c/utils.pyx"
  {

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 183 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_charge_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_charge_array, &__Pyx_TypeInfo_long, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 183, __pyx_L1_error)

#line 183 "ms_deisotope/_c/utils.pyx"
  }

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_charge_array.diminfo[0].strides = __pyx_pybuffernd_charge_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_charge_array.diminfo[0].shape = __pyx_pybuffernd_charge_array.rcbuffer->pybuffer.shape[0];

  /* "ms_deisotope/_c/utils.pyx":194
 *         DeconvolutedPeakSetIndexed peak_set
 * 
 *     peaks = []             # <<<<<<<<<<<<<<
 *     n = mz_array.shape[0]
 *     i = 0
 */

#line 194 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)

#line 194 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 194 "ms_deisotope/_c/utils.pyx"
  __pyx_v_peaks = ((PyObject*)__pyx_t_1);

#line 194 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":195
 * 
 *     peaks = []
 *     n = mz_array.shape[0]             # <<<<<<<<<<<<<<
 *     i = 0
 *     for i in range(n):
 */

#line 195 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = (__pyx_v_mz_array->dimensions[0]);

  /* "ms_deisotope/_c/utils.pyx":196
 *     peaks = []
 *     n = mz_array.shape[0]
 *     i = 0             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         mz = mz_array[i]
 */

#line 196 "ms_deisotope/_c/utils.pyx"
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/utils.pyx":197
 *     n = mz_array.shape[0]
 *     i = 0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         mz = mz_array[i]
 *         charge = charge_array[i]
 */

#line 197 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_v_n;

#line 197 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 197 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 197 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/utils.pyx":198
 *     i = 0
 *     for i in range(n):
 *         mz = mz_array[i]             # <<<<<<<<<<<<<<
 *         charge = charge_array[i]
 *         peak_neutral_mass = neutral_mass(mz, charge)
 */

#line 198 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = __pyx_v_i;

#line 198 "ms_deisotope/_c/utils.pyx"
    __pyx_v_mz = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_mz_array.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_mz_array.diminfo[0].strides));

    /* "ms_deisotope/_c/utils.pyx":199
 *     for i in range(n):
 *         mz = mz_array[i]
 *         charge = charge_array[i]             # <<<<<<<<<<<<<<
 *         peak_neutral_mass = neutral_mass(mz, charge)
 * 
 */

#line 199 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = __pyx_v_i;

#line 199 "ms_deisotope/_c/utils.pyx"
    __pyx_v_charge = (*__Pyx_BufPtrStrided1d(long *, __pyx_pybuffernd_charge_array.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_charge_array.diminfo[0].strides));

    /* "ms_deisotope/_c/utils.pyx":200
 *         mz = mz_array[i]
 *         charge = charge_array[i]
 *         peak_neutral_mass = neutral_mass(mz, charge)             # <<<<<<<<<<<<<<
 * 
 *         peak = DeconvolutedPeak._create_simple(
 */

#line 200 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peak_neutral_mass = __pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass(__pyx_v_mz, __pyx_v_charge, NULL);

    /* "ms_deisotope/_c/utils.pyx":204
 *         peak = DeconvolutedPeak._create_simple(
 *             peak_neutral_mass,
 *             intensity_array[i],             # <<<<<<<<<<<<<<
 *             charge_array[i],
 *             intensity_array[i],
 */

#line 204 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = __pyx_v_i;

    /* "ms_deisotope/_c/utils.pyx":205
 *             peak_neutral_mass,
 *             intensity_array[i],
 *             charge_array[i],             # <<<<<<<<<<<<<<
 *             intensity_array[i],
 *             mz,
 */

#line 205 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __pyx_v_i;

    /* "ms_deisotope/_c/utils.pyx":206
 *             intensity_array[i],
 *             charge_array[i],
 *             intensity_array[i],             # <<<<<<<<<<<<<<
 *             mz,
 *             None)
 */

#line 206 "ms_deisotope/_c/utils.pyx"
    __pyx_t_7 = __pyx_v_i;

    /* "ms_deisotope/_c/utils.pyx":202
 *         peak_neutral_mass = neutral_mass(mz, charge)
 * 
 *         peak = DeconvolutedPeak._create_simple(             # <<<<<<<<<<<<<<
 *             peak_neutral_mass,
 *             intensity_array[i],
 */

#line 202 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak->_create_simple(__pyx_v_peak_neutral_mass, (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_intensity_array.diminfo[0].strides)), (*__Pyx_BufPtrStrided1d(long *, __pyx_pybuffernd_charge_array.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_charge_array.diminfo[0].strides)), (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_intensity_array.diminfo[0].strides)), __pyx_v_mz, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *)Py_None))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)

#line 202 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 202 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_t_1));

#line 202 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/utils.pyx":209
 *             mz,
 *             None)
 *         peaks.append(peak)             # <<<<<<<<<<<<<<
 *     peak_set = DeconvolutedPeakSetIndexed(peaks)
 *     peak_set.reindex()
 */

#line 209 "ms_deisotope/_c/utils.pyx"
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_peaks, ((PyObject *)__pyx_v_peak)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 209, __pyx_L1_error)

#line 209 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":210
 *             None)
 *         peaks.append(peak)
 *     peak_set = DeconvolutedPeakSetIndexed(peaks)             # <<<<<<<<<<<<<<
 *     peak_set.reindex()
 *     return peak_set
 */

#line 210 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed), __pyx_v_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 210, __pyx_L1_error)

#line 210 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 210 "ms_deisotope/_c/utils.pyx"
  __pyx_v_peak_set = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *)__pyx_t_1);

#line 210 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":211
 *         peaks.append(peak)
 *     peak_set = DeconvolutedPeakSetIndexed(peaks)
 *     peak_set.reindex()             # <<<<<<<<<<<<<<
 *     return peak_set
 * 
 */

#line 211 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *)__pyx_v_peak_set->__pyx_base.__pyx_vtab)->__pyx_base.reindex(((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peak_set), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 211, __pyx_L1_error)

#line 211 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 211 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":212
 *     peak_set = DeconvolutedPeakSetIndexed(peaks)
 *     peak_set.reindex()
 *     return peak_set             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 212 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 212 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak_set));

#line 212 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_peak_set;

#line 212 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":183
 * 
 * @cython.boundscheck(False)
 * cpdef DeconvolutedPeakSetIndexed build_deconvoluted_peak_set_from_arrays(np.ndarray[double, ndim=1] mz_array,             # <<<<<<<<<<<<<<
 *                                                                          np.ndarray[double, ndim=1] intensity_array,
 *                                                                          np.ndarray[long, ndim=1] charge_array):
 */

#line 183 "ms_deisotope/_c/utils.pyx"


#line 183 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 183 "ms_deisotope/_c/utils.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_declare

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_assign

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_charge_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.build_deconvoluted_peak_set_from_arrays", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L2;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_charge_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_L2:;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_peaks);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak_set);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 183 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 183 "ms_deisotope/_c/utils.pyx"
}

#line 183 "ms_deisotope/_c/utils.pyx"


#line 183 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 183 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_15build_deconvoluted_peak_set_from_arrays(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 183 "ms_deisotope/_c/utils.pyx"
static char __pyx_doc_12ms_deisotope_2_c_5utils_14build_deconvoluted_peak_set_from_arrays[] = "build_deconvoluted_peak_set_from_arrays(ndarray mz_array, ndarray intensity_array, ndarray charge_array) -> DeconvolutedPeakSetIndexed";

#line 183 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_15build_deconvoluted_peak_set_from_arrays(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_mz_array = 0
#line 183 "ms_deisotope/_c/utils.pyx"
;
  PyArrayObject *__pyx_v_intensity_array = 0
#line 183 "ms_deisotope/_c/utils.pyx"
;
  PyArrayObject *__pyx_v_charge_array = 0
#line 183 "ms_deisotope/_c/utils.pyx"
;

#line 183 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("build_deconvoluted_peak_set_from_arrays (wrapper)", 0);

#line 183 "ms_deisotope/_c/utils.pyx"
  {

#line 183 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz_array,&__pyx_n_s_intensity_array_2,&__pyx_n_s_charge_array_2,0};

#line 183 "ms_deisotope/_c/utils.pyx"
    PyObject* values[3] = {0,0,0};

#line 183 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 183 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 183 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 183 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  3: 
#line 183 "ms_deisotope/_c/utils.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 183 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 183 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 183 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 183 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 183 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 183 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 183 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 183 "ms_deisotope/_c/utils.pyx"
      }

#line 183 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 183 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 183 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 183 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz_array)) != 0)) kw_args--;
        else 
#line 183 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 183 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 183 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 183 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_intensity_array_2)) != 0)) kw_args--;

#line 183 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("build_deconvoluted_peak_set_from_arrays", 1, 3, 3, 1); 
#line 183 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 183, __pyx_L3_error)

#line 183 "ms_deisotope/_c/utils.pyx"
        }

#line 183 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 183 "ms_deisotope/_c/utils.pyx"
        case  2:

#line 183 "ms_deisotope/_c/utils.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_array_2)) != 0)) kw_args--;

#line 183 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("build_deconvoluted_peak_set_from_arrays", 1, 3, 3, 2); 
#line 183 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 183, __pyx_L3_error)

#line 183 "ms_deisotope/_c/utils.pyx"
        }

#line 183 "ms_deisotope/_c/utils.pyx"
      }

#line 183 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 183 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "build_deconvoluted_peak_set_from_arrays") < 0)) __PYX_ERR(0, 183, __pyx_L3_error)

#line 183 "ms_deisotope/_c/utils.pyx"
      }

#line 183 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 183 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 183 "ms_deisotope/_c/utils.pyx"
    } else {

#line 183 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 183 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 183 "ms_deisotope/_c/utils.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 183 "ms_deisotope/_c/utils.pyx"
    }

#line 183 "ms_deisotope/_c/utils.pyx"
    __pyx_v_mz_array = ((PyArrayObject *)values[0]);

#line 183 "ms_deisotope/_c/utils.pyx"
    __pyx_v_intensity_array = ((PyArrayObject *)values[1]);

#line 183 "ms_deisotope/_c/utils.pyx"
    __pyx_v_charge_array = ((PyArrayObject *)values[2]);

#line 183 "ms_deisotope/_c/utils.pyx"
  }

#line 183 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("build_deconvoluted_peak_set_from_arrays", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 183 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 183, __pyx_L3_error)

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.build_deconvoluted_peak_set_from_arrays", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 183 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 183 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mz_array), __pyx_ptype_5numpy_ndarray, 1, "mz_array", 0))) __PYX_ERR(0, 183, __pyx_L1_error)

#line 183 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_intensity_array), __pyx_ptype_5numpy_ndarray, 1, "intensity_array", 0))) __PYX_ERR(0, 184, __pyx_L1_error)

#line 183 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_charge_array), __pyx_ptype_5numpy_ndarray, 1, "charge_array", 0))) __PYX_ERR(0, 185, __pyx_L1_error)
  __pyx_r = 
#line 183 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_14build_deconvoluted_peak_set_from_arrays(__pyx_self, __pyx_v_mz_array, __pyx_v_intensity_array, __pyx_v_charge_array);

#line 183 "ms_deisotope/_c/utils.pyx"


#line 183 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 183 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 183 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 183 "ms_deisotope/_c/utils.pyx"
}

#line 183 "ms_deisotope/_c/utils.pyx"


#line 183 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_14build_deconvoluted_peak_set_from_arrays(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_mz_array, PyArrayObject *__pyx_v_intensity_array, PyArrayObject *__pyx_v_charge_array) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_charge_array
#line 183 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_charge_array
#line 183 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_intensity_array
#line 183 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_intensity_array
#line 183 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mz_array
#line 183 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_mz_array
#line 183 "ms_deisotope/_c/utils.pyx"
;

#line 183 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 183 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("build_deconvoluted_peak_set_from_arrays", 0);

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_mz_array.pybuffer.buf = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_mz_array.refcount = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_mz_array.data = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_mz_array.rcbuffer = &__pyx_pybuffer_mz_array;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_intensity_array.pybuffer.buf = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_intensity_array.refcount = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_intensity_array.data = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_intensity_array.rcbuffer = &__pyx_pybuffer_intensity_array;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_charge_array.pybuffer.buf = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffer_charge_array.refcount = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_charge_array.data = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_charge_array.rcbuffer = &__pyx_pybuffer_charge_array;

#line 183 "ms_deisotope/_c/utils.pyx"
  {

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 183 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_mz_array, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 183, __pyx_L1_error)

#line 183 "ms_deisotope/_c/utils.pyx"
  }

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_mz_array.diminfo[0].strides = __pyx_pybuffernd_mz_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mz_array.diminfo[0].shape = __pyx_pybuffernd_mz_array.rcbuffer->pybuffer.shape[0];

#line 183 "ms_deisotope/_c/utils.pyx"
  {

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 183 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_intensity_array, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 183, __pyx_L1_error)

#line 183 "ms_deisotope/_c/utils.pyx"
  }

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_intensity_array.diminfo[0].strides = __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_intensity_array.diminfo[0].shape = __pyx_pybuffernd_intensity_array.rcbuffer->pybuffer.shape[0];

#line 183 "ms_deisotope/_c/utils.pyx"
  {

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 183 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_charge_array.rcbuffer->pybuffer, (PyObject*)__pyx_v_charge_array, &__Pyx_TypeInfo_long, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 183, __pyx_L1_error)

#line 183 "ms_deisotope/_c/utils.pyx"
  }

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_pybuffernd_charge_array.diminfo[0].strides = __pyx_pybuffernd_charge_array.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_charge_array.diminfo[0].shape = __pyx_pybuffernd_charge_array.rcbuffer->pybuffer.shape[0];

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_5utils_build_deconvoluted_peak_set_from_arrays(((PyArrayObject *)__pyx_v_mz_array), ((PyArrayObject *)__pyx_v_intensity_array), ((PyArrayObject *)__pyx_v_charge_array), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 183 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 183 "ms_deisotope/_c/utils.pyx"


#line 183 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 183 "ms_deisotope/_c/utils.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_declare

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_PyThreadState_assign

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_charge_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.build_deconvoluted_peak_set_from_arrays", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 183 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L2;

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_charge_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_intensity_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mz_array.rcbuffer->pybuffer);

#line 183 "ms_deisotope/_c/utils.pyx"
  __pyx_L2:;

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 183 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 183 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 183 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":216
 * 
 * @cython.binding(True)
 * cpdef double _peak_sequence_tic(self, peak_collection peaks) except -1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_17_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 216 "ms_deisotope/_c/utils.pyx"
static char __pyx_doc_12ms_deisotope_2_c_5utils_16_peak_sequence_tic[] = "_peak_sequence_tic(signatures, args, kwargs, defaults)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_5utils_17_peak_sequence_tic = 
#line 216 "ms_deisotope/_c/utils.pyx"
{"_peak_sequence_tic", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_17_peak_sequence_tic, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_16_peak_sequence_tic};

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_17_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_signatures = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_args = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  {

#line 216 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signatures,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    PyObject* values[4] = {0,0,0,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 216 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 216 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  4: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signatures)) != 0)) kw_args--;
        else 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;

#line 216 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 1); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
        }

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  2:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;

#line 216 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 2); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
        }

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  3:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;

#line 216 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 3); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
        }

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 216 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {

#line 216 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
    } else {

#line 216 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 216 "ms_deisotope/_c/utils.pyx"
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_signatures = values[0];

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_args = values[1];

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_kwargs = values[2];

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_defaults = values[3];

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 216 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_16_peak_sequence_tic(__pyx_self, __pyx_v_signatures, __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_16_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL
#line 216 "ms_deisotope/_c/utils.pyx"
;
  Py_ssize_t __pyx_v_i
#line 216 "ms_deisotope/_c/utils.pyx"
;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize
#line 216 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_arg = NULL
#line 216 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_candidates = NULL
#line 216 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_sig = NULL
#line 216 "ms_deisotope/_c/utils.pyx"
;
  int __pyx_v_match_found
#line 216 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_src_sig = NULL
#line 216 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_dst_type = NULL
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_2;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_3;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_4;

#line 216 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_5;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_7;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_8;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_9;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_13;

#line 216 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_14;

#line 216 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_15;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_16;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("_peak_sequence_tic", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(__pyx_v_kwargs);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_4) {

#line 216 "ms_deisotope/_c/utils.pyx"
  } else {

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 216 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L4_bool_binop_done;

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = ((!__pyx_t_4) != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_bool_binop_done:;

#line 216 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_v_itemsize = -1L;

#line 216 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_args == Py_None)) {

#line 216 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 216 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = ((1 < __pyx_t_5) != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 216 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 216 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(((PyObject*)__pyx_v_args), 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L6;

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_4) {

#line 216 "ms_deisotope/_c/utils.pyx"
  } else {

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 216 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L7_bool_binop_done;

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 216 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 216 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_peaks, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_t_4 != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L7_bool_binop_done:;

#line 216 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 216 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 216 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L6;

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  /*else*/ {

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 216 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 216 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_int_2);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_int_2);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_n_s_s);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_n_s_s);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L6:;

#line 216 "ms_deisotope/_c/utils.pyx"
  while (1) {

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_peak_picker__c_peak_set);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_3) {

#line 216 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L10_break;

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_ms_peak_picker__c_peak_index);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PeakIndex); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = (__pyx_t_3 != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_2) {

#line 216 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakIndex, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L10_break;

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_deisotope__c_peak_set);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_DeconvolutedPeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_3) {

#line 216 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_DeconvolutedPeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L10_break;

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_ms_deisotope__c_peak_set);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_DeconvolutedPeakSetIndexed); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = (__pyx_t_3 != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_2) {

#line 216 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_DeconvolutedPeakSetIndexed, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L10_break;

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_object, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L10_break;

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L10_break:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_signatures == Py_None)) {

#line 216 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 216 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  while (1) {

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_9 == 0)) break;

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_match_found = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_12 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 216 "ms_deisotope/_c/utils.pyx"
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);

#line 216 "ms_deisotope/_c/utils.pyx"
      if (likely(__pyx_t_12)) {

#line 216 "ms_deisotope/_c/utils.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 216 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(__pyx_t_12);

#line 216 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(function);

#line 216 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__2) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__2);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_10);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_10 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 216 "ms_deisotope/_c/utils.pyx"
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);

#line 216 "ms_deisotope/_c/utils.pyx"
      if (likely(__pyx_t_10)) {

#line 216 "ms_deisotope/_c/utils.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 216 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(__pyx_t_10);

#line 216 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(function);

#line 216 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__3) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__3);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_14 = __pyx_t_13;

#line 216 "ms_deisotope/_c/utils.pyx"
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 216 "ms_deisotope/_c/utils.pyx"
      __pyx_v_i = __pyx_t_15;

#line 216 "ms_deisotope/_c/utils.pyx"
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
      __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);

#line 216 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = (__pyx_t_2 != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
      if (__pyx_t_3) {

#line 216 "ms_deisotope/_c/utils.pyx"
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
        if (__pyx_t_3) {

#line 216 "ms_deisotope/_c/utils.pyx"
          __pyx_v_match_found = 1;

#line 216 "ms_deisotope/_c/utils.pyx"
          goto __pyx_L20;

#line 216 "ms_deisotope/_c/utils.pyx"
        }

#line 216 "ms_deisotope/_c/utils.pyx"
        /*else*/ {

#line 216 "ms_deisotope/_c/utils.pyx"
          __pyx_v_match_found = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
          goto __pyx_L18_break;

#line 216 "ms_deisotope/_c/utils.pyx"
        }

#line 216 "ms_deisotope/_c/utils.pyx"
        __pyx_L20:;

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_L18_break:;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = (__pyx_v_match_found != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_3) {

#line 216 "ms_deisotope/_c/utils.pyx"
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = ((!__pyx_t_3) != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  /*else*/ {

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_signatures == Py_None)) {

#line 216 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 216 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_candidates, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_11 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 216 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_r = __pyx_t_11;

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_t_11 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L0;

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_11);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_12);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_dest_sig);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_ndarray);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_arg);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_candidates);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_sig);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_src_sig);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_dst_type);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_23__pyx_fuse_0_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_17_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 216 "ms_deisotope/_c/utils.pyx"
static double __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 216 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 216 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_tic
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  double __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_2;

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0_peak_sequence_tic", 0);

  /* "ms_deisotope/_c/utils.pyx":223
 *         list py_peaks
 * 
 *     tic = 0.0             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()
 */

#line 223 "ms_deisotope/_c/utils.pyx"
  __pyx_v_tic = 0.0;

  /* "ms_deisotope/_c/utils.pyx":225
 *     tic = 0.0
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()             # <<<<<<<<<<<<<<
 *     else:
 *         py_peaks = list(peaks)
 */

#line 225 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);

  /* "ms_deisotope/_c/utils.pyx":232
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)
 */

#line 232 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 232 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 232 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 232 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/utils.pyx":234
 *     for i in range(n):
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)             # <<<<<<<<<<<<<<
 *         elif peak_collection is PeakIndex:
 *             peak = peaks.peaks.getitem(i)
 */

#line 234 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 234 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 234 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/utils.pyx":239
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         tic += peak.intensity             # <<<<<<<<<<<<<<
 *     return tic
 * 
 */

#line 239 "ms_deisotope/_c/utils.pyx"
    __pyx_v_tic = (__pyx_v_tic + __pyx_v_peak->intensity);

#line 239 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":240
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         tic += peak.intensity
 *     return tic             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 240 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_tic;

#line 240 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":216
 * 
 * @cython.binding(True)
 * cpdef double _peak_sequence_tic(self, peak_collection peaks) except -1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils._peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = -1.0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_23__pyx_fuse_0_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_5utils_23__pyx_fuse_0_peak_sequence_tic = 
#line 216 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_0_peak_sequence_tic", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_23__pyx_fuse_0_peak_sequence_tic, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_16_peak_sequence_tic};

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_23__pyx_fuse_0_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0_peak_sequence_tic (wrapper)", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  {

#line 216 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peaks,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 216 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 216 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 216 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_peak_sequence_tic", 1, 2, 2, 1); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
        }

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 216 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0_peak_sequence_tic") < 0)) __PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 216 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
    } else {

#line 216 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_self = values[0];

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peaks = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[1]);

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_peak_sequence_tic", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_0_peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 216 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaks), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaks", 0))) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_r = 
#line 216 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_22__pyx_fuse_0_peak_sequence_tic(__pyx_self, __pyx_v_self, __pyx_v_peaks);

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_22__pyx_fuse_0_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks) {

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0_peak_sequence_tic", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(__pyx_v_self, __pyx_v_peaks, 0); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_2;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_0_peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_25__pyx_fuse_1_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_17_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 216 "ms_deisotope/_c/utils.pyx"
static double __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peaks, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 216 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 216 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_tic
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  double __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_2;

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1_peak_sequence_tic", 0);

  /* "ms_deisotope/_c/utils.pyx":223
 *         list py_peaks
 * 
 *     tic = 0.0             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()
 */

#line 223 "ms_deisotope/_c/utils.pyx"
  __pyx_v_tic = 0.0;

  /* "ms_deisotope/_c/utils.pyx":225
 *     tic = 0.0
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()             # <<<<<<<<<<<<<<
 *     else:
 *         py_peaks = list(peaks)
 */

#line 225 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);

  /* "ms_deisotope/_c/utils.pyx":232
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)
 */

#line 232 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 232 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 232 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 232 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/utils.pyx":236
 *             peak = peaks.getitem(i)
 *         elif peak_collection is PeakIndex:
 *             peak = peaks.peaks.getitem(i)             # <<<<<<<<<<<<<<
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 */

#line 236 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->peaks->__pyx_vtab)->getitem(__pyx_v_peaks->peaks, __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 236, __pyx_L1_error)

#line 236 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 236 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 236 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/utils.pyx":239
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         tic += peak.intensity             # <<<<<<<<<<<<<<
 *     return tic
 * 
 */

#line 239 "ms_deisotope/_c/utils.pyx"
    __pyx_v_tic = (__pyx_v_tic + __pyx_v_peak->intensity);

#line 239 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":240
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         tic += peak.intensity
 *     return tic             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 240 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_tic;

#line 240 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":216
 * 
 * @cython.binding(True)
 * cpdef double _peak_sequence_tic(self, peak_collection peaks) except -1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils._peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = -1.0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_25__pyx_fuse_1_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_5utils_25__pyx_fuse_1_peak_sequence_tic = 
#line 216 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_1_peak_sequence_tic", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_25__pyx_fuse_1_peak_sequence_tic, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_16_peak_sequence_tic};

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_25__pyx_fuse_1_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peaks = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1_peak_sequence_tic (wrapper)", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  {

#line 216 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peaks,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 216 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 216 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 216 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_peak_sequence_tic", 1, 2, 2, 1); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
        }

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 216 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1_peak_sequence_tic") < 0)) __PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 216 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
    } else {

#line 216 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_self = values[0];

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peaks = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[1]);

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_peak_sequence_tic", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_1_peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 216 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaks), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peaks", 0))) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_r = 
#line 216 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_24__pyx_fuse_1_peak_sequence_tic(__pyx_self, __pyx_v_self, __pyx_v_peaks);

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_24__pyx_fuse_1_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peaks) {

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1_peak_sequence_tic", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(__pyx_v_self, __pyx_v_peaks, 0); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_2;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_1_peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_27__pyx_fuse_2_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_17_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 216 "ms_deisotope/_c/utils.pyx"
static double __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peaks, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 216 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 216 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_tic
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  double __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_2;

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2_peak_sequence_tic", 0);

  /* "ms_deisotope/_c/utils.pyx":223
 *         list py_peaks
 * 
 *     tic = 0.0             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()
 */

#line 223 "ms_deisotope/_c/utils.pyx"
  __pyx_v_tic = 0.0;

  /* "ms_deisotope/_c/utils.pyx":225
 *     tic = 0.0
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()             # <<<<<<<<<<<<<<
 *     else:
 *         py_peaks = list(peaks)
 */

#line 225 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);

  /* "ms_deisotope/_c/utils.pyx":232
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)
 */

#line 232 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 232 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 232 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 232 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/utils.pyx":234
 *     for i in range(n):
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)             # <<<<<<<<<<<<<<
 *         elif peak_collection is PeakIndex:
 *             peak = peaks.peaks.getitem(i)
 */

#line 234 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 234 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 234 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/utils.pyx":239
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         tic += peak.intensity             # <<<<<<<<<<<<<<
 *     return tic
 * 
 */

#line 239 "ms_deisotope/_c/utils.pyx"
    __pyx_v_tic = (__pyx_v_tic + __pyx_v_peak->intensity);

#line 239 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":240
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         tic += peak.intensity
 *     return tic             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 240 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_tic;

#line 240 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":216
 * 
 * @cython.binding(True)
 * cpdef double _peak_sequence_tic(self, peak_collection peaks) except -1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils._peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = -1.0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_27__pyx_fuse_2_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_5utils_27__pyx_fuse_2_peak_sequence_tic = 
#line 216 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_2_peak_sequence_tic", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_27__pyx_fuse_2_peak_sequence_tic, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_16_peak_sequence_tic};

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_27__pyx_fuse_2_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peaks = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2_peak_sequence_tic (wrapper)", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  {

#line 216 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peaks,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 216 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 216 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 216 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2_peak_sequence_tic", 1, 2, 2, 1); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
        }

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 216 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2_peak_sequence_tic") < 0)) __PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 216 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
    } else {

#line 216 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_self = values[0];

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peaks = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)values[1]);

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2_peak_sequence_tic", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_2_peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 216 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaks), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet, 1, "peaks", 0))) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_r = 
#line 216 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_26__pyx_fuse_2_peak_sequence_tic(__pyx_self, __pyx_v_self, __pyx_v_peaks);

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_26__pyx_fuse_2_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peaks) {

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2_peak_sequence_tic", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_fuse_2__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(__pyx_v_self, __pyx_v_peaks, 0); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_2;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_2_peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_29__pyx_fuse_3_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_17_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 216 "ms_deisotope/_c/utils.pyx"
static double __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_v_peaks, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 216 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 216 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_tic
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  double __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_2;

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3_peak_sequence_tic", 0);

  /* "ms_deisotope/_c/utils.pyx":223
 *         list py_peaks
 * 
 *     tic = 0.0             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()
 */

#line 223 "ms_deisotope/_c/utils.pyx"
  __pyx_v_tic = 0.0;

  /* "ms_deisotope/_c/utils.pyx":225
 *     tic = 0.0
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()             # <<<<<<<<<<<<<<
 *     else:
 *         py_peaks = list(peaks)
 */

#line 225 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *)__pyx_v_peaks->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peaks));

  /* "ms_deisotope/_c/utils.pyx":232
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)
 */

#line 232 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 232 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 232 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 232 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/utils.pyx":234
 *     for i in range(n):
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)             # <<<<<<<<<<<<<<
 *         elif peak_collection is PeakIndex:
 *             peak = peaks.peaks.getitem(i)
 */

#line 234 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *)__pyx_v_peaks->__pyx_base.__pyx_vtab)->__pyx_base.getitem(((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peaks), __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 234, __pyx_L1_error)

#line 234 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 234 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 234 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/utils.pyx":239
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         tic += peak.intensity             # <<<<<<<<<<<<<<
 *     return tic
 * 
 */

#line 239 "ms_deisotope/_c/utils.pyx"
    __pyx_v_tic = (__pyx_v_tic + __pyx_v_peak->intensity);

#line 239 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":240
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         tic += peak.intensity
 *     return tic             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 240 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_tic;

#line 240 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":216
 * 
 * @cython.binding(True)
 * cpdef double _peak_sequence_tic(self, peak_collection peaks) except -1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils._peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = -1.0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_29__pyx_fuse_3_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_5utils_29__pyx_fuse_3_peak_sequence_tic = 
#line 216 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_3_peak_sequence_tic", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_29__pyx_fuse_3_peak_sequence_tic, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_16_peak_sequence_tic};

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_29__pyx_fuse_3_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_v_peaks = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3_peak_sequence_tic (wrapper)", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  {

#line 216 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peaks,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 216 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 216 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 216 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3_peak_sequence_tic", 1, 2, 2, 1); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
        }

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 216 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3_peak_sequence_tic") < 0)) __PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 216 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
    } else {

#line 216 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_self = values[0];

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peaks = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *)values[1]);

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3_peak_sequence_tic", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_3_peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 216 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaks), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed, 1, "peaks", 0))) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_r = 
#line 216 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_28__pyx_fuse_3_peak_sequence_tic(__pyx_self, __pyx_v_self, __pyx_v_peaks);

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_28__pyx_fuse_3_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_v_peaks) {

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3_peak_sequence_tic", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_fuse_3__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(__pyx_v_self, __pyx_v_peaks, 0); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_2;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_3_peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_31__pyx_fuse_4_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_17_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 216 "ms_deisotope/_c/utils.pyx"
static double __pyx_fuse_4__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_peaks, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 216 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 216 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_tic
#line 216 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_py_peaks = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  double __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_2;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_3;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_4;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_6;

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_7;

#line 216 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_8;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_4_peak_sequence_tic", 0);

  /* "ms_deisotope/_c/utils.pyx":223
 *         list py_peaks
 * 
 *     tic = 0.0             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()
 */

#line 223 "ms_deisotope/_c/utils.pyx"
  __pyx_v_tic = 0.0;

  /* "ms_deisotope/_c/utils.pyx":227
 *         n = peaks.get_size()
 *     else:
 *         py_peaks = list(peaks)             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(py_peaks)
 *         if n > 0:
 */

#line 227 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PySequence_List(__pyx_v_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)

#line 227 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 227 "ms_deisotope/_c/utils.pyx"
  __pyx_v_py_peaks = ((PyObject*)__pyx_t_1);

#line 227 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":228
 *     else:
 *         py_peaks = list(peaks)
 *         n = PyList_GET_SIZE(py_peaks)             # <<<<<<<<<<<<<<
 *         if n > 0:
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 */

#line 228 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_py_peaks);

  /* "ms_deisotope/_c/utils.pyx":229
 *         py_peaks = list(peaks)
 *         n = PyList_GET_SIZE(py_peaks)
 *         if n > 0:             # <<<<<<<<<<<<<<
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 */

#line 229 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = ((__pyx_v_n > 0) != 0);

#line 229 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/utils.pyx":230
 *         n = PyList_GET_SIZE(py_peaks)
 *         if n > 0:
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):
 */

#line 230 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = PyList_GET_ITEM(__pyx_v_py_peaks, 0);

#line 230 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_t_3), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase); 

#line 230 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);

#line 230 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_4)) {

      /* "ms_deisotope/_c/utils.pyx":231
 *         if n > 0:
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 231 "ms_deisotope/_c/utils.pyx"
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_py_peaks, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)

#line 231 "ms_deisotope/_c/utils.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 231 "ms_deisotope/_c/utils.pyx"
      __pyx_t_5 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_interpret_r_as_a_PeakBase, ((PyObject *)Py_TYPE(__pyx_t_1))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 231, __pyx_L1_error)

#line 231 "ms_deisotope/_c/utils.pyx"
      __Pyx_GOTREF(__pyx_t_5);

#line 231 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 231 "ms_deisotope/_c/utils.pyx"
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)

#line 231 "ms_deisotope/_c/utils.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 231 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 231 "ms_deisotope/_c/utils.pyx"
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 231 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 231 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 231, __pyx_L1_error)

      /* "ms_deisotope/_c/utils.pyx":230
 *         n = PyList_GET_SIZE(py_peaks)
 *         if n > 0:
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):
 */

#line 230 "ms_deisotope/_c/utils.pyx"
    }

    /* "ms_deisotope/_c/utils.pyx":229
 *         py_peaks = list(peaks)
 *         n = PyList_GET_SIZE(py_peaks)
 *         if n > 0:             # <<<<<<<<<<<<<<
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 */

#line 229 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":232
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)
 */

#line 232 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = __pyx_v_n;

#line 232 "ms_deisotope/_c/utils.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 232 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 232 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/utils.pyx":238
 *             peak = peaks.peaks.getitem(i)
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)             # <<<<<<<<<<<<<<
 *         tic += peak.intensity
 *     return tic
 */

#line 238 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = PyList_GET_ITEM(__pyx_v_py_peaks, __pyx_v_i);

#line 238 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_3);

#line 238 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 238 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 238 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/utils.pyx":239
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         tic += peak.intensity             # <<<<<<<<<<<<<<
 *     return tic
 * 
 */

#line 239 "ms_deisotope/_c/utils.pyx"
    __pyx_v_tic = (__pyx_v_tic + __pyx_v_peak->intensity);

#line 239 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":240
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         tic += peak.intensity
 *     return tic             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 240 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_tic;

#line 240 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":216
 * 
 * @cython.binding(True)
 * cpdef double _peak_sequence_tic(self, peak_collection peaks) except -1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils._peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = -1.0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_py_peaks);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_31__pyx_fuse_4_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_12ms_deisotope_2_c_5utils_31__pyx_fuse_4_peak_sequence_tic = 
#line 216 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_4_peak_sequence_tic", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_31__pyx_fuse_4_peak_sequence_tic, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_16_peak_sequence_tic};

#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_31__pyx_fuse_4_peak_sequence_tic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_peaks = 0
#line 216 "ms_deisotope/_c/utils.pyx"
;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_4_peak_sequence_tic (wrapper)", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  {

#line 216 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peaks,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 216 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 216 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 216 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 216 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 216 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 216 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else 
#line 216 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 216 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 216 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 216 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4_peak_sequence_tic", 1, 2, 2, 1); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
        }

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 216 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4_peak_sequence_tic") < 0)) __PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
      }

#line 216 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 216 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 216 "ms_deisotope/_c/utils.pyx"
    } else {

#line 216 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 216 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 216 "ms_deisotope/_c/utils.pyx"
    }

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_self = values[0];

#line 216 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peaks = values[1];

#line 216 "ms_deisotope/_c/utils.pyx"
  }

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4_peak_sequence_tic", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 216 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 216, __pyx_L3_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_4_peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 216 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_30__pyx_fuse_4_peak_sequence_tic(__pyx_self, __pyx_v_self, __pyx_v_peaks);

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_30__pyx_fuse_4_peak_sequence_tic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_peaks) {

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 216 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_4_peak_sequence_tic", 0);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_fuse_4__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic(__pyx_v_self, __pyx_v_peaks, 0); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_2;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 216 "ms_deisotope/_c/utils.pyx"


#line 216 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_4_peak_sequence_tic", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 216 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 216 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":244
 * 
 * @cython.binding(True)
 * cpdef PeakBase _peak_sequence_bp(self, peak_collection peaks):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_19_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 244 "ms_deisotope/_c/utils.pyx"
static char __pyx_doc_12ms_deisotope_2_c_5utils_18_peak_sequence_bp[] = "_peak_sequence_bp(signatures, args, kwargs, defaults)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_5utils_19_peak_sequence_bp = 
#line 244 "ms_deisotope/_c/utils.pyx"
{"_peak_sequence_bp", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_19_peak_sequence_bp, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_18_peak_sequence_bp};

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_19_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_signatures = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_args = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  {

#line 244 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signatures,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    PyObject* values[4] = {0,0,0,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 244 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 244 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  4: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signatures)) != 0)) kw_args--;
        else 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;

#line 244 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 1); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
        }

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  2:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;

#line 244 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 2); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
        }

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  3:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;

#line 244 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 3); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
        }

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 244 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {

#line 244 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
    } else {

#line 244 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 244 "ms_deisotope/_c/utils.pyx"
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_signatures = values[0];

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_args = values[1];

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_kwargs = values[2];

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_defaults = values[3];

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 244 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_18_peak_sequence_bp(__pyx_self, __pyx_v_signatures, __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_18_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL
#line 244 "ms_deisotope/_c/utils.pyx"
;
  Py_ssize_t __pyx_v_i
#line 244 "ms_deisotope/_c/utils.pyx"
;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize
#line 244 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_arg = NULL
#line 244 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_candidates = NULL
#line 244 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_sig = NULL
#line 244 "ms_deisotope/_c/utils.pyx"
;
  int __pyx_v_match_found
#line 244 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_src_sig = NULL
#line 244 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_dst_type = NULL
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_2;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_3;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_4;

#line 244 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_5;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_7;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_8;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_9;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_11 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_13;

#line 244 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_14;

#line 244 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_15;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_16;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("_peak_sequence_bp", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(__pyx_v_kwargs);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_4) {

#line 244 "ms_deisotope/_c/utils.pyx"
  } else {

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 244 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L4_bool_binop_done;

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = ((!__pyx_t_4) != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_bool_binop_done:;

#line 244 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_v_itemsize = -1L;

#line 244 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_args == Py_None)) {

#line 244 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 244 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = ((1 < __pyx_t_5) != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 244 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 244 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(((PyObject*)__pyx_v_args), 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L6;

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_4) {

#line 244 "ms_deisotope/_c/utils.pyx"
  } else {

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 244 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L7_bool_binop_done;

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 244 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 244 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_peaks, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_t_4 != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L7_bool_binop_done:;

#line 244 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 244 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 244 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L6;

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  /*else*/ {

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 244 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 244 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_int_2);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_int_2);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_n_s_s);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_n_s_s);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L6:;

#line 244 "ms_deisotope/_c/utils.pyx"
  while (1) {

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_set);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_peak_picker__c_peak_set);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_3) {

#line 244 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L10_break;

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_peak_picker__c_peak_index);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_ms_peak_picker__c_peak_index);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PeakIndex); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = (__pyx_t_3 != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_2) {

#line 244 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_PeakIndex, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L10_break;

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_ms_deisotope__c_peak_set);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_DeconvolutedPeakSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_3) {

#line 244 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_DeconvolutedPeakSet, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L10_break;

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_kp_s_ms_deisotope__c_peak_set);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_ms_deisotope__c_peak_set);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_DeconvolutedPeakSetIndexed); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = (__pyx_t_3 != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_2) {

#line 244 "ms_deisotope/_c/utils.pyx"
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_DeconvolutedPeakSetIndexed, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L10_break;

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_object, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L10_break;

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L10_break:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_signatures == Py_None)) {

#line 244 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 244 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  while (1) {

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_9 == 0)) break;

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_match_found = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_12 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 244 "ms_deisotope/_c/utils.pyx"
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);

#line 244 "ms_deisotope/_c/utils.pyx"
      if (likely(__pyx_t_12)) {

#line 244 "ms_deisotope/_c/utils.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 244 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(__pyx_t_12);

#line 244 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(function);

#line 244 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__2) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__2);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_10);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_10 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {

#line 244 "ms_deisotope/_c/utils.pyx"
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);

#line 244 "ms_deisotope/_c/utils.pyx"
      if (likely(__pyx_t_10)) {

#line 244 "ms_deisotope/_c/utils.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);

#line 244 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(__pyx_t_10);

#line 244 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(function);

#line 244 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF_SET(__pyx_t_11, function);

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__3) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__3);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_14 = __pyx_t_13;

#line 244 "ms_deisotope/_c/utils.pyx"
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {

#line 244 "ms_deisotope/_c/utils.pyx"
      __pyx_v_i = __pyx_t_15;

#line 244 "ms_deisotope/_c/utils.pyx"
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
      __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);

#line 244 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = (__pyx_t_2 != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
      if (__pyx_t_3) {

#line 244 "ms_deisotope/_c/utils.pyx"
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
        if (__pyx_t_3) {

#line 244 "ms_deisotope/_c/utils.pyx"
          __pyx_v_match_found = 1;

#line 244 "ms_deisotope/_c/utils.pyx"
          goto __pyx_L20;

#line 244 "ms_deisotope/_c/utils.pyx"
        }

#line 244 "ms_deisotope/_c/utils.pyx"
        /*else*/ {

#line 244 "ms_deisotope/_c/utils.pyx"
          __pyx_v_match_found = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
          goto __pyx_L18_break;

#line 244 "ms_deisotope/_c/utils.pyx"
        }

#line 244 "ms_deisotope/_c/utils.pyx"
        __pyx_L20:;

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_L18_break:;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = (__pyx_v_match_found != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_3) {

#line 244 "ms_deisotope/_c/utils.pyx"
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = ((!__pyx_t_3) != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  /*else*/ {

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_signatures == Py_None)) {

#line 244 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 244 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_candidates, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_11 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_11);

#line 244 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_r = __pyx_t_11;

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_t_11 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L0;

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_11);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_12);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_dest_sig);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_ndarray);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_arg);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_candidates);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_sig);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_src_sig);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_dst_type);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_35__pyx_fuse_0_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_19_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 244 "ms_deisotope/_c/utils.pyx"
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 244 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_base_peak = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_max_intensity
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_2;

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_5;

#line 244 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_6;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0_peak_sequence_bp", 0);

  /* "ms_deisotope/_c/utils.pyx":251
 *         list py_peaks
 * 
 *     max_intensity = 0.0             # <<<<<<<<<<<<<<
 *     base_peak = None
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 251 "ms_deisotope/_c/utils.pyx"
  __pyx_v_max_intensity = 0.0;

  /* "ms_deisotope/_c/utils.pyx":252
 * 
 *     max_intensity = 0.0
 *     base_peak = None             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()
 */

#line 252 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(Py_None);

#line 252 "ms_deisotope/_c/utils.pyx"
  __pyx_v_base_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None);

  /* "ms_deisotope/_c/utils.pyx":254
 *     base_peak = None
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()             # <<<<<<<<<<<<<<
 *     else:
 *         py_peaks = list(peaks)
 */

#line 254 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);

  /* "ms_deisotope/_c/utils.pyx":261
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)
 */

#line 261 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 261 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 261 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 261 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/utils.pyx":263
 *     for i in range(n):
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)             # <<<<<<<<<<<<<<
 *         elif peak_collection is PeakIndex:
 *             peak = peaks.peaks.getitem(i)
 */

#line 263 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 263, __pyx_L1_error)

#line 263 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 263 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 263 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/utils.pyx":268
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:             # <<<<<<<<<<<<<<
 *             base_peak = peak
 *             max_intensity = peak.intensity
 */

#line 268 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = ((__pyx_v_max_intensity < __pyx_v_peak->intensity) != 0);

#line 268 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/utils.pyx":269
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:
 *             base_peak = peak             # <<<<<<<<<<<<<<
 *             max_intensity = peak.intensity
 *     return base_peak
 */

#line 269 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(((PyObject *)__pyx_v_peak));

#line 269 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_v_base_peak, __pyx_v_peak);

      /* "ms_deisotope/_c/utils.pyx":270
 *         if max_intensity < peak.intensity:
 *             base_peak = peak
 *             max_intensity = peak.intensity             # <<<<<<<<<<<<<<
 *     return base_peak
 * 
 */

#line 270 "ms_deisotope/_c/utils.pyx"
      __pyx_t_6 = __pyx_v_peak->intensity;

#line 270 "ms_deisotope/_c/utils.pyx"
      __pyx_v_max_intensity = __pyx_t_6;

      /* "ms_deisotope/_c/utils.pyx":268
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:             # <<<<<<<<<<<<<<
 *             base_peak = peak
 *             max_intensity = peak.intensity
 */

#line 268 "ms_deisotope/_c/utils.pyx"
    }

#line 268 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":271
 *             base_peak = peak
 *             max_intensity = peak.intensity
 *     return base_peak             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 271 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 271 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_base_peak));

#line 271 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_base_peak;

#line 271 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":244
 * 
 * @cython.binding(True)
 * cpdef PeakBase _peak_sequence_bp(self, peak_collection peaks):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils._peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_base_peak);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_35__pyx_fuse_0_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_5utils_35__pyx_fuse_0_peak_sequence_bp = 
#line 244 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_0_peak_sequence_bp", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_35__pyx_fuse_0_peak_sequence_bp, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_18_peak_sequence_bp};

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_35__pyx_fuse_0_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0_peak_sequence_bp (wrapper)", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  {

#line 244 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peaks,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 244 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 244 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 244 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_peak_sequence_bp", 1, 2, 2, 1); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
        }

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 244 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0_peak_sequence_bp") < 0)) __PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 244 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
    } else {

#line 244 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_self = values[0];

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peaks = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[1]);

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0_peak_sequence_bp", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_0_peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 244 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaks), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaks", 0))) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_r = 
#line 244 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_34__pyx_fuse_0_peak_sequence_bp(__pyx_self, __pyx_v_self, __pyx_v_peaks);

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_34__pyx_fuse_0_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks) {

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0_peak_sequence_bp", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(__pyx_v_self, __pyx_v_peaks, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_0_peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_37__pyx_fuse_1_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_19_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 244 "ms_deisotope/_c/utils.pyx"
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peaks, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 244 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_base_peak = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_max_intensity
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_2;

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_5;

#line 244 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_6;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1_peak_sequence_bp", 0);

  /* "ms_deisotope/_c/utils.pyx":251
 *         list py_peaks
 * 
 *     max_intensity = 0.0             # <<<<<<<<<<<<<<
 *     base_peak = None
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 251 "ms_deisotope/_c/utils.pyx"
  __pyx_v_max_intensity = 0.0;

  /* "ms_deisotope/_c/utils.pyx":252
 * 
 *     max_intensity = 0.0
 *     base_peak = None             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()
 */

#line 252 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(Py_None);

#line 252 "ms_deisotope/_c/utils.pyx"
  __pyx_v_base_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None);

  /* "ms_deisotope/_c/utils.pyx":254
 *     base_peak = None
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()             # <<<<<<<<<<<<<<
 *     else:
 *         py_peaks = list(peaks)
 */

#line 254 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);

  /* "ms_deisotope/_c/utils.pyx":261
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)
 */

#line 261 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 261 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 261 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 261 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/utils.pyx":265
 *             peak = peaks.getitem(i)
 *         elif peak_collection is PeakIndex:
 *             peak = peaks.peaks.getitem(i)             # <<<<<<<<<<<<<<
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 */

#line 265 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->peaks->__pyx_vtab)->getitem(__pyx_v_peaks->peaks, __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)

#line 265 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 265 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 265 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/utils.pyx":268
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:             # <<<<<<<<<<<<<<
 *             base_peak = peak
 *             max_intensity = peak.intensity
 */

#line 268 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = ((__pyx_v_max_intensity < __pyx_v_peak->intensity) != 0);

#line 268 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/utils.pyx":269
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:
 *             base_peak = peak             # <<<<<<<<<<<<<<
 *             max_intensity = peak.intensity
 *     return base_peak
 */

#line 269 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(((PyObject *)__pyx_v_peak));

#line 269 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_v_base_peak, __pyx_v_peak);

      /* "ms_deisotope/_c/utils.pyx":270
 *         if max_intensity < peak.intensity:
 *             base_peak = peak
 *             max_intensity = peak.intensity             # <<<<<<<<<<<<<<
 *     return base_peak
 * 
 */

#line 270 "ms_deisotope/_c/utils.pyx"
      __pyx_t_6 = __pyx_v_peak->intensity;

#line 270 "ms_deisotope/_c/utils.pyx"
      __pyx_v_max_intensity = __pyx_t_6;

      /* "ms_deisotope/_c/utils.pyx":268
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:             # <<<<<<<<<<<<<<
 *             base_peak = peak
 *             max_intensity = peak.intensity
 */

#line 268 "ms_deisotope/_c/utils.pyx"
    }

#line 268 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":271
 *             base_peak = peak
 *             max_intensity = peak.intensity
 *     return base_peak             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 271 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 271 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_base_peak));

#line 271 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_base_peak;

#line 271 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":244
 * 
 * @cython.binding(True)
 * cpdef PeakBase _peak_sequence_bp(self, peak_collection peaks):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils._peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_base_peak);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_37__pyx_fuse_1_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_5utils_37__pyx_fuse_1_peak_sequence_bp = 
#line 244 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_1_peak_sequence_bp", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_37__pyx_fuse_1_peak_sequence_bp, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_18_peak_sequence_bp};

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_37__pyx_fuse_1_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peaks = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1_peak_sequence_bp (wrapper)", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  {

#line 244 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peaks,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 244 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 244 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 244 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_peak_sequence_bp", 1, 2, 2, 1); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
        }

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 244 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1_peak_sequence_bp") < 0)) __PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 244 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
    } else {

#line 244 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_self = values[0];

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peaks = ((struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *)values[1]);

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1_peak_sequence_bp", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_1_peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 244 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaks), __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex, 1, "peaks", 0))) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_r = 
#line 244 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_36__pyx_fuse_1_peak_sequence_bp(__pyx_self, __pyx_v_self, __pyx_v_peaks);

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_36__pyx_fuse_1_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_v_peaks) {

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1_peak_sequence_bp", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(__pyx_v_self, __pyx_v_peaks, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_1_peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_39__pyx_fuse_2_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_19_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 244 "ms_deisotope/_c/utils.pyx"
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peaks, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 244 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_base_peak = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_max_intensity
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_2;

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_5;

#line 244 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_6;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2_peak_sequence_bp", 0);

  /* "ms_deisotope/_c/utils.pyx":251
 *         list py_peaks
 * 
 *     max_intensity = 0.0             # <<<<<<<<<<<<<<
 *     base_peak = None
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 251 "ms_deisotope/_c/utils.pyx"
  __pyx_v_max_intensity = 0.0;

  /* "ms_deisotope/_c/utils.pyx":252
 * 
 *     max_intensity = 0.0
 *     base_peak = None             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()
 */

#line 252 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(Py_None);

#line 252 "ms_deisotope/_c/utils.pyx"
  __pyx_v_base_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None);

  /* "ms_deisotope/_c/utils.pyx":254
 *     base_peak = None
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()             # <<<<<<<<<<<<<<
 *     else:
 *         py_peaks = list(peaks)
 */

#line 254 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);

  /* "ms_deisotope/_c/utils.pyx":261
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)
 */

#line 261 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 261 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 261 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 261 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/utils.pyx":263
 *     for i in range(n):
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)             # <<<<<<<<<<<<<<
 *         elif peak_collection is PeakIndex:
 *             peak = peaks.peaks.getitem(i)
 */

#line 263 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 263, __pyx_L1_error)

#line 263 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 263 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 263 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/utils.pyx":268
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:             # <<<<<<<<<<<<<<
 *             base_peak = peak
 *             max_intensity = peak.intensity
 */

#line 268 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = ((__pyx_v_max_intensity < __pyx_v_peak->intensity) != 0);

#line 268 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/utils.pyx":269
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:
 *             base_peak = peak             # <<<<<<<<<<<<<<
 *             max_intensity = peak.intensity
 *     return base_peak
 */

#line 269 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(((PyObject *)__pyx_v_peak));

#line 269 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_v_base_peak, __pyx_v_peak);

      /* "ms_deisotope/_c/utils.pyx":270
 *         if max_intensity < peak.intensity:
 *             base_peak = peak
 *             max_intensity = peak.intensity             # <<<<<<<<<<<<<<
 *     return base_peak
 * 
 */

#line 270 "ms_deisotope/_c/utils.pyx"
      __pyx_t_6 = __pyx_v_peak->intensity;

#line 270 "ms_deisotope/_c/utils.pyx"
      __pyx_v_max_intensity = __pyx_t_6;

      /* "ms_deisotope/_c/utils.pyx":268
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:             # <<<<<<<<<<<<<<
 *             base_peak = peak
 *             max_intensity = peak.intensity
 */

#line 268 "ms_deisotope/_c/utils.pyx"
    }

#line 268 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":271
 *             base_peak = peak
 *             max_intensity = peak.intensity
 *     return base_peak             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 271 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 271 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_base_peak));

#line 271 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_base_peak;

#line 271 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":244
 * 
 * @cython.binding(True)
 * cpdef PeakBase _peak_sequence_bp(self, peak_collection peaks):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils._peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_base_peak);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_39__pyx_fuse_2_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_5utils_39__pyx_fuse_2_peak_sequence_bp = 
#line 244 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_2_peak_sequence_bp", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_39__pyx_fuse_2_peak_sequence_bp, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_18_peak_sequence_bp};

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_39__pyx_fuse_2_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peaks = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2_peak_sequence_bp (wrapper)", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  {

#line 244 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peaks,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 244 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 244 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 244 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2_peak_sequence_bp", 1, 2, 2, 1); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
        }

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 244 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2_peak_sequence_bp") < 0)) __PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 244 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
    } else {

#line 244 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_self = values[0];

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peaks = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)values[1]);

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2_peak_sequence_bp", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_2_peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 244 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaks), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet, 1, "peaks", 0))) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_r = 
#line 244 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_38__pyx_fuse_2_peak_sequence_bp(__pyx_self, __pyx_v_self, __pyx_v_peaks);

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_38__pyx_fuse_2_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_v_peaks) {

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_2_peak_sequence_bp", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(__pyx_v_self, __pyx_v_peaks, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_2_peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_41__pyx_fuse_3_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_19_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 244 "ms_deisotope/_c/utils.pyx"
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_v_peaks, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 244 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_base_peak = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_max_intensity
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_2;

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_5;

#line 244 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_6;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3_peak_sequence_bp", 0);

  /* "ms_deisotope/_c/utils.pyx":251
 *         list py_peaks
 * 
 *     max_intensity = 0.0             # <<<<<<<<<<<<<<
 *     base_peak = None
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 251 "ms_deisotope/_c/utils.pyx"
  __pyx_v_max_intensity = 0.0;

  /* "ms_deisotope/_c/utils.pyx":252
 * 
 *     max_intensity = 0.0
 *     base_peak = None             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()
 */

#line 252 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(Py_None);

#line 252 "ms_deisotope/_c/utils.pyx"
  __pyx_v_base_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None);

  /* "ms_deisotope/_c/utils.pyx":254
 *     base_peak = None
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()             # <<<<<<<<<<<<<<
 *     else:
 *         py_peaks = list(peaks)
 */

#line 254 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *)__pyx_v_peaks->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peaks));

  /* "ms_deisotope/_c/utils.pyx":261
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)
 */

#line 261 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 261 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 261 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 261 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/utils.pyx":263
 *     for i in range(n):
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)             # <<<<<<<<<<<<<<
 *         elif peak_collection is PeakIndex:
 *             peak = peaks.peaks.getitem(i)
 */

#line 263 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *)__pyx_v_peaks->__pyx_base.__pyx_vtab)->__pyx_base.getitem(((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *)__pyx_v_peaks), __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 263, __pyx_L1_error)

#line 263 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 263 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_4));

#line 263 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/utils.pyx":268
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:             # <<<<<<<<<<<<<<
 *             base_peak = peak
 *             max_intensity = peak.intensity
 */

#line 268 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = ((__pyx_v_max_intensity < __pyx_v_peak->intensity) != 0);

#line 268 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/utils.pyx":269
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:
 *             base_peak = peak             # <<<<<<<<<<<<<<
 *             max_intensity = peak.intensity
 *     return base_peak
 */

#line 269 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(((PyObject *)__pyx_v_peak));

#line 269 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_v_base_peak, __pyx_v_peak);

      /* "ms_deisotope/_c/utils.pyx":270
 *         if max_intensity < peak.intensity:
 *             base_peak = peak
 *             max_intensity = peak.intensity             # <<<<<<<<<<<<<<
 *     return base_peak
 * 
 */

#line 270 "ms_deisotope/_c/utils.pyx"
      __pyx_t_6 = __pyx_v_peak->intensity;

#line 270 "ms_deisotope/_c/utils.pyx"
      __pyx_v_max_intensity = __pyx_t_6;

      /* "ms_deisotope/_c/utils.pyx":268
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:             # <<<<<<<<<<<<<<
 *             base_peak = peak
 *             max_intensity = peak.intensity
 */

#line 268 "ms_deisotope/_c/utils.pyx"
    }

#line 268 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":271
 *             base_peak = peak
 *             max_intensity = peak.intensity
 *     return base_peak             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 271 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 271 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_base_peak));

#line 271 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_base_peak;

#line 271 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":244
 * 
 * @cython.binding(True)
 * cpdef PeakBase _peak_sequence_bp(self, peak_collection peaks):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils._peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_base_peak);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_41__pyx_fuse_3_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_5utils_41__pyx_fuse_3_peak_sequence_bp = 
#line 244 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_3_peak_sequence_bp", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_41__pyx_fuse_3_peak_sequence_bp, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_18_peak_sequence_bp};

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_41__pyx_fuse_3_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_v_peaks = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3_peak_sequence_bp (wrapper)", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  {

#line 244 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peaks,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 244 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 244 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 244 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3_peak_sequence_bp", 1, 2, 2, 1); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
        }

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 244 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3_peak_sequence_bp") < 0)) __PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 244 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
    } else {

#line 244 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_self = values[0];

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peaks = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *)values[1]);

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3_peak_sequence_bp", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_3_peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 244 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaks), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed, 1, "peaks", 0))) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_r = 
#line 244 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_40__pyx_fuse_3_peak_sequence_bp(__pyx_self, __pyx_v_self, __pyx_v_peaks);

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_40__pyx_fuse_3_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_v_peaks) {

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_3_peak_sequence_bp", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(__pyx_v_self, __pyx_v_peaks, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_3_peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_43__pyx_fuse_4_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_19_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 244 "ms_deisotope/_c/utils.pyx"
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_fuse_4__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_peaks, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 244 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_base_peak = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_max_intensity
#line 244 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_py_peaks = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_2;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_3;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_4;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_6;

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_7;

#line 244 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_8;

#line 244 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_9;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_4_peak_sequence_bp", 0);

  /* "ms_deisotope/_c/utils.pyx":251
 *         list py_peaks
 * 
 *     max_intensity = 0.0             # <<<<<<<<<<<<<<
 *     base_peak = None
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 251 "ms_deisotope/_c/utils.pyx"
  __pyx_v_max_intensity = 0.0;

  /* "ms_deisotope/_c/utils.pyx":252
 * 
 *     max_intensity = 0.0
 *     base_peak = None             # <<<<<<<<<<<<<<
 *     if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *         n = peaks.get_size()
 */

#line 252 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(Py_None);

#line 252 "ms_deisotope/_c/utils.pyx"
  __pyx_v_base_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None);

  /* "ms_deisotope/_c/utils.pyx":256
 *         n = peaks.get_size()
 *     else:
 *         py_peaks = list(peaks)             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(py_peaks)
 *         if n > 0:
 */

#line 256 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PySequence_List(__pyx_v_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)

#line 256 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 256 "ms_deisotope/_c/utils.pyx"
  __pyx_v_py_peaks = ((PyObject*)__pyx_t_1);

#line 256 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":257
 *     else:
 *         py_peaks = list(peaks)
 *         n = PyList_GET_SIZE(py_peaks)             # <<<<<<<<<<<<<<
 *         if n > 0:
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 */

#line 257 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_py_peaks);

  /* "ms_deisotope/_c/utils.pyx":258
 *         py_peaks = list(peaks)
 *         n = PyList_GET_SIZE(py_peaks)
 *         if n > 0:             # <<<<<<<<<<<<<<
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 */

#line 258 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = ((__pyx_v_n > 0) != 0);

#line 258 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/utils.pyx":259
 *         n = PyList_GET_SIZE(py_peaks)
 *         if n > 0:
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):
 */

#line 259 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = PyList_GET_ITEM(__pyx_v_py_peaks, 0);

#line 259 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_t_3), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase); 

#line 259 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);

#line 259 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_4)) {

      /* "ms_deisotope/_c/utils.pyx":260
 *         if n > 0:
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 */

#line 260 "ms_deisotope/_c/utils.pyx"
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_py_peaks, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)

#line 260 "ms_deisotope/_c/utils.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 260 "ms_deisotope/_c/utils.pyx"
      __pyx_t_5 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_interpret_r_as_a_PeakBase, ((PyObject *)Py_TYPE(__pyx_t_1))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 260, __pyx_L1_error)

#line 260 "ms_deisotope/_c/utils.pyx"
      __Pyx_GOTREF(__pyx_t_5);

#line 260 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 260 "ms_deisotope/_c/utils.pyx"
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)

#line 260 "ms_deisotope/_c/utils.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 260 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 260 "ms_deisotope/_c/utils.pyx"
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 260 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 260 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 260, __pyx_L1_error)

      /* "ms_deisotope/_c/utils.pyx":259
 *         n = PyList_GET_SIZE(py_peaks)
 *         if n > 0:
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):
 */

#line 259 "ms_deisotope/_c/utils.pyx"
    }

    /* "ms_deisotope/_c/utils.pyx":258
 *         py_peaks = list(peaks)
 *         n = PyList_GET_SIZE(py_peaks)
 *         if n > 0:             # <<<<<<<<<<<<<<
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 */

#line 258 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":261
 *             if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
 *                 raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
 *             peak = peaks.getitem(i)
 */

#line 261 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = __pyx_v_n;

#line 261 "ms_deisotope/_c/utils.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 261 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 261 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/utils.pyx":267
 *             peak = peaks.peaks.getitem(i)
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)             # <<<<<<<<<<<<<<
 *         if max_intensity < peak.intensity:
 *             base_peak = peak
 */

#line 267 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = PyList_GET_ITEM(__pyx_v_py_peaks, __pyx_v_i);

#line 267 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_3);

#line 267 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 267 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 267 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/utils.pyx":268
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:             # <<<<<<<<<<<<<<
 *             base_peak = peak
 *             max_intensity = peak.intensity
 */

#line 268 "ms_deisotope/_c/utils.pyx"
    __pyx_t_4 = ((__pyx_v_max_intensity < __pyx_v_peak->intensity) != 0);

#line 268 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_4) {

      /* "ms_deisotope/_c/utils.pyx":269
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:
 *             base_peak = peak             # <<<<<<<<<<<<<<
 *             max_intensity = peak.intensity
 *     return base_peak
 */

#line 269 "ms_deisotope/_c/utils.pyx"
      __Pyx_INCREF(((PyObject *)__pyx_v_peak));

#line 269 "ms_deisotope/_c/utils.pyx"
      __Pyx_DECREF_SET(__pyx_v_base_peak, __pyx_v_peak);

      /* "ms_deisotope/_c/utils.pyx":270
 *         if max_intensity < peak.intensity:
 *             base_peak = peak
 *             max_intensity = peak.intensity             # <<<<<<<<<<<<<<
 *     return base_peak
 * 
 */

#line 270 "ms_deisotope/_c/utils.pyx"
      __pyx_t_9 = __pyx_v_peak->intensity;

#line 270 "ms_deisotope/_c/utils.pyx"
      __pyx_v_max_intensity = __pyx_t_9;

      /* "ms_deisotope/_c/utils.pyx":268
 *         else:
 *             peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
 *         if max_intensity < peak.intensity:             # <<<<<<<<<<<<<<
 *             base_peak = peak
 *             max_intensity = peak.intensity
 */

#line 268 "ms_deisotope/_c/utils.pyx"
    }

#line 268 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":271
 *             base_peak = peak
 *             max_intensity = peak.intensity
 *     return base_peak             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 271 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 271 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_base_peak));

#line 271 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_v_base_peak;

#line 271 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":244
 * 
 * @cython.binding(True)
 * cpdef PeakBase _peak_sequence_bp(self, peak_collection peaks):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils._peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_base_peak);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_py_peaks);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_43__pyx_fuse_4_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_12ms_deisotope_2_c_5utils_43__pyx_fuse_4_peak_sequence_bp = 
#line 244 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_4_peak_sequence_bp", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_43__pyx_fuse_4_peak_sequence_bp, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_18_peak_sequence_bp};

#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_43__pyx_fuse_4_peak_sequence_bp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_peaks = 0
#line 244 "ms_deisotope/_c/utils.pyx"
;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_4_peak_sequence_bp (wrapper)", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  {

#line 244 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_peaks,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 244 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 244 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 244 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 244 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 244 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 244 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else 
#line 244 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 244 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 244 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 244 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4_peak_sequence_bp", 1, 2, 2, 1); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
        }

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 244 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4_peak_sequence_bp") < 0)) __PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
      }

#line 244 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 244 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 244 "ms_deisotope/_c/utils.pyx"
    } else {

#line 244 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 244 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 244 "ms_deisotope/_c/utils.pyx"
    }

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_self = values[0];

#line 244 "ms_deisotope/_c/utils.pyx"
    __pyx_v_peaks = values[1];

#line 244 "ms_deisotope/_c/utils.pyx"
  }

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4_peak_sequence_bp", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 244 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 244, __pyx_L3_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_4_peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 244 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_42__pyx_fuse_4_peak_sequence_bp(__pyx_self, __pyx_v_self, __pyx_v_peaks);

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_42__pyx_fuse_4_peak_sequence_bp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_peaks) {

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 244 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_4_peak_sequence_bp", 0);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_fuse_4__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp(__pyx_v_self, __pyx_v_peaks, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 244 "ms_deisotope/_c/utils.pyx"


#line 244 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_4_peak_sequence_bp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 244 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 244 "ms_deisotope/_c/utils.pyx"
}

/* "ms_deisotope/_c/utils.pyx":274
 * 
 * 
 * cpdef double correlation(cython.floating[:] x, cython.floating[:] y):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_21correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 274 "ms_deisotope/_c/utils.pyx"
static char __pyx_doc_12ms_deisotope_2_c_5utils_20correlation[] = "correlation(signatures, args, kwargs, defaults)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_5utils_21correlation = 
#line 274 "ms_deisotope/_c/utils.pyx"
{"correlation", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_21correlation, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_20correlation};

#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_21correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_signatures = 0
#line 274 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_args = 0
#line 274 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 274 "ms_deisotope/_c/utils.pyx"
;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0
#line 274 "ms_deisotope/_c/utils.pyx"
;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  {

#line 274 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signatures,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};

#line 274 "ms_deisotope/_c/utils.pyx"
    PyObject* values[4] = {0,0,0,0};

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 274 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 274 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 274 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  4: 
#line 274 "ms_deisotope/_c/utils.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 274 "ms_deisotope/_c/utils.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 274 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 274 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 274 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 274 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 274 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 274 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 274 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signatures)) != 0)) kw_args--;
        else 
#line 274 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 274 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 274 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;

#line 274 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 1); 
#line 274 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        }

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 274 "ms_deisotope/_c/utils.pyx"
        case  2:

#line 274 "ms_deisotope/_c/utils.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;

#line 274 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 2); 
#line 274 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        }

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 274 "ms_deisotope/_c/utils.pyx"
        case  3:

#line 274 "ms_deisotope/_c/utils.pyx"
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;

#line 274 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 3); 
#line 274 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        }

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 274 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {

#line 274 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 274 "ms_deisotope/_c/utils.pyx"
    } else {

#line 274 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 274 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 274 "ms_deisotope/_c/utils.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 274 "ms_deisotope/_c/utils.pyx"
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_v_signatures = values[0];

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_v_args = values[1];

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_v_kwargs = values[2];

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_v_defaults = values[3];

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); 
#line 274 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 274 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_20correlation(__pyx_self, __pyx_v_signatures, __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 274 "ms_deisotope/_c/utils.pyx"
}

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_20correlation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL
#line 274 "ms_deisotope/_c/utils.pyx"
;
  Py_ssize_t __pyx_v_i
#line 274 "ms_deisotope/_c/utils.pyx"
;
  PyTypeObject *__pyx_v_ndarray = 0
#line 274 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_memviewslice __pyx_v_memslice
#line 274 "ms_deisotope/_c/utils.pyx"
;
  Py_ssize_t __pyx_v_itemsize
#line 274 "ms_deisotope/_c/utils.pyx"
;
  CYTHON_UNUSED int __pyx_v_dtype_signed
#line 274 "ms_deisotope/_c/utils.pyx"
;
  char __pyx_v_kind
#line 274 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_arg = NULL
#line 274 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_dtype = NULL
#line 274 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_arg_base = NULL
#line 274 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_candidates = NULL
#line 274 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_sig = NULL
#line 274 "ms_deisotope/_c/utils.pyx"
;
  int __pyx_v_match_found
#line 274 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_src_sig = NULL
#line 274 "ms_deisotope/_c/utils.pyx"
;
  PyObject *__pyx_v_dst_type = NULL
#line 274 "ms_deisotope/_c/utils.pyx"
;

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_2;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_3;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_4;

#line 274 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_5;

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  long __pyx_t_7;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_memviewslice __pyx_t_8;

#line 274 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_9;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_10;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_11;

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_13 = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_14 = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_15;

#line 274 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_16;

#line 274 "ms_deisotope/_c/utils.pyx"
  Py_ssize_t __pyx_t_17;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_18;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("correlation", 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(__pyx_v_kwargs);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_INCREF(Py_None);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(Py_None);

#line 274 "ms_deisotope/_c/utils.pyx"
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_4) {

#line 274 "ms_deisotope/_c/utils.pyx"
  } else {

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 274 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L4_bool_binop_done;

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = ((!__pyx_t_4) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_bool_binop_done:;

#line 274 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(Py_None);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_v_itemsize = -1L;

#line 274 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_args == Py_None)) {

#line 274 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 274 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = ((0 < __pyx_t_5) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 274 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 274 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(((PyObject*)__pyx_v_args), 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L6;

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_4) {

#line 274 "ms_deisotope/_c/utils.pyx"
  } else {

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = __pyx_t_4;

#line 274 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L7_bool_binop_done;

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 274 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 274 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_x, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = (__pyx_t_4 != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_t_3;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L7_bool_binop_done:;

#line 274 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_kwargs == Py_None)) {

#line 274 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 274 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_v_arg = __pyx_t_1;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L6;

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  /*else*/ {

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_args == Py_None)) {

#line 274 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 274 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_int_2);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_int_2);

#line 274 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_INCREF(__pyx_n_s_s);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_n_s_s);

#line 274 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L6:;

#line 274 "ms_deisotope/_c/utils.pyx"
  while (1) {

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = (__pyx_v_ndarray != ((PyTypeObject*)Py_None));

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_3) {

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg, __pyx_v_ndarray); 

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_2 = (__pyx_t_3 != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
      if (__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_v_dtype = __pyx_t_6;

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
        goto __pyx_L12;

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_2 = __pyx_memoryview_check(__pyx_v_arg); 

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = (__pyx_t_2 != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
      if (__pyx_t_3) {

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_base); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_v_arg_base = __pyx_t_6;

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg_base, __pyx_v_ndarray); 

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_2 = (__pyx_t_3 != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
        if (__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg_base, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_v_dtype = __pyx_t_6;

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
          goto __pyx_L13;

#line 274 "ms_deisotope/_c/utils.pyx"
        }

#line 274 "ms_deisotope/_c/utils.pyx"
        /*else*/ {

#line 274 "ms_deisotope/_c/utils.pyx"
          __Pyx_INCREF(Py_None);

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_v_dtype = Py_None;

#line 274 "ms_deisotope/_c/utils.pyx"
        }

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_L13:;

#line 274 "ms_deisotope/_c/utils.pyx"
        goto __pyx_L12;

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
      /*else*/ {

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(Py_None);

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_v_dtype = Py_None;

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_L12:;

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_v_itemsize = -1L;

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_2 = (__pyx_v_dtype != Py_None);

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = (__pyx_t_2 != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
      if (__pyx_t_3) {

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_v_itemsize = __pyx_t_5;

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_t_6); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_v_kind = __pyx_t_7;

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_v_dtype_signed = (__pyx_v_kind == 'i');

#line 274 "ms_deisotope/_c/utils.pyx"
        switch (__pyx_v_kind) {

#line 274 "ms_deisotope/_c/utils.pyx"
          case 'i':

#line 274 "ms_deisotope/_c/utils.pyx"
          case 'u':

#line 274 "ms_deisotope/_c/utils.pyx"
          break;

#line 274 "ms_deisotope/_c/utils.pyx"
          case 'f':

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_2 = (((sizeof(float)) == __pyx_v_itemsize) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
          if (__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
          } else {

#line 274 "ms_deisotope/_c/utils.pyx"
            __pyx_t_3 = __pyx_t_2;

#line 274 "ms_deisotope/_c/utils.pyx"
            goto __pyx_L16_bool_binop_done;

#line 274 "ms_deisotope/_c/utils.pyx"
          }

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_3 = __pyx_t_2;

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_L16_bool_binop_done:;

#line 274 "ms_deisotope/_c/utils.pyx"
          if (__pyx_t_3) {

#line 274 "ms_deisotope/_c/utils.pyx"
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
            goto __pyx_L10_break;

#line 274 "ms_deisotope/_c/utils.pyx"
          }

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_2 = (((sizeof(double)) == __pyx_v_itemsize) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
          if (__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
          } else {

#line 274 "ms_deisotope/_c/utils.pyx"
            __pyx_t_3 = __pyx_t_2;

#line 274 "ms_deisotope/_c/utils.pyx"
            goto __pyx_L19_bool_binop_done;

#line 274 "ms_deisotope/_c/utils.pyx"
          }

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_t_3 = __pyx_t_2;

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_L19_bool_binop_done:;

#line 274 "ms_deisotope/_c/utils.pyx"
          if (__pyx_t_3) {

#line 274 "ms_deisotope/_c/utils.pyx"
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
            goto __pyx_L10_break;

#line 274 "ms_deisotope/_c/utils.pyx"
          }

#line 274 "ms_deisotope/_c/utils.pyx"
          break;

#line 274 "ms_deisotope/_c/utils.pyx"
          case 'c':

#line 274 "ms_deisotope/_c/utils.pyx"
          break;

#line 274 "ms_deisotope/_c/utils.pyx"
          case 'O':

#line 274 "ms_deisotope/_c/utils.pyx"
          break;

#line 274 "ms_deisotope/_c/utils.pyx"
          default: break;

#line 274 "ms_deisotope/_c/utils.pyx"
        }

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
    if (!__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
    } else {

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = __pyx_t_2;

#line 274 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L22_bool_binop_done;

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(float))) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = __pyx_t_2;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_L22_bool_binop_done:;

#line 274 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_3) {

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_float(__pyx_v_arg, 0); 

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_v_memslice = __pyx_t_8;

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
      if (__pyx_t_3) {

#line 274 "ms_deisotope/_c/utils.pyx"
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 

#line 274 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        goto __pyx_L10_break;

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
      /*else*/ {

#line 274 "ms_deisotope/_c/utils.pyx"
        PyErr_Clear(); 

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
    if (!__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
    } else {

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = __pyx_t_2;

#line 274 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L26_bool_binop_done;

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(double))) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_3 = __pyx_t_2;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_L26_bool_binop_done:;

#line 274 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_3) {

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_v_arg, 0); 

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_v_memslice = __pyx_t_8;

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
      if (__pyx_t_3) {

#line 274 "ms_deisotope/_c/utils.pyx"
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 

#line 274 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        goto __pyx_L10_break;

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
      /*else*/ {

#line 274 "ms_deisotope/_c/utils.pyx"
        PyErr_Clear(); 

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L10_break;

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L10_break:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_5 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_signatures == Py_None)) {

#line 274 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");

#line 274 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_6 = __pyx_t_1;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  while (1) {

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_9, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_10);

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_11 == 0)) break;

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_v_match_found = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_13);

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_14 = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);

#line 274 "ms_deisotope/_c/utils.pyx"
      if (likely(__pyx_t_14)) {

#line 274 "ms_deisotope/_c/utils.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(__pyx_t_14);

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(function);

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF_SET(__pyx_t_13, function);

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_12 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_kp_s__2) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__2);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_12);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_split); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_13);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_12 = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);

#line 274 "ms_deisotope/_c/utils.pyx"
      if (likely(__pyx_t_12)) {

#line 274 "ms_deisotope/_c/utils.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(__pyx_t_12);

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_INCREF(function);

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF_SET(__pyx_t_13, function);

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_12, __pyx_kp_s__3) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__3);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_16 = __pyx_t_15;

#line 274 "ms_deisotope/_c/utils.pyx"
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_v_i = __pyx_t_17;

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_2 = (__pyx_t_3 != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
      if (__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
        if (__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_v_match_found = 1;

#line 274 "ms_deisotope/_c/utils.pyx"
          goto __pyx_L34;

#line 274 "ms_deisotope/_c/utils.pyx"
        }

#line 274 "ms_deisotope/_c/utils.pyx"
        /*else*/ {

#line 274 "ms_deisotope/_c/utils.pyx"
          __pyx_v_match_found = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
          goto __pyx_L32_break;

#line 274 "ms_deisotope/_c/utils.pyx"
        }

#line 274 "ms_deisotope/_c/utils.pyx"
        __pyx_L34:;

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_L32_break:;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_2 = (__pyx_v_match_found != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
    if (__pyx_t_2) {

#line 274 "ms_deisotope/_c/utils.pyx"
      __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = ((!__pyx_t_2) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_3) {

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = ((__pyx_t_9 > 1) != 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_3) {

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  /*else*/ {

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_v_signatures == Py_None)) {

#line 274 "ms_deisotope/_c/utils.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 274 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_candidates, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_13 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), __pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_GOTREF(__pyx_t_13);

#line 274 "ms_deisotope/_c/utils.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_r = __pyx_t_13;

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_t_13 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L0;

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_12);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_13);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_14);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_dest_sig);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_ndarray);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_arg);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_dtype);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_arg_base);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_candidates);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_sig);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_src_sig);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_dst_type);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 274 "ms_deisotope/_c/utils.pyx"
}

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_47__pyx_fuse_0correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_21correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 274 "ms_deisotope/_c/utils.pyx"
static double __pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils_correlation(__Pyx_memviewslice __pyx_v_x, __Pyx_memviewslice __pyx_v_y, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 274 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_xsum
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_ysum
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_xmean
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_ymean
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_cov
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_varx
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_vary
#line 274 "ms_deisotope/_c/utils.pyx"
;

#line 274 "ms_deisotope/_c/utils.pyx"
  double __pyx_r;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 274 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_1;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_2;

#line 274 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 274 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_4;

#line 274 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_5;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_6;

#line 274 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_7;

#line 274 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_8;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0correlation", 0);

  /* "ms_deisotope/_c/utils.pyx":279
 *         double xsum, ysum, xmean, ymean
 *         double cov, varx, vary
 *     n = len(x)             # <<<<<<<<<<<<<<
 *     if n == 0:
 *         return NAN
 */

#line 279 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_MemoryView_Len(__pyx_v_x); 

#line 279 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = __pyx_t_1;

  /* "ms_deisotope/_c/utils.pyx":280
 *         double cov, varx, vary
 *     n = len(x)
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return NAN
 *     xsum = 0.
 */

#line 280 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = ((__pyx_v_n == 0) != 0);

#line 280 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/utils.pyx":281
 *     n = len(x)
 *     if n == 0:
 *         return NAN             # <<<<<<<<<<<<<<
 *     xsum = 0.
 *     ysum = 0.
 */

#line 281 "ms_deisotope/_c/utils.pyx"
    __pyx_r = NPY_NAN;

#line 281 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/utils.pyx":280
 *         double cov, varx, vary
 *     n = len(x)
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return NAN
 *     xsum = 0.
 */

#line 280 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":282
 *     if n == 0:
 *         return NAN
 *     xsum = 0.             # <<<<<<<<<<<<<<
 *     ysum = 0.
 *     for i in range(n):
 */

#line 282 "ms_deisotope/_c/utils.pyx"
  __pyx_v_xsum = 0.;

  /* "ms_deisotope/_c/utils.pyx":283
 *         return NAN
 *     xsum = 0.
 *     ysum = 0.             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         xsum += x[i]
 */

#line 283 "ms_deisotope/_c/utils.pyx"
  __pyx_v_ysum = 0.;

  /* "ms_deisotope/_c/utils.pyx":284
 *     xsum = 0.
 *     ysum = 0.
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         xsum += x[i]
 *         ysum += y[i]
 */

#line 284 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 284 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_t_1;

#line 284 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 284 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/utils.pyx":285
 *     ysum = 0.
 *     for i in range(n):
 *         xsum += x[i]             # <<<<<<<<<<<<<<
 *         ysum += y[i]
 *     xmean = xsum / n
 */

#line 285 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = __pyx_v_i;

#line 285 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 285 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_5 >= (size_t)__pyx_v_x.shape[0])) __pyx_t_6 = 0;

#line 285 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 285 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 285 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 285, __pyx_L1_error)

#line 285 "ms_deisotope/_c/utils.pyx"
    }

#line 285 "ms_deisotope/_c/utils.pyx"
    __pyx_v_xsum = (__pyx_v_xsum + (*((float *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_5 * __pyx_v_x.strides[0]) ))));

    /* "ms_deisotope/_c/utils.pyx":286
 *     for i in range(n):
 *         xsum += x[i]
 *         ysum += y[i]             # <<<<<<<<<<<<<<
 *     xmean = xsum / n
 *     ymean = ysum / n
 */

#line 286 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = __pyx_v_i;

#line 286 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 286 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_5 >= (size_t)__pyx_v_y.shape[0])) __pyx_t_6 = 0;

#line 286 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 286 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 286 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 286, __pyx_L1_error)

#line 286 "ms_deisotope/_c/utils.pyx"
    }

#line 286 "ms_deisotope/_c/utils.pyx"
    __pyx_v_ysum = (__pyx_v_ysum + (*((float *) ( /* dim=0 */ (__pyx_v_y.data + __pyx_t_5 * __pyx_v_y.strides[0]) ))));

#line 286 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":287
 *         xsum += x[i]
 *         ysum += y[i]
 *     xmean = xsum / n             # <<<<<<<<<<<<<<
 *     ymean = ysum / n
 *     cov = 0.
 */

#line 287 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_n == 0)) {

#line 287 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 287 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 287, __pyx_L1_error)

#line 287 "ms_deisotope/_c/utils.pyx"
  }

#line 287 "ms_deisotope/_c/utils.pyx"
  __pyx_v_xmean = (__pyx_v_xsum / __pyx_v_n);

  /* "ms_deisotope/_c/utils.pyx":288
 *         ysum += y[i]
 *     xmean = xsum / n
 *     ymean = ysum / n             # <<<<<<<<<<<<<<
 *     cov = 0.
 *     varx = 0.
 */

#line 288 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_n == 0)) {

#line 288 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 288 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 288, __pyx_L1_error)

#line 288 "ms_deisotope/_c/utils.pyx"
  }

#line 288 "ms_deisotope/_c/utils.pyx"
  __pyx_v_ymean = (__pyx_v_ysum / __pyx_v_n);

  /* "ms_deisotope/_c/utils.pyx":289
 *     xmean = xsum / n
 *     ymean = ysum / n
 *     cov = 0.             # <<<<<<<<<<<<<<
 *     varx = 0.
 *     vary = 0.
 */

#line 289 "ms_deisotope/_c/utils.pyx"
  __pyx_v_cov = 0.;

  /* "ms_deisotope/_c/utils.pyx":290
 *     ymean = ysum / n
 *     cov = 0.
 *     varx = 0.             # <<<<<<<<<<<<<<
 *     vary = 0.
 *     for i in range(n):
 */

#line 290 "ms_deisotope/_c/utils.pyx"
  __pyx_v_varx = 0.;

  /* "ms_deisotope/_c/utils.pyx":291
 *     cov = 0.
 *     varx = 0.
 *     vary = 0.             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         cov += (x[i] - xmean) * (y[i] - ymean)
 */

#line 291 "ms_deisotope/_c/utils.pyx"
  __pyx_v_vary = 0.;

  /* "ms_deisotope/_c/utils.pyx":292
 *     varx = 0.
 *     vary = 0.
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         cov += (x[i] - xmean) * (y[i] - ymean)
 *         varx += (x[i] - xmean) ** 2
 */

#line 292 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 292 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_t_1;

#line 292 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 292 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/utils.pyx":293
 *     vary = 0.
 *     for i in range(n):
 *         cov += (x[i] - xmean) * (y[i] - ymean)             # <<<<<<<<<<<<<<
 *         varx += (x[i] - xmean) ** 2
 *         vary += (y[i] - ymean) ** 2
 */

#line 293 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = __pyx_v_i;

#line 293 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 293 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_5 >= (size_t)__pyx_v_x.shape[0])) __pyx_t_6 = 0;

#line 293 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 293 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 293 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 293, __pyx_L1_error)

#line 293 "ms_deisotope/_c/utils.pyx"
    }

#line 293 "ms_deisotope/_c/utils.pyx"
    __pyx_t_7 = __pyx_v_i;

#line 293 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 293 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 >= (size_t)__pyx_v_y.shape[0])) __pyx_t_6 = 0;

#line 293 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 293 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 293 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 293, __pyx_L1_error)

#line 293 "ms_deisotope/_c/utils.pyx"
    }

#line 293 "ms_deisotope/_c/utils.pyx"
    __pyx_v_cov = (__pyx_v_cov + (((*((float *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_5 * __pyx_v_x.strides[0]) ))) - __pyx_v_xmean) * ((*((float *) ( /* dim=0 */ (__pyx_v_y.data + __pyx_t_7 * __pyx_v_y.strides[0]) ))) - __pyx_v_ymean)));

    /* "ms_deisotope/_c/utils.pyx":294
 *     for i in range(n):
 *         cov += (x[i] - xmean) * (y[i] - ymean)
 *         varx += (x[i] - xmean) ** 2             # <<<<<<<<<<<<<<
 *         vary += (y[i] - ymean) ** 2
 *     return cov / (sqrt(varx) * sqrt(vary))
 */

#line 294 "ms_deisotope/_c/utils.pyx"
    __pyx_t_7 = __pyx_v_i;

#line 294 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 294 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 >= (size_t)__pyx_v_x.shape[0])) __pyx_t_6 = 0;

#line 294 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 294 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 294 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/utils.pyx"
    }

#line 294 "ms_deisotope/_c/utils.pyx"
    __pyx_v_varx = (__pyx_v_varx + pow(((*((float *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_7 * __pyx_v_x.strides[0]) ))) - __pyx_v_xmean), 2.0));

    /* "ms_deisotope/_c/utils.pyx":295
 *         cov += (x[i] - xmean) * (y[i] - ymean)
 *         varx += (x[i] - xmean) ** 2
 *         vary += (y[i] - ymean) ** 2             # <<<<<<<<<<<<<<
 *     return cov / (sqrt(varx) * sqrt(vary))
 */

#line 295 "ms_deisotope/_c/utils.pyx"
    __pyx_t_7 = __pyx_v_i;

#line 295 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 295 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 >= (size_t)__pyx_v_y.shape[0])) __pyx_t_6 = 0;

#line 295 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 295 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 295 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/utils.pyx"
    }

#line 295 "ms_deisotope/_c/utils.pyx"
    __pyx_v_vary = (__pyx_v_vary + pow(((*((float *) ( /* dim=0 */ (__pyx_v_y.data + __pyx_t_7 * __pyx_v_y.strides[0]) ))) - __pyx_v_ymean), 2.0));

#line 295 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":296
 *         varx += (x[i] - xmean) ** 2
 *         vary += (y[i] - ymean) ** 2
 *     return cov / (sqrt(varx) * sqrt(vary))             # <<<<<<<<<<<<<<
 */

#line 296 "ms_deisotope/_c/utils.pyx"
  __pyx_t_8 = (sqrt(__pyx_v_varx) * sqrt(__pyx_v_vary));

#line 296 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_t_8 == 0)) {

#line 296 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 296 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 296, __pyx_L1_error)

#line 296 "ms_deisotope/_c/utils.pyx"
  }

#line 296 "ms_deisotope/_c/utils.pyx"
  __pyx_r = (__pyx_v_cov / __pyx_t_8);

#line 296 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":274
 * 
 * 
 * cpdef double correlation(cython.floating[:] x, cython.floating[:] y):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.utils.correlation", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 274 "ms_deisotope/_c/utils.pyx"
}

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_47__pyx_fuse_0correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_5utils_47__pyx_fuse_0correlation = 
#line 274 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_0correlation", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_47__pyx_fuse_0correlation, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_20correlation};

#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_47__pyx_fuse_0correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_x = { 0, 0, { 0 }, { 0 }, { 0 } }
#line 274 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_memviewslice __pyx_v_y = { 0, 0, { 0 }, { 0 }, { 0 } }
#line 274 "ms_deisotope/_c/utils.pyx"
;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0correlation (wrapper)", 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  {

#line 274 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,0};

#line 274 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 274 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 274 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 274 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 274 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 274 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 274 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 274 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 274 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 274 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 274 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else 
#line 274 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 274 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 274 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;

#line 274 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0correlation", 1, 2, 2, 1); 
#line 274 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        }

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 274 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0correlation") < 0)) __PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 274 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 274 "ms_deisotope/_c/utils.pyx"
    } else {

#line 274 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 274 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_v_x = __Pyx_PyObject_to_MemoryviewSlice_ds_float(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_x.memview)) __PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_v_y = __Pyx_PyObject_to_MemoryviewSlice_ds_float(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y.memview)) __PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0correlation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 274 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_0correlation", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 274 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_46__pyx_fuse_0correlation(__pyx_self, __pyx_v_x, __pyx_v_y);

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 274 "ms_deisotope/_c/utils.pyx"
}

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_46__pyx_fuse_0correlation(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_x, __Pyx_memviewslice __pyx_v_y) {

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_0correlation", 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 274 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_v_x.memview)) { __Pyx_RaiseUnboundLocalError("x"); __PYX_ERR(0, 274, __pyx_L1_error) }

#line 274 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_v_y.memview)) { __Pyx_RaiseUnboundLocalError("y"); __PYX_ERR(0, 274, __pyx_L1_error) }

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils_correlation(__pyx_v_x, __pyx_v_y, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_0correlation", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_v_x, 1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_v_y, 1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 274 "ms_deisotope/_c/utils.pyx"
}

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_49__pyx_fuse_1correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_21correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 274 "ms_deisotope/_c/utils.pyx"
static double __pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils_correlation(__Pyx_memviewslice __pyx_v_x, __Pyx_memviewslice __pyx_v_y, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 274 "ms_deisotope/_c/utils.pyx"
;
  size_t __pyx_v_n
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_xsum
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_ysum
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_xmean
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_ymean
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_cov
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_varx
#line 274 "ms_deisotope/_c/utils.pyx"
;
  double __pyx_v_vary
#line 274 "ms_deisotope/_c/utils.pyx"
;

#line 274 "ms_deisotope/_c/utils.pyx"
  double __pyx_r;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 274 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_1;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_2;

#line 274 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_3;

#line 274 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_4;

#line 274 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_5;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_t_6;

#line 274 "ms_deisotope/_c/utils.pyx"
  size_t __pyx_t_7;

#line 274 "ms_deisotope/_c/utils.pyx"
  double __pyx_t_8;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1correlation", 0);

  /* "ms_deisotope/_c/utils.pyx":279
 *         double xsum, ysum, xmean, ymean
 *         double cov, varx, vary
 *     n = len(x)             # <<<<<<<<<<<<<<
 *     if n == 0:
 *         return NAN
 */

#line 279 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_MemoryView_Len(__pyx_v_x); 

#line 279 "ms_deisotope/_c/utils.pyx"
  __pyx_v_n = __pyx_t_1;

  /* "ms_deisotope/_c/utils.pyx":280
 *         double cov, varx, vary
 *     n = len(x)
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return NAN
 *     xsum = 0.
 */

#line 280 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = ((__pyx_v_n == 0) != 0);

#line 280 "ms_deisotope/_c/utils.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/utils.pyx":281
 *     n = len(x)
 *     if n == 0:
 *         return NAN             # <<<<<<<<<<<<<<
 *     xsum = 0.
 *     ysum = 0.
 */

#line 281 "ms_deisotope/_c/utils.pyx"
    __pyx_r = NPY_NAN;

#line 281 "ms_deisotope/_c/utils.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/utils.pyx":280
 *         double cov, varx, vary
 *     n = len(x)
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return NAN
 *     xsum = 0.
 */

#line 280 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":282
 *     if n == 0:
 *         return NAN
 *     xsum = 0.             # <<<<<<<<<<<<<<
 *     ysum = 0.
 *     for i in range(n):
 */

#line 282 "ms_deisotope/_c/utils.pyx"
  __pyx_v_xsum = 0.;

  /* "ms_deisotope/_c/utils.pyx":283
 *         return NAN
 *     xsum = 0.
 *     ysum = 0.             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         xsum += x[i]
 */

#line 283 "ms_deisotope/_c/utils.pyx"
  __pyx_v_ysum = 0.;

  /* "ms_deisotope/_c/utils.pyx":284
 *     xsum = 0.
 *     ysum = 0.
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         xsum += x[i]
 *         ysum += y[i]
 */

#line 284 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 284 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_t_1;

#line 284 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 284 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/utils.pyx":285
 *     ysum = 0.
 *     for i in range(n):
 *         xsum += x[i]             # <<<<<<<<<<<<<<
 *         ysum += y[i]
 *     xmean = xsum / n
 */

#line 285 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = __pyx_v_i;

#line 285 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 285 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_5 >= (size_t)__pyx_v_x.shape[0])) __pyx_t_6 = 0;

#line 285 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 285 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 285 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 285, __pyx_L1_error)

#line 285 "ms_deisotope/_c/utils.pyx"
    }

#line 285 "ms_deisotope/_c/utils.pyx"
    __pyx_v_xsum = (__pyx_v_xsum + (*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_5 * __pyx_v_x.strides[0]) ))));

    /* "ms_deisotope/_c/utils.pyx":286
 *     for i in range(n):
 *         xsum += x[i]
 *         ysum += y[i]             # <<<<<<<<<<<<<<
 *     xmean = xsum / n
 *     ymean = ysum / n
 */

#line 286 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = __pyx_v_i;

#line 286 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 286 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_5 >= (size_t)__pyx_v_y.shape[0])) __pyx_t_6 = 0;

#line 286 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 286 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 286 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 286, __pyx_L1_error)

#line 286 "ms_deisotope/_c/utils.pyx"
    }

#line 286 "ms_deisotope/_c/utils.pyx"
    __pyx_v_ysum = (__pyx_v_ysum + (*((double *) ( /* dim=0 */ (__pyx_v_y.data + __pyx_t_5 * __pyx_v_y.strides[0]) ))));

#line 286 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":287
 *         xsum += x[i]
 *         ysum += y[i]
 *     xmean = xsum / n             # <<<<<<<<<<<<<<
 *     ymean = ysum / n
 *     cov = 0.
 */

#line 287 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_n == 0)) {

#line 287 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 287 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 287, __pyx_L1_error)

#line 287 "ms_deisotope/_c/utils.pyx"
  }

#line 287 "ms_deisotope/_c/utils.pyx"
  __pyx_v_xmean = (__pyx_v_xsum / __pyx_v_n);

  /* "ms_deisotope/_c/utils.pyx":288
 *         ysum += y[i]
 *     xmean = xsum / n
 *     ymean = ysum / n             # <<<<<<<<<<<<<<
 *     cov = 0.
 *     varx = 0.
 */

#line 288 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_v_n == 0)) {

#line 288 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 288 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 288, __pyx_L1_error)

#line 288 "ms_deisotope/_c/utils.pyx"
  }

#line 288 "ms_deisotope/_c/utils.pyx"
  __pyx_v_ymean = (__pyx_v_ysum / __pyx_v_n);

  /* "ms_deisotope/_c/utils.pyx":289
 *     xmean = xsum / n
 *     ymean = ysum / n
 *     cov = 0.             # <<<<<<<<<<<<<<
 *     varx = 0.
 *     vary = 0.
 */

#line 289 "ms_deisotope/_c/utils.pyx"
  __pyx_v_cov = 0.;

  /* "ms_deisotope/_c/utils.pyx":290
 *     ymean = ysum / n
 *     cov = 0.
 *     varx = 0.             # <<<<<<<<<<<<<<
 *     vary = 0.
 *     for i in range(n):
 */

#line 290 "ms_deisotope/_c/utils.pyx"
  __pyx_v_varx = 0.;

  /* "ms_deisotope/_c/utils.pyx":291
 *     cov = 0.
 *     varx = 0.
 *     vary = 0.             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         cov += (x[i] - xmean) * (y[i] - ymean)
 */

#line 291 "ms_deisotope/_c/utils.pyx"
  __pyx_v_vary = 0.;

  /* "ms_deisotope/_c/utils.pyx":292
 *     varx = 0.
 *     vary = 0.
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         cov += (x[i] - xmean) * (y[i] - ymean)
 *         varx += (x[i] - xmean) ** 2
 */

#line 292 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 292 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_t_1;

#line 292 "ms_deisotope/_c/utils.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 292 "ms_deisotope/_c/utils.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/utils.pyx":293
 *     vary = 0.
 *     for i in range(n):
 *         cov += (x[i] - xmean) * (y[i] - ymean)             # <<<<<<<<<<<<<<
 *         varx += (x[i] - xmean) ** 2
 *         vary += (y[i] - ymean) ** 2
 */

#line 293 "ms_deisotope/_c/utils.pyx"
    __pyx_t_5 = __pyx_v_i;

#line 293 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 293 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_5 >= (size_t)__pyx_v_x.shape[0])) __pyx_t_6 = 0;

#line 293 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 293 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 293 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 293, __pyx_L1_error)

#line 293 "ms_deisotope/_c/utils.pyx"
    }

#line 293 "ms_deisotope/_c/utils.pyx"
    __pyx_t_7 = __pyx_v_i;

#line 293 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 293 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 >= (size_t)__pyx_v_y.shape[0])) __pyx_t_6 = 0;

#line 293 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 293 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 293 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 293, __pyx_L1_error)

#line 293 "ms_deisotope/_c/utils.pyx"
    }

#line 293 "ms_deisotope/_c/utils.pyx"
    __pyx_v_cov = (__pyx_v_cov + (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_5 * __pyx_v_x.strides[0]) ))) - __pyx_v_xmean) * ((*((double *) ( /* dim=0 */ (__pyx_v_y.data + __pyx_t_7 * __pyx_v_y.strides[0]) ))) - __pyx_v_ymean)));

    /* "ms_deisotope/_c/utils.pyx":294
 *     for i in range(n):
 *         cov += (x[i] - xmean) * (y[i] - ymean)
 *         varx += (x[i] - xmean) ** 2             # <<<<<<<<<<<<<<
 *         vary += (y[i] - ymean) ** 2
 *     return cov / (sqrt(varx) * sqrt(vary))
 */

#line 294 "ms_deisotope/_c/utils.pyx"
    __pyx_t_7 = __pyx_v_i;

#line 294 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 294 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 >= (size_t)__pyx_v_x.shape[0])) __pyx_t_6 = 0;

#line 294 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 294 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 294 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/utils.pyx"
    }

#line 294 "ms_deisotope/_c/utils.pyx"
    __pyx_v_varx = (__pyx_v_varx + pow(((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_7 * __pyx_v_x.strides[0]) ))) - __pyx_v_xmean), 2.0));

    /* "ms_deisotope/_c/utils.pyx":295
 *         cov += (x[i] - xmean) * (y[i] - ymean)
 *         varx += (x[i] - xmean) ** 2
 *         vary += (y[i] - ymean) ** 2             # <<<<<<<<<<<<<<
 *     return cov / (sqrt(varx) * sqrt(vary))
 */

#line 295 "ms_deisotope/_c/utils.pyx"
    __pyx_t_7 = __pyx_v_i;

#line 295 "ms_deisotope/_c/utils.pyx"
    __pyx_t_6 = -1;

#line 295 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_7 >= (size_t)__pyx_v_y.shape[0])) __pyx_t_6 = 0;

#line 295 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_t_6 != -1)) {

#line 295 "ms_deisotope/_c/utils.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_6);

#line 295 "ms_deisotope/_c/utils.pyx"
      __PYX_ERR(0, 295, __pyx_L1_error)

#line 295 "ms_deisotope/_c/utils.pyx"
    }

#line 295 "ms_deisotope/_c/utils.pyx"
    __pyx_v_vary = (__pyx_v_vary + pow(((*((double *) ( /* dim=0 */ (__pyx_v_y.data + __pyx_t_7 * __pyx_v_y.strides[0]) ))) - __pyx_v_ymean), 2.0));

#line 295 "ms_deisotope/_c/utils.pyx"
  }

  /* "ms_deisotope/_c/utils.pyx":296
 *         varx += (x[i] - xmean) ** 2
 *         vary += (y[i] - ymean) ** 2
 *     return cov / (sqrt(varx) * sqrt(vary))             # <<<<<<<<<<<<<<
 */

#line 296 "ms_deisotope/_c/utils.pyx"
  __pyx_t_8 = (sqrt(__pyx_v_varx) * sqrt(__pyx_v_vary));

#line 296 "ms_deisotope/_c/utils.pyx"
  if (unlikely(__pyx_t_8 == 0)) {

#line 296 "ms_deisotope/_c/utils.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 296 "ms_deisotope/_c/utils.pyx"
    __PYX_ERR(0, 296, __pyx_L1_error)

#line 296 "ms_deisotope/_c/utils.pyx"
  }

#line 296 "ms_deisotope/_c/utils.pyx"
  __pyx_r = (__pyx_v_cov / __pyx_t_8);

#line 296 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/utils.pyx":274
 * 
 * 
 * cpdef double correlation(cython.floating[:] x, cython.floating[:] y):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.utils.correlation", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_r = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 274 "ms_deisotope/_c/utils.pyx"
}

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
/* Python wrapper */

#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_49__pyx_fuse_1correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_5utils_49__pyx_fuse_1correlation = 
#line 274 "ms_deisotope/_c/utils.pyx"
{"__pyx_fuse_1correlation", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_49__pyx_fuse_1correlation, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_20correlation};

#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_5utils_49__pyx_fuse_1correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_x = { 0, 0, { 0 }, { 0 }, { 0 } }
#line 274 "ms_deisotope/_c/utils.pyx"
;
  __Pyx_memviewslice __pyx_v_y = { 0, 0, { 0 }, { 0 }, { 0 } }
#line 274 "ms_deisotope/_c/utils.pyx"
;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1correlation (wrapper)", 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  {

#line 274 "ms_deisotope/_c/utils.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,0};

#line 274 "ms_deisotope/_c/utils.pyx"
    PyObject* values[2] = {0,0};

#line 274 "ms_deisotope/_c/utils.pyx"
    if (unlikely(__pyx_kwds)) {

#line 274 "ms_deisotope/_c/utils.pyx"
      Py_ssize_t kw_args;

#line 274 "ms_deisotope/_c/utils.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 274 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {
        case  2: 
#line 274 "ms_deisotope/_c/utils.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 274 "ms_deisotope/_c/utils.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 274 "ms_deisotope/_c/utils.pyx"
        case  0: break;
        default: 
#line 274 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 274 "ms_deisotope/_c/utils.pyx"
      switch (pos_args) {

#line 274 "ms_deisotope/_c/utils.pyx"
        case  0:

#line 274 "ms_deisotope/_c/utils.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else 
#line 274 "ms_deisotope/_c/utils.pyx"
goto __pyx_L5_argtuple_error;

#line 274 "ms_deisotope/_c/utils.pyx"
        CYTHON_FALLTHROUGH;

#line 274 "ms_deisotope/_c/utils.pyx"
        case  1:

#line 274 "ms_deisotope/_c/utils.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;

#line 274 "ms_deisotope/_c/utils.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1correlation", 1, 2, 2, 1); 
#line 274 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
        }

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
      if (unlikely(kw_args > 0)) {

#line 274 "ms_deisotope/_c/utils.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1correlation") < 0)) __PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
      }

#line 274 "ms_deisotope/_c/utils.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 274 "ms_deisotope/_c/utils.pyx"
      goto __pyx_L5_argtuple_error;

#line 274 "ms_deisotope/_c/utils.pyx"
    } else {

#line 274 "ms_deisotope/_c/utils.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 274 "ms_deisotope/_c/utils.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 274 "ms_deisotope/_c/utils.pyx"
    }

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_v_x = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_x.memview)) __PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
    __pyx_v_y = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y.memview)) __PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  }

#line 274 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1correlation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 274 "ms_deisotope/_c/utils.pyx"
__PYX_ERR(0, 274, __pyx_L3_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L3_error:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_1correlation", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/utils.pyx"
  return NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 274 "ms_deisotope/_c/utils.pyx"
__pyx_pf_12ms_deisotope_2_c_5utils_48__pyx_fuse_1correlation(__pyx_self, __pyx_v_x, __pyx_v_y);

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 274 "ms_deisotope/_c/utils.pyx"
}

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_5utils_48__pyx_fuse_1correlation(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_x, __Pyx_memviewslice __pyx_v_y) {

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_r = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyDeclarations

#line 274 "ms_deisotope/_c/utils.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_lineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  const char *__pyx_filename = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  int __pyx_clineno = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannySetupContext("__pyx_fuse_1correlation", 0);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 274 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_v_x.memview)) { __Pyx_RaiseUnboundLocalError("x"); __PYX_ERR(0, 274, __pyx_L1_error) }

#line 274 "ms_deisotope/_c/utils.pyx"
  if (unlikely(!__pyx_v_y.memview)) { __Pyx_RaiseUnboundLocalError("y"); __PYX_ERR(0, 274, __pyx_L1_error) }

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils_correlation(__pyx_v_x, __pyx_v_y, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_r = __pyx_t_1;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  goto __pyx_L0;

#line 274 "ms_deisotope/_c/utils.pyx"


#line 274 "ms_deisotope/_c/utils.pyx"
  /* function exit code */

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L1_error:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.utils.__pyx_fuse_1correlation", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_r = NULL;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_L0:;

#line 274 "ms_deisotope/_c/utils.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_v_x, 1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __PYX_XDEC_MEMVIEW(&__pyx_v_y, 1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/utils.pyx"
  return __pyx_r;

#line 274 "ms_deisotope/_c/utils.pyx"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 735, __pyx_L1_error)

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 738, __pyx_L1_error)

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 741, __pyx_L1_error)

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 744, __pyx_L1_error)

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 747, __pyx_L1_error)

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /*else*/ {

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(__pyx_empty_tuple);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = __pyx_empty_tuple;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":929
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */

#line 929 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_INCREF(__pyx_v_base);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":930
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */

#line 930 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base
#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":933
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */

#line 933 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":935
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":936
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_base));

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyObject *)__pyx_v_base);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":942
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */

#line 942 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 942, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":943
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 943, __pyx_L5_except_error)

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(1, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":948
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 948 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 948, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":949
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 949, __pyx_L5_except_error)

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(1, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":954
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 954 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 954, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":955
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 955, __pyx_L5_except_error)

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":956
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(1, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":978
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":993
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_datetime __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1003
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_timedelta __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1010
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1017
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
/* Python wrapper */

#line 123 "View.MemoryView"
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 123 "View.MemoryView"
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0
#line 123 "View.MemoryView"
;
  Py_ssize_t __pyx_v_itemsize
#line 123 "View.MemoryView"
;
  PyObject *__pyx_v_format = 0
#line 123 "View.MemoryView"
;
  PyObject *__pyx_v_mode = 0
#line 123 "View.MemoryView"
;
  int __pyx_v_allocate_buffer
#line 123 "View.MemoryView"
;

#line 123 "View.MemoryView"
  int __pyx_lineno = 0;

#line 123 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 123 "View.MemoryView"
  int __pyx_clineno = 0;

#line 123 "View.MemoryView"
  int __pyx_r;

#line 123 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 123 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);

#line 123 "View.MemoryView"
  {

#line 123 "View.MemoryView"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};

#line 123 "View.MemoryView"
    PyObject* values[5] = {0,0,0,0,0};

#line 123 "View.MemoryView"
    values[3] = ((PyObject *)__pyx_n_s_c);

#line 123 "View.MemoryView"
    if (unlikely(__pyx_kwds)) {

#line 123 "View.MemoryView"
      Py_ssize_t kw_args;

#line 123 "View.MemoryView"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 123 "View.MemoryView"
      switch (pos_args) {
        case  5: 
#line 123 "View.MemoryView"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 123 "View.MemoryView"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 123 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 123 "View.MemoryView"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 123 "View.MemoryView"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  0: break;
        default: 
#line 123 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
      kw_args = PyDict_Size(__pyx_kwds);

#line 123 "View.MemoryView"
      switch (pos_args) {

#line 123 "View.MemoryView"
        case  0:

#line 123 "View.MemoryView"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else 
#line 123 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  1:

#line 123 "View.MemoryView"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;

#line 123 "View.MemoryView"
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); 
#line 123 "View.MemoryView"
__PYX_ERR(2, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  2:

#line 123 "View.MemoryView"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;

#line 123 "View.MemoryView"
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); 
#line 123 "View.MemoryView"
__PYX_ERR(2, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  3:

#line 123 "View.MemoryView"
        if (kw_args > 0) {

#line 123 "View.MemoryView"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);

#line 123 "View.MemoryView"
          if (value) { values[3] = value; kw_args--; }

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  4:

#line 123 "View.MemoryView"
        if (kw_args > 0) {

#line 123 "View.MemoryView"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);

#line 123 "View.MemoryView"
          if (value) { values[4] = value; kw_args--; }

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
      if (unlikely(kw_args > 0)) {

#line 123 "View.MemoryView"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
    } else {

#line 123 "View.MemoryView"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: 
#line 123 "View.MemoryView"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 123 "View.MemoryView"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 123 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 123 "View.MemoryView"
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 123 "View.MemoryView"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 123 "View.MemoryView"
        break;
        default: 
#line 123 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
    }

#line 123 "View.MemoryView"
    __pyx_v_shape = ((PyObject*)values[0]);

#line 123 "View.MemoryView"
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
    __pyx_v_format = values[2];

#line 123 "View.MemoryView"
    __pyx_v_mode = values[3];

#line 123 "View.MemoryView"
    if (values[4]) {

#line 123 "View.MemoryView"
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 124, __pyx_L3_error)

#line 123 "View.MemoryView"
    } else {

      /* "View.MemoryView":124
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */

#line 124 "View.MemoryView"
      __pyx_v_allocate_buffer = ((int)1);

#line 124 "View.MemoryView"
    }

#line 124 "View.MemoryView"
  }

#line 124 "View.MemoryView"
  goto __pyx_L4_argument_unpacking_done;

#line 124 "View.MemoryView"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); 
#line 124 "View.MemoryView"
__PYX_ERR(2, 123, __pyx_L3_error)

#line 124 "View.MemoryView"
  __pyx_L3_error:;

#line 124 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 124 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 124 "View.MemoryView"
  return -1;

#line 124 "View.MemoryView"
  __pyx_L4_argument_unpacking_done:;

#line 124 "View.MemoryView"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(2, 123, __pyx_L1_error)

#line 124 "View.MemoryView"
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {

#line 124 "View.MemoryView"
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(2, 123, __pyx_L1_error)

#line 124 "View.MemoryView"
  }
  __pyx_r = 
#line 124 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
  /* function exit code */

#line 123 "View.MemoryView"
  goto __pyx_L0;

#line 123 "View.MemoryView"
  __pyx_L1_error:;

#line 123 "View.MemoryView"
  __pyx_r = -1;

#line 123 "View.MemoryView"
  __pyx_L0:;

#line 123 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 123 "View.MemoryView"
  return __pyx_r;

#line 123 "View.MemoryView"
}

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx
#line 123 "View.MemoryView"
;
  Py_ssize_t __pyx_v_i
#line 123 "View.MemoryView"
;
  Py_ssize_t __pyx_v_dim
#line 123 "View.MemoryView"
;
  PyObject **__pyx_v_p
#line 123 "View.MemoryView"
;
  char __pyx_v_order
#line 123 "View.MemoryView"
;

#line 123 "View.MemoryView"
  int __pyx_r;

#line 123 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 123 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 123 "View.MemoryView"
  int __pyx_t_2;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 123 "View.MemoryView"
  int __pyx_t_4;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 123 "View.MemoryView"
  char *__pyx_t_7;

#line 123 "View.MemoryView"
  int __pyx_t_8;

#line 123 "View.MemoryView"
  Py_ssize_t __pyx_t_9;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_10 = NULL;

#line 123 "View.MemoryView"
  Py_ssize_t __pyx_t_11;

#line 123 "View.MemoryView"
  int __pyx_lineno = 0;

#line 123 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 123 "View.MemoryView"
  int __pyx_clineno = 0;

#line 123 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__", 0);

#line 123 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":130
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */

#line 130 "View.MemoryView"
  if (unlikely(__pyx_v_shape == Py_None)) {

#line 130 "View.MemoryView"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 130 "View.MemoryView"
    __PYX_ERR(2, 130, __pyx_L1_error)

#line 130 "View.MemoryView"
  }

#line 130 "View.MemoryView"
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(2, 130, __pyx_L1_error)

#line 130 "View.MemoryView"
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":131
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */

#line 131 "View.MemoryView"
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":133
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */

#line 133 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);

#line 133 "View.MemoryView"
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":134
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */

#line 134 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 134, __pyx_L1_error)

#line 134 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 134 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 134 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 134 "View.MemoryView"
    __PYX_ERR(2, 134, __pyx_L1_error)

    /* "View.MemoryView":133
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */

#line 133 "View.MemoryView"
  }

  /* "View.MemoryView":136
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */

#line 136 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);

#line 136 "View.MemoryView"
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":137
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */

#line 137 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 137, __pyx_L1_error)

#line 137 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 137 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 137 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 137 "View.MemoryView"
    __PYX_ERR(2, 137, __pyx_L1_error)

    /* "View.MemoryView":136
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */

#line 136 "View.MemoryView"
  }

  /* "View.MemoryView":139
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */

#line 139 "View.MemoryView"
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 

#line 139 "View.MemoryView"
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);

#line 139 "View.MemoryView"
  if (__pyx_t_4) {

    /* "View.MemoryView":140
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */

#line 140 "View.MemoryView"
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 140, __pyx_L1_error)

#line 140 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 140 "View.MemoryView"
    __pyx_t_6 = NULL;

#line 140 "View.MemoryView"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {

#line 140 "View.MemoryView"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 140 "View.MemoryView"
      if (likely(__pyx_t_6)) {

#line 140 "View.MemoryView"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 140 "View.MemoryView"
        __Pyx_INCREF(__pyx_t_6);

#line 140 "View.MemoryView"
        __Pyx_INCREF(function);

#line 140 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_t_5, function);

#line 140 "View.MemoryView"
      }

#line 140 "View.MemoryView"
    }

#line 140 "View.MemoryView"
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);

#line 140 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 140 "View.MemoryView"
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 140, __pyx_L1_error)

#line 140 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 140 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 140 "View.MemoryView"
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);

#line 140 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":139
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */

#line 139 "View.MemoryView"
  }

  /* "View.MemoryView":141
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */

#line 141 "View.MemoryView"
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(2, 141, __pyx_L1_error)

#line 141 "View.MemoryView"
  __pyx_t_3 = __pyx_v_format;

#line 141 "View.MemoryView"
  __Pyx_INCREF(__pyx_t_3);

#line 141 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 141 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_self->_format);

#line 141 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_self->_format);

#line 141 "View.MemoryView"
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);

#line 141 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":142
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 142 "View.MemoryView"
  if (unlikely(__pyx_v_self->_format == Py_None)) {

#line 142 "View.MemoryView"
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");

#line 142 "View.MemoryView"
    __PYX_ERR(2, 142, __pyx_L1_error)

#line 142 "View.MemoryView"
  }

#line 142 "View.MemoryView"
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(2, 142, __pyx_L1_error)

#line 142 "View.MemoryView"
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":145
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */

#line 145 "View.MemoryView"
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":146
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */

#line 146 "View.MemoryView"
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":148
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */

#line 148 "View.MemoryView"
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);

#line 148 "View.MemoryView"
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":149
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 149 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 149, __pyx_L1_error)

#line 149 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 149 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 149 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 149 "View.MemoryView"
    __PYX_ERR(2, 149, __pyx_L1_error)

    /* "View.MemoryView":148
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */

#line 148 "View.MemoryView"
  }

  /* "View.MemoryView":152
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */

#line 152 "View.MemoryView"
  __pyx_t_8 = 0;

#line 152 "View.MemoryView"
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;

#line 152 "View.MemoryView"
  for (;;) {

#line 152 "View.MemoryView"
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;

#line 152 "View.MemoryView"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 152 "View.MemoryView"
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(2, 152, __pyx_L1_error)

#line 152 "View.MemoryView"
    #else

#line 152 "View.MemoryView"
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 152, __pyx_L1_error)

#line 152 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 152 "View.MemoryView"
    #endif

#line 152 "View.MemoryView"
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 152, __pyx_L1_error)

#line 152 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 152 "View.MemoryView"
    __pyx_v_dim = __pyx_t_9;

#line 152 "View.MemoryView"
    __pyx_v_idx = __pyx_t_8;

#line 152 "View.MemoryView"
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":153
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */

#line 153 "View.MemoryView"
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);

#line 153 "View.MemoryView"
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":154
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */

#line 154 "View.MemoryView"
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 154 "View.MemoryView"
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 154 "View.MemoryView"
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_10);

#line 154 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_t_5);

#line 154 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);

#line 154 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_t_6);

#line 154 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);

#line 154 "View.MemoryView"
      __pyx_t_5 = 0;

#line 154 "View.MemoryView"
      __pyx_t_6 = 0;

#line 154 "View.MemoryView"
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 154 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 154 "View.MemoryView"
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_10);

#line 154 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 154 "View.MemoryView"
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);

#line 154 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 154 "View.MemoryView"
      __PYX_ERR(2, 154, __pyx_L1_error)

      /* "View.MemoryView":153
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */

#line 153 "View.MemoryView"
    }

    /* "View.MemoryView":155
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */

#line 155 "View.MemoryView"
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":152
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */

#line 152 "View.MemoryView"
  }

#line 152 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":158
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */

#line 158 "View.MemoryView"
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(2, 158, __pyx_L1_error)

#line 158 "View.MemoryView"
  if (__pyx_t_4) {

    /* "View.MemoryView":159
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */

#line 159 "View.MemoryView"
    __pyx_v_order = 'F';

    /* "View.MemoryView":160
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */

#line 160 "View.MemoryView"
    __Pyx_INCREF(__pyx_n_u_fortran);

#line 160 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_n_u_fortran);

#line 160 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_self->mode);

#line 160 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_self->mode);

#line 160 "View.MemoryView"
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":158
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */

#line 158 "View.MemoryView"
    goto __pyx_L10;

#line 158 "View.MemoryView"
  }

  /* "View.MemoryView":161
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */

#line 161 "View.MemoryView"
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(2, 161, __pyx_L1_error)

#line 161 "View.MemoryView"
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":162
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */

#line 162 "View.MemoryView"
    __pyx_v_order = 'C';

    /* "View.MemoryView":163
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */

#line 163 "View.MemoryView"
    __Pyx_INCREF(__pyx_n_u_c);

#line 163 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_n_u_c);

#line 163 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_self->mode);

#line 163 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_self->mode);

#line 163 "View.MemoryView"
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":161
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */

#line 161 "View.MemoryView"
    goto __pyx_L10;

#line 161 "View.MemoryView"
  }

  /* "View.MemoryView":165
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */

#line 165 "View.MemoryView"
  /*else*/ {

#line 165 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 165, __pyx_L1_error)

#line 165 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 165 "View.MemoryView"
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 165, __pyx_L1_error)

#line 165 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_10);

#line 165 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 165 "View.MemoryView"
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);

#line 165 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 165 "View.MemoryView"
    __PYX_ERR(2, 165, __pyx_L1_error)

#line 165 "View.MemoryView"
  }

#line 165 "View.MemoryView"
  __pyx_L10:;

  /* "View.MemoryView":167
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */

#line 167 "View.MemoryView"
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":170
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */

#line 170 "View.MemoryView"
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":171
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */

#line 171 "View.MemoryView"
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 171, __pyx_L1_error)

#line 171 "View.MemoryView"
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 171, __pyx_L1_error)

#line 171 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 171 "View.MemoryView"
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":172
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 172 "View.MemoryView"
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);

#line 172 "View.MemoryView"
  if (__pyx_t_4) {

    /* "View.MemoryView":175
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */

#line 175 "View.MemoryView"
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":176
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */

#line 176 "View.MemoryView"
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);

#line 176 "View.MemoryView"
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":177
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */

#line 177 "View.MemoryView"
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 177, __pyx_L1_error)

#line 177 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_10);

#line 177 "View.MemoryView"
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);

#line 177 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 177 "View.MemoryView"
      __PYX_ERR(2, 177, __pyx_L1_error)

      /* "View.MemoryView":176
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */

#line 176 "View.MemoryView"
    }

    /* "View.MemoryView":179
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */

#line 179 "View.MemoryView"
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);

#line 179 "View.MemoryView"
    if (__pyx_t_4) {

      /* "View.MemoryView":180
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */

#line 180 "View.MemoryView"
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":181
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */

#line 181 "View.MemoryView"
      if (unlikely(__pyx_v_itemsize == 0)) {

#line 181 "View.MemoryView"
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");

#line 181 "View.MemoryView"
        __PYX_ERR(2, 181, __pyx_L1_error)

#line 181 "View.MemoryView"
      }

#line 181 "View.MemoryView"
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {

#line 181 "View.MemoryView"
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");

#line 181 "View.MemoryView"
        __PYX_ERR(2, 181, __pyx_L1_error)

#line 181 "View.MemoryView"
      }

#line 181 "View.MemoryView"
      __pyx_t_1 = __Pyx_div_Py_ssize_t(__pyx_v_self->len, __pyx_v_itemsize);

#line 181 "View.MemoryView"
      __pyx_t_9 = __pyx_t_1;

#line 181 "View.MemoryView"
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {

#line 181 "View.MemoryView"
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":182
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */

#line 182 "View.MemoryView"
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":183
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */

#line 183 "View.MemoryView"
        Py_INCREF(Py_None);

#line 183 "View.MemoryView"
      }

      /* "View.MemoryView":179
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */

#line 179 "View.MemoryView"
    }

    /* "View.MemoryView":172
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 172 "View.MemoryView"
  }

  /* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
  /* function exit code */

#line 123 "View.MemoryView"
  __pyx_r = 0;

#line 123 "View.MemoryView"
  goto __pyx_L0;

#line 123 "View.MemoryView"
  __pyx_L1_error:;

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_10);

#line 123 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 123 "View.MemoryView"
  __pyx_r = -1;

#line 123 "View.MemoryView"
  __pyx_L0:;

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_format);

#line 123 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 123 "View.MemoryView"
  return __pyx_r;

#line 123 "View.MemoryView"
}

/* "View.MemoryView":186
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
/* Python wrapper */

#line 186 "View.MemoryView"
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/

#line 186 "View.MemoryView"
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {

#line 186 "View.MemoryView"
  int __pyx_r;

#line 186 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 186 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = 
#line 186 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
  /* function exit code */

#line 186 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 186 "View.MemoryView"
  return __pyx_r;

#line 186 "View.MemoryView"
}

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode
#line 186 "View.MemoryView"
;

#line 186 "View.MemoryView"
  int __pyx_r;

#line 186 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 186 "View.MemoryView"
  int __pyx_t_1;

#line 186 "View.MemoryView"
  int __pyx_t_2;

#line 186 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 186 "View.MemoryView"
  char *__pyx_t_4;

#line 186 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 186 "View.MemoryView"
  int __pyx_t_6;

#line 186 "View.MemoryView"
  Py_ssize_t *__pyx_t_7;

#line 186 "View.MemoryView"
  int __pyx_lineno = 0;

#line 186 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 186 "View.MemoryView"
  int __pyx_clineno = 0;

#line 186 "View.MemoryView"
  if (__pyx_v_info == NULL) {

#line 186 "View.MemoryView"
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");

#line 186 "View.MemoryView"
    return -1;

#line 186 "View.MemoryView"
  }

#line 186 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__", 0);

#line 186 "View.MemoryView"
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);

#line 186 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":187
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */

#line 187 "View.MemoryView"
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":188
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */

#line 188 "View.MemoryView"
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 188, __pyx_L1_error)

#line 188 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 188 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":189
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */

#line 189 "View.MemoryView"
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":188
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */

#line 188 "View.MemoryView"
    goto __pyx_L3;

#line 188 "View.MemoryView"
  }

  /* "View.MemoryView":190
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */

#line 190 "View.MemoryView"
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 190, __pyx_L1_error)

#line 190 "View.MemoryView"
  __pyx_t_1 = (__pyx_t_2 != 0);

#line 190 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":191
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */

#line 191 "View.MemoryView"
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":190
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */

#line 190 "View.MemoryView"
  }

#line 190 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":192
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */

#line 192 "View.MemoryView"
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);

#line 192 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":193
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */

#line 193 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 193, __pyx_L1_error)

#line 193 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 193 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 193 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 193 "View.MemoryView"
    __PYX_ERR(2, 193, __pyx_L1_error)

    /* "View.MemoryView":192
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */

#line 192 "View.MemoryView"
  }

  /* "View.MemoryView":194
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */

#line 194 "View.MemoryView"
  __pyx_t_4 = __pyx_v_self->data;

#line 194 "View.MemoryView"
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":195
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */

#line 195 "View.MemoryView"
  __pyx_t_5 = __pyx_v_self->len;

#line 195 "View.MemoryView"
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":196
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */

#line 196 "View.MemoryView"
  __pyx_t_6 = __pyx_v_self->ndim;

#line 196 "View.MemoryView"
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":197
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */

#line 197 "View.MemoryView"
  __pyx_t_7 = __pyx_v_self->_shape;

#line 197 "View.MemoryView"
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */

#line 198 "View.MemoryView"
  __pyx_t_7 = __pyx_v_self->_strides;

#line 198 "View.MemoryView"
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":199
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */

#line 199 "View.MemoryView"
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":200
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */

#line 200 "View.MemoryView"
  __pyx_t_5 = __pyx_v_self->itemsize;

#line 200 "View.MemoryView"
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":201
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */

#line 201 "View.MemoryView"
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":203
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */

#line 203 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);

#line 203 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":204
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */

#line 204 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->format;

#line 204 "View.MemoryView"
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":203
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */

#line 203 "View.MemoryView"
    goto __pyx_L5;

#line 203 "View.MemoryView"
  }

  /* "View.MemoryView":206
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */

#line 206 "View.MemoryView"
  /*else*/ {

#line 206 "View.MemoryView"
    __pyx_v_info->format = NULL;

#line 206 "View.MemoryView"
  }

#line 206 "View.MemoryView"
  __pyx_L5:;

  /* "View.MemoryView":208
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */

#line 208 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 208 "View.MemoryView"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 208 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_info->obj);

#line 208 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_info->obj);

#line 208 "View.MemoryView"
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":186
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
  /* function exit code */

#line 186 "View.MemoryView"
  __pyx_r = 0;

#line 186 "View.MemoryView"
  goto __pyx_L0;

#line 186 "View.MemoryView"
  __pyx_L1_error:;

#line 186 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 186 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 186 "View.MemoryView"
  __pyx_r = -1;

#line 186 "View.MemoryView"
  if (__pyx_v_info->obj != NULL) {

#line 186 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 186 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 186 "View.MemoryView"
  }

#line 186 "View.MemoryView"
  goto __pyx_L2;

#line 186 "View.MemoryView"
  __pyx_L0:;

#line 186 "View.MemoryView"
  if (__pyx_v_info->obj == Py_None) {

#line 186 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 186 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 186 "View.MemoryView"
  }

#line 186 "View.MemoryView"
  __pyx_L2:;

#line 186 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 186 "View.MemoryView"
  return __pyx_r;

#line 186 "View.MemoryView"
}

/* "View.MemoryView":212
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
/* Python wrapper */

#line 212 "View.MemoryView"
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/

#line 212 "View.MemoryView"
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {

#line 212 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 212 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);

#line 212 "View.MemoryView"
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
  /* function exit code */

#line 212 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 212 "View.MemoryView"
}

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {

#line 212 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 212 "View.MemoryView"
  int __pyx_t_1;

#line 212 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":213
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */

#line 213 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);

#line 213 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":214
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */

#line 214 "View.MemoryView"
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":213
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */

#line 213 "View.MemoryView"
    goto __pyx_L3;

#line 213 "View.MemoryView"
  }

  /* "View.MemoryView":215
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */

#line 215 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->free_data != 0);

#line 215 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":216
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */

#line 216 "View.MemoryView"
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);

#line 216 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":217
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */

#line 217 "View.MemoryView"
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":216
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */

#line 216 "View.MemoryView"
    }

    /* "View.MemoryView":219
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */

#line 219 "View.MemoryView"
    free(__pyx_v_self->data);

    /* "View.MemoryView":215
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */

#line 215 "View.MemoryView"
  }

#line 215 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":220
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 220 "View.MemoryView"
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":212
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
  /* function exit code */

#line 212 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 212 "View.MemoryView"
}

/* "View.MemoryView":223
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
/* Python wrapper */

#line 223 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 223 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {

#line 223 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 223 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 223 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 223 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
  /* function exit code */

#line 223 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 223 "View.MemoryView"
  return __pyx_r;

#line 223 "View.MemoryView"
}

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {

#line 223 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 223 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 223 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 223 "View.MemoryView"
  int __pyx_lineno = 0;

#line 223 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 223 "View.MemoryView"
  int __pyx_clineno = 0;

#line 223 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":224
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */

#line 224 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 224 "View.MemoryView"
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 224, __pyx_L1_error)

#line 224 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 224 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 224 "View.MemoryView"
  __pyx_t_1 = 0;

#line 224 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":223
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
  /* function exit code */

#line 223 "View.MemoryView"
  __pyx_L1_error:;

#line 223 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 223 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 223 "View.MemoryView"
  __pyx_r = NULL;

#line 223 "View.MemoryView"
  __pyx_L0:;

#line 223 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 223 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 223 "View.MemoryView"
  return __pyx_r;

#line 223 "View.MemoryView"
}

/* "View.MemoryView":227
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

#line 227 "View.MemoryView"


#line 227 "View.MemoryView"
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags
#line 227 "View.MemoryView"
;

#line 227 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 227 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 227 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 227 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 227 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 227 "View.MemoryView"
  int __pyx_lineno = 0;

#line 227 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 227 "View.MemoryView"
  int __pyx_clineno = 0;

#line 227 "View.MemoryView"
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":228
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */

#line 228 "View.MemoryView"
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":229
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 229 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 229 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 229 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 229 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 229 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 229 "View.MemoryView"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 229 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));

#line 229 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 229 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);

#line 229 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 229 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 229 "View.MemoryView"
  __pyx_t_1 = 0;

#line 229 "View.MemoryView"
  __pyx_t_2 = 0;

#line 229 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 229 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 229 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 229 "View.MemoryView"
  __pyx_t_2 = 0;

#line 229 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":227
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

#line 227 "View.MemoryView"


#line 227 "View.MemoryView"
  /* function exit code */

#line 227 "View.MemoryView"
  __pyx_L1_error:;

#line 227 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 227 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 227 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 227 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 227 "View.MemoryView"
  __pyx_r = 0;

#line 227 "View.MemoryView"
  __pyx_L0:;

#line 227 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 227 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 227 "View.MemoryView"
  return __pyx_r;

#line 227 "View.MemoryView"
}

/* "View.MemoryView":231
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
/* Python wrapper */

#line 231 "View.MemoryView"
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/

#line 231 "View.MemoryView"
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {

#line 231 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 231 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 231 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 231 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
  /* function exit code */

#line 231 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 231 "View.MemoryView"
  return __pyx_r;

#line 231 "View.MemoryView"
}

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {

#line 231 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 231 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 231 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":232
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */

#line 232 "View.MemoryView"
  __pyx_r = (__pyx_v_self->_shape[0]);

#line 232 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":231
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
  /* function exit code */

#line 231 "View.MemoryView"
  __pyx_L0:;

#line 231 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 231 "View.MemoryView"
  return __pyx_r;

#line 231 "View.MemoryView"
}

/* "View.MemoryView":234
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
/* Python wrapper */

#line 234 "View.MemoryView"
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/

#line 234 "View.MemoryView"
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {

#line 234 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 234 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 234 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = 
#line 234 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
  /* function exit code */

#line 234 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 234 "View.MemoryView"
  return __pyx_r;

#line 234 "View.MemoryView"
}

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {

#line 234 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 234 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 234 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 234 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 234 "View.MemoryView"
  int __pyx_lineno = 0;

#line 234 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 234 "View.MemoryView"
  int __pyx_clineno = 0;

#line 234 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":235
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */

#line 235 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 235 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 235, __pyx_L1_error)

#line 235 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 235 "View.MemoryView"
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 235, __pyx_L1_error)

#line 235 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 235 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 235 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 235 "View.MemoryView"
  __pyx_t_2 = 0;

#line 235 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":234
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
  /* function exit code */

#line 234 "View.MemoryView"
  __pyx_L1_error:;

#line 234 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 234 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 234 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 234 "View.MemoryView"
  __pyx_r = NULL;

#line 234 "View.MemoryView"
  __pyx_L0:;

#line 234 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 234 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 234 "View.MemoryView"
  return __pyx_r;

#line 234 "View.MemoryView"
}

/* "View.MemoryView":237
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
/* Python wrapper */

#line 237 "View.MemoryView"
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/

#line 237 "View.MemoryView"
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {

#line 237 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 237 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 237 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 237 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
  /* function exit code */

#line 237 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 237 "View.MemoryView"
  return __pyx_r;

#line 237 "View.MemoryView"
}

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {

#line 237 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 237 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 237 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 237 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 237 "View.MemoryView"
  int __pyx_lineno = 0;

#line 237 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 237 "View.MemoryView"
  int __pyx_clineno = 0;

#line 237 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":238
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */

#line 238 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 238 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 238, __pyx_L1_error)

#line 238 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 238 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 238, __pyx_L1_error)

#line 238 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 238 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 238 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 238 "View.MemoryView"
  __pyx_t_2 = 0;

#line 238 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":237
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
  /* function exit code */

#line 237 "View.MemoryView"
  __pyx_L1_error:;

#line 237 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 237 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 237 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 237 "View.MemoryView"
  __pyx_r = NULL;

#line 237 "View.MemoryView"
  __pyx_L0:;

#line 237 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 237 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 237 "View.MemoryView"
  return __pyx_r;

#line 237 "View.MemoryView"
}

/* "View.MemoryView":240
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
/* Python wrapper */

#line 240 "View.MemoryView"
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/

#line 240 "View.MemoryView"
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {

#line 240 "View.MemoryView"
  int __pyx_r;

#line 240 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 240 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = 
#line 240 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
  /* function exit code */

#line 240 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 240 "View.MemoryView"
  return __pyx_r;

#line 240 "View.MemoryView"
}

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {

#line 240 "View.MemoryView"
  int __pyx_r;

#line 240 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 240 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 240 "View.MemoryView"
  int __pyx_lineno = 0;

#line 240 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 240 "View.MemoryView"
  int __pyx_clineno = 0;

#line 240 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":241
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 241 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 241, __pyx_L1_error)

#line 241 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 241 "View.MemoryView"
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(2, 241, __pyx_L1_error)

#line 241 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":240
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
  /* function exit code */

#line 240 "View.MemoryView"
  __pyx_r = 0;

#line 240 "View.MemoryView"
  goto __pyx_L0;

#line 240 "View.MemoryView"
  __pyx_L1_error:;

#line 240 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 240 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 240 "View.MemoryView"
  __pyx_r = -1;

#line 240 "View.MemoryView"
  __pyx_L0:;

#line 240 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 240 "View.MemoryView"
  return __pyx_r;

#line 240 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 2 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 2 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 2 "(tree fragment)"
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
/* Python wrapper */

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 3 "(tree fragment)"
__pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 3 "(tree fragment)"
  int __pyx_lineno = 0;

#line 3 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 3 "(tree fragment)"
  int __pyx_clineno = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __pyx_L1_error:;

#line 3 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "(tree fragment)"
  __pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

/* "View.MemoryView":245
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

#line 245 "View.MemoryView"


#line 245 "View.MemoryView"
static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0
#line 245 "View.MemoryView"
;

#line 245 "View.MemoryView"
  struct __pyx_array_obj *__pyx_r = NULL;

#line 245 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 245 "View.MemoryView"
  int __pyx_t_1;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 245 "View.MemoryView"
  int __pyx_lineno = 0;

#line 245 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 245 "View.MemoryView"
  int __pyx_clineno = 0;

#line 245 "View.MemoryView"
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":249
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */

#line 249 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);

#line 249 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":250
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */

#line 250 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 250 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 250 "View.MemoryView"
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 250 "View.MemoryView"
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 250 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_shape);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_shape);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_2);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_3);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_4);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);

#line 250 "View.MemoryView"
    __pyx_t_2 = 0;

#line 250 "View.MemoryView"
    __pyx_t_3 = 0;

#line 250 "View.MemoryView"
    __pyx_t_4 = 0;

#line 250 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 250 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 250 "View.MemoryView"
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);

#line 250 "View.MemoryView"
    __pyx_t_4 = 0;

    /* "View.MemoryView":249
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */

#line 249 "View.MemoryView"
    goto __pyx_L3;

#line 249 "View.MemoryView"
  }

  /* "View.MemoryView":252
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */

#line 252 "View.MemoryView"
  /*else*/ {

#line 252 "View.MemoryView"
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 252 "View.MemoryView"
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 252 "View.MemoryView"
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 252 "View.MemoryView"
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 252 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_shape);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_shape);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_4);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_5);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_3);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);

#line 252 "View.MemoryView"
    __pyx_t_4 = 0;

#line 252 "View.MemoryView"
    __pyx_t_5 = 0;

#line 252 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":253
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */

#line 253 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 253, __pyx_L1_error)

#line 253 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 253 "View.MemoryView"
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(2, 253, __pyx_L1_error)

    /* "View.MemoryView":252
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */

#line 252 "View.MemoryView"
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 252 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 252 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 252 "View.MemoryView"
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);

#line 252 "View.MemoryView"
    __pyx_t_5 = 0;

    /* "View.MemoryView":254
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */

#line 254 "View.MemoryView"
    __pyx_v_result->data = __pyx_v_buf;

#line 254 "View.MemoryView"
  }

#line 254 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":256
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 256 "View.MemoryView"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 256 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 256 "View.MemoryView"
  __pyx_r = __pyx_v_result;

#line 256 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":245
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

#line 245 "View.MemoryView"


#line 245 "View.MemoryView"
  /* function exit code */

#line 245 "View.MemoryView"
  __pyx_L1_error:;

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 245 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 245 "View.MemoryView"
  __pyx_r = 0;

#line 245 "View.MemoryView"
  __pyx_L0:;

#line 245 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 245 "View.MemoryView"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 245 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 245 "View.MemoryView"
  return __pyx_r;

#line 245 "View.MemoryView"
}

/* "View.MemoryView":282
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
/* Python wrapper */

#line 282 "View.MemoryView"
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 282 "View.MemoryView"
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0
#line 282 "View.MemoryView"
;

#line 282 "View.MemoryView"
  int __pyx_lineno = 0;

#line 282 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 282 "View.MemoryView"
  int __pyx_clineno = 0;

#line 282 "View.MemoryView"
  int __pyx_r;

#line 282 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 282 "View.MemoryView"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 282 "View.MemoryView"
  {

#line 282 "View.MemoryView"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};

#line 282 "View.MemoryView"
    PyObject* values[1] = {0};

#line 282 "View.MemoryView"
    if (unlikely(__pyx_kwds)) {

#line 282 "View.MemoryView"
      Py_ssize_t kw_args;

#line 282 "View.MemoryView"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 282 "View.MemoryView"
      switch (pos_args) {
        case  1: 
#line 282 "View.MemoryView"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 282 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 282 "View.MemoryView"
        case  0: break;
        default: 
#line 282 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 282 "View.MemoryView"
      }

#line 282 "View.MemoryView"
      kw_args = PyDict_Size(__pyx_kwds);

#line 282 "View.MemoryView"
      switch (pos_args) {

#line 282 "View.MemoryView"
        case  0:

#line 282 "View.MemoryView"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else 
#line 282 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 282 "View.MemoryView"
      }

#line 282 "View.MemoryView"
      if (unlikely(kw_args > 0)) {

#line 282 "View.MemoryView"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(2, 282, __pyx_L3_error)

#line 282 "View.MemoryView"
      }

#line 282 "View.MemoryView"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {

#line 282 "View.MemoryView"
      goto __pyx_L5_argtuple_error;

#line 282 "View.MemoryView"
    } else {

#line 282 "View.MemoryView"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 282 "View.MemoryView"
    }

#line 282 "View.MemoryView"
    __pyx_v_name = values[0];

#line 282 "View.MemoryView"
  }

#line 282 "View.MemoryView"
  goto __pyx_L4_argument_unpacking_done;

#line 282 "View.MemoryView"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 282 "View.MemoryView"
__PYX_ERR(2, 282, __pyx_L3_error)

#line 282 "View.MemoryView"
  __pyx_L3_error:;

#line 282 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 282 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 282 "View.MemoryView"
  return -1;

#line 282 "View.MemoryView"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 282 "View.MemoryView"
__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
  /* function exit code */

#line 282 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 282 "View.MemoryView"
  return __pyx_r;

#line 282 "View.MemoryView"
}

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {

#line 282 "View.MemoryView"
  int __pyx_r;

#line 282 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 282 "View.MemoryView"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":283
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */

#line 283 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_name);

#line 283 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_name);

#line 283 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_self->name);

#line 283 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_self->name);

#line 283 "View.MemoryView"
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":282
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
  /* function exit code */

#line 282 "View.MemoryView"
  __pyx_r = 0;

#line 282 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 282 "View.MemoryView"
  return __pyx_r;

#line 282 "View.MemoryView"
}

/* "View.MemoryView":284
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
/* Python wrapper */

#line 284 "View.MemoryView"
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/

#line 284 "View.MemoryView"
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {

#line 284 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 284 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 284 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 284 "View.MemoryView"
__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
  /* function exit code */

#line 284 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 284 "View.MemoryView"
  return __pyx_r;

#line 284 "View.MemoryView"
}

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {

#line 284 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 284 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 284 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":285
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */

#line 285 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 285 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->name);

#line 285 "View.MemoryView"
  __pyx_r = __pyx_v_self->name;

#line 285 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":284
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
  /* function exit code */

#line 284 "View.MemoryView"
  __pyx_L0:;

#line 284 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 284 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 284 "View.MemoryView"
  return __pyx_r;

#line 284 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->name);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->name);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_1;

#line 6 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_2 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));

#line 8 "(tree fragment)"
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_3 = (__pyx_v_self->name != Py_None);

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_3;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_3 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_184977713);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_184977713);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_5;

#line 13 "(tree fragment)"
    __pyx_t_5 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 15 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_184977713);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_184977713);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_5);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);

#line 15 "(tree fragment)"
    __pyx_t_5 = 0;

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_4;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "View.MemoryView":299
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

#line 299 "View.MemoryView"


#line 299 "View.MemoryView"
static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p
#line 299 "View.MemoryView"
;
  size_t __pyx_v_offset
#line 299 "View.MemoryView"
;

#line 299 "View.MemoryView"
  void *__pyx_r;

#line 299 "View.MemoryView"
  int __pyx_t_1;

  /* "View.MemoryView":301
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */

#line 301 "View.MemoryView"
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":305
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */

#line 305 "View.MemoryView"
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":307
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */

#line 307 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);

#line 307 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":308
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */

#line 308 "View.MemoryView"
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":307
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */

#line 307 "View.MemoryView"
  }

  /* "View.MemoryView":310
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 310 "View.MemoryView"
  __pyx_r = ((void *)__pyx_v_aligned_p);

#line 310 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":299
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

#line 299 "View.MemoryView"


#line 299 "View.MemoryView"
  /* function exit code */

#line 299 "View.MemoryView"
  __pyx_L0:;

#line 299 "View.MemoryView"
  return __pyx_r;

#line 299 "View.MemoryView"
}

/* "View.MemoryView":346
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
/* Python wrapper */

#line 346 "View.MemoryView"
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 346 "View.MemoryView"
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0
#line 346 "View.MemoryView"
;
  int __pyx_v_flags
#line 346 "View.MemoryView"
;
  int __pyx_v_dtype_is_object
#line 346 "View.MemoryView"
;

#line 346 "View.MemoryView"
  int __pyx_lineno = 0;

#line 346 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 346 "View.MemoryView"
  int __pyx_clineno = 0;

#line 346 "View.MemoryView"
  int __pyx_r;

#line 346 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 346 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);

#line 346 "View.MemoryView"
  {

#line 346 "View.MemoryView"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};

#line 346 "View.MemoryView"
    PyObject* values[3] = {0,0,0};

#line 346 "View.MemoryView"
    if (unlikely(__pyx_kwds)) {

#line 346 "View.MemoryView"
      Py_ssize_t kw_args;

#line 346 "View.MemoryView"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 346 "View.MemoryView"
      switch (pos_args) {
        case  3: 
#line 346 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 346 "View.MemoryView"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 346 "View.MemoryView"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 346 "View.MemoryView"
        case  0: break;
        default: 
#line 346 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
      kw_args = PyDict_Size(__pyx_kwds);

#line 346 "View.MemoryView"
      switch (pos_args) {

#line 346 "View.MemoryView"
        case  0:

#line 346 "View.MemoryView"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else 
#line 346 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 346 "View.MemoryView"
        case  1:

#line 346 "View.MemoryView"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;

#line 346 "View.MemoryView"
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); 
#line 346 "View.MemoryView"
__PYX_ERR(2, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
        }

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 346 "View.MemoryView"
        case  2:

#line 346 "View.MemoryView"
        if (kw_args > 0) {

#line 346 "View.MemoryView"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);

#line 346 "View.MemoryView"
          if (value) { values[2] = value; kw_args--; }

#line 346 "View.MemoryView"
        }

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
      if (unlikely(kw_args > 0)) {

#line 346 "View.MemoryView"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
    } else {

#line 346 "View.MemoryView"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 346 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 346 "View.MemoryView"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 346 "View.MemoryView"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 346 "View.MemoryView"
        break;
        default: 
#line 346 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
    }

#line 346 "View.MemoryView"
    __pyx_v_obj = values[0];

#line 346 "View.MemoryView"
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
    if (values[2]) {

#line 346 "View.MemoryView"
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
    } else {

#line 346 "View.MemoryView"
      __pyx_v_dtype_is_object = ((int)0);

#line 346 "View.MemoryView"
    }

#line 346 "View.MemoryView"
  }

#line 346 "View.MemoryView"
  goto __pyx_L4_argument_unpacking_done;

#line 346 "View.MemoryView"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 346 "View.MemoryView"
__PYX_ERR(2, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
  __pyx_L3_error:;

#line 346 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 346 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 346 "View.MemoryView"
  return -1;

#line 346 "View.MemoryView"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 346 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
  /* function exit code */

#line 346 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 346 "View.MemoryView"
  return __pyx_r;

#line 346 "View.MemoryView"
}

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {

#line 346 "View.MemoryView"
  int __pyx_r;

#line 346 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 346 "View.MemoryView"
  int __pyx_t_1;

#line 346 "View.MemoryView"
  int __pyx_t_2;

#line 346 "View.MemoryView"
  int __pyx_t_3;

#line 346 "View.MemoryView"
  int __pyx_t_4;

#line 346 "View.MemoryView"
  int __pyx_lineno = 0;

#line 346 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 346 "View.MemoryView"
  int __pyx_clineno = 0;

#line 346 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":347
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */

#line 347 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_obj);

#line 347 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_obj);

#line 347 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_self->obj);

#line 347 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_self->obj);

#line 347 "View.MemoryView"
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":348
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */

#line 348 "View.MemoryView"
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":349
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */

#line 349 "View.MemoryView"
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));

#line 349 "View.MemoryView"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 349 "View.MemoryView"
  if (!__pyx_t_3) {

#line 349 "View.MemoryView"
  } else {

#line 349 "View.MemoryView"
    __pyx_t_1 = __pyx_t_3;

#line 349 "View.MemoryView"
    goto __pyx_L4_bool_binop_done;

#line 349 "View.MemoryView"
  }

#line 349 "View.MemoryView"
  __pyx_t_3 = (__pyx_v_obj != Py_None);

#line 349 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 349 "View.MemoryView"
  __pyx_t_1 = __pyx_t_2;

#line 349 "View.MemoryView"
  __pyx_L4_bool_binop_done:;

#line 349 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":350
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */

#line 350 "View.MemoryView"
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 350, __pyx_L1_error)

    /* "View.MemoryView":351
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */

#line 351 "View.MemoryView"
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);

#line 351 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":352
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */

#line 352 "View.MemoryView"
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":353
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 */

#line 353 "View.MemoryView"
      Py_INCREF(Py_None);

      /* "View.MemoryView":351
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */

#line 351 "View.MemoryView"
    }

    /* "View.MemoryView":349
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */

#line 349 "View.MemoryView"
  }

  /* "View.MemoryView":355
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 */

#line 355 "View.MemoryView"
  __pyx_t_1 = ((!(__PYX_CYTHON_ATOMICS_ENABLED() != 0)) != 0);

#line 355 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":357
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */

#line 357 "View.MemoryView"
    __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);

#line 357 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":358
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 */

#line 358 "View.MemoryView"
      __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

      /* "View.MemoryView":359
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 */

#line 359 "View.MemoryView"
      __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

      /* "View.MemoryView":357
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */

#line 357 "View.MemoryView"
    }

    /* "View.MemoryView":360
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */

#line 360 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);

#line 360 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":361
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *                 if self.lock is NULL:
 *                     raise MemoryError
 */

#line 361 "View.MemoryView"
      __pyx_v_self->lock = PyThread_allocate_lock();

      /* "View.MemoryView":362
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */

#line 362 "View.MemoryView"
      __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);

#line 362 "View.MemoryView"
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":363
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 *                     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */

#line 363 "View.MemoryView"
        PyErr_NoMemory(); __PYX_ERR(2, 363, __pyx_L1_error)

        /* "View.MemoryView":362
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */

#line 362 "View.MemoryView"
      }

      /* "View.MemoryView":360
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */

#line 360 "View.MemoryView"
    }

    /* "View.MemoryView":355
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 */

#line 355 "View.MemoryView"
  }

  /* "View.MemoryView":365
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */

#line 365 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);

#line 365 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":366
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */

#line 366 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);

#line 366 "View.MemoryView"
    if (__pyx_t_2) {

#line 366 "View.MemoryView"
    } else {

#line 366 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 366 "View.MemoryView"
      goto __pyx_L12_bool_binop_done;

#line 366 "View.MemoryView"
    }

#line 366 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);

#line 366 "View.MemoryView"
    __pyx_t_1 = __pyx_t_2;

#line 366 "View.MemoryView"
    __pyx_L12_bool_binop_done:;

#line 366 "View.MemoryView"
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":365
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */

#line 365 "View.MemoryView"
    goto __pyx_L11;

#line 365 "View.MemoryView"
  }

  /* "View.MemoryView":368
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */

#line 368 "View.MemoryView"
  /*else*/ {

#line 368 "View.MemoryView"
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;

#line 368 "View.MemoryView"
  }

#line 368 "View.MemoryView"
  __pyx_L11:;

  /* "View.MemoryView":370
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */

#line 370 "View.MemoryView"
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":372
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */

#line 372 "View.MemoryView"
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":346
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
  /* function exit code */

#line 346 "View.MemoryView"
  __pyx_r = 0;

#line 346 "View.MemoryView"
  goto __pyx_L0;

#line 346 "View.MemoryView"
  __pyx_L1_error:;

#line 346 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 346 "View.MemoryView"
  __pyx_r = -1;

#line 346 "View.MemoryView"
  __pyx_L0:;

#line 346 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 346 "View.MemoryView"
  return __pyx_r;

#line 346 "View.MemoryView"
}

/* "View.MemoryView":374
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
/* Python wrapper */

#line 374 "View.MemoryView"
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/

#line 374 "View.MemoryView"
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {

#line 374 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 374 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);

#line 374 "View.MemoryView"
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
  /* function exit code */

#line 374 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 374 "View.MemoryView"
}

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i
#line 374 "View.MemoryView"
;

#line 374 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 374 "View.MemoryView"
  int __pyx_t_1;

#line 374 "View.MemoryView"
  int __pyx_t_2;

#line 374 "View.MemoryView"
  int __pyx_t_3;

#line 374 "View.MemoryView"
  int __pyx_t_4;

#line 374 "View.MemoryView"
  int __pyx_t_5;

#line 374 "View.MemoryView"
  PyThread_type_lock __pyx_t_6;

#line 374 "View.MemoryView"
  PyThread_type_lock __pyx_t_7;

#line 374 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":375
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */

#line 375 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);

#line 375 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 375 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":376
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */

#line 376 "View.MemoryView"
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":375
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */

#line 375 "View.MemoryView"
    goto __pyx_L3;

#line 375 "View.MemoryView"
  }

  /* "View.MemoryView":377
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */

#line 377 "View.MemoryView"
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);

#line 377 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":379
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */

#line 379 "View.MemoryView"
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":380
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */

#line 380 "View.MemoryView"
    Py_DECREF(Py_None);

    /* "View.MemoryView":377
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */

#line 377 "View.MemoryView"
  }

#line 377 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":384
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */

#line 384 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);

#line 384 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":385
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */

#line 385 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;

#line 385 "View.MemoryView"
    __pyx_t_4 = __pyx_t_3;

#line 385 "View.MemoryView"
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {

#line 385 "View.MemoryView"
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":386
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */

#line 386 "View.MemoryView"
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);

#line 386 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":387
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */

#line 387 "View.MemoryView"
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":388
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */

#line 388 "View.MemoryView"
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);

#line 388 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":390
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */

#line 390 "View.MemoryView"
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

#line 390 "View.MemoryView"
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":389
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */

#line 389 "View.MemoryView"
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;

#line 389 "View.MemoryView"
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":388
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */

#line 388 "View.MemoryView"
        }

        /* "View.MemoryView":391
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */

#line 391 "View.MemoryView"
        goto __pyx_L6_break;

        /* "View.MemoryView":386
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */

#line 386 "View.MemoryView"
      }

#line 386 "View.MemoryView"
    }

#line 386 "View.MemoryView"
    /*else*/ {

      /* "View.MemoryView":393
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */

#line 393 "View.MemoryView"
      PyThread_free_lock(__pyx_v_self->lock);

#line 393 "View.MemoryView"
    }

#line 393 "View.MemoryView"
    __pyx_L6_break:;

    /* "View.MemoryView":384
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */

#line 384 "View.MemoryView"
  }

  /* "View.MemoryView":374
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
  /* function exit code */

#line 374 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 374 "View.MemoryView"
}

/* "View.MemoryView":395
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

#line 395 "View.MemoryView"


#line 395 "View.MemoryView"
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim
#line 395 "View.MemoryView"
;
  char *__pyx_v_itemp
#line 395 "View.MemoryView"
;
  PyObject *__pyx_v_idx = NULL
#line 395 "View.MemoryView"
;

#line 395 "View.MemoryView"
  char *__pyx_r;

#line 395 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 395 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 395 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 395 "View.MemoryView"
  Py_ssize_t __pyx_t_3;

#line 395 "View.MemoryView"
  PyObject *(*__pyx_t_4)(PyObject *);

#line 395 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 395 "View.MemoryView"
  Py_ssize_t __pyx_t_6;

#line 395 "View.MemoryView"
  char *__pyx_t_7;

#line 395 "View.MemoryView"
  int __pyx_lineno = 0;

#line 395 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 395 "View.MemoryView"
  int __pyx_clineno = 0;

#line 395 "View.MemoryView"
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":397
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */

#line 397 "View.MemoryView"
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":399
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */

#line 399 "View.MemoryView"
  __pyx_t_1 = 0;

#line 399 "View.MemoryView"
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {

#line 399 "View.MemoryView"
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;

#line 399 "View.MemoryView"
    __pyx_t_4 = NULL;

#line 399 "View.MemoryView"
  } else {
    __pyx_t_3 = -1; 
#line 399 "View.MemoryView"
__pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 399 "View.MemoryView"
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
  }

#line 399 "View.MemoryView"
  for (;;) {

#line 399 "View.MemoryView"
    if (likely(!__pyx_t_4)) {

#line 399 "View.MemoryView"
      if (likely(PyList_CheckExact(__pyx_t_2))) {

#line 399 "View.MemoryView"
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;

#line 399 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 399 "View.MemoryView"
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        #else

#line 399 "View.MemoryView"
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_5);

#line 399 "View.MemoryView"
        #endif

#line 399 "View.MemoryView"
      } else {

#line 399 "View.MemoryView"
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;

#line 399 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 399 "View.MemoryView"
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        #else

#line 399 "View.MemoryView"
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_5);

#line 399 "View.MemoryView"
        #endif

#line 399 "View.MemoryView"
      }
    } else 
#line 399 "View.MemoryView"
{

#line 399 "View.MemoryView"
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);

#line 399 "View.MemoryView"
      if (unlikely(!__pyx_t_5)) {

#line 399 "View.MemoryView"
        PyObject* exc_type = PyErr_Occurred();

#line 399 "View.MemoryView"
        if (exc_type) {

#line 399 "View.MemoryView"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 399 "View.MemoryView"
          else __PYX_ERR(2, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        }

#line 399 "View.MemoryView"
        break;

#line 399 "View.MemoryView"
      }

#line 399 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 399 "View.MemoryView"
    }

#line 399 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);

#line 399 "View.MemoryView"
    __pyx_t_5 = 0;

#line 399 "View.MemoryView"
    __pyx_v_dim = __pyx_t_1;

#line 399 "View.MemoryView"
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":400
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */

#line 400 "View.MemoryView"
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 400, __pyx_L1_error)

#line 400 "View.MemoryView"
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(2, 400, __pyx_L1_error)

#line 400 "View.MemoryView"
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":399
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */

#line 399 "View.MemoryView"
  }

#line 399 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":402
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 402 "View.MemoryView"
  __pyx_r = __pyx_v_itemp;

#line 402 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":395
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

#line 395 "View.MemoryView"


#line 395 "View.MemoryView"
  /* function exit code */

#line 395 "View.MemoryView"
  __pyx_L1_error:;

#line 395 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 395 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 395 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 395 "View.MemoryView"
  __pyx_r = NULL;

#line 395 "View.MemoryView"
  __pyx_L0:;

#line 395 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_idx);

#line 395 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 395 "View.MemoryView"
  return __pyx_r;

#line 395 "View.MemoryView"
}

/* "View.MemoryView":405
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
/* Python wrapper */

#line 405 "View.MemoryView"
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/

#line 405 "View.MemoryView"
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {

#line 405 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 405 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 405 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 405 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
  /* function exit code */

#line 405 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 405 "View.MemoryView"
  return __pyx_r;

#line 405 "View.MemoryView"
}

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL
#line 405 "View.MemoryView"
;
  PyObject *__pyx_v_indices = NULL
#line 405 "View.MemoryView"
;
  char *__pyx_v_itemp
#line 405 "View.MemoryView"
;

#line 405 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 405 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 405 "View.MemoryView"
  int __pyx_t_1;

#line 405 "View.MemoryView"
  int __pyx_t_2;

#line 405 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 405 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 405 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 405 "View.MemoryView"
  char *__pyx_t_6;

#line 405 "View.MemoryView"
  int __pyx_lineno = 0;

#line 405 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 405 "View.MemoryView"
  int __pyx_clineno = 0;

#line 405 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":406
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */

#line 406 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);

#line 406 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 406 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":407
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */

#line 407 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 407 "View.MemoryView"
    __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 407 "View.MemoryView"
    __pyx_r = ((PyObject *)__pyx_v_self);

#line 407 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":406
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */

#line 406 "View.MemoryView"
  }

  /* "View.MemoryView":409
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */

#line 409 "View.MemoryView"
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 409 "View.MemoryView"
  if (likely(__pyx_t_3 != Py_None)) {

#line 409 "View.MemoryView"
    PyObject* sequence = __pyx_t_3;

#line 409 "View.MemoryView"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 409 "View.MemoryView"
    if (unlikely(size != 2)) {

#line 409 "View.MemoryView"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 409 "View.MemoryView"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 409 "View.MemoryView"
      __PYX_ERR(2, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
    }

#line 409 "View.MemoryView"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 409 "View.MemoryView"
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 

#line 409 "View.MemoryView"
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 

#line 409 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_4);

#line 409 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_5);

#line 409 "View.MemoryView"
    #else

#line 409 "View.MemoryView"
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 409 "View.MemoryView"
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 409 "View.MemoryView"
    #endif

#line 409 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 409 "View.MemoryView"
  } else {

#line 409 "View.MemoryView"
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
  }

#line 409 "View.MemoryView"
  __pyx_v_have_slices = __pyx_t_4;

#line 409 "View.MemoryView"
  __pyx_t_4 = 0;

#line 409 "View.MemoryView"
  __pyx_v_indices = __pyx_t_5;

#line 409 "View.MemoryView"
  __pyx_t_5 = 0;

  /* "View.MemoryView":412
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */

#line 412 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 412, __pyx_L1_error)

#line 412 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":413
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */

#line 413 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 413 "View.MemoryView"
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 413, __pyx_L1_error)

#line 413 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 413 "View.MemoryView"
    __pyx_r = __pyx_t_3;

#line 413 "View.MemoryView"
    __pyx_t_3 = 0;

#line 413 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":412
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */

#line 412 "View.MemoryView"
  }

  /* "View.MemoryView":415
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */

#line 415 "View.MemoryView"
  /*else*/ {

#line 415 "View.MemoryView"
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(2, 415, __pyx_L1_error)

#line 415 "View.MemoryView"
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":416
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */

#line 416 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 416 "View.MemoryView"
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 416, __pyx_L1_error)

#line 416 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 416 "View.MemoryView"
    __pyx_r = __pyx_t_3;

#line 416 "View.MemoryView"
    __pyx_t_3 = 0;

#line 416 "View.MemoryView"
    goto __pyx_L0;

#line 416 "View.MemoryView"
  }

  /* "View.MemoryView":405
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
  /* function exit code */

#line 405 "View.MemoryView"
  __pyx_L1_error:;

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 405 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 405 "View.MemoryView"
  __pyx_r = NULL;

#line 405 "View.MemoryView"
  __pyx_L0:;

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_have_slices);

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_indices);

#line 405 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 405 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 405 "View.MemoryView"
  return __pyx_r;

#line 405 "View.MemoryView"
}

/* "View.MemoryView":418
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
/* Python wrapper */

#line 418 "View.MemoryView"
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/

#line 418 "View.MemoryView"
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {

#line 418 "View.MemoryView"
  int __pyx_r;

#line 418 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 418 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = 
#line 418 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
  /* function exit code */

#line 418 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 418 "View.MemoryView"
  return __pyx_r;

#line 418 "View.MemoryView"
}

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL
#line 418 "View.MemoryView"
;
  PyObject *__pyx_v_obj = NULL
#line 418 "View.MemoryView"
;

#line 418 "View.MemoryView"
  int __pyx_r;

#line 418 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 418 "View.MemoryView"
  int __pyx_t_1;

#line 418 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 418 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 418 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 418 "View.MemoryView"
  int __pyx_lineno = 0;

#line 418 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 418 "View.MemoryView"
  int __pyx_clineno = 0;

#line 418 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__", 0);

#line 418 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":419
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */

#line 419 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);

#line 419 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":420
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */

#line 420 "View.MemoryView"
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 420, __pyx_L1_error)

#line 420 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 420 "View.MemoryView"
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 420 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 420 "View.MemoryView"
    __PYX_ERR(2, 420, __pyx_L1_error)

    /* "View.MemoryView":419
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */

#line 419 "View.MemoryView"
  }

  /* "View.MemoryView":422
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */

#line 422 "View.MemoryView"
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 422 "View.MemoryView"
  if (likely(__pyx_t_2 != Py_None)) {

#line 422 "View.MemoryView"
    PyObject* sequence = __pyx_t_2;

#line 422 "View.MemoryView"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 422 "View.MemoryView"
    if (unlikely(size != 2)) {

#line 422 "View.MemoryView"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 422 "View.MemoryView"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 422 "View.MemoryView"
      __PYX_ERR(2, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
    }

#line 422 "View.MemoryView"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 422 "View.MemoryView"
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 

#line 422 "View.MemoryView"
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 

#line 422 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 422 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_4);

#line 422 "View.MemoryView"
    #else

#line 422 "View.MemoryView"
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 422 "View.MemoryView"
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 422 "View.MemoryView"
    #endif

#line 422 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 422 "View.MemoryView"
  } else {

#line 422 "View.MemoryView"
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
  }

#line 422 "View.MemoryView"
  __pyx_v_have_slices = __pyx_t_3;

#line 422 "View.MemoryView"
  __pyx_t_3 = 0;

#line 422 "View.MemoryView"
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);

#line 422 "View.MemoryView"
  __pyx_t_4 = 0;

  /* "View.MemoryView":424
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */

#line 424 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 424, __pyx_L1_error)

#line 424 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":425
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */

#line 425 "View.MemoryView"
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 425, __pyx_L1_error)

#line 425 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 425 "View.MemoryView"
    __pyx_v_obj = __pyx_t_2;

#line 425 "View.MemoryView"
    __pyx_t_2 = 0;

    /* "View.MemoryView":426
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */

#line 426 "View.MemoryView"
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 426, __pyx_L1_error)

#line 426 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":427
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */

#line 427 "View.MemoryView"
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 427, __pyx_L1_error)

#line 427 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_2);

#line 427 "View.MemoryView"
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 427, __pyx_L1_error)

#line 427 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 427 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 427 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":426
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */

#line 426 "View.MemoryView"
      goto __pyx_L5;

#line 426 "View.MemoryView"
    }

    /* "View.MemoryView":429
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */

#line 429 "View.MemoryView"
    /*else*/ {

#line 429 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 429, __pyx_L1_error)

#line 429 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 429 "View.MemoryView"
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(2, 429, __pyx_L1_error)

#line 429 "View.MemoryView"
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 429, __pyx_L1_error)

#line 429 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_2);

#line 429 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 429 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 429 "View.MemoryView"
    }

#line 429 "View.MemoryView"
    __pyx_L5:;

    /* "View.MemoryView":424
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */

#line 424 "View.MemoryView"
    goto __pyx_L4;

#line 424 "View.MemoryView"
  }

  /* "View.MemoryView":431
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */

#line 431 "View.MemoryView"
  /*else*/ {

#line 431 "View.MemoryView"
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 431, __pyx_L1_error)

#line 431 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 431 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 431 "View.MemoryView"
  }

#line 431 "View.MemoryView"
  __pyx_L4:;

  /* "View.MemoryView":418
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
  /* function exit code */

#line 418 "View.MemoryView"
  __pyx_r = 0;

#line 418 "View.MemoryView"
  goto __pyx_L0;

#line 418 "View.MemoryView"
  __pyx_L1_error:;

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 418 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 418 "View.MemoryView"
  __pyx_r = -1;

#line 418 "View.MemoryView"
  __pyx_L0:;

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_have_slices);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_obj);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_index);

#line 418 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 418 "View.MemoryView"
  return __pyx_r;

#line 418 "View.MemoryView"
}

/* "View.MemoryView":433
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

#line 433 "View.MemoryView"


#line 433 "View.MemoryView"
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {

#line 433 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 433 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 433 "View.MemoryView"
  int __pyx_t_1;

#line 433 "View.MemoryView"
  int __pyx_t_2;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_8 = NULL;

#line 433 "View.MemoryView"
  int __pyx_t_9;

#line 433 "View.MemoryView"
  int __pyx_lineno = 0;

#line 433 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 433 "View.MemoryView"
  int __pyx_clineno = 0;

#line 433 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_slice", 0);

#line 433 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":434
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */

#line 434 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 

#line 434 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);

#line 434 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */

#line 435 "View.MemoryView"
    {

#line 435 "View.MemoryView"
      __Pyx_PyThreadState_declare

#line 435 "View.MemoryView"
      __Pyx_PyThreadState_assign

#line 435 "View.MemoryView"
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);

#line 435 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_3);

#line 435 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_4);

#line 435 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_5);

#line 435 "View.MemoryView"
      /*try:*/ {

        /* "View.MemoryView":436
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */

#line 436 "View.MemoryView"
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 436, __pyx_L4_error)

#line 436 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */

#line 437 "View.MemoryView"
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 437, __pyx_L4_error)

#line 437 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":436
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */

#line 436 "View.MemoryView"
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 436, __pyx_L4_error)

#line 436 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_8);

#line 436 "View.MemoryView"
        __Pyx_INCREF(__pyx_v_obj);

#line 436 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_v_obj);

#line 436 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);

#line 436 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_6);

#line 436 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);

#line 436 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_7);

#line 436 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);

#line 436 "View.MemoryView"
        __pyx_t_6 = 0;

#line 436 "View.MemoryView"
        __pyx_t_7 = 0;

#line 436 "View.MemoryView"
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 436, __pyx_L4_error)

#line 436 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 436 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 436 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);

#line 436 "View.MemoryView"
        __pyx_t_7 = 0;

        /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */

#line 435 "View.MemoryView"
      }

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 435 "View.MemoryView"
      goto __pyx_L9_try_end;

#line 435 "View.MemoryView"
      __pyx_L4_error:;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":438
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */

#line 438 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);

#line 438 "View.MemoryView"
      if (__pyx_t_9) {

#line 438 "View.MemoryView"
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 438 "View.MemoryView"
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(2, 438, __pyx_L6_except_error)

#line 438 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 438 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_8);

#line 438 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":439
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */

#line 439 "View.MemoryView"
        __Pyx_XDECREF(__pyx_r);

#line 439 "View.MemoryView"
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 439 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 439 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 439 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 439 "View.MemoryView"
        goto __pyx_L7_except_return;

#line 439 "View.MemoryView"
      }

#line 439 "View.MemoryView"
      goto __pyx_L6_except_error;

#line 439 "View.MemoryView"
      __pyx_L6_except_error:;

      /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_3);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_4);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_5);

#line 435 "View.MemoryView"
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);

#line 435 "View.MemoryView"
      goto __pyx_L1_error;

#line 435 "View.MemoryView"
      __pyx_L7_except_return:;

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_3);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_4);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_5);

#line 435 "View.MemoryView"
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);

#line 435 "View.MemoryView"
      goto __pyx_L0;

#line 435 "View.MemoryView"
      __pyx_L9_try_end:;

#line 435 "View.MemoryView"
    }

    /* "View.MemoryView":434
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */

#line 434 "View.MemoryView"
  }

  /* "View.MemoryView":441
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */

#line 441 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 441 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_obj);

#line 441 "View.MemoryView"
  __pyx_r = __pyx_v_obj;

#line 441 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":433
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

#line 433 "View.MemoryView"


#line 433 "View.MemoryView"
  /* function exit code */

#line 433 "View.MemoryView"
  __pyx_L1_error:;

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_7);

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_8);

#line 433 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 433 "View.MemoryView"
  __pyx_r = 0;

#line 433 "View.MemoryView"
  __pyx_L0:;

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_obj);

#line 433 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 433 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 433 "View.MemoryView"
  return __pyx_r;

#line 433 "View.MemoryView"
}

/* "View.MemoryView":443
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

#line 443 "View.MemoryView"


#line 443 "View.MemoryView"
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice
#line 443 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_src_slice
#line 443 "View.MemoryView"
;

#line 443 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 443 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 443 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 443 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_2;

#line 443 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 443 "View.MemoryView"
  int __pyx_t_4;

#line 443 "View.MemoryView"
  int __pyx_t_5;

#line 443 "View.MemoryView"
  int __pyx_t_6;

#line 443 "View.MemoryView"
  int __pyx_lineno = 0;

#line 443 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 443 "View.MemoryView"
  int __pyx_clineno = 0;

#line 443 "View.MemoryView"
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":447
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */

#line 447 "View.MemoryView"
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(2, 447, __pyx_L1_error)

#line 447 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 447, __pyx_L1_error)

  /* "View.MemoryView":448
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */

#line 448 "View.MemoryView"
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(2, 448, __pyx_L1_error)

#line 448 "View.MemoryView"
  __pyx_t_2 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_2 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 448, __pyx_L1_error)

  /* "View.MemoryView":449
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */

#line 449 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 449 "View.MemoryView"
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 449 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 449 "View.MemoryView"
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":447
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */

#line 447 "View.MemoryView"
  __pyx_t_6 = __pyx_memoryview_copy_contents((__pyx_t_1[0]), (__pyx_t_2[0]), __pyx_t_4, __pyx_t_5, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 447, __pyx_L1_error)

  /* "View.MemoryView":443
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

#line 443 "View.MemoryView"


#line 443 "View.MemoryView"
  /* function exit code */

#line 443 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 443 "View.MemoryView"
  goto __pyx_L0;

#line 443 "View.MemoryView"
  __pyx_L1_error:;

#line 443 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 443 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 443 "View.MemoryView"
  __pyx_r = 0;

#line 443 "View.MemoryView"
  __pyx_L0:;

#line 443 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 443 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 443 "View.MemoryView"
  return __pyx_r;

#line 443 "View.MemoryView"
}

/* "View.MemoryView":451
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

#line 451 "View.MemoryView"


#line 451 "View.MemoryView"
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80]
#line 451 "View.MemoryView"
;
  void *__pyx_v_tmp
#line 451 "View.MemoryView"
;
  void *__pyx_v_item
#line 451 "View.MemoryView"
;
  __Pyx_memviewslice *__pyx_v_dst_slice
#line 451 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp_slice
#line 451 "View.MemoryView"
;

#line 451 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 451 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 451 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 451 "View.MemoryView"
  int __pyx_t_2;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 451 "View.MemoryView"
  int __pyx_t_4;

#line 451 "View.MemoryView"
  int __pyx_t_5;

#line 451 "View.MemoryView"
  char const *__pyx_t_6;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_8 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_9 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_10 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_11 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_12 = NULL;

#line 451 "View.MemoryView"
  int __pyx_lineno = 0;

#line 451 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 451 "View.MemoryView"
  int __pyx_clineno = 0;

#line 451 "View.MemoryView"
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":453
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */

#line 453 "View.MemoryView"
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":458
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */

#line 458 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 458, __pyx_L1_error)

#line 458 "View.MemoryView"
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":460
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */

#line 460 "View.MemoryView"
  __pyx_t_2 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);

#line 460 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":461
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */

#line 461 "View.MemoryView"
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":462
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */

#line 462 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);

#line 462 "View.MemoryView"
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":463
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */

#line 463 "View.MemoryView"
      PyErr_NoMemory(); __PYX_ERR(2, 463, __pyx_L1_error)

      /* "View.MemoryView":462
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */

#line 462 "View.MemoryView"
    }

    /* "View.MemoryView":464
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */

#line 464 "View.MemoryView"
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":460
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */

#line 460 "View.MemoryView"
    goto __pyx_L3;

#line 460 "View.MemoryView"
  }

  /* "View.MemoryView":466
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */

#line 466 "View.MemoryView"
  /*else*/ {

#line 466 "View.MemoryView"
    __pyx_v_item = ((void *)__pyx_v_array);

#line 466 "View.MemoryView"
  }

#line 466 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":468
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */

#line 468 "View.MemoryView"
  /*try:*/ {

    /* "View.MemoryView":469
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */

#line 469 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_self->dtype_is_object != 0);

#line 469 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":470
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */

#line 470 "View.MemoryView"
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":469
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */

#line 469 "View.MemoryView"
      goto __pyx_L8;

#line 469 "View.MemoryView"
    }

    /* "View.MemoryView":472
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 472 "View.MemoryView"
    /*else*/ {

#line 472 "View.MemoryView"
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 472, __pyx_L6_error)

#line 472 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 472 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 472 "View.MemoryView"
    }

#line 472 "View.MemoryView"
    __pyx_L8:;

    /* "View.MemoryView":476
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */

#line 476 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_self->view.suboffsets != NULL) != 0);

#line 476 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":477
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */

#line 477 "View.MemoryView"
      __pyx_t_3 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 477, __pyx_L6_error)

#line 477 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 477 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":476
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */

#line 476 "View.MemoryView"
    }

    /* "View.MemoryView":478
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */

#line 478 "View.MemoryView"
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);

#line 478 "View.MemoryView"
  }

  /* "View.MemoryView":481
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */

#line 481 "View.MemoryView"
  /*finally:*/ {

#line 481 "View.MemoryView"
    /*normal exit:*/{

#line 481 "View.MemoryView"
      PyMem_Free(__pyx_v_tmp);

#line 481 "View.MemoryView"
      goto __pyx_L7;

#line 481 "View.MemoryView"
    }

#line 481 "View.MemoryView"
    __pyx_L6_error:;

#line 481 "View.MemoryView"
    /*exception exit:*/{

#line 481 "View.MemoryView"
      __Pyx_PyThreadState_declare

#line 481 "View.MemoryView"
      __Pyx_PyThreadState_assign

#line 481 "View.MemoryView"
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;

#line 481 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 481 "View.MemoryView"
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);

#line 481 "View.MemoryView"
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_7);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_8);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_9);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_10);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_11);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_12);

#line 481 "View.MemoryView"
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;

#line 481 "View.MemoryView"
      {

#line 481 "View.MemoryView"
        PyMem_Free(__pyx_v_tmp);

#line 481 "View.MemoryView"
      }

#line 481 "View.MemoryView"
      if (PY_MAJOR_VERSION >= 3) {

#line 481 "View.MemoryView"
        __Pyx_XGIVEREF(__pyx_t_10);

#line 481 "View.MemoryView"
        __Pyx_XGIVEREF(__pyx_t_11);

#line 481 "View.MemoryView"
        __Pyx_XGIVEREF(__pyx_t_12);

#line 481 "View.MemoryView"
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);

#line 481 "View.MemoryView"
      }

#line 481 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_7);

#line 481 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_8);

#line 481 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_9);

#line 481 "View.MemoryView"
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);

#line 481 "View.MemoryView"
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;

#line 481 "View.MemoryView"
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;

#line 481 "View.MemoryView"
      goto __pyx_L1_error;

#line 481 "View.MemoryView"
    }

#line 481 "View.MemoryView"
    __pyx_L7:;

#line 481 "View.MemoryView"
  }

  /* "View.MemoryView":451
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

#line 451 "View.MemoryView"


#line 451 "View.MemoryView"
  /* function exit code */

#line 451 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 451 "View.MemoryView"
  goto __pyx_L0;

#line 451 "View.MemoryView"
  __pyx_L1_error:;

#line 451 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 451 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 451 "View.MemoryView"
  __pyx_r = 0;

#line 451 "View.MemoryView"
  __pyx_L0:;

#line 451 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 451 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 451 "View.MemoryView"
  return __pyx_r;

#line 451 "View.MemoryView"
}

/* "View.MemoryView":483
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

#line 483 "View.MemoryView"


#line 483 "View.MemoryView"
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp
#line 483 "View.MemoryView"
;

#line 483 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 483 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 483 "View.MemoryView"
  char *__pyx_t_1;

#line 483 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 483 "View.MemoryView"
  int __pyx_lineno = 0;

#line 483 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 483 "View.MemoryView"
  int __pyx_clineno = 0;

#line 483 "View.MemoryView"
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":484
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */

#line 484 "View.MemoryView"
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(2, 484, __pyx_L1_error)

#line 484 "View.MemoryView"
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":485
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */

#line 485 "View.MemoryView"
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 485, __pyx_L1_error)

#line 485 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 485 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":483
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

#line 483 "View.MemoryView"


#line 483 "View.MemoryView"
  /* function exit code */

#line 483 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 483 "View.MemoryView"
  goto __pyx_L0;

#line 483 "View.MemoryView"
  __pyx_L1_error:;

#line 483 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 483 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 483 "View.MemoryView"
  __pyx_r = 0;

#line 483 "View.MemoryView"
  __pyx_L0:;

#line 483 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 483 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 483 "View.MemoryView"
  return __pyx_r;

#line 483 "View.MemoryView"
}

/* "View.MemoryView":487
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 487 "View.MemoryView"


#line 487 "View.MemoryView"
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL
#line 487 "View.MemoryView"
;
  PyObject *__pyx_v_bytesitem = 0
#line 487 "View.MemoryView"
;
  PyObject *__pyx_v_result = NULL
#line 487 "View.MemoryView"
;

#line 487 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 487 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 487 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 487 "View.MemoryView"
  int __pyx_t_8;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_9 = NULL;

#line 487 "View.MemoryView"
  size_t __pyx_t_10;

#line 487 "View.MemoryView"
  int __pyx_t_11;

#line 487 "View.MemoryView"
  int __pyx_lineno = 0;

#line 487 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 487 "View.MemoryView"
  int __pyx_clineno = 0;

#line 487 "View.MemoryView"
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":490
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */

#line 490 "View.MemoryView"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 490, __pyx_L1_error)

#line 490 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 490 "View.MemoryView"
  __pyx_v_struct = __pyx_t_1;

#line 490 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":493
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */

#line 493 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 493, __pyx_L1_error)

#line 493 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 493 "View.MemoryView"
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);

#line 493 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */

#line 494 "View.MemoryView"
  {

#line 494 "View.MemoryView"
    __Pyx_PyThreadState_declare

#line 494 "View.MemoryView"
    __Pyx_PyThreadState_assign

#line 494 "View.MemoryView"
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);

#line 494 "View.MemoryView"
    __Pyx_XGOTREF(__pyx_t_2);

#line 494 "View.MemoryView"
    __Pyx_XGOTREF(__pyx_t_3);

#line 494 "View.MemoryView"
    __Pyx_XGOTREF(__pyx_t_4);

#line 494 "View.MemoryView"
    /*try:*/ {

      /* "View.MemoryView":495
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */

#line 495 "View.MemoryView"
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 495 "View.MemoryView"
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 495 "View.MemoryView"
      __pyx_t_7 = NULL;

#line 495 "View.MemoryView"
      __pyx_t_8 = 0;

#line 495 "View.MemoryView"
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {

#line 495 "View.MemoryView"
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);

#line 495 "View.MemoryView"
        if (likely(__pyx_t_7)) {

#line 495 "View.MemoryView"
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 495 "View.MemoryView"
          __Pyx_INCREF(__pyx_t_7);

#line 495 "View.MemoryView"
          __Pyx_INCREF(function);

#line 495 "View.MemoryView"
          __Pyx_DECREF_SET(__pyx_t_5, function);

#line 495 "View.MemoryView"
          __pyx_t_8 = 1;

#line 495 "View.MemoryView"
        }

#line 495 "View.MemoryView"
      }

#line 495 "View.MemoryView"
      #if CYTHON_FAST_PYCALL

#line 495 "View.MemoryView"
      if (PyFunction_Check(__pyx_t_5)) {

#line 495 "View.MemoryView"
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};

#line 495 "View.MemoryView"
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 495 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 495 "View.MemoryView"
      } else

#line 495 "View.MemoryView"
      #endif

#line 495 "View.MemoryView"
      #if CYTHON_FAST_PYCCALL

#line 495 "View.MemoryView"
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 495 "View.MemoryView"
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};

#line 495 "View.MemoryView"
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 495 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 495 "View.MemoryView"
      } else

#line 495 "View.MemoryView"
      #endif

#line 495 "View.MemoryView"
      {

#line 495 "View.MemoryView"
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_9);

#line 495 "View.MemoryView"
        if (__pyx_t_7) {

#line 495 "View.MemoryView"
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;

#line 495 "View.MemoryView"
        }

#line 495 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_6);

#line 495 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);

#line 495 "View.MemoryView"
        __Pyx_INCREF(__pyx_v_bytesitem);

#line 495 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_v_bytesitem);

#line 495 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);

#line 495 "View.MemoryView"
        __pyx_t_6 = 0;

#line 495 "View.MemoryView"
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 495 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 495 "View.MemoryView"
      }

#line 495 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 495 "View.MemoryView"
      __pyx_v_result = __pyx_t_1;

#line 495 "View.MemoryView"
      __pyx_t_1 = 0;

      /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */

#line 494 "View.MemoryView"
    }

    /* "View.MemoryView":499
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */

#line 499 "View.MemoryView"
    /*else:*/ {

#line 499 "View.MemoryView"
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 

#line 499 "View.MemoryView"
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);

#line 499 "View.MemoryView"
      if (__pyx_t_11) {

        /* "View.MemoryView":500
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */

#line 500 "View.MemoryView"
        __Pyx_XDECREF(__pyx_r);

#line 500 "View.MemoryView"
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 500, __pyx_L5_except_error)

#line 500 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 500 "View.MemoryView"
        __pyx_r = __pyx_t_1;

#line 500 "View.MemoryView"
        __pyx_t_1 = 0;

#line 500 "View.MemoryView"
        goto __pyx_L6_except_return;

        /* "View.MemoryView":499
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */

#line 499 "View.MemoryView"
      }

      /* "View.MemoryView":501
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */

#line 501 "View.MemoryView"
      __Pyx_XDECREF(__pyx_r);

#line 501 "View.MemoryView"
      __Pyx_INCREF(__pyx_v_result);

#line 501 "View.MemoryView"
      __pyx_r = __pyx_v_result;

#line 501 "View.MemoryView"
      goto __pyx_L6_except_return;

#line 501 "View.MemoryView"
    }

#line 501 "View.MemoryView"
    __pyx_L3_error:;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":496
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */

#line 496 "View.MemoryView"
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);

#line 496 "View.MemoryView"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 496, __pyx_L5_except_error)

#line 496 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 496 "View.MemoryView"
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);

#line 496 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 496 "View.MemoryView"
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);

#line 496 "View.MemoryView"
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;

#line 496 "View.MemoryView"
    if (__pyx_t_8) {

#line 496 "View.MemoryView"
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 496 "View.MemoryView"
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(2, 496, __pyx_L5_except_error)

#line 496 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 496 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 496 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":497
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */

#line 497 "View.MemoryView"
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 497, __pyx_L5_except_error)

#line 497 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 497 "View.MemoryView"
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 497 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 497 "View.MemoryView"
      __PYX_ERR(2, 497, __pyx_L5_except_error)

#line 497 "View.MemoryView"
    }

#line 497 "View.MemoryView"
    goto __pyx_L5_except_error;

#line 497 "View.MemoryView"
    __pyx_L5_except_error:;

    /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_4);

#line 494 "View.MemoryView"
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);

#line 494 "View.MemoryView"
    goto __pyx_L1_error;

#line 494 "View.MemoryView"
    __pyx_L6_except_return:;

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_4);

#line 494 "View.MemoryView"
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);

#line 494 "View.MemoryView"
    goto __pyx_L0;

#line 494 "View.MemoryView"
  }

  /* "View.MemoryView":487
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 487 "View.MemoryView"


#line 487 "View.MemoryView"
  /* function exit code */

#line 487 "View.MemoryView"
  __pyx_L1_error:;

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_7);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_9);

#line 487 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 487 "View.MemoryView"
  __pyx_r = 0;

#line 487 "View.MemoryView"
  __pyx_L0:;

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_struct);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_bytesitem);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_result);

#line 487 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 487 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 487 "View.MemoryView"
  return __pyx_r;

#line 487 "View.MemoryView"
}

/* "View.MemoryView":503
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 503 "View.MemoryView"


#line 503 "View.MemoryView"
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL
#line 503 "View.MemoryView"
;
  char __pyx_v_c
#line 503 "View.MemoryView"
;
  PyObject *__pyx_v_bytesvalue = 0
#line 503 "View.MemoryView"
;
  Py_ssize_t __pyx_v_i
#line 503 "View.MemoryView"
;

#line 503 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 503 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 503 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 503 "View.MemoryView"
  int __pyx_t_2;

#line 503 "View.MemoryView"
  int __pyx_t_3;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 503 "View.MemoryView"
  int __pyx_t_7;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_8 = NULL;

#line 503 "View.MemoryView"
  Py_ssize_t __pyx_t_9;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_10 = NULL;

#line 503 "View.MemoryView"
  char *__pyx_t_11;

#line 503 "View.MemoryView"
  char *__pyx_t_12;

#line 503 "View.MemoryView"
  char *__pyx_t_13;

#line 503 "View.MemoryView"
  char *__pyx_t_14;

#line 503 "View.MemoryView"
  int __pyx_lineno = 0;

#line 503 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 503 "View.MemoryView"
  int __pyx_clineno = 0;

#line 503 "View.MemoryView"
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":506
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */

#line 506 "View.MemoryView"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 506, __pyx_L1_error)

#line 506 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 506 "View.MemoryView"
  __pyx_v_struct = __pyx_t_1;

#line 506 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":511
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */

#line 511 "View.MemoryView"
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 

#line 511 "View.MemoryView"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 511 "View.MemoryView"
  if (__pyx_t_3) {

    /* "View.MemoryView":512
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */

#line 512 "View.MemoryView"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_1);

#line 512 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 512 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_4);

#line 512 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_4 = 0;

#line 512 "View.MemoryView"
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 512 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 512 "View.MemoryView"
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(2, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_4 = 0;

    /* "View.MemoryView":511
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */

#line 511 "View.MemoryView"
    goto __pyx_L3;

#line 511 "View.MemoryView"
  }

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */

#line 514 "View.MemoryView"
  /*else*/ {

#line 514 "View.MemoryView"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 514 "View.MemoryView"
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_1);

#line 514 "View.MemoryView"
    __pyx_t_5 = NULL;

#line 514 "View.MemoryView"
    __pyx_t_7 = 0;

#line 514 "View.MemoryView"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {

#line 514 "View.MemoryView"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);

#line 514 "View.MemoryView"
      if (likely(__pyx_t_5)) {

#line 514 "View.MemoryView"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);

#line 514 "View.MemoryView"
        __Pyx_INCREF(__pyx_t_5);

#line 514 "View.MemoryView"
        __Pyx_INCREF(function);

#line 514 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_t_6, function);

#line 514 "View.MemoryView"
        __pyx_t_7 = 1;

#line 514 "View.MemoryView"
      }

#line 514 "View.MemoryView"
    }

#line 514 "View.MemoryView"
    #if CYTHON_FAST_PYCALL

#line 514 "View.MemoryView"
    if (PyFunction_Check(__pyx_t_6)) {

#line 514 "View.MemoryView"
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};

#line 514 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 514 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 514 "View.MemoryView"
    } else

#line 514 "View.MemoryView"
    #endif

#line 514 "View.MemoryView"
    #if CYTHON_FAST_PYCCALL

#line 514 "View.MemoryView"
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {

#line 514 "View.MemoryView"
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};

#line 514 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 514 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 514 "View.MemoryView"
    } else

#line 514 "View.MemoryView"
    #endif

#line 514 "View.MemoryView"
    {

#line 514 "View.MemoryView"
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_8);

#line 514 "View.MemoryView"
      if (__pyx_t_5) {

#line 514 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;

#line 514 "View.MemoryView"
      }

#line 514 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_t_1);

#line 514 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);

#line 514 "View.MemoryView"
      __Pyx_INCREF(__pyx_v_value);

#line 514 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_v_value);

#line 514 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);

#line 514 "View.MemoryView"
      __pyx_t_1 = 0;

#line 514 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 514 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 514 "View.MemoryView"
    }

#line 514 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 514 "View.MemoryView"
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(2, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);

#line 514 "View.MemoryView"
    __pyx_t_4 = 0;

#line 514 "View.MemoryView"
  }

#line 514 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */

#line 516 "View.MemoryView"
  __pyx_t_9 = 0;

#line 516 "View.MemoryView"
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {

#line 516 "View.MemoryView"
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");

#line 516 "View.MemoryView"
    __PYX_ERR(2, 516, __pyx_L1_error)

#line 516 "View.MemoryView"
  }

#line 516 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_bytesvalue);

#line 516 "View.MemoryView"
  __pyx_t_10 = __pyx_v_bytesvalue;

#line 516 "View.MemoryView"
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);

#line 516 "View.MemoryView"
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));

#line 516 "View.MemoryView"
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {

#line 516 "View.MemoryView"
    __pyx_t_11 = __pyx_t_14;

#line 516 "View.MemoryView"
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":517
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */

#line 517 "View.MemoryView"
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */

#line 516 "View.MemoryView"
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":517
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */

#line 517 "View.MemoryView"
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;

#line 517 "View.MemoryView"
  }

#line 517 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":503
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 503 "View.MemoryView"


#line 503 "View.MemoryView"
  /* function exit code */

#line 503 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 503 "View.MemoryView"
  goto __pyx_L0;

#line 503 "View.MemoryView"
  __pyx_L1_error:;

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_8);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_10);

#line 503 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 503 "View.MemoryView"
  __pyx_r = 0;

#line 503 "View.MemoryView"
  __pyx_L0:;

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_struct);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_bytesvalue);

#line 503 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 503 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 503 "View.MemoryView"
  return __pyx_r;

#line 503 "View.MemoryView"
}

/* "View.MemoryView":520
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
/* Python wrapper */

#line 520 "View.MemoryView"
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/

#line 520 "View.MemoryView"
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {

#line 520 "View.MemoryView"
  int __pyx_r;

#line 520 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 520 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = 
#line 520 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
  /* function exit code */

#line 520 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 520 "View.MemoryView"
  return __pyx_r;

#line 520 "View.MemoryView"
}

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {

#line 520 "View.MemoryView"
  int __pyx_r;

#line 520 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 520 "View.MemoryView"
  int __pyx_t_1;

#line 520 "View.MemoryView"
  int __pyx_t_2;

#line 520 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 520 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 520 "View.MemoryView"
  char *__pyx_t_5;

#line 520 "View.MemoryView"
  void *__pyx_t_6;

#line 520 "View.MemoryView"
  int __pyx_t_7;

#line 520 "View.MemoryView"
  Py_ssize_t __pyx_t_8;

#line 520 "View.MemoryView"
  int __pyx_lineno = 0;

#line 520 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 520 "View.MemoryView"
  int __pyx_clineno = 0;

#line 520 "View.MemoryView"
  if (__pyx_v_info == NULL) {

#line 520 "View.MemoryView"
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");

#line 520 "View.MemoryView"
    return -1;

#line 520 "View.MemoryView"
  }

#line 520 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__", 0);

#line 520 "View.MemoryView"
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);

#line 520 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":521
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */

#line 521 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);

#line 521 "View.MemoryView"
  if (__pyx_t_2) {

#line 521 "View.MemoryView"
  } else {

#line 521 "View.MemoryView"
    __pyx_t_1 = __pyx_t_2;

#line 521 "View.MemoryView"
    goto __pyx_L4_bool_binop_done;

#line 521 "View.MemoryView"
  }

#line 521 "View.MemoryView"
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);

#line 521 "View.MemoryView"
  __pyx_t_1 = __pyx_t_2;

#line 521 "View.MemoryView"
  __pyx_L4_bool_binop_done:;

#line 521 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":522
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */

#line 522 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 522, __pyx_L1_error)

#line 522 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 522 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 522 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 522 "View.MemoryView"
    __PYX_ERR(2, 522, __pyx_L1_error)

    /* "View.MemoryView":521
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */

#line 521 "View.MemoryView"
  }

  /* "View.MemoryView":524
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */

#line 524 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);

#line 524 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":525
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */

#line 525 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->view.shape;

#line 525 "View.MemoryView"
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":524
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */

#line 524 "View.MemoryView"
    goto __pyx_L6;

#line 524 "View.MemoryView"
  }

  /* "View.MemoryView":527
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */

#line 527 "View.MemoryView"
  /*else*/ {

#line 527 "View.MemoryView"
    __pyx_v_info->shape = NULL;

#line 527 "View.MemoryView"
  }

#line 527 "View.MemoryView"
  __pyx_L6:;

  /* "View.MemoryView":529
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */

#line 529 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);

#line 529 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":530
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */

#line 530 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->view.strides;

#line 530 "View.MemoryView"
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":529
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */

#line 529 "View.MemoryView"
    goto __pyx_L7;

#line 529 "View.MemoryView"
  }

  /* "View.MemoryView":532
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */

#line 532 "View.MemoryView"
  /*else*/ {

#line 532 "View.MemoryView"
    __pyx_v_info->strides = NULL;

#line 532 "View.MemoryView"
  }

#line 532 "View.MemoryView"
  __pyx_L7:;

  /* "View.MemoryView":534
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */

#line 534 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);

#line 534 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":535
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */

#line 535 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->view.suboffsets;

#line 535 "View.MemoryView"
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":534
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */

#line 534 "View.MemoryView"
    goto __pyx_L8;

#line 534 "View.MemoryView"
  }

  /* "View.MemoryView":537
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */

#line 537 "View.MemoryView"
  /*else*/ {

#line 537 "View.MemoryView"
    __pyx_v_info->suboffsets = NULL;

#line 537 "View.MemoryView"
  }

#line 537 "View.MemoryView"
  __pyx_L8:;

  /* "View.MemoryView":539
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */

#line 539 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);

#line 539 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":540
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */

#line 540 "View.MemoryView"
    __pyx_t_5 = __pyx_v_self->view.format;

#line 540 "View.MemoryView"
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":539
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */

#line 539 "View.MemoryView"
    goto __pyx_L9;

#line 539 "View.MemoryView"
  }

  /* "View.MemoryView":542
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */

#line 542 "View.MemoryView"
  /*else*/ {

#line 542 "View.MemoryView"
    __pyx_v_info->format = NULL;

#line 542 "View.MemoryView"
  }

#line 542 "View.MemoryView"
  __pyx_L9:;

  /* "View.MemoryView":544
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */

#line 544 "View.MemoryView"
  __pyx_t_6 = __pyx_v_self->view.buf;

#line 544 "View.MemoryView"
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":545
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */

#line 545 "View.MemoryView"
  __pyx_t_7 = __pyx_v_self->view.ndim;

#line 545 "View.MemoryView"
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":546
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */

#line 546 "View.MemoryView"
  __pyx_t_8 = __pyx_v_self->view.itemsize;

#line 546 "View.MemoryView"
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":547
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */

#line 547 "View.MemoryView"
  __pyx_t_8 = __pyx_v_self->view.len;

#line 547 "View.MemoryView"
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":548
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */

#line 548 "View.MemoryView"
  __pyx_t_1 = __pyx_v_self->view.readonly;

#line 548 "View.MemoryView"
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":549
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */

#line 549 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 549 "View.MemoryView"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 549 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_info->obj);

#line 549 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_info->obj);

#line 549 "View.MemoryView"
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":520
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
  /* function exit code */

#line 520 "View.MemoryView"
  __pyx_r = 0;

#line 520 "View.MemoryView"
  goto __pyx_L0;

#line 520 "View.MemoryView"
  __pyx_L1_error:;

#line 520 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 520 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 520 "View.MemoryView"
  __pyx_r = -1;

#line 520 "View.MemoryView"
  if (__pyx_v_info->obj != NULL) {

#line 520 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 520 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 520 "View.MemoryView"
  }

#line 520 "View.MemoryView"
  goto __pyx_L2;

#line 520 "View.MemoryView"
  __pyx_L0:;

#line 520 "View.MemoryView"
  if (__pyx_v_info->obj == Py_None) {

#line 520 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 520 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 520 "View.MemoryView"
  }

#line 520 "View.MemoryView"
  __pyx_L2:;

#line 520 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 520 "View.MemoryView"
  return __pyx_r;

#line 520 "View.MemoryView"
}

/* "View.MemoryView":555
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
/* Python wrapper */

#line 555 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 555 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {

#line 555 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 555 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 555 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 555 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
  /* function exit code */

#line 555 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 555 "View.MemoryView"
  return __pyx_r;

#line 555 "View.MemoryView"
}

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0
#line 555 "View.MemoryView"
;

#line 555 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 555 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 555 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 555 "View.MemoryView"
  int __pyx_t_2;

#line 555 "View.MemoryView"
  int __pyx_lineno = 0;

#line 555 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 555 "View.MemoryView"
  int __pyx_clineno = 0;

#line 555 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":556
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */

#line 556 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 556, __pyx_L1_error)

#line 556 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 556 "View.MemoryView"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(2, 556, __pyx_L1_error)

#line 556 "View.MemoryView"
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);

#line 556 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":557
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */

#line 557 "View.MemoryView"
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 557, __pyx_L1_error)

  /* "View.MemoryView":558
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 558 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 558 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 558 "View.MemoryView"
  __pyx_r = ((PyObject *)__pyx_v_result);

#line 558 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":555
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
  /* function exit code */

#line 555 "View.MemoryView"
  __pyx_L1_error:;

#line 555 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 555 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 555 "View.MemoryView"
  __pyx_r = NULL;

#line 555 "View.MemoryView"
  __pyx_L0:;

#line 555 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 555 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 555 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 555 "View.MemoryView"
  return __pyx_r;

#line 555 "View.MemoryView"
}

/* "View.MemoryView":561
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
/* Python wrapper */

#line 561 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 561 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {

#line 561 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 561 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 561 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 561 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
  /* function exit code */

#line 561 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 561 "View.MemoryView"
  return __pyx_r;

#line 561 "View.MemoryView"
}

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 561 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 561 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 561 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":562
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 562 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 562 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->obj);

#line 562 "View.MemoryView"
  __pyx_r = __pyx_v_self->obj;

#line 562 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":561
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
  /* function exit code */

#line 561 "View.MemoryView"
  __pyx_L0:;

#line 561 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 561 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 561 "View.MemoryView"
  return __pyx_r;

#line 561 "View.MemoryView"
}

/* "View.MemoryView":565
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
/* Python wrapper */

#line 565 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 565 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {

#line 565 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 565 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 565 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 565 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
  /* function exit code */

#line 565 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 565 "View.MemoryView"
  return __pyx_r;

#line 565 "View.MemoryView"
}

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length
#line 565 "View.MemoryView"
;

#line 565 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 565 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 565 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 565 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 565 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 565 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 565 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 565 "View.MemoryView"
  int __pyx_lineno = 0;

#line 565 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 565 "View.MemoryView"
  int __pyx_clineno = 0;

#line 565 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":566
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 566 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 566 "View.MemoryView"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 566 "View.MemoryView"
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);

#line 566 "View.MemoryView"
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {

#line 566 "View.MemoryView"
    __pyx_t_2 = __pyx_t_4;

#line 566 "View.MemoryView"
    __pyx_v_length = (__pyx_t_2[0]);

#line 566 "View.MemoryView"
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 566 "View.MemoryView"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(2, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 566 "View.MemoryView"
  }

#line 566 "View.MemoryView"
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_5);

#line 566 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 566 "View.MemoryView"
  __pyx_r = __pyx_t_5;

#line 566 "View.MemoryView"
  __pyx_t_5 = 0;

#line 566 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":565
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
  /* function exit code */

#line 565 "View.MemoryView"
  __pyx_L1_error:;

#line 565 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 565 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 565 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 565 "View.MemoryView"
  __pyx_r = NULL;

#line 565 "View.MemoryView"
  __pyx_L0:;

#line 565 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 565 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 565 "View.MemoryView"
  return __pyx_r;

#line 565 "View.MemoryView"
}

/* "View.MemoryView":569
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
/* Python wrapper */

#line 569 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 569 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {

#line 569 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 569 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 569 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 569 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
  /* function exit code */

#line 569 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 569 "View.MemoryView"
  return __pyx_r;

#line 569 "View.MemoryView"
}

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride
#line 569 "View.MemoryView"
;

#line 569 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 569 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 569 "View.MemoryView"
  int __pyx_t_1;

#line 569 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 569 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 569 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 569 "View.MemoryView"
  Py_ssize_t *__pyx_t_5;

#line 569 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 569 "View.MemoryView"
  int __pyx_lineno = 0;

#line 569 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 569 "View.MemoryView"
  int __pyx_clineno = 0;

#line 569 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":570
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */

#line 570 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);

#line 570 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":572
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */

#line 572 "View.MemoryView"
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 572, __pyx_L1_error)

#line 572 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 572 "View.MemoryView"
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 572 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 572 "View.MemoryView"
    __PYX_ERR(2, 572, __pyx_L1_error)

    /* "View.MemoryView":570
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */

#line 570 "View.MemoryView"
  }

  /* "View.MemoryView":574
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 574 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 574 "View.MemoryView"
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 574 "View.MemoryView"
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);

#line 574 "View.MemoryView"
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {

#line 574 "View.MemoryView"
    __pyx_t_3 = __pyx_t_5;

#line 574 "View.MemoryView"
    __pyx_v_stride = (__pyx_t_3[0]);

#line 574 "View.MemoryView"
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 574 "View.MemoryView"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(2, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 574 "View.MemoryView"
  }

#line 574 "View.MemoryView"
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_6);

#line 574 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 574 "View.MemoryView"
  __pyx_r = __pyx_t_6;

#line 574 "View.MemoryView"
  __pyx_t_6 = 0;

#line 574 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":569
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
  /* function exit code */

#line 569 "View.MemoryView"
  __pyx_L1_error:;

#line 569 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 569 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 569 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 569 "View.MemoryView"
  __pyx_r = NULL;

#line 569 "View.MemoryView"
  __pyx_L0:;

#line 569 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 569 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 569 "View.MemoryView"
  return __pyx_r;

#line 569 "View.MemoryView"
}

/* "View.MemoryView":577
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
/* Python wrapper */

#line 577 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 577 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {

#line 577 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 577 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 577 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 577 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
  /* function exit code */

#line 577 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 577 "View.MemoryView"
  return __pyx_r;

#line 577 "View.MemoryView"
}

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset
#line 577 "View.MemoryView"
;

#line 577 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 577 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 577 "View.MemoryView"
  int __pyx_t_1;

#line 577 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 577 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 577 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 577 "View.MemoryView"
  Py_ssize_t *__pyx_t_5;

#line 577 "View.MemoryView"
  Py_ssize_t *__pyx_t_6;

#line 577 "View.MemoryView"
  int __pyx_lineno = 0;

#line 577 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 577 "View.MemoryView"
  int __pyx_clineno = 0;

#line 577 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":578
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */

#line 578 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);

#line 578 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":579
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */

#line 579 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 579 "View.MemoryView"
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 579, __pyx_L1_error)

#line 579 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 579 "View.MemoryView"
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__19, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 579, __pyx_L1_error)

#line 579 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 579 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 579 "View.MemoryView"
    __pyx_r = __pyx_t_3;

#line 579 "View.MemoryView"
    __pyx_t_3 = 0;

#line 579 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":578
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */

#line 578 "View.MemoryView"
  }

  /* "View.MemoryView":581
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 581 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 581 "View.MemoryView"
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 581 "View.MemoryView"
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);

#line 581 "View.MemoryView"
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {

#line 581 "View.MemoryView"
    __pyx_t_4 = __pyx_t_6;

#line 581 "View.MemoryView"
    __pyx_v_suboffset = (__pyx_t_4[0]);

#line 581 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 581 "View.MemoryView"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(2, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 581 "View.MemoryView"
  }

#line 581 "View.MemoryView"
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 581 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 581 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 581 "View.MemoryView"
  __pyx_t_2 = 0;

#line 581 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":577
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
  /* function exit code */

#line 577 "View.MemoryView"
  __pyx_L1_error:;

#line 577 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 577 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 577 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 577 "View.MemoryView"
  __pyx_r = NULL;

#line 577 "View.MemoryView"
  __pyx_L0:;

#line 577 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 577 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 577 "View.MemoryView"
  return __pyx_r;

#line 577 "View.MemoryView"
}

/* "View.MemoryView":584
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
/* Python wrapper */

#line 584 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 584 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {

#line 584 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 584 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 584 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 584 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
  /* function exit code */

#line 584 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 584 "View.MemoryView"
  return __pyx_r;

#line 584 "View.MemoryView"
}

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 584 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 584 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 584 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 584 "View.MemoryView"
  int __pyx_lineno = 0;

#line 584 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 584 "View.MemoryView"
  int __pyx_clineno = 0;

#line 584 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":585
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 585 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 585 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 585, __pyx_L1_error)

#line 585 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 585 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 585 "View.MemoryView"
  __pyx_t_1 = 0;

#line 585 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":584
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
  /* function exit code */

#line 584 "View.MemoryView"
  __pyx_L1_error:;

#line 584 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 584 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 584 "View.MemoryView"
  __pyx_r = NULL;

#line 584 "View.MemoryView"
  __pyx_L0:;

#line 584 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 584 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 584 "View.MemoryView"
  return __pyx_r;

#line 584 "View.MemoryView"
}

/* "View.MemoryView":588
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
/* Python wrapper */

#line 588 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 588 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {

#line 588 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 588 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 588 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 588 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
  /* function exit code */

#line 588 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 588 "View.MemoryView"
  return __pyx_r;

#line 588 "View.MemoryView"
}

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 588 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 588 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 588 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 588 "View.MemoryView"
  int __pyx_lineno = 0;

#line 588 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 588 "View.MemoryView"
  int __pyx_clineno = 0;

#line 588 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":589
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 589 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 589 "View.MemoryView"
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 589, __pyx_L1_error)

#line 589 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 589 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 589 "View.MemoryView"
  __pyx_t_1 = 0;

#line 589 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":588
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
  /* function exit code */

#line 588 "View.MemoryView"
  __pyx_L1_error:;

#line 588 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 588 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 588 "View.MemoryView"
  __pyx_r = NULL;

#line 588 "View.MemoryView"
  __pyx_L0:;

#line 588 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 588 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 588 "View.MemoryView"
  return __pyx_r;

#line 588 "View.MemoryView"
}

/* "View.MemoryView":592
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
/* Python wrapper */

#line 592 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 592 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {

#line 592 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 592 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 592 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 592 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
  /* function exit code */

#line 592 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 592 "View.MemoryView"
  return __pyx_r;

#line 592 "View.MemoryView"
}

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 592 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 592 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 592 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 592 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 592 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 592 "View.MemoryView"
  int __pyx_lineno = 0;

#line 592 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 592 "View.MemoryView"
  int __pyx_clineno = 0;

#line 592 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":593
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 593 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 593 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 593, __pyx_L1_error)

#line 593 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 593 "View.MemoryView"
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 593, __pyx_L1_error)

#line 593 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 593 "View.MemoryView"
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 593, __pyx_L1_error)

#line 593 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 593 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 593 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 593 "View.MemoryView"
  __pyx_r = __pyx_t_3;

#line 593 "View.MemoryView"
  __pyx_t_3 = 0;

#line 593 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":592
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
  /* function exit code */

#line 592 "View.MemoryView"
  __pyx_L1_error:;

#line 592 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 592 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 592 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 592 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 592 "View.MemoryView"
  __pyx_r = NULL;

#line 592 "View.MemoryView"
  __pyx_L0:;

#line 592 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 592 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 592 "View.MemoryView"
  return __pyx_r;

#line 592 "View.MemoryView"
}

/* "View.MemoryView":596
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
/* Python wrapper */

#line 596 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 596 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {

#line 596 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 596 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 596 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 596 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
  /* function exit code */

#line 596 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 596 "View.MemoryView"
  return __pyx_r;

#line 596 "View.MemoryView"
}

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL
#line 596 "View.MemoryView"
;
  PyObject *__pyx_v_length = NULL
#line 596 "View.MemoryView"
;

#line 596 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 596 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 596 "View.MemoryView"
  int __pyx_t_1;

#line 596 "View.MemoryView"
  int __pyx_t_2;

#line 596 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 596 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 596 "View.MemoryView"
  Py_ssize_t *__pyx_t_5;

#line 596 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 596 "View.MemoryView"
  int __pyx_lineno = 0;

#line 596 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 596 "View.MemoryView"
  int __pyx_clineno = 0;

#line 596 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":597
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */

#line 597 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);

#line 597 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 597 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":598
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */

#line 598 "View.MemoryView"
    __Pyx_INCREF(__pyx_int_1);

#line 598 "View.MemoryView"
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":600
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */

#line 600 "View.MemoryView"
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);

#line 600 "View.MemoryView"
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {

#line 600 "View.MemoryView"
      __pyx_t_3 = __pyx_t_5;

#line 600 "View.MemoryView"
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 600, __pyx_L1_error)

#line 600 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 600 "View.MemoryView"
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);

#line 600 "View.MemoryView"
      __pyx_t_6 = 0;

      /* "View.MemoryView":601
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */

#line 601 "View.MemoryView"
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 601, __pyx_L1_error)

#line 601 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 601 "View.MemoryView"
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);

#line 601 "View.MemoryView"
      __pyx_t_6 = 0;

#line 601 "View.MemoryView"
    }

    /* "View.MemoryView":603
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */

#line 603 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_result);

#line 603 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_result);

#line 603 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_self->_size);

#line 603 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_self->_size);

#line 603 "View.MemoryView"
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":597
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */

#line 597 "View.MemoryView"
  }

  /* "View.MemoryView":605
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 605 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 605 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->_size);

#line 605 "View.MemoryView"
  __pyx_r = __pyx_v_self->_size;

#line 605 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":596
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
  /* function exit code */

#line 596 "View.MemoryView"
  __pyx_L1_error:;

#line 596 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 596 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 596 "View.MemoryView"
  __pyx_r = NULL;

#line 596 "View.MemoryView"
  __pyx_L0:;

#line 596 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_result);

#line 596 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_length);

#line 596 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 596 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 596 "View.MemoryView"
  return __pyx_r;

#line 596 "View.MemoryView"
}

/* "View.MemoryView":607
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
/* Python wrapper */

#line 607 "View.MemoryView"
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/

#line 607 "View.MemoryView"
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {

#line 607 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 607 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 607 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 607 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
  /* function exit code */

#line 607 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 607 "View.MemoryView"
  return __pyx_r;

#line 607 "View.MemoryView"
}

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 607 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 607 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 607 "View.MemoryView"
  int __pyx_t_1;

#line 607 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":608
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */

#line 608 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);

#line 608 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":609
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */

#line 609 "View.MemoryView"
    __pyx_r = (__pyx_v_self->view.shape[0]);

#line 609 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":608
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */

#line 608 "View.MemoryView"
  }

  /* "View.MemoryView":611
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 611 "View.MemoryView"
  __pyx_r = 0;

#line 611 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":607
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
  /* function exit code */

#line 607 "View.MemoryView"
  __pyx_L0:;

#line 607 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 607 "View.MemoryView"
  return __pyx_r;

#line 607 "View.MemoryView"
}

/* "View.MemoryView":613
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
/* Python wrapper */

#line 613 "View.MemoryView"
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/

#line 613 "View.MemoryView"
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {

#line 613 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 613 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 613 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 613 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
  /* function exit code */

#line 613 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 613 "View.MemoryView"
  return __pyx_r;

#line 613 "View.MemoryView"
}

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 613 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 613 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 613 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 613 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 613 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 613 "View.MemoryView"
  int __pyx_lineno = 0;

#line 613 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 613 "View.MemoryView"
  int __pyx_clineno = 0;

#line 613 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":614
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */

#line 614 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 614 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 614 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":615
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */

#line 615 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 615, __pyx_L1_error)

#line 615 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":614
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */

#line 614 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 614 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 614 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 614 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 614 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);

#line 614 "View.MemoryView"
  __pyx_t_1 = 0;

#line 614 "View.MemoryView"
  __pyx_t_2 = 0;

#line 614 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 614 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 614 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 614 "View.MemoryView"
  __pyx_t_2 = 0;

#line 614 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":613
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
  /* function exit code */

#line 613 "View.MemoryView"
  __pyx_L1_error:;

#line 613 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 613 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 613 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 613 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 613 "View.MemoryView"
  __pyx_r = NULL;

#line 613 "View.MemoryView"
  __pyx_L0:;

#line 613 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 613 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 613 "View.MemoryView"
  return __pyx_r;

#line 613 "View.MemoryView"
}

/* "View.MemoryView":617
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
/* Python wrapper */

#line 617 "View.MemoryView"
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/

#line 617 "View.MemoryView"
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {

#line 617 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 617 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 617 "View.MemoryView"
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = 
#line 617 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
  /* function exit code */

#line 617 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 617 "View.MemoryView"
  return __pyx_r;

#line 617 "View.MemoryView"
}

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 617 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 617 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 617 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 617 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 617 "View.MemoryView"
  int __pyx_lineno = 0;

#line 617 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 617 "View.MemoryView"
  int __pyx_clineno = 0;

#line 617 "View.MemoryView"
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":618
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 618 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 618 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 618 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 618 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 618 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 618 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 618 "View.MemoryView"
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 618 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 618 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 618 "View.MemoryView"
  __pyx_t_1 = 0;

#line 618 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 618 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 618 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 618 "View.MemoryView"
  __pyx_t_1 = 0;

#line 618 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":617
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
  /* function exit code */

#line 617 "View.MemoryView"
  __pyx_L1_error:;

#line 617 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 617 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 617 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 617 "View.MemoryView"
  __pyx_r = NULL;

#line 617 "View.MemoryView"
  __pyx_L0:;

#line 617 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 617 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 617 "View.MemoryView"
  return __pyx_r;

#line 617 "View.MemoryView"
}

/* "View.MemoryView":621
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
/* Python wrapper */

#line 621 "View.MemoryView"
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 621 "View.MemoryView"
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 621 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 621 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 621 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = 
#line 621 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
  /* function exit code */

#line 621 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 621 "View.MemoryView"
  return __pyx_r;

#line 621 "View.MemoryView"
}

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice
#line 621 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp
#line 621 "View.MemoryView"
;

#line 621 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 621 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 621 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 621 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 621 "View.MemoryView"
  int __pyx_lineno = 0;

#line 621 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 621 "View.MemoryView"
  int __pyx_clineno = 0;

#line 621 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":624
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */

#line 624 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 624, __pyx_L1_error)

#line 624 "View.MemoryView"
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":625
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */

#line 625 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 625 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 625, __pyx_L1_error)

#line 625 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 625 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 625 "View.MemoryView"
  __pyx_t_2 = 0;

#line 625 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":621
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
  /* function exit code */

#line 621 "View.MemoryView"
  __pyx_L1_error:;

#line 621 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 621 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 621 "View.MemoryView"
  __pyx_r = NULL;

#line 621 "View.MemoryView"
  __pyx_L0:;

#line 621 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 621 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 621 "View.MemoryView"
  return __pyx_r;

#line 621 "View.MemoryView"
}

/* "View.MemoryView":627
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
/* Python wrapper */

#line 627 "View.MemoryView"
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 627 "View.MemoryView"
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 627 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 627 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 627 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = 
#line 627 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
  /* function exit code */

#line 627 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 627 "View.MemoryView"
  return __pyx_r;

#line 627 "View.MemoryView"
}

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice
#line 627 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp
#line 627 "View.MemoryView"
;

#line 627 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 627 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 627 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 627 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 627 "View.MemoryView"
  int __pyx_lineno = 0;

#line 627 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 627 "View.MemoryView"
  int __pyx_clineno = 0;

#line 627 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":630
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */

#line 630 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 630, __pyx_L1_error)

#line 630 "View.MemoryView"
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":631
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */

#line 631 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 631 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 631, __pyx_L1_error)

#line 631 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 631 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 631 "View.MemoryView"
  __pyx_t_2 = 0;

#line 631 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":627
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
  /* function exit code */

#line 627 "View.MemoryView"
  __pyx_L1_error:;

#line 627 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 627 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 627 "View.MemoryView"
  __pyx_r = NULL;

#line 627 "View.MemoryView"
  __pyx_L0:;

#line 627 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 627 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 627 "View.MemoryView"
  return __pyx_r;

#line 627 "View.MemoryView"
}

/* "View.MemoryView":633
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
/* Python wrapper */

#line 633 "View.MemoryView"
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 633 "View.MemoryView"
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 633 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 633 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 633 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = 
#line 633 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
  /* function exit code */

#line 633 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 633 "View.MemoryView"
  return __pyx_r;

#line 633 "View.MemoryView"
}

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice
#line 633 "View.MemoryView"
;
  int __pyx_v_flags
#line 633 "View.MemoryView"
;

#line 633 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 633 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 633 "View.MemoryView"
  __Pyx_memviewslice __pyx_t_1;

#line 633 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 633 "View.MemoryView"
  int __pyx_lineno = 0;

#line 633 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 633 "View.MemoryView"
  int __pyx_clineno = 0;

#line 633 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":635
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */

#line 635 "View.MemoryView"
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":637
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */

#line 637 "View.MemoryView"
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":638
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */

#line 638 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 638, __pyx_L1_error)

#line 638 "View.MemoryView"
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":643
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */

#line 643 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 643 "View.MemoryView"
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 643, __pyx_L1_error)

#line 643 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 643 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 643 "View.MemoryView"
  __pyx_t_2 = 0;

#line 643 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":633
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
  /* function exit code */

#line 633 "View.MemoryView"
  __pyx_L1_error:;

#line 633 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 633 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 633 "View.MemoryView"
  __pyx_r = NULL;

#line 633 "View.MemoryView"
  __pyx_L0:;

#line 633 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 633 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 633 "View.MemoryView"
  return __pyx_r;

#line 633 "View.MemoryView"
}

/* "View.MemoryView":645
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
/* Python wrapper */

#line 645 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 645 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 645 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 645 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 645 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = 
#line 645 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
  /* function exit code */

#line 645 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 645 "View.MemoryView"
  return __pyx_r;

#line 645 "View.MemoryView"
}

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src
#line 645 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_dst
#line 645 "View.MemoryView"
;
  int __pyx_v_flags
#line 645 "View.MemoryView"
;

#line 645 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 645 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 645 "View.MemoryView"
  __Pyx_memviewslice __pyx_t_1;

#line 645 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 645 "View.MemoryView"
  int __pyx_lineno = 0;

#line 645 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 645 "View.MemoryView"
  int __pyx_clineno = 0;

#line 645 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":647
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */

#line 647 "View.MemoryView"
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":649
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */

#line 649 "View.MemoryView"
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":650
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */

#line 650 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 650, __pyx_L1_error)

#line 650 "View.MemoryView"
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":655
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 655 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 655 "View.MemoryView"
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 655, __pyx_L1_error)

#line 655 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 655 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 655 "View.MemoryView"
  __pyx_t_2 = 0;

#line 655 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":645
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
  /* function exit code */

#line 645 "View.MemoryView"
  __pyx_L1_error:;

#line 645 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 645 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 645 "View.MemoryView"
  __pyx_r = NULL;

#line 645 "View.MemoryView"
  __pyx_L0:;

#line 645 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 645 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 645 "View.MemoryView"
  return __pyx_r;

#line 645 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 2 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 2 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 2 "(tree fragment)"
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
/* Python wrapper */

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 3 "(tree fragment)"
__pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 3 "(tree fragment)"
  int __pyx_lineno = 0;

#line 3 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 3 "(tree fragment)"
  int __pyx_clineno = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __pyx_L1_error:;

#line 3 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "(tree fragment)"
  __pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

/* "View.MemoryView":659
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

#line 659 "View.MemoryView"


#line 659 "View.MemoryView"
static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0
#line 659 "View.MemoryView"
;

#line 659 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 659 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 659 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 659 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 659 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 659 "View.MemoryView"
  int __pyx_lineno = 0;

#line 659 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 659 "View.MemoryView"
  int __pyx_clineno = 0;

#line 659 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":660
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */

#line 660 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 660 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 660 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 660 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_o);

#line 660 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_o);

#line 660 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);

#line 660 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 660 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);

#line 660 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 660 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 660 "View.MemoryView"
  __pyx_t_1 = 0;

#line 660 "View.MemoryView"
  __pyx_t_2 = 0;

#line 660 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 660 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 660 "View.MemoryView"
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);

#line 660 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":661
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */

#line 661 "View.MemoryView"
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":662
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */

#line 662 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 662 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 662 "View.MemoryView"
  __pyx_r = ((PyObject *)__pyx_v_result);

#line 662 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":659
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

#line 659 "View.MemoryView"


#line 659 "View.MemoryView"
  /* function exit code */

#line 659 "View.MemoryView"
  __pyx_L1_error:;

#line 659 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 659 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 659 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 659 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 659 "View.MemoryView"
  __pyx_r = 0;

#line 659 "View.MemoryView"
  __pyx_L0:;

#line 659 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 659 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 659 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 659 "View.MemoryView"
  return __pyx_r;

#line 659 "View.MemoryView"
}

/* "View.MemoryView":665
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

#line 665 "View.MemoryView"


#line 665 "View.MemoryView"
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {

#line 665 "View.MemoryView"
  int __pyx_r;

#line 665 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 665 "View.MemoryView"
  int __pyx_t_1;

#line 665 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":666
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */

#line 666 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 

#line 666 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 666 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":665
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

#line 665 "View.MemoryView"


#line 665 "View.MemoryView"
  /* function exit code */

#line 665 "View.MemoryView"
  __pyx_L0:;

#line 665 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 665 "View.MemoryView"
  return __pyx_r;

#line 665 "View.MemoryView"
}

/* "View.MemoryView":668
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

#line 668 "View.MemoryView"


#line 668 "View.MemoryView"
static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL
#line 668 "View.MemoryView"
;
  PyObject *__pyx_v_result = NULL
#line 668 "View.MemoryView"
;
  int __pyx_v_have_slices
#line 668 "View.MemoryView"
;
  int __pyx_v_seen_ellipsis
#line 668 "View.MemoryView"
;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL
#line 668 "View.MemoryView"
;
  PyObject *__pyx_v_item = NULL
#line 668 "View.MemoryView"
;
  Py_ssize_t __pyx_v_nslices
#line 668 "View.MemoryView"
;

#line 668 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 668 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 668 "View.MemoryView"
  int __pyx_t_1;

#line 668 "View.MemoryView"
  int __pyx_t_2;

#line 668 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 668 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 668 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 668 "View.MemoryView"
  PyObject *(*__pyx_t_6)(PyObject *);

#line 668 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 668 "View.MemoryView"
  Py_ssize_t __pyx_t_8;

#line 668 "View.MemoryView"
  int __pyx_t_9;

#line 668 "View.MemoryView"
  int __pyx_t_10;

#line 668 "View.MemoryView"
  PyObject *__pyx_t_11 = NULL;

#line 668 "View.MemoryView"
  int __pyx_lineno = 0;

#line 668 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 668 "View.MemoryView"
  int __pyx_clineno = 0;

#line 668 "View.MemoryView"
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":673
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */

#line 673 "View.MemoryView"
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 

#line 673 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);

#line 673 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":674
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */

#line 674 "View.MemoryView"
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 674, __pyx_L1_error)

#line 674 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 674 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_index);

#line 674 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_index);

#line 674 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);

#line 674 "View.MemoryView"
    __pyx_v_tup = __pyx_t_3;

#line 674 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":673
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */

#line 673 "View.MemoryView"
    goto __pyx_L3;

#line 673 "View.MemoryView"
  }

  /* "View.MemoryView":676
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */

#line 676 "View.MemoryView"
  /*else*/ {

#line 676 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_index);

#line 676 "View.MemoryView"
    __pyx_v_tup = __pyx_v_index;

#line 676 "View.MemoryView"
  }

#line 676 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":678
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */

#line 678 "View.MemoryView"
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 678, __pyx_L1_error)

#line 678 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 678 "View.MemoryView"
  __pyx_v_result = ((PyObject*)__pyx_t_3);

#line 678 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":679
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */

#line 679 "View.MemoryView"
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":680
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */

#line 680 "View.MemoryView"
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":681
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */

#line 681 "View.MemoryView"
  __Pyx_INCREF(__pyx_int_0);

#line 681 "View.MemoryView"
  __pyx_t_3 = __pyx_int_0;

#line 681 "View.MemoryView"
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {

#line 681 "View.MemoryView"
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;

#line 681 "View.MemoryView"
    __pyx_t_6 = NULL;

#line 681 "View.MemoryView"
  } else {
    __pyx_t_5 = -1; 
#line 681 "View.MemoryView"
__pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 681 "View.MemoryView"
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
  }

#line 681 "View.MemoryView"
  for (;;) {

#line 681 "View.MemoryView"
    if (likely(!__pyx_t_6)) {

#line 681 "View.MemoryView"
      if (likely(PyList_CheckExact(__pyx_t_4))) {

#line 681 "View.MemoryView"
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;

#line 681 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 681 "View.MemoryView"
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        #else

#line 681 "View.MemoryView"
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
        #endif

#line 681 "View.MemoryView"
      } else {

#line 681 "View.MemoryView"
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;

#line 681 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 681 "View.MemoryView"
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        #else

#line 681 "View.MemoryView"
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
        #endif

#line 681 "View.MemoryView"
      }
    } else 
#line 681 "View.MemoryView"
{

#line 681 "View.MemoryView"
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);

#line 681 "View.MemoryView"
      if (unlikely(!__pyx_t_7)) {

#line 681 "View.MemoryView"
        PyObject* exc_type = PyErr_Occurred();

#line 681 "View.MemoryView"
        if (exc_type) {

#line 681 "View.MemoryView"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 681 "View.MemoryView"
          else __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        }

#line 681 "View.MemoryView"
        break;

#line 681 "View.MemoryView"
      }

#line 681 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
    }

#line 681 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);

#line 681 "View.MemoryView"
    __pyx_t_7 = 0;

#line 681 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 681 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);

#line 681 "View.MemoryView"
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3);

#line 681 "View.MemoryView"
    __pyx_t_3 = __pyx_t_7;

#line 681 "View.MemoryView"
    __pyx_t_7 = 0;

    /* "View.MemoryView":682
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */

#line 682 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);

#line 682 "View.MemoryView"
    __pyx_t_1 = (__pyx_t_2 != 0);

#line 682 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":683
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */

#line 683 "View.MemoryView"
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);

#line 683 "View.MemoryView"
      if (__pyx_t_1) {

        /* "View.MemoryView":684
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */

#line 684 "View.MemoryView"
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(2, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 684 "View.MemoryView"
        { Py_ssize_t __pyx_temp;

#line 684 "View.MemoryView"
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {

#line 684 "View.MemoryView"
            __Pyx_INCREF(__pyx_slice__22);

#line 684 "View.MemoryView"
            __Pyx_GIVEREF(__pyx_slice__22);

#line 684 "View.MemoryView"
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__22);

#line 684 "View.MemoryView"
          }

#line 684 "View.MemoryView"
        }

#line 684 "View.MemoryView"
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":685
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */

#line 685 "View.MemoryView"
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":683
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */

#line 683 "View.MemoryView"
        goto __pyx_L7;

#line 683 "View.MemoryView"
      }

      /* "View.MemoryView":687
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */

#line 687 "View.MemoryView"
      /*else*/ {

#line 687 "View.MemoryView"
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__22); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 687, __pyx_L1_error)

#line 687 "View.MemoryView"
      }

#line 687 "View.MemoryView"
      __pyx_L7:;

      /* "View.MemoryView":688
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */

#line 688 "View.MemoryView"
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":682
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */

#line 682 "View.MemoryView"
      goto __pyx_L6;

#line 682 "View.MemoryView"
    }

    /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */

#line 690 "View.MemoryView"
    /*else*/ {

#line 690 "View.MemoryView"
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 

#line 690 "View.MemoryView"
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);

#line 690 "View.MemoryView"
      if (__pyx_t_10) {

#line 690 "View.MemoryView"
      } else {

#line 690 "View.MemoryView"
        __pyx_t_1 = __pyx_t_10;

#line 690 "View.MemoryView"
        goto __pyx_L9_bool_binop_done;

#line 690 "View.MemoryView"
      }

#line 690 "View.MemoryView"
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);

#line 690 "View.MemoryView"
      __pyx_t_1 = __pyx_t_10;

#line 690 "View.MemoryView"
      __pyx_L9_bool_binop_done:;

#line 690 "View.MemoryView"
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":691
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */

#line 691 "View.MemoryView"
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 691, __pyx_L1_error)

#line 691 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 691 "View.MemoryView"
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 691, __pyx_L1_error)

#line 691 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_11);

#line 691 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 691 "View.MemoryView"
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);

#line 691 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 691 "View.MemoryView"
        __PYX_ERR(2, 691, __pyx_L1_error)

        /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */

#line 690 "View.MemoryView"
      }

      /* "View.MemoryView":693
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */

#line 693 "View.MemoryView"
      __pyx_t_10 = (__pyx_v_have_slices != 0);

#line 693 "View.MemoryView"
      if (!__pyx_t_10) {

#line 693 "View.MemoryView"
      } else {

#line 693 "View.MemoryView"
        __pyx_t_1 = __pyx_t_10;

#line 693 "View.MemoryView"
        goto __pyx_L11_bool_binop_done;

#line 693 "View.MemoryView"
      }

#line 693 "View.MemoryView"
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 

#line 693 "View.MemoryView"
      __pyx_t_2 = (__pyx_t_10 != 0);

#line 693 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 693 "View.MemoryView"
      __pyx_L11_bool_binop_done:;

#line 693 "View.MemoryView"
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":694
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */

#line 694 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 694, __pyx_L1_error)

#line 694 "View.MemoryView"
    }

#line 694 "View.MemoryView"
    __pyx_L6:;

    /* "View.MemoryView":681
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */

#line 681 "View.MemoryView"
  }

#line 681 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 681 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":696
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */

#line 696 "View.MemoryView"
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 696, __pyx_L1_error)

#line 696 "View.MemoryView"
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":697
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */

#line 697 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_nslices != 0);

#line 697 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":698
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */

#line 698 "View.MemoryView"
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 698, __pyx_L1_error)

#line 698 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 698 "View.MemoryView"
    { Py_ssize_t __pyx_temp;

#line 698 "View.MemoryView"
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {

#line 698 "View.MemoryView"
        __Pyx_INCREF(__pyx_slice__22);

#line 698 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_slice__22);

#line 698 "View.MemoryView"
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__22);

#line 698 "View.MemoryView"
      }

#line 698 "View.MemoryView"
    }

#line 698 "View.MemoryView"
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 698, __pyx_L1_error)

#line 698 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":697
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */

#line 697 "View.MemoryView"
  }

  /* "View.MemoryView":700
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */

#line 700 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 700 "View.MemoryView"
  if (!__pyx_v_have_slices) {

#line 700 "View.MemoryView"
  } else {

#line 700 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 700 "View.MemoryView"
    __pyx_t_3 = __pyx_t_4;

#line 700 "View.MemoryView"
    __pyx_t_4 = 0;

#line 700 "View.MemoryView"
    goto __pyx_L14_bool_binop_done;

#line 700 "View.MemoryView"
  }

#line 700 "View.MemoryView"
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 700 "View.MemoryView"
  __pyx_t_3 = __pyx_t_4;

#line 700 "View.MemoryView"
  __pyx_t_4 = 0;

#line 700 "View.MemoryView"
  __pyx_L14_bool_binop_done:;

#line 700 "View.MemoryView"
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 700 "View.MemoryView"
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_11);

#line 700 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 700 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);

#line 700 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_4);

#line 700 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);

#line 700 "View.MemoryView"
  __pyx_t_3 = 0;

#line 700 "View.MemoryView"
  __pyx_t_4 = 0;

#line 700 "View.MemoryView"
  __pyx_r = ((PyObject*)__pyx_t_11);

#line 700 "View.MemoryView"
  __pyx_t_11 = 0;

#line 700 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":668
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

#line 668 "View.MemoryView"


#line 668 "View.MemoryView"
  /* function exit code */

#line 668 "View.MemoryView"
  __pyx_L1_error:;

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_7);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_11);

#line 668 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 668 "View.MemoryView"
  __pyx_r = 0;

#line 668 "View.MemoryView"
  __pyx_L0:;

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_tup);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_result);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_idx);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_item);

#line 668 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 668 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 668 "View.MemoryView"
  return __pyx_r;

#line 668 "View.MemoryView"
}

/* "View.MemoryView":702
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

#line 702 "View.MemoryView"


#line 702 "View.MemoryView"
static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset
#line 702 "View.MemoryView"
;

#line 702 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 702 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 702 "View.MemoryView"
  Py_ssize_t *__pyx_t_1;

#line 702 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 702 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 702 "View.MemoryView"
  int __pyx_t_4;

#line 702 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 702 "View.MemoryView"
  int __pyx_lineno = 0;

#line 702 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 702 "View.MemoryView"
  int __pyx_clineno = 0;

#line 702 "View.MemoryView"
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":703
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */

#line 703 "View.MemoryView"
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);

#line 703 "View.MemoryView"
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {

#line 703 "View.MemoryView"
    __pyx_t_1 = __pyx_t_3;

#line 703 "View.MemoryView"
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":704
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */

#line 704 "View.MemoryView"
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);

#line 704 "View.MemoryView"
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":705
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 705 "View.MemoryView"
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 705, __pyx_L1_error)

#line 705 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 705 "View.MemoryView"
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);

#line 705 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 705 "View.MemoryView"
      __PYX_ERR(2, 705, __pyx_L1_error)

      /* "View.MemoryView":704
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */

#line 704 "View.MemoryView"
    }

#line 704 "View.MemoryView"
  }

  /* "View.MemoryView":702
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

#line 702 "View.MemoryView"


#line 702 "View.MemoryView"
  /* function exit code */

#line 702 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 702 "View.MemoryView"
  goto __pyx_L0;

#line 702 "View.MemoryView"
  __pyx_L1_error:;

#line 702 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 702 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 702 "View.MemoryView"
  __pyx_r = 0;

#line 702 "View.MemoryView"
  __pyx_L0:;

#line 702 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 702 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 702 "View.MemoryView"
  return __pyx_r;

#line 702 "View.MemoryView"
}

/* "View.MemoryView":712
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

#line 712 "View.MemoryView"


#line 712 "View.MemoryView"
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim
#line 712 "View.MemoryView"
;
  int __pyx_v_suboffset_dim
#line 712 "View.MemoryView"
;
  int __pyx_v_dim
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_src
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_dst
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice *__pyx_v_p_src
#line 712 "View.MemoryView"
;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice *__pyx_v_p_dst
#line 712 "View.MemoryView"
;
  int *__pyx_v_p_suboffset_dim
#line 712 "View.MemoryView"
;
  Py_ssize_t __pyx_v_start
#line 712 "View.MemoryView"
;
  Py_ssize_t __pyx_v_stop
#line 712 "View.MemoryView"
;
  Py_ssize_t __pyx_v_step
#line 712 "View.MemoryView"
;
  int __pyx_v_have_start
#line 712 "View.MemoryView"
;
  int __pyx_v_have_stop
#line 712 "View.MemoryView"
;
  int __pyx_v_have_step
#line 712 "View.MemoryView"
;
  PyObject *__pyx_v_index = NULL
#line 712 "View.MemoryView"
;

#line 712 "View.MemoryView"
  struct __pyx_memoryview_obj *__pyx_r = NULL;

#line 712 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 712 "View.MemoryView"
  int __pyx_t_1;

#line 712 "View.MemoryView"
  int __pyx_t_2;

#line 712 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 712 "View.MemoryView"
  struct __pyx_memoryview_obj *__pyx_t_4;

#line 712 "View.MemoryView"
  char *__pyx_t_5;

#line 712 "View.MemoryView"
  int __pyx_t_6;

#line 712 "View.MemoryView"
  Py_ssize_t __pyx_t_7;

#line 712 "View.MemoryView"
  PyObject *(*__pyx_t_8)(PyObject *);

#line 712 "View.MemoryView"
  PyObject *__pyx_t_9 = NULL;

#line 712 "View.MemoryView"
  Py_ssize_t __pyx_t_10;

#line 712 "View.MemoryView"
  int __pyx_t_11;

#line 712 "View.MemoryView"
  Py_ssize_t __pyx_t_12;

#line 712 "View.MemoryView"
  int __pyx_lineno = 0;

#line 712 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 712 "View.MemoryView"
  int __pyx_clineno = 0;

#line 712 "View.MemoryView"
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":713
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */

#line 713 "View.MemoryView"
  __pyx_v_new_ndim = 0;

#line 713 "View.MemoryView"
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":720
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */

#line 720 "View.MemoryView"
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":724
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */

#line 724 "View.MemoryView"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 724 "View.MemoryView"
  if (unlikely(!Py_OptimizeFlag)) {

#line 724 "View.MemoryView"
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {

#line 724 "View.MemoryView"
      PyErr_SetNone(PyExc_AssertionError);

#line 724 "View.MemoryView"
      __PYX_ERR(2, 724, __pyx_L1_error)

#line 724 "View.MemoryView"
    }

#line 724 "View.MemoryView"
  }

#line 724 "View.MemoryView"
  #endif

  /* "View.MemoryView":726
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */

#line 726 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 726 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 726 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":727
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */

#line 727 "View.MemoryView"
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(2, 727, __pyx_L1_error)

#line 727 "View.MemoryView"
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);

#line 727 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 727 "View.MemoryView"
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);

#line 727 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":728
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */

#line 728 "View.MemoryView"
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":726
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */

#line 726 "View.MemoryView"
    goto __pyx_L3;

#line 726 "View.MemoryView"
  }

  /* "View.MemoryView":730
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */

#line 730 "View.MemoryView"
  /*else*/ {

#line 730 "View.MemoryView"
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":731
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 731 "View.MemoryView"
    __pyx_v_p_src = (&__pyx_v_src);

#line 731 "View.MemoryView"
  }

#line 731 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":737
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */

#line 737 "View.MemoryView"
  __pyx_t_4 = __pyx_v_p_src->memview;

#line 737 "View.MemoryView"
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":738
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 738 "View.MemoryView"
  __pyx_t_5 = __pyx_v_p_src->data;

#line 738 "View.MemoryView"
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":743
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */

#line 743 "View.MemoryView"
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":744
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */

#line 744 "View.MemoryView"
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":748
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */

#line 748 "View.MemoryView"
  __pyx_t_6 = 0;

#line 748 "View.MemoryView"
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {

#line 748 "View.MemoryView"
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;

#line 748 "View.MemoryView"
    __pyx_t_8 = NULL;

#line 748 "View.MemoryView"
  } else {
    __pyx_t_7 = -1; 
#line 748 "View.MemoryView"
__pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 748 "View.MemoryView"
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
  }

#line 748 "View.MemoryView"
  for (;;) {

#line 748 "View.MemoryView"
    if (likely(!__pyx_t_8)) {

#line 748 "View.MemoryView"
      if (likely(PyList_CheckExact(__pyx_t_3))) {

#line 748 "View.MemoryView"
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;

#line 748 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 748 "View.MemoryView"
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        #else

#line 748 "View.MemoryView"
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_9);

#line 748 "View.MemoryView"
        #endif

#line 748 "View.MemoryView"
      } else {

#line 748 "View.MemoryView"
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;

#line 748 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 748 "View.MemoryView"
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        #else

#line 748 "View.MemoryView"
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_9);

#line 748 "View.MemoryView"
        #endif

#line 748 "View.MemoryView"
      }
    } else 
#line 748 "View.MemoryView"
{

#line 748 "View.MemoryView"
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);

#line 748 "View.MemoryView"
      if (unlikely(!__pyx_t_9)) {

#line 748 "View.MemoryView"
        PyObject* exc_type = PyErr_Occurred();

#line 748 "View.MemoryView"
        if (exc_type) {

#line 748 "View.MemoryView"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 748 "View.MemoryView"
          else __PYX_ERR(2, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        }

#line 748 "View.MemoryView"
        break;

#line 748 "View.MemoryView"
      }

#line 748 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 748 "View.MemoryView"
    }

#line 748 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);

#line 748 "View.MemoryView"
    __pyx_t_9 = 0;

#line 748 "View.MemoryView"
    __pyx_v_dim = __pyx_t_6;

#line 748 "View.MemoryView"
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":749
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */

#line 749 "View.MemoryView"
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);

#line 749 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":753
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */

#line 753 "View.MemoryView"
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 753, __pyx_L1_error)

      /* "View.MemoryView":750
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */

#line 750 "View.MemoryView"
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(2, 750, __pyx_L1_error)

      /* "View.MemoryView":749
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */

#line 749 "View.MemoryView"
      goto __pyx_L6;

#line 749 "View.MemoryView"
    }

    /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */

#line 756 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_index == Py_None);

#line 756 "View.MemoryView"
    __pyx_t_1 = (__pyx_t_2 != 0);

#line 756 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":757
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */

#line 757 "View.MemoryView"
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":758
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */

#line 758 "View.MemoryView"
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":759
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */

#line 759 "View.MemoryView"
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":760
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */

#line 760 "View.MemoryView"
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */

#line 756 "View.MemoryView"
      goto __pyx_L6;

#line 756 "View.MemoryView"
    }

    /* "View.MemoryView":762
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */

#line 762 "View.MemoryView"
    /*else*/ {

#line 762 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 762, __pyx_L1_error)

#line 762 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 762 "View.MemoryView"
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 762, __pyx_L1_error)

#line 762 "View.MemoryView"
      if (!__pyx_t_1) {

#line 762 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 762 "View.MemoryView"
      } else {

#line 762 "View.MemoryView"
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 762, __pyx_L1_error)

#line 762 "View.MemoryView"
        __pyx_t_10 = __pyx_t_12;

#line 762 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 762 "View.MemoryView"
        goto __pyx_L7_bool_binop_done;

#line 762 "View.MemoryView"
      }

#line 762 "View.MemoryView"
      __pyx_t_10 = 0;

#line 762 "View.MemoryView"
      __pyx_L7_bool_binop_done:;

#line 762 "View.MemoryView"
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":763
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */

#line 763 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 763, __pyx_L1_error)

#line 763 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 763 "View.MemoryView"
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 763, __pyx_L1_error)

#line 763 "View.MemoryView"
      if (!__pyx_t_1) {

#line 763 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 763 "View.MemoryView"
      } else {

#line 763 "View.MemoryView"
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 763, __pyx_L1_error)

#line 763 "View.MemoryView"
        __pyx_t_10 = __pyx_t_12;

#line 763 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 763 "View.MemoryView"
        goto __pyx_L9_bool_binop_done;

#line 763 "View.MemoryView"
      }

#line 763 "View.MemoryView"
      __pyx_t_10 = 0;

#line 763 "View.MemoryView"
      __pyx_L9_bool_binop_done:;

#line 763 "View.MemoryView"
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":764
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */

#line 764 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 764, __pyx_L1_error)

#line 764 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 764 "View.MemoryView"
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 764, __pyx_L1_error)

#line 764 "View.MemoryView"
      if (!__pyx_t_1) {

#line 764 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 764 "View.MemoryView"
      } else {

#line 764 "View.MemoryView"
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 764, __pyx_L1_error)

#line 764 "View.MemoryView"
        __pyx_t_10 = __pyx_t_12;

#line 764 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 764 "View.MemoryView"
        goto __pyx_L11_bool_binop_done;

#line 764 "View.MemoryView"
      }

#line 764 "View.MemoryView"
      __pyx_t_10 = 0;

#line 764 "View.MemoryView"
      __pyx_L11_bool_binop_done:;

#line 764 "View.MemoryView"
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":766
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */

#line 766 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 766, __pyx_L1_error)

#line 766 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 766 "View.MemoryView"
      __pyx_t_1 = (__pyx_t_9 != Py_None);

#line 766 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 766 "View.MemoryView"
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":767
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */

#line 767 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 767, __pyx_L1_error)

#line 767 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 767 "View.MemoryView"
      __pyx_t_1 = (__pyx_t_9 != Py_None);

#line 767 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 767 "View.MemoryView"
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */

#line 768 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 768, __pyx_L1_error)

#line 768 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 768 "View.MemoryView"
      __pyx_t_1 = (__pyx_t_9 != Py_None);

#line 768 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 768 "View.MemoryView"
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":770
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */

#line 770 "View.MemoryView"
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(2, 770, __pyx_L1_error)

      /* "View.MemoryView":776
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */

#line 776 "View.MemoryView"
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

#line 776 "View.MemoryView"
    }

#line 776 "View.MemoryView"
    __pyx_L6:;

    /* "View.MemoryView":748
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */

#line 748 "View.MemoryView"
  }

#line 748 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */

#line 778 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 778 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 778 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */

#line 779 "View.MemoryView"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":780
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */

#line 780 "View.MemoryView"
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(2, 780, __pyx_L1_error) }

    /* "View.MemoryView":781
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */

#line 781 "View.MemoryView"
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(2, 781, __pyx_L1_error) }

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */

#line 779 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 779, __pyx_L1_error)

#line 779 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 779 "View.MemoryView"
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(2, 779, __pyx_L1_error)

#line 779 "View.MemoryView"
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);

#line 779 "View.MemoryView"
    __pyx_t_3 = 0;

#line 779 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */

#line 778 "View.MemoryView"
  }

  /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */

#line 784 "View.MemoryView"
  /*else*/ {

#line 784 "View.MemoryView"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":785
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 785 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 784, __pyx_L1_error)

#line 785 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */

#line 784 "View.MemoryView"
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(2, 784, __pyx_L1_error)

#line 784 "View.MemoryView"
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);

#line 784 "View.MemoryView"
    __pyx_t_3 = 0;

#line 784 "View.MemoryView"
    goto __pyx_L0;

#line 784 "View.MemoryView"
  }

  /* "View.MemoryView":712
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

#line 712 "View.MemoryView"


#line 712 "View.MemoryView"
  /* function exit code */

#line 712 "View.MemoryView"
  __pyx_L1_error:;

#line 712 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 712 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_9);

#line 712 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 712 "View.MemoryView"
  __pyx_r = 0;

#line 712 "View.MemoryView"
  __pyx_L0:;

#line 712 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);

#line 712 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_index);

#line 712 "View.MemoryView"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 712 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 712 "View.MemoryView"
  return __pyx_r;

#line 712 "View.MemoryView"
}

/* "View.MemoryView":809
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

#line 809 "View.MemoryView"


#line 809 "View.MemoryView"
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape
#line 809 "View.MemoryView"
;
  int __pyx_v_negative_step
#line 809 "View.MemoryView"
;

#line 809 "View.MemoryView"
  int __pyx_r;

#line 809 "View.MemoryView"
  int __pyx_t_1;

#line 809 "View.MemoryView"
  int __pyx_t_2;

#line 809 "View.MemoryView"
  int __pyx_t_3;

#line 809 "View.MemoryView"
  int __pyx_lineno = 0;

#line 809 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 809 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":829
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */

#line 829 "View.MemoryView"
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);

#line 829 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":831
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */

#line 831 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);

#line 831 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":832
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */

#line 832 "View.MemoryView"
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":831
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */

#line 831 "View.MemoryView"
    }

    /* "View.MemoryView":833
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */

#line 833 "View.MemoryView"
    __pyx_t_1 = (0 <= __pyx_v_start);

#line 833 "View.MemoryView"
    if (__pyx_t_1) {

#line 833 "View.MemoryView"
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);

#line 833 "View.MemoryView"
    }

#line 833 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);

#line 833 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":834
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */

#line 834 "View.MemoryView"
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 834, __pyx_L1_error)

      /* "View.MemoryView":833
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */

#line 833 "View.MemoryView"
    }

    /* "View.MemoryView":829
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */

#line 829 "View.MemoryView"
    goto __pyx_L3;

#line 829 "View.MemoryView"
  }

  /* "View.MemoryView":837
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */

#line 837 "View.MemoryView"
  /*else*/ {

#line 837 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);

#line 837 "View.MemoryView"
    if (__pyx_t_1) {

#line 837 "View.MemoryView"
    } else {

#line 837 "View.MemoryView"
      __pyx_t_2 = __pyx_t_1;

#line 837 "View.MemoryView"
      goto __pyx_L6_bool_binop_done;

#line 837 "View.MemoryView"
    }

#line 837 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);

#line 837 "View.MemoryView"
    __pyx_t_2 = __pyx_t_1;

#line 837 "View.MemoryView"
    __pyx_L6_bool_binop_done:;

#line 837 "View.MemoryView"
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":839
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */

#line 839 "View.MemoryView"
    __pyx_t_1 = (__pyx_v_have_step != 0);

#line 839 "View.MemoryView"
    if (__pyx_t_1) {

#line 839 "View.MemoryView"
    } else {

#line 839 "View.MemoryView"
      __pyx_t_2 = __pyx_t_1;

#line 839 "View.MemoryView"
      goto __pyx_L9_bool_binop_done;

#line 839 "View.MemoryView"
    }

#line 839 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);

#line 839 "View.MemoryView"
    __pyx_t_2 = __pyx_t_1;

#line 839 "View.MemoryView"
    __pyx_L9_bool_binop_done:;

#line 839 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":840
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 840 "View.MemoryView"
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 840, __pyx_L1_error)

      /* "View.MemoryView":839
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */

#line 839 "View.MemoryView"
    }

    /* "View.MemoryView":843
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */

#line 843 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_have_start != 0);

#line 843 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":844
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */

#line 844 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);

#line 844 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":845
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */

#line 845 "View.MemoryView"
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":846
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */

#line 846 "View.MemoryView"
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);

#line 846 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":847
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */

#line 847 "View.MemoryView"
          __pyx_v_start = 0;

          /* "View.MemoryView":846
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */

#line 846 "View.MemoryView"
        }

        /* "View.MemoryView":844
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */

#line 844 "View.MemoryView"
        goto __pyx_L12;

#line 844 "View.MemoryView"
      }

      /* "View.MemoryView":848
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */

#line 848 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);

#line 848 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":849
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */

#line 849 "View.MemoryView"
        __pyx_t_2 = (__pyx_v_negative_step != 0);

#line 849 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":850
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */

#line 850 "View.MemoryView"
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":849
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */

#line 849 "View.MemoryView"
          goto __pyx_L14;

#line 849 "View.MemoryView"
        }

        /* "View.MemoryView":852
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */

#line 852 "View.MemoryView"
        /*else*/ {

#line 852 "View.MemoryView"
          __pyx_v_start = __pyx_v_shape;

#line 852 "View.MemoryView"
        }

#line 852 "View.MemoryView"
        __pyx_L14:;

        /* "View.MemoryView":848
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */

#line 848 "View.MemoryView"
      }

#line 848 "View.MemoryView"
      __pyx_L12:;

      /* "View.MemoryView":843
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */

#line 843 "View.MemoryView"
      goto __pyx_L11;

#line 843 "View.MemoryView"
    }

    /* "View.MemoryView":854
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */

#line 854 "View.MemoryView"
    /*else*/ {

#line 854 "View.MemoryView"
      __pyx_t_2 = (__pyx_v_negative_step != 0);

#line 854 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":855
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */

#line 855 "View.MemoryView"
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":854
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */

#line 854 "View.MemoryView"
        goto __pyx_L15;

#line 854 "View.MemoryView"
      }

      /* "View.MemoryView":857
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */

#line 857 "View.MemoryView"
      /*else*/ {

#line 857 "View.MemoryView"
        __pyx_v_start = 0;

#line 857 "View.MemoryView"
      }

#line 857 "View.MemoryView"
      __pyx_L15:;

#line 857 "View.MemoryView"
    }

#line 857 "View.MemoryView"
    __pyx_L11:;

    /* "View.MemoryView":859
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */

#line 859 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_have_stop != 0);

#line 859 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":860
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */

#line 860 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);

#line 860 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":861
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */

#line 861 "View.MemoryView"
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":862
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */

#line 862 "View.MemoryView"
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);

#line 862 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":863
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */

#line 863 "View.MemoryView"
          __pyx_v_stop = 0;

          /* "View.MemoryView":862
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */

#line 862 "View.MemoryView"
        }

        /* "View.MemoryView":860
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */

#line 860 "View.MemoryView"
        goto __pyx_L17;

#line 860 "View.MemoryView"
      }

      /* "View.MemoryView":864
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */

#line 864 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);

#line 864 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":865
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */

#line 865 "View.MemoryView"
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":864
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */

#line 864 "View.MemoryView"
      }

#line 864 "View.MemoryView"
      __pyx_L17:;

      /* "View.MemoryView":859
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */

#line 859 "View.MemoryView"
      goto __pyx_L16;

#line 859 "View.MemoryView"
    }

    /* "View.MemoryView":867
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */

#line 867 "View.MemoryView"
    /*else*/ {

#line 867 "View.MemoryView"
      __pyx_t_2 = (__pyx_v_negative_step != 0);

#line 867 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":868
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */

#line 868 "View.MemoryView"
        __pyx_v_stop = -1L;

        /* "View.MemoryView":867
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */

#line 867 "View.MemoryView"
        goto __pyx_L19;

#line 867 "View.MemoryView"
      }

      /* "View.MemoryView":870
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */

#line 870 "View.MemoryView"
      /*else*/ {

#line 870 "View.MemoryView"
        __pyx_v_stop = __pyx_v_shape;

#line 870 "View.MemoryView"
      }

#line 870 "View.MemoryView"
      __pyx_L19:;

#line 870 "View.MemoryView"
    }

#line 870 "View.MemoryView"
    __pyx_L16:;

    /* "View.MemoryView":872
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */

#line 872 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);

#line 872 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":873
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 873 "View.MemoryView"
      __pyx_v_step = 1;

      /* "View.MemoryView":872
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */

#line 872 "View.MemoryView"
    }

    /* "View.MemoryView":877
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */

#line 877 "View.MemoryView"
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":879
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */

#line 879 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);

#line 879 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":880
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */

#line 880 "View.MemoryView"
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":879
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */

#line 879 "View.MemoryView"
    }

    /* "View.MemoryView":882
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */

#line 882 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);

#line 882 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":883
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 883 "View.MemoryView"
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":882
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */

#line 882 "View.MemoryView"
    }

    /* "View.MemoryView":886
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */

#line 886 "View.MemoryView"
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":887
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */

#line 887 "View.MemoryView"
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":888
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 888 "View.MemoryView"
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;

#line 888 "View.MemoryView"
  }

#line 888 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":891
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */

#line 891 "View.MemoryView"
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);

#line 891 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":892
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */

#line 892 "View.MemoryView"
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":891
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */

#line 891 "View.MemoryView"
    goto __pyx_L23;

#line 891 "View.MemoryView"
  }

  /* "View.MemoryView":894
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */

#line 894 "View.MemoryView"
  /*else*/ {

#line 894 "View.MemoryView"
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);

#line 894 "View.MemoryView"
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));

#line 894 "View.MemoryView"
  }

#line 894 "View.MemoryView"
  __pyx_L23:;

  /* "View.MemoryView":896
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */

#line 896 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);

#line 896 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":897
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */

#line 897 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);

#line 897 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":898
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */

#line 898 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);

#line 898 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":899
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */

#line 899 "View.MemoryView"
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":898
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */

#line 898 "View.MemoryView"
        goto __pyx_L26;

#line 898 "View.MemoryView"
      }

      /* "View.MemoryView":901
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */

#line 901 "View.MemoryView"
      /*else*/ {

        /* "View.MemoryView":902
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */

#line 902 "View.MemoryView"
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 901, __pyx_L1_error)

#line 902 "View.MemoryView"
      }

#line 902 "View.MemoryView"
      __pyx_L26:;

      /* "View.MemoryView":897
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */

#line 897 "View.MemoryView"
      goto __pyx_L25;

#line 897 "View.MemoryView"
    }

    /* "View.MemoryView":904
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */

#line 904 "View.MemoryView"
    /*else*/ {

#line 904 "View.MemoryView"
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;

#line 904 "View.MemoryView"
    }

#line 904 "View.MemoryView"
    __pyx_L25:;

    /* "View.MemoryView":896
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */

#line 896 "View.MemoryView"
  }

  /* "View.MemoryView":906
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 906 "View.MemoryView"
  __pyx_r = 0;

#line 906 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":809
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

#line 809 "View.MemoryView"


#line 809 "View.MemoryView"
  /* function exit code */

#line 809 "View.MemoryView"
  __pyx_L1_error:;

#line 809 "View.MemoryView"
  {

#line 809 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 809 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 809 "View.MemoryView"
    #endif

#line 809 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 809 "View.MemoryView"
    #ifdef WITH_THREAD

#line 809 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 809 "View.MemoryView"
    #endif

#line 809 "View.MemoryView"
  }

#line 809 "View.MemoryView"
  __pyx_r = -1;

#line 809 "View.MemoryView"
  __pyx_L0:;

#line 809 "View.MemoryView"
  return __pyx_r;

#line 809 "View.MemoryView"
}

/* "View.MemoryView":912
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

#line 912 "View.MemoryView"


#line 912 "View.MemoryView"
static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape
#line 912 "View.MemoryView"
;
  Py_ssize_t __pyx_v_stride
#line 912 "View.MemoryView"
;
  Py_ssize_t __pyx_v_suboffset
#line 912 "View.MemoryView"
;
  Py_ssize_t __pyx_v_itemsize
#line 912 "View.MemoryView"
;
  char *__pyx_v_resultp
#line 912 "View.MemoryView"
;

#line 912 "View.MemoryView"
  char *__pyx_r;

#line 912 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 912 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 912 "View.MemoryView"
  int __pyx_t_2;

#line 912 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 912 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 912 "View.MemoryView"
  int __pyx_lineno = 0;

#line 912 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 912 "View.MemoryView"
  int __pyx_clineno = 0;

#line 912 "View.MemoryView"
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":914
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */

#line 914 "View.MemoryView"
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":915
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */

#line 915 "View.MemoryView"
  __pyx_t_1 = __pyx_v_view->itemsize;

#line 915 "View.MemoryView"
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":918
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */

#line 918 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);

#line 918 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":919
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */

#line 919 "View.MemoryView"
    if (unlikely(__pyx_v_itemsize == 0)) {

#line 919 "View.MemoryView"
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");

#line 919 "View.MemoryView"
      __PYX_ERR(2, 919, __pyx_L1_error)

#line 919 "View.MemoryView"
    }

#line 919 "View.MemoryView"
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {

#line 919 "View.MemoryView"
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");

#line 919 "View.MemoryView"
      __PYX_ERR(2, 919, __pyx_L1_error)

#line 919 "View.MemoryView"
    }

#line 919 "View.MemoryView"
    __pyx_v_shape = __Pyx_div_Py_ssize_t(__pyx_v_view->len, __pyx_v_itemsize);

    /* "View.MemoryView":920
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */

#line 920 "View.MemoryView"
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":918
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */

#line 918 "View.MemoryView"
    goto __pyx_L3;

#line 918 "View.MemoryView"
  }

  /* "View.MemoryView":922
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */

#line 922 "View.MemoryView"
  /*else*/ {

#line 922 "View.MemoryView"
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":923
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */

#line 923 "View.MemoryView"
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":924
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */

#line 924 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);

#line 924 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":925
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */

#line 925 "View.MemoryView"
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":924
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */

#line 924 "View.MemoryView"
    }

#line 924 "View.MemoryView"
  }

#line 924 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":927
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */

#line 927 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);

#line 927 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":928
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */

#line 928 "View.MemoryView"
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":929
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 929 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);

#line 929 "View.MemoryView"
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":930
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */

#line 930 "View.MemoryView"
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 930, __pyx_L1_error)

#line 930 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 930 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 930, __pyx_L1_error)

#line 930 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 930 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 930 "View.MemoryView"
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 930, __pyx_L1_error)

#line 930 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 930 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 930 "View.MemoryView"
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 930 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 930 "View.MemoryView"
      __PYX_ERR(2, 930, __pyx_L1_error)

      /* "View.MemoryView":929
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 929 "View.MemoryView"
    }

    /* "View.MemoryView":927
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */

#line 927 "View.MemoryView"
  }

  /* "View.MemoryView":932
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 932 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);

#line 932 "View.MemoryView"
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":933
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */

#line 933 "View.MemoryView"
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 933, __pyx_L1_error)

#line 933 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 933 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 933, __pyx_L1_error)

#line 933 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 933 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 933 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 933, __pyx_L1_error)

#line 933 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 933 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 933 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 933 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 933 "View.MemoryView"
    __PYX_ERR(2, 933, __pyx_L1_error)

    /* "View.MemoryView":932
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 932 "View.MemoryView"
  }

  /* "View.MemoryView":935
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */

#line 935 "View.MemoryView"
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":936
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */

#line 936 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);

#line 936 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":937
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */

#line 937 "View.MemoryView"
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":936
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */

#line 936 "View.MemoryView"
  }

  /* "View.MemoryView":939
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 939 "View.MemoryView"
  __pyx_r = __pyx_v_resultp;

#line 939 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":912
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

#line 912 "View.MemoryView"


#line 912 "View.MemoryView"
  /* function exit code */

#line 912 "View.MemoryView"
  __pyx_L1_error:;

#line 912 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 912 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 912 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 912 "View.MemoryView"
  __pyx_r = NULL;

#line 912 "View.MemoryView"
  __pyx_L0:;

#line 912 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 912 "View.MemoryView"
  return __pyx_r;

#line 912 "View.MemoryView"
}

/* "View.MemoryView":945
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

#line 945 "View.MemoryView"


#line 945 "View.MemoryView"
static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim
#line 945 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_shape
#line 945 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_strides
#line 945 "View.MemoryView"
;
  int __pyx_v_i
#line 945 "View.MemoryView"
;
  int __pyx_v_j
#line 945 "View.MemoryView"
;

#line 945 "View.MemoryView"
  int __pyx_r;

#line 945 "View.MemoryView"
  int __pyx_t_1;

#line 945 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 945 "View.MemoryView"
  long __pyx_t_3;

#line 945 "View.MemoryView"
  long __pyx_t_4;

#line 945 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 945 "View.MemoryView"
  Py_ssize_t __pyx_t_6;

#line 945 "View.MemoryView"
  int __pyx_t_7;

#line 945 "View.MemoryView"
  int __pyx_t_8;

#line 945 "View.MemoryView"
  int __pyx_t_9;

#line 945 "View.MemoryView"
  int __pyx_lineno = 0;

#line 945 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 945 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":946
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */

#line 946 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;

#line 946 "View.MemoryView"
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":948
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */

#line 948 "View.MemoryView"
  __pyx_t_2 = __pyx_v_memslice->shape;

#line 948 "View.MemoryView"
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":949
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 949 "View.MemoryView"
  __pyx_t_2 = __pyx_v_memslice->strides;

#line 949 "View.MemoryView"
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":953
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */

#line 953 "View.MemoryView"
  __pyx_t_3 = __Pyx_div_long(__pyx_v_ndim, 2);

#line 953 "View.MemoryView"
  __pyx_t_4 = __pyx_t_3;

#line 953 "View.MemoryView"
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {

#line 953 "View.MemoryView"
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":954
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */

#line 954 "View.MemoryView"
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":955
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */

#line 955 "View.MemoryView"
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);

#line 955 "View.MemoryView"
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);

#line 955 "View.MemoryView"
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;

#line 955 "View.MemoryView"
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":956
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */

#line 956 "View.MemoryView"
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);

#line 956 "View.MemoryView"
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);

#line 956 "View.MemoryView"
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;

#line 956 "View.MemoryView"
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":958
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */

#line 958 "View.MemoryView"
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);

#line 958 "View.MemoryView"
    if (!__pyx_t_8) {

#line 958 "View.MemoryView"
    } else {

#line 958 "View.MemoryView"
      __pyx_t_7 = __pyx_t_8;

#line 958 "View.MemoryView"
      goto __pyx_L6_bool_binop_done;

#line 958 "View.MemoryView"
    }

#line 958 "View.MemoryView"
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);

#line 958 "View.MemoryView"
    __pyx_t_7 = __pyx_t_8;

#line 958 "View.MemoryView"
    __pyx_L6_bool_binop_done:;

#line 958 "View.MemoryView"
    if (__pyx_t_7) {

      /* "View.MemoryView":959
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */

#line 959 "View.MemoryView"
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 959, __pyx_L1_error)

      /* "View.MemoryView":958
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */

#line 958 "View.MemoryView"
    }

#line 958 "View.MemoryView"
  }

  /* "View.MemoryView":961
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 961 "View.MemoryView"
  __pyx_r = 1;

#line 961 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":945
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

#line 945 "View.MemoryView"


#line 945 "View.MemoryView"
  /* function exit code */

#line 945 "View.MemoryView"
  __pyx_L1_error:;

#line 945 "View.MemoryView"
  {

#line 945 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 945 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 945 "View.MemoryView"
    #endif

#line 945 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 945 "View.MemoryView"
    #ifdef WITH_THREAD

#line 945 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 945 "View.MemoryView"
    #endif

#line 945 "View.MemoryView"
  }

#line 945 "View.MemoryView"
  __pyx_r = 0;

#line 945 "View.MemoryView"
  __pyx_L0:;

#line 945 "View.MemoryView"
  return __pyx_r;

#line 945 "View.MemoryView"
}

/* "View.MemoryView":978
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
/* Python wrapper */

#line 978 "View.MemoryView"
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/

#line 978 "View.MemoryView"
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {

#line 978 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 978 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);

#line 978 "View.MemoryView"
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
  /* function exit code */

#line 978 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 978 "View.MemoryView"
}

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {

#line 978 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 978 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":979
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */

#line 979 "View.MemoryView"
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":978
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
  /* function exit code */

#line 978 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 978 "View.MemoryView"
}

/* "View.MemoryView":981
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

#line 981 "View.MemoryView"


#line 981 "View.MemoryView"
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {

#line 981 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 981 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 981 "View.MemoryView"
  int __pyx_t_1;

#line 981 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 981 "View.MemoryView"
  int __pyx_lineno = 0;

#line 981 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 981 "View.MemoryView"
  int __pyx_clineno = 0;

#line 981 "View.MemoryView"
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":982
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */

#line 982 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);

#line 982 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":983
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */

#line 983 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 983 "View.MemoryView"
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 983, __pyx_L1_error)

#line 983 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 983 "View.MemoryView"
    __pyx_r = __pyx_t_2;

#line 983 "View.MemoryView"
    __pyx_t_2 = 0;

#line 983 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":982
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */

#line 982 "View.MemoryView"
  }

  /* "View.MemoryView":985
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */

#line 985 "View.MemoryView"
  /*else*/ {

#line 985 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 985 "View.MemoryView"
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 985, __pyx_L1_error)

#line 985 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 985 "View.MemoryView"
    __pyx_r = __pyx_t_2;

#line 985 "View.MemoryView"
    __pyx_t_2 = 0;

#line 985 "View.MemoryView"
    goto __pyx_L0;

#line 985 "View.MemoryView"
  }

  /* "View.MemoryView":981
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

#line 981 "View.MemoryView"


#line 981 "View.MemoryView"
  /* function exit code */

#line 981 "View.MemoryView"
  __pyx_L1_error:;

#line 981 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 981 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 981 "View.MemoryView"
  __pyx_r = 0;

#line 981 "View.MemoryView"
  __pyx_L0:;

#line 981 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 981 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 981 "View.MemoryView"
  return __pyx_r;

#line 981 "View.MemoryView"
}

/* "View.MemoryView":987
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

#line 987 "View.MemoryView"


#line 987 "View.MemoryView"
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {

#line 987 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 987 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 987 "View.MemoryView"
  int __pyx_t_1;

#line 987 "View.MemoryView"
  int __pyx_t_2;

#line 987 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 987 "View.MemoryView"
  int __pyx_lineno = 0;

#line 987 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 987 "View.MemoryView"
  int __pyx_clineno = 0;

#line 987 "View.MemoryView"
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":988
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */

#line 988 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);

#line 988 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":989
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */

#line 989 "View.MemoryView"
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 989, __pyx_L1_error)

    /* "View.MemoryView":988
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */

#line 988 "View.MemoryView"
    goto __pyx_L3;

#line 988 "View.MemoryView"
  }

  /* "View.MemoryView":991
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 991 "View.MemoryView"
  /*else*/ {

#line 991 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 991, __pyx_L1_error)

#line 991 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 991 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 991 "View.MemoryView"
  }

#line 991 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":987
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

#line 987 "View.MemoryView"


#line 987 "View.MemoryView"
  /* function exit code */

#line 987 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 987 "View.MemoryView"
  goto __pyx_L0;

#line 987 "View.MemoryView"
  __pyx_L1_error:;

#line 987 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 987 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 987 "View.MemoryView"
  __pyx_r = 0;

#line 987 "View.MemoryView"
  __pyx_L0:;

#line 987 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 987 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 987 "View.MemoryView"
  return __pyx_r;

#line 987 "View.MemoryView"
}

/* "View.MemoryView":994
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
/* Python wrapper */

#line 994 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 994 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {

#line 994 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 994 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 994 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 994 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
  /* function exit code */

#line 994 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 994 "View.MemoryView"
  return __pyx_r;

#line 994 "View.MemoryView"
}

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {

#line 994 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 994 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 994 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":995
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */

#line 995 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 995 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->from_object);

#line 995 "View.MemoryView"
  __pyx_r = __pyx_v_self->from_object;

#line 995 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":994
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
  /* function exit code */

#line 994 "View.MemoryView"
  __pyx_L0:;

#line 994 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 994 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 994 "View.MemoryView"
  return __pyx_r;

#line 994 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 2 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 2 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 2 "(tree fragment)"
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
/* Python wrapper */

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 3 "(tree fragment)"
__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 3 "(tree fragment)"
  int __pyx_lineno = 0;

#line 3 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 3 "(tree fragment)"
  int __pyx_clineno = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __pyx_L1_error:;

#line 3 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "(tree fragment)"
  __pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

/* "View.MemoryView":1001
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

#line 1001 "View.MemoryView"


#line 1001 "View.MemoryView"
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0
#line 1001 "View.MemoryView"
;
  Py_ssize_t __pyx_v_suboffset
#line 1001 "View.MemoryView"
;
  PyObject *__pyx_v_length = NULL
#line 1001 "View.MemoryView"
;

#line 1001 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 1001 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1001 "View.MemoryView"
  int __pyx_t_1;

#line 1001 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1001 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1001 "View.MemoryView"
  __Pyx_TypeInfo *__pyx_t_4;

#line 1001 "View.MemoryView"
  Py_buffer __pyx_t_5;

#line 1001 "View.MemoryView"
  Py_ssize_t *__pyx_t_6;

#line 1001 "View.MemoryView"
  Py_ssize_t *__pyx_t_7;

#line 1001 "View.MemoryView"
  Py_ssize_t *__pyx_t_8;

#line 1001 "View.MemoryView"
  Py_ssize_t __pyx_t_9;

#line 1001 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1001 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1001 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1001 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1009
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

#line 1009 "View.MemoryView"
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);

#line 1009 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1010
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1010 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 1010 "View.MemoryView"
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 1010 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1009
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

#line 1009 "View.MemoryView"
  }

  /* "View.MemoryView":1015
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */

#line 1015 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1015, __pyx_L1_error)

#line 1015 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1015 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1015, __pyx_L1_error)

#line 1015 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1015 "View.MemoryView"
  __Pyx_INCREF(Py_None);

#line 1015 "View.MemoryView"
  __Pyx_GIVEREF(Py_None);

#line 1015 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);

#line 1015 "View.MemoryView"
  __Pyx_INCREF(__pyx_int_0);

#line 1015 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_int_0);

#line 1015 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);

#line 1015 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 1015 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 1015 "View.MemoryView"
  __pyx_t_2 = 0;

#line 1015 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1015, __pyx_L1_error)

#line 1015 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1015 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1015 "View.MemoryView"
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);

#line 1015 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":1017
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */

#line 1017 "View.MemoryView"
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1018
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */

#line 1018 "View.MemoryView"
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1020
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */

#line 1020 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1020, __pyx_L1_error)

#line 1020 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1020 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 1020 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_result->from_object);

#line 1020 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_result->from_object);

#line 1020 "View.MemoryView"
  __pyx_v_result->from_object = __pyx_t_2;

#line 1020 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":1021
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */

#line 1021 "View.MemoryView"
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;

#line 1021 "View.MemoryView"
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1023
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */

#line 1023 "View.MemoryView"
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;

#line 1023 "View.MemoryView"
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1024
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */

#line 1024 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1025
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */

#line 1025 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1026
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */

#line 1026 "View.MemoryView"
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1027
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */

#line 1027 "View.MemoryView"
  Py_INCREF(Py_None);

  /* "View.MemoryView":1029
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */

#line 1029 "View.MemoryView"
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);

#line 1029 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1030
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */

#line 1030 "View.MemoryView"
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1029
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */

#line 1029 "View.MemoryView"
    goto __pyx_L4;

#line 1029 "View.MemoryView"
  }

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */

#line 1032 "View.MemoryView"
  /*else*/ {

#line 1032 "View.MemoryView"
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;

#line 1032 "View.MemoryView"
  }

#line 1032 "View.MemoryView"
  __pyx_L4:;

  /* "View.MemoryView":1034
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */

#line 1034 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1035
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1035 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1038
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

#line 1038 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1039
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */

#line 1039 "View.MemoryView"
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);

#line 1039 "View.MemoryView"
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {

#line 1039 "View.MemoryView"
    __pyx_t_6 = __pyx_t_8;

#line 1039 "View.MemoryView"
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1040
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */

#line 1040 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);

#line 1040 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":1041
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */

#line 1041 "View.MemoryView"
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1042
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */

#line 1042 "View.MemoryView"
      goto __pyx_L6_break;

      /* "View.MemoryView":1040
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */

#line 1040 "View.MemoryView"
    }

#line 1040 "View.MemoryView"
  }

#line 1040 "View.MemoryView"
  __pyx_L6_break:;

  /* "View.MemoryView":1044
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */

#line 1044 "View.MemoryView"
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;

#line 1044 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1045
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */

#line 1045 "View.MemoryView"
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);

#line 1045 "View.MemoryView"
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {

#line 1045 "View.MemoryView"
    __pyx_t_6 = __pyx_t_8;

#line 1045 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1045, __pyx_L1_error)

#line 1045 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 1045 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);

#line 1045 "View.MemoryView"
    __pyx_t_2 = 0;

    /* "View.MemoryView":1046
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */

#line 1046 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1046, __pyx_L1_error)

#line 1046 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 1046 "View.MemoryView"
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1046, __pyx_L1_error)

#line 1046 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 1046 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1046 "View.MemoryView"
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 1046, __pyx_L1_error)

#line 1046 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1046 "View.MemoryView"
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

#line 1046 "View.MemoryView"
  }

  /* "View.MemoryView":1048
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */

#line 1048 "View.MemoryView"
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1049
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */

#line 1049 "View.MemoryView"
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1051
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */

#line 1051 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 1051 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 1051 "View.MemoryView"
  __pyx_r = ((PyObject *)__pyx_v_result);

#line 1051 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1001
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

#line 1001 "View.MemoryView"


#line 1001 "View.MemoryView"
  /* function exit code */

#line 1001 "View.MemoryView"
  __pyx_L1_error:;

#line 1001 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1001 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1001 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1001 "View.MemoryView"
  __pyx_r = 0;

#line 1001 "View.MemoryView"
  __pyx_L0:;

#line 1001 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 1001 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_length);

#line 1001 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 1001 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1001 "View.MemoryView"
  return __pyx_r;

#line 1001 "View.MemoryView"
}

/* "View.MemoryView":1054
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

#line 1054 "View.MemoryView"


#line 1054 "View.MemoryView"
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0
#line 1054 "View.MemoryView"
;

#line 1054 "View.MemoryView"
  __Pyx_memviewslice *__pyx_r;

#line 1054 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1054 "View.MemoryView"
  int __pyx_t_1;

#line 1054 "View.MemoryView"
  int __pyx_t_2;

#line 1054 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1054 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1054 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1054 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1054 "View.MemoryView"
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1057
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */

#line 1057 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 1057 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 1057 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1058
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */

#line 1058 "View.MemoryView"
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(2, 1058, __pyx_L1_error)

#line 1058 "View.MemoryView"
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);

#line 1058 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 1058 "View.MemoryView"
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);

#line 1058 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":1059
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */

#line 1059 "View.MemoryView"
    __pyx_r = (&__pyx_v_obj->from_slice);

#line 1059 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1057
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */

#line 1057 "View.MemoryView"
  }

  /* "View.MemoryView":1061
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */

#line 1061 "View.MemoryView"
  /*else*/ {

#line 1061 "View.MemoryView"
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1062
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */

#line 1062 "View.MemoryView"
    __pyx_r = __pyx_v_mslice;

#line 1062 "View.MemoryView"
    goto __pyx_L0;

#line 1062 "View.MemoryView"
  }

  /* "View.MemoryView":1054
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

#line 1054 "View.MemoryView"


#line 1054 "View.MemoryView"
  /* function exit code */

#line 1054 "View.MemoryView"
  __pyx_L1_error:;

#line 1054 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1054 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1054 "View.MemoryView"
  __pyx_r = NULL;

#line 1054 "View.MemoryView"
  __pyx_L0:;

#line 1054 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);

#line 1054 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1054 "View.MemoryView"
  return __pyx_r;

#line 1054 "View.MemoryView"
}

/* "View.MemoryView":1065
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

#line 1065 "View.MemoryView"


#line 1065 "View.MemoryView"
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim
#line 1065 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_shape
#line 1065 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_strides
#line 1065 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_suboffsets
#line 1065 "View.MemoryView"
;

#line 1065 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1065 "View.MemoryView"
  Py_ssize_t *__pyx_t_1;

#line 1065 "View.MemoryView"
  int __pyx_t_2;

#line 1065 "View.MemoryView"
  int __pyx_t_3;

#line 1065 "View.MemoryView"
  int __pyx_t_4;

#line 1065 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 1065 "View.MemoryView"
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1069
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */

#line 1069 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memview->view.shape;

#line 1069 "View.MemoryView"
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1070
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */

#line 1070 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memview->view.strides;

#line 1070 "View.MemoryView"
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1071
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */

#line 1071 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;

#line 1071 "View.MemoryView"
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1073
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */

#line 1073 "View.MemoryView"
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1074
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */

#line 1074 "View.MemoryView"
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1076
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */

#line 1076 "View.MemoryView"
  __pyx_t_2 = __pyx_v_memview->view.ndim;

#line 1076 "View.MemoryView"
  __pyx_t_3 = __pyx_t_2;

#line 1076 "View.MemoryView"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1076 "View.MemoryView"
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1077
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */

#line 1077 "View.MemoryView"
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1078
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */

#line 1078 "View.MemoryView"
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1079
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */

#line 1079 "View.MemoryView"
    if ((__pyx_v_suboffsets != 0)) {

#line 1079 "View.MemoryView"
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);

#line 1079 "View.MemoryView"
    } else {

#line 1079 "View.MemoryView"
      __pyx_t_5 = -1L;

#line 1079 "View.MemoryView"
    }

#line 1079 "View.MemoryView"
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;

#line 1079 "View.MemoryView"
  }

  /* "View.MemoryView":1065
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

#line 1065 "View.MemoryView"


#line 1065 "View.MemoryView"
  /* function exit code */

#line 1065 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1065 "View.MemoryView"
}

/* "View.MemoryView":1082
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

#line 1082 "View.MemoryView"


#line 1082 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice
#line 1082 "View.MemoryView"
;

#line 1082 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 1082 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1082 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 1082 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1082 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1082 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1082 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1085
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */

#line 1085 "View.MemoryView"
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1086
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */

#line 1086 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 1086 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1086, __pyx_L1_error)

#line 1086 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 1086 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 1086 "View.MemoryView"
  __pyx_t_1 = 0;

#line 1086 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1082
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

#line 1082 "View.MemoryView"


#line 1082 "View.MemoryView"
  /* function exit code */

#line 1082 "View.MemoryView"
  __pyx_L1_error:;

#line 1082 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 1082 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1082 "View.MemoryView"
  __pyx_r = 0;

#line 1082 "View.MemoryView"
  __pyx_L0:;

#line 1082 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 1082 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1082 "View.MemoryView"
  return __pyx_r;

#line 1082 "View.MemoryView"
}

/* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

#line 1089 "View.MemoryView"


#line 1089 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *)
#line 1089 "View.MemoryView"
;
  int (*__pyx_v_to_dtype_func)(char *, PyObject *)
#line 1089 "View.MemoryView"
;

#line 1089 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 1089 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1089 "View.MemoryView"
  int __pyx_t_1;

#line 1089 "View.MemoryView"
  int __pyx_t_2;

#line 1089 "View.MemoryView"
  PyObject *(*__pyx_t_3)(char *);

#line 1089 "View.MemoryView"
  int (*__pyx_t_4)(char *, PyObject *);

#line 1089 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 1089 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1089 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1089 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1089 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1096
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */

#line 1096 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 1096 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 1096 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1097
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */

#line 1097 "View.MemoryView"
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;

#line 1097 "View.MemoryView"
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1098
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */

#line 1098 "View.MemoryView"
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;

#line 1098 "View.MemoryView"
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1096
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */

#line 1096 "View.MemoryView"
    goto __pyx_L3;

#line 1096 "View.MemoryView"
  }

  /* "View.MemoryView":1100
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */

#line 1100 "View.MemoryView"
  /*else*/ {

#line 1100 "View.MemoryView"
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1101
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */

#line 1101 "View.MemoryView"
    __pyx_v_to_dtype_func = NULL;

#line 1101 "View.MemoryView"
  }

#line 1101 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1103
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */

#line 1103 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1105
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1105 "View.MemoryView"
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 1103, __pyx_L1_error)

#line 1105 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_5);

#line 1105 "View.MemoryView"
  __pyx_r = __pyx_t_5;

#line 1105 "View.MemoryView"
  __pyx_t_5 = 0;

#line 1105 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

#line 1089 "View.MemoryView"


#line 1089 "View.MemoryView"
  /* function exit code */

#line 1089 "View.MemoryView"
  __pyx_L1_error:;

#line 1089 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 1089 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1089 "View.MemoryView"
  __pyx_r = 0;

#line 1089 "View.MemoryView"
  __pyx_L0:;

#line 1089 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 1089 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1089 "View.MemoryView"
  return __pyx_r;

#line 1089 "View.MemoryView"
}

/* "View.MemoryView":1111
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

#line 1111 "View.MemoryView"


#line 1111 "View.MemoryView"
static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {

#line 1111 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 1111 "View.MemoryView"
  int __pyx_t_1;

  /* "View.MemoryView":1112
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */

#line 1112 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);

#line 1112 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1113
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */

#line 1113 "View.MemoryView"
    __pyx_r = (-__pyx_v_arg);

#line 1113 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1112
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */

#line 1112 "View.MemoryView"
  }

  /* "View.MemoryView":1115
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */

#line 1115 "View.MemoryView"
  /*else*/ {

#line 1115 "View.MemoryView"
    __pyx_r = __pyx_v_arg;

#line 1115 "View.MemoryView"
    goto __pyx_L0;

#line 1115 "View.MemoryView"
  }

  /* "View.MemoryView":1111
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

#line 1111 "View.MemoryView"


#line 1111 "View.MemoryView"
  /* function exit code */

#line 1111 "View.MemoryView"
  __pyx_L0:;

#line 1111 "View.MemoryView"
  return __pyx_r;

#line 1111 "View.MemoryView"
}

/* "View.MemoryView":1118
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

#line 1118 "View.MemoryView"


#line 1118 "View.MemoryView"
static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i
#line 1118 "View.MemoryView"
;
  Py_ssize_t __pyx_v_c_stride
#line 1118 "View.MemoryView"
;
  Py_ssize_t __pyx_v_f_stride
#line 1118 "View.MemoryView"
;

#line 1118 "View.MemoryView"
  char __pyx_r;

#line 1118 "View.MemoryView"
  int __pyx_t_1;

#line 1118 "View.MemoryView"
  int __pyx_t_2;

#line 1118 "View.MemoryView"
  int __pyx_t_3;

#line 1118 "View.MemoryView"
  int __pyx_t_4;

  /* "View.MemoryView":1123
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */

#line 1123 "View.MemoryView"
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1124
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */

#line 1124 "View.MemoryView"
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1126
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */

#line 1126 "View.MemoryView"
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {

#line 1126 "View.MemoryView"
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1127
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */

#line 1127 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);

#line 1127 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1128
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */

#line 1128 "View.MemoryView"
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1129
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */

#line 1129 "View.MemoryView"
      goto __pyx_L4_break;

      /* "View.MemoryView":1127
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */

#line 1127 "View.MemoryView"
    }

#line 1127 "View.MemoryView"
  }

#line 1127 "View.MemoryView"
  __pyx_L4_break:;

  /* "View.MemoryView":1131
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */

#line 1131 "View.MemoryView"
  __pyx_t_1 = __pyx_v_ndim;

#line 1131 "View.MemoryView"
  __pyx_t_3 = __pyx_t_1;

#line 1131 "View.MemoryView"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1131 "View.MemoryView"
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1132
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */

#line 1132 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);

#line 1132 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1133
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */

#line 1133 "View.MemoryView"
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1134
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */

#line 1134 "View.MemoryView"
      goto __pyx_L7_break;

      /* "View.MemoryView":1132
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */

#line 1132 "View.MemoryView"
    }

#line 1132 "View.MemoryView"
  }

#line 1132 "View.MemoryView"
  __pyx_L7_break:;

  /* "View.MemoryView":1136
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */

#line 1136 "View.MemoryView"
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);

#line 1136 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1137
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */

#line 1137 "View.MemoryView"
    __pyx_r = 'C';

#line 1137 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1136
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */

#line 1136 "View.MemoryView"
  }

  /* "View.MemoryView":1139
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */

#line 1139 "View.MemoryView"
  /*else*/ {

#line 1139 "View.MemoryView"
    __pyx_r = 'F';

#line 1139 "View.MemoryView"
    goto __pyx_L0;

#line 1139 "View.MemoryView"
  }

  /* "View.MemoryView":1118
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

#line 1118 "View.MemoryView"


#line 1118 "View.MemoryView"
  /* function exit code */

#line 1118 "View.MemoryView"
  __pyx_L0:;

#line 1118 "View.MemoryView"
  return __pyx_r;

#line 1118 "View.MemoryView"
}

/* "View.MemoryView":1142
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

#line 1142 "View.MemoryView"


#line 1142 "View.MemoryView"
static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i
#line 1142 "View.MemoryView"
;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent
#line 1142 "View.MemoryView"
;
  Py_ssize_t __pyx_v_dst_extent
#line 1142 "View.MemoryView"
;
  Py_ssize_t __pyx_v_src_stride
#line 1142 "View.MemoryView"
;
  Py_ssize_t __pyx_v_dst_stride
#line 1142 "View.MemoryView"
;

#line 1142 "View.MemoryView"
  int __pyx_t_1;

#line 1142 "View.MemoryView"
  int __pyx_t_2;

#line 1142 "View.MemoryView"
  int __pyx_t_3;

#line 1142 "View.MemoryView"
  Py_ssize_t __pyx_t_4;

#line 1142 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 1142 "View.MemoryView"
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1149
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */

#line 1149 "View.MemoryView"
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */

#line 1150 "View.MemoryView"
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1151
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */

#line 1151 "View.MemoryView"
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */

#line 1152 "View.MemoryView"
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1154
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */

#line 1154 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);

#line 1154 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */

#line 1155 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);

#line 1155 "View.MemoryView"
    if (__pyx_t_2) {

#line 1155 "View.MemoryView"
    } else {

#line 1155 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 1155 "View.MemoryView"
      goto __pyx_L5_bool_binop_done;

#line 1155 "View.MemoryView"
    }

#line 1155 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);

#line 1155 "View.MemoryView"
    if (__pyx_t_2) {

#line 1155 "View.MemoryView"
    } else {

#line 1155 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 1155 "View.MemoryView"
      goto __pyx_L5_bool_binop_done;

#line 1155 "View.MemoryView"
    }

    /* "View.MemoryView":1156
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */

#line 1156 "View.MemoryView"
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);

#line 1156 "View.MemoryView"
    if (__pyx_t_2) {

#line 1156 "View.MemoryView"
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));

#line 1156 "View.MemoryView"
    }

#line 1156 "View.MemoryView"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 1156 "View.MemoryView"
    __pyx_t_1 = __pyx_t_3;

#line 1156 "View.MemoryView"
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */

#line 1155 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":1157
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */

#line 1157 "View.MemoryView"
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */

#line 1155 "View.MemoryView"
      goto __pyx_L4;

#line 1155 "View.MemoryView"
    }

    /* "View.MemoryView":1159
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */

#line 1159 "View.MemoryView"
    /*else*/ {

#line 1159 "View.MemoryView"
      __pyx_t_4 = __pyx_v_dst_extent;

#line 1159 "View.MemoryView"
      __pyx_t_5 = __pyx_t_4;

#line 1159 "View.MemoryView"
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1159 "View.MemoryView"
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1160
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */

#line 1160 "View.MemoryView"
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1161
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */

#line 1161 "View.MemoryView"
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1162
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */

#line 1162 "View.MemoryView"
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);

#line 1162 "View.MemoryView"
      }

#line 1162 "View.MemoryView"
    }

#line 1162 "View.MemoryView"
    __pyx_L4:;

    /* "View.MemoryView":1154
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */

#line 1154 "View.MemoryView"
    goto __pyx_L3;

#line 1154 "View.MemoryView"
  }

  /* "View.MemoryView":1164
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */

#line 1164 "View.MemoryView"
  /*else*/ {

#line 1164 "View.MemoryView"
    __pyx_t_4 = __pyx_v_dst_extent;

#line 1164 "View.MemoryView"
    __pyx_t_5 = __pyx_t_4;

#line 1164 "View.MemoryView"
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1164 "View.MemoryView"
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1165
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */

#line 1165 "View.MemoryView"
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1169
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */

#line 1169 "View.MemoryView"
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1170
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */

#line 1170 "View.MemoryView"
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);

#line 1170 "View.MemoryView"
    }

#line 1170 "View.MemoryView"
  }

#line 1170 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1142
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

#line 1142 "View.MemoryView"


#line 1142 "View.MemoryView"
  /* function exit code */

#line 1142 "View.MemoryView"
}

/* "View.MemoryView":1172
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

#line 1172 "View.MemoryView"


#line 1172 "View.MemoryView"
static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1175
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */

#line 1175 "View.MemoryView"
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1172
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

#line 1172 "View.MemoryView"


#line 1172 "View.MemoryView"
  /* function exit code */

#line 1172 "View.MemoryView"
}

/* "View.MemoryView":1179
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

#line 1179 "View.MemoryView"


#line 1179 "View.MemoryView"
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape
#line 1179 "View.MemoryView"
;
  Py_ssize_t __pyx_v_size
#line 1179 "View.MemoryView"
;

#line 1179 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 1179 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1179 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 1179 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 1179 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1181
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */

#line 1181 "View.MemoryView"
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;

#line 1181 "View.MemoryView"
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1183
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */

#line 1183 "View.MemoryView"
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);

#line 1183 "View.MemoryView"
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {

#line 1183 "View.MemoryView"
    __pyx_t_2 = __pyx_t_4;

#line 1183 "View.MemoryView"
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1184
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */

#line 1184 "View.MemoryView"
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);

#line 1184 "View.MemoryView"
  }

  /* "View.MemoryView":1186
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */

#line 1186 "View.MemoryView"
  __pyx_r = __pyx_v_size;

#line 1186 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1179
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

#line 1179 "View.MemoryView"


#line 1179 "View.MemoryView"
  /* function exit code */

#line 1179 "View.MemoryView"
  __pyx_L0:;

#line 1179 "View.MemoryView"
  return __pyx_r;

#line 1179 "View.MemoryView"
}

/* "View.MemoryView":1189
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

#line 1189 "View.MemoryView"


#line 1189 "View.MemoryView"
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx
#line 1189 "View.MemoryView"
;

#line 1189 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 1189 "View.MemoryView"
  int __pyx_t_1;

#line 1189 "View.MemoryView"
  int __pyx_t_2;

#line 1189 "View.MemoryView"
  int __pyx_t_3;

#line 1189 "View.MemoryView"
  int __pyx_t_4;

  /* "View.MemoryView":1198
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */

#line 1198 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);

#line 1198 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1199
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */

#line 1199 "View.MemoryView"
    __pyx_t_2 = __pyx_v_ndim;

#line 1199 "View.MemoryView"
    __pyx_t_3 = __pyx_t_2;

#line 1199 "View.MemoryView"
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1199 "View.MemoryView"
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1200
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */

#line 1200 "View.MemoryView"
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1201
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */

#line 1201 "View.MemoryView"
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));

#line 1201 "View.MemoryView"
    }

    /* "View.MemoryView":1198
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */

#line 1198 "View.MemoryView"
    goto __pyx_L3;

#line 1198 "View.MemoryView"
  }

  /* "View.MemoryView":1203
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */

#line 1203 "View.MemoryView"
  /*else*/ {

#line 1203 "View.MemoryView"
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {

#line 1203 "View.MemoryView"
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1204
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */

#line 1204 "View.MemoryView"
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1205
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */

#line 1205 "View.MemoryView"
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));

#line 1205 "View.MemoryView"
    }

#line 1205 "View.MemoryView"
  }

#line 1205 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1207
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */

#line 1207 "View.MemoryView"
  __pyx_r = __pyx_v_stride;

#line 1207 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1189
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

#line 1189 "View.MemoryView"


#line 1189 "View.MemoryView"
  /* function exit code */

#line 1189 "View.MemoryView"
  __pyx_L0:;

#line 1189 "View.MemoryView"
  return __pyx_r;

#line 1189 "View.MemoryView"
}

/* "View.MemoryView":1210
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

#line 1210 "View.MemoryView"


#line 1210 "View.MemoryView"
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i
#line 1210 "View.MemoryView"
;
  void *__pyx_v_result
#line 1210 "View.MemoryView"
;
  size_t __pyx_v_itemsize
#line 1210 "View.MemoryView"
;
  size_t __pyx_v_size
#line 1210 "View.MemoryView"
;

#line 1210 "View.MemoryView"
  void *__pyx_r;

#line 1210 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1210 "View.MemoryView"
  int __pyx_t_2;

#line 1210 "View.MemoryView"
  int __pyx_t_3;

#line 1210 "View.MemoryView"
  struct __pyx_memoryview_obj *__pyx_t_4;

#line 1210 "View.MemoryView"
  int __pyx_t_5;

#line 1210 "View.MemoryView"
  int __pyx_t_6;

#line 1210 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1210 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1210 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":1221
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */

#line 1221 "View.MemoryView"
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;

#line 1221 "View.MemoryView"
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1222
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */

#line 1222 "View.MemoryView"
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1224
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */

#line 1224 "View.MemoryView"
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1225
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */

#line 1225 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);

#line 1225 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1226
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1226 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 1226, __pyx_L1_error)

    /* "View.MemoryView":1225
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */

#line 1225 "View.MemoryView"
  }

  /* "View.MemoryView":1229
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */

#line 1229 "View.MemoryView"
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1230
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */

#line 1230 "View.MemoryView"
  __pyx_t_4 = __pyx_v_src->memview;

#line 1230 "View.MemoryView"
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1231
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */

#line 1231 "View.MemoryView"
  __pyx_t_3 = __pyx_v_ndim;

#line 1231 "View.MemoryView"
  __pyx_t_5 = __pyx_t_3;

#line 1231 "View.MemoryView"
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1231 "View.MemoryView"
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1232
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */

#line 1232 "View.MemoryView"
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1233
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */

#line 1233 "View.MemoryView"
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;

#line 1233 "View.MemoryView"
  }

  /* "View.MemoryView":1235
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */

#line 1235 "View.MemoryView"
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1239
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */

#line 1239 "View.MemoryView"
  __pyx_t_3 = __pyx_v_ndim;

#line 1239 "View.MemoryView"
  __pyx_t_5 = __pyx_t_3;

#line 1239 "View.MemoryView"
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1239 "View.MemoryView"
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1240
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */

#line 1240 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);

#line 1240 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1241
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */

#line 1241 "View.MemoryView"
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1240
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */

#line 1240 "View.MemoryView"
    }

#line 1240 "View.MemoryView"
  }

  /* "View.MemoryView":1243
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */

#line 1243 "View.MemoryView"
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);

#line 1243 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1244
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */

#line 1244 "View.MemoryView"
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1243
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */

#line 1243 "View.MemoryView"
    goto __pyx_L9;

#line 1243 "View.MemoryView"
  }

  /* "View.MemoryView":1246
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */

#line 1246 "View.MemoryView"
  /*else*/ {

#line 1246 "View.MemoryView"
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);

#line 1246 "View.MemoryView"
  }

#line 1246 "View.MemoryView"
  __pyx_L9:;

  /* "View.MemoryView":1248
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1248 "View.MemoryView"
  __pyx_r = __pyx_v_result;

#line 1248 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1210
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

#line 1210 "View.MemoryView"


#line 1210 "View.MemoryView"
  /* function exit code */

#line 1210 "View.MemoryView"
  __pyx_L1_error:;

#line 1210 "View.MemoryView"
  {

#line 1210 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 1210 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1210 "View.MemoryView"
    #endif

#line 1210 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1210 "View.MemoryView"
    #ifdef WITH_THREAD

#line 1210 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1210 "View.MemoryView"
    #endif

#line 1210 "View.MemoryView"
  }

#line 1210 "View.MemoryView"
  __pyx_r = NULL;

#line 1210 "View.MemoryView"
  __pyx_L0:;

#line 1210 "View.MemoryView"
  return __pyx_r;

#line 1210 "View.MemoryView"
}

/* "View.MemoryView":1253
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

#line 1253 "View.MemoryView"


#line 1253 "View.MemoryView"
static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {

#line 1253 "View.MemoryView"
  int __pyx_r;

#line 1253 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 1253 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1253 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1253 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1253 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1253 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1253 "View.MemoryView"
  #endif

#line 1253 "View.MemoryView"
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1256
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */

#line 1256 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 1256 "View.MemoryView"
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1256 "View.MemoryView"
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1256 "View.MemoryView"
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 1256 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 1256 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);

#line 1256 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 1256 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);

#line 1256 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 1256 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);

#line 1256 "View.MemoryView"
  __pyx_t_1 = 0;

#line 1256 "View.MemoryView"
  __pyx_t_2 = 0;

#line 1256 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":1255
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */

#line 1255 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1255, __pyx_L1_error)

#line 1255 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1255 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1255 "View.MemoryView"
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1255, __pyx_L1_error)

#line 1255 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 1255 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1255 "View.MemoryView"
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 1255 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1255 "View.MemoryView"
  __PYX_ERR(2, 1255, __pyx_L1_error)

  /* "View.MemoryView":1253
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

#line 1253 "View.MemoryView"


#line 1253 "View.MemoryView"
  /* function exit code */

#line 1253 "View.MemoryView"
  __pyx_L1_error:;

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 1253 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1253 "View.MemoryView"
  __pyx_r = -1;

#line 1253 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1253 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1253 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1253 "View.MemoryView"
  #endif

#line 1253 "View.MemoryView"
  return __pyx_r;

#line 1253 "View.MemoryView"
}

/* "View.MemoryView":1259
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

#line 1259 "View.MemoryView"


#line 1259 "View.MemoryView"
static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {

#line 1259 "View.MemoryView"
  int __pyx_r;

#line 1259 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 1259 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1259 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1259 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1259 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1259 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1259 "View.MemoryView"
  #endif

#line 1259 "View.MemoryView"
  __Pyx_RefNannySetupContext("_err_dim", 0);

#line 1259 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1260
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */

#line 1260 "View.MemoryView"
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1260 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1260 "View.MemoryView"
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1260 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; 
#line 1260 "View.MemoryView"
__pyx_t_2 = NULL;

#line 1260 "View.MemoryView"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 1260 "View.MemoryView"
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);

#line 1260 "View.MemoryView"
    if (likely(__pyx_t_2)) {

#line 1260 "View.MemoryView"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 1260 "View.MemoryView"
      __Pyx_INCREF(__pyx_t_2);

#line 1260 "View.MemoryView"
      __Pyx_INCREF(function);

#line 1260 "View.MemoryView"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 1260 "View.MemoryView"
    }

#line 1260 "View.MemoryView"
  }

#line 1260 "View.MemoryView"
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);

#line 1260 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1260 "View.MemoryView"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1260 "View.MemoryView"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1260 "View.MemoryView"
  __PYX_ERR(2, 1260, __pyx_L1_error)

  /* "View.MemoryView":1259
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

#line 1259 "View.MemoryView"


#line 1259 "View.MemoryView"
  /* function exit code */

#line 1259 "View.MemoryView"
  __pyx_L1_error:;

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 1259 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1259 "View.MemoryView"
  __pyx_r = -1;

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_error);

#line 1259 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1259 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1259 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1259 "View.MemoryView"
  #endif

#line 1259 "View.MemoryView"
  return __pyx_r;

#line 1259 "View.MemoryView"
}

/* "View.MemoryView":1263
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

#line 1263 "View.MemoryView"


#line 1263 "View.MemoryView"
static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {

#line 1263 "View.MemoryView"
  int __pyx_r;

#line 1263 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1263 "View.MemoryView"
  int __pyx_t_1;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 1263 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1263 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1263 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1263 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1263 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1263 "View.MemoryView"
  #endif

#line 1263 "View.MemoryView"
  __Pyx_RefNannySetupContext("_err", 0);

#line 1263 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1264
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */

#line 1264 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);

#line 1264 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1265
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */

#line 1265 "View.MemoryView"
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1265, __pyx_L1_error)

#line 1265 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 1265 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; 
#line 1265 "View.MemoryView"
__pyx_t_5 = NULL;

#line 1265 "View.MemoryView"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 1265 "View.MemoryView"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 1265 "View.MemoryView"
      if (likely(__pyx_t_5)) {

#line 1265 "View.MemoryView"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 1265 "View.MemoryView"
        __Pyx_INCREF(__pyx_t_5);

#line 1265 "View.MemoryView"
        __Pyx_INCREF(function);

#line 1265 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_t_4, function);

#line 1265 "View.MemoryView"
      }

#line 1265 "View.MemoryView"
    }

#line 1265 "View.MemoryView"
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 1265 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 1265 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1265 "View.MemoryView"
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1265, __pyx_L1_error)

#line 1265 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 1265 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1265 "View.MemoryView"
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 1265 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1265 "View.MemoryView"
    __PYX_ERR(2, 1265, __pyx_L1_error)

    /* "View.MemoryView":1264
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */

#line 1264 "View.MemoryView"
  }

  /* "View.MemoryView":1267
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */

#line 1267 "View.MemoryView"
  /*else*/ {

#line 1267 "View.MemoryView"
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);

#line 1267 "View.MemoryView"
    __PYX_ERR(2, 1267, __pyx_L1_error)

#line 1267 "View.MemoryView"
  }

  /* "View.MemoryView":1263
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

#line 1263 "View.MemoryView"


#line 1263 "View.MemoryView"
  /* function exit code */

#line 1263 "View.MemoryView"
  __pyx_L1_error:;

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 1263 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1263 "View.MemoryView"
  __pyx_r = -1;

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_error);

#line 1263 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1263 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1263 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1263 "View.MemoryView"
  #endif

#line 1263 "View.MemoryView"
  return __pyx_r;

#line 1263 "View.MemoryView"
}

/* "View.MemoryView":1270
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

#line 1270 "View.MemoryView"


#line 1270 "View.MemoryView"
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata
#line 1270 "View.MemoryView"
;
  size_t __pyx_v_itemsize
#line 1270 "View.MemoryView"
;
  int __pyx_v_i
#line 1270 "View.MemoryView"
;
  char __pyx_v_order
#line 1270 "View.MemoryView"
;
  int __pyx_v_broadcasting
#line 1270 "View.MemoryView"
;
  int __pyx_v_direct_copy
#line 1270 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp
#line 1270 "View.MemoryView"
;
  int __pyx_v_ndim
#line 1270 "View.MemoryView"
;

#line 1270 "View.MemoryView"
  int __pyx_r;

#line 1270 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1270 "View.MemoryView"
  int __pyx_t_2;

#line 1270 "View.MemoryView"
  int __pyx_t_3;

#line 1270 "View.MemoryView"
  int __pyx_t_4;

#line 1270 "View.MemoryView"
  int __pyx_t_5;

#line 1270 "View.MemoryView"
  int __pyx_t_6;

#line 1270 "View.MemoryView"
  void *__pyx_t_7;

#line 1270 "View.MemoryView"
  int __pyx_t_8;

#line 1270 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1270 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1270 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":1278
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */

#line 1278 "View.MemoryView"
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1279
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */

#line 1279 "View.MemoryView"
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;

#line 1279 "View.MemoryView"
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1281
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */

#line 1281 "View.MemoryView"
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1282
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */

#line 1282 "View.MemoryView"
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1283
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */

#line 1283 "View.MemoryView"
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1286
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */

#line 1286 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);

#line 1286 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1287
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */

#line 1287 "View.MemoryView"
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1286
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */

#line 1286 "View.MemoryView"
    goto __pyx_L3;

#line 1286 "View.MemoryView"
  }

  /* "View.MemoryView":1288
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */

#line 1288 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);

#line 1288 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1289
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */

#line 1289 "View.MemoryView"
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1288
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */

#line 1288 "View.MemoryView"
  }

#line 1288 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1291
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */

#line 1291 "View.MemoryView"
  __pyx_t_3 = __pyx_v_dst_ndim;

#line 1291 "View.MemoryView"
  __pyx_t_4 = __pyx_v_src_ndim;

#line 1291 "View.MemoryView"
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {

#line 1291 "View.MemoryView"
    __pyx_t_5 = __pyx_t_3;

#line 1291 "View.MemoryView"
  } else {

#line 1291 "View.MemoryView"
    __pyx_t_5 = __pyx_t_4;

#line 1291 "View.MemoryView"
  }

#line 1291 "View.MemoryView"
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1293
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */

#line 1293 "View.MemoryView"
  __pyx_t_5 = __pyx_v_ndim;

#line 1293 "View.MemoryView"
  __pyx_t_3 = __pyx_t_5;

#line 1293 "View.MemoryView"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1293 "View.MemoryView"
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1294
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */

#line 1294 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);

#line 1294 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1295
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */

#line 1295 "View.MemoryView"
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);

#line 1295 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":1296
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */

#line 1296 "View.MemoryView"
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1297
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */

#line 1297 "View.MemoryView"
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1295
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */

#line 1295 "View.MemoryView"
        goto __pyx_L7;

#line 1295 "View.MemoryView"
      }

      /* "View.MemoryView":1299
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */

#line 1299 "View.MemoryView"
      /*else*/ {

#line 1299 "View.MemoryView"
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 1299, __pyx_L1_error)

#line 1299 "View.MemoryView"
      }

#line 1299 "View.MemoryView"
      __pyx_L7:;

      /* "View.MemoryView":1294
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */

#line 1294 "View.MemoryView"
    }

    /* "View.MemoryView":1301
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */

#line 1301 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);

#line 1301 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1302
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */

#line 1302 "View.MemoryView"
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 1302, __pyx_L1_error)

      /* "View.MemoryView":1301
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */

#line 1301 "View.MemoryView"
    }

#line 1301 "View.MemoryView"
  }

  /* "View.MemoryView":1304
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */

#line 1304 "View.MemoryView"
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);

#line 1304 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1306
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */

#line 1306 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);

#line 1306 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1307
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */

#line 1307 "View.MemoryView"
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1306
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */

#line 1306 "View.MemoryView"
    }

    /* "View.MemoryView":1309
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */

#line 1309 "View.MemoryView"
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(2, 1309, __pyx_L1_error)

#line 1309 "View.MemoryView"
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1310
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */

#line 1310 "View.MemoryView"
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1304
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */

#line 1304 "View.MemoryView"
  }

  /* "View.MemoryView":1312
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1312 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);

#line 1312 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1315
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */

#line 1315 "View.MemoryView"
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);

#line 1315 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1316
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */

#line 1316 "View.MemoryView"
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1315
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */

#line 1315 "View.MemoryView"
      goto __pyx_L12;

#line 1315 "View.MemoryView"
    }

    /* "View.MemoryView":1317
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */

#line 1317 "View.MemoryView"
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);

#line 1317 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */

#line 1318 "View.MemoryView"
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1317
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */

#line 1317 "View.MemoryView"
    }

#line 1317 "View.MemoryView"
    __pyx_L12:;

    /* "View.MemoryView":1320
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */

#line 1320 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_direct_copy != 0);

#line 1320 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1322
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */

#line 1322 "View.MemoryView"
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1323
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */

#line 1323 "View.MemoryView"
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1324
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */

#line 1324 "View.MemoryView"
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1325
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */

#line 1325 "View.MemoryView"
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1326
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */

#line 1326 "View.MemoryView"
      __pyx_r = 0;

#line 1326 "View.MemoryView"
      goto __pyx_L0;

      /* "View.MemoryView":1320
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */

#line 1320 "View.MemoryView"
    }

    /* "View.MemoryView":1312
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1312 "View.MemoryView"
  }

  /* "View.MemoryView":1328
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1328 "View.MemoryView"
  __pyx_t_2 = (__pyx_v_order == 'F');

#line 1328 "View.MemoryView"
  if (__pyx_t_2) {

#line 1328 "View.MemoryView"
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));

#line 1328 "View.MemoryView"
  }

#line 1328 "View.MemoryView"
  __pyx_t_8 = (__pyx_t_2 != 0);

#line 1328 "View.MemoryView"
  if (__pyx_t_8) {

    /* "View.MemoryView":1331
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */

#line 1331 "View.MemoryView"
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(2, 1331, __pyx_L1_error)

    /* "View.MemoryView":1332
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */

#line 1332 "View.MemoryView"
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(2, 1332, __pyx_L1_error)

    /* "View.MemoryView":1328
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1328 "View.MemoryView"
  }

  /* "View.MemoryView":1334
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */

#line 1334 "View.MemoryView"
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1335
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */

#line 1335 "View.MemoryView"
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1336
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */

#line 1336 "View.MemoryView"
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1338
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */

#line 1338 "View.MemoryView"
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1339
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */

#line 1339 "View.MemoryView"
  __pyx_r = 0;

#line 1339 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1270
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

#line 1270 "View.MemoryView"


#line 1270 "View.MemoryView"
  /* function exit code */

#line 1270 "View.MemoryView"
  __pyx_L1_error:;

#line 1270 "View.MemoryView"
  {

#line 1270 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 1270 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1270 "View.MemoryView"
    #endif

#line 1270 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1270 "View.MemoryView"
    #ifdef WITH_THREAD

#line 1270 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1270 "View.MemoryView"
    #endif

#line 1270 "View.MemoryView"
  }

#line 1270 "View.MemoryView"
  __pyx_r = -1;

#line 1270 "View.MemoryView"
  __pyx_L0:;

#line 1270 "View.MemoryView"
  return __pyx_r;

#line 1270 "View.MemoryView"
}

/* "View.MemoryView":1342
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

#line 1342 "View.MemoryView"


#line 1342 "View.MemoryView"
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i
#line 1342 "View.MemoryView"
;
  int __pyx_v_offset
#line 1342 "View.MemoryView"
;

#line 1342 "View.MemoryView"
  int __pyx_t_1;

#line 1342 "View.MemoryView"
  int __pyx_t_2;

#line 1342 "View.MemoryView"
  int __pyx_t_3;

  /* "View.MemoryView":1346
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */

#line 1346 "View.MemoryView"
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1348
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */

#line 1348 "View.MemoryView"
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {

#line 1348 "View.MemoryView"
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1349
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */

#line 1349 "View.MemoryView"
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1350
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */

#line 1350 "View.MemoryView"
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1351
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */

#line 1351 "View.MemoryView"
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);

#line 1351 "View.MemoryView"
  }

  /* "View.MemoryView":1353
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */

#line 1353 "View.MemoryView"
  __pyx_t_1 = __pyx_v_offset;

#line 1353 "View.MemoryView"
  __pyx_t_2 = __pyx_t_1;

#line 1353 "View.MemoryView"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 1353 "View.MemoryView"
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1354
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */

#line 1354 "View.MemoryView"
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1355
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */

#line 1355 "View.MemoryView"
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1356
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1356 "View.MemoryView"
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;

#line 1356 "View.MemoryView"
  }

  /* "View.MemoryView":1342
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

#line 1342 "View.MemoryView"


#line 1342 "View.MemoryView"
  /* function exit code */

#line 1342 "View.MemoryView"
}

/* "View.MemoryView":1364
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

#line 1364 "View.MemoryView"


#line 1364 "View.MemoryView"
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {

#line 1364 "View.MemoryView"
  int __pyx_t_1;

  /* "View.MemoryView":1368
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */

#line 1368 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);

#line 1368 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1369
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */

#line 1369 "View.MemoryView"
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1368
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */

#line 1368 "View.MemoryView"
  }

  /* "View.MemoryView":1364
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

#line 1364 "View.MemoryView"


#line 1364 "View.MemoryView"
  /* function exit code */

#line 1364 "View.MemoryView"
}

/* "View.MemoryView":1373
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

#line 1373 "View.MemoryView"


#line 1373 "View.MemoryView"
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {

#line 1373 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1373 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1373 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1373 "View.MemoryView"
  #endif

#line 1373 "View.MemoryView"
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1376
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */

#line 1376 "View.MemoryView"
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1373
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

#line 1373 "View.MemoryView"


#line 1373 "View.MemoryView"
  /* function exit code */

#line 1373 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1373 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1373 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1373 "View.MemoryView"
  #endif

#line 1373 "View.MemoryView"
}

/* "View.MemoryView":1379
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

#line 1379 "View.MemoryView"


#line 1379 "View.MemoryView"
static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i
#line 1379 "View.MemoryView"
;

#line 1379 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1379 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1379 "View.MemoryView"
  Py_ssize_t __pyx_t_2;

#line 1379 "View.MemoryView"
  Py_ssize_t __pyx_t_3;

#line 1379 "View.MemoryView"
  int __pyx_t_4;

#line 1379 "View.MemoryView"
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1383
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */

#line 1383 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_shape[0]);

#line 1383 "View.MemoryView"
  __pyx_t_2 = __pyx_t_1;

#line 1383 "View.MemoryView"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 1383 "View.MemoryView"
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1384
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */

#line 1384 "View.MemoryView"
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);

#line 1384 "View.MemoryView"
    if (__pyx_t_4) {

      /* "View.MemoryView":1385
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */

#line 1385 "View.MemoryView"
      __pyx_t_4 = (__pyx_v_inc != 0);

#line 1385 "View.MemoryView"
      if (__pyx_t_4) {

        /* "View.MemoryView":1386
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */

#line 1386 "View.MemoryView"
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1385
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */

#line 1385 "View.MemoryView"
        goto __pyx_L6;

#line 1385 "View.MemoryView"
      }

      /* "View.MemoryView":1388
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */

#line 1388 "View.MemoryView"
      /*else*/ {

#line 1388 "View.MemoryView"
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));

#line 1388 "View.MemoryView"
      }

#line 1388 "View.MemoryView"
      __pyx_L6:;

      /* "View.MemoryView":1384
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */

#line 1384 "View.MemoryView"
      goto __pyx_L5;

#line 1384 "View.MemoryView"
    }

    /* "View.MemoryView":1390
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */

#line 1390 "View.MemoryView"
    /*else*/ {

      /* "View.MemoryView":1391
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */

#line 1391 "View.MemoryView"
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);

#line 1391 "View.MemoryView"
    }

#line 1391 "View.MemoryView"
    __pyx_L5:;

    /* "View.MemoryView":1393
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1393 "View.MemoryView"
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));

#line 1393 "View.MemoryView"
  }

  /* "View.MemoryView":1379
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

#line 1379 "View.MemoryView"


#line 1379 "View.MemoryView"
  /* function exit code */

#line 1379 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1379 "View.MemoryView"
}

/* "View.MemoryView":1399
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

#line 1399 "View.MemoryView"


#line 1399 "View.MemoryView"
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1402
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */

#line 1402 "View.MemoryView"
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1403
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */

#line 1403 "View.MemoryView"
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1405
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1405 "View.MemoryView"
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1399
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

#line 1399 "View.MemoryView"


#line 1399 "View.MemoryView"
  /* function exit code */

#line 1399 "View.MemoryView"
}

/* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

#line 1409 "View.MemoryView"


#line 1409 "View.MemoryView"
static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i
#line 1409 "View.MemoryView"
;
  Py_ssize_t __pyx_v_stride
#line 1409 "View.MemoryView"
;
  Py_ssize_t __pyx_v_extent
#line 1409 "View.MemoryView"
;

#line 1409 "View.MemoryView"
  int __pyx_t_1;

#line 1409 "View.MemoryView"
  Py_ssize_t __pyx_t_2;

#line 1409 "View.MemoryView"
  Py_ssize_t __pyx_t_3;

#line 1409 "View.MemoryView"
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1413
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */

#line 1413 "View.MemoryView"
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1414
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */

#line 1414 "View.MemoryView"
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */

#line 1416 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);

#line 1416 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1417
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */

#line 1417 "View.MemoryView"
    __pyx_t_2 = __pyx_v_extent;

#line 1417 "View.MemoryView"
    __pyx_t_3 = __pyx_t_2;

#line 1417 "View.MemoryView"
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1417 "View.MemoryView"
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1418
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */

#line 1418 "View.MemoryView"
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1419
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */

#line 1419 "View.MemoryView"
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);

#line 1419 "View.MemoryView"
    }

    /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */

#line 1416 "View.MemoryView"
    goto __pyx_L3;

#line 1416 "View.MemoryView"
  }

  /* "View.MemoryView":1421
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */

#line 1421 "View.MemoryView"
  /*else*/ {

#line 1421 "View.MemoryView"
    __pyx_t_2 = __pyx_v_extent;

#line 1421 "View.MemoryView"
    __pyx_t_3 = __pyx_t_2;

#line 1421 "View.MemoryView"
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1421 "View.MemoryView"
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1422
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */

#line 1422 "View.MemoryView"
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1424
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1424 "View.MemoryView"
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);

#line 1424 "View.MemoryView"
    }

#line 1424 "View.MemoryView"
  }

#line 1424 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

#line 1409 "View.MemoryView"


#line 1409 "View.MemoryView"
  /* function exit code */

#line 1409 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__26, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  int __pyx_t_2;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_3;

#line 11 "(tree fragment)"
  int __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->name);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->name);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->name = __pyx_t_1;

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_4) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_2 = __pyx_t_4;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = (__pyx_t_4 != 0);

#line 13 "(tree fragment)"
  __pyx_t_2 = __pyx_t_5;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 14 "(tree fragment)"
    __pyx_t_8 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {

#line 14 "(tree fragment)"
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_8)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_7, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_array __pyx_vtable_array;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_array_obj *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_array_obj *)o);

#line 11 "(tree fragment)"
  p->__pyx_vtab = __pyx_vtabptr_array;

#line 11 "(tree fragment)"
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
  bad:

#line 11 "(tree fragment)"
  Py_DECREF(o); o = 0;

#line 11 "(tree fragment)"
  return NULL;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_array(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  {

#line 11 "(tree fragment)"
    PyObject *etype, *eval, *etb;

#line 11 "(tree fragment)"
    PyErr_Fetch(&etype, &eval, &etb);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);

#line 11 "(tree fragment)"
    __pyx_array___dealloc__(o);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);

#line 11 "(tree fragment)"
    PyErr_Restore(etype, eval, etb);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  Py_CLEAR(p->mode);

#line 11 "(tree fragment)"
  Py_CLEAR(p->_format);

#line 11 "(tree fragment)"
  (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {

#line 11 "(tree fragment)"
  PyObject *r;

#line 11 "(tree fragment)"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 11 "(tree fragment)"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 11 "(tree fragment)"
  Py_DECREF(x);

#line 11 "(tree fragment)"
  return r;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_array___setitem__(o, i, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_Format(PyExc_NotImplementedError,

#line 11 "(tree fragment)"
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {

#line 11 "(tree fragment)"
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);

#line 11 "(tree fragment)"
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {

#line 11 "(tree fragment)"
    PyErr_Clear();

#line 11 "(tree fragment)"
    v = __pyx_array___getattr__(o, n);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return v;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_array[] = {

#line 11 "(tree fragment)"
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},

#line 11 "(tree fragment)"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_array[] = {

#line 11 "(tree fragment)"
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PySequenceMethods __pyx_tp_as_sequence_array = {

#line 11 "(tree fragment)"
  __pyx_array___len__, /*sq_length*/

#line 11 "(tree fragment)"
  0, /*sq_concat*/

#line 11 "(tree fragment)"
  0, /*sq_repeat*/

#line 11 "(tree fragment)"
  __pyx_sq_item_array, /*sq_item*/

#line 11 "(tree fragment)"
  0, /*sq_slice*/

#line 11 "(tree fragment)"
  0, /*sq_ass_item*/

#line 11 "(tree fragment)"
  0, /*sq_ass_slice*/

#line 11 "(tree fragment)"
  0, /*sq_contains*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_concat*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_repeat*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMappingMethods __pyx_tp_as_mapping_array = {

#line 11 "(tree fragment)"
  __pyx_array___len__, /*mp_length*/

#line 11 "(tree fragment)"
  __pyx_array___getitem__, /*mp_subscript*/

#line 11 "(tree fragment)"
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyBufferProcs __pyx_tp_as_buffer_array = {

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getreadbuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getwritebuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getsegcount*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getcharbuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_array_getbuffer, /*bf_getbuffer*/

#line 11 "(tree fragment)"
  0, /*bf_releasebuffer*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type___pyx_array = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.utils.array", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_array, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  __pyx_tp_getattro_array, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/

#line 11 "(tree fragment)"
  0, /*tp_doc*/

#line 11 "(tree fragment)"
  0, /*tp_traverse*/

#line 11 "(tree fragment)"
  0, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_array, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_array, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_array, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_MemviewEnum_obj *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_MemviewEnum_obj *)o);

#line 11 "(tree fragment)"
  p->name = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_Enum(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  Py_CLEAR(p->name);

#line 11 "(tree fragment)"
  (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;

#line 11 "(tree fragment)"
  if (p->name) {

#line 11 "(tree fragment)"
    e = (*v)(p->name, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_Enum(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->name);

#line 11 "(tree fragment)"
  p->name = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_Enum[] = {

#line 11 "(tree fragment)"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type___pyx_MemviewEnum = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.utils.Enum", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_MemviewEnum___repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  0, /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_Enum, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_Enum, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_Enum, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_MemviewEnum___init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_Enum, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_memoryview_obj *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_memoryview_obj *)o);

#line 11 "(tree fragment)"
  p->__pyx_vtab = __pyx_vtabptr_memoryview;

#line 11 "(tree fragment)"
  p->obj = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->_size = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->_array_interface = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->view.obj = NULL;

#line 11 "(tree fragment)"
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
  bad:

#line 11 "(tree fragment)"
  Py_DECREF(o); o = 0;

#line 11 "(tree fragment)"
  return NULL;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_memoryview(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  {

#line 11 "(tree fragment)"
    PyObject *etype, *eval, *etb;

#line 11 "(tree fragment)"
    PyErr_Fetch(&etype, &eval, &etb);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);

#line 11 "(tree fragment)"
    __pyx_memoryview___dealloc__(o);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);

#line 11 "(tree fragment)"
    PyErr_Restore(etype, eval, etb);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  Py_CLEAR(p->obj);

#line 11 "(tree fragment)"
  Py_CLEAR(p->_size);

#line 11 "(tree fragment)"
  Py_CLEAR(p->_array_interface);

#line 11 "(tree fragment)"
  (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;

#line 11 "(tree fragment)"
  if (p->obj) {

#line 11 "(tree fragment)"
    e = (*v)(p->obj, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->_size) {

#line 11 "(tree fragment)"
    e = (*v)(p->_size, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->_array_interface) {

#line 11 "(tree fragment)"
    e = (*v)(p->_array_interface, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->view.obj) {

#line 11 "(tree fragment)"
    e = (*v)(p->view.obj, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_memoryview(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->obj);

#line 11 "(tree fragment)"
  p->obj = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->_size);

#line 11 "(tree fragment)"
  p->_size = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->_array_interface);

#line 11 "(tree fragment)"
  p->_array_interface = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  Py_CLEAR(p->view.obj);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {

#line 11 "(tree fragment)"
  PyObject *r;

#line 11 "(tree fragment)"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 11 "(tree fragment)"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 11 "(tree fragment)"
  Py_DECREF(x);

#line 11 "(tree fragment)"
  return r;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_memoryview___setitem__(o, i, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_Format(PyExc_NotImplementedError,

#line 11 "(tree fragment)"
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_memoryview[] = {

#line 11 "(tree fragment)"
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_memoryview[] = {

#line 11 "(tree fragment)"
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PySequenceMethods __pyx_tp_as_sequence_memoryview = {

#line 11 "(tree fragment)"
  __pyx_memoryview___len__, /*sq_length*/

#line 11 "(tree fragment)"
  0, /*sq_concat*/

#line 11 "(tree fragment)"
  0, /*sq_repeat*/

#line 11 "(tree fragment)"
  __pyx_sq_item_memoryview, /*sq_item*/

#line 11 "(tree fragment)"
  0, /*sq_slice*/

#line 11 "(tree fragment)"
  0, /*sq_ass_item*/

#line 11 "(tree fragment)"
  0, /*sq_ass_slice*/

#line 11 "(tree fragment)"
  0, /*sq_contains*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_concat*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_repeat*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMappingMethods __pyx_tp_as_mapping_memoryview = {

#line 11 "(tree fragment)"
  __pyx_memoryview___len__, /*mp_length*/

#line 11 "(tree fragment)"
  __pyx_memoryview___getitem__, /*mp_subscript*/

#line 11 "(tree fragment)"
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyBufferProcs __pyx_tp_as_buffer_memoryview = {

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getreadbuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getwritebuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getsegcount*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getcharbuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/

#line 11 "(tree fragment)"
  0, /*bf_releasebuffer*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type___pyx_memoryview = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.utils.memoryview", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_memoryview___repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  __pyx_memoryview___str__, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  0, /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_memoryview, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_memoryview, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_memoryview, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_memoryview, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_memoryview, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_memoryviewslice_obj *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_memoryviewslice_obj *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;

#line 11 "(tree fragment)"
  p->from_object = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->from_slice.memview = NULL;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  {

#line 11 "(tree fragment)"
    PyObject *etype, *eval, *etb;

#line 11 "(tree fragment)"
    PyErr_Fetch(&etype, &eval, &etb);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);

#line 11 "(tree fragment)"
    __pyx_memoryviewslice___dealloc__(o);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);

#line 11 "(tree fragment)"
    PyErr_Restore(etype, eval, etb);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  Py_CLEAR(p->from_object);

#line 11 "(tree fragment)"
  PyObject_GC_Track(o);

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_memoryview(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;

#line 11 "(tree fragment)"
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;

#line 11 "(tree fragment)"
  if (p->from_object) {

#line 11 "(tree fragment)"
    e = (*v)(p->from_object, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear__memoryviewslice(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;

#line 11 "(tree fragment)"
  __pyx_tp_clear_memoryview(o);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->from_object);

#line 11 "(tree fragment)"
  p->from_object = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods__memoryviewslice[] = {

#line 11 "(tree fragment)"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {

#line 11 "(tree fragment)"
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type___pyx_memoryviewslice = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.utils._memoryviewslice", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_memoryview___repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_memoryview___str__, /*tp_str*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "Internal class for passing memoryview slices to Python", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods__memoryviewslice, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets__memoryviewslice, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new__memoryviewslice, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods[] = {

#line 11 "(tree fragment)"
  {"intensity_getter", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_5utils_1intensity_getter, METH_O, __pyx_doc_12ms_deisotope_2_c_5utils_intensity_getter},

#line 11 "(tree fragment)"
  {"mz_getter", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_5utils_3mz_getter, METH_O, __pyx_doc_12ms_deisotope_2_c_5utils_2mz_getter},

#line 11 "(tree fragment)"
  {"mass_getter", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_5utils_5mass_getter, METH_O, __pyx_doc_12ms_deisotope_2_c_5utils_4mass_getter},

#line 11 "(tree fragment)"
  {"decode_envelopes", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_5utils_7decode_envelopes, METH_O, __pyx_doc_12ms_deisotope_2_c_5utils_6decode_envelopes},

#line 11 "(tree fragment)"
  {"envelopes_to_array", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_5utils_9envelopes_to_array, METH_O, __pyx_doc_12ms_deisotope_2_c_5utils_8envelopes_to_array},

#line 11 "(tree fragment)"
  {"deserialize_deconvoluted_peak_set", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_11deserialize_deconvoluted_peak_set, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_10deserialize_deconvoluted_peak_set},

#line 11 "(tree fragment)"
  {"deserialize_peak_set", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_5utils_13deserialize_peak_set, METH_O, __pyx_doc_12ms_deisotope_2_c_5utils_12deserialize_peak_set},

#line 11 "(tree fragment)"
  {"build_deconvoluted_peak_set_from_arrays", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_5utils_15build_deconvoluted_peak_set_from_arrays, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_5utils_14build_deconvoluted_peak_set_from_arrays},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
#if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
#if CYTHON_PEP489_MULTI_PHASE_INIT

#line 11 "(tree fragment)"
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/

#line 11 "(tree fragment)"
static int __pyx_pymod_exec_utils(PyObject* module); /*proto*/

#line 11 "(tree fragment)"
static PyModuleDef_Slot __pyx_moduledef_slots[] = {

#line 11 "(tree fragment)"
  {Py_mod_create, (void*)__pyx_pymod_create},

#line 11 "(tree fragment)"
  {Py_mod_exec, (void*)__pyx_pymod_exec_utils},

#line 11 "(tree fragment)"
  {0, NULL}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
#endif

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyModuleDef __pyx_moduledef = {

#line 11 "(tree fragment)"
    PyModuleDef_HEAD_INIT,

#line 11 "(tree fragment)"
    "utils",

#line 11 "(tree fragment)"
    0, /* m_doc */

#line 11 "(tree fragment)"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 11 "(tree fragment)"
    0, /* m_size */

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
    -1, /* m_size */

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
    __pyx_methods /* m_methods */,

#line 11 "(tree fragment)"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 11 "(tree fragment)"
    __pyx_moduledef_slots, /* m_slots */

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
    NULL, /* m_reload */

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
    NULL, /* m_traverse */

#line 11 "(tree fragment)"
    NULL, /* m_clear */

#line 11 "(tree fragment)"
    NULL /* m_free */

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
#endif

#line 11 "(tree fragment)"
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 0},
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_interpret_r_as_a_PeakBase, __pyx_k_Cannot_interpret_r_as_a_PeakBase, sizeof(__pyx_k_Cannot_interpret_r_as_a_PeakBase), 0, 0, 1, 0},
  {&__pyx_n_s_DeconvolutedPeakSet, __pyx_k_DeconvolutedPeakSet, sizeof(__pyx_k_DeconvolutedPeakSet), 0, 0, 1, 1},
  {&__pyx_n_s_DeconvolutedPeakSetIndexed, __pyx_k_DeconvolutedPeakSetIndexed, sizeof(__pyx_k_DeconvolutedPeakSetIndexed), 0, 0, 1, 1},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_k_Expected_at_least_d_argument_s_g, sizeof(__pyx_k_Expected_at_least_d_argument_s_g), 0, 0, 1, 0},
  {&__pyx_kp_s_Function_call_with_ambiguous_arg, __pyx_k_Function_call_with_ambiguous_arg, sizeof(__pyx_k_Function_call_with_ambiguous_arg), 0, 0, 1, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_kp_s_No_matching_signature_found, __pyx_k_No_matching_signature_found, sizeof(__pyx_k_No_matching_signature_found), 0, 0, 1, 0},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_n_s_PeakIndex, __pyx_k_PeakIndex, sizeof(__pyx_k_PeakIndex), 0, 0, 1, 1},
  {&__pyx_n_s_PeakSet, __pyx_k_PeakSet, sizeof(__pyx_k_PeakSet), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_kp_s__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 0},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_kp_s_charge_array, __pyx_k_charge_array, sizeof(__pyx_k_charge_array), 0, 0, 1, 0},
  {&__pyx_n_s_charge_array_2, __pyx_k_charge_array_2, sizeof(__pyx_k_charge_array_2), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_correlation, __pyx_k_correlation, sizeof(__pyx_k_correlation), 0, 0, 1, 1},
  {&__pyx_kp_s_deconvolution_score_array, __pyx_k_deconvolution_score_array, sizeof(__pyx_k_deconvolution_score_array), 0, 0, 1, 0},
  {&__pyx_n_s_defaults, __pyx_k_defaults, sizeof(__pyx_k_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_double, __pyx_k_double, sizeof(__pyx_k_double), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
  {&__pyx_n_s_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_include_envelopes, __pyx_k_include_envelopes, sizeof(__pyx_k_include_envelopes), 0, 0, 1, 1},
  {&__pyx_n_s_int8, __pyx_k_int8, sizeof(__pyx_k_int8), 0, 0, 1, 1},
  {&__pyx_kp_s_intensity_array, __pyx_k_intensity_array, sizeof(__pyx_k_intensity_array), 0, 0, 1, 0},
  {&__pyx_n_s_intensity_array_2, __pyx_k_intensity_array_2, sizeof(__pyx_k_intensity_array_2), 0, 0, 1, 1},
  {&__pyx_kp_s_isotopic_envelopes_array, __pyx_k_isotopic_envelopes_array, sizeof(__pyx_k_isotopic_envelopes_array), 0, 0, 1, 0},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_kp_s_m_z_array, __pyx_k_m_z_array, sizeof(__pyx_k_m_z_array), 0, 0, 1, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_kp_s_ms_deisotope__c_peak_set, __pyx_k_ms_deisotope__c_peak_set, sizeof(__pyx_k_ms_deisotope__c_peak_set), 0, 0, 1, 0},
  {&__pyx_n_s_ms_deisotope__c_utils, __pyx_k_ms_deisotope__c_utils, sizeof(__pyx_k_ms_deisotope__c_utils), 0, 0, 1, 1},
  {&__pyx_kp_s_ms_peak_picker__c_peak_index, __pyx_k_ms_peak_picker__c_peak_index, sizeof(__pyx_k_ms_peak_picker__c_peak_index), 0, 0, 1, 0},
  {&__pyx_kp_s_ms_peak_picker__c_peak_set, __pyx_k_ms_peak_picker__c_peak_set, sizeof(__pyx_k_ms_peak_picker__c_peak_set), 0, 0, 1, 0},
  {&__pyx_n_s_mz_array, __pyx_k_mz_array, sizeof(__pyx_k_mz_array), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_peak_sequence_bp, __pyx_k_peak_sequence_bp, sizeof(__pyx_k_peak_sequence_bp), 0, 0, 1, 1},
  {&__pyx_n_s_peak_sequence_tic, __pyx_k_peak_sequence_tic, sizeof(__pyx_k_peak_sequence_tic), 0, 0, 1, 1},
  {&__pyx_n_s_peaks, __pyx_k_peaks, sizeof(__pyx_k_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0_peak_sequence_bp, __pyx_k_pyx_fuse_0_peak_sequence_bp, sizeof(__pyx_k_pyx_fuse_0_peak_sequence_bp), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0_peak_sequence_tic, __pyx_k_pyx_fuse_0_peak_sequence_tic, sizeof(__pyx_k_pyx_fuse_0_peak_sequence_tic), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0correlation, __pyx_k_pyx_fuse_0correlation, sizeof(__pyx_k_pyx_fuse_0correlation), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1_peak_sequence_bp, __pyx_k_pyx_fuse_1_peak_sequence_bp, sizeof(__pyx_k_pyx_fuse_1_peak_sequence_bp), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1_peak_sequence_tic, __pyx_k_pyx_fuse_1_peak_sequence_tic, sizeof(__pyx_k_pyx_fuse_1_peak_sequence_tic), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1correlation, __pyx_k_pyx_fuse_1correlation, sizeof(__pyx_k_pyx_fuse_1correlation), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_2_peak_sequence_bp, __pyx_k_pyx_fuse_2_peak_sequence_bp, sizeof(__pyx_k_pyx_fuse_2_peak_sequence_bp), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_2_peak_sequence_tic, __pyx_k_pyx_fuse_2_peak_sequence_tic, sizeof(__pyx_k_pyx_fuse_2_peak_sequence_tic), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_3_peak_sequence_bp, __pyx_k_pyx_fuse_3_peak_sequence_bp, sizeof(__pyx_k_pyx_fuse_3_peak_sequence_bp), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_3_peak_sequence_tic, __pyx_k_pyx_fuse_3_peak_sequence_tic, sizeof(__pyx_k_pyx_fuse_3_peak_sequence_tic), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_4_peak_sequence_bp, __pyx_k_pyx_fuse_4_peak_sequence_bp, sizeof(__pyx_k_pyx_fuse_4_peak_sequence_bp), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_4_peak_sequence_tic, __pyx_k_pyx_fuse_4_peak_sequence_tic, sizeof(__pyx_k_pyx_fuse_4_peak_sequence_tic), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_scan_dict, __pyx_k_scan_dict, sizeof(__pyx_k_scan_dict), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_signatures, __pyx_k_signatures, sizeof(__pyx_k_signatures), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_kp_s_src_ms_deisotope__c_utils_pyx, __pyx_k_src_ms_deisotope__c_utils_pyx, sizeof(__pyx_k_src_ms_deisotope__c_utils_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 99, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_builtin___import__ = __Pyx_GetBuiltinName(__pyx_n_s_import); if (!__pyx_builtin___import__) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 944, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(2, 134, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(2, 149, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(2, 152, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(2, 406, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(2, 615, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(2, 834, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "ms_deisotope/_c/utils.pyx":216
 * 
 * @cython.binding(True)
 * cpdef double _peak_sequence_tic(self, peak_collection peaks) except -1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_No_matching_signature_found); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_tuple__4);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(__pyx_tuple__4);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_Function_call_with_ambiguous_arg); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_tuple__5);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 944, __pyx_L1_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__6);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 950, __pyx_L1_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__7);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "View.MemoryView":134
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */

#line 134 "View.MemoryView"
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(2, 134, __pyx_L1_error)

#line 134 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__8);

#line 134 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "View.MemoryView":137
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */

#line 137 "View.MemoryView"
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(2, 137, __pyx_L1_error)

#line 137 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__9);

#line 137 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "View.MemoryView":149
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 149 "View.MemoryView"
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(2, 149, __pyx_L1_error)

#line 149 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__10);

#line 149 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "View.MemoryView":177
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */

#line 177 "View.MemoryView"
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(2, 177, __pyx_L1_error)

#line 177 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__11);

#line 177 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "View.MemoryView":193
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */

#line 193 "View.MemoryView"
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(2, 193, __pyx_L1_error)

#line 193 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__12);

#line 193 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__13);

#line 2 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__14);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "View.MemoryView":420
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */

#line 420 "View.MemoryView"
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(2, 420, __pyx_L1_error)

#line 420 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__15);

#line 420 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "View.MemoryView":497
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */

#line 497 "View.MemoryView"
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(2, 497, __pyx_L1_error)

#line 497 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__16);

#line 497 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "View.MemoryView":522
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */

#line 522 "View.MemoryView"
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(2, 522, __pyx_L1_error)

#line 522 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__17);

#line 522 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "View.MemoryView":572
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */

#line 572 "View.MemoryView"
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(2, 572, __pyx_L1_error)

#line 572 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__18);

#line 572 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "View.MemoryView":579
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */

#line 579 "View.MemoryView"
  __pyx_tuple__19 = PyTuple_New(1); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(2, 579, __pyx_L1_error)

#line 579 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__19);

#line 579 "View.MemoryView"
  __Pyx_INCREF(__pyx_int_neg_1);

#line 579 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_int_neg_1);

#line 579 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_tuple__19, 0, __pyx_int_neg_1);

#line 579 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__20);

#line 2 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__21);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "View.MemoryView":684
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */

#line 684 "View.MemoryView"
  __pyx_slice__22 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__22)) __PYX_ERR(2, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
  __Pyx_GOTREF(__pyx_slice__22);

#line 684 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_slice__22);

  /* "View.MemoryView":705
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 705 "View.MemoryView"
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(2, 705, __pyx_L1_error)

#line 705 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__23);

#line 705 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(2, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__24);

#line 2 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__25);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__25);

#line 4 "(tree fragment)"
  __pyx_tuple__26 = PyTuple_Pack(3, __pyx_int_184977713, __pyx_int_136983863, __pyx_int_112105877); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__26);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "ms_deisotope/_c/utils.pyx":216
 * 
 * @cython.binding(True)
 * cpdef double _peak_sequence_tic(self, peak_collection peaks) except -1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_tuple__27 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_peaks); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_tuple__27);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(__pyx_tuple__27);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_ms_deisotope__c_utils_pyx, __pyx_n_s_pyx_fuse_0_peak_sequence_tic, 216, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 216, __pyx_L1_error)

  /* "ms_deisotope/_c/utils.pyx":244
 * 
 * @cython.binding(True)
 * cpdef PeakBase _peak_sequence_bp(self, peak_collection peaks):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_tuple__29 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_peaks); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_tuple__29);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(__pyx_tuple__29);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_ms_deisotope__c_utils_pyx, __pyx_n_s_pyx_fuse_0_peak_sequence_bp, 244, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 244, __pyx_L1_error)

  /* "ms_deisotope/_c/utils.pyx":274
 * 
 * 
 * cpdef double correlation(cython.floating[:] x, cython.floating[:] y):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_tuple__31 = PyTuple_Pack(2, __pyx_n_s_x, __pyx_n_s_y); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_tuple__31);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(__pyx_tuple__31);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_ms_deisotope__c_utils_pyx, __pyx_n_s_pyx_fuse_0correlation, 274, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 274, __pyx_L1_error)

  /* "View.MemoryView":287
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */

#line 287 "View.MemoryView"
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(2, 287, __pyx_L1_error)

#line 287 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__33);

#line 287 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "View.MemoryView":288
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */

#line 288 "View.MemoryView"
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(2, 288, __pyx_L1_error)

#line 288 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__34);

#line 288 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "View.MemoryView":289
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 289 "View.MemoryView"
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(2, 289, __pyx_L1_error)

#line 289 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__35);

#line 289 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "View.MemoryView":292
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */

#line 292 "View.MemoryView"
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(2, 292, __pyx_L1_error)

#line 292 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__36);

#line 292 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "View.MemoryView":293
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 293 "View.MemoryView"
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(2, 293, __pyx_L1_error)

#line 293 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__37);

#line 293 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_tuple__38 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__38);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__38);

#line 1 "(tree fragment)"
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return 0;

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return -1;

#line 1 "(tree fragment)"
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_112105877 = PyInt_FromLong(112105877L); if (unlikely(!__pyx_int_112105877)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_136983863 = PyInt_FromLong(136983863L); if (unlikely(!__pyx_int_136983863)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("ppm_error", (void (*)(void))__pyx_f_12ms_deisotope_2_c_5utils_ppm_error, "double (double, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ppm2da", (void (*)(void))__pyx_f_12ms_deisotope_2_c_5utils_ppm2da, "double (double, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("da2ppm", (void (*)(void))__pyx_f_12ms_deisotope_2_c_5utils_da2ppm, "double (double, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("intensity_getter", (void (*)(void))__pyx_f_12ms_deisotope_2_c_5utils_intensity_getter, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("mz_getter", (void (*)(void))__pyx_f_12ms_deisotope_2_c_5utils_mz_getter, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("mass_getter", (void (*)(void))__pyx_f_12ms_deisotope_2_c_5utils_mass_getter, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("decode_envelopes", (void (*)(void))__pyx_f_12ms_deisotope_2_c_5utils_decode_envelopes, "PyObject *(PyArrayObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("deserialize_deconvoluted_peak_set", (void (*)(void))__pyx_f_12ms_deisotope_2_c_5utils_deserialize_deconvoluted_peak_set, "struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_5utils_deserialize_deconvoluted_peak_set *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("build_deconvoluted_peak_set_from_arrays", (void (*)(void))__pyx_f_12ms_deisotope_2_c_5utils_build_deconvoluted_peak_set_from_arrays, "struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *(PyArrayObject *, PyArrayObject *, PyArrayObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0_peak_sequence_tic", (void (*)(void))__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic, "double (PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1_peak_sequence_tic", (void (*)(void))__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic, "double (PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_2_peak_sequence_tic", (void (*)(void))__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic, "double (PyObject *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_3_peak_sequence_tic", (void (*)(void))__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic, "double (PyObject *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_4_peak_sequence_tic", (void (*)(void))__pyx_fuse_4__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_tic, "double (PyObject *, PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0_peak_sequence_bp", (void (*)(void))__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp, "struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1_peak_sequence_bp", (void (*)(void))__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp, "struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_2_peak_sequence_bp", (void (*)(void))__pyx_fuse_2__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp, "struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(PyObject *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_3_peak_sequence_bp", (void (*)(void))__pyx_fuse_3__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp, "struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(PyObject *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_4_peak_sequence_bp", (void (*)(void))__pyx_fuse_4__pyx_f_12ms_deisotope_2_c_5utils__peak_sequence_bp, "struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(PyObject *, PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0correlation", (void (*)(void))__pyx_fuse_0__pyx_f_12ms_deisotope_2_c_5utils_correlation, "double (__Pyx_memviewslice, __Pyx_memviewslice, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1correlation", (void (*)(void))__pyx_fuse_1__pyx_f_12ms_deisotope_2_c_5utils_correlation, "double (__Pyx_memviewslice, __Pyx_memviewslice, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(2, 106, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(2, 106, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(2, 106, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(2, 280, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(2, 280, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(2, 331, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(2, 331, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(2, 331, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(2, 967, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(2, 967, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(2, 967, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(1, 770, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(1, 772, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(1, 774, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(1, 776, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(1, 778, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(1, 780, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(1, 782, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(1, 784, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(1, 786, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(1, 788, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakBase", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase) __PYX_ERR(4, 2, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase)) __PYX_ERR(4, 2, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "FittedPeak", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak)) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSet", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet)) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSetIndexed", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) __PYX_ERR(4, 83, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed)) __PYX_ERR(4, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_index"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_index", "PeakIndex", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex) __PYX_ERR(5, 5, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex)) __PYX_ERR(5, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_11composition_PyComposition = __Pyx_ImportType(__pyx_t_1, "brainpy._c.composition", "PyComposition", sizeof(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_11composition_PyComposition) __PYX_ERR(6, 123, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition = (struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_11composition_PyComposition->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition)) __PYX_ERR(6, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_distribution"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = __Pyx_ImportType(__pyx_t_1, "brainpy._c.isotopic_distribution", "TheoreticalPeak", sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak) __PYX_ERR(7, 63, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = (struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak)) __PYX_ERR(7, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "Averagine", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine) __PYX_ERR(8, 12, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine)) __PYX_ERR(8, 12, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "TheoreticalIsotopicPattern", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern) __PYX_ERR(8, 23, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern)) __PYX_ERR(8, 23, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "AveragineCache", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache) __PYX_ERR(8, 79, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache)) __PYX_ERR(8, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.deconvoluter_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "DeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase) __PYX_ERR(9, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase)) __PYX_ERR(9, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "AveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase) __PYX_ERR(9, 57, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase)) __PYX_ERR(9, 57, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "MultiAveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase) __PYX_ERR(9, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase)) __PYX_ERR(9, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "ChargeIterator", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator) __PYX_ERR(9, 91, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator)) __PYX_ERR(9, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitRecord", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord) __PYX_ERR(10, 16, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord)) __PYX_ERR(10, 16, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FitSelectorBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase) __PYX_ERR(10, 38, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase)) __PYX_ERR(10, 38, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase) __PYX_ERR(10, 50, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase)) __PYX_ERR(10, 50, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "LeastSquaresFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter) __PYX_ERR(10, 62, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter)) __PYX_ERR(10, 62, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "GTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter) __PYX_ERR(10, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter)) __PYX_ERR(10, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledGTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter) __PYX_ERR(10, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter)) __PYX_ERR(10, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "MSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter) __PYX_ERR(10, 72, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter)) __PYX_ERR(10, 72, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "PenalizedMSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter) __PYX_ERR(10, 78, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter)) __PYX_ERR(10, 78, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FunctionScorer", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer) __PYX_ERR(10, 84, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer)) __PYX_ERR(10, 84, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "InterferenceDetection", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection) __PYX_ERR(10, 89, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection)) __PYX_ERR(10, 89, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DistinctPatternFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter) __PYX_ERR(10, 95, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter)) __PYX_ERR(10, 95, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledPenalizedMSDeconvFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter) __PYX_ERR(10, 102, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter)) __PYX_ERR(10, 102, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DotProductFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter) __PYX_ERR(10, 112, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter)) __PYX_ERR(10, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "_Index", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index) __PYX_ERR(11, 5, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index)) __PYX_ERR(11, 5, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "EnvelopePair", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair) __PYX_ERR(11, 14, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair)) __PYX_ERR(11, 14, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "Envelope", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope) __PYX_ERR(11, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope)) __PYX_ERR(11, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak) __PYX_ERR(11, 40, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak)) __PYX_ERR(11, 40, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "IonMobilityDeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak) __PYX_ERR(11, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak)) __PYX_ERR(11, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSolution", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution) __PYX_ERR(11, 71, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution)) __PYX_ERR(11, 71, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet) __PYX_ERR(11, 75, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet)) __PYX_ERR(11, 75, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSetIndexed", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed) __PYX_ERR(11, 104, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed)) __PYX_ERR(11, 104, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "_CPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet) __PYX_ERR(11, 161, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet)) __PYX_ERR(11, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "nist_mass", (void **)&__pyx_vp_7brainpy_2_c_11composition_nist_mass, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "PROTON", (void **)&__pyx_vp_7brainpy_2_c_11composition_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_ElementTable", (void **)&__pyx_vp_7brainpy_2_c_11composition__ElementTable, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_constants"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "DEFAULT_ISOTOPIC_CONSTANTS_SIZE", (void **)&__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "PROTON", (void **)&__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INFINITY", (void **)&__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "neutral_mass", (void (**)(void))&__pyx_f_12ms_deisotope_2_c_9averagine_neutral_mass, "double (double, int, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initutils(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initutils(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_utils(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_utils(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_utils(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  static PyThread_type_lock __pyx_t_4[8];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'utils' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_utils(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("utils", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_deisotope___c__utils) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_deisotope._c.utils")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_deisotope._c.utils", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ms_deisotope/_c/utils.pyx":9
 * 
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * 
 */

#line 9 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)

#line 9 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 9 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 9, __pyx_L1_error)

#line 9 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":23
 * 
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 23 "ms_deisotope/_c/utils.pyx"
  __pyx_t_2 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 23, __pyx_L1_error)

  /* "ms_deisotope/_c/utils.pyx":216
 * 
 * @cython.binding(True)
 * cpdef double _peak_sequence_tic(self, peak_collection peaks) except -1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_5utils_23__pyx_fuse_0_peak_sequence_tic, 0, __pyx_n_s_pyx_fuse_0_peak_sequence_tic, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_empty_tuple);

#line 216 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_PeakSet, __pyx_t_3) < 0) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_5utils_25__pyx_fuse_1_peak_sequence_tic, 0, __pyx_n_s_pyx_fuse_1_peak_sequence_tic, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_empty_tuple);

#line 216 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_PeakIndex, __pyx_t_3) < 0) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_5utils_27__pyx_fuse_2_peak_sequence_tic, 0, __pyx_n_s_pyx_fuse_2_peak_sequence_tic, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_empty_tuple);

#line 216 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_DeconvolutedPeakSet, __pyx_t_3) < 0) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_5utils_29__pyx_fuse_3_peak_sequence_tic, 0, __pyx_n_s_pyx_fuse_3_peak_sequence_tic, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_empty_tuple);

#line 216 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_DeconvolutedPeakSetIndexed, __pyx_t_3) < 0) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_4__pyx_mdef_12ms_deisotope_2_c_5utils_31__pyx_fuse_4_peak_sequence_tic, 0, __pyx_n_s_pyx_fuse_4_peak_sequence_tic, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_empty_tuple);

#line 216 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_object, __pyx_t_3) < 0) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_mdef_12ms_deisotope_2_c_5utils_17_peak_sequence_tic, 0, __pyx_n_s_peak_sequence_tic, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_empty_tuple);

#line 216 "ms_deisotope/_c/utils.pyx"
  ((__pyx_FusedFunctionObject *) __pyx_t_3)->__signatures__ = __pyx_t_1;

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_peak_sequence_tic, __pyx_t_3) < 0) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/utils.pyx":244
 * 
 * @cython.binding(True)
 * cpdef PeakBase _peak_sequence_bp(self, peak_collection peaks):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_5utils_35__pyx_fuse_0_peak_sequence_bp, 0, __pyx_n_s_pyx_fuse_0_peak_sequence_bp, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);

#line 244 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_PeakSet, __pyx_t_1) < 0) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_5utils_37__pyx_fuse_1_peak_sequence_bp, 0, __pyx_n_s_pyx_fuse_1_peak_sequence_bp, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);

#line 244 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_PeakIndex, __pyx_t_1) < 0) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_2__pyx_mdef_12ms_deisotope_2_c_5utils_39__pyx_fuse_2_peak_sequence_bp, 0, __pyx_n_s_pyx_fuse_2_peak_sequence_bp, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);

#line 244 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_DeconvolutedPeakSet, __pyx_t_1) < 0) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_3__pyx_mdef_12ms_deisotope_2_c_5utils_41__pyx_fuse_3_peak_sequence_bp, 0, __pyx_n_s_pyx_fuse_3_peak_sequence_bp, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);

#line 244 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_DeconvolutedPeakSetIndexed, __pyx_t_1) < 0) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_4__pyx_mdef_12ms_deisotope_2_c_5utils_43__pyx_fuse_4_peak_sequence_bp, 0, __pyx_n_s_pyx_fuse_4_peak_sequence_bp, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);

#line 244 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_object, __pyx_t_1) < 0) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_mdef_12ms_deisotope_2_c_5utils_19_peak_sequence_bp, 0, __pyx_n_s_peak_sequence_bp, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);

#line 244 "ms_deisotope/_c/utils.pyx"
  ((__pyx_FusedFunctionObject *) __pyx_t_1)->__signatures__ = __pyx_t_3;

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 244 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = 0;

#line 244 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_peak_sequence_bp, __pyx_t_1) < 0) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/utils.pyx":274
 * 
 * 
 * cpdef double correlation(cython.floating[:] x, cython.floating[:] y):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_12ms_deisotope_2_c_5utils_47__pyx_fuse_0correlation, 0, __pyx_n_s_pyx_fuse_0correlation, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_empty_tuple);

#line 274 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_float, __pyx_t_3) < 0) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_12ms_deisotope_2_c_5utils_49__pyx_fuse_1correlation, 0, __pyx_n_s_pyx_fuse_1correlation, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_empty_tuple);

#line 274 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_double, __pyx_t_3) < 0) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_mdef_12ms_deisotope_2_c_5utils_21correlation, 0, __pyx_n_s_correlation, NULL, __pyx_n_s_ms_deisotope__c_utils, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_empty_tuple);

#line 274 "ms_deisotope/_c/utils.pyx"
  ((__pyx_FusedFunctionObject *) __pyx_t_3)->__signatures__ = __pyx_t_1;

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/utils.pyx"
  __pyx_t_1 = 0;

#line 274 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_correlation, __pyx_t_3) < 0) __PYX_ERR(0, 274, __pyx_L1_error)

#line 274 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/utils.pyx":1
 * # cython: embedsignature=True             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */

#line 1 "ms_deisotope/_c/utils.pyx"
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/utils.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 1 "ms_deisotope/_c/utils.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/utils.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":210
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */

#line 210 "View.MemoryView"
  __pyx_t_3 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 210, __pyx_L1_error)

#line 210 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 210 "View.MemoryView"
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_3) < 0) __PYX_ERR(2, 210, __pyx_L1_error)

#line 210 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 210 "View.MemoryView"
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":287
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */

#line 287 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 287, __pyx_L1_error)

#line 287 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 287 "View.MemoryView"
  __Pyx_XGOTREF(generic);

#line 287 "View.MemoryView"
  __Pyx_DECREF_SET(generic, __pyx_t_3);

#line 287 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 287 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":288
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */

#line 288 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 288, __pyx_L1_error)

#line 288 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 288 "View.MemoryView"
  __Pyx_XGOTREF(strided);

#line 288 "View.MemoryView"
  __Pyx_DECREF_SET(strided, __pyx_t_3);

#line 288 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 288 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":289
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 289 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 289, __pyx_L1_error)

#line 289 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 289 "View.MemoryView"
  __Pyx_XGOTREF(indirect);

#line 289 "View.MemoryView"
  __Pyx_DECREF_SET(indirect, __pyx_t_3);

#line 289 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 289 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":292
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */

#line 292 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 292, __pyx_L1_error)

#line 292 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 292 "View.MemoryView"
  __Pyx_XGOTREF(contiguous);

#line 292 "View.MemoryView"
  __Pyx_DECREF_SET(contiguous, __pyx_t_3);

#line 292 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 292 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":293
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 293 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 293, __pyx_L1_error)

#line 293 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 293 "View.MemoryView"
  __Pyx_XGOTREF(indirect_contiguous);

#line 293 "View.MemoryView"
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_3);

#line 293 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 293 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":317
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */

#line 317 "View.MemoryView"
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":318
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */

#line 318 "View.MemoryView"
  __pyx_t_4[0] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[1] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[2] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[3] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[4] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[5] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[6] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_4[7] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_4, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":551
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 551 "View.MemoryView"
  __pyx_t_3 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 551, __pyx_L1_error)

#line 551 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 551 "View.MemoryView"
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_3) < 0) __PYX_ERR(2, 551, __pyx_L1_error)

#line 551 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 551 "View.MemoryView"
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":997
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 997 "View.MemoryView"
  __pyx_t_3 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 997, __pyx_L1_error)

#line 997 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 997 "View.MemoryView"
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_3) < 0) __PYX_ERR(2, 997, __pyx_L1_error)

#line 997 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 997 "View.MemoryView"
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_3) < 0) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /*--- Wrapped vars code ---*/

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 11 "(tree fragment)"
  if (__pyx_m) {

#line 11 "(tree fragment)"
    if (__pyx_d) {

#line 11 "(tree fragment)"
      __Pyx_AddTraceback("init ms_deisotope._c.utils", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    Py_CLEAR(__pyx_m);

#line 11 "(tree fragment)"
  } else if (!PyErr_Occurred()) {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_ImportError, "init ms_deisotope._c.utils");

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 11 "(tree fragment)"
  return (__pyx_m != NULL) ? 0 : -1;

#line 11 "(tree fragment)"
  #elif PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  return __pyx_m;

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  return;

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
}

#line 1 "cython_utility"

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* ArgTypeTest */
  static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyDictVersioning */
  #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
  #if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* ExtTypeTest */
  static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* BufferIndexError */
  static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* DictGetItem */
  #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
  #if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall2Args */
  static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* RaiseDoubleKeywords */
  static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
  static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
  static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* GetItemInt */
  static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* SetItemInt */
  static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* IterFinish */
  static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectCallNoArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectGetMethod */
  static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
  static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
  static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
  static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
  static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
  static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
  static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* UnicodeAsUCS4 */
  static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject* x) {
   Py_ssize_t length;
   #if CYTHON_PEP393_ENABLED
   length = PyUnicode_GET_LENGTH(x);
   if (likely(length == 1)) {
       return PyUnicode_READ_CHAR(x, 0);
   }
   #else
   length = PyUnicode_GET_SIZE(x);
   if (likely(length == 1)) {
       return PyUnicode_AS_UNICODE(x)[0];
   }
   #if Py_UNICODE_SIZE == 2
   else if (PyUnicode_GET_SIZE(x) == 2) {
       Py_UCS4 high_val = PyUnicode_AS_UNICODE(x)[0];
       if (high_val >= 0xD800 && high_val <= 0xDBFF) {
           Py_UCS4 low_val = PyUnicode_AS_UNICODE(x)[1];
           if (low_val >= 0xDC00 && low_val <= 0xDFFF) {
               return 0x10000 + (((high_val & ((1<<10)-1)) << 10) | (low_val & ((1<<10)-1)));
           }
       }
   }
   #endif
   #endif
   PyErr_Format(PyExc_ValueError,
                "only single character unicode strings can be converted to Py_UCS4, "
                "got length %" CYTHON_FORMAT_SSIZE_T "d", length);
   return (Py_UCS4)-1;
}

/* object_ord */
  static long __Pyx__PyObject_Ord(PyObject* c) {
    Py_ssize_t size;
    if (PyBytes_Check(c)) {
        size = PyBytes_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyBytes_AS_STRING(c)[0];
        }
#if PY_MAJOR_VERSION < 3
    } else if (PyUnicode_Check(c)) {
        return (long)__Pyx_PyUnicode_AsPy_UCS4(c);
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    } else if (PyByteArray_Check(c)) {
        size = PyByteArray_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyByteArray_AS_STRING(c)[0];
        }
#endif
    } else {
        PyErr_Format(PyExc_TypeError,
            "ord() expected string of length 1, but %.200s found", Py_TYPE(c)->tp_name);
        return (long)(Py_UCS4)-1;
    }
    PyErr_Format(PyExc_TypeError,
        "ord() expected a character, but string of length %zd found", size);
    return (long)(Py_UCS4)-1;
}

/* WriteUnraisableException */
  static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* None */
  static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* MemviewSliceInit */
  static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#if PY_VERSION_HEX >= 0x030A0000 || defined(HAVE_STDARG_PROTOTYPES)
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None))
        return;
    if (unlikely(__pyx_get_slice_count(memview) < 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (unlikely(first_time)) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    if (unlikely(__pyx_get_slice_count(memview) <= 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (unlikely(last_time)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* GetTopmostException */
  #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* BytesEquals */
  static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
  static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* DivInt[Py_ssize_t] */
  static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t a, Py_ssize_t b) {
    Py_ssize_t q = a / b;
    Py_ssize_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* GetAttr */
  static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* ObjectGetItem */
  #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* decode_c_string */
  static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* GetAttr3 */
  static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* SwapException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* DivInt[long] */
  static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
  static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
  static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
  static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* FetchCommonType */
  static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
  #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* FusedFunction */
  static PyObject *
__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                        PyObject *qualname, PyObject *closure,
                        PyObject *module, PyObject *globals,
                        PyObject *code)
{
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_FusedFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        __pyx_FusedFunctionObject *fusedfunc = (__pyx_FusedFunctionObject *) op;
        fusedfunc->__signatures__ = NULL;
        fusedfunc->type = NULL;
        fusedfunc->self = NULL;
        PyObject_GC_Track(op);
    }
    return op;
}
static void
__pyx_FusedFunction_dealloc(__pyx_FusedFunctionObject *self)
{
    PyObject_GC_UnTrack(self);
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    __Pyx__CyFunction_dealloc((__pyx_CyFunctionObject *) self);
}
static int
__pyx_FusedFunction_traverse(__pyx_FusedFunctionObject *self,
                             visitproc visit,
                             void *arg)
{
    Py_VISIT(self->self);
    Py_VISIT(self->type);
    Py_VISIT(self->__signatures__);
    return __Pyx_CyFunction_traverse((__pyx_CyFunctionObject *) self, visit, arg);
}
static int
__pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self)
{
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    return __Pyx_CyFunction_clear((__pyx_CyFunctionObject *) self);
}
static PyObject *
__pyx_FusedFunction_descr_get(PyObject *self, PyObject *obj, PyObject *type)
{
    __pyx_FusedFunctionObject *func, *meth;
    func = (__pyx_FusedFunctionObject *) self;
    if (func->self || func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(self);
        return self;
    }
    if (obj == Py_None)
        obj = NULL;
    meth = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_New(
                    ((PyCFunctionObject *) func)->m_ml,
                    ((__pyx_CyFunctionObject *) func)->flags,
                    ((__pyx_CyFunctionObject *) func)->func_qualname,
                    ((__pyx_CyFunctionObject *) func)->func_closure,
                    ((PyCFunctionObject *) func)->m_module,
                    ((__pyx_CyFunctionObject *) func)->func_globals,
                    ((__pyx_CyFunctionObject *) func)->func_code);
    if (!meth)
        return NULL;
    if (func->func.defaults) {
        PyObject **pydefaults;
        int i;
        if (!__Pyx_CyFunction_InitDefaults((PyObject*)meth,
                                      func->func.defaults_size,
                                      func->func.defaults_pyobjects)) {
            Py_XDECREF((PyObject*)meth);
            return NULL;
        }
        memcpy(meth->func.defaults, func->func.defaults, func->func.defaults_size);
        pydefaults = __Pyx_CyFunction_Defaults(PyObject *, meth);
        for (i = 0; i < meth->func.defaults_pyobjects; i++)
            Py_XINCREF(pydefaults[i]);
    }
    Py_XINCREF(func->func.func_classobj);
    meth->func.func_classobj = func->func.func_classobj;
    Py_XINCREF(func->__signatures__);
    meth->__signatures__ = func->__signatures__;
    Py_XINCREF(type);
    meth->type = type;
    Py_XINCREF(func->func.defaults_tuple);
    meth->func.defaults_tuple = func->func.defaults_tuple;
    if (func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD)
        obj = type;
    Py_XINCREF(obj);
    meth->self = obj;
    return (PyObject *) meth;
}
static PyObject *
_obj_to_str(PyObject *obj)
{
    if (PyType_Check(obj))
        return PyObject_GetAttr(obj, __pyx_n_s_name_2);
    else
        return PyObject_Str(obj);
}
static PyObject *
__pyx_FusedFunction_getitem(__pyx_FusedFunctionObject *self, PyObject *idx)
{
    PyObject *signature = NULL;
    PyObject *unbound_result_func;
    PyObject *result_func = NULL;
    if (self->__signatures__ == NULL) {
        PyErr_SetString(PyExc_TypeError, "Function is not fused");
        return NULL;
    }
    if (PyTuple_Check(idx)) {
        PyObject *list = PyList_New(0);
        Py_ssize_t n = PyTuple_GET_SIZE(idx);
        PyObject *sep = NULL;
        int i;
        if (unlikely(!list))
            return NULL;
        for (i = 0; i < n; i++) {
            int ret;
            PyObject *string;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(idx, i);
#else
            PyObject *item = PySequence_ITEM(idx, i);  if (unlikely(!item)) goto __pyx_err;
#endif
            string = _obj_to_str(item);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
            Py_DECREF(item);
#endif
            if (unlikely(!string)) goto __pyx_err;
            ret = PyList_Append(list, string);
            Py_DECREF(string);
            if (unlikely(ret < 0)) goto __pyx_err;
        }
        sep = PyUnicode_FromString("|");
        if (likely(sep))
            signature = PyUnicode_Join(sep, list);
__pyx_err:
;
        Py_DECREF(list);
        Py_XDECREF(sep);
    } else {
        signature = _obj_to_str(idx);
    }
    if (!signature)
        return NULL;
    unbound_result_func = PyObject_GetItem(self->__signatures__, signature);
    if (unbound_result_func) {
        if (self->self || self->type) {
            __pyx_FusedFunctionObject *unbound = (__pyx_FusedFunctionObject *) unbound_result_func;
            Py_CLEAR(unbound->func.func_classobj);
            Py_XINCREF(self->func.func_classobj);
            unbound->func.func_classobj = self->func.func_classobj;
            result_func = __pyx_FusedFunction_descr_get(unbound_result_func,
                                                        self->self, self->type);
        } else {
            result_func = unbound_result_func;
            Py_INCREF(result_func);
        }
    }
    Py_DECREF(signature);
    Py_XDECREF(unbound_result_func);
    return result_func;
}
static PyObject *
__pyx_FusedFunction_callfunction(PyObject *func, PyObject *args, PyObject *kw)
{
     __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    int static_specialized = (cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD &&
                              !((__pyx_FusedFunctionObject *) func)->__signatures__);
    if (cyfunc->flags & __Pyx_CYFUNCTION_CCLASS && !static_specialized) {
        return __Pyx_CyFunction_CallAsMethod(func, args, kw);
    } else {
        return __Pyx_CyFunction_Call(func, args, kw);
    }
}
static PyObject *
__pyx_FusedFunction_call(PyObject *func, PyObject *args, PyObject *kw)
{
    __pyx_FusedFunctionObject *binding_func = (__pyx_FusedFunctionObject *) func;
    Py_ssize_t argc = PyTuple_GET_SIZE(args);
    PyObject *new_args = NULL;
    __pyx_FusedFunctionObject *new_func = NULL;
    PyObject *result = NULL;
    PyObject *self = NULL;
    int is_staticmethod = binding_func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD;
    int is_classmethod = binding_func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD;
    if (binding_func->self) {
        Py_ssize_t i;
        new_args = PyTuple_New(argc + 1);
        if (!new_args)
            return NULL;
        self = binding_func->self;
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_INCREF(self);
#endif
        Py_INCREF(self);
        PyTuple_SET_ITEM(new_args, 0, self);
        for (i = 0; i < argc; i++) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(args, i);
            Py_INCREF(item);
#else
            PyObject *item = PySequence_ITEM(args, i);  if (unlikely(!item)) goto bad;
#endif
            PyTuple_SET_ITEM(new_args, i + 1, item);
        }
        args = new_args;
    } else if (binding_func->type) {
        if (argc < 1) {
            PyErr_SetString(PyExc_TypeError, "Need at least one argument, 0 given.");
            return NULL;
        }
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        self = PyTuple_GET_ITEM(args, 0);
#else
        self = PySequence_ITEM(args, 0);  if (unlikely(!self)) return NULL;
#endif
    }
    if (self && !is_classmethod && !is_staticmethod) {
        int is_instance = PyObject_IsInstance(self, binding_func->type);
        if (unlikely(!is_instance)) {
            PyErr_Format(PyExc_TypeError,
                         "First argument should be of type %.200s, got %.200s.",
                         ((PyTypeObject *) binding_func->type)->tp_name,
                         Py_TYPE(self)->tp_name);
            goto bad;
        } else if (unlikely(is_instance == -1)) {
            goto bad;
        }
    }
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
    self = NULL;
#endif
    if (binding_func->__signatures__) {
        PyObject *tup;
        if (is_staticmethod && binding_func->func.flags & __Pyx_CYFUNCTION_CCLASS) {
            tup = PyTuple_Pack(3, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __Pyx_CyFunction_CallMethod(
                func, binding_func->__signatures__, tup, NULL);
        } else {
            tup = PyTuple_Pack(4, binding_func->__signatures__, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_callfunction(func, tup, NULL);
        }
        Py_DECREF(tup);
        if (unlikely(!new_func))
            goto bad;
        Py_XINCREF(binding_func->func.func_classobj);
        Py_CLEAR(new_func->func.func_classobj);
        new_func->func.func_classobj = binding_func->func.func_classobj;
        func = (PyObject *) new_func;
    }
    result = __pyx_FusedFunction_callfunction(func, args, kw);
bad:
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
#endif
    Py_XDECREF(new_args);
    Py_XDECREF((PyObject *) new_func);
    return result;
}
static PyMemberDef __pyx_FusedFunction_members[] = {
    {(char *) "__signatures__",
     T_OBJECT,
     offsetof(__pyx_FusedFunctionObject, __signatures__),
     READONLY,
     0},
    {0, 0, 0, 0, 0},
};
static PyMappingMethods __pyx_FusedFunction_mapping_methods = {
    0,
    (binaryfunc) __pyx_FusedFunction_getitem,
    0,
};
static PyTypeObject __pyx_FusedFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "fused_cython_function",
    sizeof(__pyx_FusedFunctionObject),
    0,
    (destructor) __pyx_FusedFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    0,
    0,
    0,
    &__pyx_FusedFunction_mapping_methods,
    0,
    (ternaryfunc) __pyx_FusedFunction_call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __pyx_FusedFunction_traverse,
    (inquiry) __pyx_FusedFunction_clear,
    0,
    0,
    0,
    0,
    0,
    __pyx_FusedFunction_members,
    __pyx_CyFunction_getsets,
    &__pyx_CyFunctionType_type,
    0,
    __pyx_FusedFunction_descr_get,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_FusedFunction_init(void) {
    __pyx_FusedFunctionType_type.tp_base = __pyx_CyFunctionType;
    __pyx_FusedFunctionType = __Pyx_FetchCommonType(&__pyx_FusedFunctionType_type);
    if (__pyx_FusedFunctionType == NULL) {
        return -1;
    }
    return 0;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* MemviewSliceIsContig */
  static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
  static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
  static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* TypeInfoCompare */
  static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
  static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (unlikely(buf->strides[dim] != sizeof(void *))) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (unlikely(buf->strides[dim] != buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (unlikely(stride < buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (unlikely(spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(spec & (__Pyx_MEMVIEW_PTR))) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(buf->suboffsets)) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, CYTHON_UNUSED int ndim, int spec)
{
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (unlikely(buf->suboffsets && buf->suboffsets[dim] >= 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (unlikely(!buf->suboffsets || (buf->suboffsets[dim] < 0))) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (unlikely(buf->ndim != ndim)) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (unlikely(!__Pyx_BufFmt_CheckString(&ctx, buf->format))) goto fail;
    }
    if (unlikely((unsigned) buf->itemsize != dtype->size)) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    if (buf->len > 0) {
        for (i = 0; i < ndim; i++) {
            spec = axes_specs[i];
            if (unlikely(!__pyx_check_strides(buf, i, ndim, spec)))
                goto fail;
            if (unlikely(!__pyx_check_suboffsets(buf, i, ndim, spec)))
                goto fail;
        }
        if (unlikely(buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag)))
            goto fail;
    }
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_float(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_float, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_float(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(float *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_float(const char *itemp, PyObject *obj) {
    float value = __pyx_PyFloat_AsFloat(obj);
    if ((value == (float)-1) && PyErr_Occurred())
        return 0;
    *(float *) itemp = value;
    return 1;
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_double(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(double *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_double(const char *itemp, PyObject *obj) {
    double value = __pyx_PyFloat_AsDouble(obj);
    if ((value == (double)-1) && PyErr_Occurred())
        return 0;
    *(double *) itemp = value;
    return 1;
}

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* MemviewSliceCopyTemplate */
  static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* BytesContains */
  static CYTHON_INLINE int __Pyx_BytesContains(PyObject* bytes, char character) {
    const Py_ssize_t length = PyBytes_GET_SIZE(bytes);
    char* char_start = PyBytes_AS_STRING(bytes);
    return memchr(char_start, (unsigned char)character, (size_t)length) != NULL;
}

/* ImportNumPyArray */
  static PyObject* __Pyx__ImportNumPyArray(void) {
    PyObject *numpy_module, *ndarray_object = NULL;
    numpy_module = __Pyx_Import(__pyx_n_s_numpy, NULL, 0);
    if (likely(numpy_module)) {
        ndarray_object = PyObject_GetAttrString(numpy_module, "ndarray");
        Py_DECREF(numpy_module);
    }
    if (unlikely(!ndarray_object)) {
        PyErr_Clear();
    }
    if (unlikely(!ndarray_object || !PyObject_TypeCheck(ndarray_object, &PyType_Type))) {
        Py_XDECREF(ndarray_object);
        Py_INCREF(Py_None);
        ndarray_object = Py_None;
    }
    return ndarray_object;
}
static CYTHON_INLINE PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void) {
    if (unlikely(!__pyx_numpy_ndarray)) {
        __pyx_numpy_ndarray = __Pyx__ImportNumPyArray();
    }
    Py_INCREF(__pyx_numpy_ndarray);
    return __pyx_numpy_ndarray;
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
  static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* VoidPtrImport */
  #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
  #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
